////////////////////////////////////////////////////////////////////////////////
/*!	\file		AI_ADNBase.cpp
**	\brief		This is the implementation file for AI_ADNBase.h
**	\author		John Corpening
**	\date		10/25/2002 8:09:08 PM
*/
////////////////////////////////////////////////////////////////////////////////
#pragma warning(disable:4786)

#include"AI_System.h"
using namespace AI_System;
using namespace Debugging;

// memory leak detection
#ifdef _DEBUG
    #include"memleak.h"
    #define new DEBUG_NEW
    #undef THIS_FILE
    static char THIS_FILE[] = __FILE__;
#endif

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	AI_ADN::AI_ADN
/*!	\brief		Constructor creates a blank ADN with a random goal
**	\param		type an Action_ID that specifies the type of action
**	\return		none
**	\exception	none
**	\author		John Corpening
**	\date		10/25/2002 7:44:04 PM
**	\warning	Do not use until AI_Unit connections have been made
*/
////////////////////////////////////////////////////////////////////////////////
AI_ADN::AI_ADN(TT_ID Technique)
// Added by Nate Cleveland [4/11/2003]
//  The save IDs we're being set so I set this, I'm fairly sure this is what you 
//  meant to do. 
:technique(Technique)
{
	//Ensure that only the high two bytes of the type are used
	actionClass = AC_NULL;
	hg_ID = HG_NULL;
	fitFunc = 0;
	set = false;
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	setActionClass
/*!	\brief		sets the action class of the ADN
**	\param		type, the action_ID of the action type
**	\exception	none
**	\author		John Corpening
**	\date		4/15/2003 8:53:03 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
void AI_ADN::setActionClass(Action_ID type) throw()
{
	if(set)
		return;

	//Ensure that only the high two bytes of the type are used
	actionClass = static_cast<Action_ID>(type & AC_MASK);
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	setHigherGoal
/*!	\brief		sets the higher goal of the ADN
**	\param		hg, a higher goal ID
**	\return		none
**	\exception	"Null higher goal returned" Nexception
**	\author		John Corpening
**	\date		4/15/2003 8:45:20 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
void AI_ADN::setHigherGoal(HG_ID hg)
{
	if(set)
		return;

	hg_ID = hg;
	fitFunc = getHigherGoal(hg_ID);
	if(!fitFunc)
	{
		throw(Nexception("Null Higher Goal Returned", __FILE__,__LINE__,__DATE__,__TIME__));
	}
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	initialize
/*!	\brief		initializes and sets the ADN
**	\exception	none
**	\author		John Corpening
**	\date		4/15/2003 8:43:34 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
void AI_ADN::initialize()
{
	if(set)
		return;

	if(actionClass == AC_NULL)
	{
		//Default to AC_MOVE for now
		actionClass = AC_MOVE;
	}
	if(hg_ID == HG_NULL)
	{
		hg_ID = getRandomHG_ID();
		fitFunc = getHigherGoal(hg_ID);
		if(!fitFunc)
		{
			throw(Nexception("Null Higher Goal Returned", __FILE__,__LINE__,__DATE__,__TIME__));
		}
	}
	subClassInit();
	set = true;
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	AI_ADN::getInput
/*!	\brief		Gets input from associated AI_Units
**	\return		none
**	\exception	none
**	\author		John Corpening
**	\date		10/24/2002 9:44:00 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
void AI_ADN::getInput()
{
	//Have all the AI_Units get their input and return their recommendations
	int size = AI_Units.size();
	actions.clear();
	std::map<AI_ID, AI_Unit*>::iterator pos;
	for(pos = AI_Units.begin(); pos != AI_Units.end(); pos++)
	{
		pos->second->getInput();
		
		AI_Action		temp = pos->second->determineAction();

		AI_Attribute	temp2;
		temp2.ID	= static_cast<Attribute_ID>(temp.tag.unit_ID);
		temp2.value = temp.value;

		attributes.push_back(temp2);
		actions.insert(std::make_pair(temp.tag.unit_ID, temp));
	}
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	AI_Unit::determineAction
/*!	\brief		analyzes attributes and produces an Action
**	\return		none
**	\exception	none
**	\author		John Corpening
**	\date		10/19/2002 12:54:00 PM
**	\warning	none
**	\par		Description:
**					This is the workhorse function that analyzes the attributes
**				and produces an action.
*/
////////////////////////////////////////////////////////////////////////////////
AI_Action AI_ADN::determineAction()
{
	AI_ID	unit = 0;

	fitFunc->clear();
	
	//The process function needs to return an action
	//and fill out an AI_Unit ID
	unit = process();

	//Tell the appropriate AI_Unit to save attributes
	std::map<AI_ID, AI_Unit*>::iterator pos = AI_Units.find(unit);
	if(pos != AI_Units.end())
	{
		pos->second->saveAttributes();
	}

	//An action has been determined so save the attributes for a future training example
	std::map<AI_ID, AI_Action>::iterator it = actions.find(unit);
	int size = actions.size();
	if(it == actions.end())
	{
		throw(Nexception("Action not in Action List", __FILE__,__LINE__,__DATE__,__TIME__));
	}

	//If the ADN forces an AI Unit that did not trigger to trigger
	//Then retreive a magnitude for the action
	if(it->second.value == 0.0f)
	{
		it->second.value = pos->second->getEmphasis();
	}

	saveAttributes(it->second.tag);
	//Clear out the attributes
	attributes.clear();

	Debugger::Instance()->output(AI, L2, "ADN Chooses: Unit:%d, Decision:%d, Action %s,", it->second.tag.unit_ID, it->second.tag.decisionID, getActionString(it->second.ID)); 
	return it->second;
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	AI_Unit::createTrainingExample
/*!	\brief		combines fitness and saved attributes to create a training example
**	\return		none
**	\exception	none
**	\author		John Corpening
**	\date		10/19/2002 12:54:38 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
void AI_ADN::createTrainingExample(AI_ActionTag tag)
{

	//get the attribute list
	std::map<AI_ActionTag, std::vector<AI_Attribute> >::iterator pos;
	pos = savedAttributes.find(tag);
	if(pos == savedAttributes.end())
	{
		//No attribute list, then bail
		return;	
	}

	if(tag.completed)
	{
		//Get the associated AI_Unit
		std::map<AI_ID, AI_Unit*>::iterator unit;
		unit = AI_Units.find(tag.unit_ID);
		if(unit == AI_Units.end())
		{
			//No AI unit then erase the attributes and bail
			savedAttributes.erase(pos);
			return;
		}
		//save the Training example
		Debugger::Instance()->output(AI, L3, "Training Example: Fish:%d, ADN: %s, unit:%d, \
		Response:%d", AI_PerceptGenerator::instance()->getTarget(), getActionString(actionClass), 
		tag.unit_ID, fitFunc->getResponse());
		saveTrainingExample(fitFunc->getResponse(), fitFunc->getFitness(), unit->second->getID(), pos->second);
	}

	//Clear out the saved attributes
	savedAttributes.erase(pos);
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	AI_ADN::saveTrainingExample
/*!	\brief		saves a training example
**	\param		response The fitness function determined response
**	\param		fitness The fitness function determined fitness
**	\param		action the action that was performed
**	\param		Attributes saved attributes
**	\return		none
**	\exception	none
**	\author		John Corpening
**	\date		11/7/2002 1:48:09 PM
**	\warning	none
**	\par		Description:
**				This function was created to allow sublasses to train
**				as they go.
*/
////////////////////////////////////////////////////////////////////////////////
void AI_ADN::saveTrainingExample(bool response, float fitness, AI_ID unit, std::vector<AI_Attribute>& Attributes)
{
	AI_ExampleADN ex;
	ex.result		= response;
	ex.fitness		= fitness;
	ex.ID			= unit;
	ex.attributes	= Attributes;
	trainingExamples.push_back(ex);
}
////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	AI_Unit::saveAttributes
/*!	\brief		saves current attributes to the savedAttribtes list
**	\return		none
**	\exception	none
**	\author		John Corpening
**	\date		10/19/2002 12:54:46 PM
**	\warning	none
**	\par		Description:
**					If the action from this AI unit is selected by the ADN 
**				this function will be called to save the attributes into the 
**				saved attribute list.
*/
////////////////////////////////////////////////////////////////////////////////
void AI_ADN::saveAttributes(AI_ActionTag tag)
{
	savedAttributes.insert(std::make_pair(tag, attributes));
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	AI_ADN::addAIUnit	
/*!	\brief		adds an AI Unit to the ADN's list of AI_Units
**	\param		unit a pointer to the AI_Unit to add
**	\return		none
**	\exception	none
**	\author		John Corpening
**	\date		10/28/2002 7:38:49 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
bool AI_ADN::addAIUnit(AI_Unit* unit) throw() 
{
	if(set)
		return false;
	//Check to make sure the unit is of the correct type
	if(unit->getAction() & actionClass)
	{
		AI_Units.insert(std::make_pair(unit->getID(), unit));
		return true;
	}
	return false;
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	Write
/*!	\brief		Saves the ADN to the ostream
**	\param		an ostream to write to
**	\return		none
**	\exception	none
**	\author		John Corpening
**	\date		4/15/2003 8:49:29 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
void AI_ADN::Write(std::ostream& file) const
{
	//Note: There is no need to save the technique type. The AI_Module will do this
	//Save the Higher Goal ID
	file << "\tHigher Goal ID: " << hg_ID << endl;
	//Save the Action Type
	file << "\tAction Type: " << actionClass << endl;
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	Read
/*!	\brief		Loads the ADN from the istream
**	\param		an istream to read from
**	\return		none
**	\exception	none
**	\author		John Corpening
**	\date		4/15/2003 8:49:39 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
void AI_ADN::Read(std::istream& file)
{
	const unsigned int buffSize = 512;
	char buff[buffSize] = {0};


	//Load the higher goal ID
	unsigned int temp;
	file.getline(buff, buffSize);
	sscanf(buff, "\tHigher Goal ID: %d", &temp);
	hg_ID = (HG_ID)temp;


	//Load the Action Type
	file.getline(buff, buffSize);
	sscanf(buff, "\tAction Type: %d", &temp);
	actionClass = (Action_ID)temp;

	//Get the higher goal
	fitFunc = getHigherGoal(hg_ID);
//	set = true;
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	~AI_ADN
/*!	\brief		Deletes the fitness function
**	\exception	none
**	\author		John Corpening
**	\date		11/25/2002 8:31:45 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
AI_ADN::~AI_ADN()
{
	if(fitFunc)
		delete fitFunc;
}