////////////////////////////////////////////////////////////////////////////////
/*!	\file		AI_ADNBase.h
**	\brief		Action Decision Node base class
**	\author		John Corpening
**	\date		10/21/2002 1:43:46 PM
**	\par		DESCRIPTION:
**				Action Decision Nodes (ADN's) are a similar to AI Unit. They
**				produce as an output any of their inputs. They take as inputs
**				the actions generated by the AI Units. 
**
**				There is one ADN for each type of independent action that an 
**				object can make. For example, a fish may be able to move and eat
**				but a barnacle may only be able to eat, therefore the fish will
**				have two ADN's but the barnacle will only have one. Likewise, 
**				the AI units connected associated with an ADN will only be able 
**				to specify actions that fall under the category of action that 
**				the ADN can specify. 
**
**				ADN's are evaluated by Higher Goals in the same manner that
**				AI units are evaluated by fitness function.
**	\par		REMARKS:
**				The AI unit base class is an abstract base class.
**
**	\par		TIPS FOR WRITING DERIVED CLASSES:
**				-The only functions that require implementation are the following
**					- determineAction
**					- save
**					- load
**				-The only data that needs to be accessed are the following
**					- attributes : the attributes to analyze to reach a decision
**					- trainingExamples : saved examples to train the AI
**				-Reference the AI_Unit's ID (AI_Action.tag.unit_ID) as the identifier
**					- This is because the action's specified are not unique
**				- You may also override the "train" function
**					- If so, 
**			
**		>>>>>>> !! DO NOT FORGET TO CLEAR THE TRAINING EXAMPLES !! <<<<<<<<<
**			
**					- This was not done for you since derived classes may choose
**					  to defer training until a later cycle.
*/
////////////////////////////////////////////////////////////////////////////////

#pragma once

namespace AI_System
{
//Forward declarations
class AI_Module;


class AI_ADN : public FileMemento
{
public:

	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:	AI_ADN::AI_ADN
	/*!	\brief		Constructor creates a blank ADN with a random goal
	**	\param		type an Action_ID that specifies the type of action
	**	\return		none
	**	\exception	none
	**	\author		John Corpening
	**	\date		10/25/2002 7:44:04 PM
	**	\warning	Do not use until AI_Unit connections have been made
	*/
	////////////////////////////////////////////////////////////////////////////////
	AI_ADN(TT_ID technique);

	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:	setActionClass
	/*!	\brief		sets the action class of the ADN
	**	\param		type, the action_ID of the action type
	**	\exception	none
	**	\author		John Corpening
	**	\date		4/15/2003 8:53:03 PM
	**	\warning	none
	*/
	////////////////////////////////////////////////////////////////////////////////
	void setActionClass(Action_ID type) throw();

	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:	setHigherGoal
	/*!	\brief		sets the higher goal of the ADN
	**	\param		hg, a higher goal ID
	**	\return		none
	**	\exception	"Null higher goal returned" Nexception
	**	\author		John Corpening
	**	\date		4/15/2003 8:45:20 PM
	**	\warning	none
	*/
	////////////////////////////////////////////////////////////////////////////////
	void setHigherGoal(HG_ID hg);

	//basic accessor functions
	//Gets the id of the higher goal of the ADN
	HG_ID getHigherGoalID()throw(){return hg_ID;}
	//Gets the Action Type
	Action_ID getActionType() throw() {return actionClass;}
	//gets the technique used by the ADN
	TT_ID getTechniqueType()throw(){return technique;}

	//lock the ADN from being updated
	void lockADN(){set = true;}


	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:	initialize
	/*!	\brief		initializes and sets the ADN
	**	\exception	none
	**	\author		John Corpening
	**	\date		4/15/2003 8:43:34 PM
	**	\warning	none
	*/
	////////////////////////////////////////////////////////////////////////////////
	void initialize();

	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:	~AI_ADN
	/*!	\brief		Deletes the fitness function
	**	\exception	none
	**	\author		John Corpening
	**	\date		11/25/2002 8:31:45 PM
	**	\warning	none
	*/
	////////////////////////////////////////////////////////////////////////////////
	virtual ~AI_ADN();


	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:	AI_ADN::getInput
	/*!	\brief		Gets input from associated AI_Units
	**	\return		none
	**	\exception	none
	**	\author		John Corpening
	**	\date		10/24/2002 9:44:00 PM
	**	\warning	none
	*/
	////////////////////////////////////////////////////////////////////////////////
	void getInput() throw();

	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:	AI_ADN::determineAction
	/*!	\brief		analyzes attributes and produces an Action
	**	\return		none
	**	\exception	none
	**	\author		John Corpening
	**	\date		10/19/2002 12:54:00 PM
	**	\warning	none
	**	\par		Description:
	**					This is the workhorse function that analyzes the attributes
	**				and produces an action.
	*/
	////////////////////////////////////////////////////////////////////////////////
	AI_Action determineAction() throw();

	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:	AI_ADN::createTrainingExample
	/*!	\brief		combines fitness and saved attributes to create a training example
	**	\param		tag an AI_ActionTag that identifies the AI unit and it's decision
	**	\return		none
	**	\exception	none
	**	\author		John Corpening
	**	\date		10/19/2002 12:54:38 PM
	**	\warning	none
	*/
	////////////////////////////////////////////////////////////////////////////////
	void createTrainingExample(AI_ActionTag tag) throw();


	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:	AI_ADN::addAIUnit	
	/*!	\brief		adds an AI Unit to the ADN's list of AI_Units
	**	\param		unit a pointer to the AI_Unit to add
	**	\return		none
	**	\exception	none
	**	\author		John Corpening
	**	\date		10/28/2002 7:38:49 PM
	**	\warning	none
	*/
	////////////////////////////////////////////////////////////////////////////////
	bool addAIUnit(AI_Unit* unit) throw();



//-------Virtual Functions----------------------------------------------------------
	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:	AI_ADN::Train
	/*!	\brief		Trains the AI unit based on the training examples
	**	\return		none
	**	\exception	none
	**	\author		John Corpening
	**	\date		10/19/2002 12:54:07 PM
	**	\warning	YOU MUST NOT FORGET TO CLEAR YOUR TRAINING EXAMPLES!
	*/
	////////////////////////////////////////////////////////////////////////////////
	virtual void train() throw() {trainingExamples.clear();}


	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:	Write
	/*!	\brief		Saves the ADN to the ostream
	**	\param		an ostream to write to
	**	\return		none
	**	\exception	none
	**	\author		John Corpening
	**	\date		4/15/2003 8:49:29 PM
	**	\warning	none
	*/
	////////////////////////////////////////////////////////////////////////////////
	virtual void Write(std::ostream&)const;

	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:	Read
	/*!	\brief		Loads the ADN from the istream
	**	\param		an istream to read from
	**	\return		none
	**	\exception	none
	**	\author		John Corpening
	**	\date		4/15/2003 8:49:39 PM
	**	\warning	none
	*/
	////////////////////////////////////////////////////////////////////////////////
	virtual void Read(std::istream&);

	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:	AI_ADN::save
	/*!	\brief		Saves the AI unit's configuration to a file
	**	\param		TBA
	**	\return		success or failure
	**	\exception	none
	**	\author		John Corpening
	**	\date		10/19/2002 12:55:03 PM
	**	\warning	This function is subject to change pending File I/O considerations
	*/
	////////////////////////////////////////////////////////////////////////////////
	virtual bool save() throw() = 0;

	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:	AI_ADN::load
	/*!	\brief		Loads the AI unit's configuration from a file
	**	\param		TBA
	**	\return		success or failure
	**	\exception	none
	**	\author		John Corpening
	**	\date		10/19/2002 12:55:11 PM
	**	\warning	This function is subject to change pending File I/O considerations
	*/
	////////////////////////////////////////////////////////////////////////////////
	virtual bool load() throw() = 0;

protected:
	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:	AI_ADN::process
	/*!	\brief		analyzes AI_Unit inputs, selects one as the action
	**	\param		the ID of the AI unit that will be chosen (This is a return)
	**	\return		an action to be performed
	**	\return		the ID of the AI unit that was chosen
	**	\exception	none
	**	\author		John Corpening
	**	\date		10/24/2002 9:44:16 PM
	**	\warning	none
	*/
	////////////////////////////////////////////////////////////////////////////////
	virtual AI_ID process() throw() = 0;

	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:	AI_ADN::saveTrainingExample
	/*!	\brief		saves a training example
	**	\param		response The fitness function determined response
	**	\param		fitness The fitness function determined fitness
	**	\param		action the action that was performed
	**	\param		Attributes saved attributes
	**	\return		none
	**	\exception	none
	**	\author		John Corpening
	**	\date		11/7/2002 1:48:09 PM
	**	\warning	none
	**	\par		Description:
	**				This function was created to allow sublasses to train
	**				as they go.
	*/
	////////////////////////////////////////////////////////////////////////////////
	virtual void saveTrainingExample(bool response, float fitness, AI_ID action, std::vector<AI_Attribute>& Attributes);

	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:	AI_ADN::subClassInit
	/*!	\brief		initialization function for subclasses
	**	\return		none
	**	\exception	none
	**	\author		John Corpening
	**	\date		11/25/2002 8:32:51 PM
	**	\warning	none
	*/
	////////////////////////////////////////////////////////////////////////////////
	virtual void subClassInit() {};

//-------End Virtual Functions----------------------------------------------------------
private:

	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:	AI_ADN::saveAttributes
	/*!	\brief		saves current attributes to the savedAttribtes list
	**	\return		none
	**	\exception	none
	**	\author		John Corpening
	**	\date		10/19/2002 12:54:46 PM
	**	\warning	none
	**	\par		Description:
	**					If the action from this AI unit is selected by the ADN 
	**				this function will be called to save the attributes into the 
	**				saved attribute list.
	*/
	////////////////////////////////////////////////////////////////////////////////
	void saveAttributes(AI_ActionTag tag) throw();

//----Member Data--------------------------------------------------------------------
protected:
	std::vector<AI_Attribute>	attributes;			//!< List of current attributes the ADN uses as input
	std::vector<AI_ExampleADN>	trainingExamples;	//!< List of training examples
	std::map<AI_ID, AI_Unit*>	AI_Units;			//!< AI Units associated with this ADN

private:
	std::map<AI_ID, AI_Action>	actions;			//!< List of current actions specified by the AI_Units
	std::map<AI_ActionTag, 
		std::vector<AI_Attribute> > savedAttributes;//!< Used to create a training example
	AI_HigherGoal				*fitFunc;			//!< Used to determine the fitness of an ADN 
	HG_ID						hg_ID;				//!< ID of the higher goal
	Action_ID					actionClass;		//!< Action Class
	TT_ID						technique;			//!< Technique type
	bool						set;
};
}