////////////////////////////////////////////////////////////////////////////////
/*!	\file		AI_UnitBase.cpp	
**	\brief		This is the implementation file for AI_UnitBase.h
**	\author		John Corpening
**	\date		10/25/2002 6:29:10 PM
*/
////////////////////////////////////////////////////////////////////////////////

#include"AI_System.h"
using namespace AI_System;
using namespace Debugging;

// memory leak detection
#ifdef _DEBUG
    #include"memleak.h"
    #define new DEBUG_NEW
    #undef THIS_FILE
    static char THIS_FILE[] = __FILE__;
#endif

AI_ID AI_Unit::IDassigner = 0;

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	AI_Unit::AI_Unit
/*!	\brief		This constructor creates a random AI Unit
**	\return		none
**	\exception	Construction Failure
**	\author		John Corpening
**	\date		10/25/2002 9:37:00 AM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
AI_Unit::AI_Unit(TT_ID Technique)
{
	//Get unique ID
	unitID = ++IDassigner;

	//Get the PerceptGenerator
	PG = AI_PerceptGenerator::instance();
	
	//initialize all values
	currentID = decisionID = 0;
	ff_ID = FF_NULL;
	fitFunc = 0;
	action = AC_NULL;
	technique = Technique;
	set = false;
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	setFitFunc
/*!	\brief		Sets the fitness function of the AI Unit
**	\param		FitFunc, a fitness function ID
**	\author		John Corpening
**	\date		4/15/2003 9:01:10 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
void AI_Unit::setFitFunc(FF_ID FitFunc) throw()
{
	//Don't allow changes after an AI_Unit has been set
	if(set)
		return;

	//Get the fitness function
	ff_ID = FitFunc;
	fitFunc = getFitnessFunction(ff_ID);
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	initialize
/*!	\brief		Once this is called the AI_Unit cannot be reset or re-initialized
**	\exception	none
**	\author		John Corpening
**	\date		4/15/2003 8:59:52 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
void AI_Unit::initialize()
{
	//Don't allow set AI_Units to initialize
	if(set)
		return;

	//If anything is not set then select random values for it
	if(ff_ID == FF_NULL)
	{
		ff_ID = getRandomFF_ID();
		fitFunc = getFitnessFunction(ff_ID);
	}
	if(action == AC_NULL)
	{
		//default to MOVETO for now
		action = MOVETO;
	}
	if(DataFilters.size() == 0)
	{
		std::list<DF_ID> temp; //Using this list is unfortunate but it's late in the ballgame
		//Add the maximum number of default DataFilters
		for(int i = 0; i < MAXDF && i < DF_MAX; i++)
		{
			DF_ID id = PG->getRandomDF();
			if(id == DF_NULL)// || id == DF_INVALID)
			{
				throw(Nexception("Null Fitness Function Returned", __FILE__,__LINE__,__DATE__,__TIME__));
			}
			else
			{
				temp.insert(temp.begin(), id);
				//DataFilters.push_back(id);
			}
		}

		//Ensure that the data filters are unique
		//Not the best way but this is the quick fix for now
		temp.sort();
		temp.unique();
		std::list<DF_ID>::iterator it = temp.begin();
		while(it != temp.end())
		{
			DataFilters.push_back(*it);
			it++;
		}
	}
	
	//initialize the subClasses
	subClassInit();
	set = true;
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	AI_Unit::AI_Unit
/*!	\brief		Copy Constructor
**	\param		C reference to the unit to copy
**	\return		none
**	\exception	none
**	\author		John Corpening
**	\date		11/7/2002 8:18:03 PM
**	\warning	You must provide a copy constructor for derived classes
*/
////////////////////////////////////////////////////////////////////////////////
AI_Unit::AI_Unit(const AI_Unit &C)
{
	DataFilters = C.DataFilters;
	ff_ID = C.ff_ID;
	action = C.action;
	technique = C.technique;
	fitFunc = getFitnessFunction(ff_ID);
	PG = AI_PerceptGenerator::instance();
	unitID = ++IDassigner;
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	AI_Unit::getInput
/*!	\brief		calls upon the percept generator to generate a list of attributes
**	\return		none
**	\exception	none
**	\author		John Corpening
**	\date		10/19/2002 12:54:16 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
void AI_Unit::getInput()
{
	//reset the fitness function
	fitFunc->clear();
	fitFunc->evaluateFitness();

	//Clear out the attributes
	attributes.clear();

	int size = DataFilters.size();
	
	for(int i = 0; i < size; i++)
	{
		attributes.push_back(PG->requestData(DataFilters[i]));
	}
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	AI_Unit::determineAction
/*!	\brief		analyzes attributes and produces an Action
**	\return		an AI_Action: ID: ActionID, value:emphasis, unit_ID:this->unitID,
**	\exception	none
**	\author		John Corpening
**	\date		10/19/2002 12:54:00 PM
**	\warning	none
**	\par		Description:
**				This function analyzes the attributes
**				and produces an action.
*/
////////////////////////////////////////////////////////////////////////////////
AI_Action AI_Unit::determineAction()
{
	AI_Action act;

	//Fill out action information
	act.tag.unit_ID		= unitID;
	act.tag.decisionID	= 0;		//An ID of zero means disregard
	act.tag.completed	= false;
	act.tag.fitness		= fitFunc->getFitness();
	act.ID				= action;
	act.objectID		= PG->getCurrentObject();
	act.tag.decisionID	= ++decisionID;
	
	//Note that a value of 0.0 is meant to signify no action
	//Because all fitness functions are required to
	//produce an emphasis in the range of 0 to 1
	act.value	= 0.0;				

	//Determine whether or not to trigger action
	if(triggerAction())
	{
		//The action was triggered so set the emphasis
		act.value	= fitFunc->getEmphasis();
	}	
	currentID = act.tag.decisionID;
	return act;
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	AI_Unit::createTrainingExample
/*!	\brief		combines fitness and saved attributes to create a training example
**	\return		none
**	\exception	none
**	\author		John Corpening
**	\date		10/19/2002 12:54:38 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
void AI_Unit::createTrainingExample(AI_ActionTag tag)
{
	std::map<Decision_ID, std::vector<AI_Attribute> >::iterator pos;
	pos = savedAttributes.find(tag.decisionID);
	if(pos == savedAttributes.end())
		return;

	if(tag.completed)
	{
		fitFunc->evaluateFitness(true, tag.fitness);
		Debugger::Instance()->output(AI, L3,"Fish:%d, Unit: %d, Action:%s, Response:%d", PG->getTarget(), unitID, getActionString(this->getAction()), fitFunc->getResponse());
		saveTrainingExample(fitFunc->getResponse(), fitFunc->getFitness(), pos->second);
	}

	//Clear out the saved attributes
	savedAttributes.erase(pos);
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	AI_Unit::saveTrainingExample
/*!	\brief		saves a training example
**	\param		response The fitness function determined response
**	\param		fitness The fitness function determined fitness
**	\param		Attributes saved attributes
**	\return		none
**	\exception	none
**	\author		John Corpening
**	\date		11/7/2002 1:48:09 PM
**	\warning	none
**	\par		Description:
**				This function was created to allow sublasses to train
**				as they go.
*/
////////////////////////////////////////////////////////////////////////////////
void AI_Unit::saveTrainingExample(bool response, float fitness, std::vector<AI_Attribute>& Attributes)
{
	AI_ExampleUnit	ex;

	//Get the fitness after the action was executed
	ex.result = response;
	ex.fitness = fitness;
	ex.attributes = Attributes;

	//save the example
	trainingExamples.push_back(ex);
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	AI_Unit::saveAttributes
/*!	\brief		saves current attributes to the savedAttribtes list
**	\return		none
**	\exception	none
**	\author		John Corpening
**	\date		10/19/2002 12:54:46 PM
**	\warning	none
**	\par		Description:
**					If the action from this AI unit is selected by the ADN 
**				this function will be called to save the attributes into the 
**				saved attribute list.
*/
////////////////////////////////////////////////////////////////////////////////
void AI_Unit::saveAttributes()
{
	if(!currentID)
		return;
	savedAttributes.insert(std::make_pair(currentID, attributes));
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	AI_Unit::~AI_Unit
/*!	\brief		Destructor
**	\return		none
**	\exception	none
**	\author		John Corpening
**	\date		11/25/2002 8:33:42 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
AI_Unit::~AI_Unit()
{
	//delete allocated memory
	if(fitFunc)
		delete fitFunc;
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	Write
/*!	\brief		used to save the AI unit
**	\param		ostream to write to 
**	\exception	none
**	\author		John Corpening
**	\date		4/15/2003 9:01:53 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
void AI_Unit::Write(std::ostream& file) const
{
	//Note: no need to save the technique AI_module takes care of this

	//Write the ID
	file << "Unit ID: " << unitID << endl;
	//Write the FF ID
	file << "FF ID: " << ff_ID << endl;
	//Write the Action ID
	file << "Action: " << action << endl;
	//Write the number of DataFilters
	file << "DataFilters.size: " << DataFilters.size() << endl;


	//Save all the data filters
	file << "DataFiler IDs: " << endl;
	for(unsigned int i = 0; i < DataFilters.size(); i++)
	{
		file << "\t" << DataFilters[i] << endl;		
	}
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	Read
/*!	\brief		used to load the AI unit
**	\param		istream to load from
**	\exception	none
**	\author		John Corpening
**	\date		4/15/2003 9:02:03 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
void AI_Unit::Read(std::istream& file)
{
	const unsigned int buffSize = 512;
	char buff[buffSize] = {0};

	//Read the ID
	unsigned int temp;
	file.getline(buff, buffSize);
	sscanf(buff, "Unit ID: %d", &unitID);

	//Read the FF_ID
	file.getline(buff, buffSize);
	sscanf(buff, "FF ID: %d", &temp);
	ff_ID = (FF_ID)temp;

	//Create the FF
	fitFunc = getFitnessFunction(ff_ID);
	
	//Read the Action ID
	file.getline(buff, buffSize);
	sscanf(buff, "Action: %d", &temp);
	action = (Action_ID)temp;

	//Read the number of data filters
	int size;
	file.getline(buff, buffSize);
	sscanf(buff, "DataFilters.size: %d", &size);

	//Read in all the data filters, also read in the header
	file.getline(buff, buffSize);
	for(int i = 0; i < size; i++)
	{
		file.getline(buff, buffSize);
		sscanf(buff, "\t%d", &temp);
		DataFilters.push_back((DF_ID)temp);
	}

	//Handle other initialization
	PG = AI_PerceptGenerator::instance();
	set = true;
}