////////////////////////////////////////////////////////////////////////////////
/*!	\file		AIUnitBase.h
**	\brief		Base class for AI units
**	\author		John Corpening
**	\date		10/18/2002 12:59:37 PM
**	\bug		AI units will need to have globally unique ID's
**				Currently they have ID's that are unique unless the system is 
**				saved. This is because ADN's will have to reference the AI_Units'
**				ID's as the identifier in their learning techniques.
**	\par		DESCRIPTION:
**				AI units are the base AI objects of the AI system. They take as 
**				inputs attributes generated by the data filters in the percept 
**				generator and produce an action that can be interpreted by a game 
**				object. Associated with the action generated by an AI unit is a 
**				value that is intended to signify the importance placed upon the 
**				AI unit's decision.
**
**				The AI units employ one of several types of machine learning 
**				techniques. These are trained from experience by generating 
**				training examples based on their inputs and a corresponding 
**				fitness function. The fitness functions associated with AI units 
**				will correspond to specific aspects of the game object 
**				(e.g. energy level).
**
**				Upon initial creation, the data filters and fitness function used 
**				by the AI unit are randomly chosen. This information along with 
**				the internal configuration specific to the AI technique will be 
**				able to be saved to a file for future use.
**
**				This file contains the base class from which all AI units
**				will be derived. 
**
**	\par		REMARKS:
**				The AI unit base class is an abstract base class.
**
**	\par		TIPS FOR WRITING DERIVED CLASSES:
**				-The only functions that require implementation are the following
**					- triggerAction
**					- save
**					- load
**				-The only data that needs to be accessed are the following
**					- attributes : the attributes to analyze to reach a decision
**					- trainingExamples : saved examples to train the AI
**				-The saveTrainingExample function was made virtual so that AI units
**					that can "train as they go" can do so. 
**				- You may also override the "train" function
**					- If so, 
**			
**		>>>>>>> !! DO NOT FORGET TO CLEAR THE TRAINING EXAMPLES !! <<<<<<<<<
**			
**					- This was not done for you since derived classes may choose
**					  to defer training until a later cycle.
*/
////////////////////////////////////////////////////////////////////////////////

#pragma once

namespace AI_System
{
//Forward Declarations
class AI_Module;

////////////////////////////////////////////////////////////////////////////////
/*!	\class		AI_Unit
**	\brief		abstract base class for AI units
**	\author		John Corpening
**	\date		10/19/2002 12:49:19 PM
**	\warning	This class is abstract, don't try to instantiate it
*/
////////////////////////////////////////////////////////////////////////////////
class AI_Unit : public FileMemento
{
public:

	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:	AI_Unit::AI_Unit
	/*!	\brief		This constructor creates a random AI Unit
	**	\return		none
	**	\exception	Construction Failure
	**	\author		John Corpening
	**	\date		10/25/2002 9:37:00 AM
	**	\warning	none
	*/
	////////////////////////////////////////////////////////////////////////////////
	AI_Unit(TT_ID Technique);

	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:	AI_Unit::~AI_Unit
	/*!	\brief		Destructor
	**	\return		none
	**	\exception	none
	**	\author		John Corpening
	**	\date		11/25/2002 8:33:42 PM
	**	\warning	none
	*/
	////////////////////////////////////////////////////////////////////////////////
	virtual ~AI_Unit();

	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:	initialize
	/*!	\brief		Once this is called the AI_Unit cannot be reset or re-initialized
	**	\exception	none
	**	\author		John Corpening
	**	\date		4/15/2003 8:59:52 PM
	**	\warning	none
	*/
	////////////////////////////////////////////////////////////////////////////////
	void initialize();

	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:	setFitFunc
	/*!	\brief		Sets the fitness function of the AI Unit
	**	\param		FitFunc, a fitness function ID
	**	\author		John Corpening
	**	\date		4/15/2003 9:01:10 PM
	**	\warning	none
	*/
	////////////////////////////////////////////////////////////////////////////////
	void setFitFunc(FF_ID FitFunc) throw();

	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:	addDataFilter
	/*!	\brief		Adds a data filter to the AI_Unit's list of data filters
	**	\param		df, a data filter id
	**	\author		John Corpening
	**	\date		4/15/2003 9:01:21 PM
	**	\warning	none
	*/
	////////////////////////////////////////////////////////////////////////////////
	void addDataFilter(DF_ID df) throw() {if(set) return; DataFilters.push_back(df);}

	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:	setAction
	/*!	\brief		sets the action of the AI_Unit
	**	\param		Action, the action ID of the action that the AI unit specifies
	**	\author		John Corpening
	**	\date		4/15/2003 9:01:36 PM
	**	\warning	none
	*/
	////////////////////////////////////////////////////////////////////////////////
	void setAction(Action_ID Action) throw(){if(set) return; action = Action;}

	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:	getEmphasis
	/*!	\brief		returns the emphasis as determined by the fitness function
	**	\return		the emphasis
	**	\exception	none
	**	\author		John Corpening
	**	\date		4/15/2003 9:01:45 PM
	**	\warning	none
	*/
	////////////////////////////////////////////////////////////////////////////////
	float getEmphasis() throw() {return fitFunc->getEmphasis();}

	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:	Write
	/*!	\brief		used to save the AI unit
	**	\param		ostream to write to 
	**	\exception	none
	**	\author		John Corpening
	**	\date		4/15/2003 9:01:53 PM
	**	\warning	none
	*/
	////////////////////////////////////////////////////////////////////////////////
	virtual void Write(std::ostream&)const throw();

	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:	Read
	/*!	\brief		used to load the AI unit
	**	\param		istream to load from
	**	\exception	none
	**	\author		John Corpening
	**	\date		4/15/2003 9:02:03 PM
	**	\warning	none
	*/
	////////////////////////////////////////////////////////////////////////////////
	virtual void Read(std::istream&) throw();
	
public:

	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:	AI_Unit::AI_Unit
	/*!	\brief		Copy Constructor
	**	\param		C reference to the unit to copy
	**	\return		none
	**	\exception	none
	**	\author		John Corpening
	**	\date		11/7/2002 8:18:03 PM
	**	\warning	You must provide a copy constructor for derived classes
	*/
	////////////////////////////////////////////////////////////////////////////////
	AI_Unit(const AI_Unit &C);

	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:	AI_Unit::determineAction
	/*!	\brief		analyzes attributes and produces an Action
	**	\return		an AI_Action: ID: ActionID, value:emphasis, unit_ID:this->unitID,
	**	\exception	none
	**	\author		John Corpening
	**	\date		10/19/2002 12:54:00 PM
	**	\warning	none
	**	\par		Description:
	**				This function analyzes the attributes
	**				and produces an action.
	*/
	////////////////////////////////////////////////////////////////////////////////
	AI_Action determineAction() throw();

	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:	AI_Unit::getInput
	/*!	\brief		calls upon the percept generator to generate a list of attributes
	**	\return		none
	**	\exception	none
	**	\author		John Corpening
	**	\date		10/19/2002 12:54:16 PM
	**	\warning	none
	*/
	////////////////////////////////////////////////////////////////////////////////
	void getInput() throw();

	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:	AI_Unit::createTrainingExample
	/*!	\brief		combines fitness and saved attributes to create a training example
	**	\param		ID the id for this decision (only unique within the AI_Unit)
	**	\return		none
	**	\exception	none
	**	\author		John Corpening
	**	\date		10/19/2002 12:54:38 PM
	**	\warning	none
	**	\par		Description:
	**				The AI_Module will call this function after an object
	**				has performed an action specified by this AI_Unit
	*/
	////////////////////////////////////////////////////////////////////////////////
	void createTrainingExample(AI_ActionTag tag) throw();

	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:	AI_Unit::saveAttributes
	/*!	\brief		saves current attributes to the savedAttribtes list
	**	\return		none
	**	\exception	none
	**	\author		John Corpening
	**	\date		10/19/2002 12:54:46 PM
	**	\warning	none
	**	\par		Description:
	**					If the action from this AI unit is selected by the ADN 
	**				this function will be called to save the attributes into the 
	**				saved attribute list.
	*/
	////////////////////////////////////////////////////////////////////////////////
	void saveAttributes() throw();

	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:	AI_Unit::getActionType
	/*!	\brief		action accessor
	**	\return		the AI_Unit's action
	**	\exception	none
	**	\author		John Corpening
	**	\date		10/28/2002 7:59:05 PM
	**	\warning	none
	*/
	////////////////////////////////////////////////////////////////////////////////
	Action_ID getAction() const throw() {return action;}

	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:	AI_Unit::getID
	/*!	\brief		ID accessor
	**	\return		none
	**	\exception	none
	**	\author		John Corpening
	**	\date		11/7/2002 1:47:15 PM
	**	\warning	none
	*/
	////////////////////////////////////////////////////////////////////////////////
	AI_ID	getID() const throw() {return unitID;}

	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:	AI_Unit::getTechniqueType
	/*!	\brief		returns the technique id
	**	\return		none
	**	\exception	none
	**	\author		John Corpening
	**	\date		11/7/2002 8:17:29 PM
	**	\warning	none
	*/
	////////////////////////////////////////////////////////////////////////////////
	TT_ID	getTechniqueType() const throw() {return technique;}

//--------Virtual Functions for derived classes---------------------------------------
	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:	AI_Unit::train
	/*!	\brief		Trains the AI unit based on the training examples
	**	\return		none
	**	\exception	none
	**	\author		John Corpening
	**	\date		10/19/2002 12:54:07 PM
	**	\warning	YOU MUST NOT FORGET TO CLEAR YOUR EXAMPLES AFTER YOU TRAIN!
	**	\par		Note that non-trainable AI_Units do not need to override this
	*/
	////////////////////////////////////////////////////////////////////////////////
	virtual void train() throw() {trainingExamples.clear();}

	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:	AI_Unit::save
	/*!	\brief		Saves the AI unit's configuration to a file
	**	\param		TBA
	**	\return		success or failure
	**	\exception	none
	**	\author		John Corpening
	**	\date		10/19/2002 12:55:03 PM
	**	\warning	This function is subject to change pending File I/O considerations
	*/
	////////////////////////////////////////////////////////////////////////////////
	virtual bool save() throw() = 0;

	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:	AI_Unit::load
	/*!	\brief		Loads the AI unit's configuration from a file
	**	\param		TBA
	**	\return		success or failure
	**	\exception	none
	**	\author		John Corpening
	**	\date		10/19/2002 12:55:11 PM
	**	\warning	This function is subject to change pending File I/O considerations
	*/
	////////////////////////////////////////////////////////////////////////////////
	virtual bool load() throw() = 0;

protected:
	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:	AI_Unit::triggerAction
	/*!	\brief		analyzes attributes and produces an action
	**	\return		whether to trigger action or not
	**	\exception	none
	**	\author		John Corpening
	**	\date		10/24/2002 7:32:32 PM
	**	\warning	none
	**	\par		Description:
	**				Derived classes will override this function with an AI Technique.
	*/
	////////////////////////////////////////////////////////////////////////////////
	virtual bool triggerAction() throw() = 0;

	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:	AI_Unit::saveTrainingExample
	/*!	\brief		saves a training example
	**	\param		response The fitness function determined response
	**	\param		fitness The fitness function determined fitness
	**	\param		Attributes saved attributes
	**	\return		none
	**	\exception	none
	**	\author		John Corpening
	**	\date		11/7/2002 1:48:09 PM
	**	\warning	none
	**	\par		Description:
	**				This function was created to allow sublasses to train
	**				as they go.
	*/
	////////////////////////////////////////////////////////////////////////////////
	virtual void saveTrainingExample 
		(bool response, float fitness, std::vector<AI_Attribute>& Attributes) throw();

	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:	AI_Unit::subClassInit
	/*!	\brief		initialization function for subclasses
	**	\return		none
	**	\exception	none
	**	\author		John Corpening
	**	\date		11/25/2002 8:32:51 PM
	**	\warning	none
	*/
	////////////////////////////////////////////////////////////////////////////////
	virtual void subClassInit() {};

//----------Member Data-------------------------------------------------------------
protected:
	std::vector<AI_Attribute>	attributes;			//!< List of current attributes the AI Unit uses as input
	std::vector<AI_ExampleUnit>	trainingExamples;	//!< List of training examples
	std::vector<DF_ID>			DataFilters;		//!< List of data filter IDs used by the AI_Unit

private:
	AI_FitnessFunction			*fitFunc;			//!< Used to determine the fitness of an AI unit
	AI_PerceptGenerator			*PG;				//!< Used to get data
	FF_ID						ff_ID;				//!< ID of the fitness function used to evaluate this AI_Unit
	std::map<Decision_ID, 
		std::vector<AI_Attribute> > savedAttributes;//!< Used to create a training example

	AI_ID						unitID;				//!< The ID of this AI_unit (not unique across space and time)
	Action_ID					action;				//!< The action specified
	TT_ID						technique;			//!< The AI_Technique ID
	Decision_ID					decisionID;			//!< used to assign decision ID's
	Decision_ID					currentID;			//!< id of current decision
	bool						set;				//!< if this is set to true no changes can be made to setup
	
	static AI_ID				IDassigner;			//!< used to assign AI_Unit ID's
	///Internal constants
	enum					{MAXDF = 5};			//!< Maximum data filters
};
}