////////////////////////////////////////////////////////////////////////////////
/*!	\file		AI_BrainManager.cpp
**	\brief		Implementation file for AI_BrainManager.h
**	\author		John Corpening
**	\date		10/19/2002 1:24:34 PM
**	\par		DESCRIPTION:
**	\par		REMARKS:
*/
////////////////////////////////////////////////////////////////////////////////
#include"../networking/NetData.h"
#include "AI_System.h"
#include "../interpretationModule/fishbody.hpp"
using namespace AI_System;
using namespace Debugging;

// memory leak detection
#ifdef _DEBUG
    #include"memleak.h"
    #define new DEBUG_NEW
    #undef THIS_FILE
    static char THIS_FILE[] = __FILE__;
#endif
//singleton instance
AI_BrainManager* AI_BrainManager::_instance = 0;

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	AI_BrainManager::getInstance
/*!	\brief		global singleton access point for the Brain Manager
**	\return		none
**	\exception	none
**	\author		John Corpening
**	\date		11/15/2002 4:24:57 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
AI_BrainManager* AI_BrainManager::getInstance()
{
	if(!_instance)
		_instance = new AI_BrainManager;
	return _instance;
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	AI_BrainManager::AI_BrainManager
/*!	\brief		Constructor
**	\return		none
**	\exception	none
**	\author		John Corpening
**	\date		11/15/2002 4:28:15 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
AI_BrainManager::AI_BrainManager()
{
	GenAl = AI_GeneticAlgorithm::getInstance();
	GenAl->setBrainManager(this);
	debug = Debugging::Debugger::Instance();
	MH = MsgHandler::MessageHandler::Instance();
	time = TimeObject::Instance();
//	suppressed = 0;
	client = false;
	fishMin = 0;
	fishMax = 10;
}

AI_BrainManager::~AI_BrainManager()
{
	nuke();
	delete AI_PerceptGenerator::instance();
	delete GenAl;
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	AI_BrainManager::Logic
/*!	\brief		The AI-frame function
**	\return		none
**	\exception	none
**	\author		John Corpening
**	\date		10/19/2002 1:35:00 PM
**	\warning	none
**	\par		Description:
**				-Checks list to determine whether or not a fish is ready to think
**				-Activates the AI module for a fish ready to think
**				-Sends decisions out to fish that have thought
**				-Checks the dead fish list in the tank to determine which brains to delete
**				-Deletes brains of fish that are dead and not on the mating list
**				-Inactivates brains of fish that are dead and on the mating list
*/
////////////////////////////////////////////////////////////////////////////////
void AI_BrainManager::Logic() throw()
{
	//If this is a client then everything will be handled by the server brain manager
	if(client)
		return;

	//Clean up the dead list
	{
		std::map<Object_ID, AI_Module*>::iterator pos;
		std::map<Object_ID, AI_Module*>::iterator remove;
		for(pos = deadBrains.begin(); pos != deadBrains.end(); pos++)
		{
			//If a fish is not on the mating list then delete it
			if(!GenAl->checkList(pos->first))
			{
				Object_ID temp = pos->first;
				MH->RemoveFishBody(temp);
				remove = pos++;
				deadBrains.erase(remove);
			}
		}
	}


	//Check the dead fish list in the tank to determine which brains to delete
	{
		std::list<MsgHandler::Key>& theDead = MH->GetDeleteFishListRef();
		std::list<MsgHandler::Key>::iterator pos = theDead.begin();
		std::map<Object_ID, BM_Tag>::iterator erase;

		while(pos != theDead.end())
		{
			//Find the brain
			erase = brains.find(*pos);
			if(erase != brains.end())
			{
				//Check if it is on the mating list
				if(GenAl->checkList(erase->first))
				{
					//If it is then put it on the dead brains list
					deadBrains.insert(std::make_pair(erase->first, erase->second.brain));
					//and remove it from the brains list
					brains.erase(erase->first);
				}
				else
				{
					//Otherwise delete it
					Object_ID temp = erase->first;
					MH->RemoveFishBody(temp);
					delete erase->second.brain;
					brains.erase(erase);
				}
			}
			pos++;
		}
		//All the fish have either been deleted or added to the dead fish list
		//So we can clear out the dead list
		theDead.clear();
	}
	


	//Update the mating list
	{
		std::list<MsgHandler::Key>& swingers = MH->GetMatingFishListRef();
		std::list<MsgHandler::Key>::iterator pos = swingers.begin();
		std::map<Object_ID, BM_Tag>::iterator add;

		//Add the fish that want to mate to the mating list
		while(pos != swingers.end())
		{
			add = brains.find(*pos);
			debug->output(AI, L4, "Swinger ID: %d", *pos);
			if(add != brains.end())
			{
				GenAl->addToMatingList(add->second.brain, add->first);
				debug->output(AI, L2, "BM adding to Mating List: %d", add->first);
			}
			pos++;
		}
		swingers.clear();
	}
	
	//Check to see if bodies are available for the new brains
	//If so then remove them for the new brains list and add them to brains
	//If we are at the fish limit then skip this part
	if(brains.size() < fishMax)
	{
		std::map<Object_ID, AI_Module*>::iterator it = newBrains.begin();
		std::map<Object_ID, AI_Module*>::iterator erase;
		while(it != newBrains.end() && brains.size() < fishMax)
		{
			//See if a fish body was added
			if(MH->AddFishBody(it->first))
			{
				//If so then get the body
				InterpretationModule::FishBody* body = MH->GetFishBodyPtr(it->first);
				if(body)
				{
					//Get the reaction times
					BM_Tag tag;
					tag.reaction = body->m_pAttributes->GetReactionTime();
					tag.brain = it->second;
					tag.lastReaction = 0.0;
					brains.insert(std::make_pair(it->first, tag));
					erase = it++;
					newBrains.erase(erase);
				}
			}
			else
			{
				break;
			}
		}
	}
	
	//Check list to determine whether or not a fish is ready to think
	{
		std::map<Object_ID, BM_Tag>::iterator pos;
		for(pos = brains.begin(); pos != brains.end(); pos++)
		{
			pos->second.lastReaction += time->getGameTimeStep();
			//If ready to think, activated AI module
			if(pos->second.lastReaction >= pos->second.reaction)
			{
			//	pos->second.brain->generateDecisions(pos->first);
				//If the fish is being surpressed then skip it
				if(suppressed.find(pos->first) == suppressed.end())
				{
					//send decisions out to the fish
					MH->SetFishBodyActions(pos->first, pos->second.brain->generateDecisions(pos->first));
					debug->output(AI, L1, "Fish %d has had a thought", pos->first);
					//reset lastReaction
					pos->second.lastReaction = 0;
				}
			}
			
		}
	}
	//Maintain a minimum number of fish in the tank
	{
		//The number of fish to add are the difference between the minimum and the current + the unborn
		int diff = fishMin - (brains.size() + newBrains.size());
		if(diff > 0)
		{
			CreateRandomFishBrain(diff);
		}
	}
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	AI_BrainManager::Breed
/*!	\brief		Invokes the Genetic Algorithm to breed fish 
**	\param		
**	\return		none
**	\exception	none
**	\author		John Corpening
**	\date		10/19/2002 1:35:07 PM
**	\warning	none
**	\par		Description:
**				-Invokes the Genetic Algorithm to find compatible mates
**				-Creates new brains for the fish
**				-Requests a body from the tank
**				-If no body is available, adds brain to NewBrains list until 
**				 a body is available.
*/
////////////////////////////////////////////////////////////////////////////////
void AI_BrainManager::Breed() throw()
{
	GenAl->generateOffspring();
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	AI_BrainManager::Train
/*!	\brief		Allows AI modules to train.
**	\return		none
**	\exception	none
**	\author		John Corpening
**	\date		10/19/2002 1:35:15 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
void AI_BrainManager::Train() throw()
{
	std::map<Object_ID, BM_Tag>::iterator pos;
	for(pos = brains.begin(); pos != brains.end(); pos++)
	{
		pos->second.brain->train();
	}
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	AI_BrainManager::CreateFishBrain
/*!	\brief		Creates a brain given a brain configuration
**	\param		config a brain configuration
**	\return		success or failure
**	\exception	none
**	\author		John Corpening
**	\date		10/19/2002 1:34:15 PM
**	\warning	This function is not implemented
*/
////////////////////////////////////////////////////////////////////////////////
bool AI_BrainManager::CreateFishBrain(/*AI_Module configuration*/) throw()
{
	return false;
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	AI_BrainManager::SuppressBrain
/*!	\brief		Suppresses a brain given an object ID
**	\param		ID an objectID
**	\return		none
**	\exception	none
**	\author		John Corpening
**	\date		10/19/2002 1:34:22 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
bool AI_BrainManager::confirmSuppress(Object_ID target) throw()
{
	suppressed.insert(target);
	return true;
}


////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	confirmUnsuppress
/*!	\brief		unsuppresses the brain of the target given
**	\param		target, the brain to unsuppress
**	\return		true if the fish was suppressed, false if the fish wasn't
**	\exception	none
**	\author		John Corpening
**	\date		4/16/2003 9:32:58 AM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
bool AI_BrainManager::confirmUnsuppress(Object_ID target) throw()
{
	if(suppressed.erase(target))
		return true;
	else
		return false;
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	GenerateCompleteNetMessage
/*!	\brief		Sends out the suppressed list
**	\param		ServerUpdate, netmessage to fill out with the list
**	\return		none
**	\exception	none
**	\author		John Corpening
**	\date		4/16/2003 9:38:07 AM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
void AI_BrainManager::GenerateCompleteNetMessage(UDPNet::NetMessage& ServerUpdate)
{
	std::set<Object_ID>::iterator pos;
	char size = suppressed.size();
	//Write out the number of fish in the set
	ServerUpdate.WriteNBytes(1, &size);
	for(pos = suppressed.begin(); pos != suppressed.end(); pos++)
	{
		ServerUpdate.WriteNBytes(sizeof(Object_ID), &(*pos));
	}
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	LoadNetMessage
/*!	\brief		reads the list of suppressed fish
**	\param		ServerUpdate, the update sent from the server
**	\return		none
**	\exception	none
**	\author		John Corpening
**	\date		4/16/2003 9:39:23 AM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
void AI_BrainManager::LoadNetMessage(UDPNet::NetMessage& ServerUpdate)
{
	//reinstate the pathfinding of any fish that used to be suppressed
	for(std::set<Object_ID>::iterator it = suppressed.begin(); it != suppressed.end(); ++it)
	{
		InterpretationModule::FishBody *fishToGetPathfinding = 
									MsgHandler::MessageHandler::Instance()->GetFishBodyPtr(*it);
		if(fishToGetPathfinding)//if you try to unsuppress a nonexistent fish, fail silently
			fishToGetPathfinding->SetPathing(true);
	}

	//Make it so that no fish are suppressed
	suppressed.clear();

	/* Suppress the Fish the server says to suppress */
	char size = 0;
	ServerUpdate.ReadNBytes(1, &size);
	if(!size)
		return;
	for(char i = 0; i < size; i++)
	{
		Object_ID temp = 0;
		ServerUpdate.ReadNBytes(sizeof(Object_ID), &temp);
		suppressed.insert(temp);

		//lobotomize the suppressed fish's pathfinding
		InterpretationModule::FishBody *fishToLosePathfinding = 
									MsgHandler::MessageHandler::Instance()->GetFishBodyPtr(temp);
		if(fishToLosePathfinding)//if you try to suppress a nonexistent fish, fail silently
			fishToLosePathfinding->SetPathing(false);
	}
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	AI_BrainManager::SuppressBrainWithAction
/*!	\brief		Suppresses a brain given an object ID and an actionID
**	\param		objectID	the ID of the brain to suppress
**	\param		actionID	the ID of the action to train the brain with
**	\return		none
**	\exception	none
**	\author		John Corpening
**	\date		10/19/2002 1:34:29 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
void AI_BrainManager::SuppressBrainWithAction(Object_ID target, Action_ID action) throw()
{
	//Find the fish brain
	std::map<Object_ID, BM_Tag>::iterator pos = brains.find(target);

	//bail if the brain does not exist
	if(pos == brains.end())
		return;

	//Fill out the action and value fields
	AI_Action temp;
	temp.ID = action;
	temp.objectID = 0;
	temp.value = 1.0f;
	temp.tag.unit_ID = 0;
	std::vector<AI_Action> tempV;
	tempV.push_back(temp);

	//send decisions out to the fish
	MH->SetFishBodyActions(target, tempV);
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	AI_BrainManager::AddToMatingList
/*!	\brief		Adds a given object to the mating list
**	\param		ID the id of the object to add to the mating list
**	\return		none
**	\exception	none
**	\author		John Corpening
**	\date		10/19/2002 1:34:37 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
void AI_BrainManager::AddToMatingList(Object_ID target) throw()
{
	//Check to make sure the brain exists
	std::map<Object_ID, BM_Tag>::iterator pos;
	pos = brains.find(target);
	//If not then bail
	if(pos == brains.end())
	{
		debug->output(Debugging::AI, Debugging::L2, "AddToMatingList: target %d not in list", target);
		return;
	}

	GenAl->addToMatingList(pos->second.brain, target);
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	AI_BrainManager::CreateRandomFishBrain
/*!	\brief		Creates a brain for a fish with a random configuration
**	\return		none
**	\exception	none
**	\author		John Corpening
**	\date		10/19/2002 1:34:51 PM
**	\warning	none
**	\par		Note: This function will tell tank to create a new fish body
*/
////////////////////////////////////////////////////////////////////////////////
void AI_BrainManager::CreateRandomFishBrain(int num) throw()
{
	//Prevent fish from going over the maximum
	if((brains.size() + newBrains.size()) >= fishMax)
		return;
	int numFins = 10;//, numMouth = 5, numMarker = 5;
	int j;
	
	for(int i = 0; i < num; i++)
	{
		AI_Module *brain = new AI_Module;

		//Install the appropriate number and type of AI_Units
		for(j = 0; j < numFins; j++)
		{
			brain->addAIUnit(getRandomAIUnit(AC_MOVE));
		}
//Note: It was decided to remove mouth and marker ADNs
//Update made 4/15/03 by John 
//		for(j = 0; j < numMouth; j++)
//		{
//			brain->addAIUnit(getRandomAIUnit(AC_MOUTH));
//		}
//		for(j = 0; j < numMarker; j++)
//		{
//			brain->addAIUnit(getRandomAIUnit(AC_MARKER));
//		}

		//Add the brain to the brain manager
		brain->initialize();
		addNewBrain(brain);
	}
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	AI_BrainManager::addNewBrain
/*!	\brief		adds a brain to the brain manager
**	\param		brain, pointer to the AI_Module to add to the brain manager
**	\return		none
**	\exception	none
**	\author		John Corpening
**	\date		11/22/2002 12:01:32 PM
**	\warning	none
**	\par		Note: This function will tell tank to create a new fish body
*/
////////////////////////////////////////////////////////////////////////////////
void AI_BrainManager::addNewBrain(AI_Module* brain)
{
	//Get an object id
	Object_ID id = MH->GenerateKey();
	//Set the key of the brain
	brain->setBodyID(id);
	//Check the tank to see if a body is available
	//And make sure this doesn't put us over the limit
	if((1+brains.size() <= fishMax) && MH->AddFishBody(id, brain->getMom(), brain->getDad()))
	{
		InterpretationModule::FishBody* body = MH->GetFishBodyPtr(id);
		if(body)
		{
			BM_Tag tag;
			//Get the reaction times
			tag.reaction = body->m_pAttributes->GetReactionTime();
			tag.brain = brain;
			tag.lastReaction = 0.0;
			//If so then assign the body to the brain
			brains.insert(std::make_pair(id, tag));
			//We have added it so return
			return;
		}
	}

	//If No body available or the body returned was null then add to newBrains
	newBrains.insert(std::make_pair(id, brain));
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	makeClient
/*!	\brief		prepares the brainmanager for client mode
**	\exception	none
**	\author		John Corpening
**	\date		4/16/2003 9:40:26 AM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
void AI_BrainManager::makeClient()
{
	client = true;
	nuke();
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	unmakeClient
/*!	\brief		resets the brain manager to non-client operation
**	\exception	none
**	\author		John Corpening
**	\date		4/16/2003 9:44:18 AM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
void AI_BrainManager::unmakeClient()
{
	client = false;
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	nuke
/*!	\brief		clears out the brain manager
**	\return		none
**	\exception	none
**	\author		John Corpening
**	\date		4/16/2003 9:46:01 AM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
void AI_BrainManager::nuke()
{
	//The new brains can be cleaned out
	std::map<Object_ID, AI_Module*>::iterator pos = newBrains.begin();
	while(pos != newBrains.end())
	{
		delete pos->second;
		pos++;
	}
	newBrains.clear();

	//Clear out the mating list
	GenAl->clearMatingList();

	//Clean out the dead brains list
	for(pos = deadBrains.begin(); pos != deadBrains.end(); pos++)
	{
		delete pos->second;
	}
	deadBrains.clear();

	//Clean out the brains
	std::map<Object_ID, BM_Tag>::iterator pos2 = brains.begin();
	while(pos2 != brains.end())
	{
		delete pos2->second.brain;
		pos2++;
	}
	brains.clear();

	//clean out the suppressed list
	suppressed.clear();
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	isSuppressed
/*!	\brief		checks if an object is suppressed or not
**	\param		target, the ID of the object to check
**	\return		true if suppressed, false otw
**	\exception	none
**	\author		John Corpening
**	\date		4/16/2003 9:47:34 AM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
bool AI_BrainManager::isSuppressed(Object_ID target)
{
	if(suppressed.find(target) != suppressed.end())
		return true;
	return false;
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	getMoveGoalString
/*!	\brief		gets a string that describes the given fish's movement goal
**	\param		target, the fish to check
**	\param		goal, the returned string
**	\return		false if the fish is not on the brain manager's list
**	\exception	none
**	\author		John Corpening
**	\date		4/16/2003 9:53:26 AM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
bool AI_BrainManager::getMoveGoalString(Object_ID target, std::string& goal)
{
	std::map<Object_ID, BM_Tag>::iterator	pos = brains.find(target);

	if(pos == brains.end())
	{
		return false;
	}
	goal = getHigherGoalString(pos->second.brain->getGoal(AC_MOVE));

	return true;
}

float AI_BrainManager::getMatingChoosiness()
{
	return GenAl->getCompatability();
}

void AI_BrainManager::setMatingChoosiness(float val)
{
	GenAl->setCompatability(val);
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	Write
/*!	\brief		Saves the current AI system to the given ostream
**	\param		file, the ostream to save to
**	\return		none
**	\exception	none
**	\author		John Corpening
**	\date		4/16/2003 9:49:09 AM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
void AI_BrainManager::Write(std::ostream& file) const
{
	//Pass the ostream on to each of the active brains
	file << "Number of brains: " << brains.size() << " \n";
	std::map<Object_ID, BM_Tag>::const_iterator pos;
	for(pos = brains.begin(); pos != brains.end(); pos++)
	{
		//Save the reaction times
		file << "Object_ID: " << pos->first << " \n";
		file << "Reaction time: " << pos->second.reaction << " \n";
		pos->second.brain->Write(file);
	}
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	Read
/*!	\brief		loads the AI system from the given 
**	\param		file, the stream to read from
**	\return		none
**	\exception	"Key already registered" nexception
**	\author		John Corpening
**	\date		4/16/2003 9:50:11 AM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
void AI_BrainManager::Read(std::istream& file)
{
	//Erase everything currently in the brain manager
	nuke();

	const unsigned int bufSize = 512;
	char buf[bufSize];
	unsigned int count;
	file.getline(buf, bufSize);
	sscanf(buf,"Number of brains: %d", &count);

	//Create brains and have them read from the file
	for(unsigned int i = 0; i < count; i++)
	{
		Object_ID ID;
		BM_Tag tag;
		tag.lastReaction = 0;
		tag.brain = new AI_Module;

		file.getline(buf, bufSize);
		sscanf(buf, "Object_ID: %d", &ID);

		file.getline(buf, bufSize);
		sscanf(buf,"Reaction time: %lf", &(tag.reaction));

		tag.brain->Read(file);
		tag.brain->setBodyID(ID);
		brains.insert(std::make_pair(ID, tag));
	}
}