////////////////////////////////////////////////////////////////////////////////
/*!	\file		AI_BrainManager.h
**	\brief		Manager for the game AI
**	\author		John Corpening
**	\date		10/19/2002 1:24:34 PM
**	\par		DESCRIPTION:
**	\par		REMARKS:
*/
////////////////////////////////////////////////////////////////////////////////
#pragma once

namespace AI_System
{

struct BM_Tag
{
	AI_Module*		brain;
	double			reaction;		//!< reaction time in seconds
	double			lastReaction;	//!< time since last Reaction
};
////////////////////////////////////////////////////////////////////////////////
/*!	\class		AI_BrainManager.h
**	\brief		Manager for the game AI
**	\author		John Corpening
**	\date		10/28/2002 11:48:56 AM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
class AI_BrainManager : public FileMemento
{
	std::map<Object_ID, BM_Tag>			brains;		//!< List of active AI modules in the system
	std::map<Object_ID, AI_Module*>		newBrains;	//!< List of newborn brains waiting for a body
	std::map<Object_ID, AI_Module*>		deadBrains;	//!< List of dead fish on the mating list
	AI_GeneticAlgorithm					*GenAl;		//!< A genetic algorithm module used for breeding
	TimeObject							*time;		//!< Used to check when an object is ready to think
	Debugging::Debugger					*debug;		//!< Used for debugging
	MsgHandler::MessageHandler			*MH;		//!< Message Handler, connects Brain Manager to game
	static AI_BrainManager				*_instance;	//!< singleton instance
	std::set<Object_ID>					suppressed; //!< Suppressed Fish ID
	bool								client;		//!< Flag to indicate that this is a client
	unsigned int						fishMin;	//!< the minimum number of fish in the tank
	unsigned int						fishMax;	//!< The maximum number of fish in the tank

protected:
	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:	AI_BrainManager::AI_BrainManager
	/*!	\brief		Constructor
	**	\return		none
	**	\exception	none
	**	\author		John Corpening
	**	\date		11/15/2002 4:28:15 PM
	**	\warning	none
	*/
	////////////////////////////////////////////////////////////////////////////////
	AI_BrainManager();
public:

	~AI_BrainManager();
	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:	AI_BrainManager::getInstance
	/*!	\brief		global singleton access point for the Brain Manager
	**	\return		none
	**	\exception	none
	**	\author		John Corpening
	**	\date		11/15/2002 4:24:57 PM
	**	\warning	none
	*/
	////////////////////////////////////////////////////////////////////////////////
	static AI_BrainManager* getInstance();

	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:	AI_BrainManager::Logic
	/*!	\brief		The AI-frame function
	**	\return		none
	**	\exception	none
	**	\author		John Corpening
	**	\date		10/19/2002 1:35:00 PM
	**	\warning	none
	**	\par		Description:
	**				-Checks list to determine whether or not a fish is ready to think
	**				-Activates the AI module for a fish ready to think
	**				-Sends decisions out to fish that have thought
	**				-Checks the dead fish list in the tank to determine which brains to delete
	**				-Deletes brains of fish that are dead and not on the mating list
	**				-Inactivates brains of fish that are dead and on the mating list
	*/
	////////////////////////////////////////////////////////////////////////////////
	void Logic() throw();

	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:	AI_BrainManager::Breed
	/*!	\brief		Invokes the Genetic Algorithm to breed fish 
	**	\param		
	**	\return		none
	**	\exception	none
	**	\author		John Corpening
	**	\date		10/19/2002 1:35:07 PM
	**	\warning	none
	**	\par		Description:
	**				-Invokes the Genetic Algorithm to find compatible mates
	**				-Creates new brains for the fish
	**				-Requests a body from the tank
	**				-If no body is available, adds brain to NewBrains list until 
	**				 a body is available.
	*/
	////////////////////////////////////////////////////////////////////////////////
	void Breed() throw();

	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:	AI_BrainManager::Train
	/*!	\brief		Allows AI modules to train.
	**	\return		none
	**	\exception	none
	**	\author		John Corpening
	**	\date		10/19/2002 1:35:15 PM
	**	\warning	none
	*/
	////////////////////////////////////////////////////////////////////////////////
	void Train() throw();

	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:	AI_BrainManager::CreateFishBrain
	/*!	\brief		Creates a brain given a brain configuration
	**	\param		config a brain configuration
	**	\return		success or failure
	**	\exception	none
	**	\author		John Corpening
	**	\date		10/19/2002 1:34:15 PM
	**	\warning	This function is not implemented
	*/
	////////////////////////////////////////////////////////////////////////////////
	bool CreateFishBrain(/*AI_Module configuration*/) throw();

	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:	AI_BrainManager::SuppressBrain
	/*!	\brief		Suppresses a brain given an object ID
	**	\param		ID an objectID
	**	\return		none
	**	\exception	none
	**	\author		John Corpening
	**	\date		10/19/2002 1:34:22 PM
	**	\warning	none
	*/
	////////////////////////////////////////////////////////////////////////////////
	bool confirmSuppress(Object_ID target) throw();

	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:	confirmUnsuppress
	/*!	\brief		unsuppresses the brain of the target given
	**	\param		target, the brain to unsuppress
	**	\return		true if the fish was suppressed, false if the fish wasn't
	**	\exception	none
	**	\author		John Corpening
	**	\date		4/16/2003 9:32:58 AM
	**	\warning	none
	*/
	////////////////////////////////////////////////////////////////////////////////
	bool confirmUnsuppress(Object_ID target) throw();

	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:	isSuppressed
	/*!	\brief		checks if an object is suppressed or not
	**	\param		target, the ID of the object to check
	**	\return		true if suppressed, false otw
	**	\exception	none
	**	\author		John Corpening
	**	\date		4/16/2003 9:47:34 AM
	**	\warning	none
	*/
	////////////////////////////////////////////////////////////////////////////////
	bool isSuppressed(Object_ID target) throw();

	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:	GenerateCompleteNetMessage
	/*!	\brief		Sends out the suppressed list
	**	\param		ServerUpdate, netmessage to fill out with the list
	**	\return		none
	**	\exception	none
	**	\author		John Corpening
	**	\date		4/16/2003 9:38:07 AM
	**	\warning	none
	*/
	////////////////////////////////////////////////////////////////////////////////
	void GenerateCompleteNetMessage(UDPNet::NetMessage &ServerUpdate);

	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:	LoadNetMessage
	/*!	\brief		reads the list of suppressed fish
	**	\param		ServerUpdate, the update sent from the server
	**	\return		none
	**	\exception	none
	**	\author		John Corpening
	**	\date		4/16/2003 9:39:23 AM
	**	\warning	none
	*/
	////////////////////////////////////////////////////////////////////////////////
	void LoadNetMessage(UDPNet::NetMessage &ServerUpdate);

	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:	getMoveGoalString
	/*!	\brief		gets a string that describes the given fish's movement goal
	**	\param		target, the fish to check
	**	\param		goal, the returned string
	**	\return		false if the fish is not on the brain manager's list
	**	\exception	none
	**	\author		John Corpening
	**	\date		4/16/2003 9:53:26 AM
	**	\warning	none
	*/
	////////////////////////////////////////////////////////////////////////////////
	bool getMoveGoalString(Object_ID target, std::string& goal) throw();

	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:	AI_BrainManager::SuppressBrainWithAction
	/*!	\brief		Suppresses a brain given an object ID and an actionID
	**	\param		objectID	the ID of the brain to suppress
	**	\param		actionID	the ID of the action to train the brain with
	**	\return		none
	**	\exception	none
	**	\author		John Corpening
	**	\date		10/19/2002 1:34:29 PM
	**	\warning	none
	*/
	////////////////////////////////////////////////////////////////////////////////
	void SuppressBrainWithAction(Object_ID target, Action_ID action) throw();

	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:	AI_BrainManager::AddToMatingList
	/*!	\brief		Adds a given object to the mating list
	**	\param		ID the id of the object to add to the mating list
	**	\return		none
	**	\exception	none
	**	\author		John Corpening
	**	\date		10/19/2002 1:34:37 PM
	**	\warning	none
	*/
	////////////////////////////////////////////////////////////////////////////////
	void AddToMatingList(Object_ID target) throw();

	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:	AI_BrainManager::CreateRandomFishBrain
	/*!	\brief		Creates a brain for a fish with a random configuration
	**	\return		none
	**	\exception	none
	**	\author		John Corpening
	**	\date		10/19/2002 1:34:51 PM
	**	\warning	none
	*/
	////////////////////////////////////////////////////////////////////////////////
	void CreateRandomFishBrain(int num = 1) throw();

	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:	AI_BrainManager::addNewBrain
	/*!	\brief		adds a brain to the brain manager
	**	\param		brain, pointer to the AI_Module to add to the brain manager
	**	\return		none
	**	\exception	none
	**	\author		John Corpening
	**	\date		11/22/2002 12:01:32 PM
	**	\warning	none
	**	\par		Note: This function will tell tank to create a new fish body
	*/
	////////////////////////////////////////////////////////////////////////////////
	void addNewBrain(AI_Module* brain);

	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:	makeClient
	/*!	\brief		calls nuke and sets the client flag to true
	**	\exception	none
	**	\author		John Corpening
	**	\date		4/16/2003 9:40:26 AM
	**	\warning	none
	*/
	////////////////////////////////////////////////////////////////////////////////
	void makeClient();

	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:	unmakeClient
	/*!	\brief		resets the brain manager to non-client operation
	**	\exception	none
	**	\author		John Corpening
	**	\date		4/16/2003 9:44:18 AM
	**	\warning	none
	*/
	////////////////////////////////////////////////////////////////////////////////
	void unmakeClient();

	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:	nuke
	/*!	\brief		clears out the brain manager
	**	\return		none
	**	\exception	none
	**	\author		John Corpening
	**	\date		4/16/2003 9:46:01 AM
	**	\warning	none
	*/
	////////////////////////////////////////////////////////////////////////////////
	void nuke();

	//Accessors
	void setFishMin(unsigned int val){fishMin = (val > fishMax ? fishMax : val);}
	void setFishMax(unsigned int val){fishMax = (val < fishMin ? fishMin : val);}
	unsigned int getFishMin(){return fishMin;}
	unsigned int getFishMax(){return fishMax;}
	unsigned int numActiveFish(){return brains.size();}
	unsigned int numWaitingFish(){return newBrains.size();}
	void setMatingChoosiness(float val);
	float getMatingChoosiness();

	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:	Write
	/*!	\brief		Saves the current AI system to the given ostream
	**	\param		file, the ostream to save to
	**	\return		none
	**	\exception	none
	**	\author		John Corpening
	**	\date		4/16/2003 9:49:09 AM
	**	\warning	none
	*/
	////////////////////////////////////////////////////////////////////////////////
	void Write(std::ostream&)const;

	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:	Read
	/*!	\brief		loads the AI system from the given 
	**	\param		file, the stream to read from
	**	\return		none
	**	\exception	"Key already registered" nexception
	**	\author		John Corpening
	**	\date		4/16/2003 9:50:11 AM
	**	\warning	none
	*/
	////////////////////////////////////////////////////////////////////////////////
	void Read(std::istream&);
	
};

}