////////////////////////////////////////////////////////////////////////////////
/*!	\file		AI_DataFilterBase
**	\brief		Base class for Data Filters
**	\author		John Corpening
**	\date		10/18/2002 1:11:01 PM
**	\par		DESCRIPTION:
**				Data filters are objects that the percept generator maintains to
**				present data in a meaningful manner to the AI_Units.
**				They do this by generating attribute objects. Attributes are used 
**				by the decision-making process to evaluate the game object's 
**				situation and determine actions for the game object to take.
**
**				Data filters, will only do their computations once when they are 
**				first invoked. The result will be stored and made available if 
**				the data filter is invoked again. After the AI module finishes 
**				processing, the data filters will be reset to ensure that values 
**				are re-calculated the next time the AI module is invoked. Some AI 
**				filters will require being reset with each pass of the AI module. 
**				They may do this when the percept generator notifies them to do so.
**
**	\par		REMARKS:
**				The data filter class is an abstract class, don't try to instantiate
**				it directly.
**
**	\par		Tips for writing Data Filters:
**				-The message handler pointer is the access point for data from the game
**				-Get the target (the object that is thinking)from the PerceptGenerator
*/
////////////////////////////////////////////////////////////////////////////////

#pragma once

namespace AI_System
{

////////////////////////////////////////////////////////////////////////////////
/*!	\class		AI_DataFilter
**	\brief		Class to gather data for use by AI_Units
**	\author		John Corpening
**	\date		10/18/2002 1:12:27 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
class AI_DataFilter
{
public:

	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:	AI_DataFilter::AI_DataFilter
	/*!	\brief		Constructor gets access to the MessageHandler
	**	\exception	none
	**	\author		John Corpening
	**	\date		10/18/2002 4:28:00 PM
	**	\warning	none
	*/
	////////////////////////////////////////////////////////////////////////////////
	AI_DataFilter();

	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:	AI_DataFilter::reset
	/*!	\brief		Clears the dataProcessed flag
	**	\return		none
	**	\exception	none
	**	\author		John Corpening
	**	\date		10/18/2002 4:09:27 PM
	**	\warning	none
	*/
	////////////////////////////////////////////////////////////////////////////////
	void reset(){dataProcessed = false;}

	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:	AI_DataFilter::getAttribute
	/*!	\brief		returns the attribute calculated by the DataFilter
	**	\return		an AI_Attribute to be used by the AI Units
	**	\exception	none
	**	\author		John Corpening
	**	\date		10/18/2002 4:14:38 PM
	**	\warning	none
	*/
	////////////////////////////////////////////////////////////////////////////////
	AI_Attribute getAttribute();

	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:	AI_DataFilter::processData
	/*!	\brief		virtual function to process the data
	**	\return		none
	**	\exception	none
	**	\author		John Corpening
	**	\date		10/18/2002 4:11:16 PM
	**	\warning	none
	**	\par		Description:
	**					-Queries the game for data
	**					-processes the data
	**					-Creates an attribute
	**					-sets the dataProcessed flag
	*/
	////////////////////////////////////////////////////////////////////////////////
	virtual float processData() = 0;

	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:	AI_DataFilter::endPass
	/*!	\brief		Optional virtual function to do end of pass processing
	**	\return		none
	**	\exception	none
	**	\author		John Corpening
	**	\date		10/18/2002 4:17:09 PM
	**	\warning	none
	*/
	////////////////////////////////////////////////////////////////////////////////
	virtual void endPass(){}
	
	void setPerceptGenerator(AI_PerceptGenerator *pg){PG = pg;}
	DF_ID getID(){return ID;}
protected:
	bool						dataProcessed;	//!< Indicates that the data filter has already processed the data.
	MsgHandler::MessageHandler	*MH;			//!< Reference to the message handler
	AI_PerceptGenerator			*PG;			//!< Reference to the PerceptGenerator to get data
	DF_ID						ID;				//!< An id number used to identify the Data Filter

private:
	AI_Attribute				attribute;		//!< Attribute generated by the data filter

};

}//End namespace AI_System

