////////////////////////////////////////////////////////////////////////////////
/*!	\file		AI_FitnessFunctionBase.cpp
**	\brief		This is the implementation file for AI_FitnessFunctionBase.h
**	\author		John Corpening
**	\date		10/26/2002 1:41:14 PM
*/
////////////////////////////////////////////////////////////////////////////////

#include"AI_System.h"
using namespace AI_System;
using namespace Debugging;

// memory leak detection
#ifdef _DEBUG
    #include"memleak.h"
    #define new DEBUG_NEW
    #undef THIS_FILE
    static char THIS_FILE[] = __FILE__;
#endif
////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	AI_FitnessFunction::AI_FitnessFunction
/*!	\brief		sets the message handler and percept generator pointers
**	\return		none
**	\exception	Nexception for NULL pointers returned
**	\author		John Corpening
**	\date		10/26/2002 12:58:07 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
AI_FitnessFunction::AI_FitnessFunction()
{
	PG = AI_PerceptGenerator::instance();
	if(!PG)
	{
		throw(Nexception("FF: NULL Percept Generator", __FILE__,__LINE__,__DATE__,__TIME__));
	}

	fitness = 0.0;
	emphasis = 0.0;
	dataProcessed = false;
	response = false;

}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	FitnessFunction::getEmphasis
/*!	\brief		returns the emphasis associated with an action
**	\return		the emphasis associated with an action
**	\exception	none
**	\author		John Corpening
**	\date		10/18/2002 5:33:56 PM
**	\warning	The emphasis was formerly known as the priority
**	\par		Description:
**					The emphasis is a value placed on an action
**				specified by an AI Unit to indicate how important
**				the AI Unit's decision is.
**	\par		Remarks:
**					If the data processed flag is not set then
**				this function calls the evaluateFitness function.
*/
////////////////////////////////////////////////////////////////////////////////
float AI_FitnessFunction::getEmphasis()
{
	if(!dataProcessed)
		evaluateFitness();
	return emphasis;
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	FitnessFunction::getFitness
/*!	\brief		returns the fitness evaluated by the fitness function
**	\return		the fitness evaluated by the fitness function
**	\exception	none
**	\author		John Corpening
**	\date		10/18/2002 5:34:02 PM
**	\warning	none
**	\par		Remarks:
**					If the data processed flag is not set then
**				this function calls the evaluateFitness function.
*/
////////////////////////////////////////////////////////////////////////////////
float AI_FitnessFunction::getFitness()
{
	if(!dataProcessed)
		evaluateFitness();
	if(fitness > 1.0 || fitness < 0.0)
		Debugger::Instance()->output(AI, L4, "Fitness out of range: %d", ID);
	return fitness;
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	FitnessFunction::getResponse
/*!	\brief		returns the response 
**	\return		whether the fitness function "feels" good or not
**	\exception	none
**	\author		John Corpening
**	\date		10/18/2002 5:34:14 PM
**	\warning	none
**	\par		Description:
**					The response is whether or not the fitness function
**				feels things are going well or not. It will be used
**				to create training examples.
**	\par		Remarks:
**					If the data processed flag is not set then
**				this function calls the evaluateFitness function.
*/
////////////////////////////////////////////////////////////////////////////////
bool AI_FitnessFunction::getResponse()
{
	if(!dataProcessed)
		evaluateFitness();
	return response;
}