////////////////////////////////////////////////////////////////////////////////
/*!	\file		AI_GeneticAlgorithm.cpp
**	\brief		Implementation file for AI_GeneticAlgorithm.h
**	\author		John Corpening
**	\date		11/15/2002 3:54:58 PM
*/
////////////////////////////////////////////////////////////////////////////////
#include "AI_System.h"
using namespace AI_System;
using namespace Debugging;

// memory leak detection
#ifdef _DEBUG
    #include"memleak.h"
    #define new DEBUG_NEW
    #undef THIS_FILE
    static char THIS_FILE[] = __FILE__;
#endif
//static Singleton instance
AI_GeneticAlgorithm* AI_GeneticAlgorithm::_instance = 0;

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	AI_GeneticAlgorithm::getInstance
/*!	\brief		global access point for the genetic algorithm
**	\return		pointer to the genetic algorithm singleton
**	\exception	creation failure
**	\author		John Corpening
**	\date		11/15/2002 3:36:00 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
AI_GeneticAlgorithm* AI_GeneticAlgorithm::getInstance()
{
	if(_instance)
		return _instance;
	return new AI_GeneticAlgorithm;
}


////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	AI_GeneticAlgorithm::AI_GeneticAlgorithm
/*!	\brief		initializes genetic algorithm
**	\return		none
**	\exception	none
**	\author		John Corpening
**	\date		11/15/2002 3:53:23 PM
**	\warning	Does not set the Brain Manager
*/
////////////////////////////////////////////////////////////////////////////////
AI_GeneticAlgorithm::AI_GeneticAlgorithm()
{
	//set defaults
	maxTries		= MAXTRIES;
	mutationRate	= MUTATION;	
	budMutationRate	= BUDMUTATION;
	compatability	= 0.10f;	//10 percent of higher goals must match for compatability

	//Add all higher goals to the higher goals list
	instantiateHigherGoals();

	//Get a handle to the percept generator
	PG = AI_PerceptGenerator::instance();
}

AI_GeneticAlgorithm::~AI_GeneticAlgorithm()
{
	for(unsigned int i = 0; i < higherGoals.size(); i++)
	{
		delete higherGoals[i];
	}
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	AI_GeneticAlgorithm::addToMatingList
/*!	\brief		adds a brain to the mating list
**	\param		brain a pointer to the AI_Module to add
**	\param		object the ID of the object
**	\return		none
**	\exception	none
**	\author		John Corpening
**	\date		11/15/2002 3:42:19 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
void AI_GeneticAlgorithm::addToMatingList(AI_Module *brain, Object_ID object)
{
	GA_Tag	tag;
	tag.brain = brain;
	tag.tries = 0;
	matingList.insert(std::make_pair(object, tag));
	Debugger::Instance()->output(AI, L3, "Added object %d to the mating list", object);
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	AI_GeneticAlgorithm::generateOffspring
/*!	\brief		attempts to create offspring from fish on the mating list
**	\return		none
**	\exception	none
**	\author		John Corpening
**	\date		11/15/2002 3:44:11 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
void AI_GeneticAlgorithm::generateOffspring()
{
	Debugger::Instance()->output(AI,L1,"generateOffspring, the mating list has %d", matingList.size());
	std::map<Object_ID, GA_Tag>::iterator pos = matingList.begin();
	std::map<Object_ID, GA_Tag>::iterator pos2 = pos;
	std::map<Object_ID, GA_Tag>::iterator remove;
	bool mateflag = false;
	
	while(pos != matingList.end())
	{
		mateflag = false;
		//Take the current fish
		pos2 = pos;
		pos2++;
		//Check the rest of the list
		while(pos2 != matingList.end())
		{
			//If a mate is found 
			if(findMate(pos->second.brain, pos2->second.brain))
			{
				//create offspring
				mate(pos->second.brain, pos2->second.brain);
				//remove this fish from the list
				remove = pos++;
				matingList.erase(remove);
				//remove the mate from the list
				if(pos == pos2)
					pos++;
				matingList.erase(pos2);
				
				mateflag = true;
				break;
			}
			pos2++;
		}
		//Skip the budding section
		if(mateflag)
		{
			continue;
		}
		//If a mate is not found
		//increment tries
		if(++(pos->second.tries) > maxTries)
		{
			//if tries >= maxTries
			//bud the fish
			bud(pos->second.brain);
			//remove the fish from the list
			remove = pos++;
			matingList.erase(remove);
		}
		else
		{
			pos++;
		}
	}
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	AI_GeneticAlgorithm::findMate
/*!	\brief		determines whether two fish are compatible or not
**	\param		mom pointer to one parent's AI_Module
**	\param		dad pointer to the other parent's AI_Module
**	\return		none
**	\exception	none
**	\author		John Corpening
**	\date		11/15/2002 3:45:04 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
bool AI_GeneticAlgorithm::findMate(AI_Module *mom, AI_Module *dad)
{
	//Determine the compatability count
	int target = static_cast<int>(compatability * higherGoals.size());
	//make sure we are accepting at least one
	target = (target == 0 ? 1 : target);
	int count = 0;
	std::vector<float> mResults;

	//Evaluate mom using each of the higher goals
	PG->setTarget(mom->getTarget());
	for(unsigned int i = 0; i < higherGoals.size(); i++)
	{
		mResults.push_back(higherGoals[i]->getFitness());		
	}

	//Evaluate dad using each of the higher goals
	float low, high, fitness;
	PG->setTarget(dad->getTarget());
	for(i = 0; i < higherGoals.size(); i++)
	{	
		low = mResults[i] - 0.125f;
		high = mResults[i] + 0.125f; 
		fitness = higherGoals[i]->getFitness();
		if(fitness > low && fitness < high)
		{
			count++;
		}
	}
	//Compare the results
	if(count >= target)
	{
		//If they are a close enough match then return true
		return true;
	}
	//Otherwise return false
	return false; 
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	AI_GeneticAlgorithm::mate
/*!	\brief		applies crossover to the parents to create an offspring
**	\param		mom pointer to one parent's AI_Module
**	\param		dad pointer to the other parent's AI_Module
**	\return		none
**	\exception	none
**	\author		John Corpening
**	\date		11/15/2002 3:46:24 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
void AI_GeneticAlgorithm::mate(AI_Module *mom, AI_Module *dad)
{
	//determine the number and types of AI_Modules
	//Get the parent's AI_Units
	std::multimap<Action_ID, AI_Unit*> mUnits = mom->getAIUnits();
	std::multimap<Action_ID, AI_Unit*> dUnits = dad->getAIUnits();
	
	//Create the child's AI_Module
	AI_Module *child = new AI_Module;
	
	//For each ADN Action type
	std::multimap<Action_ID, AI_Unit*>::iterator mPos, dPos, erase;
	Action_ID	type;
	unsigned int count;
	bool dadOut, momOut;
	while(mUnits.size() && dUnits.size())
	{
		mPos = mUnits.begin();
		dPos = dUnits.begin();
		type = mPos->first;
		dadOut = false;
		momOut = false;
		count = (int)(((double)mUnits.count(type) + (double)dUnits.count(type) + 0.5)/2.0);
		while(count)
		{
			//Roll the dice and determine whether to mutate or not
			if(rollDice(mutationRate))
			{
				//Create a new AI Unit of this type
				AI_Unit* newUnit = getRandomAIUnit(type);
				child->addAIUnit(newUnit);
			}
			else
			{
				//If mom is out then select from dad
				if(momOut)
				{
					AI_Unit* newUnit = copyAIUnit(dPos->second);
					child->addAIUnit(newUnit);
					//remove the unit from the list
					erase = dPos++;
					dUnits.erase(erase);
				}
				//If dad is out then select from mom
				else if(dadOut)
				{
					AI_Unit* newUnit = copyAIUnit(mPos->second);
					child->addAIUnit(newUnit);
					//remove the unit from the list
					erase = mPos++;
					mUnits.erase(erase);
				}
				//randomly select mom or dad
				else if(rand() % 2)
				{
					AI_Unit* newUnit = copyAIUnit(dPos->second);
					child->addAIUnit(newUnit);
					//remove the unit from the list
					erase = dPos++;
					dUnits.erase(erase);
				}
				else
				{
					AI_Unit* newUnit = copyAIUnit(mPos->second);
					child->addAIUnit(newUnit);
					//remove the unit from the list
					erase = mPos++;
					mUnits.erase(erase);
				}
			}
			if(!mUnits.count(type))
			{
				momOut = true;
			}
			if(!dUnits.count(type))
			{
				dadOut = true;
			}
			count--;
		}
		//Clear out any remaining units from mom and dad of this type
		mUnits.erase(mUnits.lower_bound(type), mUnits.upper_bound(type));
		dUnits.erase(dUnits.lower_bound(type), dUnits.upper_bound(type));
	}
	
	//Give the child ancesteral information
	child->setDad(dad->getBodyID());
	child->setMom(mom->getBodyID());
	child->setGeneration(mom->getGeneration(), dad->getGeneration());

	//initialize the child
	child->initialize();
	//Give the newly created brain to the brain manager to find it a body
	BM->addNewBrain(child);
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	AI_GeneticAlgorithm::bud
/*!	\brief		applies mutation to a single fish to create an offspring
**	\param		parent a pointer to the object to bud
**	\return		none
**	\exception	none
**	\author		John Corpening
**	\date		11/15/2002 3:48:45 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
void AI_GeneticAlgorithm::bud(AI_Module *parent)
{
	//Create the child Module
	AI_Module *child = new AI_Module;
	
	//Get a copy of the AI_Units from the parent
	std::multimap<Action_ID, AI_Unit*> units = parent->getAIUnits();
	
	std::multimap<Action_ID, AI_Unit*>::iterator pos;
	//For each AI_Unit
	for(pos = units.begin(); pos != units.end(); pos++)
	{
		//Roll the dice to determine whether or not to mutate
		if(rollDice(budMutationRate))
		{
			//If mutate then create a new AI_Unit and discard the old
			AI_Unit* newUnit = getRandomAIUnit(pos->second->getAction());
			child->addAIUnit(newUnit);
            delete pos->second;
		}
		else
		{
			//Otherwise place the copy into the offspring's AI_Module
			child->addAIUnit(pos->second);
		}
	}

	//Give the newly created brain to the brain manager to find it a body
	child->setMom(parent->getBodyID());
	child->setDad(parent->getBodyID());
	child->setGeneration(parent->getGeneration(), parent->getGeneration());
	child->initialize();
	BM->addNewBrain(child);
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	AI_GeneticAlgorithm::checkList
/*!	\brief		determines whether an object is on the mating list or not
**	\param		object, the objectID to check
**	\return		true if the object is on the list, otherwise false
**	\exception	none
**	\author		John Corpening
**	\date		11/26/2002 7:37:55 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
bool AI_GeneticAlgorithm::checkList(Object_ID object)
{
	if(matingList.find(object) == matingList.end())
		return false;

	return true;
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	instantiateHigherGoals
/*!	\brief		instantiates all the higher goals for use by the module
**	\return		none
**	\exception	none
**	\author		John Corpening
**	\date		4/17/2003 1:50:53 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
void AI_GeneticAlgorithm::instantiateHigherGoals()
{
	AI_HigherGoal* temp = new HigherGoals::MaxKills;
	if(temp) higherGoals.push_back(temp);

	temp = new HigherGoals::Nomad;
	if(temp) higherGoals.push_back(temp);

	temp = new HigherGoals::DukeWellington;
	if(temp) higherGoals.push_back(temp);

	temp = new HigherGoals::Random;
	if(temp) higherGoals.push_back(temp);

	temp = new HigherGoals::Glutton;
	if(temp) higherGoals.push_back(temp);

	temp = new HigherGoals::PlanktonDensity;
	if(temp) higherGoals.push_back(temp);

	temp = new HigherGoals::BestCase;
	if(temp) higherGoals.push_back(temp);

	temp = new HigherGoals::WorstCase;
	if(temp) higherGoals.push_back(temp);

	temp = new HigherGoals::AverageCase;
	if(temp) higherGoals.push_back(temp);

	temp = new HigherGoals::Herbivore;
	if(temp) higherGoals.push_back(temp);

	temp = new HigherGoals::Carnivore;
	if(temp) higherGoals.push_back(temp);
}