////////////////////////////////////////////////////////////////////////////////
/*!	\file		AI_GeneticAlgorithm,h
**	\brief		handles mating of fish
**	\author		John Corpening
**	\date		11/15/2002 3:26:38 PM
**	\par		DESCRIPTION:
**				Many of the steps involved with a genetic algorithm are taken 
**				care of as part of the nature of the game. Initialization of a 
**				population of hypotheses occurs when the initial AI modules are 
**				created for the initial population of fish. Fish are removed by 
**				their inability to survive in the game environment. The game will
**				run as long as the user wishes to run it, so there is no end 
**				condition, thereby allowing the user to evolve the system as long 
**				as they wish to do so.
**
**				The portion of the genetic algorithm that will require direct 
**				control from the genetic algorithm module (GAM) is checking for
**				"genetic" compatibility, crossover, mutation and generation of 
**				new populations.
**
**				When a fish decides it is time to "mate" it will put itself on 
**				the GAM's mating list. The GAM will then evaluate the fish and 
**				try to find a match for it from other fish on the mating list. 
**				If a match is found, it will then apply crossover and mutation 
**				techniques to generate offspring from the two fish. If no 
**				compatible match is found within a limited number of tries, 
**				the GAM will then create offspring from the single parent using 
**				mutation.
**
**	\par		REMARKS:
**				Physical bonuses are not yet implmented.
*/
////////////////////////////////////////////////////////////////////////////////

#pragma once

namespace AI_System
{
////////////////////////////////////////////////////////////////////////////////
/*!	\class		GA_Tag
**	\brief		used by the Genetic Algorithm to track mating attempts
**	\author		John Corpening
**	\date		11/15/2002 3:33:17 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
struct GA_Tag
{
	AI_Module	*brain;
	char		tries;
};

////////////////////////////////////////////////////////////////////////////////
/*!	\class		AI_GeneticAlgorithm
**	\brief		handles "mating" of game objects
**	\author		John Corpening
**	\date		11/15/2002 3:34:18 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
class AI_GeneticAlgorithm
{
public:
	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:	AI_GeneticAlgorithm::getInstance
	/*!	\brief		global access point for the genetic algorithm
	**	\return		pointer to the genetic algorithm singleton
	**	\exception	creation failure
	**	\author		John Corpening
	**	\date		11/15/2002 3:36:00 PM
	**	\warning	none
	*/
	////////////////////////////////////////////////////////////////////////////////
	static AI_GeneticAlgorithm* getInstance();

	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:	AI_GeneticAlgorithm::setBrainManager
	/*!	\brief		sets the brain manager
	**	\param		bm a pointer to the brain manager
	**	\return		none
	**	\exception	none
	**	\author		John Corpening
	**	\date		11/15/2002 3:37:07 PM
	**	\warning	none
	*/
	////////////////////////////////////////////////////////////////////////////////
	void setBrainManager(AI_BrainManager *bm) {BM = bm;}

	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:	AI_GeneticAlgorithm::addToMatingList
	/*!	\brief		adds a brain to the mating list
	**	\param		brain a pointer to the AI_Module to add
	**	\param		object the ID of the object
	**	\return		none
	**	\exception	none
	**	\author		John Corpening
	**	\date		11/15/2002 3:42:19 PM
	**	\warning	none
	*/
	////////////////////////////////////////////////////////////////////////////////
	void addToMatingList(AI_Module *brain, Object_ID object);
	
	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:	AI_GeneticAlgorithm::generateOffspring
	/*!	\brief		attempts to create offspring from fish on the mating list
	**	\return		none
	**	\exception	none
	**	\author		John Corpening
	**	\date		11/15/2002 3:44:11 PM
	**	\warning	none
	*/
	////////////////////////////////////////////////////////////////////////////////
	void generateOffspring();

	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:	AI_GeneticAlgorithm::checkList
	/*!	\brief		determines whether an object is on the mating list or not
	**	\param		object, the objectID to check
	**	\return		true if the object is on the list, otherwise false
	**	\exception	none
	**	\author		John Corpening
	**	\date		11/26/2002 7:37:55 PM
	**	\warning	none
	*/
	////////////////////////////////////////////////////////////////////////////////
	bool checkList(Object_ID object);

	void clearMatingList(){matingList.clear();}

	float getCompatability(){return compatability;}
	void setCompatability(float val){compatability = val;}
	~AI_GeneticAlgorithm();
private:

	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:	AI_GeneticAlgorithm::findMate
	/*!	\brief		determines whether two fish are compatible or not
	**	\param		mom pointer to one parent's AI_Module
	**	\param		dad pointer to the other parent's AI_Module
	**	\return		none
	**	\exception	none
	**	\author		John Corpening
	**	\date		11/15/2002 3:45:04 PM
	**	\warning	none
	*/
	////////////////////////////////////////////////////////////////////////////////
	bool findMate(AI_Module *mom, AI_Module *dad);

	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:	AI_GeneticAlgorithm::mate
	/*!	\brief		applies crossover to the parents to create an offspring
	**	\param		mom pointer to one parent's AI_Module
	**	\param		dad pointer to the other parent's AI_Module
	**	\return		none
	**	\exception	none
	**	\author		John Corpening
	**	\date		11/15/2002 3:46:24 PM
	**	\warning	none
	*/
	////////////////////////////////////////////////////////////////////////////////
	void mate(AI_Module *mom, AI_Module *dad);

	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:	AI_GeneticAlgorithm::bud
	/*!	\brief		applies mutation to a single fish to create an offspring
	**	\param		parent a pointer to the object to bud
	**	\return		none
	**	\exception	none
	**	\author		John Corpening
	**	\date		11/15/2002 3:48:45 PM
	**	\warning	none
	*/
	////////////////////////////////////////////////////////////////////////////////
	void bud(AI_Module *parent);

	std::vector<AI_HigherGoal*>	goals;		//!< used to determine compatability
	std::map<Object_ID, GA_Tag> matingList;	//!< A nice place to meet
	std::vector<AI_HigherGoal*> higherGoals;//!< used to evaluate mating compatability
	AI_BrainManager				*BM;		//!< handle to the brain manager
	AI_PerceptGenerator			*PG;		//!< handle to the percept generator

	char				maxTries;			//!< Maximum number of mating attempts
	char				mutationRate;		//!< mutation rate for crossover
	char				budMutationRate;	//!< mutation rate for budding
	float				compatability;		//!< Percent of matching higher goals to be compatable

	static AI_GeneticAlgorithm *_instance;//!< Singleton functionality
	//Internal constants
	enum 
	{
		MAXTRIES	= 10,		//!< Default number of maximum attempts to mate
		MUTATION	= 10,		//!< Default mutation rate
		BUDMUTATION = 50,		//!< Default bud mutation rate
	};

	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:	AI_GeneticAlgorithm::rollDice
	/*!	\brief		compares given value to a random number between 0 and 99
	**	\param		chance number to compare against (think %, 70 is 70% chance)
	**	\return		none
	**	\exception	none
	**	\author		John Corpening
	**	\date		11/15/2002 3:50:54 PM
	**	\warning	none
	*/
	////////////////////////////////////////////////////////////////////////////////
	bool rollDice(char chance)
	{
		if(rand() % 100  < chance)
			return true;
		return false;
	}

	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:	instantiateHigherGoals
	/*!	\brief		instantiates all the higher goals for use by the module
	**	\return		none
	**	\exception	none
	**	\author		John Corpening
	**	\date		4/17/2003 1:50:53 PM
	**	\warning	none
	*/
	////////////////////////////////////////////////////////////////////////////////
	void instantiateHigherGoals();

protected:
	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:	AI_GeneticAlgorithm::AI_GeneticAlgorithm
	/*!	\brief		initializes genetic algorithm
	**	\return		none
	**	\exception	none
	**	\author		John Corpening
	**	\date		11/15/2002 3:53:23 PM
	**	\warning	Does not set the Brain Manager
	*/
	////////////////////////////////////////////////////////////////////////////////
	AI_GeneticAlgorithm();
};
}