////////////////////////////////////////////////////////////////////////////////
/*!	\file		AI_Helpers.cpp
**	\brief		The implementation file for AI_Helpers.h
**	\author		John Corpening
**	\date		11/22/2002 11:59:12 AM
*/
////////////////////////////////////////////////////////////////////////////////

#include "AI_System.h"
using namespace AI_System;
using namespace Debugging;

// memory leak detection
#ifdef _DEBUG
    #include"memleak.h"
    #define new DEBUG_NEW
    #undef THIS_FILE
    static char THIS_FILE[] = __FILE__;
#endif

Debugger *debug = Debugger::Instance();

//Array of Technique Types
TT_ID Techniques[] = {TT_NBC, TT_C45, TT_BPNN};

//Array of all Action IDs
Action_ID ActionIDs[] = {MOVETO, FLEE, FOLLOW, MOVEWITH, ROAM, ATTACK};





////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	getRandomAIUnit
/*!	\brief		creates and returns a random AIUnit
**	\return		pointer to an AI_Unit
**	\exception	none
**	\author		John Corpening
**	\date		11/7/2002 7:47:12 PM
**	\warning	none
**	\par		UPDATED ON:			1/28/2003 1:38:17 PM
**	\par		UPDATED BY:			Nate Cleveland
**	\par		UPDATE DESCRIPTION: 
**					Added creation of C4.5 Dtrees into switch statement.
*/
////////////////////////////////////////////////////////////////////////////////
AI_Unit* AI_System::getRandomAIUnit() throw()
{
	//Switch on the types of AI Units that are possible
	switch(Techniques[rand()%(sizeof(Techniques)/sizeof(TT_ID))])
	{
	case TT_C45:	
		debug->output(AI, L2, "getRandomAIUnit(void): C4.5 Decision Tree, I CHOOSE YOU!");
		return new DTree::C45;
	case TT_NBC:
		return new Bayes::BayesUnit;
	case TT_BPNN:
		return new Neural::NeuralUnit(TT_BPNN);
	default:
		debug->output(AI, L2, "getRandomAIUnit(void): Unkown Technique Type, Using Bayes");
		return new Bayes::BayesUnit;
	}
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	getRanomAIUnit
/*!	\brief		creates a random AI Unit given an Action Tpe
**	\param		actionType an Action_ID
**	\return		a pointer to the newly created AI_Unit
**	\exception	none
**	\author		John Corpening
**	\date		11/22/2002 11:56:24 AM
**	\warning	none
**	\par		UPDATED ON:			1/28/2003 1:39:52 PM
**	\par		UPDATED BY:			Nate Cleveland
**	\par		UPDATE DESCRIPTION: 
**					Added creation of C4.5 Dtrees into switch statement. Also found
					 a trival bug where the TT_NBC case didn't have a break causing
					 two Bayes to be instantiated instead of one. 
*/
////////////////////////////////////////////////////////////////////////////////
AI_Unit* AI_System::getRandomAIUnit(Action_ID actionType) throw()
{
	AI_Unit* unit = 0;
	//Switch on the types of AI Units that are possible
	switch(Techniques[rand()%(sizeof(Techniques)/sizeof(TT_ID))])
	{
	case TT_C45:
		debug->output(AI, L2, "getRandomAIUnit: C4.5 Decision Tree, I CHOOSE YOU!");
		unit = new DTree::C45;
		break;
		
	case TT_NBC:
		debug->output(AI, L2, "getRandomAIUnit: Naive Bayes Classifier, I CHOOSE YOU!");
		unit =  new Bayes::BayesUnit;
		break;

	case TT_BPNN:
		debug->output(AI, L2, "getRandomAIUnit: Neural Network, I CHOOSE YOU!");
		unit = new Neural::NeuralUnit(TT_BPNN);
		break;

	default:
		debug->output(AI, L2, "getRandomAIUnit: Unknown Technique Type, Using Bayes");
		unit = new Bayes::BayesUnit;
	}
	unit->setAction(ActionIDs[rand() % (sizeof(ActionIDs)/sizeof(Action_ID))]);
	unit->initialize();
	return unit;
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	copyAIUnit
/*!	\brief		creates a copy of an AI Unit
**	\param		unit, the AI Unit to copy
**	\return		pointer to the newly created AI Unit
**	\exception	none
**	\author		John Corpening
**	\date		11/22/2002 11:56:28 AM
**	\warning	none
**	\par		UPDATED ON:			1/28/2003 1:41:19 PM
**	\par		UPDATED BY:			Nate Cleveland
**	\par		UPDATE DESCRIPTION: 
**					Added Dtrees to switch statement.
*/
////////////////////////////////////////////////////////////////////////////////
AI_Unit* AI_System::copyAIUnit(AI_Unit* unit) throw()
{
	switch(unit->getTechniqueType())
	{
	case TT_C45:
		debug->output(AI, L2, "copyAIUnit: C4.5 Decision Tree, I CHOOSE YOU!");
		return new DTree::C45(*dynamic_cast<DTree::C45*>(unit));
		
	case TT_NBC:
		return new Bayes::BayesUnit(*dynamic_cast<Bayes::BayesUnit*>(unit));
		break;
	case TT_BPNN:
		return new Neural::NeuralUnit(*dynamic_cast<Neural::NeuralUnit*>(unit));
	default:
		debug->output(AI, L2, "copyAIUnit: Unknown Technique Type, Using Bayes.");
		return new Bayes::BayesUnit(*dynamic_cast<Bayes::BayesUnit*>(unit));
	}
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	getRandomADN
/*!	\brief		creates and returns a random Action Decision Node
**	\param		ID Action ID of the ADN to create
**	\return		pointer to an ADN
**	\exception	none
**	\author		John Corpening
**	\date		11/7/2002 7:47:49 PM
**	\warning	none
**	\par		UPDATED ON:			1/28/2003 1:43:35 PM
**	\par		UPDATED BY:			Nate Cleveland
**	\par		UPDATE DESCRIPTION: 
**					Added C4.5 DTrees to the switch statement.
*/
////////////////////////////////////////////////////////////////////////////////
AI_ADN* AI_System::getRandomADN(Action_ID ID) throw()
{
	debug->output(AI, L1, "getRandomADN test setup");
	Action_ID	AC = static_cast<Action_ID>(ID & AC_MASK);
	AI_ADN*		retval = 0;
	
	switch(Techniques[rand()%(sizeof(Techniques)/sizeof(TT_ID))])
	{
	case TT_C45:
		debug->output(AI, L2, "getRandomADN: C4.5 Decision Tree, I CHOOSE YOU!");
		retval = new DTree::C45_ADN;
		break;	
	
	case TT_NBC:
		debug->output(AI, L2, "getRandomADN: Naive Bayes Classifier, I CHOOSE YOU!");
		retval = new Bayes::BayesADN;
		break;

	case TT_BPNN:
		debug->output(AI, L2, "getRandomADN: Neural Network, I CHOOSE YOU!");
		retval = new Neural::NeuralADN(TT_BPNN);
		break;

	default:
		debug->output(AI, L2, "getRandomADN: Unknown Action_ID using Bayes");
		retval = new Bayes::BayesADN;
	}
	retval->setActionClass(AC);
	retval->setHigherGoal(getRandomHG_ID());
//	retval->initialize();
	return retval;
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	getRandomFF_ID
/*!	\brief		used to get a random ID for a fitness function
**	\return		a fitness function ID
**	\exception	none
**	\author		John Corpening
**	\date		10/26/2002 12:30:04 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
FF_ID AI_System::getRandomFF_ID()
{
	return (FF_MAX > 0 ? static_cast<FF_ID>((rand() % FF_MAX) + 1) : FF_NULL);
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	getFitnessFunction
/*!	\brief		given an ID, creates a fitness function a returns a pointer to it
**	\param		id the ID of the fitness function to create
**	\return		a pointer to the newly created fitness function
**	\exception	none
**	\author		John Corpening
**	\date		10/26/2002 12:30:08 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
AI_FitnessFunction* AI_System::getFitnessFunction(FF_ID id)
{
	switch(id)
	{
	case FF_MAX_HEALTH:
		return new FitnessFunctions::MaxHealth;
		break;
	case FF_MAX_STOMACH:
		return new FitnessFunctions::MaxStomach;
		break;
	case FF_MAX_SPEED:
		return new FitnessFunctions::MaxSpeed;
		break;
	case FF_MAX_ENERGY:
		return new FitnessFunctions::MaxEnergyLevel;
		break;
	case FF_MAX_INCLINE:
		return new FitnessFunctions::MaxIncline;
		break;
	case FF_MAX_SPEED_ENERGY:
		return new FitnessFunctions::MaxSpeedEnergy;
		break;
	case FF_MAX_STOMACH_HEALTH:
		return new FitnessFunctions::MaxStomachHealth;
		break;
	case FF_MIN_BEEN_ATTACKED:
		return new FitnessFunctions::MinBeenAttacked;
		break;
	case FF_MAX_PLANKTON:
		return new FitnessFunctions::MaxPlankton;
		break;
	case FF_MAX_DISTANCE:
		return new FitnessFunctions::MaxDistance;
		break;
	case FF_MIN_DISTANCE:
		return new FitnessFunctions::MinDistance;
		break;
	case FF_MAX_CHUM:
		return new FitnessFunctions::MaxChum;
		break;
	case FF_MAX_PLANKTON_DENSITY:
		return new FitnessFunctions::PlanktonDensity;
		break;
	default:
		debug->output(AI, L2, "getFitnessFunction: Unknown FF_ID, using MaxHealth");
		return new FitnessFunctions::MaxHealth;
	}
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	getRandomHG_ID
/*!	\brief		used to get a random ID for a Higher Goal
**	\return		a fitness function ID
**	\exception	none
**	\author		John Corpening
**	\date		10/26/2002 12:30:04 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
HG_ID AI_System::getRandomHG_ID()
{
	return (HG_MAX > 0 ? static_cast<HG_ID>((rand() % HG_MAX) + 1) : HG_NULL);
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	getHigherGoal
/*!	\brief		given an ID, creates a Higher Goal a returns a pointer to it
**	\param		id the ID of the Higher Goal to create
**	\return		a pointer to the newly created Higher Goal
**	\exception	none
**	\author		John Corpening
**	\date		10/26/2002 12:30:08 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
AI_HigherGoal* AI_System::getHigherGoal(HG_ID id)
{
	switch(id)
	{
	case HG_KILLS:
		return new HigherGoals::MaxKills;
	case HG_NOMAD:
		return new HigherGoals::Nomad;
	case HG_DUKE_WELLINGTON:	
		return new HigherGoals::DukeWellington;
	case HG_RANDOM:			
		return new HigherGoals::Random;
	case HG_GLUTTON:
		return new HigherGoals::Glutton;
	case HG_PLANKTON_DENSITY:
		return new HigherGoals::PlanktonDensity;
	case HG_BEST_CASE:
		return new HigherGoals::BestCase;
	case HG_WORST_CASE:
		return new HigherGoals::WorstCase;
	case HG_AVERAGE_CASE:
		return new HigherGoals::AverageCase;
	case HG_HERBIVORE:	
		return new HigherGoals::Herbivore;
	case HG_CARNIVORE:
		return new HigherGoals::Carnivore;


	default:
		debug->output(AI, L2, "getHigherGoal: Unkown HG_ID, Using Nomad");
		return new HigherGoals::Nomad;
	}
}

const char* AI_System::getActionString(Action_ID action)
{
	switch(action)
	{
	case AC_MOVE:
		return "AC_MOVE";
	case AC_MOUTH:
		return "AC_MOUTH";
	case AC_MARKER:
		return "AC_MARKER";
	case MOVETO:
		return "MOVETO";
	case FLEE:
		return "FLEE";
	case FOLLOW:
		return "FOLLOW";
	case MOVEWITH:
		return "MOVEWITH";
	case ROAM:
		return "ROAM";
	case GOTO_POS_MARKER:
		return "GOTO_POS_MARKER";
	case FLEE_NEG_MARKER:
		return "FLEE_NEG_MARKER";
	case IGNORE_MARKERS:
		return "IGNORE_MARKERS";
	case HEED_MARKERS:
		return "HEED_MARKERS";	
	case ATTACK:
		return "ATTACK";
	case EAT:
		return "EAT";
	case PICKUP:
		return "PICKUP";
	case DROP:
		return "DROP";
	case DROP_POS_MARKER:
		return "DROP_POS_MARKER";
	case DROP_NEG_MARKER:
		return "DROP_NEG_MARKER";
	case EVALUATEMARKER:
		return "EVALUATEMARKER";
	default:
		return "UNKOWN ACTION TYPE";
	}
}

const char* AI_System::getDataFilterString(DF_ID df_id)
{
	switch(df_id)
	{
//	case DF_INVALID:
//		return "DF_INVALID";
	case DF_NULL:
		return "DF_NULL";
	case DF_METABOLISM:
		return "DF_METABOLISM";
	case DF_ARMOR:
		return "DF_ARMOR";
	case DF_EYES:
		return "DF_EYES";

	case DF_SIZE:
		return "DF_SIZE";
	case DF_TEETH_BALEEN:
		return "DF_TEETH_BALEEN";
	case DF_REACTIONTIME:
		return "DF_REACTIONTIME";
	case DF_MAXSPEED:
		return "DF_MAXSPEED";
	case DF_BECUM:
		return "DF_BECUM";

	case DF_LIFESPAN:
		return "DF_LIFESPAN";
	case DF_ATTACKSTRENGTH:
		return "DF_ATTACKSTRENGTH";
	case DF_SUCKABILITY:
		return "DF_SUCKABILITY";
	case DF_ENERGYLEVEL_OBJECT:
		return "DF_ENERGYLEVEL_OBJECT";
	case DF_ENERGYLEVEL_TARGET:
		return "DF_ENERGYLEVEL_TARGET";

	case DF_STOMACH_OBJECT:
		return "DF_STOMACH_OBJECT";
	case DF_STOMACH_TARGET:
		return "DF_STOMACH_TARGET";
	case DF_HEALTH_OBJECT:
		return "DF_HEALTH_OBJECT";
	case DF_HEALTH_TARGET:
		return "DF_HEALTH_TARGET";
	case DF_MATINGINCLINE_OBJECT:
		return "DF_MATINGINCLINE_OBJECT";

	case DF_MATINGINCLINE_TARGET:
		return "DF_MATINGINCLINE_TARGET";
	case DF_HEALTH_NORMALIZED_OBJECT:
		return "DF_HEALTH_NORMALIZED_OBJECT";
	case DF_HEALTH_NORMALIZED_TARGET:
		return "DF_HEALTH_NORMALIZED_TARGET";
	case DF_STOMACH_NORMALIZED_OBJECT:
		return "DF_STOMACH_NORMALIZED_OBJECT";
	case DF_STOMACH_NORMALIZED_TARGET:
		return "DF_STOMACH_NORMALIZED_TARGET";

	case DF_ENERGY_NORMALIZED_OBJECT:
		return "DF_ENERGY_NORMALIZED_OBJECT";
	case DF_ENERGY_NORMALIZED_TARGET:
		return "DF_ENERGY_NORMALIZED_TARGET";
	case DF_AGE:
		return "DF_AGE";
	case DF_DISTANCE:
		return "DF_DISTANCE";
	case DF_OBJECTSPEED:		
		return "DF_OBJECTSPEED";

	case DF_TARGETSPEED:
		return "DF_TARGETSPEED";
	case DF_OBJECTTYPE:
		return "DF_OBJECTTYPE";
	case DF_PLANKTONDENSITY_OBJECT:
		return "DF_PLANKTONDENSITY_OBJECT";
	case DF_PLANKTONDENSITY_TARGET:
		return "DF_PLANKTONDENSITY_TARGET";
	case DF_SPEED_RATE_OBJECT:
		return "DF_SPEED_RATE_OBJECT";

	case DF_SPEED_RATE_TARGET:
		return "DF_SPEED_RATE_TARGET";
	case DF_SPEED_COUNT_OBJECT:
		return "DF_SPEED_COUNT_OBJECT";
	case DF_SPEED_COUNT_TARGET:
		return "DF_SPEED_COUNT_TARGET";
	case DF_DISTANCE_RATE_OBJECT:
		return "DF_DISTANCE_RATE_OBJECT";
	case DF_DISTANCE_RATE_TARGET:
		return "DF_DISTANCE_RATE_TARGET";

	case DF_DISTANCE_COUNT_OBJECT:
		return "DF_DISTANCE_COUNT_OBJECT";
	case DF_DISTANCE_COUNT_TARGET:
		return "DF_DISTANCE_COUNT_TARGET";
	case DF_HEALTH_RATE_OBJECT:
		return "DF_HEALTH_RATE_OBJECT";
	case DF_HEALTH_RATE_TARGET:
		return "DF_HEALTH_RATE_TARGET";
	case DF_HEALTH_COUNT_OBJECT:
		return "DF_HEALTH_COUNT_OBJECT";

	case DF_HEALTH_COUNT_TARGET:
		return "DF_HEALTH_COUNT_TARGET";
	case DF_INCLINE_RATE_OBJECT:
		return "DF_INCLINE_RATE_OBJECT";
	case DF_INCLINE_RATE_TARGET:
		return "DF_INCLINE_RATE_TARGET";
	case DF_INCLINE_COUNT_OBJECT:
		return "DF_INCLINE_COUNT_OBJECT";
	case DF_INCLINE_COUNT_TARGET:
		return "DF_INCLINE_COUNT_TARGET";

	case DF_STOMACH_RATE_OBJECT:
		return "DF_STOMACH_RATE_OBJECT";
	case DF_STOMACH_RATE_TARGET:
		return "DF_STOMACH_RATE_TARGET";
	case DF_STOMACH_COUNT_OBJECT:
		return "DF_STOMACH_COUNT_OBJECT";
	case DF_STOMACH_COUNT_TARGET:
		return "DF_STOMACH_COUNT_TARGET";
	case DF_ENERGY_RATE_OBJECT:
		return "DF_ENERGY_RATE_OBJECT";

	case DF_ENERGY_RATE_TARGET:
		return "DF_ENERGY_RATE_TARGET";
	case DF_ENERGY_COUNT_OBJECT:
		return "DF_ENERGY_COUNT_OBJECT";
	case DF_ENERGY_COUNT_TARGET:
		return "DF_ENERGY_COUNT_TARGET";
	case DF_KILLS_RATE_OBJECT:
		return "DF_KILLS_RATE_OBJECT";
	case DF_KILLS_RATE_TARGET:
		return "DF_KILLS_RATE_TARGET";

	case DF_KILLS_COUNT_OBJECT:
		return "DF_KILLS_COUNT_OBJECT";
	case DF_KILLS_COUNT_TARGET:
		return "DF_KILLS_COUNT_TARGET";
	case DF_OFFSPRING_RATE_OBJECT:
		return "DF_OFFSPRING_RATE_OBJECT";
	case DF_OFFSPRING_RATE_TARGET:
		return "DF_OFFSPRING_RATE_TARGET";
	case DF_OFFSPRING_COUNT_OBJECT:
		return "DF_OFFSPRING_COUNT_OBJECT";

	case DF_OFFSPRING_COUNT_TARGET:
		return "DF_OFFSPRING_COUNT_TARGET";
	case DF_BEEN_ATTACKED_RATE_OBJECT:
		return "DF_BEEN_ATTACKED_RATE_OBJECT";
	case DF_BEEN_ATTACKED_RATE_TARGET:
		return "DF_BEEN_ATTACKED_RATE_TARGET";
	case DF_BEEN_ATTACKED_COUNT_OBJECT:
		return "DF_BEEN_ATTACKED_COUNT_OBJECT";
	case DF_BEEN_ATTACKED_COUNT_TARGET:
		return "DF_BEEN_ATTACKED_COUNT_TARGET";


	case DF_HAS_ATTACKED_RATE_OBJECT:
		return "DF_HAS_ATTACKED_RATE_OBJECT";
	case DF_HAS_ATTACKED_RATE_TARGET:
		return "DF_HAS_ATTACKED_RATE_TARGET";
	case DF_HAS_ATTACKED_COUNT_OBJECT:
		return "DF_HAS_ATTACKED_COUNT_OBJECT";
	case DF_HAS_ATTACKED_COUNT_TARGET:
		return "DF_HAS_ATTACKED_COUNT_TARGET";
	case DF_PLANKTON_RATE_OBJECT:
		return "DF_PLANKTON_RATE_OBJECT";

	case DF_PLANKTON_RATE_TARGET:
		return "DF_PLANKTON_RATE_TARGET";
	case DF_PLANKTON_COUNT_OBJECT:
		return "DF_PLANKTON_COUNT_OBJECT";
	case DF_PLANKTON_COUNT_TARGET:
		return "DF_PLANKTON_COUNT_TARGET";
	case DF_CHUM_RATE_OBJECT:
		return "DF_CHUM_RATE_OBJECT";
	case DF_CHUM_RATE_TARGET:
		return "DF_CHUM_RATE_TARGET";

	case DF_CHUM_COUNT_OBJECT:
		return "DF_CHUM_COUNT_OBJECT";
	case DF_CHUM_COUNT_TARGET:
		return "DF_CHUM_COUNT_TARGET";
	case DF_TOTAL_FOOD_RATE_OBJECT:
		return "DF_TOTAL_FOOD_RATE_OBJECT";
	case DF_TOTAL_FOOD_RATE_TARGET:
		return "DF_TOTAL_FOOD_RATE_TARGET";
	case DF_TOTAL_FOOD_COUNT_OBJECT:
		return "DF_TOTAL_FOOD_COUNT_OBJECT";

	case DF_TOTAL_FOOD_COUNT_TARGET:
		return "DF_TOTAL_FOOD_COUNT_TARGET";
	default:
		return "No string available for data filter";
	}
}

const char* AI_System::getFitnessFunctionString(FF_ID ff_id)
{
	switch(ff_id)
	{
	case FF_MAX_HEALTH:
		return "FF_MAX_HEALTH";
	case FF_MAX_STOMACH:
		return "FF_MAX_STOMACH";
	case FF_MAX_SPEED:
		return "FF_MAX_SPEED";
	default:
		return "No string available for fitness function";
	}
}

const char* AI_System::getHigherGoalString(HG_ID hg_id)
{
/*	switch(hg_id)
	{
	case HG_KILLS:
		return "HG_KILLS";
	case HG_NOMAD:
		return "HG_NOMAD";
	default:
		return "No string available for higher goal";
	}
*/
	switch(hg_id)
	{
		case HG_KILLS:
			return "Kills";
		case HG_NOMAD:
			return "Nomad";
		case HG_DUKE_WELLINGTON:	
			return "The Duke of Wellington";
		case HG_RANDOM:			
			return "Random";
		case HG_GLUTTON:
			return "Glutton";
		case HG_PLANKTON_DENSITY:
			return "Plankton Density";
		case HG_BEST_CASE:
			return "Best Case";
		case HG_WORST_CASE:
			return "Worst Case";
		case HG_AVERAGE_CASE:
			return "Average Case";
		case HG_HERBIVORE:	
			return "Herbivore";
		case HG_CARNIVORE:
			return "Carnivore";
		default:
			return "Not recognized";
	}
}