////////////////////////////////////////////////////////////////////////////////
/*!	\file		AI_HigherGoalBase.h
**	\brief		Base class for fitness functions
**	\author		John Corpening
**	\date		10/18/2002 4:58:33 PM
**	\par		DESCRIPTION:
**				Fitness functions provide a means for AI units to train 
**				by providing feedback to the various machine learning techniques
**				as to how they are doing.
**
**				Fitness functions should be written such that they consider a 
**				single aspect of the game object to maximize.
**
**				Fitness functions work in almost the same way as data filters. 
**				They evaluate data available to the percept generator and produce
**				an output based on this data. The difference is that fitness 
**				functions output either a positive or negative as well as an 
**				emphasis that is a real number in the range 0-1. The AI technique
**				may choose whether to use the fitness, the response or both for 
**				training.
**
**	\par		REMARKS:
**				This is an abstract class. Don't try to instantiate.
**
**	\par		Tips for writing Fitness Functions:
**				- Information is gathered via the Percept Generator's data filters
**				- You can use a combination of data filters for the fitness function
**				- You must return a fitness in the range of 0-1
**				- You must return a response
**					- A positive response indicates that an action should have been triggered
**					- A negative response indicates that an action should not have been triggered
**				- It is up to you to determine what the response threshold is.
**				- Each AI unit will have it's own Fitness Function so persistant data is OK
**					- Persistant data will not be saved, if the program ends it is lost
**					- This is subject to change
*/
////////////////////////////////////////////////////////////////////////////////

#pragma once

namespace AI_System
{
class AI_HigherGoal
{
protected:
	bool				dataProcessed;	//!< Indicates that the fitness function has already processed the data.
	float				fitness;		//!< Fitness generated by the fitness function
	bool				response;		//!< Tells the AI unit whether it is doing good or not
	float				emphasis;		//!< Value that determines what importance an associated action is.
	HG_ID				ID;				//!< An id number used to identify the fitness function
	AI_PerceptGenerator	*PG;			//!< Used to get the current target

public:

	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:	AI_HigherGoal::AI_HigherGoal
	/*!	\brief		sets the message handler and percept generator pointers
	**	\return		none
	**	\exception	Nexception for NULL pointers returned
	**	\author		John Corpening
	**	\date		10/26/2002 12:58:07 PM
	**	\warning	none
	*/
	////////////////////////////////////////////////////////////////////////////////
	AI_HigherGoal();

	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:	AI_HigherGoal::evaluateFitness
	/*!	\brief		Virtual function that evaluates the fitness of a fish
	**	\param		
	**	\return		Returns the response
	**	\exception	none
	**	\author		John Corpening
	**	\date		10/18/2002 5:05:16 PM
	**	\warning	Derived classes should not forget to set the emphasis and fitness!
	**	\par		Description:
	**					-Only does calculations if the dataProcessed flag is false
	**					-Queries necessary data from the game
	**					-performs necessary calculations with the data
	**					-Sets the fitness, emphasis and response
	*/
	////////////////////////////////////////////////////////////////////////////////
	virtual bool evaluateFitness() = 0;

	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:	AI_HigherGoal::clear()
	/*!	\brief		clears the dataProcessed flag
	**	\return		none
	**	\exception	none
	**	\author		John Corpening
	**	\date		10/18/2002 5:33:50 PM
	**	\warning	none
	**	\par		Description:
	**					By clearing the dataProcessed flag,
	**					a re-evaluation will occur whenever a
	**					request for data is made.
	*/
	////////////////////////////////////////////////////////////////////////////////
	void clear() throw() {dataProcessed = false;}

	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:	AI_HigherGoal::getEmphasis
	/*!	\brief		returns the emphasis associated with an action
	**	\return		the emphasis associated with an action
	**	\exception	none
	**	\author		John Corpening
	**	\date		10/18/2002 5:33:56 PM
	**	\warning	The emphasis was formerly known as the priority
	**	\par		Description:
	**					The emphasis is a value placed on an action
	**				specified by an AI Unit to indicate how important
	**				the AI Unit's decision is.
	**	\par		Remarks:
	**					If the data processed flag is not set then
	**				this function calls the evaluateFitness function.
	*/
	////////////////////////////////////////////////////////////////////////////////
	float getEmphasis() throw();

	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:	AI_HigherGoal::getFitness
	/*!	\brief		returns the fitness evaluated by the fitness function
	**	\return		the fitness evaluated by the fitness function
	**	\exception	none
	**	\author		John Corpening
	**	\date		10/18/2002 5:34:02 PM
	**	\warning	none
	**	\par		Remarks:
	**					If the data processed flag is not set then
	**				this function calls the evaluateFitness function.
	*/
	////////////////////////////////////////////////////////////////////////////////
	float getFitness() throw();

	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:	AI_HigherGoal::getResponse
	/*!	\brief		returns the response 
	**	\return		whether the fitness function "feels" good or not
	**	\exception	none
	**	\author		John Corpening
	**	\date		10/18/2002 5:34:14 PM
	**	\warning	none
	**	\par		Description:
	**					The response is whether or not the fitness function
	**				feels things are going well or not. It will be used
	**				to create training examples.
	**	\par		Remarks:
	**					If the data processed flag is not set then
	**				this function calls the evaluateFitness function.
	*/
	////////////////////////////////////////////////////////////////////////////////
	bool getResponse() throw();
};
}