////////////////////////////////////////////////////////////////////////////////
/*!	\file		AI_Module.cpp
**	\brief		Implementation file for AI_Module.h
**	\author		John Corpening
**	\date		11/7/2002 7:46:43 PM
*/
////////////////////////////////////////////////////////////////////////////////

#include"AI_System.h"
using namespace AI_System;
using namespace Debugging;

// memory leak detection
#ifdef _DEBUG
    #include"memleak.h"
    #define new DEBUG_NEW
    #undef THIS_FILE
    static char THIS_FILE[] = __FILE__;
#endif
unsigned int AI_Module::maxUnits = 20;
unsigned int AI_Module::minUnits = 1;

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	AI_Module::AI_Module
/*!	\brief		Creates a random AI_Module
**	\return		none
**	\exception	exceptions thrown for creation failure
**	\author		John Corpening
**	\date		10/28/2002 8:35:22 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
AI_Module::AI_Module() : generation(0), mom(0), dad(0)
{
	PG = AI_PerceptGenerator::instance();
	if(!PG)
	{
		throw(Nexception("Percept Generator return NULL Pointer", __FILE__,__LINE__,__DATE__,__TIME__));
	}
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	AI_Module::addAIUnit
/*!	\brief		adds an AI Unit to the AI_Module
**	\param		unit a pointer to an AI_Unit
**	\return		none
**	\exception	none
**	\author		John Corpening
**	\date		10/28/2002 8:37:18 PM
**	\warning	none
**	\par		Description:
**				-Checks ADN's for a match with the AI_Unit's action type
**				-Creates a new ADN if no match is found
*/
////////////////////////////////////////////////////////////////////////////////
bool AI_Module::addAIUnit(AI_Unit* unit)
{
	//If we have an ADN for it 
	int size = ADNs.size();
	for(int j = 0; j < size; j++)
	{
		if((unit->getAction() & ADNs[j]->getActionType()))
		{
			//then associate it with the ADN
			ADNs[j]->addAIUnit(unit);
			
			//and add it to the module's unit list
			units.insert(std::make_pair(unit->getID(), unit));
			//We have found an ADN and added it to the ADN and the Module 
			//So We're done
			return true;
		}
	}

	//If we got to here then there was no ADN that matched
	//so we have to create a new one
	AI_ADN *ADN = getRandomADN(unit->getAction());
	ADN->addAIUnit(unit);
	ADNs.push_back(ADN);
	units.insert(std::make_pair(unit->getID(), unit));
	return false;
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	AI_Module::initialize
/*!	\brief		initializes all ADN's
**	\exception	none
**	\author		John Corpening
**	\date		4/17/2003 1:55:20 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
void AI_Module::initialize() throw()
{
	for(unsigned int i = 0; i < ADNs.size(); i++)
	{
		ADNs[i]->initialize();
	}
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	AI_Module::generateDecisions
/*!	\brief		Generates a list of decisions for the critters
**	\param		Target, the object that is thinking
**	\return		Returns list of actions
**	\exception	none
**	\author		John Corpening
**	\date		10/19/2002 2:25:17 PM
**	\warning	none
**	\par		Description:
**				-runs for each object in sensor range
**				-Queries the fish for a list of actions completed
**				-If an action has been completed, it will tell that 
**					AI unit to save a training example.
**				-Calls each AI unit's determineAction function
**				-Calls each action decision node's determineAction function
**				-Generates a list of actions
*/
////////////////////////////////////////////////////////////////////////////////
std::vector<AI_Action> AI_Module::generateDecisions(Object_ID Target)
{
	//Clear out the Percept Generator
	PG->reset();

	//Tell the percept generator to get the sensed Object for the current object
	target = Target;
	PG->setTarget(target);
	PG->getSensedObjects();

	//Get the previous Actions performed and generate Training Examples
	std::vector<AI_ActionTag> performed;
	PG->getPerformedActions(performed);
//	if(performed.size())
	{
		int size = performed.size();
		Debugger::Instance()->output(AI, L2, "Receiving returned actions: %d", size);
		AI_ActionTag tag;
		for(int i = 0; i < size; i++)
		{
			tag = performed[i];
			//Find the AI Unit Responsible
			std::map<AI_ID,AI_Unit*>::iterator it = units.find(tag.unit_ID);
			if(it != units.end())
			{
//				Debugger::Instance()->output(AI, L3, "Fish: %d, AI Unit: %d, Action: %s", target, it->second->getID(), getActionString(it->second->getAction()));
				it->second->createTrainingExample(tag);
			}
			//regardless, all ADN's get to create a training example
			for(unsigned int j = 0; j < ADNs.size(); j++)
			{
				ADNs[j]->createTrainingExample(tag);
			}
		}
	}

	//Call the getInput and DetermineAction method of each Action Decision Node
	int size = ADNs.size();

	std::vector<AI_Action> actions;

	//Run the AI_System
	do
	{
		for(int i = 0; i < size; i++)
		{
			ADNs[i]->getInput();
			actions.push_back(ADNs[i]->determineAction());
		}

		//The pass has ended so notify the percept generator
		PG->endPass();
	}
	while(PG->nextObject());

	return actions;
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	AI_Module::train
/*!	\brief		Allows the AI units that are ready to train to train.
**	\param		
**	\return		none
**	\exception	none
**	\author		John Corpening
**	\date		10/19/2002 2:50:19 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
void AI_Module::train() throw()
{
	//Train ADN's
	int size = ADNs.size();
	for(int i = 0; i < size; i++)
	{
		ADNs[i]->train();
	}
	//Train AI_Units
	std::map<AI_ID, AI_Unit*>::iterator pos;
	for(pos = units.begin(); pos != units.end(); pos++)
	{
		pos->second->train();
	}
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	AI_Module::getAIUnits
/*!	\brief		returns a copy of all the AI_Units in the system
**	\param		actionClass type of AI_Unit to get
**	\return		pointer to a copy of the unit to get
**	\exception	none
**	\author		John Corpening
**	\date		10/19/2002 2:50:42 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
std::multimap<Action_ID, AI_Unit*> AI_Module::getAIUnits() throw()
{
	std::multimap<Action_ID, AI_Unit*> retval;
	std::map<AI_ID, AI_Unit*>::iterator pos;

	for(pos = units.begin(); pos != units.end(); pos++)
	{
		Action_ID	temp = static_cast<Action_ID>(pos->second->getAction() & AC_MASK);
		retval.insert(std::make_pair(temp, copyAIUnit(pos->second)));
	}
	return retval;
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	AI_Module::getGoal
/*!	\brief		returns the goal of the ADN associated with the given action class
**	\param		type, the Action class of the ADN
**	\return		the higher goal id of the ADN
**	\exception	none
**	\author		John Corpening
**	\date		4/17/2003 1:58:07 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
HG_ID AI_Module::getGoal(Action_ID type) const throw()
{
	for(unsigned int i = 0; i < ADNs.size(); i++)
	{
		if(ADNs[i]->getActionType() == type)
		{
			return ADNs[i]->getHigherGoalID();
		}
	}
	return HG_NULL;
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	AI_Module::Write
/*!	\brief		saves the AI_Module
**	\param		file, the ostream to save to
**	\return		none
**	\exception	none
**	\author		John Corpening
**	\date		4/17/2003 1:59:46 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
void AI_Module::Write(std::ostream& file) const
{
	//Save the object ID
	file << "#Begin Module" << endl;
	file << "Target: " << target << endl;

	//Save the number of ADN's
	file << "ADNs: " << ADNs.size() << endl;

	//Save all the ADN's
	file << "Techniques: " << endl;
	for(unsigned int i = 0; i < ADNs.size(); i++)
	{
		//Save the Technique type
		file << "Type: " << ADNs[i]->getTechniqueType() << endl;
		//Save the ADN
		ADNs[i]->AI_ADN::Write(file);
		ADNs[i]->Write(file);
	}

	//Save the number of AI Units
	file << "AIUnits: " << units.size() << endl;

	//Save all the AI Units
	file << "Techniques: " << endl;
	std::map<AI_ID, AI_Unit*>::const_iterator pos;
	for(pos = units.begin(); pos != units.end(); pos++)
	{
		//Save the AI_Unit Type
		file << "Type: " << pos->second->getTechniqueType() << endl;
		//Save the AI_Unit
		pos->second->AI_Unit::Write(file);
		pos->second->Write(file);
	}
	file << "#End Module" << endl;
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	AI_Module::Read
/*!	\brief		loads the AI module from the a file
**	\param		file, the istream to read from
**	\return		none
**	\exception	"AI_Module Save Error: bad technique" nexception
**	\author		John Corpening
**	\date		4/17/2003 2:02:01 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
void AI_Module::Read(std::istream& file)
{
	const unsigned int buffSize = 512;
	char buff[buffSize] = {0};

	//skip, should just be the header.
	file.getline(buff, buffSize);
	
	//Load the object ID, 
	file.getline(buff, buffSize);
	sscanf(buff, "Target: %d", &target);

	//Load number of ADNs
	int numADNs;
	file.getline(buff, buffSize);
	sscanf(buff, "ADNs: %d", &numADNs);

	//Load all the ADN's, also skip over the header line
	file.getline(buff, buffSize);
	for(int j = 0; j < numADNs; j++)
	{
		//Load the type
		unsigned int type;
		file.getline(buff, buffSize);
		sscanf(buff,"Type: %d",  &type);

		//Create an ADN
		AI_ADN* ADN = 0;
		switch((TT_ID)type)
		{
		case TT_C45:
			ADN = new DTree::C45_ADN;
			break;	
		
		case TT_NBC:
			ADN = new Bayes::BayesADN;
			break;

		case TT_BPNN:
			ADN = new Neural::NeuralADN(TT_BPNN);
			break;
		default:
			throw(Nexception("AI_Module Save Error: bad technique", __FILE__,__LINE__,__DATE__,__TIME__));
		}
		

		//Have the ADN read from the file
		ADN->AI_ADN::Read(file);
		ADN->Read(file);

		//Add the ADN to the ADN vector
		ADNs.push_back(ADN);
	}

	//Load the number of AI Units
	int numUnits;
	file.getline(buff, buffSize);
	sscanf(buff, "AIUnits: %d", &numUnits);

	//Load all the AI_Units, skip over the header
	file.getline(buff, buffSize);
	for(int i = 0; i < numUnits; i++)
	{
		//Load the technique
		unsigned int type;
		file.getline(buff, buffSize);
		sscanf(buff, "Type: %d", &type);

		//Create an AI_Unit
		AI_Unit* unit = 0;
		switch((TT_ID)type)
		{
		case TT_C45:	
			unit = new DTree::C45;
			break;
		case TT_NBC:
			unit =  new Bayes::BayesUnit;
			break;
		case TT_BPNN:
			unit = new Neural::NeuralUnit(TT_BPNN);
			break;
		default:
			throw(Nexception("AI_Module Save Error: bad technique", __FILE__,__LINE__,__DATE__,__TIME__));
		}
		

		//Have the AI Unit read from the file
		unit->AI_Unit::Read(file);
		unit->Read(file);


		//Set the AI_Units to the ADN's
		if(!addAIUnit(unit))
		{
			throw(Nexception("AI_Module Save Error: No appropriate ADN", __FILE__,__LINE__,__DATE__,__TIME__));
		}
	}

	//init everybody
	for(u32 k=0; k<ADNs.size() ;k++)
		ADNs[k]->lockADN();

	//jump over the footer.
	file.getline(buff, buffSize);
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	AI_Module::~AI_Module
/*!	\brief		destructor
**	\exception	none
**	\author		John Corpening
**	\date		4/17/2003 1:53:52 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
AI_Module::~AI_Module()
{
	//Delete the ADNs
	for(unsigned int i = 0; i < ADNs.size(); i++)
	{
		delete ADNs[i];
	}

	//Delete the AI Units
	std::map<AI_ID, AI_Unit*>::iterator	it;
	for(it = units.begin(); it != units.end(); it++)
	{
		delete it->second;
	}
}
