////////////////////////////////////////////////////////////////////////////////
/*!	\file		AI_Module.h
**	\brief		A class to make decisions based on game data
**	\author		John Corpening
**	\date		10/19/2002 2:18:19 PM
**	\par		DESCRIPTION:
**				The AI Module is the actual "brain" for an object in the
**				game. It consists of AI units, ADN's and a handle to
**				the percept generator. When given an object ID it will
**				generate decisions for that object.
**	\par		REMARKS:
**				getPerformedActions() will be implemented once the system is
**				integrated.
**				mutate was removed since the Genetic Algorithm module can take care of it
**				reset was removed since the Brain Manager can take care of that
*/
////////////////////////////////////////////////////////////////////////////////
#pragma once

namespace AI_System
{

////////////////////////////////////////////////////////////////////////////////
/*!	\class		AI_Module
**	\brief		a class that represents a critter's brain
**	\author		John Corpening
**	\date		10/19/2002 2:19:24 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
class AI_Module : public FileMemento
{
	AI_PerceptGenerator			*PG;		//!< A perceptGenerator
	std::map<AI_ID, AI_Unit*>	units;		//!< List of AI units
	std::vector<AI_ADN*>		ADNs;		//!< List of decisionNodes
	Object_ID					target;		//!< Current object
	Object_ID					mom;		//!< One fish parent
	Object_ID					dad;		//!< the other fish parent
	unsigned int				generation;	//!< The generation of the current fish
	Object_ID					bodyID;		//!< The ID of the body associated with this fish
	static unsigned int			maxUnits;	//!< Max Total AI_Units
	static unsigned int			minUnits;	//!< Min Total AI_Units


public:
	
	//basic accessors
	void setBodyID(Object_ID ID){bodyID = ID;}
	Object_ID getBodyID(){return bodyID;}
	Object_ID getMom(){return mom;}
	Object_ID getDad(){return dad;}
	void setMom(Object_ID Mom){mom = Mom;}
	void setDad(Object_ID Dad){dad = Dad;}
	void setGeneration(unsigned int gen1, unsigned int gen2){generation = 1+(gen1 < gen2 ? gen2 : gen1);}
	unsigned int getGeneration(){return generation;}

	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:	AI_Module::AI_Module
	/*!	\brief		Creates a random AI_Module
	**	\return		none
	**	\exception	exceptions thrown for creation failure
	**	\author		John Corpening
	**	\date		10/28/2002 8:35:22 PM
	**	\warning	none
	*/
	////////////////////////////////////////////////////////////////////////////////
	AI_Module();

	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:	AI_Module::~AI_Module
	/*!	\brief		destructor
	**	\exception	none
	**	\author		John Corpening
	**	\date		4/17/2003 1:53:52 PM
	**	\warning	none
	*/
	////////////////////////////////////////////////////////////////////////////////
	~AI_Module();


	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:	AI_Module::addAIUnit
	/*!	\brief		adds an AI Unit to the AI_Module
	**	\param		unit a pointer to an AI_Unit
	**	\return		none
	**	\exception	none
	**	\author		John Corpening
	**	\date		10/28/2002 8:37:18 PM
	**	\warning	none
	**	\par		Description:
	**				-Checks ADN's for a match with the AI_Unit's action type
	**				-Creates a new ADN if no match is found
	*/
	////////////////////////////////////////////////////////////////////////////////
	bool addAIUnit(AI_Unit* unit);

	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:	AI_Module::initialize
	/*!	\brief		initializes all ADN's
	**	\exception	none
	**	\author		John Corpening
	**	\date		4/17/2003 1:55:20 PM
	**	\warning	none
	*/
	////////////////////////////////////////////////////////////////////////////////
	void initialize() throw();

	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:	AI_Module::generateDecisions
	/*!	\brief		Generates a list of decisions for the critters
	**	\param		Target, the object that is thinking
	**	\return		Returns list of actions
	**	\exception	none
	**	\author		John Corpening
	**	\date		10/19/2002 2:25:17 PM
	**	\warning	none
	**	\par		Description:
	**				-runs for each object in sensor range
	**				-Queries the fish for a list of actions completed
	**				-If an action has been completed, it will tell that 
	**					AI unit to save a training example.
	**				-Calls each AI unit's determineAction function
	**				-Calls each action decision node's determineAction function
	**				-Generates a list of actions
	*/
	////////////////////////////////////////////////////////////////////////////////
	std::vector<AI_Action> generateDecisions(Object_ID Target) throw();

	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:	AI_Module::determinePerformedActions
	/*!	\brief		Gets actions performed by the fish
	**	\return		none
	**	\exception	none
	**	\author		John Corpening
	**	\date		10/19/2002 2:49:12 PM
	**	\warning	none
	**	\par		Description:
	**				Determines which AI unit performed the action and tells 
	**				that AI unit to save a training example.
	*/
	////////////////////////////////////////////////////////////////////////////////
	void getPerformedActions();


	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:	AI_Module::train
	/*!	\brief		Allows the AI units that are ready to train to train.
	**	\param		
	**	\return		none
	**	\exception	none
	**	\author		John Corpening
	**	\date		10/19/2002 2:50:19 PM
	**	\warning	none
	*/
	////////////////////////////////////////////////////////////////////////////////
	void train() throw();

	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:	AI_Module::save
	/*!	\brief		Saves the AI module to a file
	**	\param		
	**	\return		none
	**	\exception	none
	**	\author		John Corpening
	**	\date		10/19/2002 2:50:32 PM
	**	\warning	Subject to change pending file I/O considerations
	*/
	////////////////////////////////////////////////////////////////////////////////
	void save() throw();

	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:	AI_Module::load
	/*!	\brief		Loads the AI module from a file
	**	\param		
	**	\return		none
	**	\exception	none
	**	\author		John Corpening
	**	\date		10/19/2002 2:50:37 PM
	**	\warning	none
	*/
	////////////////////////////////////////////////////////////////////////////////
	void load() throw();

	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:	AI_Module::getAIUnit
	/*!	\brief		returns a copy of all the AI_Units in the system
	**	\param		actionClass type of AI_Unit to get
	**	\return		a multi-map of the AI_Units sorted by Action class
	**	\exception	none
	**	\author		John Corpening
	**	\date		10/19/2002 2:50:42 PM
	**	\warning	none
	*/
	////////////////////////////////////////////////////////////////////////////////
	std::multimap<Action_ID, AI_Unit*> getAIUnits() throw();

	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:	AI_Module::getTarget
	/*!	\brief		returns the current AI_Module's target
	**	\return		the Object_ID of the current target
	**	\exception	none
	**	\author		John Corpening
	**	\date		10/28/2002 9:27:59 AM
	**	\warning	none
	*/
	////////////////////////////////////////////////////////////////////////////////
	Object_ID	getTarget() throw() {return target;}

	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:	AI_Module::getGoal
	/*!	\brief		returns the goal of the ADN associated with the given action class
	**	\param		type, the Action class of the ADN
	**	\return		the higher goal id of the ADN
	**	\exception	none
	**	\author		John Corpening
	**	\date		4/17/2003 1:58:07 PM
	**	\warning	none
	*/
	////////////////////////////////////////////////////////////////////////////////
	HG_ID getGoal(Action_ID type) const throw();

	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:	AI_Module::Write
	/*!	\brief		saves the AI_Module
	**	\param		file, the ostream to save to
	**	\return		none
	**	\exception	none
	**	\author		John Corpening
	**	\date		4/17/2003 1:59:46 PM
	**	\warning	none
	*/
	////////////////////////////////////////////////////////////////////////////////
	void Write(std::ostream&)const;

	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:	AI_Module::Read
	/*!	\brief		loads the AI module from the a file
	**	\param		file, the istream to read from
	**	\return		none
	**	\exception	"AI_Module Save Error: bad technique" nexception
	**	\author		John Corpening
	**	\date		4/17/2003 2:02:01 PM
	**	\warning	none
	*/
	////////////////////////////////////////////////////////////////////////////////
	void Read(std::istream&);

};
}