////////////////////////////////////////////////////////////////////////////////
/*!	\file		AI_PerceptGenerator.cpp
**	\brief		The implementation file for AI_PerceptGenerator.h
**	\author		John Corpening
**	\date		10/28/2002 12:05:55 PM
*/
////////////////////////////////////////////////////////////////////////////////
#include"AI_System.h"
using namespace AI_System;

// memory leak detection
#ifdef _DEBUG
    #include"memleak.h"
    #define new DEBUG_NEW
    #undef THIS_FILE
    static char THIS_FILE[] = __FILE__;
#endif

//Initialization of singleton instance
AI_PerceptGenerator* AI_PerceptGenerator::_instance = 0;

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	AI_PerceptGenerator::instance
/*!	\brief		Singleton access point
**	\return		none
**	\exception	none
**	\author		John Corpening
**	\date		10/24/2002 4:30:56 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
AI_PerceptGenerator* AI_PerceptGenerator::instance() throw()
{
	if(!_instance)
		_instance = new AI_PerceptGenerator;
	return _instance;
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	AI_PerceptGenerator::AI_PerceptGenerator
/*!	\brief		Constructor, sets MessageHandler pointer
**	\return		none
**	\exception	Null MessageHandler pointer
**	\author		John Corpening
**	\date		10/28/2002 12:06:55 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
AI_PerceptGenerator::AI_PerceptGenerator()
{
	//Get a handle to the message handler
	MH = MsgHandler::MessageHandler::Instance();
	sensedObjects.clear();

	//Check
	if(!MH)
	{
		throw(Nexception("PerceptGenerator: MessageHandler instance failed", __FILE__,__LINE__,__DATE__,__TIME__));
	}

/*	//This is for debuggin purposes
	//Add DataFilters
	AI_DataFilter *temp = new AI_DFTest_A;
	temp->setPerceptGenerator(this);
	DataFilters.insert(std::make_pair(DF_TEST_A, temp));

	temp = new AI_DFTest_B;
	temp->setPerceptGenerator(this);
	DataFilters.insert(std::make_pair(DF_TEST_B, temp));
*/
	//Create all the data filters
	instantiateDataFilters();
}

AI_PerceptGenerator::~AI_PerceptGenerator()
{
	std::map<DF_ID, AI_DataFilter*>::iterator pos;
	for(pos = DataFilters.begin(); pos != DataFilters.end(); pos++)
	{
		delete pos->second;
	}
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	AI_PerceptGenerator::reset	
/*!	\brief		Purges sensory data and resets all data filters
**	\return		none
**	\exception	none
**	\author		John Corpening
**	\date		10/21/2002 12:20:49 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
void AI_PerceptGenerator::reset() throw()
{
	//Reset DataFilters
	std::map<DF_ID, AI_DataFilter*>::iterator i;
	for(i = DataFilters.begin(); i != DataFilters.end(); i++)
	{
		i->second->reset();
	}

	//Reset the target to the invalid key
	target = 0;

	//Reset the oIndex and currentObject
	oIndex = 0;
	currentObject = 0;

	//delete the sensed objects
	sensedObjects.clear();
/*
	if(sensedObjects.size())
	{
		delete sensedObjects;
		sensedObjects = 0;
	}
*/
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	AI_PerceptGenerator::endPass
/*!	\brief		Informs data filters that the current AI module pass has ended
**	\return		none
**	\exception	none
**	\author		John Corpening
**	\date		10/21/2002 12:22:58 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
void AI_PerceptGenerator::endPass() throw()
{
	std::map<DF_ID, AI_DataFilter*>::iterator i;
	for(i = DataFilters.begin(); i != DataFilters.end(); i++)
	{
		i->second->endPass();
	}
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	AI_PerceptGenerator::getRandomDF
/*!	\brief		returns a random DataFilter ID for the AI_Units
**	\return		a random DataFilter ID
**	\exception	none
**	\author		John Corpening
**	\date		10/25/2002 11:42:05 AM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
DF_ID AI_PerceptGenerator::getRandomDF() throw()
{
	return (DF_MAX > 0 ? static_cast<DF_ID>((rand() % DF_MAX) + 1) : DF_NULL);
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	AI_PerceptGenerator::getSensedObjects
/*!	\brief		retreives a list of objects in sensor range of the thinker
**	\param		thinker the ID of the thinker
**	\return		none
**	\exception	none
**	\author		John Corpening
**	\date		10/24/2002 4:26:46 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
bool AI_PerceptGenerator::getSensedObjects() throw()
{
	if(!target)
	{
		return false;
	}
	//Get objects that the target has sensed here
	sensedObjects.clear(); 	
	MH->GetSensedObject(target, sensedObjects);
//	if(!sensedObjects)
//		return false;

	//Set the current object to the first object sensed
	if(sensedObjects.size())
		currentObject= sensedObjects[0];

	return true;
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	AI_PerceptGenerator::nextObject()
/*!	\brief		Advances the currentObject to the next in the list
**	\return		false if the list is empty
**	\exception	none
**	\author		John Corpening
**	\date		10/24/2002 4:28:56 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
bool AI_PerceptGenerator::nextObject() throw()
{
	//If there was no target then there were no sensed objects
	if(!target)
		return false;

	//intentional assignment: end of array flag is NULL
	oIndex++;
	if(oIndex >= sensedObjects.size())
		return false;

	currentObject = sensedObjects[oIndex];
	return true;
}

void AI_PerceptGenerator::getPerformedActions(std::vector<AI_ActionTag>& ret)
{
	MH->GetPerformedActions(target, ret);
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	AI_PerceptGenerator::requestData
/*!	\brief		given a Data Filter key, returns an attribute
**	\param		key a data filter ID
**	\return		none
**	\exception	none
**	\author		John Corpening
**	\date		10/21/2002 12:18:11 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
AI_Attribute AI_PerceptGenerator::requestData(DF_ID key) throw()
{
	std::map<DF_ID, AI_DataFilter*>::iterator pos = DataFilters.find(key);
	if(pos != DataFilters.end())
	{
		return pos->second->getAttribute();
	}
	else
	{
		throw(Nexception("Can't find DataFilter", __FILE__,__LINE__,__DATE__,__TIME__));
	}
}

void AI_PerceptGenerator::instantiateDataFilters()
{
	AI_DataFilter *temp = new DataFilter::Age;
	if(temp) DataFilters.insert(std::make_pair(temp->getID(), temp));

	temp = new DataFilter::Armor;
	if(temp) DataFilters.insert(std::make_pair(temp->getID(), temp));

	temp = new DataFilter::AttackStrength;
	if(temp) DataFilters.insert(std::make_pair(temp->getID(), temp));

	temp = new DataFilter::Becum;
	if(temp) DataFilters.insert(std::make_pair(temp->getID(), temp));

	temp = new DataFilter::Been_AttackedCountObject;
	if(temp) DataFilters.insert(std::make_pair(temp->getID(), temp));

	temp = new DataFilter::Been_AttackedCountTarget;
	if(temp) DataFilters.insert(std::make_pair(temp->getID(), temp));

	temp = new DataFilter::Been_AttackedRateObject;
	if(temp) DataFilters.insert(std::make_pair(temp->getID(), temp));

	temp = new DataFilter::Been_AttackedRateTarget;
	if(temp) DataFilters.insert(std::make_pair(temp->getID(), temp));

	temp = new DataFilter::ChumCountObject;
	if(temp) DataFilters.insert(std::make_pair(temp->getID(), temp));

	temp = new DataFilter::ChumCountTarget;
	if(temp) DataFilters.insert(std::make_pair(temp->getID(), temp));

	temp = new DataFilter::ChumRateObject;
	if(temp) DataFilters.insert(std::make_pair(temp->getID(), temp));

	temp = new DataFilter::ChumRateTarget;
	if(temp) DataFilters.insert(std::make_pair(temp->getID(), temp));

	temp = new DataFilter::Distance;
	if(temp) DataFilters.insert(std::make_pair(temp->getID(), temp));

	temp = new DataFilter::DistanceCountObject;
	if(temp) DataFilters.insert(std::make_pair(temp->getID(), temp));

	temp = new DataFilter::DistanceCountTarget;
	if(temp) DataFilters.insert(std::make_pair(temp->getID(), temp));

	temp = new DataFilter::DistanceRateObject;
	if(temp) DataFilters.insert(std::make_pair(temp->getID(), temp));

	temp = new DataFilter::DistanceRateTarget;
	if(temp) DataFilters.insert(std::make_pair(temp->getID(), temp));

	temp = new DataFilter::EnergyCountObject;
	if(temp) DataFilters.insert(std::make_pair(temp->getID(), temp));

	temp = new DataFilter::EnergyCountTarget;
	if(temp) DataFilters.insert(std::make_pair(temp->getID(), temp));

	temp = new DataFilter::EnergyLevelObject;
	if(temp) DataFilters.insert(std::make_pair(temp->getID(), temp));

	temp = new DataFilter::EnergyLevelTarget;
	if(temp) DataFilters.insert(std::make_pair(temp->getID(), temp));

	temp = new DataFilter::EnergyNormalizedObject;
	if(temp) DataFilters.insert(std::make_pair(temp->getID(), temp));

	temp = new DataFilter::EnergyNormalizedTarget;
	if(temp) DataFilters.insert(std::make_pair(temp->getID(), temp));

	temp = new DataFilter::EnergyRateObject;
	if(temp) DataFilters.insert(std::make_pair(temp->getID(), temp));

	temp = new DataFilter::EnergyRateTarget;
	if(temp) DataFilters.insert(std::make_pair(temp->getID(), temp));

	temp = new DataFilter::Eyes;
	if(temp) DataFilters.insert(std::make_pair(temp->getID(), temp));

	temp = new DataFilter::Has_AttackedCountObject;
	if(temp) DataFilters.insert(std::make_pair(temp->getID(), temp));

	temp = new DataFilter::Has_AttackedCountTarget;
	if(temp) DataFilters.insert(std::make_pair(temp->getID(), temp));

	temp = new DataFilter::Has_AttackedRateObject;
	if(temp) DataFilters.insert(std::make_pair(temp->getID(), temp));

	temp = new DataFilter::Has_AttackedRateTarget;
	if(temp) DataFilters.insert(std::make_pair(temp->getID(), temp));

	temp = new DataFilter::HealthCountObject;
	if(temp) DataFilters.insert(std::make_pair(temp->getID(), temp));

	temp = new DataFilter::HealthCountTarget;
	if(temp) DataFilters.insert(std::make_pair(temp->getID(), temp));

	temp = new DataFilter::HealthNormalizedObject;
	if(temp) DataFilters.insert(std::make_pair(temp->getID(), temp));

	temp = new DataFilter::HealthNormalizedTarget;
	if(temp) DataFilters.insert(std::make_pair(temp->getID(), temp));

	temp = new DataFilter::HealthObject;
	if(temp) DataFilters.insert(std::make_pair(temp->getID(), temp));

	temp = new DataFilter::HealthRateObject;
	if(temp) DataFilters.insert(std::make_pair(temp->getID(), temp));

	temp = new DataFilter::HealthRateTarget;
	if(temp) DataFilters.insert(std::make_pair(temp->getID(), temp));

	temp = new DataFilter::HealthTarget;
	if(temp) DataFilters.insert(std::make_pair(temp->getID(), temp));

	temp = new DataFilter::InclineCountObject;
	if(temp) DataFilters.insert(std::make_pair(temp->getID(), temp));

	temp = new DataFilter::InclineCountTarget;
	if(temp) DataFilters.insert(std::make_pair(temp->getID(), temp));

	temp = new DataFilter::InclineRateObject;
	if(temp) DataFilters.insert(std::make_pair(temp->getID(), temp));

	temp = new DataFilter::InclineRateTarget;
	if(temp) DataFilters.insert(std::make_pair(temp->getID(), temp));

	temp = new DataFilter::KillsCountObject;
	if(temp) DataFilters.insert(std::make_pair(temp->getID(), temp));

	temp = new DataFilter::KillsCountTarget;
	if(temp) DataFilters.insert(std::make_pair(temp->getID(), temp));

	temp = new DataFilter::KillsRateObject;
	if(temp) DataFilters.insert(std::make_pair(temp->getID(), temp));

	temp = new DataFilter::KillsRateTarget;
	if(temp) DataFilters.insert(std::make_pair(temp->getID(), temp));

	temp = new DataFilter::LifeSpan;
	if(temp) DataFilters.insert(std::make_pair(temp->getID(), temp));

	temp = new DataFilter::MatingInclineObject;
	if(temp) DataFilters.insert(std::make_pair(temp->getID(), temp));

	temp = new DataFilter::MatingInclineTarget;
	if(temp) DataFilters.insert(std::make_pair(temp->getID(), temp));

	temp = new DataFilter::MaxSpeed;
	if(temp) DataFilters.insert(std::make_pair(temp->getID(), temp));

	temp = new DataFilter::Metabolism;
	if(temp) DataFilters.insert(std::make_pair(temp->getID(), temp));

	temp = new DataFilter::ObjectSpeed;
	if(temp) DataFilters.insert(std::make_pair(temp->getID(), temp));

	temp = new DataFilter::ObjectType;
	if(temp) DataFilters.insert(std::make_pair(temp->getID(), temp));

	temp = new DataFilter::OffspringCountObject;
	if(temp) DataFilters.insert(std::make_pair(temp->getID(), temp));

	temp = new DataFilter::OffspringCountTarget;
	if(temp) DataFilters.insert(std::make_pair(temp->getID(), temp));

	temp = new DataFilter::OffspringRateObject;
	if(temp) DataFilters.insert(std::make_pair(temp->getID(), temp));

	temp = new DataFilter::OffspringRateTarget;
	if(temp) DataFilters.insert(std::make_pair(temp->getID(), temp));

	temp = new DataFilter::PlanktonCountObject;
	if(temp) DataFilters.insert(std::make_pair(temp->getID(), temp));

	temp = new DataFilter::PlanktonCountTarget;
	if(temp) DataFilters.insert(std::make_pair(temp->getID(), temp));

	temp = new DataFilter::PlanktonDensityObject;
	if(temp) DataFilters.insert(std::make_pair(temp->getID(), temp));

	temp = new DataFilter::PlanktonDensityTarget;
	if(temp) DataFilters.insert(std::make_pair(temp->getID(), temp));

	temp = new DataFilter::PlanktonRateObject;
	if(temp) DataFilters.insert(std::make_pair(temp->getID(), temp));

	temp = new DataFilter::PlanktonRateTarget;
	if(temp) DataFilters.insert(std::make_pair(temp->getID(), temp));

	temp = new DataFilter::ReactionTime;
	if(temp) DataFilters.insert(std::make_pair(temp->getID(), temp));

	temp = new DataFilter::Size;
	if(temp) DataFilters.insert(std::make_pair(temp->getID(), temp));

	temp = new DataFilter::SpeedCountObject;
	if(temp) DataFilters.insert(std::make_pair(temp->getID(), temp));

	temp = new DataFilter::SpeedCountTarget;
	if(temp) DataFilters.insert(std::make_pair(temp->getID(), temp));

	temp = new DataFilter::SpeedRateObject;
	if(temp) DataFilters.insert(std::make_pair(temp->getID(), temp));

	temp = new DataFilter::SpeedRateTarget;
	if(temp) DataFilters.insert(std::make_pair(temp->getID(), temp));

	temp = new DataFilter::StomachCountObject;
	if(temp) DataFilters.insert(std::make_pair(temp->getID(), temp));

	temp = new DataFilter::StomachCountTarget;
	if(temp) DataFilters.insert(std::make_pair(temp->getID(), temp));

	temp = new DataFilter::StomachNormalizedObject;
	if(temp) DataFilters.insert(std::make_pair(temp->getID(), temp));

	temp = new DataFilter::StomachNormalizedTarget;
	if(temp) DataFilters.insert(std::make_pair(temp->getID(), temp));

	temp = new DataFilter::StomachObject;
	if(temp) DataFilters.insert(std::make_pair(temp->getID(), temp));

	temp = new DataFilter::StomachRateObject;
	if(temp) DataFilters.insert(std::make_pair(temp->getID(), temp));

	temp = new DataFilter::StomachRateTarget;
	if(temp) DataFilters.insert(std::make_pair(temp->getID(), temp));

	temp = new DataFilter::StomachTarget;
	if(temp) DataFilters.insert(std::make_pair(temp->getID(), temp));

	temp = new DataFilter::Suckability;
	if(temp) DataFilters.insert(std::make_pair(temp->getID(), temp));

	temp = new DataFilter::TargetSpeed;
	if(temp) DataFilters.insert(std::make_pair(temp->getID(), temp));

	temp = new DataFilter::Teeth_Baleen;
	if(temp) DataFilters.insert(std::make_pair(temp->getID(), temp));

	temp = new DataFilter::Total_FoodCountObject;
	if(temp) DataFilters.insert(std::make_pair(temp->getID(), temp));

	temp = new DataFilter::Total_FoodCountTarget;
	if(temp) DataFilters.insert(std::make_pair(temp->getID(), temp));

	temp = new DataFilter::Total_FoodRateObject;
	if(temp) DataFilters.insert(std::make_pair(temp->getID(), temp));

	temp = new DataFilter::Total_FoodRateTarget;
	if(temp) DataFilters.insert(std::make_pair(temp->getID(), temp));

	//Set the percept generator of each of these to this
	std::map<DF_ID, AI_DataFilter*>::iterator pos;
	for(pos = DataFilters.begin(); pos != DataFilters.end(); pos++)
	{
		pos->second->setPerceptGenerator(this);
	}
}
