////////////////////////////////////////////////////////////////////////////////
/*!	\file		AI_PerceptGenerator.h
**	\brief		A class to contain DataFilters and Fitness Functions
**	\author		John Corpening
**	\date		10/21/2002 11:20:55 AM
**	\par		DESCRIPTION:
**				The percept generator is the interface between the game world and
**				the decision-making processes. It contains a number of data filters
**				that query objects within the game to gather data. These data 
**				filters produce the attributes that are used by the AI units. 
**				
**				The percept generator will reset the data filters at the end of 
**				processing and notify the data filters at the end of each pass of
**				the AI module.
*/
////////////////////////////////////////////////////////////////////////////////
#pragma once

namespace AI_System
{

////////////////////////////////////////////////////////////////////////////////
/*!	\class		AI_PerceptGenerator
**	\brief		A class that provides access to game data
**	\author		John Corpening
**	\date		10/28/2002 11:58:20 AM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
class AI_PerceptGenerator
{
public:
	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:	AI_PerceptGenerator::instance
	/*!	\brief		Singleton access point
	**	\return		none
	**	\exception	none
	**	\author		John Corpening
	**	\date		10/24/2002 4:30:56 PM
	**	\warning	none
	*/
	////////////////////////////////////////////////////////////////////////////////
	static AI_PerceptGenerator*	instance();

	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:	AI_PerceptGenerator::requestData
	/*!	\brief		given a Data Filter key, returns an attribute
	**	\param		key a data filter ID
	**	\return		none
	**	\exception	none
	**	\author		John Corpening
	**	\date		10/21/2002 12:18:11 PM
	**	\warning	none
	*/
	////////////////////////////////////////////////////////////////////////////////
	AI_Attribute requestData(DF_ID key) throw();

	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:	AI_PerceptGenerator::reset	
	/*!	\brief		Purges sensory data and resets all data filters
	**	\return		none
	**	\exception	none
	**	\author		John Corpening
	**	\date		10/21/2002 12:20:49 PM
	**	\warning	none
	*/
	////////////////////////////////////////////////////////////////////////////////
	void reset() throw();

	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:	AI_PerceptGenerator::endPass
	/*!	\brief		Informs data filters that the current AI module pass has ended
	**	\return		none
	**	\exception	none
	**	\author		John Corpening
	**	\date		10/21/2002 12:22:58 PM
	**	\warning	none
	*/
	////////////////////////////////////////////////////////////////////////////////
	void endPass() throw();

	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:	AI_PerceptGenerator::getSensedObjects
	/*!	\brief		retreives a list of objects in sensor range of the thinker
	**	\param		thinker the ID of the thinker
	**	\return		none
	**	\exception	none
	**	\author		John Corpening
	**	\date		10/24/2002 4:26:46 PM
	**	\warning	none
	*/
	////////////////////////////////////////////////////////////////////////////////
	bool getSensedObjects() throw();

	void AI_PerceptGenerator::getPerformedActions(std::vector<AI_ActionTag>& ret);

	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:	AI_PerceptGenerator::setTarget
	/*!	\brief		sets the target of the percept generator
	**	\param		Target the ID of the target to set
	**	\return		none
	**	\exception	none
	**	\author		John Corpening
	**	\date		10/28/2002 11:57:23 AM
	**	\warning	none
	*/
	////////////////////////////////////////////////////////////////////////////////
	void setTarget(Object_ID Target) throw() {target = Target;}

	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:	AI_PerceptGenerator::getTarget
	/*!	\brief		returns the Percept Generator's current target
	**	\return		none
	**	\exception	none
	**	\author		John Corpening
	**	\date		10/28/2002 11:57:28 AM
	**	\warning	none
	*/
	////////////////////////////////////////////////////////////////////////////////
	Object_ID getTarget() const throw() {return target;}

	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:	AI_PerceptGenerator::nextObject()
	/*!	\brief		Advances the currentObject to the next in the list
	**	\return		false if the list is empty
	**	\exception	none
	**	\author		John Corpening
	**	\date		10/24/2002 4:28:56 PM
	**	\warning	none
	*/
	////////////////////////////////////////////////////////////////////////////////
	bool nextObject() throw();

	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:	AI_PerceptGenerator::getRandomDF
	/*!	\brief		returns a random DataFilter ID for the AI_Units
	**	\return		a random DataFilter ID
	**	\exception	none
	**	\author		John Corpening
	**	\date		10/25/2002 11:42:05 AM
	**	\warning	none
	*/
	////////////////////////////////////////////////////////////////////////////////
	DF_ID	getRandomDF() throw();

	Object_ID getCurrentObject() {return currentObject;}

	~AI_PerceptGenerator();

protected:

	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:	AI_PerceptGenerator::AI_PerceptGenerator
	/*!	\brief		Constructor, sets MessageHandler pointer
	**	\return		none
	**	\exception	Null MessageHandler pointer
	**	\author		John Corpening
	**	\date		10/28/2002 12:06:55 PM
	**	\warning	none
	*/
	////////////////////////////////////////////////////////////////////////////////
	AI_PerceptGenerator();

private:
	std::map<DF_ID, AI_DataFilter*>	DataFilters;	//!< Map of all data filters
	static AI_PerceptGenerator*		_instance;		//!< Singleton Functionality

	MsgHandler::MessageHandler	*MH;				//!< To gain access to message Handler
	std::vector<Object_ID>		sensedObjects;		//!< List of objects in fish's sensor range
	Object_ID					currentObject;		//!< The current object to be evaluated
	Object_ID					target;				//!< The object that we are retrieving data from
	unsigned int				oIndex;				//!< Index of the currently evaluated object

	void instantiateDataFilters();

};
}