////////////////////////////////////////////////////////////////////////////////
/*!	\file		BayesClassifierOpt.cpp
**	\brief		Implementation
**	\author		Nathan Frost
**	\date		11/7/2002 8:40:24 PM
**	\par		DESCRIPTION:
**					
**	\par		REMARKS:
					- Action_ID MUST be 32bit integral datatypes.
**					
*/
////////////////////////////////////////////////////////////////////////////////
#include"BayesClassifier.h"
//#include <iostream>
#include<algorithm>
#include<limits>
#include<cassert>

using namespace Bayes;

// memory leak detection
#ifdef _DEBUG
    #include"memleak.h"
    #define new DEBUG_NEW
    #undef THIS_FILE
    static char THIS_FILE[] = __FILE__;
#endif


////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	Category::Category
/*!	\brief		Constructor
**	\param		LowerBound		The lowest floating point value an Attribute can have while still
								being considered in this classification.
	\param		UpperBound		The highest floating point value (minus EPSILON) an Attribute can 
								have while still being considered in this classification.
	\param		NumOccurrences	The number of times this classification has been encountered in
								past experiences.  PASSING A NONZERO VALUE HERE IS UNDEFINED.
**	\return		none
**	\exception	none
**	\par		DESCRIPTION:
**					
**	\author		Nathan Frost
**	\date		11/7/2002 8:40:53 PM
**	\warning	See parameter list.
*/
////////////////////////////////////////////////////////////////////////////////
BayesClassifier::Category::Category(float LowerBound, float UpperBound,
									unsigned int NumOccurrences):
lowerBound(LowerBound), upperBound(UpperBound), numOccurrences(NumOccurrences)
{
	;
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	DataFilterKind::DataFilterKind
/*!	\brief		Constructor
**	\param		ID	Unique identifier for the DataFilterKind
**	\return		none
**	\exception	none
**	\par		DESCRIPTION:
**					
**	\author		Nathan Frost
**	\date		11/7/2002 8:44:16 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
BayesClassifier::DataFilterKind::DataFilterKind(unsigned int ID):
id(ID)
{
	;
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	Choice::Choice
/*!	\brief		Constructor
**	\param		ID	Unique identifier for the Choice
**	\param		NumOccurrenceChoice:	The number of times this Choice has been deemed correct in
										past experiences.  PASSING A NONZERO VALUE HERE IS 
										UNDEFINED.
**	\return		none
**	\exception	none
**	\par		DESCRIPTION:
**					
**	\author		Nathan Frost
**	\date		11/7/2002 8:44:42 PM
**	\warning	See parameter list.
*/
////////////////////////////////////////////////////////////////////////////////
BayesClassifier::Choice::Choice(Action_ID ID, unsigned int NumOccurrenceChoice):
id(ID), numOccurrenceChoice(NumOccurrenceChoice)
{
	;
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	BayesClassifier
/*!	\brief		Constructor -- the BayesClassifier begins its life in an invalid state
**	\param		none
**	\return		none
**	\exception	none
**	\par		DESCRIPTION:
**					
**	\author		Nathan Frost
**	\date		11/7/2002 8:46:13 PM
**	\warning	See parameter list.
*/
////////////////////////////////////////////////////////////////////////////////
BayesClassifier::BayesClassifier():
numTrainingExamples(0){}


////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	AddChoice
/*!	\brief		Allows the Bayesian Classifier to consider the specified Choice, in addition to 
				whatever other Choices it can already consider.
**	\param		ID				Unique identifier of the Action the Choice corresponds to (in other
								words, the Action that will be recommended if this Choice is chosen)
	\param		NumOccurrences	The number of times this Choice has been deemed correct in past
								experiences.  PASSING A NONZERO VALUE HERE IS UNDEFINED.
**	\return		none
**	\exception	none
**	\par		DESCRIPTION:
**					
**	\author		Nathan Frost
**	\date		11/7/2002 8:50:55 PM
**	\warning	See parameter list.
*/
////////////////////////////////////////////////////////////////////////////////
void BayesClassifier::AddChoice(Action_ID ID, unsigned int NumOccurrences)
{
	choice.push_back(Choice(ID, NumOccurrences));
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	FindChoice
/*!	\brief		Finds the specified Choice
**	\param		It	Iterator you would like to have point to the specified Choice
	\param		ID	Of the Choice to get It to point to
**	\return		none
**	\exception	none
**	\par		DESCRIPTION:
**					
**	\author		Nathan Frost
**	\date		4/1/2003 6:54:56 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
void BayesClassifier::FindChoice(std::vector<BayesClassifier::Choice>::iterator &It, Action_ID ID)
{
	for(It = choice.begin(); It != choice.end(); ++It)
	{
		if(It->id == ID)
			break;
	}
	assert(It != choice.end());
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	RemoveChoice
/*!	\brief		Disallows the Bayesian Classifier to consider the specified Choice.
**	\param		ID	Unique identifier of the Action the Choice corresponds to (in other words, the
					Bayesian Classifier will no longer be able to recommend this Action).  THE 
					SPECIFIED CHOICE MUST EXIST IN THE BAYESIAN CLASSIFIER.
**	\return		none
**	\exception	none
**	\par		DESCRIPTION:
**					
**	\author		Nathan Frost
**	\date		11/7/2002 8:54:11 PM
**	\warning	See parameter list.
*/
////////////////////////////////////////////////////////////////////////////////
void BayesClassifier::RemoveChoice(Action_ID ID)
{
	std::vector<BayesClassifier::Choice>::iterator it;
	FindChoice(it, ID);
	choice.erase(it);
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	AddDataFilterKindToChoice
/*!	\brief		Allows the Bayesian Classifier to, when considering whether or not to choose a
				Choice, consider the Attribute that corresponds to the specified DataFilterKind
**	\param		ChoiceID			The Choice to associate with the specified DataFilterKind.  THE
									SPECIFIED CHOICE MUST EXIST IN THE BAYESIAN CLASSIFIER.
	\param		DataFilterKindID	The Attribute that the Bayesian Classifier should start 
									considering when contemplating the specified Choice
**	\return		none
**	\exception	none
**	\par		DESCRIPTION:
**					
**	\author		Nathan Frost
**	\date		11/7/2002 9:20:50 PM
**	\warning	See parameter list.
*/
////////////////////////////////////////////////////////////////////////////////
void BayesClassifier::AddDataFilterKindToChoice(Action_ID ChoiceID, unsigned int DataFilterKindID)
{
	std::vector<Choice>::iterator it;
	FindChoice(it, ChoiceID);
	it->dataFilterKind.push_back(DataFilterKind(DataFilterKindID));
}
//TODO: Keep going from here
////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	FindDataFilterKind
/*!	\brief		
**	\param		ChoiceIt		Iterator you would like to have point to the specified Choice 
	\param		DataFilterIt	Iterator you would like to have point to the specified 
								DataFilterKind
	\param		ChoiceID		The ID of the Choice in which the DataFilterKind resides.  THE 
								SPECIFIED CHOICE MUST EXIST IN THE BAYESIAN CLASSIFIER.
	\param		DataFilterID	The ID of the DataFilter to find.
**	\return		true if the DataFilterKind DataFilterID exists in Choice ChoiceID, false if it does
				not
**	\exception	none
**	\par		DESCRIPTION:
**					
**	\author		Nathan Frost
**	\date		4/1/2003 7:01:24 PM
**	\warning	See parameter list.
*/
////////////////////////////////////////////////////////////////////////////////
bool BayesClassifier::FindDataFilterKind(std::vector<Choice>::iterator &ChoiceIt, 
										 std::vector<DataFilterKind>::iterator &DataFilterIt, 
										 Action_ID ChoiceID, unsigned int DataFilterID)
{
	//find the choice the requested DataFilterKind resides in
	FindChoice(ChoiceIt, ChoiceID);

	//find the requested DataFilterKind
	std::vector<DataFilterKind> &dataFilterKind = ChoiceIt->dataFilterKind;
	for(DataFilterIt = dataFilterKind.begin(); DataFilterIt != dataFilterKind.end(); ++DataFilterIt)
	{
		if(DataFilterIt->id == DataFilterID)
			break;
	}

	if(DataFilterIt == dataFilterKind.end())
		return false;//DataFilterKind wasn't found
	else
		return true;//DataFilterKind was found
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	RemoveDataFilterKindFromChoice
/*!	\brief		Disallows the Bayesian Classifier to, when considering whether or not to choose a
				Choice, consider the Attribute that corresponds to the specified DataFilterKind
**	\param		ChoiceID:			The Choice to disassociate with the specified DataFilterKind
	\param		DataFilterKindID:	The Attribute that the Bayesian Classifier should stop 
									considering when contemplating the specified Choice	
**	\return		none
**	\exception	Throws a Nexception if either the specified Choice object or the specified  
				DataFilterKind object does not exist in the Bayesian Classifier
**	\par		DESCRIPTION:
**					
**	\author		Nathan Frost
**	\date		11/7/2002 9:25:40 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
void BayesClassifier::RemoveDataFilterKindFromChoice(	Action_ID ChoiceID, 
														unsigned int DataFilterKindID)
{
	std::vector<Choice>::iterator choiceIt;
	std::vector<DataFilterKind>::iterator dataFilterIt;
	FindDataFilterKind(choiceIt, dataFilterIt, ChoiceID, DataFilterKindID);
	choiceIt->dataFilterKind.erase(dataFilterIt);
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	AddOrReplaceCategoryToDataFilterKindInChoice
/*!	\brief		Defines a classification for the specified DataFilterKind in the specified Choice.
**	\param		ChoiceID:			The Choice that the specified DataFilterKind resides in
	\param		DataFilterKindID:	The Attribute that is getting a new classification
**	\param		LowerBound:			The lowest floating point value an Attribute can have while 
									still being considered in this classification
	\param		UpperBound:			The highest floating point value an Attribute can have while 
									still being considered in this classification
**	\return		none
**	\exception	
				- Throws a Nexception if either the specified Choice object or the specified  
				DataFilterKind object does not exist in the Bayesian Classifier.
				- Throws a Nexception if the specified DataFilterKind's range (see the class
				definition of DataFilterKind for details) becomes invalid as a result of the 
				calling of this function
**	\par		DESCRIPTION:
**					
**	\author		Nathan Frost
**	\date		11/7/2002 9:28:28 PM
**	\warning	
				- Note that if you pass an a Category that has the same lowerBound as an already
				existing Category, but the argument Category has a different upperBound than the
				pre-existing Category, the pre-existing Category will be overwritten by the
				argument Category.  For example, if (0.5, 1.0) already exists in DataFilterKind 4
				of Choice 3, and you AddOrReplaceCategoryToDataFilterKindInChoice(3, 4, 0.5, 1.1),
				then (0.5, 1.0) will be replaced by (0.5, 1.1).
*/
////////////////////////////////////////////////////////////////////////////////
void BayesClassifier::AddOrReplaceCategoryToDataFilterKindInChoice(	Action_ID ChoiceID, 
																	unsigned int DataFilterKindID, 
																	float LowerBound, 
																	float UpperBound)
{
	//find the datafilter we want to add a category to
	std::vector<Choice>::iterator choiceIt;
	std::vector<DataFilterKind>::iterator dataFilterIt;
	FindDataFilterKind(choiceIt, dataFilterIt, ChoiceID, DataFilterKindID);

	//add the category, and sort categories
	std::vector<BayesClassifier::Category> &category = dataFilterIt->category;
	category.push_back(Category(LowerBound, UpperBound));
	std::sort(category.begin(), category.end(), Category::SmallerThan);
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	FindCategory
/*!	\brief		
**	\param		
**	\return		true if the DataFilterKind DataFilterID exists in Choice ChoiceID, false otherwise.
				not
**	\exception	none
**	\par		DESCRIPTION:
**					
**	\author		Nathan Frost
**	\date		4/1/2003 8:25:55 PM
**	\warning	
				- The Category with LowerBound must exist in DataFilterID -- otherwise this 
				function is undefined.
*/
////////////////////////////////////////////////////////////////////////////////
bool BayesClassifier::FindCategory(	std::vector<Choice>::iterator &ChoiceIt, 
									std::vector<DataFilterKind>::iterator &DataFilterIt, 
									std::vector<Category>::iterator &CategoryIt, Action_ID ChoiceID, 
									unsigned int DataFilterID, float LowerBound)
{
	//find the DataFilterKind that contains the requested Category
	if(!FindDataFilterKind(ChoiceIt, DataFilterIt, ChoiceID, DataFilterID))
		return false;//requested DataFilterKind does not exist in the specified Choice

	//find the requested Category
	std::vector<BayesClassifier::Category> &category = DataFilterIt->category;
	for(CategoryIt = category.begin(); CategoryIt != category.end(); ++CategoryIt)
	{
		if(CategoryIt->lowerBound == LowerBound)
			break;
	}

	assert(CategoryIt != category.end());
	return true;//Category (better have been) found
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	RemoveCategoryFromDataFilterKindInChoice
/*!	\brief		Removes a classification for the specified DataFilterKind in the specified Choice.
**	\param		ChoiceID:			The Choice that the specified DataFilterKind resides in
	\param		DataFilterKindID:	The Attribute that is losing a classification
**	\param		LowerBound:			The lowest floating point value an Attribute can have while 
									still being considered in the classification to be removed
	\param		UpperBound:			The highest floating point value an Attribute can have while 
									still being considered in the classification to be removed
**	\return		none
**	\exception	- Throws a Nexception if either the specified Choice object or the specified  
				DataFilterKind object does not exist in the Bayesian Classifier
				- Throws a Nexception if the specified DataFilterKind's range (see DESCRIPTION for
				details) becomes invalid as a result of the calling of this function
**	\par		DESCRIPTION:
**					
**	\author		Nathan Frost
**	\date		11/7/2002 9:34:20 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
void BayesClassifier::RemoveCategoryFromDataFilterKindInChoice(	Action_ID ChoiceID, 
																unsigned int DataFilterKindID, 
																float LowerBound, 
																float UpperBound)
{
	std::vector<Choice>::iterator choiceIt;
	std::vector<DataFilterKind>::iterator dataFilterIt;
	std::vector<Category>::iterator categIt;
	FindCategory(choiceIt, dataFilterIt, categIt, ChoiceID, DataFilterKindID, LowerBound);
	assert(categIt->upperBound == UpperBound);
	
	dataFilterIt->category.erase(categIt);
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	Categorize
/*!	\brief		Returns the lowerBound of the Category that value falls into.  Returns the trivial
				classification (Value) if there is no Category for Value to fall into.
**	\param		ChoiceID:			The Choice that the specified DataFilterKind resides in
	\param		DataFilterKindID:	The DataFilterKind to use to classify the specified Value
	\param		Value:				The Value to be classified, relative to the specified 
									DataFilterKind
**	\return		lowerBound of the Category (which resides in the specified DataFilterKind, which
				resides in the specified Choice) that the specified Value fell into
**	\exception	- Throws a Nexception if either the specified Choice object or the specified  
				DataFilterKind object does not exist in the Bayesian Classifier
**	\par		DESCRIPTION:
**					
**	\author		Nathan Frost
**	\date		11/8/2002 10:13:46 AM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
float BayesClassifier::Categorize(Action_ID ChoiceID, unsigned int DataFilterKindID, float Value)
{
	std::vector<Choice>::iterator choiceIt;
	std::vector<DataFilterKind>::iterator dataFilterIt;
	
	if(!FindDataFilterKind(choiceIt, dataFilterIt, ChoiceID, DataFilterKindID))
		return Value;

	std::vector<BayesClassifier::Category> &category = dataFilterIt->category;

	//if the Value is smaller than the lowerBound of the first (lowest ranged) Category, clamp it
	//to the first (lowest ranged) Category
	if(Value < category.begin()->lowerBound)
		return category.begin()->lowerBound;

	//find the Category the Value falls into, if any
	for(std::vector<Category>::iterator categIt = category.begin(); categIt != category.end(); 
		++categIt)
	{
		//lowerBound inclusive, upperBound exclusive
		if(Value >= categIt->lowerBound && Value < categIt->upperBound)
		{
			return categIt->lowerBound;
		}
	}
	
	//if the Value is larger than the upperBound of the last (highest ranged) Category, clamp it
	//to the last (highest ranged) Category
	--categIt;
	return categIt->lowerBound;
}


////////////////////////////////////////////////////////////////////////////////
/*!	\class		ChoiceProbability
**	\brief		Encapsulates a Choice's identifier and its computed probability.  Helper class.
**	\author		Nathan Frost
**	\date		11/8/2002 5:28:19 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
struct ChoiceProbability
{
	Action_ID id;
	float probability;

	ChoiceProbability(	Action_ID ID = static_cast<Action_ID>(0), 
						float Probability = FLT_MAX):
	id(ID), probability(Probability){}
};

//TODO: this could be optimized further (FindCategory() gets called in Categorize())
////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	MakeChoice
/*!	\brief		Given the current situation (described by the base member attributes), the Bayesian
				Classifier considers its past experiences and decides on one of its Choices.
**	\param		Attributes:	Describe every perceivable aspect of the current situation.  Let C be 
							the set of all the BayesClassifier's Choices.  Let c be any Choice in 
							C.  Then if a given Attribute is passed to which c has no corresponding
							DataFilterKind (ie the BayesClassifier does not consider the given 
							Attribute when deliberating upon c), then the given Attribute has no 
							bearing on the attractiveness of c.
**	\return		The ID of the Action that corresponds to the chosen Choice.
**	\exception	none
**	\par		DESCRIPTION:
					- Currently, the Bayesian Classifier's Certainty value is ignored.
**					
**	\author		Nathan Frost
**	\date		11/8/2002 5:26:36 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
Action_ID BayesClassifier::MakeChoice(std::vector<AI_Attribute> &Attributes)
{
	std::vector<ChoiceProbability> choiceProbability(choice.size());

	/* Calculate the probability of every available Choice */
	int i=0;
    for(std::vector<Choice>::iterator choiceIt = choice.begin(); choiceIt != choice.end(); ++choiceIt, ++i)
    {
		//calculate the percentage of times this choice has been deemed to be true
		choiceProbability[i].id = choiceIt->id;
        choiceProbability[i].probability =  static_cast<float>(
											(choiceIt->numOccurrenceChoice + 1.0) /
											(numTrainingExamples + choice.size()));

		//add the sum of the probabilities of this choice being true given every perceived aspect
		//of the current situation
        for(unsigned int j=0; j < Attributes.size(); ++j)//every aspect in the current situation
        {
			std::vector<Choice>::iterator dummyChoiceIt;
			std::vector<DataFilterKind>::iterator dataFilterIt;
			std::vector<Category>::iterator categIt;
			if(!FindCategory(	dummyChoiceIt, dataFilterIt, categIt, choiceIt->id, 
								Attributes[j].ID, 
								Categorize(choiceIt->id, Attributes[j].ID, Attributes[j].value)))
			{//we don't think about the j'th Attribute for Choice choiceIt -- so skip it!
				continue;
			}

			//(numberOfOccurrencesOfAttribute) + 1 / (numberOfOccurrencesOfChoice + numberOfChoices)
			choiceProbability[i].probability *= static_cast<float>(
												(categIt->numOccurrences + 1.0) /
												(choiceIt->numOccurrenceChoice + 
												dataFilterIt->category.size()));
        }
    }

	/* Pick the Choice that is most probable */
	Action_ID decision;
	float biggestProbability = -FLT_MAX;
	float sumOfProbabilities = 0;
	for(unsigned int j=0; j < choiceProbability.size(); ++j)
	{
		if(choiceProbability[j].probability > biggestProbability)
		{
			biggestProbability = choiceProbability[j].probability;
			decision = choiceProbability[j].id;
		}

		//sum together the choice's probability (so we can calculate Certainty)
		sumOfProbabilities += choiceProbability[j].probability;
	}
	//generate Bayesian Classifier's confidence in its decision
    float certainty = biggestProbability / sumOfProbabilities;

	return decision;
}

//TODO: this could be optimized further (FindCategory() gets called in Categorize())
//TODO: Should be able to perform multiple examples in a single call
////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	LearnFromPositiveExample
/*!	\brief		Records a new, positive past experience
**	\param		CorrectChoice:	The Action_ID of the "correct" action, given a situation described
								in Attributes.
	\param		Attributes:		Describes every perceivable aspect of a situation in which 
								CorrectChoice identifies the "correct" action.  Note that if a 
								given Attribute is passed to which CorrectChoice has no 
								corresponding DataFilterKind (ie the BayesClassifier does not 
								consider the given Attribute when deliberating upon CorrectChoice),
								then the given Attribute has no bearing on the attractiveness of 
								CorrectChoice.
**	\return		none
**	\exception	none
**	\par		DESCRIPTION:
**					
**	\author		Nathan Frost
**	\date		11/20/2002 6:50:38 PM
**	\warning	
				The Bayesian Classifier must have CorrectChoice as well as every DataFilterKind 
				listed in Attributes present in its datastructures.  Otherwise, these
				datastructures will be (erroneously) added, and the Bayesian Classifier's behaviour
				will become undefined.

				BUG: If BC is presented a DataFilterKind in Attributes that it doesn't have in
					CorrectChoice) -- UNDEFINED BEHAVIOR RESULTS.
*/
////////////////////////////////////////////////////////////////////////////////
void BayesClassifier::LearnFromPositiveExample(	Action_ID CorrectChoice, 
												std::vector<AI_Attribute>& Attributes)
{
	++numTrainingExamples;

	//record which choice is "correct" in this situation
	std::vector<BayesClassifier::Choice>::iterator it;
	FindChoice(it, CorrectChoice);
	++it->numOccurrenceChoice;

	//record the presence of every aspect of the situation that should elicit CorrectChoice
	for(std::vector<AI_Attribute>::iterator attributIt = Attributes.begin(); 
		attributIt != Attributes.end(); ++attributIt)
	{
		std::vector<Choice>::iterator choiceIt;
		std::vector<DataFilterKind>::iterator dataFilterIt;
		std::vector<BayesClassifier::Category>::iterator categIt;
		if(!FindCategory(	choiceIt, dataFilterIt, categIt, CorrectChoice, attributIt->ID, 
							Categorize(CorrectChoice, attributIt->ID, attributIt->value)))
		{//we don't think about Attribute attributIt when considering Choice choiceIt -- so skip it
			continue;
		}
		++categIt->numOccurrences;
	}
}

//TODO: this could be optimized further (FindCategory() gets called in Categorize())
//TODO: this should call LearnFromPositiveExample() instead of duplicating code
//TODO: Should be able to perform multiple examples in a single call
////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	LearnFromNegativeExample
/*!	\brief		Records a new, negative past experience
**	\param		IncorrectChoice:The Action_ID of the "incorrect" action, given a situation 
								described in Attributes.
	\param		Attributes:		Describes every perceivable aspect of a situation in which 
								IncorrectChoice identifies the "incorrect" action.  Let C be the 
								set of all the BayesClassifier's Choices that does not include 
								IncorrectChoice.  Let c be any Choice in C.  Then if a given 
								Attribute is passed to which c has no corresponding DataFilterKind
								(ie the BayesClassifier does not consider the given Attribute when
								deliberating upon c), then the given Attribute has no bearing on
								the attractiveness of c.
**	\return		none
**	\exception	none
**	\par		DESCRIPTION:
**					
**	\author		Nathan Frost
**	\date		11/26/2002 6:47:21 PM
**	\warning	
				The Bayesian Classifier must have IncorrectChoice as well as every DataFilterKind 
				listed in Attributes present in its datastructures.  Otherwise, these
				datastructures will be (erroneously) added, and the Bayesian Classifier's behaviour
				will become undefined.

				BUG: If BC is presented a DataFilterKind in Attributes that it doesn't have in every
					Choice (except IncorrectChoice) -- UNDEFINED BEHAVIOR RESULTS.
*/
////////////////////////////////////////////////////////////////////////////////
void BayesClassifier::LearnFromNegativeExample(	Action_ID IncorrectChoice, 
												std::vector<AI_Attribute>& Attributes)
{
	++numTrainingExamples;
	
	/*	Maybe do any choice other than IncorrectChoice in the situation described by Attributes */
	//in similar situations, increase the likelihood of picking every choice that was not "incorrect"
	for(std::vector<Choice>::iterator curChoiceIt = choice.begin(); curChoiceIt != choice.end(); 
		++curChoiceIt)
	{
		if(curChoiceIt->id == IncorrectChoice)
			continue;
		
		//make "not incorrect" choices more likely to occur in a similar situation
		++curChoiceIt->numOccurrenceChoice;
		//record the presence of every aspect of the situation that perhaps should elicit the Choice
		for(std::vector<AI_Attribute>::iterator attributIt = Attributes.begin(); 
			attributIt != Attributes.end(); ++attributIt)
		{
			std::vector<Choice>::iterator choiceIt;
			std::vector<DataFilterKind>::iterator dataFilterIt;
			std::vector<BayesClassifier::Category>::iterator categIt;
			if(!FindCategory(	choiceIt, dataFilterIt, categIt, curChoiceIt->id, attributIt->ID, 
								Categorize(curChoiceIt->id, attributIt->ID, attributIt->value)))
			{//we don't think about Attribute attributIt when considering Choice choiceIt -- skipit
				continue;
			}

			++categIt->numOccurrences;
		}
	}
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	DebugOut
/*!	\brief		Vomits forth the innards of the Bayesian Classifier to an ostream.
**	\param		Os:	ostream to vomit into
**	\return		none
**	\exception	none
**	\par		DESCRIPTION:
**					
**	\author		Nathan Frost
**	\date		11/8/2002 9:24:09 AM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
void BayesClassifier::DebugOut(ostream &Os) const
{
	using std::endl;


	Os << "--------------------------------------------Begin BayesClassifier" << endl;
	Os << "numTrainingExamples: " << numTrainingExamples << endl;
	for(std::vector<Choice>::const_iterator it = choice.begin(); it != choice.end(); ++it)
	{
		Os << "	Choice " << it->id << endl;
		Os << "		has occurred " << it->numOccurrenceChoice << " times" << endl;

		for(std::vector<DataFilterKind>::const_iterator jt = it->dataFilterKind.begin(); 
			jt != it->dataFilterKind.end(); ++jt)
		{
			Os << "			DataFilter " << jt->id << endl;

			int i=0;
			for(std::vector<Category>::const_iterator kt = jt->category.begin(); 
				kt != jt->category.end(); ++kt, ++i)
			{
				Os << "				Category " << i << endl;
				Os << "					lowerBound: " << kt->lowerBound << endl;
				Os << "					upperBound: " << kt->upperBound << endl;
				Os << "					has occurred " <<	kt->numOccurrences << 
																" times" << endl;
			}
		}
	}

	Os << "---------------------------------------------End Bayes Classifier" << endl;
}

////////////////////////////////////////////////////////////////////////////////
//  FUNCTION:       
/*! \brief          Returns the location of the next instance of a character
**                  in a NULL terminated string that is part of a floating
**                  point number.
**  \param          stringPtr a pointer into the string to seek through
**  \return         A pointer to the first floating point character, or NULL if
**                  there are not such characters exist in the specified string.
**  \exception      none
**  \author         Austin Spafford
**  \date           5/15/2003 1:29:48 PM
**  \warning        none
*/
////////////////////////////////////////////////////////////////////////////////
static char * SeekToFirstFloatingPointCharacter(char * stringPtr)
{
	if (NULL == stringPtr) return NULL;

	// seek until the end of the string
	while (0 != *stringPtr)
	{
		// check to see if the current character is a number
		char numericalValue = *stringPtr - '0';
		if (numericalValue >= 0 && numericalValue <= 9)
		{
			return stringPtr;
		}

		// check to see if the current character is a decimal point
		if ('.' == *stringPtr)
		{
			return stringPtr;
		}

		// check to see if the current character is a minus sign
		if ('-' == *stringPtr)
		{
			return stringPtr;
		}

		// move to the next character
		++stringPtr;
	}

	return NULL;
}


void BayesClassifier::Read(std::istream &InputStream)
{
#define HACK_SIZE 1000
	char *characterBuffer = new char[HACK_SIZE];
	char *currentCharacterPtr;

	using std::endl;


	//Os << "--------------------------------------------Begin BayesClassifier" << endl;
	//eat line
	InputStream.getline(characterBuffer, HACK_SIZE);
	
	//Os << "numTrainingExamples: " << numTrainingExamples << endl;
	InputStream.getline(characterBuffer, HACK_SIZE);
	currentCharacterPtr = SeekToFirstFloatingPointCharacter(characterBuffer);
	assert(currentCharacterPtr);
	numTrainingExamples = atoi(currentCharacterPtr);

	//Os << "Number of choices: " << choice.size() << std::endl;
	InputStream.getline(characterBuffer, HACK_SIZE);
	currentCharacterPtr = SeekToFirstFloatingPointCharacter(characterBuffer);
	assert(currentCharacterPtr);
	int numberOfChoices = atoi(currentCharacterPtr);

	choice.clear();
	for(int i=0; i < numberOfChoices; ++i)
	{
		//Os << "	Choice " << it->id << endl;
		InputStream.getline(characterBuffer, HACK_SIZE);
		currentCharacterPtr = SeekToFirstFloatingPointCharacter(characterBuffer);
		assert(currentCharacterPtr);
		Choice tempChoice((Action_ID)atoi(currentCharacterPtr));

		//Os << "		has occurred " << it->numOccurrenceChoice << " times" << endl;
		InputStream.getline(characterBuffer, HACK_SIZE);
		currentCharacterPtr = SeekToFirstFloatingPointCharacter(characterBuffer);
		assert(currentCharacterPtr);
		tempChoice.numOccurrenceChoice = atoi(currentCharacterPtr);

		//Os << "Number of DataFilters: " << it->dataFilterKind.size() << std::endl;
		InputStream.getline(characterBuffer, HACK_SIZE);
		currentCharacterPtr = SeekToFirstFloatingPointCharacter(characterBuffer);
		assert(currentCharacterPtr);
		int numberOfDataFilters = atoi(currentCharacterPtr);

		for(int j=0; j != numberOfDataFilters; ++j)
		{
			//Os << "			DataFilter " << jt->id << endl;
			InputStream.getline(characterBuffer, HACK_SIZE);
			currentCharacterPtr = SeekToFirstFloatingPointCharacter(characterBuffer);
			assert(currentCharacterPtr);
			DataFilterKind tempDataFilterKind(atoi(currentCharacterPtr));
	
			//Os << "Number of Categories: " << jt->category.size() << std::endl;
			InputStream.getline(characterBuffer, HACK_SIZE);
			currentCharacterPtr = SeekToFirstFloatingPointCharacter(characterBuffer);
			assert(currentCharacterPtr);
			int numberOfCategories = atoi(currentCharacterPtr);

			for(int k=0; k != numberOfCategories; ++k)
			{
				Category tempCategory;
				//Os << "				Category " << i << endl;
				//eat line
				InputStream.getline(characterBuffer, HACK_SIZE);

				//Os << "					lowerBound: " << kt->lowerBound << endl;
				InputStream.getline(characterBuffer, HACK_SIZE);
				currentCharacterPtr = SeekToFirstFloatingPointCharacter(characterBuffer);
				assert(currentCharacterPtr);
				tempCategory.lowerBound = atof(currentCharacterPtr);

				//Os << "					upperBound: " << kt->upperBound << endl;
				InputStream.getline(characterBuffer, HACK_SIZE);
				currentCharacterPtr = SeekToFirstFloatingPointCharacter(characterBuffer);
				assert(currentCharacterPtr);
				tempCategory.upperBound = atof(currentCharacterPtr);

				//Os << "					has occurred " <<	kt->numOccurrences << " times" << endl;
				InputStream.getline(characterBuffer, HACK_SIZE);
				currentCharacterPtr = SeekToFirstFloatingPointCharacter(characterBuffer);
				assert(currentCharacterPtr);
				tempCategory.numOccurrences = atoi(currentCharacterPtr);

				tempDataFilterKind.category.push_back(tempCategory);
			}

			tempChoice.dataFilterKind.push_back(tempDataFilterKind);
		}

		choice.push_back(tempChoice);
	}

//	Os << "---------------------------------------------End Bayes Classifier" << endl;
	//eat tail
	InputStream.getline(characterBuffer, HACK_SIZE);

	delete []characterBuffer;
}

void BayesClassifier::Write(std::ostream &Os) const
{
	using std::endl;


	Os << "--------------------------------------------Begin BayesClassifier" << endl;
	Os << "numTrainingExamples: " << numTrainingExamples << endl;
	Os << "Number of choices: " << choice.size() << std::endl;
	for(std::vector<Choice>::const_iterator it = choice.begin(); it != choice.end(); ++it)
	{
		Os << "	Choice " << it->id << endl;
		Os << "		has occurred " << it->numOccurrenceChoice << " times" << endl;

		Os << "Number of DataFilters: " << it->dataFilterKind.size() << std::endl;
		for(std::vector<DataFilterKind>::const_iterator jt = it->dataFilterKind.begin(); 
			jt != it->dataFilterKind.end(); ++jt)
		{
			Os << "			DataFilter " << jt->id << endl;

			int i=0;
			Os << "Number of Categories: " << jt->category.size() << std::endl;
			for(std::vector<Category>::const_iterator kt = jt->category.begin(); 
				kt != jt->category.end(); ++kt, ++i)
			{
				Os << "				Category " << i << endl;
				Os << "					lowerBound: " << kt->lowerBound << endl;
				Os << "					upperBound: " << kt->upperBound << endl;
				Os << "					has occurred " <<	kt->numOccurrences << 
																" times" << endl;
			}
		}
	}

	Os << "---------------------------------------------End Bayes Classifier" << endl;
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	SmallerThan
/*!	\brief		
**	\param		
**	\return		none
**	\exception	none
**	\par		DESCRIPTION:
**					
**	\author		
**	\date		4/1/2003 8:09:18 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
bool BayesClassifier::Category::SmallerThan(const BayesClassifier::Category &First, const BayesClassifier::Category &Second)
{
	return First.lowerBound < Second.lowerBound;
}


void BayesClassifier::DebugSaveNLoad()
{
#if 0//testing ChoiceIDs and DataFilterKindIDs
	for(int i=1; i < 4; ++i)
	{
		AddChoice(static_cast<Action_ID>(i));
		AddDataFilterKindToChoice(static_cast<Action_ID>(i), i*600);
		AddOrReplaceCategoryToDataFilterKindInChoice(	static_cast<Action_ID>(i),  
														i*600, 0.0, 0.5);
		AddOrReplaceCategoryToDataFilterKindInChoice(	static_cast<Action_ID>(i),
														i*600, 0.5, 1.0);
		AddDataFilterKindToChoice(static_cast<Action_ID>(i), i*650);
		AddOrReplaceCategoryToDataFilterKindInChoice(	static_cast<Action_ID>(i), 
														i*650, 0.0, 0.5);
		AddOrReplaceCategoryToDataFilterKindInChoice(	static_cast<Action_ID>(i),
														i*650, 0.5, 1.0);
		AddDataFilterKindToChoice(static_cast<Action_ID>(i), i*700);
		AddOrReplaceCategoryToDataFilterKindInChoice(	static_cast<Action_ID>(i), 
														i*700, 0.0, 0.5);
		AddOrReplaceCategoryToDataFilterKindInChoice(	static_cast<Action_ID>(i), 
														i*700, 0.5, 1.0);
	}
#endif


#if 0//testing all except ChoiceIDs and DataFilterKindIDs
	for(int i=0; i < 3; ++i)
	{
		AddChoice(static_cast<Action_ID>(i));
		
		for(int j=0; j < 4; ++j)
		{
			AddDataFilterKindToChoice(static_cast<Action_ID>(i), j);
			
			AddOrReplaceCategoryToDataFilterKindInChoice(	static_cast<Action_ID>(i), 
															j, 0.0 + i*100 + j, 0.2 + i*100 + j);
			AddOrReplaceCategoryToDataFilterKindInChoice(	static_cast<Action_ID>(i), 
															j, 0.2 + i*100 + j, 0.4 + i*100 + j);
			AddOrReplaceCategoryToDataFilterKindInChoice(	static_cast<Action_ID>(i), 
															j, 0.4 + i*100 + j, 0.6 + i*100 + j);
			AddOrReplaceCategoryToDataFilterKindInChoice(	static_cast<Action_ID>(i), 
															j, 0.6 + i*100 + j, 0.8 + i*100 + j);
			AddOrReplaceCategoryToDataFilterKindInChoice(	static_cast<Action_ID>(i), 
															j, 0.8 + i*100 + j, 1.0 + i*100 + j);
		}
	}

	numTrainingExamples = 666;

	for(int a=0; a < 3; ++a)
	{
		choice[static_cast<Action_ID>(a)].numOccurrenceChoice = a + 1000;
		
		for(int j=0; j < 4; ++j)
		{
			choice[static_cast<Action_ID>(a)].dataFilterKind[j].category[0.0 + a*100 + j].numOccurrences = a+10*j;
			choice[static_cast<Action_ID>(a)].dataFilterKind[j].category[0.2 + a*100 + j].numOccurrences = a+10*j + 100;
			choice[static_cast<Action_ID>(a)].dataFilterKind[j].category[0.4 + a*100 + j].numOccurrences = a+10*j + 200;
			choice[static_cast<Action_ID>(a)].dataFilterKind[j].category[0.6 + a*100 + j].numOccurrences = a+10*j + 300;
			choice[static_cast<Action_ID>(a)].dataFilterKind[j].category[0.8 + a*100 + j].numOccurrences = a+10*j + 400;
		}
	}
#endif
}



#if 0//debug crap code
Action_ID BayesClassifier::MakeChoice()
{
	std::vector<ChoiceProbability> choiceProbability(choice.size());

	/* Calculate the probability of every available Choice */
	int i=0;
    for(std::map<Action_ID, Choice>::iterator it = choice.begin(); it != choice.end(); ++it, ++i)
    {
		//calculate the percentage of times this choice has been deemed to be true
		choiceProbability[i].id = it->first;
        choiceProbability[i].probability =  (it->second.numOccurrenceChoice + 1.0) /
											(numTrainingExamples + choice.size());
        

		/**donkeydebug**
		std::cout << "starting probability: " << choiceProbability[i].probability << std::endl;
		**donkeydebug**/


		//add the sum of the probabilities of the choice being true given every perceived aspect
		//of the current situation
		
		
		/**donkeydebug**
		//Synthesize attributes from TDDExample
		enum dataFilterz {SIZE, SPEED, TEETH};
		std::vector<AI_Attribute> attributes;
		AI_Attribute attr;
		attr.ID = SIZE;
		attr.value = static_cast<float>(0.2);
		attributes.push_back(attr);
		attr.ID = SPEED;
		attr.value = static_cast<float>(0.6);
		attributes.push_back(attr);
		attr.ID = TEETH;
		attr.value = static_cast<float>(0.9);
		attributes.push_back(attr);
		**donkeydebug**/


        for(int j=0; j < attributes.size(); ++j)//every aspect in the current situation
        {


			/**donkeydebug**
			std::cout << std::endl;

			std::cout << "Category Probability: " << 
(static_cast<float>(it->second.dataFilterKind[attributes[j].ID].category[Categorize(it->first, attributes[j].ID, attributes[j].value)].numOccurrences)
					+ 1.0) /
					(static_cast<float>(it->second.numOccurrenceChoice) + 
					static_cast<float>(it->second.dataFilterKind[attributes[j].ID].category.size()))
					<< std::endl;


			std::cout << std::endl;
			**donkeydebug**/


            choiceProbability[i].probability *= 
(static_cast<float>(it->second.dataFilterKind[attributes[j].ID].category[Categorize(it->first, attributes[j].ID, attributes[j].value)].numOccurrences)
					+ 1.0) /
					(static_cast<float>(it->second.numOccurrenceChoice) + 
					static_cast<float>(it->second.dataFilterKind[attributes[j].ID].category.size()));
        }
    }

	/* Pick the Choice that is most probable */
	Action_ID decision;
	float biggestProbability = -std::numeric_limits<float>::max();
	float sumOfProbabilities = 0;
	for(int j=0; j < choiceProbability.size(); ++j)
	{
		if(choiceProbability[j].probability > biggestProbability)
		{
			biggestProbability = choiceProbability[j].probability;
			decision = choiceProbability[j].id;
		}

		//sum together the choice's probability
		sumOfProbabilities += choiceProbability[j].probability;


		/**donkeydebug**
		std::cout << "choice Probability: " << choiceProbability[j].probability << std::endl;
		**donkeydebug**/


	}
	//generate Bayesian Classifier's confidence in its decision
    float certainty = biggestProbability / sumOfProbabilities;

	
	/**donkeydebug**
	std::cout << "decision: " << decision << std::endl;
	std::cout << "certainty: " << certainty << std::endl;
	**donkeydebug**/


	return decision;
}
#endif
#if 0
void BayesClassifier::SetUpUsTheTDDExample()
{
	/* Constants */
	enum choicez {NOT_PREDATOR, PREDATOR};
	enum dataFilterz {SIZE, SPEED, TEETH};
	float SIZE_SMALL, SPEED_SLOW, TEETH_DULL, SIZE_MEDIUM, SPEED_AVERAGE, TEETH_AVERAGE,
				SIZE_LARGE, SPEED_FAST, TEETH_SHARP;
	SIZE_SMALL = SPEED_SLOW = TEETH_DULL = 0.15;
	SIZE_MEDIUM = SPEED_AVERAGE = TEETH_AVERAGE = 0.51;
	SIZE_LARGE = SPEED_FAST = TEETH_SHARP = 0.8;

	/* Construct the datastructures */
	AddChoice(static_cast<Action_ID>(NOT_PREDATOR));
	AddDataFilterKindToChoice(static_cast<Action_ID>(NOT_PREDATOR), SIZE);
	AddOrReplaceCategoryToDataFilterKindInChoice(	static_cast<Action_ID>(NOT_PREDATOR), SIZE, 
													0.0, 0.33);
	AddOrReplaceCategoryToDataFilterKindInChoice(	static_cast<Action_ID>(NOT_PREDATOR), SIZE, 
													0.33, 0.66);
	AddOrReplaceCategoryToDataFilterKindInChoice(	static_cast<Action_ID>(NOT_PREDATOR), SIZE, 
													0.66, 1.0);
	
	AddDataFilterKindToChoice(static_cast<Action_ID>(NOT_PREDATOR), SPEED);
	AddOrReplaceCategoryToDataFilterKindInChoice(	static_cast<Action_ID>(NOT_PREDATOR), SPEED, 
													0.0, 0.33);
	AddOrReplaceCategoryToDataFilterKindInChoice(	static_cast<Action_ID>(NOT_PREDATOR), SPEED, 
													0.33, 0.66);
	AddOrReplaceCategoryToDataFilterKindInChoice(	static_cast<Action_ID>(NOT_PREDATOR), SPEED, 
													0.66, 1.0);

	AddDataFilterKindToChoice(static_cast<Action_ID>(NOT_PREDATOR), TEETH);
	AddOrReplaceCategoryToDataFilterKindInChoice(	static_cast<Action_ID>(NOT_PREDATOR), TEETH, 
													0.0, 0.33);
	AddOrReplaceCategoryToDataFilterKindInChoice(	static_cast<Action_ID>(NOT_PREDATOR), TEETH, 
													0.33, 0.66);
	AddOrReplaceCategoryToDataFilterKindInChoice(	static_cast<Action_ID>(NOT_PREDATOR), TEETH, 
													0.66, 1.0);
	

	AddChoice(static_cast<Action_ID>(PREDATOR));
	AddDataFilterKindToChoice(static_cast<Action_ID>(PREDATOR), SIZE);
	AddOrReplaceCategoryToDataFilterKindInChoice(	static_cast<Action_ID>(PREDATOR), SIZE, 
													0.0, 0.33);
	AddOrReplaceCategoryToDataFilterKindInChoice(	static_cast<Action_ID>(PREDATOR), SIZE, 
													0.33, 0.66);
	AddOrReplaceCategoryToDataFilterKindInChoice(	static_cast<Action_ID>(PREDATOR), SIZE, 
													0.66, 1.0);
	
	AddDataFilterKindToChoice(static_cast<Action_ID>(PREDATOR), SPEED);
	AddOrReplaceCategoryToDataFilterKindInChoice(	static_cast<Action_ID>(PREDATOR), SPEED, 
													0.0, 0.33);
	AddOrReplaceCategoryToDataFilterKindInChoice(	static_cast<Action_ID>(PREDATOR), SPEED, 
													0.33, 0.66);
	AddOrReplaceCategoryToDataFilterKindInChoice(	static_cast<Action_ID>(PREDATOR), SPEED, 
													0.66, 1.0);

	AddDataFilterKindToChoice(static_cast<Action_ID>(PREDATOR), TEETH);
	AddOrReplaceCategoryToDataFilterKindInChoice(	static_cast<Action_ID>(PREDATOR), TEETH, 
													0.0, 0.33);
	AddOrReplaceCategoryToDataFilterKindInChoice(	static_cast<Action_ID>(PREDATOR), TEETH, 
													0.33, 0.66);
	AddOrReplaceCategoryToDataFilterKindInChoice(	static_cast<Action_ID>(PREDATOR), TEETH, 
													0.66, 1.0);

	
	/* Give 10 training examples */
	std::vector<AI_Attribute> attributes;
	AI_Attribute attrSize, attrSpeed, attrTeeth;

	//one
	attrSize.ID = SIZE;
	attrSize.value = SIZE_SMALL;
	attributes.push_back(attrSize);

	attrSpeed.ID = SPEED;
	attrSpeed.value = SPEED_FAST;
	attributes.push_back(attrSpeed);

	attrTeeth.ID = TEETH;
	attrTeeth.value = TEETH_AVERAGE;
	attributes.push_back(attrTeeth);

	LearnFromPositiveExample(NOT_PREDATOR, attributes);
	attributes.clear();

	//two
	attrSize.ID = SIZE;
	attrSize.value = SIZE_MEDIUM;
	attributes.push_back(attrSize);

	attrSpeed.ID = SPEED;
	attrSpeed.value = SPEED_FAST;
	attributes.push_back(attrSpeed);

	attrTeeth.ID = TEETH;
	attrTeeth.value = TEETH_SHARP;
	attributes.push_back(attrTeeth);

	LearnFromPositiveExample(PREDATOR, attributes);
	attributes.clear();

	//three
	attrSize.ID = SIZE;
	attrSize.value = SIZE_SMALL;
	attributes.push_back(attrSize);

	attrSpeed.ID = SPEED;
	attrSpeed.value = SPEED_AVERAGE;
	attributes.push_back(attrSpeed);

	attrTeeth.ID = TEETH;
	attrTeeth.value = TEETH_AVERAGE;
	attributes.push_back(attrTeeth);

	LearnFromPositiveExample(PREDATOR, attributes);
	attributes.clear();

	//four
	attrSize.ID = SIZE;
	attrSize.value = SIZE_LARGE;
	attributes.push_back(attrSize);

	attrSpeed.ID = SPEED;
	attrSpeed.value = SPEED_SLOW;
	attributes.push_back(attrSpeed);

	attrTeeth.ID = TEETH;
	attrTeeth.value = TEETH_DULL;
	attributes.push_back(attrTeeth);

	LearnFromPositiveExample(NOT_PREDATOR, attributes);
	attributes.clear();

	//five
	attrSize.ID = SIZE;
	attrSize.value = SIZE_MEDIUM;
	attributes.push_back(attrSize);

	attrSpeed.ID = SPEED;
	attrSpeed.value = SPEED_AVERAGE;
	attributes.push_back(attrSpeed);

	attrTeeth.ID = TEETH;
	attrTeeth.value = TEETH_AVERAGE;
	attributes.push_back(attrTeeth);

	LearnFromPositiveExample(NOT_PREDATOR, attributes);
	attributes.clear();

	//six
	attrSize.ID = SIZE;
	attrSize.value = SIZE_MEDIUM;
	attributes.push_back(attrSize);

	attrSpeed.ID = SPEED;
	attrSpeed.value = SPEED_FAST;
	attributes.push_back(attrSpeed);

	attrTeeth.ID = TEETH;
	attrTeeth.value = TEETH_DULL;
	attributes.push_back(attrTeeth);

	LearnFromPositiveExample(NOT_PREDATOR, attributes);
	attributes.clear();

	//seven
	attrSize.ID = SIZE;
	attrSize.value = SIZE_SMALL;
	attributes.push_back(attrSize);

	attrSpeed.ID = SPEED;
	attrSpeed.value = SPEED_SLOW;
	attributes.push_back(attrSpeed);

	attrTeeth.ID = TEETH;
	attrTeeth.value = TEETH_SHARP;
	attributes.push_back(attrTeeth);

	LearnFromPositiveExample(NOT_PREDATOR, attributes);
	attributes.clear();

	//eight
	attrSize.ID = SIZE;
	attrSize.value = SIZE_LARGE;
	attributes.push_back(attrSize);

	attrSpeed.ID = SPEED;
	attrSpeed.value = SPEED_SLOW;
	attributes.push_back(attrSpeed);

	attrTeeth.ID = TEETH;
	attrTeeth.value = TEETH_AVERAGE;
	attributes.push_back(attrTeeth);

	LearnFromPositiveExample(NOT_PREDATOR, attributes);
	attributes.clear();

	//nine
	attrSize.ID = SIZE;
	attrSize.value = SIZE_LARGE;
	attributes.push_back(attrSize);

	attrSpeed.ID = SPEED;
	attrSpeed.value = SPEED_AVERAGE;
	attributes.push_back(attrSpeed);

	attrTeeth.ID = TEETH;
	attrTeeth.value = TEETH_DULL;
	attributes.push_back(attrTeeth);

	LearnFromPositiveExample(NOT_PREDATOR, attributes);
	attributes.clear();

	//ten
	attrSize.ID = SIZE;
	attrSize.value = SIZE_SMALL;
	attributes.push_back(attrSize);

	attrSpeed.ID = SPEED;
	attrSpeed.value = SPEED_FAST;
	attributes.push_back(attrSpeed);

	attrTeeth.ID = TEETH;
	attrTeeth.value = TEETH_AVERAGE;
	attributes.push_back(attrTeeth);

	LearnFromPositiveExample(PREDATOR, attributes);
	attributes.clear();

	/* Construct the current situation */
	attrSize.ID = SIZE;
	attrSize.value = static_cast<float>(0.15);
	attributes.push_back(attrSize);

	attrSpeed.ID = SPEED;
	attrSpeed.value = static_cast<float>(0.51);
	attributes.push_back(attrSpeed);

	attrTeeth.ID = TEETH;
	attrTeeth.value = static_cast<float>(0.98);
	attributes.push_back(attrTeeth);

	DebugOut(std::cout);

	std::cout << "Choice is: " << MakeChoice(attributes) << std::endl;
	attributes.clear();

	std::cout << "Test DataFilter ignorance with LearnFromNegativeExample" << std::endl;
	attrSize.ID = 37;
	attrSize.value = SIZE_LARGE;
	attributes.push_back(attrSize);

	attrSpeed.ID = 65536;
	attrSpeed.value = 900.05;
	attributes.push_back(attrSpeed);

	attrTeeth.ID = 666;
	attrTeeth.value = -432.57;
	attributes.push_back(attrTeeth);

	LearnFromNegativeExample(NOT_PREDATOR, attributes);
	attributes.clear();
	DebugOut(std::cout);

	std::cout << "Test DataFilter ignorance with LearnFromPositiveExample" << std::endl;
	attrSize.ID = 37;
	attrSize.value = SIZE_LARGE;
	attributes.push_back(attrSize);

	attrSpeed.ID = 65536;
	attrSpeed.value = 900.05;
	attributes.push_back(attrSpeed);

	attrTeeth.ID = 666;
	attrTeeth.value = -432.57;
	attributes.push_back(attrTeeth);

	LearnFromPositiveExample(NOT_PREDATOR, attributes);
	//DON'T attributes.clear() -- we want to have some bogus Attributes mixed in for the next test
	DebugOut(std::cout);

	/* Construct the current situation */
	std::cout << "Test DataFilter ignorance with MakeChoice" << std::endl;
	attrSize.ID = SIZE;
	attrSize.value = static_cast<float>(0.15);
	attributes.push_back(attrSize);

	attrSpeed.ID = SPEED;
	attrSpeed.value = static_cast<float>(0.51);
	attributes.push_back(attrSpeed);

	attrTeeth.ID = TEETH;
	attrTeeth.value = static_cast<float>(0.98);
	attributes.push_back(attrTeeth);

	DebugOut(std::cout);

	std::cout << "Choice is: " << MakeChoice(attributes) << std::endl;
	attributes.clear();

	std::cout << "Add Choice, with two DataFilterKinds and three Categories for each" << std::endl;
	AddChoice(2);
	AddDataFilterKindToChoice(2, 0);
	AddOrReplaceCategoryToDataFilterKindInChoice(2, 0, 0.0, 0.33);
	AddOrReplaceCategoryToDataFilterKindInChoice(2, 0, 0.33, 0.66);
	AddOrReplaceCategoryToDataFilterKindInChoice(2, 0, 0.66, 1.0);
	AddDataFilterKindToChoice(2, 1);
	AddOrReplaceCategoryToDataFilterKindInChoice(2, 1, 0.0, 0.33);
	AddOrReplaceCategoryToDataFilterKindInChoice(2, 1, 0.33, 0.66);
	AddOrReplaceCategoryToDataFilterKindInChoice(2, 1, 0.66, 1.0);

	DebugOut(std::cout);

	std::cout << "A SMALL, FAST Fish with AVERAGE teeth was incorrectly classified as a PREDATOR." 
			<< std::endl;
	attrSize.ID = SIZE;
	attrSize.value = SIZE_SMALL;
	attributes.push_back(attrSize);

	attrSpeed.ID = SPEED;
	attrSpeed.value = SPEED_FAST;
	attributes.push_back(attrSpeed);

	attrTeeth.ID = TEETH;
	attrTeeth.value = TEETH_AVERAGE;
	attributes.push_back(attrTeeth);

	LearnFromNegativeExample(PREDATOR, attributes);
	attributes.clear();

	DebugOut(std::cout);

	std::cout << "A LARGE, SLOW Fish with SHARP teeth was incorrectly classified as not a PREDATOR." 
			<< std::endl;
	attrSize.ID = SIZE;
	attrSize.value = SIZE_LARGE;
	attributes.push_back(attrSize);

	attrSpeed.ID = SPEED;
	attrSpeed.value = SPEED_SLOW;
	attributes.push_back(attrSpeed);

	attrTeeth.ID = TEETH;
	attrTeeth.value = TEETH_SHARP;
	attributes.push_back(attrTeeth);

	LearnFromNegativeExample(NOT_PREDATOR, attributes);
	attributes.clear();

	DebugOut(std::cout);
}
#endif

#if 0
void BayesClassifier::subClassInit()
{
	//TODO: Move these into the derived class of BayesianClassifierAI_Unit
	Action_ID no = static_cast<Action_ID>(0);
	Action_ID yes = static_cast<Action_ID>(1);
	AddChoice(no);	//No, don't recommend the Action
	AddChoice(yes);	//Yes, recommend the Action

	//fill out the DataFilters that are considered for both Choices
	for(std::vector<DF_ID>::iterator it = DataFilters.begin(); it != DataFilters.end(); 
		++it)
	{
		AddDataFilterKindToChoice(no, *it);
		AddDataFilterKindToChoice(yes, *it);
	}

	//for each Choice, add three Categories to each DataFilterKind residing in that Choice
	for(std::map<Action_ID, Choice>::iterator jt = choice.begin(); jt != choice.end();
		++jt)
	{
		for(std::map<unsigned int, DataFilterKind>::iterator kt = jt->second.dataFilterKind.begin();
			kt != jt->second.dataFilterKind.end(); ++kt)
		{
			AddOrReplaceCategoryToDataFilterKindInChoice(jt->first, kt->first, 0.0, 0.25);
			AddOrReplaceCategoryToDataFilterKindInChoice(jt->first, kt->first, 0.25, 0.5);
			AddOrReplaceCategoryToDataFilterKindInChoice(jt->first, kt->first, 0.5, 0.75);
			AddOrReplaceCategoryToDataFilterKindInChoice(jt->first, kt->first, 0.75, 1.0);
		}
	}
}

#endif
