////////////////////////////////////////////////////////////////////////////////
/*!	\file		BayesClassifierOpt.h
**	\brief		Naive Bayes Classifier
**	\author		Nathan Frost
**	\date		11/7/2002 8:14:08 PM
**	\par		DESCRIPTION:
					- Based on Bayes theorem
					- A probabilistic method of machine learning that's based on frequencies of 
					occurrence in training data. 
					- Found to have performance comparable to neural networks or decision trees.
**					
**	\par		REMARKS:
					- Action_ID and AI_Attribute MUST be 32bit integral datatypes.
					- Every argument passed to an Action_ID or AI_Attribute parameters MUST be
					unique.

	\todo		
				- eliminate all dependencies on outside datatypes (AI_ID, to name but
				one) by switching to equivalent internal datatypes
*/
////////////////////////////////////////////////////////////////////////////////
#pragma once

#include <vector>
#include "./nexception/Nexception.h"
#include "./ai_core/AI_Keys.h"
#include "./ai_core/AI_Attribute_Action.h"
#include <map>
#include <iostream>


namespace Bayes
{
	//forward declarations
	class BayesClassifierFileMemento;
	
	
	using AI_System::Action_ID;
	using AI_System::AI_Attribute;


	////////////////////////////////////////////////////////////////////////////////
	/*!	\class		BayesClassifier
	**	\brief		Naive Bayes Classifier
	**	\author		Nathan Frost
	**	\date		11/7/2002 8:17:18 PM
	**	\warning	
					BayesClassifier is designed to throw a Nexception any time its internal 
					datastructures violate its invariant -- that is, if the BayesClassifier, 
					through the calling its interface functions, becomes messed up (such that 
					its behaviour becomes undefined), it will immediately throw a Nexception,
					allowing the client (or the writer of the BayesClassifier class) a chance to 
					rectify his or her mistake.
	*/
	////////////////////////////////////////////////////////////////////////////////
	class BayesClassifier
	{
		///nested class declarations
		////////////////////////////////////////////////////////////////////////////////
		/*!	\class		BayesClassifier::Category
		**	\brief		Encapsulates a single classification of a given DataFilter; for example, 
						"Large Teeth" could be classified to be Teeth values of 0.75 to 1.0
		**	\author		Nathan Frost
		**	\date		11/7/2002 8:20:33 PM
		**	\warning	
						- A Category's range is defined to be the half open interval of real 
						numbers from lowerBound (inclusive) to upperBound (exclusive).
						- A Category's range is considered valid if and only if its lowerBound
						is smaller than its upperBound
						- Otherwise, the Category's range is considered to be invalid
		*/
		////////////////////////////////////////////////////////////////////////////////
		struct Category
		{
			float lowerBound;
			float upperBound;
			unsigned int numOccurrences;///<Number of times classification has been encountered
			
			Category(float LowerBound, float UpperBound, unsigned int NumOccurrences = 0);
			Category(){}///<never call this-- it's provided only so it can be instantiated (by
						///<value) inside of STL containers
			static bool SmallerThan(const Category &First, 
									const Category &Second);
		};
		////////////////////////////////////////////////////////////////////////////////
		/*!	\class		BayesClassifier::DataFilterKind
		**	\brief		Encapsulates a single DataFilter ("Teeth", for example), and all the 
						classifications that have been defined for it
		**	\author		Nathan Frost
		**	\date		11/7/2002 8:23:35 PM
		**	\warning	
						- A DataFilterKind's range is defined to be the union of its Category's
						ranges
						- 1) All of a DataFilterKind's Category's must be valid
						- 2) All of a DataFilterKind's Category's ranges must be continuous
						- 3) None of a DataFilterKind's Category's ranges can have any overlap
						- If any of the above three restrictions are not met, then the 
						DataFilterKind's range is considered to be invalid, and the Bayesian
						Classifier's behaviour becomes undefined.
						- If a DataFilterKind has no Category's, then it has the Null range, which
						is considered to be valid
		*/
		////////////////////////////////////////////////////////////////////////////////
		struct DataFilterKind
		{
			unsigned int id;
			///sorted by lowerBound
			std::vector<Category> category;///<Classifications of the DataFilter
			
			DataFilterKind(unsigned int ID);
		};
		////////////////////////////////////////////////////////////////////////////////
		/*!	\class		Choice
		**	\brief		Encapsulates a single Action that the Bayesian Classifier can deliberate
						upon, and the DataFilters it will take into consideration when deciding
						whether or not to decide upon this Choice.
		**	\author		Nathan Frost
		**	\date		11/7/2002 8:24:41 PM
		**	\warning	none
		*/
		////////////////////////////////////////////////////////////////////////////////
		struct Choice
		{
			///Number of times this Choice has been deemed correct in the past
			Action_ID id;
			unsigned int numOccurrenceChoice;
			std::vector<DataFilterKind> dataFilterKind;///<Consider these when deciding
			
			Choice(Action_ID ID, unsigned int NumOccurrenceChoice = 0);
		};
		
		
		///private data
		unsigned int numTrainingExamples;///<Total number of past experiences
		std::vector<Choice> choice;///<The Actions Bayesian Classifier is capable of deciding between

		
		///private helper functions
		void FindChoice(std::vector<Choice>::iterator &It, Action_ID ID);
		bool FindDataFilterKind(std::vector<Choice>::iterator &ChoiceIt, 
								std::vector<DataFilterKind>::iterator &DataFilterIt, 
								Action_ID ChoiceID, unsigned int DataFilterID);
		bool FindCategory(	std::vector<Choice>::iterator &ChoiceIt, 
							std::vector<DataFilterKind>::iterator &DataFilterIt, 
							std::vector<Category>::iterator &CategoryIt, Action_ID ChoiceID, 
							unsigned int DataFilterID, float LowerBound);
		float Categorize(Action_ID ChoiceID, unsigned int DataFilterKindID, float value);

		
		friend BayesClassifierFileMemento;
	public:
		BayesClassifier();
		
		Action_ID MakeChoice(std::vector<AI_Attribute> &Attributes);
		void LearnFromPositiveExample(	Action_ID CorrectChoice, 
										std::vector<AI_Attribute>& Attributes);
		void LearnFromNegativeExample(	Action_ID IncorrectChoice, 
										std::vector<AI_Attribute>& Attributes);

		///data set-up
		void AddChoice(Action_ID ID, unsigned int NumOccurrences = 0);
		void RemoveChoice(Action_ID ID);
		void AddDataFilterKindToChoice(Action_ID ChoiceID, unsigned int DataFilterKindID);
		void RemoveDataFilterKindFromChoice(Action_ID ChoiceID, unsigned int DataFilterKindID);
		void AddOrReplaceCategoryToDataFilterKindInChoice(	Action_ID ChoiceID, 
															unsigned int DataFilterKindID, 
															float LowerBound, float UpperBound);
		void RemoveCategoryFromDataFilterKindInChoice(	Action_ID ChoiceID, 
														unsigned int DataFilterKindID, 
														float LowerBound, float UpperBound);
		void Write(std::ostream &OutputStream) const;
		void Read(std::istream &InputStream);
	


		void DebugOut(std::ostream &Os) const;
		void DebugSaveNLoad();
		void SetUpUsTheTDDExample();
	};
}

/*
//BEGIN_HACK
namespace MsgHandler
{
	typedef unsigned int Key;	///<Uniquely identifies every entity in the game
}

namespace AI_System
{
	typedef unsigned int AI_ID;
	typedef unsigned int Decision_ID;
		typedef unsigned int Attribute_ID;
typedef MsgHandler::Key Object_ID;

	typedef enum
	{
		AC_TEST = 0xFFFFFFFF,
		AC_NULL = 0,
		//Action classes bits 15 - 31
		AC_MASK		= 0xFFFF0000, 
		AC_MOVE		= (1 << 16),
		AC_MOUTH	= (1 << 17),
		AC_MARKER	= (1 << 18),
		//Action IDs any number up to an unsigned short
		//Movement actions
		MOVETO			= AC_MOVE | 1,
		FLEE			= AC_MOVE | 2,
		FOLLOW			= AC_MOVE | 3,
		MOVEWITH		= AC_MOVE | 4,
		ROAM			= AC_MOVE | 5,
		GOTO_POS_MARKER = AC_MOVE | 6,
		FLEE_NEG_MARKER = AC_MOVE | 7,
		IGNORE_MARKERS	= AC_MOVE | 8,
		HEED_MARKERS	= AC_MOVE | 9,
		//Mouth actions
		ATTACK			= AC_MOUTH | 1,
		EAT				= AC_MOUTH | 2,
		PICKUP			= AC_MOUTH | 3,
		DROP			= AC_MOUTH | 4,
		//Location actions
		DROP_POS_MARKER	= AC_MARKER | 1,
		DROP_NEG_MARKER = AC_MARKER | 2,
		EVALUATEMARKER	= AC_MARKER | 3,
	} Action_ID;

struct AI_ActionTag
{
	AI_ID			unit_ID;
	Decision_ID		decisionID;
	float			fitness;
	bool			completed;
	bool operator < (const AI_ActionTag& T) const throw()
	{
		//compare the unit ID's
		if(unit_ID < T.unit_ID)
			return true;
		
		//compare the decision ID's
		if(decisionID < T.decisionID)
			return true;

		return false;
	}
};

	struct AI_Action
	{
	public:
		Action_ID		ID;
		float			value;
		AI_ActionTag	tag;		
		Object_ID		objectID;	//!< Optional ID of the currently evaluated object
	};

	struct AI_Attribute
	{
		Attribute_ID	ID;		//!< Identifier for this attribute
		float			value;	//!< Value returned from Data Filter
		
		
		// Added by nate so he could use some algorithmal techniques. [11/14/2002]
		float GetValue() const throw(){return value;}
	};



}
//END_HACK
*/
