////////////////////////////////////////////////////////////////////////////////
/*!	\file		BayesUnit.cpp
**	\brief		Implementation
**	\author		Nathan Frost
**	\date		11/20/2002 9:12:45 PM
**	\par		DESCRIPTION:
**					
**	\par		REMARKS:
					- Action_ID and AI_Attribute MUST be 32bit integral datatypes.
					- Every argument passed to an Action_ID or AI_Attribute parameters MUST be
					unique.
**					
*/
////////////////////////////////////////////////////////////////////////////////
#pragma warning(disable:4786)

//#include "BayesUnit.h"
#include"../AI_System.h"
#include"BayesClassifierFileMemento.h"
using namespace Bayes;

// memory leak detection
#ifdef _DEBUG
    #include"memleak.h"
    #define new DEBUG_NEW
    #undef THIS_FILE
    static char THIS_FILE[] = __FILE__;
#endif


////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	Constructor
/*!	\brief		Calls base class's constructor
**	\param		none
**	\return		none
**	\exception	none
**	\par		DESCRIPTION:
**					
**	\author		Nathan Frost
**	\date		11/20/2002 9:44:56 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
BayesUnit::BayesUnit():
AI_Unit(AI_System::TT_NBC)
{
	;
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	triggerAction
/*!	\brief		Decides whether or not to recommend the Action associated with this AI Unit
**	\param		none
**	\return		none
**	\exception	none
**	\par		DESCRIPTION:
					- utilizes base member "attributes" in AI_Unit to describe all perceivable
					aspects of the current situation
**					
**	\author		Nathan Frost
**	\date		11/20/2002 9:16:21 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
bool BayesUnit::triggerAction() throw()
{
	//note that in subClassInit(), "Yes" is equal to 1, and "No" is equal to 0
	if(bayesClassifier.MakeChoice(attributes))//base member attributes describes current situation
		return true;
	else
		return false;
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	saveTrainingExample
/*!	\brief		Gives the AI Unit an example of a situation (described by Attributes), and what 
				response is correct in such a situation.
**	\param		response:	The fitness function determined "correct" response
**	\param		fitness:	The fitness function determined fitness (ignored)
**	\param		Attributes:	Describes every perceivable aspect of the situation in which response
							 is the "correct" reaction
**	\return		none
**	\exception	none
**	\par		DESCRIPTION:
**					
**	\author		Nathan Frost
**	\date		11/20/2002 9:16:50 PM
**	\warning	
				See warnings for BayesClassifier::LearnFromPositiveExample()
*/
////////////////////////////////////////////////////////////////////////////////
void BayesUnit::saveTrainingExample(bool response, float fitness, 
											std::vector<AI_Attribute>& Attributes) throw()
{
	if(response)
		bayesClassifier.LearnFromPositiveExample(YES, Attributes);
	else
		bayesClassifier.LearnFromPositiveExample(NO, Attributes);

	return;
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	subClassInit
/*!	\brief		Properly establishes the BayesUnit invariant.  After this function is called, the
				BayesUnit is all set to go.
**	\param		none
**	\return		none
**	\exception	none
**	\par		DESCRIPTION:
					- Adds the choices YES and NO
					- Adds every DataFilterKind in base member DataFilters to both Choices
					- Adds (arbitrarily) four Categories for every DataFilterKind in both Choices
**					
**	\author		Nathan Frost
**	\date		11/20/2002 9:26:59 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
void BayesUnit::subClassInit()
{
	bayesClassifier.AddChoice(NO);	//No, don't recommend the Action associated with this AI_Unit
	bayesClassifier.AddChoice(YES);	//Yes, recommend the Action associated with this AI_Unit

	//fill out the DataFilters that are considered for both Choices from base member DataFilters
	for(std::vector<AI_System::DF_ID>::iterator it = DataFilters.begin(); it != DataFilters.end(); 
	++it)
	{
		bayesClassifier.AddDataFilterKindToChoice(NO, *it);
		bayesClassifier.AddDataFilterKindToChoice(YES, *it);
	}
	
	//for both Choices, add four Categories to each DataFilterKind residing in that Choice
	for(std::vector<AI_System::DF_ID>::iterator kt = DataFilters.begin(); kt != DataFilters.end(); 
		++kt)
	{
		bayesClassifier.AddOrReplaceCategoryToDataFilterKindInChoice(NO, *kt, 0.0, 0.25);
		bayesClassifier.AddOrReplaceCategoryToDataFilterKindInChoice(NO, *kt, 0.25, 0.5);
		bayesClassifier.AddOrReplaceCategoryToDataFilterKindInChoice(NO, *kt, 0.5, 0.75);
		bayesClassifier.AddOrReplaceCategoryToDataFilterKindInChoice(NO, *kt, 0.75, 1.0);

		bayesClassifier.AddOrReplaceCategoryToDataFilterKindInChoice(YES, *kt, 0.0, 0.25);
		bayesClassifier.AddOrReplaceCategoryToDataFilterKindInChoice(YES, *kt, 0.25, 0.5);
		bayesClassifier.AddOrReplaceCategoryToDataFilterKindInChoice(YES, *kt, 0.5, 0.75);
		bayesClassifier.AddOrReplaceCategoryToDataFilterKindInChoice(YES, *kt, 0.75, 1.0);
	}
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	DebugOut
/*!	\brief		Vomits forth the innards of the Bayesian Classifier to an ostream.
**	\param		Os:	ostream to vomit into
**	\return		none
**	\exception	none
**	\par		DESCRIPTION:
**					
**	\author		Nathan Frost
**	\date		11/21/2002 11:14:19 AM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
void BayesUnit::DebugOut(ostream &Os)
{
	bayesClassifier.DebugOut(Os);
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	Read
/*!	\brief		Replaces the BayesClassifier stored in *this with the BayesClassifier stored in 
				InputStream.
**	\param		InputStream:	To read in the BayesClassifier from
**	\return		none
**	\exception	none
**	\par		DESCRIPTION:
**					
**	\author		Nathan Frost
**	\date		4/15/2003 6:56:49 PM
**	\warning	
				- InputStream must contain data from BayesADN::Write(), or else this function's
				behaviour is undefined
*/
////////////////////////////////////////////////////////////////////////////////
void BayesUnit::Read(std::istream &InputStream)
{
	bayesClassifier.Read(InputStream);	
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	Write
/*!	\brief		Writes out the BayesClassifier inside *this to OutputStream.
**	\param		OutputStream:	To receive the data from *this's BayesClassifier
**	\return		none
**	\exception	none
**	\par		DESCRIPTION:
**					
**	\author		Nathan Frost
**	\date		4/15/2003 6:56:49 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
void BayesUnit::Write(std::ostream &OutputStream) const
{
	bayesClassifier.Write(OutputStream);
}

//deprecated
bool BayesUnit::save() throw()
{
	return false;
}

//deprecated
bool BayesUnit::load() throw()
{
	return false;
}
