////////////////////////////////////////////////////////////////////////////////
/*!	\file		BayesADN.cpp
**	\brief		Implementation
**	\author		Nathan Frost
**	\date		11/21/2002 12:53:17 PM
**	\par		DESCRIPTION:
**					
**	\par		REMARKS:
					- Action_ID and AI_Attribute MUST be 32bit integral datatypes.
					- Every argument passed to an Action_ID or AI_Attribute parameters MUST be
					unique.
**					
*/
////////////////////////////////////////////////////////////////////////////////
//#include "BayesADN.h"
#include"../AI_System.h"
#include"BayesClassifierFileMemento.h"
using namespace Bayes;

// memory leak detection
#ifdef _DEBUG
    #include"memleak.h"
    #define new DEBUG_NEW
    #undef THIS_FILE
    static char THIS_FILE[] = __FILE__;
#endif


//forward declarations
namespace AI_System
{
	class AI_Unit;
}


////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	Constructor
/*!	\brief		Calls base class's constructor
**	\param		none
**	\return		none
**	\exception	none
**	\par		DESCRIPTION:
**					
**	\author		Nathan Frost
**	\date		11/21/2002 1:21:20 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
BayesADN::BayesADN():
AI_ADN(AI_System::TT_NBC)
{
	;
}


////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	subClassInit
/*!	\brief		Properly establishes the BayesADN invariant.  After this function is called, the
				BayesADN is all set to go.
**	\param		none
**	\return		none
**	\exception	none
**	\par		DESCRIPTION:
					- Adds as many Choices as there are AI Units in AI_Units (map the unique,
					integral AI_IDs to Action_IDs (the internal representation of unique integral
					values))
					- Inside each Choice, adds as many DataFilterKinds as there Units (again
					mapping AI_IDs to Action_IDs for internal storage)
					- Adds (arbitrarily) four Categories for every DataFilterKind in every Choice
**					
**	\author		Nathan Frost
**	\date		11/21/2002 1:26:53 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
void BayesADN::subClassInit()
{
	//Every AI_Unit connected to this BayesADN is a Choice it will deliberate on
	for(std::map<AI_ID, AI_System::AI_Unit*>::iterator it = AI_Units.begin(); it != AI_Units.end();
		++it)
	{
		//Map AI_ID to Action_ID for internal storage
		bayesClassifier.AddChoice(static_cast<Action_ID>(it->first));
		
		//Every AI_Unit connected to this BayesADN is also a perceivable aspect of a given situation
		for(std::map<AI_ID, AI_System::AI_Unit*>::iterator jt = AI_Units.begin();
			jt != AI_Units.end(); ++jt)
		{
			bayesClassifier.AddDataFilterKindToChoice(
												static_cast<Action_ID>(it->first), 
												static_cast<unsigned int>(jt->first));

			/* Arbitrarily add four categories for each aspect of a given perceivable situation */
			bayesClassifier.AddOrReplaceCategoryToDataFilterKindInChoice(
								static_cast<Action_ID>(it->first), 
								static_cast<unsigned int>(jt->first), 0.0, 0.25);
			bayesClassifier.AddOrReplaceCategoryToDataFilterKindInChoice(
								static_cast<Action_ID>(it->first), 
								static_cast<unsigned int>(jt->first), 0.25, 0.5);
			bayesClassifier.AddOrReplaceCategoryToDataFilterKindInChoice(
								static_cast<Action_ID>(it->first), 
								static_cast<unsigned int>(jt->first), 0.5, 0.75);
			bayesClassifier.AddOrReplaceCategoryToDataFilterKindInChoice(
								static_cast<Action_ID>(it->first), 
								static_cast<unsigned int>(jt->first), 0.75, 1.0);
		}
	}
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	saveTrainingExample
/*!	\brief		Gives the Action Decision Node an example of a situation (described by 
				Attributes), and, if Response is true, the "correct" Action for such a situation, 
				or, if response if false, the "incorrect" Action for such a situation
**	\param		Response:	Whether or not Action is the "correct" AI_Unit to be listened to in 
							the situation described by Attributes
**	\param		Fitness:	The fitness function determined fitness (ignored)
	\param		Action:		If response is true, then action is the "correct" AI_Unit to listen to
							in the situation described by Attributes.  Otherwise, it is an 
							"incorrect" AI_Unit -- that is, it is an AI_Unit that should not be
							listened to in the situation described by Attributes.
**	\param		Attributes:	Describes every perceivable aspect of a given situation
**	\return		none
**	\exception	none
**	\par		DESCRIPTION:
**					
**	\author		Nathan Frost
**	\date		11/21/2002 1:36:47 PM
**	\warning	
				See warnings from BayesClassifier::LearnFromPositiveExample()
				See warnings from BayesClassifier::LearnFromNegativeExample()
*/
////////////////////////////////////////////////////////////////////////////////
void BayesADN::saveTrainingExample(	bool Response, float fitness, AI_ID Action, 
									std::vector<AI_Attribute>& Attributes)
{
	//if Action is the "correct" AI_Unit, given the situation described by Attributes, then the 
	//BayesClassifier will learn positively from this example (ie: "In a similar future situation,
	//be more likely to listen to this AI_Unit")
	if(Response)
	{
		//convert to internal datatype (from AI_ID to Action_ID)
		bayesClassifier.LearnFromPositiveExample(static_cast<Action_ID>(Action), Attributes);
	}
	//otherwise, the BayesClassifier should learn negatively from this example (ie: "In a similar
	//future situation be more likely to make any Choice other than Action"
	else
	{
		bayesClassifier.LearnFromNegativeExample(static_cast<Action_ID>(Action), Attributes);
	}
}


////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	AI_ADN::process
/*!	\brief		analyzes AI_Unit inputs, selects one as the action
**	\param		attributes:	A description of the current situation
**	\return		The action to be performed
**	\exception	none
**	\author		Nathan Frost
**	\date		10/24/2002 9:44:16 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
AI_ID BayesADN::process() throw()
{
	return bayesClassifier.MakeChoice(attributes);
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	DebugOut
/*!	\brief		Vomits forth the innards of the Bayesian Classifier to an ostream.
**	\param		Os:	ostream to vomit into
**	\return		none
**	\exception	none
**	\par		DESCRIPTION:
**					
**	\author		Nathan Frost
**	\date		11/21/2002 11:14:19 AM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
void BayesADN::DebugOut(ostream &Os)
{
	bayesClassifier.DebugOut(Os);
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	Read
/*!	\brief		Replaces the BayesClassifier stored in *this with the BayesClassifier stored in 
				InputStream.
**	\param		InputStream:	To read in the BayesClassifier from
**	\return		none
**	\exception	none
**	\par		DESCRIPTION:
**					
**	\author		Nathan Frost
**	\date		4/15/2003 6:56:49 PM
**	\warning	
				- InputStream must contain data from BayesADN::Write(), or else this function's
				behaviour is undefined
*/
////////////////////////////////////////////////////////////////////////////////
void BayesADN::Read(std::istream &InputStream)
{
	bayesClassifier.Read(InputStream);
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	Write
/*!	\brief		Writes out the BayesClassifier inside *this to OutputStream.
**	\param		OutputStream:	To receive the data from *this's BayesClassifier
**	\return		none
**	\exception	none
**	\par		DESCRIPTION:
**					
**	\author		Nathan Frost
**	\date		4/15/2003 6:56:49 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
void BayesADN::Write(std::ostream &OutputStream) const
{
	bayesClassifier.Write(OutputStream);
}

//deprecated
bool BayesADN::save() throw()
{
	return false;
}

//deprecated
bool BayesADN::load() throw()
{
	return false;
}
