////////////////////////////////////////////////////////////////////////////////
/*!	\file		BayesClassifierFileMemento.cpp
**	\brief		Implementation
**	\author		Nathan Frost
**	\date		2/28/2003 7:15:21 PM
**	\par		DESCRIPTION:
**					
**	\par		REMARKS:
**					
*/
////////////////////////////////////////////////////////////////////////////////


#include "BayesClassifier.h"
#include "BayesClassifierFileMemento.h"
using namespace Bayes;


////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	operator<< for BCFMRep
/*!	\brief		Overloaded insertion operator for inserting a BCFMRep into an ostream
**	\param		std::ostream &Os:	to insert the data
	\param		TheBCFMRep:			to get the data from
**	\return		none
**	\exception	none
**	\par		DESCRIPTION:
**					
**	\author		Nathan Frost
**	\date		3/1/2003 1:07:17 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
std::ostream &operator<<(std::ostream &Os, 
						 const Bayes::BayesClassifierFileMemento::BCFMRep &TheBCFMRep)
{
	Os.write(reinterpret_cast<const char *>(&TheBCFMRep), sizeof(TheBCFMRep));

	return Os;
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	operator>> for BCFMRep
/*!	\brief		Overloaded extraction operator for extracting a BCFMRep from an istream
**	\param		std::istream &Is:	to extract the data from
	\param		TheBCFMRep:			to put the data into
**	\return		none
**	\exception	none
**	\par		DESCRIPTION:
**					
**	\author		Nathan Frost
**	\date		3/1/2003 1:07:17 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
std::istream &operator>>(std::istream &Is, Bayes::BayesClassifierFileMemento::BCFMRep &TheBCFMRep)
{
	Is.read(reinterpret_cast<char *>(&TheBCFMRep), sizeof(TheBCFMRep));

	return Is;
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	PushBack
/*!	\brief		Pushes back a BCFMRep, and increments the total number of BCFMReps that have been
				pushed back
**	\param		numBCFMReps:	The total number of BCFMReps that have been pushed back so far
	\param		bcfmRep:		To push back
**	\return		none
**	\exception	none
**	\par		DESCRIPTION:
**					
**	\author		Nathan Frost
**	\date		3/1/2003 2:04:13 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
void BayesClassifierFileMemento::PushBack(int &numBCFMReps, BCFMRep &bcfmRep)
{
	++numBCFMReps;
	rep.push_back(bcfmRep);
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	MakeFileMemento
/*!	\brief		*this BayesClassifierFileMemento stores TheBayesClassifier's state
**	\param		TheBayesClassifier: to save
**	\return		none
**	\exception	none
**	\par		DESCRIPTION:
**					
**	\author		Nathan Frost
**	\date		2/28/2003 9:06:52 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
BayesClassifierFileMemento::MakeFileMemento(BayesClassifier &TheBayesClassifier)
{
	rep.clear();
	BCFMRep bcfmRep;

	int numBCFMReps = 0;

	//record the total number of training examples
	bcfmRep.theUnsignedInt = TheBayesClassifier.numTrainingExamples;
	PushBack(numBCFMReps, bcfmRep);
	
	//record the number of Choices
	bcfmRep.theUnsignedInt = TheBayesClassifier.choice.size();
	PushBack(numBCFMReps, bcfmRep);

	/* Record all Choice information */
	for(std::map<Action_ID, BayesClassifier::Choice>::iterator it = 
		TheBayesClassifier.choice.begin(); it != TheBayesClassifier.choice.end(); ++it)
	{
		//record this Choice's ID
		bcfmRep.theUnsignedInt = it->first;
		PushBack(numBCFMReps, bcfmRep);

		//record the number of times this Choice has occurred
		bcfmRep.theUnsignedInt = it->second.numOccurrenceChoice;
		PushBack(numBCFMReps, bcfmRep);

		//record the number of DataFilterKinds in this Choice
		bcfmRep.theUnsignedInt = it->second.dataFilterKind.size();
		PushBack(numBCFMReps, bcfmRep);

		/* Record all DataFilterKind information */
		for(std::map<unsigned int, BayesClassifier::DataFilterKind>::iterator jt = 
			it->second.dataFilterKind.begin(); jt != it->second.dataFilterKind.end(); ++jt)
		{
			//record this DataFilterKind's ID
			bcfmRep.theUnsignedInt = jt->first;
			PushBack(numBCFMReps, bcfmRep);

			//record the number of Categories in this DataFilterKind
			bcfmRep.theUnsignedInt = jt->second.category.size();
			PushBack(numBCFMReps, bcfmRep);

			/* Record all Category information */
			for(std::map<double, BayesClassifier::Category>::iterator kt = 
				jt->second.category.begin(); kt != jt->second.category.end(); ++kt)
			{
				//record the lower boundary of the Category
				bcfmRep.theDouble = kt->second.lowerBound;
				PushBack(numBCFMReps, bcfmRep);

				//record the upper boundary of the Category
				bcfmRep.theDouble = kt->second.upperBound;
				PushBack(numBCFMReps, bcfmRep);

				//record the number of times this Category has occurred
				bcfmRep.theUnsignedInt = kt->second.numOccurrences;
				PushBack(numBCFMReps, bcfmRep);
			}
		}
	}

	//record the total number of BCFMReps that comprised the BayesClassifier we just stored
	bcfmRep.theUnsignedInt = numBCFMReps;
	rep.push_front(bcfmRep);
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	Write
/*!	\brief		Outputs *this BayesClassifierFileMemento's internal state data to an ostream.
**	\param		OutputStream: to output the BayesClassifierFileMemento's internal state data to
**	\return		none
**	\exception	none
**	\par		DESCRIPTION:
**					
**	\author		Nathan Frost
**	\date		2/28/2003 9:37:39 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
void BayesClassifierFileMemento::Write(std::ostream &OutputStream) const
{
	for(std::deque<BCFMRep>::const_iterator it = rep.begin(); it != rep.end(); ++it)
		OutputStream << *it;
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	Dereference
/*!	\brief		Returns the BayesClassifier in the state it was stored in.
**	\param		none
**	\return		none
**	\exception	none
**	\par		DESCRIPTION:
**					
**	\author		Nathan Frost
**	\date		3/1/2003 11:37:34 AM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
BayesClassifier BayesClassifierFileMemento::Dereference() const
{
	BayesClassifier ret;
	std::deque<BCFMRep>::const_iterator it = rep.begin();

	//skip past the total number of BCFMReps that comprised the stored BayesClassifier
	++it;

	//retrieve the total number of training examples
	ret.numTrainingExamples = (*it).theUnsignedInt;
	++it;
	
	//retrieve the number of Choices
	unsigned int numChoices = (*it).theUnsignedInt;
	++it;

	/* Retrieve all Choice information */
	for(unsigned int i=0; i < numChoices; ++i)
	{
		BayesClassifier::Choice choice;

		//retrieve choice's ID -- note that this cast should always be OK
		AI_System::Action_ID choiceID = static_cast<AI_System::Action_ID>((*it).theUnsignedInt);
		++it;

		//retrieve the number of times this Choice has occurred
		choice.numOccurrenceChoice = (*it).theUnsignedInt;
		++it;

		//retrieve the number of DataFilterKinds in this Choice
		unsigned int numDataFilterKinds = (*it).theUnsignedInt;
		++it;

		/* Retrieve all DataFilterKind information */
		for(unsigned int j=0; j < numDataFilterKinds; ++j)
		{
			BayesClassifier::DataFilterKind dataFilterKind;

			//retrieve this DataFilterKind's ID
			unsigned int dataFilterKindID = (*it).theUnsignedInt;
			++it;

			//retrieve the number of Categories in this DataFilterKind
			unsigned int numCategories = (*it).theUnsignedInt;
			++it;

			/* Retrieve all Category information */
			for(unsigned int k=0; k < numCategories; ++k)
			{
				BayesClassifier::Category category;

				//retrieve the lower boundary of the Category
				category.lowerBound = (*it).theDouble;
				++it;

				//record the upper boundary of the Category
				category.upperBound = (*it).theDouble;
				++it;

				//retrieve the number of times this Category has occurred
				category.numOccurrences = (*it).theUnsignedInt;
				++it;

				dataFilterKind.category.insert(std::make_pair(category.lowerBound, category));
			}

			choice.dataFilterKind.insert(std::make_pair(dataFilterKindID, dataFilterKind));
		}

		ret.choice.insert(std::make_pair(choiceID, choice));
	}

	return ret;
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	BayesClassifierFileMemento
/*!	\brief		Constructor
**	\param		InputStream: istream to pass to the base constructor
**	\return		none
**	\exception	none
**	\par		DESCRIPTION:
**					
**	\author		Nathan Frost
**	\date		3/1/2003 1:08:51 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
BayesClassifierFileMemento::BayesClassifierFileMemento(std::istream &InputStream):
FileMemento(InputStream)
{
	;
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	Read
/*!	\brief		Reads the stored state of a BayesClassifier into *this BayesClassifierFileMemento
**	\param		InputStream: the stream to read from
**	\return		none
**	\exception	none
**	\par		DESCRIPTION:
**					- any previous BayesClassifier state will be overwritten

**	\author		Nathan Frost
**	\date		3/1/2003 3:20:11 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
void BayesClassifierFileMemento::Read(std::istream &InputStream)
{
	BCFMRep bcfmRep;

	InputStream >> bcfmRep;
	rep.push_back(bcfmRep);

	int numBCFMReps = bcfmRep.theUnsignedInt;
	for(int i=0; i < numBCFMReps; ++i)
	{
		InputStream >> bcfmRep;
		rep.push_back(bcfmRep);
	}
}


