////////////////////////////////////////////////////////////////////////////////
/*!	\file		BayesClassifier.cpp
**	\brief		Implementation
**	\author		Nathan Frost
**	\date		11/7/2002 8:40:24 PM
**	\par		DESCRIPTION:
**					
**	\par		REMARKS:
					- Action_ID MUST be 32bit integral datatypes.
**					
*/
////////////////////////////////////////////////////////////////////////////////
#include "BayesClassifier.h"
//#include <iostream>
//#include"../AI_System.h"
#include <algorithm>
#include <limits>

using namespace Bayes;


////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	Category::Category
/*!	\brief		Constructor
**	\param		LowerBound:		The lowest floating point value an Attribute can have while still
								being considered in this classification.
	\param		UpperBound:		The highest floating point value an Attribute can have while still
								being considered in this classification.
	\param		NumOccurrences:	The number of times this classification has been encountered in
								past experiences.  PASSING A NONZERO VALUE HERE IS UNDEFINED.
**	\return		none
**	\exception	none
**	\par		DESCRIPTION:
**					
**	\author		Nathan Frost
**	\date		11/7/2002 8:40:53 PM
**	\warning	See parameter list.
*/
////////////////////////////////////////////////////////////////////////////////
BayesClassifier::Category::Category(double LowerBound, double UpperBound,
									unsigned int NumOccurrences):
lowerBound(LowerBound), upperBound(UpperBound), numOccurrences(NumOccurrences)
{
	;
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	DataFilterKind::DataFilterKind
/*!	\brief		Constructor
**	\param		none
**	\return		none
**	\exception	none
**	\par		DESCRIPTION:
**					
**	\author		Nathan Frost
**	\date		11/7/2002 8:44:16 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
BayesClassifier::DataFilterKind::DataFilterKind()
{
	;
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	Choice::Choice
/*!	\brief		Constructor
**	\param		NumOccurrenceChoice:	The number of times this Choice has been deemed correct in
										past experiences.  PASSING A NONZERO VALUE HERE IS 
										UNDEFINED.
**	\return		none
**	\exception	none
**	\par		DESCRIPTION:
**					
**	\author		Nathan Frost
**	\date		11/7/2002 8:44:42 PM
**	\warning	See parameter list.
*/
////////////////////////////////////////////////////////////////////////////////
BayesClassifier::Choice::Choice(unsigned int NumOccurrenceChoice):
numOccurrenceChoice(NumOccurrenceChoice)
{
	;
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	BayesClassifier
/*!	\brief		Constructor -- the BayesClassifier begins its life in a trivial, but valid, state
**	\param		none
**	\return		none
**	\exception	none
**	\par		DESCRIPTION:
**					
**	\author		Nathan Frost
**	\date		11/7/2002 8:46:13 PM
**	\warning	See parameter list.
*/
////////////////////////////////////////////////////////////////////////////////
BayesClassifier::BayesClassifier():
numTrainingExamples(0){}


////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	HaveChoice
/*!	\brief		Determines if the Bayesian Classifier has a specific Choice object.
**	\param		ChoiceID: Unique identifier of the Choice object to check for.
**	\return		True if the Bayesian Classifier already has the specified Choice object, false if
				not.
**	\exception	none
**	\par		DESCRIPTION:
**					
**	\author		Nathan Frost
**	\date		11/7/2002 8:49:20 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
bool BayesClassifier::HaveChoice(Action_ID ChoiceID)
{
	std::map<Action_ID, Choice>::iterator it = choice.find(ChoiceID);
	if(it == choice.end())
		return false;
	else
		return true;
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	AddChoice
/*!	\brief		Allows the Bayesian Classifier to consider the specified Choice, in addition to 
				whatever other Choices it can already consider.
**	\param		ID:				Unique identifier of the Action the Choice corresponds to (in other
								words, the Action that will be recommended if this Choice is chosen)
	\param		NumOccurrences:	The number of times this Choice has been deemed correct in past
								experiences.  PASSING A NONZERO VALUE HERE IS UNDEFINED.
**	\return		none
**	\exception	Throws an exception if the Choice object to be added already exists in the Bayesian
				Classifier.
**	\par		DESCRIPTION:
**					
**	\author		Nathan Frost
**	\date		11/7/2002 8:50:55 PM
**	\warning	See parameter list.
*/
////////////////////////////////////////////////////////////////////////////////
void BayesClassifier::AddChoice(Action_ID ID, unsigned int NumOccurrences)
{
	if(HaveChoice(ID))
	{
		throw Nexception(	"BayesClassifier::AddChoice() already has the specified Choice", 
							__FILE__, __LINE__, __DATE__, __TIME__);
	}
	choice.insert(std::make_pair(ID, Choice(NumOccurrences)));
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	RemoveChoice
/*!	\brief		Disallows the Bayesian Classifier to consider the specified Choice.
**	\param		ID: Unique identifier of the Action the Choice corresponds to (in other words, the
					Bayesian Classifier will no longer be able to recommend this Action)
**	\return		none
**	\exception	Throws a Nexception if the Choice object to be removed does not exist in the 
				Bayesian Classifier.
**	\par		DESCRIPTION:
**					
**	\author		Nathan Frost
**	\date		11/7/2002 8:54:11 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
void BayesClassifier::RemoveChoice(Action_ID ID)
{
	if(!HaveChoice(ID))
	{
		throw Nexception(	"BayesClassifier::RemoveChoice() couldn't find the specified Choice", 
							__FILE__, __LINE__, __DATE__, __TIME__);
	}

	choice.erase(ID);
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	HaveDataFilterKind
/*!	\brief		Determines whether or not the Bayesian Classifier has the specified DataFilterKind
				object (in other words, whether the Bayesian Classifier considers the Attribute 
				that corresponds to the specified DataFilterKind object when deciding whether or 
				not to choose the given Choice)
**	\param		ChoiceID:			Unique identifier of the Choice that the DataFilterKind object 
									resides in
				DataFilterKindID:	Unique identifier of the DataFilterKind that resides in the 
									given ChoiceID (that is, the Attribute that the Bayesian 
									Classifier considers when deciding whether or not to choose the
									given Choice)
**	\return		True if the specified DataFilter exists in the specified Choice of the Bayesian
				Classifier; false if it does not.
**	\exception	Throws a Nexception if the specified Choice object does not exist in the Bayesian
				Classifier.
**	\par		DESCRIPTION:
**					
**	\author		Nathan Frost
**	\date		11/7/2002 8:56:34 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
bool BayesClassifier::HaveDataFilterKind(Action_ID ChoiceID, unsigned int DataFilterKindID)
{
	if(!HaveChoice(ChoiceID))
	{
		throw Nexception(	"BayesClassifier::HaveDataFilterKind() couldn't find the specified Choice.",
							__FILE__, __LINE__, __DATE__, __TIME__);
	}

	/* find() the specified Choice, then try to find() the specified DataFilterKind */
	std::map<Action_ID, Choice>::iterator it = choice.find(ChoiceID);
	std::map<unsigned int, DataFilterKind>::iterator jt = 
		it->second.dataFilterKind.find(DataFilterKindID);
	
	if(jt == it->second.dataFilterKind.end())
		return false;
	else
		return true;
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	AddDataFilterKindToChoice
/*!	\brief		Allows the Bayesian Classifier to, when considering whether or not to choose a
				Choice, consider the Attribute that corresponds to the specified DataFilterKind
**	\param		ChoiceID:			The Choice to associate with the specified DataFilterKind
	\param		DataFilterKindID:	The Attribute that the Bayesian Classifier should start 
									considering when contemplating the specified Choice
**	\return		none
**	\exception	Throws a Nexception if the specified Choice object does not exist in the Bayesian
				Classifier, or if the specified DataFilterKind object already exists in the
				Bayesian Classifier.
**	\par		DESCRIPTION:
**					
**	\author		Nathan Frost
**	\date		11/7/2002 9:20:50 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
void BayesClassifier::AddDataFilterKindToChoice(Action_ID ChoiceID, unsigned int DataFilterKindID)
{
	if(HaveDataFilterKind(ChoiceID, DataFilterKindID))
	{
		throw Nexception(	"BayesClassifier::AddDataFilterKindToChoice() already has the specified DataFilterKind.", 
							__FILE__, __LINE__, __DATE__, __TIME__);
	}

	//since we're already assured of having the specified DataFilterKind and Choice, there's no
	//danger of accidentally inserting a blank Choice into the Bayesian Classifier
	choice[ChoiceID].dataFilterKind[DataFilterKindID] = DataFilterKind();
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	RemoveDataFilterKindFromChoice
/*!	\brief		Disallows the Bayesian Classifier to, when considering whether or not to choose a
				Choice, consider the Attribute that corresponds to the specified DataFilterKind
**	\param		ChoiceID:			The Choice to disassociate with the specified DataFilterKind
	\param		DataFilterKindID:	The Attribute that the Bayesian Classifier should stop 
									considering when contemplating the specified Choice	
**	\return		none
**	\exception	Throws a Nexception if either the specified Choice object or the specified  
				DataFilterKind object does not exist in the Bayesian Classifier
**	\par		DESCRIPTION:
**					
**	\author		Nathan Frost
**	\date		11/7/2002 9:25:40 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
void BayesClassifier::RemoveDataFilterKindFromChoice(	Action_ID ChoiceID, 
														unsigned int DataFilterKindID)
{
	if(!HaveDataFilterKind(ChoiceID, DataFilterKindID))
	{
		throw Nexception(	"BayesClassifier::RemoveDataFilterKindFromChoice() couldn't find the specified DataFilterKind.", 
							__FILE__, __LINE__, __DATE__, __TIME__);
	}
	
	//since we're already assured of having the specified DataFilterKind and Choice, there's no
	//danger of accidentally inserting a blank Choice
	choice[ChoiceID].dataFilterKind.erase(DataFilterKindID);
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	AddOrReplaceCategoryToDataFilterKindInChoice
/*!	\brief		Defines a classification for the specified DataFilterKind in the specified Choice.
**	\param		ChoiceID:			The Choice that the specified DataFilterKind resides in
	\param		DataFilterKindID:	The Attribute that is getting a new classification
**	\param		LowerBound:			The lowest floating point value an Attribute can have while 
									still being considered in this classification
	\param		UpperBound:			The highest floating point value an Attribute can have while 
									still being considered in this classification
**	\return		none
**	\exception	
				- Throws a Nexception if either the specified Choice object or the specified  
				DataFilterKind object does not exist in the Bayesian Classifier.
				- Throws a Nexception if the specified DataFilterKind's range (see the class
				definition of DataFilterKind for details) becomes invalid as a result of the 
				calling of this function
**	\par		DESCRIPTION:
**					
**	\author		Nathan Frost
**	\date		11/7/2002 9:28:28 PM
**	\warning	
				- Note that if you pass an a Category that has the same lowerBound as an already
				existing Category, but the argument Category has a different upperBound than the
				pre-existing Category, the pre-existing Category will be overwritten by the
				argument Category.  For example, if (0.5, 1.0) already exists in DataFilterKind 4
				of Choice 3, and you AddOrReplaceCategoryToDataFilterKindInChoice(3, 4, 0.5, 1.1),
				then (0.5, 1.0) will be replaced by (0.5, 1.1).
*/
////////////////////////////////////////////////////////////////////////////////
void BayesClassifier::AddOrReplaceCategoryToDataFilterKindInChoice(	Action_ID ChoiceID, 
																	unsigned int DataFilterKindID, 
																	double LowerBound, 
																	double UpperBound)
{
	if(!HaveDataFilterKind(ChoiceID, DataFilterKindID))
	{
		throw Nexception(	"BayesClassifier::AddOrReplaceCategoryToDataFilterKindInChoice() couldn't find the specified DataFilterKind.", 
							__FILE__, __LINE__, __DATE__, __TIME__);
	}

	//since we're already assured of having the specified DataFilterKind and Choice, there's no
	//danger of accidentally inserting a blank Choice or DataFilterKind into the Bayesian
	//Classifier
	choice[ChoiceID].dataFilterKind[DataFilterKindID].category[LowerBound] = 
																Category(LowerBound, UpperBound);

	if(!Validate(ChoiceID, DataFilterKindID))
	{
		throw Nexception(	"BayesClassifier::AddOrReplaceCategoryToDataFilterKindInChoice() reports that a DataFilterKind's categories are invalid.",
							__FILE__, __LINE__, __DATE__, __TIME__);
	}
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	RemoveCategoryFromDataFilterKindInChoice
/*!	\brief		Removes a classification for the specified DataFilterKind in the specified Choice.
**	\param		ChoiceID:			The Choice that the specified DataFilterKind resides in
	\param		DataFilterKindID:	The Attribute that is losing a classification
**	\param		LowerBound:			The lowest floating point value an Attribute can have while 
									still being considered in the classification to be removed
	\param		UpperBound:			The highest floating point value an Attribute can have while 
									still being considered in the classification to be removed
**	\return		none
**	\exception	- Throws a Nexception if either the specified Choice object or the specified  
				DataFilterKind object does not exist in the Bayesian Classifier
				- Throws a Nexception if the specified DataFilterKind's range (see DESCRIPTION for
				details) becomes invalid as a result of the calling of this function
**	\par		DESCRIPTION:
**					
**	\author		Nathan Frost
**	\date		11/7/2002 9:34:20 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
void BayesClassifier::RemoveCategoryFromDataFilterKindInChoice(	Action_ID ChoiceID, 
																unsigned int DataFilterKindID, 
																double LowerBound, 
																double UpperBound)
{
	if(!HaveDataFilterKind(ChoiceID, DataFilterKindID))
	{
		throw Nexception(	"BayesClassifier::RemoveCategoryFromDataFilterKindInChoice() couldn't find the specified DataFilterKind.", 
							__FILE__, __LINE__, __DATE__, __TIME__);
	}

	/* find() the specified Choice, then find() the specified DataFilterKind, then try and find() */
	/* the specified Category */
	std::map<Action_ID, Choice>::iterator it = choice.find(ChoiceID);
	std::map<unsigned int, DataFilterKind>::iterator jt = 
												it->second.dataFilterKind.find(DataFilterKindID);
	std::map<double, Category>::iterator kt = jt->second.category.find(LowerBound);

	//make sure that we really found the Category -- Categories are defined by their lowerBound
	//and upperBound
	if(kt == jt->second.category.end() || kt->second.upperBound != UpperBound)
	{
		throw Nexception(	"BayesClassifier::RemoveCategoryFromDataFilterKindInChoice() couldn't find the specified Category.", 
							__FILE__, __LINE__, __DATE__, __TIME__);		
	}

	jt->second.category.erase(LowerBound);
	if(!Validate(ChoiceID, DataFilterKindID))
	{
		throw Nexception(	"BayesClassifier::RemoveCategoryFromDataFilterKindInChoice() reports that a DataFilterKind's categories are invalid.",
							__FILE__, __LINE__, __DATE__, __TIME__);
	}
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	Validate
/*!	\brief		Checks to see if the specified DataFilterKind's range is valid.
**	\param		ChoiceID:			The Choice that the specified DataFilterKind resides in
	\param		DataFilterKindID:	The DataFilterKind whose range should be checked for validity
**	\return		True if the specified DataFilterKind's range is valid, false if it is invalid.
**	\exception	- Throws a Nexception if either the specified Choice object or the specified  
				DataFilterKind object does not exist in the Bayesian Classifier
**	\par		DESCRIPTION:
**					
**	\author		Nathan Frost
**	\date		11/7/2002 10:04:51 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
bool BayesClassifier::Validate(Action_ID ChoiceID, unsigned int DataFilterKindID)
{
	if(!HaveDataFilterKind(ChoiceID, DataFilterKindID))
	{
		throw Nexception(	"BayesClassifier::Validate() couldn't find the specified DataFilterKind.", 
							__FILE__, __LINE__, __DATE__, __TIME__);
	}

	/* Since we're already assured of having the specified DataFilterKind and Choice, there's no */
	/* danger of accidentally inserting a blank Choice or DataFilterKind into the Bayesian */
	/* Classifier */
	
	//check to see if the Category container is empty
	if(choice[ChoiceID].dataFilterKind[DataFilterKindID].category.empty())
		return true;//the Null range is considered to be a valid range

	//make sure the first Category's lowerBound is less than its upperBound
	std::map<double, Category>::iterator cur = 
		choice[ChoiceID].dataFilterKind[DataFilterKindID].category.begin();
	if(cur->second.lowerBound >= cur->second.upperBound)
	{
		return false;
	}
	//initialize next to point to the Category after the Category pointed to by cur
	std::map<double, Category>::iterator next = cur;
	++next;
	//each iteration, increment both cur and next
	//stop when next points to the end() of the Category container
	for(; next != choice[ChoiceID].dataFilterKind[DataFilterKindID].category.end(); ++cur, ++next)
	{
		if(next->second.lowerBound >= next->second.upperBound)
		{
			return false;
		}
		//check that cur->upperBound == next->lowerBound (exempting the last Category in the container)
		if(cur->second.upperBound != next->second.lowerBound)
		{
			return false;
		}
	}

	return true;
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	Categorize
/*!	\brief		Returns the lowerBound of the Category that value falls into
**	\param		ChoiceID:			The Choice that the specified DataFilterKind resides in
	\param		DataFilterKindID:	The DataFilterKind to use to classify the specified Value
	\param		Value:				The Value to be classified, relative to the specified 
									DataFilterKind
**	\return		lowerBound of the Category (which resides in the specified DataFilterKind, which
				resides in the specified Choice) that the specified Value fell into
**	\exception	- Throws a Nexception if either the specified Choice object or the specified  
				DataFilterKind object does not exist in the Bayesian Classifier
**	\par		DESCRIPTION:
**					
**	\author		Nathan Frost
**	\date		11/8/2002 10:13:46 AM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
double BayesClassifier::Categorize(Action_ID ChoiceID, unsigned int DataFilterKindID, double Value)
{
	if(!HaveDataFilterKind(ChoiceID, DataFilterKindID))
	{
		throw Nexception(	"BayesClassifier::Validate() couldn't find the specified DataFilterKind.", 
							__FILE__, __LINE__, __DATE__, __TIME__);
	}

	/* Since we're already assured of having the specified DataFilterKind and Choice, there's no */
	/* danger of accidentally inserting a blank Choice or DataFilterKind into the Bayesian */
	/* Classifier */

	std::map<double, Category>::iterator it = 
								choice[ChoiceID].dataFilterKind[DataFilterKindID].category.begin();
	//if the DataFilterKind has the Null range, then the Value is given the trivial classification
	if(it == choice[ChoiceID].dataFilterKind[DataFilterKindID].category.end())
		return Value;
	//if the Value is smaller than the lowerBound of the first (lowest ranged) Category, clamp it
	//to the first (lowest ranged) Category
	if(Value < it->second.lowerBound)
		return it->second.lowerBound;

	//find the Category the Value falls into, if any
	for(; it != choice[ChoiceID].dataFilterKind[DataFilterKindID].category.end(); ++it)
	{
		//lowerBound inclusive, upperBound exclusive
		if(Value >= it->second.lowerBound && Value < it->second.upperBound)
		{
			return it->second.lowerBound;
		}
	}
	
	//if the Value is larger than the upperBound of the last (highest ranged) Category, clamp it
	//to the last (highest ranged) Category
	--it;
	return it->second.lowerBound;
}


////////////////////////////////////////////////////////////////////////////////
/*!	\class		ChoiceProbability
**	\brief		Encapsulates a Choice's identifier and its computed probability.  Helper class.
**	\author		Nathan Frost
**	\date		11/8/2002 5:28:19 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
struct ChoiceProbability
{
	Action_ID id;
	double probability;

	ChoiceProbability(	Action_ID ID = static_cast<Action_ID>(0), 
						double Probability = DBL_MAX):
	id(ID), probability(Probability){}
};

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	MakeChoice
/*!	\brief		Given the current situation (described by the base member attributes), the Bayesian
				Classifier considers its past experiences and decides on one of its Choices.
**	\param		Attributes:	Describe every perceivable aspect of the current situation
**	\return		The ID of the Action that corresponds to the chosen Choice.
**	\exception	none
**	\par		DESCRIPTION:
					- Currently, the Bayesian Classifier's Certainty value is ignored.
**					
**	\author		Nathan Frost
**	\date		11/8/2002 5:26:36 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
Action_ID BayesClassifier::MakeChoice(vector<AI_Attribute> &Attributes)
{
	std::vector<ChoiceProbability> choiceProbability(choice.size());

	/* Calculate the probability of every available Choice */
	int i=0;
    for(std::map<Action_ID, Choice>::iterator it = choice.begin(); it != choice.end(); ++it, ++i)
    {
		//calculate the percentage of times this choice has been deemed to be true
		choiceProbability[i].id = it->first;
        choiceProbability[i].probability =  (it->second.numOccurrenceChoice + 1.0) /
											(numTrainingExamples + choice.size());

		//add the sum of the probabilities of this choice being true given every perceived aspect
		//of the current situation
        for(unsigned int j=0; j < Attributes.size(); ++j)//every aspect in the current situation
        {
			//(numberOfOccurrencesOfAttribute) + 1 / (numberOfOccurrencesOfChoice + numberOfChoices)
			choiceProbability[i].probability *= 
(static_cast<double>(it->second.dataFilterKind[Attributes[j].ID].category[Categorize(it->first, Attributes[j].ID, Attributes[j].value)].numOccurrences)
					+ 1.0) /
					(static_cast<double>(it->second.numOccurrenceChoice) + 
					static_cast<double>(it->second.dataFilterKind[Attributes[j].ID].category.size()));
        }
    }

	/* Pick the Choice that is most probable */
	Action_ID decision;
	double biggestProbability = -DBL_MAX;
	double sumOfProbabilities = 0;
	for(unsigned int j=0; j < choiceProbability.size(); ++j)
	{
		if(choiceProbability[j].probability > biggestProbability)
		{
			biggestProbability = choiceProbability[j].probability;
			decision = choiceProbability[j].id;
		}

		//sum together the choice's probability (so we can calculate Certainty)
		sumOfProbabilities += choiceProbability[j].probability;
	}
	//generate Bayesian Classifier's confidence in its decision
    double certainty = biggestProbability / sumOfProbabilities;

	return decision;
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	LearnFromPositiveExample
/*!	\brief		Records a new, positive past experience
**	\param		CorrectChoice:	The Action_ID of the "correct" action, given a situation described
								in Attributes.
	\param		Attributes:		Describes every perceivable aspect of a situation in which 
								CorrectChoice identifies the "correct" action
**	\return		none
**	\exception	none
**	\par		DESCRIPTION:
**					
**	\author		Nathan Frost
**	\date		11/20/2002 6:50:38 PM
**	\warning	
				The Bayesian Classifier must have CorrectChoice as well as every DataFilterKind 
				listed in Attributes present in its datastructures.  Otherwise, these
				datastructures will be (erroneously) added, and the Bayesian Classifier's behaviour
				will become undefined.

  				BUG: If BC is presented a DataFilterKind in Attributes that it doesn't have in
					CorrectChoice) -- UNDEFINED BEHAVIOR RESULTS.
*/
////////////////////////////////////////////////////////////////////////////////
void BayesClassifier::LearnFromPositiveExample(	AI_System::Action_ID CorrectChoice, 
												std::vector<AI_System::AI_Attribute>& Attributes)
{
	++numTrainingExamples;

	//record which choice is "correct" in this situation
	++choice[CorrectChoice].numOccurrenceChoice;

	//record the presence of every aspect of the situation that should elicit CorrectChoice
	for(std::vector<AI_System::AI_Attribute>::iterator it = Attributes.begin(); 
		it != Attributes.end(); ++it)
	{
		++choice[CorrectChoice].dataFilterKind[it->ID].category[Categorize(	CorrectChoice, 
																			it->ID, it->value)].
																			numOccurrences;
	}
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	LearnFromNegativeExample
/*!	\brief		Records a new, negative past experience
**	\param		IncorrectChoice:The Action_ID of the "incorrect" action, given a situation 
								described in Attributes.
	\param		Attributes:		Describes every perceivable aspect of a situation in which 
								IncorrectChoice identifies the "incorrect" action
**	\return		none
**	\exception	none
**	\par		DESCRIPTION:
**					
**	\author		Nathan Frost
**	\date		11/26/2002 6:47:21 PM
**	\warning	
				The Bayesian Classifier must have IncorrectChoice as well as every DataFilterKind 
				listed in Attributes present in its datastructures.  Otherwise, these
				datastructures will be (erroneously) added, and the Bayesian Classifier's behaviour
				will become undefined.

				BUG: If BC is presented a DataFilterKind in Attributes that it doesn't have in every
					Choice (except IncorrectChoice) -- UNDEFINED BEHAVIOR RESULTS.
*/
////////////////////////////////////////////////////////////////////////////////
void BayesClassifier::LearnFromNegativeExample(	AI_System::Action_ID IncorrectChoice, 
												std::vector<AI_System::AI_Attribute>& Attributes)
{
	++numTrainingExamples;
	
	/*	Maybe do any choice other than IncorrectChoice in the situation described by Attributes */
	//in similar situations, increase the likelihood of picking every choice that was not "incorrect"
	for(std::map<Action_ID, Choice>::iterator jt = choice.begin(); jt != choice.end(); ++jt)
	{
		if(jt->first == IncorrectChoice)
			continue;
		
		//make "not incorrect" choices more likely to occur in a similar situation
		++jt->second.numOccurrenceChoice;
		//record the presence of every aspect of the situation that perhaps should elicit the Choice
		for(std::vector<AI_System::AI_Attribute>::iterator kt = Attributes.begin(); 
			kt != Attributes.end(); ++kt)
		{
			++jt->second.dataFilterKind[kt->ID].category[Categorize(jt->first, kt->ID, kt->value)].
																	numOccurrences;
		}
	}
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	DebugOut
/*!	\brief		Vomits forth the innards of the Bayesian Classifier to an ostream.
**	\param		Os:	ostream to vomit into
**	\return		none
**	\exception	none
**	\par		DESCRIPTION:
**					
**	\author		Nathan Frost
**	\date		11/8/2002 9:24:09 AM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
void BayesClassifier::DebugOut(ostream &Os) const
{
	using std::endl;


	Os << "--------------------------------------------Begin BayesClassifier" << endl;
	Os << "numTrainingExamples: " << numTrainingExamples << endl;
	for(std::map<Action_ID, Choice>::const_iterator it = choice.begin(); it != choice.end(); ++it)
	{
		Os << "	Choice " << it->first << endl;
		Os << "		has occurred " << it->second.numOccurrenceChoice << " times" << endl;

		for(std::map<unsigned int, 
			DataFilterKind>::const_iterator jt = it->second.dataFilterKind.begin(); 
			jt != it->second.dataFilterKind.end(); ++jt)
		{
			Os << "			DataFilter " << jt->first << endl;

			int i=0;
			for(std::map<double, Category>::const_iterator kt = jt->second.category.begin(); 
				kt != jt->second.category.end(); ++kt, ++i)
			{
				Os << "				Category " << i << endl;
				Os << "					lowerBound: " << kt->second.lowerBound << endl;
				Os << "					upperBound: " << kt->second.upperBound << endl;
				Os << "					has occurred " <<	kt->second.numOccurrences << 
																" times" << endl;
			}
		}
	}

	Os << "---------------------------------------------End Bayes Classifier" << endl << endl;
}



void BayesClassifier::DebugSaveNLoad()
{
//#if 0//testing ChoiceIDs and DataFilterKindIDs
	for(int i=1; i < 4; ++i)
	{
		AddChoice(static_cast<AI_System::Action_ID>(i));
		AddDataFilterKindToChoice(static_cast<AI_System::Action_ID>(i), i*600);
		AddOrReplaceCategoryToDataFilterKindInChoice(	static_cast<AI_System::Action_ID>(i),  
														i*600, 0.0, 0.5);
		AddOrReplaceCategoryToDataFilterKindInChoice(	static_cast<AI_System::Action_ID>(i),
														i*600, 0.5, 1.0);
		AddDataFilterKindToChoice(static_cast<AI_System::Action_ID>(i), i*650);
		AddOrReplaceCategoryToDataFilterKindInChoice(	static_cast<AI_System::Action_ID>(i), 
														i*650, 0.0, 0.5);
		AddOrReplaceCategoryToDataFilterKindInChoice(	static_cast<AI_System::Action_ID>(i),
														i*650, 0.5, 1.0);
		AddDataFilterKindToChoice(static_cast<AI_System::Action_ID>(i), i*700);
		AddOrReplaceCategoryToDataFilterKindInChoice(	static_cast<AI_System::Action_ID>(i), 
														i*700, 0.0, 0.5);
		AddOrReplaceCategoryToDataFilterKindInChoice(	static_cast<AI_System::Action_ID>(i), 
														i*700, 0.5, 1.0);
	}
//#endif


#if 0//testing all except ChoiceIDs and DataFilterKindIDs
	for(int i=0; i < 3; ++i)
	{
		AddChoice(static_cast<AI_System::Action_ID>(i));
		
		for(int j=0; j < 4; ++j)
		{
			AddDataFilterKindToChoice(static_cast<AI_System::Action_ID>(i), j);
			
			AddOrReplaceCategoryToDataFilterKindInChoice(	static_cast<AI_System::Action_ID>(i), 
															j, 0.0 + i*100 + j, 0.2 + i*100 + j);
			AddOrReplaceCategoryToDataFilterKindInChoice(	static_cast<AI_System::Action_ID>(i), 
															j, 0.2 + i*100 + j, 0.4 + i*100 + j);
			AddOrReplaceCategoryToDataFilterKindInChoice(	static_cast<AI_System::Action_ID>(i), 
															j, 0.4 + i*100 + j, 0.6 + i*100 + j);
			AddOrReplaceCategoryToDataFilterKindInChoice(	static_cast<AI_System::Action_ID>(i), 
															j, 0.6 + i*100 + j, 0.8 + i*100 + j);
			AddOrReplaceCategoryToDataFilterKindInChoice(	static_cast<AI_System::Action_ID>(i), 
															j, 0.8 + i*100 + j, 1.0 + i*100 + j);
		}
	}

	numTrainingExamples = 666;

	for(int a=0; a < 3; ++a)
	{
		choice[static_cast<AI_System::Action_ID>(a)].numOccurrenceChoice = a + 1000;
		
		for(int j=0; j < 4; ++j)
		{
			choice[static_cast<AI_System::Action_ID>(a)].dataFilterKind[j].category[0.0 + a*100 + j].numOccurrences = a+10*j;
			choice[static_cast<AI_System::Action_ID>(a)].dataFilterKind[j].category[0.2 + a*100 + j].numOccurrences = a+10*j + 100;
			choice[static_cast<AI_System::Action_ID>(a)].dataFilterKind[j].category[0.4 + a*100 + j].numOccurrences = a+10*j + 200;
			choice[static_cast<AI_System::Action_ID>(a)].dataFilterKind[j].category[0.6 + a*100 + j].numOccurrences = a+10*j + 300;
			choice[static_cast<AI_System::Action_ID>(a)].dataFilterKind[j].category[0.8 + a*100 + j].numOccurrences = a+10*j + 400;
		}
	}
#endif
}



#if 0//debug crap code
Action_ID BayesClassifier::MakeChoice()
{
	std::vector<ChoiceProbability> choiceProbability(choice.size());

	/* Calculate the probability of every available Choice */
	int i=0;
    for(std::map<Action_ID, Choice>::iterator it = choice.begin(); it != choice.end(); ++it, ++i)
    {
		//calculate the percentage of times this choice has been deemed to be true
		choiceProbability[i].id = it->first;
        choiceProbability[i].probability =  (it->second.numOccurrenceChoice + 1.0) /
											(numTrainingExamples + choice.size());
        

		/**donkeydebug**
		std::cout << "starting probability: " << choiceProbability[i].probability << std::endl;
		**donkeydebug**/


		//add the sum of the probabilities of the choice being true given every perceived aspect
		//of the current situation
		
		
		/**donkeydebug**
		//Synthesize attributes from TDDExample
		enum dataFilterz {SIZE, SPEED, TEETH};
		std::vector<AI_System::AI_Attribute> attributes;
		AI_System::AI_Attribute attr;
		attr.ID = SIZE;
		attr.value = static_cast<float>(0.2);
		attributes.push_back(attr);
		attr.ID = SPEED;
		attr.value = static_cast<float>(0.6);
		attributes.push_back(attr);
		attr.ID = TEETH;
		attr.value = static_cast<float>(0.9);
		attributes.push_back(attr);
		**donkeydebug**/


        for(int j=0; j < attributes.size(); ++j)//every aspect in the current situation
        {


			/**donkeydebug**
			std::cout << std::endl;

			std::cout << "Category Probability: " << 
(static_cast<double>(it->second.dataFilterKind[attributes[j].ID].category[Categorize(it->first, attributes[j].ID, attributes[j].value)].numOccurrences)
					+ 1.0) /
					(static_cast<double>(it->second.numOccurrenceChoice) + 
					static_cast<double>(it->second.dataFilterKind[attributes[j].ID].category.size()))
					<< std::endl;


			std::cout << std::endl;
			**donkeydebug**/


            choiceProbability[i].probability *= 
(static_cast<double>(it->second.dataFilterKind[attributes[j].ID].category[Categorize(it->first, attributes[j].ID, attributes[j].value)].numOccurrences)
					+ 1.0) /
					(static_cast<double>(it->second.numOccurrenceChoice) + 
					static_cast<double>(it->second.dataFilterKind[attributes[j].ID].category.size()));
        }
    }

	/* Pick the Choice that is most probable */
	Action_ID decision;
	double biggestProbability = -std::numeric_limits<double>::max();
	double sumOfProbabilities = 0;
	for(int j=0; j < choiceProbability.size(); ++j)
	{
		if(choiceProbability[j].probability > biggestProbability)
		{
			biggestProbability = choiceProbability[j].probability;
			decision = choiceProbability[j].id;
		}

		//sum together the choice's probability
		sumOfProbabilities += choiceProbability[j].probability;


		/**donkeydebug**
		std::cout << "choice Probability: " << choiceProbability[j].probability << std::endl;
		**donkeydebug**/


	}
	//generate Bayesian Classifier's confidence in its decision
    double certainty = biggestProbability / sumOfProbabilities;

	
	/**donkeydebug**
	std::cout << "decision: " << decision << std::endl;
	std::cout << "certainty: " << certainty << std::endl;
	**donkeydebug**/


	return decision;
}


void BayesClassifier::SetUpUsTheTDDExample()
{
	/* Construct the datastructures */
	enum choicez {NOT_PREDATOR, PREDATOR};
	enum dataFilterz {SIZE, SPEED, TEETH};

	AddChoice(static_cast<Action_ID>(NOT_PREDATOR), 7);//hack the numOccurrences
	AddDataFilterKindToChoice(static_cast<Action_ID>(NOT_PREDATOR), SIZE);
	AddOrReplaceCategoryToDataFilterKindInChoice(	static_cast<Action_ID>(NOT_PREDATOR), SIZE, 
													0.0, 0.33);
	AddOrReplaceCategoryToDataFilterKindInChoice(	static_cast<Action_ID>(NOT_PREDATOR), SIZE, 
													0.33, 0.66);
	AddOrReplaceCategoryToDataFilterKindInChoice(	static_cast<Action_ID>(NOT_PREDATOR), SIZE, 
													0.66, 1.0);
	
	AddDataFilterKindToChoice(static_cast<Action_ID>(NOT_PREDATOR), SPEED);
	AddOrReplaceCategoryToDataFilterKindInChoice(	static_cast<Action_ID>(NOT_PREDATOR), SPEED, 
													0.0, 0.33);
	AddOrReplaceCategoryToDataFilterKindInChoice(	static_cast<Action_ID>(NOT_PREDATOR), SPEED, 
													0.33, 0.66);
	AddOrReplaceCategoryToDataFilterKindInChoice(	static_cast<Action_ID>(NOT_PREDATOR), SPEED, 
													0.66, 1.0);

	AddDataFilterKindToChoice(static_cast<Action_ID>(NOT_PREDATOR), TEETH);
	AddOrReplaceCategoryToDataFilterKindInChoice(	static_cast<Action_ID>(NOT_PREDATOR), TEETH, 
													0.0, 0.33);
	AddOrReplaceCategoryToDataFilterKindInChoice(	static_cast<Action_ID>(NOT_PREDATOR), TEETH, 
													0.33, 0.66);
	AddOrReplaceCategoryToDataFilterKindInChoice(	static_cast<Action_ID>(NOT_PREDATOR), TEETH, 
													0.66, 1.0);
	

	AddChoice(static_cast<Action_ID>(PREDATOR), 3);
	AddDataFilterKindToChoice(static_cast<Action_ID>(PREDATOR), SIZE);
	AddOrReplaceCategoryToDataFilterKindInChoice(	static_cast<Action_ID>(PREDATOR), SIZE, 
													0.0, 0.33);
	AddOrReplaceCategoryToDataFilterKindInChoice(	static_cast<Action_ID>(PREDATOR), SIZE, 
													0.33, 0.66);
	AddOrReplaceCategoryToDataFilterKindInChoice(	static_cast<Action_ID>(PREDATOR), SIZE, 
													0.66, 1.0);
	
	AddDataFilterKindToChoice(static_cast<Action_ID>(PREDATOR), SPEED);
	AddOrReplaceCategoryToDataFilterKindInChoice(	static_cast<Action_ID>(PREDATOR), SPEED, 
													0.0, 0.33);
	AddOrReplaceCategoryToDataFilterKindInChoice(	static_cast<Action_ID>(PREDATOR), SPEED, 
													0.33, 0.66);
	AddOrReplaceCategoryToDataFilterKindInChoice(	static_cast<Action_ID>(PREDATOR), SPEED, 
													0.66, 1.0);

	AddDataFilterKindToChoice(static_cast<Action_ID>(PREDATOR), TEETH);
	AddOrReplaceCategoryToDataFilterKindInChoice(	static_cast<Action_ID>(PREDATOR), TEETH, 
													0.0, 0.33);
	AddOrReplaceCategoryToDataFilterKindInChoice(	static_cast<Action_ID>(PREDATOR), TEETH, 
													0.33, 0.66);
	AddOrReplaceCategoryToDataFilterKindInChoice(	static_cast<Action_ID>(PREDATOR), TEETH, 
													0.66, 1.0);

	
	/* Synthesize the occurrences of attributes */
	numTrainingExamples = 10;


	choice[static_cast<Action_ID>(NOT_PREDATOR)].dataFilterKind[SIZE].category[0.0].numOccurrences = 2;
	choice[static_cast<Action_ID>(NOT_PREDATOR)].dataFilterKind[SIZE].category[0.33].numOccurrences = 2;
	choice[static_cast<Action_ID>(NOT_PREDATOR)].dataFilterKind[SIZE].category[0.66].numOccurrences = 3;

	choice[static_cast<Action_ID>(NOT_PREDATOR)].dataFilterKind[SPEED].category[0.0].numOccurrences = 3;
	choice[static_cast<Action_ID>(NOT_PREDATOR)].dataFilterKind[SPEED].category[0.33].numOccurrences = 2;
	choice[static_cast<Action_ID>(NOT_PREDATOR)].dataFilterKind[SPEED].category[0.66].numOccurrences = 2;

	choice[static_cast<Action_ID>(NOT_PREDATOR)].dataFilterKind[TEETH].category[0.0].numOccurrences = 3;
	choice[static_cast<Action_ID>(NOT_PREDATOR)].dataFilterKind[TEETH].category[0.33].numOccurrences = 3;
	choice[static_cast<Action_ID>(NOT_PREDATOR)].dataFilterKind[TEETH].category[0.66].numOccurrences = 1;


	choice[static_cast<Action_ID>(PREDATOR)].dataFilterKind[SIZE].category[0.0].numOccurrences = 2;
	choice[static_cast<Action_ID>(PREDATOR)].dataFilterKind[SIZE].category[0.33].numOccurrences = 1;
	choice[static_cast<Action_ID>(PREDATOR)].dataFilterKind[SIZE].category[0.66].numOccurrences = 0;

	choice[static_cast<Action_ID>(PREDATOR)].dataFilterKind[SPEED].category[0.0].numOccurrences = 0;
	choice[static_cast<Action_ID>(PREDATOR)].dataFilterKind[SPEED].category[0.33].numOccurrences = 1;
	choice[static_cast<Action_ID>(PREDATOR)].dataFilterKind[SPEED].category[0.66].numOccurrences = 2;

	choice[static_cast<Action_ID>(PREDATOR)].dataFilterKind[TEETH].category[0.0].numOccurrences = 0;
	choice[static_cast<Action_ID>(PREDATOR)].dataFilterKind[TEETH].category[0.33].numOccurrences = 2;
	choice[static_cast<Action_ID>(PREDATOR)].dataFilterKind[TEETH].category[0.66].numOccurrences = 1;


	/* Construct the current situation */
	std::vector<AI_Attribute> attributes;
	
	AI_Attribute attrSize, attrSpeed, attrTeeth;
	attrSize.ID = SIZE;
	attrSize.value = static_cast<float>(0.15);
	attributes.push_back(attrSize);

	attrSpeed.ID = SPEED;
	attrSpeed.value = static_cast<float>(0.51);
	attributes.push_back(attrSpeed);

	attrTeeth.ID = TEETH;
	attrTeeth.value = static_cast<float>(0.98);
	attributes.push_back(attrTeeth);

//	std::cout << "Choice is: " << MakeChoice(attributes) << std::endl;
}

void BayesClassifier::subClassInit()
{
	//TODO: Move these into the derived class of BayesianClassifierAI_Unit
	AI_System::Action_ID no = static_cast<Action_ID>(0);
	AI_System::Action_ID yes = static_cast<Action_ID>(1);
	AddChoice(no);	//No, don't recommend the Action
	AddChoice(yes);	//Yes, recommend the Action

	//fill out the DataFilters that are considered for both Choices
	for(std::vector<AI_System::DF_ID>::iterator it = DataFilters.begin(); it != DataFilters.end(); 
		++it)
	{
		AddDataFilterKindToChoice(no, *it);
		AddDataFilterKindToChoice(yes, *it);
	}

	//for each Choice, add three Categories to each DataFilterKind residing in that Choice
	for(std::map<AI_System::Action_ID, Choice>::iterator jt = choice.begin(); jt != choice.end();
		++jt)
	{
		for(std::map<unsigned int, DataFilterKind>::iterator kt = jt->second.dataFilterKind.begin();
			kt != jt->second.dataFilterKind.end(); ++kt)
		{
			AddOrReplaceCategoryToDataFilterKindInChoice(jt->first, kt->first, 0.0, 0.25);
			AddOrReplaceCategoryToDataFilterKindInChoice(jt->first, kt->first, 0.25, 0.5);
			AddOrReplaceCategoryToDataFilterKindInChoice(jt->first, kt->first, 0.5, 0.75);
			AddOrReplaceCategoryToDataFilterKindInChoice(jt->first, kt->first, 0.75, 1.0);
		}
	}
}

#endif
