////////////////////////////////////////////////////////////////////////////////
/*!	\file		AI_DataFilter_Other.cpp
**	\brief		Defines all methods declared in AI_DataFilter_Other.h
**	\author		Nate Cleveland
**	\date		11/25/2002 5:53:06 PM
**	\par		DESCRIPTION:
**				 Please see AI_DataFilter_Other.h for more information.	
**					
*/
////////////////////////////////////////////////////////////////////////////////
#include "../AI_System.h"
using namespace AI_System::DataFilter;
using namespace AI_System;
using namespace InterpretationModule;

// memory leak detection
#ifdef _DEBUG
    #include"memleak.h"
    #define new DEBUG_NEW
    #undef THIS_FILE
    static char THIS_FILE[] = __FILE__;
#endif



/////////////////////////////Age///////////////////////////////
//Only need to set the ID of the filter
Age::Age() throw() {ID = AI_System::DF_AGE;}

//See class documentation for details.
float Age::processData()
{
	//Get a pointer to this object iff it is a fishBody.
	const FishBody* fishBody = MH->GetFishBody(PG->getCurrentObject());

	//If the object ID wasn't a FishBody, the var fishBody will be null.
	if(fishBody)
	{
		//since this object was a fishBody, return its Age.
		return fishBody->m_pStats->GetAge();
	}

	//otherwise return NULL.
	return 0;
}

//See class documentation for details.
void Age::endPass()
{
	//nothing to do
	//...
}

/////////////////////////////Object Speed//////////////////////////////
//Set the Data Filter ID
ObjectSpeed::ObjectSpeed() throw() {ID = AI_System::DF_OBJECTSPEED;}

//Required data processing
float ObjectSpeed::processData()
{
	//Get the fish body
	const FishBody* fishBody = MH->GetFishBody(PG->getCurrentObject());

	if(fishBody)
	{
		//Return the fish speed
		return fishBody->GetVelocity() / fishBody->m_pAttributes->GetMaxSpeed();
	}
	//If no fish then return null
	return 0;
}

//Since this was the object and not the target then we need
//To reset after the pass
void ObjectSpeed::endPass()
{
	//Since this is an object being evaluated make sure to always recalculate.
	dataProcessed = false;
}


/////////////////////////////TargetSpeed///////////////////////////////
//Only need to set the ID of the filter
TargetSpeed::TargetSpeed() throw() {ID = AI_System::DF_TARGETSPEED;}

//See class documentation for details.
float TargetSpeed::processData()
{
	//Get a pointer to this Target iff it is a fishBody.
	const FishBody* fishBody = MH->GetFishBody(PG->getTarget());

	//If the Target ID wasn't a FishBody, the var fishBody will be null.
	if(fishBody)
	{
		//since this Target was a fishBody, return its TargetSpeed.
		return fishBody->GetVelocity() / fishBody->m_pAttributes->GetMaxSpeed();
	}

	//otherwise return NULL.
	return 0;
}

//See class documentation for details.
void TargetSpeed::endPass()
{
	//nothing to do
	//...
}




/////////////////////////////ObjectType///////////////////////////////
//Only need to set the ID of the filter
ObjectType::ObjectType() throw() {ID = AI_System::DF_OBJECTTYPE;}

//See class documentation for details.
float ObjectType::processData()
{
	//Get a pointer to this Target iff it is a fishBody.
	const FishBody* fishBody = MH->GetFishBody(PG->getTarget());

	//If the Target ID wasn't a FishBody, the var fishBody will be null.
	if(fishBody)
		return OBJECT_TYPE_FISH;

	//otherwise return that this object is chum.
	return OBJECT_TYPE_CHUM;
}

//See class documentation for details.
void ObjectType::endPass()
{
	//Since this is an object being evaluated make sure to always recalculate.
	dataProcessed = false;
}




/////////////////////////////Distance///////////////////////////////
//Only need to set the ID of the filter
Distance::Distance() throw() {ID = AI_System::DF_DISTANCE;}

//See class documentation for details.
float Distance::processData()
{
	//Get the positions of the other two objects
	math::linear_alg::Vector3<float>	targetPos(0,0,0), objectPos(0,0,0);


	//get position will return false if unable to deduce the position of the two keys.
	if(MH->GetPosition(PG->getTarget(), targetPos) &&  
	   MH->GetPosition(PG->getCurrentObject(), objectPos))
	{
		targetPos -= objectPos;				//get the vector between the two points
		return fabsf(targetPos.Magnitude());//Find the length of that vector.
	}
	


	//otherwise return 10000
	return 10000;
}

//See class documentation for details.
void Distance::endPass()
{
	//Since this is an object being evaluated make sure to always recalculate.
	dataProcessed = false;
}




/////////////////////////////PlanktonDensityObject///////////////////////////////
//Only need to set the ID of the filter
PlanktonDensityObject::PlanktonDensityObject() throw() {ID = AI_System::DF_PLANKTONDENSITY_OBJECT;}

//See class documentation for details.
float PlanktonDensityObject::processData()
{
	//first, get our position
	math::linear_alg::Vector3<float> posObject(0,0,0);

	//have the tanksimulator and the message handler calculate the relative plankton
	// density from that.
	if(MH->GetPosition(PG->getCurrentObject(), posObject))
		return MH->GetPlanktonDensity(posObject);

	//otherwise return 0.
	return 0;
}

//See class documentation for details.
void PlanktonDensityObject::endPass()
{
	//Since this is an object being evaluated make sure to always recalculate.
	dataProcessed = false;
}




/////////////////////////////PlanktonDensityTarget///////////////////////////////
//Only need to set the ID of the filter
PlanktonDensityTarget::PlanktonDensityTarget() throw() {ID = AI_System::DF_PLANKTONDENSITY_TARGET;}

//See class documentation for details.
float PlanktonDensityTarget::processData()
{
	//first, get our position
	math::linear_alg::Vector3<float> posTarget(0,0,0);

	//have the tanksimulator and the message handler calculate the relative plankton
	// density from that.
	if(MH->GetPosition(PG->getTarget(), posTarget))
		return MH->GetPlanktonDensity(posTarget);

	//otherwise return 0.
	return 0;
}

//See class documentation for details.
void PlanktonDensityTarget::endPass()
{
	//nothing to do
	//...
}