////////////////////////////////////////////////////////////////////////////////
/*!	\file		AI_DataFilter_Other.h
**	\brief		Declares physical and environmental DataFilters.
**	\author		Nate Cleveland
**	\date		11/25/2002 5:54:40 PM
**	\par		DESCRIPTION:
				 Declares the following physical DataFilters.
					- Age (in terms of the object).
					- Distance to object.
					- Speed of both Target and Object.

				 Declares the following environmental DataFilters.
					- Plankton Density around Target and Object.
					- Object Type (can be c	hum, fish or item).
**					
**	\par		REMARKS:
**					
*/
////////////////////////////////////////////////////////////////////////////////
//#include "AI_System.h"
#pragma once

namespace AI_System
{
	namespace DataFilter
	{
		const float OBJECT_TYPE_CHUM	 = 0;
		const float OBJECT_TYPE_FISH	 = 1;
		const float OBJECT_TYPE_ITEM	 = 0.3f;//! an item is anything a fish can pick up
		const float OBJECT_TYPE_RESERVED = 0.6f;

		
		
		////////////////////////////////////////////////////////////////////////////////
		/*!	\class		Age
		**	\brief		Processes the Age of the current object.
		**	\author		Nate Cleveland
		**	\date		11/25/2002 1:26:15 PM
		**	\warning	none
		*/
		////////////////////////////////////////////////////////////////////////////////
		class Age: public AI_System::AI_DataFilter
		{
		public:
			Age() throw();
			float processData();
			void endPass();
		};
		
		
		
		////////////////////////////////////////////////////////////////////////////////
		/*!	\class		Distance
		**	\brief		Processes the Distance to the current object.
		**	\author		Nate Cleveland
		**	\date		11/25/2002 1:26:15 PM
		**	\warning	none
		*/
		////////////////////////////////////////////////////////////////////////////////
		class Distance: public AI_System::AI_DataFilter
		{
		public:
			Distance() throw();
			float processData();
			void endPass();
		};
		
		
		
		////////////////////////////////////////////////////////////////////////////////
		/*!	\class		ObjectSpeed
		**	\brief		Processes the Speed of the current object.
		**	\author		Nate Cleveland
		**	\date		11/25/2002 1:26:15 PM
		**	\warning	none
		*/
		////////////////////////////////////////////////////////////////////////////////
		class ObjectSpeed: public AI_System::AI_DataFilter
		{
		public:
			ObjectSpeed() throw();
			float processData();
			void endPass();
		};
		
		
		
		
		////////////////////////////////////////////////////////////////////////////////
		/*!	\class		TargetSpeed
		**	\brief		Processes the Speed of the Target. Once again, target is defined
		**				 as the entity that is doing the thinking.
		**
		**	\author		Nate Cleveland
		**	\date		11/25/2002 1:26:15 PM
		**	\warning	none
		*/
		////////////////////////////////////////////////////////////////////////////////
		class TargetSpeed: public AI_System::AI_DataFilter
		{
		public:
			TargetSpeed() throw();
			float processData();
			void endPass();
		};
		
		
		
		
		
		////////////////////////////////////////////////////////////////////////////////
		/*!	\class		ObjectType
		**	\brief		Processes the Type of this item and covertly returns one of the
		**				 Object_Type_*.* values.
		**
		**	\author		Nate Cleveland
		**	\date		11/25/2002 1:26:15 PM
		**	\warning	none
		*/
		////////////////////////////////////////////////////////////////////////////////
		class ObjectType: public AI_System::AI_DataFilter
		{
		public:
			ObjectType() throw();
			float processData();
			void endPass();
		};

		
		
		////////////////////////////////////////////////////////////////////////////////
		/*!	\class		PlanktonDensity
		**	\brief		Processes the PlanktonDensity around the current object.
		**	\author		Nate Cleveland
		**	\date		11/25/2002 1:26:15 PM
		**	\warning	none
		*/
		////////////////////////////////////////////////////////////////////////////////
		class PlanktonDensityObject: public AI_System::AI_DataFilter
		{
		public:
			PlanktonDensityObject() throw();
			float processData();
			void endPass();
		};		

		
		
		////////////////////////////////////////////////////////////////////////////////
		/*!	\class		PlanktonDensity
		**	\brief		Processes the PlanktonDensity around the Target.
		**	\author		Nate Cleveland
		**	\date		11/25/2002 1:26:15 PM
		**	\warning	none
		*/
		////////////////////////////////////////////////////////////////////////////////
		class PlanktonDensityTarget: public AI_System::AI_DataFilter
		{
		public:
			PlanktonDensityTarget() throw();
			float processData();
			void endPass();
		};			
	}//end namespace DataFilter
}//end namespace AI_System