////////////////////////////////////////////////////////////////////////////////
/*! \file           NeuralADN.cpp
**  \brief          The implementation file for the NeuralNetwork based Action
**                  Decision Node.
**  \author         Austin Spafford
**  \date           2/14/2003 8:58:00 PM
**  \par            REMARKS:
**                    
*/
////////////////////////////////////////////////////////////////////////////////
#include <math.h>
#include <iomanip>
#include "NeuralUnit.h"
////////////////////////////////////////////////////////////////////////////////
using namespace std;
using namespace AI_System;
using namespace Neural;
////////////////////////////////////////////////////////////////////////////////

// memory leak detection
#ifdef _DEBUG
    #include"memleak.h"
    #define new DEBUG_NEW
    #undef THIS_FILE
    static char THIS_FILE[] = __FILE__;
#endif

////////////////////////////////////////////////////////////////////////////////
//  FUNCTION:       
/*! \brief          Returns the location of the next instance of a character
**                  in a NULL terminated string that is part of a floating
**                  point number.
**  \param          stringPtr a pointer into the string to seek through
**  \return         A pointer to the first floating point character, or NULL if
**                  there are not such characters exist in the specified string.
**  \exception      none
**  \author         Austin Spafford
**  \date           5/15/2003 1:29:48 PM
**  \warning        none
*/
////////////////////////////////////////////////////////////////////////////////
static char * SeekToFirstFloatingPointCharacter(char * stringPtr)
{
	if (NULL == stringPtr) return NULL;

	// seek until the end of the string
	while (0 != *stringPtr)
	{
		// check to see if the current character is a number
		char numericalValue = *stringPtr - '0';
		if (numericalValue >= 0 && numericalValue <= 9)
		{
			return stringPtr;
		}

		// check to see if the current character is a decimal point
		if ('.' == *stringPtr)
		{
			return stringPtr;
		}

		// check to see if the current character is a minus sign
		if ('-' == *stringPtr)
		{
			return stringPtr;
		}

		// move to the next character
		++stringPtr;
	}

	return NULL;
}

////////////////////////////////////////////////////////////////////////////////
//  FUNCTION:       
/*! \brief          Returns the location of the next instance of a character
**                  in a NULL terminated string that is not part of a floating
**                  point number.
**  \param          stringPtr a pointer into the string to seek through
**  \return         A pointer to the first non-floating-point character, or NULL if
**                  there are not such characters exist in the specified string.
**  \exception      none
**  \author         Austin Spafford
**  \date           5/15/2003 1:29:48 PM
**  \warning        none
*/
////////////////////////////////////////////////////////////////////////////////
static char * SeekToFirstNonFloatingPointCharacter(char * stringPtr)
{
	if (NULL == stringPtr) return NULL;

	// seek until the end of the string
	while (0 != *stringPtr)
	{
		// check to see if the current character is a number
		char numericalValue = *stringPtr - '0';
		if (numericalValue >= 0 && numericalValue <= 9)
		{
			// move to the next character
			++stringPtr;
			continue;
		}

		// check to see if the current character is a decimal point
		if ('.' == *stringPtr)
		{
			// move to the next character
			++stringPtr;
			continue;
		}

		// check to see if the current character is a minus sign
		if ('-' == *stringPtr)
		{
			// move to the next character
			++stringPtr;
			continue;
		}

		// the current character is not a floating point number, so we can return
		return stringPtr;
	}

	return NULL;
}

////////////////////////////////////////////////////////////////////////////////
//  FUNCTION:       
/*! \brief          Copy constructer.
**  \return         none
**  \exception      none
**  \author         Austin Spafford
**  \date           4/17/2003 8:38:11 PM
**  \warning        none
*/
////////////////////////////////////////////////////////////////////////////////
NeuralADN::NeuralADN(const NeuralADN& source)
		: AI_ADN(source),
		m_inputMapping(source.m_inputMapping),
		m_outputMapping(source.m_outputMapping),
		m_neuralNetworkPtr(new NeuralNetwork(*source.m_neuralNetworkPtr))
{
    //account for standard library deficiency
    assert(m_neuralNetworkPtr != NULL);
}

////////////////////////////////////////////////////////////////////////////////
//  FUNCTION:       
/*! \brief          Default destructer.
**  \return         none
**  \exception      none
**  \author         Austin Spafford
**  \date           4/17/2003 8:38:11 PM
**  \warning        none
*/
////////////////////////////////////////////////////////////////////////////////
NeuralADN::~NeuralADN(void)
{
	if (NULL != m_neuralNetworkPtr)
	{
		delete m_neuralNetworkPtr; 
		m_neuralNetworkPtr = NULL;
	}

	return;
}

////////////////////////////////////////////////////////////////////////////////
//  FUNCTION:       
/*! \brief          Selects an AI_Unit based on the current inputs.
**  \return         the ID of the selected unit
**  \exception      none
**  \author         Austin Spafford
**  \date           3/19/2003 11:43:36 AM
**  \warning        none
*/
////////////////////////////////////////////////////////////////////////////////
AI_ID NeuralADN::process() throw()
{
	// generate a sigmoid-"curved" vector of inputs
	vector<float> inputVector;
	inputVector.resize(attributes.size());
	for (unsigned int i=0; i<attributes.size(); i++)
	{
		float inputValue = attributes[i].value;
		inputValue = 1 / (1 + float(exp(-1*inputValue))); // sigmoid function
		unsigned int mappedInputIndex = m_inputMapping[attributes[i].ID];
		inputVector[mappedInputIndex] = inputValue;
	}

	// generate a ratings of each of the input values
	vector<float> outputVector;
	outputVector.resize(attributes.size());
	m_neuralNetworkPtr->SetAllInputValues(inputVector);
	m_neuralNetworkPtr->GenerateOutputValues();
	m_neuralNetworkPtr->GetAllOutputValues(outputVector);

	// find the index of the highest rating
	unsigned int indexOfHighestRating = 0;
	float highestRating = outputVector.front();	
	for (
		unsigned int ratingScanIndex = 1;
		ratingScanIndex < outputVector.size();
		++ratingScanIndex
		)
	{
		// check to see if the highest rating should be updated
		if (outputVector[ratingScanIndex] > highestRating)
		{
			indexOfHighestRating = ratingScanIndex;
			highestRating = outputVector[ratingScanIndex];
		}
	}

	// map to the ID of the AI_Unit who produced the highest rating
	AI_ID idOfUnitWithHighestRating = m_outputMapping[indexOfHighestRating];

	return idOfUnitWithHighestRating;
}

////////////////////////////////////////////////////////////////////////////////
//  FUNCTION:       
/*! \brief          Writes the object's contents to an ostream in a human
**                  readable format.
**  \param          outputStream the stream that will be written to
**  \return         none
**  \exception      none
**  \author         Austin Spafford
**  \date           4/17/2003 6:24:55 PM
**  \warning        none
*/
////////////////////////////////////////////////////////////////////////////////
void NeuralADN::Write(std::ostream & outputStream) const
{
	assert(NULL != m_neuralNetworkPtr);

	// make sure that any floating point numbers are printed with out
	// any scientific notation
	ios::fmtflags oldFlags = outputStream.flags();
	outputStream << std::fixed << std::setprecision(4) << std::showpoint;

	// write a small header
	outputStream << "============================================================" << endl;
	outputStream << "Begin NeuralADN Information" << endl;
	outputStream << endl;	

	// write out the neural network itself
	outputStream << "The neural network:" << endl;
	m_neuralNetworkPtr->Write(outputStream);

	// write out the input mapping
	outputStream << "The input mapping:" << endl;
	hash_map<AI_System::Attribute_ID,unsigned int>::const_iterator inputMappingIter;
	for (
		inputMappingIter = m_inputMapping.begin();
		m_inputMapping.end() != inputMappingIter;
		++inputMappingIter
		)
	{
		outputStream << "Attribute ID " << inputMappingIter->first << " is being mapped to input #" << inputMappingIter->second << endl;
	}
	outputStream << endl;

	// write out the output mapping
	outputStream << "The output mapping:" << endl;
	for (unsigned int i = 0; i<m_outputMapping.size(); ++i)
	{
		outputStream << "Output #" << i << " is being mapped to the attribute ID " << m_outputMapping[i] << endl;
	}
	outputStream << endl;

	// write a small trail
	outputStream << "End NeuralADN Information" << endl;
	outputStream << "============================================================" << endl;	
	outputStream << endl;

	// restore the old flags
	outputStream.flags(oldFlags);
}

////////////////////////////////////////////////////////////////////////////////
//  FUNCTION:       
/*! \brief          Reads the object's contents from an istream that is using
**                  a human readable format.
**  \param          inputStream the stream that will be read from
**  \return         none
**  \exception      none
**  \author         Austin Spafford
**  \date           4/17/2003 6:24:55 PM
**  \warning        none
*/
////////////////////////////////////////////////////////////////////////////////
void NeuralADN::Read(std::istream & inputStream)
{
	const int characterBufferSize = 1024;
	char * characterBuffer = new char[characterBufferSize];
	char * currentCharacterPtr;
	unsigned int i;

	// HEADER
	// eat the header
	inputStream.getline(characterBuffer, characterBufferSize);
	inputStream.getline(characterBuffer, characterBufferSize);
	inputStream.getline(characterBuffer, characterBufferSize);

	// NEURALNETWORK
	// eat a line
	inputStream.getline(characterBuffer, characterBufferSize);
	// make sure there is a neural network to load into
	if (NULL == m_neuralNetworkPtr)
	{
		m_neuralNetworkPtr = new NeuralNetwork(1,1);
	}
	// load the neural network
	m_neuralNetworkPtr->Read(inputStream);

	// INPUT MAPPING
	// eat a line
	inputStream.getline(characterBuffer, characterBufferSize);
	unsigned int numberOfInputs = m_neuralNetworkPtr->GetNumberOfInputs();
	// load each input mapping
	m_inputMapping.clear();
	for (i=0; i<numberOfInputs; ++i)
	{
		// load the input mapping entry
		inputStream.getline(characterBuffer, characterBufferSize);	
		currentCharacterPtr = SeekToFirstFloatingPointCharacter(characterBuffer);
		if (NULL == currentCharacterPtr) throw Nexception("Failed to load the Attribute ID.", __FILE__, __LINE__);
		AI_System::Attribute_ID attributeID = atoi(currentCharacterPtr);		
		currentCharacterPtr = SeekToFirstNonFloatingPointCharacter(currentCharacterPtr);
		currentCharacterPtr = SeekToFirstFloatingPointCharacter(currentCharacterPtr);
		if (NULL == currentCharacterPtr) throw Nexception("Failed to load the input index.", __FILE__, __LINE__);
		unsigned int inputIndex = atoi(currentCharacterPtr);
		// insert the input mapping entry
		m_inputMapping[attributeID] = inputIndex;
	}
	// eat a line
	inputStream.getline(characterBuffer, characterBufferSize);

	// OUTPUT MAPPING
	// eat a line
	inputStream.getline(characterBuffer, characterBufferSize);
	m_outputMapping.resize(numberOfInputs);
	for (i=0; i<numberOfInputs; ++i)
	{
		// load the output mapping entry
		inputStream.getline(characterBuffer, characterBufferSize);	
		currentCharacterPtr = SeekToFirstFloatingPointCharacter(characterBuffer);		
		currentCharacterPtr = SeekToFirstNonFloatingPointCharacter(currentCharacterPtr);
		currentCharacterPtr = SeekToFirstFloatingPointCharacter(currentCharacterPtr);
		if (NULL == currentCharacterPtr) throw Nexception("Failed to load the output mapping attribute ID.", __FILE__, __LINE__);
		m_outputMapping[i] = atoi(currentCharacterPtr);
	}
	// eat a line
	inputStream.getline(characterBuffer, characterBufferSize);

	// TAIL
	// eat the trail
	inputStream.getline(characterBuffer, characterBufferSize);
	inputStream.getline(characterBuffer, characterBufferSize);
	inputStream.getline(characterBuffer, characterBufferSize);

	delete []characterBuffer;

	return;
}

////////////////////////////////////////////////////////////////////////////////
//  FUNCTION:       
/*! \brief          Directly trains *m_neuralNetworkPtr with out saving a set of
**                  training examples.
**  \param          trainingResponse the expected output of this AI Unit
**  \param          trainingFitness the fitness function output for the
**                  actual output (ignored)
**  \param          trainingUnitID the ID of the AI_Unit that was determined to
**                  be the correct choice
**  \param          trainingInputs the inputs were given to this AI Unit when it
**                  made the decision that cooresponds to this training example
**  \return         none
**  \exception      none
**  \author         Austin Spafford
**  \date           3/19/2003 12:02:01 PM
**  \warning        none
*/
////////////////////////////////////////////////////////////////////////////////
void NeuralADN::saveTrainingExample(
		bool trainingResponse,
		float trainingFitness,
		AI_ID trainingUnitID,
		vector<AI_Attribute> & trainingInputs
		) throw()
{
	// generate a sigmoid-"curved" vector of inputs
	vector<float> inputVector;
	inputVector.resize(trainingInputs.size());
	for (unsigned int i=0; i<trainingInputs.size(); i++)
	{
		float inputValue = trainingInputs[i].value;
		inputValue = 1 / (1 + float(exp(-1*inputValue))); // sigmoid function
		unsigned int mappedInputIndex = m_inputMapping[trainingInputs[i].ID];
		inputVector[mappedInputIndex] = inputValue;
	}

	// generate a sigmoid-"curved" vector of expected outputs
	vector<float> outputVector;	
	outputVector.assign(inputVector.size(), 0.25f);
	unsigned int mappedOutputIndex = m_inputMapping[trainingUnitID];
	outputVector[mappedOutputIndex] = 0.75f; // mark the "correct" answer

	// train the NeuralNetwork using this example
	m_neuralNetworkPtr->Train(inputVector, outputVector);

	return;
}

////////////////////////////////////////////////////////////////////////////////
//  FUNCTION:       
/*! \brief          Releases *m_neuralNetworkPtr if it currently exists and
**                  instantiates a new NeuralNetwork that matches the required
**                  input vector dimensions.
**  \return         none
**  \exception      none
**  \author         Austin Spafford
**  \date           3/19/2003 12:13:37 PM
**  \warning        none
*/
////////////////////////////////////////////////////////////////////////////////
void NeuralADN::subClassInit() throw()
{
	// release the current NeuralNetwork 
	if (NULL != m_neuralNetworkPtr)
	{
		delete m_neuralNetworkPtr;
		m_neuralNetworkPtr = NULL;
	}

	// instantiate a 4 layer NeuralNetwork
	m_neuralNetworkPtr = new NeuralNetwork(AI_Units.size(), 4);
	m_neuralNetworkPtr->Randomize();

	// build the input mapping of the NeuralNetwork in an arbitrary order
	m_inputMapping.clear();
	unsigned int currentInputIndex;
	std::map<AI_ID, AI_Unit*>::iterator aiUnitIter;
	for (
		currentInputIndex = 0, aiUnitIter = AI_Units.begin();
		AI_Units.end() != aiUnitIter;
		++currentInputIndex, ++aiUnitIter
		)
	{
		// add this particular mapping to the input mapping
		m_inputMapping[aiUnitIter->first] = currentInputIndex;				
	}

	// build the output mapping using the constructed input mapping
	m_outputMapping.resize(m_inputMapping.size());
	hash_map<Attribute_ID, unsigned int>::iterator inputMappingIter;
	for (
		inputMappingIter = m_inputMapping.begin();
		m_inputMapping.end() != inputMappingIter;
		++inputMappingIter
		)
	{
		Attribute_ID aiUnitID = inputMappingIter->first;
		unsigned int neuralNetworkIndex = inputMappingIter->second;	

		// add this particular mapping to the output mapping			
		m_outputMapping[neuralNetworkIndex] = aiUnitID;
	}

	return;
}