////////////////////////////////////////////////////////////////////////////////
/*! \file           NeuralNetwork.cpp
**  \brief          the implentation file for the NeuralNetwork object
**  \author         Austin Spafford
**  \date           11/14/2002 9:29:18 PM
**                    
*/
////////////////////////////////////////////////////////////////////////////////
#include "neuralNetwork.h"
#include <stdlib.h>
#include <iomanip>
#include "nexception/Nexception.h"

// memory leak detection
#ifdef _DEBUG
    #include"memleak.h"
    #define new DEBUG_NEW
    #undef THIS_FILE
    static char THIS_FILE[] = __FILE__;
#endif
////////////////////////////////////////////////////////////////////////////////
#pragma warning (disable : 4786)
////////////////////////////////////////////////////////////////////////////////
using namespace std;
using namespace Neural;
////////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////
//  FUNCTION:       
/*! \brief          Returns the location of the next instance of a character
**                  in a NULL terminated string that is part of a floating
**                  point number.
**  \param          stringPtr a pointer into the string to seek through
**  \return         A pointer to the first floating point character, or NULL if
**                  there are not such characters exist in the specified string.
**  \exception      none
**  \author         Austin Spafford
**  \date           5/15/2003 1:29:48 PM
**  \warning        none
*/
////////////////////////////////////////////////////////////////////////////////
static char * SeekToFirstFloatingPointCharacter(char * stringPtr)
{
	if (NULL == stringPtr) return NULL;

	// seek until the end of the string
	while (0 != *stringPtr)
	{
		// check to see if the current character is a number
		char numericalValue = *stringPtr - '0';
		if (numericalValue >= 0 && numericalValue <= 9)
		{
			return stringPtr;
		}

		// check to see if the current character is a decimal point
		if ('.' == *stringPtr)
		{
			return stringPtr;
		}

		// check to see if the current character is a minus sign
		if ('-' == *stringPtr)
		{
			return stringPtr;
		}

		// move to the next character
		++stringPtr;
	}

	return NULL;
}

////////////////////////////////////////////////////////////////////////////////
//  FUNCTION:       
/*! \brief          Returns the location of the next instance of a character
**                  in a NULL terminated string that is not part of a floating
**                  point number.
**  \param          stringPtr a pointer into the string to seek through
**  \return         A pointer to the first non-floating-point character, or NULL if
**                  there are not such characters exist in the specified string.
**  \exception      none
**  \author         Austin Spafford
**  \date           5/15/2003 1:29:48 PM
**  \warning        none
*/
////////////////////////////////////////////////////////////////////////////////
static char * SeekToFirstNonFloatingPointCharacter(char * stringPtr)
{
	if (NULL == stringPtr) return NULL;

	// seek until the end of the string
	while (0 != *stringPtr)
	{
		// check to see if the current character is a number
		char numericalValue = *stringPtr - '0';
		if (numericalValue >= 0 && numericalValue <= 9)
		{
			// move to the next character
			++stringPtr;
			continue;
		}

		// check to see if the current character is a decimal point
		if ('.' == *stringPtr)
		{
			// move to the next character
			++stringPtr;
			continue;
		}

		// check to see if the current character is a minus sign
		if ('-' == *stringPtr)
		{
			// move to the next character
			++stringPtr;
			continue;
		}

		// the current character is not a floating point number, so we can return
		return stringPtr;
	}

	return NULL;
}

////////////////////////////////////////////////////////////////////////////////
//  FUNCTION:       
/*! \brief          A constructer that sets the number of neurons used by 
**                  the NeuralNetwork object. Also initializes all weights to
**                  zero.
**  \param          numberOfLayers the number of layers the NeuralNetwork will
**                  contain
**  \param          numberOfInputs the number of input and output values the 
**                  NeuralNetwork will have
**  \return         none
**  \exception      none
**  \author         Austin Spafford
**  \date           1/24/2003 9:01:43 PM
**  \warning        none
*/
////////////////////////////////////////////////////////////////////////////////
NeuralNetwork::NeuralNetwork(int numberOfInputs, int numberOfLayers)
{
	assert(numberOfLayers > 0);
	assert(numberOfInputs > 0);

	// resize the Neuron layers, which will also set them to default weight values
	m_neuronLayers.resize(numberOfLayers);
	for (int i=0; i<numberOfLayers; i++)
	{		
		m_neuronLayers[i].assign(numberOfInputs, Neuron(numberOfInputs));
	}

	// resize the input and output vectors
	m_inputValues.assign(numberOfInputs, 0);
	m_outputValues.assign(numberOfInputs, 0);

	// set a default training rate
	m_trainingRate = 20;

	return;
}

////////////////////////////////////////////////////////////////////////////////
//  FUNCTION:       
/*! \brief          randomly sets the weights of every Neuron
**  \return         none
**  \exception      none
**  \author         Austin Spafford
**  \date           1/27/2003 2:04:12 PM
**  \warning        assumes that a call to srand() has already been made
*/
////////////////////////////////////////////////////////////////////////////////
void NeuralNetwork::Randomize(void)
{
	// avoid excessive allocations, the extra weight is for the bias of the neuron
	vector<float> randomWeights(m_inputValues.size() + 1);

	// iterate through each layer
	for (
		vector<vector<Neuron> >::iterator neuronLayerIter = m_neuronLayers.begin();
		neuronLayerIter != m_neuronLayers.end();
		++neuronLayerIter
		)
	{				
		// iterate through each layer's neurons
		for (
			vector<Neuron>::iterator neuronIter = neuronLayerIter->begin();
			neuronIter != neuronLayerIter->end();
			++neuronIter
			)
		{
			// generate random weights for the neuron
			for (
				vector<float>::iterator weightIter = randomWeights.begin();
				weightIter != randomWeights.end();
				++weightIter
				)
			{
				// generate a random number between -0.05 and 0.05
				*weightIter = ((rand()%1000)/1000.0f)*0.1f - 0.05f;				
			}			

			// save the random weights
			neuronIter->SetWeights(randomWeights);
		}		
	}

	return;
}

////////////////////////////////////////////////////////////////////////////////
//  FUNCTION:       
/*! \brief          Sets the input values of the NeuralNetwork. 
**                  GenerateOutputValues() can then be called so that the output
**                  values will then represent the input values after they have
**                  sifted through all of the Neuron layers.
**  \param          inputValues the new input values
**  \return         none
**  \exception      none
**  \author         Austin Spafford
**  \date           1/24/2003 9:28:52 PM
**  \warning        none
*/
////////////////////////////////////////////////////////////////////////////////
void NeuralNetwork::SetAllInputValues(const vector<float> & newInputValues)
{
	assert(m_neuronLayers.front().size() == newInputValues.size());

	m_inputValues = newInputValues;

	return;
}

////////////////////////////////////////////////////////////////////////////////
//  FUNCTION:       
/*! \brief          Sets the training rate to a value other than the default.
**  \param          newTrainingRate the new training rate
**  \return         none
**  \exception      none
**  \author         Austin Spafford
**  \date           1/24/2003 9:28:52 PM
**  \warning        none
*/
////////////////////////////////////////////////////////////////////////////////
void NeuralNetwork::SetTrainingRate(float newTrainingRate)
{
	m_trainingRate = newTrainingRate;

	return;
}

////////////////////////////////////////////////////////////////////////////////
//  FUNCTION:       
/*! \brief          Processes all of the set input values and stores the results
**                  in the output vector.
**  \return         none
**  \exception      none
**  \author         Austin Spafford
**  \date           1/24/2003 9:34:21 PM
**  \warning        none
*/
////////////////////////////////////////////////////////////////////////////////
void NeuralNetwork::GenerateOutputValues(void)
{
	// treat the input values as the output from a neuron layer
	vector<float> previousLayerOutput = m_inputValues;

	// avoid repeated allocations
	vector<float> currentLayerOutput(previousLayerOutput.size());
	
	// iterate through each layer, generating values and passing them to the next layer
	for (
		vector<vector<Neuron> >::iterator neuronLayerIter = m_neuronLayers.begin();
		neuronLayerIter != m_neuronLayers.end();
		++neuronLayerIter
		)
	{				
		// generate the output for the current layer
		vector<Neuron>::iterator neuronIter = neuronLayerIter->begin();
		vector<float>::iterator outputValueIter = currentLayerOutput.begin();
		while (neuronIter != neuronLayerIter->end())
		{			
			*outputValueIter = neuronIter->GenerateOutputValue(previousLayerOutput);			

			++neuronIter;
			++outputValueIter;
		}

		// get ready for computing the values of the next layer
		previousLayerOutput = currentLayerOutput;
	}

	// save the results
	m_outputValues = previousLayerOutput;

	return;
}

////////////////////////////////////////////////////////////////////////////////
//  FUNCTION:       
/*! \brief          Retrieves the results of a call to GenerateOutputValues()
**  \param          outputValuesRef a reference to a vector that will be resized,
**                  and then filled with a number of elements equal to the number
**                  of inputs to the NeuralNetwork. These new values will represent
**                  the values generated by a call to GenerateOutputValues()
**  \return         none
**  \exception      none
**  \author         Austin Spafford
**  \date           1/24/2003 9:35:03 PM
**  \warning        none
*/
////////////////////////////////////////////////////////////////////////////////
void NeuralNetwork::GetAllOutputValues(vector<float> & outputValuesRef) const
{
	outputValuesRef = m_outputValues;

	return;
}

////////////////////////////////////////////////////////////////////////////////
//  FUNCTION:       
/*! \brief          Returns the current training rate.
**  \return         The current training rate.
**  \exception      none
**  \author         Austin Spafford
**  \date           1/24/2003 9:35:03 PM
**  \warning        none
*/
////////////////////////////////////////////////////////////////////////////////
float NeuralNetwork::GetTrainingRate(void) const
{
	return m_trainingRate;
}

////////////////////////////////////////////////////////////////////////////////
//  FUNCTION:       
/*! \brief          Trains the NeuralNetwork by using a single training example.
**                  Both the input and output values are what would be given to 
**                  the NeuralNetwork, while the target values are what's
**                  determined to be correct by the client. The technique used
**                  to train the network is called Backpropogation, or Backprop
**                  for short. Information on Backprop can be found 
**  \param          inputValues the vector of inputs that would be given to
**                  the NeuralNetwork
**  \param          targetValues the vector of outputs that the client expected
**  \return         none
**  \exception      none
**  \author         Austin Spafford
**  \date           1/28/2003 9:55:40 PM
**  \warning        make sure that every input vector is of the same size as the
**                  number of inputs to the neural network
*/
////////////////////////////////////////////////////////////////////////////////
void NeuralNetwork::Train(
		const vector<float> & inputValues,
		const vector<float> & targetValues)
{	
	// we must first generate the error terms that will allow us to correctly
	// adjust the input weights of each neuron
	GenerateErrorTerms(inputValues, targetValues);

	// now we can adjust the weights such that the neural network will procide
	// a more correct output vector when given this example again
	AdjustNeuronWeights(inputValues);	

	return;
}

///////////////////////////////////////////////////////////////////////////////
//  FUNCTION:       
/*! \brief          Writes the object's contents to an ostream in a human
**                  readable format.
**  \param          outputStream the stream that will be written to
**  \return         none
**  \exception      none
**  \author         Austin Spafford
**  \date           4/17/2003 6:24:55 PM
**  \warning        none
*/
////////////////////////////////////////////////////////////////////////////////
void NeuralNetwork::Write(std::ostream & outputStream) const
{
	
	// make sure that any floating point numbers are printed with out
	// any scientific notation
	ios::fmtflags oldFlags = outputStream.flags();
	outputStream << std::fixed << std::setprecision(4) << std::showpoint;

	// write a small header
	outputStream << "============================================================" << endl;
	outputStream << "Begin NeuralNetwork Information" << endl;
	outputStream << endl;

	// write the dimensions of the NeuralNetwork
	outputStream << "Number of inputs/outputs: " << m_neuronLayers.front().size() << endl;
	outputStream << "Number of layers: " << m_neuronLayers.size() << endl;
	outputStream << endl;

	// write out the neural network itself
	outputStream << "The neural network itself: " << endl;
	// iterate through each input->output "string"
	for (unsigned int inputIndex=0; inputIndex<m_neuronLayers.front().size(); ++inputIndex)
	{
		// iterate through each weight for each neuron
		for (unsigned int weightIndex=0; weightIndex<m_neuronLayers.front().size()+1; ++weightIndex)
		{
			// iterate through each layer
			for (unsigned int layerIndex=0; layerIndex<m_neuronLayers.size(); ++layerIndex)
			{
				// write the corresponding weight
				outputStream << m_neuronLayers[layerIndex][inputIndex].GetSpecificWeightValue(weightIndex) << "\t";
			}
			outputStream << endl;			
		}
		outputStream << endl;
	}	
	
	// write out the training rate
	outputStream << "The training rate: " << m_trainingRate << endl;
	outputStream << endl;

	// write a small trail	
	outputStream << "End NeuralNetwork Information" << endl;
	outputStream << "============================================================" << endl;	
	outputStream << endl;

	// restore the old flags
	outputStream.flags(oldFlags);

	return;
}

////////////////////////////////////////////////////////////////////////////////
//  FUNCTION:       
/*! \brief          Reads the object's contents from an istream that is using
**                  a human readable format.
**  \param          inputStream the stream that will be read from
**  \return         none
**  \exception      none
**  \author         Austin Spafford
**  \date           4/17/2003 6:24:55 PM
**  \warning        none
*/
////////////////////////////////////////////////////////////////////////////////
void NeuralNetwork::Read(std::istream & inputStream)
{
	const int characterBufferSize = 1024;
	char * characterBuffer = new char[characterBufferSize];
	char * currentCharacterPtr;
	unsigned int i;

	// HEADER
	// eat the header
	inputStream.getline(characterBuffer, characterBufferSize);
	inputStream.getline(characterBuffer, characterBufferSize);
	inputStream.getline(characterBuffer, characterBufferSize);

	// DIMENSIONS
	// read in the number of inputs/outputs
	inputStream.getline(characterBuffer, characterBufferSize);	
	currentCharacterPtr = SeekToFirstFloatingPointCharacter(characterBuffer);
	if (NULL == currentCharacterPtr) throw Nexception("Failed to load the number of inputs.", __FILE__, __LINE__);
	unsigned int numberOfInputs = atoi(currentCharacterPtr);
	// read in the number of layers
	inputStream.getline(characterBuffer, characterBufferSize);
	currentCharacterPtr = SeekToFirstFloatingPointCharacter(characterBuffer);
	if (NULL == currentCharacterPtr) throw Nexception("Failed to load the number of layers.", __FILE__, __LINE__);
	unsigned int numberOfLayers = atoi(currentCharacterPtr);
	// resize the Neuron layers
	m_neuronLayers.resize(numberOfLayers);
	for (i=0; i<numberOfLayers; i++)
	{
		m_neuronLayers[i].assign(numberOfInputs, Neuron(numberOfInputs));
	}
	// resize the input and output vectors
	m_inputValues.assign(numberOfInputs, 0);
	m_outputValues.assign(numberOfInputs, 0);
	// eat a line
	inputStream.getline(characterBuffer, characterBufferSize);

	// NEURONS
	// eat a line
	inputStream.getline(characterBuffer, characterBufferSize);
	// iterate through each input->output "string"
	for (unsigned int inputIndex=0; inputIndex<m_neuronLayers.front().size(); ++inputIndex)
	{
		// iterate through each weight for each neuron
		for (unsigned int weightIndex=0; weightIndex<m_neuronLayers.front().size()+1; ++weightIndex)
		{
			// grab the line to be dissected
			inputStream.getline(characterBuffer, characterBufferSize);
			currentCharacterPtr = SeekToFirstFloatingPointCharacter(characterBuffer);
			// iterate through each layer
			for (unsigned int layerIndex=0; layerIndex<m_neuronLayers.size(); ++layerIndex)
			{
				// read the corresponding weight
				m_neuronLayers[layerIndex][inputIndex].SetSpecificWeightValue(
						weightIndex,
						(float)atof(currentCharacterPtr)
						);
				// move to the next weight
				currentCharacterPtr = SeekToFirstNonFloatingPointCharacter(currentCharacterPtr);
				currentCharacterPtr = SeekToFirstFloatingPointCharacter(currentCharacterPtr);
			}			
		}
		// eat a line
		inputStream.getline(characterBuffer, characterBufferSize);
	}

	// TRAINING RATE
	// read in the training rate
	inputStream.getline(characterBuffer, characterBufferSize);	
	currentCharacterPtr = SeekToFirstFloatingPointCharacter(characterBuffer);
	if (NULL == currentCharacterPtr) throw Nexception("Failed to load the training rate.", __FILE__, __LINE__);
	m_trainingRate = (float)atof(currentCharacterPtr);
	// eat a line
	inputStream.getline(characterBuffer, characterBufferSize);

	// TAIL
	// eat the tail	
	inputStream.getline(characterBuffer, characterBufferSize);
	inputStream.getline(characterBuffer, characterBufferSize);
	inputStream.getline(characterBuffer, characterBufferSize);

	delete []characterBuffer;	

	return;
}

////////////////////////////////////////////////////////////////////////////////
//  FUNCTION:       
/*! \brief          Calculates the error terms for every neuron by using the
**                  Backpropogation algorithm. The error terms are required for
**                  the training of the NeuralNetwork.
**  \param          inputValues the vector of inputs that would be given to
**                  the NeuralNetwork
**  \param          targetValues the vector of outputs that the client expected
**  \return         none
**  \exception      none
**  \author         Austin Spafford
**  \date           1/28/2003 9:55:40 PM
**  \warning        make sure that every input vector is of the same size as the
**                  number of inputs to the neural network
*/
////////////////////////////////////////////////////////////////////////////////
void NeuralNetwork::GenerateErrorTerms(
		const vector<float> & inputValues,		
		const vector<float> & targetValues)
{
	// its handy to abstract this value for the sake of readability	
	unsigned int neuronsPerLayer = m_neuronLayers.front().size();

	// make sure the user is not passing in bogus data
	assert(inputValues.size() == neuronsPerLayer);
	assert(targetValues.size() == neuronsPerLayer);

	// we first reconstruct the origional output values that each neuron previously had
	SetAllInputValues(inputValues);
	GenerateOutputValues();

	// since we are only using vectors, direct indexing will be cleaner to implement
	// than iterators, references will still be used to allow for increased readability
	unsigned int layerIndex;
	unsigned int neuronIndex;
	unsigned int downstreamNeuronIndex;	

	// we generate error terms for the last layer
	vector<Neuron> & lastNeuronLayerRef = m_neuronLayers.back();
	for (
			neuronIndex = 0;
			neuronIndex < neuronsPerLayer;
			neuronIndex++
		)
	{
		Neuron & neuronRef = lastNeuronLayerRef[neuronIndex];

		float weight = neuronRef.GetSpecificWeightValue(0);
		float bias = neuronRef.GetSpecificWeightValue(1);
		float inputValue = inputValues[neuronIndex];
		float expectedOutput = targetValues[neuronIndex];

		// calculate the base "resposability" of the neuron, which represents how
		// much its output actually affects the system. This is directly related to
		// the fact that these neurons are sigmoid units
		float responsibility = neuronRef.GetLastOutputValue() * (1 - neuronRef.GetLastOutputValue());

		// we first calculate the "incorrectness" of the neuron, which is simple in
		// the case of the neurons in the final layer
		float incorrectness = targetValues[neuronIndex] - neuronRef.GetLastOutputValue();		

		// we now calculate the error term for the current neuron
		float errorTerm = responsibility * incorrectness;

		// store the error term for later reference when the weights are adjusted
		neuronRef.SetErrorTerm(errorTerm);
	}

	// now we move backwards towards the first layer, generating error terms as we go
	for (
			layerIndex = m_neuronLayers.size() - 1;
			layerIndex > 0;
			layerIndex--
		)
	{
		vector<Neuron> & neuronLayerRef = m_neuronLayers[layerIndex - 1];
		vector<Neuron> & downstreamNeuronLayerRef = m_neuronLayers[layerIndex];

		// now we generate an error term for each neuron in the layer
		for (
				neuronIndex = 0;
				neuronIndex < neuronsPerLayer;
				neuronIndex++
			)
		{
			Neuron & neuronRef = neuronLayerRef[neuronIndex];			

			// calculate the base "resposability" of the neuron, which represents how
			// much its output actually affects the system. This is directly related to
			// the fact that these neurons are sigmoid units
			float responsibility = neuronRef.GetLastOutputValue() * (1 - neuronRef.GetLastOutputValue());

			// now we must sum up the total "incorrectness" of the neuron, which is based
			// on the error terms of all the neurons that are directly downstream of the
			// current neuron, and their "dependence" on the current neuron (the
			// input weight that corresponds to the current neuron)
			float incorrectness = 0;
			for (
					downstreamNeuronIndex = 0;
					downstreamNeuronIndex < neuronsPerLayer;
					downstreamNeuronIndex++
				)
			{
				Neuron & downstreamNeuronRef = downstreamNeuronLayerRef[downstreamNeuronIndex];				

				// add the woes of the downstream neuron to the current "incorrectness" term
				incorrectness += 
						downstreamNeuronRef.GetErrorTerm()
						*
						downstreamNeuronRef.GetSpecificWeightValue(neuronIndex);
			}

			// now we calculate the error term for the current neuron
			float errorTerm = responsibility * incorrectness;

			// store the error term for later reference when the weights are adjusted
			neuronRef.SetErrorTerm(errorTerm);
		}		
	}

	return;
}

////////////////////////////////////////////////////////////////////////////////
//  FUNCTION:       
/*! \brief          Hones the NeuralNetwork by using the error terms that were
**                  generated by a call to GenerateErrorTerms(), and by using
**                  the input values that were used to generate the error terms.
**  \param          inputValues the vector of inputs that was previously given
**                  to the NeuralNetwork
**  \return         none
**  \exception      none
**  \author         Austin Spafford
**  \date           1/28/2003 9:55:40 PM
**  \warning        make sure that every input vector is of the same size as the
**                  number of inputs to the neural network
*/
////////////////////////////////////////////////////////////////////////////////
void NeuralNetwork::AdjustNeuronWeights(const vector<float> & inputValues)
{
	// its handy to abstract this value for the sake of readability	
	unsigned int neuronsPerLayer = m_neuronLayers.front().size();

	// make sure the user is not passing in bogus data
	assert(inputValues.size() == neuronsPerLayer);			

	// since we are only using vectors, direct indexing will be cleaner to implement
	// than iterators, references will still be used to allow for increased readability
	unsigned int layerIndex;	
	unsigned int neuronIndex;
	unsigned int weightIndex;

	// we adjust the weights for the first layer seperately, since its inputs
	// are not the outputs of another layer
	vector<Neuron> & firstNeuronLayerRef = m_neuronLayers.front();
	for (
			neuronIndex = 0;
			neuronIndex < neuronsPerLayer;
			neuronIndex++
		)
	{
		Neuron & neuronRef = firstNeuronLayerRef[neuronIndex];

		// we first adjust the weights that are dependent on input values
		for (
				weightIndex = 0;
				weightIndex < neuronsPerLayer;
				weightIndex++
			)
		{
			// now we generate an offset to the weight
			float weightOffset =
					m_trainingRate
					*
					neuronRef.GetErrorTerm()
					*
					inputValues[weightIndex];

			// now we adjust the neuron's weight
			neuronRef.OffsetWeightValue(
					weightIndex,
					weightOffset);
		}

		float bias = neuronRef.GetSpecificWeightValue(1);

		// now we adjust the bias of the neuron
		neuronRef.OffsetWeightValue(
				neuronsPerLayer,
				m_trainingRate * neuronRef.GetErrorTerm());

		bias = neuronRef.GetSpecificWeightValue(1);

		// now we have adjusted every weight of the neuron
	}

	// now we adjust the weights of all of the remaining layers
	for (
			layerIndex = 1;
			layerIndex < m_neuronLayers.size();
			layerIndex++
		)
	{
		vector<Neuron> & upstreamLayerRef = m_neuronLayers[layerIndex - 1];
		vector<Neuron> & downstreamLayerRef = m_neuronLayers[layerIndex];

		// we now adjust the weights of each neuron in the layer
		for (
				neuronIndex = 0;
				neuronIndex < neuronsPerLayer;
				neuronIndex++
			)
		{
			Neuron & downstreamNeuronRef = downstreamLayerRef[neuronIndex];

			// we first adjust the weights that are dependent on input values
			for (
					weightIndex = 0;
					weightIndex < neuronsPerLayer;
					weightIndex++
				)
			{				
				Neuron & upstreamNeuronRef = upstreamLayerRef[neuronIndex];

				// now we generate an offset to the weight
				float weightOffset =
						m_trainingRate
						*
						downstreamNeuronRef.GetErrorTerm()
						*
						upstreamNeuronRef.GetLastOutputValue();

				// now we adjust the neuron's weight
				downstreamNeuronRef.OffsetWeightValue(
						weightIndex,
						weightOffset);
			}			

			// now we adjust the bias of the neuron
			downstreamNeuronRef.OffsetWeightValue(
					neuronsPerLayer,
					m_trainingRate * downstreamNeuronRef.GetErrorTerm());			

			// now we have adjusted every weight of the neuron
		}
	}

	return;
}