////////////////////////////////////////////////////////////////////////////////
/*! \file           NeuralNetwork.h
**  \brief          the implementation file for the NeuralNetwork object
**  \author         Austin Spafford
**  \date           11/14/2002 9:28:21 PM
**                    
*/
////////////////////////////////////////////////////////////////////////////////
#pragma once
////////////////////////////////////////////////////////////////////////////////
#include <vector>
#include <assert.h>
#include <iostream>
#include <math.h>
////////////////////////////////////////////////////////////////////////////////

namespace Neural
{
	////////////////////////////////////////////////////////////////////////////////
	/*! \class          NeuralNetwork
	**  \brief          An object that is responsible for accepting a vector of input
	**                  values, and generating a vector of output values that has the
	**                  same number of elements as the input vector. The NeuralNetwork
	**                  object is also capable of recieving training examples which
	**                  will allow it to hone its prediction skills by using the
	**                  Backpropogation algorithm. This particular neural network
	**                  implementation uses sigmoid units, thus making all output
	**                  values map to the range of zero to one.
	**  \author         Austin Spafford
	**  \date           1/24/2003 6:29:14 PM
	**  \warning        none
	*/
	////////////////////////////////////////////////////////////////////////////////
	class NeuralNetwork
	{
		public:

		NeuralNetwork(int numberOfInputs, int numberOfLayers);	

        /// copy constructor
        NeuralNetwork(const NeuralNetwork& source)
            : //no base class
            m_inputValues(source.m_inputValues),
            m_neuronLayers(source.m_neuronLayers),
            m_outputValues(source.m_outputValues),
            m_trainingRate(source.m_trainingRate)
        {
        }

		////////////////////////////////////////////////////////////////////////////

		/// randomly sets the weights of every Neuron
		void Randomize(void);

		////////////////////////////////////////////////////////////////////////////

		/// call this before calling GenerateOutputValues()
		void SetAllInputValues(const std::vector<float> & newInputValues);

		/// sets the training rate to a value other than the default
		void SetTrainingRate(float newTrainingRate);	
		
		////////////////////////////////////////////////////////////////////////////

		/// call after calling SetAllInputValues(), and before
		/// calling GetAllOutputValues()
		void GenerateOutputValues(void);

		////////////////////////////////////////////////////////////////////////////

		/// call this after calling GenerateOutputValues()
		void GetAllOutputValues(std::vector<float> & outputValuesRef) const;

		/// returns the current training rate
		float GetTrainingRate(void) const;

		/// returns the number of inputs/outputs the NeuralNetwork has
		unsigned int GetNumberOfInputs(void) const { return m_neuronLayers.front().size(); }
		
		////////////////////////////////////////////////////////////////////////////

		/// trains the NeuralNetwork by using a single training example	
		void Train(
				const std::vector<float> & inputValues,			
				const std::vector<float> & targetValues);

		////////////////////////////////////////////////////////////////////////////

		// writes the object's contents to an ostream in a human readable format
		virtual void Write(std::ostream & outputStream) const;

		// reads the object's contents from an istream that is using a human readable format
		virtual void Read(std::istream & inputStream);

		////////////////////////////////////////////////////////////////////////////

		protected:

		/// generates the error terms that are needed to train the NeuralNetwork
		void GenerateErrorTerms(
				const std::vector<float> & inputValues,			
				const std::vector<float> & targetValues);

		/// hones the NeuralNetwork by using the error terms that were generated by
		/// a call to GenerateErrorTerms(), and by using the input values that were
		/// used to generate the error terms.
		void AdjustNeuronWeights(const std::vector<float> & inputValues);

		////////////////////////////////////////////////////////////////////////////

		////////////////////////////////////////////////////////////////////////////
		/*! \class          Neuron
		**  \brief          A simple Neuron that contains a vector of weights and is
		**                  capable of generating an output value when given an input
		**                  vector. There should always be one more element in the
		**                  weight vector compared to every input vector, the reason
		**                  for this is that the last element of the weight vector
		**                  is an offset to the output value. The generation of the 
		**                  ouput value is basically just a vector dot product. This
		**                  specific implementation is a sigmoid unit, which means
		**                  that every value is mapped to a sigmoid curve.
		**  \author         Austin Spafford
		**  \date           1/24/2003 7:29:39 PM
		**  \warning        none
		*/
		////////////////////////////////////////////////////////////////////////////
		class Neuron
		{
			public:

			/// the Neuron must be initialized such that it knows how many weights
			/// it will need to use
			Neuron(int numberOfInputs)
			{
				assert(numberOfInputs > 0);
				m_weights.assign(numberOfInputs + 1, 0);
			}

			////////////////////////////////////////////////////////////////////////

			/// sets the values of the weight vector, the last element will be the
			/// new offset value for the output
			void SetWeights(const std::vector<float> & newWeights)
			{
				assert(m_weights.size() == newWeights.size());			
				m_weights = newWeights;			
			}

			/// assigns an error term to the neuron, used for training
			void SetErrorTerm(float newErrorTerm)
			{
				m_errorTerm = newErrorTerm;
			}

			/// returns a specific weight value, used for training
			void SetSpecificWeightValue(unsigned int weightIndex, float newWeightValue)
			{
				m_weights[weightIndex] = newWeightValue;
			}

			////////////////////////////////////////////////////////////////////////
			
			/// returns an error term that has been assigned to the neuron, used for training
			float GetErrorTerm(void) const
			{
				return m_errorTerm;
			}

			/// returns the last output value that the neuron generated, used for training
			float GetLastOutputValue(void) const
			{
				return m_outputValue;
			}

			/// returns a specific weight value, used for training
			float GetSpecificWeightValue(unsigned int weightIndex) const
			{
				return m_weights[weightIndex];
			}

			////////////////////////////////////////////////////////////////////////

			/// offsets the specified weight by the specified value, used for training
			void OffsetWeightValue(unsigned int weightIndex, float offsetValue)
			{
				assert(weightIndex >= 0 && weightIndex < m_weights.size());
				m_weights[weightIndex] += offsetValue;
			}

			////////////////////////////////////////////////////////////////////////

			/// generates an output value based on the weights that are passed in,
			/// there must be exactly one less input value than the number of weights
			/// used by the Neuron
			float GenerateOutputValue(const std::vector<float> & inputValues)
			{
				assert(inputValues.size() == m_weights.size()-1);
				float returnValue = m_weights.back(); // start with the offset value
				// perform the dot product
				for (unsigned int i=0; i<inputValues.size(); i++)
				{
					returnValue += m_weights[i] * inputValues[i];
				}
				// "squash" the value to the range of (0,1) by means of a sigmoid function
				returnValue = 1 / (1 + float(exp(-1*returnValue)));				
				m_outputValue = returnValue;
				return returnValue;
			}				

			////////////////////////////////////////////////////////////////////////

			private:

			Neuron(); // intentional

			std::vector<float> m_weights;

			float m_errorTerm;

			float m_outputValue;			
		};

		std::vector<float> m_inputValues;

		std::vector<float> m_outputValues;

		std::vector<std::vector<Neuron> > m_neuronLayers; // a vector of neuron layers

		float m_trainingRate;

		NeuralNetwork(); // intentional
        const NeuralNetwork& operator=(const NeuralNetwork&);

		////////////////////////////////////////////////////////////////////////////
	};
}