////////////////////////////////////////////////////////////////////////////////
/*!	\file		Debug.cpp
**	\brief		Implementation file for Debug.h
**	\author		John Corpening
**	\date		10/7/2002 3:17:14 PM
**	\warning	Message length is limited to 255 characters per message.
**	\par		DESCRIPTION:
**					This file contains the Debugger class.
**					This class allows the user to display and log messages.
**	\par		REMARKS:
**					Debugger options can be set using the setOptions method.
**					Commands for this are as follows:
**					- + toggles on	
**					- - toggles off
**					- L1 level 1 messages
**					- L2 level 2 messages
**					- L3 level 3 messages
**					- L4 level 4 messages
**					- LALL All levels
**					- AI AI system messages
**					- NET Networking messages
**					- GFX graphics module messages
**					- PHY physics module messages
**					- SND sound module messages
**					- MSG message handler debug messages
**					- BOD fish body handler debug messages
**					- GEN support system and general messages
**					- ALL all systems
**					- DISPLAY Toggle display messages while preserving configuration
**					- LOGGING Toggle log messages while preserving configuration
**					- NOW Logs or displays messages immediately instead of once per frame
**					- \LOG direct commands to logger
**					- STANDARD (Console Messages only) directs to standard output
**					- SCREEN (Console Messages only) directs to in-game console
*/
////////////////////////////////////////////////////////////////////////////////
#include"Debug.h"
#include<iostream>
#include<stdarg.h>
#include<time.h>
//Edited by Nathan Frost on Oct 31, 2002
//BCB compatibility achieved
#include<ctype.h>
#include<stdio.h>
#include<stdlib.h>
//perhaps it would be better to use <cstdio>, <cstdlib>, etc so as not to pollute the global namespace?
//End Edited by Nathan Frost on Oct 31, 2002
//Except MS's libraries (and STLPort's wrappers) don't put the c-headers in the std namespace
// -M Jared Finder
using namespace Debugging;
// memory leak detection
#ifdef _DEBUG
    #include"memleak.h"
    #define new DEBUG_NEW
    #undef THIS_FILE
    static char THIS_FILE[] = __FILE__;
#endif
///Singleton Functionality
Debugger* Debugger::_instance = 0;


////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	Debugger::Debugger
/*!	\brief		Constructor
**	\exception	none
**	\author		John Corpening
**	\date		10/8/2002 11:21:24 AM
**	\warning	none
**	\par		DESCRIPTION:
**				- Sets a default buffer size and allocates memory for it.
**				- Sets all flag defaults.
**				- Prepares the log file
*/
////////////////////////////////////////////////////////////////////////////////
Debugger::Debugger() throw()
{
	//Default all flags to true
	for(int i = 0; i < DB_LEVELS; i++)
	{
		oLevel[i] = lLevel[i] = true;
	}
	for(int j = 0; j < DB_SYSTEM; j++)
	{
		oSystem[j] = lSystem[j] = true;
	}

	//Default other flags
	SCREEN = true;
	LOG = true;
	STANDARD = false;
	DISPLAY = true;			
	LOGGING = true;
	msgLog = false;
	msgOut = false;
	lNow = false;
	oNow = false;	

	prepLogFile();

	func = 0;
	//Check for a config file
	std::ifstream file("debug.atf");
	if(file)
	{
		char temp[255];
		file.getline(temp, 255);
		setOptions(temp);
	}
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	Delete
/*!	\brief		cleans up the unique instance
**	\exception	none
**	\author		John Corpening
**	\date		4/15/2003 8:30:35 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
Debugger::~Debugger()
{

}


////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	Debugger::Instance
/*!	\brief		returns the unique instance of the debugger
**	\return		the unique instance of the debugger
**	\exception	none
**	\author		John Corpening
**	\date		10/7/2002 4:14:09 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
Debugger* Debugger::Instance() throw()
{
	if(_instance == 0)
	{
		_instance = new Debugger;
	}
	return _instance;
}


////////////////////////////////////////////////////////////////////////////////
//  FUNCTION:   
/*! \brief      Cleans up the singleton and all its allocations
**  \param      
**  \return     none
**  \exception  none
**  \author     M Jared Finder
**  \date       4/10/2003 10:02:32 AM
**  \warning    none
*/
////////////////////////////////////////////////////////////////////////////////
void Debugger::Delete(void) throw()
{
    delete _instance;
    _instance = NULL;
}


////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	Debugger::Logic
/*!	\brief		Outputs messages
**	\return		none
**	\exception	none
**	\author		John Corpening
**	\date		10/7/2002 5:16:48 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
void Debugger::Logic() throw()
{
	writeToDisplay();
	writeToLogFile();
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	Debugger::setOptions
/*!	\brief		parses the options string and sets debug options appropriately.
**	\param		options the string with debugger commands
**	\return		none
**	\exception	none
**	\author		John Corpening
**	\date		10/7/2002 3:38:43 PM
**	\warning	none
**	\par		Command Summary
**				- + toggles on	
**				- - toggles off
**				- L1 level 1 messages
**				- L2 level 2 messages
**				- L3 level 3 messages
**				- L4 level 4 messages
**				- LALL All levels
**				- AI AI system messages
**				- NET Networking messages
**				- GFX graphics module messages
**				- PHY physics module messages
**				- SND sound module messages
**				- MSG message handler debug messages
**				- BOD fish body handler debug messages
**				- GEN support system and general messages
**				- ALL all systems
**				- DISPLAY Toggle display messages while preserving configuration
**				- LOGGING Toggle log messages while preserving configuration
**				- NOW Logs or displays messages immediately instead of once per frame
**				- \LOG direct commands to logger
**				- STANDARD (Console Messages only) directs to standard output
**				- SCREEN (Console Messages only) directs to in-game console
*/
////////////////////////////////////////////////////////////////////////////////
void Debugger::setOptions(const char* Options) throw()
{
	//sanity check
	if(!Options)
		return;

	char *options = new char[strlen(Options)+1];
	strcpy(options, Options);

	//upper cassify the options string
	for(int i = 0; options[i]; options[i++] = toupper(options[i]));

	//Parse the options string
	char	*tok;
	char	delimiters[] = " ,\t\n";
	bool	result = true;
	bool	logFlag = false;
	tok = strtok(options, delimiters);
	
	while(tok)
	{
		//Check to see if we are turning on or off
		if(tok[0] == '+')
		{
			result = true;
		}
		else if(tok[0] == '-')
		{
			result = false;
		}

		//Check if the remaining string contains log commands
		if(tok[0] == '\\')
		{
			if(!strcmp(tok + 1, "LOG"))
				logFlag = true;
		}
		//Handle commands
		else if(!strcmp(tok + 1, "L1"))
		{
			if(logFlag)
				lLevel[L1] = result;
			else
				oLevel[L1] = result;
		}
		else if(!strcmp(tok + 1, "L2"))
		{
			if(logFlag)
				lLevel[L2] = result;
			else
				oLevel[L2] = result;
		}
		else if(!strcmp(tok + 1, "L3"))
		{
			if(logFlag)
				lLevel[L3] = result;
			else
				oLevel[L3] = result;
		}
		else if(!strcmp(tok + 1, "L4"))
		{
			if(logFlag)
				lLevel[L4] = result;
			else
				oLevel[L4] = result;
		}
		else if(!strcmp(tok + 1, "LALL"))
		{
			if(logFlag)
			{
				lLevel[L1] = result;
				lLevel[L2] = result;
				lLevel[L3] = result;
				lLevel[L4] = result;
			}
			else
			{	
				oLevel[L1] = result;
				oLevel[L2] = result;
				oLevel[L3] = result;
				oLevel[L4] = result;
			}
		}
		else if(!strcmp(tok + 1, "AI"))
		{
			if(logFlag)
				lSystem[AI] = result;
			else
				oSystem[AI] = result;
		}
		else if(!strcmp(tok + 1, "NET"))
		{
			if(logFlag)
				lSystem[NET] = result;
			else
				oSystem[NET] = result;
		}
		else if(!strcmp(tok + 1, "GFX"))
		{
			if(logFlag)
				lSystem[GFX] = result;
			else
				oSystem[GFX] = result;
		}
		else if(!strcmp(tok + 1, "PHY"))
		{
			if(logFlag)
				lSystem[PHY] = result;
			else
				oSystem[PHY] = result;
		}
		else if(!strcmp(tok + 1, "SND"))
		{
			if(logFlag)
				lSystem[SND] = result;
			else
				oSystem[SND] = result;
		}
		else if(!strcmp(tok + 1, "MSG"))
		{
			if(logFlag)
				lSystem[MSG] = result;
			else
				oSystem[MSG] = result;
		}
		else if(!strcmp(tok + 1, "BOD"))
		{
			if(logFlag)
				lSystem[BOD] = result;
			else
				oSystem[BOD] = result;
		}
		else if(!strcmp(tok + 1, "GEN"))
		{
			if(logFlag)
				lSystem[GEN] = result;
			else
				oSystem[GEN] = result;
		}
		else if(!strcmp(tok + 1, "ALL"))
		{
			if(logFlag)
				for(int i = 0; i < DB_SYSTEM; oSystem[i++] = result);
			else
				for(int i = 0; i < DB_SYSTEM; oSystem[i++] = result);
		}
		else if(!strcmp(tok + 1, "DISPLAY"))
		{
			DISPLAY = result;
		}
		else if(!strcmp(tok + 1, "STANDARD"))
		{
			STANDARD = result;
		}
		else if(!strcmp(tok + 1, "SCREEN"))
		{
			SCREEN = result;
		}
		else if(!strcmp(tok + 1, "LOGGING"))
		{
			LOGGING = result;
		}
		else if(!strcmp(tok+1,"NOW"))
		{
			if(logFlag)
				lNow = result;
			else
				oNow = result;
		}
		tok = strtok(NULL, delimiters);
	}
	delete[] options;
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	Debugger::log
/*!	\brief		Writes message to the log file
**	\param		systemID The ID of the system that this message is associated with
**	\param		Level	 The priority level of this message (1-4, low to high)
**	\param		message	 The message to display
**	\param		...		 Formatting Variables
**	\return		none
**	\exception	none
**	\author		John Corpening
**	\date		10/7/2002 3:38:53 PM
**	\warning	Message length is limited to 255 characters per message.
*/
////////////////////////////////////////////////////////////////////////////////
void Debugger::log(DB_SystemID systemID, DB_Level Level, const char* message,...) throw()
{
	//Check to make sure we are accepting this type of message
	if(!logCheck(systemID, Level))
		return;

	char	temp[300];
	
	va_list list;
	va_start( list, message);     
	vsprintf(temp, message, list);

	//Tag the message
	char	tag[10];
	tagMessage(systemID, Level, tag);
	
	//Create the message
	logBuffer += tag;
	logBuffer += temp;

	//Insert a new line
	logBuffer += "\n";
	msgLog = true;

	//If the now option is set then spit out messages immediately
	if(lNow)
	{
		writeToLogFile();
		//Capacity does not need to be adjusted in immediate mode
		return;
	}
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	Debugger::output
/*!	\brief		Outputs messages to the in-game or standard console
**	\param		systemID	ID of system that this message is associated with
**	\param		Level		The priority level of this message (1-4, low to high)
**	\param		message		The message to display
**	\param		...			Formatting Variables
**	\return		none
**	\exception	none
**	\author		John Corpening
**	\date		10/7/2002 3:39:02 PM
**	\warning	Message length is limited to 255 characters per message.
*/
////////////////////////////////////////////////////////////////////////////////
void Debugger::output(DB_SystemID systemID, DB_Level Level, const char* message,...) throw()
{
	//Check to make sure we are accepting this type of message
	if(!displayCheck(systemID, Level))
		return;

	char	temp[256];
	
	va_list list;
	va_start( list, message);     
	vsnprintf(temp, 255, message, list);

	//Tag the message
	char	tag[10];
	tagMessage(systemID, Level, tag);

	
	//Create the message
	buffer += tag;
	buffer += temp;

	//Insert a new line
	buffer += "\n";
	msgOut = true;

	//If the now option is set then spit out messages immediately
	if(oNow)
	{
		writeToDisplay();
		//Capacity does not need to be adjusted in immediate mode
		return;
	}
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	Debugger::displayCheck
/*!	\brief		Checks to ensure the display message is currently accepted
**	\param		ID system of ID of message
**	\param		Level the priority level of the message
**	\return		none
**	\exception	none
**	\author		John Corpening
**	\date		10/9/2002 1:34:03 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
bool Debugger::displayCheck(DB_SystemID ID, DB_Level priority) throw()
{
	//Check to make sure the display is on
	if(!DISPLAY)
		return false;

	if(!SCREEN && !STANDARD)
		return false;

	//Discard garbage values
	if(priority >= DB_LEVELS)
		return false;
	if(ID > DB_SYSTEM)
		return false;
	
	//Check priority
	if(!oLevel[priority])
		return false;

	//Check system
	if(!oSystem[ID])
		return false;

	//If all checks pass return true
	return true;
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	Debugger::logCheck
/*!	\brief		Checks to ensure the log message is currently accepted
**	\param		ID system of ID of message
**	\param		Level the priority level of the message
**	\return		none
**	\exception	none
**	\author		John Corpening
**	\date		10/9/2002 1:34:07 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
bool Debugger::logCheck(DB_SystemID ID, DB_Level priority) throw()
{
	//Check to make sure the display is on
	if(!LOGGING)
		return false;

	//Discard garbage values
	if(priority >= DB_LEVELS)
		return false;
	if(ID > DB_SYSTEM)
		return false;
	
	//Check priority
	if(!lLevel[priority])
		return false;

	//Check system
	if(!lSystem[ID])
		return false;

	//If all checks pass return true
	return true;
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	Debugger::tagMessage
/*!	\brief		applies the system and level tag to a message
**	\param		ID system of ID of message
**	\param		Level the priority level of the message
**	\param		message the message buffer to tag
**	\return		none
**	\exception	none
**	\author		John Corpening
**	\date		10/9/2002 1:34:12 PM
**	\warning	Apply the tag before copying over the message
*/
////////////////////////////////////////////////////////////////////////////////
void Debugger::tagMessage(DB_SystemID ID, DB_Level Level, char* message) throw()
{
	char temp[2];
	itoa(Level+1, temp, 10);
	switch(ID)
	{
	case AI:
		strcpy(message, "AI : ");
		break;
	case NET:
		strcpy(message, "NET: ");
		break;
	case TNK:
		strcpy(message, "TNK: ");
		break;
	case SND:
		strcpy(message, "SND: ");
		break;
	case GFX:
		strcpy(message, "GFX: ");
		break;
	case BOD:
		strcpy(message, "BOD: ");
		break;
	case PHY:
		strcpy(message, "PHY: ");
		break;
	case MSG:
		strcpy(message, "MSG: ");
		break;
	case GEN:
		strcpy(message, "GEN: ");
		break;
	default:
		strcpy(message, "???: ");
		break;
	}
	strcat(message, temp);
	strcat(message," : ");
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	Debugger::writeToLogFile
/*!	\brief		writes the log buffer out to the log file
**	\return		none
**	\exception	none
**	\author		John Corpening
**	\date		10/9/2002 1:34:17 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
void Debugger::writeToLogFile() throw()
{
	//return if no messages
	if(!msgLog)
		return;

	//Open the log file
	logFile.open(fileName, std::ios::app);

	//Sanity check
	if(!logFile)
	{
		return;
	}

	//Write the log buffer
	logFile << logBuffer;

	//Close the logfile forcing a write
	logFile.close();

	//reset the log buffer
	logBuffer.clear();
	msgLog = false;
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	Debugger::writeToDisplay
/*!	\brief		writes the message buffer out to the display
**	\return		none
**	\exception	none
**	\author		John Corpening
**	\date		10/9/2002 1:34:21 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
void Debugger::writeToDisplay() throw()
{
	if(!msgOut)
		return;

	//Write to standard or screen output
	if(STANDARD)
	{
		std::cout << buffer;
	}
	if(SCREEN)
	{
		if(func)
		{
			func(buffer.c_str());
		}
	}
	//reset the message buffer
	buffer.clear();
	msgOut = false;	
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	Debugger::prepLogFile
/*!	\brief		prepares the log file for use
**	\return		none
**	\exception	none
**	\author		John Corpening
**	\date		10/9/2002 1:34:25 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
void Debugger::prepLogFile() throw()
{
	struct tm *newtime;
	time_t aclock;

	time( &aclock );                 /* Get time in seconds */

	newtime = localtime( &aclock );  /* Convert time to struct */
                                    /* tm form */
	strcpy(fileName,"log.atf");
	logFile.open(fileName);
	if(!logFile)
		return;
	logFile << "Aquarius Log File for : " << asctime( newtime ) << std::endl;
	logFile.close();
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	Debugger::logException
/*!	\brief		Logs exceptions, This option cannot be turned off.
**	\param		message the message to log
**	\return		none
**	\exception	none
**	\author		John Corpening
**	\date		10/18/2002 10:07:27 AM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
void Debugger::logException(const char* message) throw()
{
	logFile.open(fileName, std::ios::app);
	if(!logFile)
		return;
	logFile << "\n---Exception-------\n" << message << "---End Exception---\n" << std::endl;
	logFile.close();
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	setScreenOutputCallback
/*!	\brief		sets a pointer to a function that accepts output text
**	\param		f, a pointer to a function that takes the text and returns nothing
**	\exception	none
**	\author		John Corpening
**	\date		4/15/2003 8:32:39 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
void Debugger::setScreenOutputCallback(void (*f)(const char* text)) throw()
{
	if(f)
		func = f;
}
