////////////////////////////////////////////////////////////////////////////////
/*!	\file		Debug
**	\brief		Help with debugging
**	\author		John Corpening
**	\date		10/7/2002 3:17:14 PM
**	\warning	Message length is limited to 255 characters per message.
**	\par		DESCRIPTION:
**					This file contains the Debugger class.
**					This class allows the user to display and log messages.
**	\par		REMARKS:
**					Debugger options can be set using the setOptions method.
**					Commands for this are as follows:
**					- + toggles on	
**					- - toggles off
**					- L1 level 1 messages
**					- L2 level 2 messages
**					- L3 level 3 messages
**					- L4 level 4 messages
**					- LALL All levels
**					- AI AI system messages
**					- NET Networking messages
**					- GFX graphics module messages
**					- PHY physics module messages
**					- SND sound module messages
**					- MSG message handler debug messages
**					- BOD fish body handler debug messages
**					- GEN support system and general messages
**					- ALL all systems
**					- DISPLAY Toggle display messages while preserving configuration
**					- LOGGING Toggle log messages while preserving configuration
**					- NOW Logs or displays messages immediately instead of once per frame
**					- \LOG direct commands to logger
**					- STANDARD (Console Messages only) directs to standard output
**					- SCREEN (Console Messages only) directs to in-game console
*/
////////////////////////////////////////////////////////////////////////////////
#pragma once
#include<fstream>

namespace Debugging
{
//System Enumeration
typedef enum{GEN, AI, NET, TNK, SND, GFX, BOD, PHY, MSG} DB_SystemID;

//Level Enumeration
typedef enum{L1, L2, L3, L4} DB_Level;

////////////////////////////////////////////////////////////////////////////////
/*!	\class		Debugger
**	\brief		Class that displays and logs debug messages
**	\author		John Corpening
**	\date		10/7/2002 3:17:40 PM
**	\warning	none
**	\par		UPDATED ON:			10/18/2002 10:14:57 AM
**	\par		UPDATED BY:			John Corpening
**	\par		UPDATE DESCRIPTION: 
**					Added exception logging
*/
////////////////////////////////////////////////////////////////////////////////
class Debugger
{
	enum {DB_SYSTEM=9};				//!< Number of systems
	enum {DB_NAMESIZE = 30};		//!< File name size
	enum {DB_LEVELS = 4};			//!< Number of levels
public:


	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:	Debugger::Instance
	/*!	\brief		returns the unique instance of the debugger
	**	\return		the unique instance of the debugger
	**	\exception	none
	**	\author		John Corpening
	**	\date		10/7/2002 4:14:09 PM
	**	\warning	none
	*/
	////////////////////////////////////////////////////////////////////////////////
	static Debugger* Instance() throw();
    
  
	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:	Delete
	/*!	\brief		cleans up the unique instance
	**	\exception	none
	**	\author		John Corpening
	**	\date		4/15/2003 8:30:35 PM
	**	\warning	none
	*/
	////////////////////////////////////////////////////////////////////////////////
    static void Delete(void) throw();

	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:	Debugger::~Debugger
	/*!	\brief		Destructor 
	**	\return		none
	**	\exception	none
	**	\author		John Corpening
	**	\date		10/9/2002 3:03:44 PM
	**	\warning	none
	*/
	////////////////////////////////////////////////////////////////////////////////
	~Debugger();

	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:	Debugger::Logic
	/*!	\brief		Outputs messages
	**	\return		none
	**	\exception	none
	**	\author		John Corpening
	**	\date		10/7/2002 5:16:48 PM
	**	\warning	none
	*/
	////////////////////////////////////////////////////////////////////////////////
	void Logic() throw();

	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:	Debugger::setOptions
	/*!	\brief		parses the options string and sets debug options appropriately.
	**	\param		options the string with debugger commands
	**	\return		none
	**	\exception	none
	**	\author		John Corpening
	**	\date		10/7/2002 3:38:43 PM
	**	\warning	none
	**	\par		Command Summary
	**				- + toggles on	
	**				- - toggles off
	**				- L1 level 1 messages
	**				- L2 level 2 messages
	**				- L3 level 3 messages
	**				- L4 level 4 messages
	**				- LALL All levels
	**				- AI AI system messages
	**				- NET Networking messages
	**				- GFX graphics module messages
	**				- PHY physics module messages
	**				- SND sound module messages
	**				- MSG message handler debug messages
	**				- BOD fish body handler debug messages
	**				- GEN support system and general messages
	**				- ALL all systems
	**				- DISPLAY Toggle display messages while preserving configuration
	**				- LOGGING Toggle log messages while preserving configuration
	**				- NOW Logs or displays messages immediately instead of once per frame
	**				- \LOG direct commands to logger
	**				- STANDARD (Console Messages only) directs to standard output
	**				- SCREEN (Console Messages only) directs to in-game console
	*/
	////////////////////////////////////////////////////////////////////////////////
	void setOptions(const char* Options) throw();


	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:	Debugger::log
	/*!	\brief		Writes message to log file
	**	\param		fileID	 The ID received from startLogFile
	**	\param		systemID The ID of the system that this message is associated with
	**	\param		Level	 The priority level of this message (1-4, low to high)
	**	\param		message	 The message to display
	**	\param		...		 Formatting Variables
	**	\return		none
	**	\exception	none
	**	\author		John Corpening
	**	\date		10/7/2002 3:38:53 PM
	**	\warning	Message length is limited to 255 characters per message.
	*/
	////////////////////////////////////////////////////////////////////////////////
	void log(DB_SystemID systemID, DB_Level Level, const char* message,...) throw();


	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:	Debugger::output
	/*!	\brief		Outputs messages to the in-game or standard console
	**	\param		systemID	ID of system that this message is associated with
	**	\param		Level		The priority level of this message (1-4, low to high)
	**	\param		message		The message to display
	**	\param		...			Formatting Variables
	**	\return		none
	**	\exception	none
	**	\author		John Corpening
	**	\date		10/7/2002 3:39:02 PM
	**	\warning	Message length is limited to 255 characters per message.
	*/
	////////////////////////////////////////////////////////////////////////////////
	void output(DB_SystemID systemID, DB_Level Level, const char* message,...) throw();

	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:	Debugger::logException
	/*!	\brief		Logs exceptions, This option cannot be turned off.
	**	\param		message the message to log
	**	\return		none
	**	\exception	none
	**	\author		John Corpening
	**	\date		10/18/2002 10:07:27 AM
	**	\warning	none
	*/
	////////////////////////////////////////////////////////////////////////////////
	void logException(const char* message) throw();

	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:	setScreenOutputCallback
	/*!	\brief		sets a pointer to a function that accepts output text
	**	\param		f, a pointer to a function that takes the text and returns nothing
	**	\exception	none
	**	\author		John Corpening
	**	\date		4/15/2003 8:32:39 PM
	**	\warning	none
	*/
	////////////////////////////////////////////////////////////////////////////////
	void setScreenOutputCallback(void (*f)(const char* text)) throw();

//--------Private---------------------------------------------------------------------
private:
	std::string		buffer;					//!< Message buffer, output once per frame
	unsigned int	bufSize;				//!< Buffer Size
	std::string		logBuffer;				//!< log buffer, output when full
	char			fileName[DB_NAMESIZE];	//!< current log file name
	unsigned int	logSize;				//!< log buffer size
	unsigned int	amtLog;					//!< Amount of info logged last loop
	bool			oLevel[DB_LEVELS];		//!< Priority level flags for output
	bool			oSystem[DB_SYSTEM];		//!< System Flags for output
	bool			lLevel[DB_LEVELS];		//!< Priority Flags for logging
	bool			lSystem[DB_SYSTEM];		//!< System Flags for logging
	bool			SCREEN,STANDARD,LOG;	//!< Output flags
	bool			DISPLAY;				//!< Toggle for display messages
	bool			LOGGING;				//!< Toggle for log messages
	bool			msgLog;					//!< Flag log output received
	bool			msgOut;					//!< Flag display output received
	bool			lNow;					//!< write log messages immediately
	bool			oNow;					//!< write display output immediately

	std::ofstream	logFile;				//!< Log file
	char			name[DB_NAMESIZE];		//!< Log file name
	void			(*func)(const char* text);	//!< screen output callback

	///Singleton Functionality
	static	Debugger* _instance;

	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:	Debugger::displayCheck
	/*!	\brief		Checks to ensure the display message is currently accepted
	**	\param		ID system of ID of message
	**	\param		Level the priority level of the message
	**	\return		none
	**	\exception	none
	**	\author		John Corpening
	**	\date		10/9/2002 1:34:03 PM
	**	\warning	none
	*/
	////////////////////////////////////////////////////////////////////////////////
	bool displayCheck(DB_SystemID ID, DB_Level Level) throw();

	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:	Debugger::logCheck
	/*!	\brief		Checks to ensure the log message is currently accepted
	**	\param		ID system of ID of message
	**	\param		Level the priority level of the message
	**	\return		none
	**	\exception	none
	**	\author		John Corpening
	**	\date		10/9/2002 1:34:07 PM
	**	\warning	none
	*/
	////////////////////////////////////////////////////////////////////////////////
	bool logCheck(DB_SystemID ID, DB_Level Level) throw();

	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:	Debugger::tagMessage
	/*!	\brief		applies the system and level tag to a message
	**	\param		ID system of ID of message
	**	\param		Level the priority level of the message
	**	\param		message the message buffer to tag
	**	\return		none
	**	\exception	none
	**	\author		John Corpening
	**	\date		10/9/2002 1:34:12 PM
	**	\warning	Apply the tag before copying over the message
	*/
	////////////////////////////////////////////////////////////////////////////////
	void tagMessage(DB_SystemID ID, DB_Level Level, char* message) throw();

	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:	Debugger::writeToLogFile
	/*!	\brief		writes the log buffer out to the log file
	**	\return		none
	**	\exception	none
	**	\author		John Corpening
	**	\date		10/9/2002 1:34:17 PM
	**	\warning	none
	*/
	////////////////////////////////////////////////////////////////////////////////
	void writeToLogFile() throw();

	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:	Debugger::writeToDisplay
	/*!	\brief		writes the message buffer out to the display
	**	\return		none
	**	\exception	none
	**	\author		John Corpening
	**	\date		10/9/2002 1:34:21 PM
	**	\warning	none
	*/
	////////////////////////////////////////////////////////////////////////////////
	void writeToDisplay() throw();

	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:	Debugger::prepLogFile
	/*!	\brief		prepares the log file for use
	**	\return		none
	**	\exception	none
	**	\author		John Corpening
	**	\date		10/9/2002 1:34:25 PM
	**	\warning	none
	*/
	////////////////////////////////////////////////////////////////////////////////
	void prepLogFile() throw();

protected:

	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:	Debugger::Debugger
	/*!	\brief		Constructor
	**	\exception	none
	**	\author		John Corpening
	**	\date		10/8/2002 11:21:24 AM
	**	\warning	none
	**	\par		DESCRIPTION:
	**				- Sets a default buffer size and allocates memory for it.
	**				- Sets all flag defaults.
	**				- Prepares the log file
	*/
	////////////////////////////////////////////////////////////////////////////////
	Debugger() throw();
	
};

}//End namespace debugger