////////////////////////////////////////////////////////////////////////////////
/*! \file       entrypoint.cpp
**  \brief      Entrypoint of the GAME!
**  \author     M Jared Finder
**  \date       11/20/2002 11:41:43 AM
**  \par        DESCRIPTION:
**                  
**  \par        REMARKS:
**                  
*/
////////////////////////////////////////////////////////////////////////////////
#pragma warning(disable:4786)

///this include must be first
#include<winsock2.h>
#include<stdlib.h>
#include<time.h>

#include"globals/globals.h"
#include"nexception/nexception.h"
#include"gfx/gfx.h"
#include"sound/sound.h"
#include"messageHandler/MessageHandler.h"
#include"tankSimulator/TankSimulator.h"
#include"ai_core/AI_System.h"
#include"gui/GUI_Includes.h"
#include"debugging/Debug.h"
#include"math/math.h"
#include "sound/uiclient.h"

// memory leak detection
#ifdef _DEBUG
    #include"memleak.h"
    #define new DEBUG_NEW
    #undef THIS_FILE
    static char THIS_FILE[] = __FILE__;
#else
    #define _CrtDumpMemoryLeaks() ((int)0)
#endif

/// all the modules in the game
namespace 
{
    TimeObject*                 pTimer;             ///< fps timer

    gfx::Renderer*              pGraphicsDriver;    ///< graphics engine
	AcqariusSound::Sound*       pSoundDriver;       ///< sound engine
    MsgHandler::MessageHandler* pMessageHandler;    ///< message passing
    TankSimulator*              pTankSimulator;     ///< tank simulator
    AI_System::AI_BrainManager* pBrainManager;      ///< ai
    UserInput::UI_Handler*      pGUI;               ///< user interface
    Debugging::Debugger*        pDebug;             ///< debug


	//hack to get the sound class to play something
	AcqariusSound::UIClient*	uiClient;
}

void GameInit(void);
void GameLoop(void);
void GameCleanup(void);


////////////////////////////////////////////////////////////////////////////////
/*! \class      LoadingScreen
**  \brief      Holds the loading screen for display on one frame before
**              the game starts
**  \author     M Jared Finder
**  \date       4/5/2003 4:09:29 PM
**  \warning    none
*/
////////////////////////////////////////////////////////////////////////////////
class LoadingScreen : public gfx::IRenderable2D
{
public:
    //loads the loading screen
    LoadingScreen(void)
        : m_texture("Resources/Fullscreens/loading.tga")
    {}

    //display the loading screen
    void Render(void)
    {
        gfx::Sprite screen(m_texture, 1,1);
        screen.Draw(0,0);
    }
private:
    //texture used for loading
    gfx::Texture m_texture;
};


////////////////////////////////////////////////////////////////////////////////
//  FUNCTION:   
/*! \brief      
**  \param      
**  \return     none
**  \exception  none
**  \par        DESCRIPTION:
**                  
**  \author     M Jared Finder
**  \date       11/20/2002 11:44:05 AM
**  \warning    none
*/
////////////////////////////////////////////////////////////////////////////////
#if     defined(_CONSOLE)
    int main(void)
#elif   defined(_WINDOWS)
    int WINAPI WinMain(HINSTANCE,HINSTANCE,LPSTR,int)
#else 
#   error "Neither Console nor Windows mode specified."
#endif
{
    try
    {
        GameInit();
        GameLoop();
    }
    catch(gfx::Termination&)
    {}
    catch(gfx::Exception& ex)
    {
        MessageBox(NULL, ex.what(), "Fatal Graphics Error",
                   MB_OK | MB_ICONERROR | MB_SYSTEMMODAL);
    }
    catch(Nexception& ex)
    {
        MessageBox(
                NULL, ex.Remark(), "General Fatal Error",
                MB_OK | MB_ICONERROR | MB_SYSTEMMODAL);
    }
	catch(std::exception& ex)
	{
		MessageBox(
				NULL, ex.what(), "Standard Fatal Error",
				MB_OK | MB_ICONERROR | MB_SYSTEMMODAL);
	}
/*
    catch(...)
    {
        MessageBox(
                NULL,
                "The exception unknown occured.",
                "Fatal Error",
                MB_OK | MB_ICONERROR
                );
    }
*/

    GameCleanup();
    _CrtDumpMemoryLeaks();
    return 0;
}


////////////////////////////////////////////////////////////////////////////////
//  FUNCTION:   
/*! \brief      
**  \param      
**  \return     none
**  \exception  none
**  \par        DESCRIPTION:
**                  
**  \author     M Jared Finder
**  \date       11/20/2002 11:44:08 AM
**  \warning    none
*/
////////////////////////////////////////////////////////////////////////////////
void GameInit(void)
{
	srand(time(NULL));
    ///init globals
    g_hwnd = NULL;
    g_ScreenWidth   = 0;
    g_ScreenHeight  = 0;

    pTimer = TimeObject::Instance();

    ///init graphics
    gfx::Renderer::Create();
    pGraphicsDriver = &gfx::Renderer::Get();
	if(pGraphicsDriver == NULL)
	{
		throw Nexception("Fatal Initialization Error:\n"
					     "Could not create Render System.", __FILE__, __LINE__, __DATE__, __TIME__);
	}

    //we need an empty window for the game to run
    gfx::Resolution mode = { 640, 480, 32, 60 };
    pGraphicsDriver->ChangeResolution(mode, false);
	using namespace math::linear_alg;
	Vector3<float> pos(50,50,50);
    Vector3<float> lookat(0,1,0);
    Vector3<float> up(-1,0,0);
	pGraphicsDriver->SetProjection(60,45,0.1f, 500);
	pGraphicsDriver->Set3DView(pos, lookat, up);

	 ///init ui
    pGUI = UserInput::UI_Handler::instance();
	if(pGUI == NULL)
	{
		throw Nexception("Fatal Initialization Error:\n"
					     "Could not create UI System.", __FILE__, __LINE__, __DATE__, __TIME__);
	}

    //create a loading screen
    LoadingScreen loading;
    loading.SetRender();
    pGraphicsDriver->Flush();

    ///init sound
    pSoundDriver = AcqariusSound::Sound::Instance();
	if(pSoundDriver == NULL)
	{
		throw Nexception("Fatal Initialization Error:\n"
					     "Could not create Sound System.", __FILE__, __LINE__, __DATE__, __TIME__);
	}

    ///init message handler
    pMessageHandler = MsgHandler::MessageHandler::Instance();
	if(pMessageHandler == NULL)
	{
		throw Nexception("Fatal Initialization Error:\n"
					     "Could not create Messaging System.", __FILE__, __LINE__, __DATE__, __TIME__);
	}

    ///init debug
    pDebug = Debugging::Debugger::Instance();
	if(pDebug == NULL)
	{
		throw Nexception("Fatal Initialization Error:\n"
					     "Could not create Logging System.", __FILE__, __LINE__, __DATE__, __TIME__);
	}
//	pDebug->setOptions("-ALL");

    ///init tank
    pTankSimulator = new TankSimulator;
	if(pTankSimulator == NULL)
	{
		throw Nexception("Fatal Initialization Error:\n"
					     "Could not create Simulation System.", __FILE__, __LINE__, __DATE__, __TIME__);
	}
	pMessageHandler->RegisterTankSimulator(pTankSimulator);	
	if(!pTankSimulator->ImportHeightMapFromImage("Resources/HeightMaps/Canyons.tga", 50))
	{
		throw Nexception("Fatal Initialization Error:\n"
					     "Could not load heightmap.", __FILE__, __LINE__, __DATE__, __TIME__);
	}
	pTankSimulator->SetBoundary(math::linear_alg::Vector3<float>(-100,-50,-100), math::linear_alg::Vector3<float>(100,50,100));
    ///init ai
    pBrainManager = AI_System::AI_BrainManager::getInstance();
	if(pBrainManager == NULL)
	{
		throw Nexception("Fatal Initialization Error:\n"
					     "Could not create AI System.", __FILE__, __LINE__, __DATE__, __TIME__);
	}

	/////////////hack to init the sound to something /////////////////////
	uiClient = AcqariusSound::UIClient::Instance();
//	char* bg2 = "Resources/Sounds/bg/Klassen2.mp3";
//	char* bg1 = "Resources/Sounds/bg/Klassen2.mp3";
	if(!uiClient->LoadSoundEffects("Resources/Sounds/fx/winClick.wav","Resources/Sounds/fx/buzz.mp3"))
	{
		throw Nexception("Fatal Initialization Error:\n"
					     "Could not load sound effects.", __FILE__, __LINE__, __DATE__, __TIME__);
	}
//	std::vector<const char*> musicTracks;
//	musicTracks.push_back(bg2);
//	musicTracks.push_back(bg1);
//	if(!uiClient->LoadDynamicMusic(musicTracks))
//	{
//		throw Nexception("Fatal Initialization Error:\n"
//					     "Could not load background music.", __FILE__, __LINE__, __DATE__, __TIME__);
//	}
	pSoundDriver->SetVolume(100);
}


////////////////////////////////////////////////////////////////////////////////
//  FUNCTION:   
/*! \brief      
**  \param      
**  \return     none
**  \exception  none
**  \par        DESCRIPTION:
**                  
**  \author     M Jared Finder
**  \date       11/20/2002 12:01:32 PM
**  \warning    none
*/
////////////////////////////////////////////////////////////////////////////////
void GameLoop(void)
{
    while(pMessageHandler->GameOn())
    {
        ///timer logic
        pTimer->logic();

        ///debugger logic
        pDebug->Logic();

        ///message handler logic
        pMessageHandler->Logic();

        ///ui logic
        pGUI->logic();

        ///ai logic
		pDebug->output(Debugging::AI, Debugging::L1,"Start Brain Manager Logic Loop");
        pBrainManager->Logic();
		pDebug->output(Debugging::AI, Debugging::L1,"Start Brain Manager Train Loop");
        pBrainManager->Train();
		pDebug->output(Debugging::AI, Debugging::L1,"Start Brain Manager Breed Loop");
        pBrainManager->Breed();

        ///tank logic
		pDebug->output(Debugging::GEN, Debugging::L1, "Game Time Step: %f", pTimer->getGameTimeStep());
        pTankSimulator->StepSimulation(float(pTimer->getGameTimeStep()));

        if(!g_IsOffline) {
			pTankSimulator->Draw();
        }
        else {
            //disable rendering of the heightmap
            pGraphicsDriver->Enqueue(static_cast<gfx::IRenderableWorld*>(NULL));
        }
	
        ///sound logic
        pSoundDriver->Update(static_cast<float>(pTimer->getRealTimeStep()));

        ///graphic logic
        pGraphicsDriver->Flush();
    }
}


////////////////////////////////////////////////////////////////////////////////
//  FUNCTION:   
/*! \brief      
**  \param      
**  \return     none
**  \exception  none
**  \par        DESCRIPTION:
**                  
**  \author     M Jared Finder
**  \date       11/20/2002 12:02:36 PM
**  \warning    none
*/
////////////////////////////////////////////////////////////////////////////////
void GameCleanup(void)
{
    ///clean up ui
	delete pGUI;

    ///clean up ai
	delete pBrainManager;

	// Added by Nate Cleveland [4/8/2003]
	// Clean up after the fish body graphic client
	InterpretationModule::Model_Texture::Release();

    ///clean up tank
//unregister tank simulator
    delete pTankSimulator;
	InterpretationModule::AttributeGenerator::Release();

    ///clean up debug
    Debugging::Debugger::Delete();

    ///clean up message handler
	pMessageHandler->DestroyInstance();

    ///clean up sound
	AcqariusSound::Sound::Release();
	pSoundDriver = NULL;
	delete uiClient;


    ///clean up graphics
    gfx::Renderer::Destroy();

    ///clean up timer
	delete TimeObject::Instance();

///Call destruction function
//	UserInput::UI_Console::Destroy();
}