////////////////////////////////////////////////////////////////////////////////
/*!	\file		fileIO.cpp
**	\brief		Defines all the methods of the FileIO class.
**	\author		Nate Cleveland
**	\date		10/7/2002 12:57:26 PM
**	\par		DESCRIPTION: Defines all the methods of the FileIO class.
*/
////////////////////////////////////////////////////////////////////////////////
#include "fileIO.h"
#include "WaterMark.h"
#include <fstream>
#include "../debugging/Debug.h"
// memory leak detection
#ifdef _DEBUG
    #include"memleak.h"
    #define new DEBUG_NEW
    #undef THIS_FILE
    static char THIS_FILE[] = __FILE__;
#endif

#define MAKE_VER_NUMBER(month, day, year) (static_cast<float>(. ## month ## day ## year))
const float CURRENT_VERSION_NUMBER = MAKE_VER_NUMBER(03, 17, 03);


//instatiate all static members
FileIO* FileIO::m_instance = NULL;


//See header for documentation.
FileIO::FileIO()
{

}

//See header for documentation.
FileIO::~FileIO()
{

}

//See header for documentation.
FileIO* FileIO::Instance(void)
{
	//sanity check
	if(m_instance)
		return m_instance;
	return m_instance = new FileIO;
}

//See header for documentation.
void FileIO::Release(void)
{
	delete m_instance;
	m_instance = NULL;
}


//See header for documentation.
bool FileIO::SaveAll(const char* pathName, std::vector<FileMemento*>& mementoVec) const 
{
	//check for valid path name.
	if(!pathName)
		return false;

	//open the needed file with an io stream
	std::ofstream outFile(pathName);
//	outFile.open("a.sav");

	//test for a failed file opening.
	if(false == outFile.is_open())
	{
#		ifdef FILEIO_LOG_TO_CONSOLE
			//Log error into debuger.
			(Debugging::Debugger::Instance())->output(Debugging::GEN, Debugging::L1,
				"Error while opening the file [%s].", pathName);
#		endif
		return false;
	}

	//Water Mark the file.
	WaterMark mark("Aquarius Save File: Entire Game", CURRENT_VERSION_NUMBER);
	mark(outFile);
	
	//loop through the vector, pulling out each memento and calling its write.
	unsigned int i = 0;
	unsigned int vecSize = mementoVec.size();
	for(i = 0 ; i < vecSize ; i++)
		mementoVec[i]->Write(outFile);

	outFile.close();
	return true;
}//end FileIO::SaveAll


//See header for documentation.
bool FileIO::SaveFish(const char* pathName, const FileMemento &frozenFishMemento) const /*throw()*/
{
	//check for valid path name.
	if(!pathName)
		return false;

	//open the needed file with an io stream
	std::ofstream outFile(pathName);

	//test for a failed file opening.
	if(!outFile)
	{
#		ifdef FILEIO_LOG_TO_CONSOLE
			//Log error into debuger.
			(Debugging::Debugger::Instance())->output(Debugging::GEN, Debugging::L1,
				"Error while opening the file [%s].", pathName);
#		endif
		return false;
	}

	//save this fish memento.
	this->WriteMemento(outFile, frozenFishMemento);
	outFile.close();
	return true;
}//end FileIO:SaveFish


//See header for documentation.
bool FileIO::SaveTank(const char* pathName) const
{
	//check for valid path name.
	if(!pathName)
		return false;

	//open the needed file with an io stream
	std::ofstream outFile(pathName);

	//test for a failed file opening.
	if(!outFile)
	{
#		ifdef FILEIO_LOG_TO_CONSOLE
			//Log error into debuger.
			(Debugging::Debugger::Instance())->output(Debugging::GEN, Debugging::L1,
				"Error while opening the file [%s].", pathName);
#		endif
		return false;
	}

	//Water Mark the file.
	WaterMark mark("Aquarius Save File: Tank Only", CURRENT_VERSION_NUMBER);
	mark(outFile);
	
	//Get the Tank momento (no bodies). TODO
	//write to file tank momento

	outFile.close();
	return false;
}//end FileIO:SaveTank


//See header for documentation.
bool FileIO::LoadFish(const char* pathName, FileMemento &frozenFishMemento)
{
	//check for null
	if(pathName)
		return false;

	//open this file.
	std::ifstream inFile(pathName);

	//test for a failed file opening.
	if(!inFile)
	{
#		ifdef FILEIO_LOG_TO_CONSOLE
			//Log error into debuger.
			(Debugging::Debugger::Instance())->output(Debugging::GEN, Debugging::L1,
				"Error while opening the file [%s].", pathName);
#		endif
		return false;
	}

	WaterMark	mark(inFile);		//Object that'll read a water mark from a file


	this->ReadMemento(inFile, frozenFishMemento);
	inFile.close();
	return false;
}//end FileIO:LoadFish


//See header for documentation.
bool FileIO::LoadAll(const char* pathName, std::vector<FileMemento*>& mementoVec)
{
	//check for valid path name.
	if(!pathName)
		return false;

	//open the needed file with an io stream
	std::ifstream inFile(pathName);

	//test for a failed file opening.
	if(!inFile)
	{
#		ifdef FILEIO_LOG_TO_CONSOLE
			//Log error into debuger.
			(Debugging::Debugger::Instance())->output(Debugging::GEN, Debugging::L1,
				"Error while opening the file [%s].", pathName);
#		endif
		return false;
	}

	//Water Mark the file.
	WaterMark mark(inFile);

	//loop through the vector, pulling out each memento and calling its read.
	unsigned int i = 0;
	unsigned int vecSize = mementoVec.size();
	for(i = 0 ; i < vecSize ; i++)
		mementoVec[i]->Read(inFile);

	inFile.close();
	return true;
}//end FileIO:LoadAll

//See header for documentation.
bool FileIO::LoadTank(const char* pathName)
{
	//check for valid path name.
	if(!pathName)
		return false;

	//open the needed file with an io stream
	std::ifstream inFile(pathName);

	//test for a failed file opening.
	if(!inFile)
	{
#		ifdef FILEIO_LOG_TO_CONSOLE
			//Log error into debuger.
			(Debugging::Debugger::Instance())->output(Debugging::GEN, Debugging::L1,
				"Error while opening the file [%s].", pathName);
#		endif
		return false;
	}

	//Water Mark the file.
	WaterMark mark(inFile);
	
	//Get the Tank momento (no bodies). TODO
	//read to file tank momento

	inFile.close();
	return false;
}//end FileIO:LoadTank


