////////////////////////////////////////////////////////////////////////////////
/*!	\file		fileIO.h
**	\brief		Defines the fileIO class.
**	\author		Nate Cleveland
**	\date		10/7/2002 3:29:07 PM
**	\par		DESCRIPTION: The fileIO module abstracts file management for the
**				 UI module. Provides mechanisms for saving/loading a single fish
**				 or just an environment or entire game.	All save files in the game will 
**				 contain a water mark with the date, version number and a message.. 
**					
**	\todo		As of yet the FileIO module needs
**				\li A reference to the Message Handler in its ctor
**				\li Fill out and test the saving and loading.
**				\li Version checking of files.				
*/
////////////////////////////////////////////////////////////////////////////////
#pragma once
#include "memento.h"
#include <vector>

//!if this is left un-commented out the FileIO Module will log messages
//!to the console.
///#define FILEIO_LOG_TO_CONSOLE


////////////////////////////////////////////////////////////////////////////////
/*!	\class		FileIO class
**	\brief		Abstracts file management for the UI
**	\author		Nate Cleveland
**	\date		10/7/2002 3:54:27 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
class FileIO
{
	static FileIO* m_instance;


protected:

	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:	
	/*!	\brief		Ctor for the FileIO class.
	**	\param		Reference to the message passer.
	**	\exception	none
	**	\par		DESCRIPTION: Takes a reference to a message handler. No global
	**				 effects.
	**					
	**	\author		Nate Cleveland
	**	\date		10/7/2002 4:02:01 PM
	**	\warning	none
	*/
	////////////////////////////////////////////////////////////////////////////////
	FileIO() throw();


	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:	
	/*!	\brief		Dtor for the FileIO class.
	**	\exception	none
	**	\par		DESCRIPTION: Does nothing at this time. 
	**					
	**	\author		Nate Cleveland
	**	\date		10/7/2002 4:03:13 PM
	**	\warning	none
	*/
	////////////////////////////////////////////////////////////////////////////////
	~FileIO(void) throw();

public:
	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:	Instance
	/*!	\brief		Implements the Singleton logic of the FileIO object.
	**	\return		Pointer to the FileIO object.
	**	\exception	none
	**	\par		DESCRIPTION:
	**					Returns a pointer to the singleton object, will create one
						 if it doesn't already exsist.
	**	\author		Nate Cleveland
	**	\date		3/17/2003 10:54:58 AM
	**	\warning	none
	*/
	////////////////////////////////////////////////////////////////////////////////
	static FileIO* Instance(void);

	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:	Release
	/*!	\brief		Releases all memory used by the FileIO object.
	**	\return		none
	**	\exception	none
	**	\author		Nate Cleveland
	**	\date		3/17/2003 10:54:58 AM
	**	\warning	none
	*/
	////////////////////////////////////////////////////////////////////////////////
	static void Release(void);

	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:	
	/*!	\brief		Saves a single fish to file
	**
	**	\param		pathName	File Path to save the memento.
	**	\param		f			Memento of the Frozen Fish to save.
	**
	**	\return		Boolean true/false. False == file was not saved.
	**	\exception	none
	**	\par		DESCRIPTION: Takes the memento to a fish and saves it to file.
	**					
	**	\author		Nate Cleveland
	**	\date		10/7/2002 4:04:02 PM
	**	\warning	none
	*/
	////////////////////////////////////////////////////////////////////////////////
	bool SaveFish(const char* pathName, const FileMemento& f) const throw();


	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:	
	/*!	\brief		Saves the entire tank to file.
	**
	**	\param		pathName	File Path to save the game to.
	**
	**	\return		Boolean true/false. False == file was not saved.
	**	\exception	none
	**	\par		DESCRIPTION: 
	**				 Will save the entire game to file. 
	**					
	**	\author		Nate Cleveland
	**	\date		10/7/2002 4:07:03 PM
	**	\warning	none
	*/
	////////////////////////////////////////////////////////////////////////////////
	bool SaveAll(const char* pathName, std::vector<FileMemento*>& mementoVec) const throw();


	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:	
	/*!	\brief		Saves just the tank to file
	**
	**	\param		pathName	File Path to save the Tank to.
	**
	**	\return		Boolean true/false. False == file was not saved.
	**	\exception	none
	**	\par		DESCRIPTION: 
	**				 Will save just the tank (aka the environment) to file.Does not 
	**				  save graphical/sound settings or fish or anything else.
	**					
	**	\author		Nate Cleveland
	**	\date		10/7/2002 4:09:54 PM
	**	\warning	none
	*/
	////////////////////////////////////////////////////////////////////////////////
	bool SaveTank(const char* pathName) const throw();

	
	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:	
	/*!	\brief		Loads just a single fish from file
	**
	**	\param		pathName	File Path to load the fish from.
	**
	**	\return		Boolean true/false. False == file was not opened. True does not
	**				 guarentee that a file was read, only opened.
	**	\exception	none
	**	\par		DESCRIPTION: 
	**					Loads just a single fish from file. Will really get get messed 
	**					 up if the specified fill has anything before a single fish.
	**					
	**	\author		Nate Cleveland
	**	\date		10/7/2002 4:12:24 PM
	**	\warning	none
	*/
	////////////////////////////////////////////////////////////////////////////////
	bool LoadFish(const char* pathName, FileMemento &f)throw();

	
	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:	
	/*!	\brief		Loads the entire tank to file.
	**
	**	\param		pathName	File Path to load everything from.
	**	
	**	\return		Boolean true/false. False == file was not opened. True does not
	**				 guarentee that a file was read, only opened.
	**	\exception	none
	**	\par		DESCRIPTION: 
	**					Loads everything to the game from a file.
	**	\author		Nate Cleveland
	**	\date		10/7/2002 4:13:24 PM
	**	\warning	none
	*/
	////////////////////////////////////////////////////////////////////////////////
	bool LoadAll(const char* pathName, std::vector<FileMemento*>& mementoVec)throw();

	//
	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:	
	/*!	\brief		Loads just the tank to file
	**
	**	\param		pathName	File Path to save the Tank to.
	**	
	**	\return		Boolean true/false. False == file was not opened. True does not
	**				 guarentee that a file was read, only opened.
	**	\exception	none
	**	\par		DESCRIPTION: 
	**					Loads everything to the game from a file.
	**	\author		Nate Cleveland
	**	\date		10/7/2002 4:13:24 PM
	**	\warning	none
	*/
	////////////////////////////////////////////////////////////////////////////////
	bool LoadTank(const char* pathName)throw();


	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:	
	/*!	\brief		Writes the given memento to stream
	**
	**	\param		o		Stream to have the memento write to.	
	**	\param		mem		memento that we need to call write with
	**
	**	\return		none
	**	\exception	none
	**	\par		DESCRIPTION:
	**					
	**	\author		Nate Cleveland
	**	\date		10/7/2002 4:36:04 PM
	**	\warning	none
	*/
	////////////////////////////////////////////////////////////////////////////////
	void WriteMemento(std::ostream& o, const FileMemento& mem) const {mem.Write(o);}

	
	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:	
	/*!	\brief		Simple function that will call FileMemento::Read.
	**
	**	\param		i		Stream to have the memento read from.			
	**	\param		mem		memento that we need to call read with.
	**		
	**	\return		none
	**	\exception	none
	**	\par		DESCRIPTION:
	**					
	**	\author		Nate Cleveland
	**	\date		10/7/2002 4:40:15 PM
	**	\warning	none
	*/
	////////////////////////////////////////////////////////////////////////////////
	void ReadMemento(std::istream& i, FileMemento& mem){mem.Read(i);}
};



