////////////////////////////////////////////////////////////////////////////////
/*!	\file		memento.h
**	\brief		Provides the implementation of the FileMemento class.
**	\author		Nate Cleveland
**	\date		10/7/2002 12:01:55 PM
**	\par		DESCRIPTION: All modules and objects that wish to be saved to file
**				 must create an object that derives from this class. Inside this
**				 derived class will be an internal state or other such mechanism
**				 such that an object can be instatianted from it.
**
**					
**	\par		REMARKS: Users are allowed to stream in their saved data any way
**				 they find useful. Users must also be able to read in that data
**				 and load themselves from it.
**					
*/
////////////////////////////////////////////////////////////////////////////////
#pragma once
#include <iostream>

////////////////////////////////////////////////////////////////////////////////
/*!	\class		FileMemento
**	\brief		Provides the outline of what users will need to derive from in order
**				 for them to be saved to file.
**	\author		Nate Cleveland
**	\date		10/7/2002 12:25:31 PM
**	\warning	none
**	\par		Each user is expected to define Write and read methods to suite their
**				  needs.
*/
////////////////////////////////////////////////////////////////////////////////
class FileMemento
{
protected:

	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:	
	/*!	\brief		Creates a file memento from the specified stream.
	**
	**	\param \c istream&		Reference to a stream with the users data contained in it.
	**
	**	\exception	none
	**	\par		DESCRIPTION: Creates a file memento from the specified stream. The 
	**				 memento will now have a state to be passed through the Message 
	**				 Handler to the specified module. The stream will be updated to have 
	**				 read past the data. This will normally be used for initializing a 
	**				 memento from the data in a file.
	**					
	**	\author		Nate Cleveland
	**	\date		10/7/2002 12:34:41 PM
	**	\warning	none
	*/
	////////////////////////////////////////////////////////////////////////////////
	FileMemento(std::istream&){}
	FileMemento(){}

public:
	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:	
	/*!	\brief		Does nothing. This is a virtual function for overriding.
	**	\return		none
	**	\exception	none		
	**	\author		Nate Cleveland
	**	\date		10/7/2002 12:39:42 PM
	**	\warning	none
	*/
	////////////////////////////////////////////////////////////////////////////////
	virtual ~FileMemento(void) throw (){}

	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:	
	/*!	\brief		Writes out the data contained in the FileMemento to the stream.
	**
	**	\param \c ostream&		Reference to a stream that this class will write the data
	**							 to.
	**
	**	\return		none
	**	\exception	none
	**	\par		DESCRIPTION: This will normally be used for writting data to a file.
	**					
	**	\author		Nate Cleveland
	**	\date		10/7/2002 12:40:22 PM
	**	\warning	none
	*/
	////////////////////////////////////////////////////////////////////////////////
	virtual void Write(std::ostream&) const = 0;


	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:	
	/*!	\brief		Reads in the data contained in the stream to the FileMemento.
	**	\param \c istream&		Reference to a stream with the users data contained in it.		
	**	\return		none
	**	\exception	none
	**	\par		DESCRIPTION: This will normally be used for reading data from a file.
	**					
	**	\author		Nate Cleveland
	**	\date		10/7/2002 12:40:25 PM
	**	\warning	none
	*/
	////////////////////////////////////////////////////////////////////////////////
	virtual void Read(std::istream&) = 0;
};