////////////////////////////////////////////////////////////////////////////////
/*!	\file		waterMark.h
**	\brief		Defines the Date struct and the WaterMark class.
**	\author		Nate Cleveland
**	\date		10/7/2002 4:43:59 PM
**	\par		DESCRIPTION: The Data struct is just a simple object that holds the
**				 date on a water mark. The WaterMark class writes and reads file headers
**				 
**					
**	\par		REMARKS:
**					
*/
////////////////////////////////////////////////////////////////////////////////
#pragma once
#include <iostream>


////////////////////////////////////////////////////////////////////////////////
/*!	\class		Date
**	\brief		Holds a date.
**	\author		Nate Cleveland
**	\date		10/7/2002 4:52:33 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
struct Date
{
	int m_month;		//!< Month that this file was saved (1 -> 12)	
	int m_day;			//!< Day that this file was saved (1 -> 31)
	int m_year;			//!< Year that this file was saved on ie 2002

	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:	
	/*!	\brief		Reads in a date string.
	**	\param \c const \c char*		String that the date can be read from.
	**	\return		none
	**	\exception	none
	**	\par		DESCRIPTION: Takes a passed in string and reads it into the struct.
	**				 For this to work the passed in string must be in the
	**				 format of MONTH // DAY // YEAR (example: 9 // 7 // 2002). 
	**					
	**	\author		Nate Cleveland
	**	\date		10/7/2002 4:55:28 PM
	**	\warning	none
	*/
	////////////////////////////////////////////////////////////////////////////////
	void operator ()(const char*);

	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:	
	/*!	\brief		Defualt ctor.
	**	\param \c int	month, day, year.		
	**	\return		none
	**	\exception	none
	**	\par		DESCRIPTION: Sets all values to zero.
	**					
	**	\author		Nate Cleveland
	**	\date		10/7/2002 5:05:24 PM
	**	\warning	none
	*/
	////////////////////////////////////////////////////////////////////////////////
	Date(int month = 0, int day = 0, int year = 0)
		:m_month(month), m_day(day), m_year(year){}
};

////////////////////////////////////////////////////////////////////////////////
/*!	\class		
**	\brief		
**	\author		Nate Cleveland
**	\date		10/7/2002 5:06:29 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
class WaterMark
{
	char	*m_message;	//!< Message to write at top of the file.
	float	m_ver;		//!< version number of this water mark.
	Date	m_date;		//!< date of this water mark.

public:

	//instatiates a Watermark object.
	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:	
	/*!	\brief		Instatiates a WaterMark object. 
	**
	**	\param		Message			A one line message to send top a file with
	**	\param		versionNumber			A floating pnt value of the version.	
	**
	**	\exception	none
	**	\par		DESCRIPTION: Creates a water mark with the given values. Is defualted
	**				 to zeros.
	**					
	**	\author		Nate Cleveland
	**	\date		10/7/2002 5:06:42 PM
	**	\warning	none
	*/
	////////////////////////////////////////////////////////////////////////////////
	WaterMark(const char *Message = NULL, float versionNumber = 0.0f) throw();

	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:	
	/*!	\brief		Instantiates a WaterMark class from a input stream.
	**
	**	\param \c istream& i		Reads in from this stream.
	**
	**	\exception	none
	**	\par		DESCRIPTION: Will fill in a WaterMark based on what it finds in 
	**				 the passed in stream.
	**					
	**	\author		Nate Cleveland
	**	\date		10/7/2002 5:10:19 PM
	**	\warning	none
	*/
	////////////////////////////////////////////////////////////////////////////////
	WaterMark(std::istream& i) throw():m_message(NULL), m_ver(0) {(*this)(i);}

	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:	
	/*!	\brief		defualt dtor.
	**	\exception	none
	**	\par		DESCRIPTION: Frees any memory used up by the m_message member.					
	**	\author		Nate Cleveland
	**	\date		10/7/2002 5:13:14 PM
	**	\warning	none
	*/
	////////////////////////////////////////////////////////////////////////////////
	~WaterMark();


	//water marks the passed in stream
	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:	
	/*!	\brief		Stamps a stream with a water mark. 	
	**	\param \c ostream&		Stream to place the water mark on.
	**	\return		none
	**	\exception	none
	**	\par		DESCRIPTION: Streams a water mark onto the passed in stream. 
	**					
	**	\author		Nate Cleveland
	**	\date		10/7/2002 5:13:50 PM
	**	\warning	none
	*/
	////////////////////////////////////////////////////////////////////////////////
	void operator ()(std::ostream &) throw();

	//reads the watermark from a stream. 
	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:	
	/*!	\brief		Reads a water mark from the passed in stream
	**	\param \c istream&		Stream to read from.
	**	\return		none
	**	\exception	none
	**	\par		DESCRIPTION: Reads a water mark from the passed in stream.
	**					
	**	\author		Nate Cleveland
	**	\date		10/7/2002 5:13:53 PM
	**	\warning	none
	*/
	////////////////////////////////////////////////////////////////////////////////
	void operator ()(std::istream &);
};


