////////////////////////////////////////////////////////////////////////////////
/*! \file       gfx/api_loader.cpp
**  \brief      Loads the graphical API functionality
**  \author     M Jared Finder
**  \date       2/14/2003 3:29:38 PM
*/
////////////////////////////////////////////////////////////////////////////////
#include"gfx/api_loader.h"
using gfx::ApiLoader;


////////////////////////////////////////////////////////////////////////////////
// Needed includes
////////////////////////////////////////////////////////////////////////////////
//#include"stdheaders.h"
#include"gfx/opengl.h"
#include<sstream>
#include<string>
#include<string.h>
#include<stdlib.h>
#include<assert.h>
// memory leak detection
#ifdef _DEBUG
    #include"memleak.h"
    #define new DEBUG_NEW
    #undef THIS_FILE
    static char THIS_FILE[] = __FILE__;
#endif


////////////////////////////////////////////////////////////////////////////////
//  FUNCTION:   
/*! \brief      Initialize the OpenGL API entry points.
**  \param      
**  \return     none
**  \exception  none
**  \author     M Jared Finder
**  \date       2/14/2003 3:30:29 PM
**  \warning    none
*/
////////////////////////////////////////////////////////////////////////////////
ApiLoader::ApiLoader(void)
        : majorVersion(0), minorVersion(0),
        ///ARB extensions
        pointParameters(false),             vertexBlend(false),
        matrixPalette(false),               textureEnvCrossbar(false),
        textureMirroredRepeat(false),       depthTexture(false),
        shadow(false),                      shadowAmbient(false),
        windowPos(false),
        ///EXT extensions
        compiledVertexArray(false),
        ///NV extensions
        NVvertexArray(false),               NVpointSprite(false),
        ///ATI extensions
        ATIvertexArray(false)
{
    //get the version number
    char Buffer[256];
    char* pNextData;
    strncpy(&Buffer[0], (const char*)glGetString(GL_VERSION), 255);
    Buffer[255] = '\0';
    pNextData = &Buffer[0];
    majorVersion = strtol(pNextData, &pNextData, 10);
    ++pNextData;
    minorVersion = strtol(pNextData, &pNextData, 10);

    //initialize the OpenGL standard entry points
    InitializeGL_1_0();
    InitializeGL_1_1();
    InitializeGL_1_2();
    InitializeGL_1_3();

    //initialize the ARB extension entry points
    InitializeARB_point_parameters();
    InitializeARB_vertex_blend();
    InitializeARB_matrix_palette();
    InitializeARB_texture_env_crossbar();
    InitializeARB_texture_mirrored_repeat();
    InitializeARB_depth_texture();
    InitializeARB_shadow();
    InitializeARB_shadow_ambient();
    InitializeARB_window_pos();

    //initialize the EXT extension entry points
    InitializeEXT_compiled_vertex_array();
    
    //initialize the NV extension entry points
    InitializeNV_vertex_array_range();
    InitializeNV_point_sprite();

    //initialize the ATI extension entry points
    InitializeATI_vertex_array_object();
}


////////////////////////////////////////////////////////////////////////////////
//  FUNCTION:   
/*! \brief      Detects if an OpenGL extension is available
**  \param      extension   The extension to check
**  \return     none
**  \exception  none
**  \par        DESCRIPTION:
**                  
**  \author     M Jared Finder
**  \date       2/18/2003 3:53:09 PM
**  \warning    none
*/
////////////////////////////////////////////////////////////////////////////////
bool ApiLoader::IsExtensionAvailable(const char* extension)
{
    //create a stream from the extensions
    std::istringstream extensions((const char*)glGetString(GL_EXTENSIONS));
    //create a string from the desired extension
    std::string desired(extension);

    //search for the extension
    std::string anExtension;
    while(extensions)
    {
        extensions >> anExtension;
        if(anExtension == desired)
        {
            return true;
        }
    }
    

    return false;
}

////////////////////////////////////////////////////////////////////////////////
//  FUNCTION:   
/*! \brief      Load respective OpenGL version entry points.
**  \param      
**  \return     none
**  \exception  none
**  \par        DESCRIPTION:
**                  
**  \author     M Jared Finder
**  \date       2/15/2003 4:23:25 PM
**  \warning    none
*/
////////////////////////////////////////////////////////////////////////////////
void ApiLoader::InitializeGL_1_0(void)
{
    //already provided by opengl32.lib
}
void ApiLoader::InitializeGL_1_1(void)
{
    //already provided by opengl32.lib
}
void ApiLoader::InitializeGL_1_2(void)
{
    //need at least OpenGL 1.2
    if(majorVersion < 1 || (majorVersion == 1 && minorVersion < 2))
    {
        return;
    }

    ///load OpenGL 1.2
    glBlendColor                    = (PFNGLBLENDCOLORPROC                  )wglGetProcAddress("glBlendColor");
    glBlendEquation                 = (PFNGLBLENDEQUATIONPROC               )wglGetProcAddress("glBlendEquation");
    glDrawRangeElements             = (PFNGLDRAWRANGEELEMENTSPROC           )wglGetProcAddress("glDrawRangeElements");
    glColorTable                    = (PFNGLCOLORTABLEPROC                  )wglGetProcAddress("glColorTable");
    glColorTableParameterfv         = (PFNGLCOLORTABLEPARAMETERFVPROC       )wglGetProcAddress("glColorTableParameterfv");
    glColorTableParameteriv         = (PFNGLCOLORTABLEPARAMETERIVPROC       )wglGetProcAddress("glColorTableParameteriv");
    glCopyColorTable                = (PFNGLCOPYCOLORTABLEPROC              )wglGetProcAddress("glCopyColorTable");
    glGetColorTable                 = (PFNGLGETCOLORTABLEPROC               )wglGetProcAddress("glGetColorTable");
    glGetColorTableParameterfv      = (PFNGLGETCOLORTABLEPARAMETERFVPROC    )wglGetProcAddress("glGetColorTableParameterfv");
    glGetColorTableParameteriv      = (PFNGLGETCOLORTABLEPARAMETERIVPROC    )wglGetProcAddress("glGetColorTableParameteriv");
    glColorSubTable                 = (PFNGLCOLORSUBTABLEPROC               )wglGetProcAddress("glColorSubTable");
    glCopyColorSubTable             = (PFNGLCOPYCOLORSUBTABLEPROC           )wglGetProcAddress("glCopyColorSubTable");
    glConvolutionFilter1D           = (PFNGLCONVOLUTIONFILTER1DPROC         )wglGetProcAddress("glConvolutionFilter1D");
    glConvolutionFilter2D           = (PFNGLCONVOLUTIONFILTER2DPROC         )wglGetProcAddress("glConvolutionFilter2D");
    glConvolutionParameterf         = (PFNGLCONVOLUTIONPARAMETERFPROC       )wglGetProcAddress("glConvolutionParameterf");
    glConvolutionParameterfv        = (PFNGLCONVOLUTIONPARAMETERFVPROC      )wglGetProcAddress("glConvolutionParameterfv");
    glConvolutionParameteri         = (PFNGLCONVOLUTIONPARAMETERIPROC       )wglGetProcAddress("glConvolutionParameteri");
    glConvolutionParameteriv        = (PFNGLCONVOLUTIONPARAMETERIVPROC      )wglGetProcAddress("glConvolutionParameteriv");
    glCopyConvolutionFilter1D       = (PFNGLCOPYCONVOLUTIONFILTER1DPROC     )wglGetProcAddress("glCopyConvolutionFilter1D");
    glCopyConvolutionFilter2D       = (PFNGLCOPYCONVOLUTIONFILTER2DPROC     )wglGetProcAddress("glCopyConvolutionFilter2D");
    glGetConvolutionFilter          = (PFNGLGETCONVOLUTIONFILTERPROC        )wglGetProcAddress("glGetConvolutionFilter");
    glGetConvolutionParameterfv     = (PFNGLGETCONVOLUTIONPARAMETERFVPROC   )wglGetProcAddress("glGetConvolutionParameterfv");
    glGetConvolutionParameteriv     = (PFNGLGETCONVOLUTIONPARAMETERIVPROC   )wglGetProcAddress("glGetConvolutionParameteriv");
    glGetSeparableFilter            = (PFNGLGETSEPARABLEFILTERPROC          )wglGetProcAddress("glGetSeparableFilter");
    glSeparableFilter2D             = (PFNGLSEPARABLEFILTER2DPROC           )wglGetProcAddress("glSeparableFilter2D");
    glGetHistogram                  = (PFNGLGETHISTOGRAMPROC                )wglGetProcAddress("glGetHistogram");
    glGetHistogramParameterfv       = (PFNGLGETHISTOGRAMPARAMETERFVPROC     )wglGetProcAddress("glGetHistogramParameterfv");
    glGetHistogramParameteriv       = (PFNGLGETHISTOGRAMPARAMETERIVPROC     )wglGetProcAddress("glGetHistogramParameteriv");
    glGetMinmax                     = (PFNGLGETMINMAXPROC                   )wglGetProcAddress("glGetMinmax");
    glGetMinmaxParameterfv          = (PFNGLGETMINMAXPARAMETERFVPROC        )wglGetProcAddress("glGetMinmaxParameterfv");
    glGetMinmaxParameteriv          = (PFNGLGETMINMAXPARAMETERIVPROC        )wglGetProcAddress("glGetMinmaxParameteriv");
    glHistogram                     = (PFNGLHISTOGRAMPROC                   )wglGetProcAddress("glHistogram");
    glMinmax                        = (PFNGLMINMAXPROC                      )wglGetProcAddress("glMinmax");
    glResetHistogram                = (PFNGLRESETHISTOGRAMPROC              )wglGetProcAddress("glResetHistogram");
    glResetMinmax                   = (PFNGLRESETMINMAXPROC                 )wglGetProcAddress("glResetMinmax");
    glTexImage3D                    = (PFNGLTEXIMAGE3DPROC                  )wglGetProcAddress("glTexImage3D");
    glTexSubImage3D                 = (PFNGLTEXSUBIMAGE3DPROC               )wglGetProcAddress("glTexSubImage3D");
    glCopyTexSubImage3D             = (PFNGLCOPYTEXSUBIMAGE3DPROC           )wglGetProcAddress("glCopyTexSubImage3D");

    assert(glBlendColor);
    assert(glBlendEquation);
    assert(glDrawRangeElements);
    assert(glColorTable);
    assert(glColorTableParameterfv);
    assert(glColorTableParameteriv);
    assert(glCopyColorTable);
    assert(glGetColorTable);
    assert(glGetColorTableParameterfv);
    assert(glGetColorTableParameteriv);
    assert(glColorSubTable);
    assert(glCopyColorSubTable);
    assert(glConvolutionFilter1D);
    assert(glConvolutionFilter2D);
    assert(glConvolutionParameterf);
    assert(glConvolutionParameterfv);
    assert(glConvolutionParameteri);
    assert(glConvolutionParameteriv);
    assert(glCopyConvolutionFilter1D);
    assert(glCopyConvolutionFilter2D);
    assert(glGetConvolutionFilter);
    assert(glGetConvolutionParameterfv);
    assert(glGetConvolutionParameteriv);
    assert(glGetSeparableFilter);
    assert(glSeparableFilter2D);
    assert(glGetHistogram);
    assert(glGetHistogramParameterfv);
    assert(glGetHistogramParameteriv);
    assert(glGetMinmax);
    assert(glGetMinmaxParameterfv);
    assert(glGetMinmaxParameteriv);
    assert(glHistogram);
    assert(glMinmax);
    assert(glResetHistogram);
    assert(glResetMinmax);
    assert(glTexImage3D);
    assert(glTexSubImage3D);
    assert(glCopyTexSubImage3D);
}
void ApiLoader::InitializeGL_1_3(void)
{
    //need at least OpenGL 1.3(
    if(majorVersion < 1 || (majorVersion == 1 && minorVersion < 3))
    {
        return;
    }

    //load OpenGL 1.3
    glActiveTexture                 = (PFNGLACTIVETEXTUREPROC               )wglGetProcAddress("glActiveTexture");
    glClientActiveTexture           = (PFNGLCLIENTACTIVETEXTUREPROC         )wglGetProcAddress("glClientActiveTexture");
    glMultiTexCoord1d               = (PFNGLMULTITEXCOORD1DPROC             )wglGetProcAddress("glMultiTexCoord1d");
    glMultiTexCoord1dv              = (PFNGLMULTITEXCOORD1DVPROC            )wglGetProcAddress("glMultiTexCoord1dv");
    glMultiTexCoord1f               = (PFNGLMULTITEXCOORD1FPROC             )wglGetProcAddress("glMultiTexCoord1f");
    glMultiTexCoord1fv              = (PFNGLMULTITEXCOORD1FVPROC            )wglGetProcAddress("glMultiTexCoord1fv");
    glMultiTexCoord1i               = (PFNGLMULTITEXCOORD1IPROC             )wglGetProcAddress("glMultiTexCoord1i");
    glMultiTexCoord1iv              = (PFNGLMULTITEXCOORD1IVPROC            )wglGetProcAddress("glMultiTexCoord1iv");
    glMultiTexCoord1s               = (PFNGLMULTITEXCOORD1SPROC             )wglGetProcAddress("glMultiTexCoord1s");
    glMultiTexCoord1sv              = (PFNGLMULTITEXCOORD1SVPROC            )wglGetProcAddress("glMultiTexCoord1sv");
    glMultiTexCoord2d               = (PFNGLMULTITEXCOORD2DPROC             )wglGetProcAddress("glMultiTexCoord2d");
    glMultiTexCoord2dv              = (PFNGLMULTITEXCOORD2DVPROC            )wglGetProcAddress("glMultiTexCoord2dv");
    glMultiTexCoord2f               = (PFNGLMULTITEXCOORD2FPROC             )wglGetProcAddress("glMultiTexCoord2f");
    glMultiTexCoord2fv              = (PFNGLMULTITEXCOORD2FVPROC            )wglGetProcAddress("glMultiTexCoord2fv");
    glMultiTexCoord2i               = (PFNGLMULTITEXCOORD2IPROC             )wglGetProcAddress("glMultiTexCoord2i");
    glMultiTexCoord2iv              = (PFNGLMULTITEXCOORD2IVPROC            )wglGetProcAddress("glMultiTexCoord2iv");
    glMultiTexCoord2s               = (PFNGLMULTITEXCOORD2SPROC             )wglGetProcAddress("glMultiTexCoord2s");
    glMultiTexCoord2sv              = (PFNGLMULTITEXCOORD2SVPROC            )wglGetProcAddress("glMultiTexCoord2sv");
    glMultiTexCoord3d               = (PFNGLMULTITEXCOORD3DPROC             )wglGetProcAddress("glMultiTexCoord3d");
    glMultiTexCoord3dv              = (PFNGLMULTITEXCOORD3DVPROC            )wglGetProcAddress("glMultiTexCoord3dv");
    glMultiTexCoord3f               = (PFNGLMULTITEXCOORD3FPROC             )wglGetProcAddress("glMultiTexCoord3f");
    glMultiTexCoord3fv              = (PFNGLMULTITEXCOORD3FVPROC            )wglGetProcAddress("glMultiTexCoord3fv");
    glMultiTexCoord3i               = (PFNGLMULTITEXCOORD3IPROC             )wglGetProcAddress("glMultiTexCoord3i");
    glMultiTexCoord3iv              = (PFNGLMULTITEXCOORD3IVPROC            )wglGetProcAddress("glMultiTexCoord3iv");
    glMultiTexCoord3s               = (PFNGLMULTITEXCOORD3SPROC             )wglGetProcAddress("glMultiTexCoord3s");
    glMultiTexCoord3sv              = (PFNGLMULTITEXCOORD3SVPROC            )wglGetProcAddress("glMultiTexCoord3sv");
    glMultiTexCoord4d               = (PFNGLMULTITEXCOORD4DPROC             )wglGetProcAddress("glMultiTexCoord4d");
    glMultiTexCoord4dv              = (PFNGLMULTITEXCOORD4DVPROC            )wglGetProcAddress("glMultiTexCoord4dv");
    glMultiTexCoord4f               = (PFNGLMULTITEXCOORD4FPROC             )wglGetProcAddress("glMultiTexCoord4f");
    glMultiTexCoord4fv              = (PFNGLMULTITEXCOORD4FVPROC            )wglGetProcAddress("glMultiTexCoord4fv");
    glMultiTexCoord4i               = (PFNGLMULTITEXCOORD4IPROC             )wglGetProcAddress("glMultiTexCoord4i");
    glMultiTexCoord4iv              = (PFNGLMULTITEXCOORD4IVPROC            )wglGetProcAddress("glMultiTexCoord4iv");
    glMultiTexCoord4s               = (PFNGLMULTITEXCOORD4SPROC             )wglGetProcAddress("glMultiTexCoord4s");
    glMultiTexCoord4sv              = (PFNGLMULTITEXCOORD4SVPROC            )wglGetProcAddress("glMultiTexCoord4sv");
    glLoadTransposeMatrixf          = (PFNGLLOADTRANSPOSEMATRIXFPROC        )wglGetProcAddress("glLoadTransposeMatrixf");
    glLoadTransposeMatrixd          = (PFNGLLOADTRANSPOSEMATRIXDPROC        )wglGetProcAddress("glLoadTransposeMatrixd");
    glMultTransposeMatrixf          = (PFNGLMULTTRANSPOSEMATRIXFPROC        )wglGetProcAddress("glMultTransposeMatrixf");
    glMultTransposeMatrixd          = (PFNGLMULTTRANSPOSEMATRIXDPROC        )wglGetProcAddress("glMultTransposeMatrixd");
    glSampleCoverage                = (PFNGLSAMPLECOVERAGEPROC              )wglGetProcAddress("glSampleCoverage");
    glCompressedTexImage3D          = (PFNGLCOMPRESSEDTEXIMAGE3DPROC        )wglGetProcAddress("glCompressedTexImage3D");
    glCompressedTexImage2D          = (PFNGLCOMPRESSEDTEXIMAGE2DPROC        )wglGetProcAddress("glCompressedTexImage2D");
    glCompressedTexImage1D          = (PFNGLCOMPRESSEDTEXIMAGE1DPROC        )wglGetProcAddress("glCompressedTexImage1D");
    glCompressedTexSubImage3D       = (PFNGLCOMPRESSEDTEXSUBIMAGE3DPROC     )wglGetProcAddress("glCompressedTexSubImage3D");
    glCompressedTexSubImage2D       = (PFNGLCOMPRESSEDTEXSUBIMAGE2DPROC     )wglGetProcAddress("glCompressedTexSubImage2D");
    glCompressedTexSubImage1D       = (PFNGLCOMPRESSEDTEXSUBIMAGE1DPROC     )wglGetProcAddress("glCompressedTexSubImage1D");
    glGetCompressedTexImage         = (PFNGLGETCOMPRESSEDTEXIMAGEPROC       )wglGetProcAddress("glGetCompressedTexImage");

    assert(glActiveTexture);
    assert(glClientActiveTexture);
    assert(glMultiTexCoord1d);
    assert(glMultiTexCoord1dv);
    assert(glMultiTexCoord1f);
    assert(glMultiTexCoord1fv);
    assert(glMultiTexCoord1i);
    assert(glMultiTexCoord1iv);
    assert(glMultiTexCoord1s);
    assert(glMultiTexCoord1sv);
    assert(glMultiTexCoord2d);
    assert(glMultiTexCoord2dv);
    assert(glMultiTexCoord2f);
    assert(glMultiTexCoord2fv);
    assert(glMultiTexCoord2i);
    assert(glMultiTexCoord2iv);
    assert(glMultiTexCoord2s);
    assert(glMultiTexCoord2sv);
    assert(glMultiTexCoord3d);
    assert(glMultiTexCoord3dv);
    assert(glMultiTexCoord3f);
    assert(glMultiTexCoord3fv);
    assert(glMultiTexCoord3i);
    assert(glMultiTexCoord3iv);
    assert(glMultiTexCoord3s);
    assert(glMultiTexCoord3sv);
    assert(glMultiTexCoord4d);
    assert(glMultiTexCoord4dv);
    assert(glMultiTexCoord4f);
    assert(glMultiTexCoord4fv);
    assert(glMultiTexCoord4i);
    assert(glMultiTexCoord4iv);
    assert(glMultiTexCoord4s);
    assert(glMultiTexCoord4sv);
    assert(glLoadTransposeMatrixf);
    assert(glLoadTransposeMatrixd);
    assert(glMultTransposeMatrixf);
    assert(glMultTransposeMatrixd);
    assert(glSampleCoverage);
    assert(glCompressedTexImage3D);
    assert(glCompressedTexImage2D);
    assert(glCompressedTexImage1D);
    assert(glCompressedTexSubImage3D);
    assert(glCompressedTexSubImage2D);
    assert(glCompressedTexSubImage1D);
    assert(glGetCompressedTexImage);
}


////////////////////////////////////////////////////////////////////////////////
//  FUNCTION:   
/*! \brief      Load respective OpenGL ARB Extension entry points.
**  \param      
**  \return     none
**  \exception  none
**  \par        DESCRIPTION:
**                  
**  \author     M Jared Finder
**  \date       2/18/2003 2:59:48 PM
**  \warning    none
*/
////////////////////////////////////////////////////////////////////////////////
void ApiLoader::InitializeARB_point_parameters(void)
{
    if(!IsExtensionAvailable("GL_ARB_point_parameters"))
    {
        return;
    }
    pointParameters = true;

    glPointParameterfARB            = (PFNGLPOINTPARAMETERFARBPROC          )wglGetProcAddress("glPointParameterfARB");
    glPointParameterfvARB           = (PFNGLPOINTPARAMETERFVARBPROC         )wglGetProcAddress("glPointParameterfvARB");

    assert(glPointParameterfARB);
    assert(glPointParameterfvARB);
}
void ApiLoader::InitializeARB_vertex_blend(void)
{
    if(!IsExtensionAvailable("GL_ARB_vertex_blend"))
    {
        return;
    }
    vertexBlend = true;

    glWeightbvARB                   = (PFNGLWEIGHTBVARBPROC                 )wglGetProcAddress("glWeightbvARB");
    glWeightsvARB                   = (PFNGLWEIGHTSVARBPROC                 )wglGetProcAddress("glWeightsvARB");
    glWeightivARB                   = (PFNGLWEIGHTIVARBPROC                 )wglGetProcAddress("glWeightivARB");
    glWeightfvARB                   = (PFNGLWEIGHTFVARBPROC                 )wglGetProcAddress("glWeightfvARB");
    glWeightdvARB                   = (PFNGLWEIGHTDVARBPROC                 )wglGetProcAddress("glWeightdvARB");
    glWeightubvARB                  = (PFNGLWEIGHTUBVARBPROC                )wglGetProcAddress("glWeightubvARB");
    glWeightusvARB                  = (PFNGLWEIGHTUSVARBPROC                )wglGetProcAddress("glWeightusvARB");
    glWeightuivARB                  = (PFNGLWEIGHTUIVARBPROC                )wglGetProcAddress("glWeightuivARB");
    glWeightPointerARB              = (PFNGLWEIGHTPOINTERARBPROC            )wglGetProcAddress("glWeightPointerARB");
    glVertexBlendARB                = (PFNGLVERTEXBLENDARBPROC              )wglGetProcAddress("glVertexBlendARB");

    assert(glWeightbvARB);
    assert(glWeightsvARB);
    assert(glWeightivARB);
    assert(glWeightfvARB);
    assert(glWeightdvARB);
    assert(glWeightubvARB);
    assert(glWeightusvARB);
    assert(glWeightuivARB);
    assert(glWeightPointerARB);
    assert(glVertexBlendARB);
}
void ApiLoader::InitializeARB_matrix_palette(void)
{
    if(!IsExtensionAvailable("GL_ARB_matrix_palette"))
    {
        return;
    }
    matrixPalette = true;

    glCurrentPaletteMatrixARB       = (PFNGLCURRENTPALETTEMATRIXARBPROC     )wglGetProcAddress("glCurrentPaletteMatrixARB");
    glMatrixIndexubvARB             = (PFNGLMATRIXINDEXUBVARBPROC           )wglGetProcAddress("glMatrixIndexubvARB");
    glMatrixIndexusvARB             = (PFNGLMATRIXINDEXUSVARBPROC           )wglGetProcAddress("glMatrixIndexusvARB");
    glMatrixIndexuivARB             = (PFNGLMATRIXINDEXUIVARBPROC           )wglGetProcAddress("glMatrixIndexuivARB");
    glMatrixIndexPointerARB         = (PFNGLMATRIXINDEXPOINTERARBPROC       )wglGetProcAddress("glMatrixIndexPointerARB");

    assert(glCurrentPaletteMatrixARB);
    assert(glMatrixIndexubvARB);
    assert(glMatrixIndexusvARB);
    assert(glMatrixIndexuivARB);
    assert(glMatrixIndexPointerARB);
}
void ApiLoader::InitializeARB_texture_env_crossbar(void)
{
    if(!IsExtensionAvailable("GL_ARB_texture_env_crossbar"))
    {
        return;
    }
    textureEnvCrossbar = true;
}
void ApiLoader::InitializeARB_texture_mirrored_repeat(void)
{
    if(!IsExtensionAvailable("GL_ARB_texture_mirrored_repeat"))
    {
        return;
    }
    textureMirroredRepeat = true;
}
void ApiLoader::InitializeARB_depth_texture(void)
{
    if(!IsExtensionAvailable("GL_ARB_texture_depth_texture"))
    {
        return;
    }
    depthTexture = true;
}
void ApiLoader::InitializeARB_shadow(void)
{
    if(!IsExtensionAvailable("GL_ARB_shadow"))
    {
        return;
    }
    shadow = true;
}
void ApiLoader::InitializeARB_shadow_ambient(void)
{
    if(!IsExtensionAvailable("GL_ARB_shadow_ambient"))
    {
        return;
    }
    shadowAmbient = true;
}
void ApiLoader::InitializeARB_window_pos(void)
{
    if(!IsExtensionAvailable("GL_ARB_window_pos"))
    {
        return;
    }
    windowPos = true;

    glWindowPos2dARB                = (PFNGLWINDOWPOS2DARBPROC              )wglGetProcAddress("glWindowPos2dARB");
    glWindowPos2dvARB               = (PFNGLWINDOWPOS2DVARBPROC             )wglGetProcAddress("glWindowPos2dvARB");
    glWindowPos2fARB                = (PFNGLWINDOWPOS2FARBPROC              )wglGetProcAddress("glWindowPos2fARB");
    glWindowPos2fvARB               = (PFNGLWINDOWPOS2FVARBPROC             )wglGetProcAddress("glWindowPos2fvARB");
    glWindowPos2iARB                = (PFNGLWINDOWPOS2IARBPROC              )wglGetProcAddress("glWindowPos2iARB");
    glWindowPos2ivARB               = (PFNGLWINDOWPOS2IVARBPROC             )wglGetProcAddress("glWindowPos2ivARB");
    glWindowPos2sARB                = (PFNGLWINDOWPOS2SARBPROC              )wglGetProcAddress("glWindowPos2sARB");
    glWindowPos2svARB               = (PFNGLWINDOWPOS2SVARBPROC             )wglGetProcAddress("glWindowPos2svARB");
    glWindowPos3dARB                = (PFNGLWINDOWPOS3DARBPROC              )wglGetProcAddress("glWindowPos3dARB");
    glWindowPos3dvARB               = (PFNGLWINDOWPOS3DVARBPROC             )wglGetProcAddress("glWindowPos3dvARB");
    glWindowPos3fARB                = (PFNGLWINDOWPOS3FARBPROC              )wglGetProcAddress("glWindowPos3fARB");
    glWindowPos3fvARB               = (PFNGLWINDOWPOS3FVARBPROC             )wglGetProcAddress("glWindowPos3fvARB");
    glWindowPos3iARB                = (PFNGLWINDOWPOS3IARBPROC              )wglGetProcAddress("glWindowPos3iARB");
    glWindowPos3ivARB               = (PFNGLWINDOWPOS3IVARBPROC             )wglGetProcAddress("glWindowPos3ivARB");
    glWindowPos3sARB                = (PFNGLWINDOWPOS3SARBPROC              )wglGetProcAddress("glWindowPos3sARB");
    glWindowPos3svARB               = (PFNGLWINDOWPOS3SVARBPROC             )wglGetProcAddress("glWindowPos3svARB");

    assert(glWindowPos2dARB);
    assert(glWindowPos2dvARB);
    assert(glWindowPos2fARB);
    assert(glWindowPos2fvARB);
    assert(glWindowPos2iARB);
    assert(glWindowPos2ivARB);
    assert(glWindowPos2sARB);
    assert(glWindowPos2svARB);
    assert(glWindowPos3dARB);
    assert(glWindowPos3dvARB);
    assert(glWindowPos3fARB);
    assert(glWindowPos3fvARB);
    assert(glWindowPos3iARB);
    assert(glWindowPos3ivARB);
    assert(glWindowPos3sARB);
    assert(glWindowPos3svARB);
}


////////////////////////////////////////////////////////////////////////////////
//  FUNCTION:   
/*! \brief      Load respective OpenGL EXT Extension entry points.
**  \param      
**  \return     none
**  \exception  none
**  \par        DESCRIPTION:
**                  
**  \author     M Jared Finder
**  \date       2/18/2003 2:59:48 PM
**  \warning    none
*/
////////////////////////////////////////////////////////////////////////////////
void ApiLoader::InitializeEXT_compiled_vertex_array(void)
{
    if(!IsExtensionAvailable("GL_EXT_compiled_vertex_array"))
    {
        return;
    }
    compiledVertexArray = true;

    glLockArraysEXT                 = (PFNGLLOCKARRAYSEXTPROC               )wglGetProcAddress("glLockArraysEXT");
    glUnlockArraysEXT               = (PFNGLUNLOCKARRAYSEXTPROC             )wglGetProcAddress("glUnlockArraysEXT");

    assert(glLockArraysEXT);
    assert(glUnlockArraysEXT);
}


////////////////////////////////////////////////////////////////////////////////
//  FUNCTION:   
/*! \brief      Load respective OpenGL Extension entry points
**  \param      
**  \return     none
**  \exception  none
**  \par        DESCRIPTION:
**                  
**  \author     M Jared Finder
**  \date       2/27/2003 1:23:46 PM
**  \warning    none
*/
////////////////////////////////////////////////////////////////////////////////
void ApiLoader::InitializeNV_vertex_array_range(void)
{
    if(!IsExtensionAvailable("GL_NV_vertex_array_range2"))
    {
        return;
    }
    NVvertexArray = true;

    wglAllocateMemoryNV             = (PFNWGLALLOCATEMEMORYNVPROC           )wglGetProcAddress("wglAllocateMemoryNV");
    wglFreeMemoryNV                 = (PFNWGLFREEMEMORYNVPROC               )wglGetProcAddress("wglFreeMemoryNV");
    glFlushVertexArrayRangeNV       = (PFNGLFLUSHVERTEXARRAYRANGENVPROC     )wglGetProcAddress("glFlushVertexArrayRangeNV");
    glVertexArrayRangeNV            = (PFNGLVERTEXARRAYRANGENVPROC          )wglGetProcAddress("glVertexArrayRangeNV");

    assert(wglAllocateMemoryNV);
    assert(wglFreeMemoryNV);
    assert(glFlushVertexArrayRangeNV);
    assert(glVertexArrayRangeNV);
}
void ApiLoader::InitializeNV_point_sprite(void)
{
    if(!IsExtensionAvailable("GL_NV_point_sprite"))
    {
        return;
    }
    NVpointSprite = true;

    glPointParameteriNV                = (PFNGLPOINTPARAMETERINVPROC        )wglGetProcAddress("glPointParameteriNV");
    glPointParameterivNV               = (PFNGLPOINTPARAMETERIVNVPROC       )wglGetProcAddress("glPointParameterivNV");
}
void ApiLoader::InitializeATI_vertex_array_object(void)
{
    if(!IsExtensionAvailable("GL_ATI_vertex_array_object"))
    {
        return;
    }
    ATIvertexArray = true;

    glNewObjectBufferATI            = (PFNGLNEWOBJECTBUFFERATIPROC          )wglGetProcAddress("glNewObjectBufferATI");
    glIsObjectBufferATI             = (PFNGLISOBJECTBUFFERATIPROC           )wglGetProcAddress("glIsObjectBufferATI");
    glUpdateObjectBufferATI         = (PFNGLUPDATEOBJECTBUFFERATIPROC       )wglGetProcAddress("glUpdateObjectBufferATI");
    glGetObjectBufferfvATI          = (PFNGLGETOBJECTBUFFERFVATIPROC        )wglGetProcAddress("glGetObjectBufferfvATI");
    glGetObjectBufferivATI          = (PFNGLGETOBJECTBUFFERIVATIPROC        )wglGetProcAddress("glGetObjectBufferivATI");
    glDeleteObjectBufferATI         = (PFNGLDELETEOBJECTBUFFERATIPROC       )wglGetProcAddress("glDeleteObjectBufferATI");
    glArrayObjectATI                = (PFNGLARRAYOBJECTATIPROC              )wglGetProcAddress("glArrayObjectATI");
    glGetArrayObjectfvATI           = (PFNGLGETARRAYOBJECTFVATIPROC         )wglGetProcAddress("glGetArrayObjectfvATI");
    glGetArrayObjectivATI           = (PFNGLGETARRAYOBJECTIVATIPROC         )wglGetProcAddress("glGetArrayObjectivATI");
    glVariantArrayObjectATI         = (PFNGLVARIANTARRAYOBJECTATIPROC       )wglGetProcAddress("glVariantArrayObjectATI");
    glGetVariantArrayObjectfvATI    = (PFNGLGETVARIANTARRAYOBJECTFVATIPROC  )wglGetProcAddress("glGetVariantArrayObjectfvATI");
    glGetVariantArrayObjectivATI    = (PFNGLGETVARIANTARRAYOBJECTIVATIPROC  )wglGetProcAddress("glGetVariantArrayObjectivATI");


    assert(glNewObjectBufferATI);
    assert(glIsObjectBufferATI);
    assert(glUpdateObjectBufferATI);
    assert(glGetObjectBufferfvATI);
    assert(glGetObjectBufferivATI);
    assert(glDeleteObjectBufferATI);
    assert(glArrayObjectATI);
    assert(glGetArrayObjectfvATI);
    assert(glGetArrayObjectivATI);
    assert(glVariantArrayObjectATI);
    assert(glGetVariantArrayObjectfvATI);
    assert(glGetVariantArrayObjectivATI);
}