////////////////////////////////////////////////////////////////////////////////
/*! \file       gfx/exception.h
**  \brief      All exceptions the graphics engine uses.
**  \author     M Jared Finder
**  \date       2/13/2003 9:34:52 PM
*/
////////////////////////////////////////////////////////////////////////////////
#pragma once


////////////////////////////////////////////////////////////////////////////////
// Forward Declarations and needed includes
////////////////////////////////////////////////////////////////////////////////
#include<string.h>


namespace gfx {
////////////////////////////////////////////////////////////////////////////////
/*! \class      Exception
**  \brief      Base exception class for the graphics engine.
**  \author     M Jared Finder
**  \date       2/13/2003 9:38:41 PM
**  \warning    none
*/
////////////////////////////////////////////////////////////////////////////////
class Exception
{
public:
    /// generic error message
    Exception(void)
        { MessageSetup("generic graphics exception");               }
    /// user defined message
    Exception(const char* message)
        { MessageSetup(message);                                    }
    /// find out what happened, if the message was filled
    virtual const char* what(void) const
        { return m_message;                                         }

private:
    ///copy the data over
    void MessageSetup(const char* message)
        { strncpy(m_message, message, 256), m_message[255] = '\0';  }

private:
    char m_message[256];
};


////////////////////////////////////////////////////////////////////////////////
/*! \class      Termination
**  \brief      Permature termination, by the app being forcibly closed.
**  \author     M Jared Finder
**  \date       2/13/2003 9:39:30 PM
**  \warning    none
*/
////////////////////////////////////////////////////////////////////////////////
class Termination : public Exception
{
public:
    Termination(void)
        : Exception("premature termination") {}
    Termination(const char* message)
        : Exception(message) {}
};


////////////////////////////////////////////////////////////////////////////////
/*! \class      BadVersion
**  \brief      Version provided by the drivers is lower than required.
**  \author     M Jared Finder
**  \date       2/13/2003 9:44:47 PM
**  \warning    none
*/
////////////////////////////////////////////////////////////////////////////////
class BadVersion : public Exception
{
public:
    BadVersion(void)
        : Exception("expected feature not found") {}
    BadVersion(const char* message)
        : Exception(message) {}
};


////////////////////////////////////////////////////////////////////////////////
/*! \class      ApiException
**  \brief      A generic exception raised by the API.
**  \author     M Jared Finder
**  \date       2/13/2003 9:45:58 PM
**  \warning    none
*/
////////////////////////////////////////////////////////////////////////////////
class ApiException : public Exception
{
public:
    ApiException(void)
        : Exception("generic hardware failure") {}
    ApiException(const char* message)
        : Exception(message) {}
};


////////////////////////////////////////////////////////////////////////////////
/*! \class      BadParameter
**  \brief      A bad parameter was passed to the API.
**  \author     M Jared Finder
**  \date       2/13/2003 9:46:47 PM
**  \warning    none
*/
////////////////////////////////////////////////////////////////////////////////
class BadParameter : public ApiException
{
public:
    BadParameter(void)
        : ApiException("Internal Error:\n"
                       "A bad parameter was passed") {}
    BadParameter(const char* message)
        : ApiException(message) {}
};


////////////////////////////////////////////////////////////////////////////////
/*! \class      BadOperation
**  \brief      An invalid operation was attempted.
**  \author     M Jared Finder
**  \date       2/13/2003 9:47:41 PM
**  \warning    none
*/
////////////////////////////////////////////////////////////////////////////////
class BadOperation : public ApiException
{
public:
    BadOperation(void)
        : ApiException("Internal Error:\n"
                       "A bad operation was attempted") {}
    BadOperation(const char* message)
        : ApiException(message) {}
};


////////////////////////////////////////////////////////////////////////////////
/*! \class      MemoryAlloc
**  \brief      Memory allocation attempted by the subsytem has failed.
**  \author     M Jared Finder
**  \date       2/13/2003 9:48:32 PM
**  \warning    none
*/
////////////////////////////////////////////////////////////////////////////////
class MemoryAlloc : public ApiException
{
public:
    MemoryAlloc(void)
        : ApiException("out of memory") {}
    MemoryAlloc(const char* message)
        : ApiException(message) {}
};


}