////////////////////////////////////////////////////////////////////////////////
/*! \file       font.cpp
**  \brief      Implement basic font functionality
**  \author     M Jared Finder
**  \date       10/28/2002 9:05:29 PM
*/
////////////////////////////////////////////////////////////////////////////////
#include"font.h"

////////////////////////////////////////////////////////////////////////////////
// Needed includes
////////////////////////////////////////////////////////////////////////////////
//#include"stdheaders.h"
#include<windows.h>
#include<gl/gl.h>
#include"../nexception/nexception.h"
// memory leak detection
#ifdef _DEBUG
    #include"memleak.h"
    #define new DEBUG_NEW
    #undef THIS_FILE
    static char THIS_FILE[] = __FILE__;
#endif


////////////////////////////////////////////////////////////////////////////////
//  FUNCTION:   
/*! \brief      Create a font of the given name and given size
**  \param      fontName    The name of the font.  If this font does not exist,
**                          the system font will be used.
**  \param      fontSize    The height of the font in pixels.  Note that when
**                          the resolution is changed, the font will appear
**                          ``smaller''.
**  \return     none
**  \exception  none
**  \author     M Jared Finder
**  \date       10/28/2002 9:05:50 PM
**  \warning    none
*/
////////////////////////////////////////////////////////////////////////////////
static int size;
gfx::Font::Font(const char* fontName)
        : m_fontBase(glGenLists(256)), m_texture(fontName, true)
{
    //create the lists
    for(int i = 0; i < 256; ++i)
    {
        int xPos = i % 16;
        int yPos = 15 - i / 16;

        float left  = (xPos + 0) / 16.0f;
        float right = (xPos + 1) / 16.0f;
        float bottom= (yPos + 0) / 16.0f;
        float top   = (yPos + 1) / 16.0f;

        glNewList(i + m_fontBase, GL_COMPILE);
        glBegin(GL_TRIANGLE_STRIP);
            glTexCoord2f(left, bottom);
            glVertex2f(  0, 0);
            glTexCoord2f(left, top);
            glVertex2f(  0, 1);
            glTexCoord2f(right, bottom);
            glVertex2f(  1, 0);
            glTexCoord2f(right, top);
            glVertex2f(  1, 1);
        glEnd();
        glTranslatef(10/16.0, 0, 0);
        glEndList();
    }
}


////////////////////////////////////////////////////////////////////////////////
//  FUNCTION:   
/*! \brief      Clean up the font.
**  \return     none
**  \exception  none
**  \par        DESCRIPTION:
**                  
**  \author     M Jared Finder
**  \date       10/28/2002 9:12:46 PM
**  \warning    none
*/
////////////////////////////////////////////////////////////////////////////////
gfx::Font::~Font(void)
{
    glDeleteLists(m_fontBase, 256);
}


////////////////////////////////////////////////////////////////////////////////
//  FUNCTION:   
/*! \brief      Display the string in the specified spot.
**  \param      x,y             Position to render on the string
**  \param      formatString    printf style argument list
**  \return     Number of characters written.  Less than zero if there is an
**              error.
**  \exception  none
**  \author     M Jared Finder
**  \date       10/28/2002 9:13:04 PM
**  \warning    Write assumes a fixed limit of 255 characters.  If the string
**              passed in would expand to over 255 characters, characters 256
**              and up will not be rendered.
*/
////////////////////////////////////////////////////////////////////////////////
int gfx::Font::Write(
        float scaleX, float scaleY,
        float x, float y,
        float r, float g, float b,
        char formatString[], ...)
{
    //set up rendering
    glPushMatrix();
    glListBase(m_fontBase);
    glBindTexture(GL_TEXTURE_2D, m_texture.GetTextureHandle());
    glColor3f(r,g,b);
    glTranslatef(x, y, 0);
    glScalef(scaleX, scaleY, 1);

    //make the string to display
    char buffer[256];
    int  numCharacters;
    va_list varArg;

    va_start(varArg, formatString);
    numCharacters = _vsnprintf(buffer,256, formatString, varArg);
    va_end(varArg);
    if(numCharacters <= 0)
    {
        glPopMatrix();
        return -1;
    }

    //render
    glCallLists(numCharacters, GL_UNSIGNED_BYTE, buffer);

    glPopMatrix();

    return numCharacters;
}