////////////////////////////////////////////////////////////////////////////////
/*! \file       model.cpp
**  \brief      Static 3D models
**  \author     M Jared Finder
**  \date       11/26/2002 6:34:30 PM
**  \par        DESCRIPTION:
**                  Loads and renders 3D models from a file.  Currently supports
**                  MD2 format only.
*/
////////////////////////////////////////////////////////////////////////////////
#include"model.h"

////////////////////////////////////////////////////////////////////////////////
// Needed includes
////////////////////////////////////////////////////////////////////////////////
//#include"stdheaders.h"
#include<windows.h>
#include<gl/gl.h>
#include<stdio.h>
#include"texture.h"
#include"../math/linear_alg/vector3.h"
// memory leak detection
#ifdef _DEBUG
    #include"memleak.h"
    #define new DEBUG_NEW
    #undef THIS_FILE
    static char THIS_FILE[] = __FILE__;
#endif


////////////////////////////////////////////////////////////////////////////////
//  FUNCTION:   
/*! \brief      Load the specified file's frame number.
**  \param      fileName    File to load from
**  \param      frameNumber frame to load from
**  \return     none
**  \exception  none
**  \author     M Jared Finder
**  \date       11/26/2002 6:59:34 PM
**  \warning    none
*/
////////////////////////////////////////////////////////////////////////////////
gfx::Model::Model(const char fileName[], int frameNumber)
    : m_modelData(fileName), m_frameNumber(frameNumber)
{
}


////////////////////////////////////////////////////////////////////////////////
//  FUNCTION:   
/*! \brief      Clean up the model data
**  \param      
**  \return     none
**  \exception  none
**  \author     M Jared Finder
**  \date       11/26/2002 7:01:00 PM
**  \warning    none
*/
////////////////////////////////////////////////////////////////////////////////
gfx::Model::~Model(void)
{
}


////////////////////////////////////////////////////////////////////////////////
//  FUNCTION:   
/*! \brief      Draw the model with the specified data
**  \param      texture     Texture to draw the model with
**  \param      position    Position of the model in world space
**  \param      scale       scale of the model (seperate in X, Y, Z)
**  \param      axis        axis of rotation for the model
**  \param      angle       amount to rotate around the axis counterclockwise.
**  \return     none
**  \exception  none
**  \author     M Jared Finder
**  \date       11/26/2002 7:01:05 PM
**  \warning    none
*/
////////////////////////////////////////////////////////////////////////////////
void
gfx::Model::Draw(
        gfx::Texture& texture,
        math::linear_alg::Vector3<float> position,
        math::linear_alg::Vector3<float> scale,
        math::linear_alg::Vector3<float> axis, float angle,
        float r, float g, float b)
{
    //transform the model
    glPushMatrix();
    glTranslatef(position[0], position[1], position[2]);
    glRotatef(angle, axis[0], axis[1], axis[2]);
    glScalef(scale[0], scale[1], scale[2]);

    //texture the model
    glColor3f(r,g,b);
    glBindTexture(GL_TEXTURE_2D, texture.GetTextureHandle());

    //render the model
    using loader::ModelTriangle;
    using loader::ModelFrame;
    const std::vector<float>&
        texCoord    = m_modelData.GetTextureCoordArray();
    const std::vector<loader::ModelCommandData>&
        commandList = m_modelData.GetCommandList();
    const ModelFrame&
        frame       = m_modelData.GetFrameArray()[m_frameNumber];
    std::vector<loader::ModelCommandData>::const_iterator it = commandList.begin();

    while(it != commandList.end())
    {
        int numVertices = it->i;

        //render triangle strip
        if(numVertices > 0)
        {
            glBegin(GL_TRIANGLE_STRIP);
            ++it;
        }
        //render triangle fan
        else if(numVertices < 0)
        {
            glBegin(GL_TRIANGLE_FAN);
            numVertices *= -1;
            ++it;
        }
        //done
        else
        {
            break;
        }

        while(numVertices > 0)
        {
            float s = it[0].f;
            float t = it[1].f;
            int vertex = it[2].i;

            glTexCoord2f(s, t);
            glVertex3fv(&frame.vertexList[vertex * 3]);

            it += 3;
            --numVertices;
        }
        glEnd();
    }

    glPopMatrix();
}
