////////////////////////////////////////////////////////////////////////////////
/*! \file       multi_model.cpp
**  \brief      Animated 3D models with interpolation between different
**              vertices.
**  \author     M Jared Finder
**  \date       4/8/2003 1:50:55 PM
*/
////////////////////////////////////////////////////////////////////////////////
#include"multi_model.h"

////////////////////////////////////////////////////////////////////////////////
// Needed includes
////////////////////////////////////////////////////////////////////////////////
//#include"stdheaders.h"
#include<windows.h>
#include"opengl.h"
#include<stdio.h>

#include"gfx/texture.h"
#include"math/linear_alg/vector3.h"
// memory leak detection
#ifdef _DEBUG
    #include"memleak.h"
    #define new DEBUG_NEW
    #undef THIS_FILE
    static char THIS_FILE[] = __FILE__;
#endif

using gfx::MultiModel;


////////////////////////////////////////////////////////////////////////////////
//  FUNCTION:   
/*! \brief      Interpolate between two vertices, A and B creating (1-t)A + tB
**  \param      outputVertex    The output iterpolated vertex
**  \param      interpAmount    An interpolation amount from 0 to 1.
**  \param      vertexA         Starting vertex
**  \param      vertexB         Ending vertex
**  \return     none
**  \exception  none
**  \author     M Jared Finder
**  \date       1/30/2003 8:41:04 PM
**  \warning    none
*/
////////////////////////////////////////////////////////////////////////////////
inline void
Interpolate(float outputVertex[3], float interpAmount,
            const float vertexA[3], const float vertexB[3])
{
    outputVertex[0] = (1-interpAmount)*vertexA[0] + interpAmount*vertexB[0];
    outputVertex[1] = (1-interpAmount)*vertexA[1] + interpAmount*vertexB[1];
    outputVertex[2] = (1-interpAmount)*vertexA[2] + interpAmount*vertexB[2];
}


////////////////////////////////////////////////////////////////////////////////
//  FUNCTION:   
/*! \brief      Load the specified files.
**  \param      fileMinAMinB    Reference model to load, where A = 0, B = 0
**  \param      fileMinAMaxB    Reference model to load, where A = 0, B = 1
**  \param      fileMaxAMinB    Reference model to load, where A = 1, B = 0
**  \param      fileMaxAMaxB    Reference model to load, where A = 1, B = 1
**  \return     none
**  \exception  none
**  \author     M Jared Finder
**  \date       11/26/2002 6:59:34 PM
**  \warning    none
*/
////////////////////////////////////////////////////////////////////////////////
MultiModel::MultiModel(const char* fileMinAMinB, const char* fileMinAMaxB,
                       const char* fileMaxAMinB, const char* fileMaxAMaxB)
    : m_modelMinAMinB(fileMinAMinB), m_modelMinAMaxB(fileMinAMaxB),
    m_modelMaxAMinB(fileMaxAMinB), m_modelMaxAMaxB(fileMaxAMaxB)
{
}


////////////////////////////////////////////////////////////////////////////////
//  FUNCTION:   
/*! \brief      Clean up the model data
**  \param      
**  \return     none
**  \exception  none
**  \author     M Jared Finder
**  \date       11/26/2002 7:01:00 PM
**  \warning    none
*/
////////////////////////////////////////////////////////////////////////////////
MultiModel::~MultiModel(void)
{
}


////////////////////////////////////////////////////////////////////////////////
//  FUNCTION:   
/*! \brief      Draw one frame of the model with the specified data
**  \param      weightA     How much to interpolate between models with min and
**                          max A.
**  \param      weightB     How much to interpolate between models with min and
**                          max B.
**  \param      frame       The frame to render
**  \param      texture     Texture to draw the model with
**  \param      position    Position of the model in world space
**  \param      scale       scale of the model (seperate in X, Y, Z)
**  \param      axis        axis of rotation for the model
**  \param      angle       amount to rotate around the axis counterclockwise.
**  \param      r           Red component   (0.0 to 1.0)
**  \param      g           Green component (0.0 to 1.0)
**  \param      b           Blue component  (0.0 to 1.0)
**  \return     none
**  \exception  none
**  \author     M Jared Finder
**  \date       11/26/2002 7:01:05 PM
**  \warning    none
*/
////////////////////////////////////////////////////////////////////////////////
void MultiModel::Draw(float weightA, float weightB, float weightTex,
                      int frame,
                      gfx::Texture& tex1, gfx::Texture& tex2,
                      math::linear_alg::Vector3<float> position,
                      math::linear_alg::Vector3<float> scale,
                      math::linear_alg::Vector3<float> axis, float angle,
                      float r, float g, float b
                      )
{
    //transform the model
    glPushMatrix();
    glTranslatef(position[0], position[1], position[2]);
    glRotatef(angle, axis[0], axis[1], axis[2]);
    glScalef(scale[0], scale[1], scale[2]);

    //texture the model
    glColor4f(r,g,b, weightTex);
    glBindTexture(GL_TEXTURE_2D, tex1.GetTextureHandle());
    glActiveTexture(GL_TEXTURE1);
    glEnable(GL_TEXTURE_2D);
    glBindTexture(GL_TEXTURE_2D, tex2.GetTextureHandle());

    //render the model
    using loader::ModelTriangle;
    using loader::ModelFrame;
    const std::vector<float>&
        texCoord    = m_modelMinAMinB.GetTextureCoordArray();
    const std::vector<loader::ModelCommandData>&
        commandList = m_modelMinAMinB.GetCommandList();
    std::vector<loader::ModelCommandData>::const_iterator it = commandList.begin();

    while(it != commandList.end())
    {
        int numVertices = it->i;

        //render triangle strip
        if(numVertices > 0)
        {
            glBegin(GL_TRIANGLE_STRIP);
            ++it;
        }
        //render triangle fan
        else if(numVertices < 0)
        {
            glBegin(GL_TRIANGLE_FAN);
            numVertices *= -1;
            ++it;
        }
        //done
        else
        {
            break;
        }

        while(numVertices > 0)
        {
            float s = it[0].f;
            float t = it[1].f;

            //interpolate to get a vector
            int vertex = it[2].i;
            const float* vertexMinAMinB = &(m_modelMinAMinB.GetFrameArray()[frame].vertexList[vertex*3]);
            const float* vertexMinAMaxB = &(m_modelMinAMaxB.GetFrameArray()[frame].vertexList[vertex*3]);
            const float* vertexMaxAMinB = &(m_modelMaxAMinB.GetFrameArray()[frame].vertexList[vertex*3]);
            const float* vertexMaxAMaxB = &(m_modelMaxAMaxB.GetFrameArray()[frame].vertexList[vertex*3]);

            float vertexMinA[3];
            float vertexMaxA[3];
            float vertexFinal[3];
            Interpolate(vertexMinA, weightB, vertexMinAMinB, vertexMinAMaxB);
            Interpolate(vertexMaxA, weightB, vertexMaxAMinB, vertexMaxAMinB);
            Interpolate(vertexFinal, weightA, vertexMinA, vertexMaxA);

            glMultiTexCoord2f(GL_TEXTURE0, s, t);
            glMultiTexCoord2f(GL_TEXTURE1, s, t);
            glVertex3fv(&vertexFinal[0]);

            it += 3;
            --numVertices;
        }
        glEnd();
    }

    glDisable(GL_TEXTURE_2D);
    glActiveTexture(GL_TEXTURE0);
    glPopMatrix();
}


////////////////////////////////////////////////////////////////////////////////
//  FUNCTION:   
/*! \brief      Draw one frame of the model with the specified data
**  \param      frameA          Starting frame to draw with
**  \param      frameB          Ending frame to draw with
**  \param      interpAmount    Interpolation value from 0 to 1 for animation.
**  \param      texture     Texture to draw the model with
**  \param      position    Position of the model in world space
**  \param      scale       scale of the model (seperate in X, Y, Z)
**  \param      axis        axis of rotation for the model
**  \param      angle       amount to rotate around the axis counterclockwise.
**  \return     none
**  \exception  none
**  \author     M Jared Finder
**  \date       11/26/2002 7:01:05 PM
**  \warning    none
*/
////////////////////////////////////////////////////////////////////////////////
void MultiModel::Draw(float weightA, float weightB, float weightTex,
                      int frame1, int frame2, float interpAmount,
                      gfx::Texture& tex1, gfx::Texture& tex2,
                      math::linear_alg::Vector3<float> position,
                      math::linear_alg::Vector3<float> scale,
                      math::linear_alg::Vector3<float> axis, float angle,
                      float r, float g, float b
                      )
{
    //transform the model
    glPushMatrix();
    glTranslatef(position[0], position[1], position[2]);
    glRotatef(angle, axis[0], axis[1], axis[2]);
    glScalef(scale[0], scale[1], scale[2]);

    //texture the model
    glColor4f(r,g,b, weightTex);
    glBindTexture(GL_TEXTURE_2D, tex1.GetTextureHandle());
    glActiveTexture(GL_TEXTURE1);
    glEnable(GL_TEXTURE_2D);
    glBindTexture(GL_TEXTURE_2D, tex2.GetTextureHandle());

    //render the model
    using loader::ModelTriangle;
    using loader::ModelFrame;
    const std::vector<float>&
        texCoord    = m_modelMinAMinB.GetTextureCoordArray();
    const std::vector<loader::ModelCommandData>&
        commandList = m_modelMinAMinB.GetCommandList();
    std::vector<loader::ModelCommandData>::const_iterator it = commandList.begin();

    while(it != commandList.end())
    {
        int numVertices = it->i;

        //render triangle strip
        if(numVertices > 0)
        {
            glBegin(GL_TRIANGLE_STRIP);
            ++it;
        }
        //render triangle fan
        else if(numVertices < 0)
        {
            glBegin(GL_TRIANGLE_FAN);
            numVertices *= -1;
            ++it;
        }
        //done
        else
        {
            break;
        }

        while(numVertices > 0)
        {
            float s = it[0].f;
            float t = it[1].f;

            //interpolate it to get a vector
            int vertex = it[2].i;
            const float* vertexMinAMinBMinT = &(m_modelMinAMinB.GetFrameArray()[frame1].vertexList[vertex*3]);
            const float* vertexMinAMaxBMinT = &(m_modelMinAMaxB.GetFrameArray()[frame1].vertexList[vertex*3]);
            const float* vertexMaxAMinBMinT = &(m_modelMaxAMinB.GetFrameArray()[frame1].vertexList[vertex*3]);
            const float* vertexMaxAMaxBMinT = &(m_modelMaxAMaxB.GetFrameArray()[frame1].vertexList[vertex*3]);
            const float* vertexMinAMinBMaxT = &(m_modelMinAMinB.GetFrameArray()[frame2].vertexList[vertex*3]);
            const float* vertexMinAMaxBMaxT = &(m_modelMinAMaxB.GetFrameArray()[frame2].vertexList[vertex*3]);
            const float* vertexMaxAMinBMaxT = &(m_modelMaxAMinB.GetFrameArray()[frame2].vertexList[vertex*3]);
            const float* vertexMaxAMaxBMaxT = &(m_modelMaxAMaxB.GetFrameArray()[frame2].vertexList[vertex*3]);

            float vertexMinAMinB[3];
            float vertexMinAMaxB[3];
            float vertexMaxAMinB[3];
            float vertexMaxAMaxB[3];
            Interpolate(vertexMinAMinB, interpAmount, vertexMinAMinBMinT, vertexMinAMinBMaxT);
            Interpolate(vertexMinAMaxB, interpAmount, vertexMinAMaxBMinT, vertexMinAMaxBMaxT);
            Interpolate(vertexMaxAMinB, interpAmount, vertexMaxAMinBMinT, vertexMaxAMinBMaxT);
            Interpolate(vertexMaxAMaxB, interpAmount, vertexMaxAMaxBMinT, vertexMaxAMaxBMaxT);

            float vertexMinA[3];
            float vertexMaxA[3];
            Interpolate(vertexMinA, weightB, vertexMinAMinB, vertexMinAMaxB);
            Interpolate(vertexMaxA, weightB, vertexMaxAMinB, vertexMaxAMaxB);

            float vertexFinal[3];
            Interpolate(vertexFinal, weightA, vertexMinA, vertexMaxA);

            glMultiTexCoord2f(GL_TEXTURE0, s, t);
            glMultiTexCoord2f(GL_TEXTURE1, s, t);
            glVertex3fv(&vertexFinal[0]);

            it += 3;
            --numVertices;
        }
        glEnd();
    }

    //clean up
    glDisable(GL_TEXTURE_2D);
    glActiveTexture(GL_TEXTURE0);
    glPopMatrix();
}
