////////////////////////////////////////////////////////////////////////////////
/*! \file       gfx/renderable.h
**  \brief      Defines base class objects for the graphics engine.
**  \author     M Jared Finder
**  \date       10/24/2002 4:43:08 PM
**  \par        DESCRIPTION:
**                  The graphics engine has three base class objects
**                      - IRenderable
**                          The common base class for all renderable objects.
**                          IRenderable is a pure virtual class; it is only
**                          used for creating a consistent interface.
**                      - IRenderable3D
**                          Just a holder class.  All 3D renderable objects
**                          must derive from IRenderable3D
**                      - IRenderable2D
**                          Just a holder class.  All 2D renderable objects
**                          must derive from IRenderable2D
**                      - IRenderable3DAlpha
**                          Just a holder class.  All 3D renderable objects
**                          that make use of alpha blending must derive from
**                          IRenderable3DAlpha
**                      - IRenderable2DAlpha
**                          Just a holder class.  All 2D renderable objects
**                          that make use of alpha blending must derive from
**                          IRenderable2DAlpha
**
**  \par        IMPLEMENTATION DETAILS
**  \par        REMARKS:
**                  
*/
////////////////////////////////////////////////////////////////////////////////
#pragma once


////////////////////////////////////////////////////////////////////////////////
// Forward Declarations and needed includes
////////////////////////////////////////////////////////////////////////////////
#include"gfx/renderer.h"


namespace gfx {
////////////////////////////////////////////////////////////////////////////////
/*! \class      IRenderable
**  \brief      The base class for all renderable objects.
**  \author     M Jared Finder
**  \date       10/24/2002 6:24:22 PM
**  \warning    none
*/
////////////////////////////////////////////////////////////////////////////////
class IRenderable
{
public:
    ////////////////////////////////////////////////////////////////////////////////
    //  FUNCTION:   
    /*! \brief      SetRender should be called by subclasses to place an object in
    **              the drawing queue.
    **  \param      
    **  \return     none
    **  \exception  none
    **  \par        DESCRIPTION:
    **                  SetRender should be overriden in each interface subclass to
    **                  properly queue of the request.  For example, IRenderable2D's
    **                  set render will place the object in the 2D rendering queue.
    **                  for example, in IRenderable2D and IRenderable3D
    **  \author     M Jared Finder
    **  \date       10/26/2002 12:38:10 PM
    **  \warning    none
    */
    ////////////////////////////////////////////////////////////////////////////////
    void SetRender(void) {}
    ////////////////////////////////////////////////////////////////////////////////
    //  FUNCTION:   
    /*! \brief      Render is called by the graphics engine to render an object.
    **  \param      
    **  \return     none
    **  \exception  none
    **  \par        DESCRIPTION:
    **                  When Render is called by the graphics engine, the object
    **                  should immediately render itself.
    **  \author     M Jared Finder
    **  \date       10/25/2002 11:17:23 AM
    **  \warning    none
    */
    ////////////////////////////////////////////////////////////////////////////////
    virtual void Render(void) = 0;
};


////////////////////////////////////////////////////////////////////////////////
/*! \class      IRenderable*
**  \brief      Classes to derive from for rendering objects.  If there is a
**              new style of rendering, that class must be written.
**  \author     M Jared Finder
**  \date       10/25/2002 11:26:09 AM
**  \warning    none
*/
////////////////////////////////////////////////////////////////////////////////
class IRenderableWorld : public IRenderable
{
public:
    void SetRender(void)
    {
        Renderer::Get().Enqueue(this);
        IRenderable::SetRender();
    }
};
class IRenderable3D : public IRenderable
{
public:
    void SetRender(void)
    {
        Renderer::Get().Enqueue(this);
        IRenderable::SetRender();
    }
};
class IRenderable2D : public IRenderable
{
public:
    void SetRender(void)
    {
        Renderer::Get().Enqueue(this);
        IRenderable::SetRender();
    }
};
class IRenderable3DAlpha : public IRenderable
{
public:
    void SetRender(void)
    {
        Renderer::Get().Enqueue(this);
        IRenderable::SetRender();
    }
};
class IRenderable2DAlpha : public IRenderable
{
public:
    void SetRender(void)
    {
        Renderer::Get().Enqueue(this);
        IRenderable::SetRender();
    }
};

}