////////////////////////////////////////////////////////////////////////////////
/*! \file       gfx/renderer.cpp
**  \brief      Implementation of graphics engine rendering
**  \author     M Jared Finder
**  \date       10/9/2002 10:52:36 AM
*/
////////////////////////////////////////////////////////////////////////////////
#include"gfx/renderer.h"
using gfx::Renderer;
using gfx::Frustum;


////////////////////////////////////////////////////////////////////////////////
// Needed includes
////////////////////////////////////////////////////////////////////////////////
//#include"stdheaders.h"
#pragma warning(disable:4786)
#include<iostream>
#include<math.h>
#include<sstream>
#include"gfx/opengl.h"
#include"gfx/renderable.h"
#include"gfx/exception.h"
#include"loader/image.h"
#include"globals/globals.h"
#include"math/linear_alg/vector3.h"
// memory leak detection
#ifdef _DEBUG
    #include"memleak.h"
    #define new DEBUG_NEW
    #undef THIS_FILE
    static char THIS_FILE[] = __FILE__;
#endif


/// the static member
Renderer* Renderer::m_pRenderer = NULL;


////////////////////////////////////////////////////////////////////////////////
//  FUNCTION:   
/*! \brief      Renders all objects of a specific type
**  \param      objList     The objects to render
**  \return     none
**  \exception  none
**  \author     M Jared Finder
**  \date       11/8/2002 10:09:49 AM
**  \warning    none
*/
////////////////////////////////////////////////////////////////////////////////
template<typename T> __forceinline
void RenderObjects(std::list<T*>& objList)
{
    std::list<T*>::iterator it;
    for(it = objList.begin(); it != objList.end(); ++it)
    {
        glPushMatrix();
        (*it)->Render();
        glPopMatrix();
    }
    objList.clear();
}


////////////////////////////////////////////////////////////////////////////////
//  FUNCTION:   
/*! \brief      Initialize the graphics renderer.
**  \param      none
**  \return     none
**  \exception  none
**  \par        DESCRIPTION:
**                  Starts up the graphics renderer.  No other function will
**                  work correctly until this function is implemented.
**  \author     M Jared Finder
**  \date       10/9/2002 10:55:39 AM
**  \warning    none
*/
////////////////////////////////////////////////////////////////////////////////
void Renderer::Create(void)
{    
	g_hwnd          = NULL;/////////////////////////////////////////////////////
    g_ScreenWidth   = 0;////////////////////////////////////////////////////////
    g_ScreenHeight  = 0;////////////////////////////////////////////////////////

	
    //create the object
    m_pRenderer = new Renderer;

    //initalize the matrices
    math::linear_alg::Matrix4<float>& view2d = Get().m_viewMatrix2d;
    math::linear_alg::Matrix4<float>& view3d = Get().m_viewMatrix3d;
    math::linear_alg::Matrix4<float>& proj2d = Get().m_projMatrix2d;
    math::linear_alg::Matrix4<float>& proj3d = Get().m_projMatrix3d;

    //all the matrices start out as identities
    math::linear_alg::Matrix4<float> identity;
    identity[0][0] =  1; identity[0][1] =  0; identity[0][2] =  0; identity[0][3] =  0;
    identity[1][0] =  0; identity[1][1] =  1; identity[1][2] =  0; identity[1][3] =  0;
    identity[2][0] =  0; identity[2][1] =  0; identity[2][2] =  1; identity[2][3] =  0;
    identity[3][0] =  0; identity[3][1] =  0; identity[3][2] =  0; identity[3][3] =  1;
    view2d = identity;
    view3d = identity;
    proj2d = identity;
    proj3d = identity;

    //we have a default window

    Get().m_hasWindow       = true;
    Get().m_inFullscreen    = false;
    Get().m_screenshotFlag  = false;

    //we have no world
    Get().m_WorldObject = NULL;
    
    Get().InitProjectInvariants();
}


////////////////////////////////////////////////////////////////////////////////
//  FUNCTION:   
/*! \brief      Cleans up everything.  If you call this more then once, there's
**              most likely something wrong.
**  \param      none
**  \return     none
**  \exception  none
**  \par        DESCRIPTION:
**                  Cleans up after the graphics object.
**  \author     M Jared Finder
**  \date       10/9/2002 11:33:15 AM
**  \warning    none
*/
////////////////////////////////////////////////////////////////////////////////
void Renderer::Destroy(void)
{
    // detach graphics api (if not attached, doesn't do anything!)
    g_hwnd          = NULL;
    g_ScreenWidth   = 0;
    g_ScreenHeight  = 0;

    // destroy the singleton object
    delete m_pRenderer;
}


////////////////////////////////////////////////////////////////////////////////
//  FUNCTION:   
/*! \brief      Singleton accessor
**  \param      none
**  \return     Reference to the singleton object
**  \exception  none
**  \author     M Jared Finder
**  \date       10/9/2002 11:35:20 AM
**  \warning    none
*/
////////////////////////////////////////////////////////////////////////////////
Renderer& Renderer::Get(void)
{
    //return the graphics object
    return *m_pRenderer;
}


////////////////////////////////////////////////////////////////////////////////
//  FUNCTION:   
/*! \brief      Message pump and drawing.
**  \param      none
**  \return     true if there has been no errors this frame
**              false if there has been errors this frame
**  \exception  none
**  \par        DESCRIPTION:
**                  Empties the message pump and draws to the screen.  Call this
**                  once per frame!
**  \author     M Jared Finder
**  \date       10/10/2002 1:01:11 PM
**  \warning    none
*/
////////////////////////////////////////////////////////////////////////////////
bool Renderer::Flush(void)
{
    //windows message loop
    if(m_renderingWindow.MessagePump() == false)
    {
        throw Termination();
    }

    //start drawing
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT | GL_STENCIL_BUFFER_BIT);

    //3d view
    glEnable(GL_DEPTH_TEST);
    glLoadMatrixf(m_viewMatrix3d.Data());

    glDisable(GL_BLEND);
    if(m_WorldObject != NULL)   { m_WorldObject->Render();  }   ///< draw the world
    RenderObjects<IRenderable3D>(m_3DObjects);
    glEnable(GL_BLEND);
    glDepthMask(GL_FALSE);
    RenderObjects<IRenderable3DAlpha>(m_3DAlphaObjects);
    glDepthMask(GL_TRUE);

    //2D view (no zbuffer!)
    glDisable(GL_DEPTH_TEST);
    glMatrixMode(GL_PROJECTION);
    glPushMatrix();

    glLoadIdentity();
    glOrtho(0, 1, 0, 1, 1, -1);
    glMatrixMode(GL_MODELVIEW);
    glLoadMatrixf(m_viewMatrix2d.Data());

    glDisable(GL_BLEND);
    RenderObjects<IRenderable2D>(m_2DObjects);
    glEnable(GL_BLEND);
    RenderObjects<IRenderable2DAlpha>(m_2DAlphaObjects);

    glMatrixMode(GL_PROJECTION);
    glPopMatrix();
    glMatrixMode(GL_MODELVIEW);

    if(m_screenshotFlag)
    {
        SaveScreenshotInternal();
        m_screenshotFlag = false;
    }

    //double buffer
    m_renderingWindow.SwapBuffers();

    //check for errors
    CheckForErrors();

    return true;
}


////////////////////////////////////////////////////////////////////////////////
//  FUNCTION:   
/*! \brief      Checks for errors on the current state
**  \return     none
**  \exception  none
**  \author     M Jared Finder
**  \date       2/14/2003 10:02:41 AM
**  \warning    none
*/
////////////////////////////////////////////////////////////////////////////////
void Renderer::CheckForErrors(void) const
{
    // get the first error
    GLenum ApiError = glGetError();
    if(ApiError == GL_NO_ERROR)
    {
        return;
    }

    // clear the error state
    while(glGetError() != GL_NO_ERROR) {}

    // throw the correct exception
    switch(ApiError)
    {
    case GL_INVALID_ENUM:
        throw gfx::BadParameter("Fatal Internal Error:\n"
                                "GL_INVALID_ENUM\n"
                                "(An unacceptable value is specified for an enumerated argument.)"
                                );
    case GL_INVALID_VALUE:
        throw gfx::BadParameter("Fatal Internal Error:\n"
                                "GL_INVALID_VALUE\n"
                                "(A numeric argument is out of range.)");
    case GL_INVALID_OPERATION:
        throw gfx::BadOperation("Fatal Internal Error:\n"
                                "GL_INVALID_OPERATION\n"
                                "(The specified operation is not allowed in the current state.)");
    case GL_STACK_OVERFLOW:
        throw gfx::BadOperation("Fatal Internal Error:\n"
                                "GL_STACK_OVERFLOW\n"
                                "(This command would cause a stack overflow.)");
    case GL_STACK_UNDERFLOW:
        throw gfx::BadOperation("Fatal Internal Error:\n"
                                "GL_STACK_UNDERFLOW\n"
                                "(This command would cause a stack underflow.)");
    case GL_OUT_OF_MEMORY:
        throw gfx::MemoryAlloc("Fatal Internal Error:\n"
                               "GL_OUT_OF_MEMORY\n"
                               "(There is not enough memory left to execute the command.)");
    default:
        throw gfx::Exception();
    }
}


////////////////////////////////////////////////////////////////////////////////
//  FUNCTION:   
/*! \brief      Returns all supported fullscreen modes
**  \return     A vector of all supported fullscreen modes
**  \exception  none
**  \author     M Jared Finder
**  \date       10/15/2002 9:45:48 PM
**  \warning    none
*/
////////////////////////////////////////////////////////////////////////////////
const std::vector<gfx::Resolution>& Renderer::GetResolutions(void)
{
	if(m_supportedModes.size() == 0)
	{
		m_renderingWindow.GetSupportedResolutions(m_supportedModes);
	}
    return m_supportedModes;
}


////////////////////////////////////////////////////////////////////////////////
//  FUNCTION:   
/*! \brief      Changes the display mode and resizes the window
**  \param      mode        The mode to change to
**  \param      fullScreen  true if we want to use fullscreen.  If fullScreen
**                          is set to false, then only the width and height
**                          of mode are used
**  \return     true        If the resolution could be changed
**              false       If the resolution could not be changed
**  \exception  none
**  \author     M Jared Finder
**  \date       10/15/2002 9:48:38 PM
**  \warning    none
*/
////////////////////////////////////////////////////////////////////////////////
bool Renderer::ChangeResolution(const gfx::Resolution& mode, bool fullScreen)
{
    bool bReturn;

    //change to the correct window type
    bReturn = m_renderingWindow.ChangeResolution(mode, fullScreen);

    if(bReturn)
    {
        g_ScreenWidth   = mode.width;
        g_ScreenHeight  = mode.height;
        glViewport(0,0, mode.width, mode.height);
    }

    //store the current resolution
    m_currentResolution = mode;
    return bReturn;
}


////////////////////////////////////////////////////////////////////////////////
//  FUNCTION:   
/*! \brief      Sets the projection for 3D rendering
**  \param      xFOV    x-field of view angle in DEGREES!
**  \param      yFOV    y-field of view angle in DEGREES!
**  \param      zNear   Position of the near clipping plane
**  \param      zFar    Position of the far clipping plane
**  \return     none
**  \exception  none
**  \author     M Jared Finder
**  \date       12/4/2002 7:11:13 PM
**  \warning    none
*/
////////////////////////////////////////////////////////////////////////////////
void Renderer::SetProjection(
        float xFOV,     float yFOV,
        float zNear,    float zFar)
{
    //create the matrix
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    gluPerspective(yFOV, xFOV/yFOV, zNear, zFar);
    glMatrixMode(GL_MODELVIEW);

    //save the matrix
    float tempMatrix[16];
    glGetFloatv(GL_PROJECTION_MATRIX, &tempMatrix[0]);
    memcpy(m_projMatrix3d.Data(), &tempMatrix[0], sizeof(float)*16);

    CalculateViewFrustum();
}


////////////////////////////////////////////////////////////////////////////////
//  FUNCTION:   
/*! \brief      Set the 3D view position and orientation
**  \param      position   The position the camera is looking from
**  \param      lookAtPt   The point to which the camera is looking.
**  \param      upVector   An up vector
**  \param      up
**  \return     none
**  \exception  none
**  \par        DESCRIPTION:
**                  
**  \author     M Jared Finder
**  \date       12/4/2002 7:22:00 PM
**  \warning    none
*/
////////////////////////////////////////////////////////////////////////////////
void Renderer::Set3DView(
        math::linear_alg::Vector3<float> position,
        math::linear_alg::Vector3<float> lookAtPt,
        math::linear_alg::Vector3<float> upVector)
{
    upVector.Normalize();
    
    //calculate the lookat matrix
    glMatrixMode(GL_MODELVIEW);
    glPushMatrix();
        glLoadIdentity();
        gluLookAt(
                position[0],    position[1],    position[2],
                lookAtPt[0],    lookAtPt[1],    lookAtPt[2],
                upVector[0],    upVector[1],    upVector[2]);

        //store it
        float tempMatrix[16];
        glGetFloatv(GL_MODELVIEW_MATRIX, tempMatrix);
        memcpy(m_viewMatrix3d.Data(), tempMatrix, sizeof(float)*16);
    glPopMatrix();

    //store position, lookat, and up
    m_position = position;
    m_lookAtPt = lookAtPt;
    m_upVector = upVector;

    CalculateViewFrustum();
}


////////////////////////////////////////////////////////////////////////////////
//  FUNCTION:   
/*! \brief      Calculates the view frustum.
**  \return     none
**  \exception  none
**  \par        DESCRIPTION:
**                  
**  \author     M Jared Finder
**  \date       2/28/2003 10:24:43 AM
**  \warning    none
*/
////////////////////////////////////////////////////////////////////////////////
void Renderer::CalculateViewFrustum(void)
{
    //calculate the view frustum planes
    math::linear_alg::Matrix4<float> transform(m_projMatrix3d);
    transform *= m_viewMatrix3d;

    //right plane
    m_frustum.planeNormal[0][0] = transform[3][0] - transform[0][0];
    m_frustum.planeNormal[0][1] = transform[3][1] - transform[0][1];
    m_frustum.planeNormal[0][2] = transform[3][2] - transform[0][2];
    m_frustum.planeOffset[0]    = transform[3][3] - transform[0][3];
    //left plane
    m_frustum.planeNormal[1][0] = transform[3][0] + transform[0][0];
    m_frustum.planeNormal[1][1] = transform[3][1] + transform[0][1];
    m_frustum.planeNormal[1][2] = transform[3][2] + transform[0][2];
    m_frustum.planeOffset[1]    = transform[3][3] + transform[0][3];
    //top plane
    m_frustum.planeNormal[2][0] = transform[3][0] - transform[1][0];
    m_frustum.planeNormal[2][1] = transform[3][1] - transform[1][1];
    m_frustum.planeNormal[2][2] = transform[3][2] - transform[1][2];
    m_frustum.planeOffset[2]    = transform[3][3] - transform[1][3];
    //bottom plane
    m_frustum.planeNormal[3][0] = transform[3][0] + transform[1][0];
    m_frustum.planeNormal[3][1] = transform[3][1] + transform[1][1];
    m_frustum.planeNormal[3][2] = transform[3][2] + transform[1][2];
    m_frustum.planeOffset[3]    = transform[3][3] + transform[1][3];
    //far plane
    m_frustum.planeNormal[4][0] = transform[3][0] - transform[2][0];
    m_frustum.planeNormal[4][1] = transform[3][1] - transform[2][1];
    m_frustum.planeNormal[4][2] = transform[3][2] - transform[2][2];
    m_frustum.planeOffset[4]    = transform[3][3] - transform[2][3];
    //near plane
    m_frustum.planeNormal[5][0] = transform[3][0] + transform[2][0];
    m_frustum.planeNormal[5][1] = transform[3][1] + transform[2][1];
    m_frustum.planeNormal[5][2] = transform[3][2] + transform[2][2];
    m_frustum.planeOffset[5]    = transform[3][3] + transform[2][3];


    //normalize all the normals
    for(int i = 0; i < 6; ++i)
    {
        float planeLength = m_frustum.planeNormal[i].Magnitude();
        m_frustum.planeNormal[i] /= planeLength;
        m_frustum.planeOffset[i] /= planeLength;
    }

    //calculate the view frustum points
    ;
}


////////////////////////////////////////////////////////////////////////////////
//  FUNCTION:   
/*! \brief      Get the 4x4 transformation matrices for 3d / 2d viewing.
**  \param      proj        Will get the projection matrix
**  \param      view        Will get the view orientation matrix
**  \return     none
**  \exception  none
**  \par        DESCRIPTION:
**                  
**  \author     M Jared Finder
**  \date       2/14/2003 6:52:13 PM
**  \warning    none
*/
////////////////////////////////////////////////////////////////////////////////
void Renderer::GetTransform3D(  math::linear_alg::Matrix4<float>& proj,
                                math::linear_alg::Matrix4<float>& view) const
{
    //the screen in parametrized from 0 to 1
    static math::linear_alg::Matrix4<float>
        proj2dInv(0.5f,   0,      0,      0.5f,
                  0,      0.5f,   0,      0.5f,
                  0,      0,      1,      0,
                  0,      0,      0,      1);
    proj = proj2dInv;
    proj*= m_projMatrix3d;
    view = m_viewMatrix3d;
}
void Renderer::GetTransform2D(  math::linear_alg::Matrix4<float>& proj,
                                math::linear_alg::Matrix4<float>& view) const
{
    proj = m_projMatrix2d;
    view = m_viewMatrix2d;
}


////////////////////////////////////////////////////////////////////////////////
//  FUNCTION:   
/*! \brief      Returns the stored position, look-at point, and up vector.
**  \param      position    Camera position
**  \param      lookAtPt    Camera lookat point
**  \param      upVector    Relative up vector
**  \return     none
**  \exception  none
**  \par        DESCRIPTION:
**                  
**  \author     M Jared Finder
**  \date       3/12/2003 9:27:12 PM
**  \warning    none
*/
////////////////////////////////////////////////////////////////////////////////
void Renderer::Get3DView(   math::linear_alg::Vector3<float>& position,
                            math::linear_alg::Vector3<float>& lookAtPt,
                            math::linear_alg::Vector3<float>& upVector) const
{
    position = m_position;
    lookAtPt = m_lookAtPt;
    upVector = m_upVector;
}


////////////////////////////////////////////////////////////////////////////////
//  FUNCTION:   
/*! \brief      Queue a command to save a screenshot
**  \param      fileName    The file to save the last rendered frame to.
**  \return     true    The file was saved
**              false   The file could not be saved for some reason.
**  \exception  none
**  \author     M Jared Finder
**  \date       10/19/2002 3:08:17 PM
**  \warning    none!
*/
////////////////////////////////////////////////////////////////////////////////
bool Renderer::SaveScreenshot(char fileName[])
{
    //save the request
    strncpy(&m_screenshotFile[0], &fileName[0], sizeof(m_screenshotFile));
    m_screenshotFlag = true;

    //Currently nothing happens...
    return true;
}


////////////////////////////////////////////////////////////////////////////////
//  FUNCTION:   
/*! \brief      Saves a screenshot of the current screen to the file specified
**              earlier.
**  \return     none
**  \exception  none
**  \par        DESCRIPTION:
**                  
**  \author     M Jared Finder
**  \date       2/14/2003 8:57:40 PM
**  \warning    none
*/
////////////////////////////////////////////////////////////////////////////////
void Renderer::SaveScreenshotInternal(void)
{
    std::vector<unsigned char> buffer(m_currentResolution.width * m_currentResolution.height * 4);

    ///get the data
    glReadPixels(
            0,0,
            m_currentResolution.width, m_currentResolution.height,
            GL_RGBA, GL_UNSIGNED_BYTE, &buffer[0]);

    loader::Image picture(&buffer[0], m_currentResolution.width, m_currentResolution.height);
    picture.Save(m_screenshotFile);
}


////////////////////////////////////////////////////////////////////////////////
//  FUNCTION:   
/*! \brief      Queue up the rendering of the world
**  \param      world       The world object to render later
**  \return     none
**  \exception  none
**  \par        DESCRIPTION:
**                  
**  \author     M Jared Finder
**  \date       11/1/2002 11:52:52 AM
**  \warning    none
*/
////////////////////////////////////////////////////////////////////////////////
void Renderer::Enqueue(gfx::IRenderableWorld* world)
    { m_WorldObject = world;                }
void Renderer::Enqueue(gfx::IRenderable2D* client)
    { m_2DObjects.push_back(client);        }
void Renderer::Enqueue(gfx::IRenderable3D* client)
    { m_3DObjects.push_back(client);        }
void Renderer::Enqueue(gfx::IRenderable2DAlpha* client)
    { m_2DAlphaObjects.push_back(client);   }
void Renderer::Enqueue(gfx::IRenderable3DAlpha* client)
    { m_3DAlphaObjects.push_back(client);   }


////////////////////////////////////////////////////////////////////////////////
//  FUNCTION:   
/*! \brief      Initializes all assumptions made about OGL's state.
**  \param      none
**  \return     none
**  \exception  none
**  \par        DESCRIPTION:
**                  Any assumptions about the OpenGL state should be placed here.
**  \author     M Jared Finder
**  \date       10/30/2002 7:41:57 PM
**  \warning    none
*/
////////////////////////////////////////////////////////////////////////////////
void Renderer::InitProjectInvariants(void)
{
    //check for proper version
    if(     m_ApiData.majorVersion < 1
            || (m_ApiData.majorVersion == 1 && m_ApiData.minorVersion < 3))
    {
        std::ostringstream o;
        o << "Fatal Graphics Error:\n"
          << "This program requires OpenGL 1.3, but your system only supports "
          << m_ApiData.majorVersion << "." << m_ApiData.minorVersion << ".\n"
          << "Please upgrade your drivers.";

        throw gfx::BadVersion(o.str().c_str());
    }
    if(!m_ApiData.compiledVertexArray)
    {
        throw gfx::BadVersion("Fatal Graphics Error:\n"
                              "This program requires the following features that your system not support:\n"
                              " * Compiled Vertex Arrays\n"
                              "Please upgrade your drivers.");
    }

    //check for reasonable video card limitaitons
    int maxViewport[2];
    int maxTextureSize;
    int maxLights;
    int maxTextureUnits;
    glGetIntegerv(GL_MAX_TEXTURE_UNITS, &maxTextureUnits);
    glGetIntegerv(GL_MAX_VIEWPORT_DIMS, &maxViewport[0]);
    glGetIntegerv(GL_MAX_TEXTURE_SIZE,  &maxTextureSize);
    glGetIntegerv(GL_MAX_LIGHTS,        &maxLights);
    if(maxViewport[0] < 1024 || maxViewport[1] < 768
        || maxTextureSize < 256
        || maxLights < 1
        || maxTextureUnits < 2
 )
    {
        throw gfx::BadVersion("Fatal Graphics Error:\n"
                              "This program requires the following features that your system not support:\n"
                              " * 1024 x 768 Resolution\n"
                              " * 256 x 256 Textures\n"
                              " * 1 Light\n"
                              " * A multitexturing video card\n"
                              "Please upgrade your drivers.");
    }
    m_numTextureUnits = maxTextureUnits;

    //hints
    glHint(GL_FOG_HINT,                     GL_FASTEST);
    glHint(GL_LINE_SMOOTH_HINT,             GL_FASTEST);
    glHint(GL_PERSPECTIVE_CORRECTION_HINT,  GL_FASTEST);
    glHint(GL_POINT_SMOOTH_HINT,            GL_NICEST);
    glHint(GL_POLYGON_SMOOTH_HINT,          GL_FASTEST);

    //allowed action assumptions
    glEnable(GL_CULL_FACE);
    glEnable(GL_TEXTURE_2D);
    glDisable(GL_LIGHTING);
    glDisable(GL_FOG);
    glEnable(GL_POINT_SMOOTH);
    //glEnable(GL_ALPHA_TEST);

    glEnableClientState(GL_VERTEX_ARRAY);
    glEnableClientState(GL_TEXTURE_COORD_ARRAY);

    //implementation of actions
    glCullFace(GL_BACK);
    glFrontFace(GL_CW);
    glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
    glDepthFunc(GL_LESS);
    glClearDepth(1.0);
    glDepthRange(0.0, 1.0);
    glAlphaFunc(GL_GREATER, 0.01f);

    //fogging
    float fogColor[4] = {81/255.0f, 177/255.0f, 219/255.0f, 1.0};
    glClearColor(fogColor[0], fogColor[1], fogColor[2], fogColor[3]);
    glEnable(GL_FOG);
    glFogi(GL_FOG_MODE, GL_EXP);
    glFogf(GL_FOG_DENSITY, 0.01f);
    glFogfv(GL_FOG_COLOR, &fogColor[0]);
    
    //texture interpolation setup
    glActiveTexture(GL_TEXTURE1);
    glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_COMBINE);
    glTexEnvi(GL_TEXTURE_ENV, GL_COMBINE_RGB, GL_INTERPOLATE);
    glTexEnvi(GL_TEXTURE_ENV, GL_SOURCE0_RGB,   GL_TEXTURE);
    glTexEnvi(GL_TEXTURE_ENV, GL_SOURCE1_RGB,   GL_PREVIOUS);
    glTexEnvi(GL_TEXTURE_ENV, GL_SOURCE2_RGB,   GL_PRIMARY_COLOR);
    glTexEnvi(GL_TEXTURE_ENV, GL_OPERAND0_RGB,  GL_SRC_COLOR);
    glTexEnvi(GL_TEXTURE_ENV, GL_OPERAND1_RGB,  GL_SRC_COLOR);
    glTexEnvi(GL_TEXTURE_ENV, GL_OPERAND2_RGB,  GL_SRC_ALPHA);
    float color[] = { 1, 1, 1, 1 };
    glTexEnvfv(GL_TEXTURE_ENV, GL_TEXTURE_ENV_COLOR, color);
    glActiveTexture(GL_TEXTURE0);
}


////////////////////////////////////////////////////////////////////////////////
//  FUNCTION:   
/*! \brief      
**  \param      
**  \return     none
**  \exception  none
**  \par        DESCRIPTION:
**                  
**  \author     M Jared Finder
**  \date       2/21/2003 2:58:13 PM
**  \warning    none
*/
////////////////////////////////////////////////////////////////////////////////
const Frustum& Renderer::GetFrustum(void) const
{
    return m_frustum;
}


////////////////////////////////////////////////////////////////////////////////
//  FUNCTION:   
/*! \brief      Determines is the frustums contains a point.  Useful for other
**              frustum detection (quadtree, octree, etc.).
**  \param      point       Point to test
**  \return     true if the point is in the frustum.  false if the point is
**              outside of the frustum.
**  \exception  none
**  \par        DESCRIPTION:
**                  
**  \author     M Jared Finder
**  \date       2/21/2003 9:28:31 PM
**  \warning    none
*/
////////////////////////////////////////////////////////////////////////////////
bool Frustum::ContainsPoint(const math::linear_alg::Vector3<float>& point) const
{
//*
    for(int i = 0; i < 6; ++i)
    {
        if(math::linear_alg::Dot(planeNormal[i], point) + planeOffset[i] < 0)
        {
            return false;
        }
    }
//*/
    return true;
}



////////////////////////////////////////////////////////////////////////////////
//  FUNCTION:   
/*! \brief      Determines is the frustums contains a sphere.  Useful for
**              bounding sphere detection.
**  \param      center      Center of the sphere
**  \param      radius      Radius of the sphere
**  \return     true if the sphere is contained partially or wholy in the
**              frustum.  false if the sphere is completely outside of the
**              frustum.
**  \exception  none
**  \author     M Jared Finder
**  \date       2/21/2003 11:03:26 AM
**  \warning    none
*/
////////////////////////////////////////////////////////////////////////////////
bool Frustum::ContainsSphere(const math::linear_alg::Vector3<float>& center,
                             float radius) const
{
//*
    for(int i = 0; i < 6; ++i)
    {
        if(math::linear_alg::Dot(planeNormal[i], center) + planeOffset[i] < -radius)
        {
            return false;
        }
    }
//*/
    return true;
}


////////////////////////////////////////////////////////////////////////////////
//  FUNCTION:   
/*! \brief      Determines is the frustums contains an axis-aligned box.
**  \param      pt          Points that determine the extents of the box.
**  \return     true if the box is contained partially or wholy in the
**              frustum.  false if the box is completely outside of the
**              frustum.
**  \exception  none
**  \author     M Jared Finder
**  \date       2/21/2003 6:08:35 PM
**  \warning    If the frustum is contained entirely within the bounding box,
**              then this function will "co
*/
////////////////////////////////////////////////////////////////////////////////
bool Frustum::ContainsBox(const math::linear_alg::Vector3<float> pt[8]) const
{
//*   
    for(int i = 0; i < 6; ++i)
    {
        for(int j = 0; j < 8; ++j)
        {
            if(math::linear_alg::Dot(planeNormal[i], pt[j]) + planeOffset[i] < 0)
            {
                return false;
            }
        }
    }
//
    return true;
}