////////////////////////////////////////////////////////////////////////////////
/*! \file       gfx/renderer.h
**  \brief      Graphics engine rendering
**  \author     M Jared Finder
**  \date       10/8/2002 8:19:49 PM
**  \par        DESCRIPTION:
**                  The Renderer class is responsible for basic 3D functionality
**                  from changing the resolution changing to flushing queued
**                  commands.
**  \par        REMARKS:
**                  The graphics driver currently implements all its
**                  functionality using Win32 OpenGL, and expects compatible
**                  formats.
**  \warning    If you call any subrouting before calling Create or after
**              calling destroy (and before calling Create again) then undefined
**              behavior will result.
**              Think "dereferencing a NULL pointer."
**  \todo       gfx/renderer.h is becoming bloated.  It needs to be seperated
**              into seperate headers as it encompases too much functionality:
**                  - the queueing of rendering commands
**                  - the chanding the of rendering display
**                  - changing the viewport / camera
**                  - general purpose stuff like the view frustum.
**              Everybody using only one or two of these things should NOT be
**              compile-time dependent on all four features!
*/
////////////////////////////////////////////////////////////////////////////////
#pragma once


////////////////////////////////////////////////////////////////////////////////
// Forward Declarations and needed includes
////////////////////////////////////////////////////////////////////////////////
#include<vector>
#include<list>
#include"gfx/structs.h"
#include"gfx/rendering_window.h"
#include"gfx/api_loader.h"
#include"math/linear_alg/vector3.h"
#include"math/linear_alg/vector4.h"
#include"math/linear_alg/matrix4.h"
namespace math { namespace linear_alg
{
    template<typename type> class Vector3;
} }
namespace gfx
{
    class IRenderableWorld;
    class IRenderable2D;
    class IRenderable3D;
    class IRenderable2DAlpha;
    class IRenderable3DAlpha;
}


namespace gfx {
////////////////////////////////////////////////////////////////////////////////
/*! \struct     Frustum      
**  \brief      View frustum collision routines
**  \author     M Jared Finder
**  \date       2/21/2003 9:44:31 AM
**  \warning    none
*/
////////////////////////////////////////////////////////////////////////////////
struct Frustum
{
    ///the collision planes
    math::linear_alg::Vector3<float> planeNormal[6];
    float                            planeOffset[6];
    ///the points defining the collision planes
    math::linear_alg::Vector3<float> vertices[8];

    ///containment test for point
    bool ContainsPoint(const math::linear_alg::Vector3<float>&) const;
    ///containment routine with bounding sphere
    bool ContainsSphere(const math::linear_alg::Vector3<float>&, float) const;
    ///containment test with bounding box
    bool ContainsBox(const math::linear_alg::Vector3<float> pt[8]) const;
};


////////////////////////////////////////////////////////////////////////////////
/*! \class      Renderer
**  \brief      The graphics rendering interface
**  \author     M Jared Finder
**  \date       10/15/2002 7:28:18 PM
*/
////////////////////////////////////////////////////////////////////////////////
class Renderer
{
public:
    /// singleton creation
    static void Create(void);
    /// singleton destruction
    static void Destroy(void);
    /// signleton access
    static Renderer& Get(void);
    
    /// check for error states
    void CheckForErrors(void) const;

    /// flushes all queued commands and performs any processing needed on the
    /// window
    bool Flush(void);

    /// get the list of possible fullscreen resolutions
    const std::vector<Resolution>& GetResolutions(void);
    /// create a window & change resolution (possibly)
    bool ChangeResolution(const Resolution&, bool fullScreen);

    /// set the 3D projection amount
    void SetProjection(
            float xFOV,     float yFOV,
            float zNear,    float zFar);
    /// set the 3D view orientation
    void Set3DView(
            math::linear_alg::Vector3<float> position,
            math::linear_alg::Vector3<float> lookAtPt,
            math::linear_alg::Vector3<float> upVector);
    /// set the 2D view position
    //void Set2DView(void)

    /// recalculate the view frustum
    void CalculateViewFrustum(void);

    /// get the matrices used for 3d viewing
    void GetTransform3D(
        math::linear_alg::Matrix4<float>& proj,
        math::linear_alg::Matrix4<float>& view) const;

    /// get the matrcies used for 2d viewing
    void GetTransform2D(
        math::linear_alg::Matrix4<float>& proj,
        math::linear_alg::Matrix4<float>& view) const;
    
    /// get the 3D view
    void Get3DView(
            math::linear_alg::Vector3<float>& position,
            math::linear_alg::Vector3<float>& lookAtPt,
            math::linear_alg::Vector3<float>& upVector) const;

    /// get the 2D view
    //void Set2DView(void)
    
    /// check if a specific hardware accelerated feature is enabled
    /// (pass a extension paramet
    const ApiLoader& IsEnabled(void) const
        { return m_ApiData; }
    
    int GetNumTextureUnits(void) const
        { return m_numTextureUnits; }
    
    /// get the current view frustum
    const Frustum& GetFrustum(void) const;

    /// save a screenshot of the current screen
    bool SaveScreenshot(char fileName[]);

    //client interfaces
        /// world client interface
        void Enqueue(IRenderableWorld*);
        /// 2D client interface
        void Enqueue(IRenderable2D*);
        /// 3D client interface
        void Enqueue(IRenderable3D*);
        /// 2D alpha client interface
        void Enqueue(IRenderable2DAlpha*);
        /// 3D alpha client interface
        void Enqueue(IRenderable3DAlpha*);
protected:
    /// actually saves a screenshot
    void SaveScreenshotInternal(void);

    /// initializes the invariants expected for the game
    void InitProjectInvariants(void);
private:
    /// singleton
    static Renderer* m_pRenderer;

    ///OS Dependent data    -- Win32 dependent impl
    RenderingWindow m_renderingWindow;

    ///Graphical API loader -- Win32 dependent impl
    ApiLoader       m_ApiData;

    ///current resolution
    Resolution m_currentResolution;

    ///should we take a screencap?
    bool m_screenshotFlag;
    ///where should we take a screenshot from?
    char m_screenshotFile[1024];

    //display mode stuff
        ///do we have a window?
        bool m_hasWindow;
        ///are we in fullscreen?
        bool m_inFullscreen;
        ///supported modes
        std::vector<Resolution> m_supportedModes;

    //matrices
        ///the projection for 3D view
        math::linear_alg::Matrix4<float> m_projMatrix3d;
        ///the view for 3D view
        math::linear_alg::Matrix4<float> m_viewMatrix3d;
        ///the projection for 2D view
        math::linear_alg::Matrix4<float> m_projMatrix2d;
        ///the view for 2D view
        math::linear_alg::Matrix4<float> m_viewMatrix2d;
        ///current view frustum
        gfx::Frustum m_frustum;
        ///current position
        math::linear_alg::Vector3<float> m_position;
        ///current look-at point
        math::linear_alg::Vector3<float> m_lookAtPt;
        ///current up vector
        math::linear_alg::Vector3<float> m_upVector;

    //cached data
        ///the stored world
        IRenderableWorld* m_WorldObject;
        ///the stored 2D objects
        std::list<IRenderable2D*>       m_2DObjects;
        ///stored 3D objects
        std::list<IRenderable3D*>       m_3DObjects;
        ///stored 2D alpha object
        std::list<IRenderable2DAlpha*>  m_2DAlphaObjects;
        ///sorted 3D alpha objects
        std::list<IRenderable3DAlpha*>  m_3DAlphaObjects;

    //opengl limitations
        ///the number of texture units
        int m_numTextureUnits;
};


}