////////////////////////////////////////////////////////////////////////////////
/*! \file       gfx/rendering_window.cpp
**  \brief      Contains implementation of all the OS dependent Driver
**              information
**  \author     M Jared Finder
**  \date       10/15/2002 7:56:08 PM
**  \par        REMARKS:
**                  Only the driver should interface with the OS Dependent data;
**                  even the rest of the graphics library should be shielded.
*/
////////////////////////////////////////////////////////////////////////////////
#include"gfx/rendering_window.h"
using gfx::RenderingWindow;


////////////////////////////////////////////////////////////////////////////////
// Needed includes
////////////////////////////////////////////////////////////////////////////////
//#include"stdheaders.h"
#include<windows.h>
#include"gfx/opengl.h"
#include"gfx/structs.h"
#include"gfx/exception.h"
#include"globals/globals.h"
// memory leak detection
#ifdef _DEBUG
    #include"memleak.h"
    #define new DEBUG_NEW
    #undef THIS_FILE
    static char THIS_FILE[] = __FILE__;
#endif
#define IDI_APPICON 101


namespace
{
    const LONG windowedStyle        = WS_CAPTION | WS_SYSMENU | WS_MINIMIZEBOX
                                        | WS_CLIPSIBLINGS | WS_CLIPCHILDREN;
    const LONG fullscreenStyle      = WS_POPUP
                                        | WS_CLIPSIBLINGS | WS_CLIPCHILDREN;
    const char titleBar[]           = "Aquarius";
}



////////////////////////////////////////////////////////////////////////////////
//  FUNCTION:   
/*! \brief      The OSDependent information constructor.
**  \return     none
**  \exception  none
**  \par        DESCRIPTION:
**                  Initializes all OSDependent info.  For Win32, it records the
**                  procedure address (HINSTANCE) and the window class (ATOM).
**  \author     M Jared Finder
**  \date       10/15/2002 7:58:51 PM
*/
////////////////////////////////////////////////////////////////////////////////
RenderingWindow::RenderingWindow(void)
        : m_hWnd(NULL),             //< window handle
          m_hdc(NULL),              //< window context
          m_isFullscreen(false)     //< and in windowed mode
{
    MakeWindow();
    AttachGraphics();
}


////////////////////////////////////////////////////////////////////////////////
//  FUNCTION:   
/*! \brief      Destructor.  Will detach OGL, destroy the window, and change the
**              resolution back.
**  \return     none
**  \exception  none
**  \author     M Jared Finder
**  \date       10/15/2002 8:08:08 PM
**  \warning    none
*/
////////////////////////////////////////////////////////////////////////////////
RenderingWindow::~RenderingWindow(void)
{
    //detach OpenGL
    wglMakeCurrent(NULL, NULL);

    //destroy the window
    if(m_hWnd)
    {
        DestroyWindow(m_hWnd);
        m_hWnd = NULL;
    }
}


////////////////////////////////////////////////////////////////////////////////
//  FUNCTION:   
/*! \brief      Swaps the buffers in double buffering
**  \return     none
**  \exception  none
**  \author     M Jared Finder
**  \date       10/16/2002 9:58:33 AM
**  \warning    none
*/
////////////////////////////////////////////////////////////////////////////////
void RenderingWindow::SwapBuffers(void)
{
    glFlush();
    ::SwapBuffers(m_hdc);
}


////////////////////////////////////////////////////////////////////////////////
//  FUNCTION:   
/*! \brief      Creates a list of all supported resolutions (at least 16bit)
**  \param      displayModeList     The list to fill in.
**  \return     none
**  \exception  none
**  \author     M Jared Finder
**  \date       10/15/2002 9:07:10 PM
**  \warning    none
*/
////////////////////////////////////////////////////////////////////////////////
void RenderingWindow::GetSupportedResolutions(
        std::vector<gfx::Resolution>& resList) const
{
    //clear the list
    std::vector<Resolution> newList;

    //get all possible modes
    DEVMODE temp;
    Resolution newMode;
    int modeNumber = 0;
    while(EnumDisplaySettings(NULL, modeNumber, &temp) != 0)
    {
        if(temp.dmBitsPerPel >= 16)
        {
            newMode.width =     temp.dmPelsWidth;
            newMode.height =    temp.dmPelsHeight;
            newMode.frequency = temp.dmDisplayFrequency;
            newMode.bpp =       temp.dmBitsPerPel;

            newList.push_back(newMode);
        }
        ++modeNumber;
    }

    resList.swap(newList);
}


////////////////////////////////////////////////////////////////////////////////
//  FUNCTION:   
/*! \brief      Changes the resolution to the display mode specified by mode.
**  \param      mode        The display mode to change to.
**  \return     none
**  \exception  none
**  \par        DESCRIPTION:
**                  
**  \author     M Jared Finder
**  \date       10/15/2002 9:59:41 PM
**  \warning    none
*/
////////////////////////////////////////////////////////////////////////////////
bool RenderingWindow::ChangeResolution(const gfx::Resolution& mode, bool fullscreen)
{
    if(fullscreen)
    {
        ResizeWindow(mode.width, mode.height, true);

        //change the resolution
        DEVMODE newRes;
        memset(&newRes, 0, sizeof(DEVMODE));
        newRes.dmSize =             sizeof(DEVMODE);
        newRes.dmPelsWidth =        mode.width;
        newRes.dmPelsHeight =       mode.height;
        newRes.dmBitsPerPel =       mode.bpp;
        newRes.dmDisplayFrequency = mode.frequency;
        newRes.dmFields =           DM_PELSWIDTH | DM_PELSHEIGHT |
                                    DM_BITSPERPEL | DM_DISPLAYFREQUENCY;

        if(ChangeDisplaySettings(&newRes, CDS_FULLSCREEN) != DISP_CHANGE_SUCCESSFUL)
        {
            return false;
        }

        m_isFullscreen = true;
    }
    else
    {
        //make sure we have the default resolution
        ChangeDisplaySettings(NULL, 0);

        //resize the window
        ResizeWindow(mode.width, mode.height, false);

        m_isFullscreen = false;
    }

    return true;
}


////////////////////////////////////////////////////////////////////////////////
//  FUNCTION:   
/*! \brief      Pumps all the messages.  Also informs if the window has been
**              closed
**  \return     true        If the application should continue existing
**              false       If the application should terminate
**  \exception  none
**  \par        DESCRIPTION:
**                  
**  \author     M Jared Finder
**  \date       10/15/2002 9:17:39 PM
**  \warning    none
*/
////////////////////////////////////////////////////////////////////////////////
bool RenderingWindow::MessagePump(void)
{
    //do a peek message loop
    MSG msg;
    while(PeekMessage(&msg,NULL,0,0,PM_REMOVE))
    {
        //should we quit?
        if(msg.message == WM_QUIT)
        {
            return false;
        }

        TranslateMessage(&msg);
        DispatchMessage(&msg);
    }

    return true;
}


////////////////////////////////////////////////////////////////////////////////
//  FUNCTION:   
/*! \brief      Makes a window suitable for the rendering
**  \return     none
**  \exception  none
**  \author     M Jared Finder
**  \date       2/13/2003 6:33:24 PM
**  \warning    none
*/
////////////////////////////////////////////////////////////////////////////////
void RenderingWindow::MakeWindow(void)
{
    //register the window class
    WNDCLASS wndClass;
    wndClass.style =        CS_HREDRAW | CS_VREDRAW | CS_OWNDC;
    wndClass.lpfnWndProc =  &WindowProc;
    wndClass.cbClsExtra =   0;
    wndClass.cbWndExtra =   0;
    wndClass.hInstance =    GetModuleHandle(NULL);
    wndClass.hIcon =        LoadIcon(GetModuleHandle(NULL), MAKEINTRESOURCE(IDI_APPICON));
    wndClass.hCursor =      LoadCursor(NULL, IDC_ARROW);
    wndClass.hbrBackground= reinterpret_cast<HBRUSH>(COLOR_WINDOW + 1);
    wndClass.lpszMenuName = NULL;
    wndClass.lpszClassName= "GfxEngine"__DATE__;
    ATOM windowClass = RegisterClass(&wndClass);
    if(windowClass == 0)
    {
        throw gfx::ApiException("Could not create internal Window Style");
    }

    //create a window
    m_hWnd = CreateWindow(
            reinterpret_cast<char*>(windowClass),
            titleBar,
            windowedStyle,
            CW_USEDEFAULT,CW_USEDEFAULT,
            CW_USEDEFAULT,CW_USEDEFAULT,
            NULL,
            NULL,
            GetModuleHandle(NULL),
            this);
    if(m_hWnd == NULL)
    {
        throw gfx::ApiException("Could not create internal window");
    }
    m_hdc = GetDC(m_hWnd);

    ////////////////////////////////////////////////////////////////////////////
    // Store the HWND so that other people can use it.
    ////////////////////////////////////////////////////////////////////////////
    g_hwnd = m_hWnd;
}


////////////////////////////////////////////////////////////////////////////////
//  FUNCTION:   
/*! \brief      Attaches the graphics subsystem to the current window
**  \param      
**  \return     none
**  \exception  none
**  \par        DESCRIPTION:
**                  
**  \author     M Jared Finder
**  \date       2/13/2003 6:37:20 PM
**  \warning    none
*/
////////////////////////////////////////////////////////////////////////////////
void RenderingWindow::AttachGraphics(void)
{
    //create the rendering context
    PIXELFORMATDESCRIPTOR pfd = { sizeof(PIXELFORMATDESCRIPTOR) };
    pfd.nVersion        = 1;                    //< required value
    pfd.dwFlags         = PFD_DRAW_TO_WINDOW    //< draw directly to the window
                          | PFD_SUPPORT_OPENGL  //< with OGL support
                          | PFD_DOUBLEBUFFER;   //< and double buffering
    pfd.iPixelType      = PFD_TYPE_RGBA;        //< we don't want palettes
    pfd.cColorBits      = 32;                   //< BPP
    pfd.cAlphaBits      = 0;                    //< alpha channel's BPP
    pfd.cAccumBits      = 0;                    //< accumulation buffer's BPP
    pfd.cDepthBits      = 16;                   //< depth buffer's BPP
    pfd.cStencilBits    = 0;                    //< stencil buffer's BPP
    pfd.cAuxBuffers     = 0;                    //< number of axuilary buffers
    pfd.iLayerType      = PFD_MAIN_PLANE;       //< want direct rendering

    //get the pixel format number
    int nPixelFormat = ChoosePixelFormat(m_hdc, &pfd);
    if(nPixelFormat == 0)
    {
        throw gfx::BadVersion("No sutable hardware accelerated pixel format");
    }
    SetPixelFormat(m_hdc, nPixelFormat, &pfd);

    //make this the current rendering context
    wglMakeCurrent(m_hdc, wglCreateContext(m_hdc));
}


////////////////////////////////////////////////////////////////////////////////
//  FUNCTION:   
/*! \brief      Changes the window's size to be the proper style and size
**  \param      w           Desired new width
**  \param      h,          Desired new height
**  \param      fullscreen  true if the new window should be used for fullscreen
**  \return     none
**  \exception  none
**  \author     M Jared Finder
**  \date       2/13/2003 6:40:36 PM
**  \warning    none
*/
////////////////////////////////////////////////////////////////////////////////
void RenderingWindow::ResizeWindow(int w, int h, bool fullscreen)
{
    //get the new size of the window
    RECT rect = {0,0,w,h};

    if(fullscreen)
    {
        //change the window's style
        SetWindowLong(m_hWnd, GWL_STYLE, fullscreenStyle);
        SetWindowPos(m_hWnd, HWND_TOPMOST,
                     0,0,0,0,
                     SWP_NOMOVE | SWP_NOSIZE);
        
        //change the window's size
        AdjustWindowRect(&rect, fullscreenStyle, FALSE);
        SetWindowPos(
            m_hWnd,
            NULL,0,0,           ///< ignored parameters
            rect.right - rect.left,
            rect.bottom - rect.top,
            SWP_SHOWWINDOW | SWP_NOZORDER
            );
    }
    else
    {
        //change the window's style
        SetWindowLong(m_hWnd, GWL_STYLE, windowedStyle);
        SetWindowPos(m_hWnd, HWND_TOP,
                     0,0,0,0,
                     SWP_NOMOVE | SWP_NOSIZE);

        //change the window's size
        AdjustWindowRect(&rect, windowedStyle, FALSE);
        SetWindowPos(
            m_hWnd,
            NULL,0,0,           ///< ignored parameters
            rect.right - rect.left,
            rect.bottom - rect.top,
            SWP_SHOWWINDOW | SWP_NOMOVE | SWP_NOZORDER
            );
    }
}


////////////////////////////////////////////////////////////////////////////////
//  FUNCTION:   
/*! \brief      The Win32 WindowProc
**  \param      hWnd        Handle to the window getting a message
**  \param      msg         Message to handle
**  \param      wParam      Message dependent.  See MSDN.
**  \param      lParam      Message dependent.  See MSDN.
**  \return     Message dependent.  See MSDN.
**  \exception  none
**  \author     M Jared Finder
**  \date       10/15/2002 9:30:50 PM
*/
////////////////////////////////////////////////////////////////////////////////
LRESULT CALLBACK RenderingWindow::WindowProc(HWND hWnd, UINT msg,
                                             WPARAM wParam, LPARAM lParam)
{
    static RenderingWindow* pThis;

    switch(msg)
    {
    //the window is being created
    case WM_CREATE:
        {
            CREATESTRUCT* cs;
            cs = reinterpret_cast<CREATESTRUCT*>(lParam);
            pThis = reinterpret_cast<RenderingWindow*>(cs->lpCreateParams);
            return 0;
        }

    //the window is being closed
    case WM_CLOSE:
        {
            //make sure OGL can be cleaned up
            wglMakeCurrent(NULL, NULL);
            DestroyWindow(hWnd);
            return 0;
        }

    //The window is about to be destroyed
    case WM_DESTROY:
        {
            PostQuitMessage(0);
            return 0;
        }

    //the application is gaining or losing focus
    case WM_ACTIVATEAPP:
        {
            bool beingActivated = (wParam == TRUE);
            if(pThis->m_isFullscreen && !beingActivated)
            {
                Resolution res;
                res.width       = 640;
                res.height      = 480;
                res.bpp         = 32;
                res.frequency   = 0;
                pThis->ChangeResolution(res, false);
            }

            break;
        }
    }

    //we haven't handled the message, return default
    return DefWindowProc(hWnd, msg, wParam,lParam);
}
