////////////////////////////////////////////////////////////////////////////////
/*! \file       sprite.cpp
**  \brief      Sprite functionality
**  \author     M Jared Finder
**  \date       10/30/2002 2:45:48 PM
**  \par        DESCRIPTION:
**                  
**  \par        REMARKS:
**                  A sprite is stored as a vertex array and a texture array.
**                  The coordinates are ordered clockwise, starting from
**                  the bottom-left:
**
**                  1               3
**                   +-------------+
**                   |\            |
**                   |  \          |
**                   |    \        |
**                   |      \      |
**                   |        \    |
**                   |          \  |
**                   |            \|
**                   +-------------+
**                  0               2
*/
////////////////////////////////////////////////////////////////////////////////
#include"sprite.h"

////////////////////////////////////////////////////////////////////////////////
// Needed includes
////////////////////////////////////////////////////////////////////////////////
//#include"stdheaders.h"
#include<windows.h>
#include<gl/gl.h>
#include"texture.h"
// memory leak detection
#ifdef _DEBUG
    #include"memleak.h"
    #define new DEBUG_NEW
    #undef THIS_FILE
    static char THIS_FILE[] = __FILE__;
#endif


////////////////////////////////////////////////////////////////////////////////
//  FUNCTION:   
/*! \brief      Initialize a sprite.
**  \param      tex         Texture to use
**  \param      width       Width of the sprite (in normalized coords)
**  \param      height      Height of the sprite (in normalized coords)
**  \param      originX     Initial origin x-coord.  Defaults to 0
**  \param      originY     Initial origin y-coord.  Defaults to 0
**  \return     none
**  \exception  none
**  \author     M Jared Finder
**  \date       10/30/2002 2:52:17 PM
**  \warning    none
*/
////////////////////////////////////////////////////////////////////////////////
gfx::Sprite::Sprite(
        const Texture& tex,
        float width, float height,
        float originX, float originY
        )
        : m_textureHandle(tex.GetTextureHandle()),
        m_originX(originX), m_originY(originY)
{
    //set the vertex coordinates
    SetSize(width,height);

    //set the texture coordinates (m_vertexCoords is an array of 2D texture coords)
    SetTexture(tex);
}


////////////////////////////////////////////////////////////////////////////////
//  FUNCTION:   
/*! \brief      Set the texture to use.
**  \param      tex         The texture to use
**  \return     none
**  \exception  none
**  \par        DESCRIPTION:
**                  Recalculates the texture coords and saves the current texture
**  \author     M Jared Finder
**  \date       10/30/2002 3:07:31 PM
**  \warning    none
*/
////////////////////////////////////////////////////////////////////////////////
void
gfx::Sprite::SetTexture(const Texture& tex)
{
    //save the texture handle
    m_textureHandle = tex.GetTextureHandle();

    //recalculate the texture coords
    m_textureCoords[0*2 + 0] = m_textureCoords[1*2 + 0] = tex.GetLeft();
    m_textureCoords[2*2 + 0] = m_textureCoords[3*2 + 0] = tex.GetRight();

    m_textureCoords[0*2 + 1] = m_textureCoords[2*2 + 1] = tex.GetBottom();
    m_textureCoords[1*2 + 1] = m_textureCoords[3*2 + 1] = tex.GetTop();
}


////////////////////////////////////////////////////////////////////////////////
//  FUNCTION:   
/*! \brief      Sets the size of the sprite.  The sprite will be centered about
**              the current origin.
**  \param      width       Width of the sprite (in normalized coordinates)
**  \param      height      Height of the sprite (in normalized coordinates)
**  \return     none
**  \exception  none
**  \par        DESCRIPTION:
**                  
**  \author     M Jared Finder
**  \date       10/30/2002 3:10:11 PM
**  \warning    none
*/
////////////////////////////////////////////////////////////////////////////////
void
gfx::Sprite::SetSize(float width, float height)
{
    //left
    m_vertexCoords[0*2 + 0] = m_vertexCoords[1*2 + 0] = -m_originX;
    //right
    m_vertexCoords[2*2 + 0] = m_vertexCoords[3*2 + 0] = width - m_originX;
    //bottom
    m_vertexCoords[0*2 + 1] = m_vertexCoords[2*2 + 1] = -m_originY;
    //top
    m_vertexCoords[1*2 + 1] = m_vertexCoords[3*2 + 1] = height - m_originY;
}


////////////////////////////////////////////////////////////////////////////////
//  FUNCTION:   
/*! \brief      Sets the origin.  Will recalculate the vertex coordinates.
**  \param      x           The horizontal origin
**  \param      y           Vertical origin
**  \return     none
**  \exception  none
**  \par        DESCRIPTION:
**                  
**  \author     M Jared Finder
**  \date       10/30/2002 3:18:53 PM
**  \warning    none
*/
////////////////////////////////////////////////////////////////////////////////
void
gfx::Sprite::SetOrigin(float x, float y)
{
    int i;

    //adjust vertex coordinates
    for(i = 0; i < 4; ++i)
    {
        //adjust horizontally
        m_vertexCoords[i*2 + 0] += m_originX - x;
        //adjust vertically
        m_vertexCoords[i*2 + 1] += m_originY - y;
    }

    //save the new origin
    m_originX = x;
    m_originY = y;
}


////////////////////////////////////////////////////////////////////////////////
//  FUNCTION:   
/*! \brief      Draw the sprite at the specified location
**  \param      x           x coordinate of the sprite's origin
**  \param      y           y coordinate of the sprite's origin
**  \param      rotation    Rotation amount
**  \param      scaleX      x scaling amount
**  \param      scaleY      y scaling amount
**  \return     none
**  \exception  none
**  \par        DESCRIPTION:
**                  
**  \author     M Jared Finder
**  \date       10/30/2002 3:22:31 PM
**  \warning    none
*/
////////////////////////////////////////////////////////////////////////////////
void
gfx::Sprite::Draw(float x, float y,
                  float rotation,
                  float scaleX, float scaleY,
                  float r, float g, float b)
{
    glPushMatrix();
    
    //use the information
    glColor3f(r,g,b);
    glBindTexture(GL_TEXTURE_2D, m_textureHandle);

    //transform
    glTranslatef(x, y, 0);
    glRotatef(rotation, 0,0,1);
    glScalef(scaleX, scaleY, 1);

    //draw
    glTexCoordPointer(2, GL_FLOAT, 0, m_textureCoords);
    glVertexPointer(2, GL_FLOAT, 0, m_vertexCoords);
    glDrawArrays(GL_TRIANGLE_STRIP, 0, 4);

    glPopMatrix();
}
