////////////////////////////////////////////////////////////////////////////////
/*! \file       vtx_buffer.cpp
**  \brief      Places vertex buffers at accelerated memory (AGP memory or on
**              the card) for faster rendering.
**  \author     M Jared Finder
**  \date       2/27/2003 5:54:30 PM
*/
////////////////////////////////////////////////////////////////////////////////
#include"vtx_buffer.h"
using gfx::VertexBuffer;


////////////////////////////////////////////////////////////////////////////////
// Needed includes
////////////////////////////////////////////////////////////////////////////////
//#include"stdheaders.h"
#include"gfx/opengl.h"
#include<stdlib.h>
#include<assert.h>
#include"gfx/renderer.h"
// memory leak detection
#ifdef _DEBUG
    #include"memleak.h"
    #define new DEBUG_NEW
    #undef THIS_FILE
    static char THIS_FILE[] = __FILE__;
#endif


////////////////////////////////////////////////////////////////////////////////
//  FUNCTION:   
/*! \brief      Create a vertex buffer with default data.
**  \param      size        Desired size of the buffer.
**  \param      isDynamic   Can have two values, true and false.
**                          On true, the buffer is stored in AGP memory for
**                          fast writes, slow reads.
**                          On false, the buffer is stored in video memory for
**                          slow reads and writes.
**                          If you need fast reading and writing, store a local
**                          copy
**  \return     none
**  \exception  none
**  \par        DESCRIPTION:
**                  
**  \author     M Jared Finder
**  \date       2/27/2003 5:56:45 PM
**  \warning    none
*/
////////////////////////////////////////////////////////////////////////////////
VertexBuffer::VertexBuffer(int size, bool isDynamic)
    : m_pData(NULL), m_numElements(0), m_atiHandle(0),
    m_usingNV(false), m_usingATI(false)
{
    // can we use NV extensions?
    if(gfx::Renderer::Get().IsEnabled().NVvertexArray)
    {
        if(!isDynamic)
        {
            void* pTemp = wglAllocateMemoryNV(size, 0,0,1);
            if(pTemp != NULL)
            {
                glVertexArrayRangeNV(size, pTemp);
                m_pData = pTemp;
                m_usingNV = true;

                return;
            }
        }
    }

    // how about ATI?
    if(gfx::Renderer::Get().IsEnabled().ATIvertexArray)
    {
        // no additional setup
    }

    // fall back to system memory
    {
        float* pTemp = new float[size];
        if(pTemp == NULL)
        {
            throw std::bad_alloc();
        }
        m_numElements = size;
        m_pData = pTemp;

        return;
    }
}


////////////////////////////////////////////////////////////////////////////////
//  FUNCTION:   
/*! \brief      Deallocate any resources, and deallocate any data.
**  \param      
**  \return     none
**  \exception  none
**  \par        DESCRIPTION:
**                  
**  \author     M Jared Finder
**  \date       2/27/2003 6:02:46 PM
**  \warning    none
*/
////////////////////////////////////////////////////////////////////////////////
VertexBuffer::~VertexBuffer(void)
{
    // can we use NV extensions?
    if(m_usingNV)
    {
        glFlushVertexArrayRangeNV();
        wglFreeMemoryNV(m_pData);
    }
    // how about ATI?
    else if(m_usingATI)
    {
        assert(false);
    }
    // fall back to system memory
    else
    {
        if(m_numElements > 0)
        {
            delete[] m_pData;
        }
    }
}


////////////////////////////////////////////////////////////////////////////////
//  FUNCTION:   
/*! \brief      Activates the vertex buffer as the correct vertex.  Acts just
**              like glVertexPointer / glTexturePointer, etc.
**  \param      array       Which array to activate
**  \param      size        Number of coordinates per vertex
**  \param      stride      Stride between the data.
**  \return     none
**  \exception  none
**  \par        DESCRIPTION:
**                  
**  \author     M Jared Finder
**  \date       2/27/2003 6:43:51 PM
**  \warning    none
*/
////////////////////////////////////////////////////////////////////////////////
void VertexBuffer::Activate(bool vertexArray, bool texCoordArray, bool clrArray,
                            int stride)
{
    // can we use NV extensions?
    if(m_usingNV)
    {
        glEnableClientState(GL_VERTEX_ARRAY_RANGE_WITHOUT_FLUSH_NV);

        float* pOrigin = (float*)m_pData;
        if(vertexArray)
        {
            glVertexPointer(3, GL_FLOAT, stride, pOrigin);
            pOrigin += 3;
        }
        if(texCoordArray)
        {
            glTexCoordPointer(2, GL_FLOAT, stride, pOrigin);
            pOrigin += 2;
        }
        if(clrArray)
        {
            glColorPointer(4, GL_FLOAT, stride, pOrigin);
            pOrigin += 4;
        }

        return;
    }
    // how about ATI?
    else if(m_usingATI)
    {
        assert(false);
    }
    // fall back to system memory
    else
    {
        float* pOrigin = (float*)m_pData;
        if(vertexArray)
        {
            glVertexPointer(3, GL_FLOAT, stride, pOrigin);
            pOrigin += 3;
        }
        if(texCoordArray)
        {
            glTexCoordPointer(2, GL_FLOAT, stride, pOrigin);
            pOrigin += 2;
        }
        if(clrArray)
        {
            glColorPointer(4, GL_FLOAT, stride, pOrigin);
            pOrigin += 4;
        }
    }
}


////////////////////////////////////////////////////////////////////////////////
//  FUNCTION:   
/*! \brief      Deactivates the current buffer
**  \param      
**  \return     none
**  \exception  none
**  \par        DESCRIPTION:
**                  
**  \author     M Jared Finder
**  \date       3/31/2003 11:18:00 AM
**  \warning    none
*/
////////////////////////////////////////////////////////////////////////////////
void VertexBuffer::Deactivate(void)
{
    // can we use NV extensions?
    if(m_usingNV)
    {
        glDisableClientState(GL_VERTEX_ARRAY_RANGE_WITHOUT_FLUSH_NV);
    }
    // how about ATI?
    else if(m_usingATI)
    {
        assert(false);
    }
    // fall back to system memory
    else
    {
    }
}


////////////////////////////////////////////////////////////////////////////////
//  FUNCTION:   
/*! \brief      Unlock the vertex buffer, allowing changes
**  \param      
**  \return     a pointer to writable memory
**  \exception  none
**  \par        DESCRIPTION:
**                  
**  \author     M Jared Finder
**  \date       3/18/2003 8:31:36 PM
**  \warning    none
*/
////////////////////////////////////////////////////////////////////////////////
void* VertexBuffer::Unlock(void)
{
    // can we use NV extensions?
    if(m_usingNV)
    {
        return m_pData;
    }
    // how about ATI?
    else if(m_usingATI)
    {
        assert(false);
        return NULL;
    }
    // fall back to system memory
    else
    {
        return m_pData;
    }
}


////////////////////////////////////////////////////////////////////////////////
//  FUNCTION:   
/*! \brief      Lock the vertex buffer, no longer allowing changes
**  \param      
**  \return     none
**  \exception  none
**  \par        DESCRIPTION:
**                  
**  \author     M Jared Finder
**  \date       3/18/2003 8:32:57 PM
**  \warning    none
*/
////////////////////////////////////////////////////////////////////////////////
void VertexBuffer::Lock(void)
{
    // can we use NV extensions?
    if(m_usingNV)
    {
        return;
    }
    // how about ATI?
    else if(m_usingATI)
    {
        assert(false);
    }
    // fall back to system memory
    else
    {
    }
}