////////////////////////////////////////////////////////////////////////////////
/*!	\file		GUI_Action_Subclasses.h
**	\brief		Subclasses of actions for actions that require GUI Primitives
**	\author		John Corpening
**	\date		1/18/2003 1:26:07 PM
**	\par		DESCRIPTION:
**				Subclasses include:
**				-Pushbutton actions
**				-checkbox actions
**				-slider actions
**				-editbox actions
**				-stat actions
**	\par		REMARKS:
*/
////////////////////////////////////////////////////////////////////////////////

#pragma once

namespace UserInput
{
////////////////////////////////////////////////////////////////////////////////
/*!	\class		UI_Action_Pushbutton
**	\brief		Action class for pushbuttons
**	\author		John Corpening
**	\date		1/18/2003 1:27:18 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
class UI_Action_Pushbutton : public UI_Action
{
protected:
	UI_Pushbutton* button;
public:
	UI_Action_Pushbutton(UI_Screen* Screen) : UI_Action(Screen), button(0) {}
	UI_Action_Pushbutton() : button(0) {}
	virtual void takeAction() = 0;

	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:	UI_Action_Pushbutton::checkMenuItem
	/*!	\brief		determines whether or not the button has been clicked
	**	\return		true if clicked, false otherwise
	**	\exception	none
	**	\author		John Corpening
	**	\date		1/18/2003 1:29:28 PM
	**	\warning	none
	*/
	////////////////////////////////////////////////////////////////////////////////
	bool checkMenuItem() throw();

	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:	UI_Action_Pushbutton::setMenuItem
	/*!	\brief		Sets the given menuitem as a pushbutton
	**	\param		item, a UI_MenuItem to set as a pushbutton
	**	\return		none
	**	\exception	"Item is not a pushbutton" Nexcpetion
	**	\author		John Corpening
	**	\date		1/18/2003 1:30:32 PM
	**	\warning	none
	*/
	////////////////////////////////////////////////////////////////////////////////
	void setMenuItem(UI_MenuItem *item);
};

////////////////////////////////////////////////////////////////////////////////
/*!	\class		UI_Action_CheckBox
**	\brief		Action Class for checkboxes
**	\author		John Corpening
**	\date		2/5/2003 12:16:55 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
class UI_Action_CheckBox : public UI_Action
{
protected:
	UI_CheckBox	*checkbox;
public:
	UI_Action_CheckBox(UI_Screen* Screen) : UI_Action(Screen), checkbox(0) {}
	UI_Action_CheckBox() : checkbox(0) {}
	virtual void takeAction() = 0;

	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:	UI_Action_CheckBox::checkMenuItem
	/*!	\brief		determines if the checkbox has been updated or not
	**	\return		true if updated, false otherwise
	**	\exception	none
	**	\author		John Corpening
	**	\date		2/22/2003 10:47:25 AM
	**	\warning	none
	*/
	////////////////////////////////////////////////////////////////////////////////
	bool checkMenuItem() throw();

	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:	UI_Action_CheckBox::setMenuItem
	/*!	\brief		casts the paramter to a checkbox and sets it internally
	**	\param		item, a pointer to a UI_MenuItem
	**	\return		none
	**	\exception	"Item is not a checkbox" Nexception
	**	\author		John Corpening
	**	\date		2/22/2003 10:47:29 AM
	**	\warning	none
	*/
	////////////////////////////////////////////////////////////////////////////////
	void setMenuItem(UI_MenuItem *item);
};

////////////////////////////////////////////////////////////////////////////////
/*!	\class		UI_Action_Slider
**	\brief		Action class for Sliders
**	\author		John Corpening
**	\date		2/22/2003 10:47:41 AM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
class UI_Action_Slider : public UI_Action
{
protected:
	UI_Slider *slider;
public:
	UI_Action_Slider(UI_Screen* Screen) : UI_Action(Screen), slider(0) {}
	UI_Action_Slider() : slider(0) {}
	virtual void takeAction() = 0;
	virtual void onEnter() = 0;

	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:	UI_Action_Slider::checkMenuItem
	/*!	\brief		determines whether or not the slider state has been updated
	**	\return		true if updated false otherwise
	**	\exception	none
	**	\author		John Corpening
	**	\date		2/22/2003 10:47:57 AM
	**	\warning	none
	*/
	////////////////////////////////////////////////////////////////////////////////
	bool checkMenuItem();

	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:	UI_Action_Slider::setMenuItem
	/*!	\brief		casts the parameter as a UI_Slider and sets it internally
	**	\param		item, pointer to the slider
	**	\return		none
	**	\exception	"Item is not a Slider" Nexception
	**	\author		John Corpening
	**	\date		2/22/2003 10:48:02 AM
	**	\warning	none
	*/
	////////////////////////////////////////////////////////////////////////////////
	void setMenuItem(UI_MenuItem *item);
	
};

////////////////////////////////////////////////////////////////////////////////
/*!	\class		UI_Action_Edit
**	\brief		Action class for edit boxes
**	\author		John Corpening
**	\date		2/22/2003 10:48:08 AM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
class UI_Action_Edit : public UI_Action
{
protected:
	UI_EditBox *edit;
public:
	UI_Action_Edit(UI_Screen* Screen) : UI_Action(Screen), edit(0){}
	UI_Action_Edit() : edit(0) {}
	virtual void takeAction() = 0;
	virtual void onEnter() = 0;

	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:	UI_Action_Edit::checkMenuItem
	/*!	\brief		determines whether or not the edit box was updated
	**	\return		true if updated, false otherwise
	**	\exception	none
	**	\author		John Corpening
	**	\date		2/22/2003 10:48:14 AM
	**	\warning	none
	*/
	////////////////////////////////////////////////////////////////////////////////
	bool checkMenuItem();

	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:	UI_Action_Edit::setMenuItem
	/*!	\brief		casts the parameter to an editbox and sets it internally
	**	\param		item, pointer to the item to cast as an edit box
	**	\return		none
	**	\exception	"Item is not an editbox" Nexception
	**	\author		John Corpening
	**	\date		2/22/2003 10:48:20 AM
	**	\warning	none
	*/
	////////////////////////////////////////////////////////////////////////////////
	void setMenuItem(UI_MenuItem *item);
};

////////////////////////////////////////////////////////////////////////////////
/*!	\class		UI_Action_Stat
**	\brief		class for fish statistics
**	\author		John Corpening
**	\date		3/10/2003 2:35:04 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
class UI_Action_Stat : public UI_Action
{
protected:
	UI_TextBox	*textbox;
public:
	UI_Action_Stat(UI_Screen* Screen) : UI_Action(Screen), textbox(0){}
	UI_Action_Stat() : textbox(0){}
	virtual void onEnter(){}
	void takeAction(){};
	virtual bool checkMenuItem() = 0;

	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:	UI_Action_Stat::setMenuItem
	/*!	\brief		casts the parameter to a textbox and sets it internally
	**	\param		item, pointer to the item to cast as a text box
	**	\return		none
	**	\exception	"Item is not a textbox" Nexception
	**	\author		John Corpening
	**	\date		3/10/2003 2:34:01 PM
	**	\warning	none
	*/
	////////////////////////////////////////////////////////////////////////////////
	void setMenuItem(UI_MenuItem *item);
};
}