////////////////////////////////////////////////////////////////////////////////
/*!	\file		GUI_Camera.cpp
**	\brief		Class to control the camera implementation file
**	\author		John Corpening
**	\date		1/18/2003 1:40:34 PM
**	\par		DESCRIPTION:
**	\par		REMARKS:
*/
////////////////////////////////////////////////////////////////////////////////

#include"GUI_Includes.h"
#include<math.h>

// memory leak detection
#ifdef _DEBUG
    #include"memleak.h"
    #define new DEBUG_NEW
    #undef THIS_FILE
    static char THIS_FILE[] = __FILE__;
#endif

//These are hard coded field of view values in Radians (60 degrees X and 45 degrees Y)
#define FOVX	1.0472f
#define FOVY	0.7854f

//the speed of camera dolly
const float cameraDollySpeed    = 40;
const float cameraPanSpeed      = 40;


using namespace UserInput;

//Initialize Singleton Instance
UI_Camera* UI_Camera::_instance = 0;

bool checkPositionBounds(math::linear_alg::Vector3<float>& pos)//, const math::linear_alg::Vector3<float>& view)
{
	const float distTollerance = 5;

	//Check vs. Heightmap
	float temp  = MsgHandler::MessageHandler::Instance()->GetProjectedHeightOnHeightMap(pos);
	if(pos[1] < temp + distTollerance)
	{
		pos[1] = temp + distTollerance;
	}
	//Check vs. World Bounds
	math::linear_alg::Vector3<float> min, max;
	MsgHandler::MessageHandler::Instance()->getBoundary(min, max);

	if (pos[0] < min[0] + distTollerance) pos[0] = min[0] + distTollerance;
	if (pos[1] < min[1] + distTollerance) pos[1] = min[1] + distTollerance;
	if (pos[2] < min[2] + distTollerance) pos[2] = min[2] + distTollerance;
	if (pos[0] > max[0] - distTollerance) pos[0] = max[0] - distTollerance;
	if (pos[1] > max[1] - distTollerance) pos[1] = max[1] - distTollerance;
	if (pos[2] > max[2] - distTollerance) pos[2] = max[2] - distTollerance;
/*
	if((pos[0] < min[0]) ||
	   (pos[1] < min[1]) ||
	   (pos[2] < min[2]) ||
	   (pos[0] > max[0]) ||
	   (pos[1] > max[1]) ||
	   (pos[2] > max[2]))
	{
		return false;
	}
*/

	return true;
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	UI_Camera::UI_Camera
/*!	\brief		Constructor initializes all member data to defaults
**	\return		none
**	\exception	none
**	\author		John Corpening
**	\date		1/18/2003 1:42:09 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
UI_Camera::UI_Camera() : view(0,0,1), pos(0,0,0), up(0,1,0), forward(DIK_W), back(DIK_S), left(DIK_A),
						 right(DIK_D), update(false), yAxisInvert(false), controlLock(false), orbitFlag(false)
{
	//Get a handle to the input module
	input = PlayerInput::instance();

	//Get a handle to the graphics module
	gfx::Renderer::Get().Set3DView(pos,view,up);
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	UI_Camera::getInstance
/*!	\brief		Singleton Instance access point
**	\return		pointer to the singleton instance
**	\exception	none
**	\author		John Corpening
**	\date		1/18/2003 1:43:25 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
UI_Camera* UI_Camera::getInstance()
{
	if(!_instance)
		_instance = new UI_Camera;
	return _instance;
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	UI_Camera::moveForward
/*!	\brief		moves the camera in the positive view vector direction
**	\return		none
**	\exception	none
**	\author		John Corpening
**	\date		1/18/2003 1:45:38 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
void UI_Camera::moveForward()
{
	math::linear_alg::Vector3<float> temp = pos;
    temp += view * float(TimeObject::Instance()->getRealTimeStep() * cameraDollySpeed);
	if(checkPositionBounds(temp))
	{
		pos = temp;
		update = true;
	}
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	UI_Camera::moveBackward
/*!	\brief		moves the camera in the negative view vector direction
**	\return		none
**	\exception	none
**	\author		John Corpening
**	\date		1/18/2003 1:47:27 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
void UI_Camera::moveBackward()
{
	math::linear_alg::Vector3<float> temp = pos;
	temp -= view * float(TimeObject::Instance()->getRealTimeStep() * cameraDollySpeed);
	if(checkPositionBounds(temp))
	{
		pos = temp;
		update = true;
	}
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	UI_Camera::strafeLeft
/*!	\brief		moves the camera in the up X view vector direction
**	\return		none
**	\exception	none
**	\author		John Corpening
**	\date		1/18/2003 1:47:55 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
void UI_Camera::strafeLeft()
{
	math::linear_alg::Vector3<float> temp = math::linear_alg::Cross(up, view);
	math::linear_alg::Vector3<float> temp2 = pos;
	temp2 += temp.Normalize() * float(TimeObject::Instance()->getRealTimeStep() * cameraPanSpeed);
	if(checkPositionBounds(temp2))
	{
		pos = temp2;
		update = true;
	}
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	UI_Camera::strafeRight
/*!	\brief		moves the camera in the view X up view vector direction
**	\return		none
**	\exception	none
**	\author		John Corpening
**	\date		1/18/2003 1:48:33 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
void UI_Camera::strafeRight()
{
	math::linear_alg::Vector3<float> temp2 = pos;
	math::linear_alg::Vector3<float> temp = math::linear_alg::Cross(view, up);
	temp2 += temp.Normalize() * float(TimeObject::Instance()->getRealTimeStep() * cameraPanSpeed);;
	if(checkPositionBounds(temp2))
	{
		pos = temp2;
		update = true;
	}
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	UI_Camera::pitch
/*!	\brief		rotates view vector about the x axis
**	\param		a, degrees in radians to pitch
**	\return		none
**	\exception	none
**	\author		John Corpening
**	\date		1/18/2003 1:49:10 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
void UI_Camera::pitch(float a)
{
	 math::linear_alg::Matrix3<float> m(	1,	0,					0,
											0,	(float)cos(a),		(float)-sin(a),
											0,	(float)sin(a),		(float)cos(a)
										);

	view = m*view;
	update = true;
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	UI_Camera::yaw
/*!	\brief		rotates the view vecotr about the y axis
**	\param		a, degrees in radians to yaw
**	\return		none
**	\exception	none
**	\author		John Corpening
**	\date		1/18/2003 1:49:59 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
void UI_Camera::yaw(float a)
{
	 math::linear_alg::Matrix3<float> m(	(float)cos(a),		0,		(float)sin(a),
											0,					1,		0,
											(float)-sin(a),		0,		(float)cos(a)
										);
	view = m*view;
	update = true;
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	UI_Camera::chase
/*!	\brief		Sets the camera 15 units behind and looking at a target point
**	\param		target, the point to chase
**	\return		none
**	\exception	none
**	\author		John Corpening
**	\date		1/18/2003 1:51:23 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
void UI_Camera::chase( math::linear_alg::Vector3<float>& target)
{
	pos = target +  math::linear_alg::Vector3<float>(0,0,-15);
	view = target - pos;
	view.Normalize();
	update = true;
}

void UI_Camera::orbit(math::linear_alg::Vector3<float>& target)
{
	orbitPos = target;
	orbitFlag = true;
	//pos = target - 15.0f*view;
	update = true;
}

//BEGIN Added by Nathan Frost <April 12, 2003>
void UI_Camera::chaseWithStaticViewVector(math::linear_alg::Vector3<float>& target)
{
	// CHANGED by Austin Spafford such that chase movement is smoothed out.
	//   NOTE: this current implementation of smoothing is frame-rate dependant.
	pos = 0.85f*pos + 0.15f*(target + -15.0f*view);
	update = true;
}
//END Added by Nathan Frost <April 12, 2003>

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	UI_Camera::logic
/*!	\brief		Processes input and updates the camera position and orientation
**	\return		none
**	\exception	none
**	\author		John Corpening
**	\date		1/18/2003 1:52:23 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
void UI_Camera::logic()
{
	if(!controlLock)
	{
		//Handle Rotations
		float dx = (float)input->mouseDeltaX();
		float dy = (float)input->mouseDeltaY();
		if(dy || dx)
		{
			//Convert view to local coordinates
			math::linear_alg::Vector3<float>oldView = view;

			math::linear_alg::Vector3<float>temp = math::linear_alg::Cross(up, view);
			temp.Normalize();
			math::linear_alg::Vector3<float>ty = math::linear_alg::Cross(view,temp);
			math::linear_alg::Matrix3<float> m( temp[0],	temp[1],	temp[2],
												ty[0],		ty[1],		ty[2],
												view[0],	view[1],	view[2]);
			
			math::linear_alg::Matrix3<float> m1(m);
			m1.Transpose();
			view = m*view;
			
			if(dx != 0)
			{
				yaw(-(dx / g_ScreenWidth)* FOVX);
			}
			if(dy != 0)
			{
				if(yAxisInvert)
				{
					pitch(-(dy / g_ScreenHeight)*FOVY);
				}
				else
				{
					pitch((dy / g_ScreenHeight)*FOVY);
				}
			}
			
			view = m1*view;
			view.Normalize();
			math::linear_alg::Vector3<float> test = math::linear_alg::Cross(up,view);
			if(test.Magnitude() < 0.1)
			{
				view = oldView;
			}
		}
		
		//Handle Translations
		if(input->isPressed(forward))
		{
			moveForward();
		}
		else if(input->isPressed(back))
		{
			moveBackward();
		}
		if(input->isPressed(left))
		{
			strafeLeft();
		}
		else if(input->isPressed(right))
		{
			strafeRight();
		}
	}
	//Orbit
	if(orbitFlag)
	{
		//Scroll the orbit distance
		float sign = (float)input->mouseDeltaWheel();
		if(sign)
		{
			sign = (sign < 0.0f ? 2.0f : -2.0f);
			orbitDist += sign;
			if(orbitDist < 4.0f) orbitDist = 4.0f;
			if(orbitDist > 100.0f) orbitDist = 100.0f;
		}
		pos = orbitPos - orbitDist*view;
		orbitFlag = false;
	}
	//Update
	if(update)
	{
//		Debugging::Debugger::Instance()->output(Debugging::GEN, Debugging::L4,"Pre:  pos[2]=%f", pos[2]);
//		checkPositionBounds(pos, view);
//		Debugging::Debugger::Instance()->output(Debugging::GEN, Debugging::L4,"post: pos[2]=%f", pos[2]);
		AcqariusSound::UIClient::Instance()->UpdateListener((float*)(pos.Data()), (float*)(up.Data()),(float*)(view.Data()));
		gfx::Renderer::Get().Set3DView(pos, pos + view, up);
		update = false;
	}
}
