////////////////////////////////////////////////////////////////////////////////
/*!	\file		GUI_Menu.cpp
**	\brief		This is the implementation file for GUI_Menu.h
**	\author		John Corpening
**	\date		10/29/2002 6:44:27 PM
*/
////////////////////////////////////////////////////////////////////////////////

#include "GUI_Includes.h"
#include <set>

// memory leak detection
#ifdef _DEBUG
    #include"memleak.h"
    #define new DEBUG_NEW
    #undef THIS_FILE
    static char THIS_FILE[] = __FILE__;
#endif
using namespace UserInput;
////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	UI_Menu::menuLogic
/*!	\brief		logic for processing player input
**	\return		true for input processed
**	\exception	none
**	\author		John Corpening
**	\date		10/29/2002 6:33:50 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
bool UI_Menu::menuLogic()
{
	if(!active)
		return false;
	//If the menu can process the message as a menu item then return true
	if(logic())
	{
		return true;
	}

	//Otherwise pass it along to the menu items
	
	int size = menuItems.size();
	for(int i = 0; i < size; i++)
	{
		//If a menuItem is found that processes this message then return
		if(menuItems[i]->logic())
		{
			return true;
		}
	}
	return false;
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	UI_Menu::draw
/*!	\brief		draws the menu and then all the menuItems
**	\return		none
**	\exception	none
**	\author		John Corpening
**	\date		10/29/2002 6:42:37 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
void UI_Menu::draw()
{
	if(!active)
		return;
	//Draw yourself here
	drawMenu();

	//Draw everything else here
	int size = menuItems.size();
	for(int i = 0; i < size; i++)
	{
		menuItems[i]->draw();
	}
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	UI_Menu::setPosition
/*!	\brief		sets the menu position and updates all menuItem positions
**	\param		X the new X coordinate
**	\param		Y the new Y coordinate
**	\return		none
**	\exception	none
**	\author		John Corpening
**	\date		10/29/2002 6:41:16 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
void UI_Menu::setPosition(float X, float Y)
{
	x = X;
	y = Y;
	int size = menuItems.size();
	for(int i = 0; i < size; i++)
	{
		menuItems[i]->setPosition(menuItems[i]->relX() + x, menuItems[i]->relY() + y);
	}
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	UI_Menu::addMenuItem
/*!	\brief		adds a menu item to the menu
**	\param		item the item to add
**	\param		X the x coordinate relative to the menu
**	\param		Y the y coordinate relative to the menu
**	\return		none
**	\exception	none
**	\author		John Corpening
**	\date		10/29/2002 6:37:46 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
void UI_Menu::addMenuItem(UI_MenuItem *item, float rX, float rY)
{
	if(item->getWidth() + rX > width - 0.05f)
	{
		width = item->getWidth() + 0.05f + rX;
	}
	if(item->getHeight() + rY > height - 0.05f)
	{
		height = item->getHeight() + 0.05f + rY;
	}
	item->setPosition(x + rX, y + rY);
	item->setRelativePosition(rX, rY);
	menuItems.push_back(item);
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	UI_Menu::setScreen
/*!	\brief		sets the screen that owns the menu for the menu and all children
**	\param		Screen, the pointer to the UI_Screen that owns this menu
**	\return		none
**	\exception	none
**	\author		John Corpening
**	\date		11/25/2002 5:40:47 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
void UI_Menu::setScreen(UI_Screen *Screen)
{
	screen = Screen;
	int size = menuItems.size();
	for(int i = 0; i < size; i++)
	{
		menuItems[i]->setScreen(Screen);
	}
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	UI_Menu::setCursor
/*!	\brief		sets the cursor for this menu and all of its children
**	\param		Cursor, pointer to UI_Cursor
**	\return		none
**	\exception	none
**	\author		John Corpening
**	\date		11/25/2002 5:41:27 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
void UI_Menu::setCursor(UI_Cursor *Cursor)
{
	cursor = Cursor;
	int size = menuItems.size();
	for(int i = 0; i < size; i++)
	{
		menuItems[i]->setCursor(Cursor);
	}
}

UI_Menu::~UI_Menu()
{
	for(unsigned int i = 0; i < menuItems.size(); i++)
	{
		delete menuItems[i];
	}
}