////////////////////////////////////////////////////////////////////////////////
/*!	\file		GUI_Screen.cpp
**	\brief		Implementation of GUI_Screen.h
**	\author		John Corpening
**	\date		10/30/2002 8:15:52 PM
*/
////////////////////////////////////////////////////////////////////////////////
#pragma warning(disable:4786)

#include "GUI_Includes.h"

// memory leak detection
#ifdef _DEBUG
    #include"memleak.h"
    #define new DEBUG_NEW
    #undef THIS_FILE
    static char THIS_FILE[] = __FILE__;
#endif
using namespace UserInput;

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	UI_Screen::UI_Screen
/*!	\brief		Constructor for the screen
**	\return		none
**	\exception	none
**	\author		John Corpening
**	\date		11/25/2002 6:09:28 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
UI_Screen::UI_Screen() : toolTip(0), pullDown(0), focus(0), keyMap(0), camera(0), m_drawCursor(true)
{
	cursor = new UI_Cursor;
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	UI_Screen::draw
/*!	\brief		draws the menus, then the pulldown menu then the toolTip
**	\return		none
**	\exception	none
**	\author		John Corpening
**	\date		10/29/2002 6:56:41 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
void UI_Screen::draw()
{
	//First draw all the menus
	int size = menus.size();
	for(int i = 0; i < size; i++)
	{
		menus[i]->draw();
	}

	//Then draw the pullDown menu if set
	if(pullDown)
	{
		pullDown->draw();
	}

	//Then draw the toolTip if set
	if(toolTip)
	{
		toolTip->draw();
	}

	//Then draw the mouse cursor
	if(cursor && m_drawCursor)
	{
		cursor->draw();
	}
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	UI_Screen::logic
/*!	\brief		regulates processing of input for menus
**	\return		none
**	\exception	none
**	\author		John Corpening
**	\date		10/29/2002 6:56:28 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
bool UI_Screen::logic() throw()
{
	//Clear toolTips and pulldown Menus
	toolTip = 0;

	//Update the cursor
	if(cursor)
	{
		cursor->logic();
	}

	//First see if the focus can handle the message
	if(focus)
	{
		if(focus->logic())
		{
			return true;
		}
	}
	//If there is no focus the pass it off to the keymap
	//The screen should have a keymap this is just for sanity
	if(keyMap)
	{
		keyMap->menuLogic();
	}

	//Update the position of the toolTip
	if(toolTip)
	{
		toolTip->setPosition(cursor->X(), cursor->Y());
	}

	//Camera movement
	if(camera)
	{
		camera->logic();
	}

	//Then check the menus
	int size = menus.size();
	for(int i = 0; i < size; i++)
	{
		//If a menu is able to process the input then quit
		if(menus[i]->menuLogic())
		{
			return true;
		}
	}
	return false;
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	UI_Screen::addMenu
/*!	\brief		adds a menu to the screen
**	\param		menu the menu to add
**	\return		none
**	\exception	none
**	\author		John Corpening
**	\date		10/29/2002 6:58:06 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
void UI_Screen::addMenu(UI_Menu *menu, float x, float y) throw()
{
	if(!menu)
		return;
	menu->setPosition(x, y);
	menus.push_back(menu);
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	UI_Screen::onEnter
/*!	\brief		code that is executed whenever the screen is entered
**	\return		none
**	\exception	none
**	\author		John Corpening
**	\date		2/22/2003 11:43:10 AM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
void UI_Screen::onEnter()
{
	if(keyMap)
	{
		keyMap->onEnter();
	}
}

void UI_Screen::onDeselect()
{
	if(keyMap)
	{
		keyMap->onDeselect();
	}
}

bool UI_Screen::checkCursorVsMenus()
{
	if(!cursor)
		return false;

	for(unsigned int i = 0; i < menus.size(); i++)
	{
		if (menus[i]->getActiveState())
		{
			float x = menus[i]->X();
			float y = menus[i]->Y();
			float height = menus[i]->getHeight();
			float width = menus[i]->getWidth();

			if((cursor->X() >= x) 
			&& (cursor->X() < (x + width))
			&& (cursor->Y() >= y)
			&& (cursor->Y() < (y + height)))
			{
				return true;
			}
		}
	}
	return false;
}

UI_Screen::~UI_Screen()
{
	if(keyMap)
		delete keyMap;
	for(unsigned int i = 0; i < menus.size(); i++)
	{
		//HACK: Don't delete the Console
		UI_Menu* temp = menus[i];
		if(!dynamic_cast<UI_Console*>(temp))
		{
			delete menus[i];
		}
	}
	if(cursor)
		delete cursor;
}