////////////////////////////////////////////////////////////////////////////////
/*!	\file		UI_Handler.cpp
**	\brief		Implementation file for UI_Handler.h
**	\author		John Corpening
**	\date		10/30/2002 7:00:17 PM
**	\par		DESCRIPTION:
**	\par		REMARKS:
*/
////////////////////////////////////////////////////////////////////////////////

#include "GUI_Includes.h"
#include "../gfx/opengl.h"


// memory leak detection
#ifdef _DEBUG
    #include"memleak.h"
    #define new DEBUG_NEW
    #undef THIS_FILE
    static char THIS_FILE[] = __FILE__;
#endif
using namespace UserInput;

//Static member initialization
UI_Handler *UI_Handler::_instance = 0;

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	UI_Handler:: instance
/*!	\brief		access point for the UI_Handler
**	\return		none
**	\exception	none
**	\author		John Corpening
**	\date		10/30/2002 6:55:28 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
UI_Handler* UI_Handler::instance()
{
	if(!_instance)
		_instance = new UI_Handler;
	return _instance;
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	UI_Handler::UI_Handler
/*!	\brief		Constructor creates all screens for the game
**	\return		none
**	\exception	none
**	\author		John Corpening
**	\date		10/30/2002 6:54:22 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
UI_Handler::UI_Handler()
{
	//Initialize the PlayerInput module
	input = PlayerInput::initialize(g_hwnd);
	MH = MsgHandler::MessageHandler::Instance();

	myID = MH->GetMyIP();

	//Initialize the UI_Style
	UI_Style::staticMemberInit();

	//Create the screens
	screens.insert(std::make_pair(GUI_MAINMENU, createMainMenu()));
	screens.insert(std::make_pair(GUI_EDITMODE, createEditModeScreen()));
	screens.insert(std::make_pair(GUI_OBSERVERMODE, createObserverScreen()));
	screens.insert(std::make_pair(GUI_TESTMODE, createTestScreen()));
	screens.insert(std::make_pair(GUI_OFFLINEMODE, createOfflineModeScreen()));
	screens.insert(std::make_pair(GUI_GAMEOPTIONS, createGameOptionsScreen()));
	screens.insert(std::make_pair(GUI_SYSTEMOPTIONS, createSystemOptionsScreen()));
	screens.insert(std::make_pair(GUI_MULTIPLAYER, createMultiplayerScreen()));
	screens.insert(std::make_pair(GUI_CLIENTPERMISSIONS, createClientPermissionsScreen()));
	screens.insert(std::make_pair(GUI_SWIMMODE, createSwimModeScreen()));
	screens.insert(std::make_pair(GUI_SAVESCREEN, createSaveScreen()));
	screens.insert(std::make_pair(GUI_LOADSCREEN, createLoadScreen()));
	screens.insert(std::make_pair(GUI_TITLESCREEN, createTitleScreen()));
	screens.insert(std::make_pair(GUI_CREDITS, createCreditsScreen()));

	//Set the default screen
	std::map<GUI_ScreenID, UI_Screen*>::iterator pos = screens.find(GUI_TITLESCREEN);
	if(pos != screens.end())
	{
		currentScreen = pos->second;
	}
	else
	{
		currentScreen = 0;
	}
	fish = 0;
	prevScreenID = curScreenID = GUI_TITLESCREEN;
	style = new UI_Style;
	songTS = 0;
	songIG = 0;
	songBE = 0;
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	UI_Handler::logic
/*!	\brief		allows the current screen to process input
**	\return		none
**	\exception	none
**	\author		John Corpening
**	\date		10/30/2002 6:56:04 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
void UI_Handler::logic()
{
/*
	// RAY CASTING DEBUG
	for (unsigned int x=0; x<20; x++)
	{
		for (unsigned int y=0; y<20; y++)
		{
			math::linear_alg::Matrix4<float> proj;
			math::linear_alg::Matrix4<float> view;
			math::linear_alg::Vector4<float> rayOrigin(
													x/20.0f, 
													y/20.0f, 
													0, 
													1);
			gfx::Renderer::Get().GetTransform3D(proj,view);
			proj *= view;
			proj.Invert();
			rayOrigin = proj * rayOrigin;
			rayOrigin /= rayOrigin[3];
			MsgHandler::Key dummyKey;

			math::linear_alg::Vector3<float> click3(rayOrigin[0],rayOrigin[1],rayOrigin[2]);
			math::linear_alg::Vector3<float> pos = UI_Camera::getInstance()->getPosition();
			bool rayHit = MsgHandler::MessageHandler::Instance()->TraceRayAgainstAllFish(pos,click3-pos, dummyKey);

			//bool rayHit = MsgHandler::MessageHandler::Instance()->TraceRayAgainstAllFish(math::linear_alg::Vector3<float>(rayOrigin[0], rayOrigin[1], rayOrigin[2]),UI_Camera::getInstance()->getViewDir(), dummyKey);

			if (false != rayHit)
			{
				style->drawText(x/20.0f, y/20.0f, "X");
				style->draw();
			}
		}
	}
*/

	input->getUserInput();

	//Make sure the fish still exists. If it does not then deselect it
	if(fish && !MH->doesFishExist(fish))
	{
		if(curScreenID == GUI_SWIMMODE)
		{
			switchScreens(GUI_EDITMODE);
		}
		MH->DeselectRequest(fish);
	}

	static MsgHandler::Key tempFish = 0;

	//Don't be fooled, the above declaration is static
	if(tempFish)
	{
		MH->HighLightFish(tempFish, false);
		tempFish = 0;
	}

	bool val = currentScreen->logic();

	if(!val && !currentScreen->checkCursorVsMenus() && (curScreenID == GUI_EDITMODE))// && input->isTriggered(DIK_LBUTTON))
	{
		//Nothing was able to process the mouse click so see if the stray click hit a fish
		math::linear_alg::Matrix4<float> proj;
		math::linear_alg::Matrix4<float> view;
		math::linear_alg::Vector4<float> click(
												currentScreen->getCursor()->X(), 
												currentScreen->getCursor()->Y(), 
												0, 
												1);
		gfx::Renderer::Get().GetTransform3D(proj,view);
		proj *= view;
		proj.Invert();
		click = proj * click;
		click /= click[3];
		math::linear_alg::Vector3<float> click3(click[0],click[1],click[2]);
		math::linear_alg::Vector3<float> pos = UI_Camera::getInstance()->getPosition();
		MH->TraceRayAgainstAllFish(pos,click3-pos, tempFish);
		
		bool selectFlag = false;

		if(tempFish && (tempFish != fish) && !isSelected(tempFish))
		{
			MH->HighLightFish(tempFish, true, 0, 1, 0);
		}
		else
		{
			if (isSelected(tempFish))
			{
				tempFish = 0;
				selectFlag = true;
			}
		}
		//If the user clicked then try to select it
		if(input->isTriggered(DIK_LBUTTON))
		{			
			if(tempFish)
			{
				MH->DeselectRequest(fish);
				MH->SelectRequest(tempFish);				
			}
			else if(fish && !selectFlag)
			{
				MH->DeselectRequest(fish);
				//fish = 0;
			}

		}		
	}



	if(fish)
	{
		math::linear_alg::Vector3<float> lookat, pos, dir;
		if(MH->GetDirection(fish, dir))
		{
			if(MH->GetPosition(fish, lookat))
			{
				switch(curScreenID)
				{
				case GUI_EDITMODE:
					//BEGIN Changed by Nathan Frost <April 12, 2003>
					if(prevScreenID != GUI_EDITMODE && prevScreenID != GUI_SWIMMODE)
					{
//						UI_Camera::getInstance()->chase(lookat);
					}
					UI_Camera::getInstance()->chaseWithStaticViewVector(lookat);
					//END Changed by Nathan Frost <April 12, 2003>
					break;
				case GUI_SWIMMODE:
					UI_Camera::getInstance()->orbit(lookat);
					break;
				}
			}
		}
	}

/*
	// FISH SELECTION PROTECTION DEBUG
	MsgHandler::Key startingFishKey = MH->GetNextFish(0);
	if (0 != startingFishKey) // check to see if there are any fish
	{
		MsgHandler::Key iteratingFishKey = startingFishKey;
		do
		{
			iteratingFishKey = MH->GetNextFish(iteratingFishKey);
			MH->HighLightFish(iteratingFishKey, false);			
		} while (iteratingFishKey != startingFishKey);

		iteratingFishKey = startingFishKey;
		do
		{
			iteratingFishKey = MH->GetNextFish(iteratingFishKey);
			if (isSelected(iteratingFishKey))
			{
				MH->HighLightFish(iteratingFishKey, true, 1, 0, 1);
			}			
		} while (iteratingFishKey != startingFishKey);
	}
*/
	
    draw();
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	UI_Hanlder::draw
/*!	\brief		draws the current screen
**	\return		none
**	\exception	none
**	\author		John Corpening
**	\date		11/25/2002 6:16:12 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
void UI_Handler::draw()
{
		//Special Drawing
	switch(curScreenID)
	{
	case GUI_TITLESCREEN:
		style->drawTitle(0.0f, 0.0f);
		style->draw();
		break;
	case GUI_CREDITS:
		style->drawCredits();
		style->draw();
		break;
	};

	currentScreen->draw();
    
    //no more FPS counter!
/*
	char buf[128];
	char outBuf[128] = "FPS: ";
	strcat(outBuf, gcvt(TimeObject::Instance()->getAverageFramerate(), 4, buf));
	style->drawText(0.8f, 0.5f, outBuf);
	style->draw();
*/
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	UI_Handler::switchScreens
/*!	\brief		switches to the given screen if it exists
**	\param		screenID, the ID of the screen to switch to
**	\return		none
**	\exception	none
**	\author		John Corpening
**	\date		1/18/2003 1:56:22 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
void UI_Handler::switchScreens(GUI_ScreenID screenID)
{
	std::map<GUI_ScreenID, UI_Screen*>::iterator pos = screens.find(screenID);
	if(pos != screens.end())
	{
		currentScreen = pos->second;
		prevScreenID = curScreenID;
		curScreenID = screenID;
		pos->second->onEnter();
		UI_Camera::getInstance()->unlockControls();
		
		switch(prevScreenID)
		{
		case GUI_OFFLINEMODE:
			MH->offLineMode(false);
			break;
		}
		
		switch(curScreenID)
		{
		case GUI_TITLESCREEN:
			AcqariusSound::UIClient::Instance()->SetState(AcqariusSound::UISTATE_TITLE_SCREEN);
			break;
		case GUI_MAINMENU:
			if(prevScreenID == GUI_TITLESCREEN)
			{
				AcqariusSound::UIClient::Instance()->SetState(AcqariusSound::UISTATE_IN_GAME);
			}
			break;
		case GUI_OBSERVERMODE:
			//Clear out fish target in Observermode
			MH->DeselectRequest(fish);// = 0;
			break;
		case GUI_TESTMODE:
			UI_Camera::getInstance()->lockControls();
			break;
		case GUI_EDITMODE:
			UI_Camera::getInstance()->lockControls();
			break;
		case GUI_SWIMMODE:
			UI_Camera::getInstance()->setOrbitDist(15.0f);
			break;
		case GUI_OFFLINEMODE:
			MH->offLineMode(true);
			break;
		}
	}
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	UI_Handler::previousScreen
/*!	\brief		switches to the previousScreen
**	\return		none
**	\exception	none
**	\author		John Corpening
**	\date		1/18/2003 1:57:01 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
void UI_Handler::previousScreen()
{
	switchScreens(prevScreenID);
}


////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	UI_Handler::confirmSelect
/*!	\brief		Confirms the selection of a fish
**	\param		select, the key of the fish to select
**	\return		none
**	\exception	none
**	\author		John Corpening
**	\date		3/15/2003 12:51:17 PM
**	\warning	This function should only be called by the Server MessageHandler
*/
////////////////////////////////////////////////////////////////////////////////
bool UI_Handler::confirmSelect(const MsgHandler::Key& select, PlayerID id)
{
	//the key is invalid...
	if(!select || !id)
		return false;//...no good

	//Check to make sure that the player is not already on the list
	std::map<MsgHandler::Key, PlayerID>::iterator pos;
	for(pos = possessed.begin(); pos != possessed.end(); pos++)
	{
		if(pos->second == id && pos->first != select)
		{
			//The fish is trying to possess a fish when it already has one selected
			//This should not happen but guard against it
			possessed.erase(pos);
			break;
		}
	}
	//Check whether or not the fish is currently in the selected fish list
	 pos = possessed.find(select);

	//If the fish is already on the selected fish list...
	if(pos != possessed.end())
	{
		//...but we're trying to select a second time...
		if(pos->second == id)
			return true;//...then the Fish is already highlighted, so remind the caller of this
		else
			return false;//...don't try to select other people's Fish
	}

	//If it wasn't then add it to the set of fish possessed and return true
	possessed.insert(std::make_pair(select, id));

	//The selected fish needs to be highlighted
	if(myID != id)
		MH->HighLightFish(select, true, 1, 0, 0);

	return true;
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	UI_Handler::deselect
/*!	\brief		deselects fish from the set of selected fish
**	\param		select, the key of the fish to deselect
**	\return		true if the fish was selected, false otherwise
**	\exception	none
**	\author		John Corpening
**	\date		3/15/2003 1:02:34 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
bool UI_Handler::confirmDeselect(const MsgHandler::Key& select, PlayerID id)
{
	//Check whether or not the fish is currently selected
	std::map<PlayerID, MsgHandler::Key>::iterator pos = possessed.find(select);

	//If it wasn't then return false
	if(!select || (pos == possessed.end()))
	{
		return false;
	}

	//If it was then remove it from the list, unhiglight it and return true
	possessed.erase(select);

	//If it is surpressed then unsuppress it
	if(AI_System::AI_BrainManager::getInstance()->isSuppressed(select))
	{
		MH->UnsuppressRequest(select);
	}
	
	if(myID != id)
		MH->HighLightFish(select, false);

	return true;
}

void UI_Handler::GenerateCompleteNetMessage(UDPNet::NetMessage &ServerUpdate)
{
	std::map<PlayerID, MsgHandler::Key>::iterator pos;
	char size = possessed.size();
	//Write out the number of fish in the set
	ServerUpdate.WriteNBytes(1, &size);

	for(pos = possessed.begin(); pos != possessed.end(); pos++)
	{
		ServerUpdate.WriteNBytes(sizeof(MsgHandler::Key), &pos->first);
		ServerUpdate.WriteNBytes(sizeof(PlayerID), &pos->second);
	}
}

void UI_Handler::LoadNetMessage(UDPNet::NetMessage &ServerUpdate)
{
	char size;
	MsgHandler::Key tempKey;
	PlayerID	tempID;

	//Unhighlight fish in the set
	for(std::map<PlayerID,MsgHandler::Key>::iterator pos = possessed.begin(); pos != possessed.end(); pos++)
	{
		MH->HighLightFish(pos->first, false);
	}
	//Clear out the set
	possessed.clear();
	fish = 0;

	//if there's Fish highlighted
	ServerUpdate.ReadNBytes(1, &size);
	if(!size)
	{
		setFishTarget(0);
		return;
	}

	//Populate the map with the elements from the server
	for(int i = 0; i < size; ++i)
	{
		ServerUpdate.ReadNBytes(sizeof(MsgHandler::Key), &tempKey);
		ServerUpdate.ReadNBytes(sizeof(PlayerID), &tempID);
		
		Debugging::Debugger::Instance()->output(Debugging::NET, Debugging::L4, 
												"Key selected: %i", tempKey);

		possessed.insert(std::make_pair(tempKey, tempID));
		if(tempID == myID)
		{
			fish = tempKey;
			//setFishTarget(tempKey);
		}
		else
		{
			MH->HighLightFish(tempKey, true, 1, 0, 0);
		}
	}

	//Hack
	if(!fish)
	{
		setFishTarget(0);
	}
}

void UI_Handler::setTitleScreenSong(const char *song)
{
	if(songTS)
	{
		delete[] songTS;
	}
	songTS = new char[strlen(song) + 1];
	strcpy(songTS, song);
}

void UI_Handler::setBrainEditorSong(const char* song)
{
	if(songBE)
	{
		delete[] songBE;
	}
	songBE = new char[strlen(song)+1];
	strcpy(songBE, song);
}

void UI_Handler::setInGameSong(const char* song)
{
	if(songIG)
	{
		delete[] songIG;
	}
	songIG = new char[strlen(song)+1];
	strcpy(songIG, song);
}

MsgHandler::Key UI_Handler::getFishFromIP(PlayerID id)
{
	std::map<MsgHandler::Key, PlayerID>::iterator pos;
	for(pos = possessed.begin(); pos != possessed.end(); pos++)
	{
		if(pos->second == id)
		{
			return pos->first;
		}
	}
	return 0;
}

void UI_Handler::deselectAllButThisIP(PlayerID id)
{
	std::map<MsgHandler::Key, PlayerID>::iterator pos;
	MsgHandler::Key temp;
	for(pos = possessed.begin(); pos != possessed.end(); pos++)
	{
		if(pos->second == id)
		{
			temp = pos->first;
			break;
		}
	}
	possessed.clear();
	possessed.insert(std::make_pair(temp, id));
}

bool UI_Handler::isSelected(MsgHandler::Key key)
{
	if(possessed.find(key) == possessed.end())
		return false;
	return true;
}

//there must be at least one fish in the simulation
MsgHandler::Key UI_Handler::FindFirstDeselectedFish(MsgHandler::Key FirstFish)
{
	MsgHandler::Key curFish = FirstFish;
	if(!isSelected(curFish))
		return curFish;
	else
	{
		do
		{	
			curFish = MH->GetNextFish(curFish);
			if(!isSelected(curFish))
				return curFish;
		}while(curFish != FirstFish);
		return 0;//there's no fish available to select
	}
}

void UI_Handler::getNextFish()
{
	//if there are no fish in the simulation then take no action
	MsgHandler::Key temp = MH->GetNextFish(fish);
	if(!temp)
		return;

	MsgHandler::Key availableFish = FindFirstDeselectedFish(temp);
	//if we actually found a available fish...
	if(availableFish)
	{
		//and if we had a fish previously selected...
		if(fish != 0)
		{
			//and if the fish we're about to select is not the fish we have currently selected...
			if(fish != availableFish)
			{
				//then deselect the previously selected fish and select the available fish
				MH->DeselectRequest(fish);
				MH->SelectRequest(availableFish);
			}
		}
		else//we had no fish selected previously...
			MH->SelectRequest(availableFish);//...then select the availableFish
	}
	
	return;//mission accomplished
}

void UI_Handler::setFishTarget(const MsgHandler::Key &Fish)
{
	fish = Fish;
	Debugging::Debugger::Instance()->output(Debugging::GEN, Debugging::L4, "UI Fish Target: %d", fish);
	if(!fish)
	{
		currentScreen->onDeselect();
	}
}

UI_Handler::~UI_Handler()
{
	if(style)
		delete style;

	std::map<GUI_ScreenID, UI_Screen*>::iterator pos;
	for(pos = screens.begin(); pos != screens.end(); pos++)
	{
		GUI_ScreenID temp = pos->first;
		delete pos->second;
	}

	delete UI_Camera::getInstance();
	delete input;
	//	delete UI_Console::getInstance();
	UI_Console::Destroy();
}