////////////////////////////////////////////////////////////////////////////////
/*!	\file		CommandID.hpp
**	\brief		Stats the data type and values of Command IDs.
**	\author		Nate Cleveland
**	\date		10/18/2002 11:33:25 AM
**	\par		DESCRIPTION:
**				Read each command specification for documentation.	
**	\par		REMARKS:
**					
*/
////////////////////////////////////////////////////////////////////////////////
#pragma once

namespace InterpretationModule
{
	//! Typedef for the varible type used for CommandIDs.
	typedef unsigned int CommandID_type;
	
	////////////////////////////////////////////////////////////////////////////////
	/*!	
	**	\brief	Follows an object. Will maintain a reasonable distance.
	**	\par	Reasonable distance is generally NOT close enough to attack/eat an object.
	*/
	////////////////////////////////////////////////////////////////////////////////
	const CommandID_type AQUARIUS_COMMAND_FOLLOW			= 30290;
	

	////////////////////////////////////////////////////////////////////////////////
	/*!	
	**	\brief	Tells the fish body to swim where the other fish is swimming to.
	**	\par	Simmilar to flocking only far simpler and less robust.
	*/
	////////////////////////////////////////////////////////////////////////////////
	const CommandID_type AQUARIUS_COMMAND_MOVE_WITH			= 8888;
	
	
	////////////////////////////////////////////////////////////////////////////////
	/*!	
	**	\brief	Moves body to object, point or voxel.
	**	\par	Gets as close as possible to the target
	*/
	////////////////////////////////////////////////////////////////////////////////
	const CommandID_type AQUARIUS_COMMAND_GOTO				= 187240; 
	
	////////////////////////////////////////////////////////////////////////////////
	/*!	
	**	\brief	Moves away from the indicated object, point, or voxel
	**	\par	Moves away from the indicated object, point, or voxel. Would be the
			 same as giving a Goto command with a negative wieght. Conversly,
			 a negative wieght given to a flee command is the same as a goto
			 command.
	*/
	////////////////////////////////////////////////////////////////////////////////
	const CommandID_type AQUARIUS_COMMAND_FLEE				= 26671;
	
	////////////////////////////////////////////////////////////////////////////////
	/*!	
	**	\brief	Attacks the indicated object.
	**	\par	Will move towards the target and attack it.
	*/
	////////////////////////////////////////////////////////////////////////////////
	const CommandID_type AQUARIUS_COMMAND_ATTACK			= 2872;
	
	
	////////////////////////////////////////////////////////////////////////////////
	/*!	
	**	\brief	Tries to eat the incated object.
	**	\par	Moves toward the indicated object and tries to eat it.
	*/
	////////////////////////////////////////////////////////////////////////////////
	const CommandID_type AQUARIUS_COMMAND_EAT				= 3471;
	
	
	////////////////////////////////////////////////////////////////////////////////
	/*!	
	**	\brief	Attacks the indicated object.
	**	\par	Will move towards the target and attack it.
	*/
	////////////////////////////////////////////////////////////////////////////////
	const CommandID_type AQUARIUS_COMMAND_MOVE_FORWARD		= 16486;
	
	
	////////////////////////////////////////////////////////////////////////////////
	/*!	
	**	\brief	Forces all other fin related commands.
	**	\par	This command reserves the right to poison all the other commands.
	*/
	////////////////////////////////////////////////////////////////////////////////
	const CommandID_type AQUARIUS_COMMAND_STOP_FIN			= 7854;
	
	
	////////////////////////////////////////////////////////////////////////////////
	/*!	
	**	\brief	Forces all other mouth commands to stop
	**	\par	This command reserves the right to poison all the other commands.
	*/
	////////////////////////////////////////////////////////////////////////////////
	const CommandID_type AQUARIUS_COMMAND_STOP_MOUTH		= 14803;
	
	////////////////////////////////////////////////////////////////////////////////
	/*!	
	**	\brief	Tells the body to consider only positive markers.
	**	\par	Weights for this command should be positive. Negative markers will
	**			 corrupt the intension of this command. 
	*/
	////////////////////////////////////////////////////////////////////////////////
	const CommandID_type AQUARIUS_COMMAND_EVALUATE_POSITIVE	= 1965;
	
	////////////////////////////////////////////////////////////////////////////////
	/*!	
	**	\brief	Tells the body to consider only negative markers.
	**	\par	Weights for this command should be positive. Negative markers will
	**			 corrupt the intension of this command. 
	*/
	////////////////////////////////////////////////////////////////////////////////
	const CommandID_type AQUARIUS_COMMAND_EVALUATE_NEGATIVE	= 905;
	
	////////////////////////////////////////////////////////////////////////////////
	/*!	
	**	\brief	Tells the body to consider all prevously set markers.
	**	\par	Weights for this command should be positive. Negative markers will
	**			 corrupt the intension of this command. Is logically equivelent to
	**			 using both Evaluate Positive and Negative at the same time.
	*/
	////////////////////////////////////////////////////////////////////////////////
	const CommandID_type AQUARIUS_COMMAND_EVALUATE_MARKERS	= 965;
	
	
	////////////////////////////////////////////////////////////////////////////////
	/*!	
	**	\brief	Tells the body to drop a marker at this point.
	**	\par	Markers may be positive or negative depending on the value of the weight
			 given in the CommandPacket.
	*/
	////////////////////////////////////////////////////////////////////////////////
	const CommandID_type AQUARIUS_COMMAND_DROP_MARKER		= 14466;		
	
	
	////////////////////////////////////////////////////////////////////////////////
	/*!	
	**	\brief	Tells the body to swim to a semi-random location.
	**	\par	
	*/
	////////////////////////////////////////////////////////////////////////////////
	const CommandID_type AQUARIUS_COMMAND_ROAM				= 666;	
	
	
	////////////////////////////////////////////////////////////////////////////////
	/*!	
	**	\brief	Orders a fish to lift an object.
	**	\par	At this time this command is not supported.
	*/
	////////////////////////////////////////////////////////////////////////////////
	const CommandID_type AQUARIUS_COMMAND_PICK_UP			= 10083;
	
	
	////////////////////////////////////////////////////////////////////////////////
	/*!	
	**	\brief	Orders a fish to drop an object.
	**	\par	At this time this command is not supported.
	*/
	////////////////////////////////////////////////////////////////////////////////
	const CommandID_type AQUARIUS_COMMAND_DROP_OBJECT		= 1358;
}//end namespace InterpretationModule.


