////////////////////////////////////////////////////////////////////////////////
/*!	\file		action.hpp
**	\brief		Declares all the actions that a fishBody can preform.
**	\author		Nate Cleveland
**	\date		10/22/2002 2:03:28 PM
**	\par		DESCRIPTION: Currently the following actions are supported
**					- Roaming
**					- Goto
**					- Flee
**					- Follow
**					- Move Forward
**					- Attack
**					- Eat
**					- Stop Fin
**					- Stop Mouth
**					- Drop Markers
**					- Evaluate Markers
**					- Evaluate Markers Positive
**					- Evaluate Markers Negative
**
**	\par		The following actions are defined but not yet supported.
**
**
**	\par		The following actions are neither defined or supported.
**					- Drop Objects
**					- Lift Objects
**					- Move With
**				
**	\par		REMARKS:
**				Also, at this time no statistics are altered.
*/
////////////////////////////////////////////////////////////////////////////////
#pragma once
#include "commandPacket.hpp"
#include "fishBodyHandle.hpp"
#include "StrObj.hpp"
#include "../fileio/memento.h"
namespace UDPNet 
{
	class NetMessage;
};


namespace InterpretationModule
{
	struct FrozenFish;


	//Constant values used by the Action classes.
	extern const float FOLLOW_ACTION_MIN_DISTANCE;	//!< Distance that follow cmd will try to maintain.
	extern const float EVAL_MARKER_ACTION_RANGE;	//!< Max distance that a eval marker will be considered.
	extern const float ATTACK_ACTION_MIN_DISTANCE;	//!< Min distance before the fish begin an attack.
	extern const float ATTACK_ACTION_TIME_DELAY;	//!< Time delay before fish can land an attack
	extern const float EAT_ACTION_MIN_DISTANCE;		//!< Min distance before the fish begin eatting.
	extern const float EAT_ACTION_TIME_DELAY;		//!< Delay before fish can land an eat command.
	extern const float ROAM_ACTION_MIN_DISTANCE;	//!< Min distance before roam action chooses a new target.
	extern const float GOTO_ACTION_MIN_DISTANCE;	//!< Min distance before we stop moving towards the target.
	extern const float FLEE_ACTION_MIN_DISTANCE;
	extern const float MOVE_WITH_TIME;				//!< Ammount of time 

	////////////////////////////////////////////////////////////////////////////////
	/*!	\class		Action
	**	\brief		Base class for all fish body actions to derive from.
	**	\author		Nate Cleveland
	**	\date		10/21/2002 12:06:46 PM
	**	\warning	none
	*/
	////////////////////////////////////////////////////////////////////////////////
	class Action: public FileMemento
	{
		bool			m_isDone;		//!< Flag set to true when the action is done.
		bool			m_isComplete;	//!< Flag set to false iff the action WAS completed 
										//!<  but was completed unsuccesfully..
		float			m_elapsedTime;	//!< Time elapsed since this object was created.

	protected:
		HFishBody&		m_hFishBody;	//!< Handle to this action's body.
		CommandPacket	m_cmdPack;		//!< command packet for this action.


		StrObj&			m_strObj;


	public:
		friend struct FrozenFish;
		//! Sets the m_isDone flag to the passed in value.
		void SetDone(bool isDone = true, bool isComplete = true)
		{
			m_isDone = isDone;
			m_isComplete = isComplete;
		}

		////////////////////////////////////////////////////////////////////////////////
		//	FUNCTION:
		/*!	\brief		ctor for the Action class.
		**
		**	\param HFishBody& hFishBody		Handle to the fish who's actions this
		**										 object will dictate.
		**	\param CommandPacket& cmdPack
		**
		**	\exception	This class will not throw an exception, but its children might.
		**	\par		DESCRIPTION: Fills out all the members that are held within the
		**				 Action class.
		**
		**	\author		Nate Cleveland
		**	\date		10/23/2002 10:53:37 AM
		**	\warning	none
		*/
		////////////////////////////////////////////////////////////////////////////////
		Action(StrObj& strObj, HFishBody& hFishBody, CommandPacket& cmdPack);
		virtual ~Action(){}


		////////////////////////////////////////////////////////////////////////////////
		//	FUNCTION:
		/*!	\brief		Process a fish action for the caller.
		**
		**	\param InterpVector_type&		A vector that'll be filled with whatever
		**										 movement is needed for this action.
		**	\return		none
		**	\exception	none
		**	\par		DESCRIPTION: Computes an action from the given command. Will
		**				 generate a potential field for completing the action. The
		**				 generated potential field will also be added to the summation
		**				 vector contained in the HFishBody.
		**
		**	\author		Nate Cleveland
		**	\date		10/23/2002 11:04:21 AM
		**	\warning	none
		*/
		////////////////////////////////////////////////////////////////////////////////
		virtual bool operator() (InternalVector_type& ) = 0;

		////////////////////////////////////////////////////////////////////////////////
		//	FUNCTION:
		/*!	\brief		Returns true if this object thinks its done with the action.
		**	\param		none
		**	\return		boolean true/false. true == object thinks it is done.
		**	\exception	none
		**	\par		DESCRIPTION: Flag does not pass judgement on if the action was
						 completed successfully, only if it thinks it is done.
		**
		**	\author		Nate Cleveland
		**	\date		10/23/2002 11:10:45 AM
		**	\warning	none
		*/
		////////////////////////////////////////////////////////////////////////////////
		bool IsDone(void)const throw(){return m_isDone;}
		bool IsCompleted(void) const throw(){return m_isComplete;}


		////////////////////////////////////////////////////////////////////////////////
		//	FUNCTION:
		/*!	\brief		Returns to the caller how long this object has been alive.
		**	\param		none
		**	\return		Return value is time in seconds since object creation.
		**	\exception	none
		**	\par		DESCRIPTION:
		**
		**	\author		Nate Cleveland
		**	\date		10/23/2002 11:13:38 AM
		**	\warning	none
		*/
		////////////////////////////////////////////////////////////////////////////////
		float ElapsedTime(void) const throw(){return m_elapsedTime;}
		void SetElapsedTime(float val) throw(){m_elapsedTime = val;}

		////////////////////////////////////////////////////////////////////////////////
		//	FUNCTION:
		/*!	\brief		Increments how long this object has been alive.
		**	\param float incVal		How long, in seconds, since last time update.
		**	\return		none
		**	\exception	none
		**	\par		DESCRIPTION: This must be called every frame to allow child
		**				 classes' internal timers to function properly.
		**
		**	\author		Nate Cleveland
		**	\date		10/23/2002 11:13:42 AM
		**	\warning	none
		*/
		////////////////////////////////////////////////////////////////////////////////
		void IncElapsedTime(float incVal) throw() {m_elapsedTime += incVal;}


		////////////////////////////////////////////////////////////////////////////////
		//	FUNCTION:	
		/*!	\brief		Creates an action from a network stream.
		**	\param		hFishBody		handle to fishBody
		**	\param		netMessage		reference to netMessage object.
		**	\param		strObj			Object to write data strings to.
		**	\return		Pntr to new action.
		**	\exception	none
		**	\author		Nate Cleveland
		**	\date		2/6/2003 5:29:44 PM
		**	\warning	none
		*/
		////////////////////////////////////////////////////////////////////////////////
		static Action* CreateAction(StrObj& strObj, HFishBody& hFishBody, UDPNet::NetMessage& netMessage);

		////////////////////////////////////////////////////////////////////////////////
		//	FUNCTION:
		/*!	\brief		Will create the proper child class for the calling function.
		**	\param		Parameters that're passed to created object, see Action::Action
		**	\return		Pointer to the new object. NULL == unable to find proper class.
		**	\param		strObj			Object to write data strings to.
		**	\exception	none
		**	\par		DESCRIPTION: 
		**				 Great for creating Action child classes because it gaurantees 
		**				 not to throw an exception.
		**
		**	\author		Nate Cleveland
		**	\date		10/23/2002 11:16:25 AM
		**	\warning	none
		**	\example	Action *newAction = Action::CreateAction(hFishBody, cmdPack);
		*/
		////////////////////////////////////////////////////////////////////////////////
		static Action* CreateAction(StrObj& strObj, HFishBody& hFishBody, CommandPacket& cmdPack) throw();


		////////////////////////////////////////////////////////////////////////////////
		//	FUNCTION:
		/*!	\brief		Will create the proper child class for the calling function.
		**
		**	\param		hFishBody	Handle to the fish who's actions this object will dictate.
		**	\param		inStream	Stream to read in and instantiate the class from.	
		**	\param		strObj			Object to write data strings to.
		**
		**	\return		Pointer to the new object. NULL == unable to find proper class.
		**	\exception	none
		**	\par		DESCRIPTION: 
		**				 Great for creating Action child classes because it gaurantees 
		**				  not to throw an exception.
		**
		**	\author		Nate Cleveland
		**	\date		10/23/2002 11:16:25 AM
		**	\warning	none
		**	\example	Action *newAction = Action::CreateAction(hFishBody, inStream);
		*/
		////////////////////////////////////////////////////////////////////////////////
		static Action* CreateAction(StrObj& strObj, HFishBody& hFishBody, std::istream&inStream) throw();

		////////////////////////////////////////////////////////////////////////////////
		//	FUNCTION:
		/*!	\brief		Returns to the caller the location of the target.
		**
		**	\param InterpPoint_type& point 	Var that'll be filled with location of
		**										 target.
		**
		**	\return		boolean. false == unable to deduce location of target.
		**	\exception	none
		**	\par		DESCRIPTION: This method will function as long as the target
		**				 is a point, key, voxel or any other target that has a position
		**				 in the game space.
		**
		**	\author		Nate Cleveland
		**	\date		10/23/2002 11:35:09 AM
		**	\warning	At this time is unable to deduce voxel locations.
		*/
		////////////////////////////////////////////////////////////////////////////////
		bool GetTargetPoint(InternalPoint_type& point)const throw();


		////////////////////////////////////////////////////////////////////////////////
		//	FUNCTION:
		/*!	\brief		Generates a vector from the given point to target.
		**
		**	\param const InterpPoint_type& fromPoint		pnt to cast vector from.
		**	\param InterpVector_type& answerVector		place to store the answer.
		**
		**	\return		boolean true/false. true good, false bad.
		**	\exception	none
		**	\par		DESCRIPTION: The answer will NOT be a unit vector.
		**
		**	\author		Nate Cleveland
		**	\date		10/23/2002 11:49:10 AM
		**	\warning	none
		*/
		////////////////////////////////////////////////////////////////////////////////
		bool GetVectorToTarget(const InternalPoint_type& fromPoint,
							   InternalVector_type& answerVector) const throw();

		////////////////////////////////////////////////////////////////////////////////
		//	FUNCTION:
		/*!	\brief		Same as other overloaded method, but assumes fromPoint is this
		**				 fish's position.
		**
		**	\param InterpVector_type& answerVector		place to store the answer.
		**
		**	\return		boolean true/false. true good, false bad.
		**	\exception	none
		**	\par		DESCRIPTION:
		**
		**	\author		Nate Cleveland
		**	\date		10/23/2002 11:57:17 AM
		**	\warning	none
		*/
		////////////////////////////////////////////////////////////////////////////////
		bool GetVectorToTarget(InternalVector_type& answerVector) const throw()
		{
			InternalPoint_type pos;						//var to hold our position
			this->m_hFishBody.GetPos(pos);				//find our position
			return GetVectorToTarget(pos, answerVector);//Call twin method w/ our position
		}

		//! Returns to the caller the distance to the target
		////////////////////////////////////////////////////////////////////////////////
		//	FUNCTION:
		/*!	\brief		Returns the caller the distance to the target.
		**	\param		none
		**	\return		Distance to the target.
		**	\exception	none
		**	\par		DESCRIPTION:
		**
		**	\author		Nate Cleveland
		**	\date		10/23/2002 12:02:49 PM
		**	\warning	none
		*/
		////////////////////////////////////////////////////////////////////////////////
		float FindDistToTarget(void) const throw()
		{
			InternalVector_type distVec;
			this->GetVectorToTarget(distVec);
			return static_cast<float>(fabs(Magnitude<float>(distVec)));
		}

		////////////////////////////////////////////////////////////////////////////////
		//	FUNCTION:
		/*!	\brief		Method that creates a vector to the target.
		**
		**	\param InterpVector_type& answerVector	Var filled with answer.
		**
		**	\return		Distance to the target. retValue of -1.0f implies error.
		**	\exception	none
		**	\par		DESCRIPTION:
		**					- Generates a unit vector to target.
		**					- Scales unit vector by the weight in the cmdPack.
		**					- Checks for errors and logs to console/file when
		**					  there is an error.
		**					- If there is an error will return -1.0f
		**					- Finds the absolute distance to the target.
		**
		**	\author		Nate Cleveland
		**	\date		10/23/2002 12:03:31 PM
		**	\warning	none
		*/
		////////////////////////////////////////////////////////////////////////////////
		float GotoTarget(InternalVector_type& answerVector) const throw();
		
		void VectorToPoint(const InternalVector_type& toPoint, 
						   InternalVector_type& answerVector) const throw();

		void VectorToPoint(const InterpPnt_type& toPoint, 
						   InternalVector_type& answerVector) const throw()
		{
			InternalVector_type a;
			a.x = toPoint[0];
			a.y = toPoint[1];
			a.z = toPoint[2];
			VectorToPoint(a, answerVector);
		}
		//Following methods are just wrappers around command Packet methods.
		CommandID_type	 GetCommandID			   (void)					const throw(){return m_cmdPack.GetCommandID();}
		AquariusKey_type GetCommandPackBodyID	   (void)					const throw(){return m_cmdPack.GetBodyID();}
		AquariusKey_type GetCommandPackSenderID	   (void)					const throw(){return m_cmdPack.GetSenderID();}
		void			 GetCommandPackTarget	   (ActionTarget& actTarget)const throw(){m_cmdPack.GetTarget(actTarget);}
		ActionTarget::ActionTargetFlag GetCommandPackTargetFlag(void)		const throw(){return m_cmdPack.GetTargetFlag();}
		AquariusKey_type GetCommandPackTargetKey   (void)					const		 {return m_cmdPack.GetTargetKey();}
		void			 GetCommandPacktTargetPoint(InternalPoint_type& pnt)const		 {m_cmdPack.GetTargetPoint(pnt);}
		void			 GetCommandPackTargetVoxel (Voxel_type& vox)		const		 {m_cmdPack.GetTargetVoxel(vox);}
		float			 GetCommandPackWeight	   (void)					const throw(){return m_cmdPack.GetWeight();}
		void			 SetCommandPackWeight	   (float weight)			throw()		 {m_cmdPack.SetWeight(weight);}
		float			 GetCommandPackFitness	   (void)					const throw(){return m_cmdPack.GetFitness();}
		unsigned int	 GetCommandPackDecisionID  (void)					const throw(){return m_cmdPack.GetDecisionID();}		
		void			 SetCommandPackTarget(const AquariusKey_type& key)	throw()      {m_cmdPack.SetTarget(key);}
		void			 SetCommandPackTarget(const ActionTarget& newTarget) throw()	 {m_cmdPack.SetTarget(newTarget);}
		
		
		////////////////////////////////////////////////////////////////////////////////
		//	FUNCTION:
		/*!	\brief		Replaces the current target with a new way.
		**
		**	\param const  ActionTarget& newTarget		Target to replace current with.
		**
		**	\return		none
		**	\exception	none
		**	\par		DESCRIPTION: Allows an action to replace a NONE target with a
						 specified point.
		**
		**	\author		Nate Cleveland
		**	\date		10/22/2002 1:13:01 PM
		**	\warning	none
		*/
		////////////////////////////////////////////////////////////////////////////////
		void SetTarget(const ActionTarget& newTarget) throw(){m_cmdPack.SetTarget(newTarget);}


		void GetPos(InternalPoint_type& pos) {m_hFishBody.GetPos(pos);}

		//orders fish to set its string.
		virtual void SetString(void) throw();

		virtual float GiveOptional(void) const throw() {return 0;}
		virtual void SetOptional(float val) throw(){return;}
		
		
		////////////////////////////////////////////////////////////////////////////////
		//	FUNCTION:
		/*!	\brief		Equality operator. Compairs this object's cmdPack to another
		**				 CommandPacket.
		**	\author		Nate Cleveland
		**	\date		10/23/2002 12:12:17 PM
		**	\warning	none
		*/
		////////////////////////////////////////////////////////////////////////////////
		bool operator ==(const CommandPacket& cmdPack)const throw(){return m_cmdPack == cmdPack;}

 		////////////////////////////////////////////////////////////////////////////////
		//	FUNCTION:
		/*!	\brief		Non-equality operator. Compairs this object's cmdPack to another
		**				 CommandPacket.
		**	\author		Nate Cleveland
		**	\date		10/23/2002 12:12:17 PM
		**	\warning	none
		*/
		////////////////////////////////////////////////////////////////////////////////
		bool operator !=(const CommandPacket& cmdPack)const throw(){return m_cmdPack != cmdPack;}




		////////////////////////////////////////////////////////////////////////////////
		//	FUNCTION:
		/*!	\brief		Writes this class to the passed in stream
		**	\param		o			Stream to write to.
		**	\return		none
		**	\exception	none
		**	\author		Nate Cleveland
		**	\date		10/16/2002 6:30:49 PM
		**	\warning	none
		*/
		////////////////////////////////////////////////////////////////////////////////
		virtual void Write(std::ostream& o) const;



		////////////////////////////////////////////////////////////////////////////////
		//	FUNCTION:	
		/*!	\brief		Gives child a chance to read in specific data from a network stream.
		**	\param		netMessage		reference to NetMessage Object to read from.
		**	\return		none
		**	\exception	iff there is an error in the stream
		**	\author		Nate Cleveland
		**	\date		2/6/2003 5:35:07 PM
		**	\warning	Children should still call their parent's write function to make
						 sure data is read that they may not be personaly responsible for.
		*/
		////////////////////////////////////////////////////////////////////////////////
		virtual void Read(UDPNet::NetMessage& netMessage){}

		////////////////////////////////////////////////////////////////////////////////
		//	FUNCTION:	
		/*!	\brief		Gives child a chance to write specific data from a network stream.
		**	\param		netMessage		reference to NetMessage Object to write to.
		**	\return		none
		**	\exception	iff there is an error in the stream
		**	\author		Nate Cleveland
		**	\date		2/6/2003 5:35:07 PM
		**	\warning	Children should still call their parent's write function to make
						 sure data is sent that they may not be personaly responsible for.
		*/
		////////////////////////////////////////////////////////////////////////////////
		virtual void Write(UDPNet::NetMessage& netMessage);

		////////////////////////////////////////////////////////////////////////////////
		//	FUNCTION:
		/*!	\brief		Reads data from the stream.
		**	\param istream&					Stream to read from.
		**	\return		none
		**	\exception	none
		**	\par		DESCRIPTION:
		**
		**	\author		Nate Cleveland
		**	\date		10/16/2002 6:30:55 PM
		**	\warning	none
		*/
		////////////////////////////////////////////////////////////////////////////////
		void Read(std::istream&inStream)
		{
			Read(inStream, m_cmdPack, m_isDone, m_elapsedTime);
		}

		////////////////////////////////////////////////////////////////////////////////
		//	FUNCTION:
		/*!	\brief		Reads data from the stream.
		**	\param istream&					Stream to read from.
		**	\param		All other paramters will be filled with the data from the stream.
		**	\return		none
		**	\exception	none
		**	\par		DESCRIPTION:
		**
		**	\author		Nate Cleveland
		**	\date		10/16/2002 6:30:55 PM
		**	\warning	none
		*/
		////////////////////////////////////////////////////////////////////////////////
		static void Read(std::istream&, CommandPacket&, bool& isDone, float& elapsedTime);
	};//end class Action



	////////////////////////////////////////////////////////////////////////////////
	/*!	\class		GotoAction
	**	\brief		Attempts to guide a fish to the specified object.
	**	\author		Nate Cleveland
	**	\date		10/21/2002 12:21:54 PM
	**	\warning	none
	*/
	////////////////////////////////////////////////////////////////////////////////
	class GotoAction: public Action
	{
	public:
		////////////////////////////////////////////////////////////////////////////////
		//	FUNCTION:
		/*!	\brief		ctor for FollowAction.
		**	\exception	Will throw an exception if the commandID is not Goto or if
		**				 the target is set to none.
		**
		**	\author		Nate Cleveland
		**	\date		10/23/2002 10:00:52 AM
		**	\warning	none
		*/
		////////////////////////////////////////////////////////////////////////////////
		GotoAction(StrObj& strObj, HFishBody& hFishBody, CommandPacket& cmdPack);
		virtual ~GotoAction(){}
		virtual void SetString(void) throw();


		////////////////////////////////////////////////////////////////////////////////
		//	FUNCTION:
		/*!	\brief		Returns to the caller a vector that leads straght to the target.
		**
		**	\param InterpVector_type			Resultant vector is stored in this param.
		**
		**	\return		boolean true/false. If false is returned Action has been aborted.
		**				 No examples should be generated from this.
		**	\exception	none
		**	\par		DESCRIPTION: The resultant vector will be scaled by the weight of
		**				 the commandPacket. Once the object is exactly at or what is
		**				 considered a reasonable distance the resultant vector returned
		**				 will be zero, true will still be returned as well.
		**
		**	\author		Nate Cleveland
		**	\date		10/21/2002 12:24:17 PM
		**	\warning	none
		*/
		////////////////////////////////////////////////////////////////////////////////
		virtual bool operator() (InternalVector_type& potenFieldVec);
	};//end GotoAction class.

	////////////////////////////////////////////////////////////////////////////////
	/*!	\class		Follow Action
	**	\brief		Follows the indicated object, will try to maintain a set distance.
	**	\author		Nate Cleveland
	**	\date		10/21/2002 12:20:49 PM
	**	\warning	none
	*/
	////////////////////////////////////////////////////////////////////////////////
	class FollowAction: public Action
	{
	public:
		////////////////////////////////////////////////////////////////////////////////
		//	FUNCTION:
		/*!	\brief		ctor for FollowAction.
		**	\exception	Will throw an exception if the commandID is not Follow or if
		**				 the target is set to none.
		**
		**	\author		Nate Cleveland
		**	\date		10/23/2002 10:00:52 AM
		**	\warning	none
		*/
		////////////////////////////////////////////////////////////////////////////////
		FollowAction(StrObj& strObj, HFishBody& hFishBody, CommandPacket& cmdPack);
		~FollowAction(){}
		void SetString(void) throw();

		////////////////////////////////////////////////////////////////////////////////
		//	FUNCTION:
		/*!	\brief		Returns to the caller a vector that leads to near the target.
		**
		**	\param InterpVector_type			Resultant vector is stored in this param.
		**
		**	\return		boolean true/false. If false is returned Action has been aborted.
		**				 No examples should be generated from this.
		**	\exception	none
		**	\par		DESCRIPTION: Similuar to GotoAction, only will try to maintain
		**				 a distance of no less then FOLLOW_ACTION_MIN_DISTANCE. If the
		**				 fish gets closer then FOLLOW_ACTION_MIN_DISTANCE this method
		**				 will return true and the zero vector.
		**
		**	\author		Nate Cleveland
		**	\date		10/21/2002 12:27:04 PM
		**	\warning	none
		*/
		////////////////////////////////////////////////////////////////////////////////
		bool operator() (InternalVector_type& potenFieldVec);
	};//end FollowAction class.



	////////////////////////////////////////////////////////////////////////////////
	/*!	\class		FleeAction
	**	\brief		Moves away from the indicated target.
	**	\author		Nate Cleveland
	**	\date		10/21/2002 12:20:49 PM
	**	\warning	none
	*/
	////////////////////////////////////////////////////////////////////////////////
	class FleeAction: public Action
	{
	public:
		////////////////////////////////////////////////////////////////////////////////
		//	FUNCTION:
		/*!	\brief		ctor for FleeAction.
		**	\exception	Will throw an exception if the commandID is not Flee or if
		**				 the target is set to none.
		**
		**	\author		Nate Cleveland
		**	\date		10/23/2002 10:00:52 AM
		**	\warning	none
		*/
		////////////////////////////////////////////////////////////////////////////////
		FleeAction(StrObj& strObj, HFishBody& hFishBody, CommandPacket& cmdPack);
		~FleeAction(){}
		void SetString(void) throw();

		////////////////////////////////////////////////////////////////////////////////
		//	FUNCTION:
		/*!	\brief		Returns to the caller a vector that leads away from the target.
		**
		**	\param InterpVector_type			Resultant vector is stored in this param.
		**
		**	\return		boolean true/false. If false is returned Action has been aborted.
		**				 No examples should be generated from this.
		**	\exception	none
		**	\par		DESCRIPTION: Similuar to GotoAction, only moves away from the
		**				 target instead of to it.
		**
		**	\author		Nate Cleveland
		**	\date		10/21/2002 12:27:04 PM
		**	\warning	none
		*/
		////////////////////////////////////////////////////////////////////////////////
		bool operator() (InternalVector_type& potenFieldVec);
	};//end FleeAction class.


	////////////////////////////////////////////////////////////////////////////////
	/*!	\class		AttackAction
	**	\brief		Moves towards the indicated target and tries to attack it.
	**	\author		Nate Cleveland
	**	\date		10/21/2002 12:20:49 PM
	**	\warning	none
	*/
	////////////////////////////////////////////////////////////////////////////////
	class AttackAction: public Action
	{
	protected:
		float	m_attackStartTime;			//!< Time attack began relative to Action::m_elapsedTime

	public:
		////////////////////////////////////////////////////////////////////////////////
		//	FUNCTION:
		/*!	\brief		ctor for AttackAction.
		**	\exception	Will throw an exception if the commandID is not attack or if
		**				 the target is set to none.
		**
		**	\author		Nate Cleveland
		**	\date		10/23/2002 10:00:52 AM
		**	\warning	none
		*/
		////////////////////////////////////////////////////////////////////////////////
		AttackAction(StrObj& strObj, HFishBody& hFishBody, CommandPacket& cmdPack);
		~AttackAction(){}
		void SetString(void) throw();
		float GiveOptional(void) const throw() {return m_attackStartTime;}
		void SetOptional(float val) throw(){m_attackStartTime = val;}

		////////////////////////////////////////////////////////////////////////////////
		//	FUNCTION:
		/*!	\brief		Moves towards and tries to attack the indicated object.
		**
		**	\param InterpVector_type			Resultant vector is stored in this param.
		**
		**	\return		boolean true/false. If false is returned Action has been aborted.
		**				 No examples should be generated from this.
		**	\exception	none
		**	\par		DESCRIPTION: Similuar to GotoAction, only besides just swimming
						 towards the target will also try to attack it.
		**
		**	\author		Nate Cleveland
		**	\date		10/21/2002 12:27:04 PM
		**	\warning	none
		*/
		////////////////////////////////////////////////////////////////////////////////
		bool operator() (InternalVector_type& potenFieldVec);

		////////////////////////////////////////////////////////////////////////////////
		//	FUNCTION:
		/*!	\brief		Streams out data for the AttackAction class.
		**	\param		Reference to stream in from.
		**	\return		none
		**	\exception	none
		**	\par		DESCRIPTION:
		**				 Adds the streaming of the attack time to the file.
		**	\author		Nate Cleveland
		**	\date		10/28/2002 12:50:05 PM
		**	\warning	none
		*/
		////////////////////////////////////////////////////////////////////////////////
		void Write(std::ostream& outStream) const
		{
			Action::Write(outStream);
			outStream << "Attack Timer: " << m_attackStartTime <<"\n";
		}

		////////////////////////////////////////////////////////////////////////////////
		//	FUNCTION:
		/*!	\brief		Reads in the one extra line that was put into the file for this
		**				 action.
		**	\param		Reference to stream to read in from.
		**	\return		none
		**	\exception	none
		**	\par		DESCRIPTION:
		**
		**	\author		Nate Cleveland
		**	\date		10/28/2002 1:26:19 PM
		**	\warning	none
		*/
		////////////////////////////////////////////////////////////////////////////////
		void Read(std::istream& inStream);


		//!see parent class for details.
		void Read(UDPNet::NetMessage& netMessage);

		//!see parent class for details.
		void Write(UDPNet::NetMessage& netMessage) ;
	};//end AttackAction class.


	////////////////////////////////////////////////////////////////////////////////
	/*!	\class		EatAction
	**	\brief		Moves towards the indicated target and tries to eat it.
	**	\author		Nate Cleveland
	**	\date		10/21/2002 12:20:49 PM
	**	\warning	none
	*/
	////////////////////////////////////////////////////////////////////////////////
	class EatAction: public Action
	{
	protected:
		float	m_eatStartTime;			//!< Time attack began relative to Action::m_elapsedTime

	public:
		////////////////////////////////////////////////////////////////////////////////
		//	FUNCTION:
		/*!	\brief		ctor for EatAction.
		**	\exception	Will throw an exception if the commandID is not eat or if
		**				 the target is set to none.
		**
		**	\author		Nate Cleveland
		**	\date		10/23/2002 10:00:52 AM
		**	\warning	none
		*/
		////////////////////////////////////////////////////////////////////////////////
		EatAction(StrObj& strObj, HFishBody& hFishBody, CommandPacket& cmdPack);
		~EatAction(){}
		void SetString(void) throw();
		float GiveOptional(void) const throw() {return m_eatStartTime;}
		void SetOptional(float val) throw(){m_eatStartTime = val;}

		////////////////////////////////////////////////////////////////////////////////
		//	FUNCTION:
		/*!	\brief		Moves towards and tries to eat the indicated object.
		**
		**	\param InterpVector_type			Resultant vector is stored in this param.
		**
		**	\return		boolean true/false. If false is returned Action has been aborted.
		**				 No examples should be generated from this.
		**	\exception	none
		**	\par		DESCRIPTION: Similuar to GotoAction, only besides just swimming
		**				 towards the target will also try to eat it.
		**
		**	\author		Nate Cleveland
		**	\date		10/21/2002 12:27:04 PM
		**	\warning	none
		*/
		////////////////////////////////////////////////////////////////////////////////
		bool operator() (InternalVector_type& potenFieldVec);


		////////////////////////////////////////////////////////////////////////////////
		//	FUNCTION:
		/*!	\brief		Streams out data for the AttackAction class.
		**	\param		Reference to stream in from.
		**	\return		none
		**	\exception	none
		**	\par		DESCRIPTION:
		**				 Adds the streaming of the eat time to the file.
		**	\author		Nate Cleveland
		**	\date		10/28/2002 12:50:05 PM
		**	\warning	none
		*/
		////////////////////////////////////////////////////////////////////////////////
		void Write(std::ostream& outStream) const
		{
			Action::Write(outStream);
			outStream << "Eat Timer: " << m_eatStartTime<<"\n";
		}

		////////////////////////////////////////////////////////////////////////////////
		//	FUNCTION:
		/*!	\brief		Reads in the one extra line that was put into the file for this
		**				 action.
		**	\param		Reference to stream to read in from.
		**	\return		none
		**	\exception	none
		**	\par		DESCRIPTION:
		**
		**	\author		Nate Cleveland
		**	\date		10/28/2002 1:26:19 PM
		**	\warning	none
		*/
		////////////////////////////////////////////////////////////////////////////////
		void Read(std::istream& inStream);

		//!see parent class for details.
		void Read(UDPNet::NetMessage& netMessage);

		//!see parent class for details.
		void Write(UDPNet::NetMessage& netMessage) ;
	};//end EatAction class.


	////////////////////////////////////////////////////////////////////////////////
	/*!	\class		MoveForwardAction
	**	\brief		Tells the fish to just swim forward.
	**	\author		Nate Cleveland
	**	\date		10/21/2002 12:20:49 PM
	**	\warning	none
	*/
	////////////////////////////////////////////////////////////////////////////////
	class MoveForwardAction: public Action
	{
	public:
		////////////////////////////////////////////////////////////////////////////////
		//	FUNCTION:
		/*!	\brief		ctor for MoveForwardAction.
		**	\exception	Will throw an exception if the commandID is not set to moveForward.
		**
		**	\author		Nate Cleveland
		**	\date		10/23/2002 10:00:52 AM
		**	\warning	none
		*/
		////////////////////////////////////////////////////////////////////////////////
		MoveForwardAction(StrObj& strObj, HFishBody& hFishBody, CommandPacket& cmdPack);
		~MoveForwardAction(){}
		void SetString(void) throw();

		////////////////////////////////////////////////////////////////////////////////
		//	FUNCTION:
		/*!	\brief		Tells the fish to simply move forward
		**
		**	\param InterpVector_type			Resultant vector is stored in this param.
		**
		**	\return		boolean true/false. If false is returned Action has been aborted.
		**				 No examples should be generated from this.
		**	\exception	none
		**	\par		DESCRIPTION: May seem needless but method is usefull when the
		**				 player has possesd the fish.
		**
		**	\author		Nate Cleveland
		**	\date		10/21/2002 12:27:04 PM
		**	\warning	none
		*/
		////////////////////////////////////////////////////////////////////////////////
		bool operator() (InternalVector_type& potenFieldVec);
	};//end MoveForwardAction class.


	////////////////////////////////////////////////////////////////////////////////
	/*!	\class		StopFinAction
	**	\brief		Tells the fish to stop moving.
	**	\author		Nate Cleveland
	**	\date		10/21/2002 12:20:49 PM
	**	\warning	none
	*/
	////////////////////////////////////////////////////////////////////////////////
	class StopFinAction: public Action
	{
	public:
		////////////////////////////////////////////////////////////////////////////////
		//	FUNCTION:
		/*!	\brief		ctor for StopFinAction.
		**	\exception	Will throw an exception if the commandID is not set to stopFin
		**
		**	\author		Nate Cleveland
		**	\date		10/23/2002 10:00:52 AM
		**	\warning	none
		*/
		////////////////////////////////////////////////////////////////////////////////
		StopFinAction(StrObj& strObj, HFishBody& hFishBody, CommandPacket& cmdPack);
		~StopFinAction(){}
		void SetString(void) throw();

		////////////////////////////////////////////////////////////////////////////////
		//	FUNCTION:
		/*!	\brief		Stops all movement of the fish.
		**
		**	\param InterpVector_type			Will always be set to a zero vector.
		**
		**	\return		boolean true/false. If false is returned Action has been aborted.
		**				 No examples should be generated from this.
		**	\exception	none
		**	\par		DESCRIPTION: If Action is initiated it will poison the movement
		**				 of all the other actions and may prevent them from completing
		**				 their actions.
		**
		**	\author		Nate Cleveland
		**	\date		10/21/2002 12:27:04 PM
		**	\warning	none
		*/
		////////////////////////////////////////////////////////////////////////////////
		bool operator() (InternalVector_type& potenFieldVec);
	};//end StopFinAction class.


	////////////////////////////////////////////////////////////////////////////////
	/*!	\class		StopMouthAction
	**	\brief		Tells the fish to stop all mouth
	**	\author		Nate Cleveland
	**	\date		10/21/2002 12:20:49 PM
	**	\warning	none
	*/
	////////////////////////////////////////////////////////////////////////////////
	class StopMouthAction: public Action
	{
	public:
		////////////////////////////////////////////////////////////////////////////////
		//	FUNCTION:
		/*!	\brief		ctor for StopMouthAction.
		**	\exception	Will throw an exception if the commandID is not set to stopMouth
		**
		**	\author		Nate Cleveland
		**	\date		10/23/2002 10:00:52 AM
		**	\warning	none
		*/
		////////////////////////////////////////////////////////////////////////////////
		StopMouthAction(StrObj& strObj, HFishBody& hFishBody, CommandPacket& cmdPack);
		~StopMouthAction(){}
		void SetString(void) throw();

		////////////////////////////////////////////////////////////////////////////////
		//	FUNCTION:
		/*!	\brief		Prevents the fish from using its mouth.
		**
		**	\param InterpVector_type			Resultant vector is stored in this param.
		**
		**	\return		boolean true/false. If false is returned Action has been aborted.
		**				 No examples should be generated from this.
		**	\exception	none
		**	\par		DESCRIPTION: If Action is initiated it will poison the mouth
		**				 related actions of all other actions. This may stop the completion
		**				 of these aborted actions.
		**
		**	\author		Nate Cleveland
		**	\date		10/21/2002 12:27:04 PM
		**	\warning	none
		*/
		////////////////////////////////////////////////////////////////////////////////
		bool operator() (InternalVector_type& potenFieldVec);
	};//end StopMouthAction class.


	////////////////////////////////////////////////////////////////////////////////
	/*!	\class		RoamAction
	**	\brief		Picks a semi-random point and swims towards it.
	**	\author		Nate Cleveland
	**	\date		10/21/2002 12:20:49 PM
	**	\warning	none
	*/
	////////////////////////////////////////////////////////////////////////////////
	class RoamAction: public Action
	{
	protected:

	public:
		////////////////////////////////////////////////////////////////////////////////
		//	FUNCTION:
		/*!	\brief		Same as other ctors but also chooses a random point for the fish
		**				 to swim to.
		**	\exception	Will throw an exception if the commandID is not set to Roam
		**
		**	\author		Nate Cleveland
		**	\date		10/22/2002 1:27:34 PM
		**	\warning	none
		*/
		////////////////////////////////////////////////////////////////////////////////
		RoamAction(StrObj& strObj, HFishBody& hFishBody, CommandPacket& cmdPack);
		~RoamAction(){}
		void SetString(void) throw();

		////////////////////////////////////////////////////////////////////////////////
		//	FUNCTION:
		/*!	\brief		Moves the fish towards a semi-random point.
		**
		**	\param InterpVector_type			Resultant vector is stored in this param.
		**
		**	\return		boolean true/false. If false is returned Action has been aborted.
		**				 No examples should be generated from this.
		**	\exception	none
		**	\par		DESCRIPTION: Similuar to GotoAction, only it will choose a random
		**				 location to swim to.
		**
		**	\author		Nate Cleveland
		**	\date		10/21/2002 12:27:04 PM
		**	\warning	none
		*/
		////////////////////////////////////////////////////////////////////////////////
		bool operator() (InternalVector_type& potenFieldVec);


		////////////////////////////////////////////////////////////////////////////////
		//	FUNCTION:
		/*!	\brief		Streams out data for the Roam class.
		**	\param		Reference to stream in from.
		**	\return		none
		**	\exception	none
		**	\par		DESCRIPTION:
		**				 Adds the streaming of the attack time to the file.
		**	\author		Nate Cleveland
		**	\date		10/28/2002 12:50:05 PM
		**	\warning	none
		*/
		////////////////////////////////////////////////////////////////////////////////
		void Write(std::ostream& outStream) const
		{
			Action::Write(outStream);
		}
	};//end RoamAction class.

	////////////////////////////////////////////////////////////////////////////////
	/*!	\class		EvaluateMarkersAction_Pos
	**	\brief		Will evaluate any markers that are positvie and considederd within range.
	**	\author		Nate Cleveland
	**	\date		10/21/2002 12:20:49 PM
	**	\warning	none
	*/
	////////////////////////////////////////////////////////////////////////////////
	class EvaluateMarkersAction_Pos: public Action
	{
		float m_markerTimer;

	public:
		////////////////////////////////////////////////////////////////////////////////
		//	FUNCTION:
		/*!	\brief		ctor for EvaluateMarkersAction_Pos.
		**	\exception	Will throw an exception if the commandID is not set to
		**				 EvaluateMarkers_Pos.
		**
		**	\author		Nate Cleveland
		**	\date		10/23/2002 10:00:52 AM
		**	\warning	none
		*/
		////////////////////////////////////////////////////////////////////////////////
		EvaluateMarkersAction_Pos(StrObj& strObj, HFishBody& hFishBody, CommandPacket& cmdPack);
		~EvaluateMarkersAction_Pos(){}
		float GiveOptional(void) const throw() {return m_markerTimer;}
		void SetOptional(float val) throw(){m_markerTimer = val;}

		////////////////////////////////////////////////////////////////////////////////
		//	FUNCTION:
		/*!	\brief		Tells the fish to take into account only positive markers in range.
		**
		**	\param InterpVector_type			Resultant vector is stored in this param.
		**
		**	\return		boolean true/false. If false is returned Action has been aborted.
		**				 No examples should be generated from this.
		**	\exception	none
		**	\par		DESCRIPTION: Markers refered to in this method are set by the
		**				 DropMarkerAction. "In range" is defined by EVAL_MARKER_ACTION_RANGE.
		**				 These markers are finite and cumulative. If a negative weight
		**				 is set in the CommandPacket it will corrupt the meaning of the
		**				 markers that are left behind.
		**
		**	\author		Nate Cleveland
		**	\date		10/21/2002 12:27:04 PM
		**	\warning	none
		*/
		////////////////////////////////////////////////////////////////////////////////
		bool operator() (InternalVector_type& potenFieldVec);

		//!see parent class for details.
		void Read(UDPNet::NetMessage& netMessage);

		//!see parent class for details.
		void Write(UDPNet::NetMessage& netMessage) ;
	};//end EvaluateMarkersAction_Pos class.


	////////////////////////////////////////////////////////////////////////////////
	/*!	\class		EvaluateMarkersAction_Neg
	**	\brief		Will evaluate any markers that are Negative and considederd within range.
	**	\author		Nate Cleveland
	**	\date		10/21/2002 12:20:49 PM
	**	\warning	none
	*/
	////////////////////////////////////////////////////////////////////////////////
	class EvaluateMarkersAction_Neg: public Action
	{
		float m_markerTimer;


	public:
		////////////////////////////////////////////////////////////////////////////////
		//	FUNCTION:
		/*!	\brief		ctor for EvaluateMarkersAction_Neg.
		**	\exception	Will throw an exception if the commandID is not set to
		**				 EvaluateMarkers_Pos.
		**
		**	\author		Nate Cleveland
		**	\date		10/23/2002 10:00:52 AM
		**	\warning	none
		*/
		////////////////////////////////////////////////////////////////////////////////
		EvaluateMarkersAction_Neg(StrObj& strObj, HFishBody& hFishBody, CommandPacket& cmdPack);
		~EvaluateMarkersAction_Neg(){}
		float GiveOptional(void) const throw() {return m_markerTimer;}
		void SetOptional(float val) throw(){m_markerTimer = val;}

		////////////////////////////////////////////////////////////////////////////////
		//	FUNCTION:
		/*!	\brief		Tells the fish to take into account only negative markers in range.
		**
		**	\param InterpVector_type			Resultant vector is stored in this param.
		**
		**	\return		boolean true/false. If false is returned Action has been aborted.
		**				 No examples should be generated from this.
		**	\exception	none
		**	\par		DESCRIPTION: Markers refered to in this method are set by the
		**				 DropMarkerAction. "In range" is defined by EVAL_MARKER_ACTION_RANGE.
		**				 These markers are finite and cumulative. If a negative weight
		**				 is set in the CommandPacket it will corrupt the meaning of the
		**				 markers that are left behind.
		**
		**	\author		Nate Cleveland
		**	\date		10/21/2002 12:27:04 PM
		**	\warning	none
		*/
		////////////////////////////////////////////////////////////////////////////////
		bool operator() (InternalVector_type& potenFieldVec);


		//!see parent class for details.
		void Read(UDPNet::NetMessage& netMessage);

		//!see parent class for details.
		void Write(UDPNet::NetMessage& netMessage) ;
	};//end EvaluateMarkersAction_Pos class.


	////////////////////////////////////////////////////////////////////////////////
	/*!	\class		EvaluateMarkersAction
	**	\brief		Will evaluate any markers that are considederd within range.
	**	\author		Nate Cleveland
	**	\date		10/21/2002 12:20:49 PM
	**	\warning	none
	*/
	////////////////////////////////////////////////////////////////////////////////
	class EvaluateMarkersAction: public Action
	{
		float m_markerTimer;
	public:
		////////////////////////////////////////////////////////////////////////////////
		//	FUNCTION:
		/*!	\brief		ctor for EvaluateMarkersAction.
		**	\exception	Will throw an exception if the commandID is not set to
		**				 EvaluateMarkers.
		**
		**	\author		Nate Cleveland
		**	\date		10/23/2002 10:00:52 AM
		**	\warning	none
		*/
		////////////////////////////////////////////////////////////////////////////////
		EvaluateMarkersAction(StrObj& strObj, HFishBody& hFishBody, CommandPacket& cmdPack);
		~EvaluateMarkersAction(){}
		float GiveOptional(void) const throw() {return m_markerTimer;}
		void SetOptional(float val) throw(){m_markerTimer = val;}

		////////////////////////////////////////////////////////////////////////////////
		//	FUNCTION:
		/*!	\brief		Tells the fish to take into account any markers in range.
		**
		**	\param InterpVector_type			Resultant vector is stored in this param.
		**
		**	\return		boolean true/false. If false is returned Action has been aborted.
		**				 No examples should be generated from this.
		**	\exception	none
		**	\par		DESCRIPTION: Markers refered to in this method are set by the
		**				 DropMarkerAction. "In range" is defined by EVAL_MARKER_ACTION_RANGE.
		**				 These markers are finite and cumulative. If a negative weight
		**				 is set in the CommandPacket it will corrupt the meaning of the
		**				 markers that are left behind.
		**
		**	\author		Nate Cleveland
		**	\date		10/21/2002 12:27:04 PM
		**	\warning	none
		*/
		////////////////////////////////////////////////////////////////////////////////
		bool operator() (InternalVector_type& potenFieldVec);

		//!see parent class for details.
		void Read(UDPNet::NetMessage& netMessage);

		//!see parent class for details.
		void Write(UDPNet::NetMessage& netMessage) ;
	};//end EvaluateMarkersAction class.


	////////////////////////////////////////////////////////////////////////////////
	/*!	\class		DropMarkersAction
	**	\brief		Will drop markers that can later be evaluated by the EvalMarkerAction
	**	\author		Nate Cleveland
	**	\date		10/21/2002 12:20:49 PM
	**	\warning	none
	*/
	////////////////////////////////////////////////////////////////////////////////
	class DropMarkersAction: public Action
	{
	public:
		////////////////////////////////////////////////////////////////////////////////
		//	FUNCTION:
		/*!	\brief		ctor for DropMarkersAction.
		**	\exception	Will throw an exception if the commandID is not set to
		**				 DropMarkers.
		**
		**	\author		Nate Cleveland
		**	\date		10/23/2002 10:00:52 AM
		**	\warning	none
		*/
		////////////////////////////////////////////////////////////////////////////////
		DropMarkersAction(StrObj& strObj, HFishBody& hFishBody, CommandPacket& cmdPack);
		~DropMarkersAction(){}

		////////////////////////////////////////////////////////////////////////////////
		//	FUNCTION:
		/*!	\brief		Tells the fish to leave behind markers for later use.
		**
		**	\param InterpVector_type			Will always be set to a zero vector.
		**
		**	\return		boolean true/false. If false is returned Action has been aborted.
		**				 No examples should be generated from this.
		**	\exception	none
		**	\par		DESCRIPTION: Markers are something the fish can leave behind to
		**				 remember this location as either a good place or a bad one.
		**				 The location is considered bad if the weight of the commandPacket
		**				 is < 0. Thus, the location is condsidered good if the weight is
		**				 > 0. Markers are only considered if the EvaluateMarkers action
		**				 is initiated. The number of markers is finite.
		**
		**	\author		Nate Cleveland
		**	\date		10/21/2002 12:27:04 PM
		**	\warning	none
		*/
		////////////////////////////////////////////////////////////////////////////////
		bool operator() (InternalVector_type& potenFieldVec);
	};//end DropMarkersAction class.


	////////////////////////////////////////////////////////////////////////////////
	/*!	\class		MoveWithAction
	**	\brief		Will attempt to move with another fish (psuedo flocking).
	**	\author		Nate Cleveland
	**	\date		10/21/2002 12:20:49 PM
	**	\warning	none
	*/
	////////////////////////////////////////////////////////////////////////////////
	class MoveWithAction: public Action
	{
		float	m_moveWithStartTime;	//!< Time that this fish began to psuedo flock
	public:
		////////////////////////////////////////////////////////////////////////////////
		//	FUNCTION:
		/*!	\brief		ctor for MoveWithAction.
		**	\exception	Will throw an exception if the commandID is not set to
		**				 MoveWith.
		**
		**	\author		Nate Cleveland
		**	\date		10/23/2002 10:00:52 AM
		**	\warning	none
		*/
		////////////////////////////////////////////////////////////////////////////////
		MoveWithAction(StrObj& strObj, HFishBody& hFishBody, CommandPacket& cmdPack);
		~MoveWithAction(){}
		void SetString(void) throw();
		float GiveOptional(void) const throw() {return m_moveWithStartTime;}
		void SetOptional(float val) throw(){m_moveWithStartTime = val;}

		////////////////////////////////////////////////////////////////////////////////
		//	FUNCTION:
		/*!	\brief		Tells the fish to Move with another fish (psuedo flocking).
		**
		**	\param InterpVector_type			Will always be set to a zero vector.
		**
		**	\return		boolean true/false. If false is returned Action has been aborted.
		**				 No examples should be generated from this.
		**	\exception	none
		**	\par		DESCRIPTION: 
		**				 Psuedo flocking. At this time only attempts to move towards the
						  same target at the specified fish. In the future it will try to
						  simulate the same movement style.
		**
		**	\author		Nate Cleveland
		**	\date		10/21/2002 12:27:04 PM
		**	\warning	none
		*/
		////////////////////////////////////////////////////////////////////////////////
		bool operator() (InternalVector_type& potenFieldVec);

		//!see parent class for details.
		void Read(UDPNet::NetMessage& netMessage);

		//!see parent class for details.
		void Write(UDPNet::NetMessage& netMessage) ;
	};//end MoveWithAction class.
}//end namespace InterpretationModule.