#include "angleMgr.hpp"
#include "fishBodyMath.hpp"
#include <assert.h>
// memory leak detection
#ifdef _DEBUG
    #include"memleak.h"
    #define new DEBUG_NEW
    #undef THIS_FILE
    static char THIS_FILE[] = __FILE__;
#endif

using namespace InterpretationModule;

///#define ANGLE_MGR_LOG
#ifdef ANGLE_MGR_LOG
#	include "../debugging/Debug.h"
	namespace 
	{
		char buff[1024] ={0};
	}
#endif
	


////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	
/*!	\brief		Finds the angle of a vector given the value of its independant and
				 dependant elements.
**	\param		independant		Independant element of a the given vector. 
**	\param		dependant		Dependant element of a the given vector. 
**	\return		Value of the angle.
**	\exception	none	
**	\par		DESCRIPTION:
		
**	\author		Nate Cleveland
**	\date		1/24/2003 10:23:42 AM
**	\warning	Value should be negated and added to 2 pi in order to bring it into
				 the cord system of the rest of the game.
*/
////////////////////////////////////////////////////////////////////////////////
#include <iostream>
using namespace std;
inline float FindAngle(float x, float y)
{
	if(x == 0 && y == 0)
		return 0;	

	float length = sqrtf(x*x + y*y);
	return atan2f(x/length, y/length);

}


//see header for documentation. Also read documentation on function FindAngle
InterpAngle_type AngleMgr::FindXZAngle(const InterpVec_type& vec) 
{
	return FindAngle(vec[0], vec[2]);
}


//see header for documentation. Also read documentation on function FindAngle
InterpAngle_type AngleMgr::FindXYAngle(const InterpVec_type& vec) 
{
	return FindAngle(vec[0], vec[1]);
}

 
//see header for documentation. Also read documentation on function FindAngle
InterpAngle_type AngleMgr::FindYZAngle(const InterpVec_type& vec) 
{
	return -FindAngle(vec[1], vec[2]);
}

//see header for documentation
void AngleMgr::FindEulerAngles(const InterpVec_type& toVec, const InterpVec_type& fromVec, 
					 InterpAngle_type& xyAngle, InterpAngle_type& zxAngle) const throw()
{

	////check for nearly no change vectors
	float deltaX = static_cast<float>(fabs(toVec[0] - fromVec[0]));
	float deltaY = static_cast<float>(fabs(toVec[1] - fromVec[1]));
	float deltaZ = static_cast<float>(fabs(toVec[2] - fromVec[2]));

	//if all the changes are this small don't bother calling atan.
	if(deltaX < .01 && deltaY < .01 && deltaY < .01) 
	{
		xyAngle = 0;
		zxAngle = 0;
		return;
	}
	

	//create normalized versions of each vector.
	InterpVec_type	normToVec, normFromVec;

	UnitizeVector<InterpAngle_type>(toVec, normToVec);
	UnitizeVector<InterpAngle_type>(fromVec, normFromVec);


	//Find angle on the xyPlane aka rotation w.r.t. the zAxis.
	xyAngle = FindXYAngle(normToVec) - FindXYAngle(normFromVec);
	zxAngle = FindXZAngle(normToVec) - FindXZAngle(normFromVec);

#	ifdef ANGLE_MGR_LOG
		sprintf(buff, "to(%f,%f,%f) from(%f, %f, %f)",
				toVec[0],   toVec[1],   toVec[2],
				fromVec[0], fromVec[1], fromVec[2]);
		(Debugging::Debugger::Instance())->output(Debugging::BOD, Debugging::L4, buff);


		sprintf(buff, "nTo(%f,%f,%f) nFrom(%f, %f, %f)",
				normToVec[0], normToVec[1], normToVec[2],
				normFromVec[0], normFromVec[1], normFromVec[2]);
		(Debugging::Debugger::Instance())->output(Debugging::BOD, Debugging::L4, buff);
#	endif
}

//see header for documentation
void AngleMgr::ApplyEulerAngles(const InterpVec_type&v, InterpVec_type& ans, float z, float y) const throw()
{
	//fill out the multVectors
	math::linear_alg::Matrix3<float> yRot(
			static_cast<float>(cos(y)),		0,		static_cast<float>(-sin(y)),
			0,								1,		0,
			static_cast<float>(sin(y)),		0,		static_cast<float>(cos(y)));
	
	math::linear_alg::Matrix3<float> zRot(
			static_cast<float>(cos(z)),		static_cast<float>(sin(z)),		0,
			static_cast<float>(-sin(z)),	static_cast<float>(cos(z)),		0,
			0,								0,								1);

	yRot *= zRot;
	Multiply(yRot, v, ans);
}


//Converts a rotation matrix to a quaternion.
void AngleMgr::ConvertToQuaterion(const math::linear_alg::Matrix4<float>& rotMatrix, float& s, 
								  math::linear_alg::Vector4<float>& q) const throw()
#if 1
{
	//first calculate the trace of the matrix
	float rotTrace = rotMatrix[0][0] + rotMatrix[1][1] + rotMatrix[2][2] + rotMatrix[3][3];
	
	//if the trace is positive then the answer is simple.
	if(rotTrace > 0)
	{
		s			= .5f / static_cast<float>(sqrt(rotTrace));
		q[0]		= (rotMatrix[2][1] - rotMatrix[1][2]) * s;
		q[1]		= (rotMatrix[0][2] - rotMatrix[2][0]) * s;
		q[2]		= (rotMatrix[1][0] - rotMatrix[0][1]) * s;
		q[3]		= .25f / s;				
	}//end if

	//other wise execute on of the other three formulae.
	else if(rotMatrix[0][0] > rotMatrix[1][1] && rotMatrix[0][0] > rotMatrix[2][2])
	{
		s			= static_cast<float>(sqrt(1.0 + rotMatrix[0][0] - rotMatrix[1][1] - rotMatrix[2][2])) * 2;
		q[0]		= 0.5f / s;
		q[1]		= (rotMatrix[0][1] + rotMatrix[1][0]) / s;
		q[2]		= (rotMatrix[0][2] + rotMatrix[2][0]) / s;;
		q[3]		= (rotMatrix[1][2] + rotMatrix[2][1]) / s;;		
	}//end else if

	else if(rotMatrix[1][1] > rotMatrix[2][2])
	{
		s			= static_cast<float>(sqrt(1.0 + rotMatrix[1][1] - rotMatrix[0][0] - rotMatrix[2][2])) * 2;
		q[0]		= (rotMatrix[0][1] + rotMatrix[1][0]) / s;
		q[1]		= .5f / s;
		q[2]		= (rotMatrix[1][2] + rotMatrix[2][1]) / s;
		q[3]		= (rotMatrix[0][2] + rotMatrix[2][0]) / s;
	}//end else if

	else 
	{
		s			= static_cast<float>(sqrt(1.0 + rotMatrix[2][2] - rotMatrix[0][0] - rotMatrix[1][1])) * 2;
		q[0]		= (rotMatrix[0][2] + rotMatrix[2][0]) / s;
		q[1]		= (rotMatrix[1][2] + rotMatrix[2][1]) / s;
		q[2]		= .5f / s;
		q[3]		= (rotMatrix[0][1] + rotMatrix[1][0]) / s;
	}//end else if	
}
#elif 0
{
	s = sqrtf(rotMatrix[0][0] + rotMatrix[1][1] + rotMatrix[2][2] + rotMatrix[3][3]) / 2;
	float val = 1 - 2 * s * s;


	q[0] = sqrtf((rotMatrix[0][0] + val) / 2);
	q[1] = sqrtf((rotMatrix[1][1] + val) / 2);
	q[2] = sqrtf((rotMatrix[2][2] + val) / 2);



}
#else 
{    /* From Wertz pg. 415    
	 w = Math.sqrt(1.0 + m1.m00 + m1.m11 + m1.m22) / 2.0;    
	 double w4 = (4.0 * w);   
	 x = (m1.m12 - m1.m21) / w4 ;    
	 y = (m1.m20 - m1.m02) / w4 ;   
	 z = (m1.m01 - m1.m10) / w4 ;
	 */

	s = sqrtf(rotMatrix[0][0] + rotMatrix[1][1] + rotMatrix[2][2] + rotMatrix[3][3]) / 2;
	float wtf = 4 * s;
	q[0] = (rotMatrix[1][2] - rotMatrix[2][1] ) / wtf;
	q[1] = (rotMatrix[2][0] - rotMatrix[0][2] ) / wtf;
	q[2] = (rotMatrix[0][1] - rotMatrix[1][0] ) / wtf;

	float n = sqrtf(q[0]*q[0] + q[1]*q[1] + q[2]*q[2]);
	q[0] /= n;
	q[2] /= n;
	q[1] /= n;
  
}



#endif