////////////////////////////////////////////////////////////////////////////////
/*!	\file		angleMgr.hpp
**	\brief		AngleMgr class provides a simple way for interpretationModules 
**				 objects to add angle logic to themselves.
**	\author		Nate Cleveland
**	\date		11/4/2002 12:38:48 PM
**	\par		DESCRIPTION:
**					Terms used by this class and else where in angle documentation
**					 may seem a little misleading if thought is applied to the terms.
**					 Specificaly the phrase "Find the angle of a Vector" is used thoughout
**					 class documentation. This is shorthand for finding the angle of a 
**					 two element portion of a 3D vector between itself and the dependant 
**					 element's elementary axis. This is quite a mouthful and would lead
**					 to confusion when first read by clients. Therefore the term "Finding
**					 the Angle" is used.
**
**	\par		For instance,
**					When finding the "angle" of a vector on the XZ plane what is truly happening
**					 is that the angle between the vector <Vx Vz> and <1 0>.
**				
**	
**	\par		REMARKS:
**					- Object does not truncate angle values to 0 to 2*pi.	
					- All angles used by this object are in radians.
*/
////////////////////////////////////////////////////////////////////////////////
#pragma once
#include "fishBodyUtils.hpp"
#include <math.h>




namespace InterpretationModule
{
	////////////////////////////////////////////////////////////////////////////////
	/*!	\class		AngleMgr
	**	\brief		Allows a simple way for InterpretationModule object to add 
	**				 angle logic to theirselves.
	**	\author		Nate Cleveland
	**	\date		10/30/2002 4:16:33 PM
	**	\warning	none
	*/
	////////////////////////////////////////////////////////////////////////////////
	class AngleMgr
	{
	protected:
		InterpAngle_type	m_AngleZY;	//!< Roll of fish. z/y plane angle. or x-axis rotation.
		InterpAngle_type	m_AngleXY;	//!< Yaw of fish.x/y plane angle. or z-axis rotation.
		InterpAngle_type	m_AngleXZ;	//!< Pitch of fish.x/z plane angle. or y-axis rotation.

	public:
		//sets all values to zero.
		////////////////////////////////////////////////////////////////////////////////
		//	FUNCTION:	
		/*!	\brief		Defualt ctor
		**	\param		none
		**	\return		none
		**	\exception	none
		**	\par		DESCRIPTION:
		**				Sets all values to zero.	
		**	\author		Nate Cleveland
		**	\date		11/4/2002 12:42:27 PM
		**	\warning	none
		*/
		////////////////////////////////////////////////////////////////////////////////
		AngleMgr()throw() :m_AngleZY(0), m_AngleXY(0), m_AngleXZ(0){}
		~AngleMgr(){}


		//Follows are all the members that give the values of this classes members
		InterpAngle_type GetZYAngle(void)const throw(){return m_AngleZY;}
		InterpAngle_type GetXYAngle(void)const throw(){return m_AngleXY;}
		InterpAngle_type GetXZAngle(void)const throw(){return m_AngleXZ;}

		InterpAngle_type GetRoll(void) const throw(){return GetZYAngle();}
		InterpAngle_type GetPitch(void) const throw(){return GetXZAngle();}
		InterpAngle_type GetYaw(void) const throw(){return GetXYAngle();}


		//Follows are all the needed methods to change the value of this classes members
		void SetZYAngle(InterpAngle_type val)	throw(){m_AngleZY = val;}//CapAngle(m_AngleZY);}// 
		void SetXYAngle(InterpAngle_type val)	throw(){m_AngleXY = val;CapAngle(m_AngleXY);}//}//
		void SetXZAngle(InterpAngle_type val)	throw(){m_AngleXZ = val;}//CapAngle(m_AngleXZ);}//
		
		void SetRoll(InterpAngle_type val)		throw(){SetZYAngle(val);}
		void SetPitch(InterpAngle_type val)	throw(){SetXZAngle(val);}
		void SetYaw(InterpAngle_type val)		throw(){SetXYAngle(val);}


		////////////////////////////////////////////////////////////////////////////////
		//	FUNCTION:	
		/*!	\brief		Converts an angle to a value between pi and -pi.
		**	\param		
		**	\return		none
		**	\exception	none			
		**	\author		Nate Cleveland
		**	\date		1/24/2003 10:50:09 AM
		**	\warning	none
		*/
		////////////////////////////////////////////////////////////////////////////////
		static void CapAngle(InterpAngle_type& val)
		{
			//wraps it to less then 2*pi
			while(val > pi)	 val -= 2*pi;

			//wraps it to greater then 0
			while(val < -pi) val += 2*pi;	
		}


		////////////////////////////////////////////////////////////////////////////////
		//	FUNCTION:	
		/*!	\brief		Finds relevent euler angles between two vectors.
		**
		**	\param		toVec		Vector where the angle calculation stops.		
		**	\param		fromVec		Vector where the angle calculation starts.
		**	\param		xyAngle		Angle on the xyPlane aka rotation w.r.t. the zAxis 
		**							 aka 'yaw'.
		**	\param		xyAngle		Angle on the zxPlane aka rotation w.r.t. the yAxis 
		**							 aka 'pitch'.
		**
		**	\return		none
		**	\exception	none
		**	\par		DESCRIPTION:
		**					Finds the two relevent euler angle between the given two 
		**					 vectors. The roll is ignored since it is unnessicary and 
		**					 using it will only lead to gimbil lock.
		**					
		**	\author		Nate Cleveland
		**	\date		1/21/2003 10:01:28 AM
		**	\warning	none
		*/
		////////////////////////////////////////////////////////////////////////////////
		void FindEulerAngles(const InterpVec_type& toVec, const InterpVec_type& fromVec, 
							 InterpAngle_type& xyAngle, InterpAngle_type& zxAngle) const throw();


		void ApplyEulerAngles(const InterpVec_type&v, InterpVec_type& ans, float z, float y) const throw();

		////////////////////////////////////////////////////////////////////////////////
		//	FUNCTION:	
		/*!	\brief		Converts a rotation matrix into a quaterion.
		**	\param		rotMatrix		Pntr to a rotation matrix that needs converting.
		**	\param		s				solution angle
		**	\param		q				Vector to rotate around.
		**	\return		none
		**	\exception	none
		**	\par		DESCRIPTION:
		**					Nothing is computed if rotMatrix is null. Uses formula given 
		**					 at "www.martinb.com/maths/geometry/rotations/conversions/
		**					 matrixToQuaternion/index.htm". 
		**
		**	\author		Nate Cleveland
		**	\date		1/22/2003 10:16:33 AM
		**	\warning	Equation is large and ungainly, not for the weak of hear.
		*/
		////////////////////////////////////////////////////////////////////////////////
		void ConvertToQuaterion(const math::linear_alg::Matrix4<float>& rotMatrix, float& s, 
								math::linear_alg::Vector4<float>& q) const throw();


		//static methods.	
		////////////////////////////////////////////////////////////////////////////////
		//	FUNCTION:	
		/*!	\brief		Finds the angle generated by this vector.
		**
		**	\param		u		First vector to compute.
		**	\param		v		Second vector to compute.
		**
		**	\return		none
		**	\exception	none
		**	\par		DESCRIPTION: 
		**				 Returns to the caller the angle in the x/z plane between two 
		**				 vectors. Angle will be in radians.
		**
		**	\author		Nate Cleveland
		**	\date		10/30/2002 4:54:40 PM
		**	\warning	none
		*/
		////////////////////////////////////////////////////////////////////////////////
		static InterpAngle_type FindXZAngle(const InterpVec_type& vec);



		////////////////////////////////////////////////////////////////////////////////
		//	FUNCTION:	
		/*!	\brief		Finds the angle in z/y plane. aka rotation about the x-axis aka
		**				 Roll.
		**
		**	\param		vec		Vector to generate angle from
		**
		**	\return		none
		**	\exception	none
		**	\par		DESCRIPTION:
		**				 Returns to the caller the angle of the given vector.
		**
		**	\author		Nate Cleveland
		**	\date		10/30/2002 4:54:40 PM
		**	\warning	This method has NOT been tested. It may or may not work.
		*/
		////////////////////////////////////////////////////////////////////////////////
		static InterpAngle_type FindYZAngle(const InterpVec_type& vec) ;

		
		////////////////////////////////////////////////////////////////////////////////
		//	FUNCTION:	
		/*!	\brief		Finds the angle in x/y plane. aka rotation about the z-axis.
		**
		**	\param		vec		Vector to generate angle from
		**
		**	\return		none
		**	\exception	none
		**	\par		DESCRIPTION:
		**				 Returns to the caller the angle of the given vector.
		**
		**	\author		Nate Cleveland
		**	\date		10/30/2002 4:54:40 PM
		**	\warning	none
		*/
		////////////////////////////////////////////////////////////////////////////////
		static InterpAngle_type FindXYAngle(const InterpVec_type& vec) ;


		static InterpAngle_type RadianToDegree(InterpAngle_type angle){return angle*(180.0f/pi);}
		static InterpAngle_type DegreeToRadian(InterpAngle_type degree){return degree*(pi/180.0f);}
	};//end FishBodyPhysics
}//namespace