////////////////////////////////////////////////////////////////////////////////
/*!	\file		attributeGenerator.cpp
**	\brief		Defines all methods and const varibles declared in attributeGenerator.hpp
**	\author		Nate Cleveland
**	\date		10/16/2002 4:09:27 PM
**	\par		DESCRIPTION: Please see attributeGenerator.hpp for description.
**					
**					
**	\par		REMARKS: Mating Inclination is generally shorthanded to Incline. Also
**				 all times in this class are fractionaly parts of a minute, not 
**				 seconds as in the rest of the game. This is legacy from the GDD. 
**				 It is very simple to convert fractional seconds to minutes, just 
**				 divide the second value by 60.0f and viola.
**					
*/
////////////////////////////////////////////////////////////////////////////////
#include "attributeGenerator.hpp"
#include "attributes.hpp"
#include <assert.h>
#include <iostream>
#include <math.h>
#include <algorithm>
using namespace std;
using namespace InterpretationModule;
// memory leak detection
#ifdef _DEBUG
    #include"memleak.h"
    #define new DEBUG_NEW
    #undef THIS_FILE
    static char THIS_FILE[] = __FILE__;
#endif


//Below are const's that declare the value of some internally used values.
// Each are declared in attributeGenerator.hpp.
const float InterpretationModule::ENERGY_LOSS_EATING	= 0.0125f;
const float InterpretationModule::ENERGY_LOSS_SUCKING	= 0.0f;
const float InterpretationModule::ENERGY_LOSS_DEFENDING	= 0.0125f;
const float InterpretationModule::ENERGY_LOSS_ATTACKING	= 0.125f;
const float InterpretationModule::ENERGY_LOSS_SWIMMING	= 0.0005f;

const float InterpretationModule::INCLINE_BONUS_MULTIPLIER = 1.3f;
const float InterpretationModule::INCLINE_UNBONUS_MULTIPLIER = 1.3f;
const float InterpretationModule::MATEING_INCLINE_BONUS_THRESHOLD = 0.8f;
const float InterpretationModule::MATEING_INCLINE_UNBONUS_THRESHOLD = 0.1f;


#define ATTRIBUTE_GENERATOR_LOG				//!< Remove this to stop all loggin.


#ifdef ATTRIBUTE_GENERATOR_LOG

	//! Remove this to stop the attribGen class from loggin to console.
#	define ATTRIBUTE_GENERATOR_LOG_CONSOLE		

	//! Remove this to stop the attribGen class from loggin to file
#	define ATTRIBUTE_GENERATOR_LOG_FILE	.
#endif


AttributeGenerator*  AttributeGenerator::m_instance = NULL;


//See header for documentation
AttributeGenerator& AttributeGenerator::Instance(void)
{
	if(m_instance)
		return *m_instance;
	return *(m_instance = new AttributeGenerator);
}

//See header for documentation
void AttributeGenerator::Release(void)
{
	if(m_instance)
	{
		delete m_instance;
		m_instance = NULL;
	}
}


//See header for documentation
AttributeGenerator::AttributeGenerator() throw()
{
	////1. Fill out all the tables for the class.
	//Generate any ReactionTime values. Reverse the order as well.
	GenerateAttributeTable(1.20f, 1.000025f,  0.0166667f, 3.0f, 
		this->m_reactionTable, this->m_reactionLevels);
	std::reverse(this->m_reactionTable.begin(), this->m_reactionTable.end());
	
	//double each of the reaction table values.
	for(unsigned int i = 0 ; i < m_reactionTable.size(); i++)
		m_reactionTable[i] *= 2;



	//Generate any LifeSpan Values. 
	GenerateAttributeTable(1.03f*1.03f, 1.0f, 5.0f, 60.0f,
		this->m_lifeSpanTable, this->m_lifeSpanLevels);

	//Generate any Poundage Values. 
	GenerateAttributeTable(1.07f, 1.0f, 5.0f, 200.0f,
		this->m_poundageTable, this->m_poundageLevels);

	//Generate the inclination tables
	GenerateAttributeTable(1.25f, 1.0f, 0.2f, 20.0f,
		this->m_inclineTable, this->m_inclineLevels);

}


//See header for documentation
AttributeGenerator::~AttributeGenerator()
{
	//No dynamic memory to clean up after.
}



//See header for documentation
void AttributeGenerator::VerifyValues(Attributes& at)throw()
{
	//make sure the base values are capped.
	at.SetMetabolism(NormalizeAttributeValue(at.GetMetabolism()));
	at.SetArmor(NormalizeAttributeValue(at.GetArmor()));
	at.SetEyes(NormalizeAttributeValue(at.GetEyes()));
	at.SetTeeth_Baleen(NormalizeAttributeValue(at.GetTeeth_Baleen()));
	at.SetSize(NormalizeAttributeValue(at.GetSize()));

	//reset Derived values.
	at.SetReactionTime(this->FindReactionTime(at.GetMetabolism()));
	at.SetMaxSpeed(this->FindMaxSpeed(at.GetMetabolism(), at.GetArmor(), at.GetSize()));
	at.SetBecum(this->FindBecum(at.GetMetabolism(), at.GetSize()));
	at.SetLifeSpan(this->FindLifeSpan(at.GetMetabolism(), at.GetSize()));
	at.SetAttackStrength(this->FindAttackStrength(at.GetTeeth_Baleen(), at.GetSize()));
	at.SetSuckability(this->FindSuckablity(at.GetTeeth_Baleen(), at.GetSize()));

	//set reactive limit/modifiers
	at.SetMaxEnergy(this->FindMaxEnergy(at.GetSize()));
	at.SetStomachSize(this->FindStomachSize(at.GetSize()));
	at.SetInclineIncrement(this->FindInclineIncrement(at.GetMetabolism(), at.GetSize()));
	at.SetMaxHealth(this->FindMaxHealth(at.GetSize()));

	//apply those limit/modifiers to the reactive attributes.
	if(at.GetEnergyLevel_Normalized() > 1.0f)
		at.SetEnergyLevel_Normalized(1.0f);
	else if(at.GetEnergyLevel_Normalized() < 0.0f)
		at.SetEnergyLevel_Normalized(0.0f);

	if(at.GetIncline() > 1.0f)
		;//at.SetIncline(1.0f);
	else if(at.GetIncline() < 0.0f)
		at.SetIncline(0.0f);

	if(at.GetStomach_Normalized() > 1.0f)
		at.SetStomach_Normalized(1.0f);
	else if(at.GetStomach_Normalized() < 0.0f)
		at.SetStomach_Normalized(0.0f);

	if(at.GetHealth_Normalized() > 1.0f)
		at.SetHealth_Normalized(1.0f);
	else if(at.GetHealth_Normalized() < 0.0f)
		at.SetHealth_Normalized(0.0f);
	return ;
}//end AttributeGenerator::VerifyValues


//See header for documentation
void AttributeGenerator::Update(Attributes& at, float elapsedTime_minutes) throw()
{
#ifdef ATTRIBUTE_GENERATOR_LOG_CONSOLE
	//do a soft check to see if the elapsedTime_minutes member is set to 
	//seconds. Log it to the console if it seems like seconds. (ie it is too
	//big).
#endif

#ifdef ATTRIBUTE_GENERATOR_LOG_FILE
	//do a soft check to see if the elapsedTime_minutes member is set to 
	//seconds. Log it to the file if it seems like seconds. (ie it is too
	//big).
#endif
	float elapsedTime = elapsedTime_minutes;


	////1. verify the values.
	VerifyValues(at);

	////2. update any data that needs to be done on a frame by frame basis
	//2.1 deduct becum enegry loss. StaticEnergyLoss will compute the 
	//normalized static energy loss.
	float staticLoss = this->StaticEnergyLoss(at.GetBecum(), elapsedTime / 20);

	//convert this value to non-normalized and apply this to the fish.
	staticLoss *= at.GetMaxEnergy();
	at.SetEnergyLevel(at.GetEnergyLevel() - staticLoss);

	//2.2 process the food to energy conversion. Find out the max ammount of food we
	//can take. 
	float maxEnergyGain = this->EnergyFromFood(at.GetMetabolism(), at.GetMaxStomachSize(), elapsedTime);
	float actualEnergyGain = maxEnergyGain;

	//do a bounds check so we don't consume more food then we have.
	if(at.GetStomach() - maxEnergyGain <= 0.0f) 
		actualEnergyGain = at.GetStomach();

	//consume the food
	at.SetStomach(at.GetStomach() - actualEnergyGain);

	//apply the energy gain, cap the increase.
	at.SetEnergyLevel(at.GetEnergyLevel() + actualEnergyGain);	
	if(at.GetEnergyLevel() > at.GetMaxEnergy())
		at.SetEnergyLevel(at.GetMaxEnergy());

	//if our energy is less then our health, set energy to zero and deduct our energy loss
	//from our health.
	if(at.GetEnergyLevel() < 0.0f)
	{
		at.SetHealth(at.GetHealth() + at.GetEnergyLevel());
		at.SetEnergyLevel(0);
	}

	//increment health iff energy isn't zero
	if(0 != at.GetEnergyLevel())
	{
		float healthGain = actualEnergyGain;// + (1-at.GetArmor());
	//	healthGain /= 2;

		at.SetHealth_Normalized(at.GetHealth_Normalized() + elapsedTime*healthGain);
		if(at.GetHealth_Normalized() > 1)
			at.SetHealth_Normalized(1);
	}
	


	//2.3 Apply the mating inclination increase and call the methods that calculate
	//the Incline Bonus & Unbonuses
	float inclineInc = (at.GetInclineIncrement() / 20) *elapsedTime;
	inclineInc = this->ApplyInclineBonus(inclineInc, at.GetEnergyLevel_Normalized());
	inclineInc = this->ApplyInclineUnbonus(inclineInc, at.GetHealth_Normalized());
	at.SetIncline(at.GetIncline() + inclineInc);

	//cap the incline to > 0 && < 1
	if(at.GetIncline() < 0.0f)
		at.SetIncline(0.0f);

	else if(at.GetIncline() > 1) 
		at.SetIncline(1);

	
	return ;
}//end AttributeGenerator::Update







//See header for documentation
//! \note	Input = Metabolism
//!			 This of course means that influence ratios are 
//!				- Meta  100%
Attribute_type AttributeGenerator::FindReactionTime(Attribute_type metabolism)const throw()
{
	assert(metabolism <= 1.0f && metabolism >= 0.0f);
	return FindTableEntry(metabolism, this->m_reactionTable, this->m_reactionLevels);
}//end AttributeGenerator::FindReactionTime


//See header for documentation
//! \note	Equation for generation is ( (Size + Meta) / 2 - Armor    ) * MaxSpeed
//!			 This of course means that influence ratios are 
//!				- Size  25%
//!				- Meta  25%
//!				- Armor 50%
Attribute_type AttributeGenerator::FindMaxSpeed(Attribute_type metabolism, Attribute_type armor, 
												Attribute_type size)const throw()
{
	assert(metabolism <= 1.0f && metabolism >= 0.0f);
	assert(armor <= 1.0f && armor >= 0.0f);
	assert(size <= 1.0f && size >= 0.0f);
 	const float GameWorldMaxSpeed = 120.0f;	

	float input = (size*metabolism*(1-armor));
	input = input*0.60f + (size*0.33f+metabolism*0.33f + (1-armor)*.33f)*0.40f;
		

	//make sure that we never give a negative speed.
	if(input > 0)
		return __max(3, input * GameWorldMaxSpeed);
	return 0;
}//end AttributeGenerator::FindMaxSpeed

/*
 	const float GameWorldMaxSpeed = 120.0f;	

	float input = (size*metabolism*(1-armor));
	input = input*0.60f + (size*0.33f+metabolism*0.33f + (1-armor)*.33f)*0.40f;

  


	const float GameWorldMaxSpeed = 120.0f;	

	float input = sinf(size/2)*sinf(metabolism/2)*cosf(metabolism/2);
	input = sinf(input);

 


	const float GameWorldMaxSpeed = 120.0f;	

	float input = tanf(size)*tanf(metabolism)*atanf(armor);
 */


//See header for documentation
//! \note	Equation for generation is (Size + Meta) / 2)
//!			 This of course means that influence ratios are 
//!				- Size  50%
//!				- Meta  50%
Attribute_type AttributeGenerator::FindBecum(Attribute_type metabolism, Attribute_type size)const throw()
{	
	assert(metabolism <= 1.0f && metabolism >= 0.0f);
	assert(size <= 1.0f && size >= 0.0f);
	return (metabolism + size) / 2.0f;
}//end AttributeGenerator::FindBecum


//See header for documentation
//! \note	input is Size - Meta
//!			 This of course means that influence ratios are 
//!				- Size  50%
//!				- Meta  50%
Attribute_type AttributeGenerator::FindLifeSpan(Attribute_type metabolism, Attribute_type size)const throw()
{
	assert(metabolism <= 1.0f && metabolism >= 0.0f);
	assert(size <= 1.0f && size >= 0.0f);
	Attribute_type input = (size - metabolism);

	//make sure that we don't have input less then zero.
	if(input < 0)
		input = 0;

	return FindTableEntry(input, this->m_lifeSpanTable, this->m_lifeSpanLevels);
}//end AttributeGenerator::FindLifeSpan


//See header for documentation
//! \note	Equation for generation is Size * teeth_baleen
//!			 This of course means that influence ratios are 
//!				- TB    50%
//!				- Meta  50%
Attribute_type AttributeGenerator::FindAttackStrength(Attribute_type teeth_baleen, Attribute_type size)const throw()
{
	assert(teeth_baleen <= 1.0f && teeth_baleen >= 0.0f);
	assert(size <= 1.0f && size >= 0.0f);
	return (teeth_baleen + size)/2;
}//end AttributeGenerator::FindAttackStrength


//See header for documentation
//! \note	Input = (1 - teeth_balen) * size
//!			 This of course means that influence ratios are 
//!				- TB   50%
//!				- Size 50%
Attribute_type AttributeGenerator::FindSuckablity(Attribute_type teeth_baleen, Attribute_type size)const throw()
{
	assert(teeth_baleen <= 1.0f && teeth_baleen >= 0.0f);
	assert(size <= 1.0f && size >= 0.0f);
	return (1 - teeth_baleen) * size;
}//end AttributeGenerator::Suckablity


//See header for documentation
//! \note	Equation for generation is poundage / 2.0f
//!			 This of course means that influence ratios are 
//!				- Size  100%
float AttributeGenerator::FindMaxHealth(Attribute_type size)const throw()
{
	assert(size <= 1.0f && size >= 0.0f);
	return FindTableEntry(size, this->m_poundageTable, this->m_poundageLevels) / 2.0f;
}//end AttributeGenerator::MaxHealth

//See header for documentation
//! \note	Equation for generation is poundage / 2.0f
//!			 This of course means that influence ratios are 
//!				- Size  100%
float AttributeGenerator::FindStomachSize(Attribute_type size)const throw()
{
	assert(size <= 1.0f && size >= 0.0f);
	return FindTableEntry(size, this->m_poundageTable, this->m_poundageLevels) / 2.0f;
}//end AttributeGenerator::StomachSize

//See header for documentation
//! \note	At this time all matiting inclination increments are the same.
//!			 This of course means that influence ratios are 
//!				- Size  50%  (Negatively)
//!				- Meta	50%  (Positively)
float AttributeGenerator::FindInclineIncrement(Attribute_type metabolism, Attribute_type size)const throw()
{
	float ans = static_cast<float>(FindTableEntry(((1 - size)+metabolism)/2, this->m_inclineTable, this->m_inclineLevels));
	return ans ;
}//end AttributeGenerator::FindInclineIncrement

//See header for documentation
//! \note	Equation for generation is poundage / 2.0f
//!			 This of course means that influence ratios are 
//!				- Size  100%
float AttributeGenerator::FindMaxEnergy(Attribute_type size)const throw()
{
	assert(size <= 1.0f && size >= 0.0f);
	return FindTableEntry(size, this->m_poundageTable, this->m_poundageLevels) / 2.0f;
}//end AttributeGenerator::FindMaxEnergy

//See header for documentation
//! \note	Equation for generation is becum * elapsedTime
//!			 This of course means that influence ratios are 
//!				- Size  50%
//!				- Meta  50%
float AttributeGenerator::StaticEnergyLoss(Attribute_type becum, float elapsedTime)const throw()
{
	return static_cast<float>(becum) * elapsedTime;
}

//See header for documentation
//! \note	Equation for generation is const
float AttributeGenerator::EatingEnergyLoss(void)const throw()
{
	return ENERGY_LOSS_EATING;
}

//See header for documentation
//! \note	Equation for generation is const
float AttributeGenerator::SuckingEnergyLoss(void) const throw()
{
	return ENERGY_LOSS_SUCKING;
}

//See header for documentation
//! \note	Equation for generation is const
float AttributeGenerator::DefendingEnergyLoss(void) const throw()
{
	return ENERGY_LOSS_DEFENDING;
}

//See header for documentation
//! \note	Equation for generation is (CurrSpeed^2)*(5e-4) / maxSpeed
float AttributeGenerator::SwimmingEnergyLoss(Attribute_type maxSpeed, float currSpeed, float elapsedTime) const throw()
{
	return currSpeed * currSpeed * (ENERGY_LOSS_SWIMMING) / static_cast<float>(maxSpeed);
}

//See header for documentation
//! \note	Equation for generation is size * 0.125
float AttributeGenerator::AttackingEnergyLoss(Attribute_type size) const throw()
{
	assert(size <= 1.0f && size >= 0.0f);
	return ENERGY_LOSS_ATTACKING * static_cast<float>(size);
}

//See header for documentation
//! \note	Equation for generation is (AttackerStrength - AttackieArmor) * attackieMaxHealth
//!			 This of course means that influence ratios are 
//!				- Size  75ish%
//!				- Armor 25ish%
float AttributeGenerator::DamageDealt(Attribute_type attackerStrength, Attribute_type attackieArmor,
									  Attribute_type attackieMaxHealth) const throw()
{
	float ans = (attackerStrength - attackieArmor);

	if(ans <= 0.05f)
		return 0.1f;

	return attackieMaxHealth * ans;
}

//See header for documentation
//! \note	Equation for generation is poundage * planktonDensity * suckability
//!			 This of course means that influence ratios are 
//!				- Size				50%
//!				- TB				25%
//!				- Plankton Density	25%
float AttributeGenerator::FoodFromSuck(Attribute_type size, Attribute_type suckability, 
									   float PlanktonDensity) const throw()
{
	float ans = static_cast<float>(FindTableEntry(size, this->m_poundageTable, this->m_poundageLevels));
	return PlanktonDensity * suckability;
}

//See header for documentation
//! \note	Equation for generation is meta * FoodAmmount * elapsedTime
//!			 This of course means that influence ratios are 
//!				- meta  50%
//!				- Food	50%
float AttributeGenerator::EnergyFromFood(Attribute_type meta, float FoodAmmount, float elapsedTime) const throw()
{
	return meta * FoodAmmount * elapsedTime;
}

//See header for documentation
//! \note	Equation for generation is (energy% > MATEING_INCLINE_BONUS_THRESHOLD)?incline = incline^1.3f:incline = incline
//!			 This of course means that influence ratios are 
//!				- energyLevel 100%
float AttributeGenerator::ApplyInclineBonus(float inclineInc, float energy_normalized) const throw()
{
	if(energy_normalized > MATEING_INCLINE_BONUS_THRESHOLD) 
		return inclineInc + static_cast<float>(pow(inclineInc, INCLINE_BONUS_MULTIPLIER));
	return inclineInc;
}

//See header for documentation
//! \note	Equation for generation is (health% < MATEING_INCLINE_UNBONUS_THRESHOLD)?incline = 0 - incline^1.3f:incline = incline
//!			 This of course means that influence ratios are 
//!				- energyLevel 100%
float AttributeGenerator::ApplyInclineUnbonus(float inclineInc, float health_normalized) const throw()
{
	if(health_normalized < MATEING_INCLINE_UNBONUS_THRESHOLD) 
		return inclineInc - static_cast<float>(pow(inclineInc, INCLINE_UNBONUS_MULTIPLIER));
	return inclineInc;
}