////////////////////////////////////////////////////////////////////////////////
/*!	\file		attributeGenerator.hpp
**	\brief		Declares the class AttributeGenerator.
**	\author		Nate Cleveland
**	\date		10/16/2002 3:10:23 PM
**	\par		DESCRIPTION: The AttributeGenerator enforces the rule system for
**				 the Fish Attributes. Each method's documentation should be read
**				 closely before altering any part of this class. Below is a summary
**				 of the functionaity provided the AttributeGenerator class.
**					- Finding any Derived Attribute based on its related Base
**					Attributes.
**					- Boundry enforcement of all values (except MatingInclination,
**					at this time it is allowed to go above 1.0f).
**					- Methods for calculating energy loss, damage dealt, food and
**					energy consumption, etc.
**
**
**	\par		REMARKS: Mating Inclination is generally shorthanded to Incline. Also
**				 all times in this class are fractionaly parts of a minute, not
**				 seconds as in the rest of the game. This is legacy from the GDD.
**				 It is very simple to convert fractional seconds to minutes, just
**				 divide the second value by 60.0f and viola.
**
*/
////////////////////////////////////////////////////////////////////////////////
#pragma once
#include "attributeUtils.hpp"
#include <vector>

namespace InterpretationModule
{
	class Attributes;

	//Below are const's that define the value of some internally used values.
	// Each are defined in attributeGenerator.cpp.
	extern const float ENERGY_LOSS_EATING;
	extern const float ENERGY_LOSS_SUCKING;
	extern const float ENERGY_LOSS_DEFENDING;
	extern const float ENERGY_LOSS_ATTACKING;
	extern const float ENERGY_LOSS_SWIMMING;

	extern const float INCLINE_BONUS_MULTIPLIER;
	extern const float INCLINE_UNBONUS_MULTIPLIER;
	extern const float MATEING_INCLINE_BONUS_THRESHOLD;
	extern const float MATEING_INCLINE_UNBONUS_THRESHOLD;


	////////////////////////////////////////////////////////////////////////////////
	/*!	\class		AttributeGenerator
	**	\brief		Enforces the rule system for the Fish Attributes.
	**	\author		Nate Cleveland
	**	\date		10/16/2002 4:05:43 PM
	**	\warning	All time values in this class are fractionaly parts of minutes,
	**				 not seconds as in the rest of the game. This is legacy from the
	**				 GDD.
	*/
	////////////////////////////////////////////////////////////////////////////////
	class AttributeGenerator
	{
		std::vector<Attribute_type>	m_reactionTable;	//!< holds all possible reaction values.
		int							m_reactionLevels;	//!< # of levels for the react table

		std::vector<Attribute_type>	m_lifeSpanTable;	//!< holds all possible lifeSpan values.
		int							m_lifeSpanLevels;	//!< # of levels for the lifeSpan table

		std::vector<Attribute_type>	m_poundageTable;	//!< holds all possible poundage values.
		int							m_poundageLevels;	//!< # of levels for the poundage table

		std::vector<Attribute_type>	m_inclineTable;		//!< holds all possible incline values.
		int							m_inclineLevels;	//!< # of levels for the incline table

		static AttributeGenerator*  m_instance;

		////////////////////////////////////////////////////////////////////////////////
		//	FUNCTION:
		/*!	\brief		Defualt ctor.
		**	\param		none
		**	\exception	none
		**	\par		DESCRIPTION: Fills out all the tables presented in the GDD.
		**
		**	\author		Nate Cleveland
		**	\date		10/16/2002 3:56:11 PM
		**	\warning	none
		*/
		////////////////////////////////////////////////////////////////////////////////
		AttributeGenerator() throw();

		////////////////////////////////////////////////////////////////////////////////
		//	FUNCTION:
		/*!	\brief		defualt dtor
		**	\exception	none
		**	\par		DESCRIPTION: Does nothing at this time. There is no dynamic
		**				 memory to clean.
		**
		**	\author		Nate Cleveland
		**	\date		10/16/2002 3:57:35 PM
		**	\warning	none
		*/
		////////////////////////////////////////////////////////////////////////////////
		~AttributeGenerator();
	public:

		static AttributeGenerator& Instance(void);
		static void Release(void);



		////////////////////////////////////////////////////////////////////////////////
		//	FUNCTION:
		/*!	\brief		Updates the passed in Attributes class.
		**
		**	\param Attributes& at		  	  Attribute class to update.
		**	\param float elapsedTime_minutes	  How long, in <b> MINUTES </b> since
		**										   last update call on this Attribute class.
		**
		**	\return		none
		**	\exception	none
		**	\par		Calculates the following for the attribute class:
		**					- Value validation (Calls VerifyValues()).
		**					- Static (BECUM) energy loss.
		**					- Stomach to energy conversion.
		**					- Increases Mating Inclination (bonus and unbonuses included).
		**
		**	\par		Does NOT calculate the following:
		**					- Energy lost from swimming or attacking or defending.
		**
		**
		**
		**	\author		Nate Cleveland
		**	\date		10/16/2002 3:20:16 PM
		**	\warning	<b> The second parameter takes fractional parts of MINUTES NOT
		**				 SECONDS.</b>
		*/
		////////////////////////////////////////////////////////////////////////////////
		void Update(Attributes& , float elapsedTime_minutes) throw();

		////////////////////////////////////////////////////////////////////////////////
		//	FUNCTION:
		/*!	\brief		Verify the given set of attributes.
		**
		**	\param Attributes& at		Attribute class to have its values verified.
		**
		**	\return		none
		**	\exception	none
		**	\par		DESCRIPTION: This method analizes all attributes and forces them
		**				 into legal values.
		**					- Base Attributes are forced to be 0.0f <= x <= 1.0f.
		**					- Derived Attributes are recalculated from the Base Attributes.
		**					- Reactive limiter/modifiers are also recalculated.
		**					- Reactive Attributes are capped to their legal values.
		**
		**	\author		Nate Cleveland
		**	\date		10/16/2002 3:34:50 PM
		**	\warning	none
		*/
		////////////////////////////////////////////////////////////////////////////////
		void VerifyValues(Attributes& ) throw();



		////Find Value methods to numerous and simple to doxygenize. Each implementation
		//// contains a quick outline of the formula used and influence ratios.

		//Derived Attributes.
		Attribute_type FindReactionTime(Attribute_type metabolism)const throw();
		Attribute_type FindMaxSpeed(Attribute_type metabolism, Attribute_type armor,
									Attribute_type size)const throw();
		Attribute_type FindBecum(Attribute_type metabolism, Attribute_type size)const throw();
		Attribute_type FindLifeSpan(Attribute_type metabolism, Attribute_type size)const throw();
		Attribute_type FindAttackStrength(Attribute_type teeth_baleen, Attribute_type size)const throw();
		Attribute_type FindSuckablity(Attribute_type teeth_baleen, Attribute_type size)const throw();

		//reactive limiter/modifiers.
		float FindMaxHealth(Attribute_type size)const throw();
		float FindStomachSize(Attribute_type size)const throw();
		float FindInclineIncrement(Attribute_type metabolism, Attribute_type size)const throw();
		float FindMaxEnergy(Attribute_type size)const throw();


		//Energy Loss Methods.
		float StaticEnergyLoss(Attribute_type becum, float elapsedTime_minutes)const throw();
		float EatingEnergyLoss(void)const throw();
		float SuckingEnergyLoss(void) const throw();
		float DefendingEnergyLoss(void) const throw();
		float AttackingEnergyLoss(Attribute_type size) const throw();
		float SwimmingEnergyLoss(Attribute_type maxSpeed, float currSpeed, float elapsedTime) const throw();

		//Other fish formulae
		float DamageDealt(Attribute_type attackerStrength, Attribute_type attackieArmor,
						  Attribute_type attackieMaxHealth) const throw();
		float FoodFromSuck(Attribute_type size, Attribute_type suckability, float PlanktonDensity) const throw();
		float EnergyFromFood(Attribute_type meta, float FoodAmmount, float elapsedTime) const throw();
		float ApplyInclineBonus(float inclineInc, float energy_normalized) const throw();
		float ApplyInclineUnbonus(float inclineInc, float health_normalized) const throw();

	};//end class AttributeGenerator
}//end namespace In