////////////////////////////////////////////////////////////////////////////////
/*!	\file		attribute.hpp
**	\brief		Holds all of a Fish's attributes.
**	\author		Nate Cleveland
**	\date		10/16/2002 4:17:31 PM
**	\par		DESCRIPTION: Class that holds all the Attributes for a fish. Most
**				 values are kept normalized, ie from 0.0f to 1.0f, but some are not.
**				 The following are kept normalized.
**					- Metabolism
**					- Armor
**					- Eyes
**					- Teeth/Baleen
**					- Size
**					- Reaction Time
**					- Max Speed
**					- BECUM
**					- Life Span
**					- Attack Strength
**					- Suckability
**					- Mating Inclination
**					- Mating Inclination Increment
**
**	\par		The Following are NOT kept normalized
**					- Max Stomach Size
**					- Stomach
**					- Max Energy
**					- Energy
**					- Max Health
**					- Health.
**
**	\par		All of the non-normalized Reactive Attributes have two accessors,
**				 one normalized, the other not normalized. The normalized values
**				 returned are local only to this Fish. So FishA's health and FishB's
**				 normalized health values could be the same even if FishA's
**				 non-normalized health is actually much larger.
**
**	\par		REMARKS: This class does not interpet any of this data, it simply
**				 holds it. The AttributeGenerator class modifies each attribute
**				 and make's sure it keeps legal values.
**
**	\par		If this class becomes too slow it could be speed up by doing the following
**					- Have all the "Set" functions set a "HasChanged" flag.
**					- Everytime the AttributeGenerator class upates this object
**					 upates it checks the flag and only verifies this object if
**					 the flag is set to true.
**					- After every update this object would reset its HasChanged flag
**					 to false.
**
**	\par		This is not done at this time since it seems like it would take
**				 more time to implement then it would save.
**
*/
////////////////////////////////////////////////////////////////////////////////
#pragma once
#include "attributeUtils.hpp"
#include "../FileIO/memento.h"

//forward declarations.
namespace UDPNet 
{
	class NetMessage;
};


namespace InterpretationModule
{
	class AttributeGenerator;
	const Attribute_type FISH_SCALER_OFFSET = 0.5f;


	////////////////////////////////////////////////////////////////////////////////
	/*!	\class		Attributes
	**	\brief		Holds all of an individual fish's attributes.
	**	\author		Nate Cleveland
	**	\date		10/16/2002 4:32:37 PM
	**	\warning	Unlike the AttributeGenerator class all times in this class are
	*				 kept in seconds.
	*/
	////////////////////////////////////////////////////////////////////////////////
	class Attributes: public FileMemento
	{

	protected:
		//Base Attributes, all base attributes are normalized.
		Attribute_type m_metabolism;
		Attribute_type m_armor;
		Attribute_type m_eyes;
		Attribute_type m_teeth_baleen;
		Attribute_type m_size;

		//Derived Attributes, all base attributes are normalized.
		Attribute_type m_reactionTime;
		Attribute_type m_maxSpeed;
		Attribute_type m_becum;
		Attribute_type m_lifeSpan;
		Attribute_type m_attackStrength;
		Attribute_type m_suckability;

		//Reactive Attribute Modifier/Limiters. Not all kept normalized.
		float			m_maxStomachSize;			//!< NOT a normalized value.
		float			m_maxEnergy;				//!< NOT a normalized value.
		float			m_maxHealth;				//!< NOT a normalized value.
		float		m_matingInclinationIncrement;	//!< this IS a normalized value.

		//Reactive Attributes. Not all kept normalized.
		Attribute_type	m_energyLevel;				//!< NOT a normalized value.
		Attribute_type	m_stomach;					//!< NOT a normalized value.
		Attribute_type	m_health;					//!< NOT a normalized value.
		Attribute_type	m_matingInclination;		//!< This IS a normalized value.

	public:


		////////////////////////////////////////////////////////////////////////////////
		//	FUNCTION:
		/*!	\brief		Instantitates an attribute class from mutt values defined by the
		**				 GDD.
		**
		**	\param		none
		**	\exception	none
		**	\par		DESCRIPTION:
		**
		**	\author		Nate Cleveland
		**	\date		10/16/2002 4:40:29 PM
		**	\warning	none
		*/
		////////////////////////////////////////////////////////////////////////////////
		Attributes(void) throw();

		//!takes the Base values from the user and use them to generatate the rest.
		////////////////////////////////////////////////////////////////////////////////
		//	FUNCTION:
		/*!	\brief		Instantitates an attribute class with the passed in values.
		**	\param		straight forward.
		**	\exception	none
		**	\par		DESCRIPTION:
		**
		**	\author		Nate Cleveland
		**	\date		10/16/2002 4:40:29 PM
		**	\warning	none
		*/
		////////////////////////////////////////////////////////////////////////////////
		Attributes(Attribute_type metabolism, Attribute_type armor, Attribute_type eyes,
					Attribute_type teeth_baleen, Attribute_type size) throw();

		////////////////////////////////////////////////////////////////////////////////
		//	FUNCTION:
		/*!	\brief		defualt dtor.
		**	\exception	none
		**	\par		DESCRIPTION: Has nothing to do at this time.
		**
		**	\author		Nate Cleveland
		**	\date		10/16/2002 4:38:20 PM
		**	\warning	none
		*/
		////////////////////////////////////////////////////////////////////////////////
		~Attributes(void) throw ();


		////////////////////////////////////////////////////////////////////////////////
		//	FUNCTION:
		/*!	\brief		Tells this class to verify its values.
		**	\param		none
		**	\return		none
		**	\exception	none
		**	\par		DESCRIPTION: Passes itself to its internal and static
		**				 AttributeGenerator class. If any values are illegal they will be
		**				 changed.
		**
		**	\author		Nate Cleveland
		**	\date		10/16/2002 4:54:57 PM
		**	\warning	none
		*/
		////////////////////////////////////////////////////////////////////////////////
		void VerifyValues(void) throw();

		////////////////////////////////////////////////////////////////////////////////
		//	FUNCTION:
		/*!	\brief		Tells this class to update its values.
		**
		**	\param		elapsedTime_seconds		Time since last update.

		**	\return		none
		**	\exception	none
		**	\par		DESCRIPTION: Passes itself to its internal and static
		**				 AttributeGenerator class. If any values are illegal they will be
		**				 changed before the class is updated.
		**
		**	\author		Nate Cleveland
		**	\date		10/16/2002 4:54:57 PM
		**	\warning	none
		*/
		////////////////////////////////////////////////////////////////////////////////
		void Update(float elapsedTime_seconds) throw();
		

		////////////////////////////////////////////////////////////////////////////////
		//	FUNCTION:	
		/*!	\brief		Returns to the caller a reference to the AttributeGenerator class.
		**	\param		none
		**	\return		Reference to attribute generator class.
		**	\exception	none
		**	\par		DESCRIPTION:
		**					
		**	\author		Nate Cleveland
		**	\date		11/4/2002 9:53:14 AM
		**	\warning	none
		*/
		////////////////////////////////////////////////////////////////////////////////
		static AttributeGenerator& GetGenerator(void) throw();
		

		////////////////////////////////////////////////////////////////////////////////
		//	FUNCTION:	
		/*!	\brief		Returns to the caller a value they can use to scale a fish's
		**				 dimensions.
		**	\param		none
		**	\return		none
		**	\exception	none
		**	\par		DESCRIPTION:
		**					
		**	\author		Nate Cleveland
		**	\date		11/25/2002 10:09:36 AM
		**	\warning	none
		*/
		////////////////////////////////////////////////////////////////////////////////
		Attribute_type GetFishScaler(void){return (this->m_size + FISH_SCALER_OFFSET)*10;}



		////Accessers.
		//Base
		Attribute_type GetMetabolism(void) const throw(){return m_metabolism;}
		Attribute_type GetArmor(void) const throw(){return m_armor;}
		Attribute_type GetEyes(void) const throw(){return m_eyes;}
		Attribute_type GetTeeth_Baleen(void) const throw(){return m_teeth_baleen;}
		Attribute_type GetSize(void) const throw(){return m_size;}

		//Derived
		Attribute_type GetReactionTime(void) const throw();
		Attribute_type GetMaxSpeed(void) const throw(){return m_maxSpeed + 1;}
		Attribute_type GetBecum(void) const throw(){return m_becum;}
		Attribute_type GetLifeSpan(void) const throw(){return m_lifeSpan;}
		Attribute_type GetAttackStrength(void) const throw(){return m_attackStrength;}
		Attribute_type GetSuckability(void) const throw(){return m_suckability;}

		//Reactive Modifier/Limiter
		float GetMaxStomachSize(void)const throw(){return m_maxStomachSize;}
		float GetMaxEnergy(void) const throw(){return m_maxEnergy;}
		float GetInclineIncrement(void) const throw(){return m_matingInclinationIncrement;}
		float GetMaxHealth(void) const throw(){return m_maxHealth;}

		//Reactive
		Attribute_type GetEnergyLevel(void) const throw(){return m_energyLevel;}
		Attribute_type GetStomach(void) const throw(){return m_stomach;}
		Attribute_type GetIncline(void) const throw(){return m_matingInclination;}
		Attribute_type GetHealth(void) const throw(){return m_health;}

		Attribute_type GetEnergyLevel_Normalized(void) const throw(){return m_energyLevel / m_maxEnergy;}
		Attribute_type GetStomach_Normalized(void) const throw(){return m_stomach / m_maxStomachSize ;}
		Attribute_type GetHealth_Normalized(void) const throw(){return m_health / m_maxHealth;}

		//// Set Value Methods
		//Base
		void SetMetabolism(Attribute_type newVal) throw() {m_metabolism = newVal;}
		void SetArmor(Attribute_type newVal) throw() {m_armor = newVal;}
		void SetEyes(Attribute_type newVal) throw() {m_eyes = newVal;}
		void SetTeeth_Baleen(Attribute_type newVal) throw() {m_teeth_baleen = newVal;}
		void SetSize(Attribute_type newVal) throw() {m_size = newVal;}

		//Derived
		void SetReactionTime(Attribute_type newVal) throw() {m_reactionTime = newVal;}
		void SetMaxSpeed(Attribute_type newVal) throw() {m_maxSpeed = newVal;}
		void SetBecum(Attribute_type newVal) throw() {m_becum = newVal;}
		void SetLifeSpan(Attribute_type newVal) throw() {m_lifeSpan = newVal;}
		void SetAttackStrength(Attribute_type newVal) throw() {m_attackStrength = newVal;}
		void SetSuckability(Attribute_type newVal) throw() {m_suckability = newVal;}

		//Reactive Modifier/Limiter
		void SetStomachSize(float newVal) throw(){m_maxStomachSize = newVal;}
		void SetMaxEnergy(float newVal) throw() {m_maxEnergy = newVal;}
		void SetInclineIncrement(float newVal) throw() {m_matingInclinationIncrement = newVal;}
		void SetMaxHealth(float newVal) throw() {m_maxHealth = newVal;}

		//Reactive
		void SetEnergyLevel(Attribute_type newVal) throw() {m_energyLevel = newVal;}
		void SetStomach(Attribute_type newVal) throw() {m_stomach = newVal;}
		void SetIncline(Attribute_type newVal) throw() {m_matingInclination = newVal;}
		void SetHealth(Attribute_type newVal) throw() {m_health = newVal;}

		void SetEnergyLevel_Normalized(Attribute_type newVal) throw(){m_energyLevel = newVal * m_maxEnergy;}
		void SetStomach_Normalized(Attribute_type newVal) throw(){m_stomach = newVal * m_maxStomachSize ;}
		void SetHealth_Normalized(Attribute_type newVal) throw(){m_health = newVal * m_maxHealth;}


		////memento classes
		////////////////////////////////////////////////////////////////////////////////
		//	FUNCTION:
		/*!	\brief		Writes this class to the passed in stream
		**
		**	\param ostream&						Stream to write to.
		**
		**	\return		none
		**	\exception	none
		**	\par		DESCRIPTION: Simply calls the overloaded << operator.
		**
		**	\author		Nate Cleveland
		**	\date		10/16/2002 6:30:49 PM
		**	\warning	none
		*/
		////////////////////////////////////////////////////////////////////////////////
		void Write(std::ostream&) const;

		////////////////////////////////////////////////////////////////////////////////
		//	FUNCTION:
		/*!	\brief		Reads data from the stream.
		**	\param istream&					Stream to read from.
		**	\return		none
		**	\exception	none
		**	\par		DESCRIPTION: Simply calls the overloaded >> operator.
		**
		**	\author		Nate Cleveland
		**	\date		10/16/2002 6:30:55 PM
		**	\warning	none
		*/
		////////////////////////////////////////////////////////////////////////////////
		void Read(std::istream&);
		////////////////////////////////////////////////////////////////////////////////
		//	FUNCTION:	
		/*!	\brief		Reads from a netMessage and replaces itself with the data it 
						 finds there.
		**	\param		netMessage		obj to read from.
		**	\return		none
		**	\exception	Throws an exception if there is an error in the packet or the 
		**				 keys don't match.				
		**	\author		Nate Cleveland
		**	\date		2/6/2003 2:32:53 PM
		**	\warning	none
		*/
		////////////////////////////////////////////////////////////////////////////////
		void operator << (UDPNet::NetMessage& netMessage);


		////////////////////////////////////////////////////////////////////////////////
		//	FUNCTION:	
		/*!	\brief		Fills a NetMessage with all the data for these attributes.
		**	\param		netMessage		obj to write to.
		**	\return		none
		**	\exception	Throws an exception iff there is an error with netMessage.
		**	\author		Nate Cleveland
		**	\date		2/6/2003 2:29:45 PM
		**	\warning	none
		*/
		////////////////////////////////////////////////////////////////////////////////
		void operator >> (UDPNet::NetMessage& netMessage);


		////////////////////////////////////////////////////////////////////////////////
		//	FUNCTION:
		/*!	\brief		Writes all needed data to the stream
		**
		**	\param ostream&				object we'll stream.
		**	\param const  Attributes&	Reference to this class.
		**
		**	\return		A reference to the stream.
		**	\exception	none
		**	\par		DESCRIPTION: Writes all the needed data (Base and Reactive
		**				 Attributes) to the stream.
		**
		**	\author		Nate Cleveland
		**	\date		10/16/2002 6:30:59 PM
		**	\warning	none
		*/
		////////////////////////////////////////////////////////////////////////////////
		friend std::ostream& operator << (std::ostream&, const Attributes &);

		////////////////////////////////////////////////////////////////////////////////
		//	FUNCTION:
		/*!	\brief		Reads in data from the stream.
		**
		**	\param ostream&				object we'll stream from.
		**	\param const  Attributes&	Reference to this class.
		**
		**	\return		A reference to the stream.
		**	\exception	none
		**	\par		DESCRIPTION: Reads data from the stream and resets this class
		**				 accordingly.
		**
		**	\author		Nate Cleveland
		**	\date		10/16/2002 6:31:04 PM
		**	\warning	none
		*/
		////////////////////////////////////////////////////////////////////////////////
		friend std::istream& operator>> (std::istream&, Attributes &);



	};//end class FishAttributes
}//end namespace InterpretationModule