////////////////////////////////////////////////////////////////////////////////
/*!	\file		chum.hpp
**	\brief		Declares the chum object that fish may eat.
**	\author		Nate Cleveland
**	\date		11/4/2002 11:35:18 AM
**	\par		DESCRIPTION:
**				The chum object is what is generated when a fish dies or is attacked.
**
**	\par		REMARKS:
**				Once the chum ammount reaches zero it should be removed from the game.
**					
*/
////////////////////////////////////////////////////////////////////////////////

#pragma once
#include "fishBodyUtils.hpp"
#include <assert.h>


namespace InterpretationModule
{

	////////////////////////////////////////////////////////////////////////////////
	/*!	\class		Chum
	**	\brief		Object generated when a fish is attacked or killed.
	**	\author		Nate Cleveland
	**	\date		11/4/2002 11:37:03 AM
	**	\warning	none
	*/
	////////////////////////////////////////////////////////////////////////////////
	class Chum
	{
	protected:
		AquariusKey_type	m_key			//!< ID of this chunk-a-chum.
		float				m_chumAmmount;	//!< Ammount of chum in the water.
		InterpPnt_type		m_pos;			//!< Position of this chunk o chum.

	public:
		////////////////////////////////////////////////////////////////////////////////
		//	FUNCTION:	
		/*!	\brief		Ctor for chunk of chum. Defualts everything to zero.
		**	\param		none
		**	\return		none
		**	\exception	none
		**	\par		DESCRIPTION:
		**				Generates a chunk of chum with all zero for its value.
		**	\author		Nate Cleveland
		**	\date		11/4/2002 11:01:42 AM
		**	\warning	none
		*/
		////////////////////////////////////////////////////////////////////////////////
		Chum():m_key(0), m_chumAmmount(0), m_pos(0,0,0){}

		////////////////////////////////////////////////////////////////////////////////
		//	FUNCTION:	
		/*!	\brief		Ctor for a chunk of chum.
		**	\param		key			ID of this piece of chum.
		**	\param		chumAmmount	How much chum is in this chum.
		**	\param		pos			position of this chunk of chum.
		**	\return		none
		**	\exception	none
		**	\par		DESCRIPTION:
		**				Generates a chunk of chum and sets the values.	
		**	\author		Nate Cleveland
		**	\date		11/4/2002 11:08:53 AM
		**	\warning	none
		*/
		////////////////////////////////////////////////////////////////////////////////
		Chum(AquariusKey_type key, float chumAmmount, const InterpPnt_type& pos)
			:m_key(key), m_chumAmmount(chumAmmount), m_pos(pos){}

		////////////////////////////////////////////////////////////////////////////////
		//	FUNCTION:	
		/*!	\brief		Tries to consume the given ammount from this chunk of chum.
		**	\param		desiredAmmount		How much chum this fish is trying to eat.
		**	\return		none
		**	\exception	none
		**	\par		DESCRIPTION:
		**				Passed in value must be greater then 0. A negative value will 
		**				 cause an internal assertion to fail. Method ensure that the 
		**				 m_chumAmmount member never goes less then zero.
		**
		**	\author		Nate Cleveland
		**	\date		11/4/2002 11:19:44 AM
		**	\warning	none
		*/
		////////////////////////////////////////////////////////////////////////////////
		float Consume(float desiredAmmount) throw()
		{
			float consumed = min(desiredAmmount, m_chumAmmount);//find smaller
			m_chumAmmount -= consumed;							//decrement chum
			assert(desiredAmmount > 0 && m_chumAmmount > 0);	//assert that all is well.
			return consumed;									//return chum eaten.
		}



		//Accessor functions
		AquariusKey_type GetKey(void)				 const	throw(){return m_key;}
		float			 GetChumAmmount(void)	  	 const	throw(){return m_chumAmmount;}
		void			 GetPos(InterpPnt_type& pnt) const	throw(){pnt = m_pos;}
		float			 GetXPos(void)				 const	throw(){return m_pos[0];}
		float			 GetYPos(void)				 const	throw(){return m_pos[1];}
		float			 GetZPos(void)				 const	throw(){return m_pos[2];}

		AquariusKey_type SetKey(void)						throw(){m_key = val;}
		float			 SetChumAmmount(float val)			throw(){m_chumAmmount = val;}
		void			 SetPos(const InterpPnt_type& pnt)	throw(){m_pos = pnt;}
		float			 SetXPos(float val)					throw(){m_pos[0] = val;}
		float			 SetYPos(float val)					throw(){m_pos[1] = val;}
		float			 SetZPos(float val)					throw(){m_pos[2] = val;}
	};//end Chum
}//end InterpretationModule