////////////////////////////////////////////////////////////////////////////////
/*!	\file		commandPacket.hpp
**	\brief		Declares the CommandPacket sent by the brain to the body.
**	\author		Nate Cleveland
**	\date		10/21/2002 9:46:16 AM
**	\par		DESCRIPTION: The CommandPacket is an encapsulation of all commands
**				 sent by the brain to the body. The file CommandID.hpp contains a
**				 list of all possible commands with descriptions of each. All the
**				 brain needs to be concered about is filling out one of the several
**				 ctors for the CommandPacket class and sending it to the message
**				 passing system.
**
**	\par		REMARKS:
**
*/
////////////////////////////////////////////////////////////////////////////////
#pragma once
#include "CommandID.hpp"
#include "fishBodyUtils.hpp"
#include <memory.h>
#include <iostream>

namespace UDPNet 
{
	class NetMessage;
};

namespace InterpretationModule
{

	////////////////////////////////////////////////////////////////////////////////
	/*!	\class		struct ActionTarget
	**	\brief		Abstraction for the CommandPacket's target concept.
	**	\author		Nate Cleveland
	**	\date		10/21/2002 9:41:40 AM
	**	\warning	none
	*/
	////////////////////////////////////////////////////////////////////////////////
	struct ActionTarget
	{
		//! Used as a flag in the ActionTarget Union.
		enum ActionTargetFlag
		{
			ACTION_TARGET_FLAG_NONE,	//!< Target must be infered from the command.
			ACTION_TARGET_FLAG_KEY,		//!< Sets the target to a object with matching key
			ACTION_TARGET_FLAG_VOXEL,	//!< Sets the target to a specified voxel
			ACTION_TARGET_FLAG_POINT	//!< Sets the target to a specified point.
		} m_ActionFlag;//end ActionTargetFlag


		//! Memory saving union that allows the CommandPacket to abstract targets simply.
		union ActionTargetUnion
		{
			AquariusKey_type	m_key;		//!< ID or Key of the object to head for.
			Voxel_type			m_voxel;	//!< Voxel to set as the target of an action.
			InternalPoint_type	m_point;	//!< 3D Point to set as our current point.
		} m_ActionTarget;//end ActionTarget

		////////////////////////////////////////////////////////////////////////////////
		//	FUNCTION:
		/*!	\brief		Defualt ctor for ActionTarget struct.
		**	\param		NONE
		**	\exception	none
		**	\par		DESCRIPTION: Zeros out all members.
		**
		**	\author		Nate Cleveland
		**	\date		10/21/2002 9:42:23 AM
		**	\warning	none
		*/
		////////////////////////////////////////////////////////////////////////////////
		ActionTarget() throw() {memset(this, 0, sizeof(*this));}


		////////////////////////////////////////////////////////////////////////////////
		//	FUNCTION:	
		/*!	\brief		Reads data in from a netMessage.
		**	\param		netMessage		stream to read from
		**	\return		none
		**	\exception	iff there is an error in the stream or if target flag is unknown
		**	\author		Nate Cleveland
		**	\date		2/7/2003 9:58:59 AM
		**	\warning	none
		*/
		////////////////////////////////////////////////////////////////////////////////
		void operator << (UDPNet::NetMessage& netMessage);

		////////////////////////////////////////////////////////////////////////////////
		//	FUNCTION:	
		/*!	\brief		Writes its data to a netMessage as compact as possible
		**	\param		netMessage		stream to write to
		**	\exception	iff there is an error in the stream or if target flag is unknown
		**	\author		Nate Cleveland
		**	\date		2/7/2003 9:59:02 AM
		**	\warning	none
		*/
		////////////////////////////////////////////////////////////////////////////////
		void operator >> (UDPNet::NetMessage& netMessage);



		////////////////////////////////////////////////////////////////////////////////
		//	FUNCTION:
		/*!	\brief		Equality operator for this class.
		**	\param		Reference to another ActionTarget struct.
		**	\return		true == two structs are identical.
		**	\exception	none
		**	\par		DESCRIPTION:
		**
		**	\author		Nate Cleveland
		**	\date		10/21/2002 9:43:50 AM
		**	\warning	none
		*/
		////////////////////////////////////////////////////////////////////////////////
		bool operator == (const ActionTarget& rhs)const throw()
		{
			return (m_ActionFlag == rhs.m_ActionFlag) && (0 == memcmp(this, &rhs, sizeof(rhs)));
		}

		////////////////////////////////////////////////////////////////////////////////
		//	FUNCTION:
		/*!	\brief		Non-equality operator for this class.
		**	\param		Reference to another ActionTarget struct.
		**	\return		true == two structs are not identical.
		**	\exception	none
		**	\par		DESCRIPTION:
		**
		**	\author		Nate Cleveland
		**	\date		10/21/2002 9:43:50 AM
		**	\warning	none
		*/
		////////////////////////////////////////////////////////////////////////////////
		bool operator != (const ActionTarget& rhs)const throw()
		{
			return !(*this == rhs);
		}
	};///end ActionTarget



	////////////////////////////////////////////////////////////////////////////////
	/*!	\class		CommandPacket
	**	\brief		Holds all the data needed to send a body a command.
	**	\author		Nate Cleveland
	**	\date		10/18/2002 1:07:08 PM
	**	\warning	none
	*/
	////////////////////////////////////////////////////////////////////////////////
	class CommandPacket
	{
	protected:
		CommandID_type		m_commandID;	//!< ID of this particular command
		AquariusKey_type	m_bodyID;		//!< Reciever's ID
		AquariusKey_type	m_senderID;		//!< Sender's ID
		ActionTarget		m_target;		//!< Target of the command
		float				m_weight;		//!< Weight set by the brain for this cmd.
		unsigned int		m_decisionID;	//!< ID given to this packet by the brain.
		float				m_fitness;		//!< Value given to this packet by the brain.

	public:
		CommandPacket():m_commandID(0), m_bodyID(0), m_senderID(0), m_weight(0), 
						m_decisionID(0), m_fitness(0){}

		////////////////////////////////////////////////////////////////////////////////
		//	FUNCTION:
		/*!	\brief		Ctor to use iff the target is an object
		**
		**	\param CommandID_type				See CommandID.hpp.
		**	\param AquariusKey_type SenderID		ID of the brain sending this command..
		**	\param AquariusKey_type BodyID,		ID of the body to recieve this command.
		**	\param float weight					Weight brain wishes to give this cmd.
		**	\param AquariusKey_type targetKey	ID or Key of the target of this cmd.
		**
		**	\exception	none
		**	\par		DESCRIPTION: Fills out the internal struct for iff the target is
		**				 an object such as a fish, rock, or coral.
		**
		**	\author		Nate Cleveland
		**	\date		10/21/2002 9:50:17 AM
		**	\warning	none
		*/
		////////////////////////////////////////////////////////////////////////////////
		CommandPacket(CommandID_type , AquariusKey_type SenderID, AquariusKey_type BodyID,
					  float weight, unsigned int decisionID, float fitness, 
					  AquariusKey_type targetKey)throw();

		////////////////////////////////////////////////////////////////////////////////
		//	FUNCTION:
		/*!	\brief		Ctor to use iff the target is point.
		**
		**	\param CommandID_type				See CommandID.hpp.
		**	\param AquariusKey_type SenderID		ID of the brain sending this command..
		**	\param AquariusKey_type BodyID,		ID of the body to recieve this command.
		**	\param float weight					Weight brain wishes to give this cmd.
		**	\param const InterpPoint_type& point	Reference to the point in the game
		**											 that is to be the target of this cmd.
		**
		**	\exception	none
		**	\par		DESCRIPTION: Fills out the internal struct for iff the target is
		**				 a point in the game space.
		**
		**	\author		Nate Cleveland
		**	\date		10/21/2002 9:50:17 AM
		**	\warning	none
		*/
		////////////////////////////////////////////////////////////////////////////////
		CommandPacket(CommandID_type , AquariusKey_type SenderID, AquariusKey_type BodyID,
					  float weight, unsigned int decisionID, float fitness, 
					  const InternalPoint_type& point)throw();


		////////////////////////////////////////////////////////////////////////////////
		//	FUNCTION:
		/*!	\brief		Ctor to use iff the target is a voxel.
		**
		**	\param CommandID_type				See CommandID.hpp.
		**	\param AquariusKey_type SenderID		ID of the brain sending this command..
		**	\param AquariusKey_type BodyID,		ID of the body to recieve this command.
		**	\param float weight					Weight brain wishes to give this cmd.
		**	\param const Voxel_type& voxel		Reference to the voxel in the game
		**											 that is to be the target of this cmd.
		**
		**	\exception	none
		**	\par		DESCRIPTION: Fills out the internal struct for iff the target is
		**				 a voxel in the game space.
		**
		**	\author		Nate Cleveland
		**	\date		10/21/2002 9:50:17 AM
		**	\warning	none
		*/
		////////////////////////////////////////////////////////////////////////////////
		CommandPacket(CommandID_type , AquariusKey_type SenderID, AquariusKey_type BodyID,
					  float weight, unsigned int decisionID, float fitness, 
					  const Voxel_type& voxel)throw();

		//! target must be infered from the command.
		////////////////////////////////////////////////////////////////////////////////
		//	FUNCTION:
		/*!	\brief		Ctor to use iff the target can be inferred by the commandID.
		**
		**	\param CommandID_type				See CommandID.hpp.
		**	\param AquariusKey_type SenderID		ID of the brain sending this command..
		**	\param AquariusKey_type BodyID,		ID of the body to recieve this command.
		**	\param float weight					Weight brain wishes to give this cmd.
		**
		**	\exception	none
		**	\par		DESCRIPTION: Fills out the internal struct for iff the target is
		**				 something that can be inferred by the commandID.
		**
		**	\author		Nate Cleveland
		**	\date		10/21/2002 9:50:17 AM
		**	\warning	none
		*/
		////////////////////////////////////////////////////////////////////////////////
		CommandPacket(CommandID_type , AquariusKey_type SenderID, AquariusKey_type BodyID,
					 float weight, unsigned int decisionID,float fitness)throw();


		CommandID_type GetCommandID(void)const throw(){return m_commandID;}
		AquariusKey_type GetBodyID(void)const throw(){return m_bodyID;}
		AquariusKey_type GetSenderID(void)const throw(){return m_senderID;}

		////////////////////////////////////////////////////////////////////////////////
		//	FUNCTION:
		/*!	\brief		Target Getting method that guarentees no throw.
		**
		**	\param ActionTarget& actTarget		Action target that'll be filled with
		**											 the current target of this command
		**											 packet.
		**
		**	\return		none
		**	\exception	none
		**	\par		DESCRIPTION: Fills the passed in struct with the current target.
		**
		**	\author		Nate Cleveland
		**	\date		10/21/2002 10:03:29 AM
		**	\warning	none
		*/
		////////////////////////////////////////////////////////////////////////////////
		void GetTarget(ActionTarget& actTarget)const throw();
		ActionTarget::ActionTargetFlag GetTargetFlag(void) const throw(){return m_target.m_ActionFlag;}


		////////////////////////////////////////////////////////////////////////////////
		//	FUNCTION:
		/*!	\brief		Returns a target's Key. Will throw iff the target isn't a key.
		**	\param		none
		**	\return		The key or id of the target.
		**	\exception	Will throw iff the target isn't a key
		**	\par		DESCRIPTION: Returns a target's Key. Will throw iff the target
		**				 isn't a key.
		**
		**	\author		Nate Cleveland
		**	\date		10/21/2002 10:09:00 AM
		**	\warning	none
		*/
		////////////////////////////////////////////////////////////////////////////////
		AquariusKey_type GetTargetKey(void) const;


		////////////////////////////////////////////////////////////////////////////////
		//	FUNCTION:
		/*!	\brief		Fills the passed point with the point of the target.
		**	\param		pnt			Point to fill with target point.
		**	\return		none
		**	\exception	Will throw iff the target isn't a point
		**	\par		DESCRIPTION: Fills given point with the point of the target. Will
		**				 throw iff the target isn't a point.
		**
		**	\author		Nate Cleveland
		**	\date		10/21/2002 10:09:00 AM
		**	\warning	none
		*/
		////////////////////////////////////////////////////////////////////////////////
		void GetTargetPoint(InternalPoint_type& pnt) const;


		////////////////////////////////////////////////////////////////////////////////
		//	FUNCTION:
		/*!	\brief		Fills the passed voxel with the voxel of the target.
		**	\param		vox			Voxel to fill with target voxel.
		**	\return		none
		**	\exception	Will throw iff the target isn't a voxel.
		**	\par		DESCRIPTION: Fills given point with the point of the target. Will
		**				 throw iff the target isn't a point.
		**
		**	\author		Nate Cleveland
		**	\date		10/21/2002 10:09:00 AM
		**	\warning	none
		*/
		////////////////////////////////////////////////////////////////////////////////
		void GetTargetVoxel(Voxel_type& vox) const;

		////////////////////////////////////////////////////////////////////////////////
		//	FUNCTION:
		/*!	\brief		Replaces the current target with the caller specified one.
		**
		**	\param		newTarget		Target to replace current with.
		**	\author		Nate Cleveland
		**	\date		10/22/2002 1:13:01 PM
		**	\warning	none
		*/
		////////////////////////////////////////////////////////////////////////////////
		void SetTarget(const ActionTarget& newTarget) throw(){this->m_target = newTarget;}

		////////////////////////////////////////////////////////////////////////////////
		//	FUNCTION:	
		/*!	\brief		Replaces the current target with a target for the given key.
		**	\param		newTargKey			Key of the new target.
		**	\return		none
		**	\exception	none			
		**	\author		Nate Cleveland
		**	\date		2/20/2003 10:52:41 AM
		**	\warning	none
		*/
		////////////////////////////////////////////////////////////////////////////////
		void SetTarget(const AquariusKey_type& newTargKey) throw()
		{
			m_target.m_ActionFlag = ActionTarget::ACTION_TARGET_FLAG_KEY;
			m_target.m_ActionTarget.m_key = newTargKey;
		}


		float GetWeight(void) const throw(){return m_weight;}
		void  SetWeight(float weight) throw(){m_weight = weight;}
		float GetFitness(void)const throw(){return m_fitness;}
		unsigned int GetDecisionID(void) const throw(){return m_decisionID;}


		////////////////////////////////////////////////////////////////////////////////
		//	FUNCTION:
		/*!	\brief		Equality operator for CommandPacket. Ignores the weight member.
		**
		**	\param const CommandPacket&			Right Hand side of == expression.
		**
		**	\return		boolean true false.
		**	\exception	none
		**	\par		DESCRIPTION: At this time ignores the wieght member. Also ignores
		**				 the target member iff we are acting upon the ROAM command.
		**
		**	\author		Nate Cleveland
		**	\date		10/21/2002 10:14:22 AM
		**	\warning	none
		*/
		////////////////////////////////////////////////////////////////////////////////
		bool operator != (const CommandPacket& ) const throw();

		//!at this time weight is ignored by the cmp routines.
		////////////////////////////////////////////////////////////////////////////////
		//	FUNCTION:
		/*!	\brief		Non-equality operator for CommandPacket. Ignores the weight member.
		**
		**	\param const CommandPacket&			Right Hand side of != expression.
		**
		**	\return		boolean true/false.
		**	\exception	none
		**	\par		DESCRIPTION: At this time ignores the wieght member. Also ignores
		**				 the target member iff we are acting upon the ROAM command.
		**
		**	\author		Nate Cleveland
		**	\date		10/21/2002 10:17:03 AM
		**	\warning	none
		*/
		////////////////////////////////////////////////////////////////////////////////
		bool operator == (const CommandPacket& ) const throw();
		

		////////////////////////////////////////////////////////////////////////////////
		//	FUNCTION:	
		/*!	\brief		Reads data in from a netMessage.
		**	\param		netMessage		stream to read from
		**	\exception	iff there is an error in the stream or if target flag is unknown
		**	\author		Nate Cleveland
		**	\date		2/7/2003 9:58:59 AM
		**	\warning	none
		*/
		////////////////////////////////////////////////////////////////////////////////
		void operator << (UDPNet::NetMessage& netMessage);

		////////////////////////////////////////////////////////////////////////////////
		//	FUNCTION:	
		/*!	\brief		Writes its data to a netMessage as compact as possible
		**	\param		netMessage		stream to write to
		**	\exception	iff there is an error in the stream or if target flag is unknown
		**	\author		Nate Cleveland
		**	\date		2/7/2003 9:59:02 AM
		**	\warning	none
		*/
		////////////////////////////////////////////////////////////////////////////////
		void operator >> (UDPNet::NetMessage& netMessage);

		////////////////////////////////////////////////////////////////////////////////
		//	FUNCTION:
		/*!	\brief		Writes all needed data to the stream
		**
		**	\param ostream&					object we'll stream to.
		**	\param const  CommandPacket&	Reference to this class.
		**
		**	\return		A reference to the stream.
		**	\exception	none
		**	\par		DESCRIPTION: Writes all the needed data to the stream.
		**
		**	\author		Nate Cleveland
		**	\date		10/16/2002 6:30:59 PM
		**	\warning	none
		*/
		////////////////////////////////////////////////////////////////////////////////
		friend std::ostream& operator << (std::ostream&, const CommandPacket&);

		////////////////////////////////////////////////////////////////////////////////
		//	FUNCTION:
		/*!	\brief		Reads in data from the stream.
		**
		**	\param ostream&					object we'll stream from.
		**	\param const  CommandPacket&	Reference to this class.
		**
		**	\return		A reference to the stream.
		**	\exception	none
		**	\par		DESCRIPTION: Reads data from the stream and resets this class
		**				 accordingly.
		**
		**	\author		Nate Cleveland
		**	\date		10/16/2002 6:31:04 PM
		**	\warning	none
		*/
		////////////////////////////////////////////////////////////////////////////////
		friend std::istream& operator>> (std::istream&, CommandPacket &);
	};//end class CommandPacket
}//end namespace InterpretationModule.