////////////////////////////////////////////////////////////////////////////////
/*!	\file		fishBody.cpp
**	\brief		Defines all functionality in fishBody.hpp
**	\author		Nate Cleveland
**	\date		2/21/2003 9:57:51 AM
**	\par		DESCRIPTION:
**					Please refer to fishBody.hpp for details.
**	\par		REMARKS:
**					
*/
////////////////////////////////////////////////////////////////////////////////
#pragma warning(disable:4786)
#include <algorithm>
#include <fstream>
#include "mutate.hpp"
#include "SoundLogic.hpp"
#include "fishBodyHandle.hpp"
#include "action.hpp"
#include "commandPacket.hpp"
#include "graphicClient.hpp"
#include "fishBody.hpp"
#include "frozenFish.hpp"
#include <assert.h>
#include "worldBoundryTracer.hpp"
#include "../tankSimulator/TankSimulator.h"
#include "../ai_core/AI_System.h"
#include "../Nexception/Nexception.h"
#include "../networking/NetData.h"
#include "../gfx/renderer.h"
#include <stdio.h>
using namespace InterpretationModule;
using namespace std;
// memory leak detection
#ifdef _DEBUG
    #include"memleak.h"
    #define new DEBUG_NEW
    #undef THIS_FILE
    static char THIS_FILE[] = __FILE__;
#endif


//Debugging defines.
//#define FISH_BODY_DEBUG
#ifdef FISH_BODY_DEBUG
	//include the debuggin header, create a temp buffer for sprintf.
#	include "../debugging/Debug.h"
	namespace 
	{
		char buff[1024] ={0};
	}

	//wrap all uses of the debugger around this define.
#	define FISH_BODY_DEBUG_LOG
#endif


//Instatiate each of the const item mentioned in header.
const float InterpretationModule::MAX_TURN_VALUE_XZ_PER_SEC = pi;
const float InterpretationModule::MAX_TURN_VALUE_XY_PER_SEC = pi;
const float InterpretationModule::MAX_TURN_VALUE_ZY_PER_SEC = pi;
const float InterpretationModule::BASE_RADIUS				= 7;
const float InterpretationModule::MATING_INCLINE_THRESHOLD	= 0.8f;


namespace
{
	const float	DEFUALT_TOLERANCE								= 6;
	const float DEFUALT_SCALER									= 7;
	const float MIN_SWIM_SOUND_SPEED							= 0.0f;
	const int	MAX_ACTIONS										= 10;
}



/*******************************************************************************
 * Help functions and function prototypes.                                                             
 *//////////////////////////////////////////////////////////////////////////////

//A couple simple function prototypes. Each function name speaks for itself.
// functions are defined at bottom of this file.
CommandID_type CommandID_BrainToBody(AI_System::Attribute_ID actID);
AI_System::Action_ID CommandID_BodyToBrain(CommandID_type cmdID);

//Returns to the caller the ammount that a field should be scaled to push a fish
// away from an object.
////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	
/*!	\brief		This method computes how much a terrain field should be scaled.
**	\param		distToTarg		How far away we are from the target.
**	\param		tolerance		How far away the target is before we ignore it.
**	\param		scaler			All values will be multiplied by this value.
**	\return		none
**	\exception	none
**	\par		DESCRIPTION:
**				Finds how much a terrain field should be scaled based on the
**				 distance to the target and our tolerance to distance. 
**
**	\author		Nate Cleveland
**	\date		11/8/2002 10:25:33 AM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
template<class val_type>
val_type FindFieldMultiplier(val_type distToTarget, val_type tolerance = DEFUALT_TOLERANCE, 
							 val_type scaler = DEFUALT_SCALER)
{
	val_type multiplier = tolerance - fabsf(distToTarget);
	multiplier = __max(0, multiplier);					//if less then zero, set to zero
	return multiplier*scaler;
}


////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	
/*!	\brief		Checks for collision with the height map.
**	\param		pTankClient		pointer to a tank simulator client.
**				fishPos			const reference to the fish's position.
**				radius			radius of the fish's sphere.
**	\return		 0 == collision with height map
**				-1 == NULL tank client passed.
**				>0 == Y distance to the height map.
**	\exception	none
**	\par		DESCRIPTION:
**					Checks for collision with the height map. Returns 0 for a collision
**					 (no matter how far under the map it is), -1 if the tank client is
**					 invalid or the y distance to the height map.
**	\author		Nate Cleveland
**	\date		4/2/2003 10:03:03 AM
**	\warning	ALWAYS check for a -1 return, it is common for fish to get a single
**				 update without a tank client.
*/
////////////////////////////////////////////////////////////////////////////////
f32 HeightMapCollision(TankSimulatorClient* pTankClient, const InterpPnt_type& fishPos,
					   float radius = 0)
{
	//sanity check
	if(pTankClient) 
	{
		//add the radius to the pos's y element
		InterpPnt_type pos(fishPos[0], fishPos[1]+radius, fishPos[2]);

		//find the height of the height map at this location
		float hMap_y	= pTankClient->GetProjectedHeightOnHeightMap(pos);

		//if the hMap_y is greater then pos[1] we have an hMap collision, return zero.
		if(hMap_y >= pos[1])
			return 0;

		//otherwise, return the difference between the two.
		return pos[1] - hMap_y;
	}//end if

	return -1;
}




/*******************************************************************************
 * FishBody Methods.                                                             
 *//////////////////////////////////////////////////////////////////////////////

//See header for documentation.
FishBody::FishBody(void)
:m_pAttributes(&m_attributes),	m_pStats(&m_stats),	m_pTankClient(NULL),
 m_markerArrHead(0),			m_markerArrSize(0),	m_IsMouthStopped(false), 
 m_isAttacking(false),			m_isEating(false),  m_IsFinStopped(false),
 m_actListMaxSize(ACT_LIST_DEFUALT_MAX_SIZE),		m_key(0),
 m_updateTimer(0.016667f, false),					m_facingVector(0,0,1), 
 m_strObj(new StrObj),			m_IsLearn(true),	m_IsPathing(true),
 BodyClient(AcqariusSound::Sound::Instance())
{

	//create the interface object between us and the actions.
	m_hFishBody = new HFishBody(*this);

	//zero out the marker array
	memset(m_markerArr, 0, sizeof(InternalPoint_type) * MARKER_ARRAY_SIZE);
	memset(m_markerArrWeight, 0, sizeof(float) * MARKER_ARRAY_SIZE);

	//set a defualt direction for us to face.
	m_currVector.y = m_currVector.z = 0;
	m_currVector.x = 1.0f;

	//create a mutation object and mutate this fish's attributes.
	Mutate mutator;
	mutator(*this);
	m_attributes.SetHealth_Normalized(1);
	m_attributes.SetEnergyLevel_Normalized(1);

	//Create a soundLoader object to choose the proper sounds for this fish.
	SoundLoader soundLoader(m_attributes);
	LoadSounds(soundLoader.Bite(), soundLoader.Hit(), soundLoader.Swim(), NULL,//soundLoader.Think(),
			   soundLoader.Lift(), soundLoader.Drop(), soundLoader.CollideBody(), 
			   soundLoader.CollideTerrain(), soundLoader.Death(), soundLoader.Eat());
}



//See header for documentation.
void FishBody::Build(const FrozenFish& frozenFish) throw()
{
	///stats and atts
	m_attributes		= frozenFish.m_attributes;
	m_attributes.VerifyValues();
	m_stats				= frozenFish.m_stats;

	//state flags
	m_IsMouthStopped	= frozenFish.m_isMouthStopped;
	m_isAttacking		= frozenFish.m_isAttacking;
	m_isEating			= frozenFish.m_isEating;
	m_IsFinStopped		= frozenFish.m_isFinStopped;
	m_IsLearn			= frozenFish.m_isLearn;
	m_IsPathing			= frozenFish.m_isPathing;

	//physical stuff
	m_currVector.x		= frozenFish.m_currVector[0];
	m_currVector.y		= frozenFish.m_currVector[1];
	m_currVector.z		= frozenFish.m_currVector[2];

	m_facingVector[0]	= frozenFish.m_facingVector[0];
	m_facingVector[1]	= frozenFish.m_facingVector[1];
	m_facingVector[2]	= frozenFish.m_facingVector[2];
	

	//sphere data, leave zero if the tank client is null
	if(m_pTankClient) 
	{
		InterpVec_type tempVec;

		PhysicsSimulator& phySim		 = m_pTankClient->GetPhysicsSimulatorReference();
		PhysicsSimulator::Sphere& sphere = phySim.GetSphereRef(m_key);

		sphere.SetAcceleration(frozenFish.m_accel);
		sphere.SetPosition(frozenFish.m_pos);
		sphere.SetVelocity(frozenFish.m_vel);
	}


	//examples
	m_examples			= frozenFish.m_examples;

	//actions
	ClearActions();
//	m_actList			= frozenFish.m_actions;
	for(unsigned int i=0; i<frozenFish.m_actions.size(); i++)
	{
		Action* newAct=NULL;
		const FrozenFish::PsuedoAction& psuedoAct = frozenFish.m_actions[i];

		//create the correct action, no target
		if(0 == psuedoAct.m_targetFlag) 
		{
			CommandPacket cmdPack(psuedoAct.m_cmdID, psuedoAct.m_senderID, psuedoAct.m_bodyID,
								  psuedoAct.m_weight, psuedoAct.m_decisionID, psuedoAct.m_fitness);
			newAct = Action::CreateAction(*m_strObj, *m_hFishBody, cmdPack);

		}

		//target is a key
		else if(1 == psuedoAct.m_targetFlag)
		{
			CommandPacket cmdPack(psuedoAct.m_cmdID, psuedoAct.m_senderID, psuedoAct.m_bodyID,
								  psuedoAct.m_weight, psuedoAct.m_decisionID, psuedoAct.m_fitness,
								  psuedoAct.m_target.m_key);
			newAct = Action::CreateAction(*m_strObj, *m_hFishBody, cmdPack);
		}

		else if(3 == psuedoAct.m_targetFlag)
		{
			CommandPacket cmdPack(psuedoAct.m_cmdID, psuedoAct.m_senderID, psuedoAct.m_bodyID,
								  psuedoAct.m_weight, psuedoAct.m_decisionID, psuedoAct.m_fitness,
								  psuedoAct.m_target.m_point);

			newAct = Action::CreateAction(*m_strObj, *m_hFishBody, cmdPack);
		}

		assert(newAct);

		newAct->SetElapsedTime(psuedoAct.m_elapsedTime);
		newAct->SetDone(psuedoAct.m_isDone, psuedoAct.m_isComplete);

		m_actList.push_back(newAct);

	}

	//commands
	m_incomingCommands  = frozenFish.m_commands;	

}



//See header for documentation.
FishBody::~FishBody(void)
{
	//clean up the actions.
	ClearActions();

	delete m_hFishBody;
	delete m_strObj;
	
	//clean away the sphere.
	assert(m_pTankClient); 
	{
		PhysicsSimulator &phySim = m_pTankClient->GetPhysicsSimulatorReference();
		phySim.RemoveSphere(m_sphereKey);

		delete m_pTankClient;
		m_pTankClient = NULL;
	}
	

	//release keys????
	(MsgHandler::MessageHandler::Instance())->ReturnKey(m_key);
}
//See header for documentation.
void FishBody::ClearActions(void) throw()
{
	//clean up the actions.
	while(!m_actList.empty()) 
	{
		Action* actToKill = m_actList.back();	//get last element
		m_actList.pop_back();					//remove from list
		delete actToKill;						//free memory.
	}
}


//See header for documentation.
void FishBody::Breed(void) throw()
{
	Mutate mutator;
	mutator(*this);
}


//See header for documentation.
bool FishBody::Breed(AquariusKey_type parent) throw()
{
	//get a pointer to the fish to mutate from.
	FishBody* mother = GetFishBodyPtr(parent);

	//make sure that fish exsists.
	if(!mother)
		return false;

	//Mutate the fish Based on a single parent.
	Mutate mutator(*mother);
	mutator(*this);
	return true;
}


//See header for documentation.
bool FishBody::Breed(AquariusKey_type parentA, AquariusKey_type parentB) throw()
{
	FishBody* mother = GetFishBodyPtr(parentA);
	FishBody* father = GetFishBodyPtr(parentB);

	//make sure both of these fish exsist.
	if(!mother || !father)
		return false;


	//mutate the fish based on both parents.
	Mutate mutator(*mother, *father);
	mutator(*this);
	return true;
}


//See header for documentation.
void FishBody::ForceDraw(void) throw()
{
	m_graphicClient.Render();
	return;
}

//See header for documentation
void FishBody::QueueDraw(void) throw()
{
	InterpPnt_type pos;
	GetPos(pos);

	
	if(gfx::Renderer::Get().GetFrustum().ContainsSphere(pos, m_graphicClient.GetRadius()))
		m_graphicClient.SetRender();

	return;
}


//see header for documentation
void FishBody::SoundLogic(void) throw()
{
	/////Terrain and other fish collision
	if(m_pTankClient)
	{
		InterpPnt_type pos;
		GetPos(pos);
		m_pos(&(pos[0]));

		//get a reference to our sphere.
		PhysicsSimulator::Sphere& sphere = 
			m_pTankClient->GetPhysicsSimulatorReference().GetSphereRef(m_sphereKey);

		//see if there was a collision.
		if(sphere.GetCollidedDuringLastFrame())
		{
			CollideTerrain();
//			sphere.SetCollidedDuringLastFrame(false);		// Unneeded. [12/??/2003]

#			ifdef FISH_BODY_DEBUG_LOG
//				(Debugging::Debugger::Instance())->output(Debugging::BOD, Debugging::L4, "Collide");
#			endif
		}//end if
	}//end if

	/////swiming logic.
	InterpVec_type dir;

	GetDirection(dir);
	//Play only if speed is greater then 1.
	if(Magnitude<float>(dir) > MIN_SWIM_SOUND_SPEED)
		Swim();
}


//see header for documentation
void FishBody::GraphicLogic() throw()
{
	//Posision.
	InterpPnt_type pos;
	GetPos(pos);
	
	m_graphicClient.SetPos(pos);
	
	//Direction
	GetDirection(pos);
	m_graphicClient.FindRotation(pos, m_facingVector);


	//attributes.
	m_graphicClient.SetArmor(m_pAttributes->GetArmor());
	m_graphicClient.SetHealth(m_pAttributes->GetHealth_Normalized());
	m_graphicClient.SetMetabolism(m_pAttributes->GetMetabolism());
	m_graphicClient.SetRadius(BASE_RADIUS*__max(.09f, m_attributes.GetSize()));
	m_graphicClient.SetSize(__max(.09f, m_pAttributes->GetSize()));
	m_graphicClient.SetSpeedRatio(m_speedRatio);
	m_graphicClient.SetStomachNormal(m_pAttributes->GetStomach_Normalized());
	m_graphicClient.SetTeeth(m_pAttributes->GetTeeth_Baleen());
}

//See header for documentation.
void FishBody::RayToField(float x, float y, float z, InterpVec_type& field, 
						  InterpPnt_type& pos, PhysicsSimulator& phySim) const throw()
{
	//make a vector of the given values
	InterpVec_type ray(x,y,z);
	InterpPnt_type collisionPnt;

	//cast the ray against the height map
	if(false)//phySim.CastRay(pos, ray, collisionPnt))
	{
		//find the distance to the collisionPnt.
		float dist = ComputeDistance<float>(pos, collisionPnt);	
	
		//negate vector and scale it by the value returned by findFieldMult
		ray *= -1 * FindFieldMultiplier(dist);

		//apply this ray to the field.
		

		if(ray[0] != 0 && ray[1] != 0 && ray[2] != 0)
			cout << "Ray hit (" << ray[0] << ", " << ray[1] << ", " << ray[2] << ")" << endl;
		else
			field += ray;
		
	}

	//if there was no collision with the height map, check with 
	// the world itself.
	else
	{
		//create and call worldTracer functor.
		WorldBoundryTracer worldTracer(m_pTankClient);
		float dist = worldTracer(pos, x, y, z);

		//make sure we had a collision
		if(dist > 0 && dist < DEFUALT_TOLERANCE)
		{
			//negate vector and scale it by the value returned by findFieldMult
			ray *= -1 * FindFieldMultiplier(dist, DEFUALT_TOLERANCE, DEFUALT_SCALER);
			field += ray;
		}//end if
	}//end else
}

//See header for documentation
void FishBody::GenerateTerrainField(InterpVec_type& terrainField) const throw()
{
	//get the location and position for the RayToFiled method.
	InterpPnt_type pos;
	GetPos(pos);
	PhysicsSimulator& phySim = m_pTankClient->GetPhysicsSimulatorReference();
	
	//cast the rays to generate the terrainField
	RayToField( 1, 0, 0,  terrainField, pos, phySim);	//cast a ray along ( 1, 0, 0)
	RayToField(-1, 0, 0,  terrainField, pos, phySim);	//cast a ray along (-1, 0, 0)
	RayToField( 0, 1, 0,  terrainField, pos, phySim);	//cast a ray along ( 0, 1, 0)
	RayToField( 0,-1, 0,  terrainField, pos, phySim);	//cast a ray along ( 0,-1, 0)
	RayToField( 0, 0, 1,  terrainField, pos, phySim);	//cast a ray along ( 0, 0, 1)
	RayToField( 0, 0,-1,  terrainField, pos, phySim);	//cast a ray along ( 0, 0,-1)
	
	//check for collision, or near collision, with the height map.
	float dist = HeightMapCollision(m_pTankClient, pos, m_graphicClient.GetRadius());
	if(dist >= 0 && dist >= DEFUALT_TOLERANCE)
	{
		//apply a resistive force.
		InterpVec_type resist(0,1,0);
		resist *= FindFieldMultiplier(dist, DEFUALT_TOLERANCE, DEFUALT_SCALER);
		terrainField += resist;			
	}//end if
}


//See header for documentation.
void FishBody::AddMarker(const InternalPoint_type& point, float weight)
{

	//add to the array at the current head
	this->m_markerArr[this->m_markerArrHead] = point;
	this->m_markerArrWeight[this->m_markerArrHead] = weight;

	//increment the count iff we won't overflow the array
	if(MARKER_ARRAY_SIZE > this->m_markerArrSize + 1)
		this->m_markerArrSize++;

	//move to the next element in the array, be sure to wrap around
	this->m_markerArrHead++;
	this->m_markerArrHead %= MARKER_ARRAY_SIZE;	
}

//See header for documentation
bool FishBody::SetClient(TankSimulatorClient *pTankClient) throw()
{
	//check for null
	if(pTankClient)
	{
		PhysicsSimulator& phySim = pTankClient->GetPhysicsSimulatorReference();

		//Generate a sphere key, for now just set to the same as the fishBody key.
		AquariusKey_type sphereKey = this->m_key;		

		//Add a sphere to the simulation, check for failure.
		if(phySim.AddSphere(sphereKey)) 
		{
			//All went well, set the sphere key and tankClient pntr.
			m_sphereKey = sphereKey;
			m_pTankClient = pTankClient;

			//Set the position, acceleration, velocity an radius of our sphere.
			//..
			PhysicsSimulator::Sphere& sphere = phySim.GetSphereRef(sphereKey);
			InterpPnt_type pos;
			GetRandPoint(pos);


			sphere.SetPosition(pos);
			
			return true;
		}//end if

		else
			; //release the sphere key?
	}//end if
	
	return false;
}

//See header for documentation
bool FishBody::GetFishBody(AquariusKey_type key, FishBody& fishBody) 
{
	if(m_pTankClient)
	{
		fishBody = *m_pTankClient->GetFishBodyPtr(key);
		return true;
	}
	
	return false;
}

//See header for documentation
FishBody* FishBody::GetFishBodyPtr(AquariusKey_type key) throw()
{
	if(m_pTankClient) 
		return m_pTankClient->GetFishBodyPtr(key);
	return NULL;	
}

//See header for documentation
void FishBody::AddToMatingList(void) throw()
{
	//Make sure we have a tank client.
	if(m_pTankClient) 
	{
		//Add self to mating list.
		this->m_pTankClient->AddFishToMatingList(m_key);

		//set mating inclination to zero.
		this->m_attributes.SetIncline(0);
	}
}

//See header for documentation
void FishBody::ForcePosition() throw()
{
	InterpPnt_type pos;
	GetPos(pos);
	m_phyData.ForcePosition(pos[0], pos[1], pos[2]);
}


//See header for documentation.
bool FishBody::Command(CommandID_type cmdID, AquariusKey_type senderID, AquariusKey_type bodyID,
					   float weight, unsigned int decisionID, float fitness, 
					   AquariusKey_type targetKey) throw()
{
	CommandPacket cmdPack(cmdID, senderID, bodyID, weight, decisionID, fitness, targetKey);
	return Command(cmdPack);
}


//See header for documentation.
bool FishBody::Command(const CommandPacket& cmdPack) throw()
{
	//if this is true add to the vector of incoming commands.
	if(m_incomingCommands.size() < this->m_actListMaxSize) 
	{
		m_incomingCommands.push_back(cmdPack);
		return true;
	}

	return false;
}


//See header for documentation.
int FishBody::Command(const std::vector<CommandPacket>& cmdPackVec) throw()
{
	int addedCount = 0;		//keep a list of cmdPacks that're successfully added to list.

	//loop through and pull out each element one at a time.
	std::vector<CommandPacket>::const_iterator cmdIT = cmdPackVec.begin();
	for( ; cmdIT != cmdPackVec.end() ; cmdIT++)
	{
		const CommandPacket& cmdPack = *cmdIT;

		//if the returned value is false the list is full, just return out to the caller
		// how many we succefully added.
		if(!Command(cmdPack)) 
			return addedCount;
		addedCount++;
	}

	return addedCount;
}



//See header for documentation
int FishBody::GiveAICommand(const std::vector<AI_System::AI_Action>& aiActVec) throw()
{
	//make sure the fish is alive before we bother taking in any commands.
	if(m_attributes.GetHealth() <= 0)
		return 0;

	//only allow ourselves so many pending actions. Not concered with at this time. [2/10/2003]
//	if(m_incomingCommands.size() <= MAX_ACTIONS) 
//		return 0;
	
	
	//keep a count of how many cmdPacks we've added to the system.
	int addedCount = 0;		


	//loop through, pulling out each element and converting it to a cmdPack
	std::vector<const AI_System::AI_Action>::const_iterator aiActionIT = aiActVec.begin();
	for( ;aiActionIT != aiActVec.end() ; aiActionIT++, addedCount++ )
	{		
		const AI_System::AI_Action aiAct = *aiActionIT;

		//check for and ignore NULL commands.
		if(aiAct.ID == AI_System::AC_NULL)
			continue; 
		

		//check and cut all marker actions
		CommandID_type cmdID = CommandID_BrainToBody(aiAct.ID);

		if(	cmdID == AQUARIUS_COMMAND_EVALUATE_POSITIVE ||
			cmdID == AQUARIUS_COMMAND_EVALUATE_NEGATIVE ||
			cmdID == AQUARIUS_COMMAND_EVALUATE_MARKERS	||
			cmdID == AQUARIUS_COMMAND_DROP_OBJECT		||
			cmdID == AQUARIUS_COMMAND_PICK_UP			||
			cmdID == AQUARIUS_COMMAND_DROP_MARKER		||
			cmdID == AQUARIUS_COMMAND_EAT				   ) 
		{
#		ifdef FISH_BODY_DEBUG_LOG			
			(Debugging::Debugger::Instance())->output(Debugging::BOD, Debugging::L1, 
				"Depricated command given to fish body, %d in fish body lingo.", cmdID);
#		endif
			continue;
		}
		

		// changed it where all weight is assumed to be 10 [1/16/2003]. Otherwise fish 
		//  would never move.
		CommandPacket cmdPack(cmdID, aiAct.tag.unit_ID, m_key, 
		10, aiAct.tag.decisionID, aiAct.tag.fitness, 
		aiAct.objectID);

		//There is a special case for negative markers.
		if(AI_System::DROP_NEG_MARKER == aiAct.ID && aiAct.value > 0)
			cmdPack.SetWeight(0 - cmdPack.GetWeight());



#		ifdef FISH_BODY_DEBUG_LOG
			//display the packet we just loaded.
			sprintf(buff, "ID: %d   Evaluated Object %d, value %.2f, weight%.2f\n",
				m_key,  aiAct.objectID, aiAct.value, 
				aiAct.value*this->m_attributes.GetMaxSpeed()+0.001f);
			(Debugging::Debugger::Instance())->output(Debugging::BOD, Debugging::L2, buff);
#		endif

		//Add this command to internal container. if false is returned break loop
		// and return the number of items we were able to read in to the caller.
		if(!Command(cmdPack))
			break;
	}


	// Fold the action we're just given into what we're already doing. [12/??/2002]
	Flush();


	//play the "think" sound.
	Think();
	return addedCount;
}


bool FishBody::GotoPoint(const InterpPnt_type& targetPnt) throw()
{
	//create the command
	InternalPoint_type pnt;
	pnt.x = targetPnt[0];
	pnt.y = targetPnt[1];
	pnt.z = targetPnt[2];

	CommandPacket cmdPack(AQUARIUS_COMMAND_GOTO, 0, m_key, 10.0f, 0, 0.0f, pnt);

	//send the packet to the system.
	if(!Command(cmdPack))
		return false;

	//activate the current command.
	Flush();
	return true;
}


//See header for documentation.
void FishBody::Flush(void) throw()
{
	std::list<Action*> tempList;
	
	//clear the old action strings
	m_strObj->ClearStrings();

	//go through the vector of incomingCommands. If we find that we're already
	// acting on that command, add it to the temp vector.
	for( ; !m_incomingCommands.empty() ; )
	{
		//will point to an action that matches this command.
		Action *matchedAction = NULL;			
		
		CommandPacket cmdPack = (m_incomingCommands.back());

		//remove packet from vector since it is no longer needed.
		m_incomingCommands.pop_back();

		//loop through list of current actions and see if this value matches any of 
		// them.
		std::list<Action*>::iterator actListIT = this->m_actList.begin();
		while( actListIT != this->m_actList.end() )
		{
			matchedAction = *actListIT;

			//if this true we have a match.
			if(*matchedAction == cmdPack)
			{
				//remove this action from the list, catch the new actListIT
				actListIT = m_actList.erase(actListIT);

				//add it to the new list with the new weight value.
				matchedAction->SetCommandPackWeight(cmdPack.GetWeight());	
				break;			
			}//end if

			else
			{
				matchedAction = NULL;
				actListIT++;
			}//end else
		}//end for


		//if this is true then we didn't find an exsisting action and'll need to create
		// a new one.
		if(!matchedAction)
			matchedAction = Action::CreateAction(*m_strObj, *m_hFishBody, cmdPack);

		//check for null before adding action.
		if(matchedAction)
			tempList.push_back(matchedAction);
	}//end for


	//at this time none of the actions that were matched will be retained to the next
	// round. So, delete all of them after generating examples from them.
	for( ; !m_actList.empty() ; )
	{
		Action* action = this->m_actList.back();
		m_actList.pop_back();
		GenerateExample(action);
		delete action;
	}

	//now set all the generated actions into our active actions list.
	m_actList = tempList;


	//instantiate all the strings in the action list
	std::list<Action*>::iterator	actListIT = m_actList.begin();
	for( ;actListIT != m_actList.end(); actListIT++)
		(*actListIT)->SetString();
}//end FishBody::Flush



//See header for documentation.
void FishBody::GenerateExample(const Action* action) throw()
{
	if(m_IsLearn) 
	{
		//create a new ai action tag to return to the brain.
		AI_System::AI_ActionTag newActTag;

		//fill it out.
		newActTag.completed	= action->IsCompleted();
		newActTag.fitness		= action->GetCommandPackFitness();
		newActTag.unit_ID		= action->GetCommandPackSenderID();
		newActTag.decisionID	= action->GetCommandPackDecisionID();


		//check for special negative marker case
		if(action->GetCommandID() == AQUARIUS_COMMAND_DROP_MARKER && 
		   action->GetCommandPackWeight() < 0) 
		   newActTag.decisionID = AI_System::DROP_NEG_MARKER;
		

		//add it to the list.
		m_examples.push_back(newActTag);
	}
}


//See header for documentation.
void FishBody::GetExamples(std::vector<AI_System::AI_ActionTag>& vec) const throw()
{
	vec = m_examples;
}


//See header for documentation.
void FishBody::Dump(void) throw()
{
	m_examples.clear();
}


//See header for documentation.
void FishBody::Prune(void) throw()
{
	//loop through and remove any items that are marked as done.
	std::list<Action*>::iterator	actIT = this->m_actList.begin();
	while(actIT != this->m_actList.end()) 
	{
		Action* currAction = *actIT;

		//check to see if done.
		if(currAction->IsDone()) 
		{
			this->GenerateExample(currAction);		//create an example from this action.
			delete currAction;
			actIT = this->m_actList.erase(actIT);	//remove from list.
		}//end if

		else
			actIT++;
	}
}//end FishBody::Prune

//See header for documentation.
void FishBody::GetPos(InterpPnt_type& pos) const 
{
	try
	{
		//set the position of the fish for the phyData struct.
		PhysicsSimulator& phySim = m_pTankClient->GetPhysicsSimulatorReference();
		PhysicsSimulator::Sphere& sphere = phySim.GetSphereRef(m_key);
		pos = sphere.GetPosition();
	} catch(...)	{
#			ifdef FISH_BODY_DEBUG_LOG
				(Debugging::Debugger::Instance())->output(Debugging::BOD, Debugging::L4, 
					"Non fatal error when fishBody attempted to get its position.");
#			endif
	}
}


//See header for documentation.
void FishBody::GetDir(InterpVec_type& dir) const 
{
	//set the position of the fish for the phyData struct.
	PhysicsSimulator& phySim = m_pTankClient->GetPhysicsSimulatorReference();
	PhysicsSimulator::Sphere& sphere = phySim.GetSphereRef(m_key);
	dir = sphere.GetVelocity();
}
//See header for documentation.
void FishBody::SetPos(const InterpPnt_type& pos) 
{
	//set the position of the fish for the phyData struct.
	PhysicsSimulator& phySim = m_pTankClient->GetPhysicsSimulatorReference();
	PhysicsSimulator::Sphere& sphere = phySim.GetSphereRef(m_key);
	sphere.SetPosition(pos);
}



//See header for documentation.
void FishBody::SetDir(const InterpPnt_type& dir) 
{
	//set the position of the fish for the phyData struct.
	PhysicsSimulator& phySim = m_pTankClient->GetPhysicsSimulatorReference();
	PhysicsSimulator::Sphere& sphere = phySim.GetSphereRef(m_key);
	sphere.SetVelocity(dir);
}



//See header for documentation.
void FishBody::UpdateFishStats(float elapsedTime) throw()
{
	InterpPnt_type pos;
	GetPos(pos);

	m_phyData.Update(elapsedTime,pos[0], pos[1], pos[2]);		//phyData.
	m_attributes.Update(elapsedTime);							//attributes
	m_stats.Update(m_attributes, m_phyData, elapsedTime);		//stats
	
	//check to see if it is time for this fish to die of old age.
	if(m_pAttributes->GetLifeSpan() * 3 < m_stats.GetAge() / 60)
		m_attributes.SetHealth(0);

}//end FishBody::UpdateFishStats

//See header for documentation.
void FishBody::UpdateActions(InternalVector_type& summationVec, float elapsedTime) throw()
{
	summationVec.x = summationVec.y = summationVec.z = 0;
	m_hFishBody->Flush();
	Prune();

	//loop through all our actions, if any of them throw an 
	// exception delete them and do not generate an example
	std::list<Action*>::iterator	actIT = this->m_actList.begin();
	while(actIT != m_actList.end())
	{
		Action *currAction = *actIT;

		try
		{
			currAction->IncElapsedTime(elapsedTime);
			(*currAction)(summationVec);	//the () operator is the "update" method.
			actIT++;

		} catch(...){

			cout << "Non fatal exception thrown in FishBody::UpdateActions." << endl;
#			ifdef FISH_BODY_DEBUG_LOG	
				//log the abortion.
				sprintf(buff, "Fish %d's action %d was terminated\n", 
					    currAction->GetCommandPackTargetKey(), currAction->GetCommandID());
				(Debugging::Debugger::Instance())->output(Debugging::BOD, Debugging::L2, buff);
#			endif

			//abort this action. delete it and do not generate an example
			actIT = this->m_actList.erase(actIT);
		}//end catch
	}//end for

	//Get the summation vector from the hFishBody object so we can compute this
	// rounds potential fields.
	m_hFishBody->GetSummationVector(summationVec);
	m_hFishBody->Flush();							//reset the hFishBody member.
	Prune();										//remove any finished actions
}//end FishBody::UpdateActions

//See header for documentation.
void FishBody::UpdatePotenFields(InternalVector_type& summationVec) throw()
{
	//Generate a potential field to avoid nearby terrain.
	InterpVec_type terrainFields(0,0,0);
	GenerateTerrainField(terrainFields);

	//Add that vector to the summation vec.
	summationVec.x += terrainFields[0];
	summationVec.y += terrainFields[1];
	summationVec.z += terrainFields[2];
}

//See header for documentation.
void FishBody::LimitVelocity(InterpVec_type& summationVec, float& speedRatio)const throw()
{
	//caluclate the maxSpeed and scale the velocity vector by that ammount. (it should
	// be a vector that has magnitude <= 1.
	float maxSpeed		= this->m_pAttributes->GetMaxSpeed();
	ScaleVector<float>(summationVec, summationVec, maxSpeed);

	//find the magnitude of the vector after the multiplication.
	float currMagnitude = Magnitude<float>(summationVec);

	//test to see if our velocity is too great.
	if(currMagnitude > maxSpeed)
	{
		//scale the vector down to be a unit vector, then scale again by maxSpeed.
		UnitizeVector<float>(summationVec, summationVec);
		ScaleVector<float>(summationVec, summationVec, maxSpeed);
	}

	//test to see if the velocity is too small. If it is, unitize it.
	// This technique will not effect intential zero vectors.
	else if(currMagnitude < 1.0f)
		UnitizeVector<float>(summationVec, summationVec);


	speedRatio = __max(0, __min(1, currMagnitude));
}

//See header for documentation.
void FishBody::SwimEnergyLoss(float elapsedTime) throw()
{
	//Use the attributeGenerator to find out how much energy we lost.
	AttributeGenerator& atG = m_attributes.GetGenerator();
	float swimEnergyLoss	= atG.SwimmingEnergyLoss(m_attributes.GetMaxSpeed(), 
							  Magnitude<float>(this->m_currVector), elapsedTime);

	//Effect our energyLevel accordingly.
	m_attributes.SetEnergyLevel(m_attributes.GetEnergyLevel() - swimEnergyLoss);
}

//See header for documentation.
void FishBody::FormatSphere(InterpVec_type& newDir) throw()
{
	//Get a reference to our sphere.
	PhysicsSimulator& phySim = m_pTankClient->GetPhysicsSimulatorReference();
	PhysicsSimulator::Sphere& sphere = phySim.GetSphereRef(m_sphereKey);

	//stick our new direction into our accelleration.
	sphere.SetAcceleration(newDir * 2.0f);

	//Modify our spherical radius.
	sphere.SetRadius(m_graphicClient.GetRadius());
}

//See header for documentation.
void FishBody::Suck(void) throw()
{
	//Use our position to get the plankton density at this location.
	InterpPnt_type pos;
	GetPos(pos);
	float planktonDensity = 1;//m_pTankClient->GetPlanktonDensity(pos);

	//Use the attributeGenerator to see how much plankton we desire to remove.
	const AttributeGenerator& ag = m_attributes.GetGenerator();
	float desiredPlankton = ag.FoodFromSuck(m_pAttributes->GetSize(), 
										m_pAttributes->GetSuckability(), planktonDensity);

	//factor in the speed of the fish
	desiredPlankton *= m_speedRatio;

	//Try to remove some of that plankton from the environment.
	float actualPlankton = m_pTankClient->SuckPlankton(pos, desiredPlankton);
	
	//increment the fish's stats
	m_stats.IncPlankton(actualPlankton);
	

	//Credit the fish for how much plankton it removed.
	m_attributes.SetStomach(m_attributes.GetStomach() + actualPlankton);

	if(m_attributes.GetStomach_Normalized() > 1)
		m_attributes.SetStomach_Normalized(1);
	
}

//See header for documentation.
void FishBody::AddChum(float chumAmmount) throw()
{
	//sanity check
	if(m_pTankClient)
	{
		//get our position
		InterpPnt_type pos;
		GetPos(pos);


		//adds specified ammount of chum to the environment.
		m_pTankClient->DumpChum(pos, chumAmmount);
	}
}

//See header for documentation.
void FishBody::UpdatePhysical(InternalVector_type& summationVec, float elapsedTime) throw()
{	

	//Limit the turns
	InterpVec_type newDir(summationVec.x, summationVec.y, summationVec.z);
	m_graphicClient.LimitRotation(newDir, m_facingVector);
	
	//Adjust the velocity vector.
	LimitVelocity(newDir, m_speedRatio);

	//format our spherical representation.
	FormatSphere(newDir);

	//update our current direction
	PhysicsSimulator&			phySim	= m_pTankClient->GetPhysicsSimulatorReference();
	PhysicsSimulator::Sphere&	sphere	= phySim.GetSphereRef(m_sphereKey);	
	InterpVec_type				velVec	= sphere.GetVelocity();

	m_currVector.x = velVec[0];
	m_currVector.y = velVec[1];
	m_currVector.z = velVec[2];

	//update energy loss from swimming.
	SwimEnergyLoss(elapsedTime);
}

//See header for documentation.
void FishBody::Update(float elapsedTime_seconds) throw()
{
	//update our graphic logic.
	GraphicLogic();
	
	//do nothing more if we don't have a tank client yet. Also check if our update timer
	// has gone off yet.
	if(NULL == m_pTankClient)
		return;

	if(false == m_updateTimer(elapsedTime_seconds))
		return;
	
	//collect time that elapsed in other game frames that we didn't update.
	elapsedTime_seconds = m_updateTimer.GetElapsedTime();
	
	//reset the update timer's logic back to zero.
	m_updateTimer.Reset();	
	

	//check to see if the fish has died. We will need to run simplier logic if fish is dead.
	if(m_attributes.GetHealth() <= 0)	
	{
		//clear action string
		m_strObj->ClearStrings();

			////Decay the fish...
			if(m_attributes.GetHealth() <= 0.0f) 
			{
				//find out how much size is to be lost
				float sizeLost = elapsedTime_seconds / 250;

				//Send out that size as chum into the environment.
//				AddChum(__min(m_attributes.GetGenerator().FindStomachSize(sizeLost) * 2, 1));

				//decrement that size from the fish.
				m_attributes.SetSize(__max(m_attributes.GetSize() - sizeLost, 0));
				m_attributes.SetHealth(0.0f);
			}

		//next check to see if the size is <= 0, if it is we need to kill this fish 
		// and leave the function post-haste.
		if(m_attributes.GetSize() <= 0)
		{
			m_pTankClient->AddFishToDeleteFishList(m_key);
			return;
		}
		

		//check to see if we need to clear out the fish's memory.
		if(m_actList.size() || m_examples.size())
		{
			ClearActions();		//clear actions
			Dump();				//clear examples.
		}//end if

		//format our physics sphere to make our fish slowly drift upward.
		InterpVec_type decayDir(0,RandFloat() / 2, 0);
		FormatSphere(decayDir);
	}
	
	
	//otherwise, Run full out update for living fish.
	else 
	{
		//create the main object that'll be used to calculate our potential fields.
		InternalVector_type summationVec;
		summationVec.x = summationVec.y = summationVec.z = 0;
		
		////Update each area of the fish's logic.
		//Update Stats, Attributes and PhyData.
		UpdateFishStats(elapsedTime_seconds);
		
		//check to make sure the fish didn't die in the above update before going further.
		if(m_pAttributes->GetHealth() > 0)
		{
			//Run the logic for each individual action.
			UpdateActions(summationVec, elapsedTime_seconds);	
			
			//Run pathing/collision logic.
			if(m_IsPathing)
				UpdatePotenFields(summationVec);
			
			//find our rotation, velocity, etc.
			UpdatePhysical(summationVec, elapsedTime_seconds);
			
			//play any hit, think sounds.
			SoundLogic();										
			
			//Check to see if we should suck...
			if(!m_isEating && !m_isAttacking)
				Suck();							//suck. suck. SUCK!!!!
			
			//also see if we need to add self to mating list.
			if(m_pAttributes->GetIncline() >= MATING_INCLINE_THRESHOLD) 
				AddToMatingList();
		}//end if

		//otherwise play the death music
		else
			Death();
	}//end else


#	ifdef FISH_BODY_DEBUG_LOG
		//Get current position.
		InterpPnt_type pnt;
		GetPos(pnt);

		//log out any relevent stats.
		sprintf(buff, "ID: %d Pos (%.2f, %.2f, %.2f) speed %.2f\n",
			m_key, pnt[0], pnt[1], pnt[2], m_attributes.GetMaxSpeed());
		(Debugging::Debugger::Instance())->output(Debugging::BOD, Debugging::L2, buff);

		sprintf(buff, "Heal %.3f, N_Ergy %.3f, MatInc %.3f, Stom %.3f", 
			m_pAttributes->GetHealth(), m_pAttributes->GetEnergyLevel_Normalized(), 
			m_pAttributes->GetIncline(), m_pAttributes->GetStomach()); 
		(Debugging::Debugger::Instance())->output(Debugging::BOD, Debugging::L2, buff);

		sprintf(buff, "Dir (%.2f, %.2f, %.2f)\n", m_key, m_currVector.x, m_currVector.y, m_currVector.z);
		(Debugging::Debugger::Instance())->output(Debugging::BOD, Debugging::L3, buff);

		sprintf(buff, "Face(%.2f, %.2f, %.2f)\n", m_key, m_facingVector[0], m_facingVector[1], m_facingVector[2]);
		(Debugging::Debugger::Instance())->output(Debugging::BOD, Debugging::L3, buff);
#	endif
}//end FishBody::Update


//See header for documentation
void FishBody::GetRandPoint(InterpPnt_type& randPnt) throw()
{
	//sanity check
	if(!m_pTankClient)
	{
		randPnt[0] = randPnt[1] = randPnt[2] = 0;
		return;
	}

	////find the bounds of the simulation.
	InterpPnt_type lowCords, highCords;	
	m_pTankClient->GetBoundary(lowCords, highCords);


	//loop through calculating until we are sure this isn't under the hMap
	do 
	{
		//use austin's safer random float method.
		for(int i=0; i<3; i++)
			randPnt[i] = lowCords[i] + (fabsf(highCords[i] - lowCords[i]) * RandFloat());



	} while(0 == HeightMapCollision(m_pTankClient, randPnt, m_graphicClient.GetRadius()));
}


//See header for documentation
void FishBody::Highlight(bool isOn, float red/* = 1*/, float green/* = 1*/, float blue/* = 1*/) throw()
{
	m_graphicClient.Highlight(isOn, red, green, blue);
}



//See header for documentation
void FishBody::Write(std::ostream& outStream) const
{
	outStream << *this;
}

//See header for documentation
void FishBody::Read(std::istream& inStream)
{
	inStream  >> *this;
}

#pragma warning(push)
#pragma warning(disable: 4800)
//See header for documentation (read)
void FishBody::operator << (UDPNet::NetMessage& netMessage)
{
	const char *currChunk = NULL;	//temp point to store data location
	InterpPnt_type pnt_vec;			//temp pnt or vec used to store data before reading in.
	
	//changed by Nathan Frost -- silent failures lead to crashes
	assert(m_pTankClient);

	////FishKey, check for a match. throw if not.
	currChunk = netMessage.ReadNBytes(sizeof(AquariusKey_type));
	if(m_key != *reinterpret_cast<const AquariusKey_type*>(currChunk))
		throw Nexception("Fatal Error when Fishbody tried to read in a netMessage.");

	////Location
	currChunk = netMessage.ReadNBytes(sizeof(InterpPnt_type));
	memcpy(&pnt_vec[0], currChunk, sizeof(InterpPnt_type));
	SetPos(pnt_vec);


	////Facing Dir
	currChunk = netMessage.ReadNBytes(sizeof(InterpVec_type));
	memcpy(&m_facingVector[0], currChunk, sizeof(InterpVec_type));



	////Travel Dir
	currChunk = netMessage.ReadNBytes(sizeof(InterpVec_type));
	memcpy(&pnt_vec[0], currChunk, sizeof(InterpVec_type));
	SetDir(pnt_vec);
	


	////State flags
	//isMouth
	m_IsMouthStopped = *reinterpret_cast<const bool*>(netMessage.ReadNBytes(sizeof(bool)));

	//isAttack
	m_isAttacking	 = *reinterpret_cast<const bool*>(netMessage.ReadNBytes(sizeof(bool)));

	//isEat
	m_isEating		 = *reinterpret_cast<const bool*>(netMessage.ReadNBytes(sizeof(bool)));

	//isFin
	m_IsFinStopped	 = *reinterpret_cast<const bool*>(netMessage.ReadNBytes(sizeof(bool)));

	m_IsPathing		 = *reinterpret_cast<const bool*>(netMessage.ReadNBytes(sizeof(bool)));


	////Atts
	m_attributes << netMessage;

	////stats
	m_stats << netMessage;


	////MarkArraySize
///	m_markerArrSize	 = *reinterpret_cast<const bool*>(netMessage.ReadNBytes(sizeof(int)));


	////markArrayHead
//	m_markerArrHead	 = *reinterpret_cast<const bool*>(netMessage.ReadNBytes(sizeof(int)));
	

	////Mark Array
/*	for(int i = 0 ; i < m_markerArrSize && i < MARKER_ARRAY_SIZE; i++)
	{
		currChunk = netMessage.ReadNBytes(sizeof(InternalPoint_type));
		memcpy(m_markerArr + i, currChunk, sizeof(InternalPoint_type));
	}


	////Mark Array weight
	for(i = 0 ; i < m_markerArrSize && i < MARKER_ARRAY_SIZE; i++)
		m_markerArrWeight[i] = *reinterpret_cast<const bool*>(netMessage.ReadNBytes(sizeof(float)));
*/
	////Action #
	unsigned int actionNumber = *reinterpret_cast<const bool*>(netMessage.ReadNBytes(sizeof(unsigned int)));

	////Actions
	//clear old actions.
	ClearActions();

	//read in the new actions sent by the stream
	for( ; actionNumber != 0 ; --actionNumber)
		m_actList.push_back(Action::CreateAction(*m_strObj, *m_hFishBody, netMessage));
		
}
#pragma warning(pop)

//See header for documentation (write)
void FishBody::operator >> (UDPNet::NetMessage& netMessage)
{
	const char *currChunk = NULL;	//temp point to store data location
	InterpPnt_type pnt_vec;			//temp pnt or vec used to store data before writing it.
	
	if(!m_pTankClient)
		return;


	////FishKey
	netMessage.WriteNBytes(sizeof(AquariusKey_type),reinterpret_cast<char*>(&m_key));

	////Location
	GetPos(pnt_vec);
	netMessage.WriteNBytes(sizeof(InterpPnt_type),reinterpret_cast<char*>(&pnt_vec[0]));


	////Facing Dir
	netMessage.WriteNBytes(sizeof(InterpVec_type),reinterpret_cast<char*>(&m_facingVector[0]));


	////Travel Dir
	GetDir(pnt_vec);
	netMessage.WriteNBytes(sizeof(InterpPnt_type),reinterpret_cast<char*>(&pnt_vec[0]));
	

	////State flags
	//isMouth
	netMessage.WriteNBytes(sizeof(bool),reinterpret_cast<char*>(&m_IsMouthStopped));

	//isAttack
	netMessage.WriteNBytes(sizeof(bool),reinterpret_cast<char*>(&m_isAttacking));

	//isEat
	netMessage.WriteNBytes(sizeof(bool),reinterpret_cast<char*>(&m_isEating));

	//isFin
	netMessage.WriteNBytes(sizeof(bool),reinterpret_cast<char*>(&m_IsFinStopped));

	//isPathing
	netMessage.WriteNBytes(sizeof(bool), reinterpret_cast<char*>(&m_IsPathing));


	////Atts
	m_attributes >> netMessage;

	////stats
	m_stats >> netMessage;


	////MarkArraySize
/*	netMessage.WriteNBytes(sizeof(int),reinterpret_cast<char*>(&m_markerArrSize));

	////markArrayHead
	netMessage.WriteNBytes(sizeof(int),reinterpret_cast<char*>(&m_markerArrHead));
	

	////Mark Array
	netMessage.WriteNBytes(sizeof(InternalPoint_type)*m_markerArrSize,
						   reinterpret_cast<char*>(m_markerArr));

	////Mark Array weight
	netMessage.WriteNBytes(sizeof(float)*m_markerArrSize,
						   reinterpret_cast<char*>(m_markerArrWeight));
*/
	////Action #
	unsigned int actionNumber = m_actList.size();
	netMessage.WriteNBytes(sizeof(unsigned int),reinterpret_cast<char*>(&actionNumber));

	////Actions
	//write each action to the net stream
	std::list<Action*>::iterator  actIT = m_actList.begin();
	for( ; m_actList.end() != actIT ; actIT++)
		(*actIT)->Write(netMessage);	
}





//See header for documentation
ostream& InterpretationModule::operator << (ostream &outStream, const FishBody &fishBody)
{
	char buffer[256] = {0};		//buffer to sprintf to.
	int i = 0;					//loop counter

	//// Header
	outStream << "#### Begin FishBody\n";


	/*
	 *	Marker Array
	 */
	outStream << "### Begin Marker Array :" << MARKER_ARRAY_SIZE << "\n";
	for(i = 0 ; i < MARKER_ARRAY_SIZE; i++)
	{
		sprintf(buffer, "(%f, %f, %f) %f\n", fishBody.m_markerArr[i].x, fishBody.m_markerArr[i].y, 
											 fishBody.m_markerArr[i].z, fishBody.m_markerArrWeight[i]);
		outStream << buffer;		
	}
	outStream << "### End Marker Array\n";



	/*
	 *	Actions
	 */
	outStream << "### Begin Action List :" << fishBody.m_actList.size() << "\n";
	list<Action*>::const_iterator actIT = fishBody.m_actList.begin();
	for( ;actIT != fishBody.m_actList.end() ;actIT++)
	{
		const Action *const currAction = *actIT;
		currAction->Write(outStream);

	}
	outStream << "### End Action List\n";


	/*
	 *	Stats
	 */
	outStream << "### Begin Stats\n";
	outStream << fishBody.m_stats;
	outStream << "### End Stats\n";
	
	/*
	 *	Attributes.
	 */
	outStream << "### Begin Attributes\n";
	outStream << fishBody.m_attributes;
	outStream << "### End Attributes\n";

	/*
	 *	Physical Data
	 */
	outStream << "### Begin Physical Data\n";
	outStream << fishBody.m_phyData;
	outStream << "### End Physical Data\n";

	/*
	 *	Sphere stuff
	 */
	outStream << "### Begin Physics Data\n";
		PhysicsSimulator& phySim = fishBody.m_pTankClient->GetPhysicsSimulatorReference();
		PhysicsSimulator::Sphere& sphere = phySim.GetSphereRef(fishBody.m_key);

		sprintf(buffer, "Position (%f, %f, %f)\n", 
			sphere.GetPosition()[0], sphere.GetPosition()[1], sphere.GetPosition()[2]);
		cout << buffer;
	
		sprintf(buffer, "Velocity (%f, %f, %f)\n", 
			sphere.GetVelocity()[0], sphere.GetVelocity()[1], sphere.GetVelocity()[2]);
		cout << buffer;
	
		sprintf(buffer, "Acceleration (%f, %f, %f)\n", 
			sphere.GetAcceleration()[0], sphere.GetAcceleration()[1], sphere.GetAcceleration()[2]);
		cout << buffer;
	outStream << "### End Physics Data\n";
	

	////footer
	outStream << "#### End FishBody\n";

	return outStream;
}

//See header for documentation
istream& InterpretationModule::operator>> (istream& inStream, FishBody &at)
{
	char buffer[256] = {0};		//buffer to sprintf to.
	int i = 0;					//loop counter

	//// Header
	inStream.getline(buffer, 512, '\n');


	/*
	 *	Marker Array
	 */
	//read in how many marker arrays are stored.
	int markerCount = 0;
	inStream.getline(buffer, 512, '\n');
	sscanf(buffer, "### Begin Marker Array :%d\n", &markerCount);

	//bounds check
	if(MARKER_ARRAY_SIZE < markerCount) 
		at.m_markerArrSize = MARKER_ARRAY_SIZE;
	else
		at.m_markerArrSize = markerCount;
	

	for(i = 0 ; i < markerCount; i++)
	{
		inStream.getline(buffer, 512, '\n');
		if(i < MARKER_ARRAY_SIZE)
			sscanf(buffer, "(%f, %f, %f) %f\n", &at.m_markerArr[i].x, &at.m_markerArr[i].y, 
											 &at.m_markerArr[i].z, &at.m_markerArrWeight[i]);
	}//end for
	inStream.getline(buffer, 512, '\n');



	/*
	 *	Actions
	 */
	//read in the header and count of how many actions there are.
	int listSize = 0;
	inStream.getline(buffer, 512, '\n');
	sscanf(buffer, "### Begin Action List :%d\n", &listSize);

	//read in each action in the file.
	for(i = 0 ; i < listSize ; i++)
	{
		Action* newAction = Action::CreateAction(*at.m_strObj, *at.m_hFishBody, inStream);

		//if we're able to load the item push it to the back of the list.
		if(newAction)
			at.m_actList.push_back(newAction);	
	}	
	inStream.getline(buffer, 512, '\n');


	/*
	 *	Stats
	 */
	inStream.getline(buffer, 512, '\n');
	inStream >> at.m_stats;
	inStream.getline(buffer, 512, '\n');

	/*
	 *	Attributes.
	 */
	inStream.getline(buffer, 512, '\n');
	inStream >> at.m_attributes;
	inStream.getline(buffer, 512, '\n');


	/*
	 *	Physical Data
	 */
	inStream.getline(buffer, 512, '\n');
	inStream >> at.m_phyData;
	InterpPnt_type pos(at.m_phyData.GetXPos(),at.m_phyData.GetYPos(),at.m_phyData.GetZPos());
	at.SetPos(pos);
	inStream.getline(buffer, 512, '\n');
	

	/*
	 *	Sphere stuff
	 */
	inStream.getline(buffer, 512, '\n');
		PhysicsSimulator& phySim = at.m_pTankClient->GetPhysicsSimulatorReference();
		PhysicsSimulator::Sphere& sphere = phySim.GetSphereRef(at.m_key);
		InterpVec_type temp;

		inStream.getline(buffer, 512, '\n');
		sscanf(buffer, "Position (%f, %f, %f)\n", &temp[0], &temp[1], &temp[2]);
		sphere.SetPosition(temp);
	
		sprintf(buffer, "Velocity (%f, %f, %f)\n", &temp[0], &temp[1], &temp[2]);
		sphere.SetVelocity(temp);
	
		sprintf(buffer, "Acceleration (%f, %f, %f)\n", &temp[0], &temp[1], &temp[2]);
		sphere.SetAcceleration(temp);
	inStream.getline(buffer, 512, '\n');;



	

	////footer
	inStream.getline(buffer, 512, '\n');
	
	return inStream;
}



/*******************************************************************************
 * More Help Functions.                                                             
 *//////////////////////////////////////////////////////////////////////////////


////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	CommandID_BrainToBody
/*!	\brief		Converts an ai command ID into a body command id.
**	\param		actID		ID of the action in the brain format.
**	\return		ID of the command in body format.
**	\exception	none		
**	\author		Nate Cleveland
**	\date		4/2/2003 11:31:50 AM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
CommandID_type CommandID_BrainToBody(AI_System::Attribute_ID actID)
{
	using namespace AI_System;
	
	switch(actID)
	{
	case MOVETO:
		return AQUARIUS_COMMAND_GOTO;

	case FLEE:
		return AQUARIUS_COMMAND_FLEE;

	case FOLLOW:
		return AQUARIUS_COMMAND_FOLLOW;

	case MOVEWITH:
		return AQUARIUS_COMMAND_MOVE_WITH;

	case ROAM:
		return AQUARIUS_COMMAND_ROAM;

	case GOTO_POS_MARKER:
		return AQUARIUS_COMMAND_EVALUATE_POSITIVE;

	case FLEE_NEG_MARKER:
		return AQUARIUS_COMMAND_EVALUATE_NEGATIVE;

	case IGNORE_MARKERS:
		return 0;

	case HEED_MARKERS:
		return AQUARIUS_COMMAND_EVALUATE_MARKERS;

	case ATTACK:
		return AQUARIUS_COMMAND_ATTACK;

	case EAT:
		return AQUARIUS_COMMAND_EAT;

	case PICKUP:
		return AQUARIUS_COMMAND_PICK_UP;

	case DROP:
		return AQUARIUS_COMMAND_DROP_OBJECT;

	case DROP_POS_MARKER:
		return AQUARIUS_COMMAND_DROP_MARKER;

	case DROP_NEG_MARKER:
		return AQUARIUS_COMMAND_DROP_MARKER;

	case EVALUATEMARKER:
		return AQUARIUS_COMMAND_EVALUATE_MARKERS;

	default:
		return 0;
	}
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	CommandID_BodyToBrain
/*!	\brief		Converts a fish body command ID to a brain one.
**	\param		cmdID		ID of a fish body command
**	\return		ID of the command in brain format.
**	\exception	none				
**	\author		Nate Cleveland
**	\date		4/2/2003 11:33:28 AM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
AI_System::Action_ID CommandID_BodyToBrain(CommandID_type cmdID)
{
	using namespace AI_System;

	switch(cmdID)
	{
	case AQUARIUS_COMMAND_GOTO:
		return MOVETO;

	case AQUARIUS_COMMAND_FLEE:
		return FLEE;

	case AQUARIUS_COMMAND_FOLLOW:
		return FOLLOW;

	case AQUARIUS_COMMAND_MOVE_WITH:
		return MOVEWITH;

	case AQUARIUS_COMMAND_ROAM:
		return ROAM;

	case AQUARIUS_COMMAND_EVALUATE_POSITIVE:
		return GOTO_POS_MARKER;

	case AQUARIUS_COMMAND_EVALUATE_NEGATIVE:
		return FLEE_NEG_MARKER;

//	case IGNORE_MARKERS:
//		return 0;

//	case AQUARIUS_COMMAND_EVALUATE_MARKERS:
//		return HEED_MARKERS;

	case AQUARIUS_COMMAND_ATTACK:
		return ATTACK;

	case AQUARIUS_COMMAND_EAT:
		return EAT;

	case AQUARIUS_COMMAND_PICK_UP:
		return PICKUP;

	case AQUARIUS_COMMAND_DROP_OBJECT:
		return DROP;

	case AQUARIUS_COMMAND_DROP_MARKER:
		return DROP_POS_MARKER;

//	case AQUARIUS_COMMAND_DROP_MARKER:
//		return DROP_NEG_MARKER;

	case AQUARIUS_COMMAND_EVALUATE_MARKERS:
		return EVALUATEMARKER;

	default:
		return AC_NULL;
	}
}
