////////////////////////////////////////////////////////////////////////////////
/*!	\file		fishBody.hpp
**	\brief		Declares the FishBody class.
**	\author		Nate Cleveland
**	\date		10/25/2002 11:39:32 AM
**	\par		DESCRIPTION: The fish body class is currently the only object that
**				 can interpret commands sent from the brain to the body. It is broken
**				 down into several different parts.
**				 - Fish Stat/Attribute management. (PhyData, Statistics, Attributes)
**				 - Positional/physics related data storage. (undefined.)
**				 - Command interpretation. (list of Action *).
**				 - Example Generation. (undefined)
**
**				Once per frame the Update method must be called so object can have
**				 chance to process its environment. At that time it will act upon
**				 commands, update statistics, alter direction if needed, etc.
**
**
**	\par		REMARKS:
**
*/
////////////////////////////////////////////////////////////////////////////////
#pragma once

#pragma warning(push)			//save the warning state before compiling this header.
#pragma warning(disable: 4786)	//disable warning 4786 about truncating browse names.
#pragma warning(disable: 4018)	//disable warning about comparing signed/unsigned values
								// (this is an stlPort issue, not my fault).

#include "fishBodyUtils.hpp"
#include "fishBodyMath.hpp"
#include "updateTimer.hpp"
#include "commandID.hpp"
#include "../fileio/memento.h"
#include "attributes.hpp"
#include "attributeGenerator.hpp"
#include "stats.hpp"
#include "physicData.hpp"
#include "graphicClient.hpp"
#include "../sound/bodyClient.h"
#include <stdlib.h>
#include <list>
#include <vector>


//Forward Declarations.
class TankSimulatorClient;
class PhysicsSimulator;
namespace AI_System 
{
	struct AI_Action;
	struct AI_ActionTag;
}
namespace UDPNet 
{
	class NetMessage;
};



namespace InterpretationModule
{
	//Namespace specific namespace forward declarations.
	class HFishBody;
	class Action;
	class CommandPacket;
	class StrObj;
	struct FrozenFish;

	const int MARKER_ARRAY_SIZE = 10;
	const int ACT_LIST_DEFUALT_MAX_SIZE = 12;
	extern const float MAX_TURN_VALUE_XZ_PER_SEC;
	extern const float MAX_TURN_VALUE_XY_PER_SEC;
	extern const float MAX_TURN_VALUE_ZY_PER_SEC;
	extern const float MATING_INCLINE_THRESHOLD;	//!<Once the fish has a incline of
													//!< this or more it will add itself
													//!< to tank simulator's mating list.
	extern const float BASE_RADIUS;					//!<Defualt radius of our module and 
													//!< spherical representation in the
													//!< tank simulator.

	////////////////////////////////////////////////////////////////////////////////
	/*!	\class		FishBody
	**	\brief		Mother class of the InterpretationModule.
	**	\author		Nate Cleveland
	**	\date		10/28/2002 9:40:01 AM
	**	\warning	none
	*/
	////////////////////////////////////////////////////////////////////////////////
	class FishBody: public FileMemento, public AcqariusSound::BodyClient
	{
	protected:
		//Key data
		AquariusKey_type	m_key;			//!< Key for this FishBody.
		AquariusKey_type	m_sphereKey;	//!< Key for our sphere in physicsSimulation.


		//next four members are for the marker system.
		InternalPoint_type	m_markerArr[MARKER_ARRAY_SIZE];		//!< An array of makers
																//!<  that the brain
																//!<  may tell the body
																//!<  to drop.
		float			m_markerArrWeight[MARKER_ARRAY_SIZE];	//!< Weight values that
																//!<  coincide with the
																//!<  markerArr elements.
		int					m_markerArrHead;					//!< Index to next place
																//!<  to insert into the
																//!<  marker array.
		int					m_markerArrSize;					//!< Number of elements
																//!<  currently in the
																//!<  marker array.

		//State flags. First three flags are mutually exclusive
		bool		m_IsMouthStopped;		//!< If true will not allow attacking/eating.
		bool		m_isAttacking;			//!< Set to true if the fish is attacking.
		bool		m_isEating;				//!< Set to true if the fish is eating.
		bool		m_IsFinStopped;			//!< If true the fish may not move itself.
		bool		m_IsLearn;				//!< If set to false fish does not store examples.
		bool		m_IsPathing;			//!< If true the fish will try to pathfind

		//update logic timer
		UpdateTimer<float>	m_updateTimer;	//! Object that controls update logic.

		//Action related members
		std::list<Action*>	m_actList;			//!<List of all the actions we're currently
												//!<  acting on.
		unsigned int		m_actListMaxSize;	//!< Maximum size of the act list.
		std::vector<CommandPacket>	m_incomingCommands;	//!< Place to hold all incoming
														//!<  commands until the brain
														//!<  has sent all commands for
														//!<  this period.
		HFishBody*	m_hFishBody;				//!< Interface for the Action objects to
												//!<  communicate through.

		//Example related members
		std::vector<AI_System::AI_ActionTag> m_examples;	//!< A vector of all the 
															//!<  examples this body
															//!<  has generated.


		//Physic Related members
		InternalVector_type	m_currVector;	//!< Current vector fish is moving along.
		InterpVec_type		m_facingVector;	//!< If our direction is zero then we store the facing vector here.
		TankSimulatorClient*m_pTankClient;	//!< Pntr to tank simulator client.
		float				m_speedRatio;	//!< currSpeed/maxSpeed.


		//Statistic Members
		Attributes	m_attributes;			//! Object that manages all Fish Attributes.
		Statistics	m_stats;				//! Object that holds all stats of this fish.
		PhysicData	m_phyData;				//! Records physical data for Attributes/stats

		//Graphical
		GraphicClient m_graphicClient;		//! Instance of our graphic client.


		////////////////////////////////////////////////////////////////////////////////
		//	FUNCTION:	
		/*!	\brief		Empties the entire action list.
		**	\param		none
		**	\return		none
		**	\exception	none					
		**	\author		Nate Cleveland
		**	\date		2/6/2003 5:16:00 PM
		**	\warning	none
		*/
		////////////////////////////////////////////////////////////////////////////////
		void ClearActions(void) throw();
			


		
		////////////////////////////////////////////////////////////////////////////////
		//	FUNCTION:	
		/*!	\brief		Ensures that physic vector and fish vector are synched up.

		**	\param		
		**	\return		Returns true if fish is facing the same direction its traveling
		**	\exception	iff unable to get the phyVelocity member.
		**	\par		DESCRIPTION:
							If a fish collides with an object it is shot off away from the
							 object of collision. When this happens the graphical direction
							 of the fish and the direction it is actually going lose synch.
							 This method compensates for this by testing to see if their
							 is a disparency between m_currVector and the vector retained 
							 by the physicsSimulator. Object check to make sure we have a
							 tankClient before accessing any data.
		**					
		**	\author		Nate Cleveland
		**	\date		1/16/2003 11:42:01 AM
		**	\warning	none
		*/
		////////////////////////////////////////////////////////////////////////////////
		void VectorSynch(void);


		//Generates a potential field for avoiding terrain in the game.
		////////////////////////////////////////////////////////////////////////////////
		//	FUNCTION:	
		/*!	\brief		Generates Terrain Field for ths fish to avoid.
		**	\param		terrainField	This value will hold the summation of all potential 
		**								 fields generated by the method.
		**	\return		none
		**	\exception	none
		**	\par		DESCRIPTION:
		**				Uses RayToField to generate the potential fields that will guide 
		**				 the fish around obsticles.
		**	\author		Nate Cleveland
		**	\date		11/8/2002 10:21:32 AM
		**	\warning	none
		*/
		////////////////////////////////////////////////////////////////////////////////
		void GenerateTerrainField(InterpVec_type& terrainField) const throw();

		//Helper method for GenerateTerrainField.
		////////////////////////////////////////////////////////////////////////////////
		//	FUNCTION:	
		/*!	\brief		Casts a ray and generates a single Field for GenerateTerrainField().
		**	\param		x		Direction of ray to cast along x-axis.
		**	\param		y		Direction of ray to cast along y-axis.
		**	\param		z		Direction of ray to cast along z-axis.
		**	\param		field	Newly generated field is added to this parameter.
		**	\return		none
		**	\exception	none
		**	\par		DESCRIPTION:
		**				Uses the PhysicSimulator RayCasting method to generate a terrain
		**				 field for the fish to avoid.
		**
		**	\todo		This method will not function until the PhysicSimulator ray casting
		**				 method is written.
		**
		**	\author		Nate Cleveland
		**	\date		11/8/2002 10:18:05 AM
		**	\warning	none
		*/
		////////////////////////////////////////////////////////////////////////////////
		void RayToField(float x, float y, float z, InterpVec_type& field, InterpPnt_type& pos,
						PhysicsSimulator& phySim) const throw();


		//Adds a marker to this fish's circular queue of markers.
		////////////////////////////////////////////////////////////////////////////////
		//	FUNCTION:
		/*!	\brief		Adds a marker to this fish's circular queue of markers.
		**
		**	\param const  InterpPoint_type& point	Position of marker.
		**	\param float						 weight Weight of marker.
		**
		**	\return		none
		**	\exception	none
		**	\par		DESCRIPTION:
		**				 Only makes a distinction of positive and negative markers by
		**				 their wieght.
		**
		**	\author		Nate Cleveland
		**	\date		10/28/2002 9:22:30 AM
		**	\warning	none
		*/
		////////////////////////////////////////////////////////////////////////////////
		void AddMarker(const InternalPoint_type& point, float weight);

		////////////////////////////////////////////////////////////////////////////////
		//	FUNCTION:
		/*!	\brief		Returns to the caller the current direction Vector of this fish.
		**	\param		Vector to fill with current direction of this fish.
		**	\return		none
		**	\exception	none
		**	\par		DESCRIPTION:
		**
		**	\author		Nate Cleveland
		**	\date		10/28/2002 9:21:44 AM
		**	\warning	none
		*/
		////////////////////////////////////////////////////////////////////////////////
		void GetCurrVector(InternalVector_type& vec) const throw(){vec = m_currVector;}

		////////////////////////////////////////////////////////////////////////////////
		//	FUNCTION:	
		/*!	\brief		Returns to the caller a reference to another fish body.
		**	\param		key		ID of the fish to gain a reference to.
		**	\return		Returns false if unable to find fish.
		**	\exception	none
		**	\par		DESCRIPTION:
		**				Allows the Action classes, through their hFishBody member, to 
		**				 gain access to other fish.
		**
		**	\author		Nate Cleveland
		**	\date		11/4/2002 9:43:23 AM
		**	\warning	none
		*/
		////////////////////////////////////////////////////////////////////////////////
		bool GetFishBody(AquariusKey_type key, FishBody& fishBody) throw();
		FishBody* GetFishBodyPtr(AquariusKey_type key) throw();

		////////////////////////////////////////////////////////////////////////////////
		//	FUNCTION:	
		/*!	\brief		Returns true if this object is a fish.
		**	\param		key		ID of the object to check.
		**	\return		True if this is a key of a fish.
		**	\exception	none
		**	\par		DESCRIPTION:
		**				Checks to be sure that this id is that of a fish.	
		**	\author		Nate Cleveland
		**	\date		11/4/2002 9:47:50 AM
		**	\warning	none
		*/
		////////////////////////////////////////////////////////////////////////////////
		bool IsFishKey(AquariusKey_type key)
		{
			return key != 0;
		}


		//removes any finished actions. Generates examples from them.
		////////////////////////////////////////////////////////////////////////////////
		//	FUNCTION:
		/*!	\brief		Removes any finished actions from m_actList.
		**	\param		none
		**	\return		none
		**	\exception	none
		**	\par		DESCRIPTION:
		**				 Removes any finihsed actions from m_actList and generates
		**				 examples from them.
		**
		**	\author		Nate Cleveland
		**	\date		10/28/2002 9:10:59 AM
		**	\warning	none
		*/
		////////////////////////////////////////////////////////////////////////////////
		void Prune(void) throw();
		
		//Sucks plankton from the environment.
		////////////////////////////////////////////////////////////////////////////////
		//	FUNCTION:	
		/*!	\brief		Sucks plankton from the environment.
		**	\param		none
		**	\return		none
		**	\exception	none
		**	\par		DESCRIPTION:
		**				 Calls the tank simulator and sucks plankton from the environment.	
		**	\author		Nate Cleveland
		**	\date		11/26/2002 3:18:54 PM
		**	\warning	none
		*/
		////////////////////////////////////////////////////////////////////////////////
		void Suck(void) throw();

		//Adds chum to the environment from this fish's position
		////////////////////////////////////////////////////////////////////////////////
		//	FUNCTION:	
		/*!	\brief		Adds chum to the environment at this fish's position.
		**	\param		chumAmmount		How much chum you wish to add.
		**	\return		none
		**	\exception	none
		**	\par		DESCRIPTION:
		**					
		**	\author		Nate Cleveland
		**	\date		1/21/2003 1:21:53 PM
		**	\warning	none
		*/
		////////////////////////////////////////////////////////////////////////////////
		void AddChum(float chumAmmount) throw();

		//updates the PhyData, Attributes and Statistics classes.
		////////////////////////////////////////////////////////////////////////////////
		//	FUNCTION:
		/*!	\brief		Updates all Statistics, Fish Attributes and PhyData classes.
		**
		**	\param float elapsedTime		Time in seconds since last update.
		**
		**	\return		none
		**	\exception	none
		**	\par		DESCRIPTION:
		**				 Updates all Statistics, Fish Attributes and PhyData classes.
		**
		**	\author		Nate Cleveland
		**	\date		10/28/2002 9:12:10 AM
		**	\warning	none
		*/
		////////////////////////////////////////////////////////////////////////////////
		void UpdateFishStats(float elapsedTime) throw();

		//updates and prunes actions
		////////////////////////////////////////////////////////////////////////////////
		//	FUNCTION:
		/*!	\brief		Updates all fish actions.
		**	\param InterpVector_type& summationVec	Var will be filled with the
		**												 summation of all the specified
		**												 direction vectors.
		**	\return		none
		**	\exception	none
		**	\par		DESCRIPTION:
		**				 Calls update on all actions. Also calls method FishBody::Prune.
		**
		**	\author		Nate Cleveland
		**	\date		10/28/2002 9:14:48 AM
		**	\warning	none
		*/
		////////////////////////////////////////////////////////////////////////////////
		void UpdateActions(InternalVector_type& summationVec, float elapsedTime) throw();

		//updates all the potential field data
		////////////////////////////////////////////////////////////////////////////////
		//	FUNCTION:
		/*!	\brief		Updates all potential field data.
		**	\param InterpVector_type& summationVec	Var filled by FishBody::
		**												 UpdateActions
		**	\return		none
		**	\exception	none
		**	\par		DESCRIPTION:
		**				 Cruches all the potential fields together to produce movement
		**				 for a FishBody.
		**
		**	\author		Nate Cleveland
		**	\date		10/28/2002 9:17:46 AM
		**	\warning	none
		*/
		////////////////////////////////////////////////////////////////////////////////
		void UpdatePotenFields(InternalVector_type& summationVec) throw();

		//updates any physics related data sans-PhyData member
		////////////////////////////////////////////////////////////////////////////////
		//	FUNCTION:
		/*!	\brief		Updates all physics related data except PhyData.
		**	\param		summationVector		Final answer for the direction we want to go.
		**	\return		none
		**	\exception	none
		**	\par		DESCRIPTION:
		**				 Does the following.
		**					- Limits the fish velocity (see LimitVelocity).
		**					- Limits the ammount a fish can turn (See LimitTurn).
		**					- Calculates the rotation value of the turn (See LimitTurn).
		**					- Calculates the Swimming energyLoss (See SwimEnergyLoss).
		**					- Updates our spherical representation in the Physics 
		**					Simulator (See FormatSphere).
		**
		**	\author		Nate Cleveland
		**	\date		10/28/2002 9:20:30 AM
		**	\warning	none
		*/
		////////////////////////////////////////////////////////////////////////////////
		void UpdatePhysical(InternalVector_type& summationVec, float elapsedTime) throw();



		////////////////////////////////////////////////////////////////////////////////
		//	FUNCTION:	
		/*!	\brief		Calculates the energy lost through swimming.
		**	\param		elapsedTime		Seconds since last update.
		**	\return		none
		**	\exception	none
		**	\par		DESCRIPTION:
		**				 Uses the Attributes and Atrribute Generator classes to find 
		**				  how much energy should be deducted for swimming.	
		**	\author		Nate Cleveland
		**	\date		11/25/2002 10:01:27 AM
		**	\warning	none
		*/
		////////////////////////////////////////////////////////////////////////////////
		void SwimEnergyLoss(float elapsedTime) throw();

		////////////////////////////////////////////////////////////////////////////////
		//	FUNCTION:	
		/*!	\brief		Fills out the dimensions of the sphere used to represent this fish.
		**	\param		none
		**	\return		none
		**	\exception	none
		**	\par		DESCRIPTION:
		**					
		**	\author		Nate Cleveland
		**	\date		11/25/2002 10:06:19 AM
		**	\warning	none
		*/
		////////////////////////////////////////////////////////////////////////////////
		void FormatSphere(InterpVec_type& newDir) throw();

		////////////////////////////////////////////////////////////////////////////////
		//	FUNCTION:	
		/*!	\brief		Limits how fast a fish can move.
		**	\param		summationVec		Finalized potential field vector.
		**	\return		none
		**	\exception	none
		**	\par		DESCRIPTION:
		**				Takes the finalized potential field vector and makes sure the 
		**				 magnitude isn't greater then maxSpeed.	
		**	\author		Nate Cleveland
		**	\date		11/8/2002 10:33:13 AM
		**	\warning	none
		*/
		////////////////////////////////////////////////////////////////////////////////
		void LimitVelocity(InterpVec_type& summationVec, float& speedRatio)const throw();


	public:
		const Attributes* m_pAttributes;	//! Quick way to view fish Attributes
		const Statistics* m_pStats;			//! Quick way to view fish stats
		StrObj*		m_strObj;					//!< Member actions write themselves to.

		

		//HFishBody class is the only interface an action has with its fish body.
		friend class HFishBody;
		friend struct FrozenFish;

		////////////////////////////////////////////////////////////////////////////////
		//	FUNCTION:
		/*!	\brief		Defualt ctor for FishBody
		**	\param
		**	\return		none
		**	\exception	none
		**	\par		DESCRIPTION:
		**
		**	\author		Nate Cleveland
		**	\date		10/28/2002 9:34:30 AM
		**	\warning	none
		*/
		////////////////////////////////////////////////////////////////////////////////
		FishBody();
		void Build(const FrozenFish& frozenFish) throw();

		////////////////////////////////////////////////////////////////////////////////
		//	FUNCTION:
		/*!	\brief	
		**	\param
		**	\return		none
		**	\exception	none
		**	\par		DESCRIPTION:
		**
		**	\author		Nate Cleveland
		**	\date		10/28/2002 9:34:33 AM
		**	\warning	none
		*/
		////////////////////////////////////////////////////////////////////////////////
		~FishBody();

		////////////////////////////////////////////////////////////////////////////////
		//	FUNCTION:	
		/*!	\brief		Alters the Attributes of this Fish to reflect its parents.
		**	\param		parentA		Key to find the parent fish with.
		**	\param		parentB		Key to find another parent with.
		**	\return		Returns false if unable to find the two fish.
		**	\exception	none
		**	\par		DESCRIPTION:
		**				 Takes the two given Fish and sets its Attributes to reflect a
		**					combination of them.
		**				  
		**	\author		Nate Cleveland
		**	\date		11/27/2002 5:56:24 PM
		**	\warning	none
		*/
		////////////////////////////////////////////////////////////////////////////////
		bool Breed(AquariusKey_type parentA, AquariusKey_type parentB) throw();

		////////////////////////////////////////////////////////////////////////////////
		//	FUNCTION:	
		/*!	\brief		Alters the Attributes of this fish to reflect its parent.
		**	\param		parent		ID of this fish's parent.
		**	\return		False if unable to find the parent.
		**	\exception	none
		**	\par		DESCRIPTION:
		**				 Takes the stats of the given fish, slightly mutates them, and
						  sets that as its attributes.
		**	\author		Nate Cleveland
		**	\date		11/27/2002 6:01:15 PM
		**	\warning	none
		*/
		////////////////////////////////////////////////////////////////////////////////
		bool Breed(AquariusKey_type parent) throw();

		////////////////////////////////////////////////////////////////////////////////
		//	FUNCTION:	
		/*!	\brief		Alters the Attributes dramatically.
		**	\param		none
		**	\return		none
		**	\exception	none
		**	\par		DESCRIPTION:
		**				 Randomly mutates this fish's Attributes.	
		**	\author		Nate Cleveland
		**	\date		11/27/2002 6:02:09 PM
		**	\warning	none
		*/
		////////////////////////////////////////////////////////////////////////////////
		void Breed(void) throw();

		////////////////////////////////////////////////////////////////////////////////
		//	FUNCTION:	
		/*!	\brief		Forces the image to be drawn to the screen imediately.	
		**	\param		none
		**	\return		none
		**	\exception	none
		**	\par		DESCRIPTION:
		**				 Draws the image to the screen instantly via use of the 
						  m_graphics client.
		**	\author		Nate Cleveland
		**	\date		11/1/2002 11:21:21 AM
		**	\warning	Only suggested for use in testing programs.
		*/
		////////////////////////////////////////////////////////////////////////////////
		void ForceDraw(void) throw();

		////////////////////////////////////////////////////////////////////////////////
		//	FUNCTION:	
		/*!	\brief		Adds the images to the graphic systems draw list.
		**	\param		none
		**	\return		none
		**	\exception	none
		**	\par		DESCRIPTION:
		**				 Uses the fishBody graphics client to add this FishBody to the
						  graphic module's draw list.
		**	\author		Nate Cleveland
		**	\date		11/1/2002 11:21:21 AM
		**	\warning	Only suggested for use in testing programs.
		*/
		////////////////////////////////////////////////////////////////////////////////
		void QueueDraw(void) throw();

		////////////////////////////////////////////////////////////////////////////////
		//	FUNCTION:	
		/*!	\brief		Computes most the sound logic.
		**	\param		none
		**	\return		none
		**	\exception	none
		**	\par		DESCRIPTION:
		**				 Checks for and initiates the following sounds
		**					- Swim
		**					- Collision (body and terrain).
		**					
		**	\author		Nate Cleveland
		**	\date		12/4/2002 1:15:15 PM
		**	\warning	none
		*/
		////////////////////////////////////////////////////////////////////////////////
		void SoundLogic(void) throw();

		////////////////////////////////////////////////////////////////////////////////
		//	FUNCTION:	
		/*!	\brief		Updates graphic logic.
		**	\param		none
		**	\return		none
		**	\exception	none
		**	\par		DESCRIPTION:
		**				 Updates the Graphic Client's information on the fish (ie health,
						  armor, meta, size, etc).
		**					
		**	\author		Nate Cleveland
		**	\date		12/4/2002 1:15:15 PM
		**	\warning	none
		*/
		////////////////////////////////////////////////////////////////////////////////
		void GraphicLogic(void) throw();


		////////////////////////////////////////////////////////////////////////////////
		//	FUNCTION:	
		/*!	\brief		Sets this fish's key to the provided value.
		**	\param		key		New value of the key.
		**	\return		none
		**	\exception	none
		**	\par		DESCRIPTION:
		**				 Sets this fish's key to the provided value.
		**	\author		Nate Cleveland
		**	\date		11/1/2002 1:21:43 PM
		**	\warning	none
		*/
		////////////////////////////////////////////////////////////////////////////////
		void SetKey(AquariusKey_type key)throw(){m_key = key;}

		////////////////////////////////////////////////////////////////////////////////
		//	FUNCTION:	
		/*!	\brief		Returns to the caller this fish's current key value.
		**	\param		none
		**	\return		none
		**	\exception	none
		**	\par		DESCRIPTION:
		**				 Returns to the caller this fish's current key value.	
		**	\author		Nate Cleveland
		**	\date		11/1/2002 1:22:49 PM
		**	\warning	none
		*/
		////////////////////////////////////////////////////////////////////////////////
		AquariusKey_type GetKey(void)const throw(){return m_key;}


		//Gives the client a pntr to a TankSimulatorClient
		////////////////////////////////////////////////////////////////////////////////
		//	FUNCTION:	
		/*!	\brief		Sets the tank simulator client pntr to the specified value.
		**	\param		tankClient		Pntr to a tank client.
		**	\return		Returns false if call to PhysicSimulator::AddSphere fails. If 
		**				 false is returned the passed client pointer is forgotten by
		**				 this object.
		**	\exception	none
		**	\par		DESCRIPTION:
		**				Must be called before the update method is called.	
		**	\author		Nate Cleveland
		**	\date		11/4/2002 3:45:44 PM
		**	\warning	none
		*/
		////////////////////////////////////////////////////////////////////////////////
		bool SetClient(TankSimulatorClient *pTankClient) throw();

		////////////////////////////////////////////////////////////////////////////////
		//	FUNCTION:	
		/*!	\brief		Adds this fish to the TankSimulator mating list. 
		**	\param		none
		**	\return		none
		**	\exception	none
		**	\par		DESCRIPTION:
		**				 Adds this fish to the TankSimulator mating list. Resets the 
		**				  Mating inclination to zero.
		**	\author		Nate Cleveland
		**	\date		11/27/2002 12:45:18 PM
		**	\warning	Resets the mating inclination of this fish to zero.
		*/
		////////////////////////////////////////////////////////////////////////////////
		void AddToMatingList(void) throw();



		//Forces the position of this fish to move.
		////////////////////////////////////////////////////////////////////////////////
		//	FUNCTION:	
		/*!	\brief		Must be called after instantiation so internal fish logic is 
		**				 preserved.
		**	\return		none
		**	\exception	none
		**	\par		DESCRIPTION:
		**				Forces the position of the fish to match that of its mathching
		**				 sphere class. This must be under the following circumstances:
		**					- If the fish ever teleports (unlikely given our game design).
		**					- After it has been given its key and needs to configure
		**					its internal position logic.
		**					
		**	\author		Nate Cleveland
		**	\date		11/4/2002 11:41:38 AM
		**	\warning	The Fish will not be given credit for its distance traveled if this
		**				 method is called.
		*/
		////////////////////////////////////////////////////////////////////////////////
		void ForcePosition(void) throw();

		////////////////////////////////////////////////////////////////////////////////
		//	FUNCTION:	
		/*!	\brief		Returns to the caller the key to the sphere that represents this
		**				 object in the physics simulation.
		**	\param		none
		**	\return		The key used by this object or zero if it doens't yet have one.
		**	\exception	none
		**	\par		DESCRIPTION:
		**				Returns to the caller the key to the sphere that represents this
		**				 object in the physics simulation. Will return zero if it does 
		**				 not yet have a sphere.
		**	\author		Nate Cleveland
		**	\date		11/6/2002 10:04:16 AM
		**	\warning	none
		*/
		////////////////////////////////////////////////////////////////////////////////
		AquariusKey_type GetSphereKey(void) const throw()
		{
			if(m_pTankClient)			//if this is null we don't have a key yet.
				return m_sphereKey;
			return 0;
		}


		////////////////////////////////////////////////////////////////////////////////
		//	FUNCTION:	
		/*!	\brief		Takes a vector of AI_Action * and converts them to commandPacks
		**				 before giving them to the FishBody class.
		**	\param		vec			Reference to a vector of AI_Action* to convert to 
		**							 CommandPackets.
		**	\return		Number of elements successfully added to internal list.
		**	\exception	none
		**	\par		DESCRIPTION:
		**				Converts each element in the vector to a CommandPacket before
		**				 adding it to the internal container. Commands are not acted upon 
		**				 until the next period. Automatically calls FishBody::Flush(void) 
		**				 for the user.
		**	\author		Nate Cleveland
		**	\date		11/5/2002 1:38:47 PM
		**	\warning	none
		*/
		////////////////////////////////////////////////////////////////////////////////
		int GiveAICommand(const std::vector<AI_System::AI_Action>& vec) throw();


		////////////////////////////////////////////////////////////////////////////////
		//	FUNCTION:	
		/*!	\brief		Quick and dirty tells the fish to go to the specified point.
		**	\param		targetPnt		pnt the fish is to go to.
		**	\return		returns false if there is a problem.
		**	\exception	none
		**	\par		DESCRIPTION:
		**					Automatically flushes the command buffer.
		**	\author		Nate Cleveland
		**	\date		2/19/2003 9:12:47 AM
		**	\warning	No useable examples will be generated by this.
		*/
		////////////////////////////////////////////////////////////////////////////////
		bool GotoPoint(const InterpPnt_type& targetPnt) throw();

		////////////////////////////////////////////////////////////////////////////////
		//	FUNCTION:	
		/*!	\brief		Informs the caller whether or not the fish is storing examples.
		**	\return		boolean value, true == fish stores examples.
		**	\exception	none				
		**	\author		Nate Cleveland
		**	\date		4/2/2003 11:23:57 AM
		**	\warning	none
		*/
		////////////////////////////////////////////////////////////////////////////////
		bool IsLearn(void)const throw(){return m_IsLearn;}

		////////////////////////////////////////////////////////////////////////////////
		//	FUNCTION:	
		/*!	\brief		Sets fish's internal flag that turns on and off example storage.
		**	\param		isLearn		true == stores examples
		**	\return		none
		**	\exception	none				
		**	\author		Nate Cleveland
		**	\date		4/2/2003 11:25:11 AM
		**	\warning	none
		*/
		////////////////////////////////////////////////////////////////////////////////
		void IsLearn(bool isLearn) throw() {m_IsLearn = isLearn;}


		////////////////////////////////////////////////////////////////////////////////
		//	FUNCTION:
		/*!	\brief		Fills out a command packet and stores it inside the FishBody.
		**	\param		See commandPacket.hpp for detailes on parameters.
		**	\return		Will return false if we already have the maximum ammount of
		**				 incoming commands or if method is unable to generate the
		**				 commandPacket from given parameters.
		**	\exception	none
		**	\par		DESCRIPTION: This command is not acted upon until the next period.
		**				 To begin the next period the method "Flush" must be called.
		**
		**	\author		Nate Cleveland
		**	\date		10/25/2002 10:25:08 AM
		**	\warning	none
		*/
		////////////////////////////////////////////////////////////////////////////////
		bool Command(CommandID_type cmdID, AquariusKey_type senderID, AquariusKey_type bodyID,
					 float weight, unsigned int decisionID, float fitness, 
					 AquariusKey_type targetKey) throw();

		////////////////////////////////////////////////////////////////////////////////
		//	FUNCTION:
		/*!	\brief		Adds a command to the incoming command list.
		**
		**	\param CommandPacket&		Reference to the commandPacket that will be
		**									 added to list of incoming commands.
		**
		**	\return		Will return false if we already have the maximum ammount of
		**				 incoming commands.
		**	\exception	none
		**	\par		DESCRIPTION: This command is not acted upon until the next period.
		**				 To begin the next period the method "Flush" must be called.
		**
		**	\author		Nate Cleveland
		**	\date		10/25/2002 10:25:08 AM
		**	\warning	none
		*/
		////////////////////////////////////////////////////////////////////////////////
		bool Command(const CommandPacket& cmdPack) throw();
		

		////////////////////////////////////////////////////////////////////////////////
		//	FUNCTION:
		/*!	\brief		Adds a vector of commands to the incoming command list.
		**
		**	\param		vec			Reference to a vector of commandPackets that will be 
		**							 added to list of incoming commands.
		**
		**	\return		Returns the number of elements successfully added to internal list.
		**
		**	\exception	none
		**	\par		DESCRIPTION: 
		**				Commands are not acted upon until the next period.
		**				 To begin the next period the method "Flush" must be called.
		**
		**	\author		Nate Cleveland
		**	\date		10/25/2002 10:25:08 AM
		**	\warning	none
		*/
		////////////////////////////////////////////////////////////////////////////////
		int Command(const std::vector<CommandPacket>&vec) throw();



		////////////////////////////////////////////////////////////////////////////////
		//	FUNCTION:
		/*!	\brief		Processes all commands.
		**	\param		none
		**	\return		none
		**	\exception	none
		**	\par		DESCRIPTION: Takes all incoming commands and melds them with
		**				 the currently acting actionds.
		**
		**	\author		Nate Cleveland
		**	\date		10/25/2002 10:44:41 AM
		**	\warning	none
		*/
		////////////////////////////////////////////////////////////////////////////////
		void Flush(void) throw();


		////////////////////////////////////////////////////////////////////////////////
		//	FUNCTION:
		/*!	\brief		Processes all commands and take a list of commands.
		**	\param const  vector<>&	Reference to a vector of commandPackets that
		**									 will be added to list of commands.
		**	\return		none
		**	\exception	none
		**	\par		DESCRIPTION: Takes all incoming commands and melds them with
		**				 the currently acting actionds.
		**
		**	\author		Nate Cleveland
		**	\date		10/25/2002 10:44:41 AM
		**	\warning	none
		*/
		////////////////////////////////////////////////////////////////////////////////
		void Flush(const std::vector<CommandPacket>&cmdPakVec) throw()
		{
			Command(cmdPakVec);
			Flush();
		}

		////////////////////////////////////////////////////////////////////////////////
		//	FUNCTION:
		/*!	\brief		Generates an example from the given Action class.
		**	\param const Action*		Pointer to the action to learn from.
		**	\return		none
		**	\exception	none
		**	\par		DESCRIPTION:
		**				 All actions generated will be stored in a internal container. 
		**
		**	\author		Nate Cleveland
		**	\date		10/25/2002 1:49:31 PM
		**	\warning	none
		*/
		////////////////////////////////////////////////////////////////////////////////
		void GenerateExample(const Action* action) throw();


		////////////////////////////////////////////////////////////////////////////////
		//	FUNCTION:
		/*!	\brief		Fills given container with any generated examples.
		**	\param		vec		Container to fill with examples.
		**	\return		none
		**	\exception	none
		**	\par		DESCRIPTION:
		**				Gives the caller all the current examples that this module has
		**				 generated since the last call to Dump().
		**
		**	\author		Nate Cleveland
		**	\date		10/25/2002 1:53:09 PM
		**	\warning	none
		*/
		////////////////////////////////////////////////////////////////////////////////
		void GetExamples(std::vector<AI_System::AI_ActionTag>& vec) const throw();


		////////////////////////////////////////////////////////////////////////////////
		//	FUNCTION:
		/*!	\brief		Deletes all generated examples held with in this module.
		**	\param		none
		**	\return		none
		**	\exception	none
		**	\par		DESCRIPTION:
		**
		**	\author		Nate Cleveland
		**	\date		10/25/2002 1:55:35 PM
		**	\warning	none
		*/
		////////////////////////////////////////////////////////////////////////////////
		void Dump(void) throw();


		////////////////////////////////////////////////////////////////////////////////
		//	FUNCTION:
		/*!	\brief		Wrapper for calling both Dump(void) and GetExamples().
		**	\param		vec		Container to fill with examples.
		**	\return		none
		**	\exception	none
		**	\par		DESCRIPTION:
		**				 Please see FishBody::GetExamples() and FishBody::Dump().
		**	\author		Nate Cleveland
		**	\date		10/25/2002 1:57:37 PM
		**	\warning	none
		*/
		////////////////////////////////////////////////////////////////////////////////
		void Dump(std::vector<AI_System::AI_ActionTag>& vec) throw()
		{
			GetExamples(vec);
			Dump();
		}


		////////////////////////////////////////////////////////////////////////////////
		//	FUNCTION:
		/*!	\brief		Updates the module's state.
		**
		**	\param float elapsedTime		Time since last call to Update in seconds.
		**
		**	\return		none
		**	\exception	none
		**	\par		DESCRIPTION:
		**
		**	\author		Nate Cleveland
		**	\date		10/25/2002 2:15:54 PM
		**	\warning	none
		*/
		////////////////////////////////////////////////////////////////////////////////
		void Update(float elapsedTime_seconds) throw();	

		////////////////////////////////////////////////////////////////////////////////
		//	FUNCTION:	
		/*!	\brief		Fills passed in member with position of fish.
		**	\param		pos		Ref to fill with position of this fish.
		**	\return		none
		**	\exception	none
		**	\par		DESCRIPTION:
		**					
		**	\author		Nate Cleveland
		**	\date		11/4/2002 4:44:13 PM
		**	\warning	none
		*/
		////////////////////////////////////////////////////////////////////////////////
		void GetPos(InterpPnt_type& pos) const throw();
		void GetDir(InterpVec_type& dir) const ;


		////////////////////////////////////////////////////////////////////////////////
		//	FUNCTION:	
		/*!	\brief		Sets the position to given location.
		**	\param		pos		Ref to new position of fish.
		**	\return		none
		**	\exception	none
		**	\par		DESCRIPTION:
		**					
		**	\author		Nate Cleveland
		**	\date		11/4/2002 4:56:08 PM
		**	\warning	none
		*/
		////////////////////////////////////////////////////////////////////////////////
		void SetPos(const InterpPnt_type& pos);
		void SetDir(const InterpVec_type& dir);

		////////////////////////////////////////////////////////////////////////////////
		//	FUNCTION:	
		/*!	\brief		Fills the caller's given point with a random location.
		**	\param		randPnt		Reference to an obj to fill with a rand point.
		**	\return		none
		**	\exception	none
		**	\par		DESCRIPTION:
		**					(0,0,0) is set if unable to deduce a random point (ie
							 when we don't yet have a TankClient).
		
		**	\author		Nate Cleveland
		**	\date		2/14/2003 2:33:28 PM
		**	\warning	none
		*/
		////////////////////////////////////////////////////////////////////////////////
		void GetRandPoint(InterpPnt_type& randPnt) throw();



		////////////////////////////////////////////////////////////////////////////////
		//	FUNCTION:	
		/*!	\brief		Gives the caller a copy of the current Fish attributes.
		**	\param		at		Reference to the location to fill with this fish's 
		**						 Attributes.
		**	\return		none
		**	\exception	none
		**	\par		DESCRIPTION:
		**				Provides a copy of the stats that the caller can view and alter
		**				 at their will. Note, that for viewing the public member 
		**				 m_pAttributes may be quicker. If the caller wishes to
		**				 have any changes they make to the class take effect they must
		**				 pass their copy to SetAttribute.
		**	\author		Nate Cleveland
		**	\date		11/4/2002 1:23:01 PM
		**	\warning	none
		*/
		////////////////////////////////////////////////////////////////////////////////
		void GetAttributesCopy(Attributes& at)const throw() {at = this->m_attributes;}

		////////////////////////////////////////////////////////////////////////////////
		//	FUNCTION:	
		/*!	\brief		Gives the caller the magnitude of the velocity vector.
		**	\param		none
		**	\return		none
		**	\exception	none
		**	\par		DESCRIPTION:
		**					
		**	\author		Nate Cleveland
		**	\date		11/25/2002 4:49:01 PM
		**	\warning	none
		*/
		////////////////////////////////////////////////////////////////////////////////
		float GetVelocity(void) const throw() {return Magnitude<float>(this->m_currVector);}
		

		float GetVisionRadius(void) const throw()
		{
			return m_pAttributes->GetEyes() * 2000;
		}
		
		void GetDirection(InternalVector_type& dir) const throw(){dir = m_currVector;}
		void GetDirection(InterpVec_type& dir) const 
		{
			dir[0] = m_currVector.x;
			dir[1] = m_currVector.y;
			dir[2] = m_currVector.z;
		}


		////////////////////////////////////////////////////////////////////////////////
		//	FUNCTION:	
		/*!	\brief		Sets the new attribute values into this fish.
		**	\param		at		Reference of object to read values in from.
		**	\return		none
		**	\exception	none
		**	\par		DESCRIPTION:
		**				Validates the settings before accepting them.
		**	\author		Nate Cleveland
		**	\date		11/4/2002 1:28:40 PM
		**	\warning	none
		*/
		////////////////////////////////////////////////////////////////////////////////
		void SetAttributes(const Attributes& at) throw()
		{
			this->m_attributes = at;
			m_attributes.VerifyValues();
		}

		////memento classes
		////////////////////////////////////////////////////////////////////////////////
		//	FUNCTION:
		/*!	\brief		Writes this class to the passed in stream
		**
		**	\param ostream&						Stream to write to.
		**
		**	\return		none
		**	\exception	none
		**	\par		DESCRIPTION: Simply calls the overloaded << operator.
		**
		**	\author		Nate Cleveland
		**	\date		10/16/2002 6:30:49 PM
		**	\warning	none
		*/
		////////////////////////////////////////////////////////////////////////////////
		void Write(std::ostream&) const;

		////////////////////////////////////////////////////////////////////////////////
		//	FUNCTION:
		/*!	\brief		Reads data from the stream.
		**	\param istream&					Stream to read from.
		**	\return		none
		**	\exception	none
		**	\par		DESCRIPTION: Simply calls the overloaded >> operator.
		**
		**	\author		Nate Cleveland
		**	\date		10/16/2002 6:30:55 PM
		**	\warning	none
		*/
		////////////////////////////////////////////////////////////////////////////////
		void Read(std::istream&);


		////////////////////////////////////////////////////////////////////////////////
		//	FUNCTION:	Highlight
		/*!	\brief		Highlights given fish with given color.
		**	\param		isOn	If set to true the fish will become highlighted. If
		**						 set to false it will be NOT highlighted.
		**	\param		red		Value of red componet, 0 to 1.
		**	\param		green	Value of green componet, 0 to 1
		**	\param		blue	Value of blue componet, 0 to 1
		**	\return		none
		**	\exception	none
		**	\par		DESCRIPTION:
		**					Instructs Graphic client to highlight or unhighlight a fish.
		**					 Pass true and a color value to turn highlighting on, or just
		**					 false to turn it off (you can send a color if you like, but
		**					 there would be no real point.
		**	\author		Nate Cleveland
		**	\date		3/13/2003 12:36:59 PM
		**	\warning	none
		*/
		////////////////////////////////////////////////////////////////////////////////
		void Highlight(bool isOn, float red = 1, float green = 1, float blue = 1) throw();


		////////////////////////////////////////////////////////////////////////////////
		//	FUNCTION:	
		/*!	\brief		
		**	\param		
		**	\return		none
		**	\exception	none
		**	\par		DESCRIPTION:
		**					
		**	\author		Nate Cleveland
		**	\date		4/15/2003 4:05:24 PM
		**	\warning	none
		*/
		////////////////////////////////////////////////////////////////////////////////
		bool SetPathing(bool IsPathing) throw() {return m_IsPathing = IsPathing;}


		////////////////////////////////////////////////////////////////////////////////
		//	FUNCTION:	
		/*!	\brief		Reads from a netMessage and replaces itself with the data it 
						 find their.
		**	\param		
		**	\return		none
		**	\exception	Throws an exception if there is an error in the packet or the 
		**				 keys don't match.
		**	\par		DESCRIPTION:
		**					
		**	\author		Nate Cleveland
		**	\date		2/6/2003 2:32:53 PM
		**	\warning	none
		*/
		////////////////////////////////////////////////////////////////////////////////
		void operator << (UDPNet::NetMessage& netMessage);


		////////////////////////////////////////////////////////////////////////////////
		//	FUNCTION:	
		/*!	\brief		Fills a NetMessage with all the data for this fish.
		**	\param		
		**	\return		none
		**	\exception	Throws an exception iff there is an error with netMessage.
		**	\par		DESCRIPTION:
		**					
		**	\author		Nate Cleveland
		**	\date		2/6/2003 2:29:45 PM
		**	\warning	none
		*/
		////////////////////////////////////////////////////////////////////////////////
		void operator >> (UDPNet::NetMessage& netMessage);


		////////////////////////////////////////////////////////////////////////////////
		//	FUNCTION:
		/*!	\brief		Writes all needed data to the stream
		**
		**	\param ostream&				object we'll stream.
		**	\param const  Attributes&	Reference to this class.
		**
		**	\return		A reference to the stream.
		**	\exception	none
		**	\par		DESCRIPTION: Writes all the needed data (Base and Reactive
		**				 Attributes) to the stream.
		**
		**	\author		Nate Cleveland
		**	\date		10/16/2002 6:30:59 PM
		**	\warning	none
		*/
		////////////////////////////////////////////////////////////////////////////////
		friend std::ostream& operator << (std::ostream&, const FishBody &);

		////////////////////////////////////////////////////////////////////////////////
		//	FUNCTION:
		/*!	\brief		Reads in data from the stream.
		**
		**	\param ostream&				object we'll stream from.
		**	\param const  Attributes&	Reference to this class.
		**
		**	\return		A reference to the stream.
		**	\exception	none
		**	\par		DESCRIPTION: Reads data from the stream and resets this class
		**				 accordingly.
		**
		**	\author		Nate Cleveland
		**	\date		10/16/2002 6:31:04 PM
		**	\warning	none
		*/
		////////////////////////////////////////////////////////////////////////////////
		friend std::istream& operator>> (std::istream&, FishBody &);
	};//end class FishBody
}//end namespace InterpretationModule.

#pragma warning(pop)	//restore warning state to what it was before this file compiled.
