////////////////////////////////////////////////////////////////////////////////
/*!	\file		fishBodyHandle.hpp
**	\brief		Declares the class HFishBody.
**	\author		Nate Cleveland
**	\date		10/28/2002 9:41:42 AM
**	\par		DESCRIPTION:
**				 HFishBody is the only interface an action has with its body.
**				 The HFishBody provides the following services to actions.
**					- Fish State management (IsAttacking, IsEating, IsMouthStopped,
**					SetMouthStopped, SetFinStopped, EndAttack, EndEat, BeginAttack
**					and BeginEat).
**					- FinMovementStopped logic (AddVector).
**					- Physical data (GetPos, GetRandPos and GetCurrentDirectionVector).
**					- Marker Management (AddMarker, GetMarkerArray).
**
**	\par		 HFishBody Provides the following services to a fish body.
**					- FinMovementStopped logic (GetSummationVector, Flush).
**
**	\par		REMARKS:
**				 This class can access and alter any member in the FishBody class
**				 that is passed to it in its ctor.
**
*/
////////////////////////////////////////////////////////////////////////////////
#pragma once
#include "fishBody.hpp"


namespace InterpretationModule
{

	////////////////////////////////////////////////////////////////////////////////
	/*!	\class		HFishBody
	**	\brief		Only interface an action has with its body.
	**	\author		Nate Cleveland
	**	\date		10/28/2002 9:41:13 AM
	**	\warning	none
	*/
	////////////////////////////////////////////////////////////////////////////////
	class HFishBody
	{
	protected:
		FishBody&			m_fishBody;				//!< Ref to the fish that this handle connects to.
		InternalVector_type	m_vectorSummation;		//!< Summ of all vectors since last flush call.


		HFishBody(FishBody& fishBody):m_fishBody(fishBody){}

		////////////////////////////////////////////////////////////////////////////////
		//	FUNCTION:
		/*!	\brief		Resets the state of the HFishBody module
		**	\param		none
		**	\return		none
		**	\exception	none
		**	\par		DESCRIPTION:
		**				Should be called once before the first Action is updated.
		**				Currently resets the following members,
		**					- m_vectorSummation
		**
		**	\author		Nate Cleveland
		**	\date		10/21/2002 1:56:12 PM
		**	\warning	none
		*/
		////////////////////////////////////////////////////////////////////////////////
		void Flush()throw();



	public:
		~HFishBody(){}
		friend class FishBody;
		

		bool GetFishBody(AquariusKey_type key, FishBody& fishBody) const throw();

		FishBody* GetFishBodyPtr(AquariusKey_type key) throw();

		////////////////////////////////////////////////////////////////////////////////
		//	FUNCTION:
		/*!	\brief		Returns true if this fish is in the attack state.
		**	\param		none
		**	\return		none
		**	\exception	none
		**	\par		DESCRIPTION:
		**
		**	\author		Nate Cleveland
		**	\date		10/23/2002 12:28:13 PM
		**	\warning	none
		*/
		////////////////////////////////////////////////////////////////////////////////
		bool IsAttacking(void)const throw();

		////////////////////////////////////////////////////////////////////////////////
		//	FUNCTION:
		/*!	\brief		Returns true if this fish is in the eating state.
		**	\param		none
		**	\return		none
		**	\exception	none
		**	\par		DESCRIPTION:
		**
		**	\author		Nate Cleveland
		**	\date		10/23/2002 12:28:13 PM
		**	\warning	none
		*/
		////////////////////////////////////////////////////////////////////////////////
		bool IsEating(void)const throw();

		////////////////////////////////////////////////////////////////////////////////
		//	FUNCTION:
		/*!	\brief		Returns true if this fish is in the eating state.
		**	\param		none
		**	\return		none
		**	\exception	none
		**	\par		DESCRIPTION:
		**
		**	\author		Nate Cleveland
		**	\date		10/23/2002 12:28:13 PM
		**	\warning	none
		*/
		////////////////////////////////////////////////////////////////////////////////
		bool IsMouthStopped(void)const throw();

		////////////////////////////////////////////////////////////////////////////////
		//	FUNCTION:
		/*!	\brief		Sets the isMouthStopped flag to the passed in value.
		**	\param		boolean value to set the flag to. Overloaded to true.
		**	\return		none
		**	\exception	none
		**	\par		DESCRIPTION: If true is passed will stop any other mouth states.
		**
		**	\author		Nate Cleveland
		**	\date		10/23/2002 12:28:13 PM
		**	\warning	none
		*/
		////////////////////////////////////////////////////////////////////////////////
		void SetMouthStopped(bool isMouthStopped = true);

		////////////////////////////////////////////////////////////////////////////////
		//	FUNCTION:
		/*!	\brief		Sets the isFinStopped flag to the passed in value.
		**	\param		boolean value to set the flag to. Overloaded to true.
		**	\return		none
		**	\exception	none
		**	\par		DESCRIPTION: If true is passed will stop any other fin state(s).
		**
		**	\author		Nate Cleveland
		**	\date		10/23/2002 12:28:13 PM
		**	\warning	none
		*/
		////////////////////////////////////////////////////////////////////////////////
		void SetFinStopped(bool isFinStopped = true);


		//returns the value of the summation vector to the caller.
		////////////////////////////////////////////////////////////////////////////////
		//	FUNCTION:
		/*!	\brief		Gives the caller the current value of the summation vector.
		**	\param		Vector to fill with copy of current summation vector.
		**	\return		none
		**	\exception	none
		**	\par		DESCRIPTION:
		**				Summation vector is reset everytime the HFishBody::Flush method
						 is called.
		**
		**	\author		Nate Cleveland
		**	\date		10/23/2002 12:34:37 PM
		**	\warning	none
		*/
		////////////////////////////////////////////////////////////////////////////////
		void GetSummationVector(InternalVector_type&) const throw();


		////////////////////////////////////////////////////////////////////////////////
		//	FUNCTION:
		/*!	\brief		Adds this vector to the vector summation.
		**
		**	\param const  InterpVector_type&	Vector to add to the summation.
		**
		**	\return		none
		**	\exception	none
		**	\par		DESCRIPTION: Will do nothing if the Fin Stopped flag is set to
		**				 true.
		**
		**	\author		Nate Cleveland
		**	\date		10/21/2002 1:43:45 PM
		**	\warning	none
		*/
		////////////////////////////////////////////////////////////////////////////////
		void AddVector(const InternalVector_type& newField) throw();


		////////////////////////////////////////////////////////////////////////////////
		//	FUNCTION:
		/*!	\brief		Returns to caller the position of this fish.
		**	\param InterpPoint_type& point		Point to fill with loaction.
		**	\return		none
		**	\exception	none
		**	\par		DESCRIPTION: Returns to caller the position of this fish.
		**
		**	\author		Nate Cleveland
		**	\date		10/21/2002 1:46:26 PM
		**	\warning	none
		*/
		////////////////////////////////////////////////////////////////////////////////
		void GetPos(InternalPoint_type& point) const throw();

		////////////////////////////////////////////////////////////////////////////////
		//	FUNCTION:
		/*!	\brief		Returns to the caller the loaction of the specified object.
		**
		**	\param AquariusKey_type key		Key of object in question.
		**	\param	 InterpPoint_type& point	Will be filled with location of object.
		**
		**	\return		Will return false if unable to find the item.
		**	\exception	none
		**	\par		DESCRIPTION:
		**
		**	\author		Nate Cleveland
		**	\date		10/21/2002 1:47:32 PM
		**	\warning	none
		*/
		////////////////////////////////////////////////////////////////////////////////
		bool GetPos(AquariusKey_type key, InternalPoint_type& point) const throw();


		////////////////////////////////////////////////////////////////////////////////
		//	FUNCTION:
		/*!	\brief		Give the caller the current direction the fish is going.
		**	\param
		**	\return		none
		**	\exception	none
		**	\par		DESCRIPTION:
		**
		**	\author		Nate Cleveland
		**	\date		10/22/2002 1:55:10 PM
		**	\warning	none
		*/
		////////////////////////////////////////////////////////////////////////////////
		bool GetCurrentDirectionVector(InternalVector_type& vec) const throw();


		////////////////////////////////////////////////////////////////////////////////
		//	FUNCTION:
		/*!	\brief		Tells the fish to end the attacking state
		**
		**	\param AquariusKey_type key		Key to try to attack..
		**
		**	\return		none
		**	\exception	none
		**	\par		DESCRIPTION: This method should be called after a fish action
		**				 has waited for an appropriate period of time to allow for
		**				 attack animations and what not. Once this method is called
		**				 the fish's state is set to not attacking and damage is dealt
		**				 to the target. This method also generates Chum and effects
		**				 energy levels.
		**
		**	\author		Nate Cleveland
		**	\date		10/21/2002 2:01:52 PM
		**	\warning	none
		*/
		////////////////////////////////////////////////////////////////////////////////
		bool EndAttack(AquariusKey_type key) throw();


		////////////////////////////////////////////////////////////////////////////////
		//	FUNCTION:
		/*!	\brief		Tells the fish to enter the attacking state/animation.
		**	\param		none
		**	\return		Returns the values of the m_isAttacking flag.
		**	\exception	none
		**	\par		DESCRIPTION: Tries to set the is attacking flag to true.
		**
		**	\author		Nate Cleveland
		**	\date		10/21/2002 3:43:24 PM
		**	\warning	none
		*/
		////////////////////////////////////////////////////////////////////////////////
		bool BeginAttack(void) throw();


		////////////////////////////////////////////////////////////////////////////////
		//	FUNCTION:
		/*!	\brief		Tells the fish to eat the target.
		**
		**	\param AquariusKey_type key		Key to try to eat.
		**
		**	\return		none
		**	\exception	none
		**	\par		DESCRIPTION: Ends the eating state and consumes any food for the
		**				 fish.
		**	\todo		The method HFishBody::EndEat will not function until we have a
		**				 structure in the TankSimulator for holding chum.
		**
		**	\author		Nate Cleveland
		**	\date		10/21/2002 2:01:52 PM
		**	\warning	none
		*/
		////////////////////////////////////////////////////////////////////////////////
		bool EndEat(AquariusKey_type key) throw();


		////////////////////////////////////////////////////////////////////////////////
		//	FUNCTION:
		/*!	\brief		Tells the fish to enter the eating state/animation.
		**	\param		none
		**	\return		Returns the value of the m_isEating flag.
		**	\exception	none
		**	\par		DESCRIPTION: Tries to set the is attacking flag to true. Will
		**				 fail if it already has a mouth related flag set.
		**
		**	\author		Nate Cleveland
		**	\date		10/21/2002 3:56:49 PM
		**	\warning	none
		*/
		////////////////////////////////////////////////////////////////////////////////
		bool BeginEat(void) throw();



		////////////////////////////////////////////////////////////////////////////////
		//	FUNCTION:
		/*!	\brief
		**	\param
		**	\return		none
		**	\exception	none
		**	\par		DESCRIPTION:
		**
		**	\author		Nate Cleveland
		**	\date		10/22/2002 3:36:47 PM
		**	\warning	none
		*/
		////////////////////////////////////////////////////////////////////////////////
		void AddMarker(const InternalPoint_type& newMarker, float weight) throw();


		////////////////////////////////////////////////////////////////////////////////
		//	FUNCTION:
		/*!	\brief		Returns to the caller the marker array, read-only.
		**
		**	\param		elementCount		will be filled with the number of array
		**	\param		weightArr			will be filled with a pntr to the weight array.
		**	\return		A constant pntr the marker array.
		**	\exception	none
		**	\author		Nate Cleveland
		**	\date		10/22/2002 3:38:36 PM
		**	\warning	none
		*/
		////////////////////////////////////////////////////////////////////////////////
		const InternalPoint_type *GetMarkerArray(int& elementCount, const float **weightArr) const throw();

		////////////////////////////////////////////////////////////////////////////////
		//	FUNCTION:	
		/*!	\brief		Returns to the caller the location of the nearest chum chunk.
		**	\param		chumPoint		Location where the chum'll be stored.
		**	\return		none
		**	\exception	none
		**	\par		DESCRIPTION:
		**					Returns to the caller the location of the nearest chunk of 
							 chum relative to this fish.
		**	\author		Nate Cleveland
		**	\date		1/21/2003 4:45:19 PM
		**	\warning	none
		*/
		////////////////////////////////////////////////////////////////////////////////
		void GetNearestChum(InternalPoint_type& chumPoint) const throw();
		
		////////////////////////////////////////////////////////////////////////////////
		//	FUNCTION:	
		/*!	\brief		Returns to the caller the ID of the nearest fish.
		**	\param		none
		**	\return		0 if there are no other fish otherwise the key of the nearest fish.
		**	\exception	Any exceptions thrown would have to be results of tankSimulatorClient.	
		**	\author		Nate Cleveland
		**	\date		2/20/2003 10:42:06 AM
		**	\warning	none
		*/
		////////////////////////////////////////////////////////////////////////////////
		AquariusKey_type GetNearestFish(void);


		////////////////////////////////////////////////////////////////////////////////
		//	FUNCTION:	CheckCollsion
		/*!	\brief		Returns true iff this fish collides with the fish of the given key.
		**	\param		key		Key of the fish to check collision with.
		**	\return		boolean true/false, true iff collsion with other fish.
		**	\exception	none
		**	\par		DESCRIPTION:
		**					FYI, uses spherical collsion.
		**	\author		Nate Cleveland
		**	\date		3/18/2003 2:30:53 PM
		**	\warning	none
		*/
		////////////////////////////////////////////////////////////////////////////////
		bool CheckCollsion(AquariusKey_type key) const throw();
		

		void ClipPoint(InternalPoint_type& pnt) throw();


		float GetRadius(void);
	};//end class HFishBody
}//end namespace InterpretationModule.