#include "fishBodyMath.hpp"
#include <math.h>
// memory leak detection
#ifdef _DEBUG
    #include"memleak.h"
    #define new DEBUG_NEW
    #undef THIS_FILE
    static char THIS_FILE[] = __FILE__;
#endif


#define	X		0				// x component index
#define Y		1				// y component index
#define Z		2				// z component index
#define W		3				// w component index

#define XY 		2				// number of components in 2D CS
#define XYZ		3				// number of components in 3D CS
#define XYZW	4				// number of components in 4D CS

#define A		0				// A component index
#define B		1				// B component index
#define C		2				// C component index
#define D		3				// D component index
#define ABCD	4				// number of params in 3D plane






////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	RaySphereIntersection()
//!	\brief		Finds if a given point is on a given ray.
//!	\return		0 if there is NOT an intersection, 1 if there is an intersection.
//!	\date		2/4/2003
//!	\warning	Assumes legal ray.
////////////////////////////////////////////////////////////////////////////////
u8 PointOnRay3D(Ray3D ray, Point3D point)
{
	//hugmongous if statement taken from Dr. Li's Notes. 
	if( (ray.direction[X]>0 && (point[X]-ray.origin[X] >= 0)) || 
		(ray.direction[X]<0 && (point[X]-ray.origin[X] <= 0)) ||
		(ray.direction[Y]>0 && (point[Y]-ray.origin[Y] >= 0)) || 
		(ray.direction[Y]<0 && (point[Y]-ray.origin[Y] <= 0)) ||
		(ray.direction[Z]>0 && (point[Z]-ray.origin[Z] >= 0)) || 
		(ray.direction[Z]<0 && (point[Z]-ray.origin[Z] <= 0))   )
		return 1;
	
	return 0;
}




////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	LinePlaneIntersection3D
//!	\brief		Finds the point of intersection between the line l and plane p.
//!				 Stores resulting intersection in parameter m (if there is one).
//!	\return		0 if there is no intersection, 1 if there is.
//!	\date		1/26/2003
//!	\warning	none
////////////////////////////////////////////////////////////////////////////////
u8  LinePlaneIntersection3D(Line3D l, Plane3D p, Point3D m)
{
	f64		normalDOTvector;	//will hold dot product of plane's normal lines's vector.
	f64		tValue;				//will hold value of parametric line's 't'

	//plane an line never intersect if plane's normal DOT line's vector is zero.
	normalDOTvector = p[A]*l.vector[X] + p[B]*l.vector[Y] + p[C]*l.vector[Z];
	if(normalDOTvector == 0.0)
		return 0;

	//t = -(A*x0 + B*y0 + C*z0 + D) / (v DOT N)
	tValue	= -(p[A]*l.point[X] + p[B]*l.point[Y] + p[C]*l.point[Z] + p[D]) / normalDOTvector;
	
	//Formula for finding intersection is x = x0 + p*t, y = y0 + q*t, z = z0 + r*t
	m[X]	= l.point[X] + l.vector[X]*tValue;
	m[Y]	= l.point[Y] + l.vector[Y]*tValue;
	m[Z]	= l.point[Z] + l.vector[Z]*tValue;

	return 1;
}



////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	LinePlaneIntersection3D
//!	\brief		Computes the intersection point between a plane and a ray.
//!	\return		0 if there is no intersection, 1 if there is.
//!	\date		1/26/2003
//!	\warning	none
////////////////////////////////////////////////////////////////////////////////
u8 RayPlaneIntersection3D(Ray3D ray, Plane3D plane, Point3D point)
{
	//create a line equation from the ray.
	Line3D	line =
				{
					{ray.origin[X], ray.origin[Y], ray.origin[Z]},
					{ray.direction[X], ray.direction[Y], ray.direction[Z]},
				};

	//Test for collision with the plane. If they do collide make sure
	// the point is on the ray.
	if(LinePlaneIntersection3D(line, plane, point))
		return PointOnRay3D(ray, point);
	return 0;
}

void CopyPoint(const Point3D point, Point3D ans)
{
	ans[X] = point[X];
	ans[Y] = point[Y];
	ans[Z] = point[Z];
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	FindPlaneD
//!	\brief		Creates a plane in 3 space from a point and a normal.
//!	\return		nothing.
//!	\date		1/24/2003
//!	\warning	none.
////////////////////////////////////////////////////////////////////////////////
void CreatePlane3D_NormalPoint(f64 x, f64 y, f64 z, Point3D point, Plane3D plane)
{
	plane[A] = x;
	plane[B] = y;
	plane[C] = z;
	plane[D] = -(point[X]*plane[A] + point[Y]*plane[B] + point[Z]*plane[C]);
}



////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	RaySphereIntersection()
//!	\brief		Checks if point is within the AABB box.
//!	\return		0 if there is NOT an intersection, 1 if there is an intersection.
//!	\date		2/4/2003
//!	\warning	none
////////////////////////////////////////////////////////////////////////////////
u8 PointOnBox(const Box3D box, const Point3D point)
{
	if( (box.origin[X] <= point[X] && point[X] <= box.extent[X]) &&
	    (box.origin[Y] <= point[Y] && point[Y] <= box.extent[Y]) &&	
		(box.origin[Z] <= point[Z] && point[Z] <= box.extent[Z])   )
		return 1;
	return 0;
}


////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	RayBoxIntersection()
//!	\brief		Computes the intersection points between a ray and an AABB.
//!				 Stores intersection points in outPoint array.
//!	\return		The number of intersection points, 0, 1 or 2.
//!	\date		2/4/2003
//!	\warning	Assumes that the vector from box.origin and box.extend has all
//!				 positive elements.
////////////////////////////////////////////////////////////////////////////////
s8 RayBoxIntersection (Ray3D ray, Box3D box, Point3D outPoint[2])
{
	u8 retVal = 0;

	f64 xMin,yMin,zMin,xMax,yMax,zMax;			// constants
	f64 xP,yP,zP,xV,yV,zV;						// constants
	f64 kxMin,kxMax,kyMin,kyMax,kzMin,kzMax;	// constants
	Point3D x1,x2,y1,y2,z1,z2;					// instersection points with the planes of the box
	s8 xParallel, yParallel, zParallel;			// flags

	xParallel = yParallel = zParallel = 0;

	// set up the data

	xMin = box.origin[X];
	yMin = box.origin[Y];
	zMin = box.origin[Z];

	xMax = box.extent[X];
	yMax = box.extent[Y];
	zMax = box.extent[Z];

	xP = ray.origin[X];
	yP = ray.origin[Y];
	zP = ray.origin[Z];

	xV = ray.direction[X];
	yV = ray.direction[Y];
	zV = ray.direction[Z];

	// constants used in our equations

	if (xV != 0)
	{
		kxMin = (xMin - xP) / xV;
		kxMax = (xMax - xP) / xV;
	}
	else
		xParallel = 1;

	if (yV != 0)
	{
		kyMin = (yMin - yP) / yV;
		kyMax = (yMax - yP) / yV;
	}
	else 
		yParallel = 1;

	if (zV != 0)
	{
		kzMin = (zMin - zP) / zV;
		kzMax = (zMax - zP) / zV;
	}
	else 
		zParallel = 1;

	// solve for intersections with the six planes

	if(!xParallel)
	{
		// xMin
		
		x1[X] = xMin;
		x1[Y] = yP + (kxMin * yV);
		x1[Z] = zP + (kxMin * zV);

		// xMax

		x2[X] = xMax;
		x2[Y] = yP + (kxMax * yV);
		x2[Z] = zP + (kxMax * zV);

		// test to see if these are valid intersection points

		if(x1[X] >= xMin && x1[X] <= xMax)
		{
			if(x1[Y] >= yMin && x1[Y] <= yMax)
			{
				if(x1[Z] >= zMin && x1[Z] <= zMax)
				{
					// must be in ray also

					if( ( (xV != 0) && ((x1[X] - xP) / xV > 0) ) ||
						( (yV != 0) && ((x1[Y] - yP) / yV > 0) ) ||
						( (zV != 0) && ((x1[Z] - zP) / zV > 0) ) )
					{
						outPoint[retVal][X] = x1[X];
						outPoint[retVal][Y] = x1[Y];
						outPoint[retVal][Z] = x1[Z];

						++retVal;
					}
				}
			}
		}

		if(x2[X] >= xMin && x2[X] <= xMax)
		{
			if(x2[Y] >= yMin && x2[Y] <= yMax)
			{
				if(x2[Z] >= zMin && x2[Z] <= zMax)
				{
					// must be in ray also

					if( ( (xV != 0) && ((x2[X] - xP) / xV > 0) ) ||
						( (yV != 0) && ((x2[Y] - yP) / yV > 0) ) ||
						( (zV != 0) && ((x2[Z] - zP) / zV > 0) ) )
					{
						outPoint[retVal][X] = x2[X];
						outPoint[retVal][Y] = x2[Y];
						outPoint[retVal][Z] = x2[Z];

						++retVal;
					}
				}
			}
		}

		if(retVal == 2)	// early exit
		{
			if(outPoint[0][X] == outPoint[1][X] &&
			   outPoint[0][X] == outPoint[1][X] &&
			   outPoint[0][X] == outPoint[1][X])
			{
				retVal = 1;
			}

			return retVal;
		}
	}

	if(!yParallel)
	{
		// yMin

		y1[X] = xP + (kyMin * xV);
		y1[Y] = yMin;
		y1[Z] = zP + (kyMin * zV);

		// yMax

		y2[X] = xP + (kyMax * xV);
		y2[Y] = yMax;
		y2[Z] = zP + (kyMax * zV);

		// test to see if these are valid intersection points

		if(y1[X] >= xMin && y1[X] <= xMax)
		{
			if(y1[Y] >= yMin && y1[Y] <= yMax)
			{
				if(y1[Z] >= zMin && y1[Z] <= zMax)
				{
					// must be in ray also

					if( ( (xV != 0) && ((y1[X] - xP) / xV > 0) ) ||
						( (yV != 0) && ((y1[Y] - yP) / yV > 0) ) ||
						( (zV != 0) && ((y1[Z] - zP) / zV > 0) ) )
					{
						outPoint[retVal][X] = y1[X];
						outPoint[retVal][Y] = y1[Y];
						outPoint[retVal][Z] = y1[Z];

						++retVal;

						if(retVal == 2)	// early exit
						{
							if(outPoint[0][X] == outPoint[1][X] &&
							   outPoint[0][X] == outPoint[1][X] &&
							   outPoint[0][X] == outPoint[1][X])
							{
								retVal = 1;
							}

							return retVal;
						}
					}
				}
			}
		}

		if(y2[X] >= xMin && y2[X] <= xMax)
		{
			if(y2[Y] >= yMin && y2[Y] <= yMax)
			{
				if(y2[Z] >= zMin && y2[Z] <= zMax)
				{
					// must be in ray also

					if( ( (xV != 0) && ((y2[X] - xP) / xV > 0) ) ||
						( (yV != 0) && ((y2[Y] - yP) / yV > 0) ) ||
						( (zV != 0) && ((y2[Z] - zP) / zV > 0) ) )
					{
						outPoint[retVal][X] = y2[X];
						outPoint[retVal][Y] = y2[Y];
						outPoint[retVal][Z] = y2[Z];

						++retVal;

						if(retVal == 2)	// early exit
						{
							if(outPoint[0][X] == outPoint[1][X] &&
							   outPoint[0][X] == outPoint[1][X] &&
							   outPoint[0][X] == outPoint[1][X])
							{
								retVal = 1;
							}

							return retVal;
						}
					}
				}
			}
		}
	}

	if(!zParallel)
	{
		// zMin

		z1[X] = xP + (kzMin * xV);
		z1[Y] = yP + (kzMin * yV);
		z1[Z] = zMin;

		// zMax

		z2[X] = xP + (kzMax * xV);
		z2[Y] = yP + (kzMax * yV);
		z2[Z] = zMax;

		// test to see if these are valid intersection points

		if(z1[X] >= xMin && z1[X] <= xMax)
		{
			if(z1[Y] >= yMin && z1[Y] <= yMax)
			{
				if(z1[Z] >= zMin && z1[Z] <= zMax)
				{
					// must be in ray also

					if( ( (xV != 0) && ((z1[X] - xP) / xV > 0) ) ||
						( (yV != 0) && ((z1[Y] - yP) / yV > 0) ) ||
						( (zV != 0) && ((z1[Z] - zP) / zV > 0) ) )
					{
						outPoint[retVal][X] = z1[X];
						outPoint[retVal][Y] = z1[Y];
						outPoint[retVal][Z] = z1[Z];

						++retVal;

						if(retVal == 2)	// early exit
						{
							if(outPoint[0][X] == outPoint[1][X] &&
							   outPoint[0][X] == outPoint[1][X] &&
							   outPoint[0][X] == outPoint[1][X])
							{
								retVal = 1;
							}

							return retVal;
						}
					}
				}
			}
		}

		if(z2[X] >= xMin && z2[X] <= xMax)
		{
			if(z2[Y] >= yMin && z2[Y] <= yMax)
			{
				if(z2[Z] >= zMin && z2[Z] <= zMax)
				{
					// must be in ray also

					if( ( (xV != 0) && ((z2[X] - xP) / xV > 0) ) ||
						( (yV != 0) && ((z2[Y] - yP) / yV > 0) ) ||
						( (zV != 0) && ((z2[Z] - zP) / zV > 0) ) )
					{
						outPoint[retVal][X] = z2[X];
						outPoint[retVal][Y] = z2[Y];
						outPoint[retVal][Z] = z2[Z];

						++retVal;
					}
				}
			}
		}
	}

	// check to see if the two points are actually the same point (like a vertex)

	if(retVal == 2)
	{
		if(outPoint[0][X] == outPoint[1][X] &&
		   outPoint[0][X] == outPoint[1][X] &&
		   outPoint[0][X] == outPoint[1][X])
		{
			retVal = 1;
		}
	}

	return retVal;
}


////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	PointSubtract3D
//!	\brief		Subtracts two points, (p1 - p2). Answer stored in w.
//!	\return		none
//!	\date		1/14/2003 11:42:24 AM
//!	\warning	none
////////////////////////////////////////////////////////////////////////////////
void PointSubtract3D(const Point3D p1, const Point3D p2, Vector3D w)
{
	w[X] = p1[X] - p2[X];
	w[Y] = p1[Y] - p2[Y];
	w[Z] = p1[Z] - p2[Z];
}


////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	VectorLength3D
//!	\brief		Computes the magnitude of a vector.
//!	\return		The magnitude of the given vector.
//!	\date		1/24/2003
//!	\warning	none
////////////////////////////////////////////////////////////////////////////////
f64 VectorLength3D(const Vector3D v)
{
	return sqrt(v[X]*v[X] + v[Y]*v[Y] + v[Z]*v[Z]);	
}

f64 PointDistance3D(const Point3D p1, const Point3D p2)
{
	//creates a vector from one point to another, finds the magnitude of that vector.
	Vector3D distVec;
	PointSubtract3D(p2, p1, distVec);
	return VectorLength3D(distVec);	
}



//see header for doc.
bool SphereCollision(const InterpretationModule::InterpPnt_type& pntA, float radiusA,
					 const InterpretationModule::InterpPnt_type& pntB, float radiusB,
					 float tolerance/* = 0*/)
{
	//calculate the distance between the two points.
	Point3D A1 ={pntA[0], pntA[1], pntA[2]};
	Point3D B1 ={pntB[0], pntB[1], pntB[2]};
	f64 dist = PointDistance3D(A1, B1);

	//if the distance is <= radiusA + b then the sphere collide.
	f64 change = dist - (radiusA + radiusB);
	if(tolerance >= fabs(change))
		return true;
	return false;
}

