#pragma once


#include "fishBodyUtils.hpp"


typedef unsigned char		u8;
typedef unsigned short		u16;
typedef unsigned int		u32;
typedef unsigned long		u64;
typedef char				s8;
typedef short				s16;
typedef int					s32;
typedef long				s64;
typedef float				f32;
typedef double				f64;

typedef f64 Point3D[3];		// xyz components of a 3D point
typedef f64 Vector3D[3];		// xyz components of a 3D vector
typedef f64 Plane3D[4];		// params of a 3D plane equation 

// data type that defines a 3D line equation
typedef struct {
	Point3D		point;			// point the line passes
	Vector3D	vector;			// directional vector
} Line3D;

// data type that defines a 3D ray
typedef struct {
	Point3D		origin;		// the ray origin
	Vector3D	direction;		// the ray directional vector
} Ray3D;

// data type that defines a 3D bounding box
typedef struct {
	Point3D	origin;		// corner with minimum x, y and z
	Point3D	extent;		// corner with maximum x, y and z 
} Box3D;

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:
/*!	\brief		Finds the unitized version of the passed in vector.
**	\param		Two vectors. First is the Vector to unitize, second is where to
**				 store that unit vector.
**	\return		none
**	\exception	none
**	\par		DESCRIPTION:
**				 Finds the Unitized version of a vector. It is exceptable to
**				 pass the same vectors for both parameters.
**
**	\author		Nate Cleveland
**	\date		10/28/2002 9:52:26 AM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
template<class Value_type>
void UnitizeVector(const InterpretationModule::InternalVector_type& vec, InterpretationModule::InternalVector_type& unitVector)
{
	Value_type length = Magnitude<Value_type>(vec);

	if(length)
		ScaleVector<Value_type>(vec, unitVector, static_cast<Value_type>(1)/length);
	else
		ScaleVector<Value_type>(vec, unitVector, 0);
}


template<class Value_type>
void UnitizeVector(const InterpretationModule::InterpVec_type& vec, InterpretationModule::InterpVec_type& unitVector)
{
	Value_type length = Magnitude<Value_type>(vec);

	if(length)
		ScaleVector<Value_type>(vec, unitVector, static_cast<Value_type>(1)/length);
	else
		ScaleVector<Value_type>(vec, unitVector, 0);
}


////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:
/*!	\brief		Scales a vector by the given ammount.
**
**	\param vec,			Vector to scale by last parameter.
**	\param scaleVector	Place to store scaled vector
**	\param scale		How much to scale 1st param by.
**
**	\return		none
**	\exception	none
**	\par		DESCRIPTION:
**				 Scales the given vector and stores it in the second param. It is
**				 exceptable to pass the same vectors for the 1st and 2nd arguments.
**
**	\author		Nate Cleveland
**	\date		10/28/2002 9:54:31 AM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
template<class Value_type>
void ScaleVector (const InterpretationModule::InternalVector_type& vec, InterpretationModule::InternalVector_type& scaleVector, Value_type scale)
{
	scaleVector.x = vec.x*scale;
	scaleVector.y = vec.y*scale;
	scaleVector.z = vec.z*scale;
}

template<class Value_type>
void ScaleVector (const InterpretationModule::InterpVec_type& vec, InterpretationModule::InterpVec_type& scaleVector, Value_type scale)
{
	scaleVector = vec;
	scaleVector *= scale;
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:
/*!	\brief		Computes the magnitude of a vector.
**	\param const  InterpVectorValue_type& vec,	Vector to find magnitude of.
**	\return		none
**	\exception	none
**	\par		DESCRIPTION:
**
**	\author		Nate Cleveland
**	\date		10/28/2002 9:56:56 AM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
template<class Value_type>
Value_type Magnitude(const InterpretationModule::InternalVector_type& vec)
{
	return static_cast<Value_type>(sqrt(vec.x*vec.x + vec.y*vec.y + vec.z*vec.z));
}

template<class Value_type>
Value_type Magnitude(const InterpretationModule::InterpVec_type& vec)
{
	return static_cast<Value_type>(sqrt(vec[0]*vec[0] + vec[1]*vec[1] + vec[2]*vec[2]));
}


////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	
/*!	\brief		Returns to the caller the absolute distance between points.
**	\param		a		First vector to compute.
**	\param		b		Second vector to compute.
**	\return		Absolute magnitude of vector from first point to next.
**	\exception	none
**	\par		DESCRIPTION:
**					
**	\author		Nate Cleveland
**	\date		11/4/2002 10:12:18 AM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
template<class Value_type>
Value_type ComputeDistance(const InterpretationModule::InternalPoint_type& a, const InterpretationModule::InternalPoint_type& b)
{
	InterpretationModule::InternalPoint_type ans;
	ans.x = a.x - b.x;
	ans.y = a.y - b.y;
	ans.z = a.z - b.z;
	return fabs(Magnitude<float>(ans));
}

template<class Value_type>
Value_type ComputeDistance(const InterpretationModule::InterpVec_type& a, const InterpretationModule::InterpVec_type& b)
{
	InterpretationModule::InterpVec_type ans(a[0] - b[0], a[1] - b[1], a[2] - b[3]);
	return static_cast<Value_type>(fabs(Magnitude<Value_type>(ans)));
}


////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	
/*!	\brief		Multiplies a matrix3 and a vector3
**	\param		m		Matrix to multiply. lhs
**	\param		u		Vector to multiply. rhs	
**	\param		ans		Place to store answer.
**	\return		none
**	\exception	none
**	\par		DESCRIPTION:
**					
**	\author		Nate Cleveland
**	\date		11/4/2002 3:03:33 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
template<class Value_type>
void Multiply(math::linear_alg::Matrix3<Value_type>& m, const InterpretationModule::InterpVec_type& u, InterpretationModule::InterpVec_type& ans)
{
	InterpretationModule::InterpVec_type finalAns;
	finalAns[0] = m[0][0]*u[0] + m[0][1]*u[1] + m[0][2]*u[2];
	finalAns[1] = m[1][0]*u[0] + m[1][1]*u[1] + m[1][2]*u[2];
	finalAns[2] = m[2][0]*u[0] + m[2][1]*u[1] + m[2][2]*u[2];
	ans = finalAns;
}



////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	
/*!	\brief		Applies rotation along the x vector.
**	\param		vec		Vector to rotate.
**	\param		x		angle in radians to rotate
**	\param		ans		place to store result.
**	\return		none
**	\exception	none
**	\par		DESCRIPTION:
**					
**	\author		Nate Cleveland
**	\date		11/4/2002 2:30:44 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
template<class Value_type>
void RotateXAxis(const InterpretationModule::InterpVec_type& vec, float x, InterpretationModule::InterpVec_type& ans)
{
	math::linear_alg::Matrix3<Value_type> rotX(
		1,		0,									0,
		0,		static_cast<float>(cos(x)),			static_cast<float>(sin(x)),
		0,		static_cast<float>(-sin(x)),		static_cast<float>(cos(x)));
	Multiply<Value_type>(rotX, vec, ans);
}



////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	
/*!	\brief		Applies rotation along the y vector.
**	\param		vec		Vector to rotate.
**	\param		y		angle in radians to rotate
**	\param		ans		place to store result.
**	\return		none
**	\exception	none
**	\par		DESCRIPTION:
**					
**	\author		Nate Cleveland
**	\date		11/4/2002 2:30:44 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
template<class Value_type>
void RotateYAxis(const InterpretationModule::InterpVec_type& vec, float y, InterpretationModule::InterpVec_type& ans)
{
	math::linear_alg::Matrix3<Value_type> rotY(
		static_cast<float>(cos(y)),		0,		static_cast<float>(-sin(y)),
		0,								1,		0,
		static_cast<float>(sin(y)),		0,		static_cast<float>(cos(y)));
	Multiply<Value_type>(rotY, vec, ans);
}


////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	
/*!	\brief		Applies rotation along the z vector.
**	\param		vec		Vector to rotate.
**	\param		z		angle in radians to rotate
**	\param		ans		place to store result.
**	\return		none
**	\exception	none
**	\par		DESCRIPTION:
**					
**	\author		Nate Cleveland
**	\date		11/4/2002 2:30:44 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
template<class Value_type>
void RotateZAxis(const InterpretationModule::InterpVec_type& vec, float z, InterpretationModule::InterpVec_type& ans)
{
	math::linear_alg::Matrix3<Value_type> rotZ(
		static_cast<float>(cos(z)),		static_cast<float>(sin(z)),		0,
		static_cast<float>(-sin(z)),	static_cast<float>(cos(z)),		0,
		0,								0,								1);
	Multiply<Value_type>(rotZ, vec, ans);
}


////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	DotProduct
/*!	\brief		Computes the dot product of the two given vectors.
**	\param		u		first vector.
**	\param		v		second vector
**	\return		dot product of two vectors
**	\exception	none					
**	\author		Nate Cleveland
**	\date		3/18/2003 1:41:41 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
template<class Value_type>
	Value_type DotProduct(const InterpretationModule::InterpVec_type& u, const InterpretationModule::InterpVec_type& v)
{
	return u[0]*v[0] + u[1]*v[1] + u[2]*v[2];
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	
/*!	\brief		Sees if the two given spheres collide.
**	\param		pntA		center point of first sphere.
**	\param		radiusA		radius of first sphere.
**	\param		pntB		center point of second sphere.
**	\param		radiusB		radius of second sphere.
**	\param		tolerance	How close to touching spheres must be before collision
							 is declared.
**	\return		true iff two spheres collide.
**	\exception	none				
**	\author		Nate Cleveland
**	\date		3/18/2003 1:39:31 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
bool SphereCollision(const InterpretationModule::InterpPnt_type& pntA, float radiusA,
					 const InterpretationModule::InterpPnt_type& pntB, float radiusB,
					 float tolerance = 0);

s8 RayBoxIntersection (Ray3D ray, Box3D box, Point3D outPoint[2]);
f64 PointDistance3D(const Point3D p1, const Point3D p2);