#include <math.h>
#include "../Nexception/Nexception.h"
#include "graphicClient.hpp"
#include <gl/gl.h>
#include "gfx/renderable.h"     ///< Changed by Jared to lower dependencies
#include "timeObject/timer.h"   ///< Added for framerate independence
#include "fishBodyMath.hpp"
#include <assert.h>
#include <iostream>
using namespace std;
using namespace InterpretationModule;
// memory leak detection
#ifdef _DEBUG
    #include"memleak.h"
    #define new DEBUG_NEW
    #undef THIS_FILE
    static char THIS_FILE[] = __FILE__;
#endif

//Comment out the below line and the fish will no longer display their collision
// sphere and target direction.
//#define DISPLAY_DIRECTION_AND_SPHERE


//Constant variables that may be needed by other objects.
namespace 
{
	//Model and texture names.
	/*const*/char*				DEFUALT_FISH_MODEL			= "Resources/Model/Fish.md2";
	/*const*/char*				SMALL_PREY_FISH_MODEL		= "Resources/Model/Sub-Models/Small Prey.md2";	
	/*const*/char*				SMALL_PREDATOR_FISH_MODEL	= "Resources/Model/Sub-Models/Small Predator.md2";	
	/*const*/char*				LARGE_PREY_FISH_MODEL		= "Resources/Model/Sub-Models/Large Prey.md2";	
	/*const*/char*				LARGE_PREDATOR_FISH_MODEL	= "Resources/Model/Sub-Models/Large Predator.md2";	
	/*const*/char*				DEFUALT_FISH_TEXTURE		= "Resources/Model/Fish Flanks.tga";
	/*const*/char*				FISH_ARMOR_TEXTURE			= "Resources/Model/Fish Flanks - Armored.tga";

	//Other constant variables.
	const float				DEAD_FISH_ROTATION			= pi;
	const InterpAngle_type	MAX_TURN_VALUE_PER_SECOND	= 10.0f * pi;
	const float				BASE_FRAMES_PER_SEC			= 30;
	const unsigned int		MIN_FRAME_BITE				= 0;
	const unsigned int		MAX_FRAME_BITE				= 8;
	const unsigned int		MIN_FRAME_SWIM				= 19;
	const unsigned int		MAX_FRAME_SWIM				= 34;
	const float				HIGH_META_THRESHOLD			= 0.6f;
	const float				LOW_META_THRESHOLD			= 0.4f;
	const float				LOW_HEALTH_THRESHOLD		= 0.5f;
	const char				BLINK_FRAME					= 0;		//frame to blink the sick tint on.
}



#include "../debugging/Debug.h"
namespace 
{
	char buff[1024] ={0};
}




/******************************************************************************
 *	Model_Texture struct
 */////////////////////////////////////////////////////////////////////////////

Model_Texture::Texture_type*	Model_Texture::m_defualtTexture		= NULL;
Model_Texture::Model_type*		Model_Texture::m_defualtModel		= NULL;
Model_Texture::Texture_type*		Model_Texture::m_armorTexture		= NULL;
//see header for documentation.
Model_Texture::Model_Texture()
//:m_defualtTexture(NULL), m_defualtModel(NULL)
{

}

//see header for documentation.
Model_Texture::~Model_Texture()
{
	Release();
}

void Model_Texture::Release(void) throw()
{
	//release memory allocated for defualt model and texture.
	if(m_defualtModel)
	{
		delete m_defualtModel;
		//Added by John 2/15/03
		//Static Pointer being deleted by multiple objects, caused game to crash
		//Set m_defaultModel to 0 
		m_defualtModel = 0;
	}

	if(m_defualtTexture)
	{
		delete m_defualtTexture;
		//Added by John 2/15/03
		//Static Pointer being deleted by multiple objects, caused game to crash
		//Set m_defualtTexture to 0
		m_defualtTexture = 0;
	}

	if(m_armorTexture)
	{
		delete m_armorTexture;
		m_armorTexture = 0;
	}
		
}

//see header for documentation.
Model_Texture::Model_type& Model_Texture::GetModel(void) const
{
	//check to see if the model has been already allocated.
	if(m_defualtModel)
		return *m_defualtModel;
	return *(m_defualtModel = new Model_type(SMALL_PREY_FISH_MODEL, SMALL_PREDATOR_FISH_MODEL, 
											 LARGE_PREY_FISH_MODEL, LARGE_PREDATOR_FISH_MODEL));
}

//see header for documentation.
Model_Texture::Texture_type& Model_Texture::GetTexture(void) const
{
	if(m_defualtTexture)
		return *m_defualtTexture;
	return *(m_defualtTexture = new Texture_type(DEFUALT_FISH_TEXTURE));
}

//see header for documentation.
Model_Texture::Texture_type& Model_Texture::GetArmorTexture(void) const
{
	if(m_armorTexture)
		return *m_armorTexture;
	return *(m_armorTexture = new Texture_type(FISH_ARMOR_TEXTURE));
}



/*******************************************************************************
 * Helper functions                                                             
 *//////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	AdvanceFrame
/*!	\brief		Advances the animation frame for the caller
**	\param		
**	\return		none
**	\exception	none
**	\par		DESCRIPTION:
**					
**	\author		Nate Cleveland
**	\date		3/14/2003 1:32:54 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
inline unsigned int AdvanceFrame(unsigned int currFrame, float& interpAmount,
								 unsigned int maxFrame, unsigned int minFrame)
{
	//advance iff the frame count is at least one.
	if(interpAmount > 1)
	{
		interpAmount = 0;
		currFrame++;

		//wrap the animation
		if(currFrame > maxFrame || currFrame < minFrame) 
			currFrame = minFrame;
	}
	
	return currFrame;
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	FindBlinkCount
/*!	\brief		Calculates the max blink count of a fish based on its health.
**	\param		health			HEalth to evaluate of the given fish.
**	\return		Ammount of frames to elapsed between sick blinks.
**	\exception	none
**	\par		DESCRIPTION:
**					
**	\author		Nate Cleveland
**	\date		4/1/2003 11:18:16 AM
**	\warning	asserts that the given health is less then LOW_HEALTH_THRESHOLD.
*/
////////////////////////////////////////////////////////////////////////////////
inline char FindBlinkCount(float health)
{
	assert(LOW_HEALTH_THRESHOLD >= health);
	return static_cast<char>(health * 100);	
}



////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	CheckForSickTint
/*!	\brief		Calculates the sick tinting logic.
**	\param		health			Current healt of the fish.
**	\param		blinkTimer		Reference to the m_blinkTimer parameter of a 
								 graphicClient.
**	\return		returns true if this fish should flash sickly.
**	\exception	none			
**	\author		Nate Cleveland
**	\date		4/1/2003 11:04:57 AM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
inline bool CheckForSickTint(float health, char& blinkTimer)
{
	//check to see if the health is low enough to blink.
	if(health < LOW_HEALTH_THRESHOLD) 
	{
		//if the blink timer is BLINK_FRAME, return true
		if(BLINK_FRAME == blinkTimer)
		{
			blinkTimer = FindBlinkCount(health);
			return true;
		}

		//otherwise just decrement the timer.
		else
			blinkTimer--;			
	}//end if

	//if it isn't flag the blink timer.
	else
		blinkTimer = 0;
	return false;
}

/******************************************************************************
 *	GraphicClient class.
 */////////////////////////////////////////////////////////////////////////////
//instantiate state struct
Model_Texture GraphicClient::m_assetMgr;


//see header for documentation.
GraphicClient::GraphicClient()
:m_pos(0,0,0),		m_currframeNum(0),		m_totalFrameNum(35),
 m_interpAmount(0), m_isHighlight(false),	m_isBiting(false)
{
	//zero out some stuff just to ensure they are zero.
	SetXYAngle(0);
	SetZYAngle(0);
	SetXZAngle(0);
	SetMetabolism(0);
}



//see header for documentation.
void GraphicClient::AnimationLogic(float& rotation) throw()
{
	//advance the animation iff we're alive.
	if(m_graphicAtts.m_health > 0)
    {  
		////special logic for each animation type
		//Eating/Attacking
		if(IsBiting()) 
		{
			m_interpAmount += BASE_FRAMES_PER_SEC * static_cast<float>(TimeObject::Instance()->getGameTimeStep());
			m_currframeNum = AdvanceFrame(m_currframeNum, m_interpAmount, MAX_FRAME_BITE, MIN_FRAME_BITE);
			
			if(MAX_FRAME_BITE == m_currframeNum)
				m_isBiting = false;
		}

		//Everything else
		else
		{
			m_interpAmount += BASE_FRAMES_PER_SEC *  static_cast<float>(TimeObject::Instance()->getGameTimeStep()*__max(0.1f,m_graphicAtts.m_speedRatio));
			m_currframeNum = AdvanceFrame(m_currframeNum, m_interpAmount, MAX_FRAME_SWIM, MIN_FRAME_SWIM);
		}
	}
}


//see header for documentation.
void GraphicClient::Render(void)
{
	////Update the Animation logic
	InterpVec_type	axis(0,0,0);			//axis to rotate the fish around.
	InterpPnt_type  pnt = m_pos;//(0,0,0);	//position of the fish.
	float			rotation = 0;			//holds rotation of fish in degrees.		
	AnimationLogic(rotation);

	glPushMatrix();
//		glTranslatef((GLfloat)m_pos[0], (GLfloat)m_pos[1], (GLfloat)m_pos[2]); 

		if(m_graphicAtts.m_health > 0) 
		{ 
//			glRotatef(RadianToDegree(GetXZAngle()), 0,1,0);
//			glRotatef(RadianToDegree(GetZYAngle()), 1,0,0);
			axis[1] = 1;
			rotation = RadianToDegree(GetXZAngle());

		}

		else
		{	//flip the fish upside down....
			axis[2]		= 1;
			rotation	= 180;//glRotatef(180, 0,0,1);
			m_currframeNum = MAX_FRAME_BITE / 2;
		}

/////////////////////////////////////////////////////////////////
//BEGIN NEW CODE BLOCK for fish rendering by Austin Spafford
// NOTE: This block of code does not use any extra functions or
//       constants so that it can be easily removed, which of
//       course also means that it contains Magic Numbers^TM.
//
//       m_smoothBlinkingAnimationTimer was added to the class 
//       to keep track of the current animation state.
//
//       Every visible blink starts at timer = 0, and depending
//       on the health level, either terminates early and then does not 
//       effect the fish until time = 1, or the fish is constantly
//       blinking where timer = 0.5 represents the apogee of the blink.

		const float c_lowHealthThreshold = 0.3f;
		const float c_blinkingFrequencyMin = 0.2f;
		const float c_blinkingFrequencyMax = 4;
		const float c_maximumBlinkDuration = 0.3f; // <-- in seconds
		const float c_blinkColor[3] = {0.1f, 0.1f, 1.0f};		
		const float c_pi = 3.1415926535f;

		// draw the fish at full brightness if no one has specified a color
		if(false == m_isHighlight)
			m_highLightColorRGB[0] = m_highLightColorRGB[1] = m_highLightColorRGB[2] = 1;

		// calculate the current health warning color of the fish (defaults to non-masking)
		float healthWarningColor[3] = {1, 1, 1};		
		if ((m_graphicAtts.m_health < c_lowHealthThreshold) && (m_graphicAtts.m_health > 0)) // also avoid dead fish
		{
			// use a linear warning magnitude
			float blinkingEffectMagnitude = 1 - ((float)m_graphicAtts.m_health / (float)c_lowHealthThreshold);

			// linearly generate a frequency from the min/max constants
			float blinkingFrequency = c_blinkingFrequencyMin + (c_blinkingFrequencyMax - c_blinkingFrequencyMin)*(blinkingEffectMagnitude);

			// update the blink animation timer based on the current frequency
			m_smoothBlinkingAnimationTimer += blinkingFrequency * static_cast<float>(TimeObject::Instance()->getRealTimeStep());
			// wrap the timer to the range of 0 to 1 (fmod() isn't used since it can generate negative values)
			m_smoothBlinkingAnimationTimer -= floorf(m_smoothBlinkingAnimationTimer);

			// figure out whether or not the frequency is lower that the lowest blink
			// frequency allowed (this generates an intermittent at higher health levels)
			// , and if it is not, then we blink the fish as quickly as possible
			float currentBlinkProgress = m_smoothBlinkingAnimationTimer;
			if (blinkingFrequency < (1.0f/c_maximumBlinkDuration))
			{
				// figure out how much of the current cycle is occupied by the blink time
				// (this is the ratio between the current and mininum frequencies
				float blinkCycleLength = blinkingFrequency * c_maximumBlinkDuration;

				// figure out whether or not the animation timer has skipped past 
				// the active blink time yet.				
				if (m_smoothBlinkingAnimationTimer > blinkCycleLength)
				{
					// force the fish to not be altered by the blink effect since it
					// is currently between blinks
					currentBlinkProgress = 0;
				}
				else
				{
					// calculate the progress through the small blink
					currentBlinkProgress = m_smoothBlinkingAnimationTimer / blinkCycleLength;
				}
			}

			// now that we know how far we are through the blink itself (with a
			// progress of 0 also meaning that there should be no effect), we
			// can figure out what the relative magnitude of the blink should be,
			// on a scale of 0 to 1.
			float relativeBlinkMagnitude = 0.5f*(-1*cosf(2*c_pi*currentBlinkProgress) + 1);

			// generate the blink masking color!
			healthWarningColor[0] = 1 - relativeBlinkMagnitude * (1 - c_blinkColor[0]);
			healthWarningColor[1] = 1 - relativeBlinkMagnitude * (1 - c_blinkColor[1]);
			healthWarningColor[2] = 1 - relativeBlinkMagnitude * (1 - c_blinkColor[2]);
		} // end if fish needs to blink

		// generate the final masking color used to render the fish
		float finalMaskingColor[3];
		finalMaskingColor[0] = m_highLightColorRGB[0] * healthWarningColor[0];
		finalMaskingColor[1] = m_highLightColorRGB[1] * healthWarningColor[1];
		finalMaskingColor[2] = m_highLightColorRGB[2] * healthWarningColor[2];

		// render the fish
		m_assetMgr.GetModel().Draw( m_graphicAtts.m_size, 
									m_graphicAtts.m_teeth, 
									m_graphicAtts.m_armor, 
									m_currframeNum,  
									(m_currframeNum+1) % m_totalFrameNum,
									m_interpAmount, 
									m_assetMgr.GetTexture(), 
									m_assetMgr.GetArmorTexture(), 
									pnt, 
									m_scale, 
									axis,			
									rotation, 
									finalMaskingColor[0], 
									finalMaskingColor[1],   
									finalMaskingColor[2]);

//END NEW CODE BLOCK for fish rendering by Ausitn Spafford
///////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////
/* BEGIN OLD CODE BLOCK for fish rendering by Nate Cleveland

		if(false == m_isHighlight)
			m_highLightColorRGB[0] = m_highLightColorRGB[1] = m_highLightColorRGB[2] = 1;

		////Render the fish to the scene using the graphic module's functionality
		//Check to see if we need to blink the fish sick
		if(CheckForSickTint(m_graphicAtts.m_health, m_blinkTimer) && 0!=m_graphicAtts.m_health)
		{
			m_assetMgr.GetModel().Draw( m_graphicAtts.m_size, 
										m_graphicAtts.m_teeth, 
										m_graphicAtts.m_armor,
										m_currframeNum,  
									    (m_currframeNum+1) % m_totalFrameNum,
										m_interpAmount,   
										m_assetMgr.GetTexture(), 
										m_assetMgr.GetArmorTexture(),
										pnt, 
										m_scale, 
										axis,			  
										rotation, 
										m_highLightColorRGB[0]/2, 
										m_highLightColorRGB[1]/2,   
										m_highLightColorRGB[2]/2);
		}

		//Otherwise just draw normally.
		else
		{
			m_assetMgr.GetModel().Draw( m_graphicAtts.m_size, 
									    m_graphicAtts.m_teeth, 
										m_graphicAtts.m_armor, 
										m_currframeNum,  
									    (m_currframeNum+1) % m_totalFrameNum,
										m_interpAmount, 
										m_assetMgr.GetTexture(), 
										m_assetMgr.GetArmorTexture(), 
										pnt, 
										m_scale, 
										axis,			
										rotation, 
										m_highLightColorRGB[0], 
										m_highLightColorRGB[1],   
										m_highLightColorRGB[2]);
		}

END OLD CODE BLOCK for fish rendering by Nate Cleveland */
///////////////////////////////////////////////////////////////

	glPopMatrix();

#	ifdef DISPLAY_DIRECTION_AND_SPHERE
/*		glPushMatrix();
			glTranslatef((GLfloat)m_pos[0], (GLfloat)m_pos[1], (GLfloat)m_pos[2]); 
			glutWireSphere(GetRadius(), 7, 7);


			glPushMatrix();
				glTranslatef(m_dir[0] * 4, m_dir[1] * 4, m_dir[2] * 4);
				glutWireCube(1);
			glPopMatrix();	
		glPopMatrix();
*/#	endif
}


//See header for documentation.
void GraphicClient::FindRotation(const InterpVec_type& currDirection, InterpVec_type& facingVec)throw()
{
	InterpVec_type actualDirection;


	//if the current direction is zero then use the facing vector.rather then the
	// currVector. That way a still fish will point the last direction it was going.
	if(fabsf(currDirection[0]) < .01f && 
	   fabsf(currDirection[1]) < .01f && 
	   fabsf(currDirection[2]) < .01f   ) 
	   {
			//Assume player wants to stand still.
			actualDirection = facingVec;
			
			
			//log status
///			sprintf(buff, "Use facing vector (%.2f, %.2f, %.2f)", 
//				facingVec[0], facingVec[1], facingVec[2]);
//			(Debugging::Debugger::Instance())->output(Debugging::BOD, Debugging::L3, buff);
	   }

	//otherwise just use the direction we're traveling.
	else
		actualDirection = currDirection;

	//unitize the vectors for simplicity later on.
	UnitizeVector<float>(actualDirection, actualDirection);

	//Find the two rotation angle for this vector.
	InterpAngle_type	xzPlane = 0, yzPlane = 0;
	yzPlane = AngleMgr::FindYZAngle(actualDirection);
	xzPlane = AngleMgr::FindXZAngle(actualDirection);

	//pass the angles over to the graphic client. 
	SetZYAngle(yzPlane);
	SetXZAngle(xzPlane);


	//Lastly, update the facing vector. Only do this if the direction vector is not zero.
	if(fabsf(currDirection[0]) > .01f || 
	   fabsf(currDirection[1]) > .01f || 
	   fabsf(currDirection[2]) > .01f   ) 
	   {
//			(Debugging::Debugger::Instance())->output(Debugging::BOD, Debugging::L3, "Update Face Vec");
			facingVec = currDirection;
	   }

	m_dir = actualDirection;
}



//See header for documentation
void GraphicClient::LimitRotation(InterpVec_type& currDirection, InterpVec_type& facingVec) throw()
{
	//if the current direction is zero then don't bother updating the rotation logic
	if(fabsf(currDirection[0]) < .01f && 
	   fabsf(currDirection[1]) < .01f && 
	   fabsf(currDirection[2]) < .01f   ) 
		return;

	//Translate into a unit vector
	UnitizeVector<float>(currDirection, currDirection);
	   
	//Find the two rotation angles for this vector.
	InterpAngle_type	xzPlane = 0, yzPlane = 0;
	yzPlane = AngleMgr::FindYZAngle(currDirection);
	xzPlane = AngleMgr::FindXZAngle(currDirection);

	//set up a max turn value.
	const InterpAngle_type maxAngleChange = MAX_TURN_VALUE_PER_SECOND * static_cast<float>(TimeObject::Instance()->getGameTimeStep());
//	cout << "M " <<maxAngleChange << "\t";



	//Find the change on the yzPlane
	InterpAngle_type yzPlaneChange = GetZYAngle() - AngleMgr::FindYZAngle(currDirection);
//	cout << "Y " <<yzPlaneChange << "\t";

/*	//clamp the change, then rotate by that much in the yzPlane ie the X axis.
	if(yzPlaneChange < 0)
		RotateXAxis<float>(currDirection, yzPlaneChange = -__min(fabsf(yzPlaneChange), maxAngleChange), currDirection);

	else
		RotateXAxis<float>(currDirection, yzPlaneChange = __min(yzPlaneChange, maxAngleChange), currDirection);

*/
 
	//Find the change on the xzPlane
	InterpAngle_type xzPlaneChange = GetXZAngle() - AngleMgr::FindXZAngle(currDirection);
//	cout << "X " <<xzPlaneChange << "\t";


	//clamp the change, the rotate the direction vector by that clamped value. 
	// in the xzPlane aka y-Axis
	
	while (xzPlaneChange > pi) xzPlaneChange -= 2*pi;
	while (xzPlaneChange < -pi) xzPlaneChange += 2*pi;
	if (xzPlaneChange > maxAngleChange) xzPlaneChange = maxAngleChange;
	if (xzPlaneChange < -maxAngleChange) xzPlaneChange = -maxAngleChange;
	RotateYAxis<float>(facingVec, xzPlaneChange / 2, currDirection);

	while (yzPlaneChange > pi) yzPlaneChange -= 2*pi;
	while (yzPlaneChange < -pi) yzPlaneChange += 2*pi;
	if (yzPlaneChange > maxAngleChange) yzPlaneChange = maxAngleChange;
	if (yzPlaneChange < -maxAngleChange) yzPlaneChange = -maxAngleChange;
	RotateXAxis<float>(currDirection, yzPlaneChange / 2, currDirection);
/*
	if(xzPlaneChange < 0)
		RotateYAxis<float>(facingVec, xzPlaneChange = -__min(fabsf(xzPlaneChange), maxAngleChange), currDirection);
	else
		RotateYAxis<float>(facingVec,  xzPlaneChange = __min(xzPlaneChange, maxAngleChange), currDirection);

*/
	m_dir = currDirection;
	

//	cout << "Fin X "<< xzPlaneChange << "\t";
//	cout << "Fin Y" << yzPlaneChange; 
//	cout << endl;
}//end LimitRotation


//See header for documentation
void GraphicClient::Highlight(bool isOn, float red/* = 1*/, float green/* = 1*/, float blue/* = 1*/) throw()
{
	//just copy over given values, let the renderer stress over the details.
	m_isHighlight			= isOn;
	m_highLightColorRGB[0]	= red;
	m_highLightColorRGB[1]	= green;
	m_highLightColorRGB[2]	= blue;

	//if we turned the highlighting off, reset the metabolism tinting.
	if(!isOn)
		SetMetabolism(m_graphicAtts.m_metabolism);
	
}

//See header for documentation
void GraphicClient::SetRadius(float radius) throw()
{
	m_scale[0] = m_scale[1] = m_scale[2] = radius;

	//strech out the z-dimension to make the fish look sleeker if it has high metabolism
	m_scale[2] += radius*(m_graphicAtts.m_metabolism - .5f);

	//widen the fish by streching the x-dimension, shows us fish with a full tummy
	m_scale[0] += radius*(m_graphicAtts.m_stomach - .75f); // Values adjusted by Austin on [4,11,03] to reduce obese appearance
}



//See header for documentation
void GraphicClient::SetMetabolism(float val)throw()
{
	m_graphicAtts.m_metabolism = val;

	//we only care about metabolism tinting if we aren't being highlighted 
	// by the ui
	if(!m_isHighlight) 
	{
		//factor in either a high metabolism or a low one.
		if(m_graphicAtts.m_metabolism > HIGH_META_THRESHOLD) 		
		{
			m_highLightColorRGB[0] = m_graphicAtts.m_metabolism / 2;
			m_highLightColorRGB[1] = 0.3f;
			m_highLightColorRGB[2] = 0.3f;
		}

		else if(m_graphicAtts.m_metabolism < LOW_META_THRESHOLD) 
		{
			m_highLightColorRGB[0] = 0.3f;
			m_highLightColorRGB[1] = 0.3f;
			m_highLightColorRGB[2] = 1 - m_graphicAtts.m_metabolism;
		}//end else if
	}//end if
}

float GraphicClient::GetRadius(void)const throw()
{
	//return the biggest axis, which for us is always the z.
	return m_scale[2];
}


//See header for documentation
void GraphicClient::BeginBite(void)
{
	m_isBiting	   = true;
	m_interpAmount = 0;
	m_currframeNum = MIN_FRAME_BITE;	
}


//Standard accessors, taken from the header so I can alter them without forcing a huge rebuild.
void  GraphicClient::SetPos			 (const InterpPnt_type& pos)throw()	{m_pos = pos;}
void  GraphicClient::SetSize		 (float val)throw(){m_graphicAtts.m_size = val;}
void  GraphicClient::SetSpeedRatio	 (float val)throw(){m_graphicAtts.m_speedRatio = val;}
void  GraphicClient::SetHealth		 (float val)throw(){m_graphicAtts.m_health = val;}
void  GraphicClient::SetArmor		 (float val)throw(){m_graphicAtts.m_armor = val;}
void  GraphicClient::SetStomachNormal(float val)throw(){m_graphicAtts.m_stomach = val;}
void  GraphicClient::SetTeeth		 (float val)throw(){m_graphicAtts.m_teeth = val;}
bool  GraphicClient::IsBiting		 (void)			   {return m_isBiting;}









/*
 *	
		float compAng  = GetXZAngle();
		



		if(compAng < 0)
			compAng += 2*pi;	
		if(rotation < 0)
			rotation += 2*pi;
		

		cout << "XZ: " << compAng << " ZY: " << rotation << endl;


		//quad III and II
		if(compAng  < 3*pi/2 && compAng  > pi  &&
		   rotation > pi/2   && rotation < 3*pi/2) 
		{
			cout << "Adj" << endl;
			rotation -= pi/2;
		}
		
		else if(compAng > pi/2 && compAng < 3*pi/2)//GetXZAngle() < -pi/2)///|| 
			rotation += pi;

		
			

 */
