
////////////////////////////////////////////////////////////////////////////////
/*!	\file		graphicClient.hpp
**	\brief		Defines all the functionality for the fishBody Graphics client.
**	\author		Nate Cleveland
**	\date		2/6/2003 10:55:19 AM
**	\par		DESCRIPTION:
				 Contains functionality for displaying graphics via the graphics
				  module. See the graphics module for more details. 			  
		
**					
**	\par		REMARKS:
**					
*/
////////////////////////////////////////////////////////////////////////////////
#pragma warning(push)
#pragma warning(disable: 4786)
#include "../gfx/gfx.h"
#include "angleMgr.hpp"
#pragma once



namespace InterpretationModule
{
	struct GraphicAttributes;

	////////////////////////////////////////////////////////////////////////////////
	/*!	\class		contains a copy of all needed models and textures for the graphic
	**				 client class.
	**	\brief		
	**	\author		Nate Cleveland
	**	\date		2/6/2003 11:01:05 AM
	**	\warning	Order of construction is not gaurenteed with static classes.
	**				 Never use static data in your ctor.
	*/
	////////////////////////////////////////////////////////////////////////////////
	struct Model_Texture
	{
		//Typedefs of the model and texture types.
		typedef gfx::MultiModel	Model_type;
		typedef gfx::Texture	Texture_type;

		static Model_type*		m_defualtModel;			//!< A defualt model of the fish.
		static Texture_type*	m_defualtTexture;		//!< A defualt texture of the fish.
		static Texture_type*	m_armorTexture;
		

		////////////////////////////////////////////////////////////////////////////////
		//	FUNCTION:	Release
		/*!	\brief		Frees all the static models and textures.
		**	\exception	none
		**	\author		Nate Cleveland
		**	\date		4/8/2003 11:03:26 AM
		**	\warning	none
		*/
		////////////////////////////////////////////////////////////////////////////////
		static void Release(void) throw();


		////////////////////////////////////////////////////////////////////////////////
		//	FUNCTION:	Model_Texture
		/*!	\brief		Loads all models and textures.
		**	\exception	Will throw an exception if unable to load any models or textures.
		**	\author		Nate Cleveland
		**	\date		2/6/2003 11:03:26 AM
		**	\warning	none
		*/
		////////////////////////////////////////////////////////////////////////////////
		Model_Texture();

		////////////////////////////////////////////////////////////////////////////////
		//	FUNCTION:	~Model_Texture
		/*!	\brief		Frees any memory allocated for textures and models.			
		**	\author		Nate Cleveland
		**	\date		2/6/2003 11:05:19 AM
		**	\warning	none
		*/
		////////////////////////////////////////////////////////////////////////////////
		~Model_Texture();


		////////////////////////////////////////////////////////////////////////////////
		//	FUNCTION:	
		/*!	\brief		Returns to the caller the defualt model.
		**	\param		none
		**	\return		Reference to defualt model.
		**	\exception	Throws exception if there is no defualt model.
		**	\author		Nate Cleveland
		**	\date		2/6/2003 11:10:22 AM
		**	\warning	none
		*/
		////////////////////////////////////////////////////////////////////////////////
		Model_type& GetModel(void) const;

		////////////////////////////////////////////////////////////////////////////////
		//	FUNCTION:	
		/*!	\brief		Returns to caller the defualt texture.
		**	\param		none
		**	\return		Reference to defualt texture.
		**	\exception	Throws exception if there is no defualt texture.
		**	\author		Nate Cleveland
		**	\date		2/6/2003 11:11:05 AM
		**	\warning	none
		*/
		////////////////////////////////////////////////////////////////////////////////
		Texture_type& GetTexture(void) const;
		Texture_type& GetArmorTexture(void) const;

	};



	////////////////////////////////////////////////////////////////////////////////
	/*!	\class		GraphicClient
	**	\brief		Client used to display a fish on the screen.
	**	\author		Nate Cleveland
	**	\date		11/25/2002 7:52:08 PM
	**	\warning	none
	*/
	////////////////////////////////////////////////////////////////////////////////
	class GraphicClient: public gfx::IRenderable3D, public AngleMgr
	{
	protected:
		static Model_Texture	m_assetMgr;		//!< Holds all models and textures for fish.
		InterpPnt_type	m_pos;					//!< Position of this object.
		unsigned int	m_currframeNum;			//!< frame of animation we're on.
		unsigned int	m_totalFrameNum;		//!< total Number of frames for this ani.
        float           m_interpAmount;			//!< number from 0 to 1 that is the interpolation amount
		bool			m_isHighlight;			//!< If set to true will highlight fish. 
		float			m_highLightColorRGB[3];	//!< Color to highlight fish, each is 0 to 1.
		bool			m_isBiting;				//!< Set to true if we're biting.
		InterpVec_type	m_scale;				//!< The scaling dimensions of this fish.
		char			m_blinkTimer;			//!< Whenever this is set to zero the fish
												//!<  flashes green or red to diplay that
												//!<  it is unhealthy.
		float			m_smoothBlinkingAnimationTimer; //!<This value cycles through 0 to 1
												//!< at a rate that is dependant on the
												//!< current health level of the fish.
		InterpVec_type	m_dir;					//!< Stores the direction of the fish.		
		
		////////////////////////////////////////////////////////////////////////////////
		/*!	\class		GraphicAttributes
		**	\brief		FishAtts gives the graphicClient an idea how to draw the fish.
		**	\author		Nate Cleveland
		**	\date		2/7/2003 12:07:34 PM
		**	\warning	ALL attributes must be normalized!!!
		*/
		////////////////////////////////////////////////////////////////////////////////
		struct GraphicAttributes
		{
			float	m_size;
			float	m_metabolism;
			float	m_speedRatio;
			float	m_health;
			float	m_armor;
			float	m_stomach;
			float	m_teeth;
			GraphicAttributes():m_size(0), m_metabolism(0), m_speedRatio(0), m_health(0), m_armor(0), m_stomach(0), m_teeth(0){}
		}m_graphicAtts;


		////////////////////////////////////////////////////////////////////////////////
		//	FUNCTION:	
		/*!	\brief		Updates this fish's animation logic. Called once per frame.
		**	\param		rotation		how much to rotate the fish by for this frame
		**	\return		none
		**	\exception	none				
		**	\author		Nate Cleveland
		**	\date		4/3/2003 10:38:47 AM
		**	\warning	none
		*/
		////////////////////////////////////////////////////////////////////////////////
		void AnimationLogic(float& rotation) throw();


	public:
		////////////////////////////////////////////////////////////////////////////////
		//	FUNCTION:	
		/*!	\brief		Defualt ctor for GraphicClient.
		**	\param		none
		**	\return		none
		**	\exception	none
		**	\par		DESCRIPTION:
						 Defualts the rotation matrix to the identity matrix.
		**					
		**	\author		Nate Cleveland
		**	\date		11/25/2002 7:50:42 PM
		**	\warning	none
		*/
		////////////////////////////////////////////////////////////////////////////////
		GraphicClient();

		////////////////////////////////////////////////////////////////////////////////
		//	FUNCTION:	
		/*!	\brief		Draws the fish to the screen
		**	\param		
		**	\return		none
		**	\exception	none
		**	\par		DESCRIPTION:
		**				 Draws the model to the screen.	
		**	\author		Nate Cleveland
		**	\date		11/25/2002 7:51:17 PM
		**	\warning	none
		*/
		////////////////////////////////////////////////////////////////////////////////
		void Render(void);


		//Simple accessors.
		void  SetRadius			(float radius)				throw();
		void  SetPos			(const InterpPnt_type& pos)	throw();
		void  SetSize			(float val)					throw();
		void  SetMetabolism		(float val)					throw();
		void  SetSpeedRatio		(float val)					throw();
		void  SetHealth			(float val)					throw();
		void  SetArmor			(float val)					throw();
		void  SetStomachNormal	(float val)					throw();
		void  SetTeeth			(float val)					throw();
		float GetRadius			(void)				  const throw();
		


		////////////////////////////////////////////////////////////////////////////////
		//	FUNCTION:	
		/*!	\brief		Takes the current direction of the fish and translates that into
		**				 a proper rotation for our graphical model.
		**
		**	\param		currDirection		Direction we're currently going
		**	\param		facingVec			Vector that we we're last facing. Will change
		**									 this value iff our currentDirection isn't 0.
		**
		**	\return		none
		**	\exception	none
		**	\par		DESCRIPTION:
		**					
		**	\author		Nate Cleveland
		**	\date		2/21/2003 10:25:04 AM
		**	\warning	none
		*/
		////////////////////////////////////////////////////////////////////////////////
		void FindRotation(const InterpVec_type& currDirection, InterpVec_type& facingVec)throw();


		////////////////////////////////////////////////////////////////////////////////
		//	FUNCTION:	
		/*!	\brief		
		**	\param		
		**	\return		none
		**	\exception	none
		**	\par		DESCRIPTION:
		**					
		**	\author		Nate Cleveland
		**	\date		3/12/2003 1:17:48 PM
		**	\warning	none
		*/
		////////////////////////////////////////////////////////////////////////////////
		void LimitRotation(InterpVec_type& desiredDirection, InterpVec_type& facingVec) throw();


		////////////////////////////////////////////////////////////////////////////////
		//	FUNCTION:	Highlight
		/*!	\brief		Highlights given fish with given color.
		**	\param		isOn	If set to true the fish will become highlighted. If
		**						 set to false it will be NOT highlighted.
		**	\param		red		Value of red componet, 0 to 1.
		**	\param		green	Value of green componet, 0 to 1
		**	\param		blue	Value of blue componet, 0 to 1
		**	\return		none
		**	\exception	none
		**	\par		DESCRIPTION:
		**					Instructs Graphic client to highlight or unhighlight a fish.
		**					 Pass true and a color value to turn highlighting on, or just
		**					 false to turn it off (you can send a color if you like, but
		**					 there would be no real point.
		**	\author		Nate Cleveland
		**	\date		3/13/2003 12:36:59 PM
		**	\warning	none
		*/
		////////////////////////////////////////////////////////////////////////////////
		void Highlight(bool isOn, float red = 1, float green = 1, float blue = 1) throw();
		bool IsHighlight(void)const throw() {return m_isHighlight;}
		

		////////////////////////////////////////////////////////////////////////////////
		//	FUNCTION:	BeginBite
		/*!	\brief		Starts the bite animation.
		**	\param		none
		**	\return		none
		**	\exception	none
		**	\author		Nate Cleveland
		**	\date		3/14/2003 12:48:32 PM
		**	\warning	none
		*/
		////////////////////////////////////////////////////////////////////////////////
		void BeginBite(void);

		////////////////////////////////////////////////////////////////////////////////
		//	FUNCTION:	IsBiting
		/*!	\brief		Returns true iff the fish is currently biting.
		**	\param		none
		**	\return		none
		**	\exception	none					
		**	\author		Nate Cleveland
		**	\date		3/14/2003 12:49:16 PM
		**	\warning	none
		*/
		////////////////////////////////////////////////////////////////////////////////
		bool IsBiting(void);
	};

}
#pragma warning(pop)