////////////////////////////////////////////////////////////////////////////////
/*!	\file		physicData.hpp
**	\brief		Records physical data about this fish for the attributes and stat
**				 classes.
**	\author		Nate Cleveland
**	\date		11/4/2002 8:56:20 AM
**	\par		DESCRIPTION:
**				Object retains the following for other classes
**					- Position (x,y,z)
**					- Length of distance vector.
**					- Boolean check for if this is a new voxel or not.
**					
**	\par		REMARKS:
**				This is created so that none of the attributes/stats classes need
**				 to know anything about the attributes and stats classes.
**					
*/
////////////////////////////////////////////////////////////////////////////////

#include "fishBodyUtils.hpp"
#pragma once
namespace InterpretationModule
{

	////////////////////////////////////////////////////////////////////////////////
	/*!	\class		PhysicData
	**	\brief		Object that records useful physical data for teh attribute/stat
	**				 class.
	**	\author		Nate Cleveland
	**	\date		11/4/2002 9:01:02 AM
	**	\warning	none
	*/
	////////////////////////////////////////////////////////////////////////////////
	class PhysicData
	{
	protected:
		//Position of fish
		float			m_x;			//!< Position on the x-axis of this fish.
		float			m_y;			//!< Position on the y-axis of this fish.
		float			m_z;			//!< Position on the z-axis of this fish.

		//the prevous voxel of this fish and flag for if this is new.
		Voxel_type		m_prevVoxel;	//!< Voxel fish was on last frame.
		bool			m_isNewVoxel;	//!< Set to true if this is a new voxel.

		//current speed of fish
		float			m_currDist;		//!< Distance traveled since last frame.


	public:

		//simple ctor for this object.
		////////////////////////////////////////////////////////////////////////////////
		//	FUNCTION:	
		/*!	\brief		simple ctor for this object.
		**	\param		x		Position on the x-axis of this fish.
		**	\param		y		Position on the y-axis of this fish.
		**	\param		z		Position on the z-axis of this fish.
		**	\param		voxel	Current voxel of this fish.
		**	\exception	none
		**	\par		DESCRIPTION:
		**					
		**	\author		Nate Cleveland
		**	\date		11/4/2002 9:01:54 AM
		**	\warning	none
		*/
		////////////////////////////////////////////////////////////////////////////////
		PhysicData(float x=0,float y=0, float z=0, const Voxel_type& voxel=0);

		// Generic Update method for the physicData object.
		////////////////////////////////////////////////////////////////////////////////
		//	FUNCTION:	
		/*!	\brief		Generic update method for PhysicData object.
		**	\param		elapsedTime		Time passed, in secs, since last call.
		**	\param		x				Position on the x-axis of this fish.
		**	\param		y				Position on the y-axis of this fish.
		**	\param		z				Position on the z-axis of this fish.			
		**	\param		currVoxel		Reference to current voxel of fish.
		**	\return		none
		**	\exception	none
		**	\par		DESCRIPTION:
		**				This method should be called before attributes/statistics object's
		**				 update methods are called. At this time voxel is defualted to 
		**				 zero.
		**	\author		Nate Cleveland
		**	\date		11/4/2002 9:09:58 AM
		**	\warning	none
		*/
		////////////////////////////////////////////////////////////////////////////////
		void Update(float elapsedTime, float x, float y, float z, 
					const Voxel_type& currVoxel=0) throw();

		//!Returns true if we are in a different voxel then from the last update.
		////////////////////////////////////////////////////////////////////////////////
		//	FUNCTION:	
		/*!	\brief		Quick way to see if the fish is currently in a new voxel.
		**	\param		none
		**	\return		Returns true if the fish has entered a new voxel.
		**	\exception	none
		**	\par		DESCRIPTION:
		**				Quick way to see if the fish is currently in a new voxel.	
		**	\author		Nate Cleveland
		**	\date		11/4/2002 9:08:27 AM
		**	\warning	none
		*/
		////////////////////////////////////////////////////////////////////////////////
		bool IsNewVoxel(void) const throw() {return m_isNewVoxel;}

		//Accessor methods
		float GetDist(void) const throw(){return m_currDist;}
		float GetXPos(void) const throw(){return m_x;}
		float GetYPos(void) const throw(){return m_y;}
		float GetZPos(void) const throw(){return m_z;}


		// Sets the position of fish. 
		////////////////////////////////////////////////////////////////////////////////
		//	FUNCTION:	
		/*!	\brief		Sets the position of fish. 
		**	\param		x				Position on the x-axis of this fish.
		**	\param		y				Position on the y-axis of this fish.
		**	\param		z				Position on the z-axis of this fish.
		**	\return		none
		**	\exception	none
		**	\par		DESCRIPTION:
		**				Will force the position of the fish and skip the calculation of 
		**				 the distance vector. Should only be called if the fish,
		**					- Teleports (very unlikely given design of game).
		**					- Position was not initialized in ctor.
		**	\author		Nate Cleveland
		**	\date		11/4/2002 9:17:04 AM
		**	\warning	Will not calculate the distance vector.
		*/
		////////////////////////////////////////////////////////////////////////////////
		void ForcePosition(float x, float y, float z) throw()
		{
			m_x = x; 
			m_y = y;
			m_z = z;
			m_currDist = 0;
		}
		
		////////////////////////////////////////////////////////////////////////////////
		//	FUNCTION:
		/*!	\brief		Writes all needed data to the stream
		**
		**	\param ostream&				object we'll stream.
		**	\param const  Attributes&	Reference to this class.
		**
		**	\return		A reference to the stream.
		**	\exception	none
		**	\par		DESCRIPTION: Writes all the needed data (Base and Reactive
		**				 Attributes) to the stream.
		**
		**	\author		Nate Cleveland
		**	\date		10/16/2002 6:30:59 PM
		**	\warning	none
		*/
		////////////////////////////////////////////////////////////////////////////////
		friend std::ostream& operator << (std::ostream&, const PhysicData &);

		////////////////////////////////////////////////////////////////////////////////
		//	FUNCTION:
		/*!	\brief		Reads in data from the stream.
		**
		**	\param ostream&				object we'll stream from.
		**	\param const  Attributes&	Reference to this class.
		**
		**	\return		A reference to the stream.
		**	\exception	none
		**	\par		DESCRIPTION: Reads data from the stream and resets this class
		**				 accordingly.
		**
		**	\author		Nate Cleveland
		**	\date		10/16/2002 6:31:04 PM
		**	\warning	none
		*/
		////////////////////////////////////////////////////////////////////////////////
		friend std::istream& operator>> (std::istream&, PhysicData &);

	};//end class PhysicData
}//end namespace InterpretationModule.