	////////////////////////////////////////////////////////////////////////////////
/*!	\file		stats.cpp
**	\brief		Defines all the methods declared in the file stats.hpp
**	\author		Nate Cleveland
**	\date		10/18/2002 8:54:54 AM
**	\par		DESCRIPTION: See stats.hpp for description.
**					
**	\par		REMARKS:
**					
*/
////////////////////////////////////////////////////////////////////////////////
#pragma warning(disable:4786)
#include "stats.hpp"
#include "attributes.hpp"
#include "physicData.hpp"
#include "../networking/NetData.h"
#include <fstream>
#include <stdio.h>
using namespace InterpretationModule;

// memory leak detection
#ifdef _DEBUG
    #include"memleak.h"
    #define new DEBUG_NEW
    #undef THIS_FILE
    static char THIS_FILE[] = __FILE__;
#endif

#define STATISTICS_LOG				//!< Remove this to stop all logging by this obj.
#ifdef STATISTICS_LOG

	//! Remove this to stop the statistics class from loggin to console.
#	define STATISTICS_LOG_CONSOLE		

	//! Remove this to stop the statistics class from loggin to file
#	define STATISTICS_LOG_FILE	.
#endif

#define STATISTICS_THROW






//See header for documentation.
Statistics::Statistics(void) throw()
:m_age(0.0f)
{
	//zero out the stat array.
	Reset();
}

//See header for documentation.
Statistics::~Statistics() throw()
{
	//nothing to do at this time.	
}



//updates any attribute related stats.
void Statistics::Update(Attributes & at, float elapsedTime) throw()
{
	//increment the health, mating inclination, stomach and energy of a fish.
	this->IncHealth(at.GetHealth() * elapsedTime);	
	this->IncIncline(at.GetIncline() * elapsedTime);
	this->IncStomach(at.GetStomach() * elapsedTime);
	this->IncEnergy(at.GetEnergyLevel() * elapsedTime);
	return;
}

//updates any physical related stats
void Statistics::Update(PhysicData& phyData, float elapsedTime) throw()
{
	IncDistance(phyData.GetDist());				//update distance travled.
	IncSpeed(phyData.GetDist() * elapsedTime);	//update speed.
	
	if(phyData.IsNewVoxel())					//check to see if we're at a new voxel
		IncVoxels(1.0f);						//increase voxel count by 1.
}

//updates everything.
void Statistics::Update(Attributes& at, PhysicData& phyData, float elapsedTime) throw()
{
	this->Update(at, elapsedTime);		//update attribute related stats
	this->Update(phyData, elapsedTime);	//update physical stats
	this->IncAge(elapsedTime);			//update age
	return;
}

//See header for documentation
void Statistics::Reset(void)
{
	//zero out the stat array.
	memset(this->m_summationArray, 0, SUMMATION_TYPE_COUNT * sizeof(stat_type));
	m_age = 0;
}


//See header for documentation
void Statistics::Write(std::ostream& outStream) const
{
	outStream << *this;
}

//See header for documentation
void Statistics::Read(std::istream& inStream)
{
	inStream  >> *this;
}


//See header for documentation
void Statistics::operator << (UDPNet::NetMessage& netMessage)
{
	memcpy(m_summationArray, netMessage.ReadNBytes(sizeof(stat_type) * SUMMATION_TYPE_COUNT),
		   sizeof(stat_type) * SUMMATION_TYPE_COUNT);
}


//See header for documentation
void Statistics::operator >> (UDPNet::NetMessage& netMessage)
{
	netMessage.WriteNBytes(	sizeof(stat_type) * SUMMATION_TYPE_COUNT, 
							reinterpret_cast<char *>(m_summationArray));	
}








using namespace std;

//See header for documentation
ostream& InterpretationModule::operator << (ostream &outStream, const Statistics &st)
{
	//Stream out a header for this section.
	outStream << "##Begin Fish Statistics: " << SUMMATION_TYPE_COUNT << "\n";

	//loop through and just slap the stats into the stream.
	for(int i = 0 ; i < SUMMATION_TYPE_COUNT; i++)
		outStream << st.m_summationArray[i] << " ";

	//stream out a footer.
	outStream << "\n##End Fish Statistics \n";

	return outStream;
}//end Statistics friend operator <<

//See header for documentation
istream& InterpretationModule::operator>> (istream& inStream, Statistics &st)
{
	char	buffer[256] = {0};					//temp buffer to read file in from.
	int		fileSummationCount = 0;
	

	//read in the header
	inStream.getline(buffer, 256, '\n');

	sscanf(buffer, "##Begin Fish Statistics: %d", &fileSummationCount);


	//check to see if there is an error in the stream.
	if(fileSummationCount != SUMMATION_TYPE_COUNT) 
	{
	
#		ifdef STATISTICS_LOG_CONSOLE
			//log the error.
#		endif

#		ifdef STATISTICS_LOG_FILE
			//Log the error
#		endif	

#		ifdef STATISTICS_THROW
			throw 0;
#		endif
	}//end if

	//if we enter this clause then the data was read in correctly.
	else
	{
		//churn through the file and read in all the data.
		for(int i = 0 ; i < SUMMATION_TYPE_COUNT ; i++)
			inStream >> st.m_summationArray[i];


	}//end else
	
	//read in the footer and the last '\n' thrust into the stream.
	inStream.getline(buffer, 256, '\n');
	inStream.getline(buffer, 256, '\n');

	return inStream;
}//end Statistics friend operator >>


