////////////////////////////////////////////////////////////////////////////////
/*!	\file		stats.hpp
**	\brief		Declares the Statistics class used by the interpretation module.
**	\author		Nate Cleveland
**	\date		10/17/2002 4:07:38 PM
**	\par		DESCRIPTION: The statistic object holds all the statisitics for
**				 a fish body. Member of the InterpretationModule.
**
**	\par		REMARKS: There are three groups of accessor methods used by this class.
**					- "Inc" methods. All Inc methods increase the internal counters
**					by the specified ammount. These internal counters are refered to
**					as the "Summation" of a stat.
**					- "Get" methods. All Get methods return the summation of a
**				 	particular stat.
**					- "GetAvg" methods. All "GetAvg" return a rate or average of a
**					stat based either on a specified period of time or simply the age
**				 	of the fish.
**
*/
////////////////////////////////////////////////////////////////////////////////

#pragma once
#include "../FileIO/memento.h"
#include <vector>


//forward declarations.
namespace UDPNet 
{
	class NetMessage;
};



namespace InterpretationModule
{
class PhysicData;
class Attributes;

//! This enum resolves to indexes into the summation array
//!  contained inside the Statistics class.
enum Summation_Indexes
{
	//physical summations
	SPEED,
	DISTANCE,
	VOXELS,

	//reactional summations.
	HEALTH,
	INCLINE,		//short hand for mating inclination
	STOMACH,
	ENERGY,

	//Activities
	KILLS,
	OFFSPRING,
	BEEN_ATTACKED,
	HAS_ATTACKED,
	PLANKTON,
	CHUM,
	/*
	 *	Also need members for picking up individual objects
	 */

	 SUMMATION_TYPE_COUNT	//!< This member resloves to number of elements of array.
};//end enum

////////////////////////////////////////////////////////////////////////////////
/*!	\class		Statistics.
**	\brief		An object that holds all the statistical data of a fish.
**	\author		Nate Cleveland
**	\date		10/18/2002 8:51:27 AM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
class Statistics: public FileMemento
{
protected:
	typedef float	stat_type;			//!< Type for the summation array
	float			m_age;				//!< Hold old the fish is in game seconds.

	//! An array that holds a summation of all the stats.
	stat_type m_summationArray[SUMMATION_TYPE_COUNT];

public:
	Statistics(void) throw();
	~Statistics() throw();

	//Age methods. Age automatically updated in Update(Attributes&, PhysicData&, float)
	void  IncAge(float elapsedTime_seconds)	throw(){m_age += elapsedTime_seconds;}
	float GetAge(void)						const throw(){return m_age;}


	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:
	/*!	\brief		Updates all Attribute related stats for the caller.
	**
	**	\param Attributes&			Attribute class to anaylize.
	**	\param float elapsedTime		Time since last update, in seconds.
	**
	**	\return		none
	**	\exception	none
	**	\par		DESCRIPTION: Updates the following data from the Attribute class.
	**					- Health
	**					- EnergyLevel
	**					- Mating Inclination
	**					- Stomach
	**	\par		All data taken will be scaled by the elapsedTime.
	**
	**	\author		Nate Cleveland
	**	\date		10/17/2002 4:10:31 PM
	**	\warning	Do not use this method if you are already calling
	**				 Update(Attributes& , PhysicData&, float) every frame.
	*/
	////////////////////////////////////////////////////////////////////////////////
	void Update(Attributes &, float elapsedTime) throw();

	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:
	/*!	\brief		Updates all Physics related stats for the caller.
	**	\param
	**
	**	\param PhysicData&			PhysicData class to anaylize.
	**	\param float elapsedTime		Time since last update, in seconds.
	**
	**	\exception	none
	**	\par		DESCRIPTION: Updates the following data from the PhysicData class.
	**					- Distance Traveled
	**					- Voxels Visited.
	**					- Speed.
	**
	**	\author		Nate Cleveland
	**	\date		10/18/2002 8:40:07 AM
	**	\warning	Do not use this method if you are already calling
	**				 Update(Attributes& , PhysicData&, float) every frame.
	*/
	////////////////////////////////////////////////////////////////////////////////
	void Update(PhysicData& , float elapsedTime) throw();

	//updates everything.
	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:
	/*!	\brief		Calls all other Update methods for the user.
	**	\param
	**	\return		none
	**	\exception	none
	**	\par		DESCRIPTION: Calls all other update methods.
	**
	**	\author		Nate Cleveland
	**	\date		10/18/2002 8:41:17 AM
	**	\warning	none
	*/
	////////////////////////////////////////////////////////////////////////////////
	void Update(Attributes& , PhysicData& , float elapsedTime) throw();
	

	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:	
	/*!	\brief		Resets the stats to their starting values.
	**	\param		none
	**	\return		none
	**	\exception	none
	**	\author		Nate Cleveland
	**	\date		1/16/2003 9:43:39 AM
	**	\warning	none
	*/
	////////////////////////////////////////////////////////////////////////////////
	void Reset(void);

	//methods to increase the values of the stats.
	void IncSpeed					(stat_type incValue)throw(){m_summationArray[SPEED] += incValue;}
	void IncDistance				(stat_type incValue)throw(){m_summationArray[DISTANCE] += incValue;}
	void IncVoxels					(stat_type incValue)throw(){m_summationArray[VOXELS] += incValue;}
	void IncHealth					(stat_type incValue)throw(){m_summationArray[HEALTH] += incValue;}
	void IncIncline					(stat_type incValue)throw(){m_summationArray[INCLINE] += incValue;}
	void IncStomach					(stat_type incValue)throw(){m_summationArray[STOMACH] += incValue;}
	void IncEnergy					(stat_type incValue)throw(){m_summationArray[ENERGY] += incValue;}
	void IncKills					(stat_type incValue)throw(){m_summationArray[KILLS] += incValue;}
	void IncOffspring				(stat_type incValue)throw(){m_summationArray[OFFSPRING] += incValue;}
	void IncBeen_Attacked			(stat_type incValue)throw(){m_summationArray[BEEN_ATTACKED] += incValue;}
	void IncHas_Attacked			(stat_type incValue)throw(){m_summationArray[HAS_ATTACKED] += incValue;}
	void IncPlankton				(stat_type incValue)throw(){m_summationArray[PLANKTON] += incValue;}
	void IncChum					(stat_type incValue)throw(){m_summationArray[CHUM] += incValue;}

	//Methods to return the avg (or rate) of a statistic. avg will be based on time specifed by caller.
	stat_type GetAvgSpeed			(float time)const throw(){return GetSpeed() / time;}
	stat_type GetAvgDistance		(float time)const throw(){return GetDistance() / time;}
	stat_type GetAvgVoxels			(float time)const throw(){return GetVoxels()/ time;}
	stat_type GetAvgHealth			(float time)const throw(){return GetHealth() / time;}
	stat_type GetAvgIncline			(float time)const throw(){return GetIncline() / time;}
	stat_type GetAvgStomach			(float time)const throw(){return GetStomach() / time;}
	stat_type GetAvgEnergy			(float time)const throw(){return GetEnergy() / time;}
	stat_type GetAvgKills			(float time)const throw(){return GetKills() / time;}
	stat_type GetAvgOffspring		(float time)const throw(){return GetOffspring() / time;}
	stat_type GetAvgBeen_Attacked	(float time)const throw(){return GetBeen_Attacked() / time;}
	stat_type GetAvgHas_Attacked	(float time)const throw(){return GetHas_Attacked() / time;}
	stat_type GetAvgPlankton		(float time)const throw(){return GetPlankton() / time;}
	stat_type GetAvgChum			(float time)const throw(){return GetChum() / time;}
	stat_type GetAvgTotal_Food		(float time)const throw(){return GetTotal_Food() / time;}

	//Methods to return the avg (or rate) of a statistic. avg will be based on the age of the fish.
	stat_type GetAvgSpeed			(void)		const throw(){return GetSpeed() / GetAge();}
	stat_type GetAvgDistance		(void)		const throw(){return GetDistance() / GetAge();}
	stat_type GetAvgVoxels			(void)		const throw(){return GetVoxels() / GetAge();}
	stat_type GetAvgHealth			(void)		const throw(){return GetHealth() / GetAge();}
	stat_type GetAvgIncline			(void)		const throw(){return GetIncline() / GetAge();}
	stat_type GetAvgStomach			(void)		const throw(){return GetStomach() / GetAge();}
	stat_type GetAvgEnergy			(void)		const throw(){return GetEnergy() / GetAge();}
	stat_type GetAvgKills			(void)		const throw(){return GetKills() / GetAge();}
	stat_type GetAvgOffspring		(void)		const throw(){return GetOffspring() / GetAge();}
	stat_type GetAvgBeen_Attacked	(void)		const throw(){return GetBeen_Attacked() / GetAge();}
	stat_type GetAvgHas_Attacked	(void)		const throw(){return GetHas_Attacked() / GetAge();}
	stat_type GetAvgPlankton		(void)		const throw(){return GetPlankton() / GetAge();}
	stat_type GetAvgChum			(void)		const throw(){return GetChum() / GetAge();}
	stat_type GetAvgTotal_Food		(void)		const throw(){return GetTotal_Food() / GetAge();}

	//Methods to return the summation of a stat.
	stat_type GetSpeed				(void)		const throw(){return m_summationArray[SPEED];}
	stat_type GetDistance			(void)		const throw(){return m_summationArray[DISTANCE];}
	stat_type GetVoxels				(void)		const throw(){return m_summationArray[VOXELS];}
	stat_type GetHealth				(void)		const throw(){return m_summationArray[HEALTH];}
	stat_type GetIncline			(void)		const throw(){return m_summationArray[INCLINE];}
	stat_type GetStomach			(void)		const throw(){return m_summationArray[STOMACH];}
	stat_type GetEnergy				(void)		const throw(){return m_summationArray[ENERGY];}
	stat_type GetKills				(void)		const throw(){return m_summationArray[KILLS];}
	stat_type GetOffspring			(void)		const throw(){return m_summationArray[OFFSPRING];}
	stat_type GetBeen_Attacked		(void)		const throw(){return m_summationArray[BEEN_ATTACKED];}
	stat_type GetHas_Attacked		(void)		const throw(){return m_summationArray[HAS_ATTACKED];}
	stat_type GetPlankton			(void)		const throw(){return m_summationArray[PLANKTON];}
	stat_type GetChum				(void)		const throw(){return m_summationArray[CHUM];}
	stat_type GetTotal_Food			(void)		const throw(){return m_summationArray[PLANKTON] + m_summationArray[CHUM];}


	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:
	/*!	\brief		Writes this class to the passed in stream
	**
	**	\param ostream&						Stream to write to.
	**
	**	\return		none
	**	\exception	none
	**	\par		DESCRIPTION: Simply calls the overloaded << operator.
	**
	**	\author		Nate Cleveland
	**	\date		10/17/2002 3:30:49 PM
	**	\warning	none
	*/
	////////////////////////////////////////////////////////////////////////////////
	void Write(std::ostream&) const;

	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:
	/*!	\brief		Reads data from the stream.
	**	\param istream&					Stream to read from.
	**	\return		none
	**	\exception	none
	**	\par		DESCRIPTION: Simply calls the overloaded >> operator.
	**
	**	\author		Nate Cleveland
	**	\date		10/17/2002 3:30:55 PM
	**	\warning	none
	*/
	////////////////////////////////////////////////////////////////////////////////
	void Read(std::istream&);


	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:	
	/*!	\brief		Reads from a netMessage and replaces itself with the data it 
					 finds there.
	**	\param		netMessage		obj to read from.
	**	\return		none
	**	\exception	Throws an exception if there is an error in the packet or the 
	**				 keys don't match.				
	**	\author		Nate Cleveland
	**	\date		2/6/2003 2:32:53 PM
	**	\warning	none
	*/
	////////////////////////////////////////////////////////////////////////////////
	void operator << (UDPNet::NetMessage& netMessage);


	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:	
	/*!	\brief		Fills a NetMessage with all the data for these attributes.
	**	\param		netMessage		obj to write to.
	**	\return		none
	**	\exception	Throws an exception iff there is an error with netMessage.
	**	\author		Nate Cleveland
	**	\date		2/6/2003 2:29:45 PM
	**	\warning	none
	*/
	////////////////////////////////////////////////////////////////////////////////
	void operator >> (UDPNet::NetMessage& netMessage);

	
	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:
	/*!	\brief		Writes all needed data to the stream
	**
	**	\param ostream&				object we'll stream.
	**	\param const  Attributes&	Reference to this class.
	**
	**	\return		A reference to the stream.
	**	\exception	none
	**	\par		DESCRIPTION: Writes all the needed data (Base and Reactive
	**				 Attributes) to the stream.
	**
	**	\author		Nate Cleveland
	**	\date		10/17/2002 3:30:59 PM
	**	\warning	none
	*/
	////////////////////////////////////////////////////////////////////////////////
	friend std::ostream& operator << (std::ostream&, const Statistics &);

		////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:
	/*!	\brief		Reads in data from the stream.
	**
	**	\param ostream&				object we'll stream from.
	**	\param const  Attributes&	Reference to this class.
	**
	**	\return		A reference to the stream.
	**	\exception	none
	**	\par		DESCRIPTION: Reads data from the stream and resets this class
	**				 accordingly.
	**
	**	\author		Nate Cleveland
	**	\date		10/17/2002 3:31:04 PM
	**	\warning	none
	*/
	////////////////////////////////////////////////////////////////////////////////
	friend std::istream& operator>> (std::istream&, Statistics &);
};//end class Statistics
}//end namespace InterpretationModule