////////////////////////////////////////////////////////////////////////////////
/*!	\file		updateTimer.hpp
**	\brief		Allows a user a simple reusable way to space out updates or any
				 other generic timed application.
**	\author		Nate Cleveland
**	\date		1/13/2003 11:01:24 AM
**	\par		DESCRIPTION:
**					- Allows user to customize both time between updates, type 
					  of variable to track update times and when elapsed timer
					  is reset. See accessor functions.
					- Method for testing if updateTimer has triggered. See () 
					  operators.
					- All time is in seconds.
					- Methods are all inline for maximum speed.
**	\par		REMARKS:
**					Time unit is seconds.
*/
////////////////////////////////////////////////////////////////////////////////
#include "../timeObject/Timer.h"
#include "../nexception/nexception.h"


////////////////////////////////////////////////////////////////////////////////
/*!	\class		UpdateTimer
**	\brief		template parameter allows user to select varible type for tracking
				time updates.
**	\author		Nate Cleveland
**	\date		1/13/2003 11:01:28 AM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
template<typename time_type>
class UpdateTimer
{
	protected:
		time_type	m_elapsedTime;			//!< time passed since last update.
		time_type	m_timeBetweenTrigger;	//!< ammount of time between each trigger.

		bool		m_isAutoReset;			//!< If set to true object will automatically
											//!<  reset itself after each trigger.
	public:
		////////////////////////////////////////////////////////////////////////////////
		//	FUNCTION:	
		/*!	\brief		defualt ctor for update timer class.

		**	\param		timeBetweenTrigger	Time user wishes to have between each trigger.
		**	\param		isAutoReset			If this is set to true when the object triggers
											 the update counter will be reset to zero.

		**	\exception	none
		**	\par		DESCRIPTION:
		**					Allows the user to customize update timer to their 
							 specifications.	
		**	\author		Nate Cleveland
		**	\date		1/13/2003 11:06:09 AM
		**	\warning	none
		*/
		////////////////////////////////////////////////////////////////////////////////
		UpdateTimer(time_type timeBetweenTrigger = 1, bool isAutoReset = true) throw()
			:m_elapsedTime(0), m_timeBetweenTrigger(timeBetweenTrigger), 
			m_isAutoReset(isAutoReset){}
		~UpdateTimer(){}



		////////////////////////////////////////////////////////////////////////////////
		//	FUNCTION:	
		/*!	\brief		Returns true if the timer countdown has elapsed.
		**	\param		elapsedTime		Time passed since last update.
		**	\return		True if timer has elapsed, false if not.
		**	\exception	none
		**	\par		DESCRIPTION:
		**					Increments the internal timer and returns true if the timer
							 has triggered.
		**	\author		Nate Cleveland
		**	\date		1/13/2003 11:08:29 AM
		**	\warning	none
		*/
		////////////////////////////////////////////////////////////////////////////////
		bool operator ()(time_type elapsedTime) throw()
		{
			m_elapsedTime += elapsedTime;							//update internal timer.
			if(m_elapsedTime < m_timeBetweenTrigger) return false;	//check to see if timer hasn't elapsed.
			if(m_isAutoReset) m_elapsedTime = 0;					//reset timer if needed.
			return true;											//notify user.
		}

		////////////////////////////////////////////////////////////////////////////////
		//	FUNCTION:	
		/*!	\brief		Returns true if timer countdown has elapsed.
		**	\param		none
		**	\return		True if timer has elapsed, false if not.
		**	\exception	Throws exception iff unable to access timeObject.
		**	\par		DESCRIPTION:
		**					Same as other overloaded () operator, except this method
							 automatically examines the gameTimeStep for the user.
		**	\author		Nate Cleveland
		**	\date		1/13/2003 11:17:04 AM
		**	\warning	none
		*/
		////////////////////////////////////////////////////////////////////////////////
		bool operator ()(void)
		{
			//get an instance of the timeObject. If null throw exception.
			TimeObject *timeObject = NULL;
			if(timeObject = TimeObject::Instance()) 
				return (*this)(static_cast<time_type>(timeObject->getGameTimeStep()));
			throw Nexception("Unable to access TimeObject",__FILE__,__LINE__,__DATE__,__TIME__);
		}

		////////////////////////////////////////////////////////////////////////////////
		//	FUNCTION:	
		/*!	\brief		Resets internal timer. Un-needed if set to auto reset.
		**	\param		none
		**	\return		none
		**	\exception	none
		**	\par		DESCRIPTION:
		**					
		**	\author		Nate Cleveland
		**	\date		1/13/2003 11:44:10 AM
		**	\warning	none
		*/
		////////////////////////////////////////////////////////////////////////////////
		void Reset(void)throw(){m_elapsedTime = 0;}
		

		////////////////Standard accessors.///////////////////////////////////////////////
		time_type	GetTimeBetweenTrigger	(void)const throw(){return m_timeBetweenTrigger;}
		time_type	GetElapsedTime			(void)const throw(){return m_elapsedTime;}
		bool		GetIsAutoReset			(void)const throw(){return m_isAutoReset;}
		void		SetTimeBetweenTrigger	(const time_type& timeBetweenTrigger)throw(){m_timeBetweenTrigger = timeBetweenTrigger;}
		void		SetElapsedTime			(const time_type& elapsedTime) const throw(){m_elapsedTime = elapsedTime;}
		void		SetIsAutoReset			(bool isAutoReset)throw(){m_isAutoReset = isAutoReset;}

		//time_type typedef. Sometimes usefull for nested template definitions.
		typedef time_type TimeUnit_type;
};
