////////////////////////////////////////////////////////////////////////////////
/*! \file       image.cpp
**  \brief      Implementation of image processing functions
**  \author     M Jared Finder
**  \date       10/30/2002 10:58:21 AM
*/
////////////////////////////////////////////////////////////////////////////////
#pragma warning(disable:4786)
#include"image.h"

////////////////////////////////////////////////////////////////////////////////
// Needed includes
////////////////////////////////////////////////////////////////////////////////
//#include"stdheaders.h"
#include<stdio.h>
#include<string.h>
#include<assert.h>
#include<sstream>
#include"tga.h"
#include"../nexception/nexception.h"

// memory leak detection
#ifdef _DEBUG
    #include"memleak.h"
    #define new DEBUG_NEW
    #undef THIS_FILE
    static char THIS_FILE[] = __FILE__;
#endif


namespace
{
    const int BPP = 4;
}


////////////////////////////////////////////////////////////////////////////////
//  FUNCTION:   
/*! \brief      Create an empty image null image
**  \param      
**  \return     none
**  \exception  none
**  \par        DESCRIPTION:
**                  
**  \author     M Jared Finder
**  \date       10/30/2002 10:59:00 AM
**  \warning    none
*/
////////////////////////////////////////////////////////////////////////////////
loader::Image::Image(void)
        : m_width(0), m_height(0)
{
}


////////////////////////////////////////////////////////////////////////////////
//  FUNCTION:   
/*! \brief      Create an image from the specified file
**  \param      fileName    file to load from
**  \return     none
**  \exception  none
**  \author     M Jared Finder
**  \date       10/30/2002 11:01:44 AM
**  \warning    none
*/
////////////////////////////////////////////////////////////////////////////////
loader::Image::Image(const char* fileName)
{
    if(!Load(fileName))
    {
        std::ostringstream o;
        o << "Internal Error:\n"
          << "Could not load image " << fileName;

        throw Nexception(o.str().c_str(),
                         __FILE__, __LINE__, __DATE__, __TIME__);
    }
}


////////////////////////////////////////////////////////////////////////////////
//  FUNCTION:   
/*! \brief      Create an image from raw data
**  \param      pData       Data to load from, in RGB or RGBA format
**  \param      width       Width of the data
**  \param      height      Height of the data
**  \param      bpp         Resolution of the data.
**  \return     none
**  \exception  none
**  \par        DESCRIPTION:
**                  If the resolution is 3bpp, then the data is in RGB format.
**                  If the resolution is 4bpp, the data is in RGBA.  No other
**                  formats are legal.
**  \author     M Jared Finder
**  \date       11/5/2002 5:48:30 PM
**  \warning    none
*/
////////////////////////////////////////////////////////////////////////////////
loader::Image::Image(void* pData, int width, int height)
{
    if(!Load(pData, width, height))
    {
        std::ostringstream o;
        o << "Internal Error:\n"
          << "Could not create image from internal data.";

        throw Nexception(o.str().c_str(),
                         __FILE__, __LINE__, __DATE__, __TIME__);
    }
}


////////////////////////////////////////////////////////////////////////////////
//  FUNCTION:   
/*! \brief      Destuctor
**  \param      
**  \return     none
**  \exception  none
**  \par        DESCRIPTION:
**                  
**  \author     M Jared Finder
**  \date       11/5/2002 5:51:05 PM
**  \warning    none
*/
////////////////////////////////////////////////////////////////////////////////
loader::Image::~Image(void)
{
}


////////////////////////////////////////////////////////////////////////////////
//  FUNCTION:   
/*! \brief      Load an image from the specified file
**  \param      fileName    file to load from
**  \return     true if the file was succesfully loaded, false if not.
**  \exception  none
**  \author     M Jared Finder
**  \date       10/30/2002 11:01:44 AM
**  \warning    none
*/
////////////////////////////////////////////////////////////////////////////////
bool loader::Image::Load(const char* fileName)
{
    //detect the file type and load
    size_t length = strlen(fileName);
    const char* extension = fileName + length - 4;

    typedef bool (Image::*PFN_LOAD) (const char *);  ///< member function pointer
    PFN_LOAD pmLoad;

    if(stricmp(extension, ".tga") == 0)
    {
        pmLoad = &Image::LoadTGA;
    }
    else if(stricmp(extension, ".bmp") == 0)
    {
        pmLoad = &Image::LoadBMP;
    }
    else if(stricmp(extension, ".jpg") == 0)
    {
        pmLoad = &Image::LoadJPG;
    }
    else if(stricmp(extension, ".png") == 0)
    {
        pmLoad = &Image::LoadPNG;
    }
    else
    {
        return false;
    }

    return (this->*pmLoad)(fileName);
}


////////////////////////////////////////////////////////////////////////////////
//  FUNCTION:   
/*! \brief      Load an image from raw data
**  \param      pData       Data to load from, in RGB or RGBA format
**  \param      width       Width of the data
**  \param      height      Height of the data
**  \param      bpp         Resolution of the data.
**  \return     true if the data was succesfully loaded, false if not.
**  \exception  none
**  \par        DESCRIPTION:
**                  If the resolution is 3bpp, then the data is in RGB format.
**                  If the resolution is 4bpp, the data is in RGBA.  No other
**                  formats are legal.
**  \author     M Jared Finder
**  \date       11/5/2002 5:48:30 PM
**  \warning    none
*/
////////////////////////////////////////////////////////////////////////////////
bool loader::Image::Load(void* pData, int width, int height)
{
    m_data.resize(width * height * 4);
    m_width     = width;
    m_height    = height;

    memcpy(&m_data.front(), pData, width * height * 4);

    return true;
}


////////////////////////////////////////////////////////////////////////////////
//  FUNCTION:   
/*! \brief      Save an image to the specified file
**  \param      fileName    file to save to
**  \return     none
**  \exception  none
**  \author     M Jared Finder
**  \date       10/30/2002 11:01:44 AM
**  \warning    none
*/
////////////////////////////////////////////////////////////////////////////////
bool loader::Image::Save(const char* fileName)
{
    //detect the file type and load
    size_t length = strlen(fileName);
    const char* extension = fileName + length - 4;

    typedef bool (Image::*PFN_SAVE) (const char *);  ///< member function pointer
    PFN_SAVE pmSave;

    if(stricmp(extension, ".tga") == 0)
    {
        pmSave = &Image::SaveTGA;
    }
    else if(stricmp(extension, ".bmp") == 0)
    {
        pmSave = &Image::SaveBMP;
    }
    else if(stricmp(extension, ".jpg") == 0)
    {
        pmSave = &Image::SaveJPG;
    }
    else if(stricmp(extension, ".png") == 0)
    {
        pmSave = &Image::SavePNG;
    }
    else
    {
        return false;
    }

    return (this->*pmSave)(fileName);
}


////////////////////////////////////////////////////////////////////////////////
//  FUNCTION:   
/*! \brief      Load an image from a Truevision Targa file
**  \param      fileName    name of the Targa to load.
**  \return     true if the file was succesfully loaded, false if not.
**  \exception  none
**  \par        DESCRIPTION:
**                  This code currently is taken right out of
**                  OpenGL Game Programming, pg. 223
**  \author     M Jared Finder
**  \date       11/5/2002 5:53:44 PM
**  \warning    none
*/
////////////////////////////////////////////////////////////////////////////////
bool loader::Image::LoadTGA(const char* fileName)
{
    TGAData data;

    //open the file
    FILE* file = fopen(fileName, "rb");
    if(file == NULL)
    {
        return false;
    }

    unsigned char   ucharTemp;
    short           shortTemp;

    // read tfirst two byes of data we don't need
    fread(&ucharTemp, sizeof(unsigned char), 1, file);
    fread(&ucharTemp, sizeof(unsigned char), 1, file);

    //read in the image type
    fread(&data.imageTypeCode, sizeof(unsigned char), 1, file);

    // for our purposes, the image type should be either a 2 or a 3
    if(data.imageTypeCode != 2 && data.imageTypeCode != 3)
    {
        fclose(file);
        return false;
    }

    // read 13 (?) byes of data we don't need
    fread(&shortTemp, sizeof(short), 1, file);
    fread(&shortTemp, sizeof(short), 1, file);
    fread(&ucharTemp, sizeof(unsigned char), 1, file);
    fread(&shortTemp, sizeof(short), 1, file);
    fread(&shortTemp, sizeof(short), 1, file);

    // read image dimensions
    fread(&data.imageWidth, sizeof(short), 1, file);
    fread(&data.imageHeight, sizeof(short), 1, file);

    // read image bit depth
    fread(&data.bitCount, sizeof(unsigned char), 1, file);

    // read 1 byte of data we don't need 
    fread(&ucharTemp, sizeof(unsigned char), 1, file);


    // allocate memory for image data and read it in
    m_width     = data.imageWidth;
    m_height    = data.imageHeight;
    m_data.resize(m_width * m_height * BPP, 255);

    // read in data
    size_t pixelSize = m_width * m_height;
    if(data.bitCount == 32)
    {
        for(size_t i = 0; i < pixelSize; ++i)
        {
            m_data[i*BPP + 2] = fgetc(file);
            m_data[i*BPP + 1] = fgetc(file);
            m_data[i*BPP + 0] = fgetc(file);
            m_data[i*BPP + 3] = fgetc(file);
        }
    }
    else
    {
        for(size_t i = 0; i < pixelSize; ++i)
        {
            m_data[i*BPP + 2] = fgetc(file);
            m_data[i*BPP + 1] = fgetc(file);
            m_data[i*BPP + 0] = fgetc(file);
        }
    }

	fclose(file);
    return true;
}

////////////////////////////////////////////////////////////////////////////////
//  FUNCTION:       
/*! \brief          Saves an image to a Truevision Targa file
**  \param          fileName    name of the Targa to load.
**  \return         true if the file was succesfully loaded, false if not.
**  \exception      none
**  \par            DESCRIPTION:
**  \author         Austin Spafford
**  \date           1/21/2003 8:47:24 PM
**  \warning        none
*/
////////////////////////////////////////////////////////////////////////////////
bool loader::Image::SaveTGA(const char* fileName)
{
	/// open the file
	FILE* file = fopen(fileName, "wb");
	if(file == NULL)
	{
		return false;
	}

	unsigned char	ucharTemp;
	short			shortTemp;

	// write the image ID field length
	ucharTemp = 0; // no image ID field
	fwrite(&ucharTemp, sizeof(unsigned char), 1, file);

	// write the color map type
	ucharTemp = 0; // no color map
	fwrite(&ucharTemp, sizeof(unsigned char), 1, file);

	// write the image type, which is an uncompressed true-color image
	ucharTemp = 2;
	fwrite(&ucharTemp, sizeof(unsigned char), 1, file);

	// write the colormap details (all zeroed out)
	shortTemp = 0;
	ucharTemp = 0;
	fwrite(&shortTemp, sizeof(short), 1, file);
	fwrite(&shortTemp, sizeof(short), 1, file);
	fwrite(&ucharTemp, sizeof(unsigned char), 1, file);
	
	// write the origin of the image
	shortTemp = 0;
	fwrite(&shortTemp, sizeof(short), 1, file);
    fwrite(&shortTemp, sizeof(short), 1, file);

	// write image dimensions
	fwrite(&m_width, sizeof(short), 1, file);
	fwrite(&m_height, sizeof(short), 1, file);

	// write the image bit depth
	ucharTemp = 32;
	fwrite(&ucharTemp, sizeof(unsigned char), 1, file);

	// write the image descriptor byte
	ucharTemp = 0; // only specify that the data starts in the bottom left corner
	fwrite(&ucharTemp, sizeof(unsigned char), 1, file);	

	// write the pixel data
	unsigned int numPixels = m_width * m_height;
	for(unsigned int i = 0; i < numPixels; i++)
	{
		fputc(m_data[i*4 + 2], file);
		fputc(m_data[i*4 + 1], file);
		fputc(m_data[i*4 + 0], file);
		fputc(m_data[i*4 + 3], file);
	}

	fclose(file);
	return true;
}

////////////////////////////////////////////////////////////////////////////////
/// Unimplemented
////////////////////////////////////////////////////////////////////////////////
bool loader::Image::LoadBMP(const char*)     { return false; }
bool loader::Image::LoadPNG(const char*)     { return false; }
bool loader::Image::LoadJPG(const char*)     { return false; }

bool loader::Image::SaveBMP(const char*)     { return false; }
bool loader::Image::SavePNG(const char*)     { return false; }
bool loader::Image::SaveJPG(const char*)     { return false; }
