////////////////////////////////////////////////////////////////////////////////
/*! \file       image.h
**  \brief      Image processing functions.
**  \author     M Jared Finder
**  \date       10/29/2002 8:53:56 PM
**  \par        DESCRIPTION:
**                  image.h contains general image functions.  Currently it can
**                  load the following formats:
**                      - TGA (24bit, 32bit)
**                  In the future, the following formats should be added:
**                      - BMP
**                      - JPG
**                      - PNG
*/
////////////////////////////////////////////////////////////////////////////////
#pragma once


////////////////////////////////////////////////////////////////////////////////
// Forward Declarations and needed includes
////////////////////////////////////////////////////////////////////////////////
#include<vector>


namespace loader
{
    ////////////////////////////////////////////////////////////////////////////////
    /*! \class      Image
    **  \brief      Supports loading and saving to various pixmap formats.
    **  \author     M Jared Finder
    **  \date       10/29/2002 10:10:52 PM
    **  \warning    Image determines the file format by looking at the extension.
    **              A .JPG is assumed to be a JPEG file, so if a JPEG has .BMP as
    **              its extension, the file will not be loaded correctly, if at all.
    */
    ////////////////////////////////////////////////////////////////////////////////
    class Image
    {
    public:
        ///create an empty image
        Image(void);
        ///create an image from a file
        Image(const char* fileName);
        ///create an image from raw data
        Image(void* pData, int width, int height);
        ///clean up any associated memory
        ~Image(void);

        ///reload the image from a file
        bool Load(const char* fileName);
        ///reload the image from raw data
        bool Load(void* pData, int width, int height);

        ///save the image to a file
        bool Save(const char* fileName);

        //get information about the image
        ///get the raw data
        unsigned char* GetData(void)    { return &m_data.front(); }
        //get bitmap info
        size_t GetWidth(void)   { return m_width;   }
        size_t GetHeight(void)  { return m_height;  }

    //helper functions
    protected:
        //loading functions
        bool LoadTGA(const char* fileName);
        bool LoadBMP(const char* fileName);
        bool LoadPNG(const char* fileName);
        bool LoadJPG(const char* fileName);

        //saving functions
        bool SaveTGA(const char* fileName);
        bool SaveBMP(const char* fileName);
        bool SavePNG(const char* fileName);
        bool SaveJPG(const char* fileName);

    private:
        ///pixmap's width
        int m_width;
        ///pixmap's height
        int m_height;

        ///the data
        std::vector<unsigned char> m_data;
    };
}
