////////////////////////////////////////////////////////////////////////////////
/*! \file       interp_model.cpp
**  \brief      Implementation of model loading.
**  \author     M Jared Finder
**  \date       1/23/2003 1:57:19 PM
*/
////////////////////////////////////////////////////////////////////////////////
#pragma warning(disable:4786)
#include"interp_model.h"

////////////////////////////////////////////////////////////////////////////////
// Needed includes
////////////////////////////////////////////////////////////////////////////////
//#include"stdheaders.h"
#include<stdio.h>
#include<string.h>
#include<assert.h>
#include<sstream>
#include"md2.h"
#include"../nexception/nexception.h"

// memory leak detection
#ifdef _DEBUG
    #include"memleak.h"
    #define new DEBUG_NEW
    #undef THIS_FILE
    static char THIS_FILE[] = __FILE__;
#endif


////////////////////////////////////////////////////////////////////////////////
//  FUNCTION:   
/*! \brief      Create a model from the specified file.
**  \param      fileName    file to load from
**  \return     none
**  \exception  none
**  \author     M Jared Finder
**  \date       1/23/2003 2:55:57 PM
**  \warning    none
*/
////////////////////////////////////////////////////////////////////////////////
loader::InterpModel::InterpModel(const char* fileName)
{
    if(!Load(fileName))
    {
        std::ostringstream o;
        o << "Internal Error:\n"
          << "Could not load model " << fileName;

        throw Nexception(o.str().c_str(),
                         __FILE__, __LINE__, __DATE__, __TIME__);
    }
}


////////////////////////////////////////////////////////////////////////////////
//  FUNCTION:   
/*! \brief      Destructor
**  \param      
**  \return     none
**  \exception  none
**  \author     M Jared Finder
**  \date       1/23/2003 2:56:37 PM
**  \warning    none
*/
////////////////////////////////////////////////////////////////////////////////
loader::InterpModel::~InterpModel(void)
{
}


////////////////////////////////////////////////////////////////////////////////
//  FUNCTION:   
/*! \brief      Load a model from a specified file.
**  \param      fileName    file to load from
**  \return     true on successful loading, false on failure
**  \exception  none
**  \par        DESCRIPTION:
**                  
**  \author     M Jared Finder
**  \date       1/23/2003 2:58:58 PM
**  \warning    none
*/
////////////////////////////////////////////////////////////////////////////////
bool
loader::InterpModel::Load(const char* fileName)
{
    //detect the file type and load
    size_t length = strlen(fileName);
    const char* extension = fileName + length - 4;

    typedef bool (InterpModel::*PFN_LOAD) (const char *);  ///< member function pointer
    PFN_LOAD pmLoad;

    if(stricmp(extension, ".md2") == 0)
    {
        pmLoad = &InterpModel::LoadMD2;
    }
    else
    {
        return false;
    }
 
    return (this->*pmLoad)(fileName);
}


////////////////////////////////////////////////////////////////////////////////
//  FUNCTION:   
/*! \brief      Load an MD2 model.
**  \param      fileName    file to load from
**  \return     true on successful loading, false on failure
**  \exception  none
**  \author     M Jared Finder
**  \date       1/23/2003 2:59:52 PM
**  \warning    none
**  \todo       Calculate vertex normals for lighting models.
*/
////////////////////////////////////////////////////////////////////////////////
bool
loader::InterpModel::LoadMD2(const char* fileName)
{
    //open the file
    FILE* pFile = fopen(fileName, "rb");
    if(pFile == NULL)
    {
        goto FAILURE;
    }

    //validate MD2 format
    MD2Header header;
    fseek(pFile, 0, SEEK_SET);
    fread(&header, sizeof(MD2Header), 1, pFile);
    fseek(pFile, 0, SEEK_END);
    if(     header.id != MD2Header::CorrectID
            || header.version != MD2Header::CorrectVersion
            || header.offsetEnd != ftell(pFile))
    {
        goto FAILURE;
    }

    //load texture coordinates
    {
        static short texCoordBuffer[MAX_TEX_COORDS * 2];
        fseek(pFile, header.offsetTexCoords, SEEK_SET);
        fread(&texCoordBuffer[0], sizeof(short), header.numTexCoords*2, pFile);
        m_textureCoords.resize(header.numTexCoords*2);

        std::vector<float>::iterator it = m_textureCoords.begin();
        short* pScaledTexCoord = &texCoordBuffer[0];
        while(it != m_textureCoords.end())
        {
            *it++ = *pScaledTexCoord++ / static_cast<float>(header.skinWidth);
            *it++ = 1 - *pScaledTexCoord++ / static_cast<float>(header.skinHeight);
        }
    }


    //load all triangles (change to command list?)
    {
        static MD2Triangle triangleBuffer[MAX_TRIANGLES];
        fseek(pFile, header.offsetTriangles, SEEK_SET);
        fread(&triangleBuffer, sizeof(MD2Triangle), header.numTriangles, pFile);
        m_triangles.resize(header.numTriangles);

        std::vector<ModelTriangle>::iterator it = m_triangles.begin();
        MD2Triangle* pTriangle = &triangleBuffer[0];
        while(it != m_triangles.end())
        {
            it->textureIndex[0] = pTriangle->textureIndices[0];
            it->textureIndex[1] = pTriangle->textureIndices[1];
            it->textureIndex[2] = pTriangle->textureIndices[2];
            it->vertexIndex[0]  = pTriangle->vertexIndices[0];
            it->vertexIndex[1]  = pTriangle->vertexIndices[1];
            it->vertexIndex[2]  = pTriangle->vertexIndices[2];

            ++it;
            ++pTriangle;
        }
    }

    //load all frames' vertex locations
    {
        fseek(pFile, header.offsetFrames, SEEK_SET);
        MD2FrameHeader frameBuffer;
        static MD2FrameVertex vertexBuffer[MAX_VERTICES];
        m_frames.resize(header.numFrames);

        std::vector<ModelFrame>::iterator it = m_frames.begin();
        while(it != m_frames.end())
        {
            fread(&frameBuffer, sizeof(MD2FrameHeader), 1, pFile);
            fread(&vertexBuffer[0],
                  sizeof(MD2FrameVertex), header.numVertices, pFile);

            //scale and translate the vertices
            it->vertexList.resize(header.numVertices * 3);
            std::vector<float>::iterator itVertex = it->vertexList.begin();
            MD2FrameVertex* pVertex = &vertexBuffer[0];
            while(itVertex != it->vertexList.end())
            {
                //transfer data WHILE changing coordinates.
                *itVertex++ = pVertex->vertex[0] * frameBuffer.scale[0]
                                + frameBuffer.translalte[0];
                *itVertex++ = pVertex->vertex[2] * frameBuffer.scale[2]
                                + frameBuffer.translalte[2];
                *itVertex++ = -(pVertex->vertex[1] * frameBuffer.scale[1]
                                + frameBuffer.translalte[1]);

                pVertex++;
            }

            //calculate the vertex normals
            it->normalList.resize(header.numVertices * 3);
            std::vector<float>::iterator itNormal = it->normalList.begin();
            while(itNormal != it->normalList.end())
            {
                ++itNormal;
            }

            ++it;
        }
    }

    //load command list
    {
        fseek(pFile, header.offsetGLCommands, SEEK_SET);
        m_commandList.resize(header.numGLCommands);

        //read data
        int numCommandsLeft = header.numGLCommands;
        std::vector<ModelCommandData>::iterator it = m_commandList.begin();
        while(numCommandsLeft > 0)
        {
            ModelCommandData header;

            fread(&header.i, sizeof(int), 1, pFile);
            it++->i = header.i;
            header.i = abs(header.i);
            for(int i = 0; i < header.i; ++i)
            {
                ModelCommandData vertex[3];

                fread(&vertex[0], sizeof(int), 3, pFile);

                it++->f = vertex[0].f;
                it++->f = 1 - vertex[1].f;
                it++->i = vertex[2].i;;
            }

            numCommandsLeft -= header.i*3 + 1;
        }
        fread(&m_commandList[0], sizeof(int), header.numGLCommands, pFile);
    }

    //return success or failure
    fclose(pFile);
    return true;
FAILURE:
    if(pFile != NULL)   { fclose(pFile);    }
    return false;
}
