////////////////////////////////////////////////////////////////////////////////
/*! \file       interp_model.h
**  \brief      Load a vertex interpolated model format into an intermediate
**              structure.
**  \author     M Jared Finder
**  \date       1/23/2003 1:44:59 PM
**  \par        DESCRIPTION:
**                  interp_model.h contains model translation functionality.
**                  Currently is can load the following formats:
**                      - MD2 (Quake 2 Model Format)
*/
////////////////////////////////////////////////////////////////////////////////
#pragma once


////////////////////////////////////////////////////////////////////////////////
// Forward Declarations and needed includes
////////////////////////////////////////////////////////////////////////////////
#include<vector>


namespace loader
{
    ////////////////////////////////////////////////////////////////////////////////
    /*! \struct     ModelTriangle
    **  \brief      Information that fully describes the triangle
    **  \author     M Jared Finder
    **  \date       1/23/2003 9:20:28 PM
    **  \warning    none
    */
    ////////////////////////////////////////////////////////////////////////////////
    struct ModelTriangle
    {
        int textureIndex[3];    //< location in texture coord array
        int vertexIndex[3];     //< location in vertex array
    };
    

    ////////////////////////////////////////////////////////////////////////////////
    /*! \struct     ModelFrame
    **  \brief      Holds the vertices and normals for each frame of animation.
    **  \author     M Jared Finder
    **  \date       1/24/2003 12:09:47 PM
    **  \warning    none
    */
    ////////////////////////////////////////////////////////////////////////////////
    struct ModelFrame
    {
        std::vector<float> vertexList;
        std::vector<float> normalList;
    };


    ////////////////////////////////////////////////////////////////////////////////
    /*! \union      ModelCommandData
    **  \brief      Stores the vertex indexes and the texture coordinates
    **  \author     M Jared Finder
    **  \date       1/31/2003 4:11:21 PM
    **  \warning    none
    */
    ////////////////////////////////////////////////////////////////////////////////
    union ModelCommandData
    {
        int     i;
        float   f;
    };


    ////////////////////////////////////////////////////////////////////////////////
    /*! \class      InterpModel
    **  \brief      Loads and saves from various interpolated vertex model formats.
    **  \author     M Jared Finder
    **  \date       1/23/2003 1:52:34 PM
    **  \warning    none
    */
    ////////////////////////////////////////////////////////////////////////////////
    class InterpModel
    {
    public:
        ///create an empty model (unimplemented)
        InterpModel(void)
            { throw;        }
        ///create a model from a file
        InterpModel(const char* fileName);
        ///create an model from raw data (unimplemented)
        InterpModel(void* pData, int width, int height)
            { throw;        }
        ///clean up any associated memory
        ~InterpModel(void);

        ///reload the model from a file
        bool Load(const char* fileName);
        ///reload the model from raw data (unimplemented)
        bool Load(void* pData, int width, int height)
            { return false; }

        ///save the model to a file (unimplemented)
        bool Save(const char* fileName)
            { return false; }

        //get information about the model
        const std::vector<float>&
        GetTextureCoordArray(void)
            { return m_textureCoords;   }
        const std::vector<ModelTriangle>&
        GetTriangleArray(void)
            { return m_triangles;       }
        const std::vector<ModelFrame>&
        GetFrameArray(void)
            { return m_frames;          }
        const std::vector<ModelCommandData>&
        GetCommandList(void)
            { return m_commandList;     }

    //helper functions
    protected:
        //loading functions
        bool LoadMD2(const char* fileName);

        //saving functions (unimplemented)
        bool SaveMD2(const char* fileName)
            { return false; }

    private:
        ///list of texture coordinates
        std::vector<float> m_textureCoords;
        ///list of triangles (replace later with GL_COMMAND_LIST)
        std::vector<ModelTriangle> m_triangles;
        ///frame list (specifies location of each vertex and its normal)
        std::vector<ModelFrame> m_frames;
        ///command list, MD2 style
        std::vector<ModelCommandData> m_commandList;
    };

}