////////////////////////////////////////////////////////////////////////////////
/*! \file       matrix2.h
**  \brief      2d matrix
**  \author     M Jared Finder
**  \date       10/4/2002 7:00:57 PM
**  \par        DESCRIPTION:
**                  Implementation of 2d matrix type, and implementation of
**                  2d matrix - 2d vector interaction.
**  \par        REMARKS:
**                  Because the usage of the functions is obvious, there will
**                  no function headers.
*/
////////////////////////////////////////////////////////////////////////////////
#pragma once


////////////////////////////////////////////////////////////////////////////////
// Forward Declarations and needed includes
////////////////////////////////////////////////////////////////////////////////
#include"vector2.h"
namespace math
{
    typedef unsigned int size_type;
}
namespace math { namespace linear_alg
{
    template<typename type> class Matrix2;
} }


namespace math { namespace linear_alg {
////////////////////////////////////////////////////////////////////////////////
/*! \class      const_Matrix2Sub
**  \brief      A helper class to allow double subscripting of const matrices
**  \author     M Jared Finder
**  \date       10/4/2002 9:55:43 PM
**  \warning    none
*/
////////////////////////////////////////////////////////////////////////////////
template<typename type>
class const_Matrix2Sub
{
public:
    const_Matrix2Sub(const Matrix2<type>& mat, size_type subscript)
        :  m_mat(mat),
        m_sub(subscript)
    {}
    const type& operator[](size_type sub)
    {
        return m_mat.Data()[2*sub + m_sub];
    }

private:
    const Matrix2<type>&    m_mat;
    /// This may be able to be replaced with just a pointer to the array element
    const size_type         m_sub;
};


////////////////////////////////////////////////////////////////////////////////
/*! \class      Matrix2Sub
**  \brief      A helper class to allow double subscripting of matrices
**  \author     M Jared Finder
**  \date       10/4/2002 9:55:43 PM
**  \warning    none
*/
////////////////////////////////////////////////////////////////////////////////
template<typename type>
class Matrix2Sub
{
public:
    Matrix2Sub(Matrix2<type>& mat, size_type subscript)
        :  m_mat(mat),
        m_sub(subscript)
    {}
    type& operator[](size_type sub)
    {
        return m_mat.Data()[2*sub + m_sub];
    }

private:
    Matrix2<type>&    m_mat;
    const size_type         m_sub;
};

////////////////////////////////////////////////////////////////////////////////
/*! \class      Matrix2<type>
**  \brief      A templatized 2d square matrix.
**  \author     M Jared Finder
**  \date       10/4/2002 7:04:05 PM
**  \warning    none
*/
////////////////////////////////////////////////////////////////////////////////
template<typename type>
class Matrix2
{
public:
    /// Elementwise constructor
    Matrix2(type row1col1 = 0, type row1col2 = 0,
            type row2col1 = 0, type row2col2 = 0)   throw();
    /// Data array constructor
    Matrix2(type* matrixData)               throw();
    /// Copy constructor
    Matrix2(const Matrix2<type>& rhs)       throw();
    /// Assignment operator
    Matrix2<type>&
    operator= (const Matrix2<type>& rhs)    throw();

    Matrix2<type>&
    operator+=(const Matrix2<type>& rhs)    throw();
    Matrix2<type>&
    operator-=(const Matrix2<type>& rhs)    throw();
    Matrix2<type>&
    operator*=(const Matrix2<type>& rhs)    throw();
    Matrix2<type>&
    operator*=(const type& rhs)             throw();

    /// const accessor
    const_Matrix2Sub<type> operator[](size_type)    const throw();
    /// non-const accessor
    Matrix2Sub<type> operator[](size_type)          throw();
    /// OpenGL accessor
    type* Data(void)                                const throw();

    /// Determinant
    type Det(void)                          const throw();

    Matrix2<type>& Transpose(void)          throw();
    Matrix2<type>& Invert(void);

private:
    type    m_data[4];
};

template<typename type>
const Matrix2<type>
operator+ (const Matrix2<type>& lhs, const Matrix2<type>& rhs)
{
    Matrix2<type> retVal(lhs);
    retVal += rhs;
    return retVal;
}

template<typename type>
const Matrix2<type>
operator- (const Matrix2<type>& lhs, const Matrix2<type>& rhs)
{
    Matrix2<type> retVal(lhs);
    retVal -= rhs;
    return retVal;
}

template<typename type>
const Matrix2<type>
operator* (const Matrix2<type>& lhs, const Matrix2<type>& rhs)
{
    type row1col1 = lhs[0][0]*rhs[0][0] + lhs[0][1]*rhs[1][0];
    type row2col1 = lhs[1][0]*rhs[0][0] + lhs[1][1]*rhs[1][0];

    type row1col2 = lhs[0][0]*rhs[0][1] + lhs[0][1]*rhs[1][1];
    type row2col2 = lhs[1][0]*rhs[0][1] + lhs[1][1]*rhs[1][1];

    Matrix2<type> retValue(row1col1,    row1col2,
                           row2col1,    row2col2);

    return retValue;
}

template<typename type>
const Vector2<type>
operator* (const Matrix2<type>& lhs, const Vector2<type>& rhs)
{
    Vector2<type> retValue(lhs[0][0]*rhs[0] + lhs[0][1]*rhs[1],
                           lhs[1][0]*rhs[0] + lhs[1][1]*rhs[1]);

    return retValue;

}

template<typename type>
const Matrix2<type>
operator* (const Matrix2<type>& lhs, const type& rhs)
{
    Matrix2<type> retVal(lhs);
    retVal *= rhs;
    return retVal;
}

template<typename type>
const Matrix2<type>
operator* (const type& lhs, const Matrix2<type>& rhs)
{
    Matrix2<type> retVal(rhs);
    retVal *= lhs;
    return retVal;
}


//export workaround
#include "matrix2.cpp"


} }
