////////////////////////////////////////////////////////////////////////////////
/*! \file       matrix3.h
**  \brief      3d matrix
**  \author     M Jared Finder
**  \date       10/4/2002 7:00:57 PM
**  \par        DESCRIPTION:
**                  Implementation of 3d matrix type, and implementation of
**                  3d matrix - 3d vector interaction.
**  \par        REMARKS:
**                  Because the usage of the functions is obvious, there will
**                  no function headers.
*/
////////////////////////////////////////////////////////////////////////////////
#pragma once


////////////////////////////////////////////////////////////////////////////////
// Forward Declarations and needed includes
////////////////////////////////////////////////////////////////////////////////
#include"vector3.h"
namespace math
{
    typedef unsigned int size_type;
}
namespace math { namespace linear_alg
{
    template<typename type> class Matrix3;
} }


namespace math { namespace linear_alg {
////////////////////////////////////////////////////////////////////////////////
/*! \class      const_Matrix3Sub
**  \brief      A helper class to allow double subscripting of const matrices
**  \author     M Jared Finder
**  \date       10/4/2002 9:55:43 PM
**  \warning    none
*/
////////////////////////////////////////////////////////////////////////////////
template<typename type>
class const_Matrix3Sub
{
public:
    const_Matrix3Sub(const Matrix3<type>& mat, size_type subscript)
        :  m_mat(mat),
        m_sub(subscript)
    {}
    const type& operator[](size_type sub)
    {
        return m_mat.Data()[3*sub + m_sub];
    }

private:
    const Matrix3<type>&    m_mat;
    /// This may be able to be replaced with just a pointer to the array element
    const size_type         m_sub;
};


////////////////////////////////////////////////////////////////////////////////
/*! \class      Matrix3Sub
**  \brief      A helper class to allow double subscripting of matrices
**  \author     M Jared Finder
**  \date       10/4/2002 9:55:43 PM
**  \warning    none
*/
////////////////////////////////////////////////////////////////////////////////
template<typename type>
class Matrix3Sub
{
public:
    Matrix3Sub(Matrix3<type>& mat, size_type subscript)
        :  m_mat(mat),
        m_sub(subscript)
    {}
    type& operator[](size_type sub)
    {
        return m_mat.Data()[3*sub + m_sub];
    }

private:
    Matrix3<type>&    m_mat;
    const size_type   m_sub;
};

////////////////////////////////////////////////////////////////////////////////
/*! \class      Matrix3<type>
**  \brief      A templatized 3d square matrix.
**  \author     M Jared Finder
**  \date       10/4/2002 7:04:05 PM
**  \warning    none
*/
////////////////////////////////////////////////////////////////////////////////
template<typename type>
class Matrix3
{
public:
    /// Elementwise constructor
    Matrix3(type r1c1 = 0, type r1c2 = 0, type r1c3=0,
            type r2c1 = 0, type r2c2 = 0, type r2c3=0,
            type r3c1 = 0, type r3c2 = 0, type r3c3=0)  throw();
    /// Data array constructor
    Matrix3(type* matrixData)               throw();
    /// Copy constructor
    Matrix3(const Matrix3<type>& rhs)       throw();
    /// Assignment operator
    Matrix3<type>&
    operator= (const Matrix3<type>& rhs)    throw();

    Matrix3<type>&
    operator+=(const Matrix3<type>& rhs)    throw();
    Matrix3<type>&
    operator-=(const Matrix3<type>& rhs)    throw();
    Matrix3<type>&
    operator*=(const Matrix3<type>& rhs)    throw();
    Matrix3<type>&
    operator*=(const type& rhs)             throw();

    /// const accessor
    const_Matrix3Sub<type> operator[](size_type)    const throw();
    /// non-const accessor
    Matrix3Sub<type> operator[](size_type)          throw();
    /// OpenGL accessor
    type* Data(void)                                const throw();

    /// Determinant
    type Det(void)                          const throw();

    Matrix3<type>& Transpose(void)          throw();
    Matrix3<type>& Invert(void);

private:
    type    m_data[9];
};

template<typename type>
const Matrix3<type>
operator+ (const Matrix3<type>& lhs, const Matrix3<type>& rhs)
{
    Matrix3<type> retVal(lhs);
    retVal += rhs;
    return retVal;
}

template<typename type>
const Matrix3<type>
operator- (const Matrix3<type>& lhs, const Matrix3<type>& rhs)
{
    Matrix3<type> retVal(lhs);
    retVal -= rhs;
    return retVal;
}


template<typename type>
const Matrix3<type>
operator* (const Matrix3<type>& lhs, const Matrix3<type>& rhs)
{
    type r1c1 = lhs[0][0]*rhs[0][0] + lhs[0][1]*rhs[1][0] + lhs[0][2]*rhs[2][0];
    type r2c1 = lhs[1][0]*rhs[0][0] + lhs[1][1]*rhs[1][0] + lhs[1][2]*rhs[2][0];
    type r3c1 = lhs[2][0]*rhs[0][0] + lhs[2][1]*rhs[1][0] + lhs[2][2]*rhs[2][0];

    type r1c2 = lhs[0][0]*rhs[0][1] + lhs[0][1]*rhs[1][1] + lhs[0][2]*rhs[2][1];
    type r2c2 = lhs[1][0]*rhs[0][1] + lhs[1][1]*rhs[1][1] + lhs[1][2]*rhs[2][1];
    type r3c2 = lhs[2][0]*rhs[0][1] + lhs[2][1]*rhs[1][1] + lhs[2][2]*rhs[2][1];

    type r1c3 = lhs[0][0]*rhs[0][2] + lhs[0][1]*rhs[1][2] + lhs[0][2]*rhs[2][2];
    type r2c3 = lhs[1][0]*rhs[0][2] + lhs[1][1]*rhs[1][2] + lhs[1][2]*rhs[2][2];
    type r3c3 = lhs[2][0]*rhs[0][2] + lhs[2][1]*rhs[1][2] + lhs[2][2]*rhs[2][2];

    Matrix3<type> retValue(r1c1, r1c2, r1c3,
                           r2c1, r2c2, r2c3,
                           r3c1, r3c2, r3c3);

    return retValue;
}


template<typename type>
const Vector3<type>
operator* (const Matrix3<type>& lhs, const Vector3<type>& rhs)
{
    Vector3<type> retValue(lhs[0][0]*rhs[0] + lhs[0][1]*rhs[1] + lhs[0][2]*rhs[2],
                           lhs[1][0]*rhs[0] + lhs[1][1]*rhs[1] + lhs[1][2]*rhs[2],
                           lhs[2][0]*rhs[0] + lhs[2][1]*rhs[1] + lhs[2][2]*rhs[2]);

    return retValue;

}

template<typename type>
const Matrix3<type>
operator* (const Matrix3<type>& lhs, const type& rhs)
{
    Matrix3<type> retVal(lhs);
    retVal *= rhs;
    return retVal;
}

template<typename type>
const Matrix3<type>
operator* (const type& lhs, const Matrix3<type>& rhs)
{
    Matrix3<type> retVal(rhs);
    retVal *= lhs;
    return retVal;
}


//export workaround
#include "matrix3.cpp"


} }

