////////////////////////////////////////////////////////////////////////////////
/*! \file       matrix4.h
**  \brief      4d matrix
**  \author     M Jared Finder
**  \date       10/4/2002 7:00:57 PM
**  \par        DESCRIPTION:
**                  Implementation of 4d matrix type, and implementation of
**                  4d matrix - 4d vector interaction.
**  \par        REMARKS:
**                  Because the usage of the functions is obvious, there will
**                  no function headers.
*/
////////////////////////////////////////////////////////////////////////////////
#pragma once


////////////////////////////////////////////////////////////////////////////////
// Forward Declarations and needed includes
////////////////////////////////////////////////////////////////////////////////
#include"vector4.h"
namespace math
{
    typedef unsigned int size_type;
}
namespace math { namespace linear_alg
{
    template<typename type> class Matrix4;
} }



namespace math { namespace linear_alg {
////////////////////////////////////////////////////////////////////////////////
/*! \class      const_Matrix4Sub
**  \brief      A helper class to allow double subscripting of const matrices
**  \author     M Jared Finder
**  \date       10/4/2002 9:55:43 PM
**  \warning    none
*/
////////////////////////////////////////////////////////////////////////////////
template<typename type>
class const_Matrix4Sub
{
public:
    const_Matrix4Sub(const Matrix4<type>& mat, size_type subscript)
        :  m_mat(mat),
        m_sub(subscript)
    {}
    const type& operator[](size_type sub)
    {
        return m_mat.Data()[4*sub + m_sub];
    }

private:
    const Matrix4<type>&    m_mat;
    /// This may be able to be replaced with just a pointer to the array element
    const size_type         m_sub;
};


////////////////////////////////////////////////////////////////////////////////
/*! \class      Matrix4Sub
**  \brief      A helper class to allow double subscripting of matrices
**  \author     M Jared Finder
**  \date       10/4/2002 9:55:43 PM
**  \warning    none
*/
////////////////////////////////////////////////////////////////////////////////
template<typename type>
class Matrix4Sub
{
public:
    Matrix4Sub(Matrix4<type>& mat, size_type subscript)
        :  m_mat(mat),
        m_sub(subscript)
    {}
    type& operator[](size_type sub)
    {
        return m_mat.Data()[4*sub + m_sub];
    }

private:
    Matrix4<type>&    m_mat;
    const size_type   m_sub;
};

////////////////////////////////////////////////////////////////////////////////
/*! \class      Matrix4<type>
**  \brief      A templatized 4d square matrix.
**  \author     M Jared Finder
**  \date       10/4/2002 7:04:05 PM
**  \warning    none
*/
////////////////////////////////////////////////////////////////////////////////
template<typename type>
class Matrix4
{
public:
    /// Elementwise constructor
    Matrix4(type r1c1 = 0, type r1c2 = 0, type r1c3=0, type r1c4=0,
            type r2c1 = 0, type r2c2 = 0, type r2c3=0, type r2c4=0,
            type r3c1 = 0, type r3c2 = 0, type r3c3=0, type r3c4=0,
            type r4c1 = 0, type r4c2 = 0, type r4c3=0, type r4c4=0
           )  throw();
    /// Data array constructor
    Matrix4(type* matrixData)               throw();
    /// Copy constructor
    Matrix4(const Matrix4<type>& rhs)       throw();
    /// Assignment operator
    Matrix4<type>&
    operator= (const Matrix4<type>& rhs)    throw();

    Matrix4<type>&
    operator+=(const Matrix4<type>& rhs)    throw();
    Matrix4<type>&
    operator-=(const Matrix4<type>& rhs)    throw();
    Matrix4<type>&
    operator*=(const Matrix4<type>& rhs)    throw();
    Matrix4<type>&
    operator*=(const type& rhs)             throw();

    /// const accessor
    const_Matrix4Sub<type> operator[](size_type)    const throw();
    /// non-const accessor
    Matrix4Sub<type> operator[](size_type)          throw();
    /// OpenGL accessor
    type* Data(void)                                const throw();

    /// Determinant
    type Det(void)                          const throw();

    Matrix4<type>& Transpose(void)          throw();
    Matrix4<type>& Invert(void);

private:
    type    m_data[16];
};

template<typename type>
const Matrix4<type>
operator+ (const Matrix4<type>& lhs, const Matrix4<type>& rhs)
{
    Matrix3<type> retVal(lhs);
    retVal += rhs;
    return retVal;
}

template<typename type>
const Matrix4<type>
operator- (const Matrix4<type>& lhs, const Matrix4<type>& rhs)
{
    Matrix4<type> retVal(lhs);
    retVal -= rhs;
    return retVal;
}


template<typename type>
const Matrix4<type>
operator* (const Matrix4<type>& lhs, const Matrix4<type>& rhs)
{
    type r1c1 = lhs[0][0]*rhs[0][0] + lhs[0][1]*rhs[1][0] + lhs[0][2]*rhs[2][0] + lhs[0][3]*rhs[3][0];
    type r2c1 = lhs[1][0]*rhs[0][0] + lhs[1][1]*rhs[1][0] + lhs[1][2]*rhs[2][0] + lhs[1][3]*rhs[3][0];
    type r3c1 = lhs[2][0]*rhs[0][0] + lhs[2][1]*rhs[1][0] + lhs[2][2]*rhs[2][0] + lhs[2][3]*rhs[3][0];
    type r4c1 = lhs[3][0]*rhs[0][0] + lhs[3][1]*rhs[1][0] + lhs[3][2]*rhs[2][0] + lhs[3][3]*rhs[3][0];

    type r1c2 = lhs[0][0]*rhs[0][1] + lhs[0][1]*rhs[1][1] + lhs[0][2]*rhs[2][1] + lhs[0][3]*rhs[3][1];
    type r2c2 = lhs[1][0]*rhs[0][1] + lhs[1][1]*rhs[1][1] + lhs[1][2]*rhs[2][1] + lhs[1][3]*rhs[3][1];
    type r3c2 = lhs[2][0]*rhs[0][1] + lhs[2][1]*rhs[1][1] + lhs[2][2]*rhs[2][1] + lhs[2][3]*rhs[3][1];
    type r4c2 = lhs[3][0]*rhs[0][1] + lhs[3][1]*rhs[1][1] + lhs[3][2]*rhs[2][1] + lhs[3][3]*rhs[3][1];

    type r1c3 = lhs[0][0]*rhs[0][2] + lhs[0][1]*rhs[1][2] + lhs[0][2]*rhs[2][2] + lhs[0][3]*rhs[3][2];
    type r2c3 = lhs[1][0]*rhs[0][2] + lhs[1][1]*rhs[1][2] + lhs[1][2]*rhs[2][2] + lhs[1][3]*rhs[3][2];
    type r3c3 = lhs[2][0]*rhs[0][2] + lhs[2][1]*rhs[1][2] + lhs[2][2]*rhs[2][2] + lhs[2][3]*rhs[3][2];
    type r4c3 = lhs[3][0]*rhs[0][2] + lhs[3][1]*rhs[1][2] + lhs[3][2]*rhs[2][2] + lhs[3][3]*rhs[3][2];

    type r1c4 = lhs[0][0]*rhs[0][3] + lhs[0][1]*rhs[1][3] + lhs[0][2]*rhs[2][3] + lhs[0][3]*rhs[3][3];
    type r2c4 = lhs[1][0]*rhs[0][3] + lhs[1][1]*rhs[1][3] + lhs[1][2]*rhs[2][3] + lhs[1][3]*rhs[3][3];
    type r3c4 = lhs[2][0]*rhs[0][3] + lhs[2][1]*rhs[1][3] + lhs[2][2]*rhs[2][3] + lhs[2][3]*rhs[3][3];
    type r4c4 = lhs[3][0]*rhs[0][3] + lhs[3][1]*rhs[1][3] + lhs[3][2]*rhs[2][3] + lhs[3][3]*rhs[3][3];

    Matrix4<type> retValue(r1c1, r1c2, r1c3, r1c4,
                           r2c1, r2c2, r2c3, r2c4,
                           r3c1, r3c2, r3c3, r3c4,
                           r4c1, r4c2, r4c3, r4c4);

    return retValue;
}

template<typename type>
const Vector4<type>
operator* (const Matrix4<type>& lhs, const Vector4<type>& rhs)
{
    Vector4<type> retValue(
        lhs[0][0]*rhs[0] + lhs[0][1]*rhs[1] + lhs[0][2]*rhs[2] + lhs[0][3]*rhs[3],
        lhs[1][0]*rhs[0] + lhs[1][1]*rhs[1] + lhs[1][2]*rhs[2] + lhs[1][3]*rhs[3],
        lhs[2][0]*rhs[0] + lhs[2][1]*rhs[1] + lhs[2][2]*rhs[2] + lhs[2][3]*rhs[3],
        lhs[3][0]*rhs[0] + lhs[3][1]*rhs[1] + lhs[3][2]*rhs[2] + lhs[3][3]*rhs[3]);

    return retValue;

}


template<typename type>
const Matrix4<type>
operator* (const Matrix4<type>& lhs, const type& rhs)
{
    Matrix4<type> retVal(lhs);
    retVal *= rhs;
    return retVal;
}

template<typename type>
const Matrix4<type>
operator* (const type& lhs, const Matrix4<type>& rhs)
{
    Matrix4type> retVal(rhs);
    retVal *= lhs;
    return retVal;
}


//export workaround
#include "matrix4.cpp"


} }
