////////////////////////////////////////////////////////////////////////////////
/*! \file       vector3.h
**  \brief      3d vector
**  \author     M Jared Finder
**  \date       10/4/2002 2:42:23 PM
**  \par        DESCRIPTION:
**                  Implementation of a templated 3d vector type.
**  \par        REMARKS:
**                  Because the usage of the functions is obvious, there will
**                  no function headers.
*/
////////////////////////////////////////////////////////////////////////////////
#pragma once

////////////////////////////////////////////////////////////////////////////////
// Forward Declarations and needed includes
////////////////////////////////////////////////////////////////////////////////
namespace math
{
    typedef unsigned int size_type;
}


namespace math { namespace linear_alg {
////////////////////////////////////////////////////////////////////////////////
/*! \class      Vector3<type>
**  \brief      3d Vector, templatized.
**  \author     M Jared Finder
**  \date       10/4/2002 2:42:23 PM
**  \warning    none
*/
////////////////////////////////////////////////////////////////////////////////
template<typename type>
class Vector3
{
public:

    /// Element-wise constructor
    Vector3(type element1=0, type element2=0,
            type element3=0)                            throw();
    /// Data array constructor
    Vector3(type* vectorData)                           throw();
    /// Copy constructor
    Vector3(const Vector3<type>&)                       throw();
    /// Assignment operator
    Vector3& operator= (const Vector3<type>&)           throw();

    Vector3& operator+= (const Vector3<type>&)  throw();
    Vector3& operator-= (const Vector3<type>&)  throw();
    Vector3& operator*= (type)                  throw();
    Vector3& operator/= (type)                  throw();

    const type& operator[] (size_type)  const throw();
    type& operator[] (size_type)        throw();
    /// Returns the data
    const type* Data(void)              const throw();

    /// Length of vector
    type Magnitude(void)        const throw();
    /// Normalizes the vector
    Vector3& Normalize(void);

private:
    type    m_data[3];
};

template<typename type>
const Vector3<type>
operator+(const Vector3<type>& lhs, const Vector3<type>& rhs) throw();

template<typename type>
const Vector3<type>
operator-(const Vector3<type>& lhs, const Vector3<type>& rhs) throw();

template<typename type>
const type
Dot(const Vector3<type>& lhs, const Vector3<type>& rhs) throw();

template<typename type>
const Vector3<type>
Cross(const Vector3<type>& lhs, const Vector3<type>& rhs) throw();

template<typename type>
const Vector3<type>
operator*(const Vector3<type>& lhs, type rhs) throw()
{
    Vector3<type> retVal(lhs);
    retVal *= rhs;
    return retVal;
}

template<typename type>
const Vector3<type>
operator*(type rhs, const Vector3<type>& lhs) throw()
{
    Vector3<type> retVal(lhs);
    retVal *= rhs;
    return retVal;
}


// export workaround
#include "vector3.cpp"


} }
