////////////////////////////////////////////////////////////////////////////////
/*!	\file		MessageHandler.cpp
**	\brief		Implementation of all non-Key-Generation functionality
**	\author		Nathan Frost
**	\date		10/17/2002 8:17:14 PM
**	\par		DESCRIPTION:
				
**					
**	\par		REMARKS:
				Currently, GenerateKey() generates an Official Key half the time,
				and an Unofficial Key otherwise
**					
*/
////////////////////////////////////////////////////////////////////////////////
#include<math.h>
#include<sstream>
#include<string>
#include "MessageHandler.h"
#include "../tankSimulator/TankSimulator.h"
#include "../interpretationModule/fishBody.hpp"
#include "../ai_core/AI_System.h"
#include "../networking/GameNetworking.h"
#include "networking/NetData.h"
#include "networking/NetMessageID.h"
#include "ai_core/AI_BrainManager.h"
#include "../interpretationModule/commandPacket.hpp"
#include "../gui/GUI_Includes.h"
#include "../interpretationModule/StrObj.hpp"
#include "../fileIO/fileIO.h"
#include <io.h>
#include <direct.h>
using namespace MsgHandler;
// memory leak detection
#ifdef _DEBUG
    #include"memleak.h"
    #define new DEBUG_NEW
    #undef THIS_FILE
    static char THIS_FILE[] = __FILE__;
#endif

namespace 
{
	const char *LOAD_SAVE_PATH = "";//"Saved Games\\";
}


////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	MessageHandler()
/*!	\brief		Constructor
**	\param		none
**	\return		none
**	\exception	none
**	\par		DESCRIPTION:
				- Is protected since MessageHandler is a Singleton -- use Instance() to access
				the object.
**				- Initializes curKey to 1.
				- Initializes gameOn to true -- the game defaults to active
**	\author		Nathan Frost
**	\date		10/17/2002 8:23:25 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
MessageHandler::MessageHandler():curKey(1), gameOn(true)
{
	//by default, the game is a server
	gameNetworking = new UDPNet::Server();
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	MessageHandler
/*!	\brief		Destructor
**	\param		none
**	\return		none
**	\exception	none
**	\par		DESCRIPTION:
**					
**	\author		Nathan Frost
**	\date		3/11/2003 12:46:32 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
MessageHandler::~MessageHandler()
{
	delete gameNetworking;
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	NukeAllFish
/*!	\brief		Destroys all Fish in the tankSimulator
**	\param		none
**	\return		none
**	\exception	none
**	\par		DESCRIPTION:
**					
**	\author		Nathan Frost
**	\date		3/12/2003 7:06:04 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
void MessageHandler::NukeAllFish()
{
	//get the first fish
	Key key = tankSimulator->GetNextFish(0);
	while(key != 0)
	{
		tankSimulator->RemoveFishBody(key);
		key = tankSimulator->GetNextFish(0);
	}
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	Logic
/*!	\brief		Should be called once each game loop
**	\param		none
**	\return		none
**	\exception	none
**	\par		DESCRIPTION:
**					
**	\author		Nathan Frost
**	\date		3/11/2003 3:15:35 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
void MessageHandler::Logic()
{
	gameNetworking->DoNetworking();
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	GetPosition
/*!	\brief		Fills out the passed vector with the position of the object
				related to the given key.
	\param		COMMENT ME
	\param		COMMENT ME
**	\return		none
**	\exception	none
**	\par		DESCRIPTION:
**					- Fills out the passed vector with the position of the object
					related to the given key. Returns false if unable to deduce
					the position of this object.
					- Eventually this must be filled out, it would be helpful for one of the 
					data filters (distance to object).  If this is not defined then the percept 
					generator itself will need to write a similar method. Suggested to be placed 
					here since it might be usefull for other parts of the game besides data 
					filters.
**	\author		Nate Cleveland
**	\date		11/25/2002 2:56:14 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
bool MessageHandler::GetPosition(Key key, math::linear_alg::Vector3<float>& pos ) throw()
{
	InterpretationModule::FishBody *fishBody = 0;
	if(fishBody = tankSimulator->GetFishBodyPtr(key))
	{
		fishBody->GetPos(pos);
		return true;
	}
	else
	{
		return false;
	}
}


////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	GetDirection
/*!	\brief		Fills out the passed vector with the direction of the object
				related to the given key.
	\param		COMMENT ME
	\param		COMMENT ME
**	\return		none
**	\exception	none
**	\par		DESCRIPTION:
**					- Fills out the passed vector with the direction of the object
					related to the given key. Returns false if unable to deduce
					the direction of this object.
					- Eventually this must be filled out, it would be helpful for one of the 
					data filters (distance to object).  If this is not defined then the percept 
					generator itself will need to write a similar method. Suggested to be placed 
					here since it might be usefull for other parts of the game besides data 
				Nathan Frost
**	\date		11/25/2002 2:56:14 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
bool MessageHandler::GetDirection(Key key, math::linear_alg::Vector3<float>& pos) throw()
{
	InterpretationModule::FishBody *fishBody = 0;
	if(fishBody = tankSimulator->GetFishBodyPtr(key))
	{
		fishBody->GetDirection(pos);
		return true;
	}
	else
	{
		return false;
	}
}



////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	GetPlanktonDensity
/*!	\brief		Returns to the caller the plankton density at given location.
**	\param		pos			Position to find plankton density from.
**	\return		none
**	\exception	none
**	\par		DESCRIPTION:
**				 Accesses the TankSimulator's plankton density methods.	

**	\author		Nate Cleveland
**	\date		11/25/2002 5:39:38 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
float MessageHandler::GetPlanktonDensity(const math::linear_alg::Vector3<float>& pos) const throw()
{
	return 0;
/* 
	//Get a ref to the TankSimulator
	//...
	TankSimulator& tankSim = blah;
	

	//Use that to find the voxel index of this location
	math::linear_alg::Vector3<int> voxelIndex;
	voxelIndex = tankSim.GetContainingVoxelIndex(pos);


	//Use that index to find the plankton density of this location.
	return tankSim.GetPlanktonDensity(voxelIndex);
*/

}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	GetFishBody
/*!	\brief		Retrieves a const pointer to the specified FishBody
	\param		Body:	FishBody to get a pointer to
**	\return		A pointer to the FishBody; will be NULL if the requested FishBody does not exist
**	\exception	none
**	\par		DESCRIPTION:
**					
**	\author		Nathan Frost
**	\date		11/25/2002 1:48:29 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
const InterpretationModule::FishBody *MessageHandler::GetFishBody(const Key Body)const throw()
{
	return tankSimulator->GetFishBodyPtr(Body);
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	GameOn
/*!	\brief		Returns whether or not the game is active, as per usual, or if it is in the process
				of quitting.
**	\param		none
**	\return		True if the game is running, false if the game is in the process of quitting
**	\exception	none
**	\par		DESCRIPTION:
**					
**	\author		Nathan Frost
**	\date		11/25/2002 9:17:11 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
bool MessageHandler::GameOn()
{
	return gameOn;
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	QuitGame
/*!	\brief		Triggers the game's quitting process
**	\param		none
**	\return		none
**	\exception	none
**	\par		DESCRIPTION:
					- Will need to do a lot more than just this in the future (networking, 
					destruction of other objects, etc)
**					
**	\author		Nathan Frost
**	\date		11/25/2002 9:17:11 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
void MessageHandler::QuitGame()
{
	gameOn = false;
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	GetMatingFishListRef
/*! \brief      Calls the TankSimulator to retrieve a reference to the mating fish list
**  \return     a reference to the mating fish list
**  \exception  none
**  \author     Nathan Frost
**  \date       11/25/2002 9:17:11 PM
**  \warning    none
*/
////////////////////////////////////////////////////////////////////////////////
std::list<Key> &MessageHandler::GetMatingFishListRef()
{
	return tankSimulator->GetMatingFishListRef();
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	RegisterTankSimulator
/*!	\brief		Registers the TankSimulator with the MessageHandler, and provides the TankSimulator
				with a reference to the MessageHandler.  
**	\param		tankSimulator
**	\return		none
**	\exception	none
**	\par		DESCRIPTION:
					- Must be called once in and around startup for the game to function properly.
**					
**	\author		Nathan Frost
**	\date		11/25/2002 9:34:23 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
MessageHandler &MessageHandler::RegisterTankSimulator(TankSimulator *TS)
{
	tankSimulator = TS;
	return *this;
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	GetDeleteFishListRef
/*! \brief      Retrieves a reference to the delete fish list
**  \return     Reference to the delete fish list
**  \exception  none
**	\author		Nathan Frost
**	\date		11/26/2002 11:47:57 AM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
std::list<Key> &MessageHandler::GetDeleteFishListRef()
{
	return tankSimulator->GetDeleteFishListRef();
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	RemoveFishBody
/*! \brief      Removes a FishBody with the specified Key
**  \param      Body:	The Key value that will be used as a label for the FishBody that is to be 
						removed
**  \return     true if the specified FishBody was removed, false if the specified FishBody does 
				not exist in the simulation
**  \exception  none
**	\author		Nathan Frost
**	\date		11/26/2002 11:49:39 AM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
bool MessageHandler::RemoveFishBody(Key &Body)
{
	return tankSimulator->RemoveFishBody(Body);
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	AddFishBody
/*! \brief      Adds a FishBody to the simulation and assigns the specified
**              Key to it
**  \param      Body:	The Key value that will be used as a label for the new FishBody
**  \return     true if a FishBody was added to the simulation, and false if the FishBody limit was
				reached or a FishBody already exists that uses the specified Key as a label
**  \exception  none
**	\author		Nathan Frost
**	\date		11/26/2002 11:50:33 AM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
bool MessageHandler::AddFishBody(const Key &Body, Key parentA, Key parentB)
{
//	return tankSimulator->AddFishBody(Body);

	if(tankSimulator->AddFishBody(Body)) 
	{
		BreedFish(Body, parentA, parentB);
		return true;
	}
	
	return false;
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	LocalFishGotoPoint
/*! \brief      
**  \param      randPoint:		The point for the local fish to move to
	\param		act:			Action produced by the brain.
	\param		FishKey:		The identifier for the local fish that is to move
	\param		tankSimulator:	Pointer to the tankSimulator
**  \return     none
**  \exception  none
**	\author		Nathan Frost
**	\date		11/26/2002 11:50:33 AM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
void LocalFishGotoPoint(const math::linear_alg::Vector3<float>& randPoint, const AI_System::AI_Action& act,
		 const Key& FishKey, TankSimulator* tankSimulator) 
{
	assert(tankSimulator);
	
	//translate point to Interp version
	InterpretationModule::InternalPoint_type interpRandPoint;
	interpRandPoint.x = randPoint[0];
	interpRandPoint.y = randPoint[1];
	interpRandPoint.z = randPoint[2];
	
	
	//create the appropriate command packet for the local FishBody to interpret
	InterpretationModule::CommandPacket cmdPack(InterpretationModule::AQUARIUS_COMMAND_GOTO,
		act.tag.unit_ID, FishKey, 10.0f,
		act.tag.decisionID, act.tag.fitness,
		interpRandPoint);
	
	
	//give that command packet to the local fish.
	InterpretationModule::FishBody* fishBody = tankSimulator->GetFishBodyPtr(FishKey);
	
	if(fishBody)
		fishBody->Command(cmdPack);
}


////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	SetFishBodyActions
/*! \brief      Sets the specified fish body's actions to the provided action vector
**  \param      FishKey:	A Key to a FishBody
**  \param      Actions:	A vector of actions that the FishBody must process
**  \return     returns false if the specified FishBody does not exist
**  \exception  none
	\par		DESCRIPTION: 
					- Only the Server should call this function.
					- Replaces all ROAM FishCommands with a GotoPoint command, such that
					the GoToPoint command takes a random point that we, the Server generate.
					This causes all Clients to Goto the same point as the Server, instead of
					erroneously choosing their own, different, points to ROAM to.
**	\author		Nathan Frost
**	\date		11/26/2002 11:52:36 AM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
bool MessageHandler::SetFishBodyActions(const Key &FishKey, 
										std::vector<AI_System::AI_Action> &Actions)
{
	//Make sure that this particular Fish on all remote machines performs the propr actions
	UDPNet::NetMessage netMessage;
	netMessage.SetID(UDPNet::SET_FISH_BODY_ACTIONS);

	netMessage.WriteNBytes(sizeof(FishKey), &FishKey);
	int numActions = Actions.size();
	netMessage.WriteNBytes(sizeof(numActions), &numActions);
	for(std::vector<AI_System::AI_Action>::iterator it = Actions.begin(); it != Actions.end(); ++it)
	{
		//hack hack hack. this is a hack, hack hack
//		it->ID = AI_System::ROAM;
		
		/* If this is a roam command, replace it with a GotoPoint(<randomPoint>) command .*/
		if(AI_System::ROAM == it->ID) 
		{
			//choose a random point, also convert it to the interpModule's version of a point.
			math::linear_alg::Vector3<float> randPnt;
			
			//use the fish body to get a random point.
			InterpretationModule::FishBody* fishBody = GetFishBodyPtr(FishKey);
			if(fishBody)
				fishBody->GetRandPoint(randPnt);
	
			//comment me, I like it that way.
			LocalFishGotoPoint(randPnt, *it, FishKey, tankSimulator);

			//lastly, send this off to the remote copies of the local fish so they interpret it too
			netMessage.WriteNBytes(sizeof(it->ID), &it->ID);
			netMessage.WriteNBytes(sizeof(it->objectID), &it->objectID);
			netMessage.WriteNBytes(sizeof(it->tag.unit_ID), &it->tag.unit_ID);	
			netMessage.WriteNBytes(sizeof(randPnt), &randPnt[0]);


			//mark this action as NULL so that the FishInterpretation module ignores it
			it->ID = AI_System::AC_NULL;
		}
		
		//send all non-ROAM fishActions to their remote copies as usual
		else
		{
			netMessage.WriteNBytes(sizeof(it->ID), &it->ID);
			netMessage.WriteNBytes(sizeof(it->objectID), &it->objectID);
			netMessage.WriteNBytes(sizeof(it->tag.unit_ID), &it->tag.unit_ID);	
		}
	}
	
	gameNetworking->SendNetMessage(netMessage);

	return tankSimulator->SetFishBodyActions(FishKey, Actions);
}




////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	SetFishBodyActions (remote)
/*!	\brief		Relays remote brain commands to local fish bodies
**	\param		TheNetMessage:	Has ID SET_FISH_BODY_ACTIONS and therefore contains the remote 
								brain commands
**	\return		none
**	\exception	none
**	\par		DESCRIPTION:
**					- Only the Client should call this function
**	\author		Nathan Frost
**	\date		3/12/2003 9:51:34 AM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
bool MessageHandler::SetFishBodyActions(UDPNet::NetMessage &TheNetMessage)
{
	Key fishKey;
	std::vector<AI_System::AI_Action> actions;

	//get the recipient this message is destined for
	TheNetMessage.ReadNBytes(sizeof(fishKey), &fishKey);

	/* Get every action the recipient is to perform */
	int numActions = 0;
	TheNetMessage.ReadNBytes(sizeof(numActions), &numActions);
	for(int i=0; i != numActions; ++i)
	{
		AI_System::AI_Action ai_Action;
		TheNetMessage.ReadNBytes(sizeof(ai_Action.ID), &ai_Action.ID);
		TheNetMessage.ReadNBytes(sizeof(ai_Action.objectID), &ai_Action.objectID);
		TheNetMessage.ReadNBytes(sizeof(ai_Action.tag.unit_ID), &ai_Action.tag.unit_ID);

		/* If the action is ROAM, then replace this action with a GotoPoint command */
		if(AI_System::ROAM == ai_Action.ID)
		{
			//read in the random point sent over the network
			math::linear_alg::Vector3<float> randPoint;
			TheNetMessage.ReadNBytes(sizeof(randPoint), &randPoint[0]);

			//Give this command to the local fish
			LocalFishGotoPoint(randPoint, ai_Action, fishKey, tankSimulator);

			//mark this action as NULL so that the FishInterpretation module ignores it
			ai_Action.ID = AI_System::AC_NULL;
		}

		actions.push_back(ai_Action);
	}

//	Debugging::Debugger::Instance()->output(Debugging::NET, Debugging::L4, "Got FishBodyActions");
	
	return tankSimulator->SetFishBodyActions(fishKey, actions);
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	GetFishBodyPtr
/*! \brief      Retrieves a pointer to the specified FishBody
	\param		Body:	FishBody to get a pointer to
**  \return     A pointer to the FishBody; will be NULL if the requested FishBody does not exist
**	\author		Nathan Frost
**	\date		11/26/2002 11:55:01 AM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
InterpretationModule::FishBody* MessageHandler::GetFishBodyPtr(const Key &Body)
{
	return tankSimulator->GetFishBodyPtr(Body);
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	
/*!	\brief		
**	\param		COMMENT ME PLEASE!!!!	
**	\return		none
**	\exception	none
**	\par		DESCRIPTION:
**					
**	\author		
**	\date		11/27/2002 10:59:12 AM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
void MessageHandler::GetSensedObject(Key &Object, std::vector<Key>& retVal)
{
	tankSimulator->GetVectorOfFishVisibleToFish(Object, retVal);
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	GetPerformedActions
/*!	\brief		Queries an Object for all recently performed actions
**	\param		Object:	Is the Object to query for recently performed actions
**	\return		All the recently performed actions of Object
**	\exception	none
**	\par		DESCRIPTION:
**
**	\author		Nathan Frost
**	\date		10/25/2002 1:55:35 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
void MessageHandler::GetPerformedActions(Key &Object, std::vector<AI_System::AI_ActionTag>& ret)
{
	GetFishBodyPtr(Object)->Dump(ret);
}

////////////////////////////////////////////////////////////////////////////////
//  FUNCTION:       GetVectorOfFishVisibleToFish
/*! \brief          retrieves a vector of Keys to fish that are currently visible
**                  to the specified fish
**  \param          fishKey a Key to the FishBody the will be used as a point of
**                  view when figuring out when which fish are visible
**  \return         a vector of Keys for FishBodies that are visible to the
**                  specified fish
**  \exception      none
**  \author         Nathan Frost
**  \date           12/2/2002 3:20:45 PM
**  \warning        none
*/
////////////////////////////////////////////////////////////////////////////////
void MessageHandler::GetVectorOfFishVisibleToFish(Key FishKey, std::vector<MsgHandler::Key>& vec)
{
	tankSimulator->GetVectorOfFishVisibleToFish(FishKey, vec);
}


////////////////////////////////////////////////////////////////////////////////
//  FUNCTION:       GetNextFish
/*! \brief          retrieves the key to the next fish in the simulation 
**                  (arbitrary order, wraps)
**  \param          fishKey a Key to a FishBody
**  \return         the first fish's Key if the specfied fish does not exist, 
**                  and 0 if no fish are in the simulation
**  \exception      none
**  \author         Nathan Frost
**  \date           12/4/2002 10:53:42 AM
**  \warning        none
*/
////////////////////////////////////////////////////////////////////////////////
Key MessageHandler::GetNextFish(Key FishKey)
{
	return tankSimulator->GetNextFish(FishKey);
}

////////////////////////////////////////////////////////////////////////////////
//  FUNCTION:       ImportHeightMapFromImage
/*! \brief          loads the specified image file and maps its red channel to 
**                  the height values of the height map
**  \param          fileName a pointer to a NULL terminated string that specifies 
**                  the file that should be read from
**  \param          maxHeight the value that a red value of 255 should be mapped
**                  to once the image is loaded
**  \return         returns true if the image was properly imported, false otherwise
**  \exception      none
**  \author         Nathan Frost
**  \date           12/6/2002 5:24:10 PM
**  \warning        Also changes the resolution of the height map to that of the image.
*/
////////////////////////////////////////////////////////////////////////////////
bool MessageHandler::ImportHeightMapFromImage(char *FileName, float MaxHeight)
{
	return tankSimulator->ImportHeightMapFromImage(FileName, MaxHeight);
}

////////////////////////////////////////////////////////////////////////////////
//  FUNCTION:       ExportHeightMapToImage
/*! \brief          saves a gray scale image of the height map, with white being
**                  the highest elevation
**  \param          fileName a pointer to a NULL terminated string that specifies 
**                  the file that should be written to
**  \return         returns true if the image was properly exported, false otherwise
**  \exception      none
**  \author         Nathan Frost
**  \date           12/6/2002 5:29:23 PM
**  \warning        The resolution of the image will be that of the height map
*/
////////////////////////////////////////////////////////////////////////////////
bool MessageHandler::ExportHeightMapToImage(char *FileName)
{
	return tankSimulator->ExportHeightMapToImage(FileName);
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	AddRandomFish
/*!	\brief		Adds a RandomFish to the Aquarium in a random position.
**	\param		none
**	\return		none
**	\exception	none
**	\par		DESCRIPTION:
**					
**	\author		Nathan Frost
**	\date		1/18/2003 11:58:24 AM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
void MessageHandler::AddRandomFish()
{
	if(NetworkingIsServer())
		AI_System::AI_BrainManager::getInstance()->CreateRandomFishBrain();
	else//Client must ask the Server permission to add a RandomFish
	{
		UDPNet::NetMessage netMessage;
		netMessage.SetID(UDPNet::ADD_RANDOM_FISH);
		gameNetworking->SendNetMessage(netMessage);
	}
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	TraceRayAgainstAllFish
/*! \brief          Traces a ray against every fish in the simulation.
**  \param          rayOrigin: the world space origin of the ray
**  \param          rayDirection: the direction vector of the ray
**  \param          fishKeyRef: a reference to a key that will be filled out with
**                  the key of the closest fish that intersects the ray
**  \return         true if a fish was hit and therefore the value referenced to
**                  by fishKeyRef changed, false otherwise.
**  \exception      none
**  \author         Nathan Frost
**  \date           2/6/2003 8:16:47 PM
**  \warning        The intersection algorithm is slightly inaccurate, mainly
**                  due to a preference for speed over accuracy 
*/
////////////////////////////////////////////////////////////////////////////////
bool MessageHandler::TraceRayAgainstAllFish(const math::linear_alg::Vector3<float> & rayOrigin, 
											const math::linear_alg::Vector3<float> & rayDirection,
											Key & fishKeyRef)
{
	return tankSimulator->TraceRayAgainstAllFish(rayOrigin, rayDirection, fishKeyRef);
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	SwimModeGotoPoint
/*!	\brief		Causes the Fish of fishID to swim to point.
**	\param		fishID:	Fish doing the swimming
	\param		point:	the fish is swimming to
**	\return		none
**	\exception	none
**	\par		DESCRIPTION:
**					
**	\author		Nathan Frost
**	\date		2/19/2003 10:16:19 AM
**	\warning	none
**	\par		UPDATED ON:			4/10/2003 2:24:24 PM
**	\par		UPDATED BY:			Nate Cleveland
**	\par		UPDATE DESCRIPTION: 
**				Somebody forgot the "!" in their conditional. Easy mistake to make.
*/
////////////////////////////////////////////////////////////////////////////////
void MessageHandler::SwimModeGotoPoint(Key fishID, math::linear_alg::Vector3<float> point)
{
	InterpretationModule::FishBody *fishBody = tankSimulator->GetFishBodyPtr(fishID);

	//Fail silently if the fish does not exist for networking timing issues
	if(!fishBody)
		return;
	
	
	fishBody->GotoPoint(point);

	UDPNet::NetMessage netMessage;
	netMessage.SetID(UDPNet::SWIM_MODE_GOTO_POINT);
	netMessage.WriteNBytes(sizeof(fishID), &fishID);
	netMessage.WriteNBytes(sizeof(point), &point[0]);
	
	gameNetworking->SendNetMessage(netMessage);
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	SwimModeGotoPoint (remote)
/*!	\brief		Causes the fishID to swim to a point as a result of a remote message
**	\param		TheNetMessage: that contains the information to remotely instruct the Fish to move
**	\return		none
**	\exception	none
**	\par		DESCRIPTION:
**					
**	\author		Nathan Frost
**	\date		3/12/2003 12:19:32 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
void MessageHandler::SwimModeGotoPoint(UDPNet::NetMessage &TheNetMessage)
{
	Key fishID;
	math::linear_alg::Vector3<float> point;

	TheNetMessage.ReadNBytes(sizeof(fishID), &fishID);
	TheNetMessage.ReadNBytes(sizeof(point), &point[0]);

	InterpretationModule::FishBody *fishBody = tankSimulator->GetFishBodyPtr(fishID);
	
	//fail silently if the requested fish doesn't exist -- due to network timing
	if(!fishBody)
		return;

	fishBody->GotoPoint(point);
}



////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	setTankDepth
/*!	\brief		changes the depth of the tank
**	\param		depth, the new depth of the tank
**	\return		none
**	\exception	none
**	\author		John Corpening
**	\date		2/27/2003 8:58:08 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
void MessageHandler::setTankDepth(float depth)
{
	if(NetworkingIsServer())
	{
		math::linear_alg::Vector3<float> min,max;
		tankSimulator->GetBoundary(min,max);
		max[2] = depth / 2;
		min[2] = -1 * depth / 2;
		tankSimulator->SetBoundary(min,max);
	}
	else
		;//notify the user that clients can't do this
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	setTankHeight
/*!	\brief		changes the height of the tank
**	\param		height, the new height of the tank
**	\return		none
**	\exception	none
**	\author		John Corpening
**	\date		2/27/2003 8:59:05 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
void MessageHandler::setTankHeight(float height)
{
	if(NetworkingIsServer())
	{
		math::linear_alg::Vector3<float> min,max;
		tankSimulator->GetBoundary(min,max);
		max[1] = height / 2;
		min[1] = -1 * height / 2;
		tankSimulator->SetBoundary(min,max);
	}
	else
		;//notify the user that clients can't do this
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	setTankWidth
/*!	\brief		changes the width of the tank
**	\param		width, the new width of the tank
**	\return		none
**	\exception	none
**	\author		John Corpening
**	\date		2/27/2003 8:59:43 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
void MessageHandler::setTankWidth(float width)
{
	if(NetworkingIsServer())
	{
		math::linear_alg::Vector3<float> min,max;
		tankSimulator->GetBoundary(min,max);
		max[0] = width / 2;
		min[0] = -1 * width / 2;
		tankSimulator->SetBoundary(min,max);
	}
	else
		;//notify the user that clients can't do this
}


// Added by Nate [3/13/2003]
////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	Rangify
/*!	\brief		Converts the given 0 to 1 value to 0 to 100.
**	\param		value		value to ragify
**	\return		0 to 100 value.
**	\author		Nate Cleveland
**	\date		3/13/2003
*/
////////////////////////////////////////////////////////////////////////////////
template<typename return_type>
return_type Rangify(float value)
{
	//left shift by two values (not bit shift, but decimal shift).
	value *= 1000;

	//divide away the annoying trainling bits by casting to u32.
	unsigned int rangedValue = static_cast<unsigned int>(value);

	//cast to the template'd return value.
	return static_cast<return_type>(rangedValue) / 10;
	
}



// Added by Nate [3/10/2003]
////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	FishBody_Meta
/*!	\brief		Returns to the caller the value of the given fish's metabolism
**	\param		key, the key of the fish to read from
**	\return		float value of metabolism, 0 to 1
**	\author		Nate Cleveland
**	\date		3/10/2003 
**	\warning	//asserts on legal fish key value.
*/
////////////////////////////////////////////////////////////////////////////////
float MessageHandler::FishBody_Meta(const Key &key)const throw()
{
	const InterpretationModule::FishBody* fishBody = GetFishBody(key);
	//assert(fishBody);
	if(fishBody)
		return Rangify<float>(fishBody->m_pAttributes->GetMetabolism());
	return 0;
}


////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	FishBody_Armor
/*!	\brief		Returns to the caller the value of the given fish's armor
**	\param		key, the key of the fish to read from
**	\return		float value of armor, 0 to 1
**	\author		Nate Cleveland
**	\date		3/10/2003 
**	\warning	//asserts on legal fish key value.
*/
////////////////////////////////////////////////////////////////////////////////
float MessageHandler::FishBody_Armor(const Key &key)const throw()
{
	const InterpretationModule::FishBody* fishBody = GetFishBody(key);
	//assert(fishBody);
	if(fishBody)
		return Rangify<float>(fishBody->m_pAttributes->GetArmor());
	return 0;
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	FishBody_Teeth_Baleen
/*!	\brief		Returns to the caller the value of the given fish's Teeth/Baleen
**	\param		key, the key of the fish to read from
**	\return		float value of Teeth/Baleen, 0 to 1
**	\author		Nate Cleveland
**	\date		3/10/2003 
**	\warning	//asserts on legal fish key value.
*/
////////////////////////////////////////////////////////////////////////////////
float MessageHandler::FishBody_Teeth_Baleen(const Key &key)const throw()
{
	const InterpretationModule::FishBody* fishBody = GetFishBody(key);
	//assert(fishBody);
	if(fishBody)
		return Rangify<float>(fishBody->m_pAttributes->GetTeeth_Baleen());
	return 0;
}


////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	FishBody_Size
/*!	\brief		Returns to the caller the value of the given fish's size
**	\param		key, the key of the fish to read from
**	\return		float value of size, 0 to 1
**	\author		Nate Cleveland
**	\date		3/10/2003 
**	\warning	//asserts on legal fish key value.
*/
////////////////////////////////////////////////////////////////////////////////
float MessageHandler::FishBody_Size(const Key &key)const throw()
{
	const InterpretationModule::FishBody* fishBody = GetFishBody(key);
	//assert(fishBody);
	if(fishBody)
		return Rangify<float>(fishBody->m_pAttributes->GetSize());
	return 0;
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	FishBody_MatingIncline
/*!	\brief		Returns to the caller the value of the given fish's MatingIncline
**	\param		key, the key of the fish to read from
**	\return		float value of MatingIncline, 0 to 1
**	\author		Nate Cleveland
**	\date		3/10/2003 
**	\warning	//asserts on legal fish key value.
*/
////////////////////////////////////////////////////////////////////////////////
float MessageHandler::FishBody_MatingIncline(const Key &key)const throw()
{
	const InterpretationModule::FishBody* fishBody = GetFishBody(key);
	//assert(fishBody);
	if(fishBody)
		return Rangify<float>(fishBody->m_pAttributes->GetIncline());
	return 0;
}


////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	FishBody_Health
/*!	\brief		Returns to the caller the value of the given fish's Health
**	\param		key, the key of the fish to read from
**	\return		float value of Health, 0 to 1
**	\author		Nate Cleveland
**	\date		3/10/2003 
**	\warning	//asserts on legal fish key value.
*/
////////////////////////////////////////////////////////////////////////////////
float MessageHandler::FishBody_Health(const Key &key)const throw()
{
	const InterpretationModule::FishBody* fishBody = GetFishBody(key);
	//assert(fishBody);
	if(fishBody)
		return Rangify<float>(fishBody->m_pAttributes->GetHealth_Normalized());
	return 0;
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	FishBody_Stomach
/*!	\brief		Returns to the caller the value of the given fish's Stomach
**	\param		key, the key of the fish to read from
**	\return		float value of Stomach, 0 to 1
**	\author		Nate Cleveland
**	\date		3/10/2003 
**	\warning	//asserts on legal fish key value.
*/
////////////////////////////////////////////////////////////////////////////////
float MessageHandler::FishBody_Stomach(const Key &key)const throw()
{
	const InterpretationModule::FishBody* fishBody = GetFishBody(key);
	//assert(fishBody);
	if(fishBody)
		return __max(0,100 - Rangify<float>(fishBody->m_pAttributes->GetStomach_Normalized()));
	return 0;
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	FishBody_EnergyLevel
/*!	\brief		Returns to the caller the value of the given fish's energyLevel
**	\param		key, the key of the fish to read from
**	\return		float value of energyLevel, 0 to 1
**	\author		Nate Cleveland
**	\date		3/10/2003 
**	\warning	//asserts on legal fish key value.
*/
////////////////////////////////////////////////////////////////////////////////
float MessageHandler::FishBody_EnergyLevel(const Key &key)const throw()
{
	const InterpretationModule::FishBody* fishBody = GetFishBody(key);
	//assert(fishBody);
	if(fishBody)
		return Rangify<float>(fishBody->m_pAttributes->GetEnergyLevel_Normalized());
	return 0;
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	ConnectToServer
/*!	\brief		Turns your game into Client mode
**	\param		IPAddress:	String that contains the IP Address of the Server we're to try to get 
							in touch with.
**	\return		true on success (if IPAddress was a valid IP Address), false otherwise.
**	\exception	none
**	\par		DESCRIPTION:
**					- If this function fails, the game remains in Server Mode and does not attempt
					to switch to Client Mode.

**	\author		Nathan Frost
**	\date		3/11/2003 3:23:03 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
bool MessageHandler::ConnectToServer(const char *IPAddress)
{

	std::istringstream ipStringStream(IPAddress);
    int ipNum[4];
    char dummy;

	for(int i=0; i < 4; ++i)
	{
		ipStringStream >> ipNum[i];
		ipStringStream >> dummy;//get rid of that .
		
		//don't instantiate a Client with an invalid IP Address
		if(ipNum[i] > 255 || ipNum[i] < 0)
			return false;
	}

	/* Time to become a Client */
    u_long ip = 0;
    ip += ipNum[0] << 24;
    ip += ipNum[1] << 16;
    ip += ipNum[2] << 8;
    ip += ipNum[3];

	//we cannot be our own Client
	if(ip == UDPNet::UDPNetworking::Instance()->MyIPAddress())
		return false;

	//switch off all local brains -- all the thinking's going to be done on the Server
	AI_System::AI_BrainManager::getInstance()->makeClient();

	delete gameNetworking;
	gameNetworking = new UDPNet::Client(ip);


	return true;
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	WeBootedClient
/*!	\brief		We, the Server, booted a Client -- so clean up the Client's game logic!
**	\param		IPAddress: of the Client we Booted
**	\return		none
**	\exception	none
**	\par		DESCRIPTION:
**					
**	\author		Nathan Frost
**	\date		3/11/2003 3:50:45 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
void MessageHandler::WeBootedClient(u_long IPAddress)
{
	Key fishSelectedByBootedClient = UserInput::UI_Handler::instance()->getFishFromIP(IPAddress);

	//if the client had a fish selected...
	if(fishSelectedByBootedClient)
	{
		//deselect the fish
		if(UserInput::UI_Handler::instance()->confirmDeselect(	fishSelectedByBootedClient, 
																IPAddress))
		{
			//notify clients that the list of selected fish has changed
			NotifyRemoteMachinesOfFishSelectListChanging();
		}
		//and unsuppress the fish if it was suppressed
		if(AI_System::AI_BrainManager::getInstance()->confirmUnsuppress(fishSelectedByBootedClient))
		{
			//Fish is unsuppressed -- notify clients that the list of suppressed fish has changed
			NotifyRemoteMachinesOfFishSuppressionListChanging();
		}
	}
	//std::cout << "We do hereby boot 0x" << std::hex << IPAddress << std::endl;
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	BootAllClients
/*!	\brief		States that we, the Server, are going to Boot every Client that's connected to us
**	\param		none
**	\return		none
**	\exception	none
**	\par		DESCRIPTION:
**					
**	\author		Nathan Frost
**	\date		3/17/2003 10:52:00 AM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
void MessageHandler::BootAllClients()
{
	//if I'm the Server, I want to deselect all the Client's selections since they're getting booted
	if(NetworkingIsServer())
		UserInput::UI_Handler::instance()->deselectAllButThisIP(GetMyIP());

	gameNetworking->BootAllRemoteMachines();
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	WeAcceptedClient
/*!	\brief		Informs the MessageHandler that we, the Server, have accepted a Client
**	\param		IPAddress: of the accepted Client
**	\return		none
**	\exception	none
**	\par		DESCRIPTION:
**					
**	\author		Nathan Frost
**	\date		3/11/2003 3:55:59 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
void MessageHandler::WeAcceptedClient(u_long IPAddress)
{
	//std::cout << "We, the Server, do hereby accept 0x" << std::hex << IPAddress << std::endl;
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	FillOutHeightMapNetMessage
/*!	\brief		Fills out TheNetMessage with Heightmap data.
**	\param		TheNetMessage:	to be filled out with the HeightMap data.
**	\return		none
**	\exception	none
**	\par		DESCRIPTION:
**					
**	\author		Nathan Frost
**	\date		4/10/2003 9:42:52 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
void MessageHandler::FillOutHeightMapNetMessage(UDPNet::NetMessage &TheNetMessage)
{
	tankSimulator->GenerateHeightMapNetMessage(TheNetMessage);
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	OurServerIsDelinquent
/*!	\brief		Informs the MessageHandler that we, the Client, have found our Server to be
				Delinquent and are therefore no longer connected
**	\param		none
**	\return		none
**	\exception	none
**	\par		DESCRIPTION:
**					
**	\author		Nathan Frost
**	\date		3/11/2003 3:59:29 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
void MessageHandler::OurServerIsDelinquent()
{
	NukeAllFish();
	Debugging::Debugger::Instance()->output(Debugging::NET, Debugging::L4, 
											"We, the Client, have found our Server to be Delinquent."
											);
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	WeAreConnected
/*!	\brief		Informs the MessageHandler that we, the Client, have connected to the Server
**	\param		HeightMapData: NetMessage that will be filled out with the Server's heightmap data
**	\return		none
**	\exception	none
**	\par		DESCRIPTION:
**					
**	\author		Nathan Frost
**	\date		3/11/2003 4:08:19 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
void MessageHandler::WeAreConnected(UDPNet::NetMessage &HeightMapData)
{
	tankSimulator->LoadHeightMapNetMessage(HeightMapData);
	Debugging::Debugger::Instance()->output(Debugging::NET, Debugging::L4, 
											"We, the Client, are connected.");
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	WeAreBooted
/*!	\brief		Informs the MessageHandler that we, the Client, have been booted by the Server
**	\param		none
**	\return		none
**	\exception	none
**	\par		DESCRIPTION:
**					
**	\author		Nathan Frost
**	\date		3/11/2003 4:34:54 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
void MessageHandler::WeAreBooted()
{
	/*	annihilate the local Aquarium -- we are no longer connected to Mother Server from which all
		life springs */
	NukeAllFish();
	Debugging::Debugger::Instance()->output(Debugging::NET, Debugging::L4, 
											"We, the Client, are booted.");
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	NetworkingIsServer
/*!	\brief		Lets the caller know if the local machine is acting as a Server or a Client
**	\param		none
**	\return		true if the local machine is a Server, false otherwise
**	\exception	none
**	\par		DESCRIPTION:
**					
**	\author		Nathan Frost
**	\date		3/11/2003 8:19:03 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
bool MessageHandler::NetworkingIsServer()
{
	if(dynamic_cast<UDPNet::Server *>(gameNetworking))
		return true;
	else
		return false;
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	NetworkingState
/*!	\brief		Returns a string that will contain "Server" if the we are currently a Server and
				"Client" if we are currently a Client.
**	\param		none
**	\return		none
**	\exception	none
**	\par		DESCRIPTION:
**					
**	\author		Nathan Frost
**	\date		3/11/2003 4:47:45 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
std::string MessageHandler::NetworkingState()
{
	std::string ret;

	if(NetworkingIsServer())
	{
		//TODO: there's gotta be a better of way of doing this
		char dummy;
		ret += "Server -- ";
		ret += itoa(gameNetworking->GetNumConnectedClients(), &dummy, 10);
		ret += " Clients connected";
	}
	else
	{
		ret += "Client -- ";
		if(gameNetworking->Connected())
			ret += "Connected";
		else
			ret += "Disconnected";
	}

	return ret;
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	FillOutServerUpdate
/*!	\brief		Fills out a server update message with all the necessary data to give Clients a 
				snapshot of the current game situation
**	\param		ServerUpdate: The NetMessage to be filled out
**	\return		none
**	\exception	none
**	\par		DESCRIPTION:
**					
**	\author		Nathan Frost
**	\date		3/11/2003 6:21:11 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
void MessageHandler::FillOutServerUpdate(UDPNet::NetMessage &ServerUpdate)
{
	tankSimulator->GeneratePartialNetMessage(ServerUpdate);
	UserInput::UI_Handler::instance()->GenerateCompleteNetMessage(ServerUpdate);
	AI_System::AI_BrainManager::getInstance()->GenerateCompleteNetMessage(ServerUpdate);
	
	//send over the current gameSpeed
	double gameSpeed = TimeObject::Instance()->getGameSpeed();
	ServerUpdate.WriteNBytes(sizeof(gameSpeed), &gameSpeed);

#if 0
	if(ServerUpdate.GetPacketSize(ServerUpdate.GetPacket(0)))
	{
		tankSimulator->GeneratePartialNetMessage(ServerUpdate);
		UserInput::UI_Handler::instance()->GenerateCompleteNetMessage(ServerUpdate);
	}
#endif
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	GotServerUpdate
/*!	\brief		Informs the MessageHandler that the Server has sent us, the Client, a ServerUpdate
**	\param		ServerUpdate: The NetMessage we shall glean game snapshot data from
**	\return		none
**	\exception	none
**	\par		DESCRIPTION:
**					
**	\author		Nathan Frost
**	\date		3/11/2003 7:01:01 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
void MessageHandler::GotServerUpdate(UDPNet::NetMessage &ServerUpdate)
{
	Debugging::Debugger::Instance()->output(Debugging::NET, Debugging::L4, 
											"Loaded up a Server Message");
	tankSimulator->LoadNetMessage(ServerUpdate);
	UserInput::UI_Handler::instance()->LoadNetMessage(ServerUpdate);
	AI_System::AI_BrainManager::getInstance()->LoadNetMessage(ServerUpdate);
	NewGameSpeed(ServerUpdate);
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	HighLightFish
/*!	\brief		Highlights or tints a fish the given color.
**	\param			isOn	Set true if you want to highlight the fish, false
							 if you want it to NOT highlight.
**	\param			red		Red componet of the color.
**	\param			green	Green componet of the color
**	\param			blue	Blue componet of the color
**	\return		none
**	\exception	none					
**	\author		Nate Cleveland
**	\date		3/13/2003 12:59:36 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
void MessageHandler::HighLightFish(const Key& fishKey, bool isOn, float red/* = 1*/, 
								   float green/* = 1*/, float blue/* = 1*/) throw()
{
	InterpretationModule::FishBody* fishBody = NULL;

	//get the fish pntr, if it isn't null call the highlight function
	if(fishBody = GetFishBodyPtr(fishKey))
		fishBody->Highlight(isOn, red, green, blue);	
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	SelectRequest
/*!	\brief		States that a player would like to deselect a fish
**	\param		Fish	A message handler key of the fish to be deselected
**	\return		true if the Select request succeeded synchronously (ie right away), false if not.
				Note that the Select request may still succeed asynchronously (ie in a few seconds)
**	\exception	none
**	\par		DESCRIPTION:
**					- if the request succeeds, all RemoteMachines are notified that the request 
					succeeded
**	\author		Nathan Frost
**	\date		3/15/2003 12:30:57 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
bool MessageHandler::SelectRequest(const Key& Fish)
{
	if(NetworkingIsServer())
	{
		if(UserInput::UI_Handler::instance()->confirmSelect(Fish, GetMyIP()))
		{
			UserInput::UI_Handler::instance()->setFishTarget(Fish);
			//notify clients that the list of selected fish has changed
			NotifyRemoteMachinesOfFishSelectListChanging();

			return true;
		}
	}
	else//Client
	{
		//ask Server for permission to select
		UDPNet::NetMessage fishSelectRequest;
		fishSelectRequest.SetID(UDPNet::FISH_SELECT_REQUEST);
		fishSelectRequest.WriteNBytes(sizeof(Fish), &Fish);
		gameNetworking->SendNetMessage(fishSelectRequest);
	}

	return false;//success may happen asynchronously, but won't happen right away
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	SuppressRequest
/*!	\brief		States that a player would like to suppress a fish
**	\param		Fish	A message handler key of the fish to be suppressed
**	\return		true if the suppress request succeeded synchronously (ie right away), false if not.
				Note that the suppress request may still succeed asynchronously (ie in a few 
				seconds)
**	\exception	none
**	\par		DESCRIPTION:
**					- if the request succeeds, all RemoteMachines are notified that the request 
					succeeded
**	\author		Nathan Frost
**	\date		3/15/2003 12:30:57 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
bool MessageHandler::SuppressRequest(const Key& Fish) throw()
{
	if(NetworkingIsServer())
	{
		if(AI_System::AI_BrainManager::getInstance()->confirmSuppress(Fish))
		{
			//wipe the Fish's thoughts clean and lobotomize its pathfinding so it stops in its trax
			ClearFishBrain(Fish);
			InterpretationModule::FishBody *fishToLosePathfinding = GetFishBodyPtr(Fish);
			if(fishToLosePathfinding)//if you try to suppress a nonexistent fish, fail silently
				fishToLosePathfinding->SetPathing(false);

			//notify clients that the list of suppressed fish has changed
			NotifyRemoteMachinesOfFishSuppressionListChanging();

			return true;
		}
	}
	else//Client
	{
		//ask Server for permission to suppress
		UDPNet::NetMessage fishSuppressRequest;
		fishSuppressRequest.SetID(UDPNet::FISH_SUPPRESS_REQUEST);
		fishSuppressRequest.WriteNBytes(sizeof(Fish), &Fish);
		gameNetworking->SendNetMessage(fishSuppressRequest);
	}

	return false;//success may happen asynchronously, but won't happen right away
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	UnsuppressRequest
/*!	\brief		States that a player would like to unsuppress a fish
**	\param		Fish	A message handler key of the fish to be unsuppressed
**	\return		none
**	\exception	none
**	\author		Nathan Frost
**	\date		3/15/2003 12:30:57 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
void MessageHandler::UnsuppressRequest(const Key& Fish)
{
	if(NetworkingIsServer())
	{
		if(AI_System::AI_BrainManager::getInstance()->confirmUnsuppress(Fish))
		{
			/* Fish is unsuppressed */
			//notify clients that the list of suppressed fish has changed
			NotifyRemoteMachinesOfFishSuppressionListChanging();
			//and reinstate the fish's pathfinding capabilities
			InterpretationModule::FishBody *fishToGetPathfinding = GetFishBodyPtr(Fish);
			if(fishToGetPathfinding)//if you try to unsuppress a nonexistent fish, fail silently
				fishToGetPathfinding->SetPathing(true);		
		}
	}
	else//Client
	{
		//ask Server for permission to unsuppress
		UDPNet::NetMessage fishUnsuppressRequest;
		fishUnsuppressRequest.SetID(UDPNet::FISH_UNSUPPRESS_REQUEST);
		fishUnsuppressRequest.WriteNBytes(sizeof(Fish), &Fish);
		gameNetworking->SendNetMessage(fishUnsuppressRequest);
	}
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	ClientSuppressRequest
/*!	\brief		Informs the Server that a Client would like to suppress a Fish
**	\param		ClientSuppressRequest	The NetMessage that contains the relevant data
**	\return		none
**	\exception	none
**	\par		DESCRIPTION:
**					- if the request succeeds, all RemoteMachines are notified that the request 
					succeeded

**	\author		Nathan Frost
**	\date		3/15/2003 3:32:24 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
void MessageHandler::ClientSuppressRequest(UDPNet::NetMessage &ClientSuppressRequest)
{
	Key fish;
	ClientSuppressRequest.ReadNBytes(sizeof(fish), &fish);

	if(AI_System::AI_BrainManager::getInstance()->confirmSuppress(fish))
	{
		//wipe the Fish's thoughts clean and lobotomize its pathfinding so it stops in its tracks
		ClearFishBrain(fish);
		InterpretationModule::FishBody *fishToLosePathfinding = GetFishBodyPtr(fish);
		if(fishToLosePathfinding)//if you try to suppress a nonexistent fish, fail silently
			fishToLosePathfinding->SetPathing(false);

		//notify all Clients that the fish suppression list has changed
		NotifyRemoteMachinesOfFishSuppressionListChanging();
	}
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	FishSuppressListChanged
/*!	\brief		Informs the local machine of the new list of suppressed Fish.
**	\param		FishSuppressListChanged: the NetMessage that contains the relevant data
**	\return		none
**	\exception	none
**	\par		DESCRIPTION:
**					
**	\author		Nathan Frost
**	\date		3/15/2003 3:28:54 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
void MessageHandler::FishSuppressListChanged(UDPNet::NetMessage &FishSuppressListChanged)
{
	AI_System::AI_BrainManager::getInstance()->LoadNetMessage(FishSuppressListChanged);
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	ClientUnsuppressRequest
/*!	\brief		Notifies the Server that a Client would like to unsuppress a Fish
**	\param		ClientUnsuppressRequest	The NetMessage that contains relevant data
**	\return		none
**	\exception	none
**	\par		DESCRIPTION:
**					- if the request succeeds, the Client is notified that its request succeeded
**	\author		Nathan Frost
**	\date		3/15/2003 3:52:30 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
void MessageHandler::ClientUnsuppressRequest(UDPNet::NetMessage &ClientUnsuppressRequest)
{
	Key fish;
	ClientUnsuppressRequest.ReadNBytes(sizeof(fish), &fish);

	if(AI_System::AI_BrainManager::getInstance()->confirmUnsuppress(fish))
	{
		//notify Client that its request for fish unsuppression succeeded
		NotifyRemoteMachinesOfFishSuppressionListChanging();

		//and reinstate the fish's pathfinding capabilities
		InterpretationModule::FishBody *fishToGetPathfinding = GetFishBodyPtr(fish);
		assert(fishToGetPathfinding);
		fishToGetPathfinding->SetPathing(true);
	}
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	NotifyRemoteMachinesOfFishSuppressionListChanging
/*!	\brief		Tells all connected remote machines to update their fish suppression list
**	\param		none
**	\return		none
**	\exception	none
**	\par		DESCRIPTION:
**					
**	\author		Nathan Frost
**	\date		3/15/2003 3:41:00 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
void MessageHandler::NotifyRemoteMachinesOfFishSuppressionListChanging()
{
	UDPNet::NetMessage fishSuppressListChanged;
	fishSuppressListChanged.SetID(UDPNet::FISH_SUPPRESS_LIST_CHANGED);
	AI_System::AI_BrainManager::getInstance()->GenerateCompleteNetMessage(fishSuppressListChanged);
	gameNetworking->SendNetMessage(fishSuppressListChanged);
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	NotifyRemoteMachinesOfFishSelectListChanging
/*!	\brief		Tells all connected remote machines to update their fish selected list
**	\param		none
**	\return		none
**	\exception	none
**	\par		DESCRIPTION:
**					
**	\author		Nathan Frost
**	\date		3/15/2003 3:41:00 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
void MessageHandler::NotifyRemoteMachinesOfFishSelectListChanging()
{
	UDPNet::NetMessage fishSelectListChanged;
	fishSelectListChanged.SetID(UDPNet::FISH_SELECT_LIST_CHANGED);
	UserInput::UI_Handler::instance()->GenerateCompleteNetMessage(fishSelectListChanged);
	gameNetworking->SendNetMessage(fishSelectListChanged);
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	ClientSelectRequest
/*!	\brief		Informs the Server that a Client would like to select a Fish
**	\param		ClientSelectRequest	The NetMessage that contains the relevant data
	\param		SenderIP			Of the Client that is making the request.
**	\return		none
**	\exception	none
**	\par		DESCRIPTION:
**					- if the request succeeds, all RemoteMachines are notified that the request 
					succeeded

**	\author		Nathan Frost
**	\date		3/15/2003 3:32:24 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
void MessageHandler::ClientSelectRequest(UDPNet::NetMessage &ClientSelectRequest, u_long SenderIP)
{
	Key fish;
	ClientSelectRequest.ReadNBytes(sizeof(fish), &fish);

	if(UserInput::UI_Handler::instance()->confirmSelect(fish, SenderIP))
	{
		//notify all Clients that the fish selection list has changed
		NotifyRemoteMachinesOfFishSelectListChanging();
	}
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	FishSelectListChanged
/*!	\brief		Informs the local machine of the new list of selected Fish.
**	\param		fishSelectListChanged: the NetMessage that contains the relevant data
**	\return		none
**	\exception	none
**	\par		DESCRIPTION:
**					
**	\author		Nathan Frost
**	\date		3/15/2003 3:28:54 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
void MessageHandler::FishSelectListChanged(UDPNet::NetMessage &fishSelectListChanged)
{
	UserInput::UI_Handler::instance()->LoadNetMessage(fishSelectListChanged);
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	DeselectRequest
/*!	\brief		States that a player would like to deselect a fish
**	\param		Fish	A message handler key of the fish to be deselected
**	\return		none
**	\exception	none
**	\author		Nathan Frost
**	\date		3/15/2003 12:30:57 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
void MessageHandler::DeselectRequest(const Key& fish)
{
	if(NetworkingIsServer())
	{
		if(UserInput::UI_Handler::instance()->confirmDeselect(fish, GetMyIP()))
		{
			//deselect Fish
			UserInput::UI_Handler::instance()->setFishTarget(0);
			
			//notify clients that the list of selected fish has changed
			NotifyRemoteMachinesOfFishSelectListChanging();
		}
	}
	else//Client
	{
		//ask Server for permission to deselect
		UDPNet::NetMessage fishDeselectRequest;
		fishDeselectRequest.SetID(UDPNet::FISH_DESELECT_REQUEST);
		fishDeselectRequest.WriteNBytes(sizeof(fish), &fish);
		gameNetworking->SendNetMessage(fishDeselectRequest);
	}
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	ClientDeselectRequest
/*!	\brief		Notifies the Server that a Client would like to deselect a Fish
**	\param		ClientDeselectRequest	The NetMessage that contains relevant data
	\param		SenderIP				Of the Client that is making the request
**	\return		none
**	\exception	none
**	\par		DESCRIPTION:
**					- if the request succeeds, the Client is notified that its request succeeded
**	\author		Nathan Frost
**	\date		3/15/2003 3:52:30 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
void MessageHandler::ClientDeselectRequest(	UDPNet::NetMessage &ClientDeselectRequest, 
											u_long SenderIP)
{
	Key fish;
	ClientDeselectRequest.ReadNBytes(sizeof(fish), &fish);

	if(UserInput::UI_Handler::instance()->confirmDeselect(fish, SenderIP))
	{
		//notify Client that its request for fish selection succeeded
		NotifyRemoteMachinesOfFishSelectListChanging();
	}
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	GetMyIP
/*!	\brief		Returns the IP Address of the local machine
**	\param		none
**	\return		The IP Address of the local machine
**	\exception	none
**	\par		DESCRIPTION:
**					
**	\author		Nathan Frost
**	\date		3/15/2003 2:40:56 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
u_long MessageHandler::GetMyIP()
{
	return UDPNet::UDPNetworking::Instance()->MyIPAddress();
}


////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	GetActionString
/*!	\brief		Given a Key to a fish this method provides the caller with pointers 
**					to strings filled with data a bout the given fish's actions.
**	\param		fishKey		Key to the given fish.
**	\param		mouth		if not null, will be filled witha pntr to a string.
**	\param		move		if not null, will be filled witha pntr to a string.
**	\param		marker		if not null, will be filled witha pntr to a string.
**	\return		none
**	\exception	none
**	\par		DESCRIPTION:
**					Checks each pointer for null before filling it, thus if you only
					 need one string for one particular type of action, this method 
					 can handle it.
**	\author		Nate Cleveland
**	\date		3/17/2003 2:24:06 PM
**	\warning	none
*/
/////////////////////////////////////////////////////////////////////////////////
void MessageHandler::GetActionString(const Key& fishKey, std::string **mouth,
									 std::string **move, std::string **marker)
{
	//Get a pntr to the fish
	InterpretationModule::FishBody* fishBody = GetFishBodyPtr(fishKey);

	//sanity check
	if(fishBody) 
	{
		//fill in the mouth string iff they provided us with one.
		if(mouth)
			fishBody->m_strObj->GetMouth_Action(mouth);

		//fill in movement string iff they provided us with one.
		if(move) 
			fishBody->m_strObj->GetMove_Action(move);

		//fill in movement string iff they provided us with one.
		if(marker)
			fishBody->m_strObj->GetMarker_Action(marker);
	}
	

	else
	{
		if(mouth)
			*mouth = NULL;

		if(move) 
			*move = NULL;

		//fill in movement string iff they provided us with one.
		if(marker)
			*marker = NULL;
	}
}



////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	
/*!	\brief		Given a Key to a fish this method provides the caller with pointers 
**					to strings filled with data a bout the given fish's targets.
**	\param		fishKey		Key to the given fish.
**	\param		mouth		if not null, will be filled witha pntr to a string.
**	\param		move		if not null, will be filled witha pntr to a string.
**	\param		marker		if not null, will be filled witha pntr to a string.
**	\return		none
**	\exception	none
**	\par		DESCRIPTION:
**					Checks each pointer for null before filling it, thus if you only
					 need one string for one particular type of target, this method 
					 can handle it.
**	\author		Nate Cleveland
**	\date		3/17/2003 2:24:06 PM
**	\warning	none
*/
/////////////////////////////////////////////////////////////////////////////////
void MessageHandler::GetTargetString(const Key& fishKey, std::string **mouth,
									 std::string **move, std::string **marker)
{
	//Get a pntr to the fish
	InterpretationModule::FishBody* fishBody = GetFishBodyPtr(fishKey);

	//sanity check
	if(fishKey && fishBody) 
	{
		//fill in the mouth string iff they provided us with one.
		if(mouth)
			fishBody->m_strObj->GetMouth_Target(mouth);

		//fill in movement string iff they provided us with one.
		if(move) 
			fishBody->m_strObj->GetMove_Target(move);

		//fill in movement string iff they provided us with one.
		if(marker)
			fishBody->m_strObj->GetMarker_Target(marker);
	}	

	else
	{
		if(mouth)
			*mouth = NULL;

		if(move) 
			*move = NULL;

		//fill in movement string iff they provided us with one.
		if(marker)
			*marker = NULL;
	}
}


////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	FishBody_KillCount
/*!	\brief		Returns to the caller the number of kills this fish has.
**	\param		key		Key of the fish.
**	\return		guess.
**	\exception	none	
**	\author		Nate Cleveland
**	\date		3/18/2003 10:15:55 AM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
float MessageHandler::FishBody_KillCount(const Key &key)const throw()
{
	const InterpretationModule::FishBody* fishBody = GetFishBody(key);

	if(fishBody)
		return fishBody->m_pStats->GetKills();
	return 0;
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	FishBody_BeenAttacked
/*!	\brief		Returns to the caller the number of times this fish has been attacked.
**	\param		key		Key of the fish.
**	\return		guess.
**	\exception	none	
**	\author		Nate Cleveland
**	\date		3/18/2003 10:15:55 AM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
float MessageHandler::FishBody_BeenAttacked(const Key &key)const throw()
{
	const InterpretationModule::FishBody* fishBody = GetFishBody(key);

	if(fishBody)
		return fishBody->m_pStats->GetBeen_Attacked();
	return 0;
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	FishBody_HasAttacked
/*!	\brief		Returns to the caller the number of times this fish has attacked something.
**	\param		key		Key of the fish.
**	\return		guess.
**	\exception	none	
**	\author		Nate Cleveland
**	\date		3/18/2003 10:15:55 AM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
float MessageHandler::FishBody_HasAttacked(const Key &key)const throw()
{
	const InterpretationModule::FishBody* fishBody = GetFishBody(key);

	if(fishBody)
		return fishBody->m_pStats->GetHas_Attacked();
	return 0;
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	FishBody_Distance
/*!	\brief		Returns to the caller how far this fish has traveled (approx).
**	\param		key		Key of the fish.
**	\return		guess.
**	\exception	none	
**	\author		Nate Cleveland
**	\date		3/18/2003 10:15:55 AM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
float MessageHandler::FishBody_Distance(const Key &key)const throw()
{
	const InterpretationModule::FishBody* fishBody = GetFishBody(key);

	if(fishBody)
		return fishBody->m_pStats->GetDistance();
	return 0;
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	FishBody_AvgSpeed
/*!	\brief		Returns to the caller the fish's average speed.
**	\param		key		Key of the fish.
**	\return		guess.
**	\exception	none	
**	\author		Nate Cleveland
**	\date		3/18/2003 10:15:55 AM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
float MessageHandler::FishBody_AvgSpeed(const Key &key)const throw()
{
	const InterpretationModule::FishBody* fishBody = GetFishBody(key);

	if(fishBody)
		return fishBody->m_pStats->GetAvgSpeed();
	return 0;
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	FishBody_PlanktonEaten
/*!	\brief		Returns to the caller how much plankton the fish has eaten.
**	\param		key		Key of the fish.
**	\return		guess.
**	\exception	none	
**	\author		Nate Cleveland
**	\date		3/18/2003 10:15:55 AM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
float MessageHandler::FishBody_PlanktonEaten(const Key &key)const throw()
{
	const InterpretationModule::FishBody* fishBody = GetFishBody(key);

	if(fishBody)
		return fishBody->m_pStats->GetPlankton();
	return 0;
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	FishBody_ChumEaten
/*!	\brief		Returns to the caller how much chum the fish has eaten.
**	\param		key		Key of the fish.
**	\return		guess.
**	\exception	none	
**	\author		Nate Cleveland
**	\date		3/18/2003 10:15:55 AM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
float MessageHandler::FishBody_ChumEaten(const Key &key)const throw()
{
	const InterpretationModule::FishBody* fishBody = GetFishBody(key);

	if(fishBody)
		return fishBody->m_pStats->GetChum();
	return 0;
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	FishBody_Age
/*!	\brief		Returns to the caller the age of the fish in MINUTES.
**	\param		key		Key of the fish.
**	\return		guess.
**	\exception	none	
**	\author		Nate Cleveland
**	\date		3/18/2003 10:15:55 AM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
float MessageHandler::FishBody_Age(const Key &key)const throw()
{
	const InterpretationModule::FishBody* fishBody = GetFishBody(key);

	//value is given to us in seconds, convert to minutes.
	if(fishBody)
		return fishBody->m_pStats->GetAge() / 60;
	return 0;
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	FishBody_AvgEnergy
/*!	\brief		Returns to the caller the fish's average energy.
**	\param		key		Key of the fish.
**	\return		guess.
**	\exception	none	
**	\author		Nate Cleveland
**	\date		3/18/2003 10:15:55 AM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
float MessageHandler::FishBody_AvgEnergy(const Key &key)const throw()
{
	const InterpretationModule::FishBody* fishBody = GetFishBody(key);

	if(fishBody)
		return fishBody->m_pStats->GetAvgEnergy();
	return 0;
}


////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	FishBody_AvgHealth
/*!	\brief		Returns to the caller the fish's average health.
**	\param		key		Key of the fish.
**	\return		guess.
**	\exception	none	
**	\author		Nate Cleveland
**	\date		3/18/2003 10:15:55 AM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
float MessageHandler::FishBody_AvgHealth(const Key &key)const throw()
{
	const InterpretationModule::FishBody* fishBody = GetFishBody(key);

	if(fishBody)
		return fishBody->m_pStats->GetAvgHealth();
	return 0;
}


////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	FillMementoVector
/*!	\brief		Fills the given memento vector with all the object we wish to
				 save/load to/from.
**	\param		mementoVec		A vector to fill with memento*s.
**	\return		none
**	\exception	none
**	\par		DESCRIPTION:
**					
**	\author		Nate Cleveland
**	\date		3/18/2003 12:40:50 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
void MessageHandler::FillMementoVector(std::vector<FileMemento*>& mementoVec) throw()
{
	//TankSimulator
	// HACK by Austin Spafford
	//   the memory referrenced to by tankSimulatorMementoPtr is deallocated outside
	//   of this function in an ambiguous fashion.
	FileMemento * tankSimulatorMementoPtr = tankSimulator->CreateFileMemento();
	mementoVec.push_back(tankSimulatorMementoPtr);

	//Brain manager
	mementoVec.push_back(dynamic_cast<FileMemento*>(AI_System::AI_BrainManager::getInstance()));

	//UI?
	//...

	//Graphics?
	//...

	//Sound? (Probably not).
	//...

	//Anything else that wish's to save its state.
	//...

}


////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	SaveGame
/*!	\brief		Saves a game to file, automatically appends the path the 
				 front of the given file name.
**	\param		fileName		Name of the file to load from, path will be appended.	
**	\return		boolean true/false, false == file not saved.
**	\exception	none					
**	\author		Nate Cleveland
**	\date		3/18/2003 12:14:06 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
bool MessageHandler::SaveGame(const char* fileName) throw()
{
	_chdir("Aquariums");
	//first off, get an instance of the file io object.
	FileIO* fileIO = FileIO::Instance();
	
	//hopefully this should never happen, thus we assert
	assert(fileIO);	

	//next append the path name.
	char *pathName = new char[strlen(fileName) + strlen(LOAD_SAVE_PATH) + 1];
	strcpy(pathName, LOAD_SAVE_PATH);
	strcat(pathName, fileName);

	//next create a vector that'll be filled with all the mementos we want to
	// load the file from.
	std::vector<FileMemento*> mementoVec;
	FillMementoVector(mementoVec);

	//call the FileIO functionality.
	bool retVal = fileIO->SaveAll(pathName/*file*/, mementoVec);
	_chdir("..");

	// HACK by Austin Spafford
	//   to patch up a memory leak caused by incongruent designs
	//   (the memory is allocated in FillMementoVector())
	//   NOTE: the TankSimulatorFileMemento pointer MUST be the first element
	delete mementoVec.front();

	return retVal;	
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	LoadGame
/*!	\brief		Loads a game from file, automatically appends the path the 
				 front of the given file name.	
**	\param		fileName		Name of the file to load from, path will be appended.
**	\return		boolean true/false, false == file not loaded.
**	\exception	none					
**	\author		Nate Cleveland
**	\date		3/18/2003 12:14:07 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
bool MessageHandler::LoadGame(const char* fileName) throw()
{
	_chdir("Aquariums");
	//first off, get an instance of the file io object.
	FileIO* fileIO = FileIO::Instance();
	
	//hopefully this should never happen, thus we assert
	assert(fileIO);	

	//next append the path name.
	char *pathName = new char[strlen(fileName) + strlen(LOAD_SAVE_PATH) + 1];
	strcpy(pathName, LOAD_SAVE_PATH);
	strcat(pathName, fileName);

	//next create a vector that'll be filled with all the mementos we want to
	// load the file from.
	std::vector<FileMemento*> mementoVec;
	FillMementoVector(mementoVec);

	//call the FileIO functionality.
	bool retVal = fileIO->LoadAll(pathName/*file*/, mementoVec);
	_chdir("..");

	// HACK by Austin Spafford
	//   to patch up a memory leak caused by incongruent designs
	//   (the memory is allocated in FillMementoVector())
	//   NOTE: the TankSimulatorFileMemento pointer MUST be the first element
	tankSimulator->LoadFileMemento(*mementoVec.front());
	delete mementoVec.front();

	return retVal;
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	AbortClientConnection
/*!	\brief		If the local machine is a Client, then the local machine severs its connection (if
				any) to its Server, and becomes a Server itself.  If the local machine is a Server,
				calling this function has no effect.
**	\param		none
**	\return		none
**	\exception	none
**	\par		DESCRIPTION:
**					
**	\author		Nathan Frost
**	\date		3/25/2003 4:03:25 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
void MessageHandler::AbortClientConnection()
{
	if(!NetworkingIsServer())
	{
		delete gameNetworking;
		gameNetworking = new UDPNet::Server;
	}
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	NewAquarium
/*!	\brief		Destroys all Fish and FishBrains in the Aquarium
**	\param		none
**	\return		none
**	\exception	none
**	\par		DESCRIPTION:
**					
**	\author		Nathan Frost
**	\date		4/4/2003 9:50:44 AM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
void MessageHandler::NewAquarium()
{
	NukeAllFish();
	AI_System::AI_BrainManager::getInstance()->nuke();
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	GetProjectedHeightOnHeightMap
/*!	\brief		Vertically projects a point onto the HeightMap and returns the projected point's Y 
				value
**	\param		PointToProject	
**	\return		The projected point's Y value
**	\exception	none
**	\par		DESCRIPTION:
**					
**	\author		Nathan Frost
**	\date		3/26/2003 12:35:54 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
float MessageHandler::GetProjectedHeightOnHeightMap(
										const math::linear_alg::Vector3<float> & PointToProject)
{
	return tankSimulator->GetProjectedHeightOnHeightMap(PointToProject);
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	getBoundary
/*!	\brief		Gets the bounding box of the Aquarium
**	\param		
**	\return		none
**	\exception	none
**	\par		DESCRIPTION:
**					
**	\author		
**	\date		4/9/2003 9:12:26 AM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
void MessageHandler::getBoundary(math::linear_alg::Vector3<float>& min, math::linear_alg::Vector3<float>& max)
{
	tankSimulator->GetBoundary(min, max);
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	getFishMoveGoal
/*!	\brief		Gets the movement goal of the fish
**	\param		
**	\return		none
**	\exception	none
**	\par		DESCRIPTION:
**					
**	\author		
**	\date		4/9/2003 9:12:29 AM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
void MessageHandler::getFishMoveGoal(Key fish, std::string& goal)
{
	AI_System::AI_BrainManager::getInstance()->getMoveGoalString(fish, goal);
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	issueCommand
/*!	\brief		issues a command to the fish through the brain manager
	\param		fish:	To perform the action
**	\param		action:	To instruct the fish to perform
**	\return		none
**	\exception	none
**	\par		DESCRIPTION:
**					
**	\author		Nathan Frost
**	\date		4/9/2003 9:12:31 AM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
void MessageHandler::issueCommand(Key fish, AI_System::Action_ID action)
{
	//if server, do it
	if(NetworkingIsServer())
		AI_System::AI_BrainManager::getInstance()->SuppressBrainWithAction(fish, action);
	else//client, so tell server to do it
	{
		UDPNet::NetMessage issueCommand;
		issueCommand.SetID(UDPNet::ISSUE_COMMAND);
		issueCommand.WriteNBytes(sizeof(Key), &fish);
		issueCommand.WriteNBytes(sizeof(AI_System::Action_ID), &action);
		gameNetworking->SendNetMessage(issueCommand);
	}
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	ClientIssueCommandRequest
/*!	\brief		We, the Server, will get the specified fish to perform the specified action, just
				as the Client has asked
**	\param		ClientIssueCommandRequest:	NetMessage containing the relevant data
**	\return		none
**	\exception	none
**	\par		DESCRIPTION:
**					
**	\author		Nathan Frost
**	\date		4/16/2003 8:46:59 AM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
void MessageHandler::ClientIssueCommandRequest(UDPNet::NetMessage &ClientIssueCommandRequest)
{
	Key fish;
	AI_System::Action_ID action;

	ClientIssueCommandRequest.ReadNBytes(sizeof(fish), &fish);
	ClientIssueCommandRequest.ReadNBytes(sizeof(AI_System::Action_ID), &action);

	AI_System::AI_BrainManager::getInstance()->SuppressBrainWithAction(fish, action);
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	addToMatingList
/*!	\brief		Adds the specified fish to the mating list
**	\param		fish:	To add to the mating list
**	\return		none
**	\exception	none
**	\par		DESCRIPTION:
**					
**	\author		Nathan Frost
**	\date		4/9/2003 9:12:34 AM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
void MessageHandler::addToMatingList(Key fish)
{
	if(NetworkingIsServer())
		AI_System::AI_BrainManager::getInstance()->AddToMatingList(fish);
	else
	{
		UDPNet::NetMessage addFishToMatingList;
		addFishToMatingList.SetID(UDPNet::ADD_FISH_TO_MATING_LIST);
		addFishToMatingList.WriteNBytes(sizeof(Key), &fish);
		gameNetworking->SendNetMessage(addFishToMatingList);
	}
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	ClientAddToMatingListRequest
/*!	\brief		We, the Server, will add the Fish to the Mating List, as the Client requested
**	\param		ClientAddToMatingListRequest: NetMessage that contains the relevant data
**	\return		none
**	\exception	none
**	\par		DESCRIPTION:
**					
**	\author		
**	\date		4/15/2003 9:07:33 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
void MessageHandler::ClientAddToMatingListRequest(UDPNet::NetMessage &ClientAddToMatingListRequest)
{
	Key fish;
	ClientAddToMatingListRequest.ReadNBytes(sizeof(Key), &fish);

	AI_System::AI_BrainManager::getInstance()->AddToMatingList(fish);
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	setMinFish
/*!	\brief		
**	\param		
**	\return		none
**	\exception	none
**	\par		DESCRIPTION:
**					
**	\author		
**	\date		4/9/2003 9:12:37 AM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
void MessageHandler::setMinFish(unsigned int val)
{
	if(NetworkingIsServer())
		AI_System::AI_BrainManager::getInstance()->setFishMin(val);
	else
		;//notify the user that clients can't do this
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	getMinFish
/*!	\brief		
**	\param		
**	\return		none
**	\exception	none
**	\par		DESCRIPTION:
**					
**	\author		
**	\date		4/9/2003 9:12:39 AM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
unsigned int MessageHandler::getMinFish()
{
	return AI_System::AI_BrainManager::getInstance()->getFishMin();
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	setMaxFish
/*!	\brief		
**	\param		
**	\return		none
**	\exception	none
**	\par		DESCRIPTION:
**					
**	\author		
**	\date		4/9/2003 9:12:42 AM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
void MessageHandler::setMaxFish(unsigned int val)
{
	if(NetworkingIsServer())
		AI_System::AI_BrainManager::getInstance()->setFishMax(val);
	else
		;//notify the user that clients can't do this
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	getMaxFish
/*!	\brief		
**	\param		
**	\return		none
**	\exception	none
**	\par		DESCRIPTION:
**					
**	\author		
**	\date		4/9/2003 9:12:44 AM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
unsigned int MessageHandler::getMaxFish()
{
	return AI_System::AI_BrainManager::getInstance()->getFishMax();
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	offLineMode
/*!	\brief		
**	\param		
**	\return		none
**	\exception	none
**	\par		DESCRIPTION:
**					
**	\author		
**	\date		4/9/2003 9:12:46 AM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
void MessageHandler::offLineMode(bool val)
{
	//Do whatever it takes to go to offline mode here
	if(val)
	{
		g_IsOffline = true;
		TimeObject::Instance()->setGameSpeed(1.0);
	}
	else
	{
		g_IsOffline = false;
		TimeObject::Instance()->setGameSpeed(1.0);
	}
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	numFish
/*!	\brief		Returns the number of fish in the simulation
**	\param		none
**	\return		The number of fish in the simulation
**	\exception	none
**	\par		DESCRIPTION:
**					
**	\author		Nathan Frost
**	\date		4/9/2003 9:12:49 AM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
unsigned int MessageHandler::numFish()
{
	return tankSimulator->GetNumberOfFish();
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	NotifyRemoteMachinesAboutTimeChange
/*!	\brief		Notifies all connected RemoteMachines that a new (framerate independent) timestep
				has been set
**	\param		none
**	\return		none
**	\exception	none
**	\par		DESCRIPTION:
**					
**	\author		Nathan Frost
**	\date		4/16/2003 9:35:07 AM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
void MessageHandler::NotifyRemoteMachinesAboutTimeChange()
{
	double gameSpeed = TimeObject::Instance()->getGameSpeed();

	UDPNet::NetMessage newGameSpeed;
	newGameSpeed.SetID(UDPNet::NEW_GAME_SPEED);
	newGameSpeed.WriteNBytes(sizeof(gameSpeed), &gameSpeed);
	gameNetworking->SendNetMessage(newGameSpeed);
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	increaseGameSpeed
/*!	\brief		Increases the (framerate independent) game speed
**	\param		none
**	\return		none
**	\exception	none
**	\par		DESCRIPTION:
**					
**	\author		Nathan Frost
**	\date		4/9/2003 9:12:53 AM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
void MessageHandler::increaseGameSpeed()
{
	if(NetworkingIsServer())
	{
		TimeObject::Instance()->incrementGameSpeed();
		NotifyRemoteMachinesAboutTimeChange();
	}
	else
		;//notify the user that clients can't do this
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	decreaseGameSpeed
/*!	\brief		Decreases the (framerate independent) game speed
**	\param		none
**	\return		none
**	\exception	none
**	\par		DESCRIPTION:
**					
**	\author		Nathan Frost
**	\date		4/9/2003 9:12:56 AM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
void MessageHandler::decreaseGameSpeed()
{
	if(NetworkingIsServer())
	{
		TimeObject::Instance()->decrementGameSpeed();
		NotifyRemoteMachinesAboutTimeChange();
	}
	else
		;//notify the user that clients can't do this
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	NewGameSpeed
/*!	\brief		The Server is telling us, the Client, that there is a new gamespeed that we must 
				adhere to
**	\param		NewGameSpeed: NetMessage containing the relevant data
**	\return		none
**	\exception	none
**	\par		DESCRIPTION:
**					
**	\author		Nathan Frost
**	\date		4/16/2003 9:48:11 AM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
void MessageHandler::NewGameSpeed(UDPNet::NetMessage &NewGameSpeed)
{
	double newSpeed;
	NewGameSpeed.ReadNBytes(sizeof(newSpeed), &newSpeed);
	TimeObject::Instance()->setGameSpeed(newSpeed);
}

//Nathan should probably put a comment here.
void MessageHandler::ClearFishBrain(Key fishID)
{
	InterpretationModule::FishBody* fishBody = GetFishBodyPtr(fishID);

	//make sure we have a fish.
	if(fishBody)
	{
		//The doubld call the function is intentional. Don't ask or I might have to 
		// explan it and neither of us would enjoy that.
		fishBody->Flush();
		fishBody->Flush();
	}
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	MessageHandler::doesFishExist
/*!	\brief		Checks to see if a fish exists
**	\param		fish, fish key
**	\return		true if the fish exists, false otherwise
**	\exception	none
**	\author		John Corpening
**	\date		4/10/2003 6:47:16 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
bool MessageHandler::doesFishExist(Key fish)
{
	return tankSimulator->DoesFishBodyExist(fish);
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	BreedFish
/*!	\brief		Combines the physical attributes of 1,2 or 0 fish into a child.
**	\param		child		
	\param		parentA		Key to the first parent, if any. May be zero.
	\param		parentB		Key to the second parent, if any. May be zero.
**	\return		none
**	\exception	none
**	\par		DESCRIPTION:
**					Combines the attributes of a child from its parents, if any.
					 A fish with no parents will totally mutate into random 
					 attributes. A fish with only one parent will have some
					 mutation, but will mostly resemble its parent.
					 A fish with two parents will become a cross between the
					 two with an occasional mutation thrown in sometimes.
**	\author		Nate Cleveland
**	\date		4/15/2003 2:42:21 PM
**	\warning	If parentA is zero then parentB is assumed to be as well.
*/
////////////////////////////////////////////////////////////////////////////////
void MessageHandler::BreedFish(Key child, Key parentA/*=0*/, Key parentB/*=0*/)
{
	InterpretationModule::FishBody* fishBodyChild	=	GetFishBodyPtr(child);

	//check for a child, silent fail
	if(fishBodyChild) 
	{
		//check for an A parent
		InterpretationModule::FishBody* fishBodyParentA	= NULL;
		if(0 != parentA)
		{
			//check for a B parent.
			if(0 != parentB)
				fishBodyChild->Breed(parentA, parentB);

			//otherwise only parently mutate the child
			else
				fishBodyChild->Breed(parentA);
		}

		//otherwise massively mutate the fish.
		else
			fishBodyChild->Breed();
	}//end if
	
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	MessageHandler::setHeightMap
/*!	\brief		Used by the UI to change the height map
**	\param		string, the name of the height map to use
**	\return		none
**	\exception	none
**	\author		John Corpening
**	\date		4/17/2003 12:06:14 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
void MessageHandler::setHeightMap(std::string &file)
{
	if(NetworkingIsServer())
	{
		//Check if the file contains the path:
		std::string::size_type pos = file.find("Resources/HeightMaps/");
		if(pos == std::string::npos)
		{
			std::string temp("Resources/HeightMaps/");
			temp += file;
			file = temp;
		}
		// Check to see if the user forgot to append ".tga" onto the filename
		pos = file.find(".tga");
		if(pos == std::string::npos)
		{
			std::string temp(".tga");
			file += temp;			
		}
		//This copying is done due to the long trail of non-const paramaters in
		//the import funciton calls
		char *temp = new char[file.size() + 1];
		strcpy(temp, file.c_str());
		//Set the height map
		ImportHeightMapFromImage(temp, 10000);
		//Get the current tank values
		math::linear_alg::Vector3<float> min,max;
		tankSimulator->GetBoundary(min, max);
		//Set the current tank values (this formats the heightmap such that it fills half the tank)
		tankSimulator->SetBoundary(min, max);
		delete[] temp;
	}
	else
		;//notify user that Client can't do this
}