////////////////////////////////////////////////////////////////////////////////
/*!	\file		MessageHandler.h
**	\brief		Declares the MessageHandler class
**	\author		Nathan Frost
**	\date		10/17/2002 2:04:50 PM
**	\par		DESCRIPTION:
					- Decouples the data that defines the current state of the Environment
					(the Tank Simulator) from the operations that nondeterministically
					modify that data (Player Input and AI [and Networking, which acts as a
					surrogate for remote Player Input and AI]).
					- Ensures that the Networking, Player Input, AI and the Tank Simulator
					are able to use a simple interface to exchange data with the rest of the
					system.
					- Hands out, and keeps track of, the numbers that uniquely identify every
					Fish and Object in the Aquarium.
					- Uses the Singleton and Mediator patterns
**					
**	\par		REMARKS:
					None
**					
*/
////////////////////////////////////////////////////////////////////////////////
#pragma once

#include <list>
#include <vector>
#include <bitset>
#include <limits>
#include <string>
#include "networking/WinSockSynthesis.h"


/* Forward declarations */
class FileMemento;
class TankSimulatorFileMemento;
class TankSimulator;

namespace AI_System
{
	struct AI_Action;
	struct AI_ActionTag;
	enum Action_ID;
}

namespace InterpretationModule
{
	class FishBody;
	class StrObj;
}

namespace UDPNet
{
	class GameNetworking;
	class NetMessage;
}

namespace MsgHandler
{
	class MessageHandler;
	typedef unsigned int Key;	///<Uniquely identifies every entity in the game
	#define LARGEST_KEY (1 << 15)	///<one larger than the largest valid Key value (leaving the 
									///<out the mostSignificantBit, which is reserved for 
									///<signifying the Official or Unofficial status of a Key).  
									///<In other words, (LARGEST_KEY - 1) is the number of Keys that 
									///<can be allocated at once on the Server (since 0 is an  
									///<invalid Key value).  Must be a power of 2.
}

namespace math
{
	namespace linear_alg
	{
		template<typename type>
			class Vector3;
	}
	
}


namespace MsgHandler
{
////////////////////////////////////////////////////////////////////////////////
/*!	\class		MessageHandler
**	\brief		The Singleton MessageHandler is the central Mediator of the game system, and keeps
				track of entity Keys.
**	\author		Nathan Frost
**	\date		10/17/2002 2:13:52 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
	class MessageHandler
	{
		//singleton
		MessageHandler();


		/* Key Generation */
		std::bitset<((LARGEST_KEY << 1))> keyUsed;
		Key curKey;

		bool ValidateKey(const Key &key);


		bool gameOn;///<whether or not the game is active -- if gameOn==false, the game is in the 
					///<process of quitting

		
		//Pointers to modules MessageHandler must communicate with
		TankSimulator *tankSimulator;
		UDPNet::GameNetworking *gameNetworking;


		void NukeAllFish();
		void NotifyRemoteMachinesOfFishSelectListChanging();
		void NotifyRemoteMachinesOfFishSuppressionListChanging();
		void NotifyRemoteMachinesAboutTimeChange();
		void FillMementoVector(std::vector<FileMemento*>& mementoVec) throw();
	public:
		~MessageHandler();
		static MessageHandler *Instance();
		static void DestroyInstance();


		///Key management
		Key GenerateKey();
		void ReturnKey(const Key &key);
		bool RegisterKey(Key TheKey);
		bool IsOfficial(Key TheKey);


		///Game control flow
		bool GameOn();
		void QuitGame();
		void Logic();


		///Registration functions so that the MessageHandler can function
		MessageHandler &RegisterTankSimulator(TankSimulator *tankSimulator);


		///Game hub functionality
		/* Game Initialization and Save and Load Functionality */
		void NewAquarium();
		bool ImportHeightMapFromImage(char *FileName, float MaxHeight);
		bool ExportHeightMapToImage(char *FileName);
		bool SaveGame(const char* fileName) throw();
		bool LoadGame(const char* fileName) throw();

		/* Fish Functionality */
		std::list<Key> &GetMatingFishListRef();
		std::list<Key> &MessageHandler::GetDeleteFishListRef();
		bool RemoveFishBody(Key &Body);
		bool AddFishBody(const Key &Body,  Key parentA=0, Key parentB=0);
		bool SetFishBodyActions(const Key &FishKey, std::vector<AI_System::AI_Action> &Actions);
		bool SetFishBodyActions(UDPNet::NetMessage &TheNetMessage);
		InterpretationModule::FishBody* GetFishBodyPtr(const Key &Body);
		const InterpretationModule::FishBody* GetFishBody(const Key key)const throw();
		void GetSensedObject(Key &Object, std::vector<Key>& retVal);
		void GetPerformedActions(Key &Object, std::vector<AI_System::AI_ActionTag>& ret);
		void GetVectorOfFishVisibleToFish(Key FishKey, std::vector<MsgHandler::Key>& vec);
		Key GetNextFish(Key FishKey);
		bool GetPosition(Key key, math::linear_alg::Vector3<float>& pos )throw();
		bool GetDirection(Key key, math::linear_alg::Vector3<float>& pos) throw();
		void SwimModeGotoPoint(Key fishID, math::linear_alg::Vector3<float> point);
		void SwimModeGotoPoint(UDPNet::NetMessage &TheNetMessage);
		void getFishMoveGoal(Key fish, std::string& goal);//MIGHT NEED NETWORKING
		void issueCommand(Key fish, AI_System::Action_ID action);
		void addToMatingList(Key fish);
		// Added by Nate Cleveland [4/10/2003]
		// clears out all the thoughts a fish is thinking.
		void ClearFishBrain(Key fishID);
		unsigned int numFish();
		bool doesFishExist(Key fish);
		void AddRandomFish();
		bool TraceRayAgainstAllFish(const math::linear_alg::Vector3<float> & rayOrigin, 
									const math::linear_alg::Vector3<float> & rayDirection, 
									Key & fishKeyRef);


		/* UI Functionality */
		// Added by Nate [3/10/2003]
		//functions to provide UI with data on a fishBody. All of these
		//values are 0 to 1. The functionality is covered in depth in 
		//attributes.hpp
		float FishBody_Meta			(const Key &Body)const throw();
		float FishBody_Armor		(const Key &Body)const throw();
		float FishBody_Teeth_Baleen	(const Key &Body)const throw();
		float FishBody_Size			(const Key &Body)const throw();
		float FishBody_MatingIncline(const Key &Body)const throw();
		float FishBody_Health		(const Key &Body)const throw();
		float FishBody_Stomach		(const Key &Body)const throw();
		float FishBody_EnergyLevel	(const Key &Body)const throw();
		bool SuppressRequest(const Key& fish) throw();
		void UnsuppressRequest(const Key& fish) throw();
		//Added by John [3/15/03]
		//Request to select a fish
		bool SelectRequest(const Key& fish) throw();
		void DeselectRequest(const Key& fish) throw();
		// Added by nate [3/18/2003]
		//Functions provide the UI with stats on a fishBody. These may or may not
		//be zero to one values, depending on the data in question, Functionality
		//is coverved in depth in statistics.hpp
		float FishBody_KillCount	(const Key &Body)const throw();
		float FishBody_BeenAttacked	(const Key &Body)const throw();
		float FishBody_HasAttacked	(const Key &Body)const throw();
		float FishBody_Distance		(const Key &Body)const throw();
		float FishBody_AvgSpeed		(const Key &Body)const throw();
		float FishBody_PlanktonEaten(const Key &Body)const throw();
		float FishBody_ChumEaten	(const Key &Body)const throw();
		float FishBody_Age			(const Key &Body)const throw();
		float FishBody_AvgEnergy	(const Key &Body)const throw();
		float FishBody_AvgHealth	(const Key &Body)const throw();
		// Added by Nate so the UI can access info on a fish's action and target [3/17/2003]
		void GetActionString(const Key& fishKey,		std::string **mouth  = NULL, 
							 std::string **move = NULL, std::string **marker = NULL);

		void GetTargetString(const Key& fishKey,		std::string **mouth  = NULL, 
							 std::string **move = NULL, std::string **marker = NULL);
		// Added by nate so we can highlight/tint fish [3/13/2003]
		void HighLightFish(const Key& fishKey, bool isOn, float red = 1, float green = 1, float blue = 1) throw();


		/* Tank Related Functionality */
		float GetPlanktonDensity(const math::linear_alg::Vector3<float>& pos) const throw();
		float GetProjectedHeightOnHeightMap(
										const math::linear_alg::Vector3<float> & PointToProject);
		void setTankWidth(float width) throw();
		void setTankHeight(float height) throw();
		void setTankDepth(float depth) throw();
		void getBoundary(math::linear_alg::Vector3<float>& min, math::linear_alg::Vector3<float>& max);
		void setMinFish(unsigned int val);
		void setMaxFish(unsigned int val);
		unsigned int getMinFish();
		unsigned int getMaxFish();
		void offLineMode(bool val);
		void increaseGameSpeed();
		void decreaseGameSpeed();


		/* Network Functionality */
		void NewGameSpeed(UDPNet::NetMessage &NewGameSpeed);
		void ClientIssueCommandRequest(UDPNet::NetMessage &ClientIssueCommandRequest);
		void ClientAddToMatingListRequest(UDPNet::NetMessage &ClientAddToMatingListRequest);
		void ClientSuppressRequest(UDPNet::NetMessage &ClientSuppressRequest);
		void ClientUnsuppressRequest(UDPNet::NetMessage &ClientUnsuppressRequest);
		bool ConnectToServer(const char *IPAddress);
		void WeBootedClient(u_long IPAddress);
		void WeAcceptedClient(u_long IPAddress);
		void OurServerIsDelinquent();
		void WeAreConnected(UDPNet::NetMessage &HeightMapData);
		void WeAreBooted();
		bool NetworkingIsServer();
		std::string NetworkingState();
		void GotServerUpdate(UDPNet::NetMessage &ServerUpdate);
		u_long GetMyIP();
		void FishSelectListChanged(UDPNet::NetMessage &fishSelectListChanged);
		void FishSuppressListChanged(UDPNet::NetMessage &FishSuppressedListChanged);
		void ClientSelectRequest(UDPNet::NetMessage &ClientSelectRequest, u_long SenderIP);
		void ClientDeselectRequest(UDPNet::NetMessage &ClientDeselectRequest, u_long SenderIP);
		void BootAllClients();
		void FillOutHeightMapNetMessage(UDPNet::NetMessage &TheNetMessage);
		void FillOutServerUpdate(UDPNet::NetMessage &ServerUpdate);
		void AbortClientConnection();
		

		// Added by Nate Cleveland [4/15/2003]
		// Fish Body Breeding functionality. See cpp for documentation.
		void BreedFish(Key child, Key parentA=0, Key parentB=0); 

		//Added by John Corpening [4/17/2003]
		void setHeightMap(std::string &file);
	};
}
////////////////////////////////////////////////////////////////////////////////
