////////////////////////////////////////////////////////////////////////////////
/*!	\file		MessageHandlerKey
**	\brief		Implementation of all of the MessageHandler's Key Generation functionality
**	\author		Nathan Frost
**	\date		3/11/2003 8:08:10 PM
**	\par		DESCRIPTION:
**					
**	\par		REMARKS:
**					
*/
////////////////////////////////////////////////////////////////////////////////
#include<assert.h>
#include "MessageHandler.h"
using namespace MsgHandler;
// memory leak detection
#ifdef _DEBUG
    #include"memleak.h"
    #define new DEBUG_NEW
    #undef THIS_FILE
    static char THIS_FILE[] = __FILE__;
#endif


//Any number from 1 to ((LARGEST_KEY - 1)*2) can be used as a valid Key, but the most significant
//bit is reserved for flagging a given Key as Official or Unofficial
const unsigned int mostSignificantBit = (LARGEST_KEY);


//singleton implementation
static MessageHandler *instance;


////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	Instance()
/*!	\brief		Allows the client access to the MessageHandler object.
**	\param		none
**	\return		Pointer to the singleton MessageHandler object
**	\exception	none
**	\par		DESCRIPTION:
**					
**	\author		Nathan Frost
**	\date		10/17/2002 8:33:11 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
MessageHandler *MessageHandler::Instance()
{
	if(!instance)
		instance = new MessageHandler();

	return instance;
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	DestroyInstance
/*!	\brief		Destroys the instance of the MessageHandler such that the next call to Instance()
				will force the MessageHandler to be reconstructed
**	\param		none
**	\return		none
**	\exception	none
**	\par		DESCRIPTION:
**					
**	\author		Nathan Frost
**	\date		4/1/2003 3:45:31 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
void MessageHandler::DestroyInstance()
{
	delete instance;
	instance = 0;
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	GenerateKey
/*!	\brief		Generates a valid Key to identify an entity.
**	\param		none
**	\return		Valid Key (never zero)
**	\exception	none
**	\par		DESCRIPTION:
**				The generated Key is Official if the local machine is a Server; otherwise the
				generated Key is Unofficial, because the local machine is a Client. 

**	\author		Nathan Frost
**	\date		10/17/2002 8:37:06 PM
**	\warning	
				- If there are no valid Keys left, there'll be a stack overflow!  Remember to
				ReturnKey()s.
				- Currently, GenerateKey() generates an Official Key half the time,
				and an Unofficial Key otherwise
*/
////////////////////////////////////////////////////////////////////////////////
Key MessageHandler::GenerateKey()
{
	//find a curKey that is free
	while(curKey < LARGEST_KEY && keyUsed.test(curKey))
		++curKey;
	//if we need to wrap around to find a free curKey, then do so
	if(curKey >= LARGEST_KEY)
	{
		curKey = 1;
		return GenerateKey();
	}
	
	Key ret = curKey;
	//if we are a Server, then this Key is Official -- otherwise it is not
	if(NetworkingIsServer())
		ret |= mostSignificantBit;

	//mark the Key we generated as used
	keyUsed.set(ret);
	++curKey;	//we'll probably have a valid curKey next time we need to Generate one

	return ret;	//return the valid Key

}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	ReturnKey()
/*!	\brief		Frees the specified Key, so another entity can make use of it.
**	\param		key: The Key to free up
**	\return		none
**	\exception	none
**	\par		DESCRIPTION:
**				Passing an invalid key has no effect.
**	\author		Nathan Frost
**	\date		10/17/2002 8:40:17 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
void MessageHandler::ReturnKey(const Key &key)
{
	if(ValidateKey(key))
	{
		keyUsed.reset(key);
	}

	return;
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	ValidateKey()
/*!	\brief		Checks the specified Key for validity
**	\param		key: The Key to check for validity
**	\return		true if key is valid, false if it is not
**	\exception	none
**	\par		DESCRIPTION:
**				
**	\author		Nathan Frost
**	\date		10/17/2002 8:40:17 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
bool MessageHandler::ValidateKey(const Key &key)
{
	//the largest possible valid Key is LARGEST_KEY with its mostSignificantBit set -- hence the
	//key must be smaller than (LARGEST_KEY << 1)
	if(	(key < (LARGEST_KEY << 1) && keyUsed.test(key)))
		return true;
	else
		return false;
}


////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	IsOfficial
/*!	\brief		Lets the caller know if the TheKey is Official or Unofficial
**	\param		TheKey: To check for Official or Unofficial status
**	\return		true if TheKey is Official, false otherwise
**	\exception	none
**	\par		DESCRIPTION:
**					
**	\author		Nathan Frost
**	\date		3/11/2003 8:14:41 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
bool MessageHandler::IsOfficial(Key TheKey)
{
	if(TheKey & mostSignificantBit)
		return true;
	else
		return false;
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	RegisterKey
/*!	\brief		Registers a Key as used.
**	\param		TheKey	An Official Key to register (ie TheMostSignificantBit is set)
**	\return		true if TheKey was successfully registered, false if it was not registered because
				it is already in use
**	\exception	none
**	\par		DESCRIPTION:
**					
**	\author		Nathan Frost
**	\date		4/3/2003 8:54:34 AM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
bool MessageHandler::RegisterKey(Key TheKey)
{
	//if TheKey is not in use, mark TheKey as used
	if(!keyUsed.test(TheKey))
	{
		keyUsed.set(TheKey);
		return true;
	}
	else
		return false;
}
