////////////////////////////////////////////////////////////////////////////////
/*!	\file		GameNetworking.cpp
**	\brief		Implementation
**	\author		Nathan Frost
**	\date		2/11/2003 9:25:07 PM
**	\par		DESCRIPTION:
**
**	\par		REMARKS:
**
*/
////////////////////////////////////////////////////////////////////////////////
#pragma warning(disable:4786)
#include "GameNetworking.h"
#include "UDPNetworking.h"
#include "../messageHandler/MessageHandler.h"
#include "../timeObject/Timer.h"
#include "RemoteSocksToWaitOn.h"
#include "NetMessageID.h"
#include <assert.h>
#include "../nexception/Nexception.h"
//#include "../debugging/Debug.h"
// memory leak detection
#ifdef _DEBUG
    #include"memleak.h"
    #define new DEBUG_NEW
    #undef THIS_FILE
    static char THIS_FILE[] = __FILE__;
#endif
using namespace UDPNet;


const float EVENT_EXPIRY_TIME = 2;///<in seconds
const float TIMEOUT = 20;///<in seconds
const float TIME_BETWEEN_CLIENT_KEEP_ALIVES = 5;///<in seconds
const float TIME_BETWEEN_CLIENT_HELLOS = 2;///<in seconds
const float TIME_BETWEEN_SERVER_UPDATES = 0.25;///<in seconds
const float TIME_BEFORE_READY_TO_PLAY = TIME_BETWEEN_CLIENT_HELLOS*4;///<in seconds


////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	GameNetworking
/*!	\brief		Constructor
**	\return		none
**	\exception	
				- If LISTEN_PORT is already occupied by another application, an exception will be
				thrown, since this port is required for the networking module to function.
**	\par		DESCRIPTION:
**					
**	\author		Nathan Frost
**	\date		2/12/2003 8:49:26 AM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
GameNetworking::GameNetworking():
udpNetworking(*UDPNetworking::Instance()), timer(*TimeObject::Instance()), curDocumentNum(0)
{
	//otherwise a Client may time out before it actually gets truly included in the game
	assert(TIME_BEFORE_READY_TO_PLAY < TIMEOUT);

    remoteSocksToWaitOn = new RemoteSocksToWaitOn;

	receivePort = LISTEN_PORT;
	udpNetworking.CreateSock(receiveSock, receivePort);
	remoteSocksToWaitOn->AddSock(receiveSock);

	u_short anyPort = ANY_PORT;
	udpNetworking.CreateSock(sendSock, anyPort);               

	remoteSocksToWaitOn->LaunchListeningThread(udpNetworking, event);
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	~GameNetworking
/*!	\brief		Destructor
**	\param		none
**	\return		none
**	\exception	none
**	\par		DESCRIPTION:
**					
**	\author		Nathan Frost
**	\date		2/13/2003 9:41:03 AM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
GameNetworking::~GameNetworking()
{
	/* Relinquish SOCKETs initialized to listen to the machines we were staying in contact with */
	BootAllRemoteMachines();

    /* Terminate the listening thread if it is active */
	//tell the thread to terminate itself, jar it out of select(), then wait for it to die
	remoteSocksToWaitOn->KillThread();                    
    NotifySelf();
    while(remoteSocksToWaitOn->CurThreadState() != DEAD)
		Sleep(0);

	/* Relinquish memory allocated to any events still waiting to be sent to the MessageHandler */
	for(std::deque<Event>::iterator jt = event.begin(); jt != event.end(); ++jt)
		jt->netMessage.DeallocatePackets();

	udpNetworking.DestroySock(sendSock);
	udpNetworking.DestroySock(receiveSock);

    delete remoteSocksToWaitOn;


    //BEG_HACK: memory allocation
    if(numPacketsAllocated != 0 || udpNetworking.numSocketsAllocated != 0)
    {
		Beep(440,500);
		Debugging::Debugger::Instance()->log(	Debugging::NET, Debugging::L4, 
												"numPacketsAllocated: %i, numSocketsAllocated: %i", 
												numPacketsAllocated, 
												udpNetworking.numSocketsAllocated);
    }
    //END_HACK: memory allocation
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	BootAllRemoteMachines
/*!	\brief		Boots all remote machines the local machine is currently in contact with
**	\param		none
**	\return		true if one or more remote machine was booted, false if no remote machines were
				connected, and hence, no remote machines were booted
**	\exception	none
**	\par		DESCRIPTION:
**					
**	\author		Nathan Frost
**	\date		3/17/2003 10:54:34 AM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
bool GameNetworking::BootAllRemoteMachines()
{
	bool ret = false;//assume there is no remote machines to boot

	//boot everyone we're currently in contact with
	while(remoteMachine.size())
    {
		ret = true;
        //need to copy the temporary returned by begin() to a local variable so that the local
        //variable can be passed by non_const reference
        std::list<RemoteMachine>::iterator it = remoteMachine.begin();
		BootRemoteMachine(it);
    }

	return ret;
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	GenerateDocumentNum
/*!	\brief		Generates a valid DocumentNum for a given NetMessage
**	\param		none
**	\return		none
**	\exception	none
**	\par		DESCRIPTION:
**					
**	\author		Nathan Frost
**	\date		2/13/2003 10:39:28 AM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
unsigned int GameNetworking::GenerateDocumentNum()
{
	++curDocumentNum;
	return curDocumentNum;
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	ProcessEvents
/*!	\brief		Processes all received events such that the MessageHandler is informed of complete
				events and incomplete, "overripe" events are discarded.  It also updates
				RemoteSocksToWaitOn's timeLeftToModeChange value.
**	\param		none
**	\return		none
**	\exception	none
**	\par		DESCRIPTION:
**					
**	\author		Nathan Frost
**	\date		2/12/2003 12:21:51 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
void GameNetworking::ProcessEvents()
{
	/* Update time values so networking engine can function */
	remoteSocksToWaitOn->StepTimeLeftToModeChange(-static_cast<float>(timer.getRealTimeStep()));
	for(std::list<RemoteMachine>::iterator kt = remoteMachine.begin(); kt != remoteMachine.end();
		++kt)
	{
		float timeStep = static_cast<float>(timer.getRealTimeStep());
		kt->timeSinceLastUpdate += timeStep;
		if(!kt->receivedHeightmap)
		{	//if this player hasn't yet got the Heightmap...
			kt->timeSinceLastHello += timeStep;//...then wait for them to stop sending HELLO's
			if(kt->timeSinceLastHello > TIME_BEFORE_READY_TO_PLAY)
				kt->receivedHeightmap = true;//assume that this player has received the Heightmap
		}
	}

	LockMutexEvent();
	std::deque<Event>::iterator it = event.begin();
	while(it != event.end())
	{
		/* Track Events' time spent alive and remove any events that have expired */
		it->timeSpentAlive += static_cast<float>(timer.getRealTimeStep());
		if(it->timeSpentAlive > EVENT_EXPIRY_TIME)
		{
			//BEG_HACK
			if(it->netMessage.GetID() == UDPNet::YOU_ARE_ACCEPTED)
				Beep(440,250);
			else if(it->netMessage.GetID() == UDPNet::SERVER_UPDATE)
				Beep(440,250);
			Debugging::Debugger::Instance()->output(Debugging::NET, Debugging::L4, 
													"Deleted message %i.", it->netMessage.GetID());
			//END_HACK
			it->netMessage.DeallocatePackets();
			it = event.erase(it);
		}
		/* If the NetMessage has completely arrived, send it off */
		else if(it->netMessage.GetNumPackets() == it->netMessage.GetNumCurrentPackets())
		{
			//find the remoteMachine that just spoke with us
			bool care = false;//assume that some machine not playing the game sent us this data
			for(std::list<RemoteMachine>::iterator jt = remoteMachine.begin();
				jt != remoteMachine.end(); ++jt)
			{
				if(jt->ipAddress == it->senderIP)
				{
					care = true;//a connected machine in the game sent us data
					jt->timeSinceLastUpdate = 0;
				}
			}
			if(care || it->netMessage.GetID() == HELLO)
				SendEventToMessageHandler(*it);//this function must DeallocatePackets() now
            else
                it->netMessage.DeallocatePackets();//in this case we suppress event; so clean up
			it = event.erase(it);//and either way we have now dealt with this event
		}
		else
			++it;//move on to the next Event -- this one's neither expired nor fully reconstructed
	}
	UnlockMutexEvent();
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	DetectRemoteMachineDelinquency
/*!	\brief		If a RemoteMachine has not communicated in TIMEOUT seconds, the MessageHandler is
                informed.
**	\return		Whether or not one or more machines were BootRemoteMachine()'d.
**	\exception	none
**	\par		DESCRIPTION:
**					
**	\author		Nathan Frost
**	\date		2/13/2003 9:03:46 AM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
bool GameNetworking::DetectRemoteMachineDelinquency()
{
    bool ret = false;//assume no booting
    std::list<RemoteMachine>::iterator it = remoteMachine.begin();
	while(it != remoteMachine.end())
	{
		if(it->timeSinceLastUpdate > TIMEOUT)
        {
            ret = true;//we booted someone

            //this specific call is important because Client needs to retain receiveSock
            if(BootRemoteMachine(it->ipAddress))
			{
				if(MsgHandler::MessageHandler::Instance()->NetworkingIsServer())
					MsgHandler::MessageHandler::Instance()->WeBootedClient(it->ipAddress);
				else
					MsgHandler::MessageHandler::Instance()->OurServerIsDelinquent();

                it = remoteMachine.begin();//BootRemoteMachine removed a machine
			}
        }
        else
            ++it;//move on to the next remoteMachine, this one isn't delinquent
	}

    return ret;
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	NotifySelf
/*!	\brief		Causes the local machine to send a dummy NetMessage to the local machine
**	\param		none
**	\return		none
**	\exception	none
**	\par		DESCRIPTION:
**					- note that this function sends more bytes than it needs to to get the job
					done.  I do not believe this is a problem, since the hardware should route the
					message directly and therefore should not have any impact on bandwidth
**	\author		Nathan Frost
**	\date		2/13/2003 10:08:35 AM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
void GameNetworking::NotifySelf()
{
    NetMessage dummy;
    dummy.SetID(NOTIFY_SELF);
	dummy.SetDocumentNum(GenerateDocumentNum());
	udpNetworking.SendPacket(	sendSock, udpNetworking.MyIPAddress(), receivePort,
								dummy.GetPacket(0), dummy.GetPacketSize(dummy.GetPacket(0)));
    dummy.DeallocatePackets();
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	SendNetMessageToSingleRemoteMachine
/*!	\brief		Sends out a NetMessage by sending them off on their perilous journeys to the 
				specified machine.
**	\param		Message:		The NetMessage to send
	\param		IPAddress:		of the RemoteMachine to send to
	\param		PortToSendTo:	of the RemoteMachine to send to
**	\return		none
**	\exception	none
**	\par		DESCRIPTION:
**					
**	\author		Nathan Frost
**	\date		2/13/2003 10:22:38 AM
**	\warning	
				- The Message must have Message.SetDocumentNum(GenerateDocumentNum()) called on it
				first.
				- The Message must have Message.DeallocatePackets() called after this function 
				returns.
*/
////////////////////////////////////////////////////////////////////////////////
void GameNetworking::SendNetMessageToSingleRemoteMachine(NetMessage &Message, 
														 u_long IPAddress, u_short PortToSendTo)
{
	for(unsigned int i=0; i < Message.GetNumPackets(); ++i)
	{
		udpNetworking.SendPacket(	sendSock, IPAddress, PortToSendTo, Message.GetPacket(i),
									Message.GetPacketSize(Message.GetPacket(i)));
	}
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	SendNetMessage()
/*!	\brief		Sends out a NetMessage to every other machine that we are currently in contact
				with and is in the game by stamping every one of the NetMessage's Packets with a
				DocumentNum and then sending to every machine we're in contact with.  Calls 
				Message.DeallocatePackets().
**	\param		Message: The NetMessage to send
**	\return		none
**	\exception	none
**	\par		DESCRIPTION:
**					
**	\author		Nathan Frost
**	\date		2/13/2003 10:22:38 AM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
void GameNetworking::SendNetMessage(NetMessage Message)
{
    //ensure that all Packets of the NetMessage bear the same DocumentNum so it can be reconstructed
	Message.SetDocumentNum(GenerateDocumentNum());

	for(std::list<RemoteMachine>::iterator it = remoteMachine.begin(); it != remoteMachine.end(); 
		++it)
	{
		//this RemoteMachine doesn't want any data at all until it receives the Heightmap
		if(it->receivedHeightmap)
			SendNetMessageToSingleRemoteMachine(Message, it->ipAddress, it->portToSendTo);
	}

    Message.DeallocatePackets();
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	BootRemoteMachine
/*!	\brief		Removes the machine identified by It from the machines we are currently in
				contact with.
**	\param		It:                 that identifies the machine to boot.  Will point to the next
                                    RemoteMachine in the list.
                KillRelatedSock:    Whether the socket created to listen to the remote machine
                                    identified by It should be destroyed
**	\return		true.
**	\exception	none
**	\par		DESCRIPTION:
**
**	\author		Nathan Frost
**	\date		2/13/2003 10:42:09 AM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
bool GameNetworking::BootRemoteMachine( std::list<RemoteMachine>::iterator &It,
                                        bool KillRelatedSock)
{
    //tell this guy he's booted
    NetMessage booted;
    booted.SetID(UDPNet::BOOTED);
	booted.SetDocumentNum(GenerateDocumentNum());
    udpNetworking.SendPacket(   sendSock, It->ipAddress, It->portToSendTo,
                                booted.GetPacket(0),
                                booted.GetPacketSize(booted.GetPacket(0)));
    booted.DeallocatePackets();

    /* Eradicate this guy from our memory */
    remoteSocksToWaitOn->RemoveSock(It->socketToReceiveFrom);
    if(KillRelatedSock)
        remoteSocksToWaitOn->KillSock(It->socketToReceiveFrom);
    //make sure It points to the next RemoteMachine as advertised
    It = remoteMachine.erase(It);

    return true;
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	Server::BootRemoteMachine
/*!	\brief		Removes the machine identified by IPAddress from the machines we are currently in
				contact with.
**	\param		IPAddress: that identifies the machine to boot.
**	\return		true if the specified machine was successfully booted, false if not.
**	\exception	none
**	\par		DESCRIPTION:
**
**	\author		Nathan Frost
**	\date		2/13/2003 10:42:09 AM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
bool Server::BootRemoteMachine(u_long IPAddress)
{
	for(std::list<RemoteMachine>::iterator it = remoteMachine.begin(); it != remoteMachine.end();
		++it)
	{
		if(it->ipAddress == IPAddress)
		    return GameNetworking::BootRemoteMachine(it);//is equivalent to "return true"
    }

    return false;//couldn't find the machine to boot
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	Client::BootRemoteMachine
/*!	\brief		Removes the machine identified by IPAddress from the machines we are currently in
				contact with.
**	\param		IPAddress: that identifies the machine to boot.
**	\return		true if the specified machine was successfully booted, false if not.
**	\exception	none
**	\par		DESCRIPTION:
**
**	\author		Nathan Frost
**	\date		2/13/2003 10:42:09 AM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
bool Client::BootRemoteMachine(u_long IPAddress)
{
	for(std::list<RemoteMachine>::iterator it = remoteMachine.begin(); it != remoteMachine.end();
		++it)
	{
		if(it->ipAddress == IPAddress)
		    return GameNetworking::BootRemoteMachine(it, false);//is equivalent to "return true" --
                                                                //the false param tells
                                                                //BootRemoteMachine not to destroy
                                                                //receiveSock
    }

    return false;//couldn't find the machine to boot
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	TurnOnPacketLossSimulation
/*!	\brief		Turns on Packet Loss Simulation, with the specified parameters.
**	\param		none
**	\return		true if Packet Loss Simulation was engaged, false if it was not
**	\exception	none
**	\par		DESCRIPTION:
**

**	\author		Nathan Frost
**	\date		2/7/2003 5:16:40 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
bool GameNetworking::TurnOnPacketLossSimulation(float PacketLossOccurrenceBase,
                                                float PacketLossOccurrenceDelta,
                                                float PacketLossLengthBase,
                                                float PacketLossLengthDelta)
{
    return remoteSocksToWaitOn->TurnOnPacketLossSimulation( PacketLossOccurrenceBase,
                                                            PacketLossOccurrenceDelta,
                                                            PacketLossLengthBase,
                                                            PacketLossLengthDelta);
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	TurnOffPacketLossSimulation
/*!	\brief		Turns off Packet Loss Simulation.
**	\param		none
**	\return		none
**	\exception	none
**	\par		DESCRIPTION:
**

**	\author		Nathan Frost
**	\date		2/7/2003 5:16:40 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
void GameNetworking::TurnOffPacketLossSimulation()
{
    remoteSocksToWaitOn->TurnOffPacketLossSimulation();
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	Server
/*!	\brief		Constructor
**	\param		none
**	\return		none
**	\exception	none
**	\par		DESCRIPTION:
**					
**	\author		Nathan Frost
**	\date		2/13/2003 11:24:19 AM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
Server::Server():
GameNetworking(), timeUntilServerUpdate(0.0)
{
	;
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	~Server()
/*!	\brief		Destructor
**	\param		none
**	\return		none
**	\exception	none
**	\par		DESCRIPTION:
**					
**	\author		Nathan Frost
**	\date		2/13/2003 11:28:58 AM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
Server::~Server()
{
	;
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	Server::DoNetworking
/*!	\brief		Should be called every frame
**	\param		none
**	\return		none
**	\exception	none
**	\par		DESCRIPTION:
**					
**	\author		Nathan Frost
**	\date		2/13/2003 12:24:11 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
void Server::DoNetworking()
{
	ProcessEvents();
	DetectRemoteMachineDelinquency();

	timeUntilServerUpdate -= static_cast<float>(timer.getRealTimeStep());
	if(timeUntilServerUpdate < 0 && remoteMachine.size())
	{
        UDPNet::NetMessage serverUpdate;
        serverUpdate.SetID(SERVER_UPDATE);
		MsgHandler::MessageHandler::Instance()->FillOutServerUpdate(serverUpdate);
        SendNetMessage(serverUpdate);                                 
        timeUntilServerUpdate = TIME_BETWEEN_SERVER_UPDATES;
	}
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	NotifyClientOfAcceptance
/*!	\brief		Let the accepted machine know that it has been accepted
**	\param		PortToReceiveFrom:	The Port that we, the Server, will use to listen to the Client
**	\param		IPAddress:	Of the remote machine (in Little Endian)
	\param		Port:		Where the remote machine will receive the data we send it
**	\return		none
**	\exception	none
**	\par		DESCRIPTION:
**					
**	\author		Nathan Frost
**	\date		3/17/2003 6:54:51 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
void Server::NotifyClientOfAcceptance(u_short PortToReceiveFrom, u_long IPAddress, u_short Port)
{
	UDPNet::NetMessage youAreAccepted;
	youAreAccepted.SetID(YOU_ARE_ACCEPTED);
	youAreAccepted.WriteNBytes(sizeof(u_short), reinterpret_cast<char *>(&PortToReceiveFrom));
	MsgHandler::MessageHandler::Instance()->FillOutHeightMapNetMessage(youAreAccepted);

	/*	send the heightmap over to ONLY the new Client -- and be sure to initialize and clean up
		after the NetMessage */
	youAreAccepted.SetDocumentNum(GenerateDocumentNum());
	SendNetMessageToSingleRemoteMachine(youAreAccepted, IPAddress, Port);
	youAreAccepted.DeallocatePackets();
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	Server::AcceptClient
/*!	\brief		Puts the specified machine on the list of machines that we're currently staying in
				contact with if the specified machine is a new machine.  Sends a CompleteNetMessage
				to the Machine to initialize it.
**	\param		IPAddress:	Of the remote machine (in Little Endian)
	\param		Port:		Where the remote machine will receive the data we send it
**	\return		The local port which the remote machine should send data to if the specified
                machine was a new machine that was successfully added to the list of remote
                machines we are staying in contact with, or 0 if the specified machine was not a
                new machine and, hence, was not added to the list.
**	\exception	none
**	\par		DESCRIPTION:
**					
**	\author		Nathan Frost
**	\date		2/13/2003 1:05:46 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
u_short Server::AcceptClient(u_long IPAddress, u_short Port)
{
    //ensure that the machine we are about to accept has not already been accepted
    for(std::list<RemoteMachine>::iterator it = remoteMachine.begin(); it != remoteMachine.end();
        ++it)
    {
        if(it->ipAddress == IPAddress)
        {
            //we have already accepted this guy, but he clearly doesn't know that yet
			NotifyClientOfAcceptance(it->portToReceiveFrom, IPAddress, Port);
            return 0;//no duplicate machines, thank you very much
        }
    }

    //otherwise, we have accepted a new machine
	SOCKET socketToReceiveFrom;
	u_short localPort = ANY_PORT;
	udpNetworking.CreateSock(socketToReceiveFrom, localPort);
	remoteSocksToWaitOn->AddSock(socketToReceiveFrom);

	remoteMachine.push_back(RemoteMachine(IPAddress, Port, socketToReceiveFrom, localPort));
	NotifySelf();//ensure the thread is responsive to the new machine's messages right away

    /* Let the accepted machine know that it has been accepted */
    NotifyClientOfAcceptance(localPort, IPAddress, Port);

	return localPort;//client was successfully accepted
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	Server::Connected
/*!	\brief		Trivially returns true -- the server is connected to itself.
**	\param		none
**	\return		true
**	\exception	none
**	\par		DESCRIPTION:
**					
**	\author		Nathan
**	\date		2/13/2003 3:19:42 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
bool Server::Connected()
{
	return true;
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	Server::SendEventToMessageHandler
/*!	\brief		Sends an Event appropriately to the MessageHandler.
**	\param		TheEvent: to send.
**	\return		none
**	\exception	none
**	\par		DESCRIPTION:
**
**	\author		Nathan Frost
**	\date		2/12/2003 1:10:33 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
void Server::SendEventToMessageHandler(Event &TheEvent)
{
	switch(TheEvent.netMessage.GetID())
    {
        case UDPNet::DEFAULT:
        {
            break;
        }
        case UDPNet::HELLO:
        {
            if(AcceptClient(TheEvent.senderIP, LISTEN_PORT))
				MsgHandler::MessageHandler::Instance()->WeAcceptedClient(TheEvent.senderIP);
			
			/* Find the machine that we are accepting and reset its timeSinceLastHello */
            std::list<RemoteMachine>::iterator it = remoteMachine.begin();
	        for(; it != remoteMachine.end(); ++it)
            {
                if(it->ipAddress == TheEvent.senderIP)
                    break;
            }
			if(it != remoteMachine.end())
				it->timeSinceLastHello = 0.0;
			
			//BEG_HACK
			Debugging::Debugger::Instance()->output(Debugging::NET, Debugging::L4, 
													"Responded to HELLO.");
			//END_HACK
			break;
        }
        case UDPNet::BOOTED:
        {
            /* Find the machine that BOOTED us */
            std::list<RemoteMachine>::iterator it = remoteMachine.begin();
	        for(; it != remoteMachine.end(); ++it)
            {
                if(it->ipAddress == TheEvent.senderIP)
                    break;
            }
            if(it != remoteMachine.end())
            {
				MsgHandler::MessageHandler::Instance()->WeBootedClient(it->ipAddress);
                GameNetworking::BootRemoteMachine(it);
            }
            break;
        }
		case SWIM_MODE_GOTO_POINT:
		{
			MsgHandler::MessageHandler::Instance()->SwimModeGotoPoint(TheEvent.netMessage);
			break;
		}
		case ADD_RANDOM_FISH:
		{	
			MsgHandler::MessageHandler::Instance()->AddRandomFish();
			break;
		}
		case FISH_SELECT_LIST_CHANGED:
		{
			MsgHandler::MessageHandler::Instance()->FishSelectListChanged(TheEvent.netMessage);
			break;
		}
		case FISH_SELECT_REQUEST:
		{
			MsgHandler::MessageHandler::Instance()->ClientSelectRequest(TheEvent.netMessage, TheEvent.senderIP);
			break;
		}
		case FISH_DESELECT_REQUEST:
		{
			MsgHandler::MessageHandler::Instance()->ClientDeselectRequest(TheEvent.netMessage, TheEvent.senderIP);
			break;
		}
		case FISH_SUPPRESS_LIST_CHANGED:
		{
			MsgHandler::MessageHandler::Instance()->FishSuppressListChanged(TheEvent.netMessage);
			break;
		}
		case FISH_SUPPRESS_REQUEST:
		{
			MsgHandler::MessageHandler::Instance()->ClientSuppressRequest(TheEvent.netMessage);
			break;
		}
		case FISH_UNSUPPRESS_REQUEST:
		{
			MsgHandler::MessageHandler::Instance()->ClientUnsuppressRequest(TheEvent.netMessage);
			break;
		}
		case ADD_FISH_TO_MATING_LIST:
		{
			MsgHandler::MessageHandler::Instance()->ClientAddToMatingListRequest(TheEvent.netMessage);
			break;
		}
		case ISSUE_COMMAND:
		{
			MsgHandler::MessageHandler::Instance()->ClientIssueCommandRequest(TheEvent.netMessage);
			break;
		}
        default:
        {
            break;
        }
    }
    TheEvent.netMessage.DeallocatePackets();
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	Server::GetNumConnectedClients
/*!	\brief		Returns the number of Clients we, the Server, are currently in contact with
**	\param		none
**	\return		The number of Clients we, the Server, are currently in contact with
**	\exception	none
**	\par		DESCRIPTION:
**					
**	\author		Nathan Frost
**	\date		3/17/2003 4:58:53 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
unsigned int Server::GetNumConnectedClients()
{
	return remoteMachine.size();
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	Client::GetNumConnectedClients
/*!	\brief		Returns whether or not we are Connected().
**	\param		none
**	\return		1 if we are Connected(), 0 if not.
**	\exception	none
**	\par		DESCRIPTION:
**					
**	\author		Nathan Frost
**	\date		3/17/2003 4:58:53 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
unsigned int Client::GetNumConnectedClients()
{
	if(Connected())
		return 1;
	else
		return 0;
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	Client
/*!	\brief		Constructor; gets Client ready to try to connect to the Server
**	\param		IPAddress:	Of the Server (in Little Endian)
**	\return		none
**	\exception	If the Client is instantiated with IPAddress equal to the local machine's IP
				Address, a Nexception is thrown, as a local machine cannot be its own client
**	\par		DESCRIPTION:
**					
**	\author		Nathan Frost
**	\date		2/13/2003 3:23:43 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
Client::Client(u_long IPAddress):
GameNetworking(), connected(false), timeUntilKeepAlive(0)
{
	if(IPAddress == udpNetworking.MyIPAddress())
		throw Nexception("Client cannot be instantiated with its own IPAddress.", __FILE__, __LINE__, __DATE__, __TIME__);
	else
		remoteMachine.push_back(RemoteMachine(IPAddress, LISTEN_PORT, receiveSock, LISTEN_PORT));
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	~Client
/*!	\brief		Destructor
**	\param		none
**	\return		none
**	\exception	none
**	\par		DESCRIPTION:
**
**	\author		Nathan Frost
**	\date		2/13/2003 3:28:43 PM
**	\warning	
				- there should only be one RemoteMachine on the Client's list of RemoteMachines --
				the Server.  If there is ever more than one RemoteMachine on the Client's list of
				RemoteMachines, then the networking engine will crash
*/
////////////////////////////////////////////////////////////////////////////////
Client::~Client()
{
	assert(remoteMachine.size() == 1 || remoteMachine.size() == 0);
    /* If we're connected to the Server, let him know we're quitting on him */
    if(remoteMachine.size())
    {
        //need to copy the temporary returned by begin() to a local variable so that the local
        //variable can be passed by non_const reference
        std::list<RemoteMachine>::iterator it = remoteMachine.begin();
	    GameNetworking::BootRemoteMachine(it, false);   //false param says not to destroy the
                                                        //receiveSock -- we'll do that in
                                                        //~GameNetworking()
    }
	//remove the (at most one) RemoteMachine on the list of remoteMachines so that
	//~GameNetworking() only DestroySock()'s receiveSock once.
	remoteMachine.clear();
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	Client::DoNetworking
/*!	\brief		Should be called every frame.
**	\param		none
**	\return		none
**	\exception	none
**	\par		DESCRIPTION:
**					
**	\author		Nathan Frost
**	\date		2/13/2003 3:29:04 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
void Client::DoNetworking()
{
	ProcessEvents();
	
	timeUntilKeepAlive -= static_cast<float>(timer.getRealTimeStep());
	if(!connected)
	{
		if(timeUntilKeepAlive < 0)
		{
			NetMessage hello;
			hello.SetID(HELLO);
			SendNetMessage(hello);
			timeUntilKeepAlive = TIME_BETWEEN_CLIENT_HELLOS;
		}
	}
	else
    {
        if(timeUntilKeepAlive < 0)
        {
            NetMessage keepAlive;
            keepAlive.SetID(KEEP_ALIVE);
            SendNetMessage(keepAlive);                              
            timeUntilKeepAlive = TIME_BETWEEN_CLIENT_KEEP_ALIVES;
        }
        if(DetectRemoteMachineDelinquency())
        {
            connected = false;//our one and only Server was booted for delinquency!
            MsgHandler::MessageHandler::Instance()->OurServerIsDelinquent();
        }
    }
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	Client::Connected
/*!	\brief		Informs the caller whether or not this Client has successfully connected with a 
				Server.
**	\param		none
**	\return		true if the Client has connected to a Server, false if it has not.
**	\exception	none
**	\par		DESCRIPTION:
**					
**	\author		Nathan Frost
**	\date		2/13/2003 3:32:22 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
bool Client::Connected()
{
	return connected;
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	Client::AcceptClient
/*!	\brief		Trivially returns ANY_PORT -- the Client can never accept other Clients
**	\param		IPAddress:  Irrelevant
	\param		Port:       Irrelevant
**	\return		ANY_PORT
**	\exception	none
**	\par		DESCRIPTION:
**					
**	\author		Nathan Frost
**	\date		2/13/2003 3:29:30 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
u_short Client::AcceptClient(u_long IPAddress, u_short Port)
{
	return ANY_PORT;
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	Client::SendEventToMessageHandler
/*!	\brief		Sends an Event appropriately to the MessageHandler.
**	\param		TheEvent: to send.
**	\return		none
**	\exception	none
**	\par		DESCRIPTION:
**
**	\author		Nathan Frost
**	\date		2/12/2003 1:10:33 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
void Client::SendEventToMessageHandler(Event &TheEvent)
{
	switch(TheEvent.netMessage.GetID())
    {
        case DEFAULT:
        {
            break;
        }
        case YOU_ARE_ACCEPTED:
        {
			Debugging::Debugger::Instance()->output(Debugging::NET, Debugging::L4, 
													"I got accepted.");
            if(!connected)
            {
                connected = true;
                //change Port information for the one and only RemoteMachine
                u_short portToSendTo =
                        *reinterpret_cast<const u_short *>(TheEvent.netMessage.ReadNBytes(2));
                remoteMachine.begin()->portToSendTo = portToSendTo;

				Debugging::Debugger::Instance()->output(Debugging::NET, Debugging::L4, 
														"I got accepted.");
                MsgHandler::MessageHandler::Instance()->WeAreConnected(TheEvent.netMessage);
            }
            else
            {
                TheEvent.netMessage.DeallocatePackets();
                return;
            }

            break;
        }
        case BOOTED:
        {
			MsgHandler::MessageHandler::Instance()->WeAreBooted();
            connected = false;//our one and only Server booted us!
            remoteMachine.clear();//Client's usefulness has ended -- don't talk to Server anymore
            break;
        }
		case SERVER_UPDATE:
		{
			MsgHandler::MessageHandler::Instance()->GotServerUpdate(TheEvent.netMessage);
			break;
		}
		case SET_FISH_BODY_ACTIONS:
		{
			MsgHandler::MessageHandler::Instance()->SetFishBodyActions(TheEvent.netMessage);
			break;
		}
		case SWIM_MODE_GOTO_POINT:
		{
			MsgHandler::MessageHandler::Instance()->SwimModeGotoPoint(TheEvent.netMessage);
			break;
		}
		case FISH_SELECT_LIST_CHANGED:
		{
			MsgHandler::MessageHandler::Instance()->FishSelectListChanged(TheEvent.netMessage);
			break;
		}
		case FISH_SUPPRESS_LIST_CHANGED:
		{
			MsgHandler::MessageHandler::Instance()->FishSuppressListChanged(TheEvent.netMessage);
			break;
		}
		case NEW_GAME_SPEED:
		{
			MsgHandler::MessageHandler::Instance()->NewGameSpeed(TheEvent.netMessage);
			break;
		}
        default:
        {
            break;
        }
    }
    TheEvent.netMessage.DeallocatePackets();
}



#if 0//debug crap code
void GameNetworking::Debug(int Case)
{
	NetMessage expiryMessage;
	int bigSize = MAX_PACKET_SIZE - PACKET_HEADER_SIZE - 1;

	switch(Case)
	{
	case 10:
		{
			u_short portyo, portyeah;
			portyo = portyeah = ANY_PORT;
			SOCKET sockyeah, sockyo;
			udpNetworking.CreateSock(sockyeah, portyeah);
			udpNetworking.CreateSock(sockyo, portyo);
			remoteMachine.push_back(RemoteMachine(0x01231032, portyeah, sockyeah));
			//remoteMachine.push_back(RemoteMachine(udpNetworking.MyIPAddress(), portyo, sockyo));
			break;
		}
	case 11:
		{
			NetMessage netMessage;
			char firstMessage[] = "First message";
			netMessage.WriteNBytes(	strlen(firstMessage) + 1,
									reinterpret_cast<unsigned char *>(&firstMessage[0]));
			LockMutexEvent();
			event.push_back(Event(netMessage, udpNetworking.MyIPAddress()));
			UnlockMutexEvent();
			break;
		}
	case 12:
		{
			NetMessage netMessage;
			char *big = new char[bigSize];
			netMessage.WriteNBytes(	bigSize,
									reinterpret_cast<unsigned char *>(&big[0]));
			char *msg = "That should be two Packets, yo.";
			netMessage.WriteNBytes(	strlen(msg) + 1,
									reinterpret_cast<unsigned char *>(&msg[0]));
			
			LockMutexEvent();
			event.push_back(Event(netMessage, udpNetworking.MyIPAddress()));
			UnlockMutexEvent();
			break;
		}
	case 13:
		{
			NetMessage netMessage;
			char *big = new char[bigSize];
			netMessage.WriteNBytes(	bigSize,
									reinterpret_cast<unsigned char *>(&big[0]));
			netMessage.WriteNBytes(	bigSize,
									reinterpret_cast<unsigned char *>(&big[0]));
			char *msg = "That should be three Packets, yo.";
			netMessage.WriteNBytes(	strlen(msg) + 1,
									reinterpret_cast<unsigned char *>(&msg[0]));
			
			LockMutexEvent();
			event.push_back(Event(netMessage, udpNetworking.MyIPAddress()));
			UnlockMutexEvent();
			break;
		}
	case 14:
		{
			char *big = new char[bigSize];
			expiryMessage.WriteNBytes(	bigSize,
										reinterpret_cast<unsigned char *>(&big[0]));
			expiryMessage.WriteNBytes(	bigSize,
									reinterpret_cast<unsigned char *>(&big[0]));
			char *msg = "That should be three Packets, yo.";
			expiryMessage.WriteNBytes(	strlen(msg) + 1,
										reinterpret_cast<unsigned char *>(&msg[0]));

			NetMessage netMessage;
			netMessage.ConcatenatePacket(expiryMessage.GetPacket(1));
			
			LockMutexEvent();
			event.push_back(Event(netMessage, udpNetworking.MyIPAddress()));
			UnlockMutexEvent();
			break;
		}
	case 15:
		{
			NetMessage anotherExpiryMessage;
			char *big = new char[bigSize];
			anotherExpiryMessage.WriteNBytes(	bigSize,
												reinterpret_cast<unsigned char *>(&big[0]));
			anotherExpiryMessage.WriteNBytes(	bigSize,
												reinterpret_cast<unsigned char *>(&big[0]));
			char *msg = "That should be another three Packets, yo.";
			anotherExpiryMessage.WriteNBytes(	strlen(msg) + 1,
												reinterpret_cast<unsigned char *>(&msg[0]));

			NetMessage netMessage;
			netMessage.ConcatenatePacket(anotherExpiryMessage.GetPacket(1));
			netMessage.ConcatenatePacket(anotherExpiryMessage.GetPacket(0));
			
			LockMutexEvent();
			event.push_back(Event(netMessage, udpNetworking.MyIPAddress()));
			UnlockMutexEvent();

			break;
		}
	case 16:
		{
			NetMessage netMessage;
			char *big = new char[bigSize];
			netMessage.WriteNBytes(	bigSize,
									reinterpret_cast<unsigned char *>(&big[0]));
			netMessage.WriteNBytes(	bigSize,
									reinterpret_cast<unsigned char *>(&big[0]));
			char *msg = "That should be three Packets, yo.";
			netMessage.WriteNBytes(	strlen(msg) + 1,
									reinterpret_cast<unsigned char *>(&msg[0]));
			
			SendNetMessage(netMessage);
			break;
		}
	case 17:
/*		{
			BootRemoteMachine(0x00000012);
			BootRemoteMachine(udpNetworking.MyIPAddress());
			BootRemoteMachine(0x00000012);
			BootRemoteMachine(0x01231032);
			BootRemoteMachine(0x00000012);
			break;
		}*/
	default:
		{
			break;
		}
	}
}
#endif