////////////////////////////////////////////////////////////////////////////////
/*!	\file		GameNetworking
**	\brief		High level game networking logic
**	\author		Nathan Frost
**	\date		2/11/2003 9:13:56 PM
**	\par		DESCRIPTION:
**					
**	\par		REMARKS:
**
*/
////////////////////////////////////////////////////////////////////////////////
#ifndef GAME_NETWORKING_H
#define GAME_NETWORKING_H
#pragma once


#include <deque>
#include <list>
#include "WinSockSynthesis.h"
#include "Netdata.h"


//forward declarations
namespace MsgHandler
{
	class MessageHandler;
}
class TimeObject;
namespace UDPNet
{
    class RemoteSocksToWaitOn;
}



namespace UDPNet
{
    //The port number Clients send to when trying to connect to a Server -- for TCP/IP any port from
    //1024 to 5000 should be reserved by the OS for use by applications -- hence if another
    //application isn't using the following port number, we're OK
    const u_short LISTEN_PORT = 2000;


	////////////////////////////////////////////////////////////////////////////////
	/*!	\class		RemoteMachine
	**	\brief		Encapsulates all the information necessary to refer to a remotely connected
					game.
	**	\author		Nathan Frost
	**	\date		2/12/2003 10:03:43 AM
	**	\warning	none
	*/
	////////////////////////////////////////////////////////////////////////////////
	struct RemoteMachine
	{
		u_long ipAddress;			///<machine resides at this IPAddress (stored in Little Endian)
		u_short portToSendTo;       ///<data sent to this machine should be sent to this port
		SOCKET socketToReceiveFrom;	///<dedicated to receiving data from this machine
        u_short portToReceiveFrom;  ///<port socketToReceiveFrom resides at
		float timeSinceLastUpdate;	///<number of seconds since this machine spoke with us
		float timeSinceLastHello;	///<number of seconds since this machine sent us a HELLO message
		bool receivedHeightmap;		///<whether or not this Client has received the Heightmap


		RemoteMachine(  u_long IPAddress, u_short PortToSendTo, SOCKET SocketToReceiveFrom,
                        u_short PortToReceiveFrom):
		ipAddress(IPAddress), portToSendTo(PortToSendTo), socketToReceiveFrom(SocketToReceiveFrom),
        portToReceiveFrom(PortToReceiveFrom), timeSinceLastUpdate(0.0), timeSinceLastHello(0.0), 
		receivedHeightmap(false){}
	};

	////////////////////////////////////////////////////////////////////////////////
	/*!	\class		GameNetworking
	**	\brief		Encapsulates game networking functionality common to both Server and Clients.
	**	\author		Nathan Frost
	**	\date		2/12/2003 8:45:39 AM
	**	\warning	none
	*/
	////////////////////////////////////////////////////////////////////////////////
	class GameNetworking
	{
	private:
		u_short receivePort;		///<port at which data is received from remote machines
		std::deque<Event> event;	///<received Events waiting to be sent to the MessageHandler
		unsigned int curDocumentNum;///<DocumentNum that the next outgoing NetMessage will have


		virtual void SendEventToMessageHandler(Event &TheEvent) = 0;
	protected:
		TimeObject &timer;					///<Allows GameNetworking to track the passage of time
		RemoteSocksToWaitOn *remoteSocksToWaitOn;///<asynchronously gathers Events from remote data
		SOCKET receiveSock;							///<receives data from remote machines
		SOCKET sendSock;							///<the only vector to send data fro
		UDPNetworking &udpNetworking;	///<Low level networking functionality will be leveraged
		std::list<RemoteMachine> remoteMachine;		///<Machines we're staying in contact with

		
		unsigned int GenerateDocumentNum();
		void ProcessEvents();
		bool DetectRemoteMachineDelinquency();
		void SendNetMessageToSingleRemoteMachine(	NetMessage &Message, u_long IPAddress, 
													u_short PortToSendTo);
		bool BootRemoteMachine( std::list<RemoteMachine>::iterator &It,
                                bool KillRelatedSock = true);
		GameNetworking();
	public:
        bool TurnOnPacketLossSimulation(float PacketLossOccurrenceBase,
                                        float PacketLossOccurrenceDelta,
                                        float PacketLossLengthBase, float PacketLossLengthDelta);
        void TurnOffPacketLossSimulation();
		void SendNetMessage(NetMessage Message);
		void NotifySelf();
        virtual bool BootRemoteMachine(u_long IPAddress) = 0;
		bool BootAllRemoteMachines();
		virtual ~GameNetworking();


		virtual void DoNetworking() = 0;
		virtual u_short AcceptClient(u_long IPAddress, u_short Port) = 0;
		virtual bool Connected() = 0;
		virtual unsigned int GetNumConnectedClients() = 0;


		void Debug(int Case);
	};

	////////////////////////////////////////////////////////////////////////////////
	/*!	\class		Server
	**	\brief		Should be instantiated whenever the local game is to act in a Server capacity,
					and destroyed whenever the local game should not behave like a Server.
	**	\author		Nathan Frost
	**	\date		2/12/2003 10:17:53 AM
	**	\warning	none
	*/
	////////////////////////////////////////////////////////////////////////////////
	class Server:public GameNetworking
	{
	private:
		float timeUntilServerUpdate;        ///<seconds until it's time to send a server update
        //TODO: std::list<RemoteMachine> banList;   ///<list of Machines that we will ignore


        void SendEventToMessageHandler(Event &TheEvent);
		void NotifyClientOfAcceptance(u_short PortToReceiveFrom, u_long IPAddress, u_short Port);
	public:
		bool Connected();
		unsigned int GetNumConnectedClients();
		u_short AcceptClient(u_long IPAddress, u_short Port);
        bool BootRemoteMachine(u_long IPAddress);
		void DoNetworking();
		Server();
		~Server();
	};

	////////////////////////////////////////////////////////////////////////////////
	/*!	\class		Client
	**	\brief		Should be instantiated whenever the local game is to act in a Client capacity,
					and destroyed whenever the local game should not behave like a Client.
	**	\author		Nathan Frost
	**	\date		2/13/2003 3:14:58 PM
	**	\warning
                    - It should be noted that if an invalid IPAddress and/or Port is passed to
                    Client's constructor, Nexceptions will be thrown that contain the a string
                    instructing the user to "restart GameNetworking".  If this Nexception is
                    caught, the user then has the opportunity to destroy the Client instance and
                    re-instantiate it by passing a valid IPAddress and Port.  
	*/
	////////////////////////////////////////////////////////////////////////////////
	class Client:public GameNetworking
	{
	private:
		bool connected;             ///<whether or not the Client is connected to a Server
        float timeUntilKeepAlive;   ///<seconds until it's time to send a client keepalive


        void SendEventToMessageHandler(Event &TheEvent);
	public:
		bool Connected();
		unsigned int GetNumConnectedClients();
		u_short AcceptClient(u_long IPAddress, u_short Port);
        bool BootRemoteMachine(u_long IPAddress);
		void DoNetworking();
		Client(u_long IPAddress);
		~Client();
	};
}

#endif