////////////////////////////////////////////////////////////////////////////////
/*!	\file		NetData.cpp
**	\brief		Implementation
**	\author		Nathan Frost
**	\date		1/14/2003 9:13:35 AM
**	\par		DESCRIPTION:
**					
**	\par		REMARKS:
**					
*/
////////////////////////////////////////////////////////////////////////////////
#pragma warning(disable:4786)
#include "NetData.h"
#include "NetMessageID.h"
#include "NodeMgr.h"
#include "../nexception/Nexception.h"
#include <assert.h>
//#include "GUITester.h"//TODO:HACK
// memory leak detection
#ifdef _DEBUG
    #include"memleak.h"
    #define new DEBUG_NEW
    #undef THIS_FILE
    static char THIS_FILE[] = __FILE__;
#endif
using namespace UDPNet;


//constant data
const unsigned int UDPNet::PACKET_HEADER_SIZE = 11;///<in bytes
const unsigned int UDPNet::MAX_PACKET_PAYLOAD_SIZE =
                                UDPNet::MAX_PACKET_SIZE - UDPNet::PACKET_HEADER_SIZE;///<in bytes


//BEG_HACK: memory allocation
int numPacketsAllocated = 0;
//END_HACK: memory allocation



////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	NetData
/*!	\brief		Constructor
**	\param		none
**	\return		none
**	\exception	none
**	\par		DESCRIPTION:
					- curPacket is intialized to -1 because the NetMessage begins with no Packets,
					and as soon as WriteNBytes() is called, the first (zeroth) Packet is added to
					the NetMessage by calling NewPacket() -- which increments curPacket to 0, which
					is the correct Packet to begin writing to
					- The NetMessage begins empty of all data (and is therefore initialized to
                    write mode) and will create Packets with the NULL PacketID
**					
**	\author		Nathan Frost
**	\date		1/10/2003 7:16:02 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
NetMessage::NetMessage():
curPacket(-1), curByte(0), totalPages(0), bytesRemaining(0), readMode(false), packetID(0)
{}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	NewPacket
/*!	\brief		Internal helper function that allocates another Packet to this NetMessage
**	\param		none
**	\return		none
**	\exception	none
**	\par		DESCRIPTION:
					- Updates NetMessage's internal datastructures such that all subsequent calls
					to WriteNBytes() will write data to the newly created Packet
**					
**	\author		Nathan Frost
**	\date		1/10/2003 9:07:57 PM
**	\warning	
				- Should only be called when the NetMessage is in write mode (!readMode == true)
*/
////////////////////////////////////////////////////////////////////////////////
void NetMessage::NewPacket()
{
    //BEG_HACK: memory allocation
    ++numPacketsAllocated;
    //END_HACK: memory allocation

	Packet newPacket = new char[MAX_PACKET_SIZE];//TODO: Memory management
	assert(newPacket);

	/* Fill out Packet header information */
	//PageNum of newPacket
	newPacket[4] = totalPages;

	//the initial size, in bytes, of newPacket
	SetPacketSize(newPacket, PACKET_HEADER_SIZE);

	//add newPacket to the NetMessage
	pages.insert(std::pair<unsigned int, Packet>(totalPages, newPacket));

	//newPacket's PacketID
	SetID(packetID);
	
	/* Let each Packet know the total number of Packets in the NetMessage */
	for(std::map<unsigned int, Packet>::iterator it = pages.begin(); it != pages.end(); ++it)
		it->second[5] = totalPages + 1;//'+ 1' is required since we haven't yet updated totalPages

	/* Update internal data to reflect the added Packet */
	++totalPages;	//we just put a new Packet in the NetMessage
	++curPacket;	//we're now writing to the Packet with this PageNum (the newest Packet)
	curByte = PACKET_HEADER_SIZE;	//we're now writing to this byte of the newest Packet
	//We have gained this much extra space in the NetMessage
	bytesRemaining = MAX_PACKET_SIZE - PACKET_HEADER_SIZE;
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	WriteNBytes
/*!	\brief		Writes N bytes of data to this NetMessage.
**	\param		N:		The number of bytes to write
	\param		Data:	Pointer to the data to write
**	\return		true if Data was successfully written, false if it was not.
**	\exception	none
**	\par		DESCRIPTION:
**					- Writing large chunks of data (MAX_PACKET_PAYLOAD_SIZE is the biggest that can
                    occur in one call) is the most efficient use of WriteNBytes
**	\author		Nathan Frost
**	\date		1/10/2003 7:31:29 PM
**	\warning	
				- Fails if this NetMessage is in readMode
				- Fails if (N > MAX_PACKET_SIZE - PACKET_HEADER_SIZE) -- this limitation could be
				overcome with a little more programming time and effort.
                - When reconstructing a NetMessage, use ConcatenatePacket(), not WriteNBytes()!
*/
////////////////////////////////////////////////////////////////////////////////
bool NetMessage::WriteNBytes(unsigned int N, const void *Data)
{
	//not allowed to Write after Read Mode has been entered or Data is too large to be sent
	if(readMode || N > MAX_PACKET_SIZE - PACKET_HEADER_SIZE)
		return false;
	
	//ensure there's enough space in the NetMessage for the new data
	if(static_cast<signed int>(bytesRemaining - N) < 0)//casted only to avoid compiler warning
		NewPacket();
	
	//insert Data into the current Packet
	Packet packet = pages[curPacket];
	assert(packet != 0);
	memcpy(reinterpret_cast<void *>(packet + curByte), Data, N);

	/* Update the size, in bytes, of the Packet */
	unsigned int packetSize = GetPacketSize(pages[curPacket]);
	packetSize += N;
	SetPacketSize(packet, packetSize);

	//reflect the fact that we've written some bytes
	curByte += N;
	bytesRemaining -= N;

	return true;
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	FillTheVoid
/*!	\brief		Helper function ensures there is at least one (payload-lacking) Packet in the
                NetMessage
**	\param		none
**	\return		none
**	\exception	none
**	\par		DESCRIPTION:
**                  -   this function is important if one wishes to be able to SetID(), for example,
                        before calling WriteNBytes().
**	\author		Nathan Frost
**	\date		1/11/2003 4:57:41 PM
**	\warning    none
*/
////////////////////////////////////////////////////////////////////////////////
void NetMessage::FillTheVoid()
{
    if(pages.empty())
        NewPacket();
}


////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	GetPacketSize
/*!	\brief		Helper function that returns the total size, in bytes, of ThePacket
**	\param		ThePacket: has (return value) many bytes of data
**	\return		The total number of bytes, including header information, that's stored in ThePacket
**	\exception	none
**	\par		DESCRIPTION:
**
**	\author		Nathan Frost
**	\date		1/11/2003 4:57:41 PM
**	\warning    none
*/
////////////////////////////////////////////////////////////////////////////////
unsigned int NetMessage::GetPacketSize(Packet ThePacket)
{
	unsigned int ret = 0;
	//extract size
	reinterpret_cast<unsigned char *>(&ret)[0] = ThePacket[6];
	reinterpret_cast<unsigned char *>(&ret)[1] = ThePacket[7];
	reinterpret_cast<unsigned char *>(&ret)[2] = ThePacket[8];

	return ret;
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	GetPacketDocumentNum
/*!	\brief		Returns ThePacket's DocumentNum field
**	\param		ThePacket: Retrieve this Packet's DocumentNum field
**	\return		ThePacket's DocumentNum field
**	\exception	none
**	\par		DESCRIPTION:
**					
**	\author		Nathan Frost
**	\date		1/13/2003 8:28:07 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
unsigned int NetMessage::GetPacketDocumentNum(Packet ThePacket)
{
	unsigned int ret = 0;
	//extract DocumentNum
	reinterpret_cast<unsigned char *>(&ret)[0] = ThePacket[0];
	reinterpret_cast<unsigned char *>(&ret)[1] = ThePacket[1];
	reinterpret_cast<unsigned char *>(&ret)[2] = ThePacket[2];
	reinterpret_cast<unsigned char *>(&ret)[3] = ThePacket[3];

	return ret;
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	GetNetMessageDocumentNum
/*!	\brief		Returns the DocumentNum that NetMessages' Packet(s) bear
**	\param		none
**	\return		NetMessage's Packet's or Packets' DocumentNum field
**	\exception	none
**	\par		DESCRIPTION:
**					
**	\author		Nathan Frost
**	\date		1/13/2003 8:28:07 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
unsigned int NetMessage::GetNetMessageDocumentNum()
{
	if(pages.size())
		return GetPacketDocumentNum(pages.begin()->second);
	else
		throw Nexception("GetNetMessageDocumentNum() was called on a NetMessage with no Packets.",
							__FILE__, __LINE__, __DATE__, __TIME__);
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	SetDocumentNum
/*!	\brief		Sets the DocumentNum of all Packets to DocumentNum
**	\param		DocumentNum: The integer to set all Packets' DocumentNum field to
**	\return		none
**	\exception	none
**	\par		DESCRIPTION:
**					
**	\author		Nathan Frost
**	\date		1/27/2003 9:11:28 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
void NetMessage::SetDocumentNum(unsigned int DocumentNum)
{
    FillTheVoid();//ensure that there is at least one Packet in *this to accept DocumentNum
	for(std::map<unsigned int, Packet>::iterator it = pages.begin(); it != pages.end(); ++it)
	{
		//Packet's DocumentNum
		it->second[0] = reinterpret_cast<unsigned char *>(&DocumentNum)[0];
		it->second[1] = reinterpret_cast<unsigned char *>(&DocumentNum)[1];
		it->second[2] = reinterpret_cast<unsigned char *>(&DocumentNum)[2];
		it->second[3] = reinterpret_cast<unsigned char *>(&DocumentNum)[3];
	}
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	GetPageNum
/*!	\brief		Returns ThePacket's PageNum field
**	\param		ThePacket: Retrieve this Packet's PageNum field
**	\return		ThePacket's PageNum field
**	\exception	none
**	\par		DESCRIPTION:
**					
**	\author		Nathan Frost
**	\date		1/13/2003 7:21:06 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
unsigned char NetMessage::GetPageNum(Packet ThePacket)
{
	return ThePacket[4];
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	GetTotalPages
/*!	\brief		Returns ThePacket's TotalPages field
**	\param		ThePacket: Retrieve this Packet's TotalPages field
**	\return		ThePacket's TotalPages field
**	\exception	none
**	\par		DESCRIPTION:
**					
**	\author		Nathan Frost
**	\date		1/13/2003 7:21:06 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
unsigned char NetMessage::GetTotalPages(Packet ThePacket)
{
	return ThePacket[5];
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	SetPacketSize
/*!	\brief		Sets the Size field of a Packet to Size.
**	\param		ThePacket:	The Packet whose Size field you wish to modify.
	\param		Size:		The 24-bit value ThePacket's Size field should be set to.
**	\return		none
**	\exception	none
**	\par		DESCRIPTION:
**					
**	\author		Nathan Frost
**	\date		1/13/2003 6:29:26 PM
**	\warning
                - respect the 24-bit size of the Size parameter
*/
////////////////////////////////////////////////////////////////////////////////
void NetMessage::SetPacketSize(Packet ThePacket, unsigned int Size)
{
	ThePacket[6] = reinterpret_cast<const unsigned char *>(&Size)[0];
	ThePacket[7] = reinterpret_cast<const unsigned char *>(&Size)[1];
	ThePacket[8] = reinterpret_cast<const unsigned char *>(&Size)[2];
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	GetNumCurrentPackets
/*!	\brief		Returns the number of Packets currently present in the NetMessage
**	\param		none
**	\return		The current number of Packets in the NetMessage
**	\exception	none
**	\par		DESCRIPTION:
**					
**	\author		Nathan Frost
**	\date		2/12/2003 1:04:04 PM
**	\warning
                - if this NetMessage is being reconstructed, remember that GetNumCurrentPackets()
                may be less than GetNumPackets() -- which would indicate that the NetMessage is not
                completely reconstructed
*/
////////////////////////////////////////////////////////////////////////////////
unsigned int NetMessage::GetNumCurrentPackets()
{
	return pages.size();
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	GetNumPackets
/*!	\brief		Returns the number of Packets that should appear in the complete NetMessage
**	\param		none
**	\return	    The number of Packets in the NetMessage
**	\exception	none
**	\par		DESCRIPTION:
**					
**	\author		Nathan Frost
**	\date		1/11/2003 1:43:26 PM
**	\warning	- if the NetMessage has no Packets at all then this function returns 
				- if this NetMessage is being reconstructed, remember that GetNumPackets()
                may be greater than GetNumCurrentPackets() -- which would indicate that the
                NetMessage is not completely reconstructed
*/
////////////////////////////////////////////////////////////////////////////////
unsigned int NetMessage::GetNumPackets()
{
	if(!pages.size())
		return 0;
	else
		return static_cast<unsigned char>(pages.begin()->second[5]);
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	SetID
/*!	\brief		Sets all the NetMessage's Packets' PacketIDs to ID
**	\param		ID:	to set all the NetMessage's Packets' PacketIDs to
**	\return		none
**	\exception	none
**	\par		DESCRIPTION:
**					
**	\author		Nathan Frost
**	\date		1/11/2003 1:57:35 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
void NetMessage::SetID(unsigned short ID)
{
    packetID = ID;//all future Packets added to this NetMessage will be stamped with ID
    FillTheVoid();//ensure that there is at least one Packet in *this to accept ID
	for(std::map<unsigned int, Packet>::iterator it = pages.begin(); it != pages.end(); ++it)
	{
		//Packet's PacketID
		it->second[9] = reinterpret_cast<unsigned char *>(&ID)[0];
		it->second[10] = reinterpret_cast<unsigned char *>(&ID)[1];
	}
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	GetID
/*!	\brief		Gets the ID all the NetMessages' Packets bear
**	\param		none
**	\return		The ID all the NetMessages' Packets bear
**	\exception	none
**	\par		DESCRIPTION:
**					
**	\author		Nathan Frost
**	\date		1/11/2003 1:57:35 PM
**	\warning
                - if there is no Packets in the NetMessage (ie GetNumCurrentPackets() = 0),
                then a meaningless value (DEFAULT) is returned.
*/
////////////////////////////////////////////////////////////////////////////////
unsigned short NetMessage::GetID()
{
    if(pages.empty())
	    return DEFAULT;
    else
    {
        int ret = 0;
        //Packet's PacketID
		reinterpret_cast<unsigned char *>(&ret)[0] = pages.begin()->second[9];
		reinterpret_cast<unsigned char *>(&ret)[1] = pages.begin()->second[10];
        return ret;
    }
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	GetPacket
/*!	\brief		Returns the Packet that bears PageNum
**	\param		PageNum: The PageNum of the requested Packet
**	\return		The Packet that bears PageNum
**	\exception	none
**	\par		DESCRIPTION:
**
**	\author		Nathan Frost
**	\date		1/11/2003 2:39:55 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
Packet NetMessage::GetPacket(unsigned char PageNum)
{
	return pages[PageNum];
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	EnterReadMode
/*!	\brief		Places the NetMessage into Read Mode
**	\param		none
**	\return		none
**	\exception	none
**	\par		DESCRIPTION:
**
**	\author		Nathan Frost
**	\date		1/11/2003 2:39:55 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
void NetMessage::EnterReadMode()
{
    /* Enter readMode */
    readMode = true;
    //we are reading from the beginning of the NetMessage
    curPacket = 0;
    curByte = PACKET_HEADER_SIZE;
    bytesRemaining = GetPacketSize(pages[curPacket]) - PACKET_HEADER_SIZE;
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	ProcessPacketFalloff
/*!	\brief		Processes NetMessage's datastructure to account for transitioning between Packets
**	\param		none
**	\return		none
**	\exception	none
**	\par		DESCRIPTION:
**
**	\author		Nathan Frost
**	\date		1/11/2003 2:39:55 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
void NetMessage::ProcessPacketFalloff()
{
    if(curPacket < totalPages && curByte >= GetPacketSize(pages[curPacket]))
	{
		//start reading from the next Packet
		++curPacket;
		curByte = PACKET_HEADER_SIZE;

		//if we fell off of the whole NetMessage...
		if(curPacket >= totalPages)
			bytesRemaining = 0;//all subsequent reads from this NetMessage will fail
		else
			bytesRemaining = GetPacketSize(pages[curPacket]) - PACKET_HEADER_SIZE;
	}
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	ReadNBytes
/*!	\brief		Reads the next N bytes from the NetMessage.  Puts the NetMessage into Read Mode.
**	\param		N: The number of bytes to read
**	\return		A pointer to the first of the requested N bytes or 0 on failure
**	\exception	none
**	\par		DESCRIPTION:
**					
**	\author		Nathan Frost
**	\date		1/11/2003 2:41:43 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
const char *NetMessage::ReadNBytes(unsigned int N)
{
	//if there's no data in the NetMessage...
	if(pages.empty())
		return 0;

	if(!readMode)
        EnterReadMode();

	//we might have fallen off of the current Packet...
	ProcessPacketFalloff();//...so deal with it

	//if we requested more data than the NetMessage has to give...
	if(N > bytesRemaining || curPacket >= totalPages)
		return 0;

	//otherwise return the requested data
	const char *ret = pages[curPacket] + curByte;
	//reflect that we read the requested N bytes
	curByte += N;
	bytesRemaining -= N;

	return ret;
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	ReadNBytes
/*!	\brief		Reads the next N bytes from the NetMessage and copies them to Destination.  Puts 
				the NetMessage into Read Mode.
**	\param		N:				The number of bytes to read
	\param		Destination:	Pointer to the buffer to copy the bytes read into
**	\return		none
**	\exception	none
**	\par		DESCRIPTION:
**					
**	\author		Nathan Frost
**	\date		1/11/2003 2:41:43 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
void NetMessage::ReadNBytes(unsigned int N, void *Destination)
{
	memcpy(Destination, ReadNBytes(N), N);
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	DeallocatePackets
/*!	\brief		Releases the dynamic memory allocated by the NetMessage
**	\param		none
**	\return		none
**	\exception	none
**	\par		DESCRIPTION:
**
**	\author		Nathan Frost
**	\date		1/11/2003 5:38:12 PM
**	\warning	
				- Always remember to call this function once one is done with the NetMessage
				- After calling a NetMessage's DeallocatePackets() function, the NetMessage's
                behavior becomes undefined
                - Calling DeallocatePackets() on a NetMessage multiple times is a serious error
                that will cause memory damage.
*/
////////////////////////////////////////////////////////////////////////////////
void NetMessage::DeallocatePackets()
{
	for(std::map<unsigned int, Packet>::iterator it = pages.begin(); it != pages.end(); ++it)
    {
        //BEG_HACK: memory allocation
        --numPacketsAllocated;
        //END_HACK: memory allocation

		assert(it->second);
		delete[] it->second;
		it->second = 0;
    }
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	ConcatenatePacket
/*!	\brief		Puts the given Packet into its rightful spot in the NetMessage
**	\param		ThePacket: to concatenate to the NetMessage
**	\return		True.
**	\exception	none
**	\par		DESCRIPTION:
**					- ConcatenatePacket is only meant to be used to reconstruct a NetMessage on the
					receiving end -- any other use is undefined
					- Data pointed to by ThePacket is blindly copied into the NetMessage

**	\author		Nathan Frost
**	\date		1/13/2003 12:26:47 PM
**	\warning
				- Assumes that ThePacket is a valid Packet
				- If this NetMessage has (totalPages > 2), then the Packets the user
				ConcatenatePacket()'s MUST all have the same DocumentNum -- otherwise, the
				NetMessage's behavior is undefined
				- A partially reconstructed NetMessage's (that is, a NetMessage with
				(totalPages > 2) that does not have all of its Packets ConcatenatePacket()'ed yet)
				behavior is undefined, except for calling the NetMessage's ConcatenatePacket()
				function with a valid ThePacket parameter.  In other words, when reconstructing a
                NetMessage, make sure that the whole NetMessage is completely ConcatenatePacket()'d
                back together before doing anything else with it.
*/
////////////////////////////////////////////////////////////////////////////////
bool NetMessage::ConcatenatePacket(Packet ThePacket)
{
	assert(ThePacket != 0);

    //BEG_HACK: memory allocation
    ++numPacketsAllocated;
    //END_HACK: memory allocation

	/* Add ThePacket to this NetMessage */
	Packet concatPacket = new char[MAX_PACKET_SIZE];
	assert(concatPacket != 0);

	memcpy(	reinterpret_cast<void *>(concatPacket), reinterpret_cast<const void *>(ThePacket),
			GetPacketSize(ThePacket));
	pages.insert(std::pair<unsigned int, Packet>(GetPageNum(concatPacket), concatPacket));

	//if the NetMessage didn't know before, it now knows how many Packets it's supposed to have
	totalPages = GetTotalPages(concatPacket);

	return true;
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	GetSizeOfBiggestPossibleRead
/*!	\brief		Returns the number of bytes that can currently be read in one call to
                WriteNBytes().  Puts the NetMessage into Read Mode.
**	\param		none
**	\return		True.
**	\exception	none
**	\par		DESCRIPTION:

**	\author		Nathan Frost
**	\date		1/13/2003 12:26:47 PM
**	\warning    none
*/
////////////////////////////////////////////////////////////////////////////////
unsigned int NetMessage::GetSizeOfBiggestPossibleRead()
{
    if(!readMode)
        EnterReadMode();

    //we might have fallen off of the current Packet...
	ProcessPacketFalloff();//...so deal with it

    if(curPacket >= totalPages)
        return 0;
    else
        return GetPacketSize(GetPacket(curPacket)) - curByte;
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	Event
/*!	\brief		Constructor
**	\param		NetMessage: The payload of this Event
	\param		SenderIP:	IPAddress of the machine that sent this NetMessage
**	\return		none
**	\exception	none
**	\par		DESCRIPTION:
**
**	\author		Nathan Frost
**	\date		1/13/2003 9:18:12 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
Event::Event(NetMessage TheNetMessage, u_long SenderIP):
netMessage(TheNetMessage), senderIP(SenderIP), timeSpentAlive(0.0)
{
	;
}


void NetMessage::DebugOut(std::ostream &Os)
{
	using std::endl;

	for(int i=0; i < totalPages; ++i)
	{
		Packet packet = pages[i];

		//extract packetSize
		unsigned int packetSize = GetPacketSize(packet);

		unsigned int pageNum = 0;
		pageNum = static_cast<unsigned int>(packet[4]);

		unsigned int totalNumberOfPackets = 0;
		totalNumberOfPackets = static_cast<unsigned int>(packet[5]);

		Os	<<	"DocumentNum:             " << GetPacketDocumentNum(packet)
			<< endl <<
				"PageNum:                 " << pageNum
			<< endl << 
				"Total Number of Packets: " << totalNumberOfPackets
			<< endl <<
				"PacketSize:              " << packetSize 
			<< endl <<
				"PacketID:                " << *reinterpret_cast<unsigned short *>(packet + 9) 
			<< endl << endl;
	}
}
