////////////////////////////////////////////////////////////////////////////////
/*!	\file		NetData.h
**	\brief		Declaration of Packet, NetMessage and Event datastructures that are used in the
				Networking module.
**	\author		Nathan Frost
**	\date		1/10/2003 6:38:26 PM
**	\par		DESCRIPTION:
**					
**	\par		REMARKS:
					- Currently the boost::singleton_pool is not implemented -- in its stead is the
					standard new/delete paradigm of dynamic memory management.  This could very
					well lead to serious memory fragmentation and unacceptable running efficiency.
					- There appears to be a subtle off-by-one error in NetMessage such that one can
					only write MAX_PACKET_SIZE - PACKET_HEADER_SIZE - 1 bytes to a NetMessage.
					
**					
*/
////////////////////////////////////////////////////////////////////////////////
#ifndef NET_DATA_H
#define NET_DATA_H
#pragma once


#include <map>
#include "UDPNetworking.h"
#include <iostream>


//BEG_HACK: memory allocation
extern int numPacketsAllocated;
//END_HACK: memory allocation


namespace UDPNet
{
	extern const unsigned int PACKET_HEADER_SIZE;///<in bytes
    extern const unsigned int MAX_PACKET_PAYLOAD_SIZE;///<in bytes


	typedef char * Packet;///<The lowest level datatype that can be sent over the network

	
	////////////////////////////////////////////////////////////////////////////////
	/*!	\class		NetMessage
	**	\brief		The NetMessage is the high level unit of data that is 
					GameNetworking::SendMessage()ed to remotely participating machines.
	**	\author		Nathan Frost
	**	\date		1/10/2003 7:01:11 PM
	**	\warning	
					- Always remember to DeallocatePackets() once and only once after you're
                    finished with a NetMessage!
					- Always ensure that the NetMessage is valid when passing it around (this often
                    means passing NetMessages by value)!
					- A PacketID of 0 is the default PacketID, also known as the NULL PacketID --
					it is an error to receive any Packet with PacketID = 0.  Use SetID() to set all
					PacketIDs in this NetMessage to a valid Packet ID, defined in NetMessageID.h.
	*/
	////////////////////////////////////////////////////////////////////////////////
	class NetMessage
	{
	private:
		std::map<unsigned int, Packet> pages;	///<Handle to all the Packets in this NetMessage
		unsigned int curPacket;					///<The current Packet that is being written to or
												///<read from
		unsigned int curByte;					///<The byte that is currently being written or read
		unsigned char totalPages;				///<The total number of Packets
		unsigned int bytesRemaining;			///<The number of "empty" bytes left in the
												///<NetMessage -- that is, if writing,
												///<bytesRemaining = empty space left, if reading,
												///<bytesRemaining = valid bytes left to read
		bool readMode;							///<Whether or not readMode has been entered;
												///<!readMode = writeMode
		unsigned short packetID;				///<The PacketID of every Packet in the NetMessage
		//boost::singletonpool;


		void NewPacket();
		unsigned char GetPageNum(Packet ThePacket);
		unsigned int GetPacketDocumentNum(Packet ThePacket);
		unsigned char GetTotalPages(Packet ThePacket);
		void SetPacketSize(Packet ThePacket, unsigned int Size);
        void FillTheVoid();
        void EnterReadMode();
        void ProcessPacketFalloff();
	public:
		NetMessage();
		unsigned int GetNetMessageDocumentNum();
		bool WriteNBytes(unsigned int N, const void *Data);
		const char *ReadNBytes(unsigned int N);
		void ReadNBytes(unsigned int N, void *Destination);
		Packet GetPacket(unsigned char PacketNum);
		unsigned int GetNumPackets();
		unsigned int GetNumCurrentPackets();
		unsigned int GetPacketSize(Packet ThePacket);
        unsigned int GetSizeOfBiggestPossibleRead();
		unsigned short GetID();
		void SetID(unsigned short ID);
		bool ConcatenatePacket(Packet ThePacket);
		void DeallocatePackets();
        void SetDocumentNum(unsigned int DocumentNum);


		void DebugOut(std::ostream &Os);
	};

	////////////////////////////////////////////////////////////////////////////////
	/*!	\class		Event
	**	\brief		Contains a NetMessage, the IP Address of the sender, and the time the Event was
					constructed.
	**	\author		Nathan Frost
	**	\date		1/13/2003 9:11:52 PM
	**	\warning	none
	*/
	////////////////////////////////////////////////////////////////////////////////
	struct Event
	{
		u_long senderIP;		///<IPAddress of Sender (in Little Endian)
		NetMessage netMessage;	///<payload
		float timeSpentAlive;	///<time since Event was constructed
		
		explicit Event(NetMessage TheNetMessage, u_long SenderIP);
	};
}
#endif