////////////////////////////////////////////////////////////////////////////////
/*!	\file		RemoteSocksToWaitOn.cpp
**	\brief		Implementation
**	\author		Nathan Frost
**	\date		2/7/2003 4:36:08 PM
**	\par		DESCRIPTION:
**					
**	\par		REMARKS:
**					
*/
////////////////////////////////////////////////////////////////////////////////
#pragma warning(disable:4786)
#include "RemoteSocksToWaitOn.h"
#include "UDPNetworking.h"
#include <winsock2.h>
#include "../nexception/Nexception.h"
#include "NetData.h"
#include <time.h>
#include <assert.h>
using namespace UDPNet;
// memory leak detection
#ifdef _DEBUG
    #include"memleak.h"
    #define new DEBUG_NEW
    #undef THIS_FILE
    static char THIS_FILE[] = __FILE__;
#endif


//forward declarations
namespace
{
    fd_set set;
}


//file scope variables
HANDLE hMutexReady, hMutexWaitOn, hMutexEvent, hMutexDeath;
HANDLE hThread;
DWORD threadID;


////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	CreateMutex
/*!	\brief		Creates a Windows API mutex object.
**	\param		Name:	Pointer to a string that uniquely identifies the mutex object.
	\param		hMutex:	Handle value that will be filled out to refer to the created mutex.
**	\return		none
**	\exception	
				- Throws a Nexception if the mutex fails to be created.
				- Throws a Nexception if the mutex already exists.
**	\par		DESCRIPTION:
**					
**	\author		Nathan Frost
**	\date		2/7/2003 5:13:54 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
void CreateMutex(char *Name, HANDLE &hMutex)
{
	//create Mutex
	if((hMutex = CreateMutex(	NULL,	//default security attributes
								FALSE,  //no initial owner of the mutex
        	                    Name))	//name of mutex object (length must be < MAX_PATH)
            	                == NULL)
	{
    	throw Nexception("CreateMutex() returned error.", __FILE__, __LINE__, __DATE__, __TIME__);
	}
	if(GetLastError() == ERROR_ALREADY_EXISTS)
	{
		throw Nexception("Mutex object of given Name already exists.", __FILE__, __LINE__, __DATE__, __TIME__);
	}
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	DestroyMutex
/*!	\brief		Destroys a Windows API mutex object.
**	\param		hMutex: Value that refers to the mutex to be destroyed.
**	\return		none
**	\exception	
				- Throws a Nexception if the mutex is not closed.
**	\par		DESCRIPTION:
**					
**	\author		Nathan Frost
**	\date		2/7/2003 5:18:27 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
void DestroyMutex(HANDLE hMutex)
{
	if(!CloseHandle(hMutex))
	{
		throw Nexception("Mutex's handle couldn't be closed.", __FILE__, __LINE__, __DATE__, __TIME__);
	}
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	LockData
/*!	\brief		Locks the specified Windows API mutex
**	\param		hMutex: Value that refers to the mutex to lock
**	\return		none
**	\exception
				- Throws a Nexception if the caller unequivocally fails to lock the mutex
**	\par		DESCRIPTION:
**					
**	\author		Nathan Frost
**	\date		2/7/2003 5:28:11 PM
**	\warning
                - call will block until the requested mutex becomes unlocked
*/
////////////////////////////////////////////////////////////////////////////////
void LockData(HANDLE hMutex)
{
	//if hEventMutex is closed while WaitForSingleObject is still pending, the function's behavior
    //is undefined
	DWORD result = WaitForSingleObject(	hMutex, //object to gain ownership of ASAP
    									INFINITE);	//milliseconds to wait before function returns,
                                        			//even if object ownership was not attained
    //WAIT_ABANDONED means that the thread that created the Mutex object has terminated, but the
    //Mutex remains.  This should not concern us -- it is our responsibility to systematically free
    //the resources the Mutex occupies, not the creating thread's.
	if(result != WAIT_OBJECT_0 && result != WAIT_ABANDONED)
    {
    	throw Nexception("Failed to gain ownership of Mutex object.", __FILE__, __LINE__, __DATE__, __TIME__);
    }

    return;
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	UnlockData
/*!	\brief		Unlocks the specified Windows API mutex
**	\param		hMutex: Value that refers to the mutex to unlock
**	\return		none
**	\exception	
				- Throws a Nexception if the caller unequivocally fails to unlock the mutex
**	\par		DESCRIPTION:
**					
**	\author		Nathan Frost
**	\date		2/7/2003 5:33:56 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
void UnlockData(HANDLE hMutex)
{
	if(!ReleaseMutex(hMutex))
    {
        throw Nexception("Failed to release ownership of Mutex object.", __FILE__, __LINE__, __DATE__, __TIME__);
    }

    return;
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	CreateMutexes
/*!	\brief		Creates all the mutex objects necessary for RemoteSocksToWaitOn to do its job.
**	\param		none
**	\return		none
**	\exception	none
**	\par		DESCRIPTION:
**					
**	\author		Nathan Frost
**	\date		2/7/2003 5:15:58 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
void CreateMutexes()
{
    CreateMutex("WaitOn", hMutexWaitOn);
    CreateMutex("Ready", hMutexReady);
    CreateMutex("Event", hMutexEvent);
    CreateMutex("Death", hMutexDeath);
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	DestroyMutexes
/*!	\brief		Destroys all the mutex objects necessary for RemoteSocksToWaitOn to do its job.
**	\param		none
**	\return		none
**	\exception	none
**	\par		DESCRIPTION:
**					
**	\author		
**	\date		2/7/2003 5:23:58 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
void DestroyMutexes()
{
    DestroyMutex(hMutexWaitOn);
    DestroyMutex(hMutexReady);
    DestroyMutex(hMutexEvent);
    DestroyMutex(hMutexDeath);
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	RemoteSocksToWaitOn
/*!	\brief		Constructor
**	\param		none
**	\return		none
**	\exception	none
**	\par		DESCRIPTION:
**					- establishes "blank slate" invariant for RemoteSocksToWaitOn
					- threadState is initialized to ACTIVE (even though the ListenThread is not
					active yet) for thread-synchronization reasons

**	\author		Nathan Frost
**	\date		2/7/2003 5:16:40 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
UDPNet::RemoteSocksToWaitOn::RemoteSocksToWaitOn():
threadState(ACTIVE), timeLeftToModeChange(0.0), param(), packetLossOccurrenceBase(10.0),
packetLossOccurrenceDelta(4.0), packetLossLengthBase(2.0), packetLossLengthDelta(1.0),
packetLossActive(false), throwAway(false)
{
	// oh hey, srand is already called in entrypoint.cpp, void GameInit(void);, line ~132
	// nate cleveland [3/17/2003]
//    srand(time(0));

	FD_ZERO(&set);

	CreateMutexes();

    LockMutexWaitOn();
	waitOn.clear();
    UnlockMutexWaitOn();
    LockMutexReady();
	ready.clear();
    UnlockMutexReady();
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	~RemoteSocksToWaitOn
/*!	\brief		Destructor
**	\param		none
**	\return		none
**	\exception	none
**	\par		DESCRIPTION:
**
**	\author		Nathan Frost
**	\date		2/7/2003 5:24:27 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
UDPNet::RemoteSocksToWaitOn::~RemoteSocksToWaitOn()
{
	DestroyMutexes();
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	AddSock
/*!	\brief		Adds a new SOCKET to the list of SOCKETs that are listened to for data
**	\param		Sock: SOCKET to start listening for data on
**	\return		none
**	\exception	none
**	\par		DESCRIPTION:
**					
**	\author		Nathan Frost
**	\date		2/7/2003 5:24:45 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
void UDPNet::RemoteSocksToWaitOn::AddSock(SOCKET Sock)
{
    LockMutexWaitOn();
	waitOn.push_back(Sock);
    UnlockMutexWaitOn();
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	RemoveSock
/*!	\brief		Removes a SOCKET from the list of SOCKETs that are listened to for data
**	\param		Sock: SOCKET to stop listening for data on
**	\return		none
**	\exception	none
**	\par		DESCRIPTION:
**					
**	\author		Nathan Frost
**	\date		2/7/2003 5:25:46 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
void UDPNet::RemoteSocksToWaitOn::RemoveSock(SOCKET Sock)
{
    LockMutexWaitOn();
	waitOn.remove(Sock);
    UnlockMutexWaitOn();
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	RemoveAllSocks
/*!	\brief		Causes RemoteSocksToWaitOn to stop listening to all SOCKETs
**	\param		none
**	\return		none
**	\exception	none
**	\par		DESCRIPTION:
**					
**	\author		Nathan Frost
**	\date		2/7/2003 5:26:18 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
void UDPNet::RemoteSocksToWaitOn::RemoveAllSocks()
{
    LockMutexWaitOn();
	waitOn.clear();
    UnlockMutexWaitOn();
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	CurThreadState
/*!	\brief		Lets the user know what state the Thread Object is currently in.
**	\param		none
**	\return		Thread Object's state
**	\exception	none
**	\par		DESCRIPTION:
**					
**	\author		Nathan Frost
**	\date		2/7/2003 5:27:14 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
enum ThreadState UDPNet::RemoteSocksToWaitOn::CurThreadState()
{
    return threadState;
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	LockMutexReady
/*!	\brief		Locks the mutex that corresponds to the ready member of RemoteSocksToWaitOn
**	\param		none
**	\return		none
**	\exception	none
**	\par		DESCRIPTION:
**					
**	\author		Nathan Frost
**	\date		2/7/2003 5:36:00 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
void LockMutexReady()
{
    LockData(hMutexReady);
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	UnlockMutexReady
/*!	\brief		Unlocks the mutex that corresponds to the ready member of RemoteSocksToWaitOn
**	\param		none
**	\return		none
**	\exception	none
**	\par		DESCRIPTION:
**					
**	\author		Nathan Frost
**	\date		2/7/2003 5:36:04 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
void UnlockMutexReady()
{
    UnlockData(hMutexReady);
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	LockMutexWaitOn
/*!	\brief		Locks the mutex that corresponds to the waitOn member of RemoteSocksToWaitOn
**	\param		none
**	\return		none
**	\exception	none
**	\par		DESCRIPTION:
**
**	\author		Nathan Frost
**	\date		2/7/2003 5:36:06 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
void LockMutexWaitOn()
{
    LockData(hMutexWaitOn);
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	UnlockMutexWaitOn
/*!	\brief		Unlocks the mutex that corresponds to the waitOn member of RemoteSocksToWaitOn
**	\param		none
**	\return		none
**	\exception	none
**	\par		DESCRIPTION:
**
**	\author		Nathan Frost
**	\date		2/7/2003 5:36:09 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
void UnlockMutexWaitOn()
{
    UnlockData(hMutexWaitOn);
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	LockMutexEvent
/*!	\brief		Locks the mutex that corresponds to the list of event list member of GameNetworking
**	\param		none
**	\return		none
**	\exception	none
**	\par		DESCRIPTION:
**
**	\author		Nathan Frost
**	\date		2/7/2003 5:36:13 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
void LockMutexEvent()
{
    LockData(hMutexEvent);
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	UnlockMutexEvent
/*!	\brief		Unlocks the mutex that corresponds to the list of event list member of GameNetworking
**	\param		none
**	\return		none
**	\exception	none
**	\par		DESCRIPTION:
**
**	\author		Nathan Frost
**	\date		2/7/2003 5:36:16 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
void UnlockMutexEvent()
{
    UnlockData(hMutexEvent);
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	LockMutexDeath
/*!	\brief		Locks the mutex that corresponds to the list of SOCKETs pending DestroySock()'tion
**	\param		none
**	\return		none
**	\exception	none
**	\par		DESCRIPTION:
**
**	\author		Nathan Frost
**	\date		2/7/2003 5:36:13 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
void LockMutexDeath()
{
    LockData(hMutexDeath);
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	UnlockMutexDeath
/*!	\brief		Unlocks the mutex that corresponds to the list of SOCKETs pending DestroySock()'tion
**	\param		none
**	\return		none
**	\exception	none
**	\par		DESCRIPTION:
**
**	\author		Nathan Frost
**	\date		2/7/2003 5:36:16 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
void UnlockMutexDeath()
{
    UnlockData(hMutexDeath);
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	KillSock
/*!	\brief      Frees the resources associated with Sock in a threadsafe manner.
**	\param		Sock: SOCKET to give back to the OS
**	\return		none
**	\exception	none
**	\par		DESCRIPTION:
**
**	\author		Nathan Frost
**	\date		2/7/2003 5:36:16 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
void UDPNet::RemoteSocksToWaitOn::KillSock(SOCKET Sock)
{
    LockData(hMutexDeath);
    death.push_back(Sock);
    UnlockData(hMutexDeath);
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	Wait
/*!	\brief		Waits until data is available on one or more SOCKETs RemoteSocksToWaitOn is
				listening to
**	\param		none
**	\return		true if the user specified some SOCKETs to listen to, false if the user specified
                no SOCKETs to listen to
**	\exception	none
**	\par		DESCRIPTION:
**					
**	\author		Nathan Frost
**	\date		2/7/2003 5:39:56 PM
**	\warning	
	  k			- this is a blocking call
*/
////////////////////////////////////////////////////////////////////////////////
bool UDPNet::RemoteSocksToWaitOn::Wait()
{
    //select()'s behaviour is undefined if passed an empty set of sockets...
    LockMutexWaitOn();
    if(waitOn.empty())
    {
        UnlockMutexWaitOn();
        return false;//...so notify the user that there's nothing for Wait() to do
    }

    /* Clear out any old socks, and get ready to listen to the specified socks */
    LockMutexReady();
	ready.clear();
	FD_ZERO(&set);
	for(std::list<SOCKET>::iterator it = waitOn.begin(); it != waitOn.end(); ++it)
		FD_SET(*it, &set);
    UnlockMutexWaitOn();
    UnlockMutexReady();

    //listen for data on the specified socks
	select(	0,		//meaningless argument -- included only for backwards compatibility
			&set,	//return when one or more of the SOCKETs in this set become readable
			0,		//don't check any SOCKETs for writability
			0,		//don't check any SOCKETs for errors
			NULL);	//block (wait indefinitely) until one or more SOCKETs become readable

	//select() replaced the SOCKETs in set with the subset of SOCKETs that are readable (in other
	//words, set only contains SOCKETs that have data ready to be recv()'d from them)
    LockMutexWaitOn();
    LockMutexReady();
    //report all socks that have data on them
	for(std::list<SOCKET>::iterator jt = waitOn.begin(); jt != waitOn.end(); ++jt)
	{
		if(FD_ISSET(*jt, &set))
			ready.push_back(*jt);
	}
    UnlockMutexWaitOn();
    UnlockMutexReady();

    return true;
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	ListenSockFunc
/*!	\brief		Thread Object that gets data from ripe SOCKETs, puts that data in an Event and 
				shoves that Event in the specified Event container
**	\param		Param: (void *) that must point to a valid, filled out ThreadParameter object
**	\return		0 when there are no SOCKETs to listen to.
**	\exception	none
**	\par		DESCRIPTION:
**					
**	\author		Nathan Frost
**	\date		2/7/2003 5:45:33 PM
**	\warning	
				- this is a blocking call
*/
////////////////////////////////////////////////////////////////////////////////
DWORD WINAPI ListenSockFunc(LPVOID Param)
{
	//prepare for listening
    ThreadParameter *param = static_cast<ThreadParameter *>(Param);
    char *buf = new char[UDPNet::MAX_PACKET_SIZE];
	assert(buf != 0);

	//keep listening until there are no SOCKETs left to listen to
    while(param->remoteSocksToWaitOn->Wait())
    {
		//for every SOCKET that has data waiting on it...
        LockMutexReady();
        for(std::list<SOCKET>::iterator it = param->ready->begin(); it != param->ready->end(); ++it)
        {
			//...get that data!
            u_long senderIP;
            u_short senderPort;
            unsigned int numBytesReceived = 
						param->udpNetworking->ReceivePacket(*it, reinterpret_cast<void *>(&buf[0]),
													UDPNet::MAX_PACKET_SIZE, senderIP, senderPort);

            if(	param->remoteSocksToWaitOn->ThrowAway() ||		//if we're simulating PacketLoss
				numBytesReceived < UDPNet::PACKET_HEADER_SIZE)	//or the packet is smaller than any
			{													//packet the game would ever send
                continue;										//then ignore this Packet
			}

            //see if there's already an event with this document number here
            LockMutexEvent();
            bool found = false;//assume that this is a brand new event
            for(std::deque<UDPNet::Event>::iterator it = param->event->begin();
                it != param->event->end(); ++it)
            {
				if( it->netMessage.GetNetMessageDocumentNum() == 
					*reinterpret_cast<unsigned int *>(buf)&&//since DocumentNum=first 4 bytes of buf
                    it->senderIP == senderIP)//and the packets were in NetMessage from the same guy
                {
                    it->netMessage.ConcatenatePacket(buf);
                    found = true;   //we found the single event to concatenate this NetMessage to
                    break;          //so we're done
                }
            }
            if(!found)
            {
                //compose and insert the new event
                UDPNet::NetMessage netMessage;
                netMessage.ConcatenatePacket(buf);
                UDPNet::Event event(netMessage, senderIP);
                param->event->push_back(event);
            }
            UnlockMutexEvent();
        }
        UnlockMutexReady();

        /* Destroy the SOCKETs that are marked for death */
        LockMutexDeath();
        while(param->death->size())
        {
            param->udpNetworking->DestroySock(*param->death->begin());
            param->death->pop_front();
        }
        UnlockMutexDeath();

		//if the main thread wants us dead...
		if(*param->threadState == DYING)
			break;//...so be it
    }

	delete[] buf;
    *param->threadState = DEAD;//let RemoteSocksToWaitOn know we're done
    return 0;
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	LaunchListeningThread
/*!	\brief		Brings the Thread Object into existence and gets it executing.
**	\param		udpNetworking:	Reference to the UDPNetworking object
	\param		event:			Reference to the container of Events that will receive incoming 
								events
**	\return		none
**	\exception	none
**	\par		DESCRIPTION:
**					
**	\author		Nathan Frost
**	\date		2/7/2003 6:00:46 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
void UDPNet::RemoteSocksToWaitOn::LaunchListeningThread(UDPNet::UDPNetworking &udpNetworking,
                                                        std::deque<Event> &event)
{
    param.remoteSocksToWaitOn = this;
    param.udpNetworking = &udpNetworking;
    param.ready = &this->ready;
    param.waitOn =  &this->waitOn;
    param.death = &this->death;
    param.event = &event;
    param.threadState = &this->threadState;

    if(!(hThread = CreateThread(
    	NULL,						//default security attributes
        0,							//use parent process's default stack size for this thread
        ListenSockFunc,				//pointer to the function that gets called upon thread creation
        reinterpret_cast<void *>(&param),//pass a pointer to param so ListenSockFunc can do its job
        0,							//no special flags
        &threadID)))	            //get the thread ID
    {
		throw Nexception("CreateThread() failed.", __FILE__, __LINE__, __DATE__, __TIME__);
	}
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	KillThread
/*!	\brief		If the listening thread is active, calling this function will send it on the path
                to termination
**	\param		none
**	\return		true if the thread has been set to DYING (ie is in the process of terminating),
				false if the thread is already DYING or DEAD.
**	\exception	none
**	\par		DESCRIPTION:
**
**	\author		Nathan Frost
**	\date		2/14/2003 10:28:43 AM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
bool UDPNet::RemoteSocksToWaitOn::KillThread()
{
	if(threadState == ACTIVE)
	{
		threadState = DYING;
		return true;
	}
	else
		return false;
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	TurnOnPacketLossSimulation
/*!	\brief		Turns on Packet Loss Simulation, with the specified parameters.
**	\param		none
**	\return		true if Packet Loss Simulation was engaged, false if it was not
**	\exception	none
**	\par		DESCRIPTION:
**

**	\author		Nathan Frost
**	\date		2/7/2003 5:16:40 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
bool RemoteSocksToWaitOn::TurnOnPacketLossSimulation(   float PacketLossOccurrenceBase,
                                                        float PacketLossOccurrenceDelta,
                                                        float PacketLossLengthBase,
                                                        float PacketLossLengthDelta)
{
    if( PacketLossOccurrenceDelta >= PacketLossOccurrenceBase ||
        PacketLossLengthDelta >= PacketLossLengthBase)
    {
        return false;
    }
    else
    {
        packetLossOccurrenceBase = PacketLossOccurrenceBase;
        packetLossOccurrenceDelta = PacketLossOccurrenceDelta;
        packetLossLengthBase = PacketLossLengthBase;
        packetLossLengthDelta = PacketLossLengthDelta;
        packetLossActive = true;

        return true;
    }
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	TurnOffPacketLossSimulation
/*!	\brief		Turns off Packet Loss Simulation.
**	\param		none
**	\return		none
**	\exception	none
**	\par		DESCRIPTION:
**

**	\author		Nathan Frost
**	\date		2/7/2003 5:16:40 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
void RemoteSocksToWaitOn::TurnOffPacketLossSimulation()
{
    packetLossActive = false;
    throwAway = false;//since packet loss simulation is turned off, we'll never throw away packets
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	ThrowAway
/*!	\brief		Reveals whether or not the listen thread is currently tossing packets instead of
                reporting them.
**	\param		none
**	\return		none
**	\exception	none
**	\par		DESCRIPTION:
**

**	\author		Nathan Frost
**	\date		2/7/2003 5:16:40 PM
**	\warning    none
*/
////////////////////////////////////////////////////////////////////////////////
bool RemoteSocksToWaitOn::ThrowAway()
{
    return throwAway;
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	StepTimeLeftToModeChange
/*!	\brief		Modifies the value of timeLeftToModeChange.  Intended to be modified every frame.
**	\param		Value: to modify timerLeftToModeChange.
**	\return		none
**	\exception	none
**	\par		DESCRIPTION:
**					
**	\author		Nathan Frost
**	\date		2/12/2003 12:37:44 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
void UDPNet::RemoteSocksToWaitOn::StepTimeLeftToModeChange(float Value)
{
    if(packetLossActive)
    {
        timeLeftToModeChange += Value;

        if(timeLeftToModeChange < 0.0)
        {
            float deltaPercentage = static_cast<float>(rand()) / RAND_MAX;
            float sign = 1.0;
            if(rand()%2)
                sign = -1.0;

            if(throwAway == true)
            {
                throwAway = false;
                timeLeftToModeChange =  packetLossOccurrenceBase +
                                        sign*packetLossOccurrenceDelta*deltaPercentage;
            }
            else
            {
                throwAway = true;
                timeLeftToModeChange =  packetLossLengthBase +
                                        sign*packetLossLengthDelta*deltaPercentage;
            }
        }
    }
}






#if 0//testing crap code
void RemoteSocksToWaitOn::Test()
{
	using std::cout;
	using std::endl;

	SOCKET joy, happy, yaye;
	
	UDPNetworking *udpNet = UDPNetworking::Instance();
	u_short anyPort = 0;
	udpNet->CreateSock(joy, anyPort);
	anyPort = 0;
	udpNet->CreateSock(happy, anyPort);
	anyPort = 0;
	udpNet->CreateSock(yaye, anyPort);

	cout << "joy = " << joy << endl;
	cout << "happy = " << happy << endl;
	cout << "yaye = " << yaye << endl;

	cout << "Add joy and happy." << endl;
	AddSock(joy);
	AddSock(happy);

	for(std::list<SOCKET>::iterator it = waitOn.begin(); it != waitOn.end(); ++it)
		cout << *it << endl;

	cout << "Remove joy." << endl;
	RemoveSock(joy);
	for(it = waitOn.begin(); it != waitOn.end(); ++it)
		cout << *it << endl;
	
	cout << "Remove joy again." << endl;
	RemoveSock(joy);
	for(it = waitOn.begin(); it != waitOn.end(); ++it)
		cout << *it << endl;

	cout << "Add yaye." << endl;
	AddSock(yaye);
	for(it = waitOn.begin(); it != waitOn.end(); ++it)
		cout << *it << endl;

	cout << "Remove all sockets." << endl;
	RemoveAllSocks();
	for(it = waitOn.begin(); it != waitOn.end(); ++it)
		cout << *it << endl;
}
#endif