////////////////////////////////////////////////////////////////////////////////
/*!	\file		RemoteSocksToWaitOn.h
**	\brief		ReceiveSocksToWaitOn abstracts the concept of waiting for data to be received by a
				set of SOCKETs.
**	\author		Nathan Frost
**	\date		2/7/2003 4:26:49 PM
**	\par		DESCRIPTION:
**					
**	\par		REMARKS:
**					
*/
////////////////////////////////////////////////////////////////////////////////
#ifndef REMOTE_SOCKS_TO_WAIT_H
#define REMOTE_SOCKS_TO_WAIT_H
#pragma once


#include "NetData.h"
#include <deque>
#include <list>


//forward declarations
namespace UDPNet
{
    class RemoteSocksToWaitOn;
    class UDPNetworking;
}


void CreateMutexes();
void DestroyMutexes();
void LockMutexReady();
void UnlockMutexReady();
void LockMutexWaitOn();
void UnlockMutexWaitOn();
void LockMutexEvent();
void UnlockMutexEvent();


enum ThreadState {ACTIVE, DYING, DEAD};///<state of the listen thread


////////////////////////////////////////////////////////////////////////////////
/*!	\class		ThreadParameter
**	\brief		Encapsulates all the information the Thread Object requires to do its job.
**	\author		Nathan Frost
**	\date		2/7/2003 4:32:55 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
struct ThreadParameter
{
    UDPNet::RemoteSocksToWaitOn *remoteSocksToWaitOn;
    UDPNet::UDPNetworking *udpNetworking;
    std::list<SOCKET> *ready;                                   
    std::list<SOCKET> *waitOn;
    std::list<SOCKET> *death;
    std::deque<UDPNet::Event> *event;
    enum ThreadState *threadState;

    ThreadParameter(UDPNet::RemoteSocksToWaitOn *RemoteSocksToWaitOn,
                    UDPNet::UDPNetworking *UDPNetworking, std::list<SOCKET> *Ready,
                    std::list<SOCKET> *WaitOn, std::deque<UDPNet::Event> *Event,
                    enum ThreadState *TheThreadState):
                    remoteSocksToWaitOn(RemoteSocksToWaitOn), udpNetworking(UDPNetworking),
                    ready(Ready), waitOn(WaitOn), event(Event),
                    threadState(TheThreadState){}
    ThreadParameter():
        remoteSocksToWaitOn(0), udpNetworking(0), ready(0), waitOn(0), death(0), event(0),
        threadState(0){}
};


namespace UDPNet
{
	////////////////////////////////////////////////////////////////////////////////
	/*!	\class		RemoteSocksToWaitOn
	**	\brief		Provides a high level interface to specify which SOCKETs to wait on, actually 
					wait for data to be sent to one or more of the SOCKETs, and then retrieve 
					which SOCKETs have Packets ready to be received. Finally, it allows the user to
					simulate internet latency by allowing the user to specify how often, and for
					how long, batches of Packets will be discarded rather than received.
	**	\author		Nathan Frost
	**	\date		2/7/2003 4:33:38 PM
	**	\warning	none
	*/
	////////////////////////////////////////////////////////////////////////////////
	class RemoteSocksToWaitOn
	{
	private:
        enum ThreadState threadState;   ///<The current state of the listening thread
        ThreadParameter param;          ///<repository for the listening thread's interface data
        std::list<SOCKET> waitOn;       ///<all SOCKETs that are currently being listened to
		std::list<SOCKET> ready;        ///<all SOCKETs that contain data that should be read
        std::list<SOCKET> death;        ///<all SOCKETs that need to be DestroySock()'d ASAP


		/* Packet Loss Simulation Information is all measured in seconds */
        ///every packetLossOccurrenceBase (+ or -) packetLossOccurrenceDelta seconds, for
        ///packetLossLengthBase (+ or -) packetLossLengthDelta seconds, ReceiveSocksToWaitOn will
        ///throw away every incoming Packet.
        float packetLossOccurrenceBase;
        float packetLossOccurrenceDelta;
        float packetLossLengthBase;
        float packetLossLengthDelta;
		float timeLeftToModeChange;     ///<time remaining until packet loss mode changes
        bool packetLossActive;          ///<whether or not Packet Loss Simulation is activated
        bool throwAway;                 ///<if (packetLoss && throwAway) then chuck Packets
	public:
        bool TurnOnPacketLossSimulation(float PacketLossOccurrenceBase,
                                        float PacketLossOccurrenceDelta, float PacketLossLengthBase,
                                        float PacketLossLengthDelta);
        void TurnOffPacketLossSimulation();
        bool ThrowAway();

		void AddSock(SOCKET Sock);
		void RemoveSock(SOCKET Sock);
		void RemoveAllSocks();
        void KillSock(SOCKET Sock);
		bool Wait();
        void LaunchListeningThread( UDPNet::UDPNetworking &udpNetworking,
                                    std::deque<UDPNet::Event> &event);
        enum ThreadState CurThreadState();
		bool KillThread();
		void StepTimeLeftToModeChange(float Value);
		RemoteSocksToWaitOn();
		~RemoteSocksToWaitOn();


		void Test();                             
	};
}


#endif
