////////////////////////////////////////////////////////////////////////////////
/*!	\file		UDPNetworking.cpp
**	\brief		Implements UDPNetworking.h
**	\author		Nathan Frost
**	\date		10/25/2002 6:23:10 PM
**	\par		DESCRIPTION:
**
**	\par		REMARKS:
**					
*/
////////////////////////////////////////////////////////////////////////////////
#pragma warning(disable:4786)
#include <winsock2.h>
#include "UDPNetworking.h"
#include "../nexception\Nexception.h"
#include <assert.h>
using namespace UDPNet;
// memory leak detection
#ifdef _DEBUG
    #include"memleak.h"
    #define new DEBUG_NEW
    #undef THIS_FILE
    static char THIS_FILE[] = __FILE__;
#endif


////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	UDPNetworking::IdentifyError
/*!	\brief		Helper debugger function to identify WinSock errors.
**	\param		none
**	\return		none
**	\exception	none
**	\par		DESCRIPTION:

**
**	\author     Nathan Frost
**	\date		10/25/2002 9:54:47 PM
**	\warning    none
*/
////////////////////////////////////////////////////////////////////////////////
namespace UDPNet
{
    void IdentifyError()
    {
        int err;
        switch(err = WSAGetLastError())
        {
            case WSANOTINITIALISED:
            {
                Beep(440, 50);
                break;
            }
            case WSAENETDOWN:
            {
                Beep(440, 50);
                break;
            }
            case WSAEFAULT:
            {
                Beep(440, 50);
                break;
            }
            case WSAEINTR:
            {
                Beep(440, 50);
                break;
            }
            case WSAEINPROGRESS:
            {
                Beep(440, 50);
                break;
            }
            case WSAEINVAL:
            {
                Beep(440, 50);
                break;
            }
            case WSAEISCONN:
            {
                Beep(440, 50);
                break;
            }
            case WSAENETRESET:
            {
                Beep(440, 50);
                break;
            }
            case WSAENOTSOCK:
            {
                Beep(440, 50);
                break;
            }
            case WSAEOPNOTSUPP:
            {
                Beep(440, 50);
                break;
            }
            case WSAESHUTDOWN:
            {
                Beep(440, 50);
                break;
            }
            case WSAEWOULDBLOCK:
            {
                Beep(440, 50);
                break;
            }
            case WSAEMSGSIZE:
            {
                Beep(440, 50);
                break;
            }
            case WSAETIMEDOUT:
            {
                Beep(440, 50);
                break;
            }
            case WSAECONNRESET:
            {
                Beep(440, 50);
                break;
            }
            default:
            {
                Beep(440, 50);
                break;
            }
        }
    }
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	UDPNetworking::Instance
/*!	\brief		Allows the client access to the UDPNetworking object.
**	\param		none
**	\return		Pointer to the singleton UDPNetworking object
**	\exception	none
**	\par		DESCRIPTION:
**					
**	\author		Nathan Frost
**	\date		10/25/2002 6:24:14 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
UDPNetworking *UDPNetworking::Instance()
{
	static UDPNetworking instance;
	return &instance;
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	UDPNetworking::UDPNetworking()
/*!	\brief		Constructor
**	\param		none
**	\return		none
**	\exception	Throws a Nexception on failure
**	\par		DESCRIPTION:
				- Is private since UDPNetworking is a Singleton -- use Instance() to access the
				object.
				- Initializes local address information, which includes only one of the local
				machine's IP Addresses (though their could potentially be many) and the ASCII 
				representation of the host machine's name
**					
**	\author		Nathan Frost
**	\date		10/25/2002 6:26:19 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
UDPNetworking::UDPNetworking()
{
    //BEG_HACK: memory allocation
    numSocketsAllocated = 0;
    //END_HACK: memory allocation

	//initialize WinSock 2.2
    WORD versionRequested = MAKEWORD(2,2);
	WSADATA wsaData;
	int err = WSAStartup(versionRequested, &wsaData);
	
    //if there was an error...
	if(err != 0)
    {
   		//...deal with it
        throw Nexception(	"UDPNetworking::UDPNetworking() returned error.", __FILE__, __LINE__, 
							__DATE__, __TIME__);
	}
    //confirm that we are all set to use a WinSock DLL that provides WinSock 2.2's functionality
    //(ie is version 2.2 or higher)
	if(LOBYTE(wsaData.wVersion) != 2 || HIBYTE(wsaData.wVersion) != 2)
    {
		//WinSock DLL is lower than WinSock 2.2
      	WSACleanup();
        throw Nexception(	"UDPNetworking::UDPNetworking() sez: ""Didn't get WinSock 2.2.""", 
							__FILE__, __LINE__, __DATE__, __TIME__);
	}
 

    //get the local host name formatted as a 0-terminated ASCII string
	if(gethostname(localName, sizeof(localName)) == SOCKET_ERROR)
    {
    	//deal with error
	    throw Nexception(	"UDPNetworking::UDPNetworking() sez: ""gethostname() returned error.""",
							__FILE__, __LINE__, __DATE__, __TIME__);
    }
   
    //get the hostent of the local host...
	hostent *hostEnt;
	if((hostEnt = gethostbyname(localName)) == NULL)
    {
    	//deal with error
		throw Nexception(	"UDPNetworking::UDPNetworking() sez: ""gethostbyname() returned NULL.""", 
							__FILE__, __LINE__, __DATE__, __TIME__);
    }
    //...so that we can retrieve the host's first IP address (though it may have more than one),
    //which is in Big Endian -- the local IP address
	char *charLocalIP = hostEnt->h_addr_list[0];
	localIP = *(reinterpret_cast<u_long *>(charLocalIP));
}


////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	UDPNetworking::~UDPNetworking()
/*!	\brief		Destructor
**	\param		none
**	\return		none
**	\exception	none
**	\par		DESCRIPTION:
				- Frees the entire WinSock subsystem -- closing all sockets that have not been
				closesocket()'ed and discarding any data queued up to received on those sockets
**					
**	\author		Nathan Frost
**	\date		10/28/2002 9:53:54 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
UDPNetworking::~UDPNetworking()
{
    //(a more graceful shutdown might be assured by calling shutdown() first?)
    if(WSACleanup() == SOCKET_ERROR)
    {
    	//deal with error
    	throw Nexception(	"UDPNetworking::~UDPNetworking sez: ""WSACleanup() returned an error""",
							__FILE__, __LINE__, __DATE__, __TIME__);
    }
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	UDPNetworking::MyIPAddress
/*!	\brief		Returns the integral representation of the local IP Address in Little Endian
**	\param		none
**	\return		The local IP Address in Little-Endian
**	\exception	none
**	\par		DESCRIPTION:
**					
**	\author		Nathan Frost
**	\date		10/25/2002 6:36:51 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
u_long UDPNetworking::MyIPAddress()
{
	return ntohl(localIP);
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	UDPNetworking::CreateSock
/*!	\brief		Initializes Socket to a valid, ready for action SOCKET
**	\param		Socket: SOCKET to set up
	\param		Port:	The Port to bind to (in Little Endian).  Set to ANY_PORT to bind to some 
				available port.
**	\return		none
**	\exception	Throws an Nexception on failure
**	\par		DESCRIPTION:
**				- The created SOCKET will be bound to the (one) local IP Address that was recorded
				by UDPNetworking on construction -- no other IP Addresses are possible.
**	\author		Nathan Frost
**	\date		10/25/2002 7:42:48 PM
**	\warning	
				- When binding to a specific port address, always catch Nexceptions in case the 
				port number is not available (and hence the SOCKET was then not able to be properly
				initialized)
*/
////////////////////////////////////////////////////////////////////////////////
void UDPNetworking::CreateSock(SOCKET &Socket, u_short &Port)
{
    //create new socket using the datagram type (UDP). PF_INET address family is used (for
    //historical reasons {supporting multiple protocols within a single address family, which, in
    //actuality, was never implemented})
    if((Socket = socket(PF_INET, SOCK_DGRAM, 0)) == INVALID_SOCKET)
    {
    	//deal with error
        throw Nexception(	"UDPNetworking::CreateSock sez: ""socket() gave me an invalid socket!""", 
							__FILE__, __LINE__, __DATE__, __TIME__);
    }

    /* Bind socket to a port of the local machine */
    struct sockaddr_in sockAddr;
    sockAddr.sin_family = AF_INET;
	sockAddr.sin_addr.s_addr = localIP;	//the particular local IP Address to bind to, in Big Endian
	sockAddr.sin_port = htons(Port);	//converted to Big Endian
    memset(&sockAddr.sin_zero, 0, sizeof(sockAddr.sin_zero));//WinSock likes it this way
    //note that sockaddr_in objects and sockaddr objects can be used interchangeably -- they are
    //the same size, and WinSock treats them compatibly
    if(bind(Socket, reinterpret_cast<sockaddr *>(&sockAddr), sizeof(sockAddr)) == SOCKET_ERROR)
    {
    	//deal with error -- it's possible that the sockAddr specified as already bound and not
		//specified for reuse (SO_REUSEADDR)
        throw Nexception(	"UDPNetworking::CreateSock sez: ""bind() had a SOCKET_ERROR!""", 
							__FILE__, __LINE__, __DATE__, __TIME__);
    }
	//if the caller didn't specify which port to bind to...
    if(Port == 0)
    {
        //...we've now got hold of a port -- find out which port
    	struct sockaddr_in receiveSockName;
    	int sizeOfSockAddr = sizeof(receiveSockName);
    	if(getsockname(	Socket, reinterpret_cast<sockaddr *>(&receiveSockName),
    					&sizeOfSockAddr) == SOCKET_ERROR)
    	{
            //deal with error
    		throw Nexception(	"UDPNetworking::CreateSock sez: ""getsockname() had a SOCKET_ERROR!""", 
								__FILE__, __LINE__, __DATE__, __TIME__);
    	}
    	//getsockname() returns the number of bytes it actually wrote to the receiveSockName buffer
    	if(sizeOfSockAddr != sizeof(receiveSockName))
    	{
        	//getsockname() didn't return the whole name -- deal with error
    		throw Nexception(	"UDPNetworking::CreateSock sez: ""getsockname() didn't return the whole name!""", 
								__FILE__, __LINE__, __DATE__, __TIME__);
    	}
    	//the recvfrom() port number -- converted from Big Endian to Little Endian
    	Port = ntohs(receiveSockName.sin_port);
    }

    //BEG_HACK: memory allocation
    ++numSocketsAllocated;
    //END_HACK: memory allocation

	return;
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	UDPNetworking::DestroySock
/*!	\brief		Frees the resources associated with Socket.
**	\param		Socket: SOCKET to give back to the OS
**	\return		none
**	\exception	Throws a Nexception on failure
**	\par		DESCRIPTION:
**					
**	\author		Nathan Frost
**	\date		10/25/2002 7:56:43 PM
**	\warning	
				- If this function ever throws a Nexception, chances are excellent that Socket was
				invalid
*/
////////////////////////////////////////////////////////////////////////////////
void UDPNetworking::DestroySock(SOCKET &Socket)
{
	if(closesocket(Socket) == SOCKET_ERROR)
    {
    	//deal with error
    	throw Nexception(	"UDPNetworking::DestroySock sez: ""closesocket() returned SOCKET_ERROR", 
							__FILE__, __LINE__, __DATE__, __TIME__);
    }

    //BEG_HACK: memory allocation
    --numSocketsAllocated;
    //END_HACK: memory allocation
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	UDPNetworking::SendPacket
/*!	\brief		Sends the specified Data from the specified Socket to the specified Port on the 
				specified IP Address.
**	\param		Socket:		The local SOCKET to send the Data from
	\param		IPAddress:	The remote IP Address (in Little Endian) to send the Data to
	\param		Port:		The remote Port (at IPAddress, in Little Endian) to send the Data to
	\param		Data:		Pointer to the Data to send
	\param		SizeOfData: The size (in bytes) of the Data to send
**	\return		none
**	\exception	Throws a Nexception on failure
**	\par		DESCRIPTION:
**					
**	\author		Nathan Frost
**	\date		10/25/2002 8:29:58 PM
**	\warning	
				- This is UDP; we must be careful that MAX_PACKET_SIZE < SO_MAX_MSG_SIZE -- always.
				- A Nexception is thrown if SizeOfData > MAX_PACKET_SIZE
*/
////////////////////////////////////////////////////////////////////////////////
void UDPNetworking::SendPacket(	const SOCKET &Socket, const u_long &IPAddress, const u_short &Port,
								void *Data, const unsigned int SizeOfData)
{
	//BEGIN_HACK
	unsigned int docNum = 0;
	//extract DocumentNum
	reinterpret_cast<unsigned char *>(&docNum)[0] = ((char *)Data)[0];
	reinterpret_cast<unsigned char *>(&docNum)[1] = ((char *)Data)[1];
	reinterpret_cast<unsigned char *>(&docNum)[2] = ((char *)Data)[2];
	reinterpret_cast<unsigned char *>(&docNum)[3] = ((char *)Data)[3];

	unsigned char pageNum = ((char *)Data)[4];

//	std::cout << "Sent DocumentNum: " << docNum << std::endl;
	//END_HACK



	/* Fill out the address of the machine we're sending to */
	sockaddr_in dest;
    dest.sin_family = AF_INET;
    dest.sin_port = htons(Port);					//Port to send in (converted to Big Endian)
    dest.sin_addr.S_un.S_addr = htonl(IPAddress);	//IP Address to send to (converted to Big Endian)
    memset(&dest.sin_zero, 0, sizeof(dest.sin_zero));//WinSock likes it this way

    //This is UDP: be careful not to exceed the maximum packet size (SO_MAX_MSG_SIZE)
	if(SizeOfData > MAX_PACKET_SIZE)
		throw Nexception(	"UDPNetworking::SendPacket sez: ""SizeOfData > MAX_PACKET_SIZE!""",
							__FILE__, __LINE__, __DATE__, __TIME__);

    int numBytesSent;
    if((numBytesSent = sendto(	Socket, reinterpret_cast<char *>(Data), SizeOfData, 0,
    							reinterpret_cast<sockaddr *>(&dest), sizeof(dest))) < 0)
    {
        int err = WSAGetLastError();
        if(err == WSAEDESTADDRREQ || err == WSAEADDRNOTAVAIL)
        {
            //these errors can be recovered from
    	    throw Nexception(	"UDPNetworking::SendPacket sez: Invalid IP Address -- restart GameNetworking",
							    __FILE__, __LINE__, __DATE__, __TIME__);
        }
        else
        {
    	    //no idea how to handle any other error
    	    throw Nexception(	"UDPNetworking::SendPacket sez: ""sendto() failed!""",
			    				__FILE__, __LINE__, __DATE__, __TIME__);
        }
    }

	//make sure all the data was sent properly
    if(numBytesSent != SizeOfData)
    {
    	//not all the data was sent
    	throw Nexception(	"UDPNetworking::SendPacket sez: ""Not all of the data was sent!""", 
							__FILE__, __LINE__, __DATE__, __TIME__);
    }

    return;
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	UDPNetworking::ReceivePacket
/*!	\brief		Data is received off of the specified SOCKET and the sender's IP address and port
				are returned as well.  Blocks until data becomes available on the specified SOCKET.
**	\param		Socket:			SOCKET to read data from.
	\param		Buffer:			Pointer to the memory buffer to put the received data in.
	\param		SizeOfBuffer:	The size, in bytes, of the memory buffer to put the received data in
	\param		SenderIP:		The IP Address (in Little Endian) of the sender of the received data
	\param		SenderPort:		The Port (in Little Endian) the sender of the received data sent from
**	\return		The number of bytes actually read into Buffer.
**	\exception	Throws a Nexception on failure
**	\par		DESCRIPTION:
				- Blocks control flow until data becomes available to be read from the SOCKET.
**
**	\author
**	\date		10/25/2002 9:54:47 PM
**	\warning
				This is a blocking call.
*/
////////////////////////////////////////////////////////////////////////////////
int UDPNetworking::ReceivePacket(	const SOCKET &Socket, void *Buffer, 
									const unsigned int SizeOfBuffer, u_long &SenderIP, 
									u_short &SenderPort)
{
    sockaddr_in source;                               
    memset(&source, 0, sizeof(source));	//recvfrom() will tell us who sent us this message
    int sizeOfSource = sizeof(source);	//let recvfrom() know how big source is; recvfrom() will
       									//then overwrite this value with the number of bytes it
                                    	//actually wrote to the source buffer

    int err;//debug variable
	if(err = recvfrom(  Socket, reinterpret_cast<char *>(Buffer), SizeOfBuffer, 0,
				        reinterpret_cast<sockaddr *>(&source), &sizeOfSource) != SOCKET_ERROR)
	{
		SenderIP = ntohl(source.sin_addr.S_un.S_addr);
		SenderPort = ntohs(source.sin_port);
		return sizeOfSource;//mission accomplished
	}
	else
	{
		throw Nexception(	"UDPNetworking::ReceivePacket sez: ""recvfrom() returned SOCKET_ERROR!""",
							__FILE__, __LINE__, __DATE__, __TIME__);
	}

	return 0;//should never reach here
}

