////////////////////////////////////////////////////////////////////////////////
/*!	\file		Nexception.h
**	\brief		Declaration of class Nexception and its ancillary data and 
				functions.
**	\author		Nathan Frost
**	\date		07/04/2002 3:54:00 PM
**	\par		DESCRIPTION:	
								Nexception is intended to be used as a base 
								exception class.  It encapsulates some basic
								information that pertains to the context in
								which the exception was thrown.
**					
**	\par		REMARKS:	
							Nexception never throws an exception, and is 
							bullet-proof to any reasonable usage, barring a 
							stack overflow.
**	\par		UPDATED ON:			10/18/2002 10:17:57 AM
**	\par		UPDATED BY:			John Corpening
**	\par		UPDATE DESCRIPTION: 
							Added exception logging using the debugging module
**  \par        UPDATED ON:         11/7/2002 5:59:48 PM
**  \par        UPDATED BY:         M Jared Finder
**  \par        UPDATE DESCRIPTION: 
**                  Added default arguments for file, line, date, and time.
**                  This makes the Nexception easier to derive from for
**                  expected classes of exceptions.
**					
*/
////////////////////////////////////////////////////////////////////////////////
#pragma once

#include <iostream>
#include <string>
#include "../Debugging/Debug.h"		//Update: John Corpening, 10/18/2002


/*-------------------------------------------Constants-------------------------------------------*/
/* Length of text fields in Concrete Nexception */
//every text field length includes 0 terminator
const int MAX_LENGTH = 161;			///<for all text fields
const int DATE_MAX_LENGTH = 12;		///<__DATE__
const int TIME_MAX_LENGTH = 9;		///<__TIME__

const char NEWLINE[3] = "\r\n";	///<newline character for Win2K environment

/*----------------------------Class and Related Function Declarations----------------------------*/
//Abstract base class
////////////////////////////////////////////////////////////////////////////////
/*!	\class		NexceptionInterface
**	\brief		Abstract base class that defines the interface for concrete,
				derived Nexceptions
**	\author		Nathan Frost
**	\date		07/04/2002 3:54:00 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
class NexceptionInterface
{
public:
	/* accessors to data fields -- informs the programmer about the exception that occurred */
	//Note that all data fields other than Remark are intended to be passed the __FILE__, __LINE__,
	//__DATE__ and __TIME__ macros in the constructor.  Other usage is not directly supported.
	virtual const char *Remark() const throw() = 0;		//what the thrower has to say for himself
	virtual const char *File() const throw() = 0;		//__FILE__ that threw the exception
	virtual const unsigned int Line() const throw() = 0;//__LINE__ that threw the exception
	virtual const char *Date() const throw() = 0;		//__DATE__ File() was last compiled
	virtual const char *Time() const throw() = 0;		//__TIME__ File() was last compiled
	virtual ~NexceptionInterface() throw(){};
};

////////////////////////////////////////////////////////////////////////////////
/*!	\class		Nexception
**	\brief		Concrete implementation of NexceptionInterface
**	\author		Nathan Frost
**	\date		07/04/2002 3:54:00 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
class Nexception:public NexceptionInterface
{
public:
	//implementation of abstract interface
	const char *Remark() const throw();
	const char *File() const throw();
	const unsigned int Line() const throw();
	const char *Date() const throw();
	const char *Time() const throw();

	//constructor
	Nexception(	const char *Remark,
                const char *File        = "",
                const unsigned int Line = 0,
                const char *Date        = "",
				const char *Time        = ""
                ) throw();
	//copy constructor
	Nexception(const Nexception &nexception) throw();
	//overloaded assignment
 	Nexception &operator =(const Nexception &nexception) throw();

private:
	///stack-based representation guarantees that no free-store exceptions (such as bad_alloc) are 
	///thrown
	char remark[MAX_LENGTH];
	char file[MAX_LENGTH];
	unsigned int line;
	char date[DATE_MAX_LENGTH];
	char time[TIME_MAX_LENGTH];
};

/* overloaded extraction operators for ease of outputting */
//to std::ostream
using std::ostream;
ostream &operator << (ostream &os, Nexception &nexception);
using std::string;
//to std::string
string &operator << (string &str, Nexception &nexception);
/*-----------------------------------------------------------------------------------------------*/
