////////////////////////////////////////////////////////////////////////////////
/*! \file           HeightMap.cpp
**  \brief          HeightMap implementation file
**  \author         Austin Spafford
**  \date           10/11/2002 6:11:04 PM
**  \par            DESCRIPTION:
**                    The HeightMap object is capable of storing rasterized data
**                    points that describe the height a rectangular piece of
**                    terrain. 
**  \par            REMARKS:
**                    
*/
////////////////////////////////////////////////////////////////////////////////
#pragma warning (disable : 4786)
////////////////////////////////////////////////////////////////////////////////
#include <stdlib.h>
#include <memory.h>
#include <assert.h>
#include <math.h>
#include "loader/Image.h"
#include "nexception/Nexception.h"
#include "HeightMap.h"
#include "networking/NetData.h"
// memory leak detection
#ifdef _DEBUG
    #include"memleak.h"
    #define new DEBUG_NEW
    #undef THIS_FILE
    static char THIS_FILE[] = __FILE__;
#endif
////////////////////////////////////////////////////////////////////////////////
using namespace math::linear_alg;
using namespace UDPNet;
////////////////////////////////////////////////////////////////////////////////
const unsigned int c_elementsPerNetMessageChunk = 32; // used for networking
////////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////
//  FUNCTION:       
/*! \brief          returns the input values after they are restricted to a
**					minimum value pair
**  \param          inputValue the value to be limited
**  \param          minimumValue the value to be limit inputValue to
**  \return         a limited version of inputValue
**  \exception      none
**  \author         Austin Spafford
**  \date           10/11/2002 8:40:14 PM
**  \warning        none
*/
////////////////////////////////////////////////////////////////////////////////
template <typename type>
static Vector2<type> LimitedToMinimum(Vector2<type> inputValue, Vector2<type> minimumValue)
{
	Vector2<type> returnValue;

	if (inputValue[0] < minimumValue[0])
	{
		returnValue[0] = minimumValue[0];
	}
	else
	{
		returnValue[0] = inputValue[0];
	}

	if (inputValue[1] < minimumValue[1])
	{
		returnValue[1] = minimumValue[1];
	}
	else
	{
		returnValue[1] = inputValue[1];
	}

	return returnValue;
}

////////////////////////////////////////////////////////////////////////////////
//  FUNCTION:       
/*! \brief          returns the input values after they are restricted to a
**					maximum value pair
**  \param          inputValue the value to be limited
**  \param          maximumValue the value to be limit inputValue to
**  \return         a limited version of inputValue
**  \exception      none            
**  \author         Austin Spafford
**  \date           10/11/2002 8:40:14 PM
**  \warning        none
*/
////////////////////////////////////////////////////////////////////////////////
template <typename type>
static Vector2<type> LimitedToMaximum(Vector2<type> inputValue, Vector2<type> maximumValue)
{
	Vector2<type> returnValue;

	if (inputValue[0] > maximumValue[0])
	{
		returnValue[0] = maximumValue[0];
	}
	else
	{
		returnValue[0] = inputValue[0];
	}

	if (inputValue[1] > maximumValue[1])
	{
		returnValue[1] = maximumValue[1];
	}
	else
	{
		returnValue[1] = inputValue[1];
	}

	return returnValue;
}

////////////////////////////////////////////////////////////////////////////////
//  FUNCTION:       
/*! \brief          returns the input values after they are restricted to a
**					min/max value pair
**  \param          inputValue the value to be limited
**  \param          minimumValue a value to be limit inputValue to
**  \param          maximumValue a value to be limit inputValue to
**  \return         a limited version of inputValue
**  \exception      none            
**  \author         Austin Spafford
**  \date           10/11/2002 8:40:14 PM
**  \warning        none
*/
////////////////////////////////////////////////////////////////////////////////
template <typename type>
static Vector2<type> LimitedToMinMax(Vector2<type> inputValue, Vector2<type> minimumValue, Vector2<type> maximumValue)
{
	Vector2<type> returnValue;

	if (inputValue[0] < minimumValue[0])
	{
		returnValue[0] = minimumValue[0];
	}
	else 
	{
		if (inputValue[0] > maximumValue[0])
		{
			returnValue[0] = maximumValue[0];
		}
		else
		{
			returnValue[0] = inputValue[0];
		}
	}

	if (inputValue[1] < minimumValue[1])
	{
		returnValue[1] = minimumValue[1];
	}
	else
	{
		if (inputValue[1] > maximumValue[1])
		{
			returnValue[1] = maximumValue[1];
		}
		else
		{
			returnValue[1] = inputValue[1];
		}
	}

	return returnValue;
}

////////////////////////////////////////////////////////////////////////////////
//  FUNCTION:       
/*! \brief          default constructer, generates a single cell HeightMap
**  \param          
**  \return         none
**  \exception      none
**  \par            DESCRIPTION:
**                    
**  \author         Austin Spafford
**  \date           10/11/2002 8:20:34 PM
**  \warning        none
*/
////////////////////////////////////////////////////////////////////////////////
HeightMap::HeightMap() :
		m_positionOffset(0,0,0),
		m_pointCount(2,2),
		m_cellDimensions(1,1),
		m_dataPoints(NULL),
		m_normals(NULL),
        m_verticesChanged(true), m_spaceTransformationChanged(true)
{	
	ReallocateDynamicMemory();
	RecalculateAllNormals();

	return;
}

////////////////////////////////////////////////////////////////////////////////
//  FUNCTION:       
/*! \brief          detailed constructer
**  \param          footPrint designates the X/Y dimensions of the new HeightMap
**  \param          pointCount designates the X/Y resolution of the new HeightMap
**  \return         none
**  \exception      none
**  \par            DESCRIPTION:
**                    
**  \author         Austin Spafford
**  \date           10/11/2002 8:21:15 PM
**  \warning        none
*/
////////////////////////////////////////////////////////////////////////////////
HeightMap::HeightMap(Vector2<float> footPrint, Vector2<int> pointCount) :
		m_positionOffset(0,0,0),
		m_pointCount(pointCount),
		m_dataPoints(NULL),
		m_normals(NULL),
        m_verticesChanged(true), m_spaceTransformationChanged(true)
{	
	assert(pointCount == LimitedToMinimum(pointCount, Vector2<int>(2,2)));	

	m_cellDimensions = Vector2<float>(footPrint[0]/(m_pointCount[0]-1),footPrint[1]/(m_pointCount[1]-1));

	ReallocateDynamicMemory();
	RecalculateAllNormals();

	return;
}

////////////////////////////////////////////////////////////////////////////////
//  FUNCTION:       
/*! \brief          deconstructer
**  \param          
**  \return         none
**  \exception      none
**  \par            DESCRIPTION:
**                    
**  \author         Austin Spafford
**  \date           10/11/2002 9:45:51 PM
**  \warning        none
*/
////////////////////////////////////////////////////////////////////////////////
HeightMap::~HeightMap()
{
	if (NULL != m_dataPoints)
	{
		delete m_dataPoints;
		m_dataPoints = NULL;
	}

	if (NULL != m_normals)
	{
		delete m_normals;
		m_normals = NULL;
	}


	return;
}

////////////////////////////////////////////////////////////////////////////////
//  FUNCTION:       
/*! \brief          loads the specified image file and maps its red channel to 
**                  the height values of the height map
**  \param          fileName a pointer to a NULL terminated string that specifies 
**                  the file that should be read from
**  \param          maxHeight the value that a red value of 255 should be mapped
**                  to once the image is loaded
**  \return         returns true if the image was properly imported, false otherwise
**  \exception      none
**  \author         Austin Spafford
**  \date           11/9/2002 5:24:10 PM
**  \warning        Also changes the resolution of the height map to that of the image.
*/
////////////////////////////////////////////////////////////////////////////////
bool HeightMap::ImportFromImage(char * fileName, float maxHeight)
{
	loader::Image importImage;

	if(false == importImage.Load(fileName))
	{
		return false;
	}

	// save the current footprint
	Vector2<float> oldFootPrint = GetFootPrint();

	// resize the height field to match the resolution of the image
	Vector2<int> newPointCount(importImage.GetWidth(), importImage.GetHeight());
	SetPointCount(newPointCount);

	// resize the cell size so the old footprint will be used
	SetFootPrint(oldFootPrint);

	// copy over the image data
	Vector2<int> currentPointIndex;
	for (currentPointIndex[0]=0; currentPointIndex[0]<newPointCount[0]; currentPointIndex[0]++)
	{
		for (currentPointIndex[1]=0; currentPointIndex[1]<newPointCount[1]; currentPointIndex[1]++)
		{
			unsigned char pixelValue = importImage.GetData()[4*(currentPointIndex[0] + newPointCount[0]*currentPointIndex[1])]; 
			float pointValue = (maxHeight*pixelValue) / 255.0f;
			SetPointValue(currentPointIndex, pointValue);
		}
	}	

	return true;
}
	
////////////////////////////////////////////////////////////////////////////////
//  FUNCTION:       
/*! \brief          saves a gray scale image of the height map, with white being
**                  the highest elevation
**  \param          fileName a pointer to a NULL terminated string that specifies 
**                  the file that should be written to
**  \return         returns true if the image was properly exported, false otherwise
**  \exception      none
**  \author         Austin Spafford
**  \date           11/9/2002 5:29:23 PM
**  \warning        The resolution of the image will be that of the height map
*/
////////////////////////////////////////////////////////////////////////////////
bool HeightMap::ExportToImage(char * fileName)
{	
	char * heightImageDataPtr = new char [4*(GetPointCount()[0] * GetPointCount()[1])];	

	if (NULL == heightImageDataPtr)
	{
		return false;
	}

	Vector2<int> currentPointIndex;
	float maxHeightValue = 1;	

	// find the max height value	
	for (currentPointIndex[0]=0; currentPointIndex[0]<GetPointCount()[0]; currentPointIndex[0]++)
	{
		for (currentPointIndex[1]=0; currentPointIndex[1]<GetPointCount()[1]; currentPointIndex[1]++)
		{			
			float pointValue = GetPointValue(currentPointIndex);

			if (maxHeightValue < pointValue)
			{
				maxHeightValue = pointValue;
			}
		}
	}

	// copy over the height map data	
	for (currentPointIndex[0]=0; currentPointIndex[0]<GetPointCount()[0]; currentPointIndex[0]++)
	{
		for (currentPointIndex[1]=0; currentPointIndex[1]<GetPointCount()[1]; currentPointIndex[1]++)
		{			
			float pointValue = GetPointValue(currentPointIndex);
			char * currentPixelPtr = heightImageDataPtr + 4*(currentPointIndex[0] + GetPointCount()[0]*currentPointIndex[1]);

			if (pointValue < 0)
			{
				pointValue = 0;
			}

			char pixelValue = char((pointValue/maxHeightValue) * 255);			

			currentPixelPtr[0] = pixelValue;	// red
			currentPixelPtr[1] = pixelValue;	// green
			currentPixelPtr[2] = pixelValue;	// blue
			currentPixelPtr[3] = char(255);			// alpha
		}
	}

	// save the 32 bit image to the specified file
	loader::Image exportImage;
	bool loadImageResult = exportImage.Load(heightImageDataPtr, GetPointCount()[0], GetPointCount()[1]);
	delete [] heightImageDataPtr;

	// check to see if the exportImage failed to load from the temp data buffer
	if (false == loadImageResult)
	{				
		return false;
	}

	// check to see if the exportImage failed to save to disk
	if (false == exportImage.Save(fileName))
	{				
		return false;
	}
	else
	{		
		return true;
	}
}

////////////////////////////////////////////////////////////////////////////////
//  FUNCTION:       
/*! \brief          Updates the state of the HeightMap by means of a NetMessage
**                  that originated on a sister HeightMap by means of a call
**                  to GenerateNetMessage().
**  \param          netMessageRef a reference to the NetMessage to be read from
**  \return         none
**  \exception      none
**  \author         Austin Spafford
**  \date           3/5/2003 3:44:43 PM
**  \warning        none
*/
////////////////////////////////////////////////////////////////////////////////
void HeightMap::LoadNetMessage(NetMessage & netMessageRef)
{
	// We by loading the point count, cell dimensions, and position offset
	Vector2<int> newPointCount;
	Vector2<float> newCellDimensions;
	Vector3<float> newPositionOffset;
	netMessageRef.ReadNBytes(sizeof(Vector2<int>), &newPointCount);
	netMessageRef.ReadNBytes(sizeof(Vector2<float>), &newCellDimensions);
	netMessageRef.ReadNBytes(sizeof(Vector3<float>), &newPositionOffset);	
	SetPointCount(newPointCount);		
	SetCellDimensions(newCellDimensions);
	SetPositionOffset(newPositionOffset);

	// Now we just load the vertex values in small chunks until the
	// entire HeightMap has been loaded
	unsigned int indexOfFirstElementOfCurrentChunk;
	unsigned int numberOfElements = m_pointCount[0] * m_pointCount[1];	
	// load each whole vertex chunk
	for (
		indexOfFirstElementOfCurrentChunk = 0;
		indexOfFirstElementOfCurrentChunk <= numberOfElements - c_elementsPerNetMessageChunk;
		indexOfFirstElementOfCurrentChunk += c_elementsPerNetMessageChunk
		)
	{
		// obtain a pointer to the vertex chunk
		float * currentChunkPtr = m_dataPoints + indexOfFirstElementOfCurrentChunk;

		// load the vertex chunk from the NetMessage
		netMessageRef.ReadNBytes(sizeof(float) * c_elementsPerNetMessageChunk, currentChunkPtr);
	}
	// check to see if a partial chunk is left
	if (indexOfFirstElementOfCurrentChunk < numberOfElements)
	{
		// obtain a pointer to the vertex chunk
		float * currentChunkPtr = m_dataPoints + indexOfFirstElementOfCurrentChunk;

		// load the final, and partial, vertex chunk to the NetMessage
		netMessageRef.ReadNBytes(sizeof(float) * (numberOfElements - indexOfFirstElementOfCurrentChunk), currentChunkPtr);
	}

	return;
}

////////////////////////////////////////////////////////////////////////////////
//  FUNCTION:       
/*! \brief          Appends data to  a NetMessage that will encode the current 
**                  state of the HeightMap, so that a sister HeightMap can load
**                  it by means of a call to LoadNetMessage().
**  \param          netMessageRef a reference to the NetMessage to append to
**  \return         none
**  \exception      none
**  \author         Austin Spafford
**  \date           3/5/2003 3:51:59 PM
**  \warning        none
*/
////////////////////////////////////////////////////////////////////////////////
void HeightMap::GenerateNetMessage(NetMessage & netMessageRef)
{
	// We first save the point count, cell dimensions, and position offset
	netMessageRef.WriteNBytes(sizeof(Vector2<int>), &m_pointCount);
	netMessageRef.WriteNBytes(sizeof(Vector2<float>), &m_cellDimensions);
	netMessageRef.WriteNBytes(sizeof(Vector3<float>), &m_positionOffset);

	// Now we just save the vertex values in small chunks until the
	// entire HeightMap has been saved
	unsigned int indexOfFirstElementOfCurrentChunk;
	unsigned int numberOfElements = m_pointCount[0] * m_pointCount[1];	
	// write each whole vertex chunk
	for (
		indexOfFirstElementOfCurrentChunk = 0;
		indexOfFirstElementOfCurrentChunk <= numberOfElements - c_elementsPerNetMessageChunk;
		indexOfFirstElementOfCurrentChunk += c_elementsPerNetMessageChunk
		)
	{
		// obtain a pointer to the vertex chunk
		float * currentChunkPtr = m_dataPoints + indexOfFirstElementOfCurrentChunk;

		// write the vertex chunk to the NetMessage
		netMessageRef.WriteNBytes(sizeof(float) * c_elementsPerNetMessageChunk, currentChunkPtr);
	}
	// check to see if a partial chunk is left
	if (indexOfFirstElementOfCurrentChunk < numberOfElements)
	{
		// obtain a pointer to the vertex chunk
		float * currentChunkPtr = m_dataPoints + indexOfFirstElementOfCurrentChunk;

		// write the final, and partial, vertex chunk to the NetMessage
		netMessageRef.WriteNBytes(sizeof(float) * (numberOfElements - indexOfFirstElementOfCurrentChunk), currentChunkPtr);
	}

	return;
}

////////////////////////////////////////////////////////////////////////////////
//  FUNCTION:       
/*! \brief          A factory method that generates an object that encapsulates 
**                  the state of the HeightMap.
**  \return         A pointer to a new HeightMapFileMemento object.
**  \exception      none
**  \author         Austin Spafford
**  \date           5/15/2003 3:54:55 PM
**  \warning        none
*/
////////////////////////////////////////////////////////////////////////////////
HeightMapFileMemento * HeightMap::CreateFileMemento(void) const
{
	HeightMapFileMemento * newFileMomentoPtr = new HeightMapFileMemento;

	// save the point count, cell dimensions, and position offset
	newFileMomentoPtr->m_pointCount = m_pointCount;
	newFileMomentoPtr->m_cellDimensions = m_cellDimensions;
	newFileMomentoPtr->m_positionOffset = m_positionOffset;	

	// save the heightmap points
	newFileMomentoPtr->m_dataPointsBuffer = new float[m_pointCount[0] * m_pointCount[1]];
	memcpy(newFileMomentoPtr->m_dataPointsBuffer, m_dataPoints, sizeof(float) * m_pointCount[0] * m_pointCount[1]);

	return newFileMomentoPtr;
}

////////////////////////////////////////////////////////////////////////////////
//  FUNCTION:       
/*! \brief          Sets the current state of the HeightMap to that which is
**                  stored in fileMemento.
**  \param          fileMemento the object that will be used to set the current
**                  state of the HeightMap.
**  \return         none
**  \exception      none
**  \author         Austin Spafford
**  \date           5/15/2003 4:32:15 PM
**  \warning        none
*/
////////////////////////////////////////////////////////////////////////////////
void HeightMap::LoadFileMemento(const HeightMapFileMemento & fileMemento)
{
	// load the point count, cell dimensions, and position offset
	SetPointCount(fileMemento.m_pointCount);		
	SetCellDimensions(fileMemento.m_cellDimensions);
	SetPositionOffset(fileMemento.m_positionOffset);
	
	// load the heightmap points	
	memcpy(m_dataPoints, fileMemento.m_dataPointsBuffer, sizeof(float) * m_pointCount[0] * m_pointCount[1]);

	// make sure to trigger these
	m_verticesChanged = true;
	m_spaceTransformationChanged = true;
	m_normalsAreCorrect = false;
	m_dataHasChanged = true;

	return;
}

////////////////////////////////////////////////////////////////////////////////
//  FUNCTION:       
/*! \brief          retrieves a point's value
**  \param          pointIndex a two dimensional index into the HeightMap
**  \return         a float representing the specified point's value
**  \exception      none
**  \par            DESCRIPTION:
**                    
**  \author         Austin Spafford
**  \date           10/11/2002 9:46:04 PM
**  \warning        none
*/
////////////////////////////////////////////////////////////////////////////////
float HeightMap::GetPointValue(Vector2<int> pointIndex)
{	
	assert(pointIndex == LimitedToMinMax(pointIndex, Vector2<int>(0,0), Vector2<int>(m_pointCount[0]-1, m_pointCount[1]-1)));

	return m_dataPoints[pointIndex[0] + pointIndex[1]*(m_pointCount[0])];	
}

////////////////////////////////////////////////////////////////////////////////
//  FUNCTION:       
/*! \brief          Vertically projects a point onto the HeightMap and returns
**                  the resulting height.
**  \param          pointToProject a position that is projected onto the HeightMap.
**  \return         The height of the resulting projection (its Y value).
**  \exception      none
**  \author         Austin Spafford
**  \date           3/25/2003 8:46:25 PM
**  \warning        none
*/
////////////////////////////////////////////////////////////////////////////////
float HeightMap::GetProjectedHeight(const Vector3<float> & pointToProject)
{
	Vector2<float> pointPositionOnHeightMap;
	Vector2<int> cellIndex;

	// find out which cell the projected point lies in
	pointPositionOnHeightMap[0] = pointToProject[0] - m_positionOffset[0];
	pointPositionOnHeightMap[1] = pointToProject[2] - m_positionOffset[2];
	cellIndex[0] = int(floor(pointPositionOnHeightMap[0] / m_cellDimensions[0]));
	cellIndex[1] = int(floor(pointPositionOnHeightMap[1] / m_cellDimensions[1]));

	// limit the cell index
	for (int i=0; i<2; i++)
	{
		if (cellIndex[i] < 0)
		{
			cellIndex[i] = 0;
		}

		if (cellIndex[i] > m_pointCount[i] - 2)
		{
			cellIndex[i] = m_pointCount[i] - 2;
		}
	}

	// redefine the point's position in terms of the current cell's position
	pointPositionOnHeightMap[0] -= cellIndex[0] * m_cellDimensions[0];
	pointPositionOnHeightMap[1] -= cellIndex[1] * m_cellDimensions[1];

	// find out which triangle the projected point is in
	HeightMap::e_TriangleMajor triangleMajor;
	if (pointPositionOnHeightMap[0] > pointPositionOnHeightMap[1])
	{
		triangleMajor = HeightMap::e_xMajorTriangle;
	}
	else
	{
		triangleMajor = HeightMap::e_yMajorTriangle;
	}

	// obtain a reference to the triangle's normal
	if (false == m_normalsAreCorrect)
	{
		RecalculateAllNormals();
	}	
	Vector3<float> & triangleNormal = GetNormalReference(cellIndex, triangleMajor);

	// find the height of the projection relative to the index point
	float relativeHeight =
			-1*(
				triangleNormal[0] * pointPositionOnHeightMap[0]
				+
				triangleNormal[2] * pointPositionOnHeightMap[1]
			)
			/ triangleNormal[1];

	// return the height relative to the world
	return relativeHeight + GetPointValue(cellIndex) + m_positionOffset[1];
}

////////////////////////////////////////////////////////////////////////////////
//  FUNCTION:       
/*! \brief          retrieves a point's location
**  \param          pointIndex a two dimensional index into the HeightMap
**  \return         a three dimensional vector representing the specified
**                  point's position
**  \exception      none
**  \par            DESCRIPTION:
**                    
**  \author         Austin Spafford
**  \date           10/11/2002 9:47:15 PM
**  \warning        none
*/
////////////////////////////////////////////////////////////////////////////////
Vector3<float> HeightMap::GetPointLocation(Vector2<int> pointIndex)
{	
	assert(pointIndex == LimitedToMinMax(pointIndex, Vector2<int>(0,0), Vector2<int>(m_pointCount[0]-1, m_pointCount[1]-1)));

	return Vector3<float>(
			pointIndex[0] * m_cellDimensions[0],		
			GetPointValue(pointIndex),
			pointIndex[1] * m_cellDimensions[1]		
			)
			+
			m_positionOffset;
}

////////////////////////////////////////////////////////////////////////////////
//  FUNCTION:       
/*! \brief          retrieves a point's location, but with out the position 
**                  offset of the HeightMap being factored in
**  \param          pointIndex a two dimensional index into the HeightMap
**  \return         a three dimensional vector representing the specified
**                  point's position, BUT WITHOUT THE HEIGHTMAP'S OFFSET!
**  \exception      none
**  \par            DESCRIPTION:
**                 
**  \author         Austin Spafford
**  \date           11/4/2002 11:32:55 AM
**  \warning        none
*/
////////////////////////////////////////////////////////////////////////////////
Vector3<float> HeightMap::GetPointLocationWithoutOffset(Vector2<int> pointIndex)
{	
	assert(pointIndex == LimitedToMinMax(pointIndex, Vector2<int>(0,0), Vector2<int>(m_pointCount[0]-1, m_pointCount[1]-1)));

	return Vector3<float>(
			pointIndex[0] * m_cellDimensions[0],		
			GetPointValue(pointIndex),
			pointIndex[1] * m_cellDimensions[1]		
			);
}

////////////////////////////////////////////////////////////////////////////////
//  FUNCTION:       
/*! \brief          returns the value of a normal that is indicated by its 
**                  "smallest" vetice's index and its axis major
**  \param          pointIndex specifies the vertex of the triangle that has the
**                  smallest possible indice values.
**  \param          triangleMajor is used to differentiate between the two
**                  triangles that share the same index.
**  \return         none
**  \exception      none
**  \par            DESCRIPTION:
**                    
**  \author         Austin Spafford
**  \date           10/13/2002 9:58:56 PM
**  \warning        
*/
////////////////////////////////////////////////////////////////////////////////
Vector3<float> HeightMap::GetNormalVector(Vector2<int> pointIndex, e_TriangleMajor triangleMajor)
{
	if (false == m_normalsAreCorrect)
	{
		RecalculateAllNormals();
	}

	return GetNormalReference(pointIndex, triangleMajor);
}


////////////////////////////////////////////////////////////////////////////////
//  FUNCTION:   
/*! \brief      Gets the normal corresponding to a vertex.  Since points do not
**              have explicit normals, one is calculated by a weighted average
**              of the planes surrounding the normal
**  \param      pointIndex  The index of the vertex specified.  It is assumed
**                          that this vertex is inside the heightmap.
**  \return     A vertex normal
**  \exception  none
**  \author     M Jared Finder
**  \date       4/16/2003 2:18:20 PM
**  \warning    none
*/
////////////////////////////////////////////////////////////////////////////////
Vector3<float> HeightMap::GenerateVertexNormalVector(Vector2<int> pointIndex)
{
    if(!m_normalsAreCorrect)
    {
        RecalculateAllNormals();
    }

    Vector3<float> vertexNormal(0,0,0);
    float numAdjacentPlanes(0);
    const int xMin = 0;
    const int yMin = 0;
    const int xMax = m_pointCount[0] - 1;
    const int yMax = m_pointCount[1] - 1;

    //planes shared to the lower left
    if(pointIndex[0] > xMin && pointIndex[1] > yMin)
    {
        numAdjacentPlanes += 2;
        vertexNormal += GetNormalVector(
                                pointIndex - Vector2<int>(1,1),
                                e_xMajorTriangle
                                );
        vertexNormal += GetNormalVector(
                                pointIndex - Vector2<int>(1,1),
                                e_yMajorTriangle
                                );
    }
    //planes shared to the lower right
    if(pointIndex[0] < xMax && pointIndex[1] > yMin)
    {
        numAdjacentPlanes += 1;
        vertexNormal += GetNormalVector(
                            pointIndex - Vector2<int>(0,1),
                            e_yMajorTriangle
                            );
    }
    //planes shared to the upper left
    if(pointIndex[0] > xMin && pointIndex[1] < yMax)
    {
        numAdjacentPlanes += 1;
        vertexNormal += GetNormalVector(
                            pointIndex - Vector2<int>(1,0),
                            e_xMajorTriangle
                            );
    }
    //planes shared to the upper right
    if(pointIndex[0] < xMax && pointIndex[1] < yMax)
    {
        numAdjacentPlanes += 2;
        vertexNormal += GetNormalVector(
                            pointIndex - Vector2<int>(0,0),
                            e_xMajorTriangle);
        vertexNormal += GetNormalVector(
                            pointIndex - Vector2<int>(0,0),
                            e_yMajorTriangle);
    }
    
    return (vertexNormal * (1.f/numAdjacentPlanes));
}

////////////////////////////////////////////////////////////////////////////////
//  FUNCTION:       
/*! \brief          retrieves the point count of the HeightMap
**  \param          
**  \return         a two dimensional vector representing the number of point
**                  values that are stored in each dimension
**  \exception      none
**  \par            DESCRIPTION:
**                   
**  \author         Austin Spafford
**  \date           10/11/2002 9:48:25 PM
**  \warning        none
*/
////////////////////////////////////////////////////////////////////////////////
Vector2<int> HeightMap::GetPointCount(void)
{
	return m_pointCount;
}

////////////////////////////////////////////////////////////////////////////////
//  FUNCTION:       
/*! \brief          retrieves the footprint of the HeightMap
**  \param          
**  \return         a two dimensional vector representing the total space that
**                  is taken up by the HeightMap (excluding height).
**  \exception      none
**  \par            DESCRIPTION:
**                    
**  \author         Austin Spafford
**  \date           10/11/2002 9:49:40 PM
**  \warning        none
*/
////////////////////////////////////////////////////////////////////////////////
Vector2<float> HeightMap::GetFootPrint(void)
{
	return Vector2<float>(
			(m_pointCount[0] - 1) * m_cellDimensions[0], 
			(m_pointCount[1] - 1) * m_cellDimensions[1]
			);
}

////////////////////////////////////////////////////////////////////////////////
//  FUNCTION:       
/*! \brief          retrieves the cell dimensions of the HeightMap
**  \param          
**  \return         a two dimensional vector representing the space that
**                  is taken up by each cell of the HeightMap (excluding height).
**  \exception      none
**  \par            DESCRIPTION:
**                    
**  \author         Austin Spafford
**  \date           10/20/2002 9:49:40 PM
**  \warning        none
*/
////////////////////////////////////////////////////////////////////////////////
Vector2<float> HeightMap::GetCellDimensions(void)
{
	return m_cellDimensions;
}

////////////////////////////////////////////////////////////////////////////////
//  FUNCTION:       
/*! \brief          retrieves the position offset of the HeightMap
**  \param          
**  \return         a three dimensional vector representing the vector that is
**                  added to each point location that is requested by the client
**  \exception      none
**  \par            DESCRIPTION:
**                    
**  \author         Austin Spafford
**  \date           10/11/2002 9:49:40 PM
**  \warning        none
*/
////////////////////////////////////////////////////////////////////////////////
math::linear_alg::Vector3<float> HeightMap::GetPositionOffset(void)
{
	return m_positionOffset;
}

////////////////////////////////////////////////////////////////////////////////
//  FUNCTION:       
/*! \brief          sets a point's value
**  \param          pointIndex a two dimensional index into the HeightMap
**  \return         none
**  \exception      none
**  \par            DESCRIPTION:
**                    
**  \author         Austin Spafford
**  \date           10/11/2002 9:51:00 PM
**  \warning        Will invalidate normals, normals will be corrected upon reading
*/
////////////////////////////////////////////////////////////////////////////////
void HeightMap::SetPointValue(Vector2<int> pointIndex, float value)
{	
	assert(pointIndex == LimitedToMinMax(pointIndex, Vector2<int>(0,0), Vector2<int>(m_pointCount[0]-1, m_pointCount[1]-1)));

	m_dataPoints[pointIndex[0] + pointIndex[1]*m_pointCount[0]] = value;

	m_normalsAreCorrect = false;


    //cached render data is no longer correct
    m_verticesChanged = true;
	m_dataHasChanged = true;

	return;
}

////////////////////////////////////////////////////////////////////////////////
//  FUNCTION:       
/*! \brief          sets the point count of the HeightMap
**  \param          newPointCount a two dimensional vector representing the
**                  number of data points that should exist along each dimension
**  \return         none
**  \exception      none
**  \par            DESCRIPTION:
**                    
**  \author         Austin Spafford
**  \date           10/11/2002 9:51:38 PM
**  \warning        This method will erase all data point values!
*/
////////////////////////////////////////////////////////////////////////////////
void HeightMap::SetPointCount(Vector2<int> newPointCount)
{	
	assert(newPointCount == LimitedToMinimum(newPointCount, Vector2<int>(2,2)));

	Vector2<float> footPrint = GetFootPrint();

	m_pointCount = newPointCount;

	ReallocateDynamicMemory();	

	SetFootPrint(footPrint);

    //cached render data is no longer correct
	m_dataHasChanged = true;
	m_verticesChanged = true;
	
	return;
}

////////////////////////////////////////////////////////////////////////////////
//  FUNCTION:       
/*! \brief          sets the total footprint of the HeightMap, essentially 
**                  stretching it
**  \param          newFootPrint a two dimensional vector representing the
**                  new horizontal size the HeightMap should occupy
**  \return         none
**  \exception      none
**  \par            DESCRIPTION:
**                    
**  \author         Austin Spafford
**  \date           10/11/2002 9:53:29 PM
**  \warning        none
*/
////////////////////////////////////////////////////////////////////////////////
void HeightMap::SetFootPrint(Vector2<float> newFootPrint)
{
	m_cellDimensions = Vector2<float>(newFootPrint[0]/(m_pointCount[0]-1), newFootPrint[1]/(m_pointCount[1]-1));

    //cached render data is no longer correct
    m_verticesChanged = true;
	m_dataHasChanged = true;

	return;
}

////////////////////////////////////////////////////////////////////////////////
//  FUNCTION:       
/*! \brief          sets the total footprint of the HeightMap, essentially 
**                  stretching it
**  \param          newCellDimensions a two dimensional vector representing the
**                  new horizontal size each cell in the HeightMap
**  \return         none
**  \exception      none
**  \par            DESCRIPTION:
**                    
**  \author         Austin Spafford
**  \date           10/20/2002 9:53:29 PM
**  \warning        none
*/
////////////////////////////////////////////////////////////////////////////////
void HeightMap::SetCellDimensions(Vector2<float> newCellDimensions)
{
	m_cellDimensions = newCellDimensions;

    //cached render data is no longer correct
    m_verticesChanged = true;
	m_dataHasChanged = true;

	return;
}

////////////////////////////////////////////////////////////////////////////////
//  FUNCTION:       
/*! \brief          retrieves the position offset of the HeightMap
**  \param          newPositionOffset a three dimensional vector representing
**                  a value that will be added to every produced point.
**  \return         none
**  \exception      none
**  \par            DESCRIPTION:
**                    
**  \author         Austin Spafford
**  \date           10/19/2002 9:53:29 PM
**  \warning        none
*/
////////////////////////////////////////////////////////////////////////////////
void HeightMap::SetPositionOffset(math::linear_alg::Vector3<float> newPositionOffset)
{
	m_positionOffset = newPositionOffset;

    //cached render data is no longer correct
    m_spaceTransformationChanged = true;
	m_dataHasChanged = true;

	return;
}

////////////////////////////////////////////////////////////////////////////////
//  FUNCTION:       
/*! \brief          Resizes the heightmap to a new min/max pair.
**  \param          newMin the lowest possible value after the resizing
**  \param          newMax the heightest possible value after the resizing
**  \return         none
**  \exception      none
**  \author         Austin Spafford
**  \date           4/18/2003 9:28:32 AM
**  \warning        none
*/
////////////////////////////////////////////////////////////////////////////////
void HeightMap::VerticallyResizeEntireHeightMap(float newMin, float newMax)
{
	float currentMin = m_dataPoints[0];
	float currentMax = m_dataPoints[0];
	unsigned int numberOfPoints = m_pointCount[0] * m_pointCount[1];
	unsigned int i;

	// find the lowest and highest values in the HeightMap
	for (i=0; i<numberOfPoints; ++i)
	{
		if (m_dataPoints[i] < currentMin) currentMin = m_dataPoints[i];
		if (m_dataPoints[i] > currentMax) currentMax = m_dataPoints[i];
	}
	
	// normalize the heightmap
	float oneOverOldHeight = 1 / (currentMax - currentMin);
	for (i=0; i<numberOfPoints; ++i)
	{
		m_dataPoints[i] -= currentMin;
		m_dataPoints[i] *= oneOverOldHeight;
	}

	// scale the heightmap to the new min and maxs
	float newHeight = newMax - newMin;
	for (i=0; i<numberOfPoints; ++i)
	{
		m_dataPoints[i] += newMin;
		m_dataPoints[i] *= newHeight;
	}

	// find the lowest and highest values in the HeightMap
	currentMin = m_dataPoints[100];
	currentMax = m_dataPoints[100];
	for (i=0; i<numberOfPoints; ++i)
	{
		if (m_dataPoints[i] < currentMin) currentMin = m_dataPoints[i];
		if (m_dataPoints[i] > currentMax) currentMax = m_dataPoints[i];
	}

	// make sure to trigger these
	m_verticesChanged = true;
	m_spaceTransformationChanged = true;
	m_normalsAreCorrect = false;
	m_dataHasChanged = true;

	return;
}

////////////////////////////////////////////////////////////////////////////////
//  FUNCTION:       
/*! \brief          Returns true if a change to the HeightMap has occured since
**                  the flag was last cleared using ClearChangesFlag().
**  \return         true if any changes have occured
**  \exception      none
**  \author         Austin Spafford
**  \date           3/5/2003 3:38:01 PM
**  \warning        none
*/
////////////////////////////////////////////////////////////////////////////////
bool HeightMap::HasDataChanged(void)
{
	return m_dataHasChanged;
}

////////////////////////////////////////////////////////////////////////////////
//  FUNCTION:       
/*! \brief          Garuntees that calls to HasDataChanged() will be false until
**                  a change to the HeightMap is made.
**  \return         none
**  \exception      none
**  \author         Austin Spafford
**  \date           3/5/2003 3:41:28 PM
**  \warning        none
*/
////////////////////////////////////////////////////////////////////////////////
void HeightMap::ClearChangesFlag(void)
{
	m_dataHasChanged = false;

	return;
}

////////////////////////////////////////////////////////////////////////////////
//  FUNCTION:       
/*! \brief          casts a ray against the heightmap, will return true if a 
**                  collision was found
**  \param          rayOrigin a vector representing the origin of the ray
**  \param          rayDirection a vector representing the direction of the ray
**  \param          collisionLocation a reference to a vector that will represent
**                  the location of the first collision
**  \return         true if a collision was detected, and the contents of
**                  collisionLocation set, false otherwise
**  \exception      none
**  \author         Austin Spafford
**  \date           12/2/2002 6:32:40 PM
**  \warning        none
*/
////////////////////////////////////////////////////////////////////////////////
bool HeightMap::CastRay(
		Vector3<float> rayOrigin,
		Vector3<float> rayDirection,
		Vector3<float> & collisionLocation
		)
{
	Vector2<int> currentCellIndex;

	// calculate the starting cell
	currentCellIndex[0] = int((rayOrigin[0] - m_positionOffset[0]) / m_cellDimensions[0]);
	currentCellIndex[1] = int((rayOrigin[2] - m_positionOffset[2]) / m_cellDimensions[1]);

	// FARK: Trace the ray against the height map!

	return true;
}

////////////////////////////////////////////////////////////////////////////////
//  FUNCTION:       
/*! \brief          recomputes all of the normals stored by the HeightMap, call 
**                  after using block of SetPointValue()'s. If it has not been 
**                  explicitly called before a call to GetNormal(), then it will
**                  be implicitly called.
**  \param          
**  \return         none
**  \exception      none
**  \par            DESCRIPTION:
**                    
**  \author         Austin Spafford
**  \date           10/11/2002 9:54:39 PM
**  \warning        none
*/
////////////////////////////////////////////////////////////////////////////////
void HeightMap::RecalculateAllNormals(void)
{	
	for (int x=0; x<m_pointCount[0]-1; x++)
	{
		for (int y=0; y<m_pointCount[1]-1; y++)
		{
			Vector3<float> indexVertex = GetPointLocation(Vector2<int>(x,y));
			Vector3<float> diagVertex = GetPointLocation(Vector2<int>(x+1,y+1));
			Vector3<float> xMajorVertex = GetPointLocation(Vector2<int>(x+1,y));
			Vector3<float> yMajorVertex = GetPointLocation(Vector2<int>(x,y+1));

			Vector3<float> diagVector = diagVertex - indexVertex;

			Vector3<float> yMajorNormal = Cross(yMajorVertex - indexVertex, diagVector);
			yMajorNormal.Normalize();

			Vector3<float> xMajorNormal = Cross(diagVector, xMajorVertex - indexVertex);
			xMajorNormal.Normalize();

			GetNormalReference(Vector2<int>(x,y), e_yMajorTriangle) = yMajorNormal;
			GetNormalReference(Vector2<int>(x,y), e_xMajorTriangle) = xMajorNormal;
		}
	}	

	m_normalsAreCorrect = true;

	return;
}

////////////////////////////////////////////////////////////////////////////////
//  FUNCTION:       
/*! \brief          returns a reference to a normal that is indicated by its 
**                  index and major
**  \param          pointIndex specifies the vertex of the triangle that has the
**                  smallest possible indice values.
**  \param          triangleMajor is used to differentiate between the two
**                  triangles that share the same index.
**  \return         none
**  \exception      none
**  \par            DESCRIPTION:
**                    
**  \author         Austin Spafford
**  \date           10/13/2002 9:58:56 PM
**  \warning        
*/
////////////////////////////////////////////////////////////////////////////////
Vector3<float> & HeightMap::GetNormalReference(Vector2<int> pointIndex, e_TriangleMajor triangleMajor)
{	
	assert(pointIndex == LimitedToMinMax(pointIndex, Vector2<int>(0,0), Vector2<int>(m_pointCount[0]-2, m_pointCount[1]-2)));	

	return m_normals[2*(pointIndex[0] + pointIndex[1]*(m_pointCount[0]-1)) + triangleMajor];
}

////////////////////////////////////////////////////////////////////////////////
//  FUNCTION:       
/*! \brief          allocates fresh memory for the data points and normals
**  \param          
**  \return         none
**  \exception      none
**  \par            DESCRIPTION:
**                    
**  \author         Austin Spafford
**  \date           10/11/2002 9:58:56 PM
**  \warning        Will wipe out all data point values to 0
*/
////////////////////////////////////////////////////////////////////////////////
void HeightMap::ReallocateDynamicMemory(void)
{	
	if (NULL != m_dataPoints)
	{
		delete m_dataPoints;
		m_dataPoints = NULL;
	}

	m_dataPoints = new float[m_pointCount[0] * m_pointCount[1]];
	if (NULL == m_dataPoints)
	{
		throw;
	}

	m_normalsAreCorrect = false;

	memset(m_dataPoints, 0, m_pointCount[0] * m_pointCount[1] * sizeof(float)); 

	if (NULL != m_normals)
	{
		delete m_normals;
		m_normals = NULL;
	}

	m_normals = new Vector3<float>[(m_pointCount[0] - 1) * (m_pointCount[1] - 1) * 2];
	if (NULL == m_normals)
	{
		throw;
	}

	
	return;
}
