////////////////////////////////////////////////////////////////////////////////
/*! \file           HeightMap.h
**  \brief          HeightMap object interface file
**  \author         Austin Spafford
**  \date           10/11/2002 5:38:09 PM
**  \par            DESCRIPTION:
**                    The HeightMap object is capable of storing rasterized data
**                    points that describe the height a rectangular piece of
**                    terrain. 
**  \par            REMARKS:
**                                  
*/
////////////////////////////////////////////////////////////////////////////////
#pragma once
////////////////////////////////////////////////////////////////////////////////
#include "math/math.h"
#include "fileIO/memento.h"
#include "HeightMapFileMemento.h"
////////////////////////////////////////////////////////////////////////////////
namespace UDPNet
{
	class NetMessage;
}
////////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////
/*! \class          HeightMap
**  \brief          The HeightMap object is capable of storing rasterized data
**                  points that describe the height a rectangular piece of
**                  terrain. For the sake of notation, "FootPrint" refers to the 
**                  total X/Y dimensions of the HeightMap, and "PointCount"  
**                  refers to a count of how many points lie along each dimension.
**  \author         Austin Spafford
**  \date           10/11/2002 5:38:13 PM
**  \warning        none
*/
////////////////////////////////////////////////////////////////////////////////

class HeightMap
{	
	public:

	enum e_TriangleMajor
	{
		e_yMajorTriangle = 0, // these two are explicitly defined because there values
		e_xMajorTriangle = 1  // are directly used in offset computations
	};

	////////////////////////////////////////////////////////////////////////////
	////////////////////////////////////////////////////////////////////////////
	////////////////////////////////////////////////////////////////////////////

	/// default constructer, generates a 1x1 HeightMap
	HeightMap();

	/// detailed constructer
	HeightMap(math::linear_alg::Vector2<float> footPrint, math::linear_alg::Vector2<int> pointCount);

	/// deconstructer
	~HeightMap();

    ////////////////////////////////////////////////////////////////////////////
	////////////////////////////////////////////////////////////////////////////
	////////////////////////////////////////////////////////////////////////////

	/// loads the specified image file and maps its red channel to the height values of the height map
	/// returns true if the image was properly imported, false otherwise
	/// IMPORTANT: Also changes the resolution of the height map to that of the image.
	bool ImportFromImage(char * fileName, float maxHeight);

	/// saves a gray scale image of the height map, with white being the highest elevation
	/// returns true if the image was properly exported, false otherwise
	/// IMPORTANT: The reolution of the image will be that of the height map
	bool ExportToImage(char * fileName);

	////////////////////////////////////////////////////////////////////////////
	////////////////////////////////////////////////////////////////////////////
	////////////////////////////////////////////////////////////////////////////

	/// updates the state of the HeightMap by means of a NetMessage that
	/// originated on a sister HeightMap by means of a call to GenerateNetMessage()
	void LoadNetMessage(UDPNet::NetMessage & netMessageRef);

	/// generates a NetMessage that will encode the current state of the
	/// HeightMap, so that a sister HeightMap can load it by means of a
	/// call to LoadNetMessage()
	void GenerateNetMessage(UDPNet::NetMessage & netMessageRef);

	////////////////////////////////////////////////////////////////////////////
	////////////////////////////////////////////////////////////////////////////
	////////////////////////////////////////////////////////////////////////////

	// a factory method that generates an object that encapsulates the state
	// of the heightmap
	HeightMapFileMemento * CreateFileMemento(void) const;

	// sets the current state of the heightmap to that which is stored in fileMemento
	void LoadFileMemento(const HeightMapFileMemento & fileMemento);

	////////////////////////////////////////////////////////////////////////////
	////////////////////////////////////////////////////////////////////////////
	////////////////////////////////////////////////////////////////////////////

	/// retrieves a point's value
	float GetPointValue(math::linear_alg::Vector2<int> pointIndex);
	
	/// vertically projects a point onto the HeightMap and returns the projected
	/// point's Y value
	float GetProjectedHeight(const math::linear_alg::Vector3<float> & pointToProject);

	/// retrieves a point's location
	math::linear_alg::Vector3<float> GetPointLocation(math::linear_alg::Vector2<int> pointIndex);

	/// retrieves a point's location, but with out the position offset of the HeightMap being factored in
	math::linear_alg::Vector3<float> GetPointLocationWithoutOffset(math::linear_alg::Vector2<int> pointIndex);

	/// returns the value of a normal that is indicated by its "smallest" vetice's index and its axis major
	math::linear_alg::Vector3<float> GetNormalVector(math::linear_alg::Vector2<int> pointIndex, e_TriangleMajor triangleMajor);
    
    /// returns the value of the normal at the vertex specified
    math::linear_alg::Vector3<float> GenerateVertexNormalVector(math::linear_alg::Vector2<int> pointIndex);

	/// retrieves the point count of the HeightMap
	math::linear_alg::Vector2<int> GetPointCount(void);

	/// retrieves the footprint of the HeightMap
	math::linear_alg::Vector2<float> GetFootPrint(void);

	/// retrieves the cell dimensions of the HeightMap
	math::linear_alg::Vector2<float> GetCellDimensions(void);

	/// retrieves the position offset of the HeightMap
	math::linear_alg::Vector3<float> GetPositionOffset(void);

	////////////////////////////////////////////////////////////////////////////
	////////////////////////////////////////////////////////////////////////////
	////////////////////////////////////////////////////////////////////////////

	/// sets a point's value
	void SetPointValue(math::linear_alg::Vector2<int> pointIndex, float value);

	/// sets the point count of the HeightMap, THIS WILL ERASE ALL POINT VALUES!
	void SetPointCount(math::linear_alg::Vector2<int> newPointCount);

	/// sets the total footprint of the HeightMap, essentially stretching it
	void SetFootPrint(math::linear_alg::Vector2<float> newFootPrint);

	/// sets the dimensions of each cell of the HeightMap, essentially stretching it
	void SetCellDimensions(math::linear_alg::Vector2<float> newCellDimensions);

	/// retrieves the position offset of the HeightMap
	void SetPositionOffset(math::linear_alg::Vector3<float> newPositionOffset);

	////////////////////////////////////////////////////////////////////////////
	////////////////////////////////////////////////////////////////////////////
	////////////////////////////////////////////////////////////////////////////

	// resizes the heightmap to a new min/max pair
	void VerticallyResizeEntireHeightMap(float newMin, float newMax);

	////////////////////////////////////////////////////////////////////////////
	////////////////////////////////////////////////////////////////////////////
	////////////////////////////////////////////////////////////////////////////

	/// returns true if a change to the HeightMap has occured since the flag was
	/// last cleared using ClearChangesFlag()
	bool HasDataChanged(void);

	/// garuntees that calls to HasDataChanged() will be false until a change to
	/// the HeightMap is made
	void ClearChangesFlag(void); 

	////////////////////////////////////////////////////////////////////////////
	////////////////////////////////////////////////////////////////////////////
	////////////////////////////////////////////////////////////////////////////

	/// casts a ray against the heightmap, will return true if a collision was found
	bool CastRay(
			math::linear_alg::Vector3<float> rayOrigin,
			math::linear_alg::Vector3<float> rayDirection,
			math::linear_alg::Vector3<float> & collisionLocation
			);

	////////////////////////////////////////////////////////////////////////////
	////////////////////////////////////////////////////////////////////////////
	////////////////////////////////////////////////////////////////////////////

	/// recomputes all of the normals stored by the HeightMap, call after using
	/// block of SetPointValues
	void RecalculateAllNormals(void);

	////////////////////////////////////////////////////////////////////////////
	////////////////////////////////////////////////////////////////////////////
	////////////////////////////////////////////////////////////////////////////
	
	protected:	

	/// returns a reference to a normal that is indicated by its "smallest" vertice's index and its axis major
	math::linear_alg::Vector3<float> & GetNormalReference(math::linear_alg::Vector2<int> pointIndex, e_TriangleMajor triangleMajor);

	/// allocates fresh memory for the data points and normals
	void ReallocateDynamicMemory(void);	

	math::linear_alg::Vector2<int> m_pointCount;
	math::linear_alg::Vector2<float> m_cellDimensions;
	float *m_dataPoints;

	bool m_normalsAreCorrect;

	bool m_dataHasChanged;

	math::linear_alg::Vector3<float> *m_normals;

	math::linear_alg::Vector3<float> m_positionOffset; // an offset to all vector output	

    // all memebers below are used by the derived class to find out if anything
    // has changed for derived RENDERING classes.  This way they can cache the
    // data in, say, Vertex Array Ranges and achieve high framerates with over
    // 200000 tris on screen
    protected:
    /// have the vertices changed?
    bool m_verticesChanged;
    /// has the position or scale changed?
    bool m_spaceTransformationChanged;
};
