////////////////////////////////////////////////////////////////////////////////
/*! \file           HeightMap.cpp
**  \brief          HeightMap implementation file
**  \author         Austin Spafford
**  \date           5/14/2003 6:10:28 PM
**  \par            DESCRIPTION:
**                    The HeightMap object is capable of storing rasterized data
**                    points that describe the height a rectangular piece of
**                    terrain. 
**  \par            REMARKS:
**                    
*/
////////////////////////////////////////////////////////////////////////////////
#pragma warning (disable : 4786)
////////////////////////////////////////////////////////////////////////////////
#include <math.h>
#include <iostream>
#include <iomanip>
#include "math/math.h"
#include "HeightMapFileMemento.h"
#include "nexception/Nexception.h"
// memory leak detection
#ifdef _DEBUG
    #include"memleak.h"
    #define new DEBUG_NEW
    #undef THIS_FILE
    static char THIS_FILE[] = __FILE__;
#endif
////////////////////////////////////////////////////////////////////////////////
using namespace math::linear_alg;
////////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////
//  FUNCTION:       
/*! \brief          Returns the location of the next instance of a character
**                  in a NULL terminated string that is part of a floating
**                  point number.
**  \param          stringPtr a pointer into the string to seek through
**  \return         A pointer to the first floating point character, or NULL if
**                  there are not such characters exist in the specified string.
**  \exception      none
**  \author         Austin Spafford
**  \date           5/15/2003 1:29:48 PM
**  \warning        none
*/
////////////////////////////////////////////////////////////////////////////////
static char * SeekToFirstFloatingPointCharacter(char * stringPtr)
{
	if (NULL == stringPtr) return NULL;

	// seek until the end of the string
	while (0 != *stringPtr)
	{
		// check to see if the current character is a number
		char numericalValue = *stringPtr - '0';
		if (numericalValue >= 0 && numericalValue <= 9)
		{
			return stringPtr;
		}

		// check to see if the current character is a decimal point
		if ('.' == *stringPtr)
		{
			return stringPtr;
		}

		// check to see if the current character is a minus sign
		if ('-' == *stringPtr)
		{
			return stringPtr;
		}

		// move to the next character
		++stringPtr;
	}

	return NULL;
}

////////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////
//  FUNCTION:       
/*! \brief          Returns the location of the next instance of a character
**                  in a NULL terminated string that is not part of a floating
**                  point number.
**  \param          stringPtr a pointer into the string to seek through
**  \return         A pointer to the first non-floating-point character, or NULL if
**                  there are not such characters exist in the specified string.
**  \exception      none
**  \author         Austin Spafford
**  \date           5/15/2003 1:29:48 PM
**  \warning        none
*/
////////////////////////////////////////////////////////////////////////////////
static char * SeekToFirstNonFloatingPointCharacter(char * stringPtr)
{
	if (NULL == stringPtr) return NULL;

	// seek until the end of the string
	while (0 != *stringPtr)
	{
		// check to see if the current character is a number
		char numericalValue = *stringPtr - '0';
		if (numericalValue >= 0 && numericalValue <= 9)
		{
			// move to the next character
			++stringPtr;
			continue;
		}

		// check to see if the current character is a decimal point
		if ('.' == *stringPtr)
		{
			// move to the next character
			++stringPtr;
			continue;
		}

		// check to see if the current character is a minus sign
		if ('-' == *stringPtr)
		{
			// move to the next character
			++stringPtr;
			continue;
		}

		// the current character is not a floating point number, so we can return
		return stringPtr;
	}

	return NULL;
}

////////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////
//  FUNCTION:       
/*! \brief          Writes the memento's contents to an ostream in a human
**                  readable format.
**  \param          outputStream the stream that is to be written to
**  \return         none
**  \exception      none
**  \author         Austin Spafford
**  \date           5/15/2003 1:26:44 PM
**  \warning        none
*/
////////////////////////////////////////////////////////////////////////////////
void HeightMapFileMemento::Write(std::ostream & outputStream) const
{
	using std::endl;

	// make sure that any floating point numbers are printed with out
	// any scientific notation
	std::ios::fmtflags oldFlags = outputStream.flags();
	outputStream << std::fixed << std::setprecision(3) << std::showpoint;

	// write a small header
	outputStream << "============================================================" << endl;
	outputStream << "Begin HeightMap Information" << endl;
	outputStream << endl;

	// write the point count, cell dimensions, and position offset
	outputStream << "PointCount: (" << m_pointCount[0] << ", " << m_pointCount[1] << ")" << endl;
	outputStream << "CellDimensions: (" << m_cellDimensions[0] << ", " << m_cellDimensions[1] << ")" << endl;
	outputStream << "PositionOffset: (" << m_positionOffset[0] << ", " << m_positionOffset[1] << ", " << m_positionOffset[2] << ")" << endl;
	outputStream << endl;

	// write the data row by row	
	outputStream << "The data points: " << endl;
	for (int currentRow=0; currentRow<m_pointCount[1]; ++currentRow)
	{		
		for (int currentColumn=0; currentColumn<m_pointCount[0]; ++currentColumn)
		{
			outputStream << m_dataPointsBuffer[currentColumn + (currentRow*m_pointCount[0])] << " ";
		}
		outputStream << endl;
	}
	outputStream << endl;

	// write a small trail
	outputStream << "End HeightMap Information" << endl;
	outputStream << "============================================================" << endl;	
	outputStream << endl;

	// restore the old flags
	outputStream.flags(oldFlags);

	return;
}

////////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////
//  FUNCTION:       
/*! \brief          Reads the memento's contents from an istream that is using
**                  a human readable format.
**  \param          inputStream the stream that is to be read from
**  \return         none
**  \exception      none
**  \author         Austin Spafford
**  \date           5/15/2003 1:28:18 PM
**  \warning        none
*/
////////////////////////////////////////////////////////////////////////////////
void HeightMapFileMemento::Read(std::istream & inputStream)
{
	const int characterBufferSize = 4096;
	char * characterBuffer = new char[characterBufferSize];
	char * currentCharacterPtr;

	// HEADER
	// eat the header
	inputStream.getline(characterBuffer, characterBufferSize);
	inputStream.getline(characterBuffer, characterBufferSize);
	inputStream.getline(characterBuffer, characterBufferSize);

	// THE POINT COUNT, CELL DIMENSIONS, AND POSITION OFFSET
	// read in the point count
	inputStream.getline(characterBuffer, characterBufferSize);	
	currentCharacterPtr = SeekToFirstFloatingPointCharacter(characterBuffer);
	if (0 == currentCharacterPtr) throw Nexception("Failed to load PointCount[0].", __FILE__, __LINE__);
	m_pointCount[0] = atoi(currentCharacterPtr);
	currentCharacterPtr = SeekToFirstNonFloatingPointCharacter(currentCharacterPtr);
	currentCharacterPtr = SeekToFirstFloatingPointCharacter(currentCharacterPtr);
	if (0 == currentCharacterPtr) throw Nexception("Failed to load PointCount[1].", __FILE__, __LINE__);
	m_pointCount[1] = atoi(currentCharacterPtr);	
	// read in the cell dimensions
	inputStream.getline(characterBuffer, characterBufferSize);
	currentCharacterPtr = SeekToFirstFloatingPointCharacter(characterBuffer);
	if (0 == currentCharacterPtr) throw Nexception("Failed to load CellDimensions[0].", __FILE__, __LINE__);
	m_cellDimensions[0] = (float)atof(currentCharacterPtr);
	currentCharacterPtr = SeekToFirstNonFloatingPointCharacter(currentCharacterPtr);
	currentCharacterPtr = SeekToFirstFloatingPointCharacter(currentCharacterPtr);
	if (0 == currentCharacterPtr) throw Nexception("Failed to load CellDimensions[1].", __FILE__, __LINE__);
	m_cellDimensions[1] = (float)atof(currentCharacterPtr);
	// read in the position offset
	inputStream.getline(characterBuffer, characterBufferSize);
	currentCharacterPtr = SeekToFirstFloatingPointCharacter(characterBuffer);
	if (0 == currentCharacterPtr) throw Nexception("Failed to load PositionOffset[0].", __FILE__, __LINE__);
	m_positionOffset[0] = (float)atof(currentCharacterPtr);
	currentCharacterPtr = SeekToFirstNonFloatingPointCharacter(currentCharacterPtr);
	currentCharacterPtr = SeekToFirstFloatingPointCharacter(currentCharacterPtr);
	if (0 == currentCharacterPtr) throw Nexception("Failed to load PositionOffset[1].", __FILE__, __LINE__);
	m_positionOffset[1] = (float)atof(currentCharacterPtr);
	currentCharacterPtr = SeekToFirstNonFloatingPointCharacter(currentCharacterPtr);
	currentCharacterPtr = SeekToFirstFloatingPointCharacter(currentCharacterPtr);
	if (0 == currentCharacterPtr) throw Nexception("Failed to load PositionOffset[2].", __FILE__, __LINE__);
	m_positionOffset[2] = (float)atof(currentCharacterPtr);
	// eat a line
	inputStream.getline(characterBuffer, characterBufferSize);

	// THE DATA POINTS
	// eat a line
	inputStream.getline(characterBuffer, characterBufferSize);
	// read the points row by row
	for (int currentRow=0; currentRow<m_pointCount[1]; ++currentRow)
	{
		// read the entire row
		inputStream.getline(characterBuffer, characterBufferSize);
		currentCharacterPtr = SeekToFirstFloatingPointCharacter(characterBuffer);
		for (int currentColumn=0; currentColumn<m_pointCount[0]; ++currentColumn)
		{
			if (0 == currentCharacterPtr)
			{				
				throw Nexception("Failed to load a HeightMap data point.", __FILE__, __LINE__);
			}
			// read the point
			m_dataPointsBuffer[currentColumn + (currentRow*m_pointCount[0])] = (float)atof(currentCharacterPtr);
			// move to the next point
			currentCharacterPtr = SeekToFirstNonFloatingPointCharacter(currentCharacterPtr);
			currentCharacterPtr = SeekToFirstFloatingPointCharacter(currentCharacterPtr);
		}
	}
	// eat a line
	inputStream.getline(characterBuffer, characterBufferSize);

	// TAIL
	// eat the tail
	inputStream.getline(characterBuffer, characterBufferSize);
	inputStream.getline(characterBuffer, characterBufferSize);
	inputStream.getline(characterBuffer, characterBufferSize);

	delete []characterBuffer;

	return;
}
