////////////////////////////////////////////////////////////////////////////////
/*! \file           PhysicsSimulator.cpp
**  \brief          PhysicsSimulator implementation file
**  \author         Austin Spafford
**  \date           10/11/2002 6:11:04 PM
**  \par            DESCRIPTION:
**                    The PhysicsSimulator object is capable of simulating the
**                    interactions between any number of spheres and a HeightMap.
**  \par            REMARKS:
**                    The PhysicsSimulator is currently using a "fuzzy" physics
**                    model, which will slowly attempt to correct 
*/
////////////////////////////////////////////////////////////////////////////////
#pragma warning (disable : 4786)
////////////////////////////////////////////////////////////////////////////////
#include <assert.h>
#include <math.h>
#include "physics/PhysicsSimulator.h"
#include "physics/HeightMap.h"
// memory leak detection
#ifdef _DEBUG
    #include"memleak.h"
    #define new DEBUG_NEW
    #undef THIS_FILE
    static char THIS_FILE[] = __FILE__;
#endif
////////////////////////////////////////////////////////////////////////////////
using namespace math::linear_alg;
using namespace std;
////////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////
//  FUNCTION:       
/*! \brief          default constructer
**  \return         none
**  \exception      none
**  \author         Austin Spafford
**  \date           10/20/2002 8:40:14 PM
**  \warning        none
*/
////////////////////////////////////////////////////////////////////////////////
PhysicsSimulator::PhysicsSimulator() :	
	m_heightMapPtr(NULL),		
	m_boundaryMins(0,0,0),
	m_boundaryMaxs(0,0,0)
{
	return;
}

////////////////////////////////////////////////////////////////////////////////
//  FUNCTION:       
/*! \brief          deconstructer
**  \return         none
**  \exception      none
**  \author         Austin Spafford
**  \date           10/20/2002 8:40:14 PM
**  \warning        none
*/
////////////////////////////////////////////////////////////////////////////////
PhysicsSimulator::~PhysicsSimulator()
{
	// delete all of the allocated spheres
	hash_map<MsgHandler::Key, Sphere*>::iterator keySpherePairIter;
	keySpherePairIter = m_spherePtrs.begin();
	while (keySpherePairIter != m_spherePtrs.end())
	{
		delete keySpherePairIter->second;

		++keySpherePairIter;
	}
	m_spherePtrs.clear();

	return;
}

////////////////////////////////////////////////////////////////////////////////
//  FUNCTION:       
/*! \brief          adds a sphere tp the simulation and assigns the specified 
**                  Key value to it
**  \param          keyForObject the Key value that will be used as a label for
**                  the new Sphere
**  \return         true if a sphere was added to the simulation, and false if
**                  the object limit was reached or a sphere already exists that
**                  uses the specified Key as a label
**  \exception      none
**  \author         Austin Spafford
**  \date           10/20/2002 8:40:14 PM
**  \warning        none
*/
////////////////////////////////////////////////////////////////////////////////
bool PhysicsSimulator::AddSphere(MsgHandler::Key keyForObject)
{
	if (false != DoesSphereExist(keyForObject))
	{
		return false;
	}

/*	if (m_spherePtrs.size() > 100)
	{
		return false;
	}*/

	Sphere * newSpherePtr = new Sphere();

	// insert the new sphere
	m_spherePtrs[keyForObject] = newSpherePtr;

	return true;
}

////////////////////////////////////////////////////////////////////////////////
//  FUNCTION:       
/*! \brief          removes a sphere with the specified Key
**  \param          keyForObject the Key value that will be used as a label for
**                  the Sphere that is to be removed
**  \return         true if the specified Sphere was removed, and false if the
**                  specified Sphere does not exist in the simulation
**  \exception      none
**  \author         Austin Spafford
**  \date           10/20/2002 8:40:14 PM
**  \warning        none
*/
////////////////////////////////////////////////////////////////////////////////
bool PhysicsSimulator::RemoveSphere(MsgHandler::Key keyForObject)
{
	if (false == DoesSphereExist(keyForObject))
	{
		return false;
	}

	Sphere * sphereToDelete = m_spherePtrs.find(keyForObject)->second;

	delete sphereToDelete;

	m_spherePtrs.erase(keyForObject);

	return true;
}

////////////////////////////////////////////////////////////////////////////////
//  FUNCTION:       
/*! \brief          fills in minValues and maxValues with the current min and 
**                  max boundary values
**  \param          minValues a three dimensional vector that will be filled out
**                  with the minimal values used by the boundry box
**  \param          maxValues a three dimensional vector that will be filled out
**                  with the maximal values used by the boundry box
**  \return         none
**  \exception      none
**  \author         Austin Spafford
**  \date           10/20/2002 8:40:14 PM
**  \warning        none
*/
////////////////////////////////////////////////////////////////////////////////
void PhysicsSimulator::GetBoundary(math::linear_alg::Vector3<float> & minValues, math::linear_alg::Vector3<float> & maxValues)
{
	minValues = m_boundaryMins;
	maxValues = m_boundaryMaxs;

	return;
}

////////////////////////////////////////////////////////////////////////////////
//  FUNCTION:       
/*! \brief          returns a reference to the requested Sphere
**  \param          keyForObject the Key for the requested Sphere
**  \return         a reference to the specified Sphere
**  \exception      none
**  \author         Austin Spafford
**  \date           10/20/2002 8:40:14 PM
**  \warning        none
*/
////////////////////////////////////////////////////////////////////////////////
PhysicsSimulator::Sphere & PhysicsSimulator::GetSphereRef(MsgHandler::Key keyForObject)
{
	#ifndef NDEBUG
	if (false == DoesSphereExist(keyForObject))
	{
		throw;
	}
	#endif

	hash_map<MsgHandler::Key, Sphere*>::iterator keySpherePairIter;

	keySpherePairIter = m_spherePtrs.find(keyForObject);	

	return *(*keySpherePairIter).second;
}	

////////////////////////////////////////////////////////////////////////////////
//  FUNCTION:       
/*! \brief          sets the ultimate boundaries for all mobile objects, passing
**                  in the same point for both corners will remove the boundary
**  \param          cornerOne a three dimensional vector representing one corner
**                  of the new boundry box.
**  \param          cornerTwo a three dimensional vector representing one corner
**                  of the new boundry box.
**  \return         none
**  \exception      none
**  \author         Austin Spafford
**  \date           10/20/2002 8:40:14 PM
**  \warning        none
*/
////////////////////////////////////////////////////////////////////////////////
void PhysicsSimulator::SetBoundary(math::linear_alg::Vector3<float> cornerOne, math::linear_alg::Vector3<float> cornerTwo){

	for (int i=0; i<3; i++)
	{
		if (cornerOne[i] < cornerTwo[i])
		{
			m_boundaryMins[i] = cornerOne[i];
			m_boundaryMaxs[i] = cornerTwo[i];
		}
		else
		{
			m_boundaryMins[i] = cornerTwo[i];
			m_boundaryMaxs[i] = cornerOne[i];
		}
	}

	return;
}

////////////////////////////////////////////////////////////////////////////////
//  FUNCTION:       
/*! \brief          sets a reference to a HeightMap object, NULL is a valid
**                  parameter value
**  \param          newHeightMapPtr a pointer to the HeightMap that will be
**                  taken into consideration when stepping the simulation
**  \return         none
**  \exception      none
**  \author         Austin Spafford
**  \date           10/20/2002 8:40:14 PM
**  \warning        none
*/
////////////////////////////////////////////////////////////////////////////////
void PhysicsSimulator::SetHeightMap(HeightMap * newHeightMapPtr)
{
	m_heightMapPtr = newHeightMapPtr;

	return;
}

////////////////////////////////////////////////////////////////////////////////
//  FUNCTION:       
/*! \brief          casts a ray against the heightmap, will return true if a 
**                  collision was found
**  \param          rayOrigin a vector representing the origin of the ray
**  \param          rayDirection a vector representing the direction of the ray
**  \param          collisionLocation a reference to a vector that will represent
**                  the location of the first collision
**  \return         true if a collision was detected, and the contents of
**                  collisionLocation set, false otherwise
**  \exception      none
**  \author         Austin Spafford
**  \date           12/2/2002 6:32:40 PM
**  \warning        none
*/
////////////////////////////////////////////////////////////////////////////////
bool PhysicsSimulator::CastRay(
		Vector3<float> rayOrigin,
		Vector3<float> rayDirection,
		Vector3<float> & collisionLocation
		)
{
	// cast the ray against the height map if one is set
	if (NULL != m_heightMapPtr)
	{
		m_heightMapPtr->CastRay(rayOrigin, rayDirection, collisionLocation);
	}

	return true;
}

////////////////////////////////////////////////////////////////////////////////
//  FUNCTION:       
/*! \brief          checks to see if a sphere exists in the simulation that uses
**                  the specified Key as a label
**  \param          keyForObject a Key that is used as a label for the target
**                  object
**  \return         returns true if a Sphere is stored that uses the specified
**                  Key, false otherwise
**  \exception      none
**  \author         Austin Spafford
**  \date           10/20/2002 8:40:14 PM
**  \warning        none
*/
////////////////////////////////////////////////////////////////////////////////
bool PhysicsSimulator::DoesSphereExist(MsgHandler::Key keyForObject)
{
	if (m_spherePtrs.end() == m_spherePtrs.find(keyForObject)) 
	{
		return false;
	}

	return true;
}

////////////////////////////////////////////////////////////////////////////////
//  FUNCTION:       
/*! \brief          steps the simulation by "timeStep" number of seconds, call
**                  every frame
**  \param          timeStep the number of seconds that the simulation needs to
**                  step by
**  \return         none
**  \exception      none
**  \author         Austin Spafford
**  \date           10/20/2002 8:40:14 PM
**  \warning        none
*/
////////////////////////////////////////////////////////////////////////////////
void PhysicsSimulator::StepSimulation(float timeStep)
{
	// move all of the spheres
	hash_map<MsgHandler::Key, Sphere*>::iterator keySpherePairIter;
	keySpherePairIter = m_spherePtrs.begin();
	while (keySpherePairIter != m_spherePtrs.end())
	{
		Sphere *spherePtr = keySpherePairIter->second;

		assert(NULL != spherePtr);

		Vector3<float> position = spherePtr->GetPosition();
		Vector3<float> velocity = spherePtr->GetVelocity();
		Vector3<float> acceleration = spherePtr->GetAcceleration();		
		Vector3<float> tempVector;
		float radius = keySpherePairIter->second->GetRadius();

		// clear the collision flag
		spherePtr->SetCollidedDuringLastFrame(false);

		// collide against the simulation boundary
		if (m_boundaryMins != m_boundaryMaxs)
		{
			for (int i=0; i<3; i++)
			{
				if (position[i] - radius < m_boundaryMins[i])
				{
					if (velocity[i] <= 0)
					{
						velocity[i] *= -0.25f;
						position[i] =  m_boundaryMins[i] + radius;

						// record that a collision occured 
						spherePtr->SetCollidedDuringLastFrame(true);
					}
				}

				if (position[i] + radius > m_boundaryMaxs[i])
				{
					if (velocity[i] >= 0)
					{
						velocity[i] *= -0.25f;
						position[i] =  m_boundaryMaxs[i] - radius;

						// record that a collision occured 
						spherePtr->SetCollidedDuringLastFrame(true);
					}
				}
			}
		}

		// correct the spheres intersecting other objects in the simulation
		hash_map<MsgHandler::Key, Sphere*>::iterator secondKeySpherePairIter;
		secondKeySpherePairIter = m_spherePtrs.begin();
		while (secondKeySpherePairIter != m_spherePtrs.end())
		{
			assert(NULL !=  secondKeySpherePairIter->second);

			// correct for collisions with other spheres
			if (secondKeySpherePairIter != keySpherePairIter)
			{
				Vector3<float> positionDelta = (*secondKeySpherePairIter).second->GetPosition() - position;
				float distance = positionDelta.Magnitude();
				float targetDistance = radius + (*secondKeySpherePairIter).second->GetRadius();

				if (distance < targetDistance)
				{
					Vector3<float> forceVector = positionDelta;
					forceVector *= timeStep * 64.0f * (targetDistance - distance) / targetDistance;

					// push the second sphere
					spherePtr->SetVelocity(							
							spherePtr->GetVelocity()
							-
							forceVector
							);
	
					// push the second sphere
					secondKeySpherePairIter->second->SetVelocity(							
							(*secondKeySpherePairIter).second->GetVelocity()
							+
							forceVector
							);

					// record that a collision occured (the second sphere will register the collision later, oh well, temp collision code anyway)
					spherePtr->SetCollidedDuringLastFrame(true);
				}
			}

			// correct for collisions with the HeightMap
			if (NULL != m_heightMapPtr)
			{
				Vector2<float> spherePositionOnGrid;
				Vector2<int> cellIndex;
				Vector2<int> indexMaxs = m_heightMapPtr->GetPointCount();
				HeightMap::e_TriangleMajor triangleMajor;

				spherePositionOnGrid[0] = position[0] - m_heightMapPtr->GetPositionOffset()[0];
				spherePositionOnGrid[1] = position[2] - m_heightMapPtr->GetPositionOffset()[2];

				cellIndex[0] = int(spherePositionOnGrid[0] / m_heightMapPtr->GetCellDimensions()[0]);
				cellIndex[1] = int(spherePositionOnGrid[1] / m_heightMapPtr->GetCellDimensions()[1]);

				// limit the indices
				for (int i=0; i<2; i++)
				{
					if (cellIndex[i] < 0)
					{
						cellIndex[i] = 0;
					}

					if (cellIndex[i] > indexMaxs[i] - 2)
					{
						cellIndex[i] = indexMaxs[i] - 2;
					}
				}

				spherePositionOnGrid[0] -= cellIndex[0] * m_heightMapPtr->GetCellDimensions()[0];
				spherePositionOnGrid[1] -= cellIndex[1] * m_heightMapPtr->GetCellDimensions()[1];

				// detect the majority of the triangle that the sphere is over
				if (spherePositionOnGrid[0] > spherePositionOnGrid[1])
				{
					triangleMajor = HeightMap::e_xMajorTriangle;
				}
				else
				{
					triangleMajor = HeightMap::e_yMajorTriangle;
				}

				Vector3<float> triangleVertex = m_heightMapPtr->GetPointLocation(cellIndex);
				Vector3<float> triangleNormal = m_heightMapPtr->GetNormalVector(cellIndex, triangleMajor);
				float distanceFromTriangle = Dot(triangleNormal, (position - triangleVertex));

				if (distanceFromTriangle <= radius) 
				{
					float sizeFactor = 3 / radius;
					velocity += timeStep * sizeFactor * 50.0f * triangleNormal * (radius - distanceFromTriangle);

					// record that a collision occured 
					spherePtr->SetCollidedDuringLastFrame(true);
				}
			}

			++secondKeySpherePairIter;
		}
		
		// throw in gravity
//		velocity[1] -= 50*timeStep;

		// apply friction
		velocity -= 5.0f * velocity * timeStep;

		// apply acceleration
		tempVector = acceleration;
		tempVector *= timeStep;
		velocity += tempVector;

		// apply velocity
		tempVector = velocity;
		tempVector *= timeStep;
		position += tempVector;

		// save the results 
		spherePtr->SetPosition(position);
		spherePtr->SetVelocity(velocity);

		++keySpherePairIter;
	}	

	return;
}
