////////////////////////////////////////////////////////////////////////////////
/*! \file           PhysicsSimulator.h
**  \brief          PhysicsSimulator object interface file
**  \author         Austin Spafford
**  \date           10/17/2002 4:20:38 PM
**  \par            DESCRIPTION:
**                    The PhysicsSimulator object is capable of simulating the
**                    interactions between any number of spheres and a HeightMap.
**  \par            REMARKS:
**                    The PhysicsSimulator is currently using a "fuzzy" physics
**                    model, which will slowly attempt to correct 
*/
////////////////////////////////////////////////////////////////////////////////
#pragma once
////////////////////////////////////////////////////////////////////////////////
#include <hash_map>
#include "math/math.h"
#include "messageHandler/MessageHandler.h"
////////////////////////////////////////////////////////////////////////////////
class HeightMap;
////////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////
/*! \class          PhysicsSimulator
**  \brief          The PhysicsSimulator object attempts to correctly model the
**                  interactions between all of the objects that it has been
**                  assigned to store and simulate.
**  \author         Austin Spafford
**  \date           10/17/2002 9:27:08 PM
**  \warning        none
*/
////////////////////////////////////////////////////////////////////////////////
class PhysicsSimulator
{
	public:

	////////////////////////////////////////////////////////////////////////////////
	/*! \class          Sphere
	**  \brief          The Sphere stores a point and a radius.
	**  \author         Austin Spafford
	**  \date           10/17/2002 8:30:53 PM
	**  \warning        none
	*/
	////////////////////////////////////////////////////////////////////////////////
	class Sphere
	{
		public:

		/// default constructer
		Sphere() :
				m_position(0,0,0), 
				m_radius(1),
				m_collidedDuringLastFrame(false)
				{}

		/// detailed constructer
		Sphere(math::linear_alg::Vector3<float> centerPoint, float radius) :
				m_position(centerPoint),
				m_radius(radius),
				m_collidedDuringLastFrame(false)
				{}

		/// deconstructer
		~Sphere() {}

		////////////////////////////////////////////////////////////////////////////

		/// returns the center point of the sphere
		math::linear_alg::Vector3<float> GetPosition(void) { return m_position; }

		/// returns the velocity of the sphere
		math::linear_alg::Vector3<float> GetVelocity(void) { return m_velocity; }

		/// returns the acceleration of the sphere
		math::linear_alg::Vector3<float> GetAcceleration(void) { return m_acceleration; }

		/// returns the radius of the sphere
		float GetRadius(void) { return m_radius; }

		/// returns true if the sphere collided with anything during the previous frame
		bool GetCollidedDuringLastFrame(void) { return m_collidedDuringLastFrame; }

		////////////////////////////////////////////////////////////////////////////

		/// sets the center point of the sphere
		void SetPosition(math::linear_alg::Vector3<float> newCenter) { m_position = newCenter; }

		/// sets the velocity of the sphere
		void SetVelocity(math::linear_alg::Vector3<float> newVelocity) { m_velocity = newVelocity; }

		/// sets the acceleration of the sphere
		void SetAcceleration(math::linear_alg::Vector3<float> newAcceleration) { m_acceleration = newAcceleration; }

		/// sets the radius of the sphere
		void SetRadius(float newRadius) { m_radius = newRadius; }

		/// sets whether or not the sphere collided with anything during the previous frame
		void SetCollidedDuringLastFrame(bool newCollisionBool) { m_collidedDuringLastFrame = newCollisionBool; }

		////////////////////////////////////////////////////////////////////////////

		private:
		
		math::linear_alg::Vector3<float> m_position;
		math::linear_alg::Vector3<float> m_velocity;
		math::linear_alg::Vector3<float> m_acceleration;
		float m_radius;
		bool m_collidedDuringLastFrame;
	};

	////////////////////////////////////////////////////////////////////////////

	/// default constructer
	PhysicsSimulator();	

	/// deconstructer
	~PhysicsSimulator();

	////////////////////////////////////////////////////////////////////////////

	/// adds a sphere to the simulation and assigns the specified Key to it
	bool AddSphere(MsgHandler::Key keyForObject);

	////////////////////////////////////////////////////////////////////////////

	/// removes a sphere with the specified Key
	bool RemoveSphere(MsgHandler::Key keyForObject);

    ////////////////////////////////////////////////////////////////////////////

	/// fills in minValues and maxValues with the current boundary limits
	void GetBoundary(math::linear_alg::Vector3<float> & minValues, math::linear_alg::Vector3<float> & maxValues);

	/// returns a reference to the requested Sphere
	Sphere & GetSphereRef(MsgHandler::Key keyForObject);

	// returns the number of spheres that are in teh simulation
	unsigned int GetSphereCount(void) { return m_spherePtrs.size(); }

	////////////////////////////////////////////////////////////////////////////

	/// sets the ultimate boundaries for all mobile objects, passing in the same
	/// point for both corners will remove the boundary
	void SetBoundary(math::linear_alg::Vector3<float> cornerOne, math::linear_alg::Vector3<float> cornerTwo);

	/// sets a reference to a HeightMap object, NULL is a valid parameter value
	void SetHeightMap(HeightMap * newHeightMapPtr);

	////////////////////////////////////////////////////////////////////////////

	/// casts a ray against the heightmap, will return true if a collision was found
	bool CastRay(
			math::linear_alg::Vector3<float> rayOrigin,
			math::linear_alg::Vector3<float> rayDirection,
			math::linear_alg::Vector3<float> & collisionLocation
			);

	////////////////////////////////////////////////////////////////////////////

	/// returns true if a Sphere is stored that uses the specified Key
	bool DoesSphereExist(MsgHandler::Key keyForObject);

	////////////////////////////////////////////////////////////////////////////

	/// steps the simulation by "timeStep" number of seconds, call every frame
	void StepSimulation(float timeStep);

	////////////////////////////////////////////////////////////////////////////

	private:
	
	std::hash_map<MsgHandler::Key, Sphere*> m_spherePtrs;

	HeightMap * m_heightMapPtr;
		
	math::linear_alg::Vector3<float> m_boundaryMins;
	math::linear_alg::Vector3<float> m_boundaryMaxs;
};
