////////////////////////////////////////////////////////////////////////////////
/*!	\file		Keyboard.cpp
**	\brief		Implementation file for Keyboard.h
**	\author		John Corpening
**	\date		10/16/2002 8:02:48 PM
**	\par		DESCRIPTION:
**					Contains Keyboard class which abstracts the keyboard.
**					-Pressed is defined as a button is down
**					-Triggered is defined as a button was not 
**						down the last frame but is down this frame
**	\par		REMARKS:
**					You can only get an instance of the keyboard from the
**					PlayerInput class. Get an instance if you will not need
**					mouse input.
*/
////////////////////////////////////////////////////////////////////////////////
#pragma warning(disable:4786)
#include "Keyboard.h"
#include "../nexception/Nexception.h"

// memory leak detection
#ifdef _DEBUG
    #include"memleak.h"
    #define new DEBUG_NEW
    #undef THIS_FILE
    static char THIS_FILE[] = __FILE__;
#endif

using namespace UserInput;
//Singleton instance initialization
Keyboard* Keyboard::_instance = 0;

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	Keyboard::initialize
/*!	\brief		returns a pointer to the Keyboard Singleton
**	\param		hwnd required for DirectInput initialization
**	\param		DI pointer to a DirectInput instance
**	\return		pointer to the player input Singleton
**	\exception	none
**	\author		John Corpening
**	\date		10/16/2002 7:53:40 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
Keyboard* Keyboard::initialize(LPDIRECTINPUT8	DI, HWND hwnd)
{
	if(_instance)
	{
		return _instance;
	}
	else
	{
		_instance = new Keyboard(DI, hwnd);
		return _instance;
	}
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	Keyboard::instance
/*!	\brief		returns a pointer to the Keyboard Singleton
**	\return		none
**	\exception	Keyboard pointer not initialized
**	\author		John Corpening
**	\date		10/28/2002 3:12:18 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
Keyboard* Keyboard::instance()
{
	if(!_instance)
		throw(Nexception("Keyboard pointer not initialized", __FILE__,__LINE__,__DATE__,__TIME__));

	return _instance;
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	Keyboard::Keyboard
/*!	\brief		Constructor initializes DirectInput
**	\param		hwnd required for DirectInput initialization
**	\param		DI pointer to a DirectInput instance
**	\return		none
**	\exception	none
**	\author		John Corpening
**	\date		10/14/2002 2:22:33 PM
**	\warning	none
**	\par		UPDATED ON:			1/31/2003 9:37:39 AM
**	\par		UPDATED BY:			Nate Cleveland
**	\par		UPDATE DESCRIPTION: 
**				Changed the cooperation level of the keyboard to DISCL_NONEXCLUSIVE 
				| DISCL_BACKGROUND this allows the keyboard to still access the keybord
				data even when it isn't in the foreground. Slight hack that we'll
				need to change in the future, but it does allow us to have the 
				window to now be in the forground.
*/
////////////////////////////////////////////////////////////////////////////////
Keyboard::Keyboard(LPDIRECTINPUT8 DI, HWND hwnd)
    : validKeyInput(true)
{
	HRESULT hr;
	//Create the keyboard object
	hr = DI->CreateDevice(GUID_SysKeyboard, &DIKeyboard, NULL); 
	if FAILED(hr) 
	{ 
	   throw(Nexception("Keyboard initialization failed", __FILE__,__LINE__,__DATE__,__TIME__));

	} 

	//Set the Keyboard Data Format
	//Note: c_dfDIKeyboard is a DirectInput global variable
	hr = DIKeyboard->SetDataFormat(&c_dfDIKeyboard); 
 
	if FAILED(hr)
	{ 
		throw(Nexception("Keyboard set data format failed", __FILE__,__LINE__,__DATE__,__TIME__));

	} 

	//Set the Cooperative Level
	// Latest hack by Nate. Now, if the file INPUT_DEBUG.on is in the working folder
	// the game will set cooperation level such that it will not crash if the game is set 
	// to the background. [3/14/2003]
	FILE *hHack = fopen("INPUT_DEBUG.on", "r" );

	if(hHack)
	{
		hr = DIKeyboard->SetCooperativeLevel(hwnd, DISCL_NONEXCLUSIVE | DISCL_BACKGROUND); 
		fclose(hHack);
	}

	else
		hr = DIKeyboard->SetCooperativeLevel(hwnd, DISCL_FOREGROUND | DISCL_EXCLUSIVE); 


 
	if FAILED(hr)
	{ 
		throw(Nexception("Keyboard set coorperative level failed", __FILE__,__LINE__,__DATE__,__TIME__));
	} 

	//Acquire the keyboard
	hr = DIKeyboard->Acquire(); 
    if (FAILED(hr))
    {
        throw Nexception("Keyboard Acquisition failed",
                         __FILE__, __LINE__, __DATE__, __TIME__);
    }

	//Miscellaneous setup
	keyToggle = 0;
	initializeConstantsArray();
	lastCharacter = 0;
	lcChecked = false;
}


////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	Keyboard::getInput
/*!	\brief		invoked once per frame to update the state of the keyboard.
**	\return		Returns success or failure.
**	\exception	none
**	\author		John Corpening
**	\date		10/11/2002 11:16:56 AM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
bool Keyboard::getInput() throw()
{
	HRESULT hr;

	//Clear out the last character checked flag
	lcChecked = false;
	if(keyToggle = 1 - keyToggle)
	{
		pKeyCur = &keyInput1;
		pKeyPrev = &keyInput2;
	}
	else
	{
		pKeyCur = &keyInput2;
		pKeyPrev = &keyInput1;
	}
	//Clear out the current
	memset(*pKeyCur, 0, 256);
    DIKeyboard->Acquire();
	hr = DIKeyboard->GetDeviceState(256,(LPVOID)*pKeyCur); 
    if FAILED(hr) 
    { 
        return false; 
    } 
	return true;
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	Keyboard::initializeConstantArray
/*!	\brief		initializes the arrays of alphanumeric constants
**	\return		none
**	\exception	none
**	\author		John Corpening
**	\date		10/16/2002 8:04:56 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
void Keyboard::initializeConstantsArray() throw()
{
	aConstants[0]  = DIK_A;
	aConstants[1]  = DIK_B;
	aConstants[2]  = DIK_C;
	aConstants[3]  = DIK_D;
	aConstants[4]  = DIK_E;
	aConstants[5]  = DIK_F;
	aConstants[6]  = DIK_G;
	aConstants[7]  = DIK_H;
	aConstants[8]  = DIK_I;
	aConstants[9]  = DIK_J;
	aConstants[10] = DIK_K;
	aConstants[11] = DIK_L;
	aConstants[12] = DIK_M;
	aConstants[13] = DIK_N;
	aConstants[14] = DIK_O;
	aConstants[15] = DIK_P;
	aConstants[16] = DIK_Q;
	aConstants[17] = DIK_R;
	aConstants[18] = DIK_S;
	aConstants[19] = DIK_T;
	aConstants[20] = DIK_U;
	aConstants[21] = DIK_V;
	aConstants[22] = DIK_W;
	aConstants[23] = DIK_X;
	aConstants[24] = DIK_Y;
	aConstants[25] = DIK_Z;

	nConstants[0][0] = DIK_0; nConstants [0][1] = DIK_NUMPAD0;
	nConstants[1][0] = DIK_1; nConstants [1][1] = DIK_NUMPAD1;
	nConstants[2][0] = DIK_2; nConstants [2][1] = DIK_NUMPAD2;
	nConstants[3][0] = DIK_3; nConstants [3][1] = DIK_NUMPAD3;
	nConstants[4][0] = DIK_4; nConstants [4][1] = DIK_NUMPAD4;
	nConstants[5][0] = DIK_5; nConstants [5][1] = DIK_NUMPAD5;
	nConstants[6][0] = DIK_6; nConstants [6][1] = DIK_NUMPAD6;
	nConstants[7][0] = DIK_7; nConstants [7][1] = DIK_NUMPAD7;
	nConstants[8][0] = DIK_8; nConstants [8][1] = DIK_NUMPAD8;
	nConstants[9][0] = DIK_9; nConstants [9][1] = DIK_NUMPAD9;
}


////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	Keyboard::getLastCharacter
/*!	\brief		Returns the printable character pressed during the last update.
**	\return		A character or 0 if no printable character was pressed
**	\exception	none
**	\author		John Corpening
**	\date		10/11/2002 11:17:08 AM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
char Keyboard::getLastCharacter() throw()
{
	//If this check has already been performed then return the result
	if(lcChecked)
		return lastCharacter;

	//Otherwise check for printable characters
	bool shift = isPressed(DIK_LSHIFT) || isPressed(DIK_RSHIFT);
	lcChecked = true;

	//First check the space character since this will often be used
	if(isTriggered(DIK_SPACE))
	{
		return lastCharacter = ' ';
	}

	//Check letters
	for(int i = 0; i < 26; i++)
	{
		if(isTriggered(aConstants[i]))
		{
			//Determine which letter was pressed and return the appropriate key
			//Check to see if the shift key is being held
			if(shift)
			{
				return lastCharacter = 'A' + i;
			}
			else
			{
				return lastCharacter = 'a' + i;
			}
		}
	}

	//Check numbers
	if(!shift)
	{
		for(int j = 0; j < 10; j++)
		{
			if(isTriggered(nConstants[j][0]) || isTriggered(nConstants[j][1]))
			{
				return lastCharacter = '0' + j;
			}
		}
	}
	else
	{
		for(int j = 0; j < 10; j++)
		{
			if(isTriggered(nConstants[j][0]))
			{
				switch(j)
				{
				case 0:
					return lastCharacter = ')';
					break;
				case 1:
					return lastCharacter = '!';
					break;
				case 2:
					return lastCharacter = '@';
					break;
				case 3:
					return lastCharacter = '#';
					break;
				case 4:
					return lastCharacter = '$';
					break;
				case 5:
					return lastCharacter = '%';
					break;
				case 6:
					return lastCharacter = '^';
					break;
				case 7:
					return lastCharacter = '&';
					break;
				case 8:
					return lastCharacter = '*';
					break;
				case 9:
					return lastCharacter = '(';
					break;
				default:
					//Assume an error and return nothing
					return 0;
				}
			}
			else if(isTriggered(nConstants[j][1]))
			{
				return lastCharacter = '0' + j;
			}
		}
	}
	//Now check special characters, Unfortunately at this point
	//There are not any handy Ascii tricks to help out anymore
	if(isTriggered(DIK_GRAVE))
	{
		if(shift)
			return lastCharacter = '~';
		else
			return lastCharacter = '`';
	}

	if(isTriggered(DIK_MINUS))
	{
		if(shift)
			return lastCharacter = '_';
		else
			return lastCharacter = '-';
	}

	if(isTriggered(DIK_EQUALS))
	{
		if(shift)
			return lastCharacter = '+';
		else
			return lastCharacter = '=';
	}

	if(isTriggered(DIK_LBRACKET))
	{
		if(shift)
			return lastCharacter = '{';
		else
			return lastCharacter = '[';
	}

	if(isTriggered(DIK_RBRACKET))
	{
		if(shift)
			return lastCharacter = '}';
		else
			return lastCharacter = ']';
	}

	if(isTriggered(DIK_SEMICOLON))
	{
		if(shift)
			return lastCharacter = ':';
		else
			return lastCharacter = ';';
	}

	if(isTriggered(DIK_APOSTROPHE))
	{
		if(shift)
			return lastCharacter = '\"';
		else
			return lastCharacter = '\'';
	}

	if(isTriggered(DIK_BACKSLASH))
	{
		if(shift)
			return lastCharacter = '|';
		else
			return lastCharacter = '\\';
	}

	if(isTriggered(DIK_COMMA))
	{
		if(shift)
			return lastCharacter = '<';
		else
			return lastCharacter = ',';
	}

	if(isTriggered(DIK_PERIOD))
	{
		if(shift)
			return lastCharacter = '>';
		else
			return lastCharacter = '.';
	}

	if(isTriggered(DIK_SLASH))
	{
		if(shift)
			return lastCharacter = '?';
		else
			return lastCharacter = '/';
	}

	if(isTriggered(DIK_GRAVE))
	{
		if(shift)
			return lastCharacter = '~';
		else
			return lastCharacter = '`';
	}

	//Non-numeric Numpad keys
	if(isTriggered(DIK_MULTIPLY))
	{
		return lastCharacter = '*';
	}
	if(isTriggered(DIK_ADD))
	{
		return lastCharacter = '+';
	}
	if(isTriggered(DIK_SUBTRACT))
	{
		return lastCharacter = '-';
	}
	if(isTriggered(DIK_DIVIDE))
	{
		return lastCharacter = '/';
	}
	if(isTriggered(DIK_DECIMAL))
	{
		return lastCharacter = '*';
	}
	//If no printable keys have been pressed then return 0
	return lastCharacter = 0;
}