////////////////////////////////////////////////////////////////////////////////
/*!	\file		Keyboard.h
**	\brief		Class to abstract the keyboard
**	\author		John Corpening
**	\date		10/16/2002 8:02:48 PM
**	\par		DESCRIPTION:
**					Contains Keyboard class which abstracts the keyboard.
**					-Pressed is defined as a button is down
**					-Triggered is defined as a button was not 
**						down the last frame but is down this frame
**	\par		REMARKS:
**					You can only get an instance of the keyboard from the
**					PlayerInput class. Get an instance if you will not need
**					mouse input.
*/
////////////////////////////////////////////////////////////////////////////////

#pragma once
#include "PlayerInput.h"

namespace UserInput
{

////////////////////////////////////////////////////////////////////////////////
/*!	\class		Keyboard
**	\brief		Class to abstract the Keyboard
**	\author		John Corpening
**	\date		10/16/2002 8:05:13 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
class Keyboard
{
public:

	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:	Keyboard::isPressed
	/*!	\brief		Returns whether a key was pressed
	**	\param		key, an input constant
	**	\return		true if pressed, false if not pressed 
	**	\exception	none
	**	\author		John Corpening
	**	\date		10/11/2002 11:17:00 AM
	**	\warning	If getInput() fails this will return false.
	*/
	////////////////////////////////////////////////////////////////////////////////
	bool isPressed(InputConstant key) const throw()
	{return (validKeyInput ? ((*pKeyCur)[key] & HIGHBIT ? true : false) : false);}
	
	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:	Keyboard::isTriggered
	/*!	\brief		Returns whether a key was Triggered.
	**	\param		key, an input constant
	**	\return		true if triggered, false if not triggered 
	**	\exception	none
	**	\author		John Corpening
	**	\date		10/11/2002 11:17:04 AM
	**	\warning	none
	*/
	////////////////////////////////////////////////////////////////////////////////
	bool isTriggered(InputConstant key) const throw()
	{return (validKeyInput ? isPressed(key) && ((*pKeyPrev)[key] & HIGHBIT ? false : true) : false);}

	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:	Keyboard::getLastCharacter
	/*!	\brief		Returns the printable character pressed during the last update.
	**	\return		A character or 0 if no printable character was pressed
	**	\exception	none
	**	\author		John Corpening
	**	\date		10/11/2002 11:17:08 AM
	**	\warning	none
	*/
	////////////////////////////////////////////////////////////////////////////////
	char getLastCharacter() throw();
	
	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:	Keyboard::instance
	/*!	\brief		returns a pointer to the Keyboard Singleton
	**	\return		none
	**	\exception	Keyboard pointer not initialized
	**	\author		John Corpening
	**	\date		10/28/2002 3:12:18 PM
	**	\warning	none
	*/
	////////////////////////////////////////////////////////////////////////////////
	static Keyboard* instance();

	friend class PlayerInput;

protected:
	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:	Keyboard::initialize
	/*!	\brief		returns a pointer to the Keyboard Singleton
	**	\param		hwnd required for DirectInput initialization
	**	\param		DI pointer to a DirectInput instance
	**	\return		pointer to the player input Singleton
	**	\exception	none
	**	\author		John Corpening
	**	\date		10/16/2002 7:53:40 PM
	**	\warning	none
	*/
	////////////////////////////////////////////////////////////////////////////////
	static Keyboard* initialize(LPDIRECTINPUT8 DI, HWND hwnd);

	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:	Keyboard::Keyboard
	/*!	\brief		Constructor initializes DirectInput
	**	\param		hwnd required for DirectInput initialization
	**	\param		DI pointer to a DirectInput instance
	**	\return		none
	**	\exception	none
	**	\author		John Corpening
	**	\date		10/14/2002 2:22:33 PM
	**	\warning	none
	*/
	////////////////////////////////////////////////////////////////////////////////
	Keyboard(LPDIRECTINPUT8	DI, HWND hwnd);

private:
	static Keyboard*		_instance;			//!< Singleton instance
	LPDIRECTINPUTDEVICE8	DIKeyboard;			//!< DirectInput Keyboard
	char					keyInput1[256];		//!< key states
	char					keyInput2[256];		//!< second state array
	char					(*pKeyCur)[256];	//!< pointer to the current state
	char					(*pKeyPrev)[256];	//!< pointer to the previous state
	bool					validKeyInput;		//!< Flag for valid keyboard input
	char					keyToggle;			//!< Toggle to swap state arrays
	InputConstant			aConstants[26];		//!< Hard coded array of alphabetic key constants
	InputConstant			nConstants[10][2];	//!< Hard coded array of numeric key constants	
	char					lastCharacter;		//!< last printable character pressed
	bool					lcChecked;			//!< flag to indicate that the last character was pressed

	enum {HIGHBIT = 0x80};						//!< Internal constant

	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:	Keyboard::getInput
	/*!	\brief		invoked once per frame to update the state of the keyboard.
	**	\return		Returns success or failure.
	**	\exception	none
	**	\author		John Corpening
	**	\date		10/11/2002 11:16:56 AM
	**	\warning	none
	*/
	////////////////////////////////////////////////////////////////////////////////
	bool getInput() throw();


	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:	Keyboard::initializeConstantArray
	/*!	\brief		initializes the arrays of alphanumeric constants
	**	\return		none
	**	\exception	none
	**	\author		John Corpening
	**	\date		10/16/2002 8:04:56 PM
	**	\warning	none
	*/
	////////////////////////////////////////////////////////////////////////////////
	void initializeConstantsArray() throw();
};

}