////////////////////////////////////////////////////////////////////////////////
/*!	\file		Mouse.cpp
**	\brief		Mouse.h implementation file
**	\author		John Corpening
**	\date		10/16/2002 8:45:40 PM
**	\par		DESCRIPTION:
**				Contains the Mouse class which abstracts the mouse.
**				-Pressed is defined as a button is down
**				-Triggered is defined as a button was not 
**					down the last frame but is down this frame
**	\par		REMARKS:
**					You can only get an instance of the mouse from the
**					PlayerInput class. Get an instance if you will not need
**					keyboard input.
*/
////////////////////////////////////////////////////////////////////////////////
#pragma warning(disable:4786)
#include "Mouse.h"
#include "../nexception/Nexception.h"

// memory leak detection
#ifdef _DEBUG
    #include"memleak.h"
    #define new DEBUG_NEW
    #undef THIS_FILE
    static char THIS_FILE[] = __FILE__;
#endif
    
using namespace UserInput;
//Initialize Singleton instance
Mouse* Mouse::_instance = 0;

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	Mouse::initialize
/*!	\brief		returns the mouse singleton instance
**	\param		DI pointer to the DirectInput object
**	\param		hwnd handle to the window required for initialization
**	\return		a pointer to the mouse instance
**	\exception	none
**	\author		John Corpening
**	\date		10/16/2002 8:47:41 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
Mouse* Mouse::initialize(LPDIRECTINPUT8	DI, HWND hwnd)
{
	if(_instance)
	{
		return _instance;
	}
	else
	{
		_instance = new Mouse(DI, hwnd);
		return _instance;
	}
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	Mouse::instance
/*!	\brief		returns the mouse singleton instance
**	\param		DI pointer to the DirectInput object
**	\param		hwnd handle to the window required for initialization
**	\return		a pointer to the mouse instance
**	\exception	Mouse Pointer is NULL
**	\author		John Corpening
**	\date		10/16/2002 8:47:41 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
Mouse* Mouse::instance()
{
	if(!_instance)
		throw(Nexception("Mouse Pointer is NULL", __FILE__,__LINE__,__DATE__,__TIME__));

	return _instance;
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	Mouse::Mouse
/*!	\brief		Constructor creates the mouse DirectInput device
**	\param		DI pointer to the DirectInput object
**	\param		hwnd handle to the window required for initialization
**	\return		none
**	\exception	none
**	\author		John Corpening
**	\date		10/16/2002 8:47:45 PM
**	\warning	none
**	\par		UPDATED ON:			1/31/2003 9:37:39 AM
**	\par		UPDATED BY:			Nate Cleveland
**	\par		UPDATE DESCRIPTION: 
**				Changed the cooperation level of the mouse to DISCL_NONEXCLUSIVE 
				| DISCL_BACKGROUND this allows the mouse to still access the mouse
				data even when it isn't in the foreground. Slight hack that we'll
				need to change in the future, but it does allow us to have the 
				window to now be in the forground.
*/
////////////////////////////////////////////////////////////////////////////////
Mouse::Mouse(LPDIRECTINPUT8	DI, HWND hwnd)
    : validMouseInput(true)
{
	HRESULT hr;
	//Create the mouse object 
	hr = DI->CreateDevice(GUID_SysMouse, &DIMouse, NULL);

	if (FAILED(hr)) 
	{
		throw(Nexception("Mouse create device failed", __FILE__,__LINE__,__DATE__,__TIME__));
	}
	
	//Set the mouse data format
	//Note: c_dfDIMouse is a DirectInput global
	hr = DIMouse->SetDataFormat(&c_dfDIMouse);

	if (FAILED(hr)) 
	{
		throw(Nexception("Mouse SetDataFormat failed", __FILE__,__LINE__,__DATE__,__TIME__));
	}

	//Set the mouse coorperative level
	// Latest hack by Nate. Now, if the file INPUT_DEBUG.on is in the working folder
	// the game will set cooperation level such that the mouse goes to the annoying state.
	// This will though allow us to have the game in the background. [3/14/2003]
	FILE *hHack = fopen("INPUT_DEBUG.on", "r" );

	if(hHack)
	{
		hr = DIMouse->SetCooperativeLevel(hwnd, DISCL_EXCLUSIVE | DISCL_FOREGROUND);
		fclose(hHack);
	}

	else
		hr = DIMouse->SetCooperativeLevel(hwnd, DISCL_NONEXCLUSIVE | DISCL_BACKGROUND);
 
	if (FAILED(hr)) 
	{
		throw(Nexception("Mouse set coorperative level failed", __FILE__,__LINE__,__DATE__,__TIME__));
	}

	//Acquire the mouse
	hr = DIMouse->Acquire();
    if (FAILED(hr))
    {
        throw Nexception("Mouse Acquisition failed",
                         __FILE__, __LINE__, __DATE__, __TIME__);
    }
	mouseToggle = 0;
	X = Y = Z = 0;
	//return true;
}

Mouse::~Mouse()
{

}
////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	Mouse::getInput
/*!	\brief		Called once per frame to get the mouse state
**	\return		success or failure
**	\exception	none
**	\author		John Corpening
**	\date		10/16/2002 8:48:10 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
bool Mouse::getInput() throw()
{
	HRESULT hr;

	if(mouseToggle = 1 - mouseToggle)
	{
		pMouseCur = &mouseState1;
		pMousePrev = &mouseState2;
	}
	else
	{
		pMouseCur = &mouseState2;
		pMousePrev = &mouseState1;
	}
	//Clear out the current
	memset(pMouseCur, 0, sizeof(*pMouseCur));
    DIMouse->Acquire();
	hr = DIMouse->GetDeviceState(sizeof(*pMouseCur),(LPVOID)pMouseCur); 
    if FAILED(hr) 
    { 
        return false; 
    } 
	else
	{
		X += pMouseCur->lX;
		Y += pMouseCur->lY;
		Z += pMouseCur->lZ;
	}

	return true;
}

