////////////////////////////////////////////////////////////////////////////////
/*!	\file		Mouse.h
**	\brief		Class to encapsulate the mouse
**	\author		John Corpening
**	\date		10/16/2002 8:45:40 PM
**	\par		DESCRIPTION:
**				Contains the Mouse class which abstracts the mouse.
**				-Pressed is defined as a button is down
**				-Triggered is defined as a button was not 
**					down the last frame but is down this frame
**	\par		REMARKS:
**					You can only get an instance of the mouse from the
**					PlayerInput class. Get an instance if you will not need
**					keyboard input.
*/
////////////////////////////////////////////////////////////////////////////////
#pragma once
#include "PlayerInput.h"

namespace UserInput
{
////////////////////////////////////////////////////////////////////////////////
/*!	\class		Mouse
**	\brief		Mouse abstraction class
**	\author		John Corpening
**	\date		10/16/2002 8:45:47 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
class Mouse
{
	
public:

	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:	Mouse::isLeftButtonDown
	/*!	\brief		returns whether the left mouse button is down
	**	\return		down or not
	**	\exception	none
	**	\author		John Corpening
	**	\date		10/16/2002 8:45:52 PM
	**	\warning	none
	*/
	////////////////////////////////////////////////////////////////////////////////
	bool	isLeftButtonDown() const throw() {return (pMouseCur->rgbButtons[0] & HIGHBIT ? true : false);}

	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:	Mouse::isRightButtonDown
	/*!	\brief		returns whether the right mouse button is down
	**	\return		down or not
	**	\exception	none
	**	\author		John Corpening
	**	\date		10/16/2002 8:45:52 PM
	**	\warning	none
	*/
	////////////////////////////////////////////////////////////////////////////////
	bool	isRightButtonDown() const throw() {return (pMouseCur->rgbButtons[1] & HIGHBIT ? true : false);}

	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:	Mouse::isWheelDown
	/*!	\brief		returns whether the mouse wheel button is down
	**	\return		down or not
	**	\exception	none
	**	\author		John Corpening
	**	\date		10/16/2002 8:45:52 PM
	**	\warning	none
	*/
	////////////////////////////////////////////////////////////////////////////////
	bool	isWheelDown() const throw() {return (pMouseCur->rgbButtons[2] & HIGHBIT ? true : false);}

	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:	Mouse::isLeftButtonTriggered
	/*!	\brief		returns whether the left mouse button has been triggered
	**	\return		triggered or not
	**	\exception	none
	**	\author		John Corpening
	**	\date		10/16/2002 8:45:52 PM
	**	\warning	none
	*/
	////////////////////////////////////////////////////////////////////////////////
	bool	isLeftButtonTriggered() const throw()
	{return (isLeftButtonDown() && !(pMousePrev->rgbButtons[0] & HIGHBIT) ? true : false);}

	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:	Mouse::isRightButtonTriggered
	/*!	\brief		returns whether the right mouse button has been triggered
	**	\return		triggered or not
	**	\exception	none
	**	\author		John Corpening
	**	\date		10/16/2002 8:45:52 PM
	**	\warning	none
	*/
	////////////////////////////////////////////////////////////////////////////////
	bool	isRightButtonTriggered() const throw()
	{return (isRightButtonDown() && !(pMousePrev->rgbButtons[1] & HIGHBIT) ? true : false);}

	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:	Mouse::isWheelTriggered
	/*!	\brief		returns whether the mouse wheel button has been triggered
	**	\return		triggered or not
	**	\exception	none
	**	\author		John Corpening
	**	\date		10/16/2002 8:45:52 PM
	**	\warning	none
	*/
	////////////////////////////////////////////////////////////////////////////////
	bool	isWheelTriggered() const throw()
	{return (isWheelDown() && !(pMousePrev->rgbButtons[2] & HIGHBIT) ? true : false);}

	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:	Mouse::getDeltaX
	/*!	\brief		returns the relative X position of the mouse
	**	\return		the relative X position of the mouse
	**	\exception	none
	**	\author		John Corpening
	**	\date		10/16/2002 8:46:37 PM
	**	\warning	none
	*/
	////////////////////////////////////////////////////////////////////////////////
	int		getDeltaX() const throw() {return (validMouseInput ? pMouseCur->lX : 0);}

	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:	Mouse::getDeltaY
	/*!	\brief		returns the relative Y position of the mouse
	**	\return		the relative Y position of the mouse
	**	\exception	none
	**	\author		John Corpening
	**	\date		10/16/2002 8:46:37 PM
	**	\warning	none
	*/
	////////////////////////////////////////////////////////////////////////////////
	int		getDeltaY()	const throw() {return (validMouseInput ? pMouseCur->lY : 0);}

	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:	Mouse::getDeltaZ
	/*!	\brief		returns the relative Z position of the mouse
	**	\return		the relative Z position of the mouse
	**	\exception	none
	**	\author		John Corpening
	**	\date		10/16/2002 8:46:37 PM
	**	\warning	none
	*/
	////////////////////////////////////////////////////////////////////////////////
	int		getDeltaZ()	const throw() {return (validMouseInput ? pMouseCur->lZ : 0);}

	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:	Mouse::getDeltaWheel
	/*!	\brief		returns the relative Wheel position of the mouse
	**	\return		the relative Wheel position of the mouse
	**	\exception	none
	**	\author		John Corpening
	**	\date		10/16/2002 8:46:37 PM
	**	\warning	none
	*/
	////////////////////////////////////////////////////////////////////////////////
	int		getDeltaWheel() const throw() {return (validMouseInput ? pMouseCur->lZ : 0);}

	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:	Mouse::getX
	/*!	\brief		returns the absolute X position of the mouse
	**	\return		the absolute X position of the mouse
	**	\exception	none
	**	\author		John Corpening
	**	\date		10/16/2002 8:46:37 PM
	**	\warning	Because the mouse is a global entity, it is better to track
	**				the mouse position locally.
	*/
	////////////////////////////////////////////////////////////////////////////////
	int		getX() const throw() {return X;}

	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:	Mouse::getY
	/*!	\brief		returns the absolute Y position of the mouse
	**	\return		the absolute Y position of the mouse
	**	\exception	none
	**	\author		John Corpening
	**	\date		10/16/2002 8:46:37 PM
	**	\warning	Because the mouse is a global entity, it is better to track
	**				the mouse position locally.
	*/
	////////////////////////////////////////////////////////////////////////////////
	int		getY() const throw() {return Y;}

	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:	Mouse::getZ
	/*!	\brief		returns the absolute Z position of the mouse
	**	\return		the absolute Z position of the mouse
	**	\exception	none
	**	\author		John Corpening
	**	\date		10/16/2002 8:46:37 PM
	**	\warning	Because the mouse is a global entity, it is better to track
	**				the mouse position locally.
	*/
	////////////////////////////////////////////////////////////////////////////////
	int		getZ() const throw() {return Z;}

	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:	Mouse::getWheelPosition
	/*!	\brief		returns the absolute wheel position of the mouse
	**	\return		the absolute wheel position of the mouse
	**	\exception	none
	**	\author		John Corpening
	**	\date		10/16/2002 8:46:37 PM
	**	\warning	Because the mouse is a global entity, it is better to track
	**				the mouse position locally.
	*/
	////////////////////////////////////////////////////////////////////////////////
	int		getWheelPosition() const throw() {return Z;}

	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:	Mouse::setX
	/*!	\brief		sets the absolute X position of the mouse
	**	\param		x the X position to move the mouse to
	**	\return		the absolute X position of the mouse
	**	\exception	none
	**	\author		John Corpening
	**	\date		10/16/2002 8:46:37 PM
	**	\warning	Because the mouse is a global entity, it is better to track
	**				the mouse position locally.
	*/
	////////////////////////////////////////////////////////////////////////////////
	void	setX(int x) throw() {X=x;}

	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:	Mouse::setY
	/*!	\brief		sets the absolute Y position of the mouse
	**	\param		y the Y position to move the mouse to
	**	\return		the absolute Y position of the mouse
	**	\exception	none
	**	\author		John Corpening
	**	\date		10/16/2002 8:46:37 PM
	**	\warning	Because the mouse is a global entity, it is better to track
	**				the mouse position locally.
	*/
	////////////////////////////////////////////////////////////////////////////////
	void	setY(int y) throw() {Y=y;}

	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:	Mouse::setZ
	/*!	\brief		sets the absolute Z position of the mouse
	**	\param		z the Z position to move the mouse to
	**	\return		the absolute Z position of the mouse
	**	\exception	none
	**	\author		John Corpening
	**	\date		10/16/2002 8:46:37 PM
	**	\warning	Because the mouse is a global entity, it is better to track
	**				the mouse position locally.
	*/
	////////////////////////////////////////////////////////////////////////////////
	void	setZ(int z) throw() {Z=z;}

	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:	Mouse::instance
	/*!	\brief		returns the mouse singleton instance
	**	\param		DI pointer to the DirectInput object
	**	\param		hwnd handle to the window required for initialization
	**	\return		a pointer to the mouse instance
	**	\exception	Mouse Pointer is NULL
	**	\author		John Corpening
	**	\date		10/16/2002 8:47:41 PM
	**	\warning	none
	*/
	////////////////////////////////////////////////////////////////////////////////
	static Mouse* instance();

	~Mouse();

	//Allow access to the PlayerInput module
	friend class PlayerInput;

protected:

	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:	Mouse::initialize
	/*!	\brief		returns the mouse singleton instance
	**	\param		DI pointer to the DirectInput object
	**	\param		hwnd handle to the window required for initialization
	**	\return		a pointer to the mouse instance
	**	\exception	Failure to initialize
	**	\author		John Corpening
	**	\date		10/16/2002 8:47:41 PM
	**	\warning	none
	*/
	////////////////////////////////////////////////////////////////////////////////
	static Mouse*	initialize(LPDIRECTINPUT8	DI, HWND hwnd);

	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:	Mouse::Mouse
	/*!	\brief		Constructor creates the mouse DirectInput device
	**	\param		DI pointer to the DirectInput object
	**	\param		hwnd handle to the window required for initialization
	**	\return		none
	**	\exception	none
	**	\author		John Corpening
	**	\date		10/16/2002 8:47:45 PM
	**	\warning	none
	*/
	////////////////////////////////////////////////////////////////////////////////
	Mouse(LPDIRECTINPUT8	DI, HWND hwnd) throw();

private:
	static Mouse*			_instance;		//!< Singleton instance
	LPDIRECTINPUTDEVICE8	DIMouse;		//!< DirectInput Mouse
	DIMOUSESTATE			mouseState1;	//!< mouse state
	DIMOUSESTATE			mouseState2;	//!< second mouse state
	DIMOUSESTATE			*pMouseCur;		//!< pointer to the current mouse state
	DIMOUSESTATE			*pMousePrev;	//!< pointer to the previous mouse state
	bool					validMouseInput;//!< Flag for valid mouse input
	char					mouseToggle;	//!< Swaps previous and current state pointers
	int						X,Y,Z;			//!< Mouse Position
	enum {HIGHBIT = 0x80};					//!< Internal constant

	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:	Mouse::getInput
	/*!	\brief		Called once per frame to get the mouse state
	**	\return		success or failure
	**	\exception	none
	**	\author		John Corpening
	**	\date		10/16/2002 8:48:10 PM
	**	\warning	none
	*/
	////////////////////////////////////////////////////////////////////////////////
	bool getInput() throw();	
};

}