////////////////////////////////////////////////////////////////////////////////
/*!	\file		PlayerInput.cpp
**	\brief		Implementation file for PlayerInput.h
**	\author		John Corpening
**	\date		10/16/2002 7:55:42 PM
**	\par		DESCRIPTION:
**				The player input module will provide the status of input devices
**				used by project Aquarius. This information will be made available
**				to the user interface. The input module will first get the current 
**				state of all the input devices. It will then update values that 
**				have persistent components such as mouse delta positions.
**	\par		REMARKS:
**				Direct-Input will be used however the Input
**				class will encapsulate its functionality. 
*/
////////////////////////////////////////////////////////////////////////////////
#pragma warning(disable:4786)

#include"PlayerInput.h"
#include"Keyboard.h"
#include"Mouse.h"
#include "../nexception/Nexception.h"

// memory leak detection
#ifdef _DEBUG
    #include"memleak.h"
    #define new DEBUG_NEW
    #undef THIS_FILE
    static char THIS_FILE[] = __FILE__;
#endif

using namespace UserInput;
//Singleton Initialization
PlayerInput *PlayerInput::_instance = 0;

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	PlayerInput::initialize
/*!	\brief		returns a pointer to the player input Singleton
**	\param		hwnd required for DirectInput initialization
**	\return		pointer to the player input Singleton
**	\exception	none
**	\author		John Corpening
**	\date		10/16/2002 7:53:40 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
PlayerInput* PlayerInput::initialize(HWND hwnd)
{
	if(_instance)
	{
		return _instance;
	}
	else
	{
		_instance = new PlayerInput(hwnd); 
		return _instance;
	}
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	PlayerInput::instance
/*!	\brief		returns a pointer to the player input Singleton
**	\return		a pointer to the player input Singleton
**	\exception	Player Input pointer is Null
**	\author		John Corpening
**	\date		10/28/2002 3:19:15 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
PlayerInput* PlayerInput::instance()
{
	if(!_instance)
		throw(Nexception("Player Input is NULL", __FILE__,__LINE__,__DATE__,__TIME__));
	return _instance;
}

PlayerInput::~PlayerInput()
{
	if(pKeyboard)
		delete pKeyboard;
	if(pMouse)
		delete pMouse;
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	PlayerInput::PlayerInput
/*!	\brief		Constructor initializes DirectInput
**	\param		hwnd required for DirectInput initialization
**	\return		none
**	\exception	none
**	\author		John Corpening
**	\date		10/14/2002 2:22:33 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
PlayerInput::PlayerInput(HWND hwnd)
{
	//Create the DirectInput Object
	HRESULT        hr; 

	hr = DirectInput8Create( GetModuleHandle(NULL), DIRECTINPUT_VERSION, 
                                         IID_IDirectInput8, (VOID**)&DI, NULL );
	if FAILED(hr) 
	{ 
	  // DirectInput not available; take appropriate action 
		throw(Nexception("DirectInput Initialization failed",__FILE__,__LINE__,__DATE__,__TIME__));
	} 

	//Set up the keyboard
	pKeyboard = Keyboard::initialize(DI, hwnd);

	//Set up the mouse
	pMouse = Mouse::initialize(DI, hwnd);
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	PlayerInput::getUserInput
/*!	\brief		invoked once per frame to update the state of the input devices.
**	\return		It returns success or failure.
**	\exception	none
**	\author		John Corpening
**	\date		10/11/2002 11:16:56 AM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
bool PlayerInput::getUserInput() throw()
{
	if(pKeyboard->getInput() && pMouse->getInput())
		return true;
	else
		return false;
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	PlayerInput::getLastCharacter
/*!	\brief		Returns the alphanumeric character pressed during the last update.
**	\return		A character or 0 if no alphanumeric character was pressed
**	\exception	none
**	\author		John Corpening
**	\date		10/11/2002 11:17:08 AM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
char PlayerInput::getLastCharacter() const throw()
{
	return pKeyboard->getLastCharacter();
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	PlayerInput::mouseX
/*!	\brief		Returns the current x position of the mouse.
**	\return		the current x position of the mouse in screenPixelCoordinates
**	\exception	none
**	\author		John Corpening
**	\date		10/11/2002 11:17:12 AM
**	\warning	Since the mouse is a global entity, it is better to locally track
**				the mouse position locally.
*/
////////////////////////////////////////////////////////////////////////////////
int PlayerInput::mouseX( ) const throw() {return pMouse->getX();}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	PlayerInput::mouseY
/*!	\brief		Returns the current y position of the mouse.
**	\return		The current y position of the mouse
**	\exception	none
**	\author		John Corpening
**	\date		10/11/2002 11:17:16 AM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
int PlayerInput::mouseY( ) const throw() {return pMouse->getY();}
////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	PlayerInput::mouseDeltaX
/*!	\brief		Returns the change in the mouses x position.
**	\return		The change in the mouse's x position
**	\exception	none
**	\author		John Corpening
**	\date		10/11/2002 11:17:21 AM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
int	PlayerInput::mouseDeltaX( ) const throw() {return pMouse->getDeltaX();}
////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	PlayerInput::mouseDeltaY
/*!	\brief		Returns the change in the mouses y position.
**	\return		the change in the mouse's y position
**	\exception	none
**	\author		John Corpening
**	\date		10/11/2002 11:17:25 AM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
int	PlayerInput::mouseDeltaY( ) const throw() {return pMouse->getDeltaY();}
////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	PlayerInput::mouseDeltaWheel
/*!	\brief		Returns the change in the mouse wheel.
**	\return		the change in the mouse wheel
**	\exception	none
**	\author		John Corpening
**	\date		10/11/2002 11:17:30 AM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
int	PlayerInput::mouseDeltaWheel( ) const throw() {return pMouse->getDeltaZ();}


////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	PlayerInput::isPressed
/*!	\brief		Returns whether a key or mouse button was pressed.
**	\param		key an input constant
**	\return		true if pressed, false if not pressed 
**	\exception	none
**	\author		John Corpening
**	\date		10/11/2002 11:17:00 AM
**	\warning	If getUserInput() fails this will return false.
*/
////////////////////////////////////////////////////////////////////////////////
bool PlayerInput::isPressed(InputConstant key) const throw()
{
	switch(key)
	{
	case DIK_LBUTTON:
		return pMouse->isLeftButtonDown();
	case DIK_RBUTTON:
		return pMouse->isRightButtonDown();
	case DIK_WHEEL:
		return pMouse->isWheelDown();
	default:
		return pKeyboard->isPressed(key);
	}
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	PlayerInput::isTriggered
/*!	\brief		Returns whether a key or mouse button was Triggered.
**	\param		key an input constant
**	\return		true if triggered, false if not triggered 
**	\exception	none
**	\author		John Corpening
**	\date		10/11/2002 11:17:04 AM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
bool PlayerInput::isTriggered(InputConstant key) const throw() 
{
	switch(key)
	{
	case DIK_LBUTTON:
		return pMouse->isLeftButtonTriggered();
	case DIK_RBUTTON:
		return pMouse->isRightButtonTriggered();
	case DIK_WHEEL:
		return pMouse->isWheelTriggered();
	default:
		return pKeyboard->isTriggered(key);
	}
}
