////////////////////////////////////////////////////////////////////////////////
/*!	\file		Input.h
**	\brief		Functionality for getting player input
**	\author		John Corpening
**	\date		10/11/2002 10:21:57 AM
**	\par		DESCRIPTION:
**				The player input module will provide the status of input devices
**				used by project Aquarius. This information will be made available
**				to the user interface. The input module will first get the current 
**				state of all the input devices. It will then update values that 
**				have persistent components such as mouse delta positions.
**				-Pressed is defined as a button is down
**				-Triggered is defined as a button was not 
**					down the last frame but is down this frame
**	\par		REMARKS:
**				Direct-Input will be used however the Input
**				class will encapsulate its functionality. 
*/
////////////////////////////////////////////////////////////////////////////////
#pragma once
#include<windows.h>

//Direct Input versioning
#define DIRECTINPUT_VERSION		0x0800
#ifndef DIRECT3D_VERSION
#define DIRECT3D_VERSION        0x0800
#endif  
#include<dinput.h>

namespace UserInput
{
//DIK_Keys will be considered as InputConstants
typedef int InputConstant;

//Additional global KeyConstants
const int DIK_LBUTTON	= 256;
const int DIK_RBUTTON	= 257;
const int DIK_WHEEL		= 258;
const int DIK_NONE		= -1;

//Forward declarations
class Mouse;
class Keyboard;


////////////////////////////////////////////////////////////////////////////////
/*!	\class		PlayerInput
**	\brief		Class to get player input
**	\author		John Corpening
**	\date		10/11/2002 10:27:23 AM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
class PlayerInput
{
public:
	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:	PlayerInput::getUserInput
	/*!	\brief		invoked once per frame to update the state of the input devices.
	**	\return		It returns success or failure.
	**	\exception	none
	**	\author		John Corpening
	**	\date		10/11/2002 11:16:56 AM
	**	\warning	none
	*/
	////////////////////////////////////////////////////////////////////////////////
	bool	getUserInput( ) throw();

	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:	PlayerInput::isPressed
	/*!	\brief		Returns whether a key or mouse button was pressed.
	**	\param		key an input constant
	**	\return		true if pressed, false if not pressed 
	**	\exception	none
	**	\author		John Corpening
	**	\date		10/11/2002 11:17:00 AM
	**	\warning	If getUserInput() fails this will return false.
	*/
	////////////////////////////////////////////////////////////////////////////////
	bool	isPressed(InputConstant key) const throw();

	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:	PlayerInput::isTriggered
	/*!	\brief		Returns whether a key or mouse button was Triggered.
	**	\param		key an input constant
	**	\return		true if triggered, false if not triggered 
	**	\exception	none
	**	\author		John Corpening
	**	\date		10/11/2002 11:17:04 AM
	**	\warning	none
	*/
	////////////////////////////////////////////////////////////////////////////////
	bool	isTriggered(InputConstant key) const throw();

	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:	PlayerInput::getLastCharacter
	/*!	\brief		Returns the printable character pressed during the last update.
	**	\return		A character or 0 if no printable character was pressed
	**	\exception	none
	**	\author		John Corpening
	**	\date		10/11/2002 11:17:08 AM
	**	\warning	none
	*/
	////////////////////////////////////////////////////////////////////////////////
	char	getLastCharacter() const throw();

	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:	PlayerInput::mouseX
	/*!	\brief		Returns the current x position of the mouse.
	**	\return		the current x position of the mouse in screenPixelCoordinates
	**	\exception	none
	**	\author		John Corpening
	**	\date		10/11/2002 11:17:12 AM
	**	\warning	Since the mouse is a global entity, it is better to locally track
	**				the mouse position locally.
	*/
	////////////////////////////////////////////////////////////////////////////////
	int	mouseX( ) const throw();

	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:	PlayerInput::mouseY
	/*!	\brief		Returns the current y position of the mouse.
	**	\return		The current y position of the mouse
	**	\exception	none
	**	\author		John Corpening
	**	\date		10/11/2002 11:17:16 AM
	**	\warning	Since the mouse is a global entity, it is better to locally track
	**				the mouse position locally.
	*/
	////////////////////////////////////////////////////////////////////////////////
	int	mouseY( ) const throw();
	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:	PlayerInput::mouseDeltaX
	/*!	\brief		Returns the change in the mouses x position.
	**	\return		The change in the mouse's x position
	**	\exception	none
	**	\author		John Corpening
	**	\date		10/11/2002 11:17:21 AM
	**	\warning	none
	*/
	////////////////////////////////////////////////////////////////////////////////
	int		mouseDeltaX( ) const throw();
	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:	PlayerInput::mouseDeltaY
	/*!	\brief		Returns the change in the mouses y position.
	**	\return		the change in the mouse's y position
	**	\exception	none
	**	\author		John Corpening
	**	\date		10/11/2002 11:17:25 AM
	**	\warning	none
	*/
	////////////////////////////////////////////////////////////////////////////////
	int		mouseDeltaY( ) const throw();
	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:	PlayerInput::mouseDeltaWheel
	/*!	\brief		Returns the change in the mouse wheel.
	**	\return		the change in the mouse wheel
	**	\exception	none
	**	\author		John Corpening
	**	\date		10/11/2002 11:17:30 AM
	**	\warning	none
	*/
	////////////////////////////////////////////////////////////////////////////////
	int		mouseDeltaWheel( ) const throw();

	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:	PlayerInput::getMouseInstance
	/*!	\brief		returns a pointer to the mouse
	**	\return		a Pointer to the mouse instance
	**	\exception	none
	**	\author		John Corpening
	**	\date		10/16/2002 7:51:49 PM
	**	\warning	none
	*/
	////////////////////////////////////////////////////////////////////////////////
	Mouse* getMouseInstance() const throw() {return pMouse;}

	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:	PlayerInput::getKeyboardInstance
	/*!	\brief		returns a pointer to the keyboard 
	**	\return		pointer to the keyboard
	**	\exception	none
	**	\author		John Corpening
	**	\date		10/16/2002 7:52:50 PM
	**	\warning	none
	*/
	////////////////////////////////////////////////////////////////////////////////
	Keyboard* getKeyboardInstance() const throw() {return pKeyboard;}

	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:	PlayerInput::initialize
	/*!	\brief		returns a pointer to the player input Singleton
	**	\param		hwnd required for DirectInput initialization
	**	\return		pointer to the player input Singleton
	**	\exception	none
	**	\author		John Corpening
	**	\date		10/16/2002 7:53:40 PM
	**	\warning	none
	*/
	////////////////////////////////////////////////////////////////////////////////
	static PlayerInput* initialize(HWND hwnd);

	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:	PlayerInput::instance
	/*!	\brief		returns a pointer to the player input Singleton
	**	\return		a pointer to the player input Singleton
	**	\exception	Player Input pointer is Null
	**	\author		John Corpening
	**	\date		10/28/2002 3:19:15 PM
	**	\warning	none
	*/
	////////////////////////////////////////////////////////////////////////////////
	static PlayerInput* instance();

	~PlayerInput();

protected:

	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:	PlayerInput::PlayerInput
	/*!	\brief		Constructor initializes DirectInput
	**	\param		hwnd required for DirectInput initialization
	**	\return		none
	**	\exception	Nexception for DirectInput problems
	**	\author		John Corpening
	**	\date		10/14/2002 2:22:33 PM
	**	\warning	none
	*/
	////////////////////////////////////////////////////////////////////////////////
	PlayerInput(HWND hwnd);
private:

	static	PlayerInput*	_instance;			//!< Singleton instance
	LPDIRECTINPUT8			DI;					//!< DirectInput Object
	Mouse					*pMouse;			//!< Pointer to the mouse object
	Keyboard				*pKeyboard;			//!< Pointer to the keyboard object

	//Internal constants
	enum {HIGHBIT = 0x80};
};


}//end UserInput namespace