////////////////////////////////////////////////////////////////////////////////
/*!	\file		ActiveSound_Fmod.h
**	\brief		Declares the ActiveSound_Fmod class.
**	\author		Nate Cleveland
**	\date		10/14/2002 12:40:03 PM
**	\par		DESCRIPTION: See class header for description.					
*/
////////////////////////////////////////////////////////////////////////////////
#pragma once
#include "threadSafety.h"
#include "HSound.h"
#include "GivePos.h"

namespace AcqariusSound
{	
////////////////////////////////////////////////////////////////////////////////
/*!	\class		ActiveSound_Fmod
**	\brief		Holds simple data about an actively playing sound
**	\author		Nate Cleveland
**	\date		10/3/2002 2:25:20 PM
**	\warning	none
**	\par		Holds simple data about an actively playing sound. These are only
**				 deleted when the entire class is destroyed. That is because they
**				 reference a specific channel more than a actual sound. This means
**				 that the first time a sound is played we create this and give it 
**				 specific channel. However, we won't be notified when the sound is 
**				 done playing, we'll have to check for it. This means it is possible 
**				 that a new sound could be assigned to a channel before we have a 
**				 chance to mark it.
**	\par		Left as a struct rather then a class because of the simplicty of it. 
**				 A ctor is provided for ease of initialization.
*/
////////////////////////////////////////////////////////////////////////////////
struct ActiveSound_Fmod
{
	//!channel that this active sound is tied to. This relationship is permenate, thus
	//! the constness of the member.
	const int		m_channel;

	//!HSound of the last know playing sound. Set to zero if the sound is done playing.
	HSound			m_hSound;

	//!this flag tells us that this is currently a threeD sound and will need to have 
	//! its position updated.
	bool			m_isThreeD;

	//!If the above flag is true then we'll use this below member to update the position 
	//! of the sound. 
	GivePos		const *m_pos;

public:

	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:	
	/*!	\brief		Loades the data into the struct.
	**	\param		
	**	\arg  \c int channel		Fmod Handle of the sound that this is supposed to hold.
	**	\arg  \c HSound &h		HSound of the sound that is to be played.
	**	
	**
	**	\exception	Reserves the right to throw an exception.		
	**	\par		DESCRIPTION:
	**				Defualt ctor for the ActiveSound_Fmod class. Loades the data into the
	**				 struct.
	**
	**					
	**	\author		Nate Cleveland
	**	\date		10/3/2002 2:35:35 PM
	**	\warning	none
	*/
	////////////////////////////////////////////////////////////////////////////////
	ActiveSound_Fmod(int channel, HSound &h) throw();


	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:	
	/*!	\brief		Cleans up after class.  
	**	\param		NONE
	**	\exception	NONE		
	**	\par		DESCRIPTION: Doesn't have any global effects.
	**				
	**	\author		Nate Cleveland
	**	\date		10/3/2002 2:35:35 PM
	**	\warning	none
	*/
	////////////////////////////////////////////////////////////////////////////////
	~ActiveSound_Fmod() throw();

	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:	
	/*!	\brief		Calls an fmod function and checks to see if it still playing. 
	**	\param		NONE
	**	\exception	NONE		
	**	\par		DESCRIPTION:
	**				Calls an fmod function and checks to see if it still playing. This 
	**				 should be called fairly (ie every frame). If the sound is done 
	**				 playing the h_hSound member will be set to zero.
	**
	**					
	**	\author		Nate Cleveland
	**	\date		10/3/2002 2:35:35 PM
	**	\warning	none
	*/
	////////////////////////////////////////////////////////////////////////////////
	bool IsDone(void) throw();
};//end ActiveSound_Fmod class
}//end AcqariusSound::

