#pragma once
#include "sound.h"

namespace AcqariusSound
{

	////////////////////////////////////////////////////////////////////////////////
	/*!	\class		BodyClient
	**	\brief		Provides an interface into the Sound class for fish bodies.
	**	\author		Nate Cleveland
	**	\date		10/14/2002 11:18:03 AM
	**	\warning	none
	*/
	////////////////////////////////////////////////////////////////////////////////
	class BodyClient
	{

	public:

		////////////////////////////////////////////////////////////////////////////////
		//	FUNCTION:	
		/*!	\brief		Defualt ctor for bodyClient
		**	\param		none
		**	\exception	none
		**	\par		DESCRIPTION:
		**				 Sets all members to NULL/0.	
		**	\author		Nate Cleveland
		**	\date		12/2/2002 6:58:53 PM
		**	\warning	none
		*/
		////////////////////////////////////////////////////////////////////////////////
		BodyClient():m_sound(NULL), m_bite(0), m_hit(0), m_lift(0), m_think(0),
		m_drop(0), m_swim(0), m_collideTerrain(0), m_collideBody(0), m_eat(0){}
		
		////////////////////////////////////////////////////////////////////////////////
		//	FUNCTION:	
		/*!	\brief		Provides functionality for giving a client a sound pointer.
		**	\param		pSound		Pointer to sound module.
		**	\exception	none
		**	\par		DESCRIPTION:
		**				 Loads the sound pointer, sets all other members to 0.	
		**	\author		Nate Cleveland
		**	\date		12/2/2002 6:59:37 PM
		**	\warning	none
		*/
		////////////////////////////////////////////////////////////////////////////////
		BodyClient(Sound* pSound):m_sound(pSound), m_bite(0), m_hit(0), m_lift(0), m_think(0),
		m_drop(0), m_swim(0), m_collideTerrain(0), m_collideBody(0), m_eat(0){}
		

		~BodyClient()
		{
			if(m_sound)
				m_sound->Deregister(m_pos);
			
		}

		////////////////////////////////////////////////////////////////////////////////
		//	FUNCTION:	
		/*!	\brief		Connects a BodyClient to a sound object.
		**	\param		pSound		Pointer to the sound module.
		**	\return		none
		**	\exception	none
		**	\par		DESCRIPTION:
		**				 Loads the sound pointer.	
		**	\author		Nate Cleveland
		**	\date		12/2/2002 7:00:38 PM
		**	\warning	none
		*/
		////////////////////////////////////////////////////////////////////////////////
		void SetSound(Sound* pSound){m_sound = pSound;}

		////////////////////////////////////////////////////////////////////////////////
		//	FUNCTION:	
		/*!	\brief		Plays a "bite" sound once and only once.
		**	\param		none
		**	\return		False if the sound is not playing after the call.
		**	\exception	none
		**	\par		DESCRIPTION:
		**				 Plays a "bite" sound once and only once. Thus, if this fish's
		**				  bite sound is already playing another one will not be started
		**				  and the current sound will allowed to finish.
		**	\par		
		**				Sound is inteneded to be played every time this fish attacks or
						 bites something.
		**	\author		Nate Cleveland
		**	\date		12/2/2002 5:51:46 PM
		**	\warning	none
		*/
		////////////////////////////////////////////////////////////////////////////////
		bool Bite(void)throw(){return PlaySingleSound(m_bite);}

		////////////////////////////////////////////////////////////////////////////////
		//	FUNCTION:	
		/*!	\brief		Plays a "Hit" sound once and only once.
		**	\param		none
		**	\return		False if the sound is not playing after the call.
		**	\exception	none
		**	\par		DESCRIPTION:
		**				 Plays a "Hit" sound once and only once. Thus, if this fish's
		**				  Hit sound is already playing another one will not be started
		**				  and the current sound will allowed to finish.
		**	\par		
		**				Sound is inteneded to be played every time this fish is attacked.
		**	\author		Nate Cleveland
		**	\date		12/2/2002 5:51:46 PM
		**	\warning	none
		*/
		////////////////////////////////////////////////////////////////////////////////
		bool Hit(void)throw(){return PlaySingleSound(m_hit);}

		////////////////////////////////////////////////////////////////////////////////
		//	FUNCTION:	
		/*!	\brief		Plays a "Swim" sound once and only once.
		**	\param		none
		**	\return		False if the sound is not playing after the call.
		**	\exception	none
		**	\par		DESCRIPTION:
		**				 Plays a "Swim" sound once and only once. Thus, if this fish's
		**				  Swim sound is already playing another one will not be started
		**				  and the current sound will allowed to finish.
		**	\par		
		**				Sound is inteneded to be played every time this fish's speed is
						 greater then delta (a very small value).
		**	\author		Nate Cleveland
		**	\date		12/2/2002 5:51:46 PM
		**	\warning	none
		*/
		////////////////////////////////////////////////////////////////////////////////
		bool Swim(void)throw(){return PlaySingleSound(m_swim);}

		////////////////////////////////////////////////////////////////////////////////
		//	FUNCTION:	
		/*!	\brief		Stops a swiming sound from playing.
		**	\param		none
		**	\return		False if the sound is playing after the call.
		**	\exception	none
		**	\par		DESCRIPTION:
		**				 Stops this fish's swim sound from playing.
		**
		**	\author		Nate Cleveland
		**	\date		12/2/2002 5:51:46 PM
		**	\warning	none
		*/
		////////////////////////////////////////////////////////////////////////////////
		bool StopSwim(void)throw(){return m_sound->EndSound(m_swim);}

		////////////////////////////////////////////////////////////////////////////////
		//	FUNCTION:	
		/*!	\brief		Plays a "Think" sound once and only once.
		**	\param		none
		**	\return		False if the sound is not playing after the call.
		**	\exception	none
		**	\par		DESCRIPTION:
		**				 Plays a "Think" sound once and only once. Thus, if this fish's
		**				  Think sound is already playing another one will not be started
		**				  and the current sound will allowed to finish.
		**
		**	\par		
		**				Sound is inteneded to be played every time this fish recieves a
		**				 new command from the brain.
		**
		**	\author		Nate Cleveland
		**	\date		12/2/2002 5:51:46 PM
		**	\warning	none
		*/
		////////////////////////////////////////////////////////////////////////////////
		bool Think(void)throw(){return PlaySingleSound(m_think);}

		////////////////////////////////////////////////////////////////////////////////
		//	FUNCTION:	
		/*!	\brief		Plays a "Lift" sound once and only once.
		**	\param		none
		**	\return		False if the sound is not playing after the call.
		**	\exception	none
		**	\par		DESCRIPTION:
		**				 Plays a "Lift" sound once and only once. Thus, if this fish's
		**				  Lift sound is already playing another one will not be started
		**				  and the current sound will allowed to finish.
		**
		**	\par		
		**				Sound is inteneded to be played every time this fish lifts an 
						 item.
		**
		**	\author		Nate Cleveland
		**	\date		12/2/2002 5:51:46 PM
		**	\warning	none
		*/
		////////////////////////////////////////////////////////////////////////////////
		bool Lift(void)throw(){return PlaySingleSound(m_lift);}


		////////////////////////////////////////////////////////////////////////////////
		//	FUNCTION:	
		/*!	\brief		Plays a "Drop" sound once and only once.
		**	\param		none
		**	\return		False if the sound is not playing after the call.
		**	\exception	none
		**	\par		DESCRIPTION:
		**				 Plays a "Drop" sound once and only once. Thus, if this fish's
		**				  Drop sound is already playing another one will not be started
		**				  and the current sound will allowed to finish.
		**
		**	\par		
		**				Sound is inteneded to be played every time this fish drops an item.
		**
		**	\author		Nate Cleveland
		**	\date		12/2/2002 5:51:46 PM
		**	\warning	none
		*/
		////////////////////////////////////////////////////////////////////////////////
		bool Drop(void)throw(){return PlaySingleSound(m_drop);}


		////////////////////////////////////////////////////////////////////////////////
		//	FUNCTION:	
		/*!	\brief		Plays a "Death" sound once and only once.
		**	\param		none
		**	\return		False if the sound is not playing after the call.
		**	\exception	none
		**	\par		DESCRIPTION:
		**				 Plays a "Death" sound once and only once. Thus, if this fish's
		**				  Death sound is already playing another one will not be started
		**				  and the current sound will allowed to finish.
		**
		**	\par		
		**				Sound is inteneded to be played every time this fish Deaths an item.
		**
		**	\author		Nate Cleveland
		**	\date		12/2/2002 5:51:46 PM
		**	\warning	none
		*/
		////////////////////////////////////////////////////////////////////////////////
		bool Death(void)throw(){return PlaySingleSound(m_death);}


		////////////////////////////////////////////////////////////////////////////////
		//	FUNCTION:	
		/*!	\brief		Plays the body collision sound once and only once.
		**	\param		none
		**	\return		False if the sound is not playing after the call.
		**	\exception	none
		**	\par		DESCRIPTION:
		**				 Plays a collision sound once and only once. Thus, if this fish's
		**				  body collision sound is already playing and this method is called
						  only one sound per fish will be played.
		**
		**	\par		
		**				Sound is inteneded to be played every time this fish collides with
						 another fish.
		**
		**	\author		Nate Cleveland
		**	\date		12/2/2002 5:51:46 PM
		**	\warning	none
		*/
		////////////////////////////////////////////////////////////////////////////////
		bool CollideBody(void)throw(){return PlaySingleSound(m_collideBody);}

		////////////////////////////////////////////////////////////////////////////////
		//	FUNCTION:	
		/*!	\brief		Plays the terrain collision sound once and only once.
		**	\param		none
		**	\return		False if the sound is not playing after the call.
		**	\exception	none
		**	\par		DESCRIPTION:
		**				 Plays a collision sound once and only once. Thus, if this fish's
		**				  terrain collision sound is already playing and this method is called
						  only one sound per fish will be played.
		**
		**	\par		
		**				Sound is inteneded to be played every time this fish collides with
						 the terrain.
		**
		**	\author		Nate Cleveland
		**	\date		12/2/2002 5:51:46 PM
		**	\warning	none
		*/
		////////////////////////////////////////////////////////////////////////////////
		bool CollideTerrain(void)throw(){return PlaySingleSound(m_collideTerrain);}

		////////////////////////////////////////////////////////////////////////////////
		//	FUNCTION:	
		/*!	\brief		Plays the terrain collision sound once and only once.
		**	\param		none
		**	\return		False if the sound is not playing after the call.
		**	\exception	none
		**	\par		DESCRIPTION:
		**				 Plays a collision sound once and only once. Thus, if this fish's
		**				  terrain collision sound is already playing and this method is called
						  only one sound per fish will be played.
		**
		**	\par		
		**				Sound is inteneded to be played every time this fish collides with
						 the terrain.
		**
		**	\author		Nate Cleveland
		**	\date		12/2/2002 5:51:46 PM
		**	\warning	none
		*/
		////////////////////////////////////////////////////////////////////////////////
		bool Eat(void)throw(){return PlaySingleSound(m_eat);}

		////////////////////////////////////////////////////////////////////////////////
		//	FUNCTION:	
		/*!	\brief		Loads all sound files for the caller.
		**
		**	\param		bite		Bite sound file pathName. 
		**	\param		hit			Hit sound file pathName.
		**	\param		swim		Swim sound file pathName.
		**	\param		think		Think sound file pathName.
		**	\param		lift		Lift sound file pathName.
		**	\param		drop		Drop sound file pathName.
		**	\param		eat			Sound played whenever the fish eats something.
		**
		**	\return		returns false if any sound were not loaded.
		**	\exception	none
		**	\par		DESCRIPTION:
		**				 Loads this fish's sounds. Files will NOT be streamed. Will 
		**				  release any sounds that are to be replaced and any parameter
		**				  may be set to null.
		**	\author		Nate Cleveland
		**	\date		12/2/2002 6:01:12 PM
		**	\warning	none
		*/
		////////////////////////////////////////////////////////////////////////////////
		bool LoadSounds(const char* bite=NULL, const char* hit=NULL, const char* swim=NULL,
						const char* think=NULL, const char* lift=NULL, const char* drop=NULL,
						const char* collideBody=NULL, const char* collideTerrain=NULL, 
						const char* death=NULL, const char* eat = NULL) throw();


		//below are the defines for the classes that'll hold our position and direction
		//vectors.
		////////////////////////////////////////////////////////////////////////////////
		/*!	\class		BodyGivePos
		**	\brief		A give pos class for fish to use for their 3D sounds.
		**	\author		Nate Cleveland
		**	\date		12/2/2002 6:56:50 PM
		**	\warning	none
		*/
		////////////////////////////////////////////////////////////////////////////////
		class BodyGivePos: public GivePos
		{
			float m_elements[3];		//!< Vector array for class.
			
		public:			
			//!Sets position to zero.
			BodyGivePos()throw(){memset(m_elements, 0, sizeof(float)*3);}

			bool IsRelevent(void) const{return true;}
			float x(void)const throw(){return m_elements[0];}
			float y(void)const throw(){return m_elements[1];}
			float z(void)const throw(){return m_elements[2];}
			bool operator () (Point3d &pnt) const
			{
				pnt.x = x();
				pnt.y = y();
				pnt.z = z();
				return true;
			}
			
			float x(float newVal)throw(){m_elements[0] = newVal;}
			float y(float newVal)throw(){m_elements[1] = newVal;}
			float z(float newVal)throw(){m_elements[2] = newVal;}
			
			//does not check for null.
			void operator()(float *newVals){memcpy(m_elements, newVals, sizeof(float)*3);}
		}m_pos;								//!< Postion of the fish

	protected:
		Sound	*m_sound;					//!< Pointer to the sound object.
		HSound	m_bite;						//!< Handle to a loaded bite sound.
		HSound	m_hit;						//!< Handle to a loaded hit sound.
		HSound	m_swim;						//!< Handle to a loaded swim sound.
		HSound	m_think;					//!< Handle to a loaded think sound.
		HSound	m_lift;						//!< Handle to a loaded lift sound.
		HSound	m_drop;						//!< Handle to a loaded drop sound.
		HSound	m_collideBody;				//!< Sound to play if fish collides with another.
		HSound	m_collideTerrain;			//!< Sound for collision with the terrian.
		HSound	m_death;					//!< Sound to play when this fish dies.
		HSound	m_eat;						//!< Sound to play when the fish eats something.



		////////////////////////////////////////////////////////////////////////////////
		//	FUNCTION:	
		/*!	\brief		Loads a single sound for the BodyClient
		**	\param		oldSound		Reference to HSound we want to load to.
		**	\param		pathName		File name to load from.
		**	\return		returns false if unable to load the file.
		**	\exception	none
		**	\par		DESCRIPTION:
		**				 Loads a single sound for the body client. Caller must check
		**				  m_sound pointer before calling, this method assumes that it
		**				  is valid. Path Name must also be checked for null before calling.
		**				  Old sound will be freed from memory if prevous sound was safely
		**				  allocated.
		**	\author		Nate Cleveland
		**	\date		12/2/2002 6:33:19 PM
		**	\warning	m_sound pointer must be checked for null before calling method.
		*/
		////////////////////////////////////////////////////////////////////////////////
		bool BodyClient::LoadSingleSound(HSound& oldSound, const char* pathName)throw();

		////////////////////////////////////////////////////////////////////////////////
		//	FUNCTION:	
		/*!	\brief		Plays a sound once and only once.
		**	\param		hSound		Handle of sound to play.
		**	\return		false if unable to play sound.
		**	\exception	none
		**	\par		DESCRIPTION:
		**				 Plays a sound iff it isn't already playing. Checks m_sound 
		**				  pointer before using it. Returns false if m_sound == NULL in
		**				  release mode, asserts m_sound in debug.
		**	\author		Nate Cleveland
		**	\date		12/2/2002 6:41:54 PM
		**	\warning	none
		*/
		////////////////////////////////////////////////////////////////////////////////
		bool BodyClient::PlaySingleSound(HSound& hSound) throw();
	};//end BodyClient
}//end namespace AquariusSound.