////////////////////////////////////////////////////////////////////////////////
/*!	\file		FmodOps.h
**	\brief		Defines the FmodOps struct.
**	\author		Nate Cleveland
**	\date		10/14/2002 12:39:11 PM
**	\par		DESCRIPTION: See class header for description.					
*/
////////////////////////////////////////////////////////////////////////////////
#pragma once

namespace AcqariusSound
{

//!This is the minimum that the hz parameter of the FmodOps ctor can be set to
//! as defined by FMOD documentation.
const int FMOD_MIN_HZ =	4000;

//!This is the maximum that the hz parameter of the FmodOps ctor can be set to
//! as defined by FMOD documentation.
const int FMOD_MAX_HZ = 65535;

//!By defualt this is what the FmodOps ctor will set the number of channels to.
const int FMOD_DEFUALT_CHANNELS = 32;

////////////////////////////////////////////////////////////////////////////////
/*!	\class		FmodOps
**	\brief		This is the generic options struct that is passed to the Sound_Fmod::Init method.
**	\author		Nate Cleveland
**	\date		10/3/2002 2:43:35 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
struct FmodOps
{
	//!The type of output you wish the system to use. Will auto detect if you
	//! set the flag to -1. Fmod documentation claims that Win NT's usage of 
	//! DSound is just a waveout wrapper and is therefor inferior. But on all other
	//! win ops DSound is better. Best to leave at -1 so fmod can detect this for us
	//! but if we have problems with sound it might be a good idea to force one or the
	//! other.
	int		m_soundOutput;

	//!This is the mix rate for the sound. 44100 is the common stereo value but number
	//! can go between 4000 and 65535.
	int		m_hz;

	//!If set to -1 system will calculate the max number of channels we can use. This may
	//! be bad since we may not have enough hardware channels to support our game.
	//! System will defualt the ammount to FMOD_DEFUALT_CHANNELS
	int		m_channelNumber;


	//!Flags to be given to the fmod init function.
	int		m_flags;

	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:	
	/*!	\brief		Init method for Sound_Fmod. Must be called for sounds to work.
	**	\param
	**	\arg  int hz			Frequency to set output at. See m_hz
	**	\arg  int soundOut	Type of sound output. See m_soundOutput. 
	**	\arg  int channels	The max number of channels.  See m_channels
	**	\arg  int flags		See m_flags
	**
	**	\exception	Reserves the right to throw exception and will do so iff soundOut
	**				 or channels paramaters are unreasonable.	
	**	\par		DESCRIPTION:
	**				Overloaded to good defualt values.
	**
	**					
	**	\author		Nate Cleveland
	**	\date		10/3/2002 2:35:35 PM
	**	\warning	none
	*/
	////////////////////////////////////////////////////////////////////////////////
	FmodOps(int hz = 44100, int soundOut = -1, int channels = FMOD_DEFUALT_CHANNELS,
		int flags = 0)
		:m_soundOutput(soundOut), m_channelNumber(channels), m_flags(flags)
	{
		//Do bounds check on all the members
		if(hz < FMOD_MIN_HZ)
			m_hz = FMOD_MIN_HZ;
		else if(hz > FMOD_MAX_HZ)
			m_hz = FMOD_MAX_HZ;
		else
			m_hz = hz;
		
#ifdef _DEBUG
		if(soundOut < -1)
			throw 0;

		if(channels < -1)
			throw 0;
#endif
	}//end ctor
};//end FmodOps
}//end AcqariusSound::