////////////////////////////////////////////////////////////////////////////////
/*!	\file		GivePos.h
**	\brief		Defines the GivePos and GiveCntrPos classes.
**	\author		Nate Cleveland
**	\date		10/14/2002 12:32:42 PM
**	\par		DESCRIPTION: See class headers for Descriptions.
*/
////////////////////////////////////////////////////////////////////////////////
#pragma once
#include "affine.h"
namespace AcqariusSound
{

////////////////////////////////////////////////////////////////////////////////
/*!	\class		GetPos
**	\brief		Provides an easy way for the pathfinding and sound modules to
**				 update their positions.
**	\par		Users mst derive publicly from this class and write a derived class
**				 that suites their needs.
**	\par		Users MUST retain a pointer to the object and set it where IsRelevent
**				 returns false when the sound/pathAI is no longer needed.
**	\par		REMARKS: Exists within the AcqariusSound namespace.
**	\author		Nate Cleveland
**	\date		9/18/2002 12:49:37 PM
**	\warning	The method of thread safety for this class has yet to be addressed
*/
////////////////////////////////////////////////////////////////////////////////
class GivePos
{
public:


	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:
	/*!	\brief		Must return false if we can no longer call this class safely.
	**	\param		none
	**	\return		boolean true/false.
	**	\exception	none
	**	\par		DESCRIPTION:
	**					This function is introduced so that users may avoid having
						 pathfinding calculated if it is no needed at this time.
	**	\author		Nate Cleveland
	**	\date		9/18/2002 12:53:21 PM
	**	\warning	none
	*/
	////////////////////////////////////////////////////////////////////////////////
	virtual bool IsRelevent(void) const = 0;


	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:
	/*!	\brief		Must return the x position of the object GetPos is connected to.
	**	\param		none
	**	\return		float that is the x position of the object GetPos is connected to.
	**	\exception	none
	**	\par		DESCRIPTION:
	**					Must return the x position of the object GetPos is connected to.
	**
	**	\author		Nate Cleveland.
	**	\date		9/18/2002 12:56:50 PM
	**	\warning	none
	*/
	////////////////////////////////////////////////////////////////////////////////
	virtual float x(void) const = 0;

	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:
	/*!	\brief		Must return the y position of the object GetPos is connected to.
	**	\param		none
	**	\return		loat that is the y position of the object GetPos is connected to.
	**	\exception	none
	**	\par		DESCRIPTION:
	**					Must return the y position of the object GetPos is connected to.
	**	\author		Nate Cleveland
	**	\date		9/18/2002 12:58:39 PM
	**	\warning	none
	*/
	////////////////////////////////////////////////////////////////////////////////
	virtual float y(void) const = 0;

	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:
	/*!	\brief		Must return the z position of the object GetPos is connected to.
	**	\param
	**	\return		Returns the z position of the object GetPos is connected to.
	**	\exception	none
	**	\par		DESCRIPTION: Must return the z position of the object GetPos is
	**				 connected to.
	**
	**	\author		Nate Cleveland
	**	\date		10/4/2002 12:31:19 PM
	**	\warning	none
	*/
	////////////////////////////////////////////////////////////////////////////////
	virtual float z(void) const = 0;

	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:
	/*!	\brief		Must return the y position of the object GetPos is connected to.
	**	\param
	**	\return		Returns the y position of the object GetPos is connected to.
	**	\exception	none
	**	\par		DESCRIPTION: Must return the y position of the object GetPos is
	**				 connected to.
	**
	**	\author		Nate Cleveland
	**	\date		10/4/2002 12:31:19 PM
	**	\warning	none
	*/
	////////////////////////////////////////////////////////////////////////////////
	virtual bool operator () (Point3d &pntToFill) const = 0;
};


////////////////////////////////////////////////////////////////////////////////
/*!	\class		GiveCntrPos
**	\brief		Same as GivePos, only must gaurantee that it returns the cntr pnt.
**	\author		Nate Cleveland
**	\date		10/4/2002 12:32:29 PM
**	\par		REMARKS: Exists within the AcqariusSound namespace.
**	\warning	none
**	\example	To use simply create a GivePos class and use that as the template
**				 parameter. Also proved a Point3d that will be the distance from
**				 the GivePos data to the center of the object.
*/
////////////////////////////////////////////////////////////////////////////////
template <class givePos>
class GiveCntrPos
{
	//!givePos function that is provided at runtime
	givePos		m_pos;

	//!distance to the center from the givePos return values.
	Point3d		m_distToCntr;
public:

	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:
	/*!	\brief		ctor
	**	\param Point3d &distToCntr		Distance offset for this item.
	**	\return		none
	**	\exception	none
	**	\par		DESCRIPTION: Recommened ctor that gives users a chance to specifiy
					 the distance to the center point.
	**
	**	\author		Nate Cleveland
	**	\date		10/4/2002 12:33:25 PM
	**	\warning	none
	*/
	////////////////////////////////////////////////////////////////////////////////
	GiveCntrPos(Point3d &distToCntr)
		:m_distToCntr(distToCntr)
	{}

	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:
	/*!	\brief		Defualt ctor that sets the distance to the center point to zero.
	**	\param		none
	**	\return		none
	**	\exception	none
	**	\par		DESCRIPTION: Defualt ctor that sets the distance to the center point
	**				 to zero.
	**
	**	\author		Nate Cleveland
	**	\date		10/4/2002 12:35:42 PM
	**	\warning	none
	*/
	////////////////////////////////////////////////////////////////////////////////
	GiveCntrPos(void){}

	/*!
		\brief
	*/
	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:
	/*!	\brief		Calls the passed in givePos class and returns its answer to the caller.
	**	\param
	**	\return		none
	**	\exception	none
	**	\par		DESCRIPTION: Calls the passed in givePos class and returns its answer
	**				 to the caller.
	**
	**	\author		Nate Cleveland
	**	\date		10/4/2002 12:37:47 PM
	**	\warning	none
	*/
	////////////////////////////////////////////////////////////////////////////////
	bool IsRelevent(void){return m_pos.IsRelevent();}


	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:
	/*!	\brief		Calls the givePos member and adds the distance to the center before
	**				 returning the value.
	**	\param
	**	\return		none
	**	\exception	none
	**	\par		DESCRIPTION: Calls the givePos member and adds the distance to the
	**				 center before returning the value.
	**
	**	\author		Nate Cleveland
	**	\date		10/4/2002 12:38:32 PM
	**	\warning	none
	*/
	////////////////////////////////////////////////////////////////////////////////
	float x(void){return m_pos.x() + m_distToCntr.x;}

	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:
	/*!	\brief		Calls the givePos member and adds the distance to the center before
	**				 returning the value.
	**	\param
	**	\return		none
	**	\exception	none
	**	\par		DESCRIPTION: Calls the givePos member and adds the distance to the
	**				 center before returning the value.
	**
	**	\author		Nate Cleveland
	**	\date		10/4/2002 12:38:32 PM
	**	\warning	none
	*/
	////////////////////////////////////////////////////////////////////////////////
	float y(void){return m_pos.y() + m_distToCntr.y};


	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:
	/*!	\brief		Calls the givePos member and adds the distance to the center before
	**				 returning the value.
	**	\param
	**	\return		none
	**	\exception	none
	**	\par		DESCRIPTION: Calls the givePos member and adds the distance to the
	**				 center before returning the value.
	**
	**	\author		Nate Cleveland
	**	\date		10/4/2002 12:38:32 PM
	**	\warning	none
	*/
	////////////////////////////////////////////////////////////////////////////////
	float z(void){return m_pos.z() + m_distToCntr.z};

	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:
	/*!	\brief		Calls the givePos member and adds the distance to the center before
	**				 returning the values.
	**
	**	\param Point3d &pntToFill		Pnt that the user wishes to fill with the
	**										 location of this item.
	**	\return		none
	**	\exception	none
	**	\par		DESCRIPTION: Calls the givePos member and adds the distance to the
	**				 center before returning the values.
	**
	**	\author		Nate Cleveland
	**	\date		10/4/2002 12:38:32 PM
	**	\warning	none
	*/
	////////////////////////////////////////////////////////////////////////////////
	bool operator () (Point3d &pntToFill)
	{
		pntToFill.x = x() + m_distToCntr.x;
		pntToFill.y = y() + m_distToCntr.y;
		pntToFill.z = z() + m_distToCntr.z;
	}
};
}//end namespace AcqariusSound