////////////////////////////////////////////////////////////////////////////////
/*!	\file		LoadedSound_Fmod.h
**	\brief		Declares the LoadedSound_Fmod class.
**	\author		Nate Cleveland
**	\date		10/14/2002 12:37:39 PM
**	\par		DESCRIPTION: See class header for description.					
*/
////////////////////////////////////////////////////////////////////////////////
#pragma once

namespace AcqariusSound
{
////////////////////////////////////////////////////////////////////////////////
/*!	\class		LoadedSound_Fmod
**	\brief		Holds all the data for the Sound_FMOD class to use.	
**	\author		Nate Cleveland
**	\date		10/4/2002 9:16:52 AM
**	\warning	none
**	\par		Holds all the data for the Sound_FMOD class to use.	It's been given a simple
**				 ctor so that it can set the data for the user and a dtor that frees the fmod 
**				 handle of the sound
**	\par		The struct does have logic to differentiate between streamed and 
**				 non-streamed ('samples' in fmod lingo).
*/
////////////////////////////////////////////////////////////////////////////////
struct LoadedSound_Fmod
{	
	//!Handle of the sound, use to find it in containers.
	HSound			m_hSound;

	//!File name of the file that this references. Left in for debugging
	//purposes. May be removed in the future to save a word.
	char			m_pathName[128];
	
	//!Set to true if the sound is streamed, false if it isn't.
	bool			m_isStreamed;

	//!0 to 255 value of the current volume
	char			m_currVolume;

	//!This union is used to implicitly force the user to state what they
	//! are currently using this as, either a sample or a stream
	union FSOUND_STREAM_SAMPLE
	{
		//!Handle to a fmod stream.
		FSOUND_STREAM	*m_stream;

		//!Handle to a fmod sample.
		FSOUND_SAMPLE	*m_sample;
	} m_handle;
	

public:

	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:	
	/*!	\brief		A Defualt ctor for the LoadedSound_Fmod class (NON-STREAM).
	**	\param		handle		Fmod sample of the sound that this is supposed to hold. 
	**	\param		h			The local handle of this sound. This is the data that the
	**							 users of this data hold.
	**	\param		channels	The max number of channels.  See m_channels
	**	\param		pathName	The filename of this sound.
	**
	**	\exception	Throws an exception if the given fsound_sample is null or if the call 
	**				 to strcpy fails (unlikely).	
	**
	**	\par		DESCRIPTION:
	**				Loads the non-streaming data into the struct and checks the validity 
	**				 of the data. Will throw an exception if any piece of data is invalid.				
	**					
	**	\author		Nate Cleveland
	**	\date		10/3/2002 2:35:35 PM
	**	\warning	none
	*/
	////////////////////////////////////////////////////////////////////////////////
	LoadedSound_Fmod(FSOUND_SAMPLE	*handle, HSound &h, const char *pathName, char vol = 127);


	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:	
	/*!	\brief		Same functionality as other ctor, but for streaming sounds.
	**	\param		
	**	\param		handle		Fmod stream of the sound that this is supposed 
	**							 to hold.
	**	\param		h			The local handle of this sound. This is the 
	**							 data that the users of this data hold.
	**	\param		channels	The max number of channels.  See m_channels
	**	\param		pathName	The filename of this sound.
	**
	**	\exception	Throws an exception if the given fsound_stream is null or if
	**				 the call to strcpy fails.	
	**
	**	\par		DESCRIPTION:
	**				Loads the streaming data into the struct and checks the validity of
	**				 the data. Will throw an exception if any piece of data is invalid.				
	**					
	**	\author		Nate Cleveland
	**	\date		10/3/2002 2:35:35 PM
	**	\warning	none
	*/
	////////////////////////////////////////////////////////////////////////////////
	LoadedSound_Fmod(FSOUND_STREAM	*handle, HSound &h, const char *pathName, char vol = 127);
	

	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:	
	/*!	\brief		Defualt dtor for this class.
	**	\exception	none
	**	\par		DESCRIPTION: 
	**				Calls FSOUND_Sample_Free or _Stream_Close
	**				 on the fsound_sample handle iff the handle isn't null. Though NULL 
	**				 handles should be impossible.	
	**	\author		Nate Cleveland
	**	\date		10/3/2002 4:02:40 PM
	**	\warning	none
	*/
	////////////////////////////////////////////////////////////////////////////////
	~LoadedSound_Fmod();
};
}//end AcqariusSound::