#include "sound.h"
#include "sound_fmod.h"

#include "../debugging/debug.h"
using namespace AcqariusSound;
// memory leak detection
#ifdef _DEBUG
    #include"memleak.h"
    #define new DEBUG_NEW
    #undef THIS_FILE
    static char THIS_FILE[] = __FILE__;
#endif

const unsigned int AcqariusSound::MAX_SOUND_FILE_PATH_NAME = 128;

//Below are a list of the options that can be set in the Sound::LoadSound method.
const int AcqariusSound::SOUND_2D			=	0x1;
const int AcqariusSound::SOUND_3D			=	0x2;
const int AcqariusSound::SOUND_LARGE		=	0x4;
const int AcqariusSound::SOUND_SMALL		=	0x8;
const int AcqariusSound::SOUND_LOOP_NORMAL	=	0x10;
const int AcqariusSound::SOUND_LOOP_OFF		=	0x20;

///////////////////////////Singleton functionality
Sound* Sound::m_instance = NULL;
Sound* Sound::Instance(SoundOutput type)
{
	if(m_instance)
		return m_instance;
	return m_instance = new Sound(type);	
}

void Sound::Release(void)
{
	if(m_instance)
		delete m_instance;	
	m_instance = NULL;
}

//!See Header...
Sound::Sound(SoundOutput type/* = FMOD */)
:m_listenerDirUp(NULL), m_listenerPos(NULL), m_listenerDirForward(NULL)
{
	//switch on the type so we can create the proper sound system.
	switch(type)
	{
	case FMOD:
		{
			//init the fmod system.
			Sound_Fmod *fmod = new Sound_Fmod;
			if(false == fmod->Init())
				throw 0;
			
			//set the system pntr to fmod.
			m_system = fmod;			
			break;
		}//end case
	default:
		throw 0;
	}//end switch
}//end ctor

//!See Header...
Sound::~Sound()
{
	m_system->Release();
	delete m_system;
}//end dtor

//!See Header...
bool Sound::LoadSound(const char *pathName, HSound &h, int loadOps)
{
	if(m_system->FindSound(pathName, h))
		return true;
	
	return m_system->LoadSound(pathName, h, loadOps);
}

//!See Header...
bool Sound::FreeSound(HSound &h)
{
	return m_system->FreeSound(h);
}

//!See header...
void Sound::SetListener(GivePos *pos, GivePos *dirUp, GivePos *dirForward)
{
	//only sets the data if they aren't null
	if(pos)
		this->m_listenerPos = pos;

	if(dirForward)
		this->m_listenerDirForward = dirForward;

	if(dirUp)
		this->m_listenerDirUp = dirUp;


	return;
}//end Sound::SetListener

//!See header...
void Sound::Update(float elapsedTime)
{
	//check for nulls.
	if(!this->m_listenerDirForward || !this->m_listenerDirUp || !this->m_listenerPos)
		return;

	//Make sure all this data is still relevent.
	if(this->m_listenerDirForward->IsRelevent() && this->m_listenerDirUp->IsRelevent()
		&& this->m_listenerPos->IsRelevent())
	{	
		//Create our update data struct
		UpdateData up;

		//fill in the position, and direction vectors
		(*this->m_listenerPos)(up.listenerPos);
		(*this->m_listenerDirForward)(up.listenerForwardDir);
		(*this->m_listenerDirUp)(up.listenerUpDir);

		//last and least, set up the speed
		up.listenerSpeed.x = elapsedTime * (up.listenerPos.x - this->m_oldListenerPos.x);
		up.listenerSpeed.y = elapsedTime * (up.listenerPos.y - this->m_oldListenerPos.y);
		up.listenerSpeed.z = elapsedTime * (up.listenerPos.z - this->m_oldListenerPos.z);


		//call update and pass the sound system the Update data struct we just filled out
		//display any errors recieved.
		if(!m_system->Update(up))
		{
//			cout << "error in update" << endl;
//			(Debugging::Debugger::Instance())->output(Debugging::SND, Debugging::L4, 
//				"minor error in Sound::Update\n");
		}
		
		//Save the current position into the old position.
		this->m_oldListenerPos.x = up.listenerPos.x;
		this->m_oldListenerPos.y = up.listenerPos.y;
		this->m_oldListenerPos.z = up.listenerPos.z;
	}//end if

	return;
}//end Sound::Update

//!See header...
bool Sound::PlaySound(HSound &h, bool StartPaused/* = false*/)
{
	return m_system->PlaySound(h, StartPaused);
}//end PlaySound
	
//!See header...
bool Sound::PlaySound(HSound &h, GivePos &pos, bool StartPaused/* = false*/)
{
	return m_system->PlaySound(h, pos, StartPaused);
}//end PlaySound

//!See documentation in header
bool Sound::SetVolume(HSound &h, int volume)
{
	return m_system->SetVolume(h, volume);
}//end Sound::SetVolume


//!See documentation in header
bool Sound::SetVolume(int volume)
{
	return m_system->SetVolume(volume);
}//end Sound::SetVolume

//!See documentation in header
int Sound::GetVolume(void)
{
	return m_system->GetVolume();
}//end Sound::GetVolume

//!See documentation in header
int Sound::GetVolume(HSound &h)
{
	return m_system->GetVolume(h);
}//end Sound::GetVolume

//!See documentation in header
bool Sound::Pause(HSound &h, bool SetPaused)
{
	return m_system->Pause(h, SetPaused);
}//end Sound::Pause

//!See documentation in header
bool Sound::IsPlaying(HSound &h) throw()
{
	return m_system->IsPlaying(h);
}//end Sound::IsPlaying

//!See documentation in header
void Sound::SetShortVolume(int volume)throw()
{
	m_system->SetShortVolume(volume);
}//end Sound::IsPlaying

bool Sound::EndSound(HSound &h) throw()
{
	return m_system->EndSound(h);
}

void Sound::Deregister(GivePos& pos)
{
	m_system->Deregister(pos);
}