////////////////////////////////////////////////////////////////////////////////
/*!	\file		SoundSytem.h
**	\brief		Provides the declaration of UpdateData struct and SoundSystem class.
**	\author		Nate Cleveland
**	\date		10/14/2002 12:27:16 PM
**	\par		DESCRIPTION: See class header for desricptions of each object.					
*/
////////////////////////////////////////////////////////////////////////////////
#pragma once
#include "threadSafety.h"
#include "HSound.h"


namespace AcqariusSound
{
////////////////////////////////////////////////////////////////////////////////
/*!	\class		UpdateData
**	\brief		
**	\author		Nate Cleveland
**	\date		10/4/2002 12:27:01 PM
**	\warning	none
**	\par		This is a simple struct that is passed to the Sound_Fmod update method.
**				 A simple inline ctor has been provided for ease of filling out the 
**				 struct but it is not forced. Exists within the AcqariusSound namespace.
*/
////////////////////////////////////////////////////////////////////////////////
struct UpdateData
{
	//!The current position of the listener.
	Point3d listenerPos;

	//!The current up vector of the listener.
	Point3d listenerUpDir;

	//!The current forward vector of the listener.
	Point3d listenerForwardDir;

	//!The current speed of the listener.
	Point3d listenerSpeed;


	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:	
	/*!	\brief		ctor for the UpdateData struct. Eases the initialization of the
	**				 data.
	**	\param		
	**	\exception	none
	**	\par		DESCRIPTION: ctor for the UpdateData struct. Eases the initialization 
	**				 of the data.
	**					
	**	\author		Nate Cleveland
	**	\date		10/4/2002 12:27:41 PM
	**	\warning	none
	*/
	////////////////////////////////////////////////////////////////////////////////
	UpdateData(Point3d &Pos, Point3d &UpDir, Point3d &ForwardDir, Point3d &Speed) throw()
		:listenerPos(Pos), listenerUpDir(UpDir), listenerForwardDir(ForwardDir),
		listenerSpeed(Speed)
	{}

	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:	
	/*!	\brief		only dtor for a class. Does nothing special.
	**	\exception	none
	**	\par		DESCRIPTION: Does nothing special.
	**					
	**	\author		Nate Cleveland
	**	\date		10/4/2002 12:28:17 PM
	**	\warning	none
	*/
	////////////////////////////////////////////////////////////////////////////////
	UpdateData() throw(){}
};//end struct UpdateData.

////////////////////////////////////////////////////////////////////////////////
/*!	\class		SoundSystem
**	\brief		
**	\author		Nate Cleveland
**	\date		10/4/2002 1:01:58 PM
**	\warning	none
**	\par		The soundSystem class lays down the outline of what all sound system
**				 used by the sound class must follow. All specific documentation on
**				 method behaviour must be provided by the derived class. Exists within 
**				 the AcqariusSound namespace.
**
**	\par		No init function is required by the soundSystem. Each class may
**				 design one that suites their sound library best. 
*/
////////////////////////////////////////////////////////////////////////////////
class SoundSystem
{

public:
	//!see documentation in sound.h or derived class.
	virtual bool LoadSound(const char *pathName, HSound &h, int loadOps) = 0;

	//!see documentation in sound.h or derived class.
	virtual bool FindSound(const char *pathName, HSound &h) = 0;

	//!see documentation in sound.h or derived class.
	virtual bool FreeSound(const HSound &h) = 0;

	//!see documentation in sound.h or derived class.
	virtual bool Release(void) = 0;

	//!see documentation in sound.h or derived class.
	virtual bool Update(UpdateData &up) = 0;

	//!see documentation in sound.h or derived class.
	virtual bool PlaySound(HSound &h, bool StartPaused = false) = 0;

	//!see documentation in sound.h or derived class.
	virtual bool PlaySound(HSound &h, const GivePos &pos, bool StartPaused = false) = 0;

	//!see documentation in sound.h or derived class.
	virtual bool SetVolume(HSound &h, int volume) = 0;

	//!see documentation in sound.h or derived class.
	virtual bool SetVolume(int volume) = 0;

	//!see documentation in sound.h or derived class.
	virtual bool EndSound(HSound &h) = 0;

	//!see documentation in sound.h or derived class.
	virtual int GetVolume(HSound &h) = 0;

	//!see documentation in sound.h or derived class.
	virtual int GetVolume(void) = 0;
	
	//!see documentation in sound.h or derived class.
	virtual bool Pause(HSound &h, bool SetPaused) = 0;

	//!see documentation in sound.h or derived class.
	virtual bool IsPlaying(HSound &h) = 0;

	virtual void SetShortVolume(int volume) = 0;

	virtual void Deregister(GivePos& pos) = 0;
};//end class SoundSytem
}//end AcqariusSound


