////////////////////////////////////////////////////////////////////////////////
/*!	\file		Sound_FMOD.cpp
**	\brief		Defines all the methods presented in the Sound_FMOD.h.
**	\author		Nate Cleveland
**	\date		10/14/2002 12:41:46 PM
**	\par		DESCRIPTION: see header for documentation.				
*/
////////////////////////////////////////////////////////////////////////////////
#pragma warning(disable:4786)
#include "Sound_Fmod.h"
using namespace AcqariusSound;
// memory leak detection
#ifdef _DEBUG
    #include"memleak.h"
    #define new DEBUG_NEW
    #undef THIS_FILE
    static char THIS_FILE[] = __FILE__;
#endif

//!The m_soundCnt is used to set values for the hSound references.
int Sound_Fmod::m_soundCnt = 0;

//!See header...
bool Sound_Fmod::Init(const FmodOps *ops/* = NULL*/)
{
	////1. Check the init op
	//This will be used if a null is passed. It is auto inited with defualt
	//values in its ctor.
	FmodOps defualt;
	
	if(!ops)
		ops = &defualt;
	
	//if the ops' channel member is set to -1 we need to get the max avaible channels
	int channels = ops->m_channelNumber;
	
	if(-1 == channels)
		channels = FSOUND_GetMaxChannels();
	
	//return false if we have zero channels selected.
	if(!channels)
		return false;
	
	
	////2. Init Fmod
	//set the sound type. if we get a false return false.
	if(FALSE == FSOUND_SetOutput(ops->m_soundOutput))
		return false;
	
	if(FALSE == FSOUND_Init(ops->m_hz,channels,ops->m_flags))
		return false;
	
	
	////3. Now set the other effects of fmod.
	FSOUND_3D_Listener_SetDistanceFactor(1000.00f);
	//FSOUND_3D_Listener_SetRolloffFactor(1000);
	FSOUND_3D_Listener_SetDopplerFactor(20.5f);
	return true;
}//end Sound_Fmod::Init

//!See header...
bool Sound_Fmod::LoadSound(const char *pathName, HSound &h, int loadOps)
{
	////1. Load the sound
	int	fmodLoadOps = FSOUND_16BITS;			//var passed to the fmodLoad
	
	//check for null path name or long a file name.
	if(!pathName || strlen(pathName) > MAX_SOUND_FILE_PATH_NAME - 1)
		return false;
	
	//check to see if this is a 3D sound or not. If it is we'll need to set certain
	//options an vice versa.
	if(loadOps & SOUND_3D)
		fmodLoadOps |= FSOUND_FORCEMONO | FSOUND_HW3D;
	else
		fmodLoadOps |= FSOUND_2D;

	//!Defual loopin to off. ie check for normal and if it isn't there 
	//! set off to the ops.
	if(loadOps & SOUND_LOOP_NORMAL)
		fmodLoadOps |= FSOUND_LOOP_NORMAL;
	else
		fmodLoadOps |= FSOUND_LOOP_OFF;


	
	//Let fmod load the file. if they set in large sound we'll load it as a stream
	//otherwise we'll load as a sample
	if(loadOps & SOUND_LARGE)
	{
		FSOUND_STREAM	*handle=NULL;				
		
		
		if(NULL == (handle = FSOUND_Stream_OpenFile(pathName, fmodLoadOps, 0)))
			return false;
		
		////2. Save the file for later retrival and set the hSound
		h = ++this->m_soundCnt;
		
		//the LoadedSound_Fmod ctor can throw so be sure to catch any errors.
		try
		{
			LoadedSound_Fmod *pHS = new LoadedSound_Fmod(handle, h, pathName);
			this->m_loadedSoundContainer.push_back(pHS);
		} catch(...)
		{
			return false;
		}//end try/catch		
	}//end stream if
	
	//this will be executed if the sound is NOT streamed.
	else
	{
		FSOUND_SAMPLE	*handle=NULL;	
		
		if(NULL == (handle = FSOUND_Sample_Load(FSOUND_FREE, pathName, fmodLoadOps, 0)))
			return false;
		
		////2. Save the file for later retrival and set the hSound
		h = ++this->m_soundCnt;
		
		//the LoadedSound_Fmod ctor can throw so be sure to catch any errors.
		try
		{
			LoadedSound_Fmod *pHS = new LoadedSound_Fmod(handle, h, pathName);
			this->m_loadedSoundContainer.push_back(pHS);
		} catch(...)
		{
			return false;
		}//end try/catch
	}//end sample else



	return true;
}//end Sound_Fmod::LoadSound


//!See documentation in header
bool Sound_Fmod::FindSound(const char *pathName, HSound &h)
{
	//Get an iterator with the first item in the container already loaded.
	vector<LoadedSound_Fmod *>::iterator	hSoundIT = this->m_loadedSoundContainer.begin();


	//keep looping until we either go through the whole list or we find it.
	bool	isFound = false;			//this'll be set true if we find it.

	for( ; hSoundIT != this->m_loadedSoundContainer.end() && !isFound ; hSoundIT++)
	{
		//deref the current loaded sound in the container.
		LoadedSound_Fmod *curr = *hSoundIT;

		//check to see if they match
		if(0 == strcmp(curr->m_pathName, pathName))
		{
			isFound = true;
			h = curr->m_hSound;
		}//end if
	}//end while

	return isFound;
}//Sound_Fmod::FindSound

//!See documentation in header
LoadedSound_Fmod *Sound_Fmod::FindLoadedSound(const HSound &h) 
{
	//Get an iterator with the first item in the container already loaded.
	vector<LoadedSound_Fmod *>::iterator	hSoundIT = this->m_loadedSoundContainer.begin();


	//keep looping until we either go through the whole list or we find it.
	for( ; hSoundIT != this->m_loadedSoundContainer.end() ; hSoundIT++)
	{
		//deref the current loaded sound in the container.
		LoadedSound_Fmod *curr = *hSoundIT;

		//check to see if they match
		if(curr->m_hSound == h)
			return curr;
	}//end for

	return NULL;
}//Sound_Fmod::FindLoadedSound

//!See header for documentation
ActiveSound_Fmod *Sound_Fmod::FindActiveSound(int channel) 
{
	//Get an iterator with the first item in the container already loaded.
	vector<ActiveSound_Fmod *>::iterator	hActSoundIT = this->m_activeSoundContainer.begin();


	//keep looping until we either go through the whole list or we find it.
	for( ; hActSoundIT != this->m_activeSoundContainer.end() ; hActSoundIT++)
	{
		//deref the current loaded sound in the container.
		ActiveSound_Fmod *curr = *hActSoundIT;

		//check to see if they match
		if(curr->m_channel == channel)
			return curr;
	}//end for

	return NULL;
}//Sound_Fmod::FindActiveSound


//!See documentation in header
bool Sound_Fmod::FreeSound(const HSound &h)
{
	//create a var to hold the sound we're about to free
	LoadedSound_Fmod *loadedSound = NULL;		

	//find the sound in the list.
	vector<LoadedSound_Fmod *>::iterator hSoundIT = this->m_loadedSoundContainer.begin();

	//we keep looking for the sound till we find it an free it.
	for( ;hSoundIT != this->m_loadedSoundContainer.end() ; hSoundIT++ )
	{
		//deref the current sound.
		loadedSound = *hSoundIT;

		//check to see if this is the sound we want.
		if(loadedSound->m_hSound == h)
		{
			//check the sounds mutex or whatever is used for thread safety...
			//...

			//Get a iterator to the last element in the container.
			vector<LoadedSound_Fmod *>::iterator containerEnd = this->m_loadedSoundContainer.end();
			--containerEnd;

			//swap it so its at the end of the container
			iter_swap(hSoundIT, containerEnd);

			//pop the last element from the container before deleting it
			this->m_loadedSoundContainer.pop_back();
			delete loadedSound;
			return true;
		}//end if
	}//end for

	//we'll only get here if we couldn't find the sound.
	return false;
}//end Sound_Fmod::FreeSound


//!See documentation in header
bool Sound_Fmod::Release(void)
{
	////1. call stop on all the active sounds.
	FSOUND_StopSound(FSOUND_ALL);
		

	////2. Free all our loaded and active sounds.
	for( ; false == this->m_loadedSoundContainer.empty() ; )
	{
		//get a handle on the next sound.
		LoadedSound_Fmod *loadedSound = this->m_loadedSoundContainer.back();
		
		//remove sound from container
		this->m_loadedSoundContainer.pop_back();
		
		//delete this sound, its dtor will free the sound.
		delete loadedSound;	
	}//end for

	//This loop behaves just as the above one.
	for( ; this->m_activeSoundContainer.size() ; )
	{
		ActiveSound_Fmod *activeSound = this->m_activeSoundContainer.back();
		this->m_activeSoundContainer.pop_back();
		delete activeSound;	
	}//end for
	
	////3. Close down the fmod system.
	FSOUND_Close();
	return true;
}//end Sound_Fmod::ReleaseMutex

//!See documentation in header
bool Sound_Fmod::Update(AcqariusSound::UpdateData &up)
{

	////1. update the listener
	FSOUND_3D_Listener_SetAttributes(&up.listenerPos.x,&up.listenerSpeed.x,
		up.listenerForwardDir.x, up.listenerForwardDir.y, up.listenerForwardDir.z,
		up.listenerUpDir.x, up.listenerUpDir.y, up.listenerUpDir.z);

	////2. update all the other sounds.
	//create an iterator to cycle threw the container of active sounds.
	vector<ActiveSound_Fmod *>::iterator hActSoundIT = this->m_activeSoundContainer.begin();

	//loop threw all the active sounds in the container.
	for( ;hActSoundIT != this->m_activeSoundContainer.end(); hActSoundIT++)
	{
		ActiveSound_Fmod *currActSound = *hActSoundIT;

		//check to see if this is a threeD sound AND has a relevent
		// givePos member. order of operation is very important for this.
		if(currActSound->m_isThreeD && currActSound->m_pos->IsRelevent())
		{
			//get the current position and this sound
			Point3d		pnt(currActSound->m_pos->x(), currActSound->m_pos->y(), currActSound->m_pos->z()); 

			//At this time we are NOT caring about speed and doplar effects. May add them
			// later. TODO TODO. Thus, pass null for the last arg of SetAttributes. 
			if(FALSE == FSOUND_3D_SetAttributes(currActSound->m_channel, &pnt.x, NULL))
				return false;
		}//end if
	}//end for


	//update the overal  3D sound system
	FSOUND_3D_Update();

	return true;
}//end Sound_Fmod::Update



//!This is the member method that the play functions call when 
//! they wish to play a sound.
ActiveSound_Fmod *Sound_Fmod::InternalPlaySound(HSound &h, bool StartPaused)
{
	LoadedSound_Fmod *loadedSound = NULL;		//will hold the sound the user wishes to play
	
	//find the sound in our container of sounds.
	if(!(loadedSound = this->FindLoadedSound(h)))
		return NULL;
	

	//call play on this sound. Eventually we'll throw
	//in some wet sounds onto the playback. TODO TODO
	int channel = -1;
	
	//switch between playing stream sounds and playing sample sounds
	if(loadedSound->m_isStreamed)
	{
		if(-1 == (channel = FSOUND_Stream_PlayEx(FSOUND_FREE, 
			loadedSound->m_handle.m_stream, NULL, StartPaused)))
			return NULL;
	}//end if

	//will be executed
	else
	{
		if(-1 == (channel = FSOUND_PlaySoundEx(FSOUND_FREE, 
			loadedSound->m_handle.m_sample, NULL, StartPaused)))
			return NULL;
	}//end else

	//next see if we already have this channel.
	ActiveSound_Fmod *activeSound = NULL;		//will hold the sound the user wishes to play
	
	//see if we already have used this channel before. This will be true
	//if we haven't.
	if(!(activeSound = this->FindActiveSound(channel)))
	{
		//allocate a new sound
		activeSound = new ActiveSound_Fmod(channel, h);

		//push it into the list
		this->m_activeSoundContainer.push_back(activeSound);
	}//end if

	//otherwise we just need to set the new handle
	else
	{
		activeSound->m_hSound = h;
	}//end else


	//also set the volume of this sound.
	FSOUND_SetVolume(activeSound->m_channel, loadedSound->m_currVolume);

	return activeSound;
}//end Sound_Fmod::InternalPlaySound

//!See header for documentation
bool Sound_Fmod::PlaySound(HSound &h, bool StartPaused/* = false*/)
{
	ActiveSound_Fmod *activeSound = NULL;		//will hold the sound the user wishes to play

	//call the internal function that does the work for us.
	if(NULL == (activeSound = this->InternalPlaySound(h, StartPaused)))
		return false;

	//make sure the ThreeD flag is set to false and return out
	activeSound->m_isThreeD = false;		
	return true;
}//end Sound_Fmod::PlaySound

//!See header for documentation
bool Sound_Fmod::PlaySound(HSound &h, const GivePos &pos, bool StartPaused/* = false*/)
{
	ActiveSound_Fmod *activeSound = NULL;		//will hold the sound the user wishes to play


	//The first step is the same as PlaySound(HSound), so just call it.
	if(NULL == (activeSound = this->InternalPlaySound(h, StartPaused)))
		return false;

	//next step is to simple plug in the GivePos reference and set the ThreeD
	// flag to true
	activeSound->m_isThreeD = true;
	activeSound->m_pos = &pos;

	return true;
}//end Sound_Fmod::PlaySound



//!See header for documentation
bool Sound_Fmod::SetVolume(HSound &h, int volume)
{
	//do a real quick bounds check.
	if(volume > 255)
		volume = 255;
	else if(volume < 0)
		volume = 0;

	//find the related loaded sound and set its volume
	vector<LoadedSound_Fmod *>::iterator	loadSoundIT = this->m_loadedSoundContainer.begin();
	for( ; loadSoundIT != this->m_loadedSoundContainer.end() ; loadSoundIT++)
	{
		LoadedSound_Fmod* curr = *loadSoundIT;

		//set the sound to given volume if it is our sound.
		if(curr->m_hSound == h) 
		{
			curr->m_currVolume = volume;
			break;
		}
	}


	//Now we must do a search and see if this sound is currently playing, if it is reset the vol.
	vector<ActiveSound_Fmod *>::iterator	hActSoundIT = this->m_activeSoundContainer.begin();
	for( ; hActSoundIT != this->m_activeSoundContainer.end() ; hActSoundIT++)
	{
		//deref the current loaded sound in the container.
		ActiveSound_Fmod *curr = *hActSoundIT;

		//check to see if they match
		if(curr->m_hSound == h)
		{
			if(!FSOUND_SetVolume(curr->m_channel, volume))
				return false;
		}//end if
	}//end for

	return true;
}//end Sound_Fmod::SetVolume

//!See header for documentation
int Sound_Fmod::GetVolume(HSound &h)
{
	vector<ActiveSound_Fmod *>::iterator hActSoundIT = this->m_activeSoundContainer.begin();

	//keep looping until we either go through the whole list or we find it.
	for( ; hActSoundIT != this->m_activeSoundContainer.end() ; hActSoundIT++)
	{
		//deref the current loaded sound in the container.
		ActiveSound_Fmod *curr = *hActSoundIT;

		//check to see if they match
		if(curr->m_hSound == h)
		{
			return FSOUND_GetVolume(curr->m_channel);
		}//end if
	}//end for

	return 0;
}//end Sound_Fmod::GetVolume

//!See header for documentation
bool Sound_Fmod::SetVolume(int volume)
{
	//do a real quick bounds check.
	if(volume > 255)
		volume = 255;
	else if(volume < 0)
		volume = 0;

	FSOUND_SetSFXMasterVolume(volume);

	return true;
}//end Sound_Fmod::SetVolume

//!See header for documentation
int Sound_Fmod::GetVolume(void)
{
	return FSOUND_GetSFXMasterVolume();
}//end Sound_Fmod::GetVolume

//!See header for documentation
bool Sound_Fmod::EndSound(HSound &h)
{
	//Get an iterator with the first item in the container already loaded.
	vector<ActiveSound_Fmod *>::iterator hActSoundIT = this->m_activeSoundContainer.begin();


	//keep looping until we either go through the whole list or we find it.
	for( ; hActSoundIT != this->m_activeSoundContainer.end() ; hActSoundIT++)
	{
		ActiveSound_Fmod *curr = *hActSoundIT;

		//check to see if they match
		if(curr->m_hSound == h)
		{
			//stop the sound from playing
			if(!FSOUND_StopSound(curr->m_channel))
				return false;
		}//end if
	}//end for


	vector<LoadedSound_Fmod *>::iterator hLoadIT = m_loadedSoundContainer.begin();
	for( ; hLoadIT != m_loadedSoundContainer.end() ; hLoadIT++)
	{
		LoadedSound_Fmod* curr = *hLoadIT;

		if(curr->m_hSound == h)
		{
			if(curr->m_isStreamed)
			{		
				if(!FSOUND_Stream_SetPosition(curr->m_handle.m_stream, 0))
					return false;			
			}//end if
			break;
		}//end if
	}

	return true;
}//end Sound_Fmod::EndSound

//See documentation in header
bool Sound_Fmod::Pause(HSound &h, bool SetPaused)
{
	//Get an iterator with the first item in the container already loaded.
	vector<ActiveSound_Fmod *>::iterator hActSoundIT = this->m_activeSoundContainer.begin();


	//keep looping until we either go through the whole list or we find it.
	for( ; hActSoundIT != this->m_activeSoundContainer.end() ; hActSoundIT++)
	{
		ActiveSound_Fmod *curr = *hActSoundIT;

		//check to see if they match
		if(curr->m_hSound == h)
		{
			//I get funky warnings when casting from bool to char sooo, we'll hack
//			return FSOUND_SetPaused(curr->m_channel, (char)SetPaused);


			signed char check = FALSE;

			if(SetPaused)
				check = FSOUND_SetPaused(curr->m_channel, 1);

			else
				check = FSOUND_SetPaused(curr->m_channel, 0);

			if(check)
				return true;
			return false;
		}//end if
	}//end for
		
	return false;
}//end Sound_Fmod::Pause

//See documentation in header
bool Sound_Fmod::IsPlaying(HSound &h) throw()
{
	LoadedSound_Fmod* pLoadedSound = this->FindLoadedSound(h);

	//check for null.
	if(pLoadedSound) 
	{
		//search through container of active sounds and try to find this sound.
		vector<ActiveSound_Fmod *>::iterator actSoundIT = this->m_activeSoundContainer.begin();

		for( ; actSoundIT != this->m_activeSoundContainer.end() ; actSoundIT++)
		{
			ActiveSound_Fmod* currActSound = *actSoundIT;

			//will be true if this is the sound we want.
			if(currActSound->m_hSound == h)
			{
				//check to see if this sound is still playing. Func will return false
				//if the sound is still playing.
				return !currActSound->IsDone();
			}//end if
		}//end for
	}//end if
	return false;
}//end Sound_Fmod::IsPlaying


//See documentation in header
void Sound_Fmod::SetShortVolume(int volume) throw()
{
	//do a real quick bounds check.
	if(volume > 255)
		volume = 255;
	else if(volume < 0)
		volume = 0;

	//Get an iterator with the first item in the container already loaded.
	vector<ActiveSound_Fmod *>::iterator	hActSoundIT = this->m_activeSoundContainer.begin();


	//keep looping until we either go through the whole list or we find it.
	for( ; hActSoundIT != this->m_activeSoundContainer.end() ; hActSoundIT++)
	{
		//deref the current loaded sound in the container.
		ActiveSound_Fmod *curr		= *hActSoundIT;
		LoadedSound_Fmod *lSound	= NULL;

		//if this channel is still active, find the sound it refers to.
		if(curr->m_hSound && (lSound = FindLoadedSound(curr->m_hSound)) && !lSound->m_isStreamed)
			FSOUND_SetVolume(curr->m_channel, volume);

	}//end for
}


void Sound_Fmod::Deregister(GivePos& pos)
{
	//look for any instances of this give pos and remove them
	//Get an iterator with the first item in the container already loaded.
	vector<ActiveSound_Fmod *>::iterator	hActSoundIT = this->m_activeSoundContainer.begin();


	//keep looping until we either go through the whole list or we find it.
	for( ; hActSoundIT != this->m_activeSoundContainer.end() ; )
	{
		ActiveSound_Fmod* currAct	= *hActSoundIT;


		//we have a match, 
		if(currAct->m_pos == &pos)
		{
			//stop this sound if it
			if(FSOUND_IsPlaying(currAct->m_channel)) 
				FSOUND_StopSound(currAct->m_channel);

			hActSoundIT = m_activeSoundContainer.erase(hActSoundIT);
			delete currAct;
		}

		else
			hActSoundIT++;
	}//end for 
	
}