////////////////////////////////////////////////////////////////////////////////
/*!	\file		Sound_FMOD.h
**	\brief		Declares the Sound_FMOD class.
**	\author		Nate Cleveland
**	\date		10/14/2002 12:35:32 PM
**	\par		DESCRIPTION: See class header for description.
*/
////////////////////////////////////////////////////////////////////////////////
#pragma once
#include "Sound.h"
#include "fmod.h"
#include "fmod_errors.h"
#include "wincompat.h"
#include "FmodOps.h"
#include "LoadedSound_Fmod.h"
#include "ActiveSound_Fmod.h"
struct UpdateData;

namespace AcqariusSound
{

////////////////////////////////////////////////////////////////////////////////
/*!	\class		Sound_Fmod
**	\brief		Wrapper around the FMOD sound library.
**	\author		Nate Cleveland
**	\date		10/4/2002 11:52:34 AM
**	\warning	none
**	\par		DESCRIPTION: Sole interface into the FMOD sound library.
*/
////////////////////////////////////////////////////////////////////////////////
class Sound_Fmod: public SoundSystem
{
protected:

	//!This count is incremented every time we load a sound and is used
	//! and used as the HSound.
	static int	m_soundCnt;

	//!Our stl container that holds all our sounds.
	vector<LoadedSound_Fmod *> m_loadedSoundContainer;

	//!this stl container holds all the active sounds in the game
	vector<ActiveSound_Fmod *> m_activeSoundContainer;

	//internal methods used to find sounds via their HSound or channel respectively.
	LoadedSound_Fmod *FindLoadedSound(const HSound &h);
	ActiveSound_Fmod *FindActiveSound(int channel);

	//!This is the member method that the play functions call when
	//! they wish to play a sound.
	ActiveSound_Fmod *InternalPlaySound(HSound &h, bool StartPaused);


public:


	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:
	/*!	\brief		Init method for Sound_Fmod. Must be called for sounds to work.
	**
	**	\param		ops		Pntr to the ops struct. Defaults to NULL. If null
	**						 is passed then the function will use defualt
	**						 values to init the method.
	**	\return		Boolean true/false. true good, false fatal. Will return false if
	**				 fmod init functions fail or if channels are set to zero.
	**	\exception	Reserves the right to throw exception.
	**	\par		DESCRIPTION: Must be called for sounds to work.
	**
	**	\author		Nate Cleveland
	**	\date		10/4/2002 11:54:53 AM
	**	\warning	none
	*/
	////////////////////////////////////////////////////////////////////////////////
	bool Init(const FmodOps *ops = NULL);




	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:
	/*!	\brief		Closes down the entire sound system and frees all the memory used
	**				 by the class.
	**
	**	\param		NONE
	**	\return		Boolean true/false. true good, false really bad.
	**	\exception	none
	**
	**	\par		DESCRIPTION: Closes down the entire sound system and frees all the
	**				 memory used by the class.
	**
	**	\author		Nate Cleveland
	**	\date		10/4/2002 11:57:32 AM
	**	\warning	none
	*/
	////////////////////////////////////////////////////////////////////////////////
	bool Release(void) throw();


	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:
	/*!	\brief		Preloads a sound for the caller VIA FMOD.
	**	\param		pathName		File name for the sound to be loaded. can NOT
	**								 be larger then MAX_SOUND_FILE_PATH_NAME - 1.
	**	\param		h				Will be loaded with a handle to the sound.
	**	\param		loadOps			See Sound::LoadSound method in sound.h for
	**								 details on how to fill this param out.
	**
	**	\return		Returns false if something goes wrong. Will NOT return false if the
	**				 sound is already loaded. Common failure would be that the max number
	**				 of sounds are already loaded. Another common fail is if the pathName
	**				 is longer then the #def'd value of MAX_SOUND_FILE_PATH_NAME - 1.
	**				 Returns true if all goes well.
	**
	**	\exception	none
	**
	**	\par		DESCRIPTION:Preloads a sound for the caller VIA FMOD. If the sound is
	**				 already loaded this call will NOT fail. <b>It will instead load another
	**				 copy of the sound.</b>
	**
	**	\author		Nate Cleveland
	**	\date		10/4/2002 11:59:30 AM
	**	\warning	none
	*/
	////////////////////////////////////////////////////////////////////////////////
	bool LoadSound(const char *pathName, HSound &h, int loadOps) throw();


	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:
	/*!	\brief		Attempts to find a copy of an already loaded sound.
	**
	**	\param		pathName		File name for the sound to be located.
	**	\param		h					Will be loaded with a handle to the sound.
	**
	**	\return		Returns false if it couldn't find the sound. Returns true if it did
	**				 find the sound.
	**	\exception	none
	**	\par		DESCRIPTION: Does NOT search to see if a sound is playing, only if
	**				 is loaded.
	**
	**	\author		Nate Cleveland
	**	\date		10/4/2002 12:05:16 PM
	**	\warning	none
	*/
	////////////////////////////////////////////////////////////////////////////////
	bool FindSound(const char *pathName, HSound &h) throw();


	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:
	/*!	\brief		Frees a preloaded sound.
	**
	**	\param		h		Handle of the sound to be freed.
	**
	**	\return		Returns false if it couldn't find the sound or if it's mutex
	**				 is in use by somebody. Returns true if it did nothing goes wrong.
	**	\exception	none
	**
	**	\par		DESCRIPTION: Frees the indicated sound VIA fmod.
	**
	**	\author		Nate Cleveland
	**	\date		10/4/2002 12:06:34 PM
	**	\warning	none
	*/
	////////////////////////////////////////////////////////////////////////////////
	bool FreeSound(const HSound &h) throw();


	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:
	/*!	\brief		Updates the sound module.
	**
	**	\param		up	An update data struct that contains all the relevent
	**					 fmod data. See UpdateData struct documentation.
	**
	**	\return		Returns false if it couldn't find the sound or if it's mutex
	**				 is in use by somebody. Returns true if it did nothing goes wrong.
	**	\exception	none
	**	\par		DESCRIPTION: Must be called once a frame so that the positions of all
	**				 the 3D sounds in the system.
	**
	**	\author		Nate Cleveland
	**	\date		10/4/2002 12:07:46 PM
	**	\warning	none
	*/
	////////////////////////////////////////////////////////////////////////////////
	bool Update(UpdateData &up) throw();


	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:
	/*!	\brief		Plays a given sound.
	**	\param		h					Handle of the sound to be played.
	**	\param		StartPaused			Users can set this to true if they want
	**									 the sound to start up paused. Defaulted to
	**									 false.
	**	\return		Returns false if the handle is invalid or something else goes wrong.
	**				 Common failure would be that the max number of sounds are already
	**				 playing.
	**	\exception	none
	**	\par		DESCRIPTION: Plays the indicated sound, the last member is a
	**				 toggle for starting paused or not.
	**
	**	\author		Nate Cleveland
	**	\date		10/4/2002 12:10:05 PM
	**	\warning	none
	*/
	////////////////////////////////////////////////////////////////////////////////
	bool PlaySound(HSound &h, bool StartPaused = false) throw();


	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:
	/*!	\brief		Plays a given 3D sound.
	**	\\param		h				Handle of the sound to be played.
	**	\param		pos				Referce to the function object that provides
	**								 the sound object with the position of the
	**								 sound.
	**	\param		StartPaused		Users can set this to true if they want
	**								 the sound to start up paused. Defaulted to
	**								 false.
	**	\return		Returns false if the handle is invalid or something else goes wrong.
	**				 Common failure would be that the max number of sounds are already
	**				 playing
	**	\exception	none
	**	\par		DESCRIPTION: 3D effects will not work if the sound wasn't loaded with
	**				 the 3D options set. Besides 3D treament functions just as
	**				Sound_Fmod::PlaySound(HSound &, bool);
	**
	**	\author		Nate Cleveland
	**	\date		10/4/2002 12:11:50 PM
	**	\warning	none
	*/
	////////////////////////////////////////////////////////////////////////////////
	bool PlaySound(HSound &h, const GivePos &pos, bool StartPaused = false) throw();

	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:
	/*!	\brief		Alters the volume of the given sound relative to the global sound.
	**
	**	\param		&h					Handle of the sound to be altered.
	**	\param		volume				A 0(silent) to 255(full on) value for the
	**									 new volume.
	**
	**	\return		Boolean true/false. True good, false bad. Common failure would
	**				 be that the hSound's mutex is used by somebody else.
	**	\exception	none
	**
	**	\par		DESCRIPTION: It will alter the volume of ALL active sounds with the
	**				 given handle. The new volume will be relative to the global sound.
	**				 This function will only effect a sound that is actively playing. So,
	**				 every time you play a sound you will have to recall SetVolume for
	**				 that sound. This is a feature of FMOD.
	**
	**	\author		Nate Cleveland
	**	\date		10/4/2002 12:15:09 PM
	**	\warning	none
	*/
	////////////////////////////////////////////////////////////////////////////////
	bool SetVolume(HSound &h, int volume) throw();

	

	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:	
	/*!	\brief		Sets the volume of all non-streamed sounds.
	**	\param		volume		Volume to set all non-streamed sounds.
	**	\return		none
	**	\exception	none			
	**	\author		Nate Cleveland
	**	\date		2/27/2003 9:43:18 AM
	**	\warning	none
	*/
	////////////////////////////////////////////////////////////////////////////////
	void SetShortVolume(int volume) throw();


	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:
	/*!	\brief		Returns the volume of the given sound.
	**
	**	\param		&h					Handle of the sound.
	**
	**	\return		Returns the current volume of the given sound. Range will be
	**				 0(silent) to 255(full on). If something goes wrong will follow
	**				 fmod's example and returns 0 to the caller.
	**
	**	\exception	none
	**	\par		DESCRIPTION: Returns the volume of the given sound. If there are
	**				 multiple instances of this sound will return the value of the first
	**				 that is found. This is not an issue since SetVolume(HSound &, int)
	**				 forces all sounds of that handle type to be of the same volume.
	**
	**	\author		Nate Cleveland
	**	\date		10/4/2002 12:17:05 PM
	**	\warning	none
	*/
	////////////////////////////////////////////////////////////////////////////////
	int GetVolume(HSound &h) throw();


	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:
	/*!	\brief		Alters the global volume.
	**	\param		volume			A 0(silent) to 255(full on) value for the
	**								 new volume.
	**	\return		Boolean true/false. True good, false bad and non-fatal.
	**	\exception	none
	**	\par		DESCRIPTION: Alters the effect of all sound in the game. All other
	**				 sounds are set relative to this one.
	**
	**	\author		Nate Cleveland
	**	\date		10/4/2002 12:21:22 PM
	**	\warning	none
	*/
	////////////////////////////////////////////////////////////////////////////////
	bool SetVolume(int volume) throw();


	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:
	/*!	\brief		Returns the global volume.
	**	\param		NONE
	**	\return		Returns the current global volume. Range will be 0(silent) to
	**				 255(full on). If something goes wrong will follow fmod's example
	**				 and returns 0 to the caller.
	**	\exception	none
	**	\par		DESCRIPTION:
	**
	**	\author		Nate Cleveland
	**	\date		10/4/2002 12:22:57 PM
	**	\warning	none
	*/
	////////////////////////////////////////////////////////////////////////////////
	int GetVolume(void) throw();


	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:
	/*!	\brief		Stops all instances of a sound from being played.
	**
	**	\param		h					Handle of the sound to be stoped.
	**
	**	\return		Returns false if the handle is invalid or something else goes wrong.
	**	\exception	none
	**	\par		DESCRIPTION: Stops all instances of a sound from playing.
	**
	**	\author		Nate Cleveland
	**	\date		10/4/2002 12:23:32 PM
	**	\warning	none
	*/
	////////////////////////////////////////////////////////////////////////////////
	bool EndSound(HSound &h) throw();


	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:
	/*!	\brief		Pauses or unpauses a sound
	**
	**	\param		h					Handle of the sound to be paused or unpaused.
	**	\param		SetPaused			Set to true to pause the sound, false to
	**									 unpause it.
	**
	**	\return		Returns false if the handle is invalid or something else goes wrong.
	**	\exception	none
	**	\par		DESCRIPTION: Sets an internal paused flag to either true or false.
	**
	**	\author		Nate Cleveland
	**	\date		10/4/2002 12:25:08 PM
	**	\warning	none
	*/
	////////////////////////////////////////////////////////////////////////////////
	bool Pause(HSound &h, bool SetPaused) throw();


	bool IsPlaying(HSound &h) throw();
	void Deregister(GivePos& pos);



};//end Sound_Fmod class

}//end AcqariusSound