////////////////////////////////////////////////////////////////////////////////
/*!	\file		UIClient.cpp
**	\brief		Defines all the methods presented in UIClient.h
**	\author		Nate Cleveland
**	\date		10/14/2002 12:25:34 PM
*/
////////////////////////////////////////////////////////////////////////////////
#include "UIClient.h"
#include <string.h>
#include "../nexception/nexception.h"
#include <fstream>
// memory leak detection
#ifdef _DEBUG
    #include"memleak.h"
    #define new DEBUG_NEW
    #undef THIS_FILE
    static char THIS_FILE[] = __FILE__;
#endif

using namespace AcqariusSound;

#define UI_SOUND_CLIENT_LOG_CONSOLE	//!< Comment this out to prevent UI from logging to console.
#define UI_SOUND_CLIENT_LOG_FILE	//!< Comment this out to prevent UI from logging to file.

UIClient* UIClient::m_instance = NULL;


namespace 
{
	const char*		DEFUALT_SONG_FILE = "Resources\\Sounds\\defualt.txt";
}





#ifdef NAMESPACE_ACQARIUSSOUND_HAS_SOUND_INSTANCE
//See header for documentation
UIClient::UIClient()
:m_sound(s), m_musicCount(0), m_currMusicHandleIndex(0), m_beep(0), m_buzz(0), 
m_currMusicVolume(255), m_currEffectsVolume(255)
{
{
	throw 0;
}
#endif

//See header for documentation
UIClient::UIClient(Sound& s) throw()
:m_sound(s), m_musicCount(0), m_currMusicHandleIndex(0), m_beep(0), m_buzz(0),
m_currMusicVolume(255), m_currEffectsVolume(255), m_state(UISTATE_TITLE_SCREEN)
{
	s.SetListener(&m_pos, &m_upVec, &m_forwardVec);


	//load all the sound effects.
	ifstream inFile(DEFUALT_SONG_FILE);

	//read in both lines
	char buff[2][256];
	inFile.getline(buff[0], 256, '\n');
	inFile.getline(buff[1], 256, '\n');

	std::vector<const char*> vec;
	vec.push_back(buff[0]);
	vec.push_back(buff[1]);

	LoadDynamicMusic(vec);
}

//!Cleans up any dynamic data
UIClient::~UIClient() throw()
{
	//clean dynamic memory.
	//let the sound client clean our memory. (it is faster this way, O(n) vs O(n^2)
}

//See header for documentation.
void UIClient::FreeSound(HSound &h)
{
	//make sure non-zero.
//	if(h) 
	{
		if(!this->m_sound.FreeSound(h))
		{
#ifdef UI_SOUND_CLIENT_LOG_CONSOLE
			//log to the console...
#endif

#ifdef UI_SOUND_CLIENT_LOG_FILE
			//log to file.
#endif	
		}//end if		
	}//end if

	//set to zero to prevent future errors.
	h = 0;
}//end UIClient::FreeSound

//See header for documentation
UIClient* UIClient::Instance() throw()
{
	if(m_instance)
		return m_instance;
	return m_instance = new UIClient(*Sound::Instance());
}

//See header for documentation
bool UIClient::LoadSoundEffects(const char* beepPathName, const char* buzzPathName)throw()
{
	HSound newBuzz = 0;
	HSound newBeep = 0;

	//check for nulls, load both the sounds, check their return values.
	if(beepPathName && !this->m_sound.LoadSound(beepPathName, newBeep, SOUND_2D | SOUND_SMALL | SOUND_LOOP_OFF))
		return false;

	if(beepPathName && !this->m_sound.LoadSound(buzzPathName, newBuzz, SOUND_2D | SOUND_SMALL | SOUND_LOOP_OFF))
		return false;

	//free both the old sounds.
	this->FreeSound(this->m_beep);
	this->FreeSound(this->m_buzz);

	//connect the new sound handles.
	this->m_beep = newBeep;
	this->m_buzz = newBuzz;
	return true;
}//end LoadSoundEffects

//See header for documentation
bool UIClient::LoadMusic(const char* backGroundMusic, UIState state)throw()
{
	char appendedBuffer[256] = {0};

	//check for null
	if(!backGroundMusic) 
		return false;


	strcpy(appendedBuffer, "Resources/Sounds/bg/");
	strcat(appendedBuffer, backGroundMusic);



	//create a handle so we can reference the new sound.
	HSound hSoundNew = 0;

	//load the sound
	if(!this->m_sound.LoadSound(appendedBuffer, hSoundNew, SOUND_2D | SOUND_LARGE | SOUND_LOOP_NORMAL))
		return false;


	HSound& hSoundOld = m_musicHandles[state];
	FreeSound(hSoundOld);
	m_musicHandles[state] = hSoundNew;
	
	//play the new sound at the required volume, check for errors.
	if(!this->m_sound.PlaySound(hSoundNew, false))
		return false;

	int newVol = (state == m_state) ? (GetMusicVolume()) : (0);
	this->m_sound.SetVolume(hSoundNew, newVol);
	return true;
}//end UIClient::LoadMusic


//See header for documentation
bool UIClient::LoadDynamicMusic(const std::vector<const char *>& stringVec)throw()
{
	std::vector<HSound> tempHandleVec;						//Temp Var to store loaded handles
	std::vector<const char *>::const_iterator stringVecIT;	//iterator for stringVec
	int newMusicHandleCount = 0;					//count of tracks we've loaded.
	

	////1. Load all the music tracks from the passed in vector, store them in the temp vec.
	for(stringVecIT = stringVec.begin() ; stringVecIT != stringVec.end() ; stringVecIT++)
	{
		HSound newSound = 0;
		const char* musicPathName = *stringVecIT;

		//load the music track, check for error.
		if(!this->m_sound.LoadSound(musicPathName, newSound, SOUND_2D | SOUND_LARGE | SOUND_LOOP_NORMAL))
		{
			//log the error.
			//....
			return false;
		}//end if

		//store the new handle and increment the new handle count.
		tempHandleVec.push_back(newSound);
		newMusicHandleCount++;
	}//end for

	////2. free the old handles.
	while(!this->m_musicHandles.empty()) 
	{
		//Remove the current handle from the vector, retain a copy.
		HSound hSoundOld = this->m_musicHandles.back();
		this->m_musicHandles.pop_back();

		//pass handle the freeing method.
		this->FreeSound(hSoundOld);		
	}


	////3. Connect the new handles and set the count and index vars..
	this->m_musicHandles = tempHandleVec;
	this->m_musicCount = newMusicHandleCount;
	this->m_currMusicHandleIndex = 0;


	////4. Call play and mute all the sounds. Must be done in such an order to, hopefully,
	////keep them in synch.
	std::vector<HSound>::iterator handleIT;	

	//loop threw and play all the sounds. Have them start paused.
	for(handleIT = this->m_musicHandles.begin() ; handleIT != this->m_musicHandles.end() ; handleIT++)
	{
		HSound currSound = *handleIT;
		this->m_sound.PlaySound(currSound, true);
	}

	//loop thru and mute all the sounds (except the 0th sound, set it to the nominal level)
	for(handleIT = this->m_musicHandles.begin() ; handleIT != this->m_musicHandles.end() ; handleIT++)
	{
		HSound currSound = *handleIT;

		//if this is the first, set it to the nominal volume
		if(handleIT == this->m_musicHandles.begin()) 
			this->m_sound.SetVolume(currSound, this->m_currMusicVolume);

		//otherwise mute the sound.
		else
			this->m_sound.SetVolume(currSound, 0);
	}

	//lastly, un-pause all the sounds.
	for(handleIT = this->m_musicHandles.begin() ; handleIT != this->m_musicHandles.end() ; handleIT++)
	{
		HSound currSound = *handleIT;
		this->m_sound.Pause(currSound, false);
	}

	return true;
}//end UIClient::LoadDynamicMusic

//See header for documentation
bool UIClient::Beep(void)throw()
{
	//check to see if the sound is already playing. If it is, just return true
	if(this->m_sound.IsPlaying(this->m_beep))
		return true;

	//play the sound.
	if(!this->m_sound.PlaySound(this->m_beep))
		return false;
	
	//set the volume and return true.
	this->m_sound.SetVolume(this->m_beep, this->GetEffectsVolume()); 
	return true;
}

//See header for documentation
bool UIClient::Buzz(void) throw()
{
	//check to see if the sound is already playing. If it is, just return true
	if(this->m_sound.IsPlaying(this->m_buzz))
		return true;

	//play the sound.
	if(!this->m_sound.PlaySound(this->m_buzz))
		return false;

	//set the volume and return true.
	this->m_sound.SetVolume(this->m_buzz, this->GetEffectsVolume());
	return true;
}

//See header for documentation
void UIClient::FadeUp()throw()
{
	//before incrementing the level do a quick bounds check.
	if(this->m_currMusicHandleIndex + 1 < this->GetMusicCount())
	{
		HSound oldSound = this->m_musicHandles[this->m_currMusicHandleIndex];
		HSound newSound = this->m_musicHandles[this->m_currMusicHandleIndex+1];

		//mute the current sound 
		if(this->m_sound.SetVolume(oldSound, 0))
		{
			//un-mute the new sound
			if(this->m_sound.SetVolume(newSound, this->m_currMusicVolume))
			{
				//no failures up to this point, set this as the new sound
				this->m_currMusicHandleIndex++;
				return;
			}//end if
		}//end if

		//only get here if there is an error, log the error
#ifdef UI_SOUND_CLIENT_LOG_CONSOLE
			//log to the console...
#endif

#ifdef UI_SOUND_CLIENT_LOG_FILE
			//log to file.
#endif		
	}//end if
	return;
}//end UIClient::FadeUp

//See header for documentation
void UIClient::FadeDown()throw()
{
	//before incrementing the level do a quick bounds check.
	if(this->m_currMusicHandleIndex - 1 >= 0)
	{
		HSound oldSound = this->m_musicHandles[this->m_currMusicHandleIndex];
		HSound newSound = this->m_musicHandles[this->m_currMusicHandleIndex-1];

		//mute the current sound 
		if(this->m_sound.SetVolume(oldSound, 0))
		{
			//un-mute the new sound
			if(this->m_sound.SetVolume(newSound, this->m_currMusicVolume))
			{
				//no failures up to this point, set this as the new sound
				this->m_currMusicHandleIndex--;
				return;
			}//end if
		}//end if

		//only get here if there is an error, log the error
#ifdef UI_SOUND_CLIENT_LOG_CONSOLE
			//log to the console...
#endif

#ifdef UI_SOUND_CLIENT_LOG_FILE
			//log to file.
#endif		
	}//end if
	return;
}

//see documentation in header
void UIClient::SetMusicVolume(int val)
{
	//check for invalid param
	if(val < 0 || val > 255 || GetMusicCount() <= 0)
		return ;

	//set this as the new volume
	this->m_currMusicVolume = val;

	//update the volume of the current actively playing sound.
	this->m_sound.SetVolume(this->m_musicHandles[this->GetCurrMusicLevel()],
							this->m_currMusicVolume);
}

//see documentation in header
void UIClient::SetEffectsVolume(int val)
{
	//check for invalid param
	if(val < 0 || val > 255)
		return ;

	//set this as the new volume
	this->m_currEffectsVolume = val;

	m_sound.SetShortVolume(val);

	return;
}


//see documentation in header
void UIClient::UpdateListener(float* pos, float* upVec, float* forwardVec) throw()
{
	//update position iff a null wasn't given.
	if(pos)
		m_pos(pos);

	//update direction vectors iff a null wasn't given.
	if(upVec)
		m_upVec(upVec);

	if(forwardVec)
		m_forwardVec(forwardVec);
}


void UIClient::SetState(UIState state)throw()
{
	if(m_state != state) 
	{
		if(m_state == UISTATE_TITLE_SCREEN)
		{
			m_sound.EndSound(m_musicHandles[UISTATE_TITLE_SCREEN]);
			m_sound.PlaySound(m_musicHandles[m_state = UISTATE_IN_GAME]);
			m_sound.SetVolume(m_musicHandles[m_state], m_currMusicVolume);
		}

		else
		{
			m_sound.EndSound(m_musicHandles[UISTATE_IN_GAME]);
			m_sound.PlaySound(m_musicHandles[m_state = UISTATE_TITLE_SCREEN]);
			m_sound.SetVolume(m_musicHandles[m_state], m_currMusicVolume);			
		}
	}


}
UIState UIClient::GetState(void)   const throw(){return m_state; }