////////////////////////////////////////////////////////////////////////////////
/*!	\file		UIClient.h
**	\brief		Declare the class UIClient.
**	\author		Nate Cleveland
**	\date		10/14/2002 10:52:37 AM
**	\par		DESCRIPTION: The UIClient class is the interface the UI module has
**				 with the Sound class. UIClient provides functionality for the
**				 following:
**				 - Playing standard background music (BGM). (LoadMusic)
**				 - Dynamic or Interactive BGM (LoadDynamicMusic, FadeUp, FadeDown,
**				 GetMusicCount and GetCurrMusicLevel).
**				 - Playing Beeps and Buzzs once and only once (LoadSoundEffects, Beep,
**				 Buzz)]
**				 - Volume control for BGM (Set/GetMusicVolume)
**				 - Volume control for all sounds (Set/GetGlobalVolume)
**
**	\par		The UIClient exists within the AcqariusSound namespace.
**
**	\par		REMARKS: The GDD states as a higher goal that the BGM will dynamically
**				 toggle between different BGMs depending on the current danger level in
**				 the area of the player. At this time the UIClient does not contain logic
**				 for detecting what the current Interactive BGM should be. The UIClient does
**				 provide methods that will help any module that wishes to undertake this task
**				 (see FadeUP, FadeDown, GetMusicCount and GetCurrMusicLevel).
**
*/
////////////////////////////////////////////////////////////////////////////////
#pragma once
#include "sound.h"

namespace AcqariusSound
{
	enum UIState
	{
		UISTATE_TITLE_SCREEN=0,
		UISTATE_IN_GAME=1,
	};


	////////////////////////////////////////////////////////////////////////////////
	/*!	\class		UIClient
	**	\brief		Provides an interface into the Sound class for the UI Module.
	**	\author		Nate Cleveland
	**	\date		10/14/2002 11:18:03 AM
	**	\warning	none
	*/
	////////////////////////////////////////////////////////////////////////////////
	class UIClient
	{

	public:
		//!Automatically tries to find its own reference to the sound client.
#ifdef NAMESPACE_ACQARIUSSOUND_HAS_SOUND_INSTANCE
		////////////////////////////////////////////////////////////////////////////////
		//	FUNCTION:
		/*!	\brief		Defualt ctor. Exisits iff the AcqariusSound namespace defines a
		**				 non-dynamic sound class.
		**	\param		none
		**	\exception	Reserves the right to throw exceptions.
		**	\par		DESCRIPTION: Exisits iff the AcqariusSound namespace defines a
		**				 non-dynamic sound class.
		**
		**	\author		Nate Cleveland
		**	\date		10/14/2002 11:21:07 AM
		**	\warning	none
		*/
		////////////////////////////////////////////////////////////////////////////////
		UIClient();
#endif

		////////////////////////////////////////////////////////////////////////////////
		//	FUNCTION:
		/*!	\brief		Instantiates the UIClient class.
		**
		**	\param Sound&	Reference to a sound class that the UIClient will use
		**						 for all of its lifetime.
		**	\exception	none
		**	\par		DESCRIPTION: Sets all of the UIClient members to defualt values.
		**				 Music will be defualted to 255. Defualt Global volume is chosen
		**				 by the Sound Class.
		**
		**	\author		Nate Cleveland
		**	\date		10/14/2002 11:22:45 AM
		**	\warning	none
		*/
		////////////////////////////////////////////////////////////////////////////////
		UIClient(Sound& ) throw();

		//!Cleans up any dynamic data
		////////////////////////////////////////////////////////////////////////////////
		//	FUNCTION:
		/*!	\brief		Defualt dtor
		**	\exception	none
		**	\par		DESCRIPTION: Cleans up any dynamic memory.
		**
		**	\author		Nate Cleveland
		**	\date		10/14/2002 11:26:49 AM
		**	\warning	none
		*/
		////////////////////////////////////////////////////////////////////////////////
		~UIClient() throw();
		
		//singleton functionality.
		static UIClient* Instance() throw();

		////////////////////////////////////////////////////////////////////////////////
		//	FUNCTION:
		/*!	\brief		Loads the "Beep" and "Buzz" sound effects.
		**	\param		pathNames		Path names that the sound module will
		**											 use to load the sounds.
		**	\return		none
		**	\exception	none
		**	\par		DESCRIPTION: Loads the first param and maps it to the "Beep" method,
		**				 loads the second param and maps it to the "Buzz" method. Naming
		**				 convention is taken from the GDD.
		**
		**	\author		Nate Cleveland
		**	\date		10/14/2002 11:27:44 AM
		**	\warning	Calls to the Beep and Buzz methods will do nothing until both params
		**				 have been mapped by the method.
		*/
		////////////////////////////////////////////////////////////////////////////////
		bool LoadSoundEffects(const char* beepPathName, const char* buzzPathName)throw();

		////////////////////////////////////////////////////////////////////////////////
		//	FUNCTION:
		/*!	\brief		Loads and plays the given music track.
		**
		**	\param		pathName	Pathname to the track to load.
		**
		**	\return		boolean true false. False == track was not loaded. Last track is
		**				 kept active..
		**
		**	\exception	none
		**	\par		DESCRIPTION: Loads one and only one BGM track. Instantly plays
		**				 the track.
		**
		**	\author		Nate Cleveland
		**	\date		10/14/2002 11:39:39 AM
		**	\warning	If this method is called Dynamic BGM methods will no longer
		**				 function until LoadDynamicMusic is called.
		*/
		////////////////////////////////////////////////////////////////////////////////
		bool LoadMusic(const char* pathName, UIState state) throw();

		////////////////////////////////////////////////////////////////////////////////
		//	FUNCTION:
		/*!	\brief		Loads a vector of dynamically toggleable sounds.
		**
		**	\param		vec		Const reference to a vector of sounds to load.
		**
		**	\return		Boolean true/false. False == at least one track failed to load
		**				 and therefore old sound tracks are retained.
		**
		**	\exception	none
		**	\par		DESCRIPTION: Loads a vector of sounds. Each vector member should be
		**				 a pathName to the desired music track. After this method has been
		**				 called Dynamic BGM methods will function correctly. They will
		**				 cease functioning iff LoadMusic is called afterwards.
		**
		**	\par		The first parameter in the vector will be level one, the second
		**				 will be level two, etc. Users of this class can change the current
		**				 Dynamic BGM level by calling FadeUp or FadeDown. Level one will be
		**				 instantly brought to the foreground and all others will remain muted.
		**
		**	\author		Nate Cleveland
		**	\date		10/14/2002 11:43:20 AM
		**	\warning	none
		*/
		////////////////////////////////////////////////////////////////////////////////
		bool LoadDynamicMusic(const std::vector<const char *>& vec)throw();

		////////////////////////////////////////////////////////////////////////////////
		//	FUNCTION:
		/*!	\brief		Plays the beep sound iff it isn't already playing.
		**	\param		none
		**
		**	\return		Boolean true/false. false == the sound is not playing. Will return
		**				 true if the sound is already playing.
		**
		**	\exception	none
		**	\par		DESCRIPTION: Plays the beep sound iff it isn't already playing.
		**
		**	\author		Nate Cleveland
		**	\date		10/14/2002 11:48:19 AM
		**	\warning	none
		*/
		////////////////////////////////////////////////////////////////////////////////
		bool Beep(void)throw();

		//! Plays the buzz sound. Will not play the buzz sound if it is already playing.
		////////////////////////////////////////////////////////////////////////////////
		//	FUNCTION:
		/*!	\brief		Plays the buzz sound iff it isn't already playing.
		**	\param		none
		**
		**	\return		Boolean true/false. false == the sound is not playing. Will return
		**				 true if the sound is already playing.
		**
		**	\exception	none
		**	\par		DESCRIPTION: Plays the buzz sound iff it isn't already playing.
		**
		**	\author		Nate Cleveland
		**	\date		10/14/2002 11:48:19 AM
		**	\warning	none
		*/
		////////////////////////////////////////////////////////////////////////////////
		bool Buzz(void)throw();

		////////////////////////////////////////////////////////////////////////////////
		//	FUNCTION:
		/*!	\brief		Raises the current Dynamic BGM level by one.
		**	\param		none
		**	\return		none
		**	\exception	none
		**	\par		DESCRIPTION: If LoadDyamicMusic has been called this method will
		**				 raise the current sound level by one. This means that the BGM
		**				 track that is currently in the foreground will be muted and the
		**				 track in the next sound level will be brough to the foreground.
		**
		**	\par		Does a bounds check to make sure we don't overstep our limits.
		**
		**	\author		Nate Cleveland
		**	\date		10/14/2002 11:51:04 AM
		**	\warning	Will do nothing if LoadMusic is called after LoadDynamicMusic
		*/
		////////////////////////////////////////////////////////////////////////////////
		void FadeUp()throw();

		////////////////////////////////////////////////////////////////////////////////
		//	FUNCTION:
		/*!	\brief		Lowers the current Dynamic BGM level by one.
		**	\param		none
		**	\return		none
		**	\exception	none
		**	\par		DESCRIPTION: If LoadDyamicMusic has been called this method will
		**				 lower the current sound level by one. This means that the BGM
		**				 track that is currently in the foreground will be muted and the
		**				 track in the next sound level down will be brough to the foreground.
		**
		**	\par		Does a bounds check to make sure we don't overstep our limits.
		**
		**	\author		Nate Cleveland
		**	\date		10/14/2002 11:51:04 AM
		**	\warning	Will do nothing if LoadMusic is called after LoadDynamicMusic
		*/
		////////////////////////////////////////////////////////////////////////////////
		void FadeDown()throw();

		////////////////////////////////////////////////////////////////////////////////
		//	FUNCTION:
		/*!	\brief		Returns to the caller the count of Dynamic BGM levels.
		**	\param		none
		**
		**	\return		A returned value of 0 implies that no BGM tracks have been loaded.
		**				 A returned value of 1 implies that Dynamic BGM is not currently
		**				 in use.
		**
		**	\exception	none
		**	\par		DESCRIPTION:
		**
		**	\author		Nate Cleveland
		**	\date		10/14/2002 11:59:27 AM
		**	\warning	none
		*/
		////////////////////////////////////////////////////////////////////////////////
		int GetMusicCount(void)const throw(){return this->m_musicCount;}

		////////////////////////////////////////////////////////////////////////////////
		//	FUNCTION:
		/*!	\brief		Returns to the caller the current level of Dynamic BGM.
		**	\param		none
		**
		**	\return		Interger value representing the current level of music playBack.
		**				 Ranges from 0 to GetMusicCount() - 1.
		**
		**	\exception	none
		**	\par		DESCRIPTION: Returns current level of playback. When in non-
		**				 Dynamic playback will always be 0.
		**
		**	\author		Nate Cleveland
		**	\date		10/14/2002 12:01:40 PM
		**	\warning	none
		*/
		////////////////////////////////////////////////////////////////////////////////
		int GetCurrMusicLevel(void)const throw() {return this->m_currMusicHandleIndex;}

		////////////////////////////////////////////////////////////////////////////////
		//	FUNCTION:
		/*!	\brief		Returns to the caller the volume of the BGM tracks.
		**	\param		none
		**
		**	\return		Volume of the BGM tracks, 0(mute) to 255(full on).
		**
		**	\exception	none
		**	\par		DESCRIPTION: The returned volume, like all volumes in the game,
						 are relative to the Global Volume.
		**
		**	\author		Nate Cleveland
		**	\date		10/14/2002 12:14:05 PM
		**	\warning	none
		*/
		////////////////////////////////////////////////////////////////////////////////
		int GetMusicVolume(void){return this->m_currMusicVolume;}

		////////////////////////////////////////////////////////////////////////////////
		//	FUNCTION:
		/*!	\brief		Change the volume of the BGM tracks.
		**
		**	\param		val		New volume for BGM tracks, 0(mute) to 255(full on).
		**
		**	\return		none
		**	\exception	none
		**	\par		DESCRIPTION: The new volume, like all volumes in the game,
		**				 are relative to the Global Volume.
		**
		**	\author		Nate Cleveland
		**	\date		10/14/2002 12:14:05 PM
		**	\warning	Will ignore the value if it is above 255 or less then 0.
		*/
		////////////////////////////////////////////////////////////////////////////////
		void SetMusicVolume(int val);

		////////////////////////////////////////////////////////////////////////////////
		//	FUNCTION:
		/*!	\brief		Returns to the caller the volume of the "Beep" and "Buzz" effects.
		**	\param		none
		**
		**	\return		Volume of the "Beep" and "Buzz" effects, 0(mute) to 255(full on).
		**
		**	\exception	none
		**	\par		DESCRIPTION: The returned volume, like all volumes in the game,
		**				 are relative to the Global Volume.
		**
		**	\author		Nate Cleveland
		**	\date		10/14/2002 12:14:05 PM
		**	\warning	none
		*/
		////////////////////////////////////////////////////////////////////////////////
		int GetEffectsVolume(void){return this->m_currEffectsVolume;}

		////////////////////////////////////////////////////////////////////////////////
		//	FUNCTION:
		/*!	\brief		Change the volume of the "Beep" and "Buzz" effects.
		**
		**	\param		val		New volume for "Beep" and "Buzz" effects, 0(mute) to
		**						 255(full on).
		**
		**	\return		none
		**	\exception	none
		**	\par		DESCRIPTION: The new volume, like all volumes in the game,
		**				 are relative to the Global Volume.
		**
		**	\author		Nate Cleveland
		**	\date		10/14/2002 12:14:05 PM
		**	\warning	Will ignore the value if it is above 255 or less then 0. Will also
		**				 not alter a "Beep" and "Buzz" effect if it is currently playing.
		**				 Will take effect the next time the sound is played.
		*/
		////////////////////////////////////////////////////////////////////////////////
		void SetEffectsVolume(int val);

		////////////////////////////////////////////////////////////////////////////////
		//	FUNCTION:
		/*!	\brief		Returns to the caller the global volume
		**	\param		none
		**
		**	\return		Global volume, 0(mute) to 255(full on).
		**
		**	\exception	none
		**	\par		DESCRIPTION: All other volumes in the game are reletive to the
		**				 returned value. See Sound::GetVolume(void) in sound.h for more
		**				 details.
		**
		**	\author		Nate Cleveland
		**	\date		10/14/2002 12:14:05 PM
		**	\warning	none
		*/
		////////////////////////////////////////////////////////////////////////////////
		int GetGlobalVolume(void) throw(){return this->m_sound.GetVolume();}

		////////////////////////////////////////////////////////////////////////////////
		//	FUNCTION:
		/*!	\brief		Changes the global volume
		**
		**	\param		val		New Global volume., 0(mute) to 255(full on).
		**
		**	\return		none
		**	\exception	none
		**	\par		DESCRIPTION: All other volumes in the game will be effected by this
		**				 new volume. See Sound::SetVolume(int) in sound.h for more details.
		**
		**	\author		Nate Cleveland
		**	\date		10/14/2002 12:14:05 PM
		**	\warning	Will ignore the value if it is above 255 or less then 0. Will also
		**				 not alter a "Beep" and "Buzz" effect if it is currently playing.
		**				 Will take effect the next time the sound is played.
		*/
		////////////////////////////////////////////////////////////////////////////////
		void SetGlobalVolume(int val){this->m_sound.SetVolume(val);}


		////////////////////////////////////////////////////////////////////////////////
		//	FUNCTION:	
		/*!	\brief		Update the listener position and direction.
		**	\param		pos			Pointer to a triple of float holding position of listener.
		**	\param		upVec		Pntr to a triple of floats containing up vector.
		**	\param		forwardVec	Pntr to a triple of floats containing the forward vector.
		**	\return		none
		**	\exception	none
		**	\par		DESCRIPTION:
		**				 Updates the listener position and direction. Null may be provided
		**				  for either direction vectors.
		**	\author		Nate Cleveland
		**	\date		12/2/2002 1:51:30 PM
		**	\warning	none
		*/
		////////////////////////////////////////////////////////////////////////////////
		void UpdateListener(float* pos, float* upVec = NULL, float* forwardVec = NULL) throw();


		void    SetState(UIState state)throw();
		UIState GetState(void)   const throw();


	protected:
		int				m_musicCount;				//!< Count of how many music tracks are loaded.
		Sound&			m_sound;					//!< Reference to the sound class.

		HSound			m_beep;						//!< Reference to "beep" sound.
		HSound			m_buzz;						//!< Reference to "buzz" sound.
		std::vector<HSound>		m_musicHandles;		//!< Vector of BGM handles.
		int				m_currMusicHandleIndex;		//!< Curr index to m_musicHandles Vec

		int				m_currMusicVolume;			//!< Current volume of the BGM
		int				m_currEffectsVolume;		//!< Current volume of "Beep" and "Buzz".
		static UIClient* m_instance;				
		UIState			m_state;					//!< Holds current state of uiclient


		////////////////////////////////////////////////////////////////////////////////
		//	FUNCTION:
		/*!	\brief		Frees a sound handle via m_sound and logs an error if there is one.
		**
		**	\param		h		Reference to the HSound that needs to be freed.
		**
		**	\return		none
		**	\exception	none
		**
		**	\par		DESCRIPTION: Besides freeing sounds and logging any errors the free
		**				 sound method also sets the HSound paramerter to 0.
		**
		**	\author		Nate Cleveland
		**	\date		10/14/2002 11:18:42 AM
		**	\warning	Will set HSound& h to zero after call is complete.
		*/
		////////////////////////////////////////////////////////////////////////////////
		void FreeSound(HSound& h);



		//below are the defines for the classes that'll hold our position and direction
		//vectors.
		class UIGivePos: public GivePos
		{
			float m_elements[3];		//!< Vector array for class.
			
		public:			
			//!Sets position to zero.
			UIGivePos()throw(){memset(m_elements, 0, sizeof(float)*3);}

			bool IsRelevent(void) const{return true;}
			float x(void)const throw(){return m_elements[0];}
			float y(void)const throw(){return m_elements[1];}
			float z(void)const throw(){return m_elements[2];}
			bool operator () (Point3d &pnt) const
			{
				pnt.x = x();
				pnt.y = y();
				pnt.z = z();
				return true;
			}
			
			float x(float newVal)throw(){m_elements[0] = newVal;}
			float y(float newVal)throw(){m_elements[1] = newVal;}
			float z(float newVal)throw(){m_elements[2] = newVal;}
			
			//does not check for null.
			void operator()(float *newVals){memcpy(m_elements, newVals, sizeof(float)*3);}
		}
		m_pos,										//!< Postion of the listener.
		m_upVec,									//!< Up pointing vec of listener.
		m_forwardVec;								//!< Forward vector of listener.
	};//end class UIClient
}//end namespace AcqariusSound