////////////////////////////////////////////////////////////////////////////////
/*!	\file		sound.h
**	\brief		Declares the Sound Class.
**	\author		Nate Cleveland
**	\date		10/14/2002 12:29:20 PM
**	\par		DESCRIPTION: See sound class header for description.					
*/
////////////////////////////////////////////////////////////////////////////////

#pragma once
#include "threadSafety.h"
#include "stlLibs.h"
#include "affine.h"
#include "GivePos.h"
#include "SoundSystem.h"

namespace AcqariusSound
{

//!An enumeration that tells the sound class what output type to use.
enum SoundOutput{DIRECT_SOUND, FMOD};

//!Max size that a sound file pathname can be.
extern const unsigned int MAX_SOUND_FILE_PATH_NAME/* = 32 */;

//Below are a list of the options that can be set in the Sound::LoadSound method.
//!See Sound::LoadSound for details.
extern const int SOUND_2D/* =			0x1*/; //!<See Sound::LoadSound for details.
extern const int SOUND_3D/* =			0x2*/; //!<See Sound::LoadSound for details.
extern const int SOUND_LARGE/* =		0x4*/; //!<See Sound::LoadSound for details.
extern const int SOUND_SMALL/* =		0x8*/; //!<See Sound::LoadSound for details.
extern const int SOUND_LOOP_NORMAL/* =	0x10*/;//!<See Sound::LoadSound for details.
extern const int SOUND_LOOP_OFF/* =		0x20*/;//!<See Sound::LoadSound for details.


////////////////////////////////////////////////////////////////////////////////
/*!	\class		Sound
**	\brief		
**	\author		Nate Cleveland
**	\date		10/4/2002 12:40:44 PM
**	\warning	none
**	\par		The sound class is a generalized interface into the fmod sound library.
**				 It has been created in such a way that it should be releatively 
**				 simple to add support for other sound libraries (ie DirectSound).
**
**	\par		For the sound class to support 3D clients must proved GivePos pointers
**				 via the method SetListener. 
**
**	\par		Exists within the AcqariusSound namespace.
*/
////////////////////////////////////////////////////////////////////////////////
class Sound
{
protected:
	//!This function object will return the current position of the  
	//! listener.
	GivePos		*m_listenerPos;

	//!This is the point that'll hold the old position of the player
	Point3d		m_oldListenerPos;

	//!This function object must return the direction that the player is
	//! facing. 
	GivePos		*m_listenerDirUp;
	GivePos		*m_listenerDirForward;


	//!polymorphic pntr to the sound system that we are currently using
	SoundSystem *m_system;
	

	//!Singleton functionality
	static Sound* m_instance;

public:

	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:	
	/*!	\brief		Defualt ctor for the Sound class.
	**	\arg  SoundOutput	type	An enum that tells the class what type of sound
	**								 output to work with.		
	**	\exception	Reserves the right to throw exception. Will throw exception if
	**				 the type member is set to something it doesn't understand. Currently
	**				 it only understands FMOD.
	**	\par		DESCRIPTION: Initializes the relevelent sound type.
	**					
	**	\author		Nate Cleveland
	**	\date		10/4/2002 12:45:49 PM
	**	\warning	none
	*/
	////////////////////////////////////////////////////////////////////////////////
	Sound(SoundOutput type = FMOD);


	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:	
	/*!	\brief		Defualt dtor for the Sound class. 
	**	\exception	none
	**	\par		DESCRIPTION: Calls release on the m_system member and then deletes it.
	**					
	**	\author		Nate Cleveland
	**	\date		10/4/2002 12:47:10 PM
	**	\warning	none
	*/
	////////////////////////////////////////////////////////////////////////////////
	~Sound();

	//!Singleton functionality
	static Sound* Instance(SoundOutput type = FMOD);
	static void Release(void);


	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:	
	/*!	\brief		Updates the Sound Module.
	**
	**	\arg  float elapsedTime			Time passed since Update was last called.
	**
	**	\return		none
	**	\exception	none
	**	\par		DESCRIPTION: Must be called once a frame to update all streaming 
	**				 and 3D sounds. 
	**					
	**	\author		Nate Cleveland
	**	\date		10/4/2002 12:47:59 PM
	**	\warning	none
	*/
	////////////////////////////////////////////////////////////////////////////////
	void Update(float elapsedTime) throw();


	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:	
	/*!	\brief		Must be called in order for 3D sound to work.
	**
	**	\arg  GivePos *pos				A pntr to a GivePos class that allows the 
	**										 sound class to easily find the position of 
	**										 the listener.
	**	\arg  GivePos *dirUp, *dirFor		Two more pntrs toe GivePos classes that allow
	**										 the sound class to easily find the 
	**										 orientation of the listener.		
	**	\return		none
	**	\exception	none
	**	\par		DESCRIPTION: Must be called at least once and provide the Sound class
	**				 with the position and direction of the listener. Must be called again 
	**				 if the listener ever changes. 3D Sound will only work if all are set.
	**					
	**	\author		Nate Cleveland
	**	\date		10/4/2002 12:48:59 PM
	**	\warning	none
	*/
	////////////////////////////////////////////////////////////////////////////////
	void SetListener(GivePos *pos, GivePos *dirUp, GivePos *dirForward) throw();



	////////////////////////Utility Methods /
	/////////////////////////////////////////
	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:	
	/*!	\brief		Loads a sound.
	**
	**	\arg  const  char *pathName		File name for the sound to be loaded.
	**	\arg  hSound &h					Reference to a hSound var that will become
	**										 the handle to this sound.	
	**	\return		Boolean true/false. true == SoundLoaded. false == NotLoaded.
	**	\exception	none
	**	\par		DESCRIPTION: Loads a sound for the caller. If the sound is already 
	**				 loaded this call will NOT fail. It reserves the right to either load 
	**				 another copy of the sound or give the caller a reference to the
	**				 already loaded copy.
	**	\par		Looping must be specified at this location. Currently there is no way
	**				 to loop a sound that isn't set to looping at load time.
	**
	**	\author		Nate Cleveland
	**	\date		10/4/2002 12:50:17 PM
	**	\warning	none
	**	\note
	**	\b Load \b Options:
	**	\li	SOUND_2D		Force 2D treatement of this sound.		
	**	\li	SOUND_3D		Force 3D effects on this sound.
	**	\li	SOUND_LARGE		Suggest streaming to the internal class.
	**	\li	SOUND_SMALL		Force non-streaming to the internal class.
	**	\li	SOUND_LOOP_OFF	Leave looping off for this sound.
	**	\li	SOUND_LOOP_ON	Turn looping on for this sound.
	*/
	////////////////////////////////////////////////////////////////////////////////
	bool LoadSound(const char *pathName, HSound &h, int loadOps) throw();


	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:	
	/*!	\brief		Attempts to find a copy of an already loaded sound. 
	**	\arg  const  char *pathName		File name for the sound to be located.
	**	\arg  hSound &h					Will be loaded with a handle to the sound.
	**
	**	\return		Returns false if it couldn't find the sound. Returns true if it did 
	**				 find the sound.
	**
	**	\exception	none
	**	\par		DESCRIPTION: Finds a loaded sound. Does not care if the sound is 
	**				 playing or not.
	**					
	**	\author		Nate Cleveland
	**	\date		10/4/2002 12:52:54 PM
	**	\warning	none
	*/
	////////////////////////////////////////////////////////////////////////////////
	bool FindSound(const char *pathName, HSound &h) throw();


	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:	
	/*!	\brief		Frees a loaded sound. 
	**
	**	\arg  hSound &h					Handle of the sound to be freed.	
	**
	**	\return		Returns false if it couldn't find the sound or if it's mutex
	**				 is in use by somebody. Returns true if it did nothing goes wrong.
	**	\exception	none
	**	\par		DESCRIPTION: Uses library specified method to free the sound.
	**					
	**	\author		Nate Cleveland
	**	\date		10/4/2002 12:54:49 PM
	**	\warning	none
	*/
	////////////////////////////////////////////////////////////////////////////////
	bool FreeSound(HSound &h) throw();


	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:	
	/*!	\brief		Frees all loaded sounds. 
	**	\param		none
	**	\return		Boolean true/false. True good, false bad. Common failure would be that
	**				 somebodies mutex was in use.
	**	\exception	none
	**	\par		DESCRIPTION: Uses library specified method to free all loaded sounds.
	**					
	**	\author		Nate Cleveland
	**	\date		10/4/2002 12:55:26 PM
	**	\warning	none
	*/
	////////////////////////////////////////////////////////////////////////////////
	bool FreeAll(void) throw();


	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:	
	/*!	\brief		Alters the volume of the given sound in relation to the global sound
	**				 level.
	**	\arg  hSound &h			Handle of the sound to be altered.
	**	\arg  int volume			A 0(silent) to 255(full on) value for the 
	**								 new volume.		
	**	\return		Boolean true/false. True good, false bad. Common failure would
	**				 be that the hSound's mutex is used by somebody else
	**	\exception	none
	**	\par		DESCRIPTION: Will only effect sounds that are currently playing.
	**					
	**	\author		Nate Cleveland
	**	\date		10/4/2002 12:57:23 PM
	**	\warning	none
	*/
	////////////////////////////////////////////////////////////////////////////////
	bool SetVolume(HSound &h, int volume) throw();


	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:	
	/*!	\brief		Alters the global volume, all playing sounds will be effected by this..
	**	\arg  int volume					A 0(silent) to 255(full on) value for the 
	**										 new volume.		
	**	\return		Boolean true/false. True good, false bad. 
	**	\exception	none
	**	\par		DESCRIPTION:
	**					
	**	\author		Nate Cleveland
	**	\date		10/4/2002 12:58:05 PM
	**	\warning	none
	*/
	////////////////////////////////////////////////////////////////////////////////
	bool SetVolume(int volume) throw();

	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:	
	/*!	\brief		Returns the volume of the given sound. 
	**
	**	\arg  hSound &h					Handle of the sound.
	**
	**	\return		Returns the current volume of the given sound. Range will be 
	**				 0(silent) to 255(full on). If something goes wrong will follow 
	**				 fmod's example and returns 0 to the caller.
	**
	**	\exception	none
	**	\par		DESCRIPTION: Returns the volume of the given sound. If there are 
	**				 multiple instances of this sound will return the value of the first
	**				 that is found. This is not an issue since SetVolume(HSound &, int)
	**				 forces all sounds of that handle type to be of the same volume.
	**					
	**	\author		Nate Cleveland
	**	\date		10/4/2002 12:17:05 PM
	**	\warning	none
	*/
	////////////////////////////////////////////////////////////////////////////////
	int GetVolume(HSound &h) throw();

	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:	
	/*!	\brief		Returns the global volume.
	**	\param		NONE
	**	\return		Returns the current global volume. Range will be 0(silent) to 
	**				 255(full on). If something goes wrong will follow fmod's example 
	**				 and returns 0 to the caller.
	**	\exception	none
	**	\par		DESCRIPTION:
	**					
	**	\author		Nate Cleveland
	**	\date		10/4/2002 12:22:57 PM
	**	\warning	none
	*/
	////////////////////////////////////////////////////////////////////////////////
	int GetVolume(void) throw();

	////////////////////////Playing Methods /
	/////////////////////////////////////////
	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:	
	/*!	\brief		Plays a given sound.
	**	\arg  hSound &h					Handle of the sound to be played.
	**	\arg  bool StartPaused			Users can set this to true if they want
	**										 the sound to start up paused. Defaulted to
	**										 false.		
	**	\return		Returns false if the handle is invalid or something else goes wrong.
	**				 Common failure would be that the max number of sounds are already 
	**				 playing.
	**	\exception	none
	**	\par		DESCRIPTION: Plays the indicated sound, the last member is a 
	**				 toggle for starting paused or not.
	**					
	**	\author		Nate Cleveland
	**	\date		10/4/2002 12:10:05 PM
	**	\warning	none
	*/
	////////////////////////////////////////////////////////////////////////////////
	bool PlaySound(HSound &h, bool StartPaused = false) throw();

	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:	
	/*!	\brief		Plays a given 3D sound.
	**	\arg  hSound &h					Handle of the sound to be played.
	**	\arg  const  GivePos &pos		Referce to the function object that provides
	**										 the sound object with the position of the 
	**										 sound.	
	**	\arg  bool StartPaused			Users can set this to true if they want
	**										 the sound to start up paused. Defaulted to
	**										 false.
	**	\return		Returns false if the handle is invalid or something else goes wrong.
	**				 Common failure would be that the max number of sounds are already 
	**				 playing
	**	\exception	none
	**	\par		DESCRIPTION: 3D effects will not work if the sound wasn't loaded with
	**				 the 3D options set. Besides 3D treament functions just as 
	**				Sound_Fmod::PlaySound(HSound &, bool);
	**					
	**	\author		Nate Cleveland
	**	\date		10/4/2002 12:11:50 PM
	**	\warning	none
	*/
	////////////////////////////////////////////////////////////////////////////////
	bool PlaySound(HSound &h, GivePos &pos, bool StartPaused = false) throw();



	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:	
	/*!	\brief		Stops all instances of a sound from being played.
	**
	**	\arg  hSound &h					Handle of the sound to be stoped.
	**
	**	\return		Returns false if the handle is invalid or something else goes wrong.
	**	\exception	none
	**	\par		DESCRIPTION: Stops all instances of a sound from playing.
	**					
	**	\author		Nate Cleveland
	**	\date		10/4/2002 12:23:32 PM
	**	\warning	none
	*/
	////////////////////////////////////////////////////////////////////////////////
	bool EndSound(HSound &h) throw();


	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:	
	/*!	\brief		Pauses or unpauses a sound
	**
	**	\arg  hSound &h					Handle of the sound to be paused or unpaused.
	**	\arg  bool SetPaused				Set to true to pause the sound, false to 
	**										 unpause it.
	**
	**	\return		Returns false if the handle is invalid or something else goes wrong.
	**	\exception	none
	**	\par		DESCRIPTION: Sets an internal paused flag to either true or false.
	**					
	**	\author		Nate Cleveland
	**	\date		10/4/2002 12:25:08 PM
	**	\warning	none
	*/
	////////////////////////////////////////////////////////////////////////////////
	bool Pause(HSound &h, bool SetPaused) throw();
	
	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:	
	/*!	\brief		Finds if this sound is playing.
	**
	**	\arg  hSound &h					Handle of the sound in quesion
	**
	**	\return		Returns true if the sound is playing.
	**	\exception	none
	**	\par		DESCRIPTION: O(n). Searches through a vector of active sounds and
	**				 returns true if this sound is playing.
	**					
	**	\author		Nate Cleveland
	**	\date		10/4/2002 12:25:08 PM
	**	\warning	none
	*/
	////////////////////////////////////////////////////////////////////////////////
	bool IsPlaying(HSound &h) throw();
	

	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:	
	/*!	\brief		Sets the volume of all non-streamed sounds.
	**	\param		volume		Volume to set all non-streamed sounds.
	**	\return		none
	**	\exception	none			
	**	\author		Nate Cleveland
	**	\date		2/27/2003 9:43:18 AM
	**	\warning	none
	*/
	////////////////////////////////////////////////////////////////////////////////
	void SetShortVolume(int volume) throw();
	


	void Deregister(GivePos& pos);
	
	};//end Sound Class

}//end AcqariusSound namespace