////////////////////////////////////////////////////////////////////////////////
/*! \file           DrawableHeightMap.h
**  \brief          DrawableHeightMap object inplementation file
**  \author         Austin Spafford
**  \date           11/1/2002 11:32:35 AM
**  \par            DESCRIPTION:
**                    The DrawableHeightMap is derived from the HeightMap object
**                    and adds drawing functionality.
**  \par            REMARKS:
**                    
*/
////////////////////////////////////////////////////////////////////////////////
#pragma warning (disable : 4786)
////////////////////////////////////////////////////////////////////////////////
#include "tankSimulator/DrawableHeightMap.h"
#include "messageHandler/MessageHandler.h"

#include "gfx/renderer.h"
#include "gfx/opengl.h"
#include "nexception/nexception.h"
#include <sstream>
////////////////////////////////////////////////////////////////////////////////
using namespace math::linear_alg;
using namespace std;
////////////////////////////////////////////////////////////////////////////////



////////////////////////////////////////////////////////////////////////////////
// Needed includes
////////////////////////////////////////////////////////////////////////////////
#include<gl/gl.h>
#include<gl/glu.h>
#include<math.h>
#include<windows.h>
#define for if(0);else for
// memory leak detection
#ifdef _DEBUG
    #include"memleak.h"
    #define new DEBUG_NEW
    #undef THIS_FILE
    static char THIS_FILE[] = __FILE__;
#endif



////////////////////////////////////////////////////////////////////////////////
/*! \struct     HeightMapVertex
**  \brief      Vertex structure used in the height map.
**  \author     M Jared Finder
**  \date       3/31/2003 10:35:53 AM
**  \warning    none
*/
////////////////////////////////////////////////////////////////////////////////
struct HeightMapVertex
{
    float pos[3];
    float tex[2];   //< shared between all texture units
    float clr[4];
};


namespace
{
    ////////////////////////////////////////////////////////////////////////////////
    //  FUNCTION:   
    /*! \brief      Calculates an interpolation ammount based on the vertex normal
    **  \param      vertexNormal    The vertex normal
    **  \return     none
    **  \exception  none
    **  \author     M Jared Finder
    **  \date       4/16/2003 4:36:39 PM
    **  \warning    none
    */
    ////////////////////////////////////////////////////////////////////////////////
    float CalculateTextureInterpolation(const Vector3<float> vertexNormal)
    {
        float rawInterpValue(1 - vertexNormal[1]);

        //normal range (0.15 to 0.5) is scaled to 0 to 1
        if(rawInterpValue < 0.15f)
        {
            return 0;
        }
        else if(rawInterpValue < 0.5f)
        {
            return (rawInterpValue - 0.15f) / (0.5f - 0.15f);
        }
        else
        {
            return 1;
        }
    }
}


////////////////////////////////////////////////////////////////////////////////
//  FUNCTION:       
/*! \brief          default constructer
**  \return         none
**  \exception      none
**  \author         Austin Spafford
**  \date           11/1/2002 8:40:14 PM
**  \warning        none
*/
////////////////////////////////////////////////////////////////////////////////
DrawableHeightMap::DrawableHeightMap()	
    :
    m_tex1("Resources/TankObj/Sand.tga", false, true),
    m_tex2("Resources/TankObj/RoughRock.tga", false, true),
    m_texDetail("Resources/TankObj/Detail.tga", false, true),
    m_texWall("Resources/TankObj/wall.tga", false, true),
    m_texCeiling("Resources/TankObj/ceiling.tga", false, true),
    m_vertexArray(256*256*sizeof(HeightMapVertex), false)
{
	return;
}

////////////////////////////////////////////////////////////////////////////////
//  FUNCTION:       
/*! \brief          deconstructer
**  \return         none
**  \exception      none
**  \author         Austin Spafford
**  \date           11/1/2002 8:40:14 PM
**  \warning        none
*/
////////////////////////////////////////////////////////////////////////////////
DrawableHeightMap::~DrawableHeightMap()
{
	return;
}

////////////////////////////////////////////////////////////////////////////////
//  FUNCTION:       
/*! \brief          draws the height map using its member graphics client
**  \return         none
**  \exception      none
**  \author         M Jared Finder
**  \date           11/1/2002 11:55:27 AM
**  \warning        none
*/
////////////////////////////////////////////////////////////////////////////////
void DrawableHeightMap::Render(void)
{
    //refresh any changed data
    RefreshData();

    //draw the background
    const float logicalTexWallWidth(30);
    const float logicalTexWallHeight(30);
    const float logicalTexCeilWidth(30);
    const float logicalTexCeilHeight(30);
    math::linear_alg::Vector3<float> min;
    math::linear_alg::Vector3<float> max;
    math::linear_alg::Vector3<float> delta;
    MsgHandler::MessageHandler::Instance()->getBoundary(min, max);
    delta[0] = fabsf(min[0] - max[0]);
    delta[1] = fabsf(min[1] - max[1]);
    delta[2] = fabsf(min[2] - max[2]);
    glColor3f(1,1,1);
    glDisable(GL_DEPTH_TEST);

    glBindTexture(GL_TEXTURE_2D, m_texWall.GetTextureHandle());
    glBegin(GL_QUADS);
        //draw left
        glTexCoord2f(0,                             delta[1]/logicalTexWallHeight);
        glVertex3f(min[0], max[1], max[2]);
        glTexCoord2f(delta[2]/logicalTexWallWidth,  delta[1]/logicalTexWallHeight);
        glVertex3f(min[0], max[1], min[2]);
        glTexCoord2f(delta[2]/logicalTexWallWidth,  0);
        glVertex3f(min[0], min[1], min[2]);
        glTexCoord2f(0,                             0);
        glVertex3f(min[0], min[1], max[2]);
        //draw right
        glTexCoord2f(delta[2]/logicalTexWallWidth,  delta[1]/logicalTexWallWidth);
        glVertex3f(max[0], max[1], max[2]);
        glTexCoord2f(delta[2]/logicalTexWallWidth,  0);
        glVertex3f(max[0], min[1], max[2]);
        glTexCoord2f(0,                             0);
        glVertex3f(max[0], min[1], min[2]);
        glTexCoord2f(0,                             delta[1]/logicalTexWallWidth);
        glVertex3f(max[0], max[1], min[2]);
        //draw front
        glTexCoord2f(0,                             delta[1]/logicalTexWallWidth);
        glVertex3f(max[0], max[1], max[2]);
        glTexCoord2f(delta[0]/logicalTexWallWidth,  delta[1]/logicalTexWallWidth);
        glVertex3f(min[0], max[1], max[2]);
        glTexCoord2f(delta[0]/logicalTexWallWidth,  0);
        glVertex3f(min[0], min[1], max[2]);
        glTexCoord2f(0,                             0);
        glVertex3f(max[0], min[1], max[2]);
        //draw back
        glTexCoord2f(delta[0]/logicalTexWallWidth,  delta[1]/logicalTexWallWidth);
        glVertex3f(max[0], max[1], min[2]);
        glTexCoord2f(delta[0]/logicalTexWallWidth,  0);
        glVertex3f(max[0], min[1], min[2]);
        glTexCoord2f(0,                             0);
        glVertex3f(min[0], min[1], min[2]);
        glTexCoord2f(0,                             delta[1]/logicalTexWallWidth);
        glVertex3f(min[0], max[1], min[2]);
    glEnd();
    glBindTexture(GL_TEXTURE_2D, m_texCeiling.GetTextureHandle());
    glBegin(GL_QUADS);
        //draw top
        glTexCoord2f(delta[2]/logicalTexCeilWidth,  delta[0]/logicalTexCeilHeight);
        glVertex3f(max[0], max[1], max[2]);
        glTexCoord2f(0,                             delta[0]/logicalTexCeilHeight);
        glVertex3f(max[0], max[1], min[2]);
        glTexCoord2f(0,                             0);
        glVertex3f(min[0], max[1], min[2]);
        glTexCoord2f(delta[2]/logicalTexCeilWidth,  0);
        glVertex3f(min[0], max[1], max[2]);
    glEnd();

    glEnable(GL_DEPTH_TEST);

    //transform to the space we want
    glPushMatrix();
    glTranslatef(m_position[0], m_position[1], m_position[2]);
    glScalef(m_scale[0], m_scale[1], m_scale[2]);

    //activate the base heightmap texture and vertices
    glBindTexture(GL_TEXTURE_2D, m_tex1.GetTextureHandle());
    glColor4f(1,1,1,1);
    m_vertexArray.Activate(true, true, true, sizeof(HeightMapVertex));
    
    //activate the alternate heightmap texture, if the texture unit is available
    glActiveTexture(GL_TEXTURE1);
    glClientActiveTexture(GL_TEXTURE1);
    glEnable(GL_TEXTURE_2D);
    glEnableClientState(GL_TEXTURE_COORD_ARRAY);
    glEnableClientState(GL_COLOR_ARRAY);
    glBindTexture(GL_TEXTURE_2D, m_tex2.GetTextureHandle());
    //NOTE: I don't have time to place this in the vertex buffer class.
    //      Because of this, I have to do this manually.
    glTexCoordPointer(
            2, GL_FLOAT, sizeof(HeightMapVertex),
            ((float*)m_vertexArray.m_pData) + 3
            );

    //draw the heightmap
    glLockArraysEXT(0, m_vertexDimensionality[0]*m_vertexDimensionality[1]);
    glDrawElements(GL_TRIANGLES, m_indexBuffer.size(), GL_UNSIGNED_SHORT,
                   &m_indexBuffer.front());
    glUnlockArraysEXT();

    //remove the alternate heightmap texture
    glBindTexture(GL_TEXTURE_2D, 0);
    glDisable(GL_TEXTURE_2D);
    glDisableClientState(GL_TEXTURE_COORD_ARRAY);
    glDisableClientState(GL_COLOR_ARRAY);
    glActiveTexture(GL_TEXTURE0);
    glClientActiveTexture(GL_TEXTURE0);

    m_vertexArray.Deactivate();

    glPopMatrix();
}


////////////////////////////////////////////////////////////////////////////////
//  FUNCTION:   
/*! \brief      Informs the graphics driver it should render the world
**  \return     none
**  \exception  none
**  \par        DESCRIPTION:
**                  
**  \author     M Jared Finder
**  \date       11/1/2002 1:23:21 PM
**  \warning    none
*/
////////////////////////////////////////////////////////////////////////////////
void DrawableHeightMap::SetRender(void)
{
    IRenderableWorld::SetRender();
}


////////////////////////////////////////////////////////////////////////////////
//  FUNCTION:   
/*! \brief      Refreshes all changed data since the last render of the height
**              map
**  \param      
**  \return     none
**  \exception  none
**  \author     M Jared Finder
**  \date       2/19/2003 7:51:59 PM
**  \warning    none
*/
////////////////////////////////////////////////////////////////////////////////
void DrawableHeightMap::RefreshData(void)
{
    //size in x and y of the texture;
    const float tex1xSize = 30 / HeightMap::m_cellDimensions[0];
    const float tex1ySize = 30 / HeightMap::m_cellDimensions[1];
    const float texDetailxSize = 12 / HeightMap::m_cellDimensions[0];
    const float texDetailySize = 12 / HeightMap::m_cellDimensions[1];

    //has the current space transformation changed?
    if(HeightMap::m_spaceTransformationChanged)
    {
        HeightMap::m_spaceTransformationChanged = false;

        m_position[0]   = HeightMap::m_positionOffset[0];
        m_position[1]   = HeightMap::m_positionOffset[1];
        m_position[2]   = HeightMap::m_positionOffset[2];
        m_scale[0]      = HeightMap::m_cellDimensions[0];
        m_scale[1]      = 1;
        m_scale[2]      = HeightMap::m_cellDimensions[1];
    }

    //have the vertices changed?
    if(HeightMap::m_verticesChanged)
    {
        m_verticesChanged = false;

        //get the new size
        m_vertexDimensionality[0] = HeightMap::m_pointCount[0];
        m_vertexDimensionality[1] = HeightMap::m_pointCount[1];
        if(m_vertexDimensionality[0] > 256 || m_vertexDimensionality[1] > 256)
        {
            std::ostringstream ex;
            ex << "Height Map too big!" << endl
               << m_vertexDimensionality[0] << ',' << m_vertexDimensionality[1];
            throw Nexception(
                ex.str().c_str(),
                __FILE__, __LINE__, __DATE__, __TIME__);
        }

        //create the vertex array
        HeightMapVertex* pData = (HeightMapVertex*)m_vertexArray.Unlock();
        for(int j = 0; j < m_vertexDimensionality[1]; ++j)
        {
            for(int i = 0; i < m_vertexDimensionality[0]; ++i)
            {
                //set the positions
                float& xCoord = pData[i+m_vertexDimensionality[0]*j].pos[0];
                float& yCoord = pData[i+m_vertexDimensionality[0]*j].pos[1];
                float& zCoord = pData[i+m_vertexDimensionality[0]*j].pos[2];
                xCoord = float(i);
                yCoord = HeightMap::GetPointValue(math::linear_alg::Vector2<int>(i,j));
                zCoord = float(j);

                //set the texture coordinates
                float& u = pData[i+m_vertexDimensionality[0]*j].tex[0];
                float& v = pData[i+m_vertexDimensionality[0]*j].tex[1];
                u = i / tex1xSize;
                v = j / tex1ySize;

                //set the color array
                float& r = pData[i+m_vertexDimensionality[0]*j].clr[0];
                float& g = pData[i+m_vertexDimensionality[0]*j].clr[1];
                float& b = pData[i+m_vertexDimensionality[0]*j].clr[2];
                float& a = pData[i+m_vertexDimensionality[0]*j].clr[3];
                Vector3<float> vertexNormal(GenerateVertexNormalVector(Vector2<int>(i,j)));
                r = 1;
                g = 1;
                b = 1;
                a = CalculateTextureInterpolation(vertexNormal);
            }
        }
        m_vertexArray.Lock();

        //create the index buffer
        m_indexBuffer.resize(
            (m_vertexDimensionality[0]-1) * (m_vertexDimensionality[1]-1) * 6
            );
        std::vector<unsigned short>::iterator it(m_indexBuffer.begin());
        for(unsigned short j = 0; j < m_vertexDimensionality[1]-1; ++j)
        {
            for(unsigned short i = 0; i < m_vertexDimensionality[0]-1; ++i)
            {
                //holds a pointer to the "quad" indecies
                *it++ = (i+0) + (j+0)*m_vertexDimensionality[0];
                *it++ = (i+1) + (j+1)*m_vertexDimensionality[0];
                *it++ = (i+0) + (j+1)*m_vertexDimensionality[0];

                *it++ = (i+0) + (j+0)*m_vertexDimensionality[0];
                *it++ = (i+1) + (j+0)*m_vertexDimensionality[0];
                *it++ = (i+1) + (j+1)*m_vertexDimensionality[0];
            }
        }
    }
}