////////////////////////////////////////////////////////////////////////////////
/*! \file           TankSimulator.cpp
**  \brief          TankSimulator implementation file
**  \author         Austin Spafford
**  \date           10/11/2002 6:11:04 PM
**  \par            DESCRIPTION:
**                    The TankSimulator object stores a collection of objects and
**                    provides an interface that allows them to interact with each
**                    other. It also simulates a volumetric plankton ecology.
**  \par            REMARKS:
**                    
*/
////////////////////////////////////////////////////////////////////////////////
#pragma warning (disable : 4786)
////////////////////////////////////////////////////////////////////////////////
#include <assert.h>
#include <math.h>
#include "nexception/Nexception.h"
#include "physics/PhysicsSimulator.h"
#include "tankSimulator/TankSimulator.h"
#include "interpretationModule/fishBody.hpp"
#include "interpretationModule/frozenFish.hpp"
#include "networking/NetData.h"
#include "ai_core/AI_System.h"
#include "TankSimulatorFileMemento.h"
// memory leak detection
#ifdef _DEBUG
    #include"memleak.h"
    #define new DEBUG_NEW
    #undef THIS_FILE
    static char THIS_FILE[] = __FILE__;
#endif
////////////////////////////////////////////////////////////////////////////////
using namespace math::linear_alg;
using namespace std;
using namespace InterpretationModule;
using namespace MsgHandler;
using namespace AI_System;
using namespace UDPNet;
////////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////
//  FUNCTION:       
/*! \brief          returns the input values after they are restricted to a
**					min/max value pair
**  \param          inputValue the value to be limited
**  \param          minimumValue a value to be limit inputValue to
**  \return         a limited version of inputValue
**  \exception      none            
**  \author         Austin Spafford
**  \date           10/11/2002 8:40:14 PM
**  \warning        none
*/
////////////////////////////////////////////////////////////////////////////////
template <typename type>
static Vector3<type> LimitedToMinimum(const Vector3<type> & inputValue, const Vector3<type> & minimumValue)
{
	Vector3<type> returnValue;

	for (int i=0; i<3; i++)
	{
		if (inputValue[i] < minimumValue[i])
		{
			returnValue[i] = minimumValue[i];
		}
		else
		{
			returnValue[i] = inputValue[i];
		}
	}

	return returnValue;
}

////////////////////////////////////////////////////////////////////////////////
//  FUNCTION:       
/*! \brief          returns the input values after they are restricted to a
**					min/max value pair
**  \param          inputValue the value to be limited
**  \param          minimumValue a value to be limit inputValue to
**  \param          maximumValue a value to be limit inputValue to
**  \return         a limited version of inputValue
**  \exception      none            
**  \author         Austin Spafford
**  \date           10/11/2002 8:40:14 PM
**  \warning        none
*/
////////////////////////////////////////////////////////////////////////////////
template <typename type>
static Vector3<type> LimitedToMinMax(const Vector3<type> & inputValue, const Vector3<type> & minimumValue, const Vector3<type> & maximumValue)
{
	Vector3<type> returnValue;

	for (int i=0; i<3; i++)
	{
		if (inputValue[i] < minimumValue[i])
		{
			returnValue[i] = minimumValue[i];
		}
		else 
		{
			if (inputValue[i] > maximumValue[i])
			{
				returnValue[i] = maximumValue[i];
			}
			else
			{
				returnValue[i] = inputValue[i];
			}
		}
	}

	return returnValue;
}

////////////////////////////////////////////////////////////////////////////////
//  FUNCTION:       
/*! \brief          default constructer
**  \return         none
**  \exception      none
**  \author         Austin Spafford
**  \date           10/30/2002 8:40:14 PM
**  \warning        none
*/
////////////////////////////////////////////////////////////////////////////////
TankSimulator::TankSimulator() :	
	m_planktonUpkeepCounter(0),
	m_planktonDensityVoxelCount(20,20,20),	
	m_planktonDensityBufferPtr(NULL),
	m_chumUpkeepCounter(0),
	m_chumVoxelCount(20,20,20),
	m_chumVoxelBufferPtr(NULL),
    m_gfxClient()
{
    // set the gfx client parent pointer
    m_gfxClient.SetParent(this);

	SetBoundary(
			Vector3<float>(0,0,0),
			Vector3<float>(1,1,1)
			);

	m_physicsSimulator.SetHeightMap(&m_heightMap);	

	ReallocatePlanktonVoxels();

	ReallocateChumVoxels();

	return;
}

////////////////////////////////////////////////////////////////////////////////
//  FUNCTION:       
/*! \brief          deconstructer
**  \return         none
**  \exception      none
**  \author         Austin Spafford
**  \date           10/30/2002 8:40:14 PM
**  \warning        none
*/
////////////////////////////////////////////////////////////////////////////////
TankSimulator::~TankSimulator()
{
	// delete all of the FishBodies
	hash_map<MsgHandler::Key, InterpretationModule::FishBody*>::iterator fishBody_key_pair_Iter;
	fishBody_key_pair_Iter = m_fishBodyPtrs.begin();
	while (fishBody_key_pair_Iter != m_fishBodyPtrs.end())
	{
		delete fishBody_key_pair_Iter->second;

		++fishBody_key_pair_Iter;
	}
	m_fishBodyPtrs.clear();	

	// release the plankton voxels
	delete [] m_planktonDensityBufferPtr;

	// release the chum voxels
	delete [] m_chumVoxelBufferPtr;

	return;
}

////////////////////////////////////////////////////////////////////////////////
//  FUNCTION:       
/*! \brief          adds a FishBody to the simulation and assigns the specified
**                  Key to it
**  \param          keyForObject the Key value that will be used as a label for
**                  the new FishBody
**  \return         true if a FishBody was added to the simulation, and false if
**                  the FishBody limit was reached or a FishBody already exists 
**                  that uses the specified Key as a label
**  \exception      none
**  \author         Austin Spafford
**  \date           11/1/2002 9:46:58 PM
**  \warning        none
*/
////////////////////////////////////////////////////////////////////////////////
bool TankSimulator::AddFishBody(MsgHandler::Key keyForObject)
{
	// If the fish already exists, blow off the request
	if (false != DoesFishBodyExist(keyForObject))
	{
		return false;
	}

	InterpretationModule::FishBody * newFishBodyPtr = new InterpretationModule::FishBody();	

	// insert the new FishBody
	m_fishBodyPtrs[keyForObject] = newFishBodyPtr;

	// Added by nate. [12/2/2002]
	//  swaped the order of SetKey and SetClient. The FishBody makes the assumption that
	//  it has a key before it has a client.
	// assign the FishBody Key
	newFishBodyPtr->SetKey(keyForObject);
	
	// give the new FishBody a new TankSimulatorClient
	newFishBodyPtr->SetClient(new TankSimulatorClient(*this));

	return true;
}

////////////////////////////////////////////////////////////////////////////////
//  FUNCTION:       
/*! \brief          adds a plankton node to the simulation
**  \param          position the position of the node in the simulation
**  \param          nodeStrength the strength of the node, a value of 1.0 should
**                  provide reasonable nurishment.
**  \return         none
**  \exception      none
**  \par            DESCRIPTION:
**  \author         Austin Spafford
**  \date           1/17/2003 6:47:15 PM
**  \warning        none
*/
////////////////////////////////////////////////////////////////////////////////
void TankSimulator::AddPlanktonNode(Vector3<float> position, float nodeStrength)
{
	PlanktonNode newPlanktonNode;

	newPlanktonNode.position = position;
	newPlanktonNode.nodeStrength = nodeStrength;

	// insert the new plankton node
	m_planktonNodes.push_back(newPlanktonNode);

	return;
}

////////////////////////////////////////////////////////////////////////////////
//  FUNCTION:       
/*! \brief          removes a FishBody with the specified Key
**  \param          keyForObject the Key value that will be used as a label for
**                  the FishBody that is to be removed
**  \return         true if the specified FishBody was removed, and false if the
**                  specified FishBody does not exist in the simulation
**  \exception      none
**  \author         Austin Spafford
**  \date           11/1/2002 9:46:58 PM
**  \warning        none
*/
////////////////////////////////////////////////////////////////////////////////
bool TankSimulator::RemoveFishBody(Key keyForObject)
{
	hash_map<Key,FishBody*>::iterator fishBodyPtrMapIter = m_fishBodyPtrs.find(keyForObject);

	if (m_fishBodyPtrs.end() == fishBodyPtrMapIter)
	{
		return false;
	}
		
	delete fishBodyPtrMapIter->second;

	m_fishBodyPtrs.erase(fishBodyPtrMapIter);

	return true;
}

////////////////////////////////////////////////////////////////////////////////
//  FUNCTION:       
/*! \brief          loads the specified image file and maps its red channel to 
**                  the height values of the height map
**  \param          fileName a pointer to a NULL terminated string that specifies 
**                  the file that should be read from
**  \param          maxHeight the value that a red value of 255 should be mapped
**                  to once the image is loaded
**  \return         returns true if the image was properly imported, false otherwise
**  \exception      none
**  \author         Austin Spafford
**  \date           12/6/2002 5:24:10 PM
**  \warning        Also changes the resolution of the height map to that of the image.
*/
////////////////////////////////////////////////////////////////////////////////
bool TankSimulator::ImportHeightMapFromImage(char * fileName, float maxHeight)
{
	return m_heightMap.ImportFromImage(fileName, maxHeight);
}
	
////////////////////////////////////////////////////////////////////////////////
//  FUNCTION:       
/*! \brief          saves a gray scale image of the height map, with white being
**                  the highest elevation
**  \param          fileName a pointer to a NULL terminated string that specifies 
**                  the file that should be written to
**  \return         returns true if the image was properly exported, false otherwise
**  \exception      none
**  \author         Austin Spafford
**  \date           12/6/2002 5:29:23 PM
**  \warning        The resolution of the image will be that of the height map
*/
////////////////////////////////////////////////////////////////////////////////
bool TankSimulator::ExportHeightMapToImage(char * fileName)
{	
	return m_heightMap.ExportToImage(fileName);
}

////////////////////////////////////////////////////////////////////////////////
//  FUNCTION:       
/*! \brief          Updates the state of the simulation by means of a NetMessage
**                  that originated on a sister TankSimulation by means of a call
**                  to GenerateNetMessage().
**  \param          netMessageRef a reference to a net message that is to
**                  read from
**  \return         none
**  \exception      none
**  \author         Austin Spafford
**  \date           2/17/2003 5:29:23 PM
*/
////////////////////////////////////////////////////////////////////////////////
void TankSimulator::LoadNetMessage(NetMessage & netMessageRef)
{	
	// We need a pointer that will point to various locations within the NetMessage
	char const * dataBufferPtr = NULL;

	// First we read the simulation boundaries from the NetMessage
	Vector3<float> newMinValues;
	Vector3<float> newMaxValues;
	dataBufferPtr = netMessageRef.ReadNBytes(sizeof(Vector3<float>));
	memcpy(&newMinValues, dataBufferPtr, sizeof(Vector3<float>));
	dataBufferPtr = netMessageRef.ReadNBytes(sizeof(Vector3<float>));
	memcpy(&newMaxValues, dataBufferPtr, sizeof(Vector3<float>));
	SetBoundary(newMinValues, newMaxValues);

	// We begin by wiping out any fish in the simulation that have been marked as
	// being temporary fish (unofficial fish)
	vector<Key> fishKeysToDelete;
	hash_map<Key, FishBody*>::iterator fishBody_key_pair_Iter;
	MessageHandler * messageHandlerPtr = MessageHandler::Instance();
	for (
		fishBody_key_pair_Iter = m_fishBodyPtrs.begin();
		fishBody_key_pair_Iter != m_fishBodyPtrs.end();
		++fishBody_key_pair_Iter
		)	
	{
		// check to see if the current fish is a temporary fish that needs
		// to be removed from the simulation since a complete list of official
		// fish has arrived
		if (false == messageHandlerPtr->IsOfficial(fishBody_key_pair_Iter->first))
		{
			fishKeysToDelete.push_back(fishBody_key_pair_Iter->first);
		}		
	}
	for (
		vector<Key>::iterator fishKeysToDeleteIter = fishKeysToDelete.begin();
		fishKeysToDeleteIter != fishKeysToDelete.end();
		++fishKeysToDeleteIter
		)
	{
		// remove the temporary fish from the simulation
		RemoveFishBody(*fishKeysToDeleteIter);
	}	
	// Read in the number of fish that have been included in the update
	unsigned int fishCountInUpdate;
	netMessageRef.ReadNBytes(sizeof(unsigned int), &fishCountInUpdate);
	// Create a copy of the fish ptr hash map that will keep track of which fish
	// have not been included in the update.
	hash_map<Key, FishBody*> fishNotIncludedInUpdate = m_fishBodyPtrs;
	// For each fish in the netmessage, we will:
	for (
		unsigned int currentFishInNetMessage = 0;
		currentFishInNetMessage < fishCountInUpdate;
		++currentFishInNetMessage
		)
	{
		// read the key of the fish
		Key fishKey;
		netMessageRef.ReadNBytes(sizeof(Key), &fishKey);

		// check to see if the fish is currently in the simulation
		FishBody * fishBodyPtr = GetFishBodyPtr(fishKey);
		if (NULL != fishBodyPtr)
		{
			// force the fish to update its current state
			*fishBodyPtr << netMessageRef;
			
			// remove the fish from the not-included list
			fishNotIncludedInUpdate.erase(fishKey);
		}
		else
		{
			// create a new fish
			AddFishBody(fishKey);

			// initialize the new fish
			*GetFishBodyPtr(fishKey) << netMessageRef;
		}
	}	
	// Delete any fish that are left unaccounted for
	for (
		fishBody_key_pair_Iter = fishNotIncludedInUpdate.begin();
		fishBody_key_pair_Iter != fishNotIncludedInUpdate.end();
		++fishBody_key_pair_Iter
		)
	{
		// we can directly remove this fish body since in order to recieve a
		// NetMessage, it is assumed that we are running on a client, and therefore
		// with out worry about a corresponding brain to remove.
		RemoveFishBody(fishBody_key_pair_Iter->first);
	}
	
	// Now we load the HeightMap if it is present
	dataBufferPtr = netMessageRef.ReadNBytes(sizeof(bool));
	if (false != *reinterpret_cast<const bool*>(dataBufferPtr))
	{
		// force the HeightMap to load its state from the NetMessage
		m_heightMap.LoadNetMessage(netMessageRef);
	}

	// LOAD PLANKTON	
	// check to see if plankton info has been included

	// LOAD CHUM
	// check to see if any chum info has been included
		
	return;
}

////////////////////////////////////////////////////////////////////////////////
//  FUNCTION:       
/*! \brief          Updates the state of the heightmap using a netmessage.
**  \param          netMessageRef a reference to a net message that is to
**                  read from
**  \return         none
**  \exception      none
**  \author         Austin Spafford
**  \date           5/10/2003 9:52:18 PM
**  \warning        none
*/
////////////////////////////////////////////////////////////////////////////////
void TankSimulator::LoadHeightMapNetMessage(UDPNet::NetMessage & netMessageRef)
{
	m_heightMap.LoadNetMessage(netMessageRef);
}

////////////////////////////////////////////////////////////////////////////////
//  FUNCTION:       
/*! \brief          Generates a NetMessage that will encode the current state of
**                  the simulation, so that a sister TankSimulation can load it
**                  by means of a call to LoadNetMessage(). The generated
**                  NetMessage will most likely be missing information depending
**                  on what information has changed since the last call to
**                  GenerateNetMessage().
**  \param          netMessageRef a reference to a net message that is to be
**                  written to
**  \return         none
**  \exception      none
**  \author         Austin Spafford
**  \date           2/17/2003 5:29:23 PM
*/
////////////////////////////////////////////////////////////////////////////////
void TankSimulator::GeneratePartialNetMessage(NetMessage & netMessageRef)
{
	// Save the simulation boundraries
	netMessageRef.WriteNBytes(sizeof(Vector3<float>), &m_boundaryMins);
	netMessageRef.WriteNBytes(sizeof(Vector3<float>), &m_boundaryMaxs);	

	// We begin by writing the number of fish present in the simulation
	// to the NetMessage
	unsigned int fishCount = m_fishBodyPtrs.size();
	netMessageRef.WriteNBytes(sizeof(unsigned int), &fishCount);
	// We will now force every fish to write to the netmessage
	for (
		hash_map<Key, FishBody*>::iterator fishBody_key_pair_Iter = m_fishBodyPtrs.begin();
		fishBody_key_pair_Iter != m_fishBodyPtrs.end();
		++fishBody_key_pair_Iter
		)	
	{
		// write the key of the fish		
		Key fishKey = fishBody_key_pair_Iter->second->GetKey();
		netMessageRef.WriteNBytes(sizeof(Key), &fishKey);
		
		// write the fish itself
		*(fishBody_key_pair_Iter->second) >> netMessageRef;
	}
	
	// Now we save the HeightMap, we begin by checking to see if any changes
	// have been made to it, if not, then we will skip it
	if (false/* != m_heightMap.HasDataChanged()*/)
	{
		// mark the presence of a HeightMap update in the NetMessage
		bool trueBool = true;
		netMessageRef.WriteNBytes(sizeof(bool), &trueBool);

		// force the HeightMap to append its state to the NetMessage
		m_heightMap.GenerateNetMessage(netMessageRef);
	}
	else
	{
		// mark the lack of a HeightMap update in the NetMessage
		bool falseBool = false;
		netMessageRef.WriteNBytes(sizeof(bool), &falseBool);
	}
	m_heightMap.ClearChangesFlag();

	// SAVE PLANKTON	
	// check to see if plankton info has been included

	// SAVE CHUM
	// check to see if any chum info has been included

	return;
}

////////////////////////////////////////////////////////////////////////////////
//  FUNCTION:       
/*! \brief          Generates a NetMessage that will encode the current state of
**                  the simulation, so that a sister TankSimulation can load it
**                  by means of a call to LoadNetMessage(). This version of
**                  GenerateNetMessage() will always include every piece of
**                  data regardless of what has recently changed.
**  \param          netMessageRef a reference to a net message that is to be
**                  written to
**  \return         none
**  \exception      none
**  \author         Austin Spafford
**  \date           3/11/2003 5:29:23 PM
*/
////////////////////////////////////////////////////////////////////////////////
void TankSimulator::GenerateCompleteNetMessage(NetMessage & netMessageRef)
{
	// Save the simulation boundraries
	netMessageRef.WriteNBytes(sizeof(Vector3<float>), &m_boundaryMins);
	netMessageRef.WriteNBytes(sizeof(Vector3<float>), &m_boundaryMaxs);	

	// We begin by writing the number of fish present in the simulation
	// to the NetMessage
	unsigned int fishCount = m_fishBodyPtrs.size();
	netMessageRef.WriteNBytes(sizeof(unsigned int), &fishCount);
	// We will now force every fish to write to the netmessage
	for (
		hash_map<Key, FishBody*>::iterator fishBody_key_pair_Iter = m_fishBodyPtrs.begin();
		fishBody_key_pair_Iter != m_fishBodyPtrs.end();
		++fishBody_key_pair_Iter
		)	
	{
		// write the key of the fish		
		Key fishKey = fishBody_key_pair_Iter->second->GetKey();
		netMessageRef.WriteNBytes(sizeof(Key), &fishKey);
		
		// write the fish itself
		*(fishBody_key_pair_Iter->second) >> netMessageRef;
	}
		
	// mark the presence of a HeightMap update in the NetMessage
	bool trueBool = true;
	netMessageRef.WriteNBytes(sizeof(bool), &trueBool);
	// force the HeightMap to append its state to the NetMessage
	m_heightMap.GenerateNetMessage(netMessageRef);	

	// SAVE PLANKTON	
	// check to see if plankton info has been included

	// SAVE CHUM
	// check to see if any chum info has been included

	return;
}

////////////////////////////////////////////////////////////////////////////////
//  FUNCTION:       
/*! \brief          Generates a NetMessage that can ONLY be loaded by
**                  LoadHeightMapNetMessage()
**  \param          netMessageRef a reference to a net message that is to be
**                  written to
**  \return         none
**  \exception      none
**  \author         Austin Spafford
**  \date           5/10/2003 9:50:33 PM
**  \warning        none
*/
////////////////////////////////////////////////////////////////////////////////
void TankSimulator::GenerateHeightMapNetMessage(NetMessage & netMessageRef)
{
	m_heightMap.GenerateNetMessage(netMessageRef);
}

////////////////////////////////////////////////////////////////////////////////
//  FUNCTION:       
/*! \brief          A factory method that generates an object that encapsulates 
**                  the state of the simulation.
**  \return         A pointer to a new FileMemento object.
**  \exception      none
**  \author         Austin Spafford
**  \date           5/15/2003 3:54:55 PM
**  \warning        none
*/
////////////////////////////////////////////////////////////////////////////////
FileMemento * TankSimulator::CreateFileMemento(void) const
{
	TankSimulatorFileMemento * newFileMomentoPtr = new TankSimulatorFileMemento;

	// initialize the boundaries
	newFileMomentoPtr->m_boundaryMins = m_boundaryMins;
	newFileMomentoPtr->m_boundaryMaxs = m_boundaryMaxs;

	// copy all of the fish
	newFileMomentoPtr->m_frozenFish.clear();
	for (
		hash_map<Key,FishBody*>::const_iterator fishBodyIter = m_fishBodyPtrs.begin();
		m_fishBodyPtrs.end() != fishBodyIter;
		++fishBodyIter
		)
	{
		newFileMomentoPtr->m_frozenFish.push_back(*fishBodyIter->second);
	}

	// copy the height map
	newFileMomentoPtr->m_heightMapFileMementoPtr = m_heightMap.CreateFileMemento();

	// copy over the fish key lists
	newFileMomentoPtr->m_matingFishList = m_matingFishList;	
	newFileMomentoPtr->m_deleteFishList = m_deleteFishList;

	return newFileMomentoPtr;
}

////////////////////////////////////////////////////////////////////////////////
//  FUNCTION:       
/*! \brief          Sets the current state of the simulation to that which is
**                  stored in fileMemento.
**  \param          fileMemento the object that will be used to set the current
**                  state of the simulation.
**  \return         none
**  \exception      none
**  \author         Austin Spafford
**  \date           5/15/2003 4:32:15 PM
**  \warning        fileMemento MUST actually reference a TankSimulatorFileMemento!
*/
////////////////////////////////////////////////////////////////////////////////
void TankSimulator::LoadFileMemento(const FileMemento & fileMemento)
{
	const TankSimulatorFileMemento & tankSimulatorFileMemento = dynamic_cast<const TankSimulatorFileMemento &>(fileMemento);

	// load the boundaries
	SetBoundary(tankSimulatorFileMemento.m_boundaryMins, tankSimulatorFileMemento.m_boundaryMaxs);

	// we must first wipe out all of the fish before we load the ones in the memento
	while (0 != GetNextFish(0))
	{
		// it is assumed that the coordesponding brains have been also been replaced
		// by the caller of this function
		RemoveFishBody(GetNextFish(0));
	}	
	// load the fish bodies	
	for (
		vector<FrozenFish>::const_iterator frozenFishIter = tankSimulatorFileMemento.m_frozenFish.begin();
		tankSimulatorFileMemento.m_frozenFish.end() != frozenFishIter;
		++frozenFishIter
		)
	{

		// create a new fish
		AddFishBody(frozenFishIter->m_key);

		// initialize the new fish
		FishBody * fishBodyPtr = GetFishBodyPtr(frozenFishIter->m_key);
		assert(NULL != fishBodyPtr);
		fishBodyPtr->Build(*frozenFishIter);
	}
	
	// load the heightmap
	assert(NULL != tankSimulatorFileMemento.m_heightMapFileMementoPtr);	
	m_heightMap.LoadFileMemento(*tankSimulatorFileMemento.m_heightMapFileMementoPtr);

	// load the fish key lists
	m_matingFishList = tankSimulatorFileMemento.m_matingFishList;
	m_deleteFishList = tankSimulatorFileMemento.m_deleteFishList;

	return;
}

////////////////////////////////////////////////////////////////////////////////
//  FUNCTION:       
/*! \brief          Traces a ray against every fish in the simulation.
**  \param          rayOrigin the world space origin of the ray
**  \param          rayDirection the direction vector of the ray
**  \param          fishKeyRef a reference to a key that will be filled out with
**                  the key of the closest fish that intersects the ray, its value
**                  will be set to 0 if no fish were hit by the ray
**  \return         true if a fish was hit and therefore the value referenced to
**                  by fishKeyRef changed, false otherwise.
**  \exception      none
**  \author         Austin Spafford
**  \date           2/6/2003 8:16:47 PM
**  \warning        This algorithm is built for speed and not accuracy, it
**                  assumes that the fish are not intersecting each other.
*/
////////////////////////////////////////////////////////////////////////////////
bool TankSimulator::TraceRayAgainstAllFish(
		const Vector3<float> & rayOrigin,
		const Vector3<float> & rayDirection,
		Key & fishKeyRef
		)
{
	float smallestIntersectingFishCenterDistanceSquared = -1; // mark the current intersection as being bogus

	// we must make sure that the ray direction vector is normalized
	Vector3<float> normalizedRayDirection = rayDirection;
	normalizedRayDirection.Normalize();

	// check to see if any of the fish are intersecting the ray
	hash_map<Key, FishBody*>::iterator fishBody_key_pair_Iter = m_fishBodyPtrs.begin();
	while (fishBody_key_pair_Iter != m_fishBodyPtrs.end())
	{
		FishBody & fishBodyRef = *fishBody_key_pair_Iter->second;

		// obtain a reference to the sphere that is representing the fish
		PhysicsSimulator::Sphere & fishSphereRef = m_physicsSimulator.GetSphereRef(fishBodyRef.GetSphereKey());

		// check to see if the fish is behind the ray origin
		Vector3<float> vectorFromFishCenterToRayOrigin =
				rayOrigin
				- fishSphereRef.GetPosition();
		float dotProductOfRayDirectionAndVectorFromFishCenterToRayOrigin =
				Dot (
					normalizedRayDirection,
					vectorFromFishCenterToRayOrigin
					);
		if (dotProductOfRayDirectionAndVectorFromFishCenterToRayOrigin < 0)
		{
			// this value is used in two calculations, and thus has been 
			// pulled out for efficiency's sake
			float distanceToFishSquared = Dot(
					vectorFromFishCenterToRayOrigin,
					vectorFromFishCenterToRayOrigin
					);

			// figure out what the intersection time of the projected vector is, using
			// the quadratic formula		
			float quadFormula_a = Dot(normalizedRayDirection, normalizedRayDirection);
			float quadFormula_b = 2 * dotProductOfRayDirectionAndVectorFromFishCenterToRayOrigin;
			float quadFormula_c = 
					distanceToFishSquared
					- fishSphereRef.GetRadius() * fishSphereRef.GetRadius();
			float quadeFormula_discriminant = 
					quadFormula_b * quadFormula_b
					- 4 * quadFormula_a * quadFormula_c;

			// now we can check to see if the ray intersects with the fish's sphere			
			if (quadeFormula_discriminant >= 0)
			{				
				// check to see if this is the first fish to intersect the ray
				if (smallestIntersectingFishCenterDistanceSquared < 0)
				{
					// record that this is the closest fish encountered
					smallestIntersectingFishCenterDistanceSquared = distanceToFishSquared;
					fishKeyRef = fishBody_key_pair_Iter->first;
				}
				else
				{
					// another fish has already been hit, so we have to compare the two
					if (distanceToFishSquared < smallestIntersectingFishCenterDistanceSquared)
					{
						// record that this is the closest fish encountered
						smallestIntersectingFishCenterDistanceSquared = distanceToFishSquared;
						fishKeyRef = fishBody_key_pair_Iter->first;
					}
					else
					{
						// this fish is farther than the closest, and will be rejected
					}
				}
			}
			else
			{
				// the discriminant is negative, and thus the
				// intersection points are imaginary
			}
		}
		else
		{
			// the fish is behind the ray origin, and thus is culled away
		}		

		++fishBody_key_pair_Iter;
	}

	// return true if the ray is intersection any of the fish
	if (smallestIntersectingFishCenterDistanceSquared >= 0)
	{
		return true;
	}
	else
	{
		fishKeyRef = 0;
		return false;
	}
}

////////////////////////////////////////////////////////////////////////////////
//  FUNCTION:       
/*! \brief          fills in minValues and maxValues with the current min and 
**                  max boundary values
**  \param          minValues a three dimensional vector that will be filled out
**                  with the minimal values used by the boundry box
**  \param          maxValues a three dimensional vector that will be filled out
**                  with the maximal values used by the boundry box
**  \return         none
**  \exception      none
**  \author         Austin Spafford
**  \date           11/13/2002 8:40:14 PM
**  \warning        none
*/
////////////////////////////////////////////////////////////////////////////////
void TankSimulator::GetBoundary(math::linear_alg::Vector3<float> & minValues, math::linear_alg::Vector3<float> & maxValues)
{
	minValues = m_boundaryMins;
	maxValues = m_boundaryMaxs;

	return;
}

////////////////////////////////////////////////////////////////////////////////
//  FUNCTION:       
/*! \brief          retrieves a pointer to the specified FishBody
**  \param          fishKey a Key to a FishBody
**  \return         a pointer to the FishBody, will be NULL if the requested 
**                  FishBody does not exist
**  \exception      none
**  \author         Austin Spafford
**  \date           11/4/2002 2:20:24 PM
**  \warning        none
*/
////////////////////////////////////////////////////////////////////////////////
InterpretationModule::FishBody * TankSimulator::GetFishBodyPtr(Key fishKey)
{
	hash_map<Key, FishBody*>::iterator fishBody_key_pair_Iter = m_fishBodyPtrs.find(fishKey);

	if (m_fishBodyPtrs.end() == fishBody_key_pair_Iter)
	{
		return NULL;
	}

	return fishBody_key_pair_Iter->second;
}

////////////////////////////////////////////////////////////////////////////////
//  FUNCTION:       
/*! \brief          retrieves the number of voxels that lie along each axis
**  \return         a vector that represents the number of voxels that lie along
**                  each axis
**  \exception      none
**  \author         Austin Spafford
**  \date           11/14/2002 1:50:55 PM
**  \warning        none
*/
////////////////////////////////////////////////////////////////////////////////
math::linear_alg::Vector3<int> TankSimulator::GetPlanktonVoxelCount(void)
{
	return m_planktonDensityVoxelCount;
}
	
////////////////////////////////////////////////////////////////////////////////
//  FUNCTION:       
/*! \brief          retrieves the value of a specific voxel
**  \param          voxelIndex the axis-aligned index of the voxel that the
**                  value should be retrieved from. Make sure the index is legal!
**  \return         the value stored in the specified voxel
**  \exception      none
**  \author         Austin Spafford
**  \date           11/14/2002 1:50:59 PM
**  \warning        none
*/
////////////////////////////////////////////////////////////////////////////////
float TankSimulator::GetPlanktonDensityFromVoxel(math::linear_alg::Vector3<int> voxelIndex)
{
	assert(voxelIndex == LimitedToMinMax(voxelIndex, Vector3<int>(0,0,0), Vector3<int>(m_planktonDensityVoxelCount[0]-1, m_planktonDensityVoxelCount[1]-1, m_planktonDensityVoxelCount[2] - 1)));

	int rowPitch = m_planktonDensityVoxelCount[0];
	int planePitch = m_planktonDensityVoxelCount[0] * m_planktonDensityVoxelCount[1];

	float * currentVoxelPtr = 
		m_planktonDensityBufferPtr
		+
		voxelIndex[0]
		+
		voxelIndex[1] * rowPitch
		+
		voxelIndex[2] * planePitch;

	return currentVoxelPtr[0];
}

////////////////////////////////////////////////////////////////////////////////
//  FUNCTION:       
/*! \brief          Retrieves the value of a specific position within plankton
**                  ecology. If the location is outside the ecology, then the
**                  closest point inside the simulation will be sampled from.
**  \param          position the location that is to be sampled from
**  \return         the density of the plankton at the specified location
**  \exception      none
**  \author         Austin Spafford
**  \date           11/14/2002 1:50:59 PM
**  \warning        none
*/
////////////////////////////////////////////////////////////////////////////////
float TankSimulator::GetPlanktonDensityAtLocation(const Vector3<float> & position)
{
	Vector3<float> boundPosition = LimitedToMinMax(position, m_boundaryMins, m_boundaryMaxs);

	// generate a density field that is goes from 0.1 to 0.8, with the highest
	// density near the ground
	float tankHeight = m_boundaryMaxs[1] - m_boundaryMins[1];
	return 0.1f + 0.7f*(1 - ((boundPosition[1] - m_boundaryMins[1]) / tankHeight));
}

////////////////////////////////////////////////////////////////////////////////
//  FUNCTION:       
/*! \brief          retrieves the number of chum voxels that lie along each axis
**  \return         a vector that represents the number of voxels that lie along
**                  each axis
**  \exception      none
**  \author         Austin Spafford
**  \date           11/14/2002 1:50:55 PM
**  \warning        none
*/
////////////////////////////////////////////////////////////////////////////////
math::linear_alg::Vector3<int> TankSimulator::GetChumVoxelCount(void)
{
	return m_chumVoxelCount;
}

////////////////////////////////////////////////////////////////////////////////
//  FUNCTION:       
/*! \brief          retrieves the values of a specific chum voxel
**  \param          
**  \return         none
**  \exception      none
**  \par            DESCRIPTION:
**                    It tastes like chicken.
**  \author         Austin Spafford
**  \date           1/18/2003 3:39:16 PM
**  \warning        none
*/
////////////////////////////////////////////////////////////////////////////////
void TankSimulator::GetChumVoxelValues(Vector3<int> voxelIndex, Vector3<float> & chumPositionRef, float & chumDensityRef)
{
	assert(voxelIndex == LimitedToMinMax(voxelIndex, Vector3<int>(0,0,0), Vector3<int>(m_chumVoxelCount[0]-1, m_chumVoxelCount[1]-1, m_chumVoxelCount[2] - 1)));

	int rowPitch = m_chumVoxelCount[0];
	int planePitch = m_chumVoxelCount[0] * m_chumVoxelCount[1];

	ChumVoxel * currentVoxelPtr = 
		m_chumVoxelBufferPtr
		+
		voxelIndex[0]
		+
		voxelIndex[1] * rowPitch
		+
		voxelIndex[2] * planePitch;

	// return the specified voxel's values
	chumPositionRef = currentVoxelPtr->chumPosition;
	chumDensityRef = currentVoxelPtr->chumDensity;

	return;
}

////////////////////////////////////////////////////////////////////////////////
//  FUNCTION:       
/*! \brief          retrieves a reference to the mating fish list
**  \return         a reference to the mating fish list
**  \exception      none
**  \author         Austin Spafford
**  \date           11/25/2002 4:40:02 AM
**  \warning        none
*/
////////////////////////////////////////////////////////////////////////////////
std::list<MsgHandler::Key> & TankSimulator::GetMatingFishListRef(void)
{
	return m_matingFishList;
}	

////////////////////////////////////////////////////////////////////////////////
//  FUNCTION:       
/*! \brief          retrieves a reference to the delete fish list
**  \return         a reference to the delete fish list
**  \exception      none
**  \author         Austin Spafford
**  \date           11/25/2002 4:40:51 AM
**  \warning        none
*/
////////////////////////////////////////////////////////////////////////////////
list<Key> & TankSimulator::GetDeleteFishListRef(void)
{
	return m_deleteFishList;
}

////////////////////////////////////////////////////////////////////////////////
//  FUNCTION:       
/*! \brief          retrieves a vector of Keys to fish that are currently visible
**                  to the specified fish
**  \param          fishKey a Key to the FishBody the will be used as a point of
**                  view when figuring out when which fish are visible
**  \return         a vector of Keys for FishBodies that are visible to the
**                  specified fish
**  \exception      none
**  \author         Austin Spafford
**  \date           12/2/2002 3:20:45 PM
**  \warning        none
*/
////////////////////////////////////////////////////////////////////////////////
void TankSimulator::GetVectorOfFishVisibleToFish(Key fishKey, vector<Key>& returnValue)
{
	hash_map<Key, FishBody*>::iterator fishBody_key_pair_Iter = m_fishBodyPtrs.begin();
	FishBody *specifiedFishBodyPtr = GetFishBodyPtr(fishKey);

	if (NULL == specifiedFishBodyPtr)
	{
		throw Nexception("The specified fish does not exist.");
	}

	float specifiedFishBody_VisionRadiusSquared = specifiedFishBodyPtr->GetVisionRadius();
	specifiedFishBody_VisionRadiusSquared *= specifiedFishBody_VisionRadiusSquared;

	// scan for visible fish
	while (fishBody_key_pair_Iter != m_fishBodyPtrs.end())
	{
		assert(NULL != fishBody_key_pair_Iter->second);

		// avoid having the specified fish be on the visible fish list
		if (fishBody_key_pair_Iter->second == specifiedFishBodyPtr)
		{
			++fishBody_key_pair_Iter;
			continue;
		}

		FishBody *currentFishBodyPtr = fishBody_key_pair_Iter->second;

		Vector3<float> currentFishBodyPosition;
		Vector3<float> specfiedFishBodyPosition;

		currentFishBodyPtr->GetPos(currentFishBodyPosition);
		specifiedFishBodyPtr->GetPos(specfiedFishBodyPosition);

		Vector3<float> displacementVector = currentFishBodyPosition - specfiedFishBodyPosition;

		float fishDistanceSquared = 
				displacementVector[0]*displacementVector[0]
				+
				displacementVector[1]*displacementVector[1]
				+
				displacementVector[2]*displacementVector[2];		

		// check to see if the current fish being examined is visible
		if (fishDistanceSquared <= specifiedFishBody_VisionRadiusSquared)
		{
			returnValue.push_back(fishBody_key_pair_Iter->first);
		}

		++fishBody_key_pair_Iter;
	}

}

////////////////////////////////////////////////////////////////////////////////
//  FUNCTION:       
/*! \brief          retrieves the key to the next fish in the simulation 
**                  (arbitrary order, wraps)
**  \param          fishKey a Key to a FishBody
**  \return         the first fish's Key if the specfied fish does not exist, 
**                  and 0 if no fish are in the simulation
**  \exception      none
**  \author         Austin Spafford
**  \date           12/4/2002 10:53:42 AM
**  \warning        none
*/
////////////////////////////////////////////////////////////////////////////////
Key TankSimulator::GetNextFish(Key fishKey)
{
	hash_map<Key,FishBody*>::iterator fishBody_key_pair_Iter;
	
	// find the specified fish
	fishBody_key_pair_Iter = m_fishBodyPtrs.find(fishKey);

	// check to see if the fish exists
	if (m_fishBodyPtrs.end() == fishBody_key_pair_Iter)
	{
		// check to see if there are no fish are in the simulation
		if (0 == m_fishBodyPtrs.size())
		{
			// return the error code for an empty simulation
			return 0;
		}
		else
		{
			// return the first fish's Key
			return m_fishBodyPtrs.begin()->first;
		}
	}

	// move to the next fish
	++fishBody_key_pair_Iter;

	// check for wrapping
	if (m_fishBodyPtrs.end() == fishBody_key_pair_Iter)
	{		
		fishBody_key_pair_Iter = m_fishBodyPtrs.begin();
	}

	// return the next fish's Key
	return fishBody_key_pair_Iter->first;
}

////////////////////////////////////////////////////////////////////////////////
//  FUNCTION:       
/*! \brief          Vertically projects a point onto the HeightMap and returns
**                  the resulting height.
**  \param          pointToProject a position that is projected onto the HeightMap.
**  \return         The height of the resulting projection (its Y value).
**  \exception      none
**  \author         Austin Spafford
**  \date           3/25/2003 8:46:25 PM
**  \warning        none
*/
////////////////////////////////////////////////////////////////////////////////
float TankSimulator::GetProjectedHeightOnHeightMap(const Vector3<float> & pointToProject)
{
	return m_heightMap.GetProjectedHeight(pointToProject);
}

////////////////////////////////////////////////////////////////////////////////
//  FUNCTION:       
/*! \brief          sets the ultimate boundaries for all mobile objects, make
**                  sure the bounding box has a non-zero volume!
**  \param          cornerOne a three dimensional vector representing one corner
**                  of the new boundry box.
**  \param          cornerTwo a three dimensional vector representing one corner
**                  of the new boundry box.
**  \return         none
**  \exception      none
**  \author         Austin Spafford
**  \date           11/13/2002 8:40:14 PM
**  \warning        none
*/
////////////////////////////////////////////////////////////////////////////////
void TankSimulator::SetBoundary(math::linear_alg::Vector3<float> cornerOne, math::linear_alg::Vector3<float> cornerTwo)
{
	// make sure the bounding box has a non-zero volume
	assert(cornerOne[0] != cornerTwo[0]);
	assert(cornerOne[1] != cornerTwo[1]);
	assert(cornerOne[2] != cornerTwo[2]);

	// find the maxs and mins of the specified bounding box
	for (int i=0; i<3; i++)
	{
		if (cornerOne[i] < cornerTwo[i])
		{
			m_boundaryMins[i] = cornerOne[i];
			m_boundaryMaxs[i] = cornerTwo[i];
		}
		else
		{
			m_boundaryMins[i] = cornerTwo[i];
			m_boundaryMaxs[i] = cornerOne[i];
		}
	}

	// update the physics simulator's boundaries as well
	m_physicsSimulator.SetBoundary(m_boundaryMins, m_boundaryMaxs);

	// resize the heightmap to the size of the simulation
	m_heightMap.SetFootPrint(Vector2<float>(m_boundaryMaxs[0] - m_boundaryMins[0], m_boundaryMaxs[2] - m_boundaryMins[2]));
	m_heightMap.VerticallyResizeEntireHeightMap(0, (m_boundaryMaxs[1] - m_boundaryMins[1]) / 2.0f);

	// move the heightmap such that's on the bottom of the simulation
	m_heightMap.SetPositionOffset(m_boundaryMins);

	return;
}

////////////////////////////////////////////////////////////////////////////////
//  FUNCTION:       
/*! \brief          sets the number of voxels that lie along each axis
**  \param          newVoxelCount a vector representing the number of voxels that
**                  lie along each axis. Make sure to make each dimension contain
**                  at least one voxel!
**  \exception      none
**  \author         Austin Spafford
**  \date           11/14/2002 2:00:08 PM
**  \warning        none
*/
////////////////////////////////////////////////////////////////////////////////
void TankSimulator::SetPlanktonVoxelCount(math::linear_alg::Vector3<int> newVoxelCount)
{
	assert(newVoxelCount == LimitedToMinimum(newVoxelCount, Vector3<int>(1,1,1)));

	m_planktonDensityVoxelCount = newVoxelCount;

	ReallocatePlanktonVoxels();

	return;
}

////////////////////////////////////////////////////////////////////////////////
//  FUNCTION:       
/*! \brief          sets the plankton density of a specific voxel
**  \param          voxelIndex the axis-aligned index of the voxel that is to 
**                  have its value set. Make sure the index is legal!
**  \param          value the new value for the voxel
**  \exception      none
**  \author         Austin Spafford
**  \date           11/14/2002 2:02:27 PM
**  \warning        none
*/
////////////////////////////////////////////////////////////////////////////////
void TankSimulator::SetPlanktonDensity(math::linear_alg::Vector3<int> voxelIndex, float value)
{
	assert(voxelIndex == LimitedToMinMax(voxelIndex, Vector3<int>(0,0,0), Vector3<int>(m_planktonDensityVoxelCount[0]-1, m_planktonDensityVoxelCount[1]-1, m_planktonDensityVoxelCount[2] - 1)));

	int rowPitch = m_planktonDensityVoxelCount[0];
	int planePitch = m_planktonDensityVoxelCount[0] * m_planktonDensityVoxelCount[1];

	float * currentVoxelPtr = 
		m_planktonDensityBufferPtr
		+
		voxelIndex[0]
		+
		voxelIndex[1] * rowPitch
		+
		voxelIndex[2] * planePitch;

	currentVoxelPtr[0] = value;

	return;
}

////////////////////////////////////////////////////////////////////////////////
//  FUNCTION:       
/*! \brief          sets the number of chum voxels that lie along each axis
**  \param          newVoxelCount a vector representing the number of voxels that
**                  lie along each axis. Make sure to make each dimension contain
**                  at least one voxel!
**  \return         none
**  \exception      none
**  \par            DESCRIPTION:
**  \author         Austin Spafford
**  \date           1/18/2003 3:27:10 PM
**  \warning        none
*/
////////////////////////////////////////////////////////////////////////////////
void TankSimulator::SetChumVoxelCount(Vector3<int> newVoxelCount)
{
	assert(newVoxelCount == LimitedToMinimum(newVoxelCount, Vector3<int>(1,1,1)));

	m_chumVoxelCount = newVoxelCount;

	ReallocateChumVoxels();

	return;
}

////////////////////////////////////////////////////////////////////////////////
//  FUNCTION:       
/*! \brief          sets the chum position and density of a specific voxel
**  \param          voxelIndex the axis-aligned index of the voxel that is to 
**                  have its value set. Make sure the index is legal!
**  \param          chumPosition the new position of the chum chunk in the voxel
**  \param          density the new density value of the chum voxel
**  \exception      none
**  \author         Austin Spafford
**  \date           11/14/2002 2:02:27 PM
**  \warning        none
*/
////////////////////////////////////////////////////////////////////////////////
void TankSimulator::SetChumVoxelValues(Vector3<int> voxelIndex, Vector3<float> chumPosition, float density)
{
	assert(voxelIndex == LimitedToMinMax(voxelIndex, Vector3<int>(0,0,0), Vector3<int>(m_chumVoxelCount[0]-1, m_chumVoxelCount[1]-1, m_chumVoxelCount[2] - 1)));

	int rowPitch = m_chumVoxelCount[0];
	int planePitch = m_chumVoxelCount[0] * m_chumVoxelCount[1];

	ChumVoxel * currentVoxelPtr = 
		m_chumVoxelBufferPtr
		+
		voxelIndex[0]
		+
		voxelIndex[1] * rowPitch
		+
		voxelIndex[2] * planePitch;

	currentVoxelPtr->chumPosition = chumPosition;
	currentVoxelPtr->chumDensity = density;

	return;
}

////////////////////////////////////////////////////////////////////////////////
//  FUNCTION:       
/*! \brief          sets the specified fish body's actions to the provided action vector
**  \param          fishKey a Key to a FishBody
**  \param          fishActions a vector of actions that the FishBody must process
**  \return         returns false if the specified FishBody does not exist
**  \exception      none
**  \author         Austin Spafford
**  \date           11/25/2002 4:41:52 AM
**  \warning        none
*/
////////////////////////////////////////////////////////////////////////////////
bool TankSimulator::SetFishBodyActions(Key fishKey, const vector<AI_Action> & fishActions)
{
	FishBody * fishBodyPtr = GetFishBodyPtr(fishKey);

	if (NULL == fishBodyPtr)
	{
		return false;
	}

	fishBodyPtr->GiveAICommand(fishActions);

	return true;
}

////////////////////////////////////////////////////////////////////////////////
//  FUNCTION:       
/*! \brief          checks to see if a FishBody exists in the simulation that uses
**                  the specified Key as a label
**  \param          fishKey a Key to a FishBody
**  \return         returns true if a FishBody is stored that uses the specified
**                  Key, false otherwise
**  \exception      none
**  \author         Austin Spafford
**  \date           10/20/2002 8:40:14 PM
**  \warning        none
*/
////////////////////////////////////////////////////////////////////////////////
bool TankSimulator::DoesFishBodyExist(Key fishKey)
{
	if (m_fishBodyPtrs.end() == m_fishBodyPtrs.find(fishKey)) 
	{
		return false;
	}

	return true;
}

////////////////////////////////////////////////////////////////////////////////
//  FUNCTION:       
/*! \brief          steps the simulation by "timeStep" number of seconds, and
**                  draws all objects in the simulation, call every frame
**  \param          timeStep describes the number of seconds that have elapsed
**                  since the simulation was last stepped. Should be between 0
**                  and 1.
**  \return         none
**  \exception      none
**  \par            DESCRIPTION:
**                    It tastes like chicken.
**  \author         Austin Spafford
**  \date           10/31/2002 7:35:09 PM
**  \warning        none
*/
////////////////////////////////////////////////////////////////////////////////
void TankSimulator::StepSimulation(float timeStep)
{
	// CHECK THE GAME STATE

	// step the fish
	hash_map<MsgHandler::Key, InterpretationModule::FishBody*>::iterator fishBody_key_pair_Iter;
	fishBody_key_pair_Iter = m_fishBodyPtrs.begin();
	while (fishBody_key_pair_Iter != m_fishBodyPtrs.end())
	{
		fishBody_key_pair_Iter->second->Update(timeStep);		
   
		++fishBody_key_pair_Iter;
	}

	// add plankton to the simulation
	vector<PlanktonNode>::iterator planktonNodeIter;
	planktonNodeIter = m_planktonNodes.begin();
	while (planktonNodeIter != m_planktonNodes.end())
	{
		// find out which voxel the plankton node is in.
		Vector3<int> voxelIndex = GetContainingPlanktonVoxelIndex(planktonNodeIter->position);
		voxelIndex = LimitedToMinMax(voxelIndex, Vector3<int>(0,0,0), Vector3<int>(m_planktonDensityVoxelCount[0]-1, m_planktonDensityVoxelCount[1]-1, m_planktonDensityVoxelCount[2] - 1));

		// add some plankton to the voxel
		float planktonDensity = GetPlanktonDensityFromVoxel(voxelIndex);		
		planktonDensity += 100 * planktonNodeIter->nodeStrength * timeStep;
		SetPlanktonDensity(voxelIndex, planktonDensity);
	
		++planktonNodeIter;	
	}

	// step the plankton densities
//	BlurPlanktonDensities(timeStep);

	// step the physics simulation
	m_physicsSimulator.StepSimulation(timeStep);	

	return;
}

////////////////////////////////////////////////////////////////////////////////
//  FUNCTION:       
/*! \brief          draws the contents of the simulation
**  \return         none
**  \exception      none
**  \author         Austin Spafford
**  \date           11/25/2002 8:34:18 AM
**  \warning        none
*/
////////////////////////////////////////////////////////////////////////////////
void TankSimulator::Draw(void)
{
	// draw the height map
	m_heightMap.SetRender();

    // draw the generic tank simulator data
    m_gfxClient.SetRender();

	// call the fish body drawing functions
	hash_map<MsgHandler::Key, InterpretationModule::FishBody*>::iterator fishBody_key_pair_Iter;
	fishBody_key_pair_Iter = m_fishBodyPtrs.begin();
	while (fishBody_key_pair_Iter != m_fishBodyPtrs.end())		
	{
		fishBody_key_pair_Iter->second->QueueDraw();
   
		++fishBody_key_pair_Iter;
	}

	return;
}

////////////////////////////////////////////////////////////////////////////////
//  FUNCTION:       
/*! \brief          averages the values of the plankton voxels
**  \param          timeStep describes the number of seconds that have elapsed
**                  since the simulation was last stepped. Should be between 0
**                  and 1.
**  \return         none
**  \exception      none
**  \author         Austin Spafford
**  \date           11/13/2002 7:27:06 PM
**  \warning        none
*/
////////////////////////////////////////////////////////////////////////////////
void TankSimulator::BlurPlanktonDensities(float timeStep)
{
	// take into account that time that has elapsed
	m_planktonUpkeepCounter += 20.0f * timeStep;

	// throw away plankton compution frame counts that are dangerously high (terminal slow down problems)
	if (m_planktonUpkeepCounter > 3)
	{
		m_planktonUpkeepCounter = 3;
	}

	// compute all required frames, most likely just one
	while (m_planktonUpkeepCounter > 1)
	{
		m_planktonUpkeepCounter -= 1;

		Vector3<int> currentVoxelIndex;

		int rowPitch = m_planktonDensityVoxelCount[0];
		int planePitch = m_planktonDensityVoxelCount[0] * m_planktonDensityVoxelCount[1];

		float * currentVoxelPtr;

		for (
			currentVoxelIndex[0] = 1;
			currentVoxelIndex[0] < m_planktonDensityVoxelCount[0] - 1;
			currentVoxelIndex[0]++)
		{
			for (
				currentVoxelIndex[1] = 1;
				currentVoxelIndex[1] < m_planktonDensityVoxelCount[1] - 1;
				currentVoxelIndex[1]++)
			{
				for (
					currentVoxelIndex[2] = 1;
					currentVoxelIndex[2] < m_planktonDensityVoxelCount[2] - 1;
					currentVoxelIndex[2]++)
				{
					currentVoxelPtr = 
						m_planktonDensityBufferPtr
						+
						currentVoxelIndex[0]
						+
						currentVoxelIndex[1] * rowPitch
						+
						currentVoxelIndex[2] * planePitch;

					// average the voxel's value with it neighbors
					currentVoxelPtr[0] = 
						24 * currentVoxelPtr[0]
						+
						currentVoxelPtr[-1] // x - 1
						+
						currentVoxelPtr[1] // x + 1
						+
						currentVoxelPtr[-rowPitch] // y - 1
						+
						currentVoxelPtr[rowPitch] // y + 1
						+
						currentVoxelPtr[-planePitch] // z - 1
						+
						currentVoxelPtr[planePitch]; // z + 1						

					currentVoxelPtr[0] /= 30;

					currentVoxelPtr[0] -= 0.001f; // clean up the voxel a bit

					if (currentVoxelPtr[0] < 0)
					{
						currentVoxelPtr[0] = 0;
					}
				}
			}
		}

	} // end while frames still need to computed
	
	return;
}
	
////////////////////////////////////////////////////////////////////////////////
//  FUNCTION:       
/*! \brief          reallocates all dynamic memory associated with the plankton
**                  voxels, uses m_planktonDensityVoxelCount to figure out how
**                  much memory must be allocated.
**  \return         none
**  \exception      none
**  \author         Austin Spafford
**  \date           11/13/2002 7:27:53 PM
**  \warning        none
*/
////////////////////////////////////////////////////////////////////////////////
void TankSimulator::ReallocatePlanktonVoxels(void)
{
	if (NULL != m_planktonDensityBufferPtr)
	{
		delete [] m_planktonDensityBufferPtr;
		m_planktonDensityBufferPtr = NULL;
	}

	int voxelCount = 
			m_planktonDensityVoxelCount[0]
			*
			m_planktonDensityVoxelCount[1]
			*
			m_planktonDensityVoxelCount[2];

	m_planktonDensityBufferPtr = new float [voxelCount];

	if (NULL == m_planktonDensityBufferPtr)
	{
		throw Nexception("Could not allocate memory for the plankton density voxels.");
	}

	for (int i=0; i<voxelCount; i++)
	{
		m_planktonDensityBufferPtr[i] = 0;
	}

	return;
}

////////////////////////////////////////////////////////////////////////////////
//  FUNCTION:       
/*! \brief          reallocates all dynamic memory associated with the chum
**                  voxels, uses m_chumVoxelCount to figure out how
**                  much memory must be allocated.
**  \return         none
**  \exception      none
**  \author         Austin Spafford
**  \date           1/17/2003 7:27:53 PM
**  \warning        none
*/
////////////////////////////////////////////////////////////////////////////////
void TankSimulator::ReallocateChumVoxels(void)
{
	if (NULL != m_chumVoxelBufferPtr)
	{
		delete [] m_chumVoxelBufferPtr;
		m_chumVoxelBufferPtr = NULL;
	}

	int voxelCount = 
			m_chumVoxelCount[0]
			*
			m_chumVoxelCount[1]
			*
			m_chumVoxelCount[2];

	m_chumVoxelBufferPtr = new ChumVoxel [voxelCount];

	if (NULL == m_chumVoxelBufferPtr)
	{
		throw Nexception("Could not allocate memory for the chum voxels.");
	}

	for (int i=0; i<voxelCount; i++)
	{
		m_chumVoxelBufferPtr[i].chumDensity = 0;
	}

	return;
}

////////////////////////////////////////////////////////////////////////////////
//  FUNCTION:       
/*! \brief          retrieves the index of a voxel that contains (or is closest
**                  to containing) the specified point
**  \param          position a vector representing the point that is checked for
**                  for containment within each voxel.
**  \return         the index of the containing (or closest to containing) voxel
**  \exception      none
**  \author         Austin Spafford
**  \date           11/23/2002 3:31:20 AM
**  \warning        none
*/
////////////////////////////////////////////////////////////////////////////////
Vector3<int> TankSimulator::GetContainingPlanktonVoxelIndex(Vector3<float> position)
{	
	// translate the point such that it is relative to the tank
	position -= m_boundaryMins;

	// scale the position such that it is terms of a unit cube
	Vector3<float> boundingBoxDimensions = m_boundaryMaxs - m_boundaryMins;
	position[0] /= boundingBoxDimensions[0];
	position[1] /= boundingBoxDimensions[1];
	position[2] /= boundingBoxDimensions[2];

	// scale the position such that it is terms of a voxel index	
	position[0] *= m_planktonDensityVoxelCount[0];
	position[1] *= m_planktonDensityVoxelCount[1];
	position[2] *= m_planktonDensityVoxelCount[2];

	// round the position down so that it represents the index of the containing voxel
	Vector3<int> voxelIndex;
	voxelIndex[0] = int(position[0]);
	voxelIndex[1] = int(position[1]);
	voxelIndex[2] = int(position[2]);

	// make sure to correct any purely theoretical indices to represent correct indice approximations
	for (int i=0; i<3; i++)
	{
		if (voxelIndex[i] < 0)
		{
			voxelIndex[i] = 0;
		}
		if (voxelIndex[i] >= m_planktonDensityVoxelCount[i])
		{
			voxelIndex[i] = m_planktonDensityVoxelCount[i] - 1;
		}
	}

	return voxelIndex;
}
