////////////////////////////////////////////////////////////////////////////////
/*! \file           TankSimulator.h
**  \brief          TankSimulator object interface file
**  \author         Austin Spafford
**  \date           10/29/2002 4:20:38 PM
**  \par            DESCRIPTION:
**                    The TankSimulator object stores a collection of objects and
**                    provides an interface that allows them to interact with each
**                    other. It also simulates a volumetric plankton ecology.
**  \par            REMARKS:
*/
////////////////////////////////////////////////////////////////////////////////
#pragma once
////////////////////////////////////////////////////////////////////////////////
#include <hash_map>
#include <list>
#include <vector>
#include "DrawableHeightMap.h"
#include "math/math.h"
#include "physics/PhysicsSimulator.h"
#include "messageHandler/MessageHandler.h"
#include "TankSimulatorClient.h"
#include "tank_sim_client_gfx.h"
////////////////////////////////////////////////////////////////////////////////
namespace AI_System { struct AI_Action; }
namespace InterpretationModule { class FishBody; }
namespace UDPNet { class NetMessage; }
class FileMemento;
////////////////////////////////////////////////////////////////////////////////


////////////////////////////////////////////////////////////////////////////////
/*! \class          TankSimulator
**  \brief          The TankSimulator object stores a collection of objects and
**                  provides an interface that allows them to interact with each
**                  other. It also simulates a volumetric plankton ecology.
**  \author         Austin Spafford
**  \date           10/29/2002 9:27:08 PM
**  \warning        none
*/
////////////////////////////////////////////////////////////////////////////////
class TankSimulator
{	
	public:

	////////////////////////////////////////////////////////////////////////////
	////////////////////////////////////////////////////////////////////////////
	////////////////////////////////////////////////////////////////////////////

	/// default constructer
	TankSimulator();	

	/// deconstructer
	~TankSimulator();

	////////////////////////////////////////////////////////////////////////////
	////////////////////////////////////////////////////////////////////////////
	////////////////////////////////////////////////////////////////////////////

	/// adds a FishBody to the simulation and assigns the specified Key to it
	/// you can then format the fish body by calling FormatFishBody(), returns
	/// false if instantiation failed
	bool AddFishBody(MsgHandler::Key keyForObject);

	/// adds a plankton node to the simulation
	/// NOTE: THIS IS A TEMPORARY INTERFACE
	void AddPlanktonNode(math::linear_alg::Vector3<float> position, float nodeStrength);

	////////////////////////////////////////////////////////////////////////////
	////////////////////////////////////////////////////////////////////////////
	////////////////////////////////////////////////////////////////////////////

	/// removes a FishBody from the simulation, returns false if the specified fish
	/// does not exist in the simulation
	bool RemoveFishBody(MsgHandler::Key keyForObject);

    ////////////////////////////////////////////////////////////////////////////
	////////////////////////////////////////////////////////////////////////////
	////////////////////////////////////////////////////////////////////////////

	/// loads the specified image file and maps its red channel to the height values of the height map
	/// returns true if the image was properly imported, false otherwise
	/// IMPORTANT: Also changes the resolution of the height map to that of the image.
	bool ImportHeightMapFromImage(char * fileName, float maxHeight);

	/// saves a gray scale image of the height map, with white being the highest elevation
	/// returns true if the image was properly exported, false otherwise
	/// IMPORTANT: The reolution of the image will be that of the height map
	bool ExportHeightMapToImage(char * fileName);

	////////////////////////////////////////////////////////////////////////////
	////////////////////////////////////////////////////////////////////////////
	////////////////////////////////////////////////////////////////////////////

	/// updates the state of the simulation by means of a NetMessage that
	/// originated on a sister TankSimulation by means of a call to GenerateNetMessage()
	/// NOTE: this will load the heightmap if it is present
	void LoadNetMessage(UDPNet::NetMessage & netMessageRef);

	/// updates the state of the heightmap using a netmessage
	void LoadHeightMapNetMessage(UDPNet::NetMessage & netMessageRef);

	/// generates a NetMessage that will encode the current state of the
	/// simulation, so that a sister TankSimulation can load it by means of a
	/// call to LoadNetMessage(). The generated NetMessage will most likely
	/// be missing information depending on what information has changed since
	/// the last call to GenerateNetMessage.
	void GeneratePartialNetMessage(UDPNet::NetMessage & netMessageRef);

	/// generates a NetMessage that will encode the current state of the
	/// simulation, so that a sister TankSimulation can load it by means of a
	/// call to LoadNetMessage(). This version of GenerateNetMessage() will
	/// always include every piece of data regardless of what has recently changed.
	void GenerateCompleteNetMessage(UDPNet::NetMessage & netMessageRef);

	/// generates a NetMessage that can ONLY be loaded by LoadHeightMapNetMessage()
	void GenerateHeightMapNetMessage(UDPNet::NetMessage & netMessageRef);

	////////////////////////////////////////////////////////////////////////////
	////////////////////////////////////////////////////////////////////////////
	////////////////////////////////////////////////////////////////////////////

	// a factory method that generates an object that encapsulates the state
	// of the simulation
	FileMemento * CreateFileMemento(void) const;

	// sets the current state of the simulation to that which is stored in fileMemento
	// NOTE: fileMemento MUST actually reference a TankSimulatorFileMemento!
	void LoadFileMemento(const FileMemento & fileMemento);

	////////////////////////////////////////////////////////////////////////////
	////////////////////////////////////////////////////////////////////////////
	////////////////////////////////////////////////////////////////////////////

	/// Traces a ray against every fish in the simulation, and returns by 
	/// reference a key if a fish was hit. Returns true if a fish was hit,
	/// else it will return false if no fish were hit, and will also set fishKeyRef to 0
	bool TraceRayAgainstAllFish(
			const math::linear_alg::Vector3<float> & rayOrigin,
			const math::linear_alg::Vector3<float> & rayDirection,
			MsgHandler::Key & fishKeyRef
			);

	////////////////////////////////////////////////////////////////////////////
	////////////////////////////////////////////////////////////////////////////
	////////////////////////////////////////////////////////////////////////////

	/// fills in minValues and maxValues with the current boundary limits
	void GetBoundary(math::linear_alg::Vector3<float> & minValues, math::linear_alg::Vector3<float> & maxValues);	

	/// retrieves a pointer to the requested FishBody, returns NULL if no such body exists
	InterpretationModule::FishBody * GetFishBodyPtr(MsgHandler::Key fishKey);

	/// retrieves the number of plankton voxels that lie along each axis
	math::linear_alg::Vector3<int> GetPlanktonVoxelCount(void);

	/// retrieves the value of a specific plankton voxel
	float GetPlanktonDensityFromVoxel(math::linear_alg::Vector3<int> voxelIndex);

	/// retrieves the value of a specific position within plankton ecology
	float GetPlanktonDensityAtLocation(const math::linear_alg::Vector3<float> & position);
	
	/// retrieves the number of chum voxels that lie along each axis
	math::linear_alg::Vector3<int> GetChumVoxelCount(void);

	/// retrieves the values of a specific chum voxel
	void GetChumVoxelValues(math::linear_alg::Vector3<int> voxelIndex, math::linear_alg::Vector3<float> & chumPositionRef, float & chumDensityRef);

	/// retrieves a reference to the mating fish list
	std::list<MsgHandler::Key> & GetMatingFishListRef(void);

	/// vertically projects a point onto the HeightMap and returns the projected
	/// point's Y value
	float GetProjectedHeightOnHeightMap(const math::linear_alg::Vector3<float> & pointToProject);

	// FARK: GetAddFishListRef(void)

	/// retrieves a reference to the delete fish list
	std::list<MsgHandler::Key> & GetDeleteFishListRef(void);

	/// retrieves a vector of Keys to fish that are currently visible to the specified fish
	void GetVectorOfFishVisibleToFish(MsgHandler::Key fishKey, std::vector<MsgHandler::Key>& returnValue);

	/// retrieves the key to the next fish in the simulation (arbitrary order, wraps), returns the first fish's Key if the specfied fish does not exist, and 0 if no fish are in the simulation
	MsgHandler::Key GetNextFish(MsgHandler::Key fishKey);

	/// retrieves the total number of live and dead fish in the system
	unsigned int GetNumberOfFish(void) { return m_fishBodyPtrs.size(); }

	////////////////////////////////////////////////////////////////////////////
	////////////////////////////////////////////////////////////////////////////
	////////////////////////////////////////////////////////////////////////////

	/// sets the ultimate boundaries for all mobile objects, make sure the 
	/// bounding box has a non-zero volume!
	void SetBoundary(math::linear_alg::Vector3<float> cornerOne, math::linear_alg::Vector3<float> cornerTwo);

	/// sets the number of plankton voxels that lie along each axis
	void SetPlanktonVoxelCount(math::linear_alg::Vector3<int> newVoxelCount);

	/// sets the plankton density of a specific voxel
	void SetPlanktonDensity(math::linear_alg::Vector3<int> voxelIndex, float value);

	/// sets the number of chum voxels that lie along each axis
	void SetChumVoxelCount(math::linear_alg::Vector3<int> newVoxelCount);

	/// sets the chum position and density of a specific voxel
	void SetChumVoxelValues(math::linear_alg::Vector3<int> voxelIndex, math::linear_alg::Vector3<float> chumPosition, float density);

	/// sets the specified fish body's actions to the provided action vector
	bool SetFishBodyActions(MsgHandler::Key fishKey, const std::vector<AI_System::AI_Action> & fishActions);

	////////////////////////////////////////////////////////////////////////////
	////////////////////////////////////////////////////////////////////////////
	////////////////////////////////////////////////////////////////////////////

	/// returns true if a FishBody is stored that uses the specified Key to identify itself
	bool DoesFishBodyExist(MsgHandler::Key keyForObject);

	////////////////////////////////////////////////////////////////////////////
	////////////////////////////////////////////////////////////////////////////
	////////////////////////////////////////////////////////////////////////////	

	/// steps the simulation by "timeStep" number of seconds, and 
	/// draws all objects in the simulation, call every frame
	void StepSimulation(float timeStep);

	////////////////////////////////////////////////////////////////////////////
	////////////////////////////////////////////////////////////////////////////
	////////////////////////////////////////////////////////////////////////////

	/// draws the contents of the simulation
	void Draw(void);

	////////////////////////////////////////////////////////////////////////////
	////////////////////////////////////////////////////////////////////////////
	////////////////////////////////////////////////////////////////////////////	

	private:

	friend TankSimulatorClient;

	struct PlanktonNode
	{		
		/// the position of the plankton node
		math::linear_alg::Vector3<float> position;
		/// the output strength of the plankton node, 1.0 should be a reasonable value
		float nodeStrength;
	};

	struct ChumVoxel
	{
		/// the position of the piece of chum contained in the voxel (is in global coordinates)
		math::linear_alg::Vector3<float> chumPosition;
		/// the amount of chum stored in the voxel
		float chumDensity;
	};

	/// averages the values of the plankton voxels
	void BlurPlanktonDensities(float timeStep);

	/// adjusts the actual buffer to what the current dimensions demand
	void ReallocatePlanktonVoxels(void);
	void ReallocateChumVoxels(void);	

	/// returns the index of a voxel that contains the specified point
	math::linear_alg::Vector3<int> GetContainingPlanktonVoxelIndex(math::linear_alg::Vector3<float> position);	

	////////////////////////////////////////////////////////////////////////////
	////////////////////////////////////////////////////////////////////////////
	////////////////////////////////////////////////////////////////////////////

	std::hash_map<MsgHandler::Key, InterpretationModule::FishBody*> m_fishBodyPtrs;

	PhysicsSimulator m_physicsSimulator;

	DrawableHeightMap m_heightMap;
	
	float m_planktonUpkeepCounter; /// a plankton distribution frame will be computed every time this value wraps over 1.0
	math::linear_alg::Vector3<int> m_planktonDensityVoxelCount;
	float * m_planktonDensityBufferPtr;

	std::vector<PlanktonNode> m_planktonNodes;

	float m_chumUpkeepCounter; /// a chum decay frame will be computed every time this value wraps over 1.0
	math::linear_alg::Vector3<int> m_chumVoxelCount;
	ChumVoxel * m_chumVoxelBufferPtr;

	math::linear_alg::Vector3<float> m_boundaryMins;
	math::linear_alg::Vector3<float> m_boundaryMaxs;

	std::list<MsgHandler::Key> m_matingFishList;
	// FARK: fish body/brain pair list for adding fish
	std::list<MsgHandler::Key> m_deleteFishList;	

    /// tank simulator graphics client
    TankSimClientGfx m_gfxClient;
};
