////////////////////////////////////////////////////////////////////////////////
/*! \file           TankSimulatorClient.cpp
**  \brief          TankSimulatorClient implementation file
**  \author         Austin Spafford
**  \date           11/11/2002 6:11:04 PM
**  \par            DESCRIPTION:
**                    The TankSimulatorClient object provides an interface for
**                    stored objects that allow them to interact with other
**                    members of the simulation
**  \par            REMARKS:
**                    
*/
////////////////////////////////////////////////////////////////////////////////
#pragma warning (disable : 4786)
////////////////////////////////////////////////////////////////////////////////
#include "math/math.h"
#include "tankSimulator/TankSimulator.h"
#include "tankSimulator/TankSimulatorClient.h"
#include "interpretationModule/fishBody.hpp"
#include "ai_core/AI_System.h"
#include "interpretationModule/commandPacket.hpp"
// memory leak detection
#ifdef _DEBUG
    #include"memleak.h"
    #define new DEBUG_NEW
    #undef THIS_FILE
    static char THIS_FILE[] = __FILE__;
#endif
////////////////////////////////////////////////////////////////////////////////
using namespace math::linear_alg;
using namespace std;
using namespace InterpretationModule;
using namespace MsgHandler;
using namespace AI_System;
////////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////
//  FUNCTION:       
/*! \brief          constructer, must use this one so that the client knows its
**                  parent
**  \return         none
**  \exception      none
**  \author         Austin Spafford
**  \date           11/1/2002 8:40:14 PM
**  \warning        none
*/
////////////////////////////////////////////////////////////////////////////////
TankSimulatorClient::TankSimulatorClient(TankSimulator &newParentTankSimulatorRef) :
		m_parentTankSimulatorRef(newParentTankSimulatorRef)
{	 
	return;
}

////////////////////////////////////////////////////////////////////////////////
//  FUNCTION:       
/*! \brief          deconstructer
**  \return         none
**  \exception      none
**  \author         Austin Spafford
**  \date           10/30/2002 8:40:14 PM
**  \warning        none
*/
////////////////////////////////////////////////////////////////////////////////
TankSimulatorClient::~TankSimulatorClient()
{
	return;
}

////////////////////////////////////////////////////////////////////////////////
//  FUNCTION:       
/*! \brief          adds a fish to the mating list
**  \param          fishKey a key to add to the mating fish list
**  \return         none
**  \exception      none
**  \author         Austin Spafford
**  \date           11/25/2002 8:48:52 AM
**  \warning        none
*/
////////////////////////////////////////////////////////////////////////////////
void TankSimulatorClient::AddFishToMatingList(Key fishKey)
{
	m_parentTankSimulatorRef.m_matingFishList.push_back(fishKey);
}

////////////////////////////////////////////////////////////////////////////////
//  FUNCTION:       
/*! \brief          adds a fish to the delete fish list
**  \param          fishKey a key to add to the delete fish list
**  \return         none
**  \exception      none
**  \author         Austin Spafford
**  \date           11/25/2002 8:46:52 AM
**  \warning        none
*/
////////////////////////////////////////////////////////////////////////////////
void TankSimulatorClient::AddFishToDeleteFishList(Key fishKey)
{
	m_parentTankSimulatorRef.m_deleteFishList.push_back(fishKey);
}

////////////////////////////////////////////////////////////////////////////////
//  FUNCTION:       
/*! \brief          retrieves a reference to the physics simulator
**  \return         a reference to the physics simulator
**  \exception      none
**  \author         Austin Spafford
**  \date           11/4/2002 2:20:24 PM
**  \warning        none
*/
////////////////////////////////////////////////////////////////////////////////
PhysicsSimulator & TankSimulatorClient::GetPhysicsSimulatorReference(void)
{
	return m_parentTankSimulatorRef.m_physicsSimulator;
}

////////////////////////////////////////////////////////////////////////////////
//  FUNCTION:       
/*! \brief          retrieves a pointer to the specified FishBody
**  \return         a pointer to the FishBody, will be NULL if the requested 
**                  FishBody does not exist
**  \exception      none
**  \author         Austin Spafford
**  \date           11/4/2002 2:20:24 PM
**  \warning        none
*/
////////////////////////////////////////////////////////////////////////////////
FishBody * TankSimulatorClient::GetFishBodyPtr(Key keyForObject)
{
	return m_parentTankSimulatorRef.GetFishBodyPtr(keyForObject);
}

////////////////////////////////////////////////////////////////////////////////
//  FUNCTION:       
/*! \brief          Retrieves the boundary values of the simulation.
**  \param          minValuesRef a reference to a vector that willl store the
**                  minimum values of the simulation's boundary.
**  \param          maxValuesRef a reference to a vector that willl store the
**                  maximum values of the simulation's boundary.
**  \return         none
**  \exception      none
**  \author         Austin Spafford
**  \date           2/14/2003 1:29:46 PM
**  \warning        none
*/
////////////////////////////////////////////////////////////////////////////////
void TankSimulatorClient::GetBoundary(
		Vector3<float> & minValuesRef,
		Vector3<float> & maxValuesRef
		)
{
	m_parentTankSimulatorRef.GetBoundary(minValuesRef, maxValuesRef);

	return;
}

////////////////////////////////////////////////////////////////////////////////
//  FUNCTION:       
/*! \brief          retrieves the value of the voxel that contains (or is 
**                  closest to containing) the specified point
**  \param          position a vector representing a point that will be used for
**                  selecting the voxel to sample the density value from
**  \return         the value stored in the voxel that contains (or is closest
**                  to containing) the specified point
**  \exception      none
**  \author         Austin Spafford
**  \date           11/26/2002 1:50:59 PM
**  \warning        none
*/
////////////////////////////////////////////////////////////////////////////////
float TankSimulatorClient::GetPlanktonDensity(Vector3<float> position)
{
	return m_parentTankSimulatorRef.GetPlanktonDensityAtLocation(position);
}

////////////////////////////////////////////////////////////////////////////////
//  FUNCTION:       
/*! \brief          retrieves the position of the closest chunk of chum to a 
**                  reference point, returns false if no nearby chum could be found
**  \param          referencePoint the point that will be used when finding the
**                  the closest chunk of chum
**  \param          chumPositionRef a reference to the location at which the 
**                  closest chunk of chum was found
**  \return         true if a nearby chunk of chum was found, false otherwise
**  \exception      none
**  \par            DESCRIPTION:
**  \author         Austin Spafford
**  \date           1/17/2003 9:52:24 PM
**  \warning        none
*/
////////////////////////////////////////////////////////////////////////////////
bool TankSimulatorClient::GetClosestChumPosition(
		Vector3<float> referencePoint,
		Vector3<float> & chumPositionRef
		)
{
	// find the voxel containing the reference point
	Vector3<int> voxelIndex = m_parentTankSimulatorRef.GetContainingPlanktonVoxelIndex(referencePoint);

	// find out how much chum is left in the voxel containing the reference point
	Vector3<float> chumPosition;
	float chumDensity;
	m_parentTankSimulatorRef.GetChumVoxelValues(voxelIndex, chumPosition, chumDensity);

	// FARK: check neighboring voxels instead of just the containing voxel

	// check to see if there is a chum chunk nearby
	if (chumDensity > 0)
	{
		// return the position of the chum chunk
		chumPositionRef = chumPosition;
		return true;
	}
	else
	{
		// report that no chum chunks were nearby
		return false;
	}	
}

////////////////////////////////////////////////////////////////////////////////
//  FUNCTION:       
/*! \brief          Returns the Key of the fish that is closest to the reference
**                  fish, returns 0 if either the specified fish is the only fish
**                  in the simulation, or if it doesn't exist
**  \param          referenceFishKey the key of the fish that's being used as a
**                  reference point
**  \return         the key of the fish that is closest to the reference fish
**  \exception      none
**  \author         Austin Spafford
**  \date           2/18/2003 2:10:16 PM
**  \warning        none
*/
////////////////////////////////////////////////////////////////////////////////
Key TankSimulatorClient::GetKeyForClosestFish(const Key & referenceFishKey)
{
	// check to see if the reference fish is valid
	FishBody * referenceFishPtr = m_parentTankSimulatorRef.GetFishBodyPtr(referenceFishKey);
	if (NULL == referenceFishPtr)
	{
		return 0;
	}	

	// otherwise we know that there are at least two fish in the simulation, so
	// lets go find the closest fish
	Key keyOfClosestFish = 0;
	float distanceSquaredToClosestFish;	
	Vector3<float> referenceFishPosition;
	referenceFishPtr->GetPos(referenceFishPosition);
	Vector3<float> currentFishPosition;
	Vector3<float> displacementVector;
	hash_map<Key, FishBody*>::iterator fishIter = m_parentTankSimulatorRef.m_fishBodyPtrs.begin();
	while (fishIter != m_parentTankSimulatorRef.m_fishBodyPtrs.end())
	{
		// check to make sure the current fish is not the reference fish
		if (fishIter->first != referenceFishKey)
		{
			// find the square of the distance between the two fish
			fishIter->second->GetPos(currentFishPosition);
			displacementVector = currentFishPosition - referenceFishPosition;
			float currentDistanceSquared = Dot(displacementVector, displacementVector);

			// check to see if this is the first distance to be calculated
			if (0 == keyOfClosestFish)
			{
				// record this as the closest fish
				keyOfClosestFish = fishIter->first;
				distanceSquaredToClosestFish = currentDistanceSquared;
			}
			else
			{
				// since other distances have already been calculated, we have
				// to check to see if the current fish is closer than the fish
				// that is currently the closest
				if (currentDistanceSquared < distanceSquaredToClosestFish)
				{
					// record this as the closest fish
					keyOfClosestFish = fishIter->first;
					distanceSquaredToClosestFish = currentDistanceSquared;
				}
			}
		}

		++fishIter;
	}

	return keyOfClosestFish;
}

////////////////////////////////////////////////////////////////////////////////
//  FUNCTION:       
/*! \brief          Vertically projects a point onto the HeightMap and returns
**                  the resulting height.
**  \param          pointToProject a position that is projected onto the HeightMap.
**  \return         The height of the resulting projection (its Y value).
**  \exception      none
**  \author         Austin Spafford
**  \date           3/25/2003 8:46:25 PM
**  \warning        none
*/
////////////////////////////////////////////////////////////////////////////////
float TankSimulatorClient::GetProjectedHeightOnHeightMap(const Vector3<float> & pointToProject)
{
	return m_parentTankSimulatorRef.GetProjectedHeightOnHeightMap(pointToProject);
}

////////////////////////////////////////////////////////////////////////////////
//  FUNCTION:       
/*! \brief          removes the requested amount of plankton from the simulation
**                  if possible, returns the actual amount of plankton sucked
**  \param          position a vector representing the "point of suckage"
**  \param          amountToSuck the amount of plankton the client wishes to 
**                  remove from the enviroment
**  \return         the actual amount of plankton removed from the enviroment
**  \exception      none
**  \author         Austin Spafford
**  \date           11/23/2002 3:17:38 AM
**  \warning        none
*/
////////////////////////////////////////////////////////////////////////////////
float TankSimulatorClient::SuckPlankton(Vector3<float> position, float amountToSuck)
{
	// The plankton now uses continous plankton
	float availablePlankton = GetPlanktonDensity(position);
	if (availablePlankton < amountToSuck)
	{
		return availablePlankton;
	}
	else
	{
		return amountToSuck;
	}
/*
	Vector3<int> voxelIndex = m_parentTankSimulatorRef.GetContainingPlanktonVoxelIndex(position);	

	float planktonDensity = m_parentTankSimulatorRef.GetPlanktonDensityFromVoxel(voxelIndex);

	// figure out whether or not the voxel will be completely drained
	if (planktonDensity > amountToSuck)
	{		
		// suck some, but not all of the plankton
		planktonDensity -= amountToSuck;
		m_parentTankSimulatorRef.SetPlanktonDensity(voxelIndex, planktonDensity);

		return amountToSuck;
	}
	else
	{
		// suck all of the remaining plankton
		m_parentTankSimulatorRef.SetPlanktonDensity(voxelIndex, 0);

		return planktonDensity;
	}
*/
}

////////////////////////////////////////////////////////////////////////////////
//  FUNCTION:       
/*! \brief          removes the requested amount of chum from the simulation if 
**                  possible, returns the actual amount of chum removed
**  \param          position the location of the chum eater
**  \param          amountToEat the amount of chum the chum eater wishes to eat
**  \return         the amount of chum that was actually eaten
**  \exception      none
**  \par            DESCRIPTION:
**  \author         Austin Spafford
**  \date           1/17/2003 9:35:31 PM
**  \warning        none
*/
////////////////////////////////////////////////////////////////////////////////
float TankSimulatorClient::EatChum(Vector3<float> position, float amountToEat)
{
	// find the voxel containing the chum eater
	Vector3<int> voxelIndex = m_parentTankSimulatorRef.GetContainingPlanktonVoxelIndex(position);

	// find out how much chum is left in the voxel containing the eater
	Vector3<float> chumPosition;
	float chumDensity;
	m_parentTankSimulatorRef.GetChumVoxelValues(voxelIndex, chumPosition, chumDensity);

	// check to see if there is enough chum left to satisfy the request
	if (amountToEat < chumDensity)
	{
		// eat some, but not all of the chum
		chumDensity -= amountToEat;
		m_parentTankSimulatorRef.SetChumVoxelValues(voxelIndex, chumPosition, chumDensity);
		return amountToEat;
	}
	else
	{
		// eat all of the remaining chum		
		m_parentTankSimulatorRef.SetChumVoxelValues(voxelIndex, chumPosition, 0);
		return chumDensity;
	}	
}

////////////////////////////////////////////////////////////////////////////////
//  FUNCTION:       
/*! \brief          adds the requested amount of chum to the simulation if 
**                  possible, returns the actual amount of chum added
**  \param          position the location of the chum dumper
**  \param          amountToDump the amount of chum the chum dumper wishes to dump
**  \return         the amount of chum that was actually dumped
**  \exception      none
**  \par            DESCRIPTION:
**  \author         Austin Spafford
**  \date           1/17/2003 9:35:31 PM
**  \warning        none
*/
////////////////////////////////////////////////////////////////////////////////
float TankSimulatorClient::DumpChum(Vector3<float> position, float amountToDump)
{
    //spawn some particles at the specified position
    m_parentTankSimulatorRef.m_gfxClient.SpawnChum(position, amountToDump);
    
    return amountToDump;
/*
	// find the voxel containing the chum dumper
	Vector3<int> voxelIndex = m_parentTankSimulatorRef.GetContainingPlanktonVoxelIndex(position);

	// find out if there is already a chum chunk in the voxel
	Vector3<float> chumPosition;
	float chumDensity;
	m_parentTankSimulatorRef.GetChumVoxelValues(voxelIndex, chumPosition, chumDensity);

	// check for a chum chunk
	if (chumDensity > 0)
	{
		// add to the current chum chunk
		chumDensity += amountToDump;
		m_parentTankSimulatorRef.SetChumVoxelValues(voxelIndex, chumPosition, chumDensity);
		return amountToDump;
	}
	else
	{
		// create a new chum chunk
		m_parentTankSimulatorRef.SetChumVoxelValues(voxelIndex, position, amountToDump);
		return amountToDump;
	}	
*/
}