////////////////////////////////////////////////////////////////////////////////
/*! \file           TankSimulatorClient.h
**  \brief          TankSimulatorClient object interface file
**  \author         Austin Spafford
**  \date           11/1/2002 4:20:38 PM
**  \par            DESCRIPTION:
**                    The TankSimulatorClient object provides an interface for
**                    stored objects that allow them to interact with other
**                    members of the simulation
**  \par            REMARKS:
*/
////////////////////////////////////////////////////////////////////////////////
#pragma once
////////////////////////////////////////////////////////////////////////////////
#include "messageHandler/MessageHandler.h"
#include "physics/PhysicsSimulator.h"
////////////////////////////////////////////////////////////////////////////////
class TankSimulator;
class InterpretationModule::FishBody;
////////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////
/*! \class          TankSimulatorClient
**  \brief          The TankSimulatorClient object provides an interface for
**                    stored objects that allow them to interact with other
**                    members of the simulation
**  \author         Austin Spafford
**  \date           11/1/2002 9:27:08 PM
**  \warning        none
*/
////////////////////////////////////////////////////////////////////////////////
class TankSimulatorClient
{
	public:

	////////////////////////////////////////////////////////////////////////////

	/// constructer, must use this one so that the client knows its parent
	TankSimulatorClient(TankSimulator &newParentTankSimulatorRef);	

	/// deconstructer
	~TankSimulatorClient();		

	////////////////////////////////////////////////////////////////////////////

	/// adds a fish to the mating list	
	void AddFishToMatingList(MsgHandler::Key fishKey);

	// FARK: AddFishList function

	/// adds a fish to the delete fish list
	void AddFishToDeleteFishList(MsgHandler::Key fishKey);	

	////////////////////////////////////////////////////////////////////////////

	// REMOVE FUNCS

    ////////////////////////////////////////////////////////////////////////////

	/// retrieves a reference to the physics simulator
	PhysicsSimulator & GetPhysicsSimulatorReference(void);		
	
	/// retrieves a pointer to another fish body in the simulation
	InterpretationModule::FishBody * GetFishBodyPtr(MsgHandler::Key keyForObject);

	/// retrieves the boundary values of the simulation
	void GetBoundary(
			math::linear_alg::Vector3<float> & minValuesRef,
			math::linear_alg::Vector3<float> & maxValuesRef
			);

	/// retrieves the value of the voxel that contains (or is closest to containing) the specified point
	float GetPlanktonDensity(math::linear_alg::Vector3<float> position);

	/// retrieves the position of the closest chunk of chum to a reference point, returns false if no nearby chum could be found
	bool GetClosestChumPosition(
			math::linear_alg::Vector3<float> referencePoint,
			math::linear_alg::Vector3<float> & chumPositionRef
			);

	/// returns the Key of the fish that is closest to the reference fish, returns 
	/// 0 if either the specified fish is the only fish in the simulation, or if
	/// it doesn't exist
	MsgHandler::Key GetKeyForClosestFish(const MsgHandler::Key & referenceFishKey);

	/// vertically projects a point onto the HeightMap and returns the projected
	/// point's Y value
	float GetProjectedHeightOnHeightMap(const math::linear_alg::Vector3<float> & pointToProject);

	////////////////////////////////////////////////////////////////////////////	

	/// removes the requested amount of plankton from the simulation if possible, returns the actual amount of plankton sucked
	float SuckPlankton(math::linear_alg::Vector3<float> position, float amountToSuck);	

	////////////////////////////////////////////////////////////////////////////

	/// removes the requested amount of chum from the simulation if possible, returns the actual amount of chum removed
	float EatChum(math::linear_alg::Vector3<float> position, float amountToEat);

	/// adds the requested amount of chum to the simulation if possible, returns the actual amount of chum added
	float DumpChum(math::linear_alg::Vector3<float> position, float amountToDump);

	////////////////////////////////////////////////////////////////////////////

	private:

	TankSimulator &m_parentTankSimulatorRef;

	////////////////////////////////////////////////////////////////////////////

	/// default constructer, intentional, use the reference-setting constructer
	TankSimulatorClient();	
};
