////////////////////////////////////////////////////////////////////////////////
/*! \file           TankSimulatorFileMemento.cpp
**  \brief          TankSimulatorFileMemento implementation file
**  \author         Austin Spafford
**  \date           5/14/2003 6:10:28 PM
**  \par            DESCRIPTION:
**                    The TankSimulatorFileMemento object encapsulates the
**                    state of a TankSimulator object and can read and write
**                    itself to i/o streams.
**  \par            REMARKS:
**                    
*/
////////////////////////////////////////////////////////////////////////////////
#pragma warning (disable : 4786)
////////////////////////////////////////////////////////////////////////////////
#include <assert.h>
#include <iostream>
#include <iomanip>
#include "TankSimulatorFileMemento.h"
#include "nexception/Nexception.h"
#include "ai_core/AI_System.h"
// memory leak detection
#ifdef _DEBUG
    #include"memleak.h"
    #define new DEBUG_NEW
    #undef THIS_FILE
    static char THIS_FILE[] = __FILE__;
#endif
////////////////////////////////////////////////////////////////////////////////
using namespace math::linear_alg;
using namespace InterpretationModule;
////////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////
//  FUNCTION:       
/*! \brief          Returns the location of the next instance of a character
**                  in a NULL terminated string that is part of a floating
**                  point number.
**  \param          stringPtr a pointer into the string to seek through
**  \return         A pointer to the first floating point character, or NULL if
**                  there are not such characters exist in the specified string.
**  \exception      none
**  \author         Austin Spafford
**  \date           5/15/2003 1:29:48 PM
**  \warning        none
*/
////////////////////////////////////////////////////////////////////////////////
static char * SeekToFirstFloatingPointCharacter(char * stringPtr)
{
	if (NULL == stringPtr) return NULL;

	// seek until the end of the string
	while (0 != *stringPtr)
	{
		// check to see if the current character is a number
		char numericalValue = *stringPtr - '0';
		if (numericalValue >= 0 && numericalValue <= 9)
		{
			return stringPtr;
		}

		// check to see if the current character is a decimal point
		if ('.' == *stringPtr)
		{
			return stringPtr;
		}

		// check to see if the current character is a minus sign
		if ('-' == *stringPtr)
		{
			return stringPtr;
		}

		// move to the next character
		++stringPtr;
	}

	return NULL;
}

////////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////
//  FUNCTION:       
/*! \brief          Returns the location of the next instance of a character
**                  in a NULL terminated string that is not part of a floating
**                  point number.
**  \param          stringPtr a pointer into the string to seek through
**  \return         A pointer to the first non-floating-point character, or NULL if
**                  there are not such characters exist in the specified string.
**  \exception      none
**  \author         Austin Spafford
**  \date           5/15/2003 1:29:48 PM
**  \warning        none
*/
////////////////////////////////////////////////////////////////////////////////
static char * SeekToFirstNonFloatingPointCharacter(char * stringPtr)
{
	if (NULL == stringPtr) return NULL;

	// seek until the end of the string
	while (0 != *stringPtr)
	{
		// check to see if the current character is a number
		char numericalValue = *stringPtr - '0';
		if (numericalValue >= 0 && numericalValue <= 9)
		{
			// move to the next character
			++stringPtr;
			continue;
		}

		// check to see if the current character is a decimal point
		if ('.' == *stringPtr)
		{
			// move to the next character
			++stringPtr;
			continue;
		}

		// check to see if the current character is a minus sign
		if ('-' == *stringPtr)
		{
			// move to the next character
			++stringPtr;
			continue;
		}

		// the current character is not a floating point number, so we can return
		return stringPtr;
	}

	return NULL;
}

////////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////
//  FUNCTION:       
/*! \brief          Writes the memento's contents to an ostream in a human
**                  readable format.
**  \param          outputStream the stream that is to be written to
**  \return         none
**  \exception      none
**  \author         Austin Spafford
**  \date           5/15/2003 1:26:44 PM
**  \warning        none
*/
////////////////////////////////////////////////////////////////////////////////
void TankSimulatorFileMemento::Write(std::ostream & outputStream) const
{
	using std::endl;

	// make sure that any floating point numbers are printed with out
	// any scientific notation
	ios::fmtflags oldFlags = outputStream.flags();
	outputStream << std::fixed << std::setprecision(3) << std::showpoint;

	// write a small header
	outputStream << "============================================================" << endl;
	outputStream << "Begin TankSimulator Information" << endl;
	outputStream << endl;

	// write the boundaries
	outputStream << "Boundary Mins: (" << m_boundaryMins[0] << ", " << m_boundaryMins[1] << ", " << m_boundaryMins[2] << ")" << endl;
	outputStream << "Boundary Maxs: (" << m_boundaryMaxs[0] << ", " << m_boundaryMaxs[1] << ", " << m_boundaryMaxs[2] << ")"  << endl;
	outputStream << endl;

	// write the fish
	outputStream << "Number of fish bodies: " << m_frozenFish.size() << endl;	
	outputStream << endl;	
	for (unsigned int i=0; i<m_frozenFish.size(); ++i)
	{
		const FrozenFish & fishRef = m_frozenFish[i];
		// write out a fish
		outputStream << "-----------------------------" << endl;
		outputStream << "Begin Fish #" << i << ": " << endl;
		outputStream << endl;
		outputStream << "Attributes:" << endl;
		outputStream << fishRef.m_attributes << endl;
		outputStream << "Statistics:" << endl;
		outputStream << fishRef.m_stats << endl;
		outputStream << "Key:" << endl;
		outputStream << fishRef.m_key << endl;
		outputStream << "Commands:" << endl;
		outputStream << "Number of commands:" << fishRef.m_commands.size() << endl;
		std::vector<CommandPacket>::const_iterator commandIter;
		for (
			commandIter = fishRef.m_commands.begin();
			fishRef.m_commands.end() != commandIter;
			++commandIter
			)
		{
			outputStream << *commandIter << endl;
		}
/*
		outputStream << "Examples:" << endl;
		outputStream << "Number of examples:" << fishRef.m_examples.size() << endl;
		std::vector<AI_System::AI_ActionTag>::const_iterator actionTagIter;
		for (
			actionTagIter = fishRef.m_examples.begin();
			fishRef.m_examples.end() != actionTagIter;
			++actionTagIter
			)
		{
			outputStream
					<< actionTagIter->unit_ID << " "
					<< actionTagIter->decisionID << " "
					<< actionTagIter->fitness << " "
					<< actionTagIter->completed
					<< endl;
		}
		outputStream << endl;
*/
		outputStream << "Actions:" << endl;
		outputStream << "Number of actions:" << fishRef.m_actions.size() << endl;
		std::vector<FrozenFish::PsuedoAction>::const_iterator actionIter;
		for (
			actionIter = fishRef.m_actions.begin();
			fishRef.m_actions.end() != actionIter;
			++actionIter
			)
		{
			outputStream
					<< actionIter->m_decisionID << " "
					<< actionIter->m_fitness << " "
					<< actionIter->m_cmdID << " "
					<< actionIter->m_bodyID << " "
					<< actionIter->m_senderID << " "
					<< actionIter->m_weight << " "
					<< actionIter->m_isDone << " "
					<< actionIter->m_isComplete << " "
					<< actionIter->m_elapsedTime << " "
					<< actionIter->m_targetFlag << " "
					<< "("
					<< actionIter->m_target.m_point.x << ", "
					<< actionIter->m_target.m_point.y << ", "
					<< actionIter->m_target.m_point.z
					<< ") "
					<< actionIter->m_optional << " "
					<< endl;
		}
		outputStream << endl;
		outputStream << "Is Mouth Stopped: " << fishRef.m_isMouthStopped << endl;
		outputStream << "Is Attacking: " << fishRef.m_isAttacking << endl;
		outputStream << "Is Eating: " << fishRef.m_isEating << endl;
		outputStream << "Is Fin Stopped: " << fishRef.m_isFinStopped << endl;
		outputStream << "Is Learning: " << fishRef.m_isLearn << endl;
		outputStream << "Is Pathing: " << fishRef.m_isPathing << endl;
		outputStream << "Current Movement Vector: (" << fishRef.m_currVector[0] << ", " << fishRef.m_currVector[1] << ", " << fishRef.m_currVector[2] << ")" << endl;
		outputStream << "Current Facing Vector: (" << fishRef.m_facingVector[0] << ", " << fishRef.m_facingVector[1] << ", " << fishRef.m_facingVector[2] << ")" << endl;
		outputStream << "Current Position: (" << fishRef.m_pos[0] << ", " << fishRef.m_pos[1] << ", " << fishRef.m_pos[2] << ")" << endl;
		outputStream << "Current Velocity: (" << fishRef.m_vel[0] << ", " << fishRef.m_vel[1] << ", " << fishRef.m_vel[2] << ")" << endl;
		outputStream << "Current Acceleration: (" << fishRef.m_accel[0] << ", " << fishRef.m_accel[1] << ", " << fishRef.m_accel[2] << ")" << endl;
		outputStream << endl;
		outputStream << "End Fish #" << i << ": " << endl;
		outputStream << "-----------------------------" << endl;
		outputStream << endl;
	}

	// write the height map
	outputStream << endl;
	outputStream << "The height map:" << endl;
	m_heightMapFileMementoPtr->Write(outputStream);	

	// write the mating fish list
	std::list<MsgHandler::Key>::const_iterator keyIter;
	outputStream << "The mating fish list:" << endl;	
	for (
		keyIter = m_matingFishList.begin();
		m_matingFishList.end() != keyIter;
		++keyIter
		)
	{
		cout << *keyIter << " ";
	}
	outputStream << endl;

	// write the delete fish list
	outputStream << "The delete fish list:" << endl;	
	for (
		keyIter = m_deleteFishList.begin();
		m_deleteFishList.end() != keyIter;
		++keyIter
		)
	{
		outputStream << *keyIter << " ";
	}
	outputStream << endl;

	// write a small trail
	outputStream << endl;
	outputStream << "End TankSimulator Information" << endl;
	outputStream << "============================================================" << endl;	
	outputStream << endl;

	// restore the old flags
	outputStream.flags(oldFlags);

	return;
}

////////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////
//  FUNCTION:       
/*! \brief          Reads the memento's contents from an istream that is using
**                  a human readable format.
**  \param          inputStream the stream that is to be read from
**  \return         none
**  \exception      none
**  \author         Austin Spafford
**  \date           5/15/2003 1:28:18 PM
**  \warning        none
*/
////////////////////////////////////////////////////////////////////////////////
void TankSimulatorFileMemento::Read(std::istream & inputStream)
{
	const int characterBufferSize = 1024;
	char * characterBuffer = new char[characterBufferSize];
	char * currentCharacterPtr;
	unsigned int i, k;

	// HEADER
	// eat the header
	inputStream.getline(characterBuffer, characterBufferSize);
	inputStream.getline(characterBuffer, characterBufferSize);
	inputStream.getline(characterBuffer, characterBufferSize);

	// BOUNDARIES
	// read in the boundary mins
	inputStream.getline(characterBuffer, characterBufferSize);	
	currentCharacterPtr = SeekToFirstFloatingPointCharacter(characterBuffer);
	if (NULL == currentCharacterPtr) throw Nexception("Failed to load BoundaryMins[0].", __FILE__, __LINE__);
	m_boundaryMins[0] = (float)atof(currentCharacterPtr);
	currentCharacterPtr = SeekToFirstNonFloatingPointCharacter(currentCharacterPtr);
	currentCharacterPtr = SeekToFirstFloatingPointCharacter(currentCharacterPtr);
	if (NULL == currentCharacterPtr) throw Nexception("Failed to load BoundaryMins[1].", __FILE__, __LINE__);
	m_boundaryMins[1] = (float)atof(currentCharacterPtr);
	currentCharacterPtr = SeekToFirstNonFloatingPointCharacter(currentCharacterPtr);
	currentCharacterPtr = SeekToFirstFloatingPointCharacter(currentCharacterPtr);
	if (NULL == currentCharacterPtr) throw Nexception("Failed to load BoundaryMins[2].", __FILE__, __LINE__);
	m_boundaryMins[2] = (float)atof(currentCharacterPtr);
	// read in the boundary maxs
	inputStream.getline(characterBuffer, characterBufferSize);
	currentCharacterPtr = SeekToFirstFloatingPointCharacter(characterBuffer);
	if (NULL == currentCharacterPtr) throw Nexception("Failed to load BoundaryMaxs[0].", __FILE__, __LINE__);
	m_boundaryMaxs[0] = (float)atof(currentCharacterPtr);
	currentCharacterPtr = SeekToFirstNonFloatingPointCharacter(currentCharacterPtr);
	currentCharacterPtr = SeekToFirstFloatingPointCharacter(currentCharacterPtr);
	if (NULL == currentCharacterPtr) throw Nexception("Failed to load BoundaryMaxs[1].", __FILE__, __LINE__);
	m_boundaryMaxs[1] = (float)atof(currentCharacterPtr);
	currentCharacterPtr = SeekToFirstNonFloatingPointCharacter(currentCharacterPtr);
	currentCharacterPtr = SeekToFirstFloatingPointCharacter(currentCharacterPtr);
	if (NULL == currentCharacterPtr) throw Nexception("Failed to load BoundaryMaxs[2].", __FILE__, __LINE__);
	m_boundaryMaxs[2] = (float)atof(currentCharacterPtr);
	// eat a line
	inputStream.getline(characterBuffer, characterBufferSize);

	// FISH BODIES
	// read in the number of fish
	unsigned int numberOfFishBodies;
	inputStream.getline(characterBuffer, characterBufferSize);
	currentCharacterPtr = SeekToFirstFloatingPointCharacter(characterBuffer);
	if (NULL == currentCharacterPtr) throw Nexception("Failed to load the fish body count.", __FILE__, __LINE__);
	numberOfFishBodies = atoi(currentCharacterPtr);
	// eat a line
	inputStream.getline(characterBuffer, characterBufferSize);
	m_frozenFish.resize(numberOfFishBodies);
	for (i=0; i<numberOfFishBodies; ++i)
	{
		FrozenFish & fishRef = m_frozenFish[i];
		// eat four lines
		inputStream.getline(characterBuffer, characterBufferSize);
		inputStream.getline(characterBuffer, characterBufferSize);
		inputStream.getline(characterBuffer, characterBufferSize);
		inputStream.getline(characterBuffer, characterBufferSize);
		// load the attributes		
		inputStream >> fishRef.m_attributes;
		// eat two lines
		inputStream.getline(characterBuffer, characterBufferSize);
		inputStream.getline(characterBuffer, characterBufferSize);
		// load the statistics
		inputStream >> fishRef.m_stats;
		// eat two lines
		inputStream.getline(characterBuffer, characterBufferSize);
		inputStream.getline(characterBuffer, characterBufferSize);
		// load the key
		inputStream >> fishRef.m_key;
		// eat two lines
		inputStream.getline(characterBuffer, characterBufferSize);
		inputStream.getline(characterBuffer, characterBufferSize);
		// load the number of commands		
		unsigned int numberOfCommands;
		inputStream.getline(characterBuffer, characterBufferSize);
		currentCharacterPtr = SeekToFirstFloatingPointCharacter(characterBuffer);
		if (NULL == currentCharacterPtr) throw Nexception("Failed to load the command count.", __FILE__, __LINE__);
		numberOfCommands = atoi(currentCharacterPtr);
		// load the commands
		fishRef.m_commands.resize(numberOfCommands);		
		for (k=0; k<numberOfCommands; ++k)
		{
			// read the command
			inputStream >> fishRef.m_commands[k];
		}
/*
		// eat a line
		inputStream.getline(characterBuffer, characterBufferSize);
		// load the number of examples		
		unsigned int numberOfExamples;
		inputStream.getline(characterBuffer, characterBufferSize);
		currentCharacterPtr = SeekToFirstFloatingPointCharacter(characterBuffer);
		if (NULL == currentCharacterPtr) throw Nexception("Failed to load the example count.", __FILE__, __LINE__);
		numberOfExamples = atoi(currentCharacterPtr);
		// load the examples
		fishRef.m_examples.resize(numberOfExamples);		
		for (k=0; k<numberOfExamples; ++k)
		{
			AI_System::AI_ActionTag & exampleRef = fishRef.m_examples[k];
			// get ready to read the entire line
			inputStream.getline(characterBuffer, characterBufferSize);
			// load the unit ID					
			currentCharacterPtr = SeekToFirstFloatingPointCharacter(characterBuffer);
			if (NULL == currentCharacterPtr) throw Nexception("Failed to load the unit ID in an example.", __FILE__, __LINE__);
			exampleRef.unit_ID = atoi(currentCharacterPtr);
			// load the decision ID
			currentCharacterPtr = SeekToFirstNonFloatingPointCharacter(currentCharacterPtr);
			currentCharacterPtr = SeekToFirstFloatingPointCharacter(currentCharacterPtr);
			if (NULL == currentCharacterPtr) throw Nexception("Failed to load the decision ID in an example.", __FILE__, __LINE__);
			exampleRef.decisionID = atoi(currentCharacterPtr);
			// load the fitness
			currentCharacterPtr = SeekToFirstNonFloatingPointCharacter(currentCharacterPtr);
			currentCharacterPtr = SeekToFirstFloatingPointCharacter(currentCharacterPtr);
			if (NULL == currentCharacterPtr) throw Nexception("Failed to load the fitness in an example.", __FILE__, __LINE__);
			exampleRef.fitness = (float)atof(currentCharacterPtr);
			// load the completion flag
			currentCharacterPtr = SeekToFirstNonFloatingPointCharacter(currentCharacterPtr);
			currentCharacterPtr = SeekToFirstFloatingPointCharacter(currentCharacterPtr);
			if (NULL == currentCharacterPtr) throw Nexception("Failed to load the completion flag in an example.", __FILE__, __LINE__);
			if (atoi(currentCharacterPtr))
			{
				exampleRef.completed = true;
			}
			else
			{
				exampleRef.completed = false;
			}
		}
		// eat a line
		inputStream.getline(characterBuffer, characterBufferSize);		
*/
		// eat a line
		inputStream.getline(characterBuffer, characterBufferSize);
		// load the number of actions		
		unsigned int numberOfActions;
		inputStream.getline(characterBuffer, characterBufferSize);
		currentCharacterPtr = SeekToFirstFloatingPointCharacter(characterBuffer);
		if (NULL == currentCharacterPtr) throw Nexception("Failed to load the action count.", __FILE__, __LINE__);
		numberOfActions = atoi(currentCharacterPtr);
		// load the actions
		fishRef.m_actions.resize(numberOfActions);		
		for (k=0; k<numberOfActions; ++k)
		{
			FrozenFish::PsuedoAction & actionRef = fishRef.m_actions[k];
			// get ready to read the entire line
			inputStream.getline(characterBuffer, characterBufferSize);
			// load the decision ID	
			currentCharacterPtr = SeekToFirstFloatingPointCharacter(characterBuffer);
			if (NULL == currentCharacterPtr) throw Nexception("Failed to load the decision ID in an action.", __FILE__, __LINE__);
			actionRef.m_decisionID = atoi(currentCharacterPtr);
			// load the fitness
			currentCharacterPtr = SeekToFirstNonFloatingPointCharacter(currentCharacterPtr);
			currentCharacterPtr = SeekToFirstFloatingPointCharacter(currentCharacterPtr);
			if (NULL == currentCharacterPtr) throw Nexception("Failed to load the fitness in an action.", __FILE__, __LINE__);
			actionRef.m_fitness = (float)atof(currentCharacterPtr);
			// load the command ID
			currentCharacterPtr = SeekToFirstNonFloatingPointCharacter(currentCharacterPtr);
			currentCharacterPtr = SeekToFirstFloatingPointCharacter(currentCharacterPtr);
			if (NULL == currentCharacterPtr) throw Nexception("Failed to load the command ID in an action.", __FILE__, __LINE__);
			actionRef.m_cmdID = atoi(currentCharacterPtr);
			// load the body ID
			currentCharacterPtr = SeekToFirstNonFloatingPointCharacter(currentCharacterPtr);
			currentCharacterPtr = SeekToFirstFloatingPointCharacter(currentCharacterPtr);
			if (NULL == currentCharacterPtr) throw Nexception("Failed to load the body ID in an action.", __FILE__, __LINE__);
			actionRef.m_bodyID = atoi(currentCharacterPtr);
			// load the sender ID
			currentCharacterPtr = SeekToFirstNonFloatingPointCharacter(currentCharacterPtr);
			currentCharacterPtr = SeekToFirstFloatingPointCharacter(currentCharacterPtr);
			if (NULL == currentCharacterPtr) throw Nexception("Failed to load the sender ID in an action.", __FILE__, __LINE__);
			actionRef.m_senderID = atoi(currentCharacterPtr);
			// load the weight
			currentCharacterPtr = SeekToFirstNonFloatingPointCharacter(currentCharacterPtr);
			currentCharacterPtr = SeekToFirstFloatingPointCharacter(currentCharacterPtr);
			if (NULL == currentCharacterPtr) throw Nexception("Failed to load the weight in an action.", __FILE__, __LINE__);
			actionRef.m_weight = (float)atof(currentCharacterPtr);
			// load the "is done" flag
			currentCharacterPtr = SeekToFirstNonFloatingPointCharacter(currentCharacterPtr);
			currentCharacterPtr = SeekToFirstFloatingPointCharacter(currentCharacterPtr);
			if (NULL == currentCharacterPtr) throw Nexception("Failed to load the \"is done\" flag in an action.", __FILE__, __LINE__);			
			actionRef.m_isDone = atoi(currentCharacterPtr) ? true : false;			
			// load the "is complete" flag
			currentCharacterPtr = SeekToFirstNonFloatingPointCharacter(currentCharacterPtr);
			currentCharacterPtr = SeekToFirstFloatingPointCharacter(currentCharacterPtr);
			if (NULL == currentCharacterPtr) throw Nexception("Failed to load the \"is complete\" flag in an action.", __FILE__, __LINE__);			
			actionRef.m_isComplete = atoi(currentCharacterPtr) ? true : false;			
			// load the elapsed time
			currentCharacterPtr = SeekToFirstNonFloatingPointCharacter(currentCharacterPtr);
			currentCharacterPtr = SeekToFirstFloatingPointCharacter(currentCharacterPtr);
			if (NULL == currentCharacterPtr) throw Nexception("Failed to load the elapsed time in an action.", __FILE__, __LINE__);
			actionRef.m_elapsedTime = (float)atof(currentCharacterPtr);
			// load the target flag
			currentCharacterPtr = SeekToFirstNonFloatingPointCharacter(currentCharacterPtr);
			currentCharacterPtr = SeekToFirstFloatingPointCharacter(currentCharacterPtr);
			if (NULL == currentCharacterPtr) throw Nexception("Failed to load the target flag in an action.", __FILE__, __LINE__);
			actionRef.m_targetFlag = atoi(currentCharacterPtr);
			// load the target[0]
			currentCharacterPtr = SeekToFirstNonFloatingPointCharacter(currentCharacterPtr);
			currentCharacterPtr = SeekToFirstFloatingPointCharacter(currentCharacterPtr);
			if (NULL == currentCharacterPtr) throw Nexception("Failed to load target[0] in an action.", __FILE__, __LINE__);
			actionRef.m_target.m_point.x = (float)atof(currentCharacterPtr);
			// load the target[1]
			currentCharacterPtr = SeekToFirstNonFloatingPointCharacter(currentCharacterPtr);
			currentCharacterPtr = SeekToFirstFloatingPointCharacter(currentCharacterPtr);
			if (NULL == currentCharacterPtr) throw Nexception("Failed to load target[1] in an action.", __FILE__, __LINE__);
			actionRef.m_target.m_point.y = (float)atof(currentCharacterPtr);
			// load the target[2]
			currentCharacterPtr = SeekToFirstNonFloatingPointCharacter(currentCharacterPtr);
			currentCharacterPtr = SeekToFirstFloatingPointCharacter(currentCharacterPtr);
			if (NULL == currentCharacterPtr) throw Nexception("Failed to load target[2] in an action.", __FILE__, __LINE__);
			actionRef.m_target.m_point.z = (float)atof(currentCharacterPtr);
			// load the optional
			currentCharacterPtr = SeekToFirstNonFloatingPointCharacter(currentCharacterPtr);
			currentCharacterPtr = SeekToFirstFloatingPointCharacter(currentCharacterPtr);
			if (NULL == currentCharacterPtr) throw Nexception("Failed to load the optional component in an action.", __FILE__, __LINE__);
			actionRef.m_optional = (float)atof(currentCharacterPtr);
		}
		// eat a line
		inputStream.getline(characterBuffer, characterBufferSize);		
		// load the "is mouth stopped" flag
		inputStream.getline(characterBuffer, characterBufferSize);
		currentCharacterPtr = SeekToFirstFloatingPointCharacter(characterBuffer);
		if (NULL == currentCharacterPtr) throw Nexception("Failed to load the \"is mouth stopped\" flag.", __FILE__, __LINE__);
		fishRef.m_isMouthStopped = atoi(currentCharacterPtr) ? true : false;
		// load the "is attacking" flag
		inputStream.getline(characterBuffer, characterBufferSize);
		currentCharacterPtr = SeekToFirstFloatingPointCharacter(characterBuffer);
		if (NULL == currentCharacterPtr) throw Nexception("Failed to load the \"is attacking\" flag.", __FILE__, __LINE__);
		fishRef.m_isAttacking = atoi(currentCharacterPtr) ? true : false;
		// load the "is eating" flag
		inputStream.getline(characterBuffer, characterBufferSize);
		currentCharacterPtr = SeekToFirstFloatingPointCharacter(characterBuffer);
		if (NULL == currentCharacterPtr) throw Nexception("Failed to load the \"is eating\" flag.", __FILE__, __LINE__);
		fishRef.m_isEating = atoi(currentCharacterPtr) ? true : false;
		// load the "is fin stopped" flag
		inputStream.getline(characterBuffer, characterBufferSize);
		currentCharacterPtr = SeekToFirstFloatingPointCharacter(characterBuffer);
		if (NULL == currentCharacterPtr) throw Nexception("Failed to load the \"is fin stopped\" flag.", __FILE__, __LINE__);
		fishRef.m_isFinStopped = atoi(currentCharacterPtr) ? true : false;
		// load the "is learning" flag
		inputStream.getline(characterBuffer, characterBufferSize);
		currentCharacterPtr = SeekToFirstFloatingPointCharacter(characterBuffer);
		if (NULL == currentCharacterPtr) throw Nexception("Failed to load the \"is learning\" flag.", __FILE__, __LINE__);
		fishRef.m_isLearn = atoi(currentCharacterPtr) ? true : false;
		// load the "is pathing" flag
		inputStream.getline(characterBuffer, characterBufferSize);
		currentCharacterPtr = SeekToFirstFloatingPointCharacter(characterBuffer);
		if (NULL == currentCharacterPtr) throw Nexception("Failed to load the \"is pathing\" flag.", __FILE__, __LINE__);
		fishRef.m_isPathing = atoi(currentCharacterPtr) ? true : false;
		// load movementVector[0]
		inputStream.getline(characterBuffer, characterBufferSize);
		currentCharacterPtr = SeekToFirstFloatingPointCharacter(characterBuffer);
		if (NULL == currentCharacterPtr) throw Nexception("Failed to load movementVector[0].", __FILE__, __LINE__);
		fishRef.m_currVector[0] = (float)atof(currentCharacterPtr);
		// load movementVector[1]
		currentCharacterPtr = SeekToFirstNonFloatingPointCharacter(currentCharacterPtr);
		currentCharacterPtr = SeekToFirstFloatingPointCharacter(currentCharacterPtr);
		if (NULL == currentCharacterPtr) throw Nexception("Failed to load movementVector[1].", __FILE__, __LINE__);
		fishRef.m_currVector[1] = (float)atof(currentCharacterPtr);
		// load movementVector[2]
		currentCharacterPtr = SeekToFirstNonFloatingPointCharacter(currentCharacterPtr);
		currentCharacterPtr = SeekToFirstFloatingPointCharacter(currentCharacterPtr);
		if (NULL == currentCharacterPtr) throw Nexception("Failed to load movementVector[2].", __FILE__, __LINE__);
		fishRef.m_currVector[2] = (float)atof(currentCharacterPtr);
		// load facingVector[0]
		inputStream.getline(characterBuffer, characterBufferSize);
		currentCharacterPtr = SeekToFirstFloatingPointCharacter(characterBuffer);
		if (NULL == currentCharacterPtr) throw Nexception("Failed to load facingVector[0].", __FILE__, __LINE__);
		fishRef.m_facingVector[0] = (float)atof(currentCharacterPtr);
		// load facingVector[1]
		currentCharacterPtr = SeekToFirstNonFloatingPointCharacter(currentCharacterPtr);
		currentCharacterPtr = SeekToFirstFloatingPointCharacter(currentCharacterPtr);
		if (NULL == currentCharacterPtr) throw Nexception("Failed to load facingVector[1].", __FILE__, __LINE__);
		fishRef.m_facingVector[1] = (float)atof(currentCharacterPtr);
		// load facingVector[2]
		currentCharacterPtr = SeekToFirstNonFloatingPointCharacter(currentCharacterPtr);
		currentCharacterPtr = SeekToFirstFloatingPointCharacter(currentCharacterPtr);
		if (NULL == currentCharacterPtr) throw Nexception("Failed to load facingVector[2].", __FILE__, __LINE__);
		fishRef.m_facingVector[2] = (float)atof(currentCharacterPtr);
		// load position[0]
		inputStream.getline(characterBuffer, characterBufferSize);
		currentCharacterPtr = SeekToFirstFloatingPointCharacter(characterBuffer);
		if (NULL == currentCharacterPtr) throw Nexception("Failed to load position[0].", __FILE__, __LINE__);
		fishRef.m_pos[0] = (float)atof(currentCharacterPtr);
		// load position[1]
		currentCharacterPtr = SeekToFirstNonFloatingPointCharacter(currentCharacterPtr);
		currentCharacterPtr = SeekToFirstFloatingPointCharacter(currentCharacterPtr);
		if (NULL == currentCharacterPtr) throw Nexception("Failed to load position[1].", __FILE__, __LINE__);
		fishRef.m_pos[1] = (float)atof(currentCharacterPtr);
		// load position[2]
		currentCharacterPtr = SeekToFirstNonFloatingPointCharacter(currentCharacterPtr);
		currentCharacterPtr = SeekToFirstFloatingPointCharacter(currentCharacterPtr);
		if (NULL == currentCharacterPtr) throw Nexception("Failed to load position[2].", __FILE__, __LINE__);
		fishRef.m_pos[2] = (float)atof(currentCharacterPtr);
		// load velocity[0]
		inputStream.getline(characterBuffer, characterBufferSize);
		currentCharacterPtr = SeekToFirstFloatingPointCharacter(characterBuffer);
		if (NULL == currentCharacterPtr) throw Nexception("Failed to load velocity[0].", __FILE__, __LINE__);
		fishRef.m_vel[0] = (float)atof(currentCharacterPtr);
		// load velocity[1]
		currentCharacterPtr = SeekToFirstNonFloatingPointCharacter(currentCharacterPtr);
		currentCharacterPtr = SeekToFirstFloatingPointCharacter(currentCharacterPtr);
		if (NULL == currentCharacterPtr) throw Nexception("Failed to load velocity[1].", __FILE__, __LINE__);
		fishRef.m_vel[1] = (float)atof(currentCharacterPtr);
		// load velocity[2]
		currentCharacterPtr = SeekToFirstNonFloatingPointCharacter(currentCharacterPtr);
		currentCharacterPtr = SeekToFirstFloatingPointCharacter(currentCharacterPtr);
		if (NULL == currentCharacterPtr) throw Nexception("Failed to load velocity[2].", __FILE__, __LINE__);
		fishRef.m_vel[2] = (float)atof(currentCharacterPtr);
		// load acceleration[0]
		inputStream.getline(characterBuffer, characterBufferSize);
		currentCharacterPtr = SeekToFirstFloatingPointCharacter(characterBuffer);
		if (NULL == currentCharacterPtr) throw Nexception("Failed to load acceleration[0].", __FILE__, __LINE__);
		fishRef.m_accel[0] = (float)atof(currentCharacterPtr);
		// load acceleration[1]
		currentCharacterPtr = SeekToFirstNonFloatingPointCharacter(currentCharacterPtr);
		currentCharacterPtr = SeekToFirstFloatingPointCharacter(currentCharacterPtr);
		if (NULL == currentCharacterPtr) throw Nexception("Failed to load acceleration[1].", __FILE__, __LINE__);
		fishRef.m_accel[1] = (float)atof(currentCharacterPtr);
		// load acceleration[2]
		currentCharacterPtr = SeekToFirstNonFloatingPointCharacter(currentCharacterPtr);
		currentCharacterPtr = SeekToFirstFloatingPointCharacter(currentCharacterPtr);
		if (NULL == currentCharacterPtr) throw Nexception("Failed to load acceleration[2].", __FILE__, __LINE__);
		fishRef.m_accel[2] = (float)atof(currentCharacterPtr);
		// eat four lines
		inputStream.getline(characterBuffer, characterBufferSize);
		inputStream.getline(characterBuffer, characterBufferSize);
		inputStream.getline(characterBuffer, characterBufferSize);
		inputStream.getline(characterBuffer, characterBufferSize);
	}
	// eat a line
	inputStream.getline(characterBuffer, characterBufferSize);

	// HEIGHTMAP
	// eat a line
	inputStream.getline(characterBuffer, characterBufferSize);
	// read in the height map
	if (NULL == m_heightMapFileMementoPtr)
	{
		m_heightMapFileMementoPtr = new HeightMapFileMemento;
	}
	m_heightMapFileMementoPtr->Read(inputStream);	

	// MATING FISH LIST
	// eat a line
	inputStream.getline(characterBuffer, characterBufferSize);
	// read in all of the fish keys in the line
	m_matingFishList.clear();
	inputStream.getline(characterBuffer, characterBufferSize);
	currentCharacterPtr = SeekToFirstFloatingPointCharacter(characterBuffer);
	while (NULL != currentCharacterPtr)
	{
		// read the key
		m_matingFishList.push_back(atoi(currentCharacterPtr));
		// move to the next key
		currentCharacterPtr = SeekToFirstNonFloatingPointCharacter(currentCharacterPtr);
		currentCharacterPtr = SeekToFirstFloatingPointCharacter(currentCharacterPtr);
	}	
	

	// DELETE FISH LIST
	// eat a line
	inputStream.getline(characterBuffer, characterBufferSize);
	// read in all of the fish keys in the line
	m_deleteFishList.clear();
	inputStream.getline(characterBuffer, characterBufferSize);
	currentCharacterPtr = SeekToFirstFloatingPointCharacter(characterBuffer);
	while (NULL != currentCharacterPtr)
	{
		// read the key
		m_deleteFishList.push_back(atoi(currentCharacterPtr));
		// move to the next key
		currentCharacterPtr = SeekToFirstNonFloatingPointCharacter(currentCharacterPtr);
		currentCharacterPtr = SeekToFirstFloatingPointCharacter(currentCharacterPtr);
	}	

	// TAIL
	// eat the trail
	inputStream.getline(characterBuffer, characterBufferSize);
	inputStream.getline(characterBuffer, characterBufferSize);
	inputStream.getline(characterBuffer, characterBufferSize);
	inputStream.getline(characterBuffer, characterBufferSize);

	delete []characterBuffer;

	return;
}
