////////////////////////////////////////////////////////////////////////////////
/*! \file       tankSimulator/tank_sim_client_gfx.h
**  \brief      The TankSimulator graphics client
**  \author     M Jared Finder
**  \date       3/12/2003 6:59:47 PM
**  \par        DESCRIPTION:
**                  Displays non-heightmap TankSimulator data, such as plankton
**                  and chum.
*/
////////////////////////////////////////////////////////////////////////////////
#pragma warning(disable:4786)
#include"tank_sim_client_gfx.h"
#include"timeObject/timer.h"


////////////////////////////////////////////////////////////////////////////////
// Needed includes
////////////////////////////////////////////////////////////////////////////////
//#include"stdheaders.h"
#include<iostream>
#include<stdlib.h>
#include<assert.h>
#include<math.h>
#include"tankSimulator/TankSimulator.h"
#include"gfx/renderable.h"
#include"gfx/renderer.h"
#include"gfx/opengl.h"
#include"math/linear_alg/vector3.h"
#include"math/linear_alg/matrix4.h"
// memory leak detection
#ifdef _DEBUG
    #include"memleak.h"
    #define new DEBUG_NEW
    #undef THIS_FILE
    static char THIS_FILE[] = __FILE__;
#endif

///half the width of the particle
const float halfWidth       = 2.0f;
///half the height of the particle
const float halfHeight      = 2.0f;
///minimum value for particle alpha
const float alphaMin        = 0;
///maxmimum value for particle alpha
const float alphaMax        = 0.28;
///half the width of the particle box
const float boxHalfWidth    = 10;
///half the height of the particle box
const float boxHalfHeight   = 10;
///half the depth of the particle box
const float boxHalfDepth    = 10;
///when blending start
const float blendStart      = 0.5f;

///initial age
const float chumInitialAge  = 2;

struct PlanktonParticleData
{
    float pos[3];
    float tex[2];
    float clr[4];
};

////////////////////////////////////////////////////////////////////////////////
//  FUNCTION:   
/*! \brief      
**  \param      
**  \return     none
**  \exception  none
**  \par        DESCRIPTION:
**                  
**  \author     M Jared Finder
**  \date       3/12/2003 7:39:14 PM
**  \warning    none
*/
////////////////////////////////////////////////////////////////////////////////
TankSimClientGfx::TankSimClientGfx(void)
    : m_pParent(NULL), m_planktonData(numPlanktonParticles*4*sizeof(PlanktonParticleData), true),
    m_planktonTex("Resources/TankObj/plankton.tga", true),
    m_chumTex("Resources/TankObj/chum.tga", true)
{
    ///create the plankton particles
    for(int i = 0; i < numPlanktonParticles; ++i)
    {
        //calculate a center for each particle
        m_planktonParticles[i].position[0] = boxHalfWidth  - (rand()/float(RAND_MAX) * 2*boxHalfWidth);
        m_planktonParticles[i].position[1] = boxHalfHeight - (rand()/float(RAND_MAX) * 2*boxHalfHeight);
        m_planktonParticles[i].position[2] = boxHalfDepth  - (rand()/float(RAND_MAX) * 2*boxHalfDepth);
    }

    ///create the chum particles
    for(i = 0; i < numChumSystems; ++i)
    {
        m_chumSystems[i].age = 0;
    }
}


////////////////////////////////////////////////////////////////////////////////
//  FUNCTION:   
/*! \brief      Sets the TankSimulator the client refers to.  Must be called
**              before attempting to draw.
**  \param      pParent     Pointer to containing TankSimulator
**  \return     none
**  \exception  none
**  \author     M Jared Finder
**  \date       3/12/2003 7:39:32 PM
**  \warning    none
*/
////////////////////////////////////////////////////////////////////////////////
void TankSimClientGfx::SetParent(TankSimulator* pParent)
{
    m_pParent = pParent;
}


////////////////////////////////////////////////////////////////////////////////
//  FUNCTION:   
/*! \brief      Draws all the tank simulator data
**  \param      
**  \return     none
**  \exception  none
**  \par        DESCRIPTION:
**                  
**  \author     M Jared Finder
**  \date       3/12/2003 7:39:53 PM
**  \warning    none
*/
////////////////////////////////////////////////////////////////////////////////
void TankSimClientGfx::Render(void)
{
    //we need to have a parent before this function is called!
    assert(m_pParent != NULL);

    UpdatePlankton();
    UpdateChum();

    RenderChum();
    RenderPlankton();
}


////////////////////////////////////////////////////////////////////////////////
//  FUNCTION:   
/*! \brief      Renders plankton based on the current plankton level
**  \param      
**  \return     none
**  \exception  none
**  \author     M Jared Finder
**  \date       3/18/2003 2:23:33 PM
**  \warning    none
*/
////////////////////////////////////////////////////////////////////////////////
void TankSimClientGfx::RenderPlankton(void)
{
    using math::linear_alg::Vector3;
    using math::linear_alg::Matrix4;

    //we need to have a parent before this function is called!
    assert(m_pParent != NULL);

    //calculate direction vectors
    Matrix4<float> projMatrix;
    Matrix4<float> viewMatrix;
    gfx::Renderer::Get().GetTransform3D(projMatrix, viewMatrix);
    Vector3<float> xDirection;
    Vector3<float> yDirection;
    xDirection[0] = viewMatrix[0][0];
    xDirection[1] = viewMatrix[0][1];
    xDirection[2] = viewMatrix[0][2];
    yDirection[0] = viewMatrix[1][0];
    yDirection[1] = viewMatrix[1][1];
    yDirection[2] = viewMatrix[1][2];

    //calculate alpha blend amount
    Vector3<float> cameraPos;
    Vector3<float> dummy;
    gfx::Renderer::Get().Get3DView(cameraPos, dummy, dummy);
    const float planktonDensity = m_pParent->GetPlanktonDensityAtLocation(cameraPos);

    //set the vertecies
    PlanktonParticleData* pData
        = reinterpret_cast<PlanktonParticleData*>(m_planktonData.Unlock());
    for(int i = 0; i < numPlanktonParticles; ++i)
    {
        //calculate alpha amount
        float alphaAmount;
        if(m_planktonParticles[i].distance > blendStart)
        {
            float interpAmount = (1 - m_planktonParticles[i].distance)/(1-blendStart);
            alphaAmount = alphaMin*(1 - interpAmount) + alphaMax*interpAmount;
        }
        else
        {
            alphaAmount = alphaMax;
        }
        alphaAmount *= planktonDensity;

        //first vertex
        pData[i*4 + 0].pos[0] = m_planktonParticles[i].position[0]
                              - halfWidth *xDirection[0]
                              - halfHeight*yDirection[0];
        pData[i*4 + 0].pos[1] = m_planktonParticles[i].position[1]
                              - halfWidth *xDirection[1]
                              - halfHeight*yDirection[1];
        pData[i*4 + 0].pos[2] = m_planktonParticles[i].position[2]
                              - halfWidth *xDirection[2]
                              - halfHeight*yDirection[2];
        pData[i*4 + 0].tex[0] = 0;
        pData[i*4 + 0].tex[1] = 0;
        pData[i*4 + 0].clr[0] = 1;
        pData[i*4 + 0].clr[1] = 1;
        pData[i*4 + 0].clr[2] = 1;
        pData[i*4 + 0].clr[3] = alphaAmount;
        //second vertex
        pData[i*4 + 1].pos[0] = m_planktonParticles[i].position[0]
                              - halfWidth *xDirection[0]
                              + halfHeight*yDirection[0];
        pData[i*4 + 1].pos[1] = m_planktonParticles[i].position[1]
                              - halfWidth *xDirection[1]
                              + halfHeight*yDirection[1];
        pData[i*4 + 1].pos[2] = m_planktonParticles[i].position[2]
                              - halfWidth *xDirection[2]
                              + halfHeight*yDirection[2];
        pData[i*4 + 1].tex[0] = 0;
        pData[i*4 + 1].tex[1] = 1;
        pData[i*4 + 1].clr[0] = 1;
        pData[i*4 + 1].clr[1] = 1;
        pData[i*4 + 1].clr[2] = 1;
        pData[i*4 + 1].clr[3] = alphaAmount;
        //third vertex
        pData[i*4 + 2].pos[0] = m_planktonParticles[i].position[0]
                              + halfWidth *xDirection[0]
                              + halfHeight*yDirection[0];
        pData[i*4 + 2].pos[1] = m_planktonParticles[i].position[1]
                              + halfWidth *xDirection[1]
                              + halfHeight*yDirection[1];
        pData[i*4 + 2].pos[2] = m_planktonParticles[i].position[2]
                              + halfWidth *xDirection[2]
                              + halfHeight*yDirection[2];
        pData[i*4 + 2].tex[0] = 1;
        pData[i*4 + 2].tex[1] = 1;
        pData[i*4 + 2].clr[0] = 1;
        pData[i*4 + 2].clr[1] = 1;
        pData[i*4 + 2].clr[2] = 1;
        pData[i*4 + 2].clr[3] = alphaAmount;
        //fourth vertex
        pData[i*4 + 3].pos[0] = m_planktonParticles[i].position[0]
                              + halfWidth *xDirection[0]
                              - halfHeight*yDirection[0];
        pData[i*4 + 3].pos[1] = m_planktonParticles[i].position[1]
                              + halfWidth *xDirection[1]
                              - halfHeight*yDirection[1];
        pData[i*4 + 3].pos[2] = m_planktonParticles[i].position[2]
                              + halfWidth *xDirection[2]
                              - halfHeight*yDirection[2];
        pData[i*4 + 3].tex[0] = 1;
        pData[i*4 + 3].tex[1] = 0;
        pData[i*4 + 3].clr[0] = 1;
        pData[i*4 + 3].clr[1] = 1;
        pData[i*4 + 3].clr[2] = 1;
        pData[i*4 + 3].clr[3] = alphaAmount;
    }
    m_planktonData.Lock();

    //make sure we can alpha blend
    glPushAttrib(GL_BLEND);
    glEnable(GL_BLEND);
    glPushClientAttrib(GL_COLOR_ARRAY);
    glEnableClientState(GL_COLOR_ARRAY);
    glDepthMask(GL_FALSE);

    //render plankton
    glColor4f(1,1,1,1);
    glBindTexture(GL_TEXTURE_2D, m_planktonTex.GetTextureHandle());
    m_planktonData.Activate(true, true, true, sizeof(PlanktonParticleData));
    glLockArraysEXT(0, numPlanktonParticles*4);
    glDrawArrays(GL_QUADS, 0, numPlanktonParticles*4);
    glUnlockArraysEXT();
    m_planktonData.Deactivate();

    //restore previous state
    glDepthMask(GL_TRUE);
    glPopClientAttrib();
    glPopAttrib();
}


////////////////////////////////////////////////////////////////////////////////
//  FUNCTION:   
/*! \brief      Renders chum based on the chum the tank holds.
**  \param      
**  \return     none
**  \exception  none
**  \author     M Jared Finder
**  \date       3/18/2003 2:23:33 PM
**  \warning    none
*/
////////////////////////////////////////////////////////////////////////////////
void TankSimClientGfx::RenderChum( void)
{
    using math::linear_alg::Vector3;

    //set up for rendering
    glVertexPointer(3, GL_FLOAT, sizeof(TankSimChumData), m_chumParticles);
    if(gfx::Renderer::Get().IsEnabled().pointParameters)
    {
        const float quadraticAttenuation[] = { 1.0f, 0.0f, 0.01f };
        glPointParameterfARB(GL_POINT_SIZE_MIN_ARB, 5);
        glPointParameterfARB(GL_POINT_SIZE_MAX_ARB, 20);
        glPointParameterfvARB(GL_POINT_DISTANCE_ATTENUATION_ARB, quadraticAttenuation);
    }
    if(gfx::Renderer::Get().IsEnabled().NVpointSprite)
    {
        glTexEnvi(GL_POINT_SPRITE_NV, GL_COORD_REPLACE_NV, GL_TRUE);
        glEnable(GL_POINT_SPRITE_NV);
        glBindTexture(GL_TEXTURE_2D, m_chumTex.GetTextureHandle());
    }
    else
    {
        glBindTexture(GL_TEXTURE_2D, 0);
    }
    glPointSize(20);
    glColor3f(1,0,0);
    //glColor3ub(204,74,94);
    

    //render all chum
    for(int i = 0; i < numChumSystems; ++i)
    {
        //only render chum systems that aren't dead
        if(m_chumSystems[i].age <= 0)
        {
            continue;
        }

        //render the particles
        glDrawArrays(
                GL_POINTS,
                m_chumSystems[i].start,
                m_chumSystems[i].length
                );
    }
    if(gfx::Renderer::Get().IsEnabled().NVpointSprite)
    {
        glDisable(GL_POINT_SPRITE_NV);
    }
}


////////////////////////////////////////////////////////////////////////////////
//  FUNCTION:   
/*! \brief      Updates the position of plankton particles via wrapping
**  \param      
**  \return     none
**  \exception  none
**  \par        DESCRIPTION:
**                  
**  \author     M Jared Finder
**  \date       3/18/2003 3:21:35 PM
**  \warning    none
*/
////////////////////////////////////////////////////////////////////////////////
void TankSimClientGfx::UpdatePlankton(void)
{
    using math::linear_alg::Vector3;
    Vector3<float> cameraPos;
    Vector3<float> lookAt;
    Vector3<float> upVector;

    gfx::Renderer::Get().Get3DView(cameraPos, lookAt, upVector);
    //if any are outside the 10x10x10 box around the player, move them!
    for(int i = 0; i < numPlanktonParticles; ++i)
    {
        Vector3<float> particleDelta(
            m_planktonParticles[i].position[0] - cameraPos[0],
            m_planktonParticles[i].position[1] - cameraPos[1],
            m_planktonParticles[i].position[2] - cameraPos[2]
            );

        //bound in width
        if(particleDelta[0] < -boxHalfWidth)
        {
            float widthDelta = fmodf(particleDelta[0]-boxHalfWidth, 2*boxHalfWidth)
                                + boxHalfWidth;
            m_planktonParticles[i].position[0] += widthDelta - particleDelta[0];
            particleDelta[0] = widthDelta;
        }
        else if(particleDelta[0] > boxHalfWidth)
        {
            float widthDelta = fmodf(particleDelta[0]+boxHalfWidth, 2*boxHalfWidth)
                                - boxHalfWidth;
            m_planktonParticles[i].position[0] += widthDelta - particleDelta[0];
            particleDelta[0] = widthDelta;
        }

        //bound in height
        if(particleDelta[1] < -boxHalfHeight)
        {
            float heightDelta = fmodf(particleDelta[1]-boxHalfHeight, 2*boxHalfHeight)
                                + boxHalfHeight;
            m_planktonParticles[i].position[1] += heightDelta - particleDelta[1];
            particleDelta[1] = heightDelta;
        }
        else if(particleDelta[1] > boxHalfHeight)
        {
            float heightDelta = fmodf(particleDelta[1]+boxHalfHeight, 2*boxHalfHeight)
                                - boxHalfHeight;
            m_planktonParticles[i].position[1] += heightDelta - particleDelta[1];
            particleDelta[1] = heightDelta;
        }

        //bound in depth
        if(particleDelta[2] < -boxHalfDepth)
        {
            float depthDelta = fmodf(particleDelta[2]-boxHalfDepth, 2*boxHalfDepth)
                                + boxHalfDepth;
            m_planktonParticles[i].position[2] += depthDelta - particleDelta[2];
            particleDelta[2] = depthDelta;
        }
        else if(particleDelta[2] > boxHalfHeight)
        {
            float depthDelta = fmodf(particleDelta[2]+boxHalfDepth, 2*boxHalfDepth)
                                - boxHalfDepth;
            m_planktonParticles[i].position[2] += depthDelta - particleDelta[2];
            particleDelta[2] = depthDelta;
        }

        ///calculate the particle distance
        particleDelta[0] = fabsf(particleDelta[0]);
        particleDelta[1] = fabsf(particleDelta[1]);
        particleDelta[2] = fabsf(particleDelta[2]);
        if(particleDelta[0] > particleDelta[1] && particleDelta[0] > particleDelta[2])
        {
            m_planktonParticles[i].distance = particleDelta[0] / boxHalfWidth;
        }
        else if(particleDelta[1] > particleDelta[0] && particleDelta[1] > particleDelta[2])
        {
            m_planktonParticles[i].distance = particleDelta[1] / boxHalfHeight;
        }
        else if(particleDelta[2] > particleDelta[0] && particleDelta[2] > particleDelta[1])
        {
            m_planktonParticles[i].distance = particleDelta[2] / boxHalfDepth;
        }
    }
}



////////////////////////////////////////////////////////////////////////////////
//  FUNCTION:   
/*! \brief      
**  \param      
**  \return     none
**  \exception  none
**  \par        DESCRIPTION:
**                  
**  \author     M Jared Finder
**  \date       4/10/2003 4:00:57 PM
**  \warning    none
*/
////////////////////////////////////////////////////////////////////////////////
void TankSimClientGfx::UpdateChum(void)
{
    float timeStep = float(TimeObject::Instance()->getGameTimeStep());
    float dragFactor(5.f);

    ///update each particle
    for(int i = 0; i < numChumPerSystem*numChumSystems; ++i)
    {
        ///update the position
        m_chumParticles[i].position[0] += m_chumParticles[i].velocity[0]*timeStep;
        m_chumParticles[i].position[1] += m_chumParticles[i].velocity[1]*timeStep;
        m_chumParticles[i].position[2] += m_chumParticles[i].velocity[2]*timeStep;

        ///update the speed
        float velocityScale(1 - dragFactor*timeStep);
        if(velocityScale < 0)
        {
            m_chumParticles[i].velocity[0] = 0;
            m_chumParticles[i].velocity[1] = 0;
            m_chumParticles[i].velocity[2] = 0;
        }
        else
        {
            m_chumParticles[i].velocity[0] *= velocityScale;
            m_chumParticles[i].velocity[1] *= velocityScale;
            m_chumParticles[i].velocity[2] *= velocityScale;
        }
    }

    ///advance each simulation
    for(i = 0; i < numChumSystems; ++i)
    {
        m_chumSystems[i].age -= timeStep;
    }
}



////////////////////////////////////////////////////////////////////////////////
//  FUNCTION:   
/*! \brief      
**  \param      
**  \return     none
**  \exception  none
**  \par        DESCRIPTION:
**                  
**  \author     M Jared Finder
**  \date       4/10/2003 6:17:45 PM
**  \warning    none
*/
////////////////////////////////////////////////////////////////////////////////
void TankSimClientGfx::SpawnChum(math::linear_alg::Vector3<float> position, float ammount)
{
    //find the first empty chum system
    int i;
    for(i = 0; i < numChumSystems; ++i)
    {
        if(m_chumSystems[i].age <= 0)
        {
            break;
        }
    }

    //so was there an empty chum system?
    if(i == numChumSystems)
    {
        return;
    }

    //calculate the number of particles to spawn
    int numParticles = 0;
    if(ammount < 0.2)
    {
        numParticles = numChumPerSystem / 2;
    }
    else if(ammount < 0.8)
    {
        numParticles = ((ammount - 0.2)/(0.8 - 0.2)/2 + 0.5)*numChumPerSystem;
    }
    else
    {
        numParticles = numChumPerSystem;
    }


    //create a system at that point
    int initialParticleIndex = int(numChumPerSystem*i);
    m_chumSystems[i].age    = chumInitialAge;
    m_chumSystems[i].start  = initialParticleIndex;
    m_chumSystems[i].length = numParticles;
    for(i = 0; i < numParticles; ++i)
    {
        m_chumParticles[initialParticleIndex + i].position[0] = position[0];
        m_chumParticles[initialParticleIndex + i].position[1] = position[1];
        m_chumParticles[initialParticleIndex + i].position[2] = position[2];

        m_chumParticles[initialParticleIndex + i].velocity[0] = float(rand() % 100 - 50);
        m_chumParticles[initialParticleIndex + i].velocity[1] = float(rand() % 100 - 50);
        m_chumParticles[initialParticleIndex + i].velocity[2] = float(rand() % 100 - 50);
    }
}
