////////////////////////////////////////////////////////////////////////////////
/*! \file       tankSimulator/tank_sim_client_gfx.h
**  \brief      The TankSimulator graphics client
**  \author     M Jared Finder
**  \date       3/12/2003 6:59:47 PM
**  \par        DESCRIPTION:
**                  Displays non-heightmap TankSimulator data, such as plankton
**                  and chum.
*/
////////////////////////////////////////////////////////////////////////////////
#pragma once


////////////////////////////////////////////////////////////////////////////////
// Forward Declarations and needed includes
////////////////////////////////////////////////////////////////////////////////
#include"gfx/renderable.h"
#include"gfx/vtx_buffer.h"
#include"gfx/texture.h"
class TankSimulator;


////////////////////////////////////////////////////////////////////////////////
/*! \struct     TankSimParticleData    
**  \brief      Holds all data relating to the particles
**  \author     M Jared Finder
**  \date       3/19/2003 8:06:37 PM
**  \warning    none
*/
////////////////////////////////////////////////////////////////////////////////
struct TankSimParticleData
{
    ///x,y,z of the center of the particle
    float position[3];

    ///rotation amount (in degrees) of the particle
    float rotation;

    ///scaled distance from the camera of the particle.
    ///0 = at camera, 1 = about to be clipped
    float distance;
};

struct TankSimChumData
{
    ///x, y, z of the chum particle
    float position[3];

    ///velocity of the particle
    float velocity[3];
};

struct TankSimChumParticleSys
{
    /// TTL for the system (at 0 it dies)
    float age;
    /// starting index
    int start;
    /// number of used indecies
    int length;
};


////////////////////////////////////////////////////////////////////////////////
/*! \class      TankSimClientGfx
**  \brief      Displays tank simulator data such as plankton and chum
**  \author     M Jared Finder
**  \date       3/12/2003 7:01:03 PM
**  \warning    none
*/
////////////////////////////////////////////////////////////////////////////////
class TankSimClientGfx : gfx::IRenderable3DAlpha
{
public:
    ///ctor
    TankSimClientGfx(void);
    ///set the parent
    void SetParent(TankSimulator* pParent);

    ///draw tank simulator stuff
    using IRenderable3DAlpha::SetRender;

    ///spawns chum particles at the specified position
    void SpawnChum(math::linear_alg::Vector3<float> position, float ammount);
    ///render function called by graphics engine
    void Render(void);
private:
    ///update the position of plankton particles
    void UpdatePlankton(void);
    ///update the positions of the chum particles
    void UpdateChum(void);
    ///render plankton, currently as particle system
    void RenderPlankton(void);
    ///render chum, currently as models
    void RenderChum(void);

private:
    ///the tank simulator this is attached to
    TankSimulator* m_pParent;

    ///the plankton texture
    gfx::Texture m_planktonTex;
    gfx::Texture m_chumTex;

    enum { numPlanktonParticles = 500 };
    enum { numChumPerSystem     = 20  };
    enum { numChumSystems       = 8   };
    ///plankton particle position, rotation, etc.
    TankSimParticleData m_planktonParticles[numPlanktonParticles];
    
    ///chum particles
    TankSimChumData m_chumParticles[numChumPerSystem*numChumSystems];
    TankSimChumParticleSys m_chumSystems[numChumSystems];

    ///vertex buffer for rendering
    gfx::VertexBuffer m_planktonData;
};