////////////////////////////////////////////////////////////////////////////////
/*!	\file		Timer.cpp
**	\brief		implementation file for Timer.h
**	\author		John Corpening
**	\date		10/7/2002 2:06:28 PM
**	\par		DESCRIPTION:
**					This file contains the implementation of the TimeObject class
**	\par		REMARKS:
*/
////////////////////////////////////////////////////////////////////////////////

#include"Timer.h"
#include<windows.h>
// memory leak detection
#ifdef _DEBUG
    #include"memleak.h"
    #define new DEBUG_NEW
    #undef THIS_FILE
    static char THIS_FILE[] = __FILE__;
#endif

TimeObject* TimeObject::_instance = 0;



////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	TimeObject::TimeObject
/*!	\brief		Constructor sets defaults
**	\return		none
**	\exception	none
**	\author		John Corpening
**	\date		10/7/2002 1:13:12 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
TimeObject::TimeObject() throw()
{
	///Get the counts per second
	LARGE_INTEGER	freq;
	QueryPerformanceFrequency(&freq);
	frequency = (double)freq.LowPart;

	///Set defaults
	frameCount = 0;
	gameSpeed = 1.0;
	timeSample = 2.0;
	maxSpeed = 10.0;
	minSpeed = 0.0;
	increment = (maxSpeed - minSpeed) / 10.0;

	QueryPerformanceCounter(&lastTime);
	lastAverage = lastTime;
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	TimeObject::Instance
/*!	\brief		returns the pointer to the single timer instance
**	\return		the pointer to the single timer instance
**	\exception	none
**	\author		John Corpening
**	\date		10/7/2002 3:23:01 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
TimeObject* TimeObject::Instance() throw()
{
	if(_instance == 0)
	{
		_instance = new TimeObject;
	}
	return _instance;
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	TimeObject::logic
/*!	\brief		This function is called every frame for timing logic
**	\return		none
**	\exception	none
**	\author		John Corpening
**	\date		10/7/2002 12:28:58 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
void TimeObject::logic() throw()
{
	LARGE_INTEGER	temp;

	///Get the current time
	///Assume that the hardware supports a high performance counter
	QueryPerformanceCounter(&temp);

	///Calculate the real time step in seconds
	realTimeStep = (double)(temp.QuadPart - lastTime.QuadPart)/frequency;

	///Calculate the game time step
	gameTimeStep = realTimeStep * gameSpeed;

	///Set the actual frame rate
	actFPS = 1.0/realTimeStep;

	///Increment the frame counter
	frameCount++;

	///Calculate the average frame rate
	/// test is the time since last average
	double test = (temp.QuadPart - lastAverage.QuadPart)/frequency;

	if(test >= timeSample)
	{
		///Average: (frames/counts)*(counts/sec)
		avgFPS = (double)frameCount/test;

		///Reset the frameCounter
		frameCount = 0;
		lastAverage = temp;
	}

	/// set the lastTime as the current time
	lastTime = temp;
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	TimeObject::setTimeSample
/*!	\brief		sets the timeSample
**	\param		time the new time sample
**	\return		the old time sample
**	\exception	none
**	\author		John Corpening
**	\date		10/7/2002 12:03:53 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
double TimeObject::setTimeSample(double time) throw()
{
	double old = timeSample;
	timeSample = time;
	return old;
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	TimeObject::setGameSpeed
/*!	\brief		sets the game speed
**	\param		speed the new game speed
**	\return		the old game speed
**	\exception	none
**	\author		John Corpening
**	\date		10/7/2002 12:27:01 PM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
double TimeObject::setGameSpeed(double speed) throw()
{
	double old = gameSpeed;

	if((speed <= maxSpeed) && (speed >= minSpeed))
	{
		gameSpeed = speed;
	}

	return old;
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	TimeObject::setMaxSpeed
/*!	\brief		sets the upper limit for the game speed
**	\param		max the upper limit for the game speed
**	\return		the old game speed upper limit
**	\exception	none
**	\author		John Corpening
**	\date		10/7/2002 12:57:37 PM
**	\warning	There is no check on the upper limit that is set!
*/
////////////////////////////////////////////////////////////////////////////////
double TimeObject::setMaxSpeed(double max) throw()
{
	double old = maxSpeed;
	maxSpeed = max;
	increment = (maxSpeed - minSpeed) / 10.0;
	return old;
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	TimeObject::setMinSpeed
/*!	\brief		sets the lower limit for the game speed
**	\param		min the lower limit for the game speed
**	\return		the old lower limit for the game speed
**	\exception	none
**	\author		John Corpening
**	\date		10/7/2002 12:57:42 PM
**	\warning	There is no check on the lower limit that is set!
*/
////////////////////////////////////////////////////////////////////////////////
double TimeObject::setMinSpeed(double min) throw()
{
	double old = minSpeed;
	minSpeed = min;
	increment = (maxSpeed - minSpeed) / 10.0;
	return old;
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	TimeObject::incrementGameSpeed
/*!	\brief		increments the game speed by 1/10 the max-min interval
**	\return		true if increment is possible false otherwise
**	\exception	none
**	\author		John Corpening
**	\date		10/11/2002 9:49:02 AM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
bool TimeObject::incrementGameSpeed() throw()
{
	if(gameSpeed >= maxSpeed)
		return false;

	gameSpeed += increment;
	return true;
}

////////////////////////////////////////////////////////////////////////////////
//	FUNCTION:	TimeObject::decrementGameSpeed
/*!	\brief		decrements the game speed by 1/10 the max-min interval
**	\return		true if decrement is possible false otherwise
**	\exception	none
**	\author		John Corpening
**	\date		10/11/2002 9:49:02 AM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
bool TimeObject::decrementGameSpeed() throw()
{
	if(gameSpeed <= minSpeed)
		return false;

	gameSpeed -= increment;
	return true;
}