////////////////////////////////////////////////////////////////////////////////
/*!	\file		Timer.h
**	\brief		Time handling code
**	\author		John Corpening
**	\date		10/7/2002 11:38:58 AM
**	\par		DESCRIPTION:
**					Modules that need support for time keeping will use the 
**					timer object. It will provide functionality for checking 
**					the frame rate, time step and game time step.
**	\par		REMARKS:
*/
////////////////////////////////////////////////////////////////////////////////
#pragma once

#include <windows.h>

////////////////////////////////////////////////////////////////////////////////
/*!	\class		TimeObject
**	\brief		The time handling object
**	\author		John Corpening
**	\date		10/7/2002 11:39:58 AM
**	\warning	none
*/
////////////////////////////////////////////////////////////////////////////////
class TimeObject
{
	double			timeSample;		//!< Amount of seconds over which to average the frame rate
	LARGE_INTEGER	lastTime;		//!< Last time count
	LARGE_INTEGER	lastAverage;	//!< Last time count since last average
	unsigned int	frameCount;		//!< frame count since last avgFPD calculation
	double			avgFPS;			//!< Average frame rate over last timeSample seconds
	double			actFPS;			//!< Current actual frame rate
	double			realTimeStep;	//!< Time since last frame in seconds
	double			gameTimeStep;	//!< Time since last frame modified by game speed
	double			gameSpeed;		//!< gameTimeStep modifier
	double			maxSpeed;		//!< Maximum gameSpeed
	double			minSpeed;		//!< Minimum gameSpeed
	double			frequency;		//!< The counts per second
	double			increment;		//!< gameSpeed increment, max - min/10

	///Singleton functionality
	static	TimeObject*	_instance;	//!< unique reference

protected:

	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:	TimeObject::TimeObject
	/*!	\brief		Constructor sets the frequency
	**	\return		none
	**	\exception	none
	**	\author		John Corpening
	**	\date		10/7/2002 1:13:12 PM
	**	\warning	none
	*/
	////////////////////////////////////////////////////////////////////////////////
	TimeObject() throw();

public:

	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:	TimeObject::Instance
	/*!	\brief		returns the pointer to the single timer instance
	**	\return		the pointer to the single timer instance
	**	\exception	none
	**	\author		John Corpening
	**	\date		10/7/2002 3:23:01 PM
	**	\warning	none
	*/
	////////////////////////////////////////////////////////////////////////////////
	static TimeObject* Instance() throw();

	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:	TimeObject::logic
	/*!	\brief		This function is called every frame for timing logic
	**	\return		none
	**	\exception	none
	**	\author		John Corpening
	**	\date		10/7/2002 12:28:58 PM
	**	\warning	This should only be called by the main loop.
	*/
	////////////////////////////////////////////////////////////////////////////////
	void logic() throw();

	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:	TimeObject::getAverageFramerate
	/*!	\brief		returns the frame rate averaged over timeSample milliseconds
	**	\return		average FPS
	**	\exception	none
	**	\author		John Corpening
	**	\date		10/7/2002 12:26:39 PM
	**	\warning	none
	*/
	////////////////////////////////////////////////////////////////////////////////
	double getAverageFramerate()const throw() {return avgFPS;}

	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:	TimeObject::getActualFramerate
	/*!	\brief		returns the current frame rate since last frame
	**	\return		actual FPS
	**	\exception	none
	**	\author		John Corpening
	**	\date		10/7/2002 12:26:45 PM
	**	\warning	none
	*/
	////////////////////////////////////////////////////////////////////////////////
	double getActualFramerate()const throw(){return actFPS;}

	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:	TimeObject::getRealTimeStep
	/*!	\brief		returns the time since last frame
	**	\return		time since last frame
	**	\exception	none
	**	\author		John Corpening
	**	\date		10/7/2002 12:26:49 PM
	**	\warning	none
	*/
	////////////////////////////////////////////////////////////////////////////////
	double getRealTimeStep()const throw(){return realTimeStep;}

	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:	TimeObject::getGameTimeStep
	/*!	\brief		returns the real time step * game speed
	**	\return		realTimeStep * gameSpeed
	**	\exception	none
	**	\author		John Corpening
	**	\date		10/7/2002 12:26:53 PM
	**	\warning	none
	*/
	////////////////////////////////////////////////////////////////////////////////
	double getGameTimeStep()const throw(){return gameTimeStep;}

	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:	TimeObject::getGameSpeed
	/*!	\brief		returns the time step modifier
	**	\return		gameSpeed, the time step modifier
	**	\exception	none
	**	\author		John Corpening
	**	\date		10/7/2002 12:26:56 PM
	**	\warning	none
	*/
	////////////////////////////////////////////////////////////////////////////////
	double getGameSpeed() const throw(){return gameSpeed;}

	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:	TimeObject::setTimeSample
	/*!	\brief		sets the timeSample
	**	\param		time the new time sample
	**	\return		the old time sample
	**	\exception	none
	**	\author		John Corpening
	**	\date		10/7/2002 12:03:53 PM
	**	\warning	none
	*/
	////////////////////////////////////////////////////////////////////////////////
	double setTimeSample(double time) throw();

	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:	TimeObject::setGameSpeed
	/*!	\brief		sets the game speed
	**	\param		speed the new game speed
	**	\return		the old game speed
	**	\exception	none
	**	\author		John Corpening
	**	\date		10/7/2002 12:27:01 PM
	**	\warning	none
	*/
	////////////////////////////////////////////////////////////////////////////////
	double setGameSpeed(double speed) throw();

	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:	TimeObject::setMaxSpeed
	/*!	\brief		sets the upper limit for the game speed
	**	\param		max the upper limit for the game speed
	**	\return		the old game speed upper limit
	**	\exception	none
	**	\author		John Corpening
	**	\date		10/7/2002 12:57:37 PM
	**	\warning	There is no check on the upper limit that is set!
	*/
	////////////////////////////////////////////////////////////////////////////////
	double setMaxSpeed(double max) throw();

	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:	TimeObject::setMinSpeed
	/*!	\brief		sets the lower limit for the game speed
	**	\param		min the lower limit for the game speed
	**	\return		the old lower limit for the game speed
	**	\exception	none
	**	\author		John Corpening
	**	\date		10/7/2002 12:57:42 PM
	**	\warning	There is no check on the lower limit that is set!
	*/
	////////////////////////////////////////////////////////////////////////////////
	double setMinSpeed(double min) throw();

	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:	TimeObject::incrementGameSpeed
	/*!	\brief		increments the game speed by 1/10 the max-min interval
	**	\return		none
	**	\exception	none
	**	\author		John Corpening
	**	\date		10/11/2002 9:49:02 AM
	**	\warning	none
	*/
	////////////////////////////////////////////////////////////////////////////////
	bool incrementGameSpeed() throw();

	////////////////////////////////////////////////////////////////////////////////
	//	FUNCTION:	TimeObject::decrementGameSpeed
	/*!	\brief		decrements the game speed by 1/10 the max-min interval
	**	\return		none
	**	\exception	none
	**	\author		John Corpening
	**	\date		10/11/2002 9:49:02 AM
	**	\warning	none
	*/
	////////////////////////////////////////////////////////////////////////////////
	bool decrementGameSpeed() throw();
};

